# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
#==============================================================================
#
#      E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#      TableEntry
#            
#------------------------------------------------------------------------------
#            
#      Data table for data entry
#
#==============================================================================
#       (C) energyXperts.NET (E4-Experts GmbH, Berlin, Germany)
#============================================================================== 
import wx
from GUI.status import *
from globals import Globals
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, toUTF8, noneCheck
from GUI.numCtrl import format_int_wrapper, convertDoubleToString
from GUI.GUITools import GRAPH_BACKGROUND_COLOR, GRID_BACKGROUND_COLOR, \
     GRID_LETTER_COLOR_HIGHLIGHT, GRID_BACKGROUND_COLOR_HIGHLIGHT, TITLE_COLOR, \
     GRID_LETTER_COLOR, GRID_LETTER_SIZE, GRID_LABEL_SIZE, \
     COLOUR_DATAENTRY_BG

from modules.messageLogger import (logDebug, showMessage, showError, 
                                            logTrack, askConfirmation)
from globals import logger

from copy import copy

from GUI.dataEntry.editCell import EditCell


DEFAULTMENU = [ _U("Insert above"),
                _U("Insert below"),
                _U("Split"),
                _U("Delete") ]


class TableEntry(wx.Panel):
    """ Generic class for I/O of data via a table (wx.grid)
    """

    def __init__(self, parent, titles, changeData=None, 
                 rowSize=20,     # default row height
                 colSize=100,    # default col width
                 colLabelSize=24,
                 size=wx.Size(600, 200),
                 hAlignment=None,   # list of horizontal alignment for each column
                 readOnly=None,     # list of readOnly - values for each column
                 unitSelectors=None,       # list of unit formats (ctype, unitType, decimals)
                 unitCol = None,        # if not None, units are assigned row-wise
                 totalRow=False):       # True if last row of table should be displayed with a special format 
        """
        :titles     list of titles for each column (-> defines the number of columns)
        :changeData function changeData(index, mode, data) called if a grid cell is changed or on popUp-Menu actions (e.g.row is added or deleted)
                    note: other actions such as mouse-click actions should be handled by an external event handler (see setExternalEventHandler)
        """
        self.prnt = parent
        wx.Panel.__init__(self, id=-1, name='TableEntry', parent=parent,
              style=0, size=size)

        self.titles = titles
        self.nCols = len(titles)
        self.changeData = changeData

        self.unitSelectors = unitSelectors
        self.unitCol = unitCol
        if unitSelectors is not None:
            ####TODO: here choices could be assigned similar to units
            self.units = [Globals.unitConversion.defaultUnit(entry[1]) if (entry[0]=="float" and entry[1] is not None) else "" for entry in unitSelectors]
            self.unitLists = [Globals.unitConversion.units[entry[1]]   if (entry[0]=="float" and entry[1] is not None) else \
                              (entry[1] if entry[0]=="choice" else []) for entry in unitSelectors]

            if unitCol is None:
                self.nSubTitleRows = 1
            else:
                self.nSubTitleRows = 0
        else:
            self.nSubTitleRows = 0

        self.totalRow = totalRow
        self.specialAttributes = None
            
        self.size = size
        self.ncols = len(titles)

        self.udata = []     # space for EditCells

        self.allowGridCellChangeEvent = False
        self._init_ctrls(parent)
        self.__do_layout()

        self.selectedRow = None
        self.hasChanges = False
        self.hasChangesFromChoice = False
        self.userChange = False             # flag for symmetry with DisplayClasses 
        self.rowOfLastChange = None

        self.eventHandler = None
        self.autoDetectSelection = False    # distinction necessary, because selection with single left click
                                            # w/o CTRL or SHIFT does NOT activate the TopLeft and BottomRight - Info
                                            # of the grid

        self.blockEventHandlers = False
        
        # data cell attributes
        attr = wx.grid.GridCellAttr()
        attr.SetTextColour(GRID_LETTER_COLOR)
        attr.SetBackgroundColour(GRID_BACKGROUND_COLOR)
        attr.SetFont(wx.Font(GRID_LETTER_SIZE, wx.SWISS, wx.NORMAL, wx.NORMAL))
        attr.SetAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)

        self.cellAttr = attr

        self.grid.CreateGrid(0, self.ncols)
        self.grid.EnableGridLines(True)
        
        self.grid.SetDefaultRowSize(rowSize)
        self.grid.SetRowLabelSize(1)
        self.grid.SetColLabelSize(colLabelSize)
        self.grid.SetDefaultColSize(colSize)
        
##        panelcolour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE)
        self.grid.SetDefaultCellBackgroundColour(GRID_BACKGROUND_COLOR)

        self.grid.EnableEditing(True)
        
        self.grid.SetSelectionMode(wx.grid.Grid.wxGridSelectRows)

        editor = wx.grid.GridCellAutoWrapStringEditor() 
        self.grid.SetDefaultEditor(editor)
        self.defaultEditor = editor

        self.grid.GetGridWindow().Bind(wx.EVT_MOTION, self.onMouseOver)

        # wrap text in cells (unfortunately does not work for labels)
        renderer = wx.grid.GridCellAutoWrapStringRenderer()
        self.grid.SetDefaultRenderer(renderer)

        # write column titles
        self.grid.SetLabelFont(wx.Font(GRID_LABEL_SIZE, wx.DEFAULT, wx.NORMAL, wx.BOLD))
        for col, title in enumerate(self.titles):
            maxChar = self.grid.GetColSize(col) / 8
            wTitle = simpleWrapper(title, maxChar)
            self.grid.SetColLabelValue(col, U(wTitle))

        # define grid cell attributes
        for c in range(self.ncols):
            cellAttr = self.cellAttr.Clone()
            if hAlignment:
                cellAttr.SetAlignment(hAlignment[c], wx.ALIGN_CENTRE)
            if readOnly:
                cellAttr.SetReadOnly(readOnly[c])
                if not readOnly[c]: cellAttr.SetBackgroundColour(COLOUR_DATAENTRY_BG)
                self.readOnly = readOnly
            else:
                self.readOnly = [False]*self.ncols
                cellAttr.SetBackgroundColour(COLOUR_DATAENTRY_BG)

            self.grid.SetColAttr(c, cellAttr)

        # assign special attribute for subtitle rows (unit selectors) or unit selector column
        for r in range(self.nSubTitleRows):
            for c in range(self.ncols):
                self.grid.SetCellFont(r, c, wx.Font(GRID_LETTER_SIZE, wx.SWISS, wx.ITALIC, wx.NORMAL))

        if self.unitCol is not None:
            cellAttr = self.cellAttr.Clone()
            cellAttr.SetFont(wx.Font(GRID_LETTER_SIZE, wx.SWISS, wx.ITALIC, wx.NORMAL))
            self.grid.SetColAttr(self.unitCol, cellAttr)
            

##        self.grid.SetGridCursor(0, 0)
        self.selection = []

        self.data=[]
        self.choices = None
        self.toolTips = [["tool-tip %s, %s"%(line+1, col+1) for col in xrange(4)] for line in xrange(20)]

        self.setPopUpMenu(DEFAULTMENU)

        self.tipCoords = [None, None]

    def _init_ctrls(self, prnt):
        # generated method, don't edit
##        self.SetClientSize(wx.Size(800, 600))

        self.grid = wx.grid.Grid(id=-1, name='TableEntryGrid',
              parent=self,
              pos=wx.Point(0, 0),
              size = self.size,
              style=0)

        self.grid.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.OnGridCellLeftDclick)
        self.grid.Bind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK, self.OnGridCellRightClick)
        self.grid.Bind(wx.grid.EVT_GRID_CELL_LEFT_CLICK,  self.OnGridCellLeftClick)
        self.grid.Bind(wx.grid.EVT_GRID_CELL_CHANGE, self.OnGridCellChange)
        self.grid.Bind(wx.grid.EVT_GRID_EDITOR_HIDDEN, self.OnGridEditorHidden)
        self.grid.Bind(wx.grid.EVT_GRID_RANGE_SELECT, self.OnSelectRange)
        self.grid.Bind(wx.grid.EVT_GRID_LABEL_RIGHT_CLICK, self.OnGridLabelRightClick)

    def __do_layout(self):
        # global sizer for panel.
        sizerGlobal = wx.BoxSizer(wx.VERTICAL)
        sizerGlobal.Add(self.grid, 1, wx.ALIGN_TOP | wx.EXPAND, 0)
        self.SetSizer(sizerGlobal)

#=================================================================================
# functions for customisation of TableEntry
#=================================================================================
    def setColSize(self, col, size):
        if col is None:
            self.grid.SetDefaultColSize(size)
        else:
            self.grid.SetColSize(col, size)
            maxChar = self.grid.GetColSize(col) / 8
            title = self.titles[col]
            wTitle = simpleWrapper(title, maxChar)
            self.grid.SetColLabelValue(col, U(wTitle))


    def setExternalEventHandler(self, method):
        """ defines an external event handler of the type:

        some_method(eventType, row, col, selection)
        """
        self.eventHandler = method

    def setTitles(self, titles):
        for col, title in enumerate(titles):
            maxChar = self.grid.GetColSize(col) / 8
            wTitle = simpleWrapper(title, maxChar)
            self.grid.SetColLabelValue(col, U(wTitle))
            
    def setAttribute(self, attribute, values, row=None, col=None):
        """
        possible attributes:
        edit:       True/False if cell can be edited or not
        colour:     Text colour
        background: Background colour of the cell
        style:      normal, bold, italic, bold+italic

        value:      can be a list -> values by column or a single parameter
        """
        self.blockEventHandlers = True
        if row is None:
            nrows = self.grid.GetNumberRows() - self.nSubTitleRows
            rows = range(nrows)
        else:
            rows = row if isinstance(row, list) else [row]

        if col is None:
            cols = range(self.ncols)
        else:
            cols = col if isinstance(col, list) else [col]

        if not hasattr(values, "__iter__"):
            values = [values]*self.ncols

        for r in rows:
            rgrid = r + self.nSubTitleRows
            for c in cols:
                if attribute == "edit":
                    self.grid.SetReadOnly(rgrid, c, isReadOnly=not values[c])
                    colour = COLOUR_DATAENTRY_BG if values[c] else GRID_BACKGROUND_COLOR
                    self.grid.SetCellBackgroundColour(rgrid, c, colour)
                        
                elif attribute == "colour":
                    self.grid.SetCellTextColour(rgrid, c, values[c])

                elif attribute == "background":
                    self.grid.SetCellBackgroundColour(rgrid, c, values[c])

                elif attribute == "style":
                    style1 = wx.ITALIC if "italic" in values[c] else wx.NORMAL
                    style2 = wx.BOLD if "bold" in values[c] else wx.NORMAL
                    #TODO: retreive current size and font type instead of overwriting with defaults
                    self.grid.SetCellFont(rgrid, c, wx.Font(GRID_LETTER_SIZE, wx.SWISS, style1, style2))

                elif attribute == "unit2":
                    self.udata[r][c].updateSecondaryUnit(values[c])                                                                                                      
                    self.grid.SetCellValue(rgrid, c, self.udata[r][c].dstr)

##        elif attribute == "font":
##             attr.SetFont(wx.Font(GRID_LETTER_SIZE, wx.SWISS, wx.NORMAL, wx.NORMAL))
##
##        elif attribute == "alignment":
##            attr.SetAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
        self.blockEventHandlers = False

    def setToolTips(self, toolTips):
        self.toolTips = toolTips

    def setPopUpMenu(self, titles):
        """ definition of pop-up menu
        the index of the title in titles is identical to the
        actionCode received by changeData

        The PopUpMenu is set to DEFAULTMENU in __init__ but can be personalized
        """
        
        self.popUpActive = (len(titles) > 0)
            
        self.popUpMenu = titles
        self.popUpMenuIds = [wx.NewId() for title in titles]

#=================================================================================
# functions for reading from / writing to GUI
#=================================================================================
    def display(self, data=None, selection=None, choices=None, toolTips=None, funit2=None, titles=None, changedCell=None):
        """ displays the content of self.data in the table

        changedCell: hack for temporarily solving problems with GridCellChoiceEditor on Windows 10
                     blocks the update of the GridCellEditor for the indicated cell
        """
        self.setToolTips(toolTips)
        self.blockEventHandlers = True
        # update of table

        if data is not None:
            if self.unitCol is not None: # case of different units for different rows
                    #update unitselectors if sequence of rows has changed
                    self.unitSelectors = [row[self.unitCol] for row in data]
                    defaultUnits = [Globals.unitConversion.defaultUnit(entry[1]) if entry[1] is not None else "" for entry in self.unitSelectors]
                    self.unitLists = [Globals.unitConversion.units[entry[1]] if entry[1] is not None else [] for entry in self.unitSelectors]
                    for r, unitList in enumerate(self.unitLists):
                        if len(self.units) <= r: self.units.append(None)
                        if self.units[r] not in unitList: self.units[r] = defaultUnits[r]

            if choices is not None:
                self.choices = choices
                
            if self.unitSelectors:
                self.data = self.unitConv(data)
                self.data = self.data[:len(data)]
            else:
                self.data = data
                
        elif self.unitSelectors:
            self.data = self.unitConv()

        if selection is not None:
            self.selection = selection
            
        rows = 0 if self.data is None else len(self.data)

        # expand or shrink grid to desired size
        nrows = self.grid.GetNumberRows() - self.nSubTitleRows
        oldTotals = None
            
        if rows > nrows:
            self.grid.InsertRows(nrows-1+self.nSubTitleRows, rows-nrows)

            # set default attributes for new rows (seems that this is not done by default, even if colAttributes had been set)
            for r in range(max(nrows-1, 0), rows):
                editable = [not ro for ro in self.readOnly]
                self.setAttribute("edit", editable, row=r)

                # restore font for rows (necessary for the case of expanding the table from a table with only the subTitleRow and no content
                for c in range(self.ncols):
                    if c != self.unitCol: self.grid.SetCellFont(r+self.nSubTitleRows, c, wx.Font(GRID_LETTER_SIZE, wx.SWISS, wx.NORMAL, wx.NORMAL))
                
            oldTotals = nrows-1

        elif nrows > rows:
            self.grid.DeleteRows(rows+self.nSubTitleRows, nrows-rows)

        if self.unitSelectors:

            # set secondary units appropriate to list specified
            if funit2 is not None:
                for entry in funit2:
                    r, c, f = entry
                    self.setAttribute("unit2", f, row=r, col=c)
            
            if self.unitCol is None: # unit selectors for columns
                for c in range(self.ncols):
                    if changedCell is None or changedCell[0] >= 0 or changedCell[1]!=c:
                        if self.unitSelectors[c][0]=="float":
                            editor = wx.grid.GridCellChoiceEditor([U(entry) for entry in self.unitLists[c]], allowOthers=False)
                            self.grid.SetCellEditor(0, c, editor)
                            self.grid.SetCellValue(0, c, U(self.units[c]))
                            self.grid.SetReadOnly(0, c, False)
                        elif self.unitSelectors[c][0]=="choice":
                            editor = wx.grid.GridCellChoiceEditor([U(entry) for entry in self.unitLists[c]], allowOthers=False)
                            for r in range(rows):
                                rgrid = r + self.nSubTitleRows

                                self.grid.SetCellEditor(rgrid, c, editor)
                        
                    self.grid.SetCellBackgroundColour(0, c, GRID_BACKGROUND_COLOR)
                    
            else: # unit selectors for rows
                for r in range(rows):
                    if self.unitSelectors[r][0]=="float":
                        editor = wx.grid.GridCellChoiceEditor([U(entry) for entry in self.unitLists[r]], allowOthers=False)
                        if changedCell is None or (r, self.unitCol) != changedCell:
                            self.grid.SetCellEditor(r, self.unitCol, editor)
                        unit = self.units[r]
                        if unit not in self.unitLists[r]:
                            unit = defaultUnits[r]
                            self.units[r] = unit
                        self.grid.SetCellValue(r, self.unitCol, U(self.units[r]))
                        
                    self.grid.SetCellBackgroundColour(r, self.unitCol, GRID_BACKGROUND_COLOR)
                    
        self.nRows = rows
        for r in range(rows):                
            for c in range(self.ncols):
                if c == self.unitCol: continue

                if changedCell is None or (r, c) !=changedCell: # don´t update the cell editor for that cell which just has been changed
                                                                # leads to crash on Windows 10 otherwise
                    if self.choices is not None and self.choices[r][c] is not None:
                        editor = wx.grid.GridCellChoiceEditor(self.choices[r][c], allowOthers=False)
                        self.grid.SetCellEditor(r+self.nSubTitleRows, c, editor)
                        
                    elif self.unitCol is None and self.unitSelectors is not None and self.unitSelectors[c][0]=="choice":
                        pass   # keep cell editor which has been set above
                    
                    else:
                        currentEditor = self.grid.GetCellEditor(r+self.nSubTitleRows, c)
                        if currentEditor != self.defaultEditor:
                            self.grid.SetCellEditor(r+self.nSubTitleRows, c, self.defaultEditor)

                dstr = U(str(noneCheck(self.data[r][c], "")))

                self.grid.SetCellValue(r+self.nSubTitleRows, c, dstr)

        if self.totalRow:
            rTotals = rows-1
            self.setAttribute("style", "bold", row=rTotals)
            self.setAttribute("background", GRID_BACKGROUND_COLOR, row=rTotals)

##            if oldTotals is not None:    # reset old total row to default
##                self.setAttribute("style", "normal", row=oldTotals)
##                self.setAttribute("background", GRID_BACKGROUND_COLOR, row=oldTotals)

        if selection is not None:
            self.grid.ClearSelection()
            for index in selection:
##                self.grid.SelectRow(index, addToSelected=True)
                self.grid.SelectBlock(index+self.nSubTitleRows, 0, index+self.nSubTitleRows, self.ncols-1, addToSelected=True)
                self.grid.MakeCellVisible(index+self.nSubTitleRows, 0)
            
        elif self.selectedRow:
##            self.grid.SelectRow(self.selectedRow)
            self.grid.SelectBlock(self.selectedRow+self.nSubTitleRows, 0, self.selectedRow+self.nSubTitleRows, self.ncols-1)
            self.grid.MakeCellVisible(self.selectedRow+self.nSubTitleRows, 0)

        if titles:
            self.setTitles(titles)
            
        self.Show()

        self.blockEventHandlers = False
        
    def getRow(self, r=None):
        """ returns a row of data """
        if r is None: r= self.selectedRow

        data = []
        for c in range(self.ncols):
            data.append(toUTF8(self.grid.GetCellValue(r+self.nSubTitleRows, c)))
        return data
            
    def getCol(self, c):
        """ returns a column of data """
        data = []
        for r in range(self.nrows):
            data.append(toUTF8(self.grid.GetCellValue(r+self.nSubTitleRows, c)))
        return data

    def getTable(self):
        """ returns all data """
        data = []
        for r in range(self.nrows):
            data.append(self.getRow(r))
        return data
            
#==============================================================================
#   EVENT HANDLERS
#==============================================================================
   
    def OnGridCellLeftClick(self, event):
        """ Detects selected row
        """
        selectedRow = event.GetRow() - self.nSubTitleRows
        if selectedRow < 0:
            if event: event.Skip()
            return
                              

        # handling of selection
        ctrlDown = event.ControlDown()
        shftDown = event.ShiftDown()
##        print "\nOnLeftClick=(row=%s ctrl=%s shift=%s ========================================"%(selectedRow, ctrlDown, shftDown)
        if not (ctrlDown or shftDown):
            self.autoDetectSelection = False
            self.selection = [selectedRow]
        else:
            self.selection = None   # unknown -> to be determined in OnSelectRange-EventHandler
            self.autoDetectSelection = True

##        print "TEST TEST TableEntry.OnGridLeftClick: autoDetect %s selection = %s"%(self.autoDetectSelection, self.selection)
##        print "========================================================================="
        
        self.selectedCol = event.GetCol()
        err=0
        if self.hasChanges:
            r = self.rowOfLastChange
            displayRow = self.getRow(r)
            if self.unitSelectors:
##                c = event.GetCol()
                c = self.colOfLastChange
                self.udata[r][c].read(displayRow[c])
                dataRow = [u.val for u in self.udata[r]]
            else:
                dataRow = displayRow
                
            if self.changeData is not None:
                err, self.data, self.choices, selection = expandResults(self.changeData(self.rowOfLastChange, -1, data=dataRow))
##            if self.choices is not None:
#                self.display(choices=self.choices)
            self.display(data=self.data, choices=self.choices)
            ### TODO - development notes 25.01.2019 - change had been necessary for TC1.01 (eliminating QOpProc by setting QOpProc = 0
            ### -> editing of a table cell changes table size.
            ### -> this is not 100% consistent yes. what if in hypothetical cases the same change appears, but self.choices is None ?
                
            self.hasChanges=False

        if selectedRow != self.selectedRow:
            if err==0:
                self.selectedRow = selectedRow
            else:
                pass
            
        if self.eventHandler:
            self.eventHandler("LeftClick", selectedRow, self.selectedCol, self.selection)

        if event: event.Skip()

       
    def OnGridCellLeftDclick(self, event):
        """ double click identical to single click
        """

##        print "\nTEST TEST TableEntry.OnGridLeftDclick: rows = %s\n-------------"%(event.GetRow())

        if self.eventHandler:
            self.eventHandler("LeftDclick", event.GetRow()-self.nSubTitleRows, event.GetCol(), self.selection)
            
        event.Skip()

    def OnGridCellRightClick(self, event):
        """ right click of mouse for carrying out any action
        """

        self.OnGridCellLeftClick(event)

        if self.popUpActive:

            if hasattr(event, "who"): # RobbieEvent
                # emulates the action that would be done by the OnPopUp event handler
                row = event.GetRow()-self.nSubTitleRows
                if self.changeData is not None:
                    err, data, choices, selection = expandResults(self.changeData(row, event.actionCode))
                
                self.display(data, choices=choices, selection=[row])
            else:
                # create pop-up
                menu = wx.Menu()
                for i, title in enumerate(self.popUpMenu):
                    id = self.popUpMenuIds[i]
                    menu.Append( id, U(title))
                    wx.EVT_MENU(menu, id, self.OnPopUp)

                # Display menu with call to PopupMenu, invoked on the source component, passing event's GetPoint. ###
                self.PopupMenu(menu, event.GetPosition())
                menu.Destroy() # destroy to avoid mem leak

            if self.eventHandler:
                self.eventHandler("RightClick", event.GetRow()-self.nSubTitleRows, event.GetCol(), self.selection)

            event.Skip()

    def OnGridLabelRightClick(self, event):
        """ right click of mouse for carrying out any action
        """

        self.OnGridCellRightClick(event)
        
    def OnGridCellChange(self, event):
        """Event handler activated when a change in the alternative table data
            is carried out
        Reads out the new data in the table and saves into the active alternative
        """
        if self.blockEventHandlers: return

        self.rowOfLastChange = event.GetRow() - self.nSubTitleRows
        self.colOfLastChange = event.GetCol()
        
        r = self.rowOfLastChange
        row = self.getRow(r)
        
        if self.hasChangesFromChoice:
            if r < 0: # change of units ?
                c = event.GetCol()
                self.units[c] = row[c]
                self.display(changedCell=(r,c))
            elif self.colOfLastChange==self.unitCol: # change of units in unit-column
                self.units[r] = row[self.unitCol]
                self.display(changedCell=(r, self.unitCol))
            else:
                # TODO: could be simplified here by using udata for ALL (1:1 conversion if no unitselector)
                if self.unitSelectors:
                    c = event.GetCol()
                    self.udata[r][c].read(row[c])
                    dataRow = [u.val for u in self.udata[r]]              
                else:
                    dataRow = row
                    
                if self.changeData is not None:
                    err, self.data, self.choices, selection = expandResults(self.changeData(self.rowOfLastChange, -1, data=dataRow))
                    
                if self.choices is not None:
        
                    self.display(data=self.data, choices=self.choices, changedCell=(r, c))
            self.hasChangesFromChoice = False

        else:
            c = event.GetCol()
            if self.unitSelectors:
                self.udata[r][c].read(row[c])
                
            self.userChange = True
            self.hasChanges = True

    def OnGridEditorHidden(self, event):
        if self.blockEventHandlers: return

        row = event.GetRow() - self.nSubTitleRows
        col = event.GetCol()

        isChoice = False
        if self.unitSelectors:
            if col==self.unitCol or self.unitCol==None and row<0:
                isChoice = True
                
        if self.choices is not None and self.choices[row][col] is not None:
            isChoice = True

        if isChoice:       
            self.hasChangesFromChoice = True
            self.rowOfLastChange = event.GetRow() - self.nSubTitleRows
            
        if event: event.Skip()

    def OnSelectRange(self, event):
        if self.blockEventHandlers: return        # block event handler while changing the grid automatically
        
        topLeft = self.grid.GetSelectionBlockTopLeft()
        bottomRight = self.grid.GetSelectionBlockBottomRight()
##        print "TEST TEST TableEntry.OnSelectRange(%s): topLeft = %s bottomRight = %s"%(self.autoDetectSelection, topLeft, bottomRight)
        if self.autoDetectSelection:
            self.selection = []
            for i, top in enumerate(topLeft):
                bottom = bottomRight[i]
                self.selection += range(top[0]- self.nSubTitleRows, bottom[0]+1-self.nSubTitleRows)
##            print "TEST TEST TableEntry.OnSelectRange(%s): selection = %s"%(self.autoDetectSelection, self.selection)
                
        if self.eventHandler:
            self.eventHandler("SelectRange", None, None, self.selection)

        if event: event.Skip()
        
    def OnPopUp(self, event):
        actionCode = self.popUpMenuIds.index(event.GetId())

        if self.changeData is not None:
            err, data, choices, selection = expandResults(self.changeData(self.selectedRow, actionCode))
        
            self.display(data, choices=choices, selection=selection)

    def GetValue(self):
        """
        function for symmetry with displayClasses - entry classes
        returns the entire table
        """
        if self.unitSelectors:
            nRows = len(self.udata)
            nCols = len(self.udata[0]) if nRows > 0 else 0
            
            # first write a row with the selected units
            data = [[self.udata[r][c].val for c in xrange(nCols)] for r in xrange(nRows)]
            return data            
        else:
            return self.data

    def SetValue(self, dataDict, emitEvent=True):
        """
        function for symmetry with displayClasses - entry classes
        returns the entire table
        """
        data = dataDict.get("data")
        choices = dataDict.get("choices")
        toolTips = dataDict.get("toolTips")
        selection = dataDict.get("selection")
        titles = dataDict.get("titles")

        self.display(data=data, choices=choices, toolTips=toolTips, selection=selection, titles=titles)

        self.userChange = False
        
    def SetCellValue(self, coords, value):
        """
        helper function for Robbie Robot for writing into table cells
        """
        r, c = coords
        previous = self.grid.GetCellValue(r+self.nSubTitleRows, c)
        dstr = str(value)
        self.grid.SetCellValue(r+self.nSubTitleRows, c, dstr)
        if self.unitSelectors:
            self.udata[r][c].read(dstr)
            self.data[r][c] = self.udata[r][c].val
            dataRow = [u.val for u in self.udata[r]]
        else:
            self.data[r][c] = dstr
            displayRow = self.getRow(r)
            dataRow = displayRow

        if self.changeData is not None:
            err, self.data, self.choices, selection = expandResults(self.changeData(r, -1, data=dataRow))
        
    def unitConv(self, data=None):
        """
        unit conversion from internal units (data) into display units (
        if data is None: use the last data as stored in udata """

        ctypes = [entry[0] for entry in self.unitSelectors]
        unitTypes = [entry[1] if entry[0]=="float" else None for entry in self.unitSelectors]
        decimals = [entry[2] for entry in self.unitSelectors]
        choices = [entry[1] if entry[0]=="choice" else None for entry in self.unitSelectors]

        if data is not None:
            for r, row in enumerate(data):
                # add new editCell rows only if necessary
                if r >= len(self.udata):
                    self.udata.append([])
                    for c, val in enumerate(row):
                        if c <= self.unitCol:
                            cell = EditCell("text")
                        else:
                            u = c if self.unitCol is None else r
                            
                            cell = EditCell(ctypes[u],
                                            unitType=unitTypes[u],
                                            decimals=decimals[u],
                                            choices=choices[u])
                            if unitTypes[u] is not None: cell.setUnit(self.units[u])
                        self.udata[r].append(cell)

                for c, val in enumerate(row):
                    if self.udata[r][c].ctype=="choice" and self.choices is not None:
                        self.udata[r][c].choice = self.choices[r][c]
                    self.udata[r][c].write(val)

        else:
            for r, row in enumerate(self.udata):
                for c, cell in enumerate(row):
                    if c == self.unitCol:
                        continue
                    else:
                        u = c if self.unitCol is None else r
                        if unitTypes[u] is not None and cell.ctype=="float":
                            cell.setUnit(self.units[u])

        nRows = len(self.udata)
        nCols = len(self.udata[0]) if nRows > 0 else 0
        
        # first write a row with the selected units
        data = [[self.udata[r][c].dstr for c in xrange(nCols)] for r in xrange(nRows)]
        return data


    def onMouseOver(self, event):
        '''
        Method to calculate where the mouse is pointing and
        then set the tooltip dynamically.
        '''

        # Use CalcUnscrolledPosition() to get the mouse position
        # within the
        # entire grid including what's offscreen
        x, y = self.grid.CalcUnscrolledPosition(event.GetX(),event.GetY())
        coords = self.grid.XYToCell(x, y)
        
        # you only need these if you need the value in the cell
        row, col = coords
        if row <0 or col<0:
            return

        if self.grid.GetCellValue(row, col) is not None and self.toolTips:
            rgrid = row - self.nSubTitleRows
            if rgrid >= 0 and rgrid < self.nRows and col >= 0 and col < self.nCols and coords != self.tipCoords:
                self.tipCoords = coords[:]
                tip = self.toolTips[row][col]
                if tip is None: tip = ""

                tt = wx.ToolTip(tip)
                self.grid.GetGridWindow().SetToolTip(tt)

        event.Skip()
                
def expandResults(res):
    """ auxiliary function for backward compatibility with old changeData functions which
        had less than the three required results err, data, choices
    """
    if hasattr(res, "__iter__"):
        err = res[0]
        data = res[1] if len(res) > 1 else [[]]
        choices = res[2] if len(res) > 2 else None
        selection = res[3] if len(res) > 3 else None
        
    else:
        err = res
        data = [[]]
        choices = None
        selection = None
        
    return err, data, choices, selection

def simpleWrapper(text, maxChar):
    """
    helper function for auto-wrap of labels.
    seems that wx.grid has the possibility of auto-wrap only for table cells, but not for the label row
    """
    
    words = text.split(" ")
    newText = ""
    nChars = 0
    for word in words:
        if nChars==0 or (nChars + 1 + len(word) <= maxChar): # first word in line or word fits
            newText += " " + word
            nChars += len(word) + 1
        else:
            newText += "\n" + word
            nChars = len(word)
    return newText
        
        
        
