# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#	displayClasses: Some useful classes for panel display
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import locale
import wx.combo
import wx.lib.intctrl
import wx.lib.stattext
import wx.lib.newevent
import wx.calendar
from math import log10
from GUI.status import Status
from globals import Globals

from fonts import FontProperties
from auxiliary.auxiliary import toUnicode as U, toGettextAndUnicode as _U, toUTF8
from constants import TRANSYESNO
from modules.messageLogger import *

CHOOSERBCGCOLOR = (255, 255, 255)
FGCOLOR = (0, 0, 0)
TEXTBKGCOLOR = (255, 255, 255)
EMPTYBKGCOLOR = (255, 255, 255)
INVALIDBKGCOLOR = (255, 255, 0)
LOWERACCEPTEDDATE = '01/01/1900'
UPPERACCEPTEDDATE = '12/31/2050'
LIGHTGREY          = (236,233,216,255)

ENCODING = "utf-8" # internal encoding

ChoiceItemsChanged, EVT_CHOICE_NEW_ALTERNATIVES = wx.lib.newevent.NewEvent()
EnableChanged, EVT_WIDGET_ENABLE_SWITCH = wx.lib.newevent.NewEvent()

def error(text):
    dlg = wx.MessageDialog(None, text, 'Error', wx.OK | wx.ICON_ERROR)
    dlg.ShowModal()
    dlg.Destroy()

def smallFloatEntryParameters(entryWidget, parent, pos):
    """Generate all parameters for a FloatEntry object without a label
    and with an absolute position in its parent as a dict.
    
    Good for entries that are positioned on a bitmap background.
    
    For example self.startTemp is the original widget, then
    
    FloatEntry(**smallFloatEntryParameters(self.startTemp, somePanel, (80, 160)))
    
    will create a FloatEntry in somePanel at position (x=80, y=160) with the
    rest of the parameters like self.startTemp except that the tool-tip
    string is prefixed with the former label string. 
    
    :param entryWidget: a FloatEntry whose parameters are to be copied
    :param parent: the value for "parent" for the result dict 
    :param pos: the value for "pos" in the result dict
    
    """
    params = {"pos": pos}
    params["label"]=""
    try:
        tip = entryWidget.label.GetLabel() + "\n"
    except AttributeError:
        tip = " "
    try:
        tip += entryWidget.entry.GetToolTip().GetTip()
    except AttributeError:
        pass
    params["tip"]= tip
    params["parent"] = parent
    params["decimals"] = entryWidget.decimals
    params["minval"] = entryWidget.entry.min
    params["maxval"] = entryWidget.entry.max
    params["value"] = entryWidget.GetValue()
    params["unitdict"] = entryWidget.unitdict
    params["isStatic"] = entryWidget.isStatic
    return params

def smallChoiceEntryParameters(entryWidget, parent, pos):
    """Generate all parameters for a FloatEntry object without a label
    and with an absolute position in its parent as a dict.
    
    Good for entries that are positioned on a bitmap background.
    
    For example self.startTemp is the original widget, then
    
    FloatEntry(**smallFloatEntryParameters(self.startTemp, somePanel, (80, 160)))
    
    will create a FloatEntry in somePanel at position (x=80, y=160) with the
    rest of the parameters like self.startTemp except that the tool-tip
    string is prefixed with the former label string. 
    
    :param entryWidget: a FloatEntry whose parameters are to be copied
    :param parent: the value for "parent" for the result dict 
    :param pos: the value for "pos" in the result dict
    
    """
    params = {"pos": pos}
    params["label"]=""
    try:
        tip = entryWidget.label.GetLabel() + "\n"
    except AttributeError:
        tip = ""
    try:
        tip += entryWidget.entry.GetToolTip().GetTip()
    except AttributeError:
        pass
    params["tip"]= tip
    params["parent"] = parent

    choices = entryWidget.entry.GetStrings()
    params["values"] = choices[1:] if entryWidget.allowNones else choices
    params["value"] = entryWidget.GetValue()
    params["wUnits"] = 0
    params["allowNones"] = entryWidget.allowNones
    return params
    
def smallFloatEntryClone(entryWidget, parent, pos): 
    """Return a Float entry without a label and with an absolute 
    position in parent. All other parameters are taken from 
    entryWidget and the entry slots of both are kept in sync.
    
    Good for entries that are positioned on a bitmap background and
    have corresponding entries somewhere else in the panel.
    
    For example if self.startTemp is the original widget, then
    
    smallFloatEntryClone(self.startTemp, somePanel, (80, 160))
    
    will return a small FloatEntry in somePanel at position (x=80, y=160) 
    with the rest of the parameters like self.startTemp and any entry
    or unit change in one will propagate to the other. 
    
    :param entryWidget: a FloatEntry whose parameters are to be copied
    :param parent: the parent wx.widget for the resulting widget 
    :param pos: the value for "pos" for the resulting
    
    """
    result = FloatEntry(**smallFloatEntryParameters(entryWidget, parent, pos))
    
    entryWidget.Bind(wx.EVT_TEXT, result.SyncValueTo, entryWidget.entry)
    result.Bind(wx.EVT_TEXT, entryWidget.SyncValueTo, result.entry)
    entryWidget.Bind(wx.EVT_CHOICE, result.SyncUnitTo, entryWidget.units)
    result.Bind(wx.EVT_CHOICE, entryWidget.SyncUnitTo, result.units)
    entryWidget.Bind(EVT_WIDGET_ENABLE_SWITCH, result.SyncEnableTo)
    result.Bind(EVT_WIDGET_ENABLE_SWITCH, entryWidget.SyncEnableTo)
    
    return result

def smallChoiceEntryClone(entryWidget, parent, pos): 
    """Return a ChoiceEntry without a label and with an absolute 
    position in parent. All other parameters are taken from 
    entryWidget and the choice slot of both is kept in sync.
    
    Good for entries that are positioned on a bitmap background and
    have corresponding entries somewhere else in the panel.
    
    For example if self.medium is the original widget, then
    
    smallFloatEntryClone(self.medium, somePanel, (80, 160))
    
    will return a small ChoiceEntry in somePanel at position (x=80, y=160) 
    with the rest of the parameters like self.medium and any change in
    the choices of one will propagate to the other. 
    
    :param entryWidget: a ChoiceEntry whose parameters are to be copied
    :param parent: the value for "parent" for the result dict 
    :param pos: the value for "pos" in the result dict
    
    """
    result = ChoiceEntry(**smallChoiceEntryParameters(entryWidget, parent, pos))
    
    entryWidget.Bind(wx.EVT_CHOICE, result.SyncValueTo, entryWidget.entry)
    result.Bind(wx.EVT_CHOICE, entryWidget.SyncValueTo, result.entry)
    entryWidget.Bind(EVT_CHOICE_NEW_ALTERNATIVES, result.SyncAlternativesTo)
    result.Bind(EVT_CHOICE_NEW_ALTERNATIVES, entryWidget.SyncAlternativesTo)
    entryWidget.Bind(EVT_WIDGET_ENABLE_SWITCH, result.SyncEnableTo)
    result.Bind(EVT_WIDGET_ENABLE_SWITCH, entryWidget.SyncEnableTo)
    
    return result

def smallCheckBoxCloneFromChoiceEntry(entryWidget, parent, pos):
    """Return a wx.CheckBox without a label and with an absolute 
    position in parent. All other parameters are taken from 
    entryWidget and the checked status vs the Yes/No status of both 
    is kept in sync.
    
    Good for entries that are positioned on a bitmap background and
    have corresponding entries somewhere else in the panel.
    
    For example if self.mandatory is the original widget, then
    
    smallCheckBoxClone(self.mandatory, somePanel, (80, 160))
    
    will return a small ChoiceEntry in somePanel at position (x=80, y=160) 
    with the rest of the parameters like self.mandatory and any change in
    the choices of one will propagate to the other. 
    
    :param entryWidget: a ChoiceEntry whose parameters are to be copied
    :param parent: the value for "parent" for the result dict 
    :param pos: the value for "pos" in the result dict
    
    """
    try:
        tip = entryWidget.label.GetLabel() + "\n"
    except AttributeError:
        tip = " " # no label
    try:
        tip += entryWidget.entry.GetToolTip().GetTip()
    except AttributeError:
        pass

    items = [toUTF8(val) for val in entryWidget.entry.GetItems()]
    if items == TRANSYESNO.values():
        style = 0 # only yes or no
    else:
        style = wx.CHK_3STATE # may have undefined states
        
    result = CheckBoxEntry(parent, label="", style=style, pos=pos)
    
    if toUTF8(entryWidget.GetValue(text=True)) not in TRANSYESNO.values():
        result.Set3StateValue(wx.CHK_UNDETERMINED)
    else:
        result.SetValue(toUTF8(entryWidget.GetValue(text=True)) ==  TRANSYESNO["yes"])
    
    result.SetToolTip(wx.ToolTip(tip))
    
    entryWidget.Bind(wx.EVT_CHOICE, result.SyncValueTo, entryWidget.entry)
    result.Bind(wx.EVT_CHECKBOX, entryWidget.SyncValueTo, result)
    entryWidget.Bind(EVT_WIDGET_ENABLE_SWITCH, result.SyncEnableTo)
    result.Bind(EVT_WIDGET_ENABLE_SWITCH, entryWidget.SyncEnableTo)
    
    return result
    
def smallCheckBoxClone(entryWidget, parent, pos):
    """Return a wx.CheckBox without a label and with an absolute 
    position in parent. All other parameters are taken from 
    entryWidget and the checked status of both is kept in sync.
    
    Good for entries that are positioned on a bitmap background and
    have corresponding entries somewhere else in the panel.
    
    For example if self.mandatory is the original widget, then
    
    smallCheckBoxClone(self.mandatory, somePanel, (80, 160))
    
    will return a small ChoiceEntry in somePanel at position (x=80, y=160) 
    with the rest of the parameters like self.mandatory and any change in
    the choices of one will propagate to the other. 
    
    :param entryWidget: a ChoiceEntry whose parameters are to be copied
    :param parent: the value for "parent" for the result dict 
    :param pos: the value for "pos" in the result dict
    
    """
    try:
        tip = entryWidget.GetLabel()
    except AttributeError:
        tip = " " # no label
    
    result = CheckBoxEntry(parent, pos=pos)
    result.SetValue(entryWidget.GetValue())
    result.SetToolTip(wx.ToolTip(tip))
    
    entryWidget.Bind(wx.EVT_CHECKBOX, result.SyncValueTo, entryWidget)
    result.Bind(wx.EVT_CHECKBOX, entryWidget.SyncValueTo, result)
    entryWidget.Bind(EVT_WIDGET_ENABLE_SWITCH, result.SyncEnableTo)
    result.Bind(EVT_WIDGET_ENABLE_SWITCH, entryWidget.SyncEnableTo)
    
    return result

def mirrorTextCtrl(entryWidget, parent, pos=wx.DefaultPosition, size=wx.DefaultSize, tip=""):
    """Return a read only TextCtrlEntry that mirrors the content of 
    entryWidget and has parent as parent.
    
    :param entryWidget: the source widget
    :type entryWidget: wx.TextCtrl
    :param parent: the parent for the mirror
    :param pos: the position of the mirror in parent
    
    """
    result = TextCtrlEntry(parent, pos=pos, style=wx.TE_READONLY, size=size)
    tip = wx.ToolTip(U(tip))
    result.SetToolTip(tip)
    entryWidget.Bind(wx.EVT_TEXT, result.SyncValueTo, entryWidget)
    
    
class FieldSizes(object):
    # sizes of subwidgets
    labelWidth = 100
    dataWidth = 200
    unitsWidth = 80
    totalWidth = labelWidth + dataWidth + unitsWidth
    Height = 32
    
    def __init__(self, wHeight=None, wLabel=None, wData=None, wUnits=None):
        #
        # size properties
        #
        try:
            if wHeight is not None: FieldSizes.Height = int(wHeight)
            if wLabel is not None: FieldSizes.labelWidth = int(wLabel)
            if wData is not None: FieldSizes.dataWidth = int(wData)
            if wUnits is not None: FieldSizes.unitsWidth = int(wUnits)
            FieldSizes.totalWidth = int(FieldSizes.labelWidth + FieldSizes.dataWidth + FieldSizes.unitsWidth + 6)
        except:
            error('Error in FieldSizes: wHeight=%s, wLabel=%s,wData=%s,wUnits=%s' % \
                                   (wHeight, wLabel, wData, wUnits))

        
class Generics(object):
    def makeColour(self, something):
        if isinstance(something, tuple):
            try:
                newcolor = wx.Colour(something[0], something[1], something[2])
                something = newcolor
            except:
                error(repr(something) + ' is not a valid color, sorry')
                something = wx.Colour(100, 100, 100)
                
        return something

    def setTooltips(self, other, tip):
        if tip:
            t = (tip.strip())
            if len(t) > 0:
                other.entry.SetToolTipString(t)
                try:
                    # the label can exist or not
                    other.label.SetToolTipString(t)
                except:
                    pass

    def setUnits(self, other, tip, unitdict, defaultDisplayUnit=None):
        """ sets up the DEFAULT display unit and the unit list of the entry """
        
        if unitdict is not None:
            other.units.Clear()
            try:
                try:
                    # find the whole list of possible units of this class
                    unitlist = Globals.unitConversion.units[unitdict].keys()
                    # find default display unit
                    if ((defaultDisplayUnit is None) or
                        (defaultDisplayUnit not in unitlist)):
                        defaultDisplayUnit = Globals.unitConversion.defaultUnit(unitdict)
                    # load the choice
                    for u in unitlist:
                            other.units.Append(U(u))
                    # set the choice to the default unit
                    other.units.SetStringSelection(U(defaultDisplayUnit))
                except KeyError:
                    error(unitdict + ' is not a defined measurement class')
            except TypeError:
                error(repr(unitdict) + ' is not a string')
            other.units.SetToolTipString(tip)
        return defaultDisplayUnit

    def setSizes(self, wLabel, wData, hasunits, wUnits):
        h = FieldSizes.Height

        # sets label size
        if wLabel is None:
            lW = FieldSizes.labelWidth
        else:
            lW = wLabel
        lblSize = wx.Size(lW, h)

        # sets data size
        if hasunits:
            if wData is None:
                lD = FieldSizes.dataWidth
            else:
                lD = wData
        else:
            if wData is None:
                lD = FieldSizes.dataWidth + FieldSizes.unitsWidth + 1
            else:
                lD = wData
        datSize = wx.Size(lD, h)
        # sets units size
        if wUnits is None:
            lU = FieldSizes.unitsWidth
        else:
            lU = wUnits
        uniSize = wx.Size(lU, h)

        totSize = wx.Size(lW + lD + lU + 2, h + 1)
        return (totSize, lblSize, datSize, uniSize)


    def getChoiceValues(self, other, multiple, text):
        if multiple:
            # multiple values can be selected
            if text:
                # return a list with the texts of the selections
                slist = other.GetSelection()
                return tuple(slist)
                    
            else:
                # we return a tuple with the indices of the selections
                return other.GetSelection()
        else:
            # single values can be selected
            if text:
                # return a tuple with the texts of the selections
                return other.GetStringSelection()
            else:
                # we return the index of the (single) selection
                return other.GetCurrentSelection()



class ListCombo(wx.ListCtrl, wx.combo.ComboPopup):
    def __init__(self):
        # Since we are using multiple inheritance, and don't know yet
        # which window is to be the parent, we'll do 2-phase create of
        # the ListCtrl instead, and call its Create method later in
        # our Create method.  (See Create below.)
        self.PostCreate(wx.PreListCtrl())
        # Init the ComboPopup base class.
        wx.combo.ComboPopup.__init__(self)

    # This is called immediately after construction finishes.  You can
    # use self.GetCombo if needed to get to the ComboCtrl instance.
    def Init(self):
        self.value = -1
        self.curitem = -1


    def AddItem(self, txt):
        self.InsertStringItem(self.GetItemCount(), txt)

    def OnMotion(self, evt):
        item, flags = self.HitTest(evt.GetPosition())
        if item >= 0:
            #self.Select(item)
            self.curitem = item

    def OnLeftDown(self, evt):
        pass
        #self.value = self.curitem
        #self.Dismiss()

    def OnLeftDClick(self, evt):
        # close the listbox
        self.Dismiss()

    # The following methods are those that are overridable from the
    # ComboPopup base class.

    # Create the popup child control.  Return true for success.
    def Create(self, parent):
        wx.ListCtrl.Create(self, parent, style=wx.LC_LIST | wx.SIMPLE_BORDER)
        self.Bind(wx.EVT_MOTION, self.OnMotion)
        #self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_DCLICK, self.OnLeftDClick)
        return True


    # Return the widget that is to be used for the popup
    def GetControl(self):
        return self

    # Called just prior to displaying the popup, you can use it to
    # 'select' the current item.
    def SetStringValue(self, val):
        idx = self.FindItem(-1, val)
        if idx != wx.NOT_FOUND:
            self.Select(idx)

    # Return a string representation of the current item.
    def GetStringValue(self):
        if self.value >= 0:
            return self.GetItemText(self.value)
        return ""

    # Called immediately after the popup is shown
    def OnPopup(self):
        wx.combo.ComboPopup.OnPopup(self)

    # Called when popup is dismissed
    def OnDismiss(self):
        s = ';'.join(self.GetSelection())
        combo = self.GetCombo()
        combo.SetText(s)
        wx.combo.ComboPopup.OnDismiss(self)

    # Receives key events from the parent ComboCtrl.  Events not
    # handled should be skipped.
    def OnComboKeyEvent(self, event):
        wx.combo.ComboPopup.OnComboKeyEvent(self, event)

    # Implement if you need to support special action when user
    # double-clicks on the parent wxComboCtrl.
    def OnComboDoubleClick(self):
        wx.combo.ComboPopup.OnComboDoubleClick(self)

    def Clear(self):
        self.ClearAll()

    # utility functions for implementing a GetSelection method
    def GetFirstSelected(self):
        """return first selected item, or -1 when none"""
        return self.GetNextSelected(-1)

    def GetNextSelected(self, item):
        """return subsequent selected items, or -1 when no more"""
        return self.GetNextItem(item, wx.LIST_NEXT_ALL, wx.LIST_STATE_SELECTED)

    def GetSelection(self):
        sel = []
        i = self.GetFirstSelected()
        while i >= 0:
            s = self.GetItemText(i)
            sel.append(s)
            i = self.GetNextSelected(i)
        return sel
        
    def UnselectAll(self):
        """Set all items to unselected state"""
        i = self.GetFirstSelected()
        while i >= 0:
            self.SetItemState(i, 0, wx.LIST_STATE_SELECTED)
            i = self.GetNextSelected(i)

class CFloat(wx.TextCtrl):
    def __init__(self, prnt,
                 id= -1,
                 pos=wx.DefaultPosition,
                 size=(100, 32),
                 min= -1e38,
                 max= +1e38,
                 foregroundColour=FGCOLOR,
                 emptyBackgroundColour=EMPTYBKGCOLOR,
                 validBackgroundColour=TEXTBKGCOLOR,
                 invalidBackgroundColour=INVALIDBKGCOLOR,
                 style=0,
                 value=0.0,
                 decimals=2,
                 nosep=False):
        self.prnt = prnt
        self.min = min
        self.max = max
        self.decimals = decimals

        u = Generics()
        self.foregroundColour = u.makeColour(foregroundColour)
        self.emptyBackgroundColour = u.makeColour(emptyBackgroundColour)
        self.validBackgroundColour = u.makeColour(validBackgroundColour)
        self.invalidBackgroundColour = u.makeColour(invalidBackgroundColour)
        
        wx.TextCtrl.__init__(self, id=id, parent=prnt,
                             pos=pos,
                             size=size,
                             style=style)
        # bind some events
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_TEXT, self.OnEntry)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)

        # get a dictionary of local parameters
        loc = locale.localeconv()
        self.dc = loc['decimal_point']
        if nosep:
            self.tsep = ''
        else:
            self.tsep = loc['thousands_sep']
        self.tsep = ' ' # just for testing
        self.dc = '.'   # testing

        if self.tsep == self.dc:
            # if decimal dot is the same as thousands separator,
            # default to something sensible
            self.tsep = ','
            self.dc = '.'

    def OnChar(self, event):
        # filter all characters except deletes, positioning,
        # digits, minus sign, decimal point
        c = event.GetKeyCode()
        if c == wx.WXK_BACK or c == wx.WXK_DELETE or c == wx.WXK_RIGHT or c == wx.WXK_LEFT:
            event.Skip()
        elif (c >= ord('0') and c <= ord('9')) or c == ord('-') or c == ord(self.dc):
            event.Skip()
        else:
            # ignore invalid keys
            event.Skip(False)


    def OnEntry(self, event):
        s = event.GetString()
        # sets value in parent (FloatEntry instance)
        self.prnt.lastInternalValue = s 
        if s.strip() == '':
            # don't touch null value
            self.SetBackgroundColour(self.emptyBackgroundColour)
        else:
            t = self.__clean_dot(s)
            if self.__isValidFloat(t):
                f = float(t)
                ignore = self.verifyLimits(f)
            else:
                self.SetBackgroundColour(self.invalidBackgroundColour)
                
        event.Skip()

    def OnFocus(self, event):
        # clean thousands separators for editing
        s = self.GetValue()
        if s.strip() == '':
            # don't touch null value
            return
        s1 = self.__clean_tsep(s)
        self.ChangeValue(s1)
        event.Skip()

    
    def OnKillFocus(self, event):
        value = self.GetValue()
        if value.strip() == '':
            self.SetBackgroundColour(self.emptyBackgroundColour)
        else:
            f = self.toFloat(value)
            if f is not None:
                value = self.verifyLimits(f)
        self.showFormatted(str(value), emitEvent=True)
        event.Skip()

    def toFloat(self, value):
        s1 = self.__clean_tsep(value) 
        s2 = self.__clean_dot(s1)
        try:
            f = float(s2)
            return f
        except:
            return None
        
    def verifyLimits(self, value):
        if value > self.max:
            self.SetBackgroundColour(self.invalidBackgroundColour)
            return self.max
        elif value < self.min:
            self.SetBackgroundColour(self.invalidBackgroundColour)
            return self.min
        else:
            self.SetBackgroundColour(self.validBackgroundColour)
        return value
        
    def SetValue(self, value, emitEvent=False):
        # several empty field conditions
        if value is None or (isinstance(value, str) and value == ''):
            value = ''
            if emitEvent is True:
                super(self.__class__, self).SetValue(value)
            else:
                self.ChangeValue(value)
            self.SetBackgroundColour(self.emptyBackgroundColour)
        elif isinstance(value, str) and value != '':
            # value is a string. must convert to float
            value = self.toFloat(value)
            if value is None:
                self.SetBackgroundColour(self.invalidBackgroundColour)

        if isinstance(value, float):
            # verify limits
            self.verifyLimits(value)
            self.showFormatted(str(value), emitEvent)

        return value
    
    def showFormatted(self, s, emitEvent=False):
        if s.strip() == '':
            # don't touch null value
            return
        f = self.toFloat(s)
        sign = ''
        if f < 0.0:
            sign = '-'
            f = abs(f)
        snum = "%.*f" % (self.decimals, f)
        try:
            (integ, frac) = snum.split(".")
        except ValueError:
            integ = snum
            frac = ''
        # set thousands separator
        sn = ''
        r0 = integ[::-1] # reverse
        for i, c in enumerate(r0):
            if (i % 3) == 0 and len(sn) > 1:
                sn += self.tsep
            sn += c
        s3 = sign + sn[::-1] + self.dc + frac
        if emitEvent is True:
            super(self.__class__, self).SetValue(s3)
        else:
            self.ChangeValue(s3)

    def __isValidFloat(self, s):
        try:
            float(str(s))
            return True
        except ValueError:
            return False

    def __clean_tsep(self, s):
        if not isinstance(s, str):
            s = str(s)
        if not self.__isValidFloat(s):
            return s.replace(self.tsep, '')
        return s
    
    def __clean_dot(self, s):
        s = str(s) # take care of unicode strings
        if not self.__isValidFloat(s):
            s = s.replace(self.dc, '.')
            s = s.replace(' ', '')
        return s


class CDate(wx.TextCtrl):
    def __init__(self, prnt, prnt2,
                 id= -1,
                 pos=wx.DefaultPosition,
                 size=(100, 32),
                 min=LOWERACCEPTEDDATE,
                 max=UPPERACCEPTEDDATE,
                 foregroundColour=FGCOLOR,
                 emptyBackgroundColour=EMPTYBKGCOLOR,
                 validBackgroundColour=TEXTBKGCOLOR,
                 invalidBackgroundColour=INVALIDBKGCOLOR,
                 style=0,
                 value=wx.DateTimeFromDMY(1, 0, 2007)):

        self.prnt = prnt
        self.prnt2 = prnt2
        self.cal = None
        self.min = wx.DateTime()
        self.min.ParseDate(min)
        self.max = wx.DateTime()
        self.max.ParseDate(max)
        if isinstance(value, wx.DateTime):
            self.initialdate = value
        else:
            self.initialdate = wx.DateTime()
            self.initialdate.ParseDate(value)
            
        u = Generics()
        self.foregroundColour = u.makeColour(foregroundColour)
        self.emptyBackgroundColour = u.makeColour(emptyBackgroundColour)
        self.validBackgroundColour = u.makeColour(validBackgroundColour)
        self.invalidBackgroundColour = u.makeColour(invalidBackgroundColour)

        wx.TextCtrl.__init__(self, id=id, parent=prnt,
                             pos=pos,
                             size=size,
                             style=style)
        # bind some events
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_TEXT, self.OnEntry)
        self.Bind(wx.EVT_SET_FOCUS, self.OnFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)

    def OnCalSelected(self, event):
        newdate = event.GetDate()
        self.showFormatted(newdate)
        #cal = event.GetEventObject()
        self.calpanel.Destroy()
        self.cal = None
        self.prnt2.SetFocus()
        
    def OnCalSelChanged(self, event):
        cal = event.GetEventObject()
        date = cal.GetDate()
        event.Skip()

    def OnCharCal(self, event):
        # any keypress on the calendar closes the window
        # this works in Windows but not in FreeBSD
        self.calpanel.Destroy()
        self.cal = None
        event.Skip()

    def OnChar(self, event):
        # if the calendar is displayed, delete it
        if self.cal is not None:
            self.calpanel.Destroy()
            self.cal = None
        
        # filter all characters except digits, '-', '/'
        c = event.GetKeyCode()
        if c == wx.WXK_BACK or c == wx.WXK_DELETE or c == wx.WXK_RIGHT or c == wx.WXK_LEFT:
            event.Skip()
        elif (c >= ord('0') and c <= ord('9')) or c == ord('-') or c == ord('/'):
            event.Skip()
        else:
            # ignore invalid keys
            event.Skip(False)

    def OnEntry(self, event):
        s = event.GetString()
        if s.strip() == '':
            # don't touch null value
            self.SetBackgroundColour(self.emptyBackgroundColour)
        else:
            self.verifyLimits(s)
                
        event.Skip()

    def OnFocus(self, event):
        if self.cal is None:
            # show calendar control
            (x, y) = self.GetPositionTuple()
            (xp, yp) = self.prnt.GetPositionTuple()
            (w, h) = self.prnt.GetSizeTuple()
            # frame for the calendar
            self.calpanel = wx.Frame(self.prnt2, id= -1,
                                     style=wx.STAY_ON_TOP)
            self.cal = wx.calendar.CalendarCtrl(self.calpanel, -1,
                                                self.initialdate, pos=(0, 0),
                                                style=wx.calendar.CAL_NO_YEAR_CHANGE|wx.calendar.CAL_SHOW_HOLIDAYS)
            (wc, hc) = self.cal.GetSizeTuple()
            self.calpanel.SetDimensions(x + xp, yp + h, wc, hc)
            self.calpanel.Show()
            self.cal.Bind(wx.calendar.EVT_CALENDAR_SEL_CHANGED, self.OnCalSelChanged)
            self.cal.Bind(wx.calendar.EVT_CALENDAR, self.OnCalSelected)
            self.cal.Bind(wx.EVT_CHAR, self.OnCharCal)

        event.Skip()
    
    def OnKillFocus(self, event):
        # if there is input from the calendar control
        value = self.GetValue()
        if value.strip() == '':
            self.SetBackgroundColour(self.emptyBackgroundColour)
        else:
            self.verifyLimits(value)
        self.showFormatted(value)
        event.Skip()

    def verifyLimits(self, value):
        if isinstance(value, wx.DateTime):
            d = value
        else:
            # value is a string
            d = wx.DateTime()
            rsp = d.ParseDate(value)
            if rsp == -1:
                self.SetBackgroundColour(self.invalidBackgroundColour)
                return

        if d.IsLaterThan(self.max):
            self.SetBackgroundColour(self.invalidBackgroundColour)
        elif d.IsEarlierThan(self.min):
            self.SetBackgroundColour(self.invalidBackgroundColour)
        else:
            self.SetBackgroundColour(self.validBackgroundColour)
        
    def SetValue(self, value):
        # several empty field conditions
        if value is None:
            value = ''
            self.ChangeValue(value)
            self.SetBackgroundColour(self.emptyBackgroundColour)
        elif isinstance(value, wx.DateTime):
            s = value.FormatISODate()
            value = s
            self.verifyLimits(value)
            self.ChangeValue(value)
        elif isinstance(value, str):
            if value.strip() == '' or value.strip() == 'None':
                self.ChangeValue('')
                self.SetBackgroundColour(self.emptyBackgroundColour)
            else:
                # verify limits
                self.verifyLimits(value)
                self.showFormatted(value)

        return value
    
    def showFormatted(self, s):
        if isinstance(s, wx.DateTime):
            s1 = s.FormatISODate()
        elif s.strip() == '':
            # don't touch null value
            s1 = ''
        else:
            d = wx.DateTime()
            d.ParseDate(s)
            s1 = d.FormatISODate()
        self.ChangeValue(s1)


class CInt(wx.lib.intctrl.IntCtrl):
    def __init__(self,
                 prnt,
                 id= -1,
                 pos=wx.DefaultPosition,
                 allowLong=True,
                 allowNone=True,
                 size=(100, 32),
                 min=None,
                 max=None,
                 foregroundcolour=(0, 0, 100),
                 backgroundcolour=TEXTBKGCOLOR,
                 invalidBackgroundColour=(255, 250, 250),
                 style=0,
                 value=0):

        u = Generics()
        foregroundcolour = u.makeColour(foregroundcolour)
        backgroundcolour = u.makeColour(backgroundcolour)
        invalidBackgroundColour = u.makeColour(invalidBackgroundColour)
        
        wx.lib.intctrl.IntCtrl.__init__(self, id=id,
                                        parent=prnt, oob_color=invalidBackgroundColour,
                                        pos=pos,
                                        allow_long=allowLong, allow_none=allowNone,
                                        size=size, style=style, value=value)
        self.SetMaxSize(size)
        # allowable range of number
        self.SetMax(max)
        self.SetMin(min)
        # background color
        self.SetBackgroundColour(backgroundcolour)
        # foreground (letters) color
        self.SetForegroundColour(foregroundcolour)
        #self.SetValue(initialvalue)



class FloatEntry(wx.Panel):
    def __init__(self, parent=None,
                 decimals=2, # digits in the fraction
                 nosep=False, # ignore thousands separator
                 minval= -1e38, # min value
                 maxval= +1e38, # max value
                 value=0., # initial value
                 unitdict=None, # unit dict for the unit selector
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None, # fontsize for subwidgets
                 isStatic=False, # can user modify entry
                 pos=wx.DefaultPosition,
                 defaultDisplayUnit=None): # explicitly place the widget
                                        
        self.prnt = parent
        self.unitdict = unitdict
        self.lastInternalValue = None
        self.defaultDisplayUnit = defaultDisplayUnit
        self.displayUnit = defaultDisplayUnit
        self.funit2 = None
        self.decimals = decimals
        self.isStatic = isStatic
        self._userinput = False # Flag whether user input has happened
        self.userChange = False # idem, for external use (detection of changes in GUI-panels)
        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()

        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, True, wUnits)
        
        wx.Panel.__init__(self, parent, id= -1, size=size, style=style, pos=pos)

        if label.strip():
            # creates label
            sty = wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT | wx.BORDER_NONE
            self.label = wx.StaticText(self, id= -1, label=label, size=lblSize, pos=(0, 0), style=sty)
            self.f.setFont(self.label, size=fontsize)
            self.label.Wrap(lblSize.GetWidth())
            self.label.Center(wx.VERTICAL)

        # create a masked float-point control
        entryStyle = wx.ALIGN_RIGHT | wx.ST_NO_AUTORESIZE
        entryEmptyBackgroundColour = EMPTYBKGCOLOR
        entryValidBackgroundColour = EMPTYBKGCOLOR
        entryInvalidBackgroundColour = INVALIDBKGCOLOR
        if self.isStatic is True:
            entryStyle = entryStyle | wx.TE_READONLY
            entryEmptyBackgroundColour = LIGHTGREY
            entryValidBackgroundColour = LIGHTGREY
            entryInvalidBackgroundColour = LIGHTGREY

        self.entry = CFloat(self, pos=(lblSize[0] + 1, 0),
                           size=datSize, # size of control
                           min=minval, # minimum value admitted
                           max=maxval, # maximum value admitted
                           decimals=decimals, # number of decimals
                           nosep=nosep,
                           style=entryStyle,
                           emptyBackgroundColour=entryEmptyBackgroundColour,
                           validBackgroundColour=entryValidBackgroundColour,
                           invalidBackgroundColour=entryInvalidBackgroundColour)

        self.f.setFont(self.entry, size=fontsize)
        self.entry.Center(wx.VERTICAL)
        self.entry.Bind(wx.EVT_CONTEXT_MENU, self.OnShowPopup)
        self.entry.Bind(wx.EVT_TEXT, self.OnEntry)

        # load and show a unit selector
        # also sets the default display unit
        if self.unitdict is not None:
            self.units = wx.Choice(self, -1, pos=(lblSize[0] + datSize[0] + 2, 0),
                                     choices=[], size=uniSize)
            self.setUnits(_U('Select a measurement unit'), unitdict, 
                          self.defaultDisplayUnit)
            backgroundcolour = self.g.makeColour(CHOOSERBCGCOLOR)
            self.units.SetBackgroundColour(backgroundcolour)
            self.f.setFont(self.units, size=fontsize)
            self.units.Center(wx.VERTICAL)
            self.units.Bind(wx.EVT_CHOICE, self.OnUnits)
            

        # set initial value (must be after loading the selector)
        self.SetValue(value, emitEvent=False)

        # set tooltips
        self.g.setTooltips(self, tip)

        # create popup menu for easy clearing/setting to zero
        self.popupmenu = wx.Menu()
        for text in "Clear Zero".split():
            item = self.popupmenu.Append(-1, text)
            self.Bind(wx.EVT_MENU, self.OnPopupItemSelected, item)

    def OnShowPopup(self, event):
        pos = event.GetPosition()
        pos = self.ScreenToClient(pos)
        self.PopupMenu(self.popupmenu, pos)

    def OnEntry(self, event):
        """Set the flag, that user input has happened."""
        self._userinput = True
        self.userChange = True
        event.Skip()

    def OnPopupItemSelected(self, event):
        item = self.popupmenu.FindItemById(event.GetId())
        text = item.GetText()
        if text == 'Clear':
            self.entry.SetValue(None)
        elif text == 'Zero':
            self.entry.SetValue(0.0)
            
    def OnUnits(self, event):
        """ activated when user changes display unit """
        if self._userinput is True: # read in user input if any happened
            self.lastInternalValue = self.GetValue()
            self._userinput = False
        # read new measurement unit and convert to internal encoding
        self.displayUnit = event.GetString().encode(ENCODING)
        # recalculate to display units
        self.SetValue(self.lastInternalValue, emitEvent=False)
        # allow descendant handlers
        event.Skip()

    def setUnits(self, tip, unitdict, defaultDisplayUnit=None):
        # only allows change in units when units were initially declared
        # and the units choice control was created
        if self.unitdict:
            self.defaultDisplayUnit = self.g.setUnits(self, tip, unitdict, 
                                                      defaultDisplayUnit)
            if self.displayUnit is None:
                self.displayUnit = self.defaultDisplayUnit

    def updateUnitDict(self, f2):
        """ auxiliary function for use of dynamic unitdicts where conversion
            factors may vary, e.g. with selected fluid or currency ...
            unitdict1 is the main unit-dictionary as stored in database
            unitdict2 is the secondary unit system
            e.g. unitdict1 = mass, unitdict2 = volume, f2 = density """

        if f2 is not None and f2 > 0:
            self.funit2 = f2
        else:
            logError("DisplayClasses.FloatEntry.updateUnitDict %s: invalid conversion factor %s"%(self.unitdict,self.funit2))
        
    def GetValue(self):
        """ returns the INTERNAL value corresponding to the
            currently set DISPLAY VALUE and selected displayUnit """
        
        dValue = self.entry.GetValue().encode(ENCODING)
        if dValue is None:
            return None
        elif isinstance(dValue, str):
            if dValue.strip() == '':
                return None
            f = self.entry.toFloat(dValue)
        elif isinstance(dValue, float) or isinstance(dValue, int):
            f = dValue
            
        if self.displayUnit is None:
            return f
        try:
            iValue = Globals.unitConversion.internalValue(f,
                                         self.displayUnit,
                                         self.unitdict, self.funit2)
            return self.entry.toFloat(iValue)
        except:
            print 'FloatEntry: error in conversion display->internal ' \
                  'display=%s class=%s default=%s' % (dValue,
                                                      self.unitdict,
                                                      self.displayUnit)
            
    def Clear(self):
        self.SetValue(None)
    
    def Enable(self, enable=True, emitEvent=True):
        super(self.__class__, self).Enable(enable)
        if emitEvent is True:
            evt = EnableChanged(enableStatus=enable)
            wx.PostEvent(self.GetEventHandler(), evt)
        
    def Disable(self):
        self.Enable(False)
    
    def SyncValueTo(self, event):
        """Adapt the shown value to the one of the sender of the event.
        
        :param event: A wx.Event from a CFloat entry containing the 
            new value of the entry.
        
        """
        self.SetValue(event.EventObject.Value, emitEvent=False, direct=True)
        self._userinput = True # some input beyond this widgets at least
        event.Skip()

    def SyncUnitTo(self, event):
        """Adapt the selected unit to the one of the sender of the event.
        
        :param event: A wx.Event from a wx.Choice widget containing the 
            new value of the entry.
        
        """
        self.OnUnits(event)
        self.setUnit(event.EventObject.GetSelection())
        event.Skip()
    
    def SyncEnableTo(self, event):
        self.Enable(event.enableStatus, emitEvent=False)
        
    def SetValue(self, iValue, emitEvent=True, direct=False):
        """Set the value shown to iValue. 
        
        If emitEvent is True a EVT_TEXT event is emitted upon 
        setting the new value, otherwise no event is emitted.
        If direct is True the IValue is not converted to an 
        other unit, but taken as is.
        
        """
        # save internal value for an eventual unit change by user
        self.lastInternalValue = iValue
        
        # reset view precision if necessary
        if self.displayUnit is not None:
            decimals = (self.decimals + 
                        log10(Globals.unitConversion.conversionFactor(self.displayUnit, 
                                                     self.unitdict, self.funit2)))
            decimals = int(round(decimals))
            self.entry.decimals = decimals
        
        # find display value
        if iValue is None or iValue == '' or iValue == 'None':
            self.entry.SetValue(None, emitEvent)
        elif (self.displayUnit is None) or direct:
            # no unit conversion necessary
            f = self.entry.toFloat(iValue)
            f = self.entry.SetValue(f, emitEvent)
            if f is None:
                print 'FloatEntry: bad value for SetValue %s' % repr(iValue)
                self.entry.SetValue(None, emitEvent)
                self.lastInternalValue = None
        else:
            # convert to user units
            try:
                f = self.entry.toFloat(iValue)
                if f is not None:
                    dValue = Globals.unitConversion.displayValue(f, self.displayUnit, self.unitdict, self.funit2)
                    self.entry.SetValue(dValue, emitEvent) 
                else:
                    logDebug('FloatEntry: bad value for SetValue %s' % repr(iValue))
                    self.entry.SetValue(None, emitEvent)
                    self.lastInternalValue = None
            except:
                print 'FloatEntry: error in conversion internal->display ' \
                      'internal=%s class=%s default=%s' % (iValue, self.unitdict, self.displayUnit)

        self.userChange = False

    def GetUnit(self, text=False):
        result = None
        if self.unitdict:
            if text:
                result = self.g.getChoiceValues(self.units, False, text).encode('utf-8')
            else:
                result = self.g.getChoiceValues(self.units, False, text)
        return result
        
    def SetEditable(self, editable):
        """Set whether the widget should allow editing of its entry"""
        self.entry.SetEditable(editable)
        if editable:
            self.units.Enable()
        else: 
            self.units.Disable()

    def setUnit(self, n):
        if self.unitdict:
            # n is the 0-based index to the contents
            self.units.SetSelection(n)

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        self.entry.SetBackgroundColour(self.g.makeColour(bgColor))
        self.entry.SetForegroundColour(self.g.makeColour(fgColor))


class IntEntry(wx.Panel):
    def __init__(self, parent=None,
                 minval=None, # min value
                 maxval=None, # max value
                 value=0, # initial value
                 unitdict=None, # unit dict for the unit selector
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None):

        self.prnt = parent
        
        self.unitdict = unitdict
        self.lastInternalValue = None
        self.defaultDisplayUnit = None
        self.displayUnit = None
        self.funit2 = None

        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()
        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, True, wUnits)

        wx.Panel.__init__(self, parent, id= -1, size=size, style=style)

        if label.strip():
            # creates label
            self.label = wx.lib.stattext.GenStaticText(self, ID= -1, label=label, size=lblSize, pos=(0, 0),
                                                       style=wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT)
            self.f.setFont(self.label, size=fontsize)

        # create a masked fixed-point control
        self.entry = CInt(self, pos=(lblSize[0] + 1, 0),
                           size=datSize, # size of control
                           min=minval, # minimum value admitted
                           max=maxval, # maximum value admitted
                           style=wx.ALIGN_RIGHT)
        self.f.setFont(self.entry, size=fontsize)
        self.entry.Bind(wx.EVT_TEXT, self.OnEntry)
        self.entry.Bind(wx.EVT_CONTEXT_MENU, self.OnShowPopup)

        # load and show a unit selector
        # also sets the default display unit
        if self.unitdict is not None:
            self.units = wx.Choice(self, -1, pos=(lblSize[0] + datSize[0] + 2, 0),
                                     choices=[], size=uniSize)
            self.setUnits(_U('Select a measurement unit'), unitdict)
            backgroundcolour = self.g.makeColour(CHOOSERBCGCOLOR)
            self.units.SetBackgroundColour(backgroundcolour)
            self.f.setFont(self.units, size=fontsize)
            self.units.Bind(wx.EVT_CHOICE, self.OnUnits)

        # set initial value (must be after loading the selector
        self.SetValue(value)

        # set tooltips
        self.g.setTooltips(self, tip)

        # create popup menu for easy clearing/setting to zero
        self.popupmenu = wx.Menu()
        for text in "Clear Zero".split():
            item = self.popupmenu.Append(-1, text)
            self.Bind(wx.EVT_MENU, self.OnPopupItemSelected, item)

    def OnShowPopup(self, event):
        pos = event.GetPosition()
        pos = self.ScreenToClient(pos)
        self.PopupMenu(self.popupmenu, pos)

    def OnPopupItemSelected(self, event):
        item = self.popupmenu.FindItemById(event.GetId())
        text = item.GetText()
        if text == 'Clear':
            self.entry.SetValue(None)
        elif text == 'Zero':
            self.entry.SetValue(0)

    def OnEntry(self, event):
        self.lastInternalValue = self.GetValue()

    def OnUnits(self, event):
        # read new measurement unit and convert to internal encoding
        self.displayUnit = event.GetString().encode(ENCODING)
        # recalculate to display units
        self.SetValue(self.lastInternalValue)

    def setUnits(self, tip, unitdict):
        if self.unitdict:
            self.defaultDisplayUnit = self.g.setUnits(self, tip, unitdict)
            if self.displayUnit is None:
                self.displayUnit = self.defaultDisplayUnit

    def GetValue(self):
        dValue = self.entry.GetValue()
        if dValue is None:
            return None
        elif isinstance(dValue, str):
            if dValue.strip() == '':
                return None
            f = self.entry.toFloat(dValue)
        elif isinstance(dValue, float) or isinstance(dValue, int):
            f = dValue

        if self.displayUnit is None:
            return int(f)
        try:
            iValue = Globals.unitConversion.internalValue(f, self.displayUnit, self.unitdict, self.funit2)
            return int(iValue)
        except:
            print 'IntEntry: error in conversion disp->int '\
                  'disp=%s class=%s default=%s' % (dValue,
                                                   self.unitdict,
                                                   repr(self.displayUnit))

    def Clear(self):
        self.SetValue(None)
    
    def SetEditable(self, editable):
        """Set whether the widget should allow editing of its entry"""
        self.entry.SetEditable(editable)
        if editable:
            self.units.Enable()
        else: 
            self.units.Disable()
  
    def SetValue(self, iValue, emitEvent=True):
        #
        # saves internal value for an eventual unit change by user
        #
        self.lastInternalValue = iValue
        #
        # find display value
        #
        if iValue is None or iValue == '' or iValue == 'None':
            self.entry.SetValue(None)
        else:
            try:
                f = int(float(iValue))
            except:
                print 'IntEntry: bad value for SetValue %s' % (repr(iValue),)
                self.entry.SetValue(0)
                return
            
            if self.displayUnit is None:
                dValue = f
            else:
                try:
                    dValue = Globals.unitConversion.displayValue(float(f), self.displayUnit, self.unitdict, self.funit2)
                except:
                    print 'IntEntry: error in conversion int->disp ' \
                          'int=%s class=%s default=%s' % (f, self.unitdict,
                                                          repr(self.displayUnit))

                self.entry.SetValue(int(dValue))



    def GetUnit(self, text=False):
        if self.unitdict:
            return self.g.getChoiceValues(self.units, False, text).encode('utf-8')
        else:
            return None

    def setUnit(self, n):
        if self.unitdict:
            # n is the 0-based index to the contents
            self.units.SetSelection(n)

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        self.entry.SetBackgroundColour(self.g.makeColour(bgColor))
        self.entry.SetForegroundColour(self.g.makeColour(fgColor))
 


class TextEntry(wx.Panel):
    def __init__(self, parent=None,
                 maxchars=None, # max characters accepted
                 value='', # initial value
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector (not used)
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None,
                 isStatic=False,
                 isMultiline=False,
                 eventHandlerLeftDown=None,
                 specialHeight = None):

        self.prnt = parent

        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()

        self.eventHandlerLeftDown=eventHandlerLeftDown

        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, False, wUnits)
        if specialHeight is not None:
            size = wx.Size(size[0], specialHeight)
            datSize = wx.Size(datSize[0], specialHeight)

        wx.Panel.__init__(self, parent, id= -1, size=size, style=style)

        if label.strip():
            # creates label
            self.label = wx.lib.stattext.GenStaticText(self, ID= -1, label=U(label), size=lblSize, pos=(0, 0),
                                                       style=wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT)
            self.f.setFont(self.label, size=fontsize)

        # create a text control
        entryStyle = wx.ALIGN_LEFT
        if isStatic:
            entryStyle = entryStyle | wx.TE_READONLY
        if isMultiline:
            entryStyle = entryStyle | wx.TE_MULTILINE
        self.entry = wx.TextCtrl(self, -1, pos=(lblSize[0] + 1, 0),
                                 value=value,size=datSize,style=entryStyle)
        self.f.setFont(self.entry, size=fontsize)
        self.entry.Bind(wx.EVT_CONTEXT_MENU, self.OnShowPopup)
        self.entry.Bind(wx.EVT_TEXT, self.OnEntry)

        foregroundColour = self.g.makeColour(FGCOLOR)
        validBackgroundColour = self.g.makeColour(TEXTBKGCOLOR)
        self.entry.SetForegroundColour(foregroundColour)

        entryValidBackgroundColour = validBackgroundColour
        if isStatic:
            entryValidBackgroundColour = LIGHTGREY
        self.entry.SetBackgroundColour(entryValidBackgroundColour)
        if maxchars is not None:
            self.entry.SetMaxLength(maxchars)

        # set tooltips
        self.g.setTooltips(self, tip)

        # create popup menu for easy clearing
        self.popupmenu = wx.Menu()
        for text in "Clear Unknown".split():
            item = self.popupmenu.Append(-1, text)
            self.Bind(wx.EVT_MENU, self.OnPopupItemSelected, item)

        if self.eventHandlerLeftDown is not None:
            self.entry.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)

        self.userChage = False

    def OnLeftDown(self, event):
        if self.eventHandlerLeftDown is not None:
            self.entry.SetValue(self.eventHandlerLeftDown(self.entry, event.GetPosition()))
        
    def OnShowPopup(self, event):
        pos = event.GetPosition()
        pos = self.ScreenToClient(pos)
        self.PopupMenu(self.popupmenu, pos)

    def OnPopupItemSelected(self, event):
        item = self.popupmenu.FindItemById(event.GetId())
        text = item.GetText()
        if text == 'Clear':
            self.entry.SetValue('')
        if text == 'Unknown':
            self.entry.SetValue('None')

    def SetEditable(self, editable):
        self.entry.SetEditable(editable)
    
    def setUnits(self, tip, unitdict):
        # this method is just for compatibility
        pass

    def GetValue(self):
        return toUTF8(self.entry.GetValue())

    def Clear(self):
        self.SetValue('')

    def SetValue(self, value, emitEvent=True):
        if value is None:
            self.entry.SetValue('')
        else:
            self.entry.SetValue(U(value))
        self.userChange = False
        
    def getUnit(self):
        # this method is just for compatibility
        return None

    def setUnit(self, value):
        # this method is just for compatibility
        pass

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        self.entry.SetBackgroundColour(self.g.makeColour(bgColor))
        self.entry.SetForegroundColour(self.g.makeColour(fgColor))

    def OnEntry(self, event):
        self.userChange = True


class DateEntry(wx.Panel):
    def __init__(self, parent=None,
                 value=wx.DateTime_Now(), # initial value
                 min=LOWERACCEPTEDDATE, # lower accepted date
                 max=UPPERACCEPTEDDATE, # upper    "      "
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector (not used)
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None):

        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()

        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, True, wUnits)

        wx.Panel.__init__(self, parent, id= -1, size=size, style=style)

        if label.strip():
            # creates label
            self.label = wx.lib.stattext.GenStaticText(self, ID= -1, label=label, size=lblSize, pos=(0, 0),
                                                       style=wx.ST_NO_AUTORESIZE | wx.ALIGN_LEFT)
            self.f.setFont(self.label, size=fontsize)


        self.entry = CDate(self, parent, id= -1, pos=(lblSize[0] + 1, 0), size=datSize)

        self.SetValue(value)

        self.f.setFont(self.entry, size=fontsize)

        # set tooltips
        self.g.setTooltips(self, tip)

    def setUnits(self, tip, unitdict):
        # this method is just for compatibility
        pass

    def GetValue(self):
        value = self.entry.GetValue()
        if not value:
            return None
        date = wx.DateTime()
        date.ParseDate(value)
        return date.FormatISODate()

    def Clear(self):
        self.SetValue(None)

    def SetValue(self, value, emitEvent=True):
        self.entry.SetValue(str(value))
    
    def getUnit(self):
        # this method is just for compatibility
        return None

    def setUnit(self, value):
        # this method is just for compatibility
        pass

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        self.entry.SetBackgroundColour(self.g.makeColour(bgColor))
        self.entry.SetForegroundColour(self.g.makeColour(fgColor))
 


class ChoiceEntry(wx.Panel):
    def __init__(self, parent=None,
                 values=[], # initial values list
                 value=None, # initial value
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector (not used)
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None,
                 pos=wx.DefaultPosition, # explicitly place the widget
                 allowNones=True,  # If True, automatically an empty line is added at index 0
                 registerUserChange=True):  # Detection of user interaction can be deactivated if setting to False
        
        self.prnt = parent

        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()
        self.allowNones = allowNones
        self.registerUserChange = registerUserChange

        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, 0, 0)

        wx.Panel.__init__(self, parent, id= -1, size=size, style=style, pos=pos)

        if label.strip() != "":
            # creates label
            self.label = wx.StaticText(self, id= -1, pos=(0, 0),
                                       label=label, size=lblSize,
                                       style=wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT)
            self.f.setFont(self.label, size=fontsize)

        #create a choice control
##        u_values = [U(value) for value in values]
##        if self.allowNones: u_values = [""] + u_values
##        self.entry = wx.Choice(self, -1, pos=(lblSize[0] + 1, 0), choices=u_values, size=datSize)
        self.entry = wx.Choice(self, -1, pos=(lblSize[0] + 1, 0), size=datSize)
        self.entry.Bind(wx.EVT_CHOICE, self.OnEntry)
        backgroundcolour = self.g.makeColour(CHOOSERBCGCOLOR)
        self.entry.SetBackgroundColour(backgroundcolour)
        self.f.setFont(self.entry, size=fontsize)

        # set tooltips
        self.g.setTooltips(self, U(tip))
        
        # set list of choices and initial value
        self.SetValue(values, emitEvent=False)        
        self.SetValue(value, emitEvent=False)

        self.userChange = False
        
    def setUnits(self, tip, unitdict):
        # this method is just for compatibility
        pass

    def GetValue(self, text=False):

        ret =  self.g.getChoiceValues(self.entry, False, text)
        if self.allowNones and not text and ret is not None:
            ret -= 1
        if ret < 0: ret = None
        return ret

    def Clear(self):
        self.entry.Clear()
    
    def Enable(self, enable=True, emitEvent=True):
        if emitEvent is True:
            evt = EnableChanged(enableStatus=enable)
            wx.PostEvent(self.GetEventHandler(), evt)
        super(self.__class__, self).Enable(enable)
        
    def Disable(self):
        self.Enable(False)
        
    def SetValue(self, thing=0, emitEvent=True):
        """this method has triple functionality:
        if 'thing' is an integer n, the choice will show the nth element. 
        if 'thing' is a string, the choice will show the element that contains the string
        if 'thing' is a list, the elements of the list will be loaded in the choice.
        
        By default it emits a wx.EVT_CHOICE event upon being called.
        
        """
        try:
            if isinstance(thing, int):
                if self.allowNones and thing is not None:
                    thing += 1
                try:
                    self.entry.SetSelection(thing)
                except:
                    self.entry.SetSelection(0)
            elif isinstance(thing, str) or isinstance(thing, unicode):
                thing = U(thing)    
                try:
                    self.entry.SetSelection(self.entry.FindString(thing))
                except:
                    try:
                        self.entry.SetSelection(max(self.entry.FindString("None"), 0))
                    except:
                        self.entry.SetSelection(0)
            elif thing is None:
                # unselecting everything and showing an empty field does not seem to work. workaround: add a white row to the list
                self.entry.SetSelection(self.entry.FindString(""))
            elif isinstance(thing, list):
                # thing is a list of values for the choice control
                uthing = [U(elem) for elem in thing]
                if self.allowNones:
                    uthing = [""] + uthing # add an empty line
                self.SetAlternatives(uthing, emitEvent)
                emitEvent = False # taken care of by SetAlternative
        except:
            # possibly a numeric constant
            try:
                nn = int(thing)
                if self.allowNones: nn += 1
                try:
                    self.entry.SetSelection(nn)
                except:
                    self.entry.SetSelection(0)
            except:
                self.entry.SetSelection(0)
        
        if emitEvent is True:
            evt = wx.CommandEvent(wx.EVT_CHOICE.evtType[0])
            evt.SetEventObject(self.entry)
            evt.SetId((self.entry.GetId()))
            try:
                thing = str(thing)
            except UnicodeEncodeError:
                pass # is already a string with some special characters
            evt.SetString(thing)
            wx.PostEvent(self.GetEventHandler(), evt)

        self.userChange = False
        
    def SetChoice(self, item):
        """Allow explicit choice without using the overloading mechanism
        of SetValue.
        
        """
        assert(isinstance(item, int))
        self.SetValue(item)

    def SyncValueTo(self, event):
        if isinstance(event.EventObject, CheckBoxEntry):
            if event.EventObject.Get3StateValue() == wx.CHK_UNDETERMINED:
                self.SetValue(" ", emitEvent=False)
            elif event.EventObject.Get3StateValue() == wx.CHK_CHECKED:
                self.SetValue(TRANSYESNO["yes"], emitEvent=False)
            elif event.EventObject.Get3StateValue() == wx.CHK_UNCHECKED:
                self.SetValue(TRANSYESNO["no"], emitEvent=False)
        else:
            val = event.EventObject.GetSelection()
            if self.allowNones:
                val = None if (val==0 or val is None) else val-1     
            self.SetValue(val, emitEvent=False)
        event.Skip()
        
    def SyncAlternativesTo(self, event):
        self.SetAlternatives(event.items, emitEvent=False)
        event.Skip()
        
    def SyncEnableTo(self, event):
        self.Enable(event.enableStatus, emitEvent=False)
    
    def SetAlternatives(self, alternatives, emitEvent=True):
        """Set the alternatives to the given list. Choose the first."""
        self.entry.Clear()
        for item in alternatives:
            if item is None:
                self.entry.Append(' ')
            else:
                self.entry.Append(item)

        if emitEvent is True:
            evt = ChoiceItemsChanged(items=alternatives)
            wx.PostEvent(self.GetEventHandler(), evt)
            
        self.entry.SetSelection(0)
    
    def getUnit(self):
        # this method is just for compatibility
        return None

    def setUnit(self, value):
        # this method is just for compatibility
        pass

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        self.entry.SetBackgroundColour(self.g.makeColour(bgColor))
        self.entry.SetForegroundColour(self.g.makeColour(fgColor))
 
    def OnEntry(self, event):
        if self.registerUserChange: self.userChange = True
        event.Skip()

class MultipleChoiceEntry(wx.Panel):
    def __init__(self, parent=None,
                 values=None, # initial values list
                 selected=None, # initial selected values (list of strings)
                 wLabel=None, # width of the label
                 wData=None, # width of the data entry
                 wUnits=None, # width of the unit selector (not used)
                 label='', # text of the label
                 tip=' ', # text of the tip
                 fontsize=None):

        self.prnt = parent

        style = wx.NO_BORDER | wx.TAB_TRAVERSAL
        self.g = Generics()
        self.f = FontProperties()

        (size, lblSize, datSize, uniSize) = self.g.setSizes(wLabel, wData, False, wUnits)

        wx.Panel.__init__(self, parent, id= -1, size=size, style=style)

        if label.strip():
            # creates label
            self.label = wx.lib.stattext.GenStaticText(self, ID= -1, pos=(0, 0),
                                                       label=label, size=lblSize,
                                                       style=wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT)
            self.f.setFont(self.label, size=fontsize)

        #create a ListCombo control
        # 1. Create a ComboCtrl
        self.cc = wx.combo.ComboCtrl(self, pos=(lblSize[0] + 1, 0), size=datSize)        
        # 2. Create a Popup
        self.entry = ListCombo()
        # 3. Associate them with each other.  This also triggers the
        #    creation of the ListCtrl.
        self.cc.SetPopupControl(self.entry)

        backgroundcolour = self.g.makeColour(CHOOSERBCGCOLOR)
        self.cc.SetBackgroundColour(backgroundcolour)
        self.entry.SetBackgroundColour(backgroundcolour)
        self.f.setFont(self.cc, size=fontsize)
        self.f.setFont(self.entry, size=fontsize)

        # set tooltips
        self.g.setTooltips(self, U(tip))

        # set values
        if values is not None:
            self.SetValue(values)

        # set initially selected
        if selected is not None:
            u_selected = [U(item) for item in selected]
            self.SetSelection(u_selected)
        
    def setUnits(self, tip, unitdict):
        # this method is just for compatibility
        pass

    def GetValue(self, text=False):
        #print self.entry.GetSelectedItemCount()

        ret = self.g.getChoiceValues(self.entry, True, text)
            
        return ret 

    def Clear(self):
        self.entry.Clear()

    def SetSelection(self, selection):
        combo = self.entry.GetCombo()
        self.entry.UnselectAll()
        if isinstance(selection, list):
            for val in selection:
                self.entry.SetStringValue(U(val))
            combo.SetText(U(';'.join(selection)))
        else:
            self.entry.SetStringValue(U(selection))
            combo.SetText(U(selection))
         
    def SetValue(self, thing=0, emitEvent=True):
        try:
            if thing is None:
                self.entry.UnselectAll()
##                self.entry.Append(' ')
            elif isinstance(thing, str) or isinstance(thing, unicode):
                self.entry.Append(U(thing))
            elif isinstance(thing, list):
                # thing is a list of values for the control
                self.entry.Clear()
                for item in thing:
                    if item is None:
                        self.entry.AddItem(' ')
                    else:
                        self.entry.AddItem(U(item))
                if len(thing) > 0:
                    self.SetSelection(thing[0])
        except:
            logTrack('DisplayClasses (MultipleChoiceEntry): Bad SetValue %s' % thing)

    
    def getUnit(self):
        # this method is just for compatibility
        return None

    def setUnit(self, value):
        # this method is just for compatibility
        pass

    def setColor(self, bgColor=(255, 255, 255), fgColor=(0, 0, 0)):
        #combo = self.entry.GetCombo()
        self.cc.SetBackgroundColour(self.g.makeColour(bgColor))
        self.cc.SetForegroundColour(self.g.makeColour(fgColor))

class CheckBoxEntry(wx.CheckBox):
    """A CheckBox that allows cloning of dependent CheckBoxes and
    can also be used as a clone of a ChoiceEntry with values from
    TRANSYESNO.values().
    
    """
    def __init__(self, *args, **kwargs):
        super(CheckBoxEntry, self).__init__(*args, **kwargs)

        self.f = FontProperties()
        self.f.setFont(self)
    
    def SyncValueTo(self, event):
        """Event handler to sync the state of two CheckBox instances
        or of this CheckBox and a ChoiceEntry with values from 
        TRANSYESNO.values()
        
        """
        if isinstance(event.EventObject, CheckBoxEntry):
            if self.GetValue() != event.EventObject.GetValue():
                self.SetValue(event.EventObject.GetValue())
        else: # should be a wx.Choice element with a yes/no list otherwise raise
            value = toUTF8(event.EventObject.GetStringSelection())
            if value not in TRANSYESNO.values():
                value = wx.CHK_UNDETERMINED
            else:
                value = wx.CHK_CHECKED if value == TRANSYESNO["yes"] else wx.CHK_UNCHECKED
            if self.Get3StateValue() != value:
                try:
                    self.Set3StateValue(value)
                except:
                    logTrack("DisplayClasses: ERROR in CheckBoxEntry.SyncValueTo. Cannot set to 3-state value %r"%value)
                evt = wx.CommandEvent(wx.EVT_CHECKBOX.evtType[0])
                evt.SetEventObject(self)
                evt.SetId((self.GetId()))
                wx.PostEvent(self.GetEventHandler(), evt)
        event.Skip()
    
    def SetValue(self, state, emitEvent=True):
        """New SetValue function that by default emits events if the
        state changes.         
        
        :param state: If True set state to checked.
        :param emitEvent: If False don't emit an event about this operation.
        
        """
        super(self.__class__, self).SetValue(state)
        if emitEvent is True:
            evt = wx.CommandEvent(wx.EVT_CHECKBOX.evtType[0])
            evt.SetEventObject(self)
            evt.SetId((self.GetId()))
            wx.PostEvent(self.GetEventHandler(), evt)
        
    def SyncEnableTo(self, event):
        self.Enable(event.enableStatus, emitEvent=False)
        
    def Enable(self, enable=True, emitEvent=True):
        if emitEvent is True:
            evt = EnableChanged(enableStatus=enable)
            wx.PostEvent(self.GetEventHandler(), evt)
        super(self.__class__, self).Enable(enable)
        
    def Disable(self):
        self.Enable(False)

class TextCtrlEntry(wx.TextCtrl):
    """A wx.TextCtrl that allows mirroring of a master wx.TextCtrl"""
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
    
    def SyncValueTo(self, event):
        """Event handler to sync the state of TextCtrlEntry instances.
        """
        self.SetValue(event.EventObject.GetValue())
        event.Skip()

class Label(wx.lib.stattext.GenStaticText):
    # ***** deprecated *****
    # this class is set to disappear soon!
    #
    # auxiliary class for labels (static text)
    # will show a short descriptive string and
    # generate a longer tooltip.
    # the tooltip is also associated to the text control(s)
    # 'txtlist' can be a scalar (just one control) or a list
    # of controls. 'tiplist' is expected to be the same type
    # and length as 'txtlist'
    # a default length is also managed.
    w0 = None
    w1 = None
    def __init__(self, parent, txtlist, text, tiplist, width0=None, width1=None, style=0):
        wx.lib.stattext.GenStaticText.__init__(self, ID= -1, parent=parent, label='',
                                              style=wx.ST_NO_AUTORESIZE | wx.ALIGN_RIGHT)
        self.SetLabel(text)
        # sets sizes
        h = self.GetMinHeight()
        if width0 is None:
            if Label.w0 is not None:
                self.SetMinSize((Label.w0, h))
        else:
            Label.w0 = width0
            self.SetMinSize((Label.w0, h))
        if width1 is None:
            if Label.w1 is not None:
                if 'list' in str(type(txtlist)):
                    # list of controls
                    for tc  in txtlist:
                        tc.SetMinSize((Label.w1, h))
                else:
                    # just one control
                    txtlist.SetMinSize((Label.w1, h))
        else:
            Label.w1 = width1
            if 'list' in str(type(txtlist)):
                for tc  in txtlist:
                    tc.SetMinSize((width1, h))
            else:
                txtlist.SetMinSize((width1, h))
        # sets tooltips
        if 'list' in str(type(txtlist)):
            self.SetToolTipString(tiplist[0])
            for i in range(len(txtlist)):
                if len(tiplist[i].strip()) > 0:
                    txtlist[i].SetToolTipString(tiplist[i])
        else:
            if len(tiplist.strip()) > 0:
                self.SetToolTipString(tiplist)
                txtlist.SetToolTipString(tiplist)



class WallpaperedPanel(wx.Panel):
    def __init__(self, parent, id=wx.ID_ANY, wallpaper=None ):
        """A simple extension of the wx.Panel, that has the picture from 
        the wx.Image wallpaper as a background.
        
        :param wallpaper: the background image
        :type wallpaper: wx.Image
        
        """
        self.wallpaper = wx.BitmapFromImage(wallpaper)
        super(self.__class__, self).__init__(parent, id)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        
    def OnEraseBackground(self, evt):
        """Repaint the wallpaper."""
        dc = evt.GetDC()
 
        if dc is None:
            dc = wx.ClientDC(self)
            rect = self.GetUpdateRegion().GetBox()
            dc.SetClippingRect(rect)

        dc.Clear()
        dc.DrawBitmap(self.wallpaper, 0, 0)
 
    def SetWallpaper(self, wallpaper):
        """Set an other wallpaper.
        
        :type wallpaper: wx.Image
        
        """
        self.wallpaper = wx.BitmapFromImage(wallpaper)
        self.Refresh()
