# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       www.einstein-energy.net
#
#------------------------------------------------------------------------------
#
#	EinsteinPanel: Generic GUI Panel
#
#==============================================================================
#
#       EINSTEIN Version No.: 3.0
#
#------------------------------------------------------------------------------
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany
#	http://www.energyxperts.net/
#
#==============================================================================
import wx
from status import Status
from globals import logger
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, toUTF8
from GUI.fonts import FontProperties
from GUI.displayClasses import FieldSizes
from GUI.GUITools import check, TITLE_COLOR
import database.dbio as dbio

# constants that control the default sizes
# 1. font sizes
TYPE_SIZE = 9
TYPE_SIZE_TITLES = 10

# 2. field sizes
HEIGHT = 24
LABEL_WIDTH = 560
DATA_ENTRY_WIDTH = 120
UNITS_WIDTH = 100

# 3. vertical separation between fields
VSEP_LEFT = 2
VSEP_RIGHT = 2

flagText = wx.ALIGN_CENTER_VERTICAL | wx.TOP



class EinsteinPanel(wx.Panel):
    def __init__(self, parent, main, layout, module=None, size=(780, 600),
                 key=None,
                 title=None,
                 bottomRow="default"):

        self.bottomRow = bottomRow
        
        self.main = main
        self.prnt = parent
        self.prnt.panel = self
        self.key = key
        
        self.objectName = None  # name of the object being edited
        self.mod = module

        self.size = size

        self.blockEventHandlers = True

        self._init_ctrls(parent, layout)
        self.__do_layout()

        self.blockEventHandlers = False

        self.data = None                # dictionary of data to be displayed

        self.unsavedChanges = False

        self.autoClose = False          # flag is set when self-closing from OK- or cancel- button.

        if title is not None:
            self.setFrameTitle(title)

        self.activePage = None          # indicates the number of the active notebook panel

        
    def _init_ctrls(self, parent, layout):
        """
        adds all controls to the panel, including containers (notebook, panels)
        """
        import platform # this is a hack to the GUI for gtk platforms
        (dummy, arch) = platform.architecture()

        if "gtk2" in wx.PlatformInfo:
            wx.Panel.__init__(self, id= -1, name='PanelQ6', parent=parent,
                              pos=wx.Point(0, 0), size=wx.Size(self.size[0], self.size[1]), style=0)
        else:
            wx.Panel.__init__(self, id= -1, name='PanelQ6', parent=parent,
                              pos=wx.Point(0, 0), size=wx.Size(self.size[0], self.size[1]-20), style=0)
        self.Hide()

        # access to font properties object
        fp = FontProperties()

        self.nbPage = []
        self.notebook = None
        if "notebook" in layout.keys():
            pageTitles = layout["notebook"]
            if pageTitles is not None and len(pageTitles) > 0:
                self.notebook = wx.Notebook(self, -1, style=0)
                self.notebook.SetFont(fp.getFont())
                self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.OnNotebookPageChanged, self.notebook)
                self.activePage = 0

                for title in pageTitles:
                    self.nbPage.append(wx.Panel(self.notebook))
                    self.notebook.AddPage(self.nbPage[-1], U(title))

        self.frame = []
        self.pageOfFrame = []
        
        if "frames" in layout.keys():
            # set font for titles
            # 1. save actual font parameters on the stack
            fp.pushFont()
            # 2. change size and weight
            fp.changeFont(size=TYPE_SIZE_TITLES, weight=wx.BOLD)

            for frame in layout["frames"]:
                page, title = frame

                nbPage = self.nbPage[page] if len(self.nbPage)>0 else self

                self.frame.append(wx.StaticBox(nbPage, -1, U(title)))
                self.frame[-1].SetForegroundColour(TITLE_COLOR)
##                self.frame[-1].SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))
                self.frame[-1].SetFont(fp.getFont())

                self.pageOfFrame.append(page)
            
            # 3. recover previous font state
            fp.popFont()
      
        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH,
                       wData=DATA_ENTRY_WIDTH, wUnits=UNITS_WIDTH)

        # left tab controls
        # tab 0 - Heat recovery from thermal equipment
        fp.pushFont()
        fp.changeFont(size=TYPE_SIZE)

        self.entry = {}

        if self.bottomRow == "default":                                  
            self.buttonOK = wx.Button(self, wx.ID_OK, "", size=(100, 24))
            self.Bind(wx.EVT_BUTTON, self.OnButtonOK, self.buttonOK)
            self.buttonOK.SetDefault()

            self.buttonCancel = wx.Button(self, wx.ID_CANCEL, "", size=(100, 24))
            self.Bind(wx.EVT_BUTTON, self.OnButtonCancel, self.buttonCancel)

            self.activeObject = wx.StaticText(self,id=-1,label="",style = wx.ALIGN_LEFT)
            self.activeObject.SetMinSize((580, 24)) #a hack in order to achieve alignment at the left. can surely be done better.


    def addEntry(self, label, entry, frameNo=None, eventHandlers = None):
        """
        adds any entry to the panel

        label: identifier of the entry
        frameNo: number of frame to which the entry belongs
        """
        self.entry.update({label: entry})
        
        if frameNo == -1 and self.bottomRow=="control":
            self.sizerBottomRow.Add(entry, 0, flagText, 0)
        elif frameNo is not None:
            self.sizerFrame[frameNo].Add(entry, 0, flagText, 0)
        else:
            self.sizerGlobal.Add(entry, 0, flagText, 0)
                                  
    def __do_layout(self):
        """
        First part of layout. Carries out the layout of default entries
        """

        # global sizer for panel. Contains notebook w/two tabs + buttons Cancel and Ok
        self.sizerGlobal = wx.BoxSizer(wx.VERTICAL)

        # sizer for each notebook page
        self.sizerPage = []
        for nbPage in self.nbPage:
            self.sizerPage.append(wx.BoxSizer(wx.VERTICAL))
                                  
        # sizer for each frame
        self.sizerFrame = []
        for frame in self.frame:
            self.sizerFrame.append(wx.StaticBoxSizer(frame, wx.VERTICAL))

        if self.bottomRow in ["default", "control"]:
            self.sizerBottomRow = wx.BoxSizer(wx.HORIZONTAL)
            
        

    def conclude_layout(self):
        """
        Second part of the layout - to be called after a possible personalised layout function
        """

        if self.notebook: self.sizerGlobal.Add(self.notebook, 3, wx.EXPAND, 0)

        for i, sizerFrame in enumerate(self.sizerFrame):
            pageNo = self.pageOfFrame[i]
            if pageNo is None:
                self.sizerGlobal.Add(sizerFrame, 0, wx.ALIGN_TOP | wx.EXPAND, 0)
            else:
                self.sizerPage[pageNo].Add(sizerFrame, 0, wx.TOP | wx.ALIGN_LEFT, 0)
            
        for pageNo, page in enumerate(self.nbPage):
            page.SetSizer(self.sizerPage[pageNo])

        if self.bottomRow=="default":
            self.sizerBottomRow.Add(self.activeObject, 0, wx.ALL | wx.EXPAND, 2)
            self.sizerBottomRow.Add(self.buttonCancel, 0, wx.ALL | wx.EXPAND, 2)
            self.sizerBottomRow.Add(self.buttonOK, 0, wx.ALL | wx.EXPAND, 2)

        if self.bottomRow in ["default", "control"]:
            self.sizerGlobal.Add(self.sizerBottomRow, 0, wx.ALIGN_BOTTOM | wx.ALIGN_RIGHT, 0)
        
        self.SetSizer(self.sizerGlobal)

        self.Layout()


#------------------------------------------------------------------------------
# GUI actions - 0. OK/Cancel buttons
#------------------------------------------------------------------------------		

    def OnButtonOK(self, event):
        """
        Default event handler for ok button
        Reads and stores the data
        Closes the panel (and the calling window)
        """ 

        newData = self.readPanel()
        if self.mod and hasattr(self.mod, "saveToDB"):
            ret = self.mod.saveToDB(data=newData)
        else:
            ret = True
            
        if ret is None or ret: # accept ret=None for backward compatibility with saveToDB functions which do not return any value
            self.prnt.ret = True
            self.unsavedChanges = False                 # not really necessary, because panel closes. but for being sure.
            if event is not None:
                wx.SafeYield()
                self.autoClose = True
                self.prnt.Close()     # event None indicates that called from OnClose function
        else:
            msg = _("Data could not be saved. Please revise your data")
            if hasattr(self.mod, "errMsgSave") and self.mod.errMsgSave is not None:
                msg += "\n"+self.mod.errMsgSave

            logger.showWarning(msg)
            self.display()

    def OnButtonCancel(self, event):
        """
        default event handler: closes without saving
        """
        if hasattr(self.mod, "cancel"): self.mod.cancel()
        self.prnt.ret = None
        if event is not None:
            wx.SafeYield()
            self.autoClose = True
            self.prnt.Close()
        
#------------------------------------------------------------------------------
# Functions for read/write/refresh of the display
#------------------------------------------------------------------------------		
                                  
    def OnNotebookPageChanged(self, event):
        self.activePage = self.notebook.GetSelection()
        
        name = self.objectName
        if name is None:
            name = ""
            
        # this is a hack. seems that this event handler is already activated in the very beginning, when the notebook is created
        # and self.activeObject is not yet existing.
        if hasattr(self, "activeObject"):
            self.activeObject.SetLabel(U(name))
            
    def display(self, no=None):
        """
        displays data if available, otherwise fetches data from the backend
        """
        if self.mod:
            self.data = self.mod.updatePanel()

        permit = not self.blockEventHandlers
        self.blockEventHandlers = True

        self.getUserChange()        # catching changes before overwriting
        self.fillPanel()

##        wx.SafeYield()
        self.blockEventHandlers = not permit

        if self.notebook:
            self.OnNotebookPageChanged(None)

        self.Show()
        self.Refresh()

    def fillPanel(self):
        """
        helper function for display
        
        writes data to the corresponding panel entries
        this function should be personalised for the specific panel entries where necessary
        """
        if self.data is None: return
        
        for label in self.entry.keys():

            # if there are values for choices, set choices first and then set the value
            choicesLabel =  "Choices" + label
            if choicesLabel in self.data.keys():
                self.entry[label].SetValue(self.data[choicesLabel], emitEvent=False)
                                           
            # if there are values for FUnit2, set secondary unit first and then set the value
            funit2Label =  "FUnit2" + label
            if funit2Label in self.data.keys():
                self.entry[label].updateUnitDict(self.data[funit2Label])
                                           
            if label in self.data.keys():
                self.entry[label].SetValue(self.data[label], emitEvent=False)

    def setFrameTitle(self, title):
        self.prnt.SetTitle(U(title))

    def readPanel(self):
        """
        inverse of fill panel
        reads out all the entries on the display
        """
        newData = {}
        for label in self.entry.keys():
            if hasattr(self.entry[label], "GetValue"):          # security check. there are entries such as ScheduleEditor which do not have a GetValue function
                newData[label] = self.entry[label].GetValue()

        self.getUserChange()
        return newData

    
    def getUserChange(self):
        if self.autoClose: return False
        
        userChange = False
        for label in self.entry.keys():

            if hasattr(self.entry[label], "userChange") and self.entry[label].userChange:
                userChange = True

        if userChange: self.unsavedChanges = True
        return self.unsavedChanges
        
    def goTo(self, key):
        """
        tries to open the notebook page appropriate
        for a given entry key

        :params key: shortname of variable to be displayed
        """
        if not self.notebook: return True    # nothing to do

        if key in self.entry.keys():
            entry = self.entry[key]
            
            nbPage = entry.prnt if hasattr(entry, "prnt") else None
            if nbPage in self.nbPage:
                pageNo = self.nbPage.index(nbPage)
                self.notebook.SetSelection(pageNo)
                return True

        return False
            

#=====================================================================================================
