# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#    editHXNet: Visualise the heat exchanger and stream network
#
#==============================================================================
#
#	Version 3.0 (2018)
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import wx
from globals import logger, Globals
from GUI.status import Status
from constants import EPSILON, INFINITE, KELVIN
import numpy as np
from modules.energyStats.moduleEditHXNet import ModuleEditHXNet
from auxiliary.auxiliary import noneCheck

from GUI.GUITools import ORANGE
from lib.GUI.drawing.drawingPanel import DrawingPanel
from lib.GUI.drawing.moduleIcon import ModuleIcon, connectIcons

POPUPMENU =[_("matrix"), _("linear"), _("logarithmic")]
POPUPMENUIDS = [wx.NewId() for item in POPUPMENU]

# x-size of stream plot 
STREAMPLOTXMIN = 200    #x-position of the start of the plot
STREAMPLOTXWIDTH = 800  #x-width of the plot
STREAMPLOTXMAX = STREAMPLOTXMIN+STREAMPLOTXWIDTH
STREAMPLOTWINDOWWIDTH = 1000

# y-dimensions of stream plot
YPOSMARGIN = 20
STREAMPLOTYMIN = 40     #y-position of the start of the plot

# grid 
STREAMPLOTDELTAT = 40   # equivalent T-separation between grid positions
NMATRIXSLOTS = 12
STREAMPLOTDELTAY = 30   # y-separation between stream lines

# icons and labels
HXICONSIZE = wx.Size(4,4)
HXCONNHEIGHT = 2      # height of the rectangle showing the heat exchanger temperature range
HXCOLOURS = ['GREEN', 'ORANGE', 'MAGENTA', 'GREEN YELLOW', 'DARK GREEN', 'PINK'] # vary colours of heat exchangers for better readability
HXLABELWIDTH = 80

STREAMPLOTXLABEL = 20   #x-position of the stream label
STREAMPLOTLABELWIDTH = 160 #width of labels

def downStreamHX(node):
    hxNodes = []
    for q in node.qout:
        ds = q.toNode
        if ds.objectCode[0]=="HX":
            hxNodes.append(ds)
        else:
            hxNodes += downStreamHX(ds)
    return hxNodes

class EditHXNet(DrawingPanel):

    def __init__(self, parent, main, size=(1200, 400)):

# init-block from stream plot. to be updated.
##    def __init__(self, parent, srcStreams, snkStreams, Tmin=-60, Tmax = 400):

        Tmin = -20.0
        Tmax = 400.0
        srcStreams = []
        snkStreams = []
        
        if Tmin >= Tmax:
            logError("EditHXNet.__init__: called with Tmax < Tmin)")
            return
                     
        self.Tmin = Tmin
        self.Tmax = Tmax
        self.TWidth = float(self.Tmax - self.Tmin)

        self.hxColourIndex = 0
        
        self.occupiedSlots = []

#--------------------------------------------------------------
        self.main = main
        self.prnt = parent

        super(EditHXNet, self).__init__(parent, ModuleEditHXNet(), size=size)

        self.conLabelWidth = 35
        self.conSize = (35, 12)
        self.conPosIn = (25, 0)
        self.conPosOut = (-25, 0)

        self.TScale = "off" if Status.ANo < 0 else "lin"  # 0: matrix view, 1: linear T-Scale, 2: logarithmic T-Scale
        
##        self.display()
##        parent.Show()

    def buildDrawing(self):

        if not self.data: return # security feature if no data are loaded

        self.ymaxDrawing = 600 # minimum window size
        self.bitMaps = {"WARN": wx.Bitmap("GUI/img/icon_warning.png")}

        srcNet, snkNet = self.data["net"]
        self.posMaxSrc, self.posMaxSnk, self.posMax = self.data["grid"]

        if self.TScale != "off":
            self.Tmin, self.Tmax = self.data["TScale"]
            self.TWidth = float(self.Tmax - self.Tmin) if (self.Tmax is not None and self.Tmin is not None) else None

            if self.TWidth is None: # security feature against none-values
                self.Tmin = 0
                self.Tmax = 400
                self.TWidth = 400
        else:
            self.Tmin = 0
            self.Tmax = 400
            self.TWidth = 400

        self.hxTable = self.data["hx"]

        self.NSrc = self.posMaxSrc[1]+1
        self.NSnk = self.posMaxSnk[1]+1

        # calculate axis limits and build grid
        self.yAxis1 = STREAMPLOTYMIN - STREAMPLOTDELTAY
        self.yAxis2 = STREAMPLOTYMIN + STREAMPLOTDELTAY*(self.NSrc+self.NSnk+1)
        self.yBaseSnk = STREAMPLOTYMIN + (self.NSrc + 1)*STREAMPLOTDELTAY  # base line for cold stream plot

        if self.TScale=="off":
            self.gridTMin = 0
            self.gridTMax = STREAMPLOTDELTAT*max(self.posMaxSrc[0], self.posMaxSnk[0])
            self.gridDeltaT = STREAMPLOTDELTAT
        elif self.TScale in ["lin", "log"]:
            self.gridTMin = self.Tmin
            self.gridTMax = self.Tmax
            self.gridDeltaT = 20


# 0. initialise and draw the grid
#---------------------------------------

        # initialise arrays containing the drawing info           
        self.lines = []
        self.labels = []
        self.icons = []
            
        self.buildGrid()

# 1. draw lines for hot and cold streams
#---------------------------------------

        srcNodes = [node for node in srcNet.node]
        snkNodes = [node for node in snkNet.node]

        for node in srcNodes:

            if not hasattr(node, "pos"): continue     # plot only nodes with a position assigned

            j0 = 9999
            j1 = -9999
            for flow in node.qout:
                if not hasattr(flow, "pos"): continue     # plot only flows with a position assigned
            
                j = flow.pos[1]
                j0 = min(j, j0)
                j1 = max(j, j1)

                up = flow.fromNode                
                TH = self.TEqSrc(up.ipos) if self.TScale=="off" else up.T0

                down = flow.toNode
                TL = self.TEqSrc(down.ipos) if self.TScale=="off" else down.T1

                name = node.name if node.nodeType=="SRC" else None
                                    
                self.plotStream(name, TL, TH, j, heating=False, colour=wx.RED, flowNo=flow.no)

            if node.objectCode[0]=="SPLIT" and j1>j0:   # j1>j0: at least one flow could be plotted
                T = TH
                self.plotSplitter("source", T, j0, j1, node.split)

        
        for node in snkNodes:

            if not hasattr(node, "pos"): continue     # plot only nodes with a position assigned

            j0 = 9999
            j1 = -9999
            for flow in node.qout:
                if not hasattr(flow, "pos"): continue     # plot only flows with a position assigned
            
                j = flow.pos[1]
                j0 = min(j, j0)
                j1 = max(j, j1)

                up = flow.fromNode                
                TL = self.TEqSnk(up.ipos) if self.TScale=="off" else up.T1

                down = flow.toNode
                TH = self.TEqSnk(down.ipos) if self.TScale=="off" else down.T0

                name = node.name if node.nodeType=="SRC" else None
                
                self.plotStream(name, TL, TH, j, heating=True, colour=wx.BLUE, flowNo=flow.no)

            if node.objectCode[0]=="SPLIT" and j1>j0:  # j1>j0: at least one flow could be plotted
                T = TL
                self.plotSplitter("sink", T, j0, j1, node.split)
   


# 2. draw heat exchangers
#---------------------------------------

        self.occupiedSlots = []
        self.hxIcons = []
        for h, data in enumerate(self.hxTable):
            self.plotHX(h, data)

        self.icons = self.hxIcons + self.icons        

#-------------------- helper functions

        if Status.ANo > 0 and Status.StatusEnergy==0:
            warningIcon = ModuleIcon(parent=self,pos=wx.Point(150, self.yBaseSnk - 50),\
                                          name = "\n\n\n             " \
                                                 + _("temperatures not up to date") + ". \n" + _("run simulation"),
                                          code = ("WARN", 0),
                                          colour = wx.RED,
                                          outFlow = [],
                                          bitmap = self.bitMaps["WARN"])
            self.icons.append(warningIcon)



    def plotStream(self, name, TL, TH, j, heating=True, colour = wx.BLACK, flowNo=None):
        """
        nhx: vertical shift of outgoing horizontal line towards standard position
        nhy: vertical shift of ingoing horizontal line towards standard position
        nv:  horizontal position of vertical line
        """

        if heating:
            ypos = self.yposSnk(j)
        else:
            ypos = self.yposSrc(j)

        x1 = self.xpos(TL)
        x2 = self.xpos(TH)

        x1 = min(max(x1,STREAMPLOTXMIN),2*STREAMPLOTXMAX)
        x2 = min(max(x2,STREAMPLOTXMIN),2*STREAMPLOTXMAX)
        
        if abs(x2 - x1) < 8:     #minimum line width for visibility (avoid dots)
            x1 = 0.5*(x1+x2) - 4
            x2 = 0.5*(x1+x2) + 4
            
        r0 = (x1, ypos)
        r1 = (x2, ypos)
        rL = (STREAMPLOTXLABEL, ypos - 10)
        rIcon = (x1, ypos - 5)

        lineWidth = 1 if self.TScale=="off" else 2      # in schematic plot, information contained in stream lines is less relevant
        self.buildLine(r0,r1,colour=colour, width=lineWidth)
        if name is not None: self.buildLabel(rL, name)

        code = "CS" if heating else "HS"
        self.icons.append(StreamIcon(self,
                                     name="",
                                     pos=rIcon,
                                     iconSize=(int(x2-x1+0.5), 10),
                                     code = (code, flowNo),
                                     inFlow=[""] if heating else [],
                                     outFlow=[] if heating else [""]))

    def buildLabel(self,r, text, colour=wx.BLACK, align=wx.ALIGN_LEFT):
        self.labels.append((text, r[0], r[1], colour, align))
        pass

    def buildGrid(self):
        for T in xrange(int(self.gridTMin), int(self.gridTMax+self.gridDeltaT), int(self.gridDeltaT)):

            x = self.xpos(T)
            r0 = (x-10, self.yAxis1)
            r0Line = (x, self.yAxis1 + 0.5*STREAMPLOTDELTAY)
            r1 = (x-10, self.yAxis2)
            r1Line = (x, self.yAxis2 - 0.5*STREAMPLOTDELTAY)

            self.buildLine(r0Line, r1Line, width=1, colour='LIGHT GREY')

            if self.TScale=="off":
                if T==self.gridTMin:
                    for r in (r0, r1): self.buildLabel(r, "T_min")
                elif T==self.gridTMax:
                    for r in (r0, r1): self.buildLabel(r, "T_max")
                
            else:
                if T%100 == 0 or T == self.gridTMin or T == self.gridTMax:
                    label = "%3.0f"%T
                    self.buildLabel(r0,label)
                    self.buildLabel(r1,label)
                
    def buildLine(self, r1, r2, colour=wx.BLACK, width=1):
        #TODO: width is not use. other panels use style and not width
        self.lines.append((r1, r2, colour, wx.SOLID, width))
        

    def buildRectangle(self,x1, x2, colour=wx.BLACK, width=1):

        x3 = (x1[0], x2[1])
        x4 = (x2[0], x1[1])
        self.buildLine(x1, x3, colour, width)
        self.buildLine(x3, x2, colour, width)
        self.buildLine(x2, x4, colour, width)
        self.buildLine(x4, x1, colour, width)

    def xpos(self, T):
        """ gets the x-coordinate corresponding to the temperature
        """
        if T is not None:
            if self.TScale=="log":
                logMin = np.log(self.Tmin+KELVIN)
                logMax = np.log(self.Tmax+KELVIN)
                logTWidth = logMax - logMin
                return int(STREAMPLOTXMIN + float(np.log(T+KELVIN) - np.log(self.Tmin+KELVIN))/logTWidth*STREAMPLOTXWIDTH + 0.5)                
            else:
                return int(STREAMPLOTXMIN + float(T - self.Tmin)/self.TWidth*STREAMPLOTXWIDTH + 0.5)
        else:
            return STREAMPLOTXMIN

    def TEqSrc(self, i):
        return self.TEqSnk(i)
##        return STREAMPLOTDELTAT * (0 if i < 0 else (self.posMax[0] - i))
    
    def TEqSnk(self, i):
        return STREAMPLOTDELTAT * (self.posMax[0] if i < 0 else i)
        
    def yposSrc(self, j):
        """ gets the y-coordinate corresponding to the stream-line
            of the n-th stream in the plot
        """
        return int(STREAMPLOTYMIN + j*STREAMPLOTDELTAY)
        
    def yposSnk(self, j):
        """ gets the y-coordinate corresponding to the stream-line
            of the n-th stream in the plot
        """
        return int(self.yBaseSnk + (self.NSnk - j - 1)*STREAMPLOTDELTAY)
        

    def plotHX(self, h, data):

        srcNode, snkNode, Thi, Tho, Tci, Tco, name = data
        """ draws a heat exchanger within a stream plot
        the src and snk numbers have to be determined beforehand
        """
        if srcNode is None or snkNode is None or not hasattr(srcNode, "pos") or not hasattr(snkNode, "pos"):
            logger.logError("EditHXNet.plotHX: received corrupt information for heat exchanger %s (%s)"%(h,name))
            return

        hxColour = HXCOLOURS[self.hxColourIndex]
        self.hxColourIndex += 1
        if self.hxColourIndex >= len(HXCOLOURS): self.hxColourIndex = 0

        # plot source
        jsrc = srcNode.pos[1]
        ypos1 = self.yposSrc(jsrc) - 0.5*HXCONNHEIGHT
        ypos2 = ypos1 + HXCONNHEIGHT


        if self.TScale=="off":
            TH = self.TEqSrc(srcNode.ipos)

            Tho = TH - 0.25*STREAMPLOTDELTAT
            Thi = TH + 0.25*STREAMPLOTDELTAT
        
        xpos1 = self.xpos(Tho)
        xpos2 = self.xpos(Thi)
        
        self.buildRectangle((xpos1, ypos1), (xpos2, ypos2), colour=hxColour, width = 2)
        self.buildRectangle((xpos1-1, ypos1-1), (xpos2+1, ypos2+1), colour=wx.BLACK, width = 1)

        xSrc = int(0.5*(xpos1 + xpos2) + 0.5)
        if self.TScale=="off":
            xSrcAsym = xpos1 + int((xpos2-xpos1)*(self.NSrc-jsrc-0.5)/self.NSrc)
        else:
            xSrcAsym = min(xpos1 + 10, xSrc) # shift xSrc towards the lower temp
        ySrc = ypos2

        # plot sink
        jsnk = snkNode.pos[1]
        ypos1 = self.yposSnk(jsnk) + 0.5*HXCONNHEIGHT
        ypos2 = ypos1 - HXCONNHEIGHT

        if self.TScale=="off":
            TC = self.TEqSnk(snkNode.ipos)

            Tci = TC - 0.25*STREAMPLOTDELTAT
            Tco = TC + 0.25*STREAMPLOTDELTAT

        xpos1 = self.xpos(Tci)
        xpos2 = self.xpos(Tco)
        
        self.buildRectangle((xpos1, ypos1), (xpos2, ypos2), colour=hxColour, width = 2)
        self.buildRectangle((xpos1-1, ypos1+1), (xpos2+1, ypos2-1), colour=wx.BLACK, width = 1)

        xSnk = int(0.5*(xpos1 + xpos2) + 0.5)
        if self.TScale=="off":
            xSnkAsym = xpos1 + int((xpos2-xpos1)*(jsnk+0.5)/self.NSnk)
        else:
            xSnkAsym = max(xpos1, xpos2-10) # shift xSnk towards the higher temp
        ySnk = ypos2


        # plot HXIcon
        if xSrcAsym < xSnkAsym and self.TScale != "off":
            xLine = int(0.5*(xSrcAsym + xSnkAsym) + 0.5)
            xLine = min(xSrc, xLine) # not at higher T than the symmetric xSrc
            xLine = max(xSnk, xLine) # not at lower T than the symmetric xSnk

            xSrcAsym = xLine
            xSnkAsym = xLine
            
        ypos1 = self.yposSnk(jsnk)                       
        xHX = int(0.5*(xSrcAsym + xSnkAsym) + 0.5)
        yHX = int(ypos1 - 0.5 * STREAMPLOTDELTAY + 0.5)

        jSlotHX = jsnk

        while (not self.freeSlot(jSlotHX, xHX, xHX)) and (jSlotHX < self.NSnk + self.NSrc - jsrc - 1):
##        while (not self.freeSlot(jSlotHX, xSnkAsym, xSrcAsym)) and (jSlotHX < self.NSnk + self.NSrc - jsrc - 1):
            jSlotHX += 1
            yHX -= STREAMPLOTDELTAY

        pos=(int(xHX), int(yHX))
        posIcon = (xHX-10, yHX-10)
        
        hx = HXIcon(self.prnt, self, pos=pos, name=name, colour = hxColour)
        # add an invisible icon (20x20) area for receiving clicks
        self.hxIcons.append(ModuleIcon(parent=self, \
                                           pos=posIcon, \
                                           iconSize=(20, 20), \
                                           name = "",\
                                           code = ("HX", h),\
                                           colour = hxColour, \
                                           inFlow = [""],
                                           outFlow = [""]))
##                                           bitmap = self.bitMaps["HX"]))
        

        # connect HXIcon with source and sink
        dx = xSnkAsym - hx.pos[0]
        r0 = (xSnkAsym, ySnk)
        r2 = hx.snkPos(dx)
        r1 = (r0[0], r2[1])

        conWidth = 2 if self.TScale=="off" else 1        
        self.buildLine(r0, r1, colour=hxColour, width=conWidth)
        self.buildLine(r1, r2, colour=hxColour, width=conWidth)

        dx = xSrcAsym - hx.pos[0]
        r0 = (xSrcAsym, ySrc)
        r2 = hx.srcPos(dx)
        r1 = (r0[0], r2[1])

        self.buildLine(r0, r1, colour=hxColour, width=conWidth)
        self.buildLine(r1, r2, colour=hxColour, width=conWidth)

        self.occupiedSlots.append((jSlotHX,                                 # row where HX is located (-> something like an jSlot)
                                   min(xSnkAsym, xSrcAsym), 
                                   max(xSrcAsym, xSnkAsym)))


    def plotSplitter(self, splitterType, T, j1, j2, split):
        """ represents a flow splitter in the graphic
        :params splitterType: source or sink
        :params data: tuple of (TSplit, n1, n2) where n1, n2 is the extendes source index of the upper and lower stream
        """

        xpos = self.xpos(T)
        
        if splitterType == "source":
            ypos1 = self.yposSrc(j1)
            ypos2 = self.yposSrc(j2)
            colour = wx.RED
            dxLabel = -28
            alignLabel = wx.ALIGN_LEFT

        else:
            ypos2 = self.yposSnk(j1)
            ypos1 = self.yposSnk(j2)
            colour = wx.BLUE
            dxLabel = +6
            alignLabel = wx.ALIGN_RIGHT

        r0 = (xpos-1, ypos1)
        r1 = (xpos-1, ypos2)

        self.buildLine(r0, r1, colour)

        r0 = (xpos, ypos1)
        r1 = (xpos, ypos2)

        self.buildLine(r0, r1, colour)
        
        r0 = (xpos+1, ypos1)
        r1 = (xpos+1, ypos2)

        self.buildLine(r0, r1, colour)

        r0 = (xpos-4, ypos1-4)
        r1 = (xpos+4, ypos2+4)

        rLabel = (xpos+dxLabel, ypos1-12)


        self.buildRectangle(r0, r1, colour='BLACK')

        splitFractions = split[:] # break link in order to avoid reversing the original
        if splitterType == "sink":
            splitFractions.reverse()
        
        for split in splitFractions:
            self.buildLabel(rLabel, ("%.3f"%split).lstrip("0") if split is not None else "?", align=alignLabel)
            rLabel = (rLabel[0], rLabel[1]+STREAMPLOTDELTAY)

##        if splitterType == "sink":
##            for ySlot in (data[1], data[2]):
##                self.occupiedSlots.append((ySlot, xpos-4 , xpos+4))
        

    def freeSlot(self, jSlot, x0, x1):
        """
        checks if the space to be occupied by the logo and/or horizontal line of the
        HX is already occupied

        """
        xmin = min(x0, x1)
        xmax = max(x0, x1)
        
        for slot in self.occupiedSlots:
            if jSlot == slot[0]:    # check only the slots with the same j-position
                dxMin = HXICONSIZE[0] + HXLABELWIDTH
                slotMin = min(slot[1:3])
                slotMax = max(slot[1:3])
                if xmax > slotMin - dxMin and xmin < slotMax + dxMin:
                    return False

        return True


#################################################
    def editIcon(self, icon):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function

            function should return None if icon is not editable,
            otherwise any object (node) with attributes:
            - nodeType
            - scaleFactor
            - splitFraction
        """
        if len(icon.code)>=2: # security control
            if icon.code[0]=="HX":
                self.main.control(icon.code[0], icon.code[1])
            else:
                pass # do nothing with stream icons

        return None

    def insertIcon(self, pos):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        if not self.editAllowed(): return
        
        # create pop-up
        menu = wx.Menu()
        for i, title in enumerate(POPUPMENU):
            id = POPUPMENUIDS[i]
            menu.Append( id, title )
            wx.EVT_MENU(menu, id, self.OnPopUpIcons)

        # Display menu with call to PopupMenu, invoked on the source component, passing event's GetPoint. ###
        self.PopupMenu(menu, pos)
        menu.Destroy() # destroy to avoid mem leak
        wx.SafeYield()

        if self.actionCode==0:
            self.TScale = "off"
        elif self.actionCode==1:
            self.TScale = "lin"
        elif self.actionCode==2:
            self.TScale = "log"

        self.display()

        return None

    def deleteIcon(self, icon):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        if not self.editAllowed(): return False

        if len(icon.code)>=2 and icon.code[0]=="HX": # security control
            h = icon.code[1]
            name = Status.int.hx[h].name

            ret = logger.askConfirmation(_("Do You want to delete %s ?"%(name)))
            if ret:
                Status.prj.deleteObject(icon.code[0], icon.code[1])
                Globals.GUI.updateDrawings()
                self.display()

        return False    # false indicates that no default action is taken by the drawing panel

    def editAllowed(self):
        if Status.ANo <= 0 and Status.StatusCC > 0:
            logger.showWarning(_("You cannot modify the present state of the system after the data has been checked.\n") + 
                        _("If you want to change the present state you may unblock the data by ") + 
                        _("choosing to confirm the deletion ") + 
                        _("of any alternative proposals in this project the next time you switch to the present state."))
            return False
        return True
        
    def OnPopUpIcons(self, event):
        actionCode = POPUPMENUIDS.index(event.GetId())
        self.actionCode = actionCode

    def buildConnection(self, startIcon, startNo, endIcon, endNo):

        if not self.editAllowed():
            return False

        srcCode = startIcon.code[0]
        snkCode = endIcon.code[0]
        if not(srcCode=="HS" and snkCode=="CS"):
            return False

        srcNo = startIcon.code[1]
        snkNo = endIcon.code[1]

        srcNet, snkNet = self.data["net"]
        srcFlow = srcNet.flow[srcNo]
        snkFlow = snkNet.flow[snkNo]

        Globals.GUI.control("HX", None)
        self.modQ6 = Globals.GUI.editPanel.mod

        # upstream node
        srcNode = srcFlow.fromNode
        oType, oNo, cNo, heating = srcNode.objectCode
        fraction = 1.0
        while oType=="SPLIT":
            srcCon = srcFlow.fromCon
            fraction *= srcNode.split[srcCon]
            oType, oNo, cNo, heating = srcNode.qin[0].fromNode.objectCode   # look for upstream node
            srcNode = srcNode.qin[0].fromNode
            print "Upstream node: %s fraction %s"%(srcNode.objectCode, fraction)

        codeDS = srcFlow.toNode.objectCode
        if codeDS[0] in ["HX", "SPLIT"]:
            dsHX = [node.objectCode for node in downStreamHX(srcNode)]
            self.modQ6.changeConnection(codeDS[0], codeDS[1], heating, dsHX)
            
        self.modQ6.addConnection(heating=False)
        self.modQ6.activeConn.objectType=oType
        self.modQ6.activeConn.objectNo = oNo
        self.modQ6.activeConn.componentNo = cNo
        self.modQ6.activeConn.fraction = fraction
        self.modQ6.activeConn.streamName, self.modQ6.activeConn.objectName = self.modQ6.activeConn.getDefaultStreamName()
        self.modQ6.storeConnection(fraction)

        snkNode = snkFlow.fromNode

        oType, oNo, cNo, heating = snkNode.objectCode
        fraction = 1.0
        while oType=="SPLIT":
            snkCon = snkFlow.fromCon
            fraction *= snkNode.split[snkCon]
            oType, oNo, cNo, heating = snkNode.qin[0].fromNode.objectCode   # look for upstream node
            snkNode = snkNode.qin[0].fromNode
    
        codeDS = snkFlow.toNode.objectCode
        if codeDS[0] in ["HX", "SPLIT"]:
            dsHX = [node.objectCode for node in downStreamHX(snkNode)]
            self.modQ6.changeConnection(codeDS[0], codeDS[1], heating, dsHX)

        self.modQ6.addConnection(heating=True)
        self.modQ6.activeConn.objectType=oType
        self.modQ6.activeConn.objectNo = oNo
        self.modQ6.activeConn.componentNo = cNo
        self.modQ6.activeConn.fraction = fraction
        self.modQ6.activeConn.streamName, self.modQ6.activeConn.objectName = self.modQ6.activeConn.getDefaultStreamName()
        self.modQ6.storeConnection(fraction)

        hx = self.modQ6.hx
        Globals.GUI.editPanel.hx = self.modQ6.hx
        Globals.GUI.editPanel.showConnections()
        Globals.GUI.editPanel.display()
        
        return False
        
    def deleteConnection(self, n):
        if not self.editAllowed():
            return False

        if self.flow[n].deleteFromDB():
            self.display()
            self.prnt.Hide()
            self.prnt.Show()        
            
        return False

        
##def strRep(val, prefix="", minVal=-INFINITE):
##    return " " if (val is None or val <= minVal) else prefix+"%.0f"%(val/1000)
##

class HXIcon(wx.Panel):
    """ Icon for the heat exchanger
    """
    
    def __init__(self, parent, main, pos=wx.Point(0,0),name = "---",colour = wx.GREEN):
        self.pos = pos  # center of the heat exchanger
        self.name = name
        self.main = main
        self.colour = colour

        r0 = (pos[0] - 0.5*HXICONSIZE[0], pos[1]-0.5*HXICONSIZE[1])
        r1 = (pos[0] + 0.5*HXICONSIZE[0], pos[1]+0.5*HXICONSIZE[1])
        self.main.buildRectangle(r0, r1, colour=self.colour, width = 8)
        self.main.buildRectangle((r0[0]-4, r0[1]-4), (r1[0]+4, r1[1]+4), colour=wx.BLACK, width = 1)
        

        r2 = (pos[0]+0.5*HXICONSIZE[0]+6, pos[1]-10)
        self.main.buildLabel(r2, self.name)
       
    def srcPos(self, dx=0):
        (x,y) = self.pos
        x += int(0.5*HXICONSIZE[0])
        y += int(0.5*HXICONSIZE[1])

##        if abs(dx) < 0.5*HXICONSIZE[0]:
##            x += int(0.5*HXICONSIZE[0])
##        elif dx > 0:
##            x += HXICONSIZE[0]
##            y += int(0.5*HXICONSIZE[1])
##        else:
##            y += int(0.5*HXICONSIZE[1])
            
        return (x,y)
    
    def snkPos(self, dx=0):
        (x,y) = self.pos
        x += int(0.5*HXICONSIZE[0])
        y += int(0.5*HXICONSIZE[1])
##        if abs(dx) < 0.5*HXICONSIZE[0]:
##            x += int(0.5*HXICONSIZE[0])
##            y += HXICONSIZE[0]
##        elif dx > 0:
##            x += HXICONSIZE[0]
##            y += int(0.5*HXICONSIZE[1])
##        else:
##            y += int(0.5*HXICONSIZE[1])
        return (x,y)

class StreamIcon(ModuleIcon):
    """ modified icon class for representation of streams """

    def __init__(self, parent, pos=wx.Point(0,0), iconSize=None, name = "---", code=None, colour = wx.GREEN, inFlow = [], outFlow=[], bitmap=None):

        super(StreamIcon, self).__init__(parent, pos=pos,
                                         iconSize=iconSize, name = name, code=code,
                                         colour = colour, inFlow = inFlow, outFlow=outFlow, bitmap=bitmap,
                                         conSize=iconSize,        # personalise to icon size (all the icon is connection)
                                         conPosIn=(iconSize[0], -1.5*parent.conYSep),
                                         conPosOut=(-iconSize[0], -1.5*parent.conYSep),
                                         conLabelWidth=0, conLabelShift=0, conTextShift=0, conStyle=0)

       
