# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#    editMF: Visualise the material streams (mass flows) in the production line
#
#==============================================================================
#
#    Version No.: 0.01
#    Created by:         Hans Schweiger 07/07/2011
#
#------------------------------------------------------------------------------
#    (C) copyleft energyXperts.BCN (E4-Experts SL), Barcelona, Spain 2011-2012
#    http://www.energyxperts.net/
#
#    This program is free software: you can redistribute it or modify it under
#    the terms of the GNU general public license as published by the Free
#    Software Foundation (www.gnu.org).
#
#==============================================================================
import wx
from modules.guiModules.moduleEditMF import ModuleEditMF
from GUI.status import Status
from modules.fluids import Fuel
from constants import EPSILON, INFINITE
from GUI.GUITools import ORANGE
from modules.messageLogger import askConfirmation

YPOSMARGIN = 40

XPOS0 = 600     # xposition of product icons (pos = 0)
DXPOS = 200     # separation of icon columns

from lib.GUI.drawing.drawingPanel import DrawingPanel
from lib.GUI.drawing.moduleIcon import ModuleIcon, connectIcons
from lib.GUI.drawing.nodeEditor import NodeData
from lib.GUI.drawing.iconGrid import IconGrid


class EditMF(DrawingPanel):

    def __init__(self, parent, main=None):
        self.activeNode = None
        self.activeIcon = None

        super(EditMF, self).__init__(parent, ModuleEditMF(), conStyle=1)
        self.prnt = parent
        self.main = main

        self.display()
        self.prnt.Show()

        self.gridOrigin = (self.gridOrigin[0], 40) # hack for making recirculating mass flows in first row better visible



    def buildDrawing(self):
        self.bitMaps = {"Fuel": wx.Bitmap("GUI/img/fuelIcon.png"),
                      "INT11": wx.Bitmap("GUI/img/nodeIcon-INT11.png"),
                      "INT12": wx.Bitmap("GUI/img/nodeIcon-INT12.png"),
                      "INT13": wx.Bitmap("GUI/img/nodeIcon-INT13.png"),
                      "INT21": wx.Bitmap("GUI/img/nodeIcon-INT21.png"),
                      "INT31": wx.Bitmap("GUI/img/nodeIcon-INT31.png"),
                      "ProcH": wx.Bitmap("GUI/img/processHIcon.png"),
                      "ProcC": wx.Bitmap("GUI/img/processCIcon.png")}

        self.ymaxDrawing = 600 # minimum window size
        
        self.lines = []
        
        self.node = self.data["Nodes"]
        self.flow = self.data["Flows"]

        self.buildScheme()  # distributes the nodes on the grid of the drawing

        NN = len(self.node)

        self.icons = []

        for n, node in enumerate(self.node):
            colour = wx.RED
            xpos = self.gridOrigin[0] + node.xpos * self.gridSize[0]
            ypos = self.gridOrigin[1] + node.ypos * self.gridSize[1]

            if node.objectType=="PROC":
                bmpKey = "ProcH"
            elif node.objectType=="INT":
                if len(node.qin) <= 1 and len(node.qout) <= 1:
                    bmpKey = "INT11"
                elif len(node.qin) <= 1 and len(node.qout) == 2:
                    bmpKey = "INT12"
                elif len(node.qin) <= 1 and len(node.qout) >= 3:
                    bmpKey = "INT13"
                elif len(node.qin) == 2 and len(node.qout) <= 1:
                    bmpKey = "INT21"
                elif len(node.qin) >= 3 and len(node.qout) <= 1:
                    bmpKey = "INT31"
                else: 
                    bmpKey = "INT11"
                    
            else:
                bmpKey = "Fuel"
            self.icons.append(ModuleIcon(parent=self,
                                               pos=wx.Point(xpos,ypos),
                                               name=node.name,
                                               colour = colour,
                                               inFlow = node.inputs,
                                               outFlow = node.outputs,
                                               bitmap = self.bitMaps[bmpKey]))

            self.icons[-1].qin = [None if q is None else strRep(q) for q in node.qinVal]
            self.icons[-1].qout = [None if q is None else strRep(q) for q in node.qoutVal]

        
        # now build connections between icons
        self.lines = []
        self.connections = []
        
        nnv = [0]*(self.ncols+1)
        nnh = [0]*self.nrows
        
        for n, flow in enumerate(self.flow):
            fromIcon = self.icons[self.node.index(flow.fromNode)]
            toIcon = self.icons[self.node.index(flow.toNode)]

            vcs = flow.fromNode.xpos+1
            nvs = nnv[vcs]
            nnv[vcs] += 1
            
            vct = flow.toNode.xpos
            nvt = nnv[vct]
            nnv[vct] += 1

            hc = flow.fromNode.ypos
            nh = nnh[hc]
            nnh[hc] += 1

            if flow.real:
                colour = ORANGE
                style = wx.SOLID
            else:
                colour = wx.BLACK
                style = wx.SHORT_DASH
                
            conLines = connectIcons(fromIcon, toIcon, inputNo=flow.toCon, outputNo=flow.fromCon,
                                       nv=nvs, nvt=nvt, nhback=nh, colour=colour, style=style)

            self.connections.append(conLines)       
            self.lines += conLines               


    def buildScheme(self):

        self.iconGrid = IconGrid(len(self.node))
        
        # initialise node settings
        for node in self.node:
            node.clearPosition()

        # put product nodes to the right
        for n, node in enumerate(self.node):
            if node.objectType == "PROD":
##                node.xpos = 0
##
                colPtr = max(self.iconGrid.ncols-1, 0)
                rowPtr = self.iconGrid.nrows


                self.iconGrid.addNode(n, rowPtr, colPtr)    #add the node
                
                rowPtr, colPtr = self.addSourceNodes(n, rowPtr, colPtr)

        # now add any other remaining node (except raw materials)
                
        remainingNodes = self.iconGrid.remainingNodes[:]
        for n in remainingNodes:
            node = self.node[n]
            if node.objectType != "RAW":
                colPtr = self.iconGrid.ncols-2
                rowPtr = self.iconGrid.nrows
                if colPtr < 0:
                    self.iconGrid.insertCol(0)
                    colPtr = 0

                ret = self.iconGrid.addNode(n, rowPtr, colPtr)    #add the node                
##                rowPtr, colPtr = self.addSourceNodes(n, rowPtr, colPtr)
##
        # finally add the raw materials which had not been input to any other node to the left
        remainingNodes = self.iconGrid.remainingNodes[:]
        for n in remainingNodes:
            node = self.node[n]
            colPtr = 0
            rowPtr = self.iconGrid.nrows
            self.iconGrid.addNode(n, rowPtr, colPtr)
        
        for r, row in enumerate(self.iconGrid.grid):
            for c, n in enumerate(row):
                if n is not None:
                    self.node[n].xpos = c
                    self.node[n].ypos = r

        self.ncols = self.iconGrid.ncols
        self.nrows = self.iconGrid.nrows

    def addSourceNodes(self, nt, rowt, colt):
        """ helper function for build-scheme
        nt: number of the target node
        colt: column of the target node
        """
        colPtr = colt - 1
        rowPtr = rowt


        remainingNodes = self.iconGrid.remainingNodes[:]
        for nsrc in remainingNodes:
            node = self.node[nsrc]
            for qout in node.qout:
                if qout is None:
                    continue
                elif qout.toNode == self.node[nt]:
                    if colPtr < 0:
                        colPtr = 0
                        self.iconGrid.insertCol(0)
                    ret = self.iconGrid.addNode(nsrc, rowPtr, colPtr)
                    if ret: rowPtr, colPtr = self.addSourceNodes(nsrc, rowPtr, colPtr)
                    break

        return (rowPtr + 1), (colPtr + 1)

        

    def buildConnection(self, startIcon, startNo, endIcon, endNo):
        startNode = self.node[self.icons.index(startIcon)]
        endNode = self.node[self.icons.index(endIcon)]
        self.mod.buildConnection(startNode, startNo, endNode, endNo)
        
    def deleteConnection(self, n):
        ret = askConfirmation(_("Do you want to delete this connection ?"))
        if ret:
            self.mod.deleteConnection(n)
            return True
        else:
            return False

    def changeConnection(self, n, mode):
        return self.mod.changeConnection(n, mode)

    def insertIcon(self, pos):
        node = self.mod.insertNode()
        self.activeNode = node
        return NodeData(node.nodeType,
                            name = node.name,
                            scaleFactor=node.scaleFactor,
                            splitFraction=node.splitFraction)

    def deleteIcon(self, icon):
        ret = askConfirmation(_("Do you want to delete this node ?"))
        if ret:
            node = self.node[self.icons.index(icon)]
            self.mod.deleteNode(node)
            return True
        else:
            return False

    def editIcon(self, icon):
        node = self.node[self.icons.index(icon)]
        self.activeNode = node
        self.activeIcon = icon
        if node.objectType in ["RAW", "INT"]:                
            return NodeData(node.nodeType,
                            name = node.name,
                            scaleFactor=node.scaleFactor,
                            splitFraction=node.splitFraction)
        else:
            return None
        
    def saveIcon(self, nodeData):
        if self.activeNode is not None:
            self.mod.saveNode(self.activeNode, nodeData)
        
            if self.activeIcon is not None:
                self.activeIcon.name = self.activeNode.name
                self.activeIcon.inFlow = self.activeNode.inputs
                self.activeIcon.outFlow = self.activeNode.outputs
                self.activeIcon.qin = [None if q is None else strRep(q) for q in self.activeNode.qinVal]
                self.activeIcon.qout = [None if q is None else strRep(q) for q in self.activeNode.qoutVal]
                    

def strRep(val, prefix="", minVal=-INFINITE):
    return " " if (val is None or val <= minVal) else prefix+"%.0f"%(val/1000)


