# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQ6: Heat exchangers
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import copy
import os
import wx
from status import Status
from constants import *
from globals import Globals, logger
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, toUTF8
from GUI.fonts import FontProperties
from GUI.displayClasses import FieldSizes, TextEntry, ChoiceEntry, FloatEntry
from GUI.GUITools import fillChoice
from database.dbio import check
import database.dbio as dbio
from GUI.conflictFrame import conflictFrame
from GUI.displayCC import PanelCCPars
from modules.fluids import Fluid

# constants that control the default sizes
# 1. font sizes
TYPE_SIZE_LEFT = 9
TYPE_SIZE_MIDDLE = 9
TYPE_SIZE_RIGHT = 9
TYPE_SIZE_TITLES = 10

# 2. field sizes
HEIGHT = 24
HEIGHT_RIGHT = 24

LABEL_WIDTH_LEFT = 400
LABEL_WIDTH_RIGHT = 400
LABEL_WIDTH_LEFT_TABLE = 200
LABEL_WIDTH_RIGHT_TABLE = 0
LABEL_WIDTH_ARROW = 40


DATA_ENTRY_WIDTH = 80
DATA_ENTRY_WIDTH_LEFT = 80
DATA_ENTRY_WIDTH_TABLE = 60

UNITS_WIDTH = 80
UNITS_WIDTH_TABLE = 50

# 3. vertical separation between fields
VSEP_LEFT = 2
VSEP_RIGHT = 2


ORANGE = '#FF6000'
TITLE_COLOR = ORANGE


class PanelQ6B(wx.Panel):
    def __init__(self, parent, main):
        self.main = main
        self.prnt = parent

        self.WHEEID = None

        self.WHEENo = None

        self.WHEEName = None

        self._init_ctrls(parent)
        self.__do_layout()

##        self.mod = ModuleQ6B("Q6B")
        

        #TODO: check wether this stuff is still needed
##        self.selectedSStream = None
##        self.selectedStreamNo = None
##        self.selectedStream = None
        
##        self.temperatureIsChanged = False
        
        self.WHEEFluid = None

    def _init_ctrls(self, parent):
        """UI setup"""
        import platform # this is a hack to the GUI for gtk platforms
        (dummy, arch) = platform.architecture()
        if "gtk2" in wx.PlatformInfo:
            wx.Panel.__init__(self, id= -1, name='PanelQ6', parent=parent,
                              pos=wx.Point(0, 0), size=wx.Size(780, 600), style=0)
        else:
            wx.Panel.__init__(self, id= -1, name='PanelQ6', parent=parent,
                              pos=wx.Point(0, 0), size=wx.Size(780, 580), style=0)
        #TODO: check why this "GTK2-Hack" is necessary and - either delete or - if really necessary - introduce to ALL panels
        self.Hide()

        # access to font properties object
        fp = FontProperties()

        self.frame_equipment_data = wx.StaticBox(self, -1, _U("Equipment data"))
        self.frame_equipment_data.SetForegroundColour(TITLE_COLOR)
        self.frame_equipment_data.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        self.frame_schedule = wx.StaticBox(self, -1, _U("Schedule"))
        self.frame_schedule.SetForegroundColour(TITLE_COLOR)
        self.frame_schedule.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        # set font for titles
        # 1. save actual font parameters on the stack
        fp.pushFont()
        # 2. change size and weight
        fp.changeFont(size=TYPE_SIZE_TITLES, weight=wx.BOLD)
        self.frame_schedule.SetFont(fp.getFont())
        self.frame_equipment_data.SetFont(fp.getFont())
        # 3. recover previous font state
        fp.popFont()
      
        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH_LEFT,
                       wData=DATA_ENTRY_WIDTH_LEFT, wUnits=UNITS_WIDTH)

        # left tab controls
        # tab 0 - Heat recovery from thermal equipment
        fp.pushFont()
        fp.changeFont(size=TYPE_SIZE_LEFT)

        # right tab controls
        # tab 1 - Heat recovery from electrical equipment
        fp.changeFont(size=TYPE_SIZE_RIGHT)
        f = FieldSizes(wHeight=HEIGHT_RIGHT, wLabel=LABEL_WIDTH_RIGHT)

        # right upper static box: Equipment data
        self.tc101 = TextEntry(self, maxchars=255, value='',
                               label=_U("Short name of electrical equipment"),
                               tip=_U("Give a short name of the equipment"))

        self.tc102 = ChoiceEntry(self,
                                 values=TRANSWHEEEQTYPES.values(),
                                 label=_U("Equipment type"),
                                 tip=_U("specify type of equipment, e.g. compressor, electric motor,..."))

        self.tc103 = ChoiceEntry(self,
                                 values=TRANSWHEEWASTEHEATTYPES.values(),
                                 label=_U("Waste heat type"),
                                 tip=_U("Specify type of waste heat (e.g. Recooling of compressed air, cooling water of motor/compressor, ...)"))
        
        self.tc104 = FloatEntry(self,
                                decimals=1, minval=0., maxval=1.e+12, value=0.,
                                unitdict='POWER',
                                label=_U("Available waste heat"),
                                tip=_U("Estimated quantity"))

        self.tc105 = ChoiceEntry(self,
                                 values=[],
                                 label=_U("Medium"),
                                 tip=_U("Waste heat carrying medium (fluid)"))

        self.Bind(wx.EVT_CHOICE, self.OnFluidChange, self.tc105.entry)

        self.tc106 = FloatEntry(self,
                                decimals=1, minval=0., maxval=1.e+12, value=0.,
                                unitdict='MASSORVOLUMEFLOW',
                                label=_U("Flow rate"),
                                tip=_U("Specify the flow rate of the waste heat carrying medium"))

        self.tc107 = FloatEntry(self,
                                decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                                unitdict='TEMPERATURE',
                                label=_U("Waste heat temperature"),
                                tip=_U("Specify the temperature of the waste heat medium at the outlet"))


        self.tc108 = ChoiceEntry(self,
                                 values=TRANSYESNO.values(),
                                 label=_U("Present use of waste heat"),
                                 tip=_U("If yes, specify distribution pipe / duct or heat exchanger where waste heat is used at present"))

        # right lower static box Schedule

        self.tc110 = FloatEntry(self,
                                decimals=1, minval=0., maxval=999., value=0.,
                                unitdict='TIME',
                                label=_U("Hours of  operation per day"),
                                tip=_U(" "))

        self.tc111 = FloatEntry(self,
                                decimals=1, minval=0., maxval=99., value=0.,
                                label=_U("Number of batches per day"),
                                tip=_U(" "))

        self.tc112 = FloatEntry(self,
                                decimals=1, minval=0., maxval=24 * 60 * 60., value=0.,
                                unitdict='TIME',
                                label=_U("Duration of 1 batch"),
                                tip=_U(" "))

        self.tc113 = FloatEntry(self,
                                decimals=1, minval=0., maxval=365., value=0.,
                                label=_U("Days of operation per year"),
                                tip=_U(" "))

        #=================================== source/sink panel



        #
        # buttons
        #
        self.buttonOK = wx.Button(self, wx.ID_OK, "")
        self.Bind(wx.EVT_BUTTON, self.OnButtonOK, self.buttonOK)
        self.buttonOK.SetDefault()

        self.activeObject = wx.StaticText(self,id=-1,label="----- ----- -----",style = wx.ALIGN_LEFT)
        self.activeObject.SetMinSize((580, 24)) #a hack in order to achieve alignment at the left. can surely be done better.

        self.buttonCancel = wx.Button(self, wx.ID_CANCEL, "")
        self.Bind(wx.EVT_BUTTON, self.OnButtonCancel, self.buttonCancel)

        # recover previous font parameters from the stack
        fp.popFont()


    def __do_layout(self):
        flagText = wx.ALIGN_CENTER_VERTICAL | wx.TOP

        # global sizer for panel. Contains notebook w/two tabs + buttons Cancel and Ok
        sizerGlobal = wx.BoxSizer(wx.VERTICAL)
        
        # sizer for right tab
        # tab 1, Heat recovery from electrical equipment

        sizerTop = wx.StaticBoxSizer(self.frame_equipment_data, wx.VERTICAL)
        sizerTop.Add(self.tc101, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc102, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc103, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc104, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc105, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc106, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc107, 0, flagText, VSEP_RIGHT)
        sizerTop.Add(self.tc108, 0, flagText, VSEP_RIGHT)
        
        sizerBottom = wx.StaticBoxSizer(self.frame_schedule, wx.VERTICAL)
        sizerBottom.Add(self.tc110, 0, flagText, VSEP_RIGHT)
        sizerBottom.Add(self.tc111, 0, flagText, VSEP_RIGHT)
        sizerBottom.Add(self.tc112, 0, flagText, VSEP_RIGHT)
        sizerBottom.Add(self.tc113, 0, flagText, VSEP_RIGHT)

        sizerGlobal.Add(sizerTop, 8, wx.ALIGN_TOP, 0)
        sizerGlobal.Add(sizerBottom, 4, wx.ALIGN_TOP, 0)

        sizerOKCancel = wx.BoxSizer(wx.HORIZONTAL)
        sizerOKCancel.Add(self.activeObject, 0, wx.ALL, 2)
        sizerOKCancel.Add(self.buttonCancel, 0, wx.ALL, 2)
        sizerOKCancel.Add(self.buttonOK, 0, wx.ALL, 2)

        sizerGlobal.Add(sizerOKCancel, 0, wx.TOP | wx.ALIGN_RIGHT, 0)
        self.SetSizer(sizerGlobal)
        self.Layout()

#------------------------------------------------------------------------------
# GUI actions - 0. OK/Cancel buttons
#------------------------------------------------------------------------------		
    def editAllowed(self):
        """Return True if the data may be modified in the current state, 
        inform the user and return False if not.
        
        """
        if Status.ANo <= 0 and Status.StatusCC > 0:
            logger.showWarning(_U("You cannot modify the heat exchangers after the data has been checked.\n") + 
                        _U("If you want to change the present state you may unblock the data by ") + 
                        _U("choosing to confirm the deletion ") + 
                        _U("of any alternative proposals in this project the next time you switch to the present state."))
            return False
        else:   
            return True

    def OnButtonCancel(self, event):
        self.clear()
        self.prnt.Close()
            
    def OnButtonOK(self, event):
        if not self.editAllowed():
            return
        wheeName = toUTF8(self.tc101.GetValue())
        # assure that a name has been entered before continuing
        if len(wheeName) == 0 or wheeName is None:
            logger.showWarning(_("You have to enter a name for the new equipment before saving"))
            return

        mediumId = self.tc105.GetValue()
        if mediumId is None:
            logger.showWarning(_("Selection of a fluid is mandatory"))
            return            

        fluidDict = Status.prj.getFluidDict()
        medium = check(findKey(fluidDict, self.tc105.GetValue(text=True)))
        
        whee = Status.prj.getWHEE(PId=Status.PId, ANo=Status.ANo, name=dbio.check(wheeName))
        if whee is None:
            whee = Status.prj.addWHEEDummy()
            self.WHEEName = wheeName
                        
        tmp = {
            "WHEEName":check(self.tc101.GetValue()),
            "WHEEEqType":check(findKey(TRANSWHEEEQTYPES, self.tc102.GetValue(text=True))),
            "WHEEWasteHeatType":check(findKey(TRANSWHEEWASTEHEATTYPES, self.tc103.GetValue(text=True))),
            "QWHEE":check(self.tc104.GetValue()),
            "WHEEMedium": str(medium),
            "WHEEFlow":check(self.tc106.GetValue()),
            "WHEETOutlet":check(self.tc107.GetValue()),
            
            "WHEEPresentUse":check(findKey(TRANSYESNO, self.tc108.GetValue(text=True))),
            
            "HPerDayWHEE":check(self.tc110.GetValue()),
            "NBatchWHEE":check(self.tc111.GetValue()),
            "HBatchWHEE":check(self.tc112.GetValue()),
            "NDaysWHEE":check(self.tc113.GetValue()),
            }
        
        whee.set(**tmp)
        
        dbio.commit()
        Status.int.changeInCascade(0)

        Status.prj.projectChanged()
        
        
        self.fillPage()
        self.ShowWHEE()

        if Status.ANo <> 0: # we modified an "real" alternative proposal
            n = whee.WHEENo-1
            conflicts = Status.mod.moduleCC.wheeCheck(n)
            Status.mod.moduleCC.initPanel()
            if conflicts > 0:
                self.cf = conflictFrame(self)
                self.cf.Show()
            else:
                self.display()

        self.prnt.Close()

    def fillPage(self):
        """Screens the SQL tables and fills the lists of HX's and WHEE's"""
        
        ### select WHEE here (self.WHEEName)

        fluidDict = Status.prj.getFluidDict()
        self.tc105.SetValue(fluidDict.values())

    def ShowWHEE(self):
        """Show the data of the currently selected heat exchanger."""

        if self.whee is not None:
            q = self.whee
            self.WHEEID = q.QWasteHeatElEquip_ID
        
            self.tc101.SetValue(q.WHEEName)
            
            if str(q.WHEEEqType) in TRANSWHEEEQTYPES.keys():
                self.tc102.SetValue(TRANSWHEEEQTYPES[q.WHEEEqType])
                
            if str(q.WHEEWasteHeatType) in TRANSWHEEWASTEHEATTYPES.keys():
                self.tc103.SetValue(TRANSWHEEWASTEHEATTYPES[str(q.WHEEWasteHeatType)])
                
            self.tc104.SetValue(str(q.QWHEE))
            
            if q.WHEEMedium is not None:
                WHEEMediumID = int(q.WHEEMedium)
            else:
                WHEEMediumID = None
            
            fluidDict = Status.prj.getFluidDict(includeZeroWater=True)
            if WHEEMediumID in fluidDict.keys():
                fluidName = fluidDict[WHEEMediumID]
                self.tc105.SetValue(fluidName)
            else:
                self.tc105.SetValue("None")
                
            self.tc106.SetValue(str(q.WHEEFlow))
            self.tc107.SetValue(str(q.WHEETOutlet))
            
            if str(q.WHEEPresentUse) in TRANSYESNO.keys():
                self.tc108.SetValue(TRANSYESNO[str(q.WHEEPresentUse)])

            self.tc110.SetValue(str(q.HPerDayWHEE))
            self.tc111.SetValue(str(q.NBatchWHEE))
            self.tc112.SetValue(str(q.HBatchWHEE))
            self.tc113.SetValue(str(q.NDaysWHEE))

            n = q.WHEENo - 1
            try:
                checkWHEE = Status.mod.moduleCC.ccWHEE[n]
            except:
##                logWarning("No check object available for WHEE %s"%n)
                checkWHEE = None

            if checkWHEE is not None:
                ccPars = PanelCCPars({"QWHEE":(checkWHEE.QWHEE, self.tc104),
                                      "WHEEFlow":(checkWHEE.WHEEFlow, self.tc106),
                                      "WHEETOutlet":(checkWHEE.WHEETOutlet, self.tc107),
                                      "HPerDayWHEE":(checkWHEE.HPerDayWHEE, self.tc110),
                                      "NBatchWHEE":(checkWHEE.NBatchWHEE, self.tc111),
                                      "HBatchWHEE":(checkWHEE.HBatchWHEE, self.tc112),
                                      "NDaysWHEE":(checkWHEE.NDaysWHEE, self.tc113)
                                     },
                                     dataGroup = ("WHEE[" + str(n+1) + "]"))
                ccPars.displayCC()
        
    def OnFluidChange(self, event):
        """activated if some fluid selection is changed.
            recalls the updateUnitDict-function for the combined-unit-entries"""

        fluidDict = Status.prj.getFluidDict()
        fluidID = findKey(fluidDict, self.tc105.GetValue(text=True))
        if fluidID is not None:
            self.WHEEFluid = Fluid(fluidID)
            old_funit2 = self.tc106.funit2
            self.tc106.updateUnitDict(self.WHEEFluid.rho)
            if self.tc106.funit2 != old_funit2:
                iValue = self.tc106.GetValue() #takes over display value into internal value
                self.tc106.SetValue(iValue)

    def display(self, n=None):

        if n is None or n<0:
            self.whee = None
            self.WHEEName = None
            self.activeObject.SetLabel(U("--- --- ---"))
            Globals.GUI.editWindow.SetTitle(_U("no equipment selected"))
        else:
            self.whee = Status.int.systemModel.whee[n]
            self.WHEEName = self.whee.WHEEName
            self.activeObject.SetLabel(U(self.WHEEName))
            Globals.GUI.editWindow.SetTitle(U(self.WHEEName))

        self.clear()
        self.fillPage()
        self.ShowWHEE()

        self.Show()

    def clear(self):
        self.tc101.SetValue('')
        self.tc104.SetValue('')
        self.tc106.SetValue('')
        self.tc107.SetValue('')

        self.tc110.SetValue('')
        self.tc111.SetValue('')
        self.tc112.SetValue('')
        self.tc113.SetValue('')

    def goTo(self, key):
        """ tries to open the page appropriate
        for a given parameter

        :params no: number of item in the list to be selected (from 0 to n-1)
        :params key: shortname of variable to be displayes
        """
        ### select WHEE 
        # then open the right page in the panel
        parList = {
            "WHEEName" : 0,
            "WHEEEqType" : 0,
            "WHEEWasteHeatType" : 0,
            "QWHEE" : 0,
            "WHEEMedium" : 0,
            "WHEEFlow" : 0,
            "WHEETOutlet" : 0,
            "WHEEPresentUse" : 0,
            "HPerDayWHEE" : 0,
            "NBatchWHEE" : 0,
            "HBatchWHEE" : 0,
            "NDaysWHEE" : 0,
        }

        if key in parList:
            panelNo = parList[key]
            return True
        else:
            return False


