# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQ8: Building data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import wx
from modules.buildings import Building, BuildingNotFoundError
from GUI.status import *
from constants import AIR_HANDLING_UNIT_STATE, TRANS_AIR_HANDLING_UNIT_STATE, findKey, KELVIN
from auxiliary.auxiliary import toGettextAndUnicode as _U
from auxiliary.auxiliary import toUnicode as U, toUTF8
from GUI.fonts import FontProperties
from GUI.displayClasses import (FieldSizes, TextEntry, FloatEntry,
                                         DateEntry, ChoiceEntry)
from auxiliary.auxiliary import noneCheck
from GUI.GUITools import check
from modules.messageLogger import *
from GUI.conflictFrame import conflictFrame
import database.dbio as dbio

# constants that control the default sizes
# 1. font sizes
TYPE_SIZE_LEFT = 9
TYPE_SIZE_MIDDLE = 9
TYPE_SIZE_RIGHT = 9
TYPE_SIZE_TITLES = 10

# 2. field sizes
HEIGHT = 32
HEIGHT_RIGHT = 32

LABEL_WIDTH_LEFT = 300
LABEL_WIDTH_RIGHT = 200
WIDE_LABEL_WIDTH = 400

DATA_ENTRY_WIDTH_RIGHT = 100
DATA_ENTRY_WIDTH_LEFT = 100

UNITS_WIDTH = 90

# 3. vertical separation between fields
VSEP_LEFT = 2
VSEP_RIGHT = 2

ORANGE = '#FF6000'
TITLE_COLOR = ORANGE


class PanelQ8(wx.Panel):
    def __init__(self, parent, main):
        self.main = main
        self.prnt = parent
        
        self.allowListBoxEventHandler = True

        self._init_ctrls(parent)
        self.__do_layout()

        self.buildingID = None
        self.buildingName = "" # the currently selected building

        self.airDensity = 1.012 #default used before EINSTEIN V2.1.94

    def _init_ctrls(self, parent):
        """UI setup"""
        wx.Panel.__init__(self, id= -1, name='PanelQ8', parent=parent,
              pos=wx.Point(0, 0), size=wx.Size(780, 580), style=0)
        self.Hide()

        # access to font properties object
        fp = FontProperties()

        self.notebook = wx.Notebook(self, -1, style=0)
        self.notebook.SetFont(fp.getFont())

        self.page0 = wx.Panel(self.notebook)
        self.notebook.AddPage(self.page0, _U('General data'))
        
        self.page1 = wx.Panel(self.notebook)
        self.notebook.AddPage(self.page1, _U('Energy demand'))

        self.page2 = wx.Panel(self.notebook)
        self.notebook.AddPage(self.page2, _U('Occupation'))
        
        self.frame_building_list = wx.StaticBox(self.page0, -1, _U("Building list"))
        self.frame_building_list.SetForegroundColour(TITLE_COLOR)
        self.frame_building_list.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        #self.frame_building = wx.StaticBox(self.page0, -1, _U("Building (or part of building)"))
        self.frame_general_data = wx.StaticBox(self.page0, -1, _U("General data"))
        self.frame_general_data.SetForegroundColour(TITLE_COLOR)
        self.frame_general_data.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        self.frame_air_handling = wx.StaticBox(self.page1, -1, _U("Air Handling"))
        self.frame_air_handling.SetForegroundColour(TITLE_COLOR)
        self.frame_air_handling.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))
        
        self.frame_HandC_demand = wx.StaticBox(self.page1, -1, _U("Heating and Cooling"))
        self.frame_HandC_demand.SetForegroundColour(TITLE_COLOR)
        self.frame_HandC_demand.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))
        
        self.frame_hotwater_demand = wx.StaticBox(self.page1, -1, _U("Domestic Hot Water"))
        self.frame_hotwater_demand.SetForegroundColour(TITLE_COLOR)
        self.frame_hotwater_demand.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        self.frame_occupation = wx.StaticBox(self.page2, -1, _U("Period of occupation"))
        self.frame_occupation.SetForegroundColour(TITLE_COLOR)
        self.frame_occupation.SetFont(wx.Font(8, wx.SWISS, wx.NORMAL, wx.BOLD))

        # set font for titles
        # 1. save actual font parameters on the stack
        fp.pushFont()
        # 2. change size and weight
        fp.changeFont(size=TYPE_SIZE_TITLES, weight=wx.BOLD)
        self.frame_building_list.SetFont(fp.getFont())
        #self.frame_building.SetFont(fp.getFont())
        for frame in (self.frame_general_data, self.frame_occupation,
                      self.frame_air_handling, self.frame_HandC_demand):
            frame.SetFont(fp.getFont())
        # 3. recover previous font state
        fp.popFont()


        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH_LEFT,
                       wData=DATA_ENTRY_WIDTH_LEFT, wUnits=UNITS_WIDTH)

        # left tab controls
        # tab 0 - general data
        #
        # tab 0 left side. building list
        self.listBoxBuildingList = wx.ListBox(self.page0, -1, choices=[])
        self.Bind(wx.EVT_LISTBOX, self.OnListBoxBuildingListClick, self.listBoxBuildingList)

        # tab 0 right side. data entries
        # tab 0 top staticbox: General data
        self.tc1 = TextEntry(self.page0, maxchars=45, value='',
                             label=_U("Building short name"),
                             tip=_U("Give some brief name of the buildings to identify them in the reports"))
        self.tc2 = FloatEntry(self.page0, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='AREA',
                              label=_U("Constructed surface"),
                              tip=_U("Surface limited by building's perimeter multiplied by number of floors"))
        self.tc3 = FloatEntry(self.page0, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='AREA',
                              label=_U("Useful surface"),
                              tip=_U("Total useful surface of building (excluding walls)"))
        self.tc4 = TextEntry(self.page0, maxchars=255, value='',
                             label=_U("Use of the building"),
                             tip=_U("Specify use, e.g. offices, production, storage,..."))
        
        FieldSizes(wLabel=WIDE_LABEL_WIDTH)
        # tab 1 Energy demands
        # first the airHandlingUnit
        self.AirHandling = ChoiceEntry(self.page1,
                                       values=[],
                                       label=_U("Air Handling"),
                                       tip=_U("Select from predefined list"),
                                       allowNones=False)
        self.Bind(wx.EVT_CHOICE, self.OnAHUSelect, self.AirHandling.entry)
        
        self.airFlowRateH = FloatEntry(self.page1, decimals=3, minval=0., 
                                        maxval=1.e+12, value=None,
                                        unitdict='VOLUMEORMASSFLOW',
                                        label=_U("Inflow of air through the AHU while heating"),
                                        tip=_U("Air exchange rate that the air handling unit achieves."))
        self.airFlowRateC = FloatEntry(self.page1, decimals=3, minval=0., 
                                        maxval=1.e+12, value=None,
                                        unitdict='VOLUMEORMASSFLOW',
                                        label=_U("Inflow of air through the AHU while cooling"),
                                        tip=_U("Air exchange rate that the air handling unit achieves."))
        self.TInFlowRecC = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Temperature of air coming from the AHU while cooling the building"),
                              tip=_U(" "))
        self.TInFlowRecH = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Temperature of air coming from the AHU while heating the building"),
                              tip=_U(" "))
        # heating and cooling demands
        FieldSizes(wLabel=LABEL_WIDTH_LEFT)
        self.heatingHeader = wx.StaticText(self.page1, label=_U("Heating"))
        self.tc5 = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='POWER',
                              label=_U("Maximum power"),
                              tip=_U("Maximum heating power (without including the security coefficient of the equipment)"),
                              isStatic=True)
        self.tc7 = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='ENERGY',
                              label=_U("Annual demand"),
                              tip=_U("Annual demand of heating"))
        self.tc8_1 = FloatEntry(self.page1, decimals=1, minval=-KELVIN, maxval=999., value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Required temperature for system"),
                              tip=_U("Specify the required forward temperature of the heating medium (hot water, air) to the heating system (radiators, radiant floor, fan coils, etc.)"))
        self.indoorTempHeating = FloatEntry(self.page1, decimals=1, minval=-KELVIN, maxval=999., value=0.,
                                            unitdict='TEMPERATURE',
                                            label=_U("Target temperature for the building"),
                                            tip=_U("Specify the minimal desired indoor temperature of the building."))
        FieldSizes(wLabel=5)
        self.coolingHeader= wx.StaticText(self.page1, label=_U("Cooling"))
        self.tc6 = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='POWER',
                              tip=_U("The maximum cooling power is a result of the consistency check."), 
                              isStatic=True)
        self.tc8 = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='ENERGY',
                              tip=_U("Annual demand of cooling"))
        self.tc8_2 = FloatEntry(self.page1, decimals=1, minval=-KELVIN, maxval=999., value=0.,
                              unitdict='TEMPERATURE',
                              tip=_U("Specify the required forward temperature of the cooling medium (cold water, air) to the air conditioning system (fan coils, radiant ceiling, etc.)"))
        self.indoorTempCooling = FloatEntry(self.page1, decimals=1, minval=-KELVIN, maxval=999., value=0.,
                                            unitdict='TEMPERATURE',
                                            tip=_U("Specify the maximal desired indoor temperature of the building."))
        # domestic hot water
        FieldSizes(wLabel=WIDE_LABEL_WIDTH)
        self.tc9 = FloatEntry(self.page1, decimals=1, minval=0., maxval=1.e+12, value=0.,
                              unitdict='VOLUME',
                              label=_U("Daily consumption of  DHW"),
                              tip=_U("Only consumption of hot water that is not included yet in 'Processes'"))

        # right tab controls
        # tab 0 - Period of occupation
        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH_RIGHT,
                        wData=DATA_ENTRY_WIDTH_RIGHT, wUnits=UNITS_WIDTH)

        #SD:change type of entry?? -> time start - time stop?? or change tip: hours of occupation per day??
        self.tc10 = FloatEntry(self.page2, decimals=1, minval=0., maxval=8760, value=0.,
                               unitdict=None,
                               label=_U("Hours of occupation"),
                               tip=_U("Hours of occupation of the building, during which heating and air conditionning is active"))
        self.tc11 = FloatEntry(self.page2, decimals=1, minval=0., maxval=365., value=0.,
                               unitdict=None,
                               label=_U("Days of use per year"),
                               tip=_U(" "))
        self.tc12_10 = DateEntry(self.page2, value='',
                              label=_U("Start holiday period 1"),
                              tip=_U("Period of year that the building is not used"))
        self.tc12_11 = DateEntry(self.page2, value='',
                              label=_U("Stop holiday period 1"),
                              tip=_U("Period of year that the building is not used"))
        self.tc12_20 = DateEntry(self.page2, value='',
                              label=_U("Start holiday period 2"),
                              tip=_U("Period of year that the building is not used"))
        self.tc12_21 = DateEntry(self.page2, value='',
                              label=_U("Stop holiday period 2"),
                              tip=_U("Period of year that the building is not used"))
        self.tc12_30 = DateEntry(self.page2, value='',
                              label=_U("Start holiday period 3"),
                              tip=_U("Period of year that the building is not used"))
        self.tc12_31 = DateEntry(self.page2, value='',
                              label=_U("Stop holiday period 3"),
                              tip=_U("Period of year that the building is not used"))

        self.tc13_1 = DateEntry(self.page2, value='',
                              label=_U("Start of heating period"),
                              tip=_U(" "))
        self.tc13_2 = DateEntry(self.page2, value='',
                              label=_U("Stop of heating period"),
                              tip=_U(" "))
        self.tc14_1 = DateEntry(self.page2, value='',
                              label=_U("Start of air\nconditioning period"),
                              tip=_U(" "))
        self.tc14_2 = DateEntry(self.page2, value='',
                              label=_U("Stop of air\nconditioning period"),
                              tip=_U(" "))

        # buttons
        self.buttonDeleteBuilding = wx.Button(self.page0, -1, _U("Delete building"))
        self.Bind(wx.EVT_BUTTON, self.OnButtonDeleteBuilding, self.buttonDeleteBuilding)
        self.buttonDeleteBuilding.SetMinSize((136, 32))
        self.buttonDeleteBuilding.SetFont(fp.getFont())
        
        self.buttonAddBuilding = wx.Button(self.page0, -1, _U("New building"))
        self.Bind(wx.EVT_BUTTON, self.OnButtonAddBuilding, self.buttonAddBuilding)
        self.buttonAddBuilding.SetMinSize((136, 32))
        self.buttonAddBuilding.SetFont(fp.getFont())
        
        self.buttonOK = wx.Button(self, wx.ID_OK, label=_U("OK"))
        self.Bind(wx.EVT_BUTTON, self.OnButtonOK, self.buttonOK)
        self.buttonOK.SetDefault()

        self.activeObject = wx.StaticText(self,id=-1,label="----- ----- -----",style = wx.ALIGN_LEFT)
        self.activeObject.SetMinSize((580, 24)) #a hack in order to achieve alignment at the left. can surely be done better.

        self.buttonCancel = wx.Button(self, wx.ID_CANCEL, label=_U("Cancel"))
        self.Bind(wx.EVT_BUTTON, self.OnButtonOK, self.buttonOK)

        # recover previous font parameters from the stack
        fp.popFont()
        
    def __do_layout(self):
        flagText = wx.ALIGN_CENTER_VERTICAL | wx.TOP

        # global sizer for panel. Contains notebook w/two tabs + buttons Cancel and Ok
        sizerGlobal = wx.BoxSizer(wx.VERTICAL)
        
        # sizer for left tab
        # tab 0, general data
        sizerPage0 = wx.BoxSizer(wx.HORIZONTAL)
        # left part: listbox
        sizerP0Left = wx.StaticBoxSizer(self.frame_building_list, wx.VERTICAL)
        sizerP0Left.Add(self.listBoxBuildingList, 1, wx.EXPAND, 0)
        sizerP0Left.Add(self.buttonAddBuilding, 0, wx.ALIGN_RIGHT, 0)
        sizerP0Left.Add(self.buttonDeleteBuilding, 0, wx.ALIGN_RIGHT, 0)
        sizerPage0.Add(sizerP0Left, 2, wx.EXPAND | wx.TOP, 10)

        # right part: data entries
        sizerP0Right = wx.BoxSizer(wx.VERTICAL)
        sizerP0RightTop = wx.StaticBoxSizer(self.frame_general_data, wx.VERTICAL)
        sizerP0RightTop.Add(self.tc1, 0, flagText, VSEP_LEFT)
        sizerP0RightTop.Add(self.tc2, 0, flagText, VSEP_LEFT)
        sizerP0RightTop.Add(self.tc3, 0, flagText, VSEP_LEFT)
        sizerP0RightTop.Add(self.tc4, 0, flagText, VSEP_LEFT)
        sizerP0Right.Add(sizerP0RightTop, 4, wx.EXPAND, 0)
        sizerPage0.Add(sizerP0Right, 5, wx.EXPAND | wx.TOP, 10)
        self.page0.SetSizer(sizerPage0)

        # tab 1, energy demands
        sizerPage1 = wx.BoxSizer(wx.VERTICAL)
        sizerAirHandling = wx.StaticBoxSizer(self.frame_air_handling, wx.VERTICAL)
        sizerPage1HandC = wx.StaticBoxSizer(self.frame_HandC_demand, wx.HORIZONTAL)
        sizerHeating = wx.BoxSizer(wx.VERTICAL)
        sizerCooling = wx.BoxSizer(wx.VERTICAL)
        sizerPage1HandC.Add(sizerHeating, 0, flagText, VSEP_LEFT)
        sizerPage1HandC.Add(sizerCooling, 0, flagText, VSEP_LEFT)
        framedSizerDHW = wx.StaticBoxSizer(self.frame_hotwater_demand, wx.VERTICAL)
        sizerPage1.Add(sizerAirHandling, 0, flagText, VSEP_LEFT)
        sizerPage1.Add(sizerPage1HandC, 0, flagText, VSEP_LEFT)
        sizerPage1.Add(framedSizerDHW, 0, flagText, VSEP_LEFT)
        
        sizerAirHandling.Add(self.AirHandling)
        sizerAirHandling.Add(self.airFlowRateH)
        sizerAirHandling.Add(self.airFlowRateC)
        sizerAirHandling.Add(self.TInFlowRecC)
        sizerAirHandling.Add(self.TInFlowRecH)
        
        sizerHeating.Add(self.heatingHeader, 0, wx.LEFT, LABEL_WIDTH_LEFT + 5)
        sizerHeating.Add(self.tc5, 0, flagText, VSEP_LEFT)
        sizerHeating.Add(self.tc7, 0, flagText, VSEP_LEFT)
        sizerHeating.Add(self.tc8_1, 0, flagText, VSEP_LEFT)
        sizerHeating.Add(self.indoorTempHeating, 0, flagText, VSEP_LEFT)
        
        sizerCooling.Add(self.coolingHeader, 0, wx.LEFT, 5 + 5)
        sizerCooling.Add(self.tc6, 0, flagText, VSEP_LEFT)
        sizerCooling.Add(self.tc8, 0, flagText, VSEP_LEFT)
        sizerCooling.Add(self.tc8_2, 0, flagText, VSEP_LEFT)
        sizerCooling.Add(self.indoorTempCooling, 0, flagText, VSEP_LEFT)
        
        framedSizerDHW.Add(self.tc9, 0, flagText, VSEP_LEFT)
        
        self.page1.SetSizer(sizerPage1)
        
        
        # sizer for right tab
        # tab 2, occupation
        sizerPage2 = wx.StaticBoxSizer(self.frame_occupation, wx.VERTICAL)
        grid_sizer_P2 = wx.FlexGridSizer(6, 2, 1, 2) #r,c,vsep,hsep
        grid_sizer_P2.Add(self.tc10, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc11, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_10, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_11, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_20, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_21, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_30, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc12_31, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc13_1, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc13_2, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc14_1, 0, flagText, 0)
        grid_sizer_P2.Add(self.tc14_2, 0, flagText, 0)
        sizerPage2.Add(grid_sizer_P2, 0, wx.EXPAND | wx.TOP, 10)

        self.page2.SetSizer(sizerPage2)

        sizerOKCancel = wx.BoxSizer(wx.HORIZONTAL)
        sizerOKCancel.Add(self.activeObject, 0, wx.ALL | wx.EXPAND, 2)
        sizerOKCancel.Add(self.buttonCancel, 0, wx.ALL | wx.EXPAND, 2)
        sizerOKCancel.Add(self.buttonOK, 0, wx.ALL | wx.EXPAND, 2)

        sizerGlobal.Add(self.notebook, 3, wx.EXPAND, 0)
        sizerGlobal.Add(sizerOKCancel, 0, wx.TOP | wx.ALIGN_RIGHT, 0)
        self.SetSizer(sizerGlobal)
        self.Layout()

    #--- UI actions
    def OnListBoxBuildingListClick(self, event):

        if not self.allowListBoxEventHandler: return

        self.buildingName = toUTF8(self.listBoxBuildingList.GetStringSelection())
        if self.buildingName is not None: self.activeObject.SetLabel(U(self.buildingName))
        
        q = Building(self.buildingName)
        
        try:
            q.loadFromDB()
        except BuildingNotFoundError:
            return # this happens for every Clear() call on the building list
        
        self.buildingID = q.dbId
        self.tc1.SetValue(q.name)
        self.tc2.SetValue(q.buildConstructSurface)
        self.tc3.SetValue(q.buildUsefulSurface)
        self.tc4.SetValue(q.buildUsage)
        self.tc5.SetValue(q.buildMaxHP)
        self.tc6.SetValue(q.buildMaxCP)
        self.tc7.SetValue(q.buildAnnualHeating)
        self.tc8.SetValue(q.buildAnnualAirCond)
        self.tc8_1.SetValue(q.buildTHeating)
        self.tc8_2.SetValue(q.buildTAirCond)
        self.tc9.SetValue(q.buildDailyDHW)
        self.tc10.SetValue(q.buildHoursOccup)
        self.tc11.SetValue(q.buildDaysInUse)
        self.tc12_10.SetValue(q.buildHolidaysPeriodStart_1)
        self.tc12_11.SetValue(q.buildHolidaysPeriodStop_1)
        self.tc12_20.SetValue(q.buildHolidaysPeriodStart_2)
        self.tc12_21.SetValue(q.buildHolidaysPeriodStop_2)
        self.tc12_30.SetValue(q.buildHolidaysPeriodStart_3)
        self.tc12_31.SetValue(q.buildHolidaysPeriodStop_3)
        self.tc13_1.SetValue(q.buildHeatingPeriodStart)
        self.tc13_2.SetValue(q.buildHeatingPeriodStop)
        self.tc14_1.SetValue(q.buildAirCondPeriodStart)
        self.tc14_2.SetValue(q.buildAirCondPeriodStop)
        self.AirHandling.SetValue(noneCheck(q.airHandlingUnit, 0))
        self.TInFlowRecH.SetValue(q.TInflowRecH)
        self.TInFlowRecC.SetValue(q.TInflowRecC)

##        n = q.airHandlingUnit
##        if n in range(len(AIR_HANDLING_UNIT_STATE)):
##            ahu = AIR_HANDLING_UNIT_STATE[n]
##            
##            self.AirHandling.SetValue(TRANS_AIR_HANDLING_UNIT_STATE[ahu])
##        else:
##            self.AirHandling.SetValue(None) # in case of error, set some default
##            logTrack("PanelQ8: erroneous value for air handling unit in building with id %s"%self.buildingID)

        if q.fluidAir is not None and q.fluidAir.rho > 0:
            self.airDensity = q.fluidAir.rho
            
        self.airFlowRateH.updateUnitDict(1./self.airDensity)
        if q.airFlowRateH is not None:
            self.airFlowRateH.SetValue(q.airFlowRateH/self.airDensity)
        else:
            self.airFlowRateH.SetValue(None)
            
        self.airFlowRateC.updateUnitDict(1./self.airDensity)
        if q.airFlowRateC is not None:
            self.airFlowRateC.SetValue(q.airFlowRateC/self.airDensity)
        else:
            self.airFlowRateC.SetValue(None)
            
        self.indoorTempHeating.SetValue(q.TIndoorH)
        self.indoorTempCooling.SetValue(q.TIndoorC)
        
    def OnAHUSelect(self, event):
        allAHUwidgets = set((self.airFlowRateC, self.airFlowRateH, 
                             self.TInFlowRecC, self.TInFlowRecH))
        sensibleWidgets = set()
        
        ahuStr = toUTF8(event.EventObject.GetStringSelection())
        key = findKey(TRANS_AIR_HANDLING_UNIT_STATE, ahuStr)
        
        if key == "no AHU":
            sensibleWidgets = set()
        elif key == "only air inlet":
            sensibleWidgets.update(allAHUwidgets)
        elif key == "in- and outlet":
            sensibleWidgets.update(allAHUwidgets)
        else:
            showError("no valid AHU %r"%ahuStr)
##            raise ValueError(event.GetString())

        if self.AirHandling.IsEnabled():
            for widget in sensibleWidgets:
                widget.Enable()
            for widget in allAHUwidgets - sensibleWidgets:
                widget.Disable()

    def OnButtonAddBuilding(self, event):
        if not self.editAllowed():
            return
        self.clear()
        if event: event.Skip()

    def OnButtonCancel(self, event):
        self.clear()
        if event: event.Skip()

    def OnButtonDeleteBuilding(self, event):
        if not self.editAllowed():
            return
        Status.prj.deleteBuilding(self.buildingID)
        self.clear()
        self.fillPage()
        if event: event.Skip()
   
    def editAllowed(self):
        """Return True if the data may be modified in the current state, 
        inform the user and return False if not.
        
        """
        if Status.ANo <= 0 and Status.StatusCC > 0:
            showWarning(_U("You cannot modify buildings after the data has been checked.\n")+
                        _U("If you want to change the present state you may unblock the data by ")+
                        _U("choosing to confirm the deletion ")+ 
                        _U("of any alternative proposals in this project the next time you switch to the present state."))
            return False
        else:   
            return True
   
    def OnButtonOK(self, event):
        """Check entries and if they are fine add create a new building or
        update an existing with the same name."""
        if not self.editAllowed():
            return

        if Status.PId != 0:
            buildingName = self.tc1.GetValue()

            attachedprocessnames = [buildingName + "_" + name for name in [("heating"), ("cooling"), ("HW")]]
            # assure that a name has been entered before continuing
            if len(buildingName) == 0 or buildingName is None:
                showWarning(_("You have to enter a name for the new building before saving"))
                return

            buildings = dbio.select("qbuildings", BuildName=dbio.check(buildingName), \
                        Questionnaire_id=Status.PId, \
                        AlternativeProposalNo=Status.ANo)

            if len(buildings) in (0, 1):
                if len(buildings) == 0:
                    # check whether processes exist that occupy the standard names for the house processes. 
                    for processname in attachedprocessnames:
                        q = dbio.select("qprocessdata", Questionnaire_id=Status.PId, \
                             AlternativeProposalNo=Status.ANo, Process=processname)
                        if len(q) > 0:
                            showError("There is already a process named %s. Please either rename this building or the process."
                                       % (processname,))
                            return # Break out from saving, it's not possible.
                    building = Building(buildingName) # create new building
                elif len(buildings) == 1:
                    building = Building(buildingName) # recreate existing building
                    building.loadFromDB()
                    
                self.buildingName = buildingName # select the new one
                lateWarnings = []
                
##                if Status.ANo > 0: # later changes to buildings
##                    # don't allow a qualitative change, zero stays zero and
##                    # non zero stays non zero for certain entries
##                    for entry, oldValue, label in ((self.tc7, building.buildAnnualHeating, self.tc7.label.GetLabel()),
##                                                   (self.tc8, building.buildAnnualAirCond, self.tc7.label.GetLabel()),
##                                                   (self.tc9, building.buildDailyDHW, self.tc9.label.GetLabel())):
##                        if (oldValue != 0) and (entry.GetValue() == 0):
##                                print oldValue, entry.GetValue()
##                                entry.SetValue(oldValue)
##                                lateWarnings.append(_U("Sorry, setting \"%s\" down to 0 is not supported in alternative proposals.") % label)
##                        elif (oldValue == 0) and (entry.GetValue() != 0):
##                                print oldValue, entry.GetValue()
##                                print type(oldValue), type(entry.GetValue())
##                                entry.SetValue(0)
##                                lateWarnings.append(_U("Sorry, changing \"%s\" from 0 is not supported in alternative proposals.") % label)
                                
                # validation of the input is done in the Building object
                building.buildConstructSurface = self.tc2.GetValue()
                building.buildUsefulSurface = self.tc3.GetValue()
                building.buildUsage = self.tc4.GetValue()
                building.buildMaxHP = self.tc5.GetValue()
                building.buildMaxCP = self.tc6.GetValue()

                ahu = self.AirHandling.GetValue()
                building.airHandlingUnit = noneCheck(ahu, 0)
##                key = findKey(TRANS_AIR_HANDLING_UNIT_STATE, self.AirHandling.GetValue(text=True))
##                if key is not None:
##                    ahu = AIR_HANDLING_UNIT_STATE.index(key)
##                    building.airHandlingUnit = ahu
##                else:
##                    logError("PanelQ8 (buildings): no valid air handling unit type %s"%check(self.AirHandling.GetValue()))
##                    building.airHandlingUnit = 0
                
                if self.airFlowRateH.IsEnabled():
                    try:
                        building.airFlowRateH = self.airFlowRateH.GetValue()*self.airDensity
                    except:
                        building.airFlowRateH = None

                    try:    
                        building.airFlowRateC = self.airFlowRateC.GetValue()*self.airDensity
                    except:
                        building.airFlowRateC = None
                else:
                    building.airFlowRateH = None
                    building.airFlowRateC = None
                if self.TInFlowRecH.IsEnabled():
                    building.TInflowRecH = self.TInFlowRecH.GetValue()
                    building.TInflowRecC = self.TInFlowRecC.GetValue()
                else:
                    building.TInflowRecH = None
                    building.TInflowRecC = None
                building.buildAnnualHeating = self.tc7.GetValue()
                building.buildAnnualAirCond = self.tc8.GetValue()
                building.buildTHeating = self.tc8_1.GetValue()
                building.buildTAirCond = self.tc8_2.GetValue()
                building.TIndoorH = self.indoorTempHeating.GetValue()
                building.TIndoorC = self.indoorTempCooling.GetValue()
                building.buildDailyDHW = self.tc9.GetValue()
                building.buildHoursOccup = self.tc10.GetValue()
                building.buildDaysInUse = self.tc11.GetValue()
                building.buildHolidaysPeriodStart_1 = self.tc12_10.GetValue()
                building.buildHolidaysPeriodStart_2 = self.tc12_20.GetValue()
                building.buildHolidaysPeriodStart_3 = self.tc12_30.GetValue()
                building.buildHolidaysPeriodStop_1 = self.tc12_11.GetValue()
                building.buildHolidaysPeriodStop_2 = self.tc12_21.GetValue()
                building.buildHolidaysPeriodStop_3 = self.tc12_31.GetValue()
                building.buildHeatingPeriodStart = self.tc13_1.GetValue()
                building.buildHeatingPeriodStop = self.tc13_2.GetValue()
                building.buildAirCondPeriodStart = self.tc14_1.GetValue()
                building.buildAirCondPeriodStop = self.tc14_2.GetValue()
                building.saveToDB(Status.PId, Status.ANo)

                if Status.ANo > 0:
                    # we modified an "real" alternative proposal
                    for processID in building.processIDs:
                        if processID is not None:
                            k = Status.prj.getProcessNo(processID)-1
                            proc = Status.int.systemModel.processObject[k]
                            conflicts = Status.mod.moduleCC.procCheck(proc)
                            Status.mod.moduleCC.initPanel()
                            if conflicts > 0:
                                self.cf = conflictFrame(self)
                                self.cf.Show()
                                break
                            else:
                                proc.saveToDB()
                        
                for warning in lateWarnings:
                    logWarning(warning)
                self.fillBuildingList()

                Status.processData.changeInProcess()
                Status.prj.projectChanged()

                self.prnt.Close()

            else:
                self.showError(_U("BuildingName has to be an unique value!"))


    # Public methods
    def display(self, data=None):
        self.clear()
        self.fillPage()
        self.activateAppropriateEntries()
        self.Show()

    def fillBuildingList(self):
        buildingName = self.buildingName
        self.allowListBoxEventHandler = False
        self.listBoxBuildingList.Clear()
        buildings = Status.prj.getBuildings()
        if len(buildings) > 0:
            for building in buildings:
                self.listBoxBuildingList.Append(U(building.BuildName))

        self.allowListBoxEventHandler = True

        if self.listBoxBuildingList.SetStringSelection(U(buildingName)) is True:
            self.buildingName = buildingName
            self.OnListBoxBuildingListClick(None)
        elif len(buildings) > 0:
            self.listBoxBuildingList.SetSelection(0)
            self.buildingName = toUTF8(self.listBoxBuildingList.GetStringSelection())
            self.OnListBoxBuildingListClick(None)

    def fillPage(self):
        ahuChoices = [TRANS_AIR_HANDLING_UNIT_STATE[key] for key in AIR_HANDLING_UNIT_STATE]
        self.AirHandling.SetValue(ahuChoices)
        
        self.fillBuildingList()

    def clear(self):
        for item in (self.tc1,
                     self.tc2,
                     self.tc3,
                     self.tc4,
                     self.tc5,
                     self.tc6,
                     self.tc7,
                     self.tc8,
                     self.tc8_1,
                     self.tc8_2,
                     self.tc9,
                     self.tc10,
                     self.tc11,
                     self.tc12_10,
                     self.tc12_11,
                     self.tc12_20,
                     self.tc12_21,
                     self.tc12_30,
                     self.tc12_31,
                     self.tc13_1,
                     self.tc13_2,
                     self.tc14_1,
                     self.tc14_2,
                     self.indoorTempCooling,
                     self.indoorTempHeating,
                     self.airFlowRateH,
                     self.airFlowRateC,
                     self.TInFlowRecH,
                     self.TInFlowRecC,):
            item.Clear()
        self.AirHandling.SetValue(0)

    def activateAppropriateEntries(self):
        """Enable and disable the entry fields that are only sensible
        for a certain cases, according to the current alternative.
        
        """
        allEntries = set((self.tc1, self.tc10, self.tc11, self.tc12_10,
                          self.tc12_11, self.tc12_20, self.tc12_21,
                          self.tc12_30, self.tc12_31, self.tc13_1, 
                          self.tc13_2, self.tc14_1, self.tc14_2,
                          self.AirHandling, self.airFlowRateC, 
                          self.airFlowRateH, self.TInFlowRecC, 
                          self.TInFlowRecH))
        sensibleEntries = set()
        if Status.ANo <> 0:
            sensibleEntries.update(allEntries)
        for entry in sensibleEntries:
            entry.Enable()
        for entry in allEntries - sensibleEntries:
            entry.Disable()
