# -*- coding: utf-8 -*-
#==============================================================================
#
#       E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#       PanelDBElectricityMixEntry: Editor for fluid data
#
#==============================================================================
#
#	Version 3.0, 2018-2019
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from GUI.displayClasses import *
from constants import INFINITE

from GUI.einsteinPanel import EinsteinPanel
from GUI.editWindow import EditWindow

from modules.guiModules.DB.moduleDB import ModuleDBEntry

from auxiliary.auxiliary import toGettextAndUnicode as _U

class PanelDBElectricityMixEntry(EinsteinPanel):

    def __init__(self, prnt, main):

        layout = {"notebook": [_('General Information'),
                               _("Data")],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("General Information")),
                             (1, _("Emissions")),
                             (1, _("Sources of Primary Energy"))]}           # list of frames (pageno, title). if empty, no frame is created

        super(PanelDBElectricityMixEntry, self).__init__(prnt, main, layout, module=None, size=(780, 600), key=None)

        self.mod = ModuleDBEntry(prnt.mod, self)

        #
        # frame 0 - General information
        #

        self.addEntry("name",
                      TextEntry(self.nbPage[0], maxchars=45, value='',
                             label=_U("Shortname of this profile"),
                             tip=_U("This name will be used to identify this mix to users throughout the program.")),
                      0)

        self.addEntry("Country",
                      TextEntry(self.nbPage[0], maxchars=45, value='',
                             label=_U("Country or region"),
                             tip=_U("Country or region")),
                      0)

        self.addEntry("Year",
                      FloatEntry(self.nbPage[0],
                              decimals=0, minval=1900, maxval=2100, value=2010,
                              label=_U("Year of data"),
                              tip=_U("Year of data")),
                      0)

        self.addEntry("Type",
                      TextEntry(self.nbPage[0], maxchars=20, value='',
                             label=_U("Type of electricity net (user, source, supplier, etc.)"),
                             tip=_U("Specify e.g. average of national grid, high voltage grid, ...")),
                      0)

        self.addEntry("Reference",
                      TextEntry(self.nbPage[0], maxchars=200, value='',
                             isMultiline=True, specialHeight=100,
                             label=_U("Bibliographic reference / data source"),
                             tip=_U("Please give a full reference of the data source that allows for tracking back the data")),
                      0)

        self.addEntry("AuditorID",
                      TextEntry(self.nbPage[0], value='',
                             isStatic=True,
                             label=_U("Auditor responsible for data set"),
                             tip=_U("Auditor responsible for data set")),
                      0)


        #
        # frame 1 - Environmental parameters
        #

        self.addEntry("PE2ConvEl",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                              unitdict='FRACTION',
                              label=_U("Primary energy per unit of electricity"),
                              tip=_U("Only non-renewable part of primary energy")),
                      1)

        self.addEntry("CO2ConvEl",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                              unitdict='CO2RATIO',
                              label=_U("CO2 generation per unit of electricity"),
                              tip=_U("CO2 generation per unit of electricity")),
                      1)

        self.addEntry("NoNukesConvEl",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                              unitdict='NUKESRATIO',
                              label=_U("Highly radiactive nuclear waste per unit of electricity"),
                              tip=_U("Highly radiactive nuclear waste per unit of electricity")),
                      1)

        #
        # frame 2 - Composition of PE sources
        #

        self.addEntry("PercNaturalGas",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                              unitdict='FRACTION',
                              label=_U("Natural Gas"),
                              tip=_U("Fraction of total generation of electricity")),
                      2)

        self.addEntry("PercCarbon",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("Carbon"),
                               tip=_U("Fraction of total generation of electricity")),
                      2)

        self.addEntry("PercOil",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("Oil"),
                               tip=_U("Fraction of total generation of electricity")),
                      2)

        self.addEntry("PercRenewables",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("Renewable Energy"),
                               tip=_U("Fraction of total generation of electricity")),
                      2)

        self.addEntry("PercNukes",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("Nuclear Energy"),
                               tip=_U("Fraction of total generation of electricity")),
                      2)

        self.addEntry("PercCHP",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("CHP"),
                               tip=_U("Fraction of total generation of electricity (except CHP using renewables)")),
                      2)

        self.addEntry("PercOther",
                      FloatEntry(self.nbPage[1],
                               decimals=4, minval= -INFINITE, maxval=INFINITE, value=0.,
                               unitdict='FRACTION',
                               label=_U("Other"),
                               tip=_U("Fraction of total generation of electricity")),
                      2)


        # conclude layout
        self.conclude_layout()


class DialogDBElectricityMixEntry(EditWindow):
    def __init__(self, prnt, main):
        self.prnt = prnt
        self.mod = prnt.mod
        if self.mod.editor: return  # do not open a second dialog

        super(DialogDBElectricityMixEntry, self).__init__(main)

        self.panel = PanelDBElectricityMixEntry(self, main)
        self.ret = None

        self.panel.display()
        self.Show()

        self.mod.editor = self
        self.prnt.panel.Disable()

    def OnClose(self, event):
        self.mod.editor = None
        self.prnt.panel.display()
        self.prnt.panel.Enable()
        super(DialogDBElectricityMixEntry, self).OnClose(event)
        
