# -*- coding: utf-8 -*-
#==============================================================================
#
#       E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#       PanelDBFuelEntry: Editor for fluid data
#
#==============================================================================
#
#	Version 3.0, 2018-2019
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from GUI.displayClasses import *
from constants import FUELTYPES, TRANSFUELTYPES, INFINITE

from GUI.einsteinPanel import EinsteinPanel
from GUI.editWindow import EditWindow

from modules.guiModules.DB.moduleDB import ModuleDBEntry

from auxiliary.auxiliary import toGettextAndUnicode as _U

class PanelDBFuelEntry(EinsteinPanel):

    def __init__(self, prnt, main):

        layout = {"notebook": [_('General data'),
                               _("Offgas calculations")],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("General information on fuel")),
                             (0, _("Main physical properties")),
                             (1, _("Properties for offgas calculations")),
                             (0, _("Environment parameters"))]}           # list of frames (pageno, title). if empty, no frame is created

        super(PanelDBFuelEntry, self).__init__(prnt, main, layout, module=None, size=(780, 600), key=None)

        self.mod = ModuleDBEntry(prnt.mod, self)

        self.addEntry("name",
                      TextEntry(self.nbPage[0], maxchars = 45, value = '',
                             label = _U("Name of the fuel"),
                             tip = _U("Give a short name for identification of the fuel")),
                      0)

        self.addEntry("FuelType",
                      ChoiceEntry(self.nbPage[0],
                               values = [TRANSFUELTYPES[ftype] for ftype in FUELTYPES],
                               label = _U("Fuel type"),
                               tip = _U(" ")),
                      0)

##        self.addEntry("DBFuelUnit",
##                      ChoiceEntry(self.nbPage[0],
##                               values = [],
##                               label = _U("Default measurement unit"),
##                               tip = _U(" ")),
##                      0)
##
        self.addEntry("FuelDataSource",
                      TextEntry(self.nbPage[0], maxchars = 200, value = '',
                             isMultiline = True, specialHeight=100,
                             label = _U("Source of data"),
                             tip = _U("Please give a full specification of data sources / literature references")),
                      0)

        self.addEntry("FuelComment",
                      TextEntry(self.nbPage[0], maxchars = 200, value = '',
                             isMultiline = True, specialHeight=100,
                             label = _U("Additional comments"),
                             tip = _U(" ")),
                      0)

        self.addEntry("FuelLCV",
                      FloatEntry(self.nbPage[0],
                              decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                              unitdict = 'SPECIFICENTHALPY',
                              label = _U("Lower calorific value"),
                              tip = _U("Calorific value not including condensation")),
                      1)

        self.addEntry("FuelHCV",
                      FloatEntry(self.nbPage[0],
                              decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                              unitdict = 'SPECIFICENTHALPY',
                              label = _U("Higher calorific value"),
                              tip = _U("Calorific value including condensation")),
                      1)

        self.addEntry("FuelDensity",
                      FloatEntry(self.nbPage[0],
                              decimals = 2, minval = -INFINITE, maxval = INFINITE, value = 0.,
                              unitdict = 'DENSITY',
                              label = _U("Density"),
                              tip = _U(" ")),
                      1)

        #
        # frame 2 - Properties for offgas calculations
        #

        self.addEntry("OffgasHeatCapacity",
                      FloatEntry(self.nbPage[1],
                              decimals = 7, minval = 0, maxval = INFINITE, value = 0.,
                              unitdict = 'SPECIFICHEAT',
                              label = _U("Specific heat of offgas"),
                              tip = _U("Specific heat of offgas")),
                      2)

        self.addEntry("OffgasDensity",
                      FloatEntry(self.nbPage[1],
                               decimals = 2, minval = 0, maxval = INFINITE, value = 0.,
                               unitdict = 'DENSITY',
                               label = _U("Offgas Density"),
                               tip = _U("Density of offgas")),
                      2)

        self.addEntry("Humidity",
                      FloatEntry(self.nbPage[1],
                               decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                               unitdict = 'FRACTION',
                               label = _U("Offgas Humidity"),
                               tip = _U("Humidity ratio (kg H20/kg offgas) in the offgas")),
                      2)

        self.addEntry("CombAir",
                      FloatEntry(self.nbPage[1],
                               decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                               unitdict = 'FRACTION',
                               label = _U("Combustion Air Ratio"),
                               tip = _U("Mass ratio of combustion air to fuel")),
                      2)

        self.addEntry("Offgas",
                      FloatEntry(self.nbPage[1],
                               decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                               unitdict = 'FRACTION',
                               label = _U("Offgas ratio"),
                               tip = _U("Mass ratio of offgas to fuel")),
                      2)

        #
        # frame 3 - Environment parameters
        #

        self.addEntry("PEConvFuel",
                      FloatEntry(self.nbPage[0],
                               decimals = 4, minval = -INFINITE, maxval = INFINITE, value = 0.,
                               unitdict = 'FRACTION',
                               label = _U("Primary energy conversion ratio"),
                               tip = _U("Primary energy conversion ratio including transport")),
                      3)

        self.addEntry("CO2ConvFuel",
                      FloatEntry(self.nbPage[0],
                               decimals = 3, minval = -INFINITE, maxval = INFINITE, value = 0.,
                               unitdict = 'CO2RATIO',
                               label = _U("Ratio of CO2 generation"),
                               tip = _U("Ratio of CO2 generation including transport")),
                      3)


        # conclude layout
        self.conclude_layout()


class DialogDBFuelEntry(EditWindow):
    def __init__(self, prnt, main):
        self.prnt = prnt
        self.mod = prnt.mod
        if self.mod.editor: return  # do not open a second dialog

        super(DialogDBFuelEntry, self).__init__(main)

        self.panel = PanelDBFuelEntry(self, main)
        self.ret = None

        self.panel.display()
        self.Show()

        self.mod.editor = self
        self.prnt.panel.Disable()

    def OnClose(self, event):
        self.mod.editor = None
        self.prnt.panel.display()
        self.prnt.panel.Enable()
        super(DialogDBFuelEntry, self).OnClose(event)
        
