# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.einstein-energy.net/" target="_blank">www.einstein-energy.net</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQ1: general data entry
#
#==============================================================================
#
#       EINSTEIN Version No.: 3.0
#
#------------------------------------------------------------------------------
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany
#	http://www.energyxperts.net/
#
#==============================================================================

from constants import EPSILON, KELVIN, TRANSYESNO, PRODUCTCODES, TRANSPRODUCTCODES

from globals import logger, Globals

from GUI.einsteinPanel import EinsteinPanel
from GUI.tools.einsteinSimpleDialog import EinsteinSimpleDialog
from GUI.tools.dialogOKCancel import DialogOKCancel

from GUI.dataEntry.tableEntry import TableEntry, DEFAULTMENU
from GUI.displayClasses import FloatEntry, TextEntry, ChoiceEntry, CheckBoxEntry, DateEntry, FieldSizes

from GUI.GUITools import GRID_BACKGROUND_COLOR
from GUI.status import Status

from lib.text.textConversion import toGettextAndUnicode as _U, toUnicode as U
from modules.guiModules.moduleQ1 import ModuleQ1

from GUI.displayCC import PanelCCPars, TableCellEntry, findRC

import wx

EDIT_DATA =  [False, True, True, True]
EDIT_TEXT =  [False, False, False, False]



class PanelQ1(EinsteinPanel):

    def __init__(self, parent, main):

        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        layout = {"notebook": [_('EINSTEIN project data'),
                               _('General information'),
                               _('Statistical and economical data'),
                               _('Period of operation'),
                               _('Information on products')],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("EINSTEIN project data")),
                             (1, _("General information")),
                             (2, _("Statistical and economical data")),
                             (3, _("Period of operation")),
                             (3, _("Principal periods of holidays or stops for maintenance")),
                             (4, _("Information on products"))]}           # list of frames (pageno, title). if empty, no frame is created

        super(PanelQ1, self).__init__(parent, main, layout)

        self.mod = ModuleQ1(self)
        initData = self.mod.initPanel()
        NACEList = initData["NACEChoices"]
        NACESubList = initData["NACESubChoices"]

        # panel 0. EINSTEIN project data

        HEIGHT = 24
        LABEL_WIDTH = 280
        DATA_ENTRY_WIDTH = 400
        UNITS_WIDTH = 100
        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH,
                       wData=DATA_ENTRY_WIDTH, wUnits=UNITS_WIDTH)

        self.addEntry("ProjectName", TextEntry(self.nbPage[0], maxchars=45,value='',
                             label=_U("Name of the EINSTEIN project"),
                             tip=_U("Internal name with which you identify your EINSTEIN project")), 0)


        self.addEntry("CreationDate", DateEntry(self.nbPage[0],
                              #value='',
                              label=_U("Project created at: "),
                              tip="Date at which this project was created"), 0)

        self.addEntry("CreatedBy", TextEntry(self.nbPage[0], maxchars=45,value='',
                             label=_U("Created by Auditor:"),
                             tip=_U("Auditor who created the project")), 0)

        self.addEntry("LastModified", DateEntry(self.nbPage[0],
                              #value='',
                              label=_U("Last modified at: "),
                              tip="Date at which this project has been modified the last time"), 0)

        self.addEntry("ModifiedBy", TextEntry(self.nbPage[0], maxchars=45,value='',
                             label=_U("Modified by Auditor:"),
                             tip=_U("Auditor who die the last modification on the project")), 0)


        # panel 1: General information

        self.addEntry("Name", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("Name of the company"),
                             tip=_U("Legal name of the company")), 1)

        self.addEntry("City", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("City / Country"),
                             tip=_U("City where production is located")), 1)

        self.addEntry("Contact", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("Name of contact person"),
                             tip=" "), 1)

        self.addEntry("Role", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("Position of contact person in the company"),
                             tip=" "), 1)

        self.addEntry("Address", TextEntry(self.nbPage[1],maxchars=200,value='',
                             label=_U("Address"),
                             tip=" ",
                                           isMultiline=True,
                                           specialHeight=100), 1)

        self.addEntry("Phone", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("Telephone No"),
                             tip=" "), 1)

        self.addEntry("Fax", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("Fax No"),
                             tip=" "), 1)

        self.addEntry("Email", TextEntry(self.nbPage[1],maxchars=45,value='',
                             label=_U("E-mail"),
                             tip=" "), 1)

        self.addEntry("DescripIndustry",
                      TextEntry(self.nbPage[1],maxchars=200,value='',
                                label=_U("Description of the industry"),
                                tip=" ",
                                isMultiline=True,
                                specialHeight=100),
                      1)

        self.addEntry("Branch", ChoiceEntry(self.nbPage[1], 
                               values=NACEList,
                               label=_U("NACE code branch"),
                               tip=" "), 1)
        self.Bind(wx.EVT_CHOICE, self.onEntryBranch, self.entry["Branch"].entry)

        self.addEntry("SubBranch", ChoiceEntry(self.nbPage[1], 
                               values=NACESubList,
                               label=_U("NACE code sub-branch"),
                               tip=" "), 1)

        # panel 2:Statistical and economical data

        HEIGHT = 24
        LABEL_WIDTH = 560
        DATA_ENTRY_WIDTH = 120
        UNITS_WIDTH = 100
        fs = FieldSizes(wHeight=HEIGHT, wLabel=LABEL_WIDTH,
                       wData=DATA_ENTRY_WIDTH, wUnits=UNITS_WIDTH)

        self.addEntry("NEmployees", FloatEntry(self.nbPage[2],
                              decimals=0, minval=0, maxval=999999, value=0,
                              unitdict=None,
                              label=_U("Number of employees"),
                              tip=" "), 2)

        self.addEntry("Turnover", FloatEntry(self.nbPage[2],
                              decimals=2, minval=0., maxval=1.0e+9, value=0.,
                              unitdict='PRICE',
                              label=_U("Annual turnover"),
                              tip=_U("Million of euro per year")), 2)

        self.addEntry("ProdCost", FloatEntry(self.nbPage[2],
                              decimals=1, minval=0., maxval=1.0e+9, value=0.,
                              unitdict='PRICE',
                              label=_U("Annual production cost"),
                              tip=_U("Specify total factor inputs for production")), 2)

        self.addEntry("BaseYear", FloatEntry(self.nbPage[2],
                              decimals=0,minval=2000, maxval=2050, value=0,
                              unitdict=None,
                              label=_U("Base year for economic data"),
                              tip=_U("Specify the reference year for economic parameters")), 2)

        self.addEntry("Growth", FloatEntry(self.nbPage[2],
                              decimals=1, minval=0., maxval=100, value=0.,
                              unitdict="GROWTHRATE",
                              label=_U("Growth rate of the production volume foreseen for the next 5 years [%/year]"),
                              tip=" "), 2)

        self.addEntry("Independent", ChoiceEntry(self.nbPage[2],
                               values=[TRANSYESNO[val] for val in ["no", "yes"]],
                               label=_U("Is the company independent?"),
                               tip=" "), 2)

        self.addEntry("OMThermal", FloatEntry(self.nbPage[2],
                              decimals=1, minval=0., maxval=1.0e+9, value=0.,
                              unitdict='PRICE',
                              label=_U("Yearly O&M heat & cold"),
                              tip=" "), 2)

        self.addEntry("OMElectrical", FloatEntry(self.nbPage[2],
                              decimals=1, minval=0., maxval=1.0e+9, value=0.,
                              unitdict='PRICE',
                              label=_U("Yearly O&M electrical"),
                              tip=" "), 2)

        self.addEntry("PercentFuelTotcost", FloatEntry(self.nbPage[2],
                              decimals=4, minval=0., maxval=100.0, value=0.,
                              unitdict="PERCENTAGE",
                              label=_U("Percentage of fuel cost on overall production cost"),
                              tip=" "), 2)

        self.addEntry("PercentElTotcost", FloatEntry(self.nbPage[2],
                              decimals=4, minval=0., maxval=100.0, value=0.,
                              unitdict="PERCENTAGE",
                              label=_U("Percentage of electricity cost on overall production cost"),
                              tip=" "), 2)        

        # panel 3: Period of operation

        self.addEntry("HPerDayInd", FloatEntry(self.nbPage[3],
                              decimals=2, minval=0., maxval=24, value=0.,
                              unitdict=None,
                              label=_U("Total hours of operation per working day"),
                              tip=" "), 3)

        self.addEntry("NShifts", FloatEntry(self.nbPage[3],
                               decimals=0, minval=0., maxval=10., value=0.,
                               unitdict=None,
                               label=_U("Number of shifts"),
                               tip=" "), 3)

        self.addEntry("NDaysInd", FloatEntry(self.nbPage[3],
                               decimals=0, minval=0., maxval=365., value=0.,
                               unitdict=None,
                               label=_U("Days of production / operation per year"),
                               tip=" "), 3)

        self.addEntry("NoProdStart_1", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 1 - start date"),
                              tip=" "), 4)

        self.addEntry("NoProdStop_1", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 1 - stop date"),
                              tip=" "), 4)

        self.addEntry("NoProdStart_2", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 2 - start date"),
                              tip=" "), 4)

        self.addEntry("NoProdStop_2", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 2 - stop date"),
                              tip=" "), 4)

        self.addEntry("NoProdStart_3", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 3 - start date"),
                              tip=" "), 4)

        self.addEntry("NoProdStop_3", DateEntry(self.nbPage[3],
                              #value='',
                              label=_U("Period of holidays No. 3 - stop date"),
                              tip=" "), 4)

##        # panel 3.Left. Product list
##        fp.changeFont(size=TYPE_SIZE_RIGHT)
##        f = FieldSizes(wHeight=HEIGHT_RIGHT,wLabel=LABEL_WIDTH_RIGHT)
##
##        self.listBoxProducts = wx.ListBox(self.nbPage[4],-1, choices=[])
##        self.Bind(wx.EVT_LISTBOX, self.OnListBoxProductsListboxClick, self.listBoxProducts)

        # panel 3.Right top. Information on products


                # In StaticBox "General data" within "Distribution of heat/cold"
        titles = [_("Type of product"),
                  _("Product code"),
                  _("Quantity of product(s) per year [product-units/year]"),
                  _("Measurement unit for\nproduct quantity"),
                  _("Annual turnover per product"),
                  _("Electricity consumption per product"),
                  _("Fuel consumption per product(LCV)")]

        codeChoices = [TRANSPRODUCTCODES[code] for code in PRODUCTCODES]
        unitSelectors = [("text", None, None), ("choice", codeChoices, None), ("float", "PU", 0), ("text", None, None),
                         ("float", "CURRENCY", 0), ("float", "ENERGY", -1), ("float", "ENERGY", -1)]
        hAlignment = [wx.ALIGN_LEFT, wx.ALIGN_LEFT, wx.ALIGN_RIGHT, wx.ALIGN_CENTRE, wx.ALIGN_RIGHT, wx.ALIGN_RIGHT, wx.ALIGN_RIGHT]
        
        readOnly = [False]*7 if Status.ANo==-1 else [True]*7 # products can only be changed in raw data entry
                          
        self.addEntry("ProductTable",
                      TableEntry(self.nbPage[4], titles, self.mod.changeDataProductTable, size=wx.Size(780, 260),
                                 rowSize=24, colSize=100, colLabelSize=64,
                                 hAlignment = hAlignment,
                                 readOnly = readOnly,
                                 unitSelectors = unitSelectors),
                      5)

        self.entry["ProductTable"].setColSize(0, 140)
        self.entry["ProductTable"].setColSize(1, 140)
        self.entry["ProductTable"].setPopUpMenu([_('add'),_('delete')])

        # conclude layout
        self.conclude_layout()

        self.notebook.SetSelection(0)


    def fillPanel(self):
        print "PanelQ1.fillPanel"
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        super(PanelQ1, self).fillPanel()

        readOnlyEntries = ["CreationDate", "CreatedBy", "LastModified", "ModifiedBy"]

        for key in readOnlyEntries:
            self.entry[key].Disable()

##        self.displayCC()
##        wx.SafeYield()    gives problems on Windows
        self.blockEventHandlers = popBlocker

    def onEntryBranch(self, event):
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        branchIndex = self.entry["Branch"].GetValue()
        NACESubList = self.mod.onEntry("Branch", branchIndex)
        self.entry["SubBranch"].SetValue(NACESubList)
        self.blockEventHandlers = popBlocker
        
    def goTo(self, key):
        """ tries to open the page appropriate
        for a given parameter key
        """
        if key in self.ccKeys:
            entryKey = self.entryKeys[self.ccKeys.index[key]]
        else:
            entryKey = key
        super(PanelQ1, self).goTo(entryKey)


