# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.einstein-energy.net/" target="_blank">www.einstein-energy.net</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQM: data entry for monitoring data
#
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#
#------------------------------------------------------------------------------
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany
#	http://www.energyxperts.net/
#
#==============================================================================

from constants import KELVIN, HEATING_COOLING, TRANSYESNO, YESNO, TRANSPROCTYPES, TRANSSCHEDULETYPES, SCHEDULETYPES, PLATFORM
from globals import logger, Globals

from GUI.einsteinPanel import EinsteinPanel
from GUI.tools.einsteinSimpleDialog import EinsteinSimpleDialog
from GUI.tools.dialogOKCancel import DialogOKCancel

from GUI.dataEntry.tableEntry import TableEntry, DEFAULTMENU
from GUI.displayClasses import FloatEntry, TextEntry, ChoiceEntry, CheckBoxEntry
from GUI.schedules.scheduleEditor import ScheduleEditor
from GUI.GUITools import GRID_BACKGROUND_COLOR

from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U
from modules.guiModules.moduleQ3 import ModuleQ3

from GUI.displayCC import PanelCCPars, TableCellEntry

import wx

EDIT_MAINTENANCE = [False, False, False, False , True, False, False, False]
EDIT_STARTUP =  [False, True, False , True, False, False, False, True]
EDIT_INFLOW =  [True, True, True, True, True, False, False, True]
EDIT_TOTAL = [False, False, False, False , False, True, False, False]

class PanelQ3(EinsteinPanel):

    def __init__(self, parent, main, no):
        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        layout = {"notebook": ["Description",
                               "Energy demand",
                               "Residual energy",
                               "Schedule",
                               "Detailed schedule"],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("Descriptive data")),
                             (1, _("Process and process supply temperature")),
                             (1, _("Demand components")),
                             (2, _("Residual heat / cooling")),
                             (3, _("Process schedule")),
                             (3, _("Additional schedule parameters")),
                             (4, ""),
                             (2, _("Supply temperature for mandatory cooling"))]}           # list of frames (pageno, title). if empty, no frame is created
        super(PanelQ3, self).__init__(parent, main, layout)

        self.mod = ModuleQ3(self)
        initData = self.mod.initPanel(no)

        # general data
        self.addEntry("name",
                      TextEntry(self.nbPage[0], maxchars=255, value='',
                             label=_U("Process short name")),
                      0)
        self.addEntry("Description",
                      TextEntry(self.nbPage[0], maxchars=200, value='',
                             label=_U("Description"),
                             tip=_U("Give a short description of the process")),
                      0)

        self.addEntry("UnitOp",
                      ChoiceEntry(self.nbPage[0],
                               values=[],
                               label=_U("Unit operation type"),
                               tip=_U("Select from predefined list")),
                      0)
        self.entry["UnitOp"].SetValue(initData["unitOps"])     # fill list of choices

        self.addEntry("heating",
                      ChoiceEntry(self.nbPage[0],
                                  values=[U(HEATING_COOLING[key]) for key in ["cooling", "heating"]],
                                  label=_U("heating/cooling"),
                                  tip=_U("Select whether this is a heating or cooling process"),
                                  allowNones=False),
                      0)

        # process and process supply temperature

        self.addEntry("PT",
                      FloatEntry(self.nbPage[1],
                              decimals=2, minval=-KELVIN, maxval=999., value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Typical (final) temperature of the  process medium during operation"),
                              tip=_U("Give the temperature of the process medium and not that of the heat supplying medium.")),
                      1)

        self.addEntry("DPSTmin",
                      FloatEntry(self.nbPage[1],
                               decimals=2, minval=0., maxval=1.e+12, value=0.,
                               unitdict='TEMPERATUREDIFF',
                               label=_U("Minimum possible temperature difference between supply and process"),
                               tip=_U("The minimum temperature difference that is required between the (inlet of) the heat or cold supply medium and the process medium")),
                      1)

        self.addEntry("ExternalSupply",
                      ChoiceEntry(self.nbPage[1],
                                  [TRANSYESNO[key] for key in YESNO], value=None,
                                  label=_U("Is external supply for in-flows/out-flows possible?"),
                                  tip=_U("May the heating up/cooling down of the inflows happen before they " \
                                         "enter the process chamber?"),
                                  allowNones=False),
                      1)


        # Table for demand components


        titlesUPX = [_("Component"), _("Medium"), _("Flow rate"), _("Inlet/Initial temp."), _("Nominal power"),
                     _("Yearly energy"), _("Operating Hours"), _("Additional data")]

        self.addEntry("UPXTable",
                      TableEntry(self.nbPage[1], titlesUPX, self.mod.changeDataUPXTable, size=wx.Size(780, 200),
                                 rowSize=24, colSize=80, colLabelSize=32,
                                 hAlignment = [wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE,
                                              wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE],
                                 readOnly = [not edit for edit in EDIT_INFLOW],
                                 unitSelectors = [("text", None, None), ("text", None, None), ("float", "MASSORVOLUMEFLOW", 0), ("float", "TEMPERATURE", 2),
                                            ("float", "POWER", 2), ("float", "ENERGY", -1), ("float", "TIME", 0), ("text", None, None)],
                                 totalRow = True),
                      2)

##        self.entry["UPHComponents"].setRowSize(0, 48) ###TODO: increase height o title row
        self.entry["UPXTable"].setColSize(0, 180)
        self.entry["UPXTable"].setColSize(1, 112)
        self.entry["UPXTable"].setPopUpMenu([_("Add maintenance"), _("Add start-up"), _("Add inflow"), _("Delete")])
        self.entry["UPXTable"].setExternalEventHandler(self.onTableClickUPX)


        # Waste heat components

        titlesQWX = [_("Component"), _("Medium"), _("Flow rate"), _("Outlet temp."), _("Final temp."), \
                     _("Nominal power"), _("Yearly energy"), _("Operating Hours"), _("Additional data")]

        self.addEntry("QWXTable",
                      TableEntry(self.nbPage[2], titlesQWX, self.mod.changeDataQWXTable, size=wx.Size(780, 160),
                                 rowSize=24, colSize=80, colLabelSize=32,
                                 hAlignment = [wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE,
                                               wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE],
                                 readOnly = [False, False, False, False, False,
                                             False, True, True, False],
                                 unitSelectors = [("text", None, None), ("text", None, None), ("float", "MASSORVOLUMEFLOW", 0), ("float", "TEMPERATURE", 2), ("float", "TEMPERATURE", 2),
                                            ("float", "POWER", 2), ("float", "ENERGY", -1), ("float", "TIME", 0), ("text", None, None)]),
                      3)

##        self.entry["UPHComponents"].setRowSize(0, 48) ###TODO: increase height o title row
        self.entry["QWXTable"].setColSize(0, 132)
        self.entry["QWXTable"].setPopUpMenu([_("Add"), _("Delete")])
        self.entry["QWXTable"].setExternalEventHandler(self.onTableClickQWX)


        # supply temperature for mandatory cooling
        self.addEntry("DPSTminMCD",
                      FloatEntry(self.nbPage[2],
                                 decimals=2, minval=0., maxval=1.e+12, value=0.,
                                 unitdict='TEMPERATUREDIFF',
                                 label=_U("Minimum possible temperature difference between supply and out flow"),
                                 tip=_U("The minimum temperature difference that is required between (the outlet of) " \
                                        "the heat or cold supply medium and the out-flow media that have mandatory cooling down demand.")),
                     7)


        # Schedule data 

        self.addEntry("ProcType",
                      ChoiceEntry(self.nbPage[3],
                                  values=TRANSPROCTYPES.values(),
                                  label=_U("Process type"),
                                  tip=_U("Specify if process is continuous or batch"),
                                  allowNones=False),
                      4)
        self.Bind(wx.EVT_CHOICE, self.onEntryProcType, self.entry["ProcType"].entry)


        self.addEntry("scheduleType",
                      ChoiceEntry(self.nbPage[3],
                                  values=[TRANSSCHEDULETYPES[key] for key in SCHEDULETYPES],
                                  label=_U("Schedule type"),
                                  tip=_U("Select type of schedule"),
                                  allowNones=False),
                      4)
        self.Bind(wx.EVT_CHOICE, self.onEntryScheduleType, self.entry["scheduleType"].entry)
        
        # instantiate custom controls
        self.addEntry("HPerDay",
                      FloatEntry(self.nbPage[3],
                                                         decimals=2, minval=0., maxval=9999., value=0.,
                                                         unitdict='TIME',
                                                         label=_U("Effective operation hours per day"),
                                                         tip=_U("Specify the total duration of process, e.g. 3 cycles/day x 2 hrs/cycle = 6 hrs.")),
                      4)
        self.addEntry("NCycles",
                      FloatEntry(self.nbPage[3],
                                                         decimals=2, minval=0., maxval=999., value=0.,
                                                         label=_U("Number of cycles per day"),
                                                         tip=_U(" ")),
                      4)
        self.addEntry("HCycle", FloatEntry(self.nbPage[3],
                                                         decimals=2, minval=0., maxval=9999., value=0.,
                                                         unitdict='TIME',
                                                         label=_U("Duration of 1 cycle"),
                                                         tip=_U(" ")),
                      4)
        self.addEntry("NDays",
                      FloatEntry(self.nbPage[3],
                                                         decimals=0, minval=0., maxval=365., value=0.,
                                                         label=_U("Days of process operation per year"),
                                                         tip=_U(" ")),
                      4)


        self.addEntry("DTStartUp",
                      FloatEntry(self.nbPage[3],
                                                         decimals=2, minval=0., maxval=9999., value=0.,
                                                         unitdict='TIME',
                                                         label=_U("Duration of start-up period"),
                                                         tip=_U(" ")),
                      5)
        
        self.addEntry("DTInFlow",
                      FloatEntry(self.nbPage[3],
                                 decimals=2, minval=0., maxval=9999., value=0.,
                                 unitdict='TIME',
                                 label=_U("Duration of inflow period"),
                                 tip=_U("only for batch processes")),
                      5)
        
        self.addEntry("DTOutFlow",
                      FloatEntry(self.nbPage[3],
                                                         decimals=2, minval=0., maxval=9999., value=0.,
                                                         unitdict='TIME',
                                                         label=_U("Duration of outflow period"),
                                                         tip=_U("only for batch processes")),
                      5)
        

        #detailed schedule
        self.tsEditor = ScheduleEditor(self.nbPage[4], changeData=self.OnScheduleChange)
        self.mod.modTSEditor = self.tsEditor.mod   # export the tsEditor module to the backend module
        
        self.addEntry("ScheduleEditor",
                      self.tsEditor,
                      6)

        # conclude layout
        self.conclude_layout()

        self.tabScheduleActive = False                       # indicates that detailed schedule page is active and included in the notebook
        self.notebook.RemovePage(4)

        if initData["BUI"]:  # building process
            for nbPage in self.nbPage[:-1]: # disable all but the detailed schedule entry panel
                nbPage.Disable()
                
                try:    # some entries (tableEntry, ScheduleEditor) do not have a disable-function
                    self.entry[key].Disable()
                except:
                    pass  

    def fillPanel(self):
        self.blockEventHandlers = True
        super(PanelQ3, self).fillPanel()

        # display table entries
        self.entry["UPXTable"].display(data=self.data["UPXTable"]["data"], choices=self.data["UPXTable"]["choices"])
        self.onTableClickUPX("RightClick", None, None, None)       # apply special formats for maintenance and start-up rows
        
        self.entry["QWXTable"].display(data=self.data["QWXTable"]["data"], choices=self.data["QWXTable"]["choices"])

        self.setSecondaryUnits()    # have to be set AFTER call to display of the tables

        # display detailed schedule editor
        self.onEntryScheduleType(None)
##        sType = self.data["scheduleType"]
##        self.nbPage[4].Enable(sType==1)
##        if sType==1:
##            self.tsEditor.display()
##        else:
##            try:
##                self.notebook.RemovePage(4)
##            except:
##                pass
##            self.nbPage[4].Hide()
        
        # display detailed schedule editor
        anyMCD = self.data["anyMCD"]
        self.entry["DPSTminMCD"].Enable(anyMCD)

        self.setFrameTitle(_("Process: ")+ self.data["name"])

        self.displayCC()
        self.blockEventHandlers = False

    def setSecondaryUnits(self):
        """
        Sets the unit conversion factor for the secondary unit for MASSORVOLUME entries
        """
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True
        funit2Data = self.mod.getSecondaryUnits()
        self.data["funit2"] = funit2Data # store for pop-ups
        
        nrows = len(self.entry["UPXTable"].data)
        pointers = self.data ["UPXTable"]["pointers"]

        ptr = pointers["startUp"]
        if ptr is not None:
            funit2 = funit2Data["startUp"]
            self.entry["UPXTable"].setAttribute("unit2", funit2, row=ptr, col=2)
            
        for ptr in range(pointers["inFlow"], pointers["total"]):
            s = ptr - pointers["inFlow"]    # stream number
            funit2 = funit2Data["inFlow"][s]
            self.entry["UPXTable"].setAttribute("unit2", funit2, row=ptr, col=2)

        for ptr in range(len(self.data["QWXTable"]["data"])):
            s = ptr    # stream number
            funit2 = funit2Data["outFlow"][s]
            self.entry["QWXTable"].setAttribute("unit2", funit2, row=ptr, col=2)

        self.blockEventHandlers = popBlocker

    def setPointers(self, pointers):
        """
        Helper function for updating pointers after adding/deleting rows in UPXTable
        """
        self.data["UPXTable"]["pointers"] = pointers

    def onTableClickUPX(self, eventType, row, col, selection, emulate=None):

        nrows = len(self.entry["UPXTable"].data)        
        if eventType == "RightClick" or (eventType=="LeftClick" and col < len(EDIT_INFLOW)-1):                
            for r in range(nrows-1):
                # row 0
                if r==0:
                    if self.mod.hasMaintenance:
                        self.entry["UPXTable"].setAttribute("edit", EDIT_MAINTENANCE, row=0)
                    elif self.mod.hasStartUp:
                        self.entry["UPXTable"].setAttribute("edit", EDIT_STARTUP, row=0)
                    else:
                        self.entry["UPXTable"].setAttribute("edit", EDIT_INFLOW, row=0)

                elif r==1:  
                    if self.mod.hasMaintenance and self.mod.hasStartUp:
                        self.entry["UPXTable"].setAttribute("edit", EDIT_STARTUP, row=1)
                    else:  # restore to normal only if this is not the TOTAL-row
                        self.entry["UPXTable"].setAttribute("edit", EDIT_INFLOW, row=1)
                else:
                    self.entry["UPXTable"].setAttribute("edit", EDIT_INFLOW, row=r)
                    
            self.entry["UPXTable"].setAttribute("edit", EDIT_TOTAL, row=nrows-1)

        elif eventType=="LeftDclick" or (eventType=="LeftClick" and col == len(EDIT_INFLOW)-1):
            cNo = self.mod.getComponentNo(row)
            rowName = self.entry["UPXTable"].data[row][0]

            if cNo is not None and cNo != 0 and row < nrows-1: # nothing to do for maintenance heat demand and unit or total row
                dialog = PanelQ3_DialogDemand(self, cNo, title=_("Additional data for %s"%rowName),
                                              mod = self.mod,
                                              data = self.mod.getDemandData(cNo))
                self.dialog = dialog

                if emulate:
                    self.dialog = dialog
                    self.activeIndex = cNo
                    dialog.Show()
                    return
                else:
                    ret = dialog.ShowModal()

                if ret==wx.ID_OK: self.mod.setDemandData(cNo)

        self.displayCC()

    def onTableClickQWX(self, eventType, row, col, selection, emulate=None):

        ncols = self.entry["QWXTable"].ncols
        nrows = len(self.entry["QWXTable"].data)
        if eventType=="RightClick":
            pass
        elif eventType=="LeftDclick" or (eventType=="LeftClick" and col == ncols-1):
            rowName = self.entry["QWXTable"].data[row][0]

            if row >= 0 and row < nrows:
                cNo = row+1
                dialog = PanelQ3_DialogResidual(self,
                                                cNo,
                                                title=_("Additional data for %s"%rowName),
                                                mod = self.mod,
                                                data = self.mod.getOutStreamData(row))

                self.dialog = dialog

                if emulate:
                    self.activeIndex = cNo
                    self.dialog = dialog
                    dialog.Show()
                    return
                else:
                    if dialog.ShowModal()==wx.ID_OK:
                        self.mod.setOutStreamData(row)
                    else:
                        pass

        anyMCD = self.mod.getMCD()
        self.entry["DPSTminMCD"].Enable(anyMCD)

        self.displayCC()

    def onEntryProcType(self, event):
        if self.blockEventHandlers: return

        pType = self.entry["ProcType"].GetValue()
        sType = self.entry["scheduleType"].GetValue()
        
        self.entry["DTStartUp"].Enable(sType>0)
        if sType==0:
            self.entry["DTStartUp"].SetValue(None)
            
        if pType==0 or sType==0: # continuous process
            self.entry["DTInFlow"].SetValue(None)
            self.entry["DTOutFlow"].SetValue(None)
            self.entry["DTInFlow"].Disable()
            self.entry["DTOutFlow"].Disable()
        else:
            self.entry["DTInFlow"].Enable()
            self.entry["DTOutFlow"].Enable()

    def onEntryScheduleType(self, event):
        if event is not None and self.blockEventHandlers: return

        sType = self.entry["scheduleType"].GetValue()
        newData = self.mod.setScheduleType(sType)

        for key in ["HCycle", "NCycles", "HPerDay", "NDays"]:
            self.entry[key].Enable(sType==0)
            if sType != 0 and newData is not None: self.entry[key].SetValue(newData[key])

        detailedSchedule = (sType==1)
        self.nbPage[4].Enable(detailedSchedule)
        if not detailedSchedule:
            if self.tabScheduleActive:
                self.notebook.RemovePage(4)
                self.tabScheduleActive = False

##            if self.entry["DTStartUp"].GetValue() == 0: self.entry["DTStartUp"].SetValue(None)
##            if self.entry["DTInFlow" ].GetValue() == 0: self.entry["DTInFlow" ].SetValue(None)
##            if self.entry["DTOutFlow"].GetValue() == 0: self.entry["DTOutFlow"].SetValue(None)

            self.frame[5].Hide()
            self.entry["DTStartUp"].Hide()
            self.entry["DTInFlow"].Hide()
            self.entry["DTOutFlow"].Hide()

        else:
            if not self.tabScheduleActive:
                self.tsEditor.display()
                self.nbPage[4].Show()
                self.notebook.InsertPage(4, self.nbPage[4], _U('Detailed schedule'), True, -1)
                self.tabScheduleActive = True

            self.frame[5].Show()
            self.entry["DTStartUp"].Show()
            self.entry["DTInFlow"].Show()
            self.entry["DTOutFlow"].Show()

    def OnScheduleChange(self):
        newData = self.mod.updateOperationTimes()
        if newData is not None:
            for key in ["HCycle", "NCycles", "HPerDay", "NDays"]:
                self.entry[key].SetValue(newData[key])

    def displayCC(self):
        """
        If a checkProc instance is available on moduleCC
        (happens after consistency checking)
        then the results are displayes (in form of colours) in the panel
        """
        
        k = self.mod.proc.no
        from GUI.status import Status
        moduleCC = Status.mod.moduleCC
        if hasattr(moduleCC, "ccProc") and moduleCC.ccProc is not None and len(moduleCC.ccProc) > k:
            checkProc = moduleCC.ccProc[k]

            entryKeys = ["HPerDay", "NCycles", "HCycle", "NDays", "PT", "DPSTmin", "DPSTminMCD"]
            ccKeys =    ["HPerDayProc", "NBatch", "HBatch", "NDaysProc", "PT", "DPSTmin", "DPSTminMCD"]


            entryIndexUPXTable = []
            ccKeysUPXTable = []

            entryIndexUPXTableStreams = []
            ccKeysUPXTableStreams = []
            pointers = self.data ["UPXTable"]["pointers"]

            # maintenance row
            ptr = pointers["maintenance"]
            if ptr is not None:
                entryIndexUPXTable += [(0, 4), (0, 5)]
                ccKeysUPXTable += ["dotUPHm", "HPerYearProc"]

            ptr = pointers["startUp"]
            if ptr is not None:
                entryIndexUPXTable += [(ptr, 3)]
                ccKeysUPXTable += ["PTStartUp"]
                


            ptr = pointers["total"]
            entryIndexUPXTable += [(ptr, 5)]
            ccKeysUPXTable += ["UPH"]

            entryUPXTable = [TableCellEntry(self.entry["UPXTable"], r, c) for r, c in entryIndexUPXTable] # select table cell corresponding to index r,c

            colIndexUPXTableStreams = [2, 3, 4]
            ccKeysUPXTableStreams = ["mInFlowNom", "PTInFlowRec", "dotUPHc"]

            entryUPXTableStreams = []
            for ptr in range(pointers["inFlow"], pointers["total"]):
                s = ptr - pointers["inFlow"]    # stream number
                entryUPXTableStreams += [[TableCellEntry(self.entry["UPXTable"], ptr, c) for c in colIndexUPXTableStreams]] # select table cell corresponding to index r,c

            # outflow row

            ccKeysQWXTable = []
            colIndexQWXTableStreams = [2, 3, 4, 5]
            ccKeysQWXTableStreams = ["mOutFlowNom", "PTOutFlowRec", "PTFinal", "dotUPHw"]

            entryQWXTableStreams = []
            for ptr in range(len(self.data["QWXTable"]["data"])):
                s = ptr    # stream number
                entryQWXTableStreams += [[TableCellEntry(self.entry["QWXTable"], ptr, c) for c in colIndexQWXTableStreams]]

        # TODO: Highlighting also parameters in pop-ups ??
####                                  "VolProcMed": (checkProc.VolProcMed, self.tc9),
####                                  "ThermalMass": (checkProc.ThermalMass, self.ThermalMassFloatEntry),

##                                      "PTInFlow ["+streamName+"]":(self.checkProc.PTInFlowRec[streamNo-1], self.InputStreamsPTInFlowRecFloatEntry),
##                                      "VInFlowCycle ["+streamName+"]":(self.checkProc.VInFlowCycle[streamNo-1], self.InputStreamsVInFlowCycleFloatEntry),

##                                      "VOutFlowCycle ["+streamName+"]":(self.checkProc.VOutFlowCycle[streamNo-1], self.OutputStreamsVOutFlowCycleFloatEntry),
##                                      "HOutFlow ["+streamName+"]":(self.checkProc.HOutFlow[streamNo-1], self.OutputStreamsHOutFlowFloatEntry),
##                                      "PTFinalRequired ["+streamName+"]":(self.checkProc.PTFinalRequired[streamNo-1], self.OutputStreamsPTFinalReqFloatEntry)
        
            ccList = [(key, (getattr(checkProc, ccKeys[i]), self.entry[key])) for i, key in enumerate(entryKeys)]
            ccList += [(ccKey, (getattr(checkProc, ccKey), entryUPXTable[i])) for i, ccKey in enumerate(ccKeysUPXTable)]

            for ptr in range(pointers["inFlow"], pointers["total"]):
                s = ptr - pointers["inFlow"]    # stream number
                ccList += [(ccKey, (getattr(checkProc, ccKey)[s], entryUPXTableStreams[s][i])) for i, ccKey in enumerate(ccKeysUPXTableStreams)]
                
            for ptr in range(len(self.data["QWXTable"]["data"])):
                s = ptr    # stream number
                ccList += [(ccKey, (getattr(checkProc, ccKey)[s], entryQWXTableStreams[s][i])) for i, ccKey in enumerate(ccKeysQWXTableStreams)]
                
            ccPars = PanelCCPars(dict(ccList), dataGroup = ("Process[" + str(k+1) + "]"))
            ccPars.displayCC()


            self.entryKeys = entryKeys
            self.ccKeys = ccKeys

            self.ccKeysUPXTable = ccKeysUPXTable + ccKeysUPXTableStreams
            self.ccKeysQWXTable = ccKeysQWXTable + ccKeysQWXTableStreams

#### WARNING: WHERE IS THIS ???
######                                  "DPSTMinMCD": (checkProc.DPSTminMCD, self.entryDPSTminMCD),

    def goTo(self, key):
        """ tries to open the page appropriate
        for a given parameter key
        """
        if key in self.ccKeys:
            entryKey = self.entryKeys[self.ccKeys.index[key]]
        elif key in self.ccKeysUPXTable:
            entryKey = "UPXTable"
        elif key in self.ccKeysQWXTable:
            entryKey = "QWXTable"
        else:
            entryKey = key
        super(PanelQ3, self).goTo(entryKey)


#===========================================================================================================
# Dialogs for in- and outstream additional data editing
        
class PanelQ3_DialogDemand(EinsteinSimpleDialog):

    def __init__(self, parent, cNo, mod=None, title="", data=None):
        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        if cNo > 0:
            layout = {"frames": [(None, _("Inflow per cycle")),
                                 (None, _("Internal heat recovery"))]}           # list of frames (pageno, title). if empty, no frame is created
        elif cNo == 0:  # maintenance
            layout = {"frames": []}                      
        else:           # start-up
            layout = {"frames": [(None, _("Process thermal mass"))]}           

        dialogSize = (800, 300) if PLATFORM=="WIN" else (800, 200)
        super(PanelQ3_DialogDemand, self).__init__(parent, layout, mod=mod, size=dialogSize, title=title)

##        self.mod = ModuleQ3()

        if cNo > 0:
            self.addEntry("VInFlowCycle",
                          FloatEntry(self,
                                     decimals=3, minval=0., maxval=1.e+12, value=None,
                                     unitdict='VOLUMEORMASS',
                                     label=_U("Inflow of process medium per cycle"),
                                     tip=_U("Continuous process: Fluid flow rate times hours of circulation. Batch process with fluid renewal: volume of renewal.")),
                          0)


            # general data
            self.addEntry("HeatRecExist",
                          ChoiceEntry(self,
                                      [TRANSYESNO[key] for key in YESNO], value=None,
                                      label=_U("Exists internal heat recovery for the process?"),
                                      tip=_U("If affirmative, specify the inlet temperature after heat recovery"),
                                      allowNones=False),
                          1)
            self.Bind(wx.EVT_CHOICE, self.onEntryHeatRecExist, self.entry["HeatRecExist"].entry)

            self.addEntry("PTInFlow",
                          FloatEntry(self,
                                     decimals=1, minval=-KELVIN, maxval=999., value=None,
                                     unitdict='TEMPERATURE',
                                     label=_U("Inlet temperature of the process medium"),
                                     tip=_U("Inlet temperature of the process medium before heat recovery")),
                          1)

            self.addEntry("PTInFlowRec",
                          FloatEntry(self,
                                                                    decimals=1, minval=-KELVIN, maxval=999., value=None,
                                                                    unitdict='TEMPERATURE',
                                                                    label=_U("Inlet temperature of the process medium  (after heat recovery)"),
                                                                    tip=_U("Inlet temperature (towards the system) of the process medium after the heat recovery")),
                          1)
        elif cNo == 0:  # maintenance
            pass
        else: # start-up
            self.addEntry("VolProcMed",
                          FloatEntry(self, decimals=3, minval=0., maxval=1.e+12, value=0.,
                                  unitdict='VOLUMEORMASS',
                                  label=_U("Volume of the process medium within the equipment or storage"),
                                  tip=_U("e.g. volume of liquid in a bottle for cleaning")),
                          0)
            self.addEntry("ThermalMass",
                          FloatEntry(self,
                                  decimals=3, minval=0., maxval=1.e+12, value=0.,
                                  unitdict='THERMALMASS',
                                  label=_U("effective thermal mass of the process medium"),
                                  tip=_U("effective thermal mass of the process medium")),
                          0)

        # conclude layout
        self.conclude_layout()

        self.blockEventHandlers = True
        self.setSecondaryUnits(cNo)
        self.fillPanel(data)
        if "PTInFlowRec" in self.entry.keys(): self.entry["PTInFlowRec"].Disable()
        self.blockEventHandlers = False

    def setSecondaryUnits(self, cNo):
        if "VolProcMed" in self.entry.keys():
            funit2 = self.prnt.data["funit2"]["startUp"]
            funit2 = 1/funit2 if funit2 > 0 else 1.0    # 1.0 just for avoiding crash
            self.entry["VolProcMed"].updateUnitDict(funit2)
            
        if "VInFlowCycle" in self.entry.keys():
            funit2 = self.prnt.data["funit2"]["inFlow"][cNo-1]
            funit2 = 1/funit2 if funit2 > 0 else 1.0    # 1.0 just for avoiding crash
            self.entry["VInFlowCycle"].updateUnitDict(funit2)
        
    def onEntryHeatRecExist(self, event):
        if self.blockEventHandlers: return

        val = self.entry["HeatRecExist"].GetValue()

        self.entry["PTInFlow"].Enable(bool(val))
        if not val:
            self.entry["PTInFlow"].SetValue(self.entry["PTInFlowRec"].GetValue())


class PanelQ3_DialogResidual(EinsteinSimpleDialog):

    def __init__(self, parent, cNo, mod=None, data=None, title=""):
        self.prnt = parent
        
        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        layout = {"frames": [(None, _("Possibility of heat recovery")),
                             (None, _("Outflow per cycle")),
                             (None, _("Additional data for two-phase streams")),
                             (None, _("Internal heat recovery")),
                             (None, _("Mandatory cool-down"))]}           # list of frames (pageno, title). if empty, no frame is created

        dialogSize = (800, 460) if PLATFORM=="WIN" else (800, 360)
        super(PanelQ3_DialogResidual, self).__init__(parent, layout, mod=mod, size=dialogSize, title=title)

##        self.mod = ModuleQ3()

        self.addEntry("HeatRecOk",
                      ChoiceEntry(self,
                                  values=[TRANSYESNO[key] for key in YESNO], value=None,
                                  label=_U("Can heat be recovered from the outflowing medium?"),
                                  tip=_U("If NO, specify why: e.g. contamination with substances which can affect the heat exchanger,..."),
                                  allowNones=False),
                      0)
        self.Bind(wx.EVT_CHOICE, self.onEntryHeatRecOk, self.entry["HeatRecOk"].entry)

        self.addEntry("VOutFlowCycle",
                      FloatEntry(self,
                                 decimals=3, minval=0., maxval=1.e+12, value=None,
                                 unitdict='VOLUMEORMASS',
                                 label=_U("Outflow of process medium per cycle"),
                                 tip=_U("Can be different from the incoming flow if e.g. there is evaporation or some chemical reaction.")),
                      1)


        self.addEntry("HOutFlow",
                      FloatEntry(self,
                                 decimals=3, minval=0., maxval=999999., value=None,
                                 unitdict='SPECIFICENTHALPY',
                                 label=_U("Specific enthalpy of outgoing (waste) heat flows"),
                                 tip=_U("Enthalpy of the outgoing waste heat flow (e.g. water or hot humid air at the outlet of a drying process)")),
                      2)
        self.addEntry("XOutFlow",
                      FloatEntry(self,
                                 decimals=3, minval=0., maxval=999999., value=None,
                                 unitdict='FRACTION',
                                 label=_U("Humidity ratio of outgoing (waste) heat flows"),
                                 tip=_U("Specify humidity ratio in kg water per kg dry air/gas")),
                      2)

        self.addEntry("HeatRecExist",
                      ChoiceEntry(self,
                                  values=[TRANSYESNO[key] for key in YESNO], value=0,
                                  label=_U("Exists internal heat recovery for the process?"),
                                  tip=_U("If affirmative, specify the outlet temperature after heat recovery"),
                                  allowNones=False),
                      3)
        self.Bind(wx.EVT_CHOICE, self.onEntryHeatRecExist, self.entry["HeatRecExist"].entry)

        self.addEntry("PTOutFlow",
                      FloatEntry(self,
                                 decimals=1, minval=-KELVIN, maxval=999., value=None,
                                 unitdict='TEMPERATURE',
                                 label=_U("Temperature of outgoing (waste) heat flows"),
                                 tip=_U("Temperature of the outgoing waste heat flow (e.g. water or hot humid air at the outlet of a drying process)")),
                      3)

        self.addEntry("PTOutFlowRec",
                      FloatEntry(self,
                                 decimals=1, minval=-KELVIN, maxval=999., value=None,
                                 unitdict='TEMPERATURE',
                                 label=_U("Outlet temperature of the process medium  (after heat recovery)"),
                                 tip=_U("Outlet temperature of the process medium after internal heat recovery")),
                      3)


        self.addEntry("MCD",
##                      CheckBoxEntry(self, label=_U("A mandatory cooling down exists."),
##                                    style=wx.ALIGN_RIGHT),
                      ChoiceEntry(self,
                                  values=[TRANSYESNO[key] for key in YESNO], value=None,
                                  label=_U("A mandatory cooling down exists."),
                                  tip=_U("If YES the cooling down is considered as cooling demand and the process has to be connected to an appropriate cooling supply"),
                                  allowNones=False),
                      4)
        self.Bind(wx.EVT_CHOICE, self.onEntryMCD, self.entry["MCD"].entry)

        self.addEntry("PTFinalRequired",
                      FloatEntry(self,
                                 decimals=1, minval=-KELVIN, maxval=999., value=None,
                                 unitdict='TEMPERATURE',
                                 label=_U("Required final temperature of outgoing (waste) heat flows"),
                                 tip=_U("Maximal temperature to which the waste heat flow must be cooled down at least.")),
                      4)

        # conclude layout
        self.conclude_layout()

        self.blockEventHandlers = True
        self.setSecondaryUnits(cNo)
        self.fillPanel(data)

        # mark parameters as read-only
        self.entry["PTOutFlowRec"].Disable()
        self.blockEventHandlers = False

    def setSecondaryUnits(self, cNo):
        funit2 = self.prnt.data["funit2"]["outFlow"][cNo-1]
        funit2 = 1/funit2 if funit2 > 0 else 1.0    # 1.0 just for avoiding crash
        self.entry["VOutFlowCycle"].updateUnitDict(funit2)

    def onEntryHeatRecOk(self, event):
        if self.blockEventHandlers: return

        val = self.entry["HeatRecOk"].GetValue()
        pass # nothing to do here

    def onEntryHeatRecExist(self, event):
        if self.blockEventHandlers: return

        val = self.entry["HeatRecExist"].GetValue()
        self.entry["PTOutFlow"].Enable(bool(val))
        if not val:
            self.entry["PTOutFlow"].SetValue(self.entry["PTOutFlowRec"].GetValue())

    def onEntryMCD(self, event):
        if self.blockEventHandlers: return

        val = self.entry["MCD"].GetValue()

        self.entry["PTFinalRequired"].Enable(bool(val))
        if not val:
            self.entry["PTFinalRequired"].SetValue(None)
        
