# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.einstein-energy.net/" target="_blank">www.einstein-energy.net</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQM: data entry for monitoring data
#
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#
#------------------------------------------------------------------------------
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany
#	http://www.energyxperts.net/
#
#==============================================================================

from constants import INFINITE, KELVIN, HEATING_COOLING, TRANSYESNO, YESNO, TRANSPROCTYPES, TRANSSCHEDULETYPES, SCHEDULETYPES, TRANSEQUIPTYPE
from constants import EQCLASSES, TRANSEQCLASSES
from globals import logger, Globals

from GUI.einsteinPanel import EinsteinPanel
from GUI.tools.einsteinSimpleDialog import EinsteinSimpleDialog
from GUI.tools.dialogOKCancel import DialogOKCancel

from GUI.dataEntry.tableEntry import TableEntry, DEFAULTMENU
from GUI.displayClasses import FloatEntry, TextEntry, ChoiceEntry, CheckBoxEntry
from GUI.schedules.scheduleEditor import ScheduleEditor
from GUI.GUITools import GRID_BACKGROUND_COLOR

from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U
from modules.guiModules.moduleQ4 import ModuleQ4

from GUI.displayCC import PanelCCPars, TableCellEntry, findRC

from lib.GUI.plot.XYPlot import XYPlot

import wx

EDIT_DATA =  [False, True, True, True]
EDIT_TEXT =  [False, False, False, False]

class PanelQ4(EinsteinPanel):

    def __init__(self, parent, main, no):

        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        layout = {"notebook": [_('Descriptive data'),
                               _("Technical data"),
                               _('Schedule and Control'),
                               _('Control'),                               
                               _("Costs"),
                               _("Efficiency Curves"),
                               _("Part load behaviour")],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("Descriptive data")),
                             (0, _("Equipment type")),
                             (1, _("Technical data")),
                             (1, _("Model parameters for off-nominal operation conditions")),
                             (2, _("Schedule")),
                             (2, _("Control")),
                             (3, _("Master schedule")),
                             (4, _("Economic parameters")),
                             (5, _(" ")),
                             (6, _(" "))]}           # list of frames (pageno, title). if empty, no frame is created

        super(PanelQ4, self).__init__(parent, main, layout)
        self.tabScheduleActive = True                       # indicates that detailed schedule page is active and included in the notebook
        self.effPlots = True                                # indicates that efficiency plot plage is active

        self.mod = ModuleQ4(self)
        initData = self.mod.initPanel(no)

        # general data
        self.addEntry("name",
                      TextEntry(self.nbPage[0], maxchars=255,value='',
                                label=_U("Short name of equipment"),
                                tip=_U("Give some brief name of the equipments to identify them in the reports")),
                      0)

        self.addEntry("Manufact",
                      TextEntry(self.nbPage[0],maxchars=255,value='',
                                label=_U("Manufacturer"),
                                tip=_U("Attach the technical data if available")),
                      0)

        self.addEntry("YearManufact",
                      FloatEntry(self.nbPage[0],
                                 decimals=0, minval=1950, maxval=2050, value=2000,
                                 label=_U("Year of  manufacturing or/and installation?"),
                                 tip=_U("Year of manufacturing or installation")),
                      0)

        self.addEntry("Model",
                      TextEntry(self.nbPage[0],maxchars=255,value='',
                                label=_U("Model"),
                                tip=_U("Model according manufacturer nomenclature")),
                      0)
   
        # technical data
        eqSubClassChoices = [U(entry) for entry in initData["eqSubClassChoices"]]
        
        self.addEntry("eqClass",
                      ChoiceEntry(self.nbPage[0],
                                  values=[TRANSEQCLASSES[entry] for entry in EQCLASSES],
                                  label=_U("Class of equipment"),
                                  tip=_U("e.g. boiler / burner / chiller / compressor / CHP motor"),
                                  allowNones=False),
                      1)
        self.Bind(wx.EVT_CHOICE, self.onEntryEqClass, self.entry["eqClass"].entry)

        self.addEntry("eqSubClass",
                      ChoiceEntry(self.nbPage[0],
                                  values=eqSubClassChoices,
                                  label=_U("Type of equipment"),
                                  tip=_U("e.g. boiler / burner / chiller / compressor / CHP motor"),
                                  allowNones=False),
                      1)
        self.Bind(wx.EVT_CHOICE, self.onEntryEqSubClass, self.entry["eqSubClass"].entry)

##        self.addEntry("FuelType",
##                      ChoiceEntry(self.nbPage[1],
##                               values=[],
##                               label=_U("Fuel type"),
##                               tip=_U("Select fuel type from predefined list")),
##                      2)
##
##        self.Bind(wx.EVT_CHOICE, self.onFuelChange, self.entry["FuelType"])

        # Table for technical data
        titles = [_("Parameter"), _("Unit"), _("Nominal"), _("Actual (average)")]

        self.addEntry("DataTable",
                      TableEntry(self.nbPage[1], titles, self.mod.changeDataDataTable, size=wx.Size(784, 440),
                                 rowSize=24, colSize=120, colLabelSize=32,
                                 hAlignment = [wx.ALIGN_LEFT, wx.ALIGN_CENTRE, wx.ALIGN_RIGHT, wx.ALIGN_RIGHT],
                                 readOnly = [not edit for edit in EDIT_DATA],
                                 unitSelectors = [("float", "POWER", 2)]*20,
                                 unitCol = 1),
                      2)

        self.entry["DataTable"].setColSize(0, 420)
        self.entry["DataTable"].setPopUpMenu([])
        self.entry["DataTable"].setExternalEventHandler(self.onTableClick)



        self.addEntry("COPex",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval=0., maxval=100, value=0.,
                              unitdict='FRACTION',
                              label=_U("Exergetic efficiency"),
                              tip=_U("Ratio of nominal efficiency to Carnot efficiency at nominal working conditions")),
                      3)

        self.addEntry("CPartLoad",
                      FloatEntry(self.nbPage[1],
                              decimals=4, minval=0., maxval=1.0, value=0.,
                              unitdict='FRACTION',
                              label=_U("Coefficient for part load correction"),
                              tip=_U("Coefficient for part load correction function (see technical manual for details)")),
                      3)
                              
##        #next from Q4C
##        self.tc102 = ChoiceEntry(self.page1,
##                               values=[],
##                               label=_U("Refrigerant"),
##                               tip=_U("Refrigerant or working fluid") + _U(" (HP or Chiller only)"))


##        self.tc6 = FloatEntry(self.page0,
##                            decimals=0,minval=0, maxval=100, value=0,
##                            label=_U("Number of units of the same type"),
##                            tip=_U("Specify how many units of this type exist"))


        # panel 3. Schedule
        self.addEntry("HPerDay",
                      FloatEntry(self.nbPage[2],
                                 decimals=1, minval=0., maxval=24., value=0.,
                                 unitdict='TIME',
                                 label=_U("Hours of operation per day"),
                                 tip=_U("Specify representative mean values")),
                      4)

        self.addEntry("NDays",
                      FloatEntry(self.nbPage[2],
                                 decimals=1, minval=0., maxval=365., value=0.,
                                 label=_U("Days of operation per year"),
                                 tip=_U("Specify representative mean values")),
                      4)

        self.addEntry("HPerYear",
                      FloatEntry(self.nbPage[2],
                                 decimals=1, minval=0., maxval=8760., value=0.,
                                 unitdict='TIME',
                                  label=_U("Hours of operation per year"),
                                  tip=_U("Real operating hours per year")),
                      4)

        self.addEntry("HPerYearNom",
                      FloatEntry(self.nbPage[2],
                                 decimals=1, minval=0., maxval=8760., value=0.,
                                 unitdict='TIME',
                                 label=_U("Full load operating hous per year"),
                                 tip=_U("Nominal operating hours per year (equivalent hours at full load)")),
                      4)

        self.addEntry("PartLoad",
                      FloatEntry(self.nbPage[2],
                                 decimals=3, minval=0., maxval=100., value=0.,
                                 unitdict='FRACTION',
                                 label=_U("Mean utilisation factor (full capacity = 100%)"),
                                 tip=_U("Specify the mean supplied power of the boiler/cooler/etc. with respect to its nominal power")),
                      4)

        self.addEntry("controlSchedule",
                      ChoiceEntry(self.nbPage[2],
                                  values = ["no", "yes"],
                                  label=_U("Master Control schedule"),
                                  allowNones = False),
                      5)

        self.Bind(wx.EVT_CHOICE, self.onEntryControlSchedule, self.entry["controlSchedule"].entry)

        self.addEntry("ControlPar1",
                      FloatEntry(self.nbPage[2],
                              decimals=3, minval=0., maxval=100., value=0.,
                              unitdict='FRACTION',
                              label=_U("Minimum eff. electrical efficiency"),
                              tip=_U("Minimum efficiency required for operating the equipment.") + "\n" + _U("See technical manual for details")),
                      5)


        #detailed schedule
        self.tsEditor = ScheduleEditor(self.nbPage[3], changeData=self.onScheduleChange)
        self.mod.modTSEditor = self.tsEditor.mod   # export the tsEditor module to the backend module
        
        self.addEntry("ScheduleEditor",
                      self.tsEditor,
                      6)

        #page 6: cost parameters

        self.addEntry("TurnKeyPrice",
                      FloatEntry(self.nbPage[4],
                                 decimals = 2, minval = -INFINITE, maxval = INFINITE, value = 0.,
                                 unitdict = 'PRICE',
                                 label = _U("Turn-key price"),
                                 tip = _U("Full cost including transport, installation, piping and auxiliary equipment") + \
                                       _U("Only new equipment")),
                      7)

        self.addEntry("OMFix",
                      FloatEntry(self.nbPage[4],
                                 decimals = 2, minval = -INFINITE, maxval = INFINITE, value = 0.,
                                 unitdict = 'PRICE',
                                 label = _U("Annual operational and maintenance fixed costs"),
                                 tip = _U("Fixed costs independent of actual operation hours")),
                      7)

        self.addEntry("OMVar",
                      FloatEntry(self.nbPage[4],
                                 decimals = 6, minval = -INFINITE, maxval = INFINITE, value = 0.,
                                 unitdict = 'ENERGYTARIFF',
                                 label = _U("Annual operational and maintenance variable costs dependant on usage"),
                                 tip = _U("Annual operational and maintenance variable costs dependant on usage (approximate average per unit of generated heat)")),
                      7)


        # add panel specific entries

        plotSize = (780, 520)        
        self.addEntry("EfficiencyCurves",
                      XYPlot(self.nbPage[5], size=plotSize),
                      8)

        self.entry["EfficiencyCurves"].xAxis(label=_("driving temperature [°C]"))
        self.entry["EfficiencyCurves"].yAxis(label=_("electric efficiency [-]"))

        self.addEntry("PartLoadBehaviour",
                      XYPlot(self.nbPage[6], size=plotSize),
                      9)

        self.entry["PartLoadBehaviour"].xAxis(label=_("ratio of actual vs. nominal electricity generation"))
        self.entry["PartLoadBehaviour"].yAxis(label=_("electric efficiency"))

        # conclude layout
        self.conclude_layout()


    def fillPanel(self):
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        super(PanelQ4, self).fillPanel()

        self.entry["eqClass"].SetValue(self.data["eqClass"])
        # display table entries
        self.entry["DataTable"].display(data=self.data["DataTable"]["data"],
                                        choices=self.data["DataTable"]["choices"],
                                        toolTips = self.data["DataTable"]["toolTips"])
        self.updateTableFormat()

        self.setSecondaryUnits()    # have to be set AFTER call to display of the tables
        # display detailed schedule editor
        self.entry["controlSchedule"].SetValue(self.data["controlSchedule"])
        self.onEntryControlSchedule(None)
##        sType = self.data["controlSchedule"]
##        self.nbPage[3].Enable(sType==1)
##        if sType==1:
##            self.tsEditor.display()
##            self.nbPage[3].Show()
##        else:
##            self.nbPage[3].Hide()

        # enable/disable entries for part-load and temperature correction
        self.entry["CPartLoad"].Enable(self.data["PartLoadCorr"])
        self.entry["COPex"].Enable(self.data["TCorr"])

        if self.data["PartLoadCorr"] or self.data["TCorr"]:
            self.entry["CPartLoad"].Show()
            self.entry["COPex"].Show()
            self.frame[3].Show()
        else:
            self.entry["CPartLoad"].Hide()
            self.entry["COPex"].Hide()
            self.frame[3].Hide()
        self.nbPage[1].Layout()
                

        # enable/disable control parameters
        self.entry["ControlPar1"].Enable(self.data["NControlPars"]>0)
        if self.data["NControlPars"] > 0:
            self.entry["ControlPar1"].Show()
        else:
            self.entry["ControlPar1"].Hide()
        self.nbPage[2].Layout()
        
        # display detailed schedule editor
        self.setFrameTitle(_("Equipment: ")+ self.data["name"])

        # display efficiency plots
        effPlots = self.data["EffPlots"]
        pageIndex = 4 + self.tabScheduleActive
        if effPlots and not self.effPlots:
            self.nbPage[5].Show()
            self.nbPage[6].Show()
            self.notebook.InsertPage(pageIndex, self.nbPage[5], _U("Efficiency Curves"), True, -1)
            self.notebook.InsertPage(pageIndex+1, self.nbPage[6], _U("Part load behaviour"), True, -1)
                
        elif self.effPlots and not effPlots:
            self.notebook.RemovePage(pageIndex+1)
            self.notebook.RemovePage(pageIndex)

        self.effPlots = effPlots

        self.displayCC()
        self.blockEventHandlers = popBlocker

    def setSecondaryUnits(self):
        """
        Sets the unit conversion factor for the secondary unit for MASSORVOLUME entries
        """
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        for row, funit2 in self.data["funit2"]:
            self.entry["DataTable"].setAttribute("unit2", funit2, row=row, col=[2, 3])

        self.blockEventHandlers = popBlocker

    def setPointers(self, pointers):
        """
        Helper function for updating pointers after adding/deleting rows in DataTable
        """
        self.data["DataTable"]["pointers"] = pointers


    def updateTableFormat(self):
        """
        Helper function for updating pointers after adding/deleting rows in DataTable
        """
        for r, row in enumerate(self.mod.editableDataTable):
            if r in self.mod.subTitlesDataTable:
                self.entry["DataTable"].setAttribute("style", ["italic, bold"]*4, row=r)
                self.entry["DataTable"].setAttribute("edit", [False]*4, row=r)
                self.entry["DataTable"].grid.SetRowSize(r, 28)
            else:
                self.entry["DataTable"].setAttribute("edit", [False, True] + row, row=r, col=[2,3])
                self.entry["DataTable"].grid.SetRowSize(r, 24)

    def onTableClick(self, eventType, row, col, selection):
        if self.blockEventHandlers: return

        nrows = len(self.entry["DataTable"].data)        
        if eventType == "RightClick" or (eventType=="LeftClick" and col < len(EDIT_DATA)-1):
            self.updateTableFormat()

    def onScheduleChange(self):
        pass
    
    def displayCC(self):
        """
        If a checkProc instance is available on moduleCC
        (happens after consistency checking)
        then the results are displayes (in form of colours) in the panel
        """
##        if self.blockEventHandlers: return

        j = self.mod.eq.no
        from GUI.status import Status
        moduleCC = Status.mod.moduleCC

        if Status.ANo > 0 and hasattr(moduleCC, "ccLocal") and moduleCC.ccLocal.dataGroup=="Equipment": # local check
            checkEq = moduleCC.ccLocal    
        elif hasattr(moduleCC, "ccEq") and moduleCC.ccEq is not None and len(moduleCC.ccEq) > j:
            checkEq = moduleCC.ccEq[j]
        else:
            return

        entryKeys = ["COPex", "TurnKeyPrice", "OMFix", "OMVar", "HPerDay", "NDays", "HPerYear", "HPerYearNom"]
        ccKeys    = ["COPex", "TurnKeyPrice", "OandMfix", "OandMvar", "HPerDayEq", "NDaysEq", "HPerYearEq", "HPerYearEqNom"]

        entryIndexDataTable = []
        ccKeysDataTable = []

        entryIndexDataTableStreams = []
        ccKeysDataTableStreams = []
        pointers = self.data["DataTable"]["pointers"]
        parCodes = self.data["DataTable"]["parCodes"]

        # access to entries in data table
        ptr = pointers["FET"]
        if ptr is not None:
            entryIndexDataTable += [findRC("mdotFuel", parCodes), findRC("Pel", parCodes), findRC("ThermalConsum", parCodes), findRC("TSupply", parCodes)]
            ccKeysDataTable += ["FuelConsumNom", "ElectriConsumNom", "ThermalConsumNom", "TSupply"]

        ptr = pointers["USH"]
        if ptr is not None:
            entryIndexDataTable += [findRC("Pnom", parCodes), findRC("HCGTEfficiency", parCodes), findRC("HCGTEffReal", parCodes),
                                    findRC("ElectriProduction", parCodes), findRC("HCGEEfficiency", parCodes)]
            ccKeysDataTable += ["HCGPnom", "HCGTEfficiencyNom", "HCGTEfficiency",
                                "ElectriProduction", "HCGEEfficiency"]
            
        ptr = pointers.get("QWH")
        if ptr is not None:
            entryIndexDataTable += [findRC("TemperatureReCooling", parCodes)]
            ccKeysDataTable += ["TemperatureReCooling"]

##                                     "TExhaustGas":(checkEq.TExhaustGas, self.tc16),
##                                     "ExcessAirRatio":(checkEq.ExcessAirRatio, self.tc16_2),

        # filter Nones in entryIndex
        indexList = []
        ccKeyList = []
        for i, index in enumerate(entryIndexDataTable):
            if index is not None:
                indexList.append(index)
                ccKeyList.append(ccKeysDataTable[i])

        entryIndexDataTable = indexList
        ccKeysDataTable = ccKeyList
        
        entryDataTable = [TableCellEntry(self.entry["DataTable"], r, c+2) for r, c in entryIndexDataTable] # select table cell corresponding to index r,c

        ccList = [(key, (getattr(checkEq, ccKeys[i]), self.entry[key])) for i, key in enumerate(entryKeys)]
        ccList += [(ccKey, (getattr(checkEq, ccKey), entryDataTable[i])) for i, ccKey in enumerate(ccKeysDataTable)]
            
        ccPars = PanelCCPars(dict(ccList), dataGroup = ("Equipment[" + str(j+1) + "]"))
        ccPars.displayCC()

        self.ccKeysDataTable = ccKeysDataTable
        self.ccKeys = ccKeys
        self.entryKeys = entryKeys


    def onEntryEqClass(self, event):
        """
        updates choices for equiment sub-class in case of a change of eqClass
        """
        if self.blockEventHandlers: return
        eqSubClassChoices, changed = self.mod.setEqClass(self.entry["eqClass"].GetValue())
        if changed:
            self.entry["eqSubClass"].SetValue(eqSubClassChoices)
#            self.display()
        
    def onEntryEqSubClass(self, event):
        if self.blockEventHandlers: return

        changed = self.mod.setEqSubClass(self.entry["eqSubClass"].GetValue())
        if changed:
            self.display()
        self.notebook.SetSelection(0)
        
    def onFuelChange(self, event):
        pass

    def onEntryControlSchedule(self, event):
        if self.blockEventHandlers and event is not None: return
        
        self.controlSchedule = self.entry["controlSchedule"].GetValue()
        self.mod.setScheduleType(self.controlSchedule)

        
        self.nbPage[3].Enable(self.controlSchedule==1)
        
        if self.controlSchedule==1:
            if not self.tabScheduleActive:
                self.tsEditor.display()
                self.nbPage[3].Show()
                self.notebook.InsertPage(3, self.nbPage[3], _U('Control'), True, -1)
                self.tabScheduleActive = True
                self.notebook.SetSelection(3)
            elif event is None:
                self.tsEditor.display()
                
        else:
            if self.tabScheduleActive:
                self.notebook.RemovePage(3)
                self.tabScheduleActive = False


    def goTo(self, key):
        """ tries to open the page appropriate
        for a given parameter key
        """

        if key in self.ccKeys:
            entryKey = self.entryKeys[self.ccKeys.index(key)]
        elif key in ["USHj", "USHConv"]: # data which cannot be entered directly -> define operation schedule
            entryKey = "PartLoad"
        elif key in self.ccKeysDataTable:
            entryKey = "DataTable"
        else:
            entryKey = key
        super(PanelQ4, self).goTo(entryKey)
