# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.einstein-energy.net/" target="_blank">www.einstein-energy.net</a>)
#
#------------------------------------------------------------------------------
#
#	PanelQ5: PIPE data entry
#
#==============================================================================
#
#       EINSTEIN Version No.: 3.0
#
#------------------------------------------------------------------------------
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany
#	http://www.energyxperts.net/
#
#==============================================================================

from constants import EPSILON, KELVIN
from globals import logger, Globals

from GUI.einsteinPanel import EinsteinPanel
from GUI.tools.einsteinSimpleDialog import EinsteinSimpleDialog
from GUI.tools.dialogOKCancel import DialogOKCancel

from GUI.dataEntry.tableEntry import TableEntry, DEFAULTMENU
from GUI.displayClasses import FloatEntry, TextEntry, ChoiceEntry, CheckBoxEntry

from GUI.displayClasses import FieldSizes, WallpaperedPanel, smallFloatEntryClone, smallChoiceEntryClone, mirrorTextCtrl

from GUI.schedules.scheduleEditor import ScheduleEditor
from GUI.GUITools import GRID_BACKGROUND_COLOR

from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U
from modules.guiModules.moduleQ5 import ModuleQ5

from GUI.displayCC import PanelCCPars, TableCellEntry, findRC

import wx

EDIT_DATA =  [False, True, True, True]
EDIT_TEXT =  [False, False, False, False]

class PanelQ5(EinsteinPanel):

    def __init__(self, parent, main, no):

        self.blockEventHandlers = True

        # generate list of notebook pages and frames
        layout = {"notebook": [_('Pipe data'),
                               _("Graphical view"),
                               "hidden"],         # list of notebook-page titles. if empty, no notebook is created
                  "frames": [(0, _("General data")),
                             (0, _("Heat losses and distribution efficiency")),
                             (0, _("Temperatures and flow rates")),
                             (0, _("Piping specifications")),
                             (2, "hidden")]}           # list of frames (pageno, title). if empty, no frame is created

        super(PanelQ5, self).__init__(parent, main, layout)
        self.tabScheduleActive = True                       # indicates that detailed schedule page is active and included in the notebook

        self.mod = ModuleQ5(self)
        initData = self.mod.initPanel(no)

        # In StaticBox "General data" within "Distribution of heat/cold"    
        self.addEntry("name",
                      TextEntry(self.nbPage[0], maxchars=255, value='',
                             label=_U("Name of the branch / distribution system"),
                             tip=_U("Give some brief name or number of the distribution tube consistent with the hydraulic scheme")),
                      0)

        self.addEntry("pipeType",
                      ChoiceEntry(self.nbPage[0],
                                  values=[U(name) for name in initData["pipeTypeChoices"]],
                                  label=_U("Type of distribution circuit"),
                                  tip=_U("specify closed / open cuircuit or built-in heater/cooler without distribution circuit"),
                                  allowNones=False),
                      0)

        self.addEntry("fluidId",
                      ChoiceEntry(self.nbPage[0],
                                  values=[U(name) for name in initData["FluidChoices"]],
                                  label=_U("Heat or cold distribution medium"),
                                  tip=_U("e.g. air for drying process, vapour, hot water, refrigerant,..."),
                                  allowNones=True),
                      0)

        self.addEntry("UAPipe",
                      FloatEntry(self.nbPage[0],
                              decimals=2, minval=0., maxval=1.e+12, value=0.,
                              unitdict='HEATTRANSFERCOEF',
                              label=_U("Total coefficient of heat losses for piping or ducts"),
                              tip=_U("For the whole duct: go and return")),
                      1)
        self.addEntry("HDEffAvg",
                      FloatEntry(self.nbPage[0],
                                   decimals=3, minval=0., maxval=100., value=0.,
                                   unitdict='FRACTION',
                                   label=_U('Energy efficiency of distribution'),
                                   tip=_U(" ")),
                      1)
        self.entry["HDEffAvg"].Disable()
        
        self.addEntry("DistribCircFlow",
                      FloatEntry(self.nbPage[0],
                              decimals=2, # max n. of characters right of decimal point
                              minval=0., # min value accepted
                              maxval=999999., # max value accepted
                              value=0., # initial value
                              unitdict='MASSORVOLUMEFLOW', # values for the units chooser
                              #label=_U("Nominal production"), # label
                              label=_U("Nominal flow rate, forward"),
                              tip=_U(" ")),
                      2)
        self.addEntry("ToutDistrib",
                      FloatEntry(self.nbPage[0],
                              decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Outlet temperature (to distribution)"),
                              tip=_U("Temperature of supply medium from equipment")),
                      2)
        self.addEntry("TreturnDistrib",
                      FloatEntry(self.nbPage[0],
                              decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Return temperature"),
                              tip=_U("Temperature of return of the supply medium from distribution (e.g. return temperature of condensate in a vapour system)")),
                      2)
        self.addEntry("PercentRecirc",
                      FloatEntry(self.nbPage[0],
                              decimals=3, minval=0., maxval=100., value=0.,
                              unitdict='FRACTION',
                              label=_U("Rate of recirculation"),
                              tip=_U("Specify the rate of recirculation of the heat/cold supply medium (100% = totally closed circuit)")),
                      2)
        self.addEntry("Tfeedup",
                      FloatEntry(self.nbPage[0],
                              decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                              unitdict='TEMPERATURE',
                              label=_U("Temperature of feed-up in open circuit"),
                              tip=_U("Temperature of medium of distribution of heat/cold entering in open circuit (e.g. temperature of water entering from network...)")),
                      2)
        self.addEntry("PressDistMedium",
                      FloatEntry(self.nbPage[0],
                              decimals=1, minval=0., maxval=999999., value=0.,
                              unitdict='PRESSURE',
                              label=_U("Pressure of heat or cold distribution medium"),
                              tip=_U("Working pressure for the heat/cold supply medium")),
                      2)
        self.addEntry("TotLengthDistPipe",
                      FloatEntry(self.nbPage[0],
                              decimals=2, minval=0., maxval=999999., value=0.,
                              unitdict='LENGTH',
                              label=_U("Total length of distribution piping or ducts (one way)"),
                              tip=_U("Only distance one way")),
                      3)
        self.addEntry("DDistPipe",
                      FloatEntry(self.nbPage[0],
                              decimals=2, minval=0., maxval=999999., value=0.,
                              unitdict='LENGTH',
                              label=_U("Mean pipe diameter"),
                              tip=_U(" ")),
                      3)
        self.addEntry("DeltaDistPipe",
                      FloatEntry(self.nbPage[0],
                              decimals=2, minval=0., maxval=999999., value=0.,
                              unitdict='LENGTH',
                              label=_U("Insulation thickness"),
                              tip=_U(" ")),
                      3)

        # hidden fields with consistency check results


        self.addEntry("TRetOut",
                      FloatEntry(self.nbPage[2],
                                         decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                                         unitdict='TEMPERATURE',
                                         label=_U("Return temperature after mixing with feed up"),
                                         tip=_U(" "),
                                         isStatic=True),
                      4)
        self.addEntry("TRetIn",
                      FloatEntry(self.nbPage[2],
                                 decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                                 unitdict='TEMPERATURE',
                                 label=_U("Return temperature"),
                                 tip=_U(" "),
                                 isStatic=True),
                      4)
        self.addEntry("TForwOut",
                      FloatEntry(self.nbPage[2],
                                   decimals=1, minval= -KELVIN, maxval=2000., value=0.,
                                   unitdict='TEMPERATURE',
                                   label=_U("Outlet temperature from distribution to process"),
                                   tip=_U(" "),
                                   isStatic=True),
                      4)
        self.addEntry("FlowRateReturn",
                      FloatEntry(self.nbPage[2],
                                         decimals=2,
                                         minval=0., 
                                         maxval=999999.,
                                         value=None, 
                                         unitdict='MASSORVOLUMEFLOW',
                                         label=_U("Nominal flow rate, return"),
                                         tip=_U(" "),
                                         isStatic=True),
                      4)

        self.Bind(wx.EVT_CHOICE, self.onPipeTypeChange, self.entry["pipeType"].entry)  
        self.Bind(wx.EVT_CHOICE, self.onFluidChange, self.entry["fluidId"].entry)  
        self.Bind(wx.EVT_TEXT, self.onRecirculationChange, self.entry["PercentRecirc"].entry)

        self.entry["TRetOut"].Hide()
        self.entry["TRetIn"].Hide()
        self.entry["TForwOut"].Hide()
        self.entry["FlowRateReturn"].Hide()


        import os

        self.wallPaper = [wx.Image(os.path.join("GUI/img", "PanelQ5_Graphics_Closed_Circuit.png")),
                          wx.Image(os.path.join("GUI/img", "PanelQ5_Graphics_Open_Circuit.png"))]

        self.nbPage[1] = WallpaperedPanel(self.notebook, wallpaper=self.wallPaper[0])
        self.notebook.DeletePage(1)
        self.notebook.InsertPage(1, self.nbPage[1], _("Graphical view"))
        self.graphicalView = True

                # graphical input page
        FieldSizes(wLabel=0, wData=80, wUnits=70)
        for widget, position in ((self.entry["fluidId"], (315, 40)), # medium
                                 (self.entry["DistribCircFlow"], (315, 90)), # DistribCircFlow
                                 (self.entry["ToutDistrib"], (5, 45)), # TOutDistrib
                                 (self.entry["PercentRecirc"], (400, 405)), # recirculation
                                 (self.entry["Tfeedup"], (100, 405)), # feed-up
                                 (self.entry["UAPipe"], (450, 185)), # UA Pipe
                                 (self.entry["DDistPipe"], (255, 135)),
                                 (self.entry["DeltaDistPipe"], (255, 165)),
                                 (self.entry["TotLengthDistPipe"], (255, 200)),
                                 (self.entry["TRetOut"], (5, 335)),
                                 (self.entry["HDEffAvg"], (325, 460)),
                                 (self.entry["TRetIn"], (600, 335)),
                                 (self.entry["TForwOut"], (600, 45)),):
            if isinstance(widget, FloatEntry):
                clone = smallFloatEntryClone(widget, self.nbPage[1], position)
            elif isinstance(widget, ChoiceEntry):
                clone = smallChoiceEntryClone(widget, self.nbPage[1], position)
            else:
                raise TypeError("Invalid widget type.")
            
            if widget == self.entry["DistribCircFlow"]: self.cloneDistribCircFlow = clone
            elif widget == self.entry["fluidId"]: self.cloneFluidId = clone
            elif widget == self.entry["PercentRecirc"]: self.clonePercentRecirc = clone
        

        self.cloneTreturnDistrib = smallFloatEntryClone(self.entry["TreturnDistrib"],
                                                        self.nbPage[1],
                                                        (220, 335)) # TRetDistrib
        self.cloneFlowRateReturn = smallFloatEntryClone(self.entry["FlowRateReturn"], 
                                                        self.nbPage[1],
                                                        (315, 290))



##        # Right panel controls
##        fp.changeFont(size=TYPE_SIZE_RIGHT)
##        FieldSizes(wHeight=HEIGHT_RIGHT, wLabel=LABEL_WIDTH_RIGHT)
##
##        self.tc15 = IntEntry(self.page1,
##                             minval=0, maxval=100, value=0,
##                             label=_U("Number of storage units"),
##                             tip=_U("Specify the number of storage units of the same type"))
##        self.tc16 = FloatEntry(self.page1,
##                              decimals=1, minval=0., maxval=1.e+12, value=0.,
##                              unitdict='VOLUME',
##                              label=_U("Volume of one storage unit"),
##                              tip=_U("Volume of the storage medium of a single single storage unit"))
##        self.tc17 = ChoiceEntry(self.page1,
##                               values=TRANSSTORAGETYPES.values(),
##                               label=_U("Type of heat storage"),
##                               tip=_U("Select from predefined list"))
##        self.tc18 = FloatEntry(self.page1,
##                              decimals=1, minval=0., maxval=9999., value=0.,
##                              unitdict='PRESSURE',
##                              label=_U("Pressure of heat storage medium"),
##                              tip=_U("Pressure of the process medium entering the storage unit if different from storage medium"))
##        self.tc19 = FloatEntry(self.page1,
##                              decimals=1, minval= -KELVIN, maxval=2000., value=0.,
##                              unitdict='TEMPERATURE',
##                              label=_U("Maximum temperature of the storage"),
##                              tip=_U("The maximum temperature to which storage unit can be operated"))
        
        # conclude layout
        self.conclude_layout()

        self.nbPage[1].Hide()
        self.nbPage[2].Hide()
        self.entry["Tfeedup"].Hide()
        self.entry["PressDistMedium"].Hide()
        self.notebook.RemovePage(2)

        self.notebook.SetSelection(0)


    def fillPanel(self):
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        super(PanelQ5, self).fillPanel()

        self.setSecondaryUnits()    # have to be set AFTER call to display of the tables

        pipeType = self.data["pipeType"]

        self.entry["PercentRecirc"].Enable(pipeType==0)

        optionalEntries = ["fluidId", "DistribCircFlow", "ToutDistrib", "PercentRecirc", "TreturnDistrib", "Tfeedup",
                           "TotLengthDistPipe", "DDistPipe", "DeltaDistPipe"]
        optionalClones = (self.cloneFlowRateReturn, self.cloneTreturnDistrib, self.clonePercentRecirc)
        
        if pipeType==2:
            for key in optionalEntries:
                self.entry[key].Hide()
            self.entry["UAPipe"].Disable()
            self.frame[2].Hide()
            self.frame[3].Hide()

##            self.nbPage[1].Hide()
            if self.graphicalView:
                self.notebook.RemovePage(1)
                self.graphicalView = False
        else:
            for key in optionalEntries:
                self.entry[key].Show()

            for widget in optionalClones:
                if pipeType==0:
                    widget.Show()
                else:
                    widget.Hide()

            recirc = self.data["PercentRecirc"]
            if recirc is not None and abs(recirc - 1.0) < EPSILON:
                self.entry["Tfeedup"].Hide()
            else:
                self.entry["Tfeedup"].Show()
                    
            self.entry["UAPipe"].Enable()
            self.frame[2].Show()
            self.frame[3].Show()

            self.nbPage[1].SetWallpaper(self.wallPaper[pipeType])
            if not self.graphicalView:
                self.notebook.InsertPage(1, self.nbPage[1], _U('Graphical view'))##            self.nbPage[1].Show()
                self.graphicalView = True

        self.nbPage[0].Layout()

        self.displayCC()
##        wx.SafeYield()    gives problems on Windows
        self.blockEventHandlers = popBlocker

    def setSecondaryUnits(self):
        """
        Sets the unit conversion factor for the secondary unit for MASSORVOLUME entries
        """
        popBlocker = self.blockEventHandlers
        self.blockEventHandlers = True

        funit2 = self.data["funit2"]
        self.entry["DistribCircFlow"].updateUnitDict(funit2)
        self.entry["FlowRateReturn"].updateUnitDict(funit2)
        self.cloneDistribCircFlow.updateUnitDict(funit2)
        self.cloneFlowRateReturn.updateUnitDict(funit2)

##        wx.SafeYield()        commented out. gives problems on Windows
        self.blockEventHandlers = popBlocker

    def displayCC(self):
        """
        If a checkProc instance is available on moduleCC
        (happens after consistency checking)
        then the results are displayes (in form of colours) in the panel
        """
        m = self.mod.pipe.no
        from GUI.status import Status
        moduleCC = Status.mod.moduleCC

        if Status.ANo > 0 and hasattr(self.mod.pipe, "cc") and self.mod.pipe.cc is not None: # local check
            cc = self.mod.pipe.cc
##            cc = moduleCC.ccLocal
            if cc.index != m:
                return
        elif hasattr(moduleCC, "ccPipe") and moduleCC.ccPipe is not None and len(moduleCC.ccPipe) > m:
            cc = moduleCC.ccPipe[m]
        else:
            return

        entryKeys = ['HDEffAvg', 'UAPipe',
                     'FlowRateReturn','DistribCircFlow', 'PercentRecirc',
                     'ToutDistrib', 'TForwOut', 'TreturnDistrib', 'TRetIn', 'TRetOut', 'Tfeedup',
                     'TotLengthDistPipe', 'DDistPipe', 'DeltaDistPipe']
        
        ccKeys =    ['HDEffAvg', 'UAPipe',
                     'RecFlow','DistribCircFlow', 'PercentRecirc',
                     'ToutDistrib', 'TForwOut', 'TreturnDistrib', 'TRetIn', 'TRetOut', 'TFeedUp',
                     'TotLengthDistPipe', 'DDistPipe', 'DeltaDistPipe']

        ccList = [(key, (getattr(cc, ccKeys[i]), self.entry[key])) for i, key in enumerate(entryKeys)]
            
        ccPars = PanelCCPars(dict(ccList), dataGroup = ("Pipe/Duct[" + str(m+1) + "]"))
        ccPars.displayCC()

        self.entryKeys = entryKeys
        self.ccKeys = ccKeys

    def onFluidChange(self, event):
        if self.blockEventHandlers: return
        popBlockers = self.blockEventHandlers
        self.blockEventHandlers = True

        newData = self.readPanel()
        self.mod.readPanel(newData)
        
        fluidId = self.entry["fluidId"].GetValue()
        funit2 = self.mod.changeFluidId(fluidId)
        
        self.display()

##        wx.SafeYield()    problems on windows
        self.blockEventHandlers = popBlockers
        
    def onRecirculationChange(self, event):    
        if self.blockEventHandlers: return

        return
    
##        recirc = self.entry["PercentRecirc"].GetValue()
##        pipeType = self.entry["pipeType"].GetValue()
        
##        if recirc is None or float(recirc) > 0 and pipeType==1:
##            self.entry["pipeType"].SetValue(0)
##            
##        if recirc==0 and pipeType==0:
##            self.entry["pipeType"].SetValue(1)

            

    def onPipeTypeChange(self, event):
        if self.blockEventHandlers: return

        popBlockers = self.blockEventHandlers
        self.blockEventHandlers = True

        newData = self.readPanel()
        self.mod.readPanel(newData)

##        index = self.entry["pipeType"].GetValue()
##        self.mod.setPipeType(index)
        
        self.display()

        self.mod.pipe.cc = None

        wx.SafeYield()
        self.blockEventHandlers = popBlockers
 

    def goTo(self, key):
        """ tries to open the page appropriate
        for a given parameter key
        """
        if key in self.ccKeys:
            entryKey = self.entryKeys[self.ccKeys.index[key]]
        else:
            entryKey = key
        super(PanelQ5, self).goTo(entryKey)

