import wx

from auxiliary.auxiliary import toUnicode as U, toGettextAndUnicode as _U
import modules.matPanel as Mp

import matplotlib.font_manager as font
from matplotlib.ticker import FuncFormatter, FixedLocator
from GUI.GUITools import *
from GUI.numCtrl import format_int_wrapper, convertDoubleToString
import textwrap

# fix parameters for plot
axeslabel_fontsize = 8
axesticks_fontsize = 8
legend_fontsize = 8
spacing_left = 0.12
spacing_right = 0.98
spacing_bottom = 0.16
spacing_top = 0.98

def drawFigure(self, data=None):
    """ Plotting method
    will be dynamically added as method of class XYPlot by Mp.matPanel
    """
    
    self.figure.clf()
    self.subplot = self.figure.add_subplot(1, 1, 1)
    self.figure.subplots_adjust(left=spacing_left, right=spacing_right, bottom=spacing_bottom, top=spacing_top)
    
    nx = len(self.xdata)
    ny = len(self.ydata)
    
    for j in range(ny):

        # common or separate x-data for each of the series in y ?
        jx = 0 if nx==1 else j
            
        self.subplot.plot(self.xdata[jx],
                          self.ydata[j],
                          LINETYPES[j%len(LINETYPES)],
                          color=ORANGECASCADE[j%len(ORANGECASCADE)],
                          label=U(self.dataLabel[j]),
                          linewidth=2)
        
    # set limits of x- and y-axes
    self.subplot.axes.set_xlim([self.xmin, self.xmax])
    self.subplot.axes.set_ylim([self.ymin, self.ymax])

    if self.xTicks is not None:
        ticks = [tick for tick in self.xTicks if tick >= self.xmin and tick <= self.xmax]
        self.subplot.xaxis.set_ticks(ticks)

    if self.minorXTicks is not None:
        ticks = [tick for tick in self.minorXTicks if tick >= self.xmin and tick <= self.xmax]
        minor_locator = FixedLocator(ticks)
        self.subplot.xaxis.set_minor_locator(minor_locator)

    if self.grid is not None:
        if self.minorXTicks is not None:
            self.subplot.grid(which='minor')
        self.subplot.grid()

    # set labels of axes
    major_formatter = FuncFormatter(format_int_wrapper)
    self.subplot.axes.xaxis.set_major_formatter(major_formatter)
    self.subplot.axes.yaxis.set_major_formatter(major_formatter)
    
    fp = font.FontProperties(size=axeslabel_fontsize)

    self.subplot.axes.set_xlabel(U(self.xlabel), fontproperties=fp)
    
    ylabel = self.ylabel
    padding = 10 if len(ylabel) > 40 else None
    self.subplot.axes.set_ylabel(textwrap.fill(U(ylabel), 40, break_long_words=False),
                                 fontproperties=fp,
                                 horizontalalignment='center',
                                 labelpad=padding)

    for label in self.subplot.axes.get_yticklabels():
#        label.set_color(self.params['ytickscolor'])
        label.set_fontsize(axesticks_fontsize)
#        label.set_rotation(self.params['yticksangle'])

    # properties of labels on the x axis
    for label in self.subplot.axes.get_xticklabels():
#        label.set_color(self.params['xtickscolor'])
        label.set_fontsize(axesticks_fontsize)
#        label.set_rotation(self.params['xticksangle'])

    # legend
    self.subplot.legend(loc=0)   #4: left lower; 0: best
    try:
        lg = self.subplot.get_legend()
        ltext = lg.get_texts()             # all the text.Text instance in the legend
        for txt in ltext:
            txt.set_fontsize(legend_fontsize)  # the legend text fontsize
        # legend line thickness
        llines = lg.get_lines()             # all the lines.Line2D instance in the legend
        for lli in llines:
            lli.set_linewidth(1.5)          # the legend linewidth
        # color of the legend frame
        # this only works when the frame is painted (see below draw_frame)
        frame = lg.get_frame()             # the patch.Rectangle instance surrounding the legend
        frame.set_facecolor('#F0F0F0')      # set the frame face color to light gray
        # should the legend frame be painted
        lg.draw_frame(False)
    except:
        # no legend
        pass

    lines = [line for line in self.axvLines if line >= self.xmin and line <= self.xmax]
    for xval in self.axvLines:
        self.subplot.axvline(x=xval)

    self.canvas.draw()
    self.canvas.flush_events()



class XYPlot(wx.Panel):

    def __init__(self, parent, style=wx.SUNKEN_BORDER, size=wx.Size(800, 600)):
        super(wx.Panel, self).__init__(parent, style=style, size=size)
        
        # remap drawing methods to the wx widgets.
        paramList = {'backcolor'   : GRAPH_BACKGROUND_COLOR} # graph background color

        Mp.MatPanel(self, wx.Panel, drawFigure)
        self.subplot = None

        # default values for axis-limits
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None

        self.axvLines = []
        self.xTicks = None
        self.minorXTicks = None

        self.grid = True

    def xAxis(self, label="x", xmin=None, xmax=None):
        """ sets Parameters of x-Axis
        """
        self.xlabel = label
        self.xmin = xmin
        self.xmax = xmax
    
    def yAxis(self, label="y", ymin=None, ymax=None):
        """ sets Parameters of y-Axis
        """
        self.ylabel = label
        self.ymin = ymin
        self.ymax = ymax

    def xData(self, data):
        self.xdata = data

    def yData(self, data, labels=None):
        self.ydata = data

        if labels is not None:
            self.dataLabel = labels
        else:
            self.dataLabel = ["Row %s"%(i+1) for i in range(len(self.ydata))]

    def setAxvLines(self, xvals):
        """ sets vertical lines within the plot """
        self.axvLines = xvals

    def setXTicks(self, xvals):
        """ sets vertical lines within the plot """
        self.xTicks = xvals

    def setMinorXTicks(self, xvals):
        """ sets vertical lines within the plot """
        self.minorXTicks = xvals

    def SetValue(self, val):
        pass

    
