# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#    scheduleEditor: Editor for time schedules
#
#==============================================================================
#
#    Version No.: 0.01
#    Created by:         Hans Schweiger 07/07/2011
#
#------------------------------------------------------------------------------
#    (C) copyleft energyXperts.BCN (E4-Experts SL), Barcelona, Spain 2011-2012
#    http://www.energyxperts.net/
#
#    This program is free software: you can redistribute it or modify it under
#    the terms of the GNU general public license as published by the Free
#    Software Foundation (www.gnu.org).
#
#==============================================================================
import wx
from GUI.status import *
from auxiliary.auxiliary import toUnicode as U, toGettextAndUnicode as _U
from GUI.displayClasses import FloatEntry
from GUI.dataEntry.tableEntry import TableEntry
from scheduleViewer import PeriodView, ScheduleView
from profileSelector import ProfileSelector
from modules.guiModules.moduleScheduleEditor import ModuleScheduleEditor

class ScheduleEditor(wx.Panel):
    def __init__(self, prnt, mod=None, changeData=None):
        """
        mod: ModuleScheduleEditor object if created before. Otherwise it will be created here
        changeData: eventHandler in the case of data change.
        """
        self.prnt = prnt
        self.changeData = changeData
        
        wx.Panel.__init__(self, id=wx.ID_ANY, name='', parent=prnt,
            style=wx.NO_BORDER | wx.TAB_TRAVERSAL,
            pos = wx.DefaultPosition)

        if mod is not None:
            self.mod=mod
        else:
            self.mod = ModuleScheduleEditor(self)

        self._init_ctrls(prnt)
        self.__do_layout()
        self.display()

    def _init_ctrls(self, prnt):
        # generated method, don't edit
        titles = [_U("start"), _U("stop"), _U("scale"), _U("profiles")]
        self.entryPeriods = TableEntry(self, titles, self.mod.changeData, size=wx.Size(604, 230),
                                       hAlignment = [wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_CENTRE, wx.ALIGN_LEFT],
                                       readOnly = [False, False, False, True],)
        self.entryPeriods.setColSize(3, 300)
        self.entryPeriods.setExternalEventHandler(self.eventHandler)

        self.entryHolidayScaling = FloatEntry(self,
                               decimals = 3, minval = 0, maxval = 1, value = 0.,
                               unitdict = 'FRACTION',
                               label = _U("Scaling general holidays"),
                               tip = _U("Rate of operation of equipment during general holidays"))
        self.entryHolidayScaling.Bind(wx.EVT_KILL_FOCUS, self.OnDataEntry)

        self.periodView = PeriodView(self, size=wx.Size(380, 200))
        self.scheduleView = ScheduleView(self, size=wx.Size(380, 200))
            

    def __do_layout(self):
        flagText = wx.ALIGN_CENTER_VERTICAL|wx.TOP

        # global sizer for panel.
        sizerGlobal = wx.BoxSizer(wx.VERTICAL)
        sizerGlobal.Add(self.entryPeriods, 1, wx.ALIGN_LEFT, 0)
        sizerGlobal.AddSpacer(10)
        sizerGlobal.Add(self.entryHolidayScaling, 0, wx.ALIGN_LEFT, 0)
        sizerGlobal.AddSpacer(10)

        sizerViews = wx.BoxSizer(wx.HORIZONTAL)
        sizerViews.Add(self.periodView, 1, wx.ALIGN_CENTRE, 8)
        sizerViews.AddSpacer(4)
        sizerViews.Add(self.scheduleView, 1, wx.ALIGN_CENTRE, 8)

        sizerGlobal.Add(sizerViews, 0, wx.ALIGN_LEFT, 0)
        self.SetSizer(sizerGlobal)

    def display(self):
        if self.mod.scheduleType=="detailed":
            
            data = self.mod.getTable()
           
            self.entryPeriods.display(data)

            self.entryHolidayScaling.SetValue(self.mod.schedule.holidayScale)

            PVf, PVt = self.mod.getPeriodView()
            self.periodView.display(data=([PVt], [PVf]))

            SVf, SVt = self.mod.getScheduleView()
            xmin, xmax = self.mod.defaultPlotRange
            self.scheduleView.display(data=([SVt], [SVf]), xmin=xmin, xmax=xmax)
##        else:
##            print "ScheduleEditor.display: Simple Schedule"


    def load(self, objectId, objectName, createNew=False):
        """ loads a schedule from the dataBase
        """
        selectedProfile = None
        if createNew:
            ps = ProfileSelector(self)
            ps.display(data=self.mod.dbProfileTable, selection=self.mod.getProfileSelection())
            ret = ps.ShowModal()
            if ret == wx.ID_OK:
                selectedProfile = self.selection

        self.mod.load(objectId, objectName, createNew, defaultProfile=selectedProfile)
        
    def save(self, ):
        """ saves the data edited
        """
        self.mod.schedule.holidayScale = self.entryHolidayScaling.GetValue()
        
    def eventHandler(self, eventType, r, c, selection):
        """ Handling of events in the period editing table:

        Click on profile-column -> opens profile selector        
        """
##        print "ScheduleEditor.eventHandler: eventType %s (%s, %s) %s"%(eventType, r, c, selection)
        
        if eventType in ["LeftClick", "LeftDclick"]:
            self.blockEventHandler = True
            if c==3:
                if self.mod.setActivePeriod(r):
                    ps = ProfileSelector(self)
                    ps.display(data=self.mod.dbProfileTable, selection=self.mod.getProfileSelection())

                    ret = ps.ShowModal()
                    if ret == wx.ID_OK:
                        self.mod.selectProfiles(self.selection)
                        self.display()
            else:
                if self.mod.setActivePeriod(r):
                    SVf, SVt = self.mod.getScheduleView()
                    xmin, xmax = self.mod.defaultPlotRange
                    self.scheduleView.display(data=([SVt], [SVf]), xmin=xmin, xmax=xmax)

            self.blockEventHandler = False

    def OnDataEntry(self, event):
        self.mod.schedule.holidayScale = self.entryHolidayScaling.GetValue()
        self.display()
        
