# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	AUXILIARY - Auxiliary functions (general)
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Module for calculation of energy statistics
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from math import floor, ceil
import bisect
import numpy as np

def almostEqual(a, b, tol):
    if a is None or b is None:
        return False
    else:
        return abs(a - b) <= tol

#TODO: eliminate all the references to text conversion functions here
# maintained only for backward compatibility
def toGettext(text):
    return _(text)

def toGettextAndUnicode(text):
    try:
        return unicode(_(text), "utf-8")
    except:
        return _(text)
    
def toUnicode(text):
    try:
        return unicode((text), "utf-8")
    except:
        return (text)

def toUTF8(text):
    try:
        return text.encode("utf-8")
    except:
        return "---"


def findKey(dict, val, default=None):
    """ returns the key corresponding to a value in a dictionary
    """
    
    if val in dict.values():
        index = dict.values().index(val)
        return dict.keys()[index]
    else:
        return default

#------------------------------------------------------------------------------		
#   2. Table look-up and interpolation routines
#       - lists: lists of values ylist[i]
#       - tables: pairs of values y-x in lists ylist[i],xlist[i]
#------------------------------------------------------------------------------		

#------------------------------------------------------------------------------
def lastNonZero(ylist):
#------------------------------------------------------------------------------
#finds the last filled position of the list (different from zero)
#------------------------------------------------------------------------------
    n = len(ylist)
    idx = 0
    for i in range(n - 1, -1, -1):
        if ylist[i] <> 0.0:
            idx = i
            break
    return idx
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
def firstNonZero(ylist):
#------------------------------------------------------------------------------
#finds the first filled position of the list (different from zero)
#------------------------------------------------------------------------------
    i = 0
    n = len(ylist)
    idx = n
    for i in range(n):
        if ylist[i] <> 0:
            idx = i
            break
    return idx
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
def findInListASC(y, ylist):
#------------------------------------------------------------------------------
#   selects the table entry with a value closest, but lower to val.
#   returns the index
#------------------------------------------------------------------------------
    return bisect.bisect_right(ylist, y, hi=len(ylist) - 1) - 1
        
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
def interpolateList(xi, ylist):
    """
    #   interpolates the y-values in a table y(x) for non-integer values of x0
    """
    if xi <= 0:
        return ylist[0]
    elif xi >= len(ylist)-1:
        return ylist[-1]
    else:
        i0 = int(xi)
        f = xi-i0
        y0 = ylist[i0]
        y1 = ylist[i0+1]
        return y0 + f*(y1-y0)
    
def interpolateTable(x, xlist, ylist):
    """Return the interpolated table values ylist(xlist) for values of 
    x not in xlist. It only works for montonous curves 
    (both ascending and descending)

    if x is out of the bounds, the maximum or minimum value of ylist will be returned

    Note: search in ascending xlist returns first occurency of x from the left,
          search in descending xlist returns first occurency of x from the right
    
    """
## TODO: np.interp would be fine, but handling of descending lists is different; for flat curves seems to behave like bisect_right
##    return np.interp(x, np.array(xlist, dtype=float), np.array(ylist, dtype=float))
        
    n = len(xlist)
    
    if xlist[0] < xlist[-1]: # ascending list
        i = bisect.bisect_left(xlist, x, hi=n - 1) - 1

        if x < xlist[0] or i < 0:
            y = ylist[0]
            
        elif x > xlist[-1] or i>n-2:
            y = ylist[n - 1]

        else:
            y = (ylist[i]*(xlist[i + 1] - x) + ylist[i + 1]*(x - xlist[i])) / (xlist[i + 1] - xlist[i])
            
            
    else: # descending list
        i = n - bisect.bisect_left(xlist[::-1], x, hi=n - 1) - 1

        if x > xlist[0] or i < 0:
            y = ylist[0]
            
        elif x < xlist[-1] or i>n-2:
            y = ylist[n - 1]
            
        else:
            y = (ylist[i]*(xlist[i + 1] - x) + ylist[i + 1]*(x - xlist[i])) / (xlist[i + 1] - xlist[i])
            
    return y

def transpose(T):
    """ Transposes a matrix (list of lists) interchanging rows with columns
    """
    invT = []

    for j in range(len(T[0])):
        row = []
        for i in range(len(T)):
            row.append(T[i][j])
        invT.append(row)

    return invT

def findFirstGE(val, ASClist):
    """Return the index of the first higher value than val in the
    ascending values list ASClist. Return len(ASClist) if val > max(ASClist)
    ASCList[result] >= val

    if val <= min(list) -> result = 0
    if val >= max(list) -> result = ?
    
    """
    return bisect.bisect_left(ASClist, val)

def findFirstGT(val, ASClist):
    """ like findFirstGE, but looks ASClist[result] > val
    """
    return bisect.bisect_right(ASClist, val)
    

#------------------------------------------------------------------------------		
#   3. Others
#------------------------------------------------------------------------------		
def checkLimits(x, xmin, xmax, default=None):   
    """Checks if a value is within limits
    If no value is given, the default value is assigned
    
    :param x: the value
    :param xmin: the lower limit
    :param xmax: the upper limit
    :param default: the default value for invalid x

    """
    if x is None:
        return default
    xl = min(x, xmax)
    xl = max(xl, xmin)
    return xl

def noneFilter(datalist, substitute=" "):   
    return [noneFilterNumber(item, substitute) for item in datalist]

def noneFilterNumber(val, substitute=0.0):   
    if val is None:
        val = substitute
    return val

def noneCheck(val, substitute):
    if val is None:
        return substitute
    else:
        return val

def toFloat(val):
    """
    converts ISO31-0 strings to a floating point number:
    
    - comma or dot as decimal separator
    - space as optional thousands separator
    """

    if isinstance(val, list):
        return [toFloat(elem) for elem in val]
    else:
        try:
            return float(val)
        except:
            try:
                return float(val.replace(",",".").replace(" ",""))
            except:
                return None

def scale(val, fscale):
    try:
        x = float(val) * fscale
    except:
        x = val
    return x

def cutInterval(x, x0, x1):
    """return the proportional part of the interval [x0,x1] or [x1,x0] 
    that is below x
    
    """
    if x0 is None or x1 is None:
        return 0

    xl = min(x0, x1)
    xh = max(x0, x1)
    
    if x <= xl:
        return 0.0
    elif x <= xh and xh > xl:
        return (x - xl) / (xh - xl)
    else:
        return 1.0

def normalize(v):
    """Return a normalised numpy array v, such that the sum over all 
    entries equals 1.
    
    """
    n = v.sum()
    if n == 0.:
        return v.copy()
    else:
        return v / n

def reorder(pos, pos1, n):
    """reorders a list of items with original no. i = 0 - n-1
        into a new numbering j = 0 - n-1
        top = 0
        bottom = n-1

    Used in several places for re-ordering equipment numbers etc.

    :returns newNo: list for lookup of new numbers j = newNo[i}
    :returns oldNo: list for lookup of old numbers i = oldNo[j]

    """
    oldNo = list(range(n))
    newNo = list(range(n))

    if pos < 0 or pos >= n or pos is None:
        return newNo, oldNo

    if pos1 < 0 or pos1 >= n or pos1 is None:
            return newNo, oldNo

    if pos1 < pos: #moving up
        oldNo[pos1] = pos
        for k1 in range(pos1+1,pos+1):
            oldNo[k1] -= 1
    elif pos1 > pos: #moving down
        oldNo[pos1]= pos
        for k1 in range(pos,pos1):
            oldNo[k1] += 1

    for k1 in range(n):
        k0 = oldNo[k1]
        newNo[k0] = k1
        
    return newNo, oldNo

def nanmean(x):
    suma = np.sum(np.nan_to_num(x))
    n = np.sum(np.isnan(x)==False)
    return suma/n if n > 0 else None
#-------------------------------------
# some unit tests

def test_interpolateTable():

    import numpy as np
    Q = np.array([0, 0, 1, 1, 2, 2], dtype=float)
    QDesc = np.array([2, 2, 1, 1, 0, 0], dtype=float)
    T = np.linspace(0, 10, 6)
    print "Q = %s"%Q
    print "T = %s"%T

    values = [-0.5, 0, 0.5, 1, 1.5, 2, 2.5]

    print "\nTesting interpolateTable(val, Q, T) - ascending Q"
    targets = [0, 0, 3, 4, 7, 8, 10]
    for i, val in enumerate(values):
        target = targets[i]
        res = interpolateTable(val, Q, T)
        diff = res - target
        print "val: %10.2f\tres %10.2f\ttarget %10.2f\tdiff %10.2f"%(val, res, target, diff)

    print "\nTesting interpolateTable(val, Q, T) - descending Q"
    targets = [10, 10, 7, 6, 3, 2, 0]
    for i, val in enumerate(values):
        target = targets[i]
        res = interpolateTable(val, QDesc, T)
        diff = res - target
        print "val: %10.2f\tres %10.2f\ttarget %10.2f\tdiff %10.2f"%(val, res, target, diff)

    print "\nTesting interpolateTable(val, Q, T) - ascending Q - negative "
    targets = [0, 2, 3, 6, 7, 10, 10]
    for i, val in enumerate(values):
        target = targets[i]
        res = interpolateTable(-val, -Q, T)
        diff = res - target
        print "val: %10.2f\tres %10.2f\ttarget %10.2f\tdiff %10.2f"%(val, res, target, diff)


        
if __name__=="__main__":
    test_interpolateTable()
    
    
        
