# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	DBIO
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	SQL Bridge classes for access to database
#       Specific functions for implementation of SQLObject / SQLite
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import DBSERVER, APPDATA
if DBSERVER == "SQLITE":
    import sqlite3
else:
    import MySQLdb
    
from sqlobject import *
from math import isnan, isinf
import numpy

SQLOBJECTWRAPPER = "SQLOBJECT"
from GUI.status import Status
import os

from auxiliary.auxiliary import toUnicode as U
from modules.messageLogger import logWarning

MYSQLKEYS = {"auditor": "Auditor_ID",
             "dbfluid": "DBFluid_ID",
             "dbfuel": "DBFuel_ID",
             "dbelectricitymix": "DBElectricityMix_ID"}

#==============================================================================
# SQL Bridge
#
# Generic functions independent of DBServer (MySQL or SQLite) and of
# SQL-Object-Wrapper (pSQL or SQLObject)
#
#==============================================================================

def getTableClass(table):
    """ auxiliary function in order to make dbio functions accessible for tables
    defined outside from this package
    """
    if isinstance(table, str):
        try:
            return eval(table)
        except NameError:
            return None
    else:
        return table

def select(table, **criteria):
    """selects a list of rows from an SQL table fulfilling a set of criteria

    :params table: name of the sql - table
    :params criteria: dictionary of COLNAME:COLVALUE pairs
                        empty dictionary: the whole table is selected
    """

    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        rows = list(tableClass.selectBy(**criteria))

        return rows
    else:
        return None
        
            

def selectByQuery(table, query):
    """selects a list of rows from an SQL table using an explicit SQL query

    :params table: name of the sql - table
    :params criteria: dictionary of COLNAME:COLVALUE pairs
                        empty dictionary: the whole table is selected
    """

    if DBSERVER == "MYSQL" and table in MYSQLKEYS.keys():
        key = MYSQLKEYS[table]
        query = query.replace("id IN", key+" IN")
    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        rows = list(tableClass.select(query))

        return rows
    else:
        return None
        
def selectByList(table, par, valList):
    query = "%s IN (%s)" % (par, ','.join([str(val) for val in valList]))
    return selectByQuery(table, query)
                            

def delete(table, **criteria):
    """selects a list of rows from an SQL table fulfilling a set of criteria
    and then deletes them

    :params table: name of the sql - table
    :params criteria: dictionary of COLNAME:COLVALUE pairs
                        empty dictionary: the whole table is selected
    """

    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        tableClass.deleteBy(**criteria)
    else:
        pass
            

def deleteMany(table, query):
    """selects a list of rows from an SQL table fulfilling a set of criteria
    and then deletes them

    :params table: name of the sql - table
    :params criteria: dictionary of COLNAME:COLVALUE pairs
                        empty dictionary: the whole table is selected
    """

    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        tableClass.deleteMany(query)
    else:
        pass

def deleteByQuery(table, query):
    """selects a list of rows from an SQL table fulfilling a set of criteria
    and then deletes them

    :params table: name of the sql - table
    :params criteria: dictionary of COLNAME:COLVALUE pairs
                        empty dictionary: the whole table is selected
    """

    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        tableClass.delete(query)
    else:
        pass

def deleteByList(table, par, valList):
##    query = "%s IN (%s)" % (par, ','.join([str(val) for val in valList]))
##    return deleteByQuery(table, query)
    for val in valList:
        parDict = {str(par): val}
        delete(table, **parDict)
            

def get(table, **criteria):
    """ like select, but returns only one (the first) row of possible fits,
    or None, if there is no match
    """

    rows = select(table, **criteria)
    if len(rows) > 0:
        return rows[0]
    else:
        return None

def insertRow(table, **data):
    """ inserts a new row in the data table and returns its id
    """
    if SQLOBJECTWRAPPER=="SQLOBJECT":
        tableClass = getTableClass(table)
        if tableClass is None: return []

        new = tableClass(**data)
        return new.id
    
    else:
        pass
        return None

def commit():
    if SQLOBJECTWRAPPER=="SQLOBJECT":
        if DBSERVER=="MYSQL":
            return
        #TODO: there are problems with the commit command for the MySQLConnection
        
        sqlhub.processConnection.commit()
    else:
        Status.SQL.commit()
    

def check(value, length=None, replacement=None, string=True, forceType=None):
    """Substitute "", None and NaN by 'NULL' and truncate to length, if 
    it's a string longer than length and length is given.
    
    You may replace "NULL" by any string you like via the replacement 
    parameter. 

    (should be moved some day to a separate file with sql-tools ...)

    :params string: if False, forces conversion of string to float number, otherwise returns None
                    equivalent behaviour to forceType="float", but latter works for all input
                    data types
    :params forceType: if set to "int" or "float", a conversion to int/float is attempted, otherwise
                        the replacement is returned

    """
    if value != "" and value is not None:
        if forceType == "int":
            try:
                return int(value)
            except:
                return replacement
        elif forceType == "float":
            try:
                return float(value)
            except:
                return replacement
            
        if isinstance(value, str) or isinstance(value, unicode):
            if string:
                return U(value).encode("utf-8")[0:length]
            else:
                try:
                    return float(value)
                except:
                    return replacement
        else:
            try:
                if isnan(value):
                    logWarning("Check encountered a NaN value. This should not happen.")
                    return replacement
                elif isinf(value):
                    logWarning("Check encountered an inf value. This should not happen.")
                    return replacement
            except TypeError: # is not a number type checking for NaN is senseless
                pass
            if isinstance(value, numpy.float64):
                return float(value)
            else:
                return value
        
    else:
        return replacement

def copyRow(table, row):
    """ copies a row of a data table changing only the primary key 'id'
        returns the id of the new row

        TODO: there should be a more efficient way to do this !!!
    """
    rowvars = vars(row)
    rowdict = {}
    for varKey in rowvars.keys():
        if varKey.startswith("_SO_val_"):
            key = varKey[8:]
            rowdict.update({key: getattr(row, key)})
    return insertRow(table, **rowdict)

def getColInfo(table):

    tableClass = getTableClass(table)
    if tableClass is None: return []

    validatorKeys = [key for key in tableClass.__dict__.keys() if key.startswith('_SO_to_python_')]
    colInfo = [("id", "int")] # first entry is the primary key
    for key in validatorKeys:
        method = getattr(tableClass, key)
        validatorClass = method.__self__.__class__
        typeInfo = str(validatorClass).split(".")[2]
        
        if typeInfo.startswith("Int"):
            colType = "int"
        elif typeInfo.startswith("String"):
            colType = "char"
        elif typeInfo.startswith("Float"):
            colType = "float"
        elif typeInfo.startswith("Date"):
            colType = "date"
        elif typeInfo.startswith("Enum"):
            colType = "enum"
        else:
            colType = typeInfo[0:4]

        colKey = key[14:]
        ####type enum("heating", "cooling")

        colInfo.append((colKey, colType))

    return colInfo


def listToQueryString(lst):
    """
    helper function connecting a list of values into a string
    to be used in an SQL query: "for XY IN (%s)"
    """

    strList = ["'%s'"%elem for elem in lst]
    qStr = ",".join(strList)

    print qStr

    return qStr
    
#==============================================================================
#
#   EINSTEIN specific functions
#
#==============================================================================

class DBConnector():

    def __init__(self, DBHost, DBUser, DBPass, DBName):
        self.DBHost = DBHost
        self.DBUser = DBUser
        self.DBPass = DBPass
        self.DBName = DBName
        self.encoding = 'utf8'
    
    def close(self):
        sqlhub.processConnection.commit(close=True)
        
    def connect(self):
        """Return a connection handler to the database."""

        #----- Connect to the Database
        if DBSERVER == "SQLITE":
            self.filepath = os.path.abspath(APPDATA+"/"+self.DBName+".db")
            if os.name == "nt":
                fp = "/" + self.filepath.replace(":","|")  # windows modification required for sqlobject
            else:
                fp = self.filepath
            conn = connectionForURI("sqlite:"+fp, debug=False, autoCommit=False)
            conn.style = MixedCaseStyle()
            trans = conn.transaction()
            sqlhub.processConnection = trans # passing trans instead of conn directly is currently the only way to avoid auto-commits. the flag above in connectionForURI seems not to work.
        else:
            MySQLConnection = mysql.builder()
            conn = MySQLConnection(host=self.DBHost, user=self.DBUser, password=self.DBPass, db=self.DBName)
            conn.style = MixedCaseStyle()
            conn.dbEncoding = "utf8"
            sqlhub.processConnection = conn
##            try:
##                conn = MySQLdb.connect(host=self.DBHost, user=self.DBUser, passwd=self.DBPass, db=self.DBName)
##            except MySQLdb.Error, e:
##                self.error = e

##        if conn:
##            # set mysql encoding translation
##            if DBSERVER == "SQLITE":
##                pass
##            else:
##                try:
##                    cursor = conn.cursor(MySQLdb.cursors.DictCursor)
##                    cursor.execute("SET NAMES '%s'" % self.encoding)
##                except MySQLdb.Error, e:
##                    self.doLog.LogThis('Could not set encoding %s: %s' % (self.encoding, str(e)))
##                    self.error = e

        if self.testConnection():
            return conn
        else:
            return None

    def testConnection(self):
        """ In SQLite, connection for URI does not give an error even if the filpath is not found
        Therefore, the test if everything is ok is the existence of an stool - table in the db
        """
        self.error = "---"
        try:
            stool = get("stool")
            return True
        except:
            if DBSERVER=="SQLITE":
                self.error = _("%s not found or not a valid EINSTEIN database"%self.filepath)
            return False
        
##    def createTables(self):
##        print "DBIO.__init__: defining stool-class"
##        self.stool = stool()
##        print "DBIO.__init__: end ================"
 
#==============================================================================
#
#   Representation of EINSTEIN SQL-Tables as SQLObject - Classes
#
#==============================================================================

class auditor(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "Auditor_ID" if DBSERVER == "MYSQL" else "id"

    Name = StringCol(length=45, default=None)
    City = StringCol(length=45, default=None)
    Country = StringCol(length=45, default=None)
    Company = StringCol(length=45, default=None)
    CompanyType = StringCol(length=45, default=None)
    Adress = StringCol(length=100, default=None)
    Phone = StringCol(length=45, default=None)
    Fax = StringCol(length=45, default=None)
    EMail = StringCol(length=45, default=None)
    Web = StringCol(length=45, default=None)

class cgeneraldata(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "CGeneralData_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    OMThermal = FloatCol(default=None)
    OMElectrical = FloatCol(default=None)
    HPerDayInd = FloatCol(default=None)
    HPerYearInd = FloatCol(default=None)
    NShifts = FloatCol(default=None)
    NDaysInd = FloatCol(default=None)
    NoProdStart = StringCol(length=6, default=None)
    NoProdStop = StringCol(length=6, default=None)
    PercentElTotcost = FloatCol(default=None)
    PercentFuelTotcost = FloatCol(default=None)
    InflationRate = FloatCol(default=None)
    FuelPriceRate = FloatCol(default=None)
    InterestExtFinancing = FloatCol(default=None)
    CompSpecificDiscountRate = FloatCol(default=None)
    PercentExtFinancing = FloatCol(default=None)
    AmortisationTime = FloatCol(default=None)
    PublicFundType = StringCol(length=45, default=None)
    EnergyCost = FloatCol(default=None)
    EnergySystemCost = FloatCol(default=None)
    AddCost = FloatCol(default=None)
    AddCostperSavedPE = FloatCol(default=None)
    TotalInvCost = FloatCol(default=None)
    InvestmentEffective = FloatCol(default=None)
    RevenueReplaceEquipment = FloatCol(default=None)
    OwnInvCost = FloatCol(default=None)
    Subsidies = FloatCol(default=None)
    RevenueSaleEquipment = FloatCol(default=None)
    Amortization = FloatCol(default=None)
    IRR = FloatCol(default=None)
    BCR = FloatCol(default=None)
    PayBack = FloatCol(default=None)
    NPV = FloatCol(default=None)
    NEquipe = IntCol(default=None)
    NThProc = IntCol(default=None)
    NPipeDuct = IntCol(default=None)
    NProducts = IntCol(default=None)
    NHX = IntCol(default=None)
    NWHEE = IntCol(default=None)
    NFuels = IntCol(default=None)
    PETOTAL_EXT = FloatCol(default=None)
    PE_INT = FloatCol(default=None)
    EL_OWN = FloatCol(default=None)
    EL_OWN_EXT = FloatCol(default=None)
    EL_INT = FloatCol(default=None)
    FUEL_OWN_EXT = FloatCol(default=None)
    FUEL_INT = FloatCol(default=None)
    PEConvEl = FloatCol(default=None)
    NoNukesConvEl = FloatCol(default=None)
    CO2ConvEl = FloatCol(default=None)
    PEConvFuelMix = FloatCol(default=None)
    CO2ConvFuelMix = FloatCol(default=None)
    PEConvFuelCHP = FloatCol(default=None)
    CO2ConvFuelCHP = FloatCol(default=None)
    FuelTariffMix = FloatCol(default=None)
    FuelTariffCHP = FloatCol(default=None)
    USH = FloatCol(default=None)
    UPH = FloatCol(default=None)
    USC = FloatCol(default=None)
    UPC = FloatCol(default=None)
    HDEffAvg = FloatCol(default=None)
    FET = FloatCol(default=None)
    FETel = FloatCol(default=None)
    FEC = FloatCol(default=None)
    FECel = FloatCol(default=None)
    ElectricityGen = FloatCol(default=None)
    ElectricitySales = FloatCol(default=None)
    FEO = FloatCol(default=None)
    FEOel = FloatCol(default=None)
    PEC = FloatCol(default=None)
    PET = FloatCol(default=None)
    PETel = FloatCol(default=None)
    PETFuels = FloatCol(default=None)
    PECel = FloatCol(default=None)
    PECFuels = FloatCol(default=None)
    PEO = FloatCol(default=None)
    ProdNoNukesEl = FloatCol(default=None)
    ProdCO2el = FloatCol(default=None)
    ProdCO2Fuels = FloatCol(default=None)
    WaterConsum = FloatCol(default=None)

class dbboiler(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBBoiler_ID" if DBSERVER == "MYSQL" else "id"
        
    BoilerManufacturer = StringCol(length=45, default=None)
    BoilerModel = StringCol(length=45, default=None)
    BoilerType = StringCol(length=45, default=None)
    Reference = StringCol(length=200, default=None)
    BBPnom = FloatCol(default=None)
    Economiser = StringCol(length=45, default=None)
    Preheater = StringCol(length=45, default=None)
    ExcessAirRatio = FloatCol(default=None)
    BBEfficiency = FloatCol(default=None)
    FuelConsum = FloatCol(default=None)
    FuelType = StringCol(length=45, default=None)
    ElConsum = FloatCol(default=None)
    BoilerTemp = FloatCol(default=None)
    BBA1 = FloatCol(default=None)
    BBA2 = FloatCol(default=None)
    BBK1 = FloatCol(default=None)
    BBK2 = FloatCol(default=None)
    BoilerPrice = FloatCol(default=None)
    BoilerTurnKeyPrice = FloatCol(default=None)
    BoilerOandMfix = FloatCol(default=None)
    BoilerOandMvar = FloatCol(default=None)
    YearUpdate = IntCol(default=None)

class dbbenchmark(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBBenchmark_ID" if DBSERVER == "MYSQL" else "id"

    NACECode = StringCol(length=45, default=None)
    UnitOp = StringCol(length=45, default=None)
    ProductCode = StringCol(length=45, default=None)
    Product = StringCol(length=45, default=None)
    ProductUnit = StringCol(length=45, default=None)
    E_EnergyInt_MIN_PC = FloatCol(default=None)
    E_EnergyInt_MAX_PC = FloatCol(default=None)
    E_EnergyInt_TARG_PC = FloatCol(default=None)
    E_EnergyInt_MIN_T = FloatCol(default=None)
    E_EnergyInt_MAX_T = FloatCol(default=None)
    E_EnergyInt_TARG_T = FloatCol(default=None)
    E_SEC_MIN = FloatCol(default=None)
    E_SEC_MAX = FloatCol(default=None)
    E_SEC_TARG = FloatCol(default=None)
    E_Unit = StringCol(length=45, default=None)
    H_EnergyInt_MIN_PC = FloatCol(default=None)
    H_EnergyInt_MAX_PC = FloatCol(default=None)
    H_EnergyInt_TARG_PC = FloatCol(default=None)
    H_EnergyInt_MIN_T = FloatCol(default=None)
    H_EnergyInt_MAX_T = FloatCol(default=None)
    H_EnergyInt_TARG_T = FloatCol(default=None)
    H_SEC_MIN = FloatCol(default=None)
    H_SEC_MAX = FloatCol(default=None)
    H_SEC_TARG = FloatCol(default=None)
    H_Unit = StringCol(length=45, default=None)
    T_EnergyInt_MIN_PC = FloatCol(default=None)
    T_EnergyInt_MAX_PC = FloatCol(default=None)
    T_EnergyInt_TARG_PC = FloatCol(default=None)
    T_EnergyInt_MIN_T = FloatCol(default=None)
    T_EnergyInt_MAX_T = FloatCol(default=None)
    T_EnergyInt_TARG_T = FloatCol(default=None)
    T_SEC_MIN = FloatCol(default=None)
    T_SEC_MAX = FloatCol(default=None)
    T_SEC_TARG = FloatCol(default=None)
    T_Unit = StringCol(length=45, default=None)
    Comments = StringCol(default=None)
    YearReference = IntCol(default=None)
    Reference = StringCol(default=None)
    Literature = StringCol(default=None)
    DataRelevance = StringCol(length=200, default=None)
    TurnoverMin = FloatCol(default=None)
    TurnoverMax = FloatCol(default=None)
    ProductionMin = FloatCol(default=None)
    ProductionMax = FloatCol(default=None)
    ProductionUnit = FloatCol(default=None)

class dbchp(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBCHP_ID" if DBSERVER == "MYSQL" else "id"
        
    Manufacturer = StringCol(length=45, default=None)
    CHPequip = StringCol(length=45, default=None)
    Type = StringCol(length=45, default=None)
    SubType = StringCol(length=45, default=None)
    Reference = StringCol(length=200, default=None)
    CHPPe = FloatCol(default=None)
    CHPPt = FloatCol(default=None)
    FuelType = StringCol(length=45, default=None)
    FuelConsum = FloatCol(default=None)
    Eta_e = FloatCol(default=None)
    FluidSupply = IntCol(default=None)
    TSupply = FloatCol(default=None)
    FlowRateSupply = FloatCol(default=None)
    FluidSupply2 = IntCol(default=None)
    TSupply2 = FloatCol(default=None)
    FlowRateSupply2 = FloatCol(default=None)
    Price = FloatCol(default=None)
    Eta_t = FloatCol(default=None)
    InvRate = FloatCol(default=None)
    OMRateFix = FloatCol(default=None)
    OMRateVar = FloatCol(default=None)
    YearUpdate = IntCol(default=None)
    
class dbelectricitymix(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBElectricityMix_ID" if DBSERVER == "MYSQL" else "id"
    
    Year = IntCol(default=None)
    ShortName = StringCol(length=45, default=None)
    Country = StringCol(length=45, default=None)
    Reference = StringCol(length=200, default=None)
    AuditorID = IntCol(default=None)
    Type = StringCol(length=20, default=None)
    PE2ConvEl = FloatCol(default=None)
    CO2ConvEl = FloatCol(default=None)
    NoNukesConvEl = FloatCol(default=None)
    PercNaturalGas = FloatCol(default=None)
    PercCarbon = FloatCol(default=None)
    PercOil = FloatCol(default=None)
    PercRenewables = FloatCol(default=None)
    PercNukes = FloatCol(default=None)
    PercCHP = FloatCol(default=None)
    PercOther = FloatCol(default=None)

class dbfluid(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBFluid_ID" if DBSERVER == "MYSQL" else "id"
##    class sqlMeta:
##        fromDatabase = True

    FluidName = StringCol(length=45, default=None)
    FluidType = StringCol(length=20, default=None)
    FluidCp = FloatCol(default=None)
    VaporCp = FloatCol(default=None)
    FluidDensity = FloatCol(default=None)
    SensibleHeat = FloatCol(default=None)
    LatentHeat = FloatCol(default=None)
    Viscosity = FloatCol(default=None)
    Conductivity = FloatCol(default=None)
    RefrigerantCode = StringCol(length=6, default=None)
    FluidDataSource = StringCol(length=200, default=None)
    SpecificMassFlow = FloatCol(default=None)
    THighP = FloatCol(default=None)
    TCond = FloatCol(default=None)
    FluidCpG = FloatCol(default=None)
    FluidComment = StringCol(length=200, default=None)

class dbfuel(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBFuel_ID" if DBSERVER == "MYSQL" else "id"

    FuelName = StringCol(length=45, default=None)
    DBFuelUnit = StringCol(length=45, default=None)
    FuelType = StringCol(length=20, default=None)
    FuelLCV = FloatCol(default=None)
    FuelHCV = FloatCol(default=None)
    CO2ConvFuel = FloatCol(default=None)
    FuelDensity = FloatCol(default=None)
    OffgasHeatCapacity = FloatCol(default=None)
    OffgasDensity = FloatCol(default=None)
    Humidity = FloatCol(default=None)
    CombAir = FloatCol(default=None)
    Offgas = FloatCol(default=None)
    PEConvFuel = FloatCol(default=None)
    FuelDataSource = StringCol(length=200, default=None)
    FuelComment = StringCol(length=200, default=None)
    FuelCode = StringCol(length=45, default=None)

class dbheatpump(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBHeatPump_ID" if DBSERVER == "MYSQL" else "id"

    HPManData = StringCol(length=45, default=None)
    HPManufacturer = StringCol(length=45, default=None)
    HPModel = StringCol(length=45, default=None)
    DBFuel_id = IntCol(default=None)
    FuelType = StringCol(length=45, default=None)
    HPFuelConsum = FloatCol(default=None)
    HPUnitsFuelConsum = StringCol(length=45, default=None)
    HPYearManufact = FloatCol(default=None)
    HPType = StringCol(length=45, default=None)
    HPSubType = StringCol(length=45, default=None)
    Reference = StringCol(length=200, default=None)
    HPSourceSink = StringCol(length=45, default=None)
    HPAbsEffects = StringCol(length=45, default=None)
    HPAbsHeatMed = StringCol(length=45, default=None)
    HPWorkFluid = StringCol(length=45, default=None)
    HPCoolCap = FloatCol(default=None)
    HPCoolCOP = FloatCol(default=None)
    HPExCoolCOP = FloatCol(default=None)
    HPThCoolCOP = FloatCol(default=None)
    HPConstExCoolCOP = FloatCol(default=None)
    HPAbsTinC = FloatCol(default=None)
    HPCondTinC = FloatCol(default=None)
    HPGenTinC = FloatCol(default=None)
    HPEvapTinC = FloatCol(default=None)
    HPHeatCap = FloatCol(default=None)
    HPHeatCOP = FloatCol(default=None)
    HPExHeatCOP = FloatCol(default=None)
    HPThHeatCOP = FloatCol(default=None)
    HPConstExHeatCOP = FloatCol(default=None)
    HPAbsTinH = FloatCol(default=None)
    HPCondTinH = FloatCol(default=None)
    HPGenTinH = FloatCol(default=None)
    HPEvapTinH = FloatCol(default=None)
    HPLimDT = FloatCol(default=None)
    HPGenTmin = FloatCol(default=None)
    HPCondTmax = FloatCol(default=None)
    HPEvapTmin = FloatCol(default=None)
    HPElectConsum = FloatCol(default=None)
    HPPrice = FloatCol(default=None)
    HPTurnKeyPrice = FloatCol(default=None)
    HPOandMfix = FloatCol(default=None)
    HPOandMvar = FloatCol(default=None)
    HPYearUpdate = IntCol(default=None)

class dbnacecode(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBNaceCode_ID" if DBSERVER == "MYSQL" else "id"
    
    CodeNACE = StringCol(length=45, default=None)
    CodeNACEsub = StringCol(length=45, default=None)
    NameNACE = StringCol(length=200, default=None)
    NameNACEsub = StringCol(length=200, default=None)
    ProductName = StringCol(length=200, default=None)
    NationalCode = StringCol(length=45, default=None)
    NationalSubCode = StringCol(length=45, default=None)
    NameNationalCode = StringCol(length=200, default=None)
    NameNationalSubCode = StringCol(length=200, default=None)

class dbsolarthermal(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBSolarThermal_ID" if DBSERVER == "MYSQL" else "id"
            
    STManufacturer = StringCol(length=45, default=None)
    STModel = StringCol(length=45, default=None)
    STType = StringCol(length=45, default=None)
    STc0 = FloatCol(default=None)
    STc1 = FloatCol(default=None)
    STc2 = FloatCol(default=None)
    K50L = FloatCol(default=None)
    K50T = FloatCol(default=None)
    STLengthGross = FloatCol(default=None)
    STHeightGross = FloatCol(default=None)
    STAreaGross = FloatCol(default=None)
    STLengthAper = FloatCol(default=None)
    STHeightAper = FloatCol(default=None)
    STAreaAper = FloatCol(default=None)
    STAreaFactor = FloatCol(default=None)
    STPnomColl = FloatCol(default=None)
    STWeightFactor = FloatCol(default=None)
    STMassFlowRate = FloatCol(default=None)
    STUnitPrice300kW = FloatCol(default=None)
    STUnitTurnKeyPrice30kW = FloatCol(default=None)
    STUnitTurnKeyPrice300kW = FloatCol(default=None)
    STUnitTurnKeyPrice3000kW = FloatCol(default=None)
    STOMUnitFix = FloatCol(default=None)
    STYearUpdate = IntCol(default=None)
    STReference = StringCol(length=200, default=None)

class dbunitoperation(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "DBUnitOperation_ID" if DBSERVER == "MYSQL" else "id"
                
    UnitOperation = StringCol(length=45, default=None)
    UnitOperationCode = StringCol(length=45, default=None)
    UnitOperationDescrip = StringCol(length=200, default=None)

class heatexchanger_pinchstream(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    qheatexchanger_QHeatExchanger_Id = IntCol(default=None)
    pinchstream_id = IntCol(default=None)
    inletTemp = FloatCol(default=None)
    outletTemp = FloatCol(default=None)
    outletOfHX_id = IntCol(default=None)
    inletOfHX_id = IntCol(default=None)
    HeatFlowPercent = FloatCol(default=None)
    PercentAverage = FloatCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)

class intervals(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    start = TimeCol(default=None)
    stop = TimeCol(default=None)
    scale = FloatCol() #WARNING: in einstein-MySQL database this is IntCol !

class periods(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    start = DateCol(default=None)
    stop = DateCol(default=None)

class pinchstream(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    name = StringCol(length=255, default=None)
    Hot_Cold = StringCol(length=6, default=None)
    Type = StringCol(length=8, default=None)
    source_id = IntCol(default=None)
    source_type = StringCol(length=64, default=None)
    medium_id = IntCol(default=None)
    StartTemp = FloatCol(default=None)
    EndTemp = FloatCol(default=None)
    StreamType = StringCol(length=64, default=None)
    HeatCapacity = FloatCol(default=None)
    MassFlowNom = FloatCol(default=None)
    SpecHeatCapacity = FloatCol(default=None)
    SpecEnthalpy = FloatCol(default=None)
    EnthalpyNom = FloatCol(default=None)
    HeatTransferCoeff = FloatCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    pID = IntCol(default=None)

class process_period_profiles(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    process_periods_id = IntCol(default=None)
    profiles_id = IntCol(default=None)

class process_periods(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    step = IntCol(default=None)
    scale = FloatCol(default=None)
    periods_id = IntCol(default=None)
    qprocessdata_QProcessData_ID = IntCol(default=None)

class process_schedules(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    startHour = IntCol(default=None)
    stopHour = IntCol(default=None)
    scale = FloatCol(default=None)
    qprocessdata_QProcessData_ID = IntCol(default=None)

class process_streams_in(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    qprocessdata_QProcessData_Id = IntCol(default=None)
    streams_in_id = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)

class process_streams_out(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    qprocessdata_QProcessData_Id = IntCol(default=None)
    streams_out_id = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)

class profile_intervals(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    profiles_id = IntCol(default=None)
    intervals_id = IntCol(default=None)

class profiles(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    name = StringCol(length=255, default=None)
    monday = IntCol(default=None)
    tuesday = IntCol(default=None)
    wednesday = IntCol(default=None)
    thursday = IntCol(default=None)
    friday = IntCol(default=None)
    saturday = IntCol(default=None)
    sunday = IntCol(default=None)

class projectinfo(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "ProjectInfo_ID" if DBSERVER == "MYSQL" else "id"

    ProjectID = IntCol(default=None)
    ProjectCreationDate = DateCol(default=None)
    ProjectFinalisationDate = DateCol(default=None)
    LastModificationDate = DateCol(default=None)
    Auditor_ID = IntCol(default=None)
    OriginalProjectID = IntCol(default=None)
    Campaign = StringCol(length=45, default=None)
    ConfidentialityLevel = IntCol(default=None)
    ToolVersion = StringCol(length=45, default=None)

class preferences(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()

    Font_facename = StringCol(length=45, default=None)
    Font_family = IntCol(default=None)
    Font_style = IntCol(default=None)
    Font_weight = IntCol(default=None)
    Font_size = IntCol(default=None)
    Font_color = StringCol(length=16, default=None)
    user = StringCol(length=64, default=None)

class psetupdata(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "PSetUpData_ID" if DBSERVER == "MYSQL" else "id"

    ElectricityMix = IntCol(default=None)
    PEConvFuels = FloatCol(default=None)
    CO2ConvEl = FloatCol(default=None)
    HRInvRate = FloatCol(default=None)
    HROMRate = FloatCol(default=None)
    FeedInCHP = FloatCol(default=None)
    SolarInvRateL = FloatCol(default=None)
    SolarInvRateS = FloatCol(default=None)
    SolarOMRate = FloatCol(default=None)
    SolarFunding = FloatCol(default=None)
    QuSolar = FloatCol(default=None)
    Inflation = FloatCol(default=None)
    InflationFuels = FloatCol(default=None)
    InterestRate = FloatCol(default=None)
    ExternalFinance = FloatCol(default=None)
    Amortisation = FloatCol(default=None)
    Annuity = FloatCol(default=None)
    EEFunding = FloatCol(default=None)
    UHPType = StringCol(length=45, default=None)
    UHPMinHop = FloatCol(default=None)
    UHPDTMax = FloatCol(default=None)
    UHPTgenIn = FloatCol(default=None)
    UHPmaxT = FloatCol(default=None)
    UHPminT = FloatCol(default=None)

class qbuildings(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QBuildings_ID" if DBSERVER == "MYSQL" else "id"
        
    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    BuildingNo = IntCol(default=None)
    BuildName = StringCol(length=45, default=None)
    BuildConstructSurface = FloatCol(default=None)
    BuildUsefulSurface = FloatCol(default=None)
    BuildUsage = StringCol(length=45, default=None)
    BuildEnergyDemand = StringCol(length=45, default=None)
    BuildMaxHP = FloatCol(default=None)
    BuildMaxCP = FloatCol(default=None)
    BuildAnnualHeating = FloatCol(default=None)
    BuildAnnualAirCond = FloatCol(default=None)
    BuildTHeating = FloatCol(default=None)
    IndoorTHeating = FloatCol(default=None)
    BuildTAirCond = FloatCol(default=None)
    IndoorTAirCond = FloatCol(default=None)
    AirHandlingUnit = IntCol(length=45, default=None)
    AirFlowRateC = FloatCol(default=None)
    AirFlowRateH = FloatCol(default=None)
    TInflowRecC = FloatCol(default=None)
    TInflowRecH = FloatCol(default=None)
    BuildDailyDHW = FloatCol(default=None)
    BuildHoursOccup = FloatCol(default=None)
    BuildDaysInUse = FloatCol(default=None)
    BuildHolidaysPeriodStart_1 = DateCol(default=None)
    BuildHolidaysPeriodStart_2 = DateCol(default=None)
    BuildHolidaysPeriodStart_3 = DateCol(default=None)
    BuildHolidaysPeriodStop_1 = DateCol(default=None)
    BuildHolidaysPeriodStop_2 = DateCol(default=None)
    BuildHolidaysPeriodStop_3 = DateCol(default=None)
    BuildHeatingPeriodStart = DateCol(default=None)
    BuildHeatingPeriodStop = DateCol(default=None)
    BuildAirCondPeriodStart = DateCol(default=None)
    BuildAirCondPeriodStop = DateCol(default=None)
    EnclBuildDraw = StringCol(length=45, default=None)

    @property
    def QBuildings_ID(self):
        return self.id

class qcombschedules(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    ProcessId = IntCol(default=None)
    Weight = FloatCol(default=None)
    Type = IntCol(default=None)
    ObjectId = IntCol(default=None)
    ComponentNo = IntCol(default=None)

class qdistributionhc(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QDistributionHC_ID" if DBSERVER == "MYSQL" else "id"
    
    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    PipeDuctNo = IntCol(default=None)
    Pipeduct = StringCol(length=45, default=None)
    HeatFromQGenerationHC_id = IntCol(default=None)
    HeatDistMedium = StringCol(length=45, default=None)
    DistribCircFlow = FloatCol(default=None)
    FlowRateReturn = FloatCol(default=None)
    ToutDistrib = FloatCol(default=None)
    TreturnDistrib = FloatCol(default=None)
    TForwOut = FloatCol(default=None)
    TRetIn = FloatCol(default=None)
    TRetOut = FloatCol(default=None)
    PercentRecirc = FloatCol(default=None)
    Tfeedup = FloatCol(default=None)
    PressDistMedium = FloatCol(default=None)
    TotLengthDistPipe = FloatCol(default=None)
    UAPipe = FloatCol(default=None)
    DDistPipe = FloatCol(default=None)
    DeltaDistPipe = FloatCol(default=None)
    NumStorageUnits = IntCol(default=None)
    VUnitStorage = FloatCol(default=None)
    VtotStorage = FloatCol(default=None)
    TypeStorage = StringCol(length=45, default=None)
    PmaxStorage = FloatCol(default=None)
    TmaxStorage = FloatCol(default=None)
    IsAlternative = IntCol(default=None)
    HDEffAvg = FloatCol(default=None)
    USHm = FloatCol(default=None)
    USHPipe = FloatCol(default=None)
    QHXPipe = FloatCol(default=None)
    UPHProcm = FloatCol(default=None)
    QLossPipe = FloatCol(default=None)
    QWHPipe = FloatCol(default=None)
    heating_cooling = EnumCol(enumValues = ['heating','cooling'], default='heating')

    @property
    def QDistributionHC_ID(self):
        return self.id

class qelectricity(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QElectricity_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    dbelectricitymix_id = IntCol(default=None)
    PowerContrTot = FloatCol(default=None)
    PowerContrStd = FloatCol(default=None)
    PowerContrPeak = FloatCol(default=None)
    PowerContrVall = FloatCol(default=None)
    ElectricityTotYear = FloatCol(default=None)
    ElectricityPeakYear = FloatCol(default=None)
    ElectricityStandYear = FloatCol(default=None)
    ElectricityValleyYear = FloatCol(default=None)
    ElGenera = FloatCol(default=None)
    ElSales = FloatCol(default=None)
    ElectricityRef = FloatCol(default=None)
    ElectricityAC = FloatCol(default=None)
    ElectricityThOther = FloatCol(default=None)
    ElectricityMotors = FloatCol(default=None)
    ElectricityChem = FloatCol(default=None)
    ElectricityLight = FloatCol(default=None)
    ElTariffClassTot = StringCol(length=45, default=None)
    ElTariffClassStd = StringCol(length=45, default=None)
    ElTariffClassPeak = StringCol(length=45, default=None)
    ElTariffClassTotVall = StringCol(length=45, default=None)
    ElTariffClassCHP = StringCol(length=45, default=None)
    ElTariffPowTot = FloatCol(default=None)
    ElTariffPowStd = FloatCol(default=None)
    ElTariffPowPeak = FloatCol(default=None)
    ElTariffPowVall = FloatCol(default=None)
    ElTariffPowCHP = FloatCol(default=None)
    ElTariffCTot = FloatCol(default=None)
    ElTariffCStd = FloatCol(default=None)
    ElTariffCPeak = FloatCol(default=None)
    ElTariffCVall = FloatCol(default=None)
    ElTariffFeedIn = FloatCol(default=None)
    ETariffCHP = FloatCol(default=None)
    ElCostYearTot = FloatCol(default=None)
    ElCostYearStd = FloatCol(default=None)
    ElCostYearPeak = FloatCol(default=None)
    ElCostYearVall = FloatCol(default=None)
    ElSalesYearFeedIn = FloatCol(default=None)
    ElSalesYearCHP = FloatCol(default=None)
    ELECTRICITY_Type = FloatCol(default=None)
    DataFile = StringCol(length=200, default=None)

    @property
    def QElectricity_ID(self):
        return self.id

class qflow(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()

    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    FromId = IntCol(default=None)
    FromCon = IntCol(default=None)
    ToId = IntCol(default=None)
    ToCon = IntCol(default=None)
    Flow = FloatCol(default=None)


class qfuel(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QFuel_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    FuelNo = IntCol(default=None)
    FuelUnit = StringCol(length=45, default=None)
    DBFuel_id = IntCol(default=None)
    MFuelYear = FloatCol(default=None)
    FuelOwn = FloatCol(default=None)
    FuelTariff = FloatCol(default=None)
    FuelCostYear = FloatCol(default=None)
    FECFuel = FloatCol(default=None)
    FETFuel = FloatCol(default=None)
    FEOFuel = FloatCol(default=None)
    PECFuel = FloatCol(default=None)
    PETFuel = FloatCol(default=None)
    ProdCO2Fuel = FloatCol(default=None)

    @property
    def QFuel_ID(self):
        return self.id

class qgenerationhc(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QGenerationHC_ID" if DBSERVER == "MYSQL" else "id"
        
    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Equipment = StringCol(length=45, default=None)
    Manufact = StringCol(length=45, default=None)
    YearManufact = IntCol(default=None)
    Model = StringCol(length=45, default=None)
    EquipType = StringCol(length=45, default=None)
    NumEquipUnits = IntCol(default=None)
    DBFuel_id = IntCol(default=None)
    HCGPnom = FloatCol(default=None)
    FuelConsum = FloatCol(default=None)
    UnitsFuelConsum = StringCol(length=45, default=None)
    ElectriConsum = FloatCol(default=None)
    HCGTEfficiency = FloatCol(default=None)
    HCGTEffReal = FloatCol(default=None)
    HCGEEfficiency = FloatCol(default=None)
    ElectriProduction = FloatCol(default=None)
    TExhaustGas = FloatCol(default=None)
    TMaxSupply = FloatCol(default=None)
    TSupply = FloatCol(default=None)
    TMaxSupply2 = FloatCol(default=None)
    PartLoad = FloatCol(default=None)
    HPerDayEq = FloatCol(default=None)
    NDaysEq = FloatCol(default=None)
    PipeDuctEquip = StringCol(length=200, default=None)
    CoolTowerType = StringCol(length=45, default=None)
    IsSelectedFromDB = IntCol(default=None)
    DatabaseNameSelection = StringCol(length=45, default=None)
    EquipTypeFromDB = StringCol(length=45, default=None)
    EquipIDFromDB = IntCol(default=None)
    EqNo = IntCol(default=None)
    ExcessAirRatio = FloatCol(default=None)
    HeatSourceLT = StringCol(length=200, default=None)
    HeatSourceHT = IntCol(default=None)
    THeatSourceLT = FloatCol(default=None)
    ThermalConsum = FloatCol(default=None)
    THeatSourceHT = FloatCol(default=None)
    Refrigerant = IntCol(default=None)
    DestinationWasteHeat = StringCol(length=200, default=None)
    TemperatureReCooling = FloatCol(default=None)
    FlowExhaustGas = FloatCol(default=None)
    QExhaustGas = FloatCol(default=None)
    QHXEq = FloatCol(default=None)
    QWHEq = FloatCol(default=None)
    QWHj = FloatCol(default=None)
    QHXj = FloatCol(default=None)
    USHj = FloatCol(default=None)
    FETj = FloatCol(default=None)
    FETFuel_j = FloatCol(default=None)
    FETel_j = FloatCol(default=None)
    FETHeat_j = FloatCol(default=None)
    HTilted = FloatCol(default=None)
    HGEffAvg = FloatCol(default=None)
    CascadeIndex = IntCol(default=None)
    HPerYearEq = FloatCol(default=None)
    HPExHeatCOP = FloatCol(default=None)
    HPExCoolCOP = FloatCol(default=None)
    Price = FloatCol(default=None)
    TurnKeyPrice = FloatCol(default=None)
    OandM = FloatCol(default=None)
    OandMvar = FloatCol(default=None)
    OandMfix = FloatCol(default=None)
    ST_C0 = FloatCol(default=None)
    ST_C1 = FloatCol(default=None)
    ST_C2 = FloatCol(default=None)
    ST_K50L = FloatCol(default=None)
    ST_K50T = FloatCol(default=None)
    TAvColl = FloatCol(default=None)
    STAreaFactor = FloatCol(default=None)
    ST_Volume = FloatCol(default=None)
    ST_SysEff = FloatCol(default=None)
    ST_Surfaces = StringCol(length=45, default=None)

    @property
    def QGenerationHC_ID(self):
        return self.id

class qheatexchanger(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QHeatExchanger_ID" if DBSERVER == "MYSQL" else "id"
        
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    HXNo = IntCol(default=None)
    HXName = StringCol(length=300, default=None)
    HXType = StringCol(length=300, default=None)
    QdotHX = FloatCol(default=None)
    QdotHXNom = FloatCol(default=None)
    HXLMTD = FloatCol(default=None)
    HXLMTDNom = FloatCol(default=None)
    UA = FloatCol(default=None)
    Area = FloatCol(default=None)
    QHX = FloatCol(default=None)
    HPerYearHX = FloatCol(default=None)
    HXSource = StringCol(length=300, default=None)
    FluidIDSource = IntCol(default=None)
    FlowSource = FloatCol(default=None)
    HXTSourceInlet = FloatCol(default=None)
    HXhSourceInlet = FloatCol(default=None)
    HXTSourceOutlet = FloatCol(default=None)
    HXhSourceOutlet = FloatCol(default=None)
    HXSink = StringCol(length=300, default=None)
    FluidIDSink = IntCol(default=None)
    FlowSink = FloatCol(default=None)
    HXTSinkInlet = FloatCol(default=None)
    HXTSinkOutlet = FloatCol(default=None)
    TurnKeyPrice = FloatCol(default=None)
    OMFix = FloatCol(default=None)
    OMVar = FloatCol(default=None)
    StorageSize = FloatCol(default=None)
    StreamStatusSource = StringCol(length=200, default=None)
    StreamStatusSink = StringCol(length=200, default=None)
    StreamTypeSink = StringCol(length=200, default=None)
    StreamTypeSource = StringCol(length=200, default=None)

    @property
    def QHeatExchanger_ID(self):
        return self.id

class qhxcontrol(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    HXId = IntCol(default=None)
    Kc0 = FloatCol(default=None)
    Kc1 = FloatCol(default=None)
    Kh0 = FloatCol(default=None)
    Kh1 = FloatCol(default=None)
    CcMin = FloatCol(default=None)
    ChMin = FloatCol(default=None)
    TsMax = FloatCol(default=None)

class qnode(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()

    ObjectId = IntCol(default=None)
    ObjectType = StringCol(length=4, default=None)
    Equation = StringCol(length=200, default=None)

class qmdatafiles(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    ProjectID = IntCol(default=None)
    Path = StringCol(length=300, default=None)
    FileName = StringCol(length=300, default=None)
    ShortName = StringCol(length=45, default=None)

class qmdatasets(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    DataFileId = IntCol(default=None)
    ColNo = IntCol(default=None)
    ObjectId = IntCol(default=None)
    ParCode = StringCol(length=20, default=None)
    Scale = FloatCol(default=None)
    

class qprocessdata(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QProcessData_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    ProcNo = IntCol(default=None)
    Process = StringCol(length=45, default=None)
    Description = StringCol(length=200, default=None)
    DBUnitOperation_id = IntCol(default=None)
    ProcType = StringCol(length=45, default=None)
    ProcMedDBFluid_id = IntCol(default=None)
    PT = FloatCol(default=None)
    PTInFlow = FloatCol(default=None)
    PTStartUp = FloatCol(default=None)
    VInFlowDay = FloatCol(default=None)
    VInFlowCycle = FloatCol(default=None)
    mInFlowNom = FloatCol(default=None)
    VolProcMed = FloatCol(default=None)
    ThermalMass = FloatCol(default=None)
    UAProc = FloatCol(default=None)
    HPerDayProc = FloatCol(default=None)
    HPerYearProc = FloatCol(default=None)
    HPerYearInFlow = FloatCol(default=None)
    HPerYearOutFlow = FloatCol(default=None)
    PartLoad = FloatCol(default=None)
    NBatch = FloatCol(default=None)
    HBatch = FloatCol(default=None)
    NDaysProc = FloatCol(default=None)
    PTOutFlow = FloatCol(default=None)
    PTOutFlowRec = FloatCol(default=None)
    PTFinal = FloatCol(default=None)
    VOutFlow = FloatCol(default=None)
    VOutFlowCycle = FloatCol(default=None)
    mOutFlowNom = FloatCol(default=None)
    HeatRecOK = StringCol(length=4, default=None)
    HeatRecExist = StringCol(length=4, default=None)
    SourceWasteHeat = StringCol(length=45, default=None)
    PTInFlowRec = FloatCol(default=None)
    EnclChartHCSupply = IntCol(default=None)
    SupplyMedDBFluid_id = IntCol(default=None)
    PipeDuctProc = StringCol(length=45, default=None)
    TSupply = FloatCol(default=None)
    SupplyMedFlow = FloatCol(default=None)
    DPSTmin = FloatCol(default=None)
    DPSTminMCD = FloatCol(default=None)
    ExternalSupply = StringCol(length=4, default='yes')
    PTInMax = FloatCol(default=None)
    QOpProc = FloatCol(default=None)
    ProcMedOut = IntCol(default=None)
    HOutFlow = FloatCol(default=None)
    XOutFlow = FloatCol(default=None)
    QEvapProc = FloatCol(default=None)
    QHXProcInt = FloatCol(default=None)
    QHXProc = FloatCol(default=None)
    UPHcGross = FloatCol(default=None)
    UPH = FloatCol(default=None)
    UPHProc = FloatCol(default=None)
    UPHc = FloatCol(default=None)
    UPHm = FloatCol(default=None)
    UPHs = FloatCol(default=None)
    UPHw = FloatCol(default=None)
    QWHProc = FloatCol(default=None)
    UPHp = FloatCol(default=None)
    UPHmin = FloatCol(default=None)
    ProcHC = IntCol(default=None)
    UPCtot = FloatCol(default=None)
    UPCnet = FloatCol(default=None)
    PMU = StringCol(length=45, default=None)
    MProcMed = FloatCol(default=None)
    PE_SEC = FloatCol(default=None)
    EL_SEC = FloatCol(default=None)
    UPH_SEC = FloatCol(default=None)
    TEnvProc = FloatCol(default=None)
    QdotProc_c = FloatCol(default=None)
    QdotProc_m = FloatCol(default=None)
    QdotProc_s = FloatCol(default=None)
    QdotProc_w = FloatCol(default=None)
    InFlowDuration = FloatCol(default=None)
    OutFlowDuration = FloatCol(default=None)
    StartUpDuration = FloatCol(default=None)
    ScheduleTolerance = FloatCol(default=None)
    HolidayScale = IntCol(default=None)
    QBuildings_ID = IntCol(default=None)
    heating_cooling = EnumCol(enumValues=['heating','cooling'], default='heating')
    Building_purpose = EnumCol(enumValues=['heating','cooling','domestic_hot_water'], default=None)
    UPHwReq = FloatCol(default=None)
    PipeDuctForMandatoryCoolingDown = StringCol(length=45, default=None)

    @property
    def QProcessData_ID(self):
        return self.id


class qproduct(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QProduct_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Product = StringCol(length=45, default=None)
    ProductCode= StringCol(length=45, default=None)
    QProdYear = FloatCol(default=None)
    ProdUnit= StringCol(length=45, default=None)
    TurnoverProd = FloatCol(default=None)
    ElProd = FloatCol(default=None)
    FuelProd = FloatCol(default=None)
    PE_SEC = FloatCol(default=None)
    EL_SEC = FloatCol(default=None)
    FUEL_SEC = FloatCol(default=None)

    @property
    def QProduct_ID(self):
        return self.id

class qrenewables(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QRenewables_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    REMotivation = StringCol(length=45, default=None)
    REInterest = IntCol(default=None)
    REReason = StringCol(length=45, default=None)
    BiomassFromProc = StringCol(length=45, default=None)
    PriceBiomassProc = FloatCol(default=None)
    PeriodBiomassProcStart = DateCol(default=None)
    PeriodBiomassProcStop = DateCol(default=None)
    QBiomassProc = FloatCol(default=None)
    QBiomassRegion = FloatCol(default=None)
    NDaysBiomassProc = FloatCol(default=None)
    QBiomassProcDay = FloatCol(default=None)
    SpaceBiomassProc = FloatCol(default=None)
    LCVBiomassProc = FloatCol(default=None)
    HumidBiomassProc = FloatCol(default=None)
    BiomassFromRegion = StringCol(length=45, default=None)
    PriceBiomassRegion = FloatCol(default=None)
    PeriodBiomassRegionStart = DateCol(default=None)
    PeriodBiomassRegionStop = DateCol(default=None)
    LCVBiomassRegion = FloatCol(default=None)
    HumidBiomassRegion = FloatCol(default=None)
    NDaysBiomassRegion = FloatCol(default=None)
    Latitude = FloatCol(default=None)
    ST_I = FloatCol(default=None)
    ST_Idn = FloatCol(default=None)
    TAmb = FloatCol(default=None)
    WDataFile = StringCol(length=1024, default=None)
    
class qsurfarea(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        
    ProjectID = IntCol(default=None)
    SurfAreaNo = IntCol(default=None)
    SurfAreaName = StringCol(length=45, default=None)
    SurfArea = FloatCol(default=None)
    Inclination = FloatCol(default=None)
    Azimuth = FloatCol(default=None)
    AzimuthClass = StringCol(length=4, default=None)
    Shading = StringCol(length=20, default=None)
    Distance = FloatCol(default=None)
    RoofType = StringCol(length=45, default=None)
    RoofStaticLoadCap = FloatCol(default=None)
    Sketch = IntCol(default=None)
    ST_IbT = FloatCol(default=None)
    ST_IT = FloatCol(default=None)

    @property
    def QSurfArea_ID(self):
        return self.id

class questionnaire(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "Questionnaire_ID" if DBSERVER == "MYSQL" else "id"

    PSetUpData_id = IntCol(default=None)
    Name = StringCol(length=45, default=None)
    City = StringCol(length=45, default=None)
    Country = StringCol(length=45, default=None)
    DescripIndustry = StringCol(length=200, default=None)
    Branch = StringCol(length=100, default=None)
    SubBranch = StringCol(length=100, default=None)
    DBNaceCode_id = IntCol(default=None)
    Contact = StringCol(length=45, default=None)
    Role = StringCol(length=45, default=None)
    Address = StringCol(length=200, default=None)
    Phone = StringCol(length=45, default=None)
    Fax = StringCol(length=45, default=None)
    Email = StringCol(length=45, default=None)
    NEmployees = IntCol(default=None)
    Turnover = FloatCol(default=None)
    ProdCost = FloatCol(default=None)
    BaseYear  = IntCol(default=None)
    Growth = FloatCol(default=None)
    Independent = IntCol(default=None)
    OMThermal = FloatCol(default=None)
    OMElectrical = FloatCol(default=None)
    HPerDayInd = FloatCol(default=None)
    NShifts = FloatCol(default=None)
    NDaysInd = FloatCol(default=None)
    NoProdStart_1 = DateCol(default=None)
    NoProdStart_2 = DateCol(default=None)
    NoProdStart_3 = DateCol(default=None)
    NoProdStop_1 = DateCol(default=None)
    NoProdStop_2 = DateCol(default=None)
    NoProdStop_3 = DateCol(default=None)
    PercentElTotcost = FloatCol(default=None)
    PercentFuelTotcost = FloatCol(default=None)
    EnclMonthlyElBills = IntCol(default=None)
    EnclMonthlyFuelBills = IntCol(default=None)
    InflationRate= FloatCol(default=None)
    FuelPriceRate = FloatCol(default=None)
    InterestExtFinancing = FloatCol(default=None)
    CompSpecificDiscountRate = FloatCol(default=None)
    PercentExtFinancing = FloatCol(default=None)
    AmortisationTime = FloatCol(default=None)
    PublicFundType = StringCol(length=45, default=None)
    OMGenTot = FloatCol(default=None)
    OMGenUtilities = FloatCol(default=None)
    OMGenLabour = FloatCol(default=None)
    OMGenExternal = FloatCol(default=None)
    OMGenRegulatory = FloatCol(default=None)
    OMBuildTot = FloatCol(default=None)
    OMBuildUtilities = FloatCol(default=None)
    OMBuildLabour = FloatCol(default=None)
    OMBuildExternal = FloatCol(default=None)
    OMBuildRegulatory = FloatCol(default=None)
    OMBiuildFung = FloatCol(default=None)
    OMMachEquipTot = FloatCol(default=None)
    OMMachEquipLabour = FloatCol(default=None)
    OMMachEquipExternal = FloatCol(default=None)
    OMMachEquipRegulatory = FloatCol(default=None)
    OMMachEquipUtilities = FloatCol(default=None)
    OMHCGenDistTot = FloatCol(default=None)
    OMHCGenDistUtilities = FloatCol(default=None)
    OMHCGenDistExternal = FloatCol(default=None)
    OMHCGenDistRegulatory = FloatCol(default=None)
    OMHCGenDistLabour = FloatCol(default=None)
    OMTotalTot = FloatCol(default=None)
    OMTotalLabour = FloatCol(default=None)
    OMTotalExternal = FloatCol(default=None)
    OMTotalRegulatory = FloatCol(default=None)
    OMTotalUtilities = FloatCol(default=None)
    EnergyManagExisting = IntCol(default=None)
    EnergyManagExternal = IntCol(default=None)
    EnclChartHCDistribSys = IntCol(default=None)
    NBuild = IntCol(default=None)

    @property
    def Questionnaire_ID(self):
        return self.id



class qwasteheatelequip(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "QWasteHeatElEquip_ID" if DBSERVER == "MYSQL" else "id"

    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    WHEENo = IntCol(default=None)
    WHEEName = StringCol(length=45, default=None)
    WHEEEqType = StringCol(length=45, default=None)
    WHEEWasteHeatType = StringCol(length=45, default=None)
    QWHEE = FloatCol(default=None)
    WHEEMedium = StringCol(length=45, default=None)
    WHEEFlow = FloatCol(default=None)
    WHEETOutlet = FloatCol(default=None)
    WHEEPresentUse = StringCol(length=45, default=None)
    HPerDayWHEE = FloatCol(default=None)
    NBatchWHEE = FloatCol(default=None)
    HBatchWHEE = FloatCol(default=None)
    NDaysWHEE = FloatCol(default=None)

    @property
    def QWasteHeatElEquip_ID(self):
        return self.id

class salternatives(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "SAlternative_ID" if DBSERVER == "MYSQL" else "id"

        
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    ShortName = StringCol(length=45, default=None)
    Description = StringCol(length=1000, default=None)
    StatusA = IntCol(default=None)
    StatusPO = IntCol(default=None)
    StatusPI = IntCol(default=None)
    StatusHX = IntCol(default=None)
    StatusHC = IntCol(default=None)
    StatusEnergy = IntCol(default=None)
    StatusECO = IntCol(default=None)

    @property
    def SAlternatives_ID(self):
        return self.id


class sproject(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "SProject_ID" if DBSERVER == "MYSQL" else "id"

    ProjectID = IntCol(default=None)
    ProjectName = StringCol(length=80, default=None)
    NoOfAlternatives = IntCol(default=None)
    ActiveAlternative = IntCol(default=None)
    FinalAlternative = IntCol(default=None)
    WriteProtected = IntCol(default=None)
    StatusQ = IntCol(default=None)
    StatusCC = IntCol(default=None)
    StatusCA = IntCol(default=None)
    StatusR = IntCol(default=None)
    LanguageReport = StringCol(length=45, default=None)
    UnitsReport = StringCol(length=45, default=None)
    Summary = StringCol(default=None)
    HRTool = StringCol(length=45, default=None)
    Auditor_ID = IntCol(default=None)
    CreationDate = DateCol(default=None)
    CreatedBy = IntCol(default=None)
    LastModified = DateCol(default=None)
    ModifiedBy = IntCol(default=None)
    Attributes = StringCol(length=10, default=None)

    @property
    def SProject_ID(self):
        return self.id
    
class stool(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "STool_ID" if DBSERVER == "MYSQL" else "id"

    Version = StringCol(length=20)
    ActiveProject = IntCol(default=None)
    UserType = StringCol(length=45)
    Auditor_ID = IntCol(default=None)
    UserInteractionLevel = StringCol(length=45, default=None)
    LanguageTool = StringCol(length=45, default=None)
    UnitsTool = StringCol(length=45, default=None)

class streams_in(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    Name = StringCol(length=256, default=None)
    dbfluid_id = IntCol(default=None)
    PTInFlow = FloatCol(default=None)
    PTInFlowRec = FloatCol(default=None)
    VInFlowCycle = FloatCol(default=None)
    mInFlowNom = FloatCol(default=None)
    HeatRecExist = IntCol(default=None)
    HInFlow = FloatCol(default=None)
    XInFlow = FloatCol(default=None)
    UPHc = FloatCol(default=None)
    QdotProc_c = FloatCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)

class streams_out(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    Name = StringCol(length=256, default=None)
    dbfluid_id = IntCol(default=None)
    PTOutFlow = FloatCol(default=None)
    PTOutFlowRec = FloatCol(default=None)
    PTFinal = FloatCol(default=None)
    VOutFlowCycle = FloatCol(default=None)
    mOutFlowNom = FloatCol(default=None)
    HeatRecExist = IntCol(default=None)
    HeatRecOk = IntCol(default=None)
    HOutFlow = FloatCol(default=None)
    XOutFlow = FloatCol(default=None)
    UPHw = FloatCol(default=None)
    QdotProc_w = FloatCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    PTFinalRequired = FloatCol(default=None)
    UPHwReq = FloatCol(default=None)

class tcacontingencies(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    Value = FloatCol(default=None)
    TimeFrame = IntCol(default=None)

class tcadetailedopcost(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    Value = FloatCol(default=None)
    Category = IntCol(default=None)

class tcadetailedrevenue(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    InitialInvestment = FloatCol(default=None)
    DeprecationPeriod = IntCol(default=None)
    RemainingPeriod = IntCol(default=None)

class tcaenergy(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    EnergyDemand = FloatCol(default=None)
    EnergyPrice = FloatCol(default=None)
    DevelopmentOfEnergyPrice = FloatCol(default=None)

class tcageneraldata(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "IDTca" if DBSERVER == "MYSQL" else "id"

    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    InflationRate = FloatCol(default=None)
    NominalInterestRate = FloatCol(default=None)
    CompSpecificDiscountRate = FloatCol(default=None)
    FulePriceRate = FloatCol(default=None)
    AmotisationTime = FloatCol(default=None)    #TODO: changed to FloatCol here, in R2.2 was IntCol !
    TotalOperatingCost = FloatCol(default=None)
    TotalRevenue = FloatCol(default=None)

class tcainvestments(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    Investment = FloatCol(default=None)
    FundingPerc = FloatCol(default=None)
    FundingFix = FloatCol(default=None)

class tcanonreoccuringcosts(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
    TcaID = IntCol(default=None)
    ProjectID = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    Description = StringCol(length=200, default=None)
    Value = FloatCol(default=None)
    Year = IntCol(default=0)
    Type = StringCol(length=200, default=None)

class uheatpump(SQLObject):
    class sqlmeta:
        style = MixedCaseStyle()
        idName = "UHeatPump_ID" if DBSERVER == "MYSQL" else "id"

    Questionnaire_id = IntCol(default=None)
    AlternativeProposalNo = IntCol(default=None)
    CHPMaintain = IntCol(default=None)
    CHPType  = StringCol(length=45, default=None)
    CHPFuelType  = StringCol(length=45, default=None)
    CHPHOp = FloatCol(default=None)
    CHPEff = FloatCol(default=None)
    UHPType  = StringCol(length=45, default=None)
    UHPMinHop = FloatCol(default=None)
    UHPDTMax = FloatCol(default=None)
    UHPTgenIn = FloatCol(default=None)
    UHPmaxT = FloatCol(default=None)
    UHPminT = FloatCol(default=None)
    BBMaintain = IntCol(default=None)
    BBSafety = FloatCol(default=None)
    BBRedundancy = IntCol(default=None)
    BBFuelType  = StringCol(length=45, default=None)
    BBHOp = FloatCol(default=None)
    BBPmin = FloatCol(default=None)
    BBEff = FloatCol(default=None)
    UHPMaintain = IntCol(default=None)
    STSolFra = FloatCol(default=None)
    STCollType  = StringCol(length=45, default=None)
    STMinYield = FloatCol(default=None)
 
