# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	PipeObject
#			
#------------------------------------------------------------------------------
#			
#       Basic object describing a pipe or duct
#
#==============================================================================
#
#	Version 3.0
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import EPSILON

from globals import logger
from GUI.status import Status

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck
from auxiliary.dataTransfer import DataTransfer

from modules.ccheck.dataSet import readDataFile

from genericObject import GenericObject

from lib.auxiliary.float import toFloat

import numpy as np

# keys for parameters which are 1:1 passed to data
# struture: sqlLabel: (attr.-name [None if identical to sqlLabel], dictionary for conversion)


# TODO:
# currently this key is used for
# a) list those parameters to be written to SQL
# b) list those parameters to be written to GUI
# subset of (a) and (b) may be different, with conversions in between, which may not be only 1:1 conversions

yesNoDict = {"no": False, "yes": True}
HCDict = {"heating": True, "cooling": False}

SQLKEYS = {"dbelectricitymix_id": (None, None),
           "PowerContrTot": (None, None),
           "PowerContrStd": (None, None),
           "PowerContrPeak": (None, None),
           "PowerContrVall": (None, None),
           "ElectricityTotYear": (None, None),
           "ElectricityPeakYear": (None, None),
           "ElectricityStandYear": (None, None),
           "ElectricityValleyYear": (None, None),
           "ElGenera": (None, None),
           "ElSales": (None, None),
           "ElectricityRef": (None, None),
           "ElectricityAC": (None, None),
           "ElectricityThOther": (None, None),
           "ElectricityMotors": (None, None),
           "ElectricityChem": (None, None),
           "ElectricityLight": (None, None),
           "ElTariffClassTot": (None, None),
           "ElTariffClassStd": (None, None),
           "ElTariffClassPeak": (None, None),
           "ElTariffClassTotVall": (None, None),
           "ElTariffClassCHP": (None, None),
           "ElTariffPowTot": (None, None),
           "ElTariffPowStd": (None, None),
           "ElTariffPowPeak": (None, None),
           "ElTariffPowVall": (None, None),
           "ElTariffPowCHP": (None, None),
           "ElTariffCTot": (None, None),
           "ElTariffCStd": (None, None),
           "ElTariffCPeak": (None, None),
           "ElTariffCVall": (None, None),
           "ElTariffFeedIn": (None, None),
           "ETariffCHP": ("ElTariffCHP", None),
           "ElCostYearTot": (None, None),
           "ElCostYearStd": (None, None),
           "ElCostYearPeak": (None, None),
           "ElCostYearVall": (None, None),
           "ElSalesYearFeedIn": (None, None),
           "ElSalesYearCHP": (None, None),
           "DataFile": (None, None)}

SQLKEYS_CGENERALDATA = ["FECel",
                        "FEOel",
                        "FETel"]


MANDATORYDATA = []
OPTIONALDATA = []
                                 
class ElObject(GenericObject):
    """
    Base object representing a process in the system model
    """
    def __init__(self, prnt, id=None, no=None, sqlData=None):
        """
        :param inStreams, outStreams: streamObjects for inflowing and outflowing streams
        """

        super(ElObject, self).__init__(prnt, id, no, sqlData,
                 sqlKeys=SQLKEYS, mandatoryData=MANDATORYDATA, optionalData=OPTIONALDATA,
                 sqlTableName="qelectricity", sqlObjectList=prnt.el,
                 keyPId="Questionnaire_id", colObjectNo=None)

        if self.new:
            self.name = _("electricity mix (default)")

        self.FETel_t = None
        self.FECel_t = None

    def __repr__(self):
        info = "El(-): id %s mix %s "%(self.id, self.name)
        return info

    def loadFromDB(self, sqlData):
        """
        fetches data from DB
        """
        super(ElObject, self).loadFromDB(sqlData)

        # fetch some data stored in cgeneraldata table
        (dummy, generalData) = Status.prj.getProjectData()
        for key in SQLKEYS_CGENERALDATA:
            setattr(self, key, getattr(generalData, key))

        # calculate derived properties: environmental parameters
        dbemix = dbio.get("dbelectricitymix", id=self.dbelectricitymix_id)

        # first define and assign None to all relevant variables ...
        self.PEConv = None
        self.CO2Conv = None
        self.NoNukesConv = None

        for key in ["TariffC", "TariffCHP", "PEConv", "CO2Conv", "NoNukesConv"]:
            setattr(self, key + "_t", None)

        # ... then overwrite the relevant ones ...
        if self.DataFile:
            data = readDataFile(self.DataFile)
            for key in ["TariffC", "TariffCHP", "PEConv", "CO2Conv", "NoNukesConv"]:
                if data.keys is not None and key in data.keys:
                    index = data.keys.index(key)
                    setattr(self, key + "_t", data.targets[index])

                    avg = np.average(data.targets[index])

            self.name = self.DataFile
            
        elif dbemix:
            self.name = dbemix.ShortName
            self.PEConv = dbemix.PE2ConvEl  # NOTE: there is a typing error in the dbemix table !!!
            self.CO2Conv = dbemix.CO2ConvEl
            self.NoNukesConv = dbemix.NoNukesConvEl
            
        else:
            self.name = _("electricity mix (default)")
            
        # ... and finally assure that there is any (default) value in order to avoid crashes.
        self.PEConv = noneCheck(self.PEConv, 3.0)
        self.CO2Conv = noneCheck(self.CO2Conv, 0.5)
        self.NoNukesConv = noneCheck(self.NoNukesConv, 5.00)

    def setName(self):
        """
        helper function. defines the name depending on the state of attributes dbelectricitymix_id and DataFile
        """
        if self.DataFile:
            self.name = self.DataFile
            
        elif self.dbelectricitymix_id is not None:
            dbemix = dbio.get("dbelectricitymix", id=self.dbelectricitymix_id)
            self.name = dbemix.ShortName
            
        else:
            self.name = _("electricity mix (default)")
            
    def saveToDB(self, ANo=None, warn=True):
        """
        stores the process in the DB
        or creates a new one, if id = None or <= 0

        ANo: special case for ANo=0 (used in consistency check). Copies process data from ANo=-1 to ANo 0
        """
        # re-calculate all derived parameters
        # Information is kept redundant for now:
        generalData = Status.prj.getGeneralData(ANo=ANo)

        if not (self.dbelectricitymix_id or self.DataFile):
            logger.logWarning(_("No electricity mix was specified.\nDefault values assumed"))

        if Status.ANo > 0: #allow for changing of FEO data in alternatives
            try:
                FEOel = float(self.ElectricityChem) + float(self.ElectricityLight) + float(self.ElectricityMotors)
            except:
                pass
            else:
                generalData.FEOel = dbio.check(FEOel)

        for key in SQLKEYS_CGENERALDATA:
            setattr(generalData, key, toFloat(getattr(self, key)))

        super(ElObject, self).saveToDB(ANo=ANo, derivedData=None, warn=warn)

        return True


    def aggregateData(self, FETel_t, FEOel, ElGen_t):
        """
        Builds the aggregate of self generation of each equipment, ElGen_t
        and the total (net) consumption FECel_t based on the following inputs

        FETel_t (aggregate net electricity consumption)
        ElGen_t (list of generation in each equipment)
        FEO     (aggregate of electricity consumption for non-thermal uses, suppose to
                 follow a schedule identical to the plant Schedule, with a constant part of 10%)
        """
        self.ElGen_t = np.sum(np.array(ElGen_t), axis=0)
        self.FETel_t = FETel_t
        self.FECel_t = np.array(FETel_t) + FEOel*(0.9*Status.schedules.plantSchedule.fav + 0.1/Status.Nt*np.ones(Status.Nt))

        # calculate yearly updates
        self.ElGenera = toFloat(np.sum(self.ElGen_t))
        self.FETel = toFloat(np.sum(self.FETel_t))
        self.FECel = toFloat(np.sum(self.FECel_t))
        self.FEOel = FEOel
        self.ElectricityTotYear = toFloat(self.FECel)
            
        
    def distributeSelfGen(self):
        """
        uses FECel_t and ElGen_t as input
        distributes the self generated electricity to
        selfConsumption and sales following economic criteria (difference of hourly tariffs)

        Mathematics:
        FECel = FECel_c (total consumption) - ElGen (self generation)
        FECel_c = purchase + ElGen - sales
        FECel = purchase - sales

        NOTE: requires previous call of self.aggregateData
        """

        self.tariffCHP = self.ElTariffCHP if self.TariffCHP_t is None else self.TariffCHP_t
        self.tariffC = self.ElTariffCTot if self.TariffC_t is None else self.TariffC_t
        
        export = (self.tariffCHP > self.tariffC + EPSILON)     # can be a constant boolean value or a np.array depending on the tariffs
        
        self.sales_t = np.maximum(-self.FECel_t, export*self.ElGen_t)
        ownConsumption_t = self.ElGen_t - self.sales_t
        
        self.purchase_t = self.FECel_t + self.sales_t

        self.ElSales = float(np.sum(self.sales_t))
        self.ElectricityTotYear = float(np.sum(self.purchase_t))

    def calculateAverageTariffs(self):
        """
        Calculates the weighted average of tariffs for sales and purchase
        """

        energy_sales = float(np.sum(self.sales_t))      # note: explicit conversion to float necessary as np.float64 gives problems when writing to sql     
        cost_sales = float(np.sum(self.sales_t * noneCheck(self.tariffCHP, self.tariffC))) if energy_sales > EPSILON else 0.0

        self.tariffCHPAvg = cost_sales/energy_sales if energy_sales > EPSILON else self.ElTariffCHP
        
        cost_purchase = float(np.sum(self.purchase_t * self.tariffC))
        energy_purchase = float(np.sum(self.purchase_t))

        self.tariffCAvg = cost_purchase/energy_purchase if energy_purchase > EPSILON else self.ElTariffCTot

        # overwrite the tariffs
        self.ElTariffCTot = self.tariffCAvg
        self.ElTariffCHP = self.tariffCHPAvg

    def estimateAverageTariffs(self):
        """
        estimates the average tariffs (for present state analyisis, when FEC/FET(t) is not known yet
        and calculates min/max values in order to estimate the accuracy
        """
        schedule = Status.schedules.plantSchedule.fav
        self.tariffCAvg = self.ElTariffCTot if self.TariffC_t is None else np.sum(self.TariffC_t*schedule)
        self.tariffCHPAvg = self.ElTariffCHP if self.TariffCHP_t is None else np.sum(self.TariffCHP_t*schedule)

        self.tariffCMin = self.ElTariffCTot if self.TariffC_t is None else np.min(self.TariffC_t)
        self.tariffCMax = self.ElTariffCTot if self.TariffC_t is None else np.max(self.TariffC_t)

        self.tariffCHPMin = self.ElTariffCHP if self.TariffCHP_t is None else np.min(self.TariffCHP_t)
        self.tariffCHPMax = self.ElTariffCHP if self.TariffCHP_t is None else np.max(self.TariffCHP_t)
        

    def calculateENV(self):
        """
        calculates environmental parameters based on FEC/Tel(t)
        """

        if self.FETel is None: # TODO: security checks probably necessary only in present state, if at all ... -> maybe check BEFORE call to calculateENV in ModuleEnergy
            self.PETel = 0.0
        else:
            self.PETel = float(np.sum(self.FETel_t * self.PEConv_t)) if (self.PEConv_t is not None and self.FETel_t is not None) else self.FETel * self.PEConv

        if self.FECel is None:
            self.PECel = 0.0
            self.ProdCO2 = 0.0
            self.ProdNoNukes = 0.0
        else:
            self.PECel = float(np.sum(self.FECel_t * self.PEConv_t)) if (self.PEConv_t is not None and self.FECel_t is not None) else self.FECel * self.PEConv 

            self.ProdCO2 = float(np.sum(self.FECel_t * self.CO2Conv_t)) if (self.CO2Conv_t is not None and self.FECel_t is not None) else self.FECel * self.CO2Conv 
            self.ProdNoNukes = float(np.sum(self.FECel_t * self.NoNukesConv_t)) if (self.NoNukesConv_t is not None and self.FECel_t is not None) else self.FECel * self.NoNukesConv
