# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	EquipmentObject
#			
#------------------------------------------------------------------------------
#			
#       Basic object describing an equipment
#
#==============================================================================
#
#	Version 3.0
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import EQCLASSES, EQUIPMENT, getEquipmentClass, getEquipmentSubClass, BBTYPES, EQUIPTYPE, EPSILON

from globals import logger
from GUI.status import Status

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck
from auxiliary.dataTransfer import DataTransfer

from modules.schedules.periodSchedule import PeriodSchedule, ScheduleNotFoundError

from modules.fluids import Fuel

DEFAULTCLASS = "BB"
DEFAULTSUBCLASS = BBTYPES[0]


# keys for parameters which are 1:1 passed to data
# struture: sqlLabel: (attr.-name [None if identical to sqlLabel], dictionary for conversion)


# TODO:
# currently this key is used for
# a) list those parameters to be written to SQL
# b) list those parameters to be written to GUI
# subset of (a) and (b) may be different, with conversions in between, which may not be only 1:1 conversions

yesNoDict = {"no": False, "yes": True}
HCDict = {"heating": True, "cooling": False}

SQLKEYS = {"Equipment": ("name", None),
           "Manufact": (None, None),
           "YearManufact": (None, None),
           "Model": (None, None),
           "EquipType": ("eqType", None),
           "DBFuel_id": (None, None),
           "HCGPnom": ("Pnom", None),
           "FuelConsum": ("mdotFuel", None),
           "ElectriConsum": ("Pel", None),
           "HCGTEfficiency": (None, None),
           "HCGTEffReal": (None, None),
           "HCGEEfficiency": (None, None),
           "ElectriProduction": (None, None),
           "HPerDayEq": ("HPerDay", None),
           "NDaysEq": ("NDays", None),
           "HPerYearEq": ("HPerYear", None),
           "TExhaustGas": (None, None),
           "TMaxSupply": (None, None),
           "TSupply": (None, None),
           "TMaxSupply2": (None, None),
           "PartLoad": (None, None),
           "PipeDuctEquip": (None, None),
           "CoolTowerType": (None, None),
           "EqNo": (None, None),
           "HeatSourceLT": (None, None),
           "HeatSourceHT": (None, None),
           "THeatSourceLT": (None, None),
           "ThermalConsum": (None, None),
           "THeatSourceHT": (None, None),
           "ExcessAirRatio": (None, None),
           "DestinationWasteHeat": (None, None),
           "TemperatureReCooling": (None, None),
           "FlowExhaustGas": (None, None),
           "QExhaustGas": (None, None),
           "QHXEq": (None, None),
           "QWHEq": (None, None),
           "QWHj": (None, None),
           "QHXj": (None, None),
           "USHj": (None, None),
           "FETj": (None, None),
           "FETFuel_j": (None, None),
           "FETel_j": (None, None),
           "FETHeat_j": (None, None),
           "HTilted": (None, None),
           "HGEffAvg": (None, None),
##           "CascadeIndex": (None, None),
           "HPExHeatCOP": (None, None),
           "HPExCoolCOP": (None, None),
           "Price": (None, None),
           "TurnKeyPrice": (None, None),
           "OandM": ("OM", None),
           "OandMvar": ("OMVar", None),
           "OandMfix": ("OMFix", None),
           "ST_C0": (None, None),
           "ST_C1": (None, None),
           "ST_C2": (None, None),
           "ST_K50L": (None, None),
           "ST_K50T": (None, None),
           "TAvColl": (None, None),
           "STAreaFactor": (None, None),
           "ST_Volume": (None, None),
           "ST_SysEff": (None, None),
           "ST_Surfaces": (None, None),
           "NumEquipUnits": (None, None),
           "UnitsFuelConsum": (None, None),
           "IsSelectedFromDB": (None, None),
           "DatabaseNameSelection": (None, None),
           "EquipTypeFromDB": (None, None),
           "EquipIDFromDB": (None, None),
           "Refrigerant": (None, None),
           }        # sqlkeys of data from SQL tables to be read; in case of different names for

MANDATORYDATA = ["HCGPnom", "HCGTEfficiency"]
MANDATORYDATA_CHP = ["HCGEEfficiency"]
MANDATORYDATA_ST = ["ST_C0", "ST_C1", "ST_C2", "ST_K50T", "ST_K50L", "STAreaFactor", "ST_Volume", "ST_SysEff"]
OPTIONALDATA_ST = ["HCGTEfficiency"]
OPTIONALDATA = []
OPTIONALDATA_CT = ["HCGTEfficiency"]
                                 
class EquipmentObject():
    """
    Base object representing a process in the system model
    """
    def __init__(self, prnt, id=None, no=None, sqlData=None):
        """
        :param inStreams, outStreams: streamObjects for inflowing and outflowing streams
        """
        # generic object properties
        self.prnt = prnt    # parent system model
        self.id = id        # identifier for process in data base (None or <=0 if local object)
        self.no = no        # number of the equipment in the system (from 0 to n-1)
        self.name = None if no is None else prnt.equipment[no].Equipment

        # equipment object specific properties
        self.eqClass = None
        self.eqType = None

        # used only for chillers
        self.TcAvg = None
        self.ThAvg = None

        # data sets
        self.TDrive = None
        self.TReject = None
    
##        # TO DO: scheduleType, detailedSchedule and combSchedule are redundant -> unify !!!
##        self.scheduleType = 0   # 0: simple, 1: detailed, 2: fit to data
##
##        self.process_periods = process_periods
##        self.detailedSchedule = (len(process_periods)>0)
##        
##        self.combSchedules = combSchedules
##        self.combSchedule = (len(combSchedules)>0) and not self.detailedSchedule

        # space for StreamGlobals - objects
        self.USHGlb = None
        self.QWHGlb = None
        self.QHXGlb = None
        self.FETHeatGlb = None

        # load data from sqlData
        if self.id is not None and sqlData is not None:
            self.loadFromDB(sqlData)
        else:

            # assign None to all pars
##            for parDict in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
            for parDict in [SQLKEYS]:
                for sqlKey in parDict:
                    key = noneCheck(parDict[sqlKey][0], sqlKey)
                    setattr(self, key, None)

            self.name = _("new equipment")
            self.heating = True

            self.eqType = "steam boiler"   # continuous
            self.eqClass = "BB"
            self.eqSubClass = "steam boiler"

            self.controlSchedule = None

            self.calculateDerivedParameters()


        self.deletedSchedules = []

        self.linkKeys = ["PipeDuctEquip", "HeatSourceHT", "HeatSourceLT"]

    def calculateDerivedParameters(self):
        # calculated derived parameters
        self.COPex = self.HPExCoolCOP if self.eqClass == "CH" else self.HPExHeatCOP

        if self.HPerYear is not None and self.PartLoad is not None:
            self.HPerYearNom = self.HPerYear*float(self.PartLoad)
        else:
            self.HPerYearNom = None

        self.CPartLoad = self.ST_C0 if self.eqSubClass in ["CHP ORC", "CHP TFC"] or self.eqClass in ["CH"] else None            
        self.ControlPar1 = self.ST_C1 if self.eqClass == "CHP" else None

        self.fuelId = self.DBFuel_id
        self.fuel = None if self.fuelId is None else Fuel(self.fuelId)

        self.calculateActualPerformance()

        self.heating = (self.eqClass != "CH")
        
    def loadFromDB(self, sqlData):
        """
        fetches data from DB
        """
        allKeys = {}
##        for keys in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
        for keys in [SQLKEYS]:
            allKeys.update(keys)
            
        transfer = DataTransfer(None, allKeys)
        newData = transfer.write(sqlData)

        for key in newData:
            setattr(self, key, newData[key])



        # calculate derived properties

        if self.eqType is None or self.eqType not in EQUIPTYPE:
            logger.logWarning("Erroneous equipment type [%s] for Equipment %s"%(self.eqType, self.name))
            self.eqType = "steam boiler"
            
        self.eqClass = getEquipmentClass(self.eqType)
        self.eqSubClass = getEquipmentSubClass(self.eqType)

        assert(self.eqClass is not None)
        assert(self.eqSubClass is not None)
        
        try:
            schedule = PeriodSchedule(self.name)
            schedule.loadFromDB(-self.id, withDetails=True)
            
        except ScheduleNotFoundError:
            schedule = None
        self.controlSchedule = schedule

        # repair severe inconsistencies or missing parameters for ANo >= 0 in order to avoid crashes
        if hasattr(sqlData, "AlternativeProposalNo") and sqlData.AlternativeProposalNo != -1:
            pass
##            if self.PT is None:
##                self.PT = -99.99    # assign something in order to avoid crashing

        self.calculateDerivedParameters()

    def setEqClass(self, eqClass):

        if eqClass in EQCLASSES:
            self.eqClass = eqClass
            self.heating = (self.eqClass != "CH")

    def setEqSubClass(self, eqSubClass):

        self.eqSubClass = eqSubClass

        subClasses = [entry[2] for entry in EQUIPMENT[self.eqClass]]
        if self.eqSubClass in subClasses:
            index = subClasses.index(self.eqSubClass)
            self.eqType = EQUIPMENT[self.eqClass][index][0]
        
    def calculateActualPerformance(self):
        """
        Restores the actual performance data from data stored in the SQL
        (helper function of loadFromDB)
        """
        
        if self.HPerYear is not None and self.HPerYear > EPSILON:
            dotFETel = self.FETel_j / self.HPerYear if self.FETel_j is not None else None
            if self.eqClass=="CHP":
                self.PelAvg = 0
                self.PelGenAvg = -dotFETel if dotFETel is not None else None
            else:
                self.PelAvg = dotFETel
                self.PelGenAvg = 0
                
            self.PfuelAvg = self.FETFuel_j / self.HPerYear if self.FETFuel_j is not None else None
            self.PheatAvg = self.FETHeat_j / self.HPerYear if self.FETHeat_j is not None else None
            if self.eqType in ["CHP ORC", "CHP TFC", "thermal chiller (air cooled)", "thermal chiller (water cooled)"]:
                if self.QHXEq is not None and self.PheatAvg is not None:
                    qhx = self.QHXEq / self.HPerYear
                    self.PheatAvg += qhx
                else:
                    self.PheatAvg = None
                
            self.Pavg = self.USHj / self.HPerYear if self.USHj is not None else None
        else:
            self.PelAvg = None
            self.PelGenAvg = None
            self.PfuelAvg = None
            self.PheatAvg = None
            self.Pavg = None
        
    def saveToDB(self, ANo = None, warn=True, simResultsOnly=False):
        """
        stores the process in the DB
        or creates a new one, if id = None or <= 0

        ANo: special case for ANo=0 (used in consistency check). Copies process data from ANo=-1 to ANo 0
        """

        if ANo is None and not Status.prj.editAllowed(warn=warn):
            return False

        if Status.ANo >= 0 and not self.checkCompleteness() : # allow saving of incomplete data for ANo < 0!
            return False

        Status.prj.projectChanged()

        if self.id is None or self.id <=0:
            # TO DO: create a new process if necessary
            self.id = dbio.insertRow("qgenerationhc")
            self.no = len(self.prnt.equipment)
            self.prnt.equipment.append(dbio.get("qgenerationhc", id=self.id))
            Status.int.systemModel.changed(0) # need for rebuilding connections
            Status.int.changeInCascade(self.no+1, keepModel=True)
            
        if ANo==0: # special case for consistency check
            sqlTable = dbio.get("qgenerationhc", Questionnaire_id=Status.PId, AlternativeProposalNo=0, CascadeIndex = self.no+1)
            data = {"Questionnaire_id":Status.PId,
                    "AlternativeProposalNo": 0,
                    "CascadeIndex": self.no+1}

            if sqlTable is not None:
                newId = sqlTable.id
            else:
                logger.logError("EquipmentObject.saveToDB: corresponding equipment in alternative no. 0 does not exist")
                return False
        else:
            sqlTable = self.prnt.equipment[self.no]
            data = {"Questionnaire_id": Status.PId,
                    "AlternativeProposalNo": Status.ANo,
                    "CascadeIndex": self.no+1}

        # re-calculate all derived parameters

        self.subClasses = [entry[2] for entry in EQUIPMENT[self.eqClass]]
        if self.eqSubClass in self.subClasses:
            index = self.subClasses.index(self.eqSubClass)
            self.eqType = EQUIPMENT[self.eqClass][index][0]
        
        if self.eqClass == "CH":
            self.HPExCoolCOP = self.COPex
        else:
            self.HPExHeatCOP = self.COPex

        if self.YearManufact is not None:
            self.YearManufact = int(self.YearManufact+0.5)
    
        # collect data defined in SQLKEYS, transform where necessary, and write to qprocessDAta

        if self.eqSubClass in ["CHP ORC", "CHP TFC"] or self.eqClass in ["CH"]:
            data.update({"ST_C0": self.CPartLoad})

        if self.eqClass == "CHP":
            data.update({"ST_C1": self.ControlPar1})

        sqlTable.set(**data)

        allKeys = {}
        for keys in [SQLKEYS]:
            allKeys.update(keys)

        if ANo==0: # special case ccheck - do NOT copy link information which is specific for each alternative
            for key in self.linkKeys:
                if key in allKeys.keys():
                    allKeys.pop(key)
                    
        transfer = DataTransfer(None, allKeys)
        transfer.read(self, sqlTable)

        # save detailed schedule
        if not simResultsOnly:
            if self.controlSchedule:
                self.controlSchedule.saveToDB(-self.id, withHolidays=True, saveAll=True)
                
            for schedule in self.deletedSchedules:
                schedule.deleteFromDB()
        dbio.commit()

        Status.int.systemModel.build()  # rebuild connections if a new object has been added

        return True

    def checkCompleteness(self):
        """
        checks if the data are complete and correct
        """
        mandatoryData = MANDATORYDATA[:]
        optionalData = OPTIONALDATA[:]
        if self.eqClass == "ST":
            mandatoryData += MANDATORYDATA_ST
            optionalData += OPTIONALDATA_ST

            from modules.solar.moduleST import equipmentDefaults
            defaults = equipmentDefaults()
            self.STAreaFactor = noneCheck(self.STAreaFactor, defaults["STAreaFactor"])

        elif self.eqClass == "CHP":
            mandatoryData += MANDATORYDATA_CHP
        elif self.eqClass == "CH":
            if self.eqType in ["cooling tower (dry)",
                               "cooling tower (wet)",
                               "fresh or ground water"]:
                optionalData += OPTIONALDATA_CT
            
##        for sqlKey in SQLKEYS.keys():
        for sqlKey in mandatoryData:
            key, trans = SQLKEYS[sqlKey]
            key = noneCheck(key, sqlKey)
            
            if key in optionalData:
                continue
            elif not hasattr(self, key) or getattr(self, key) is None:
                logger.logError("EquipmentObject.checkCompleteness: missing parameter %s"%key)
                return False
        return True     

    def set(self, **kwargs):
        """
        Helper function for assigning a series of keys
        """
        for key in kwargs.keys():
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
