# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	GenericObject
#			
#------------------------------------------------------------------------------
#			
#       Basic object describing some object within the system
#
#==============================================================================
#
#	Version 3.0
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import EPSILON

from globals import logger
from GUI.status import Status

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck
from auxiliary.dataTransfer import DataTransfer

from modules.fluids import Fluid

# keys for parameters which are 1:1 passed to data
# struture: sqlLabel: (attr.-name [None if identical to sqlLabel], dictionary for conversion)


# TODO:
# currently this key is used for
# a) list those parameters to be written to SQL
# b) list those parameters to be written to GUI
# subset of (a) and (b) may be different, with conversions in between, which may not be only 1:1 conversions

                                 
class GenericObject(object):
    """
    Base object representing a process in the system model
    """
    def __init__(self, prnt, id=None, no=None, sqlData=None,
                 sqlKeys=None, dynamicTrans=None, mandatoryData=None, optionalData=None, linkKeys=[],
                 sqlTableName=None, sqlObjectList=None,
                 keyPId="ProjectID", colObjectNo="no"):
    
        """
        :param keyPId:  name of key for ProjectID (in old tables to be set to Questionnaire_id).
                        special case database objects: keyPId = None
        :param objectList: list of sql-objects in the systemModel (e.g. systemModel.equipment, systemModel.pipe)
        linkKeys: list of keys that are not overwritten in special case ccheck: overwriting from ANo - 1 to ANo 0
        """
        # set-up for specific object type
        self.sqlKeys = sqlKeys
        self.linkKeys = linkKeys
        
        self.dynamicTrans = dynamicTrans
        self.mandatoryData = mandatoryData
        self.optionalData = optionalData
        self.new = False

        self.sqlTableName = sqlTableName
        self.sqlObjectList = sqlObjectList

        self.keyPId = keyPId
        self.colObjectNo = colObjectNo
        
        # generic object properties
        self.prnt = prnt    # parent system model
        self.id = id        # identifier for process in data base (None or <=0 if local object)
        self.no = no        # number of the process in the system (from 0 to n-1)
##        self.name = None if no is None else sqlObjectList[no].Equipment

        # load data from sqlData
        if self.id is not None and sqlData is not None:
            self.loadFromDB(sqlData)
        else:
            # assign None to all pars
##            for parDict in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
            for parDict in [self.sqlKeys]:
                for sqlKey in parDict:
                    key = noneCheck(parDict[sqlKey][0], sqlKey)
                    setattr(self, key, None)

            self.new = True

        self.changed = self.new

    def __repr__(self):
        info = "GenericObject: \n"
        for parDict in [self.sqlKeys]:
            for sqlKey in parDict:
                key = noneCheck(parDict[sqlKey][0], sqlKey)
                info += "%s\t=%s\n"%(key, getattr(self, key))
        return info
 
    def loadFromDB(self, sqlData):
        """
        fetches data from DB
        """
        transfer = DataTransfer(self.dynamicTrans, self.sqlKeys)
        newData = transfer.write(sqlData)

        for key in newData:
            setattr(self, key, newData[key])
        
    def saveToDB(self, ANo=None, derivedData=None, checkIfChanged=False, warn=True):
        """
        stores the process in the DB
        or creates a new one, if id = None or <= 0

        ANo: special case for ANo=0 (used in consistency check). Copies process data from ANo=-1 to ANo 0
        derivedData: data which are not 1:1 contained within SQLKEYS-attributes
        """

        if checkIfChanged and not self.changed: return True
        
        # check if editing is allowed and completeness only for project related data
        if self.keyPId is None:     # special case for use in database objects (not assigned to any project)
            if self.id is None or self.id <=0:
                self.id = dbio.insertRow(self.sqlTableName)
            sqlTable = dbio.get(self.sqlTableName, id=self.id)
            data =  {}


        else:                       # regular case for objects of systemModel
            if ANo is None and not Status.prj.editAllowed(warn=warn):
                return False

            if Status.ANo >= 0 and not self.checkCompleteness() : # allow saving of incomplete data for ANo < 0!
                return False

            Status.prj.projectChanged()

            if self.id is None or self.id <=0:
                self.id = dbio.insertRow(self.sqlTableName)
                self.no = len(self.sqlObjectList)
                self.sqlObjectList.append(dbio.get(self.sqlTableName, id=self.id))


            data = {self.keyPId:Status.PId,
                    "AlternativeProposalNo": 0 if ANo==0 else Status.ANo}

            if self.colObjectNo is not None:
                data.update({self.colObjectNo: self.no+1})
                
            if ANo==0: # special case for consistency check
                sqlTable = dbio.get(self.sqlTableName, **data)

                if sqlTable is not None:
                    newId = sqlTable.id
                else:
                    logger.logError("GenericObject.saveToDB: corresponding object in alternative no. 0 does not exist")
                    return False
            else:
                sqlTable = self.sqlObjectList[self.no]

        # re-calculate all derived parameters
        if derivedData:
            data.update(derivedData)
            
        sqlTable.set(**data)

        sqlKeys = self.sqlKeys
        if ANo==0:
            for key in self.linkKeys:
                if key in sqlKeys.keys():
                    sqlKeys.pop(key)

        transfer = DataTransfer(None, sqlKeys)
        transfer.read(self, sqlTable)

        dbio.commit()        

        return True
    
    def deleteFromDB(self):
        if self.id is not None:
            dbio.delete(self.sqlTableName, id=self.id)
        
    def checkCompleteness(self):
        """
        checks if the data are complete and correct
        """
        for sqlKey in self.mandatoryData:
            key, trans = self.sqlKeys[sqlKey]
            key = noneCheck(key, sqlKey)
            
            if key in self.optionalData:
                continue
            elif not hasattr(self, key) or getattr(self, key) is None:
                logger.logTrack("GenericObject.checkCompleteness: missing parameter %s"%key)
                return False
        return True     

    def set(self, **kwargs):
        """
        Helper function for assigning a series of keys
        """
        for key in kwargs.keys():
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

        self.changed = True
