# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	PipeObject
#			
#------------------------------------------------------------------------------
#			
#       Basic object describing a pipe or duct
#
#==============================================================================
#
#	Version 3.0
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import EPSILON, PIPETYPES

from globals import logger
from GUI.status import Status

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck
from auxiliary.dataTransfer import DataTransfer

from modules.fluids import Fluid

from genericObject import GenericObject

# keys for parameters which are 1:1 passed to data
# struture: sqlLabel: (attr.-name [None if identical to sqlLabel], dictionary for conversion)


# TODO:
# currently this key is used for
# a) list those parameters to be written to SQL
# b) list those parameters to be written to GUI
# subset of (a) and (b) may be different, with conversions in between, which may not be only 1:1 conversions

yesNoDict = {"no": False, "yes": True}
HCDict = {"heating": True, "cooling": False}

SQLKEYS = {"Pipeduct" : ("name", None),
           "HeatDistMedium" : (None, None),
           "DistribCircFlow" : (None, None),
           "ToutDistrib" : (None, None),
            "TreturnDistrib" : (None, None),
            "PercentRecirc" : (None, None),
            "Tfeedup" : (None, None),
            "PressDistMedium" : (None, None),
            "TotLengthDistPipe" : (None, None),
            "UAPipe" : (None, None),
            "DDistPipe" : (None, None),
            "DeltaDistPipe" : (None, None),
            "HDEffAvg" : (None, None),
            "QHXPipe" : (None, None),
            "QWHPipe" : (None, None),
            "USHPipe" : (None, None),
            "USHm" : (None, None),
            "UPHProcm" : (None, None)}        # sqlkeys of data from SQL tables to be read; in case of different names for


MANDATORYDATA_ALL = ["Pipeduct", "HDEffAvg", "UAPipe"]
MANDATORYDATA_CLOSED_AND_OPEN = ["HeatDistMedium"]

OPTIONALDATA = []
                                 
class PipeObject(GenericObject):
    """
    Base object representing a process in the system model
    """
    def __init__(self, prnt, id=None, no=None, sqlData=None):
        """
        :param inStreams, outStreams: streamObjects for inflowing and outflowing streams
        """

        super(PipeObject, self).__init__(prnt, id, no, sqlData,
                 sqlKeys=SQLKEYS, mandatoryData=MANDATORYDATA_ALL, optionalData=OPTIONALDATA,
                 sqlTableName="qdistributionhc", sqlObjectList=prnt.pipe,
                 keyPId="Questionnaire_id", colObjectNo="PipeDuctNo")

        # space for StreamGlobals - objects
        self.USHGlb = None
        self.QWHGlb = None
        self.QHXGlb = None
        self.heating = None

        if self.new:
            self.name = _("new pipe/duct")
            self.heating = True

            self.pipeType = None
            self.PercentRecirc = 1.0
            self.setPipeType("closed")

        # calculated derived parameters

        if self.PercentRecirc is None or self.PercentRecirc > EPSILON:
            if self.UAPipe is not None and self.UAPipe < EPSILON and \
               self.HDEffAvg is not None and abs(self.HDEffAvg - 1.0) < EPSILON and \
               self.PercentRecirc is not None and abs(self.PercentRecirc - 1.0) < EPSILON and \
               self.HeatDistMedium is None:
                self.pipeType = "direct"
            else:
                self.pipeType = "closed" 
        else:
            self.pipeType = "open"
            self.PercentRecirc = 0.0
        
        self.HPerYearPipe = None

        self.fluidId = int(self.HeatDistMedium) if self.HeatDistMedium is not None else None
        self.updateFluid()
        
####        self.COPex = self.HPExCoolCOP if self.eqClass == "CH" else self.HPExHeatCOP
####
####        if self.HPerYear is not None and self.PartLoad is not None:
####            self.HPerYearNom = self.HPerYear*float(self.PartLoad)
####        else:
####            self.HPerYearNom = None
####
####        self.CPartLoad = self.ST_C0 if self.eqSubClass in ["CHP ORC", "CHP TFC"] or self.eqClass in ["CH"] else None            
####        self.ControlPar1 = self.ST_C1 if self.eqClass == "CHP" else None
####
####        self.fuelId = self.DBFuel_id
####        self.fuel = None if self.fuelId is None else Fuel(self.fuelId)
####
####        self.calculateActualPerformance()

        self.heating = None

    def __repr__(self):
        info = "Pipe(%s): id %s name %s"%(self.no, self.id, self.name)
        return info

    def updateFluid(self):
        self.fluid = Fluid(self.fluidId) if self.fluidId is not None else None
        self.HeatDistMedium = str(self.fluidId) if self.fluidId is not None else None   # conversion to str needed for storage in SQL
        
    def loadFromDB(self, sqlData):
        """
        fetches data from DB
        """
        super(PipeObject, self).loadFromDB(sqlData)

        # calculate derived properties
        self.fluidId = int(self.HeatDistMedium) if self.HeatDistMedium is not None else None
        self.updateFluid()

        # repair severe inconsistencies or missing parameters for ANo >= 0 in order to avoid crashes
        if hasattr(sqlData, "AlternativeProposalNo") and sqlData.AlternativeProposalNo != -1:
            pass

    def setPipeType(self, pipeType=None, previous=None):
        """
        updates the derived parameters consistent with the type
        """
        
        if pipeType is not None and pipeType not in PIPETYPES:
            print "PipeObject.setPipeType: error setting pipeType. PIPETYPES = %s"%PIPETYPES
            return

        if previous is None:
            previous = None if pipeType is None else self.pipeType
            
        if pipeType is not None: self.pipeType = pipeType
        changed = (previous != self.pipeType) 

        if self.pipeType=="open":
            self.PercentRecirc = 0.0
            if changed:
                self.HDEffAvg = None
                if previous=="direct":
                    self.UAPipe = None

        elif self.pipeType=="closed": 
            if changed:
                self.HDeffAvg = None
                self.PercentRecirc = None
                if previous=="direct":
                    self.UAPipe = None
            elif self.PercentRecirc==0:
                self.pipeType = "open"

        elif self.pipeType=="direct":
            self.UAPipe = 0.0
            self.PercentRecirc = 1.0
            self.HDEffAvg = 1.0

        self.mandatoryData = MANDATORYDATA_ALL[:]
        if self.pipeType != "direct":
            self.mandatoryData += MANDATORYDATA_CLOSED_AND_OPEN

    def saveToDB(self, ANo=None, warn=True):
        """
        stores the process in the DB
        or creates a new one, if id = None or <= 0

        ANo: special case for ANo=0 (used in consistency check). Copies process data from ANo=-1 to ANo 0
        """
        # re-calculate all derived parameters

        self.HeatDistMedium = str(self.fluidId) if self.fluidId is not None else None
        self.setPipeType(self.pipeType) # recalculate for security, if has not been done before

        super(PipeObject, self).saveToDB(ANo=ANo, derivedData=None, warn=warn)

        return True