# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ProcessObject
#			
#------------------------------------------------------------------------------
#			
#       Basic object describing a process
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from constants import TRANSPROCTYPES, HEATING_COOLING, TRANSYESNO, SCHEDULETYPES, TRANSSCHEDULETYPES
from constants import EPSILON
from globals import logger
from GUI.status import Status

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck
from auxiliary.dataTransfer import DataTransfer

from model.objects.processStreams import InflowingStream, OutflowingStream


# keys for parameters which are 1:1 passed to data
# struture: sqlLabel: (attr.-name [None if identical to sqlLabel], dictionary for conversion)


# TODO:
# currently this key is used for
# a) list those parameters to be written to SQL
# b) list those parameters to be written to GUI
# subset of (a) and (b) may be different, with conversions in between, which may not be only 1:1 conversions

yesNoDict = {"no": False, "yes": True}
HCDict = {"heating": True, "cooling": False}
procTypeDict = {"continuous": 0, "batch": 1}

SQLKEYS = {"Process": ("name", None),
           "Description": (None, None),
           "ProcType": (None, procTypeDict),
           "DBUnitOperation_id": ("UnitOp", None),
           "heating_cooling": ("heating", HCDict),
           "PT": (None, None),
           "DPSTmin": (None, None),
           "DPSTminMCD": (None, None),
           "ExternalSupply": (None, yesNoDict),
           "UPHc": (None, None),
           "UPHs": (None, None),
           "UPHm": (None, None),
           "UPHw": (None, None),
           "UPHwReq": (None, None),
           "UPHProc": (None, None),
           "QHXProc": (None, None),
           "QWHProc": (None, None),
           "HPerDayProc": ("HPerDay", None),
           "NBatch": ("NCycles", None),
           "HBatch": ("HCycle", None),
           "NDaysProc": ("NDays", None),
           "HPerYearProc": ("HPerYear", None),
           "HPerYearInFlow": (None, None),
           "HPerYearOutFlow": (None, None),
           "InFlowDuration": (None, None),
           "OutFlowDuration": (None, None),
           "StartUpDuration": (None, None),
           "TEnvProc": (None, None),
           "QBuildings_ID": ("buildingId", None),
           "UPH_SEC": (None, None),
           "PE_SEC": (None, None),
           "EL_SEC": (None, None),
           "VInFlowDay": (None, None),
           }        # sqlkeys of data from SQL tables to be read; in case of different names for



#TODO: check if this separate list is necessary. are included above !!!
SQLKEYS_SCHEDULE = {"HPerDayProc": ("HPerDay", None),
           "NBatch": ("NCycles", None),
           "HBatch": ("HCycle", None),
           "NDaysProc": ("NDays", None),
           "HPerYearProc": ("HPerYear", None),
           }        # sqlkeys of data from SQL tables to be read; in case of different names for

### process properties read-in from table
SQLKEYS_TABLE = {"ProcMedDBFluid_id": (None, None),
                 "PTStartUp": (None, None), 
                 "VolProcMed": (None, None),
                 "ThermalMass": (None, None),
                 "QOpProc": ("dotUPHm", None),
                 "UPH": (None, None)}

### derived process properties
SQLKEYS_DERIVEDPROPS = {"PartLoad": (None, None),
                        "EnclChartHCSupply": ("autoFit", None)}       
##           "heating_cooling": ("Heating_Cooling", HEATING_COOLING),

OPTIONALDATA = ["Description", "UnitOp", "buildingId", "StartUpDuration", "InFlowDuration", "OutFlowDuration",
                "UPH_SEC", "PE_SEC", "EL_SEC", "VInFlowDay"]
                                 
class ProcessObject():
    """
    Base object representing a process in the system model
    """
    
    def __init__(self, prnt, id=None, no=None,
                 inStreams=[], outStreams=[],
                 process_periods=[], combSchedules = [],
                 sqlData=None):

        """
        :param inStreams, outStreams: streamObjects for inflowing and outflowing streams
        """
        
        # generic object properties
        self.prnt = prnt    # parent system model
        self.id = id        # identifier for process in data base (None or <=0 if local object)
        self.no = no        # number of the process in the system (from 0 to n-1)
        self.name = None if no is None else prnt.process[no].Process

        # process object specific properties

        self.heating = True     # true or false depending on heating or cooling process
        self.MCD = None

        # TO DO: scheduleType, detailedSchedule and combSchedule are redundant -> unify !!!
        self.scheduleType = 0   # 0: simple, 1: detailed, 2: fit to data

        self.process_periods = process_periods[:]
        self.detailedSchedule = (len(process_periods)>0)
        
        self.combSchedules = combSchedules[:]
        self.combSchedule = (len(combSchedules)>0) and not self.detailedSchedule

        # in- and outStreams
        self.inStreams = inStreams[:]   #TODO: the [:] is necessary, as otherwise in/outstreams are kept as class attributes for the next instance
                                        #       don't understand why python behaves like this, but it does (-> systemModel.addObject !!!)
        self.outStreams = outStreams[:]

        # buildingId if related to building
        self.buildingId = None
        
        # space for StreamGlobals - objects
        self.UPHmGlb = None
        self.UPHsGlb = None
        self.UPHcGlb = [None]*self.nIn
        self.UPHwGlb = [None]*self.nOut

        if self.id is not None and sqlData is not None:
            self.loadFromDB(sqlData)
        else:
            # assign None to all pars
            for parDict in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
                for sqlKey in parDict:
                    key = noneCheck(parDict[sqlKey][0], sqlKey)
                    setattr(self, key, None)

            self.name = _("new process")
            self.heating = True
            self.ProcType = 0   # continuous
            self.DPSTmin = 10.0
            self.ExternalSupply = True

            self.ThermalMass = 0.0
            self.VolProcMed = 0.0
            self.dotUPHm = 0.0

            self.ScheduleType = 0
            self.HPerDay = 12.0
            self.HCycle = 12.0
            self.NCycles = 1
            self.NDays = 365

        self.deleted = []   # temporary buffer for streams (in- and out) to be deleted at next saveToDB.

        self.linkKeys = ["QBuildings_ID"]


    @property
    def nIn(self):
        return len(self.inStreams)

    @property
    def nOut(self):
        return len(self.outStreams)

    def loadFromDB(self, sqlData):
        """
        fetches data from DB
        """
        allKeys = {}
        for keys in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
            allKeys.update(keys)
        transfer = DataTransfer(None, allKeys)
        newData = transfer.write(sqlData)
        for key in newData:
            setattr(self, key, newData[key])
        # calculate derived properties
        if self.autoFit:
            self.scheduleType = 2
        elif self.detailedSchedule:
            self.scheduleType = 1
        else:
            self.scheduleType = 0

        if (self.ThermalMass is not None and self.ThermalMass < EPSILON) or \
           (self.VolProcMed is not None and self.VolProcMed < EPSILON):
            self.ThermalMass=0.0
            self.VolProcMed=0.0

        # repair severe inconsistencies or missing parameters for ANo >= 0 in order to avoid crashes
        if hasattr(sqlData, "AlternativeProposalNo") and sqlData.AlternativeProposalNo != -1:
            if self.PT is None:
                self.PT = -99.99    # assign something in order to avoid crashing
        
    def saveToDB(self, ANo = None, warn=True, simResultsOnly=False):
        """
        stores the process in the DB
        or creates a new one, if id = None or <= 0

        ANo: special case for ANo=0 (used in consistency check). Copies process data from ANo=-1 to ANo 0
        """

        if ANo is None and not Status.prj.editAllowed(warn=warn):
            return False

        if (not simResultsOnly) and (Status.ANo >= 0 and not self.checkCompleteness()): # allow saving of incomplete data for ANo < 0!
            return False

        Status.prj.projectChanged()

        if self.id is None or self.id <=0:
            # TO DO: create a new process if necessary
            self.id = dbio.insertRow("qprocessdata")
            self.no = len(self.prnt.process)
            self.prnt.process.append(dbio.get("qprocessdata", id=self.id))
            
        if ANo==0: # special case for consistency check
            procTable = dbio.get("qprocessdata", Questionnaire_id=Status.PId, AlternativeProposalNo=0, ProcNo = self.no+1)
            data = {"Questionnaire_id":Status.PId,
                    "AlternativeProposalNo": 0,
                    "ProcNo": self.no+1}

            if procTable is not None:
                newProcessId = procTable.id
            else:
                logger.logError("ProcessObject.saveToDB: corresponding process in alternative no. 0 does not exist")
                return False
        else:
            procTable = self.prnt.process[self.no]
            data = {"Questionnaire_id": Status.PId,
                    "AlternativeProposalNo": Status.ANo,
                    "ProcNo": self.no+1}

        # re-calculate all derived parameters
        if self.scheduleType == 0:
            self.detailedSchedule = False
            self.autoFit = False
        elif self.scheduleType == 1:
            self.detailedSchedule = True
            self.autoFit = False
        else:
            self.detailedSchedule = True
            self.autoFit = True

        if self.detailedSchedule:
            try:
                self.HPerYearOutFlow = self.scheduleOut.HPerYear
                self.HPerYearInFlow  = self.scheduleIn.HPerYear
            except:
                self.HPerYearOutFlow = None
                self.HPerYearInFlow = None
        
        # collect data defined in SQLKEYS, transform where necessary, and write to qprocessDAta

        data.update({"heating_cooling": "heating" if self.heating else "cooling"})

        procTable.set(**data)

        allKeys = {}
        for keys in [SQLKEYS, SQLKEYS_TABLE, SQLKEYS_DERIVEDPROPS]:
            allKeys.update(keys)

        for key in self.linkKeys:
            if key in allKeys.keys():
                allKeys.pop(key)
                
        transfer = DataTransfer(None, allKeys)
        transfer.read(self, procTable)


        if not simResultsOnly:
        # save all process streams
            if ANo==0:
                for stream in self.inStreams + self.outStreams:
                    stream.saveToDB(ANo=ANo, processId = newProcessId)

            else:
                for stream in self.inStreams + self.outStreams:
                    stream.procId = self.id
                    stream.saveToDB()

            for stream in self.deleted:
                stream.deleteFromDB()
            self.deleted = []

##        # save detailed schedule
##        if self.detailedSchedule:
##            self.periodSchedule.saveToDB()
##        else:
##            if self.periodSchedule:
##                self.periodSchedule.deleteFromDB()

        dbio.commit()        

        if not simResultsOnly:
            Status.processData.changeInProcess()

        return True

    def checkCompleteness(self):
        """
        checks if the data are complete and correct
        """
##        for stream in self.inStreams + self.outStreams:
        for stream in self.outStreams:
            if not stream.checkCompleteness():
                return False
            
        for sqlKey in SQLKEYS.keys():
            key, trans = SQLKEYS[sqlKey]
            key = noneCheck(key, sqlKey)
            
            if key in OPTIONALDATA:
                continue
            elif not hasattr(self, key) or getattr(self, key) is None:
                print "ProcessObject.checkCompleteness: missing parameter %s"%key
                logger.logError("ProcessObject.checkCompleteness: missing parameter %s"%key)
                return False
        return True     

    def getScheduleGlobals(self, schedule):
        """ temporary helper function
        for bringing to work TC2.02 again
        """
        self.set(**{"StartUpDuration"   : schedule.startup,
                    "InFlowDuration"    : schedule.inflow,
                    "OutFlowDuration"   : schedule.outflow,
                    "ScheduleTolerance" : schedule.tolerance,
                    "HolidayScale"      : int(schedule.holidayScale*100),
                    "HPerDayProc"       : float(schedule.HPerDay),
                    "HBatch"            : float(schedule.HBatch),
                    "NBatch"            : schedule.NBatch,
                    "NDaysProc"         : schedule.NDays,
                    "PartLoad"          : float(schedule.PartLoad),
                    "HPerYearInFlow"     : float(schedule.HPerYearInflow),
                    "HPerYearOutFlow"    : float(schedule.HPerYearOutflow)})

    def addInFlow(self):
        """
        adds an empty inflow
        """
        dummyName = _("inflow no. %s")%(len(self.inStreams)+1)
        self.inStreams.append(InflowingStream(dummyName, procId = self.id))

    def addOutFlow(self):
        """
        adds an empty inflow
        """
        dummyName = _("outflow no. %s")%(len(self.outStreams)+1)
        self.outStreams.append(OutflowingStream(dummyName, procId=self.id))

    def deleteInFlow(self, s):
        """
        deletes the inflow locally (moved from self.inStreams to self.deleted)
        """
        stream = self.inStreams.pop(s)
        self.deleted.append(stream)

    def deleteOutFlow(self, s):
        """
        deletes the outflow locally (moved from self.outStreams to self.deleted)
        """
        stream = self.outStreams.pop(s)
        self.deleted.append(stream)

    def set(self, **kwargs):
        """
        Helper function for assigning a series of keys
        """
        for key in kwargs.keys():
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

    @property
    def hasStartUp(self):
        return self.ThermalMass is None or self.ThermalMass > EPSILON

    @property
    def hasMaintenance(self):
        return self.dotUPHm > EPSILON or self.dotUPHm is None

    @property
    def hasMCD(self):
        return any([stream.MCD for stream in self.outStreams])


