# -*- coding: utf-8 -*-
#==============================================================================
#
#   E I N S T E I N
#
#   Expert System for an Intelligent Supply of Thermal Energy in Industry
#   (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   STREAMS
#
#------------------------------------------------------------------------------
#
#   Streams class for process in/outflowing media
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
"""
:mod:`einstein.modules.streams` -- In- and out-flowing streams
==============================================================

   :synopsis: Multiple streams of process input and output
   :author:   David Baehrens <david.baehrens@energyxperts.net>
"""
from globals import logger

from GUI.status import *
from GUI.GUITools import check
from constants import findKey, EPSILON
from _mysql_exceptions import ProgrammingError
from auxiliary.auxiliary import noneCheck
from modules.fluids import Fluid
import database.dbio as dbio

class InconsistentDataBaseError(Exception):
    """BUG: Data base structure is inconsistent."""
    pass

class ProcessStreamsNotFoundError(Exception):
    """No streams for process with given id found in DB."""
    pass

class StreamNotFoundError(Exception):
    """No stream with given name for process found in DB."""
    pass

class FluidNotFoundError(Exception):
    """No fluid with given id found in DB."""

def getInflowingStreamNamesFromDB(processId):
    """Retrieve all known in-flowing stream names for given processId from data base.
    
    :param processId: Id of process with in-flowing streams to look for.
    :type  processId: Integer

    :returns: All in-flowing stream names stored in data base for processId.
    :rtype:   List of strings
    
    """
    try:
##        streamsInIds = Status.DB.process_streams_in.qprocessdata_QProcessData_Id[processId].streams_in_id.column()
        streamsInIds = [entry.streams_in_id for entry in dbio.select("process_streams_in", qprocessdata_QProcessData_Id=processId)]
    except LookupError:
        return []

    try:
##        return Status.DB.streams_in.sql_select('id IN (%s) ORDER BY Name' % ', '.join(str(id) for id in streamsInIds)).Name.column()
        streams = [dbio.get("streams_in", id=inId) for inId in streamsInIds]

        streamnames = [stream.Name for stream in streams]
        streamnames.sort()
            
        return streamnames
    except ProgrammingError:
        raise InconsistentDataBaseError

def getOutflowingStreamNamesFromDB(processId):
    """Retrieve all known out-flowing stream names for given processId from data base.
    
    :param processId: Id of process with out-flowing streams to look for.
    :type  processId: Integer

    :returns: All out-flowing stream names stored in data base for processId.
    :rtype:   List of strings
    
    """
    try:
##        streamsOutIds = Status.DB.process_streams_out.qprocessdata_QProcessData_Id[processId].streams_out_id.column()
        streamsOutIds = [entry.streams_out_id for entry in dbio.select("process_streams_out", qprocessdata_QProcessData_Id=processId)]
    except LookupError:
        return []
    try:
##        return Status.DB.streams_out.sql_select('id IN (%s) ORDER BY Name' % ', '.join(str(id) for id in streamsOutIds)).Name.column()
        streams = [dbio.get("streams_out", id=outId) for outId in streamsOutIds]
        streamnames = [stream.Name for stream in streams]
        streamnames.sort()
        return streamnames
    except ProgrammingError:
        raise InconsistentDataBaseError

def getStreams(processId):
    """Retrieve all streams associated with the given processId form the
    data base.
    
    :param processId: Id of the process whose streams are to retrieved.
    :type processId: Integer
    
    :returns: A 2-Tuple of lists of InflowingStream and 
        OutflowingStream objects, (inflowing, outflowing). 
    :rtype: a pair of lists, the first of InflowingStreams, the second 
        of OutflowinStreams
    
    """
    result = ([],[])
    if processId is not None:
        (instreams, outstreams) = Status.prj.getStreams(processId)
        for streams, generator, target in ((instreams, InflowingStream, result[0]), 
                                           (outstreams, OutflowingStream, result[1])):
            for stream in streams:
                newstream = generator(stream.Name)
                try:
                    newstream.loadFromDB(processId)
                except FluidNotFoundError:
                    logger.logTrack("Fluid not found for stream: %s" % (stream.Name))
                    raise
                target.append(newstream)
    return result

KEYS_INFLOW =  ["Name",
                   "dbfluid_id",
                   "PTInFlow",
                   "PTInFlowRec",
                   "VInFlowCycle",
                   "mInFlowNom",
                   "HeatRecExist",
##                   "HInFlow",
##                   "XInFlow",
                   "UPHc",
                   "dotUPHc"]

KEYS_OUTFLOW =  ["Name",
                   "dbfluid_id",
                   "PTOutFlow",
                   "PTOutFlowRec",
                   "PTFinal",
                   "PTFinalRequired",
                   "VOutFlowCycle",
                   "mOutFlowNom",
                   "HeatRecExist",
                   "HeatRecOk",
                   "HOutFlow",
                   "XOutFlow",
                   "UPHw",
                   "UPHwReq",
                   "dotUPHw",
                   "MCD"]

SQLKEYS = {"dotUPHc": "QdotProc_c",
           "dotUPHw": "QdotProc_w"}

DEFAULTS_OUTFLOW = {"HeatRecExist": False,
                    "HeatRecOk": True,
                    "MCD": False}

DEFAULTS_INFLOW = {"HeatRecExist": False}

def substitute(key):
    """ substitutes SQL-key by key in stream object
    """
    if key=="QdotProc_c":
        return "dotUPHc"
    elif key=="QdotProc_w":
        return "dotUPHw"
    else:
        return key

class InflowingStream(object):
    '''
    Stream objects (local working copy)
    '''


    def __init__(self, Name, Medium=None, dbfluid_id=None, PTInFlow=None, PTInFlowRec=None, VInFlowCycle=None, mInFlowNom=None, HeatRecExist=None, HInFlow=None, XInFlow=None, UPHc=None, QdotProc_c=None,
                 ExternalSupply=True,
                 procId=None, sqlData=None):
        '''
        Constructor
        '''
        if sqlData is None:
            self.Name         = Name
            self.Medium       = Medium
            self.dbfluid_id   = dbfluid_id
            self.PTInFlow     = PTInFlow
            self.PTInFlowRec  = PTInFlowRec
            self.VInFlowCycle = VInFlowCycle
            self.mInFlowNom   = mInFlowNom
            self.HeatRecExist = HeatRecExist
            self.HInFlow      = HInFlow
            self.XInFlow      = XInFlow
            self.UPHc         = UPHc
            self.dotUPHc      = QdotProc_c
            self.ExternalSupply = ExternalSupply
            
            self.procId = None
            self.id = None

            self.setDefaults()
            
        else:
            for key in KEYS_INFLOW:
                sqlKey = SQLKEYS[key] if key in SQLKEYS.keys() else key
                if hasattr(sqlData, sqlKey):
                    val = getattr(sqlData, sqlKey)
                    setattr(self, key, val)

            if self.HeatRecExist in ["no", "yes"]:
                self.HeatRecExist = (self.HeatRecExist=="yes")
                
            self.id = sqlData.id
            self.procId = procId

            self.localCheck()

    def __repr__(self):
        string = "InStream %s (%s): VCycle %s"%(self.Name, self.id, self.VInFlowCycle)
        return string
        
    def deleteFromDB(self):
        """Delete all streams with our name associated with given process"""
        if self.id is not None:
            dbio.delete("streams_in", id=self.id)
            dbio.delete("process_streams_in", streams_in_id=self.id)
    
    def saveToDB(self, processId=None, ANo=None, PId=None):
        """Save the stream attached to the given processId."""
        ANo = noneCheck(ANo, Status.ANo)
        assert(ANo is not None)

        PId = noneCheck(PId, Status.PId)
        assert(PId is not None)

        self.setDefaults()
        
        dbHeatRecExist = None if self.HeatRecExist is None else 1*self.HeatRecExist
        
        dbStream  = {'Name'                 : dbio.check(self.Name),
                     'dbfluid_id'           : dbio.check(self.dbfluid_id),
                     'PTInFlow'             : dbio.check(self.PTInFlow),
                     'PTInFlowRec'          : dbio.check(self.PTInFlowRec),
                     'VInFlowCycle'         : dbio.check(self.VInFlowCycle),
                     'mInFlowNom'           : dbio.check(self.mInFlowNom),
                     'HeatRecExist'         : dbio.check(dbHeatRecExist),
##                     'HInFlow'              : dbio.check(self.HInFlow),
##                     'XInFlow'              : dbio.check(self.XInFlow),
                     'UPHc'                 : dbio.check(self.UPHc),
                     'QdotProc_c'           : dbio.check(self.dotUPHc),
                     'ProjectID'            : PId,
                     'AlternativeProposalNo': ANo}
        
        streamRow = None
        if processId is not None:   # special case: saving a stream identified by processId
            streamRow = findStream(self.Name, processId, "in")
            self.procId = processId
        elif self.id is not None:
            streamRow = dbio.get("streams_in", id=self.id)

        if streamRow is None:
            self.id = dbio.insertRow("streams_in", **dbStream)
            dbio.insertRow("process_streams_in", **{'qprocessdata_QProcessData_Id' : self.procId,
                                                 'streams_in_id'                : self.id,
                                                 'ProjectID'                    : PId,
                                                 'AlternativeProposalNo'        : ANo})
        else:
            streamRow.set(**dbStream)
        
        
    def loadFromDB(self, processId):
        """Load stream with our name associated with given processId"""
        streamRow = findStream(self.Name, processId, "in")
        if streamRow is None:
            raise StreamNotFoundError

        self.id = streamRow.id        
        self.dbfluid_id = streamRow.dbfluid_id
        if self.dbfluid_id is None:
            self.Medium = None
        else:
            try:
                self.Medium = Fluid(self.dbfluid_id).name
            except IndexError:
                raise FluidNotFoundError("Unknown fluid id %s" % (self.dbfluid_id))
        self.PTInFlow     = streamRow.PTInFlow
        self.PTInFlowRec  = streamRow.PTInFlowRec
        self.VInFlowCycle = streamRow.VInFlowCycle
        self.mInFlowNom   = streamRow.mInFlowNom
        if streamRow.HeatRecExist is None:
            self.HeatRecExist = None
        elif streamRow.HeatRecExist in ["yes", "no"]:
            self.HeatRecExist = (streamRow.HeatRecExist=="yes")
        else:
            self.HeatRecExist = (streamRow.HeatRecExist == 1)
        self.HInFlow      = streamRow.HInFlow
        self.XInFlow      = streamRow.XInFlow
        self.UPHc         = streamRow.UPHc
        self.dotUPHc   = streamRow.QdotProc_c
        self.ExternalSupply = True
        self.procId = processId

        self.localCheck()

    def setDefaults(self):
        self.localCheck()
        
        for key in DEFAULTS_INFLOW.keys():
            val = getattr(self, key)
            if val is None:
                setattr(self, key, DEFAULTS_INFLOW[key])

    def localCheck(self):
        if self.HeatRecExist == False:
            if self.PTInFlow is None:
                self.PTInFlow = self.PTInFlowRec
            elif self.PTInFlowRec is None:
                self.PTInFlowRec = self.PTInFlow

        elif self.HeatRecExist is None and (self.PTInFlow is not None) and (self.PTInFlowRec is not None):
            self.HeatRecExist == (abs(self.PTInFlow - self.PTInFlowRec) >= EPSILON)

    def checkCompleteness(self):
        """
        checks if the data are complete and correct
        """
        self.setDefaults()
                       
        missingOptionalKeys = []
        for key in KEYS_INFLOW:
            if not hasattr(self, key) or getattr(self, key) is None:
                if key in []:
                    missingOptionalKeys.append(key)
                else:
                    logger.logError("Stream.checkCompleteness: missing parameter %s in stream %s"%(key, self.Name))
                    print "Stream.checkCompleteness: missing parameter %s in stream %s"%(key, self.Name)
                    return False
        return True     


class OutflowingStream(object):
    '''
    classdocs
    '''

    def __init__(self, Name, Medium=None, dbfluid_id=None, PTOutFlow=None, PTOutFlowRec=None, 
                 PTFinal=None, VOutFlowCycle=None, mOutFlowNom=None, 
                 HeatRecExist=None, HeatRecOK=None, HOutFlow=None, 
                 XOutFlow=None, UPHw=None, QdotProc_w=None, 
                 PTFinalRequired=None, UPHwReq=None, ExternalSupply=True,
                 procId=None, sqlData=None):
        '''
        Constructor
        '''
        if sqlData is None:
            self.Name          = Name
            self.Medium        = Medium
            self.dbfluid_id    = dbfluid_id
            self.PTOutFlow     = PTOutFlow
            self.PTOutFlowRec  = PTOutFlowRec
            self.PTFinal       = PTFinal
            self.VOutFlowCycle = VOutFlowCycle
            self.mOutFlowNom   = mOutFlowNom
            self.HeatRecExist  = HeatRecExist
            self.HeatRecOk     = HeatRecOK
            self.HOutFlow      = HOutFlow 
            self.XOutFlow      = XOutFlow
            self.UPHw          = UPHw
            self.dotUPHw       = QdotProc_w
            self.PTFinalRequired = PTFinalRequired
            self.UPHwReq       = UPHwReq
            self.ExternalSupply = ExternalSupply

            self.MCD = None
            
            self.procId = procId
            self.id = None
            
            self.setDefaults()

        else:
            for key in KEYS_OUTFLOW:
                sqlKey = SQLKEYS[key] if key in SQLKEYS.keys() else key
                if hasattr(sqlData, sqlKey):
                    val = getattr(sqlData, sqlKey)
                    setattr(self, key, val)

            if self.HeatRecExist in ["no", "yes"]:
                self.HeatRecExist = (self.HeatRecExist=="yes")
            elif self.HeatRecExist is not None:
                self.HeatRecExist = (self.HeatRecExist==1)
            self.id = sqlData.id                
            self.procId = procId

            self.localCheck()
            
        self.MCD = hasattr(self, "PTFinalRequired") and (self.PTFinalRequired is not None)


    def __repr__(self):
        string = "OutStream %s (%s): VCycle %s HeatRecExist %s"%(self.Name, self.id, self.VOutFlowCycle, self.HeatRecExist)
        return string
        
    def deleteFromDB(self):
        """Delete all streams with our name associated with given process"""
        if self.id is not None:
            dbio.delete("streams_out",id=self.id)
            dbio.delete("process_streams_out", streams_out_id=self.id)            
    
    def saveToDB(self, processId=None, ANo=None, PId=None):
        """Save the stream attached to the given processId."""
        if ANo is None:
            ANo = Status.ANo
            assert(ANo is not None)
        if PId is None:
            PId = Status.PId
            assert(PId is not None)

        self.setDefaults()
                       
        dbHeatRecExist = None if self.HeatRecExist is None else 1*self.HeatRecExist
        dbHeatRecOk = None if self.HeatRecOk is None else 1*self.HeatRecOk

        PTFinalRequired = None if self.MCD is None else self.PTFinalRequired

        dbStream  = {'Name'                 : dbio.check(self.Name),
                     'dbfluid_id'           : dbio.check(self.dbfluid_id),
                     'PTOutFlow'            : dbio.check(self.PTOutFlow), 
                     'PTOutFlowRec'         : dbio.check(self.PTOutFlowRec), 
                     'PTFinal'              : dbio.check(self.PTFinal), 
                     'VOutFlowCycle'        : dbio.check(self.VOutFlowCycle),
                     'mOutFlowNom'          : dbio.check(self.mOutFlowNom),
                     'HeatRecExist'         : dbio.check(dbHeatRecExist), 
                     'HeatRecOk'            : dbio.check(dbHeatRecOk), 
                     'HOutFlow'             : dbio.check(self.HOutFlow),
                     'XOutFlow'             : dbio.check(self.XOutFlow), 
                     'UPHw'                 : dbio.check(self.UPHw), 
                     'QdotProc_w'           : dbio.check(self.dotUPHw),
                     'PTFinalRequired'      : dbio.check(PTFinalRequired),
                     'UPHwReq'              : dbio.check(self.UPHwReq),
                     'ProjectID'            : PId,
                     'AlternativeProposalNo': ANo} 

        streamRow = None
        if processId is not None:
            streamRow = findStream(self.Name, processId, "out")
            self.procId = processId
        elif self.id is not None:
            streamRow = dbio.get("streams_out", id = self.id)

        if streamRow is None: # no input stream with our name found for given processId
            self.id = dbio.insertRow("streams_out", **dbStream)
            dbio.insertRow("process_streams_out", **{'qprocessdata_QProcessData_Id' : self.procId,
                                                 'streams_out_id'                : self.id,
                                                 'ProjectID'                     : PId,
                                                 'AlternativeProposalNo'         : ANo})
        else:   
            streamRow.set(**dbStream)        
        
    def loadFromDB(self, processId):
        """Load stream with our name associated with given processId"""

        streamRow = findStream(self.Name, processId, "out")
        if streamRow is None:
            raise StreamNotFoundError

        self.id = streamRow.id
        self.dbfluid_id = streamRow.dbfluid_id
        if self.dbfluid_id is None:
            self.Medium = None
        else:
            try:
                self.Medium = Fluid(self.dbfluid_id).name
            except IndexError:
                raise FluidNotFoundError
        self.PTOutFlow     = streamRow.PTOutFlow
        self.PTOutFlowRec  = streamRow.PTOutFlowRec
        self.PTFinal       = streamRow.PTFinal
        self.VOutFlowCycle = streamRow.VOutFlowCycle
        self.mOutFlowNom   = streamRow.mOutFlowNom
        if streamRow.HeatRecExist is None:
            self.HeatRecExist = None
        elif streamRow.HeatRecExist in ["yes", "no"]:
            self.HeatRecExist = (streamRow.HeatRecExist=="yes")
        else:
            self.HeatRecExist = (streamRow.HeatRecExist == 1)
        if streamRow.HeatRecOk is None:
            self.HeatRecOk = None
        else:
            self.HeatRecOk = (streamRow.HeatRecOk == 1)
            
        self.HOutFlow      = streamRow.HOutFlow 
        self.XOutFlow      = streamRow.XOutFlow
        self.UPHw          = streamRow.UPHw
        self.dotUPHw    = streamRow.QdotProc_w
        self.PTFinalRequired = streamRow.PTFinalRequired
        self.UPHwReq       = streamRow.UPHwReq
        self.ExternalSupply = True
        self.procId = processId

        self.MCD = hasattr(self, "PTFinalRequired") and (self.PTFinalRequired is not None)

        self.localCheck()

    def setDefaults(self):
        for key in DEFAULTS_OUTFLOW.keys():
            val = getattr(self, key)
            if val is None:
                setattr(self, key, DEFAULTS_OUTFLOW[key])

    def localCheck(self):
        if self.HeatRecExist == False:
            if self.PTOutFlow is None:
                self.PTOutFlow = self.PTOutFlowRec
            elif self.PTOutFlowRec is None:
                self.PTOutFlowRec = self.PTOutFlow

        elif self.HeatRecExist is None and (self.PTOutFlow is not None) and (self.PTOutFlowRec is not None):
            self.HeatRecExist == (abs(self.PTOutFlow - self.PTOutFlowRec) >= EPSILON)

    def checkCompleteness(self):
        """
        checks if the data are complete and correct
        """
        self.setDefaults()

        optionalKeys = ["PTFinalRequired", "XOutFlow", "QdotProc_w"]
##        if not self.HeatRecOk:
##            optionalKeys.append("HOutFlow")
            
        missingOptionalKeys = []
        for key in KEYS_OUTFLOW:
            if not hasattr(self, key) or getattr(self, key) is None:
                if key in optionalKeys:
                    missingOptionalKeys.append(key)
                else:
                    print "Stream.checkCompleteness: missing parameter %s in stream %s"%(key, self.Name)
                    logger.logError("Stream.checkCompleteness: missing parameter %s in stream %s"%(key, self.Name))
                    return False
        return True     

def findStream(name=None, processId=None, streamType="in"):
    """Currently streams are referred to by a unique name and the processId
    """
    try:
        tableName = "process_streams_"+streamType
        colName = "streams_%s_id"%streamType
        streamIds = [getattr(ps, colName) for ps in dbio.select(tableName, qprocessdata_QProcessData_Id=processId)]
    except LookupError:
        return None
    tableName = "streams_"+streamType
    table = dbio.select(tableName, Name=dbio.check(name))
    for row in table:
        if row.id in streamIds:
            return row
    return None

