# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CONNECT
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Builds the connection matrices between the different subsystems:
#       Fuels - Equipes - Pipes/Ducts - Processes
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from GUI.status import *
import numpy as np
from modules.messageLogger import *
import database.dbio as dbio
from globals import logger
from auxiliary.auxiliary import toUnicode as U


def loadConnections(silent=False):
    """Deduce the connections of equipment, pipes and processes from 
    the entries in the SQL tables and write them into the list
    systemModel.FET/USH/UPHLink

    Objects have to be loaded into SQL before
    """

    systemModel = Status.int.systemModel    
    # get list of sub-systems
    fuels = systemModel.fuelObject
    Status.NFuels = len(fuels)
    Status.NFET = Status.NFuels+1
    fuelIDs = [fuel.fuelId for fuel in fuels]
    
    equipment = systemModel.eqObject
    equipmentSQL = systemModel.equipment
    eqNameList = [eq.name for eq in equipment]
    
    Status.NEquipe = len(equipment)
    
    pipes = systemModel.pipeObject
    pipesSQL = systemModel.pipe

    pipeIDList = [pipe.id for pipe in pipes]
    pipeNameList = [pipe.name for pipe in pipes]
    
    Status.NPipeDuct = len(pipes)

    processes = systemModel.processObject
    processesSQL = systemModel.process
    procNameList = [proc.name for proc in processes]
    
    Status.NThProc = len(processes)
    
    hxes = systemModel.hx
    
    Status.NHX = len(hxes)
    
    whees = systemModel.wheeObject
    wheesSQL = systemModel.whee
    wheeNameList = [whee.name for whee in whees]
    
    Status.NWHEE = len(whees)


    systemModel.FETFuelLink = []
    systemModel.USHLink = []
    systemModel.UPHLink = []
    systemModel.UPCReqLink = []
    systemModel.FEThLink = []
    
    # 1. FETi-FETj-Link: fuels used in equipment
    for j, eq in enumerate(equipment):
        fuelId = eq.fuelId

        if fuelId in fuelIDs:
            i = fuelIDs.index(fuelId)
            systemModel.FETFuelLink.append((i,j))
        elif fuelId is not None:
            showError(_("Fuel used in equipment no. %s is not specified or is not in fuel list")%(j+1))
        
    # 2. USHj-USHm-Link: connection of equipment to pipe
    for j, eqSQL in enumerate(equipmentSQL):
        eq = equipment[j]
        pipeIDStr = eqSQL.PipeDuctEquip
        pipeNames, pipeIds = Status.prj.getPipeNames(pipeIDStr)
        connectedPipes = []
        for pipeId in pipeIds:
            if pipeId in pipeIDList:
                m = pipeIDList.index(pipeId)
                pipe = pipes[m]
                systemModel.USHLink.append((j, m))
                connectedPipes.append(m)

                if pipe.heating is None:
                    pipe.heating = eq.heating  
                elif pipe.heating != eq.heating:
                    if not silent:
                        logWarning(_("Pipe no. %s (%s) is connected to heat and cold suppliers. Please check")
                                  % (m + 1, pipe.name))
                    
        if len(connectedPipes)==0 and eq.eqType not in ["CHP ORC", "CHP TFC"]: # some CHP types may be used for el.-generation only
            logWarning(_("Equipment no. %s is not connected to any pipe")%(j+1))
                
    # 3. UPHm-UPHk Link and 4. UPHm-UPCk Link: connection of pipe to process (incl. mandatory cooling down)
    for k, proc in enumerate(processes):
        procSQL = processesSQL[k]

        pipeName = procSQL.PipeDuctProc
        if pipeName in pipeNameList:
            m = pipeNameList.index(pipeName)
            pipe = pipes[m]
            systemModel.UPHLink.append((m, k))
            if pipe.heating is None:
                pipe.heating = proc.heating
            elif pipe.heating != proc.heating:
                if not silent:
                    if pipe.heating:
                        logError(_("Pipe no. %s is connected to a heat supplier but also to the cooling process %s.")
                                  % (m + 1, proc.name))
                    else:
                        logError(_("Pipe no. %s is connected to a cold supplier but also to the heating process %s.")
                                  % (m + 1, proc.name))
                
        pipeNameMCD = procSQL.PipeDuctForMandatoryCoolingDown            
        if pipeNameMCD in pipeNameList:
            m = pipeNameList.index(pipeNameMCD)
            pipe = pipes[m]
            systemModel.UPCReqLink.append((m, k))
            if pipe.heating is None:
                pipe.heating = False
            elif pipe.heating != False:
                if not silent:
                    if pipe.heating:
                        logError(_("Pipe no. %s is connected to a heat supplier but also as mandatory cooling down to the heating process %s.")
                                  % (m + 1, proc.name))
                    else:
                        logError(_("Pipe no. %s is connected to the cooling process %s as mandatory cooling down supply. That is not sensible.")
                                  % (m + 1, proc.name))
    
    
    # 5. UPHm-FETh Link: connection of pipe to heat consuming equipment
    for j, eq in enumerate(equipment):
        if eq.HeatSourceHT is not None:
            pipeId = int(eq.HeatSourceHT)
            if pipeId in pipeIDList:
                m = pipeIDList.index(pipeId)
                systemModel.FEThLink.append((m, j))

                if pipe.heating is None:
                    pipe.heating = True
                elif not pipe.heating:
                    if not silent:
                        logWarning(_("Pipe no. %s is connected to a cold supplier but also to the heat supply for equipment %s.")
                                  % (m + 1, eq.name))
        
    # 6. QHX/QWH - Eq/Pipe/Proc Links
    systemModel.QWHEqLink = []
    systemModel.QWHPipeLink = []
    systemModel.QWHProcLink = []
    systemModel.QWHEELink = []

    systemModel.QHXEqLink = []
    systemModel.QHXPipeLink = []
    systemModel.QHXProcLink = []
    
    for h, hx in enumerate(hxes):
        sourceName = hx.HXSource
        sinkName = hx.HXSink
        
        sourceFluidID = None
        sinkFluidID = None
        
        for j, eq in enumerate(equipment):
            if sourceName == eq.name:
                systemModel.QWHEqLink.append((j, h))
                
            if sinkName == eq.name:
                systemModel.QHXEqLink.append((h, j))

        #TODO: assign here appropriate fluid IDs for the equipment heat exchangers
        #       or alternatively skip all, if fluidIds of HXes are determined in another place
        
        for m, pipe in enumerate(pipes):
            if sourceName == pipe.name:
                systemModel.QWHPipeLink.append((m, h))
                if sourceFluidID is None: sourceFluidID = pipe.fluidId
                
            if sinkName == pipe.name:
                Status.QHXPipeLink[m][h] = 1
                systemModel.QHXPipeLink.append((h, m))
                if sinkFluidID is None: sinkFluidID = pipe.fluidId

        for k, proc in enumerate(processes):
            procSQL = processesSQL[k]
            if sourceName == proc.name:
                systemModel.QWHProcLink.append((k, h))

##                if len(sourceprocesses) > 0:
##                    sourceFluidID = sourceprocesses[0].ProcMedOut
                
            if sinkName == proc.name:
                systemModel.QHXProcLink.append((h, k))
                sinkFluidID = procSQL.SupplyMedDBFluid_id
        
        for n, whee in enumerate(whees):                         #WHEE can only be source, not sink !!!
            if sourceName == whee.name:
                wheeSQL = wheesSQL[n]
                systemModel.QWHEELink.append((n, h))
                sourceFluidID = wheeSQL.WHEEMedium

        try:
            sourceFluidID = int(sourceFluidID)
        except:
            sourceFluidID = None
            
        try:
            sinkFluidID = int(sinkFluidID)
        except:
            sinkFluidID = None
        
        hxes[h].FluidIDSource = dbio.check(sourceFluidID)
        hxes[h].FluidIDSink = dbio.check(sourceFluidID)
    
##    if not silent:
##        for k in xrange(NK):
##            if check_k[k] == 0:
##                showError(_("No heat supply (pipe or heat exchanger) connected to process no. %s")%(k+1))
##        for k in xrange(NK):
##            if (checkReq_k[k] == 0) and systemModel.processObject[k].hasMCD:
##                showError(_("Process no. %s has a mandatory cooling down demand but no cooling pipe is connected.")%(k+1))

    createLinkTables()
    
def createLinkTables():
    """
    Function for backward compatibility

    Transforms the information contained in systemModel.FET/USH/UPH/...-Link
    into the corresponding tables in Status.int.FET/USH/UPH/...-Link
    """

    systemModel = Status.int.systemModel
    NI = systemModel.NI
    NJ = systemModel.NJ
    NK = systemModel.NK
    NM = systemModel.NM
    NN = systemModel.NN
    NH = systemModel.NH
    
    Status.FETFuelLink  = buildLinkTable(NJ, NI, systemModel.FETFuelLink)
    Status.FETelLink    = buildLinkTable(NJ, 1, [])
    Status.FETelLink += 1 #all equipment potentially consume something of (parasitic) electricity.

##    Status.int.FEThLink = buildLinkTable(NJ, NM, systemModel.FEThLink)

    ## TODO: USHLink and the others should be eliminated too and substitued by the corresponding matrices in systemModel   
    Status.int.USHLink = buildLinkTable(NM, NJ, systemModel.USHLink)
##    Status.int.UPHLink = buildLinkTable(NK, NM, systemModel.UPHLink)
##    Status.int.UPCReqLink = buildLinkTable(NK, NM, systemModel.UPCReqLink)

    Status.QWHEqLink   = buildLinkTable(NJ, NH, systemModel.QWHEqLink, transpose=True)  
    Status.QWHPipeLink = buildLinkTable(NM, NH, systemModel.QWHPipeLink, transpose=True)  
    Status.QWHProcLink = buildLinkTable(NK, NH, systemModel.QWHProcLink, transpose=True)
    Status.QWHEELink   = buildLinkTable(NN, NH, systemModel.QWHEELink, transpose=True)

    Status.QHXEqLink   = buildLinkTable(NJ, NH, systemModel.QHXEqLink)
    Status.QHXPipeLink = buildLinkTable(NM, NH, systemModel.QHXPipeLink)
    Status.QHXProcLink = buildLinkTable(NK, NH, systemModel.QHXProcLink)

def buildLinkTable(n1, n2, link, transpose=False):
    linkTable = np.zeros((n1, n2), dtype=int)
    for i1, i2 in link:
        if transpose:
            linkTable[i1][i2] = 1
        else:
            linkTable[i2][i1] = 1
    return linkTable

def getAllProcessesConnectedTo(m):
    """Return the list of indices of process connected to the pipe with
    the index m.
    
    """
    systemModel = Status.int.systemModel
    result = [link[1] for link in systemModel.UPHLink if link[0]==m]
    return result

def getAllEquipmentsConnectedTo(m):
    """Return the list of indices of equipments (FETh) connected to the
    pipe with the index m.
    
    """
    systemModel = Status.int.systemModel
    result = [link[1] for link in systemModel.FEThLink if link[0]==m]
    return result

def getAllMandCoolDownProcessesConnectedTo(m):
    """Return the list of indices of processes that have a mandatory
    cooling down demand and are connected to the pipe with the index m.
    
    """
    systemModel = Status.int.systemModel
    result = [link[1] for link in systemModel.UPCReqLink if link[0]==m]
    return result
    


class EnergyFlow():
    """
    class adapted from Flow in CheckMassFlows
    represents any connection between objects in the EnergyFlow plot
    (no heat exchanger connections !!!)
    """
    def __init__(self, fromNode, fromNo, toNode, toNo, flow=None, id=None, real=True):
        """ fromNode and toNode are the fuel, equipment, pipe and process objects in systemModel !!!
        """
        self.fromNode = fromNode
        self.toNode = toNode
        
        self.fromNo = fromNo
        self.toNo = toNo

        self.real = real    # False: information flow only
        
        self.flow = flow
        self.id = id

    def __repr__(self):
        string = "%s[%s]=>%s[%s]"%(self.fromNode,
                                      self.fromNo,
                                      self.toNode,
                                      self.toNo)
        return string

    def saveToDB(self):

        fromId = Status.int.systemModel.objectId(self.fromNode, self.fromNo)
        toId = Status.int.systemModel.objectId(self.toNode, self.toNo)
        
        if self.fromNode=="FUEL" and self.toNode=="EQ": #FETFuelLink
            eqClass = Status.int.systemModel.eqObject[self.toNo].eqClass
            if eqClass in ["ST", "CH", "HTP"]:
                logger.showWarning(_("Fuel supply is not possible for this type of equipment"))
                return False
            else:
                eq = dbio.get("qgenerationhc", id=toId)
                fuelId = Status.int.systemModel.fuelObject[self.fromNo].fuelId
                eq.DBFuel_id = fuelId
            
        elif self.fromNode=="EL":
            logger.showWarning(_("Cannot delete this connection.") + " " + \
                               _("Electricity connections are built automatically"))
            return False
            
        elif self.fromNode == "EQ" and self.toNode=="PIPE": #USHLink
            eq = dbio.get("qgenerationhc", id=fromId)
            pipeIDString = eq.PipeDuctEquip
            pipeNames, pipeIds = Status.prj.getPipeNames(pipeIDString)
            if toId not in pipeIds:
                heatingPipe = Status.int.systemModel.pipeObject[self.toNo].heating
                if (heatingPipe and Status.int.systemModel.eqObject[self.fromNo].eqClass in ["BB", "CHP", "HTP", "ST", "HP"]) or \
                   (not heatingPipe and Status.int.systemModel.eqObject[self.fromNo].eqClass in ["CH"]) or heatingPipe is None:
                    pipeIds.append(toId)
                else:
                    logger.showWarning(_("Cannot connect cooling pipe to heating equipment or vice versa"))
                    return False
            eq.PipeDuctEquip = Status.prj.getPipeIDString(eq, pipeIds)
            
        elif self.fromNode == "PIPE":
            pipe = dbio.get("qdistributionhc", id=fromId)
            heatingPipe = Status.int.systemModel.pipeObject[self.fromNo].heating
            
            if self.toNode == "PROC":
                proc = dbio.get("qprocessdata", id=toId)
                heatingProc = (proc.heating_cooling=="heating")
                if (heatingProc and (heatingPipe != False)) or (not heatingProc and heatingPipe != True) or heatingPipe is None:
                    proc.PipeDuctProc = pipe.Pipeduct
                elif not heatingPipe and heatingProc and Status.int.systemModel.processObject[self.toNo].MCD:
                    proc.PipeDuctForMandatoryCoolingDown = pipe.Pipeduct
                else:
                    logger.showWarning(_("Cannot connect cooling pipe to heating process or vice versa"))
                    return False
                
            elif self.toNode == "EQ":
                if heatingPipe:
                    eq = dbio.get("qgenerationhc", id=toId)
                    if eq.EquipType in ["CHP ORC", "CHP TFC", "thermal chiller (air cooled)", "thermal chiller (water cooled)"]:                   
                        eq.HeatSourceHT = fromId
                    else:
                        logger.showWarning(_("Equipment %s cannot be connected to a heat supply")%eq.Equipment)
                        return False
                        
                else:
                    logger.showWarning(_("Cannot connect cooling pipe to heat input of an equipment"))
                    return False

        else:
            return False

        Status.int.systemModel.changed()
        dbio.commit()
        return True
            
    def deleteFromDB(self):
        fromId = Status.int.systemModel.objectId(self.fromNode, self.fromNo)
        toId = Status.int.systemModel.objectId(self.toNode, self.toNo)

        if self.fromNode=="FUEL" and self.toNode=="EQ": #FETFuelLink
            eq = dbio.get("qgenerationhc", id=toId)
            eq.DBFuel_id = None

        elif self.fromNode=="EL":
            logger.showWarning(_("Cannot delete this connection.") + " " + \
                               _("Electricity connections are built automatically"))
            return False
            
        elif self.fromNode == "EQ" and self.toNode=="PIPE": #USHLink
            eq = dbio.get("qgenerationhc", id=fromId)
            pipeIDString = eq.PipeDuctEquip
            pipeNames, pipeIds = Status.prj.getPipeNames(pipeIDString)
            if toId in pipeIds:
                index = pipeIds.index(toId)
                pipeIds.pop(index)
            else:
                return False
            eq.PipeDuctEquip = Status.prj.getPipeIDString(eq, pipeIds)
            
        elif self.fromNode == "PIPE":            
            pipe = dbio.get("qdistributionhc", id=fromId)
            heatingPipe = Status.int.systemModel.pipeObject[self.fromNo].heating

            if self.toNode == "PROC":
                proc = dbio.get("qprocessdata", id=toId)
                if not heatingPipe and (proc.heating_cooling=="heating"):
                    proc.PipeDuctForMandatoryCoolingDown = None
                else:
                    proc.PipeDuctProc = None
                
            elif self.toNode == "EQ":
                eq = dbio.get("qgenerationhc", id=toId)
                eq.HeatSourceHT = None
            else:
                return False

        Status.int.systemModel.changed()
        dbio.commit()
        return True
        
##    def setMode(self, mode):
##        real = (mode==0)
##        if real != self.real:
##            if self.real:           # change from real to virtual
##                self.exportToNode(delete=True) # delete as 
##            self.real = real
##            self.exportToNode()
##            return True
##        else:
##            return False
