# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	SYSTEMMODEL
#			
#------------------------------------------------------------------------------
#			
#	Definition of system components and connections
#       This class should gradually substitute the "interfaces" class
#       collecting a mirror of all database objects and its connections.
#
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from modules.heatRecovery.hxConnection import StreamGlobals, HXConnection, genericStreamName
from modules.heatRecovery.hxConnection import buildHXNet
from systemConnections import loadConnections
from GUI.status import Status
from constants import getEquipmentClass, EPSILON
from modules.fluids import getDefaultAir
from modules.messageLogger import logTrack, logError
from auxiliary.auxiliary import noneCheck
from modules.schedules.periodSchedule import getProcessPeriods, PeriodSchedule, ScheduleNotFoundError
from modules.schedules.schedules import getProcessCombSchedules
from modules.ccheck.dataSet import loadDataSets

from model.objects.elObject import ElObject
from model.objects.equipmentObject import EquipmentObject
from model.objects.pipeObject import PipeObject
from model.objects.processObject import ProcessObject
from model.objects.processStreams import InflowingStream, OutflowingStream

class SystemModel():
    """
    status: indicates the current state of the model
            -1: no data loaded
            0: objects have been loaded from DB
            1: connections, stream list and HX network built
    """

    def __init__(self):
        self.process = []
        self.processStreamsIn = []
        self.processStreamsOut = []
        self.equipment = []
        self.eqTypes = []
        self.pipe = []
        self.whee = []
        self.hx = []
        self.bui = []
        
        self.streams = []

        self.status = -1 # -1: when created, 0 after loading components, 1 after setting up connections and building stream list 
        Status.int.pa = None

        self.projectData, self.generalData = Status.prj.getProjectData()

        self.elObject = []
        self.fuelObject = []
        self.processObject = []
        self.eqObject = []
        self.pipeObject = []
        self.wheeObject = []
        self.hxObject = []

    @property
    def NI(self):
        return len(self.fuelObject)
        
    @property
    def NJ(self):
        return len(self.eqObject)
        
    @property
    def NK(self):
        return len(self.processObject)
        
    @property
    def NM(self):
        return len(self.pipeObject)
        
    @property
    def NN(self):
        return len(self.wheeObject)
        
    @property
    def NH(self):
        return len(self.hxObject)
        
    def getComponentsFromDB(self):

        logTrack("SystemModel.getComponentsFromDB")


        self.holidays = Status.prj.getHolidays()

        self.process = Status.prj.getProcesses()

        self.processObject = []

        self.processStreamsIn = []
        self.processStreamsOut = []
        for k, proc in enumerate(self.process):
            processId = proc.QProcessData_ID
            (inStreams, outStreams) = Status.prj.getStreams(processId)

            self.processStreamsIn.append(inStreams)
            self.processStreamsOut.append(outStreams)

            inStreamObjects =  [InflowingStream("", procId=processId, sqlData=stream) for stream in inStreams]
            outStreamObjects = [OutflowingStream("", procId=processId, sqlData=stream) for stream in outStreams]
            
            self.processObject.append(ProcessObject(prnt=self,
                                                    id = processId,
                                                    no = k,
                                                    inStreams = inStreamObjects,
                                                    outStreams = outStreamObjects,
                                                    process_periods = getProcessPeriods(processId),
                                                    combSchedules = getProcessCombSchedules(processId),
                                                    sqlData = proc))

        self.el = [Status.prj.getElectricity()]
        self.fuel = Status.prj.getFuels()

        self.equipment = Status.prj.getEquipment(cascade=True)
        self.eqTypes = set([getEquipmentClass(eq.EquipType) for eq in self.equipment])                
        
        self.pipe = Status.prj.getPipes()        
        self.whee = Status.prj.getWHEEs()        
        self.hx = Status.prj.getHXes()

        self.status = 0
        Status.int.pa = None

        self.elObject = [ElObject(prnt=self, id = None if self.el[0] is None else self.el[0].id, sqlData=self.el[0], no=0)]
        self.fuelObject = [FuelObject(prnt=self, id = fuel.QFuel_ID, fuelId = fuel.DBFuel_id, no=i) for i, fuel in enumerate(self.fuel)]
        self.pipeObject = [PipeObject(prnt=self, id = pipe.QDistributionHC_ID, no=m, sqlData = pipe) for m, pipe in enumerate(self.pipe)]
        self.eqObject =   [EquipmentObject(prnt=self, id = eq.QGenerationHC_ID, no=j, sqlData=eq) for j, eq in enumerate(self.equipment)]
        self.wheeObject = [WHEEObject(prnt=self, id=whee.id, no=n) for n, whee in enumerate(self.whee)]
        self.hxObject = self.hx

        self.bui = Status.prj.getBuildings()
        self.NB = len(self.bui)

        self.dataSets, self.dataFiles = loadDataSets()

        eqDataSets = [ds for ds in self.dataSets if ds.parCode in ["TDrive", "TReject"]]
        
        for ds in eqDataSets:
            ds.getData()
            j = ds.objectNo
            if j >= 0 and j < len(self.eqObject):
                setattr(self.eqObject[j], ds.parCode, ds.values)
            else:
                logError("Erroneous equipment number in %s"%ds)
                

    def __connectObjects(self):

        logTrack("SystemModel.__connectObjects")
        loadConnections(silent=True) # silent flag avoids warnings if something is not connected
                                    # TODO: assure that connection testing is done in CCheck and when adding/deleting new objects
        self.__buildStreamList()
        buildHXNet(split=True)

        self.status = 1


    def build(self):
        """
        Builds up the model. Controls that only those steps are carried out
        which are not yet up to date
        """        
        if self.status < 0: self.getComponentsFromDB()
        if self.status < 1: self.__connectObjects()

    def update(self, requireConverged=True):
        """
        Assures that all energy data in the model are up to date
        Re-calls system simulation if necessary

        requireConverged (relevant for NP-mode only: require converged results)
        """
        Status.mod.moduleEnergy.calculate(requireConverged=requireConverged)

    @property
    def upToDate(self):
        """
        Indicates that converged simulation results for the current model parameters are available
        """
        return Status.int.cascadeUpdateLevel >= self.NJ

    def changed(self, typeOfChange=None):
        """
        Function to be called from out side for indicating that the system has
        changed
        """

        #TODO: default killer from changeInCascade maybe not necessary, if certain panels / design assistants maintain up to date the object lists
        
        if typeOfChange is None:
            self.status = -1
        elif typeOfChange in [-1, 0]:
            self.status = typeOfChange
        
# Functions for setting up and working with stream lists
        
    def __buildStreamList(self, filterZeros = True):
        """ works in parallel to the stream list building function in SimTool-Plus.
            Sequence should be the same

        filterZeros: option for eliminating zero energy streams

        NOTE: for backward compatibility with the AEE-Mode the naming conventions of the stream names should not be changed !
        """

        #TODO: probably eliminate the option "filterZeros" in this basic function when building the system model, and do filtering in the post-processing where appropriate
        
        self.streams = []
        
        for k, proc in enumerate(self.process):


            heating = (proc.heating_cooling == "heating")
            procFluid = int(noneCheck(proc.ProcMedDBFluid_id, 0))
            
            if heating:
                objectType = "ProcessH"
            else:
                objectType = "ProcessC"

            if (proc.UPHm > 0 or proc.UPHm is None) or not filterZeros:
                streamName = genericStreamName(objectType, 0, heating)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName=proc.Process,
                                                                      streamName=streamName,
                                                                      objectType=objectType,
                                                                      objectNo=k,
                                                                      componentNo=0,
                                                                      heating=heating),

                                                  fluidId = procFluid,
                                                  name = streamName))

                self.processObject[k].UPHmGlb = self.streams[-1]
                
            if (proc.UPHs > 0 or proc.UPHs is None) or not filterZeros:
                streamName = genericStreamName(objectType, -1, heating)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName=proc.Process,
                                                                      streamName=streamName,
                                                                      objectType=objectType,
                                                                      objectNo=k,
                                                                      componentNo=-1,
                                                                      heating=heating),

                                                  fluidId = procFluid,
                                                  name = streamName))

                self.processObject[k].UPHsGlb = self.streams[-1]
            
            for s, stream in enumerate(self.processStreamsIn[k]):
                if (stream.UPHc > 0 or stream.UPHc is None) or not filterZeros:
                    self.streams.append(StreamGlobals(origin=HXConnection(objectName=proc.Process,
                                                                          streamName=stream.Name,
                                                                          objectType=objectType,
                                                                          objectNo=k,
                                                                          componentNo=s+1,
                                                                          heating=heating),
                                                      fluidId = int(noneCheck(stream.dbfluid_id, 0)),
                                                      name = stream.Name))

                    self.processObject[k].UPHcGlb[s] = self.streams[-1]


            self.processObject[k].MCD = False
            for s, stream in enumerate(self.processStreamsOut[k]):
                if (stream.UPHw > 0 or stream.UPHw is None) or not filterZeros:
                    if heating and stream.UPHwReq > EPSILON:
                        required = True
                    else:
                        required = False
                        
                    self.streams.append(StreamGlobals(origin=HXConnection(objectName=proc.Process,
                                                                          streamName=stream.Name,
                                                                          objectType=objectType,
                                                                          objectNo=k,
                                                                          componentNo=s+1,
                                                                          heating=not heating),
                                                      fluidId = int(noneCheck(stream.dbfluid_id, 0)),
                                                      name = stream.Name,
                                                      required=required))

                    self.processObject[k].UPHwGlb[s] = self.streams[-1]

                    if heating and stream.PTFinalRequired is not None:
                        self.processObject[k].MCD = True

                    #TODO: here the split into mandatory and optional cool-down

        for m, pipe in enumerate(self.pipe):
            heating = self.pipeObject[m].heating
            fluidId = noneCheck(pipe.HeatDistMedium, 0)
            
            if heating:
                objectType = "PipeH"
            else:
                objectType = "PipeC"

            streamName = genericStreamName(objectType, 0, heating)                                    
            self.streams.append(StreamGlobals(origin=HXConnection(objectName = pipe.Pipeduct,
                                                                  streamName = streamName,
                                                                  objectType=objectType,
                                                                  objectNo=m,
                                                                  componentNo=0,
                                                                  heating=heating),
                                              fluidId = int(noneCheck(fluidId, 0)),
                                              pinchMask=False,  # mask as False in order to avoid double counting of heat demands
                                              name=streamName))  
            self.pipeObject[m].USHGlb = self.streams[-1]

            streamName = genericStreamName(objectType, 0, not heating)                                    
            self.streams.append(StreamGlobals(origin=HXConnection(objectName = pipe.Pipeduct,
                                                                  streamName = streamName,
                                                                  objectType=objectType,
                                                                  objectNo=m,
                                                                  componentNo=0,
                                                                  heating= not heating),
                                              fluidId = int(noneCheck(fluidId, 0)),
                                              name = streamName,
                                              required=False))

            self.pipeObject[m].QWHGlb = self.streams[-1]
            self.pipeObject[m].heating = heating

        for n, whee in enumerate(self.whee):
            streamName = genericStreamName(objectType, 0, False)                                    
            self.streams.append(StreamGlobals(origin=HXConnection(objectName = whee.WHEEName,
                                                                  streamName = streamName,
                                                                  objectType="WHEE",
                                                                  objectNo=n,
                                                                  componentNo=0,
                                                                  heating=False),
                                              fluidId = int(noneCheck(whee.WHEEMedium,0)),
                                              name = streamName,
                                              required=False))

            self.wheeObject[n].QWHEEGlb = self.streams[-1]

        for j, eq in enumerate(self.equipment):

            eqClass = self.eqObject[j].eqClass
            fuelId = eq.DBFuel_id
            
            if eqClass == "BB":
                air = getDefaultAir()

                streamName = genericStreamName("Equipment", 0, True, eqClass=eqClass, fuelId=fuelId)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName = eq.Equipment,
                                                                      streamName = streamName,
                                                                      objectType="Equipment",
                                                                      objectNo=j,
                                                                      componentNo=0,
                                                                      heating=True),
                                                  fluidId = int(air.id),
                                                  pinchMask = False,        # do not include combustion air by default for pinch analysis (may disappear when optimising equipment and heat rec.)
                                                  name = streamName))

                self.eqObject[j].QHXGlb = self.streams[-1]
            elif eqClass == "HP":
                streamName = genericStreamName("Equipment", 0, True, eqClass=eqClass, fuelId=fuelId)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName = eq.Equipment,
                                                                      streamName = streamName,
                                                                      objectType="Equipment",
                                                                      objectNo=j,
                                                                      componentNo=0,
                                                                      heating=True),
                                                  fluidId = 0,
                                                  name = streamName))

                self.eqObject[j].QHXGlb = self.streams[-1]

            if (eqClass=="CH" and eq.EquipType.startswith("thermal chiller")) or \
               (eqClass=="CHP" and eq.EquipType in ["CHP ORC", "CHP TFC"]):
                cNo = 0 if eq.EquipType in ["CHP ORC", "CHP TFC"] else 1
                streamName = genericStreamName("Equipment", cNo, True, eqClass=eqClass, fuelId=fuelId)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName = eq.Equipment,
                                                                      streamName = streamName,
                                                                      objectType="Equipment",
                                                                      objectNo=j,
                                                                      componentNo=cNo,
                                                                      heating=True),
                                                  fluidId = int(0),
                                                  name = streamName))

                self.eqObject[j].FETHeatGlb = self.streams[-1]

            if eqClass not in ["ST"]:
                if eqClass == "BB" and fuelId is not None:
                    mask = False
                    fluidId = -fuelId
                else:
                    mask = True
                    fluidId = 0
                streamName = genericStreamName("Equipment", 0, False, eqClass=eqClass, fuelId=fuelId)
                self.streams.append(StreamGlobals(origin=HXConnection(objectName = eq.Equipment,
                                                                      streamName = streamName,
                                                                      objectType="Equipment",
                                                                      objectNo=j,
                                                                      componentNo=0,
                                                                      heating=False),
                                                  fluidId = int(noneCheck(fluidId, 0)),
                                                  pinchMask = mask,
                                                  name = streamName,
                                                  required=False))

                self.eqObject[j].QWHGlb = self.streams[-1]

    def getStream(self, conn):
        """ looks for the stream in the stream list specified by the connection

        """

        filtered = [stream for stream in self.streams if stream.hasSameSource(conn)]
        if len(filtered) > 0:
            return filtered[0]
        else:
            return None

    def updateEquipmentData(self, j=None, eqTable=None):
        """ used for iterative call of simulationEquipment from design assistants
        in SimTool-Plus mode

        changes the parameters of one equipment without rebuilding everything

        WARNING: this function has to be used very carefully and only for the designed
        purpose, as possible changes in the number of connections (streams) etc.
        will not be updated
        """

        self.equipment = Status.prj.getEquipment(cascade=True)

        if j is None:
            for jj, eq in enumerate(self.eqObject):
                eq.loadFromDB(self.equipment[jj])
        else:
            self.eqObject[j].loadFromDB(eqTable)
                
        #TO DO: current version works only if SQL-Table is updated !
##        self.equipment[j] = eqTable
        if Status.SimTool == "Plus" and j is not None:
            try:
                Status.mod.moduleEnergy.sim.systemModel.equipment[j].setPars(eqTable)
            except:
                print "Could not update equipment %s in simulation-plus model"%eqTable.Equipment


    def objectList(self, oType):
        if oType=="FUEL":
            return self.fuelObject
        elif oType=="EQ":
            return self.eqObject
        elif oType=="PIPE":
            return self.pipeObject
        elif oType in ["PROC"]:
            return self.processObject
        elif oType=="HX":
            return self.hxObject
        elif oType=="WHEE":
            return self.wheeObject
        else:
            return []

    def objectId(self, oType, oNo):
        """ returns the id of an object as a function of objectType and no. """
        oList = self.objectList(oType)
        nmax = len(oList)
        try:
            return oList[oNo].id
        except:
            return None

    def getObject(self, oType, id):
        oList = self.objectList(oType)
        oIds = [o.id for o in oList]
        return oList[oIds.index(id)] if id in oIds else None            


## TODO: in older versions there was the possibility to re-order objects
## on the SQL level. Should be possible here on systemModel level, too
## -> check this old function
##    def shiftFromTo(self, k0, k1):
##        Status.prj.reorderProcesses(k0,k1)
##        self.display()

    def addObject(self, objectType):
        if objectType=="EQ":
            objectClass = EquipmentObject
            objectList = self.eqObject
        elif objectType=="PIPE":
            objectClass = PipeObject
            objectList = self.pipeObject
        elif objectType=="PROC":
            objectClass = ProcessObject
            objectList = self.processObject

        obj = objectClass(prnt=self)
        obj.no = len(objectList)
        if objectType=="EQ": obj.EqNo = obj.no
        objectList.append(obj)
        return obj
            
        
    def addEquipment(self):
        return self.addObject("EQ")
##        eq = EquipmentObject(prnt=self)
##        self.eqObject.append(eq)
##        eq.no = len(self.eqObject) - 1
##        return eq
    
#Embryos for object classes from which the system model is going to be built of
#Currently contain only a pointer to the related StreamGlbs in the general stream list
# already implemented more extensively:
##class ProcessObject():

##class PipeObject():
##class EquipmentObject():

        self.eqClass = None
        self.eqType = None

        # data sets
        self.TDrive = None
        self.TReject = None

class FuelObject():
    def __init__(self, prnt, id=None, no=None, fuelId=None):
        self.prnt = prnt
        self.id = id
        self.fuelId = fuelId
        self.no = no
        self.name = "Fuel(%s)"%no

class HXObject():
    def __init__(self, prnt, id=None, no=None):
        self.prnt = prnt
        self.id = id
        self.no = no

class WHEEObject():
    def __init__(self, prnt, id=None, no=None):
        self.prnt = prnt
        self.id = id
        self.no = no
        self.QWHEEGlb = None
        self.name = None if no is None else prnt.whee[no].WHEEName

# Helper functions for mapping and identifying objects


class ModelComponent():
    """ represents and object (component) within the system:
    fuels, equipment, pipes, processes, hx, whee, nodes, ...
    """
    
    def __init__(self, model, oType, oId=None, oNo=None):
        self.model = model
        self.oType = objectType
        if oId is None:
            self.oId = self.getObjectId(oNo)
            self.oNo = oNo
        else:
            self.oId = oId
            self.oNo = self.getObjectNo(oId)

    def getOId(self, oNo):
        oList = self.model.objectList(self.oType)
        try:
            self.oId = self.model.objectList[oNo].id
        except:
            pass
        return self.oId
            
    def getONo(self, oId):
        oList = self.model.objectList(self.oType)
        try:
            self.oNo = [o.id for o in oList].index(oId)
        except:
            pass            

class ModelComponentPin():
    """
    Represents a specific connection pin (input, output, state) within a component
    
    connection numbering for objects
    General in/out: 0 .. n-1

    processes: maintenance 0, start-up 1

    cdir: 1: inflow, 0: internal, -1: outflow
    """

    def __init__(self, obj, cId=None, cNo=None, cDir=None):
        self.cDir = cDir
        self.obj = obj
        if cId is None:
            self.cNo = cNo
            self.cId = self.getCId(cNo)
        else:
            self.cId = cId
            self.cNo = self.getCNo(cId)

    def getCId(self, cNo):
        if self.obj.oType=="PROC":
            o = self.obj.model.processObject[self.oNo]
            if self.cDir==1:
                oList=self.obj.model.processStreamsIn
            elif self.cDir==-1:
                oList=self.obj.model.processStreamsOut

            try:
                self.cId = oList[cNo].id
            except:
                self.cId = None

    def getCNo(self, cId):
        if self.oType=="PROC":
            if self.cDir==1:
                oList=self.obj.model.processStreamsIn
            elif self.cDir==-1:
                oList=self.obj.model.processStreamsOut

            try:
                self.cNo = oList.index(cId)
            except:
                self.cNo = None
    


    
