# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	Combustion 
#			
#------------------------------------------------------------------------------
#			
#	Module for exhaust gas calculations in NP mode
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import numpy as np
from GUI.status import *
from modules.interfaces import createQ_Tt, createQ_t, calcQ_T
from constants import *
from modules.fluids import Fuel
from modules.messageLogger import *
from auxiliary.auxiliary import noneCheck

TMINCOMBAIR = 18.0 # default inlet temperature for combustion air
TMAXCOMBAIR = 80.0 # default maximum external preheating temperature for combustion air

def getCombustionParameters(equipe, PNom, PEl, COPh_nom, air):
    """
    """
    # f_QWH: proportion waste heat vs. USH, f_QHXcombAir
    if equipe.DBFuel_id > 0:
        fuel = Fuel(equipe.DBFuel_id)
        if fuel.id is None:
            fuel = None
            f_QWHu_T = Status.int.createQ_T()
            f_QWH_HX_u_T = Status.int.createQ_T()
            f_combAir = 0
            LossFactEq = 0
            TExhaustGas = None
            TMinOffGas = None
        else:
            TEnvEq = 18.0
            LossFactEq = 0.01 # includes the losses of electricity consumption not transferred to the working fluid or exhaust gas

            FETRatio = 1.0 if PNom <= 0 else 1. + PEl*COPh_nom/PNom # ratio of total FET to FETFuel
            
            if COPh_nom > 0:
                f_QWH = max (0, FETRatio - COPh_nom - LossFactEq*COPh_nom) / COPh_nom
            else:
                f_QWH = 0

            xAir = noneCheck(equipe.ExcessAirRatio, 1.0)    # TODO. this is a hack in order to avoid problems that arose with CHP equipment w/o ExcessAirRatio
            f_combAirFET = (fuel.CombAir * xAir)*air.cp*(TMAXCOMBAIR - TMINCOMBAIR) / fuel.LCV # ratio QHX_combAir / FET
            if f_combAirFET > 0:
                f_combAir = 1./(1 + COPh_nom/f_combAirFET)  # ratio QHX_combAir / USH
            else:
                f_combAir = 0

            # temperature distribution of waste heat. assumed as fix
            TExhaustGas = equipe.TExhaustGas
            
            if TExhaustGas == None:
                logDebug("Boiler exhaust gas temperature not specified. 200 °C assumed")
                TExhaustGas = 200

            fuel = Fuel(equipe.DBFuel_id)
            TMinOffGas = max(fuel.TCondOffGas()+0.001, TEnvEq)
            dTtot = max(TExhaustGas - TEnvEq, 1.e-10)
            dTmin = max(TExhaustGas - TMinOffGas, 1.e-10)

            dT_T = np.linspace(0, Status.NT + 1, Status.NT + 2)
            dT_T = dT_T * Status.TemperatureInterval - TMinOffGas
            dT_T = dT_T.clip(min=0)
            TDist_QWH = (1.0 - dT_T / dTmin).clip(min=0)
            f_QWHu_T = f_QWH * (dTmin/dTtot) * TDist_QWH # useful waste heat above TMinOffGas
            f_QWH_HX_u_T = -LossFactEq * (dTmin/dTtot) * TDist_QWH

    else:
        fuel = None
        f_QWHu_T = Status.int.createQ_T()
        f_QWH_HX_u_T = Status.int.createQ_T()
        f_combAir = 0
        TExhaustGas = None
        TMinOffGas = None

    return fuel, f_QWHu_T, f_QWH_HX_u_T, f_combAir, TExhaustGas, TMinOffGas
            
        
if __name__ == "__main__":
    pass
