# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleBB (Boilers and Burners)
#			
#------------------------------------------------------------------------------
#			
#	Module for calculation of boilers
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
from GUI.status import *
from modules.boiler.combustion import getCombustionParameters, TMINCOMBAIR, TMAXCOMBAIR
from modules.interfaces import createQ_Tt, createQ_t, calcQ_T
from constants import *
from modules.modulePipes import getAllHeatingPipes
from modules.heatFlow import HeatFlow_t
from modules.fluids import Fuel, getDefaultAir
from modules.messageLogger import *
from modules import equipments
from lib.auxiliary.auxiliary import noneCheck
import database.dbio as dbio

DEFAULTSAUTODESIGNBB = { "BBMaintain":0,
                         "BBSafety":10.0,
                         "BBRedundancy":1,
                         "BBFuelType":"Natural Gas",
                         "BBHOp":1000.0,
                         "BBPmin":500,
                         "BBEff":0.85}

class ModuleBB(object):
    BBList = []

    def __init__(self, keys):
        self.keys = keys # the key to the data is sent by the panel

        self.neweqs = 0 #new equips added
        self.maxTemp = 0 # maximum temperature level of the heat demand

        self.air = getDefaultAir()

    def initPanel(self):
        if Status.int.cascadeUpdateLevel < 0:
            Status.int.initCascadeArrays(0)

    def updatePanel(self):
        """Prepare all the needed information so that it can be 
        plotted on the panel.
        
        """
        logTrack("ModuleBB: updatePanel")

        # get list of equipments and boilers in the system
        # and update the energetic calculation up to the level needed for representation
        (BBList, BBTableDataList) = self.screenEquipments()

        if Status.int.cascadeUpdateLevel < self.cascadeIndex:
            Status.mod.moduleEnergy.runSimulation(last=self.cascadeIndex)

        # 1. List of equipments
        matrix = []
        for row in BBTableDataList:
            matrix.append(row)

        data = np.array(matrix)

        Status.int.setGraphicsData('BB Table', data)

        # 2. Preparing data
        self.findmaxTemp(Status.int.USHTotal_T)
        PowerSum80 = 0
        PowerSum140 = 0
        PowerSumTmax = 0

        if self.maxTemp > 80:
            if self.maxTemp > 160: # the minimum temperature difference is now setted at 20°C but could even be a parameter
                for i in BBList:
                    bb = dbio.get("qgenerationhc", id=i["equipeID"])
                    if bb is not None:
                        if bb.TMaxSupply > 80 and bb.TMaxSupply <= 140:
                            PowerSum140 += i["equipePnom"]
                        if bb.TMaxSupply > 140:
                            PowerSumTmax += i["equipePnom"]
            else :
                for i in BBList:
                    bb = dbio.get("qgenerationhc", id=i["equipeID"])
                    if bb is not None:
                        if bb.TMaxSupply > 80:
                            PowerSumTmax += i["equipePnom"]

            for i in BBList:
                bb = dbio.get("qgenerationhc", id=i["equipeID"])
                if bb is not None:
                    if bb.TMaxSupply <= 80:
                        PowerSum80 += i["equipePnom"]
        else:
            for i in BBList:
                PowerSumTmax += i["equipePnom"]

        if len(BBList) == 0:
            index = max(len(self.equipment) + 1, 1)     # assure that index >= 0
                                                    # even if there's NO equipment at all 
        else:
            bb = dbio.get("qgenerationhc", id=i["equipeID"])
            if bb is not None:
                index = bb.CascadeIndex

        #TODO: this does not work well if no boilers are in the equipment or chillers are within the cascade
        # apply same procedure as in design assistant
        
        QD80C = Status.int.QD_Tt_mod[index - 1][int(80 / Status.TemperatureInterval + 0.5)].tolist()
        QD80C.sort(reverse=True)
        QD140C = Status.int.QD_Tt_mod[index - 1][int(140 / Status.TemperatureInterval)].tolist()
        QD140C.sort(reverse=True)

        iT_maxTemp = int(self.maxTemp / Status.TemperatureInterval)
        QDmaxTemp = Status.int.QD_Tt_mod[index - 1][iT_maxTemp].tolist()
        QDmaxTemp.sort(reverse=True)
        QResidualMaxTemp = Status.int.QD_Tt_mod[self.cascadeIndex][iT_maxTemp].tolist()
        QResidualMaxTemp.sort(reverse=True)

        # 2. XY Plot
        TimeIntervals = []
        for it in range(Status.Nt):
            TimeIntervals.append(Status.TimeStep * (it + 1) * Status.EXTRAPOLATE_TO_YEAR)

        Status.int.setGraphicsData('BB Plot', [TimeIntervals,
                                               QD80C,
                                               QD140C,
                                               QDmaxTemp])

        # 3. Configuration design assistant
        config = self.getUserDefinedPars()
        Status.int.setGraphicsData('BB Config', config)

        # 4. additional information (Info field right side of panel)
        info = []
        info.append(self.maxTemp)  #first value to be displayed
        info.append(QResidualMaxTemp[0])
        info.append(max(0, QD80C[0]))
        info.append(max(0, (QD140C[0] - QD80C[0])))
        info.append(max(0, QDmaxTemp[0]))
        Status.int.setGraphicsData('BB Info', info)

    def getUserDefinedPars(self):
        """Get the user defined data from UHeatPump and store 
        it to interfaces to be shown in HP panel.
        
        """
        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)

            config = [1, 10, 0, "Natural Gas", 100, 500, 80]
            Status.int.setGraphicsData('BB Config', config)

            self.setUserDefinedPars()
        else:
            u = urow
            config = [u.BBMaintain,
                      u.BBSafety,
                      u.BBRedundancy,
                      u.BBFuelType,
                      u.BBHOp,
                      u.BBPmin,
                      u.BBEff]
        return config

    def setUserDefinedPars(self):
        config = Status.int.GData['BB Config']
        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        if urow is None:
            dummy = {"Questionnaire_id":Status.PId, "AlternativeProposalNo":Status.ANo}
            dbio.insertRow("uheatpump", **dummy)
            urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        u = urows[0]
        u.BBMaintain = dbio.check(config[0])
        u.BBSafety = dbio.check(config[1])
        u.BBRedundancy = dbio.check(config[2])
        u.BBFuelType = dbio.check(config[3])
        u.BBHOp = dbio.check(config[4], string=False)
        u.BBPmin = dbio.check(config[5], string=False)
        u.BBEff = dbio.check(config[6], string=False)

        dbio.commit()

    def screenEquipments(self, setIndex=True):
        """Screen existing equipment, whether there are already boilers"""
        Status.int.systemModel.build()
        self.equipment = Status.int.systemModel.equipment
        Status.int.getEquipmentCascade()
        self.NEquipe = len(self.equipment)

        self.BBList = []
        maxIndex = self.NEquipe
        index = 0
        for row in Status.int.cascade:
            index += 1
            if getEquipmentClass(row["equipeType"]) == "BB":
                self.BBList.append(row)
                maxIndex = index

        if setIndex == True:
            self.cascadeIndex = maxIndex

        BBTableDataList = []
        for row in Status.int.EquipTableDataList:
            if getEquipmentClass(row[3]) == "BB":
                BBTableDataList.append(row)

        #screen list and substitute None with "not available"
        for i in range(len(BBTableDataList)):
            for j in range(len(BBTableDataList[i])):
                if BBTableDataList[i][j] == None:
                    BBTableDataList[i][j] = 'not available'

                if j == 1: #Pnom entry
                    try:
                        BBTableDataList[i][j] = "%.2f"%BBTableDataList[i][j]
                    except:
                        pass
                elif j == 2: #efficiency entry
                    try:
                        BBTableDataList[i][j] = "%.3f"%BBTableDataList[i][j]
                    except:
                        pass
                    
                    
        return (self.BBList, BBTableDataList)

    def getEqId(self, rowNo):
        """Get the EqId from the rowNo in the BBList."""
        BBId = self.BBList[rowNo]["equipeID"]
        return BBId

    def deleteEquipment(self, rowNo, automatic=False):
        if automatic == False:
            if rowNo == None:   #indicates to delete last added equipment dummy
                BBid = self.dummyEqId
            else:
                # delete BB from the equipment list under current alternative #from C&QGenerationHC under ANo
                BBid = self.getEqId(rowNo)
        else:
            BBid = rowNo
        Status.prj.deleteEquipment(BBid)

        Status.int.systemModel.changed()
        Status.int.systemModel.build()
        self.screenEquipments()

    def addEquipmentDummy(self):
        """Add a new dummy equipment to the equipment list and return 
        its position in the cascade.
        
        """
        self.equipe = Status.prj.addEquipmentDummy()
        self.dummyEqId = self.equipe.QGenerationHC_ID

        self.neweqs += 1 #No of last equip added
        NewEquipmentName = "New boiler %s" % (self.neweqs)

        # connect the dummy to all heating pipes
        dummy, pipeIDs = getAllHeatingPipes()
        pipeIDs = [str(id) for id in pipeIDs ]
        IDString = ";".join(pipeIDs)
        
        equipeData = {"Equipment":NewEquipmentName,
                      "EquipType":"Boiler (specify subtype)",
                      "PipeDuctEquip":IDString}
        self.equipe.set(**equipeData)
        dbio.commit()
        # update the interconnection tables
        Status.int.systemModel.changed()
        Status.int.systemModel.build()
        
        self.screenEquipments()
        self.cascadeIndex = self.NEquipe
        return(self.equipe)

    def setEquipmentFromDB(self, equipe, modelID):
        """Take an equipment from the data base and store it under the
        given Id in the equipment data base.
        
        """
        model = dbio.get("dbboiler", id=modelID)

        if model.BBPnom != None:
            equipe.HCGPnom=model.BBPnom
            Pnom = model.BBPnom
        else:
            equipe.HCGPnom=0.0
            logWarning("ModuleBB: boiler with 0 kW nominal power added from database. Check your boiler database")
            Pnom = 0

        if model.BBEfficiency != None:
            if model.BBEfficiency > 1.3 and model.BBEfficiency < 130.0:
                logTrack("ModuleBB: Efficiency data should be stored internally as fractions of 1")
                eff = model.BBEfficiency / 100.0
                model.BBEfficiency=eff
            else:
                eff = model.BBEfficiency
            equipe.HCGTEfficiency=eff

        if model.BoilerTemp is not None: equipe.TMaxSupply=model.BoilerTemp
        if model.BoilerManufacturer != None: equipe.Manufact=model.BoilerManufacturer
        if model.BoilerModel is not None: equipe.Model=model.BoilerModel
        equipe.EquipType=getEquipmentType("BB", model.BoilerType)
        equipe.NumEquipUnits=1
        if model.BoilerType is not None: equipe.EquipTypeFromDB=model.BoilerType
        if model.id is not None: equipe.EquipIDFromDB=model.id
        equipe.DBFuel_id=1  #use Natural Gas as default -> should later on be adjusted to type of equipment
        equipe.ExcessAirRatio=1.1

        # equipment parameters that are set defined by default if not specified
        fuel_number = equipe.DBFuel_id   # import from the fuelDB
        if fuel_number is not None and fuel_number >= 0:
            eq_fuel = Fuel(fuel_number)

            try:
                equipe.FuelConsum = equipe.HCGPnom / (eq_fuel.LCV * equipe.HCGTEfficiency)
            except:
                pass

            try:
                equipe.FlowExhaustGas = equipe.FuelConsum * (1.0 + eq_fuel.CombAir * equipe.ExcessAirRatio)
            except:
                pass

            try:
                equipe.ElectriConsum = 0.01 * equipe.HCGPnom
            except:
                pass

            try:
                LossFactEq = 0.01
                TExhaustGas = eq_fuel.LCV * (1.0 - equipe.HCGTEfficiency - LossFactEq) \
                              / ((1.0 + eq_fuel.CombAir) * equipe.ExcessAirRatio * eq_fuel.OffgasHeatCapacity)
                equipe.TExhaustGas = TExhaustGas
            except:
                pass
        else:
            equipe.FuelConsum = 0
            equipe.FlowExhaustGas = 0
            try:
                equipe.ElectriConsum = equipe.HCGPnom / equipe.HCGTEfficiency
            except:
                pass
            
            equipe.TExhaustGas = 0

        equipe.set(**{"TurnKeyPrice": model.BoilerTurnKeyPrice,
                       "OandMfix": model.BoilerOandMfix,
                       "OandMvar": model.BoilerOandMvar})
        (turnKey, OMFix, OMVar) = self.estimateEconomicParameters(equipe)
        equipe.set(**{"TurnKeyPrice": turnKey,
                       "OandMfix":OMFix,
                       "OandMvar":OMVar})

        dbio.commit()
        logTrack("moduleBB (setEquipmentFromDB): boiler added:'%s',type:'%s',Pow:'%s',T'%s'" % \
                 (model.BoilerManufacturer, model.BoilerType, model.BBPnom, model.BoilerTemp))
        Status.int.systemModel.updateEquipmentData()

    def calculateEnergyFlows(self, equipe, demand_Tt, cascadeIndex):
        """Calculate the energy flows in the equipment identified by 
        "cascadeIndex" 
        
        :returns: a pair consisting of the time and temperature 
        dependent numpy array containing the covered demand. Also known
        as USHj_Tt of the given equipment and the same data aggregated 
        over time and temperature, USHj.
        """
        j = cascadeIndex - 1
        
        if Status.SimTool == "Plus":
            logError("ModuleBB.calculateEnergyFlows: call to this function is not allowed in SimTool=Plus mode")
            return
        if Status.int.cascadeUpdateLevel < (cascadeIndex - 1):
            logTrack("ModuleBB (calculateEnergyFlows): updating previous levels in cascade until %s" % (cascadeIndex - 1))
            Status.mod.moduleEnergy.runSimulation(last=(cascadeIndex - 1))
            logTrack("ModuleBB (calculateEnergyFlows): now starting calculations")
        Status.int.extendCascadeArrays(cascadeIndex)

        if cascadeIndex > 0 and cascadeIndex <= Status.NEquipe:
            logTrack("ModuleBB (calculateEnergyFlows): starting (cascade no: %s)" % cascadeIndex)
        else:
            logError("ModuleBB (calculateEnergyFlows): cannot simulate index %s: out of cascade [%s]" % \
                     (cascadeIndex, Status.NEquipe))
            return (createQ_Tt(), 0)

        # get equipment data from equipment list in SQL
        BBModel = equipe.Model
        BBType = equipe.EquipType
        PNom = equipe.HCGPnom
        PEl = equipe.ElectriConsum
        COPh_nom = equipe.HCGTEfficiency
        TMax = equipe.TMaxSupply
        EquipmentNo = equipe.EqNo
        controlSchedule = Status.int.systemModel.eqObject[cascadeIndex-1].controlSchedule
        control = controlSchedule.schedule if controlSchedule else 1.0
        LossFactEq = 0.01

        if PNom is None:
            PNom = 0.0
            logWarning("ModuleBB (calculateEnergyFlows): No nominal power specified for equipe no. %s" % \
                 (EquipmentNo))

        if PEl is None:
            PEl = 0.01 * PNom
            logWarning("ModuleBB (calculateEnergyFlows): No electrical consumption specified for equipe no. %s" % \
                 (EquipmentNo))

        if TMax is None:
            TMax = INFINITE
            logDebug("ModuleBB (calculateEnergyFlows): no Tmax specified for equipe no. %s" % EquipmentNo)

        logTrack("ModuleBB (calculateEnergyFlows): Model = %s Type = %s PNom = %s" % \
                 (BBModel, BBType, PNom))

        fuel, f_QWHu_T, fQWH_HX_u_T, f_combAir, TExhaustGas, TMinOffGas = getCombustionParameters(equipe, PNom, PEl, COPh_nom, self.air)
        
        # get demand data for CascadeIndex/EquipmentNo from Interfaces
        # and create arrays for storing heat flow in equipment
        QD_Tt = demand_Tt.copy() # copy the given demand array for internal use
        QA_Tt = Status.int.QA_Tt_mod[cascadeIndex - 1].copy()

        USHj_Tt = createQ_Tt()
##        QHXj_Tt = Status.int.QHXj_Tt[cascadeIndex - 1] # is an input from HR calculatinos
        QHXj_t = Status.int.QHXEq_Tt[cascadeIndex-1].getQ()
        QWHj_Tt = createQ_Tt()
        QWHj_t = createQ_t()

        USHj = 0
        QHXj = Status.int.QHXEq_Tt[cascadeIndex-1].Q # is an input from HR calculatinos
        QWHj = 0
        # Calculate heat delivered by the given equipment for each temperature step
        for iT in xrange(Status.NT + 2):
            if TMax >= Status.int.T[iT] : # TMax is the max operating temperature of the boiler 
                USHj_Tt[iT] = QD_Tt[iT].clip(max=PNom * Status.TimeStep * control) # from low to high T
            else:
                if (iT > 0):
                    USHj_Tt[iT] = USHj_Tt[iT - 1] # no additional heat supply at high temp.
                else:
                    USHj_Tt[iT] = 0
           
        # count the working time intervals
        HPerYear = np.ma.array(USHj_Tt[Status.NT + 1])
        HPerYear[HPerYear > 0] = np.ma.masked # mask all working intervals
        # count the unmasked(=idle) intervals
        HPerYear = (Status.Nt - HPerYear.count()) * Status.TimeStep 
        
        QWHj_Tt += (USHj_Tt[Status.NT + 1] - QHXj_t)# broadcast to all temperature bins
        QWHj_Tt *= f_QWHu_T.reshape((Status.NT + 2, 1)) # reshape to multiply column wise

        if QHXj > 0 and LossFactEq > 0:
            QWHj_Tt += QHXj_t*fQWH_HX_u_T.reshape((Status.NT + 2, 1))

        QWHj_t += QWHj_Tt[0]
        
        USHj += np.sum(USHj_Tt[Status.NT + 1])
        QWHj += np.sum(QWHj_Tt[0])
        print "TEST TEST Module BB: USHj %s QHXj %s QWHj %s"%(USHj, QHXj, QWHj)
        
        # End of year reached. Store results in interfaces
        # remaining heat demand and availability for next equipment in cascade
        # TODO: QD_T and Tt shouldn't be needed for other calculations.
        Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
        Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)

        # heat delivered by present equipment
        Status.int.USHj_Tt[cascadeIndex - 1] = USHj_Tt.copy()
        Status.int.USHj_T[cascadeIndex - 1] = Status.int.calcQ_T(USHj_Tt)
        Status.int.USHj_t[cascadeIndex - 1] = USHj_Tt[Status.NT + 1].copy()

        Status.int.USHjChanged[cascadeIndex - 1] = True

##        Status.int.QHXj_Tt[cascadeIndex - 1] = QHXj_Tt
##        Status.int.QHXj_T[cascadeIndex - 1] = Status.int.calcQ_T(QHXj_Tt)
        Status.int.QHXj_t[cascadeIndex - 1] = QHXj_t.copy()

        # maximum waste heat that can be absorbed by present equipment for combustion air preheating
        if f_combAir > 0:
            QHXEqReq_t = (USHj_Tt[Status.NT+1] - QHXj_t) * f_combAir
            QHXEqReq = f_combAir * (USHj - QHXj)

        if f_combAir == 0:
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = HeatFlow_t()
            Status.int.QHXEqReqChanged[cascadeIndex-1] = False
        else:
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = HeatFlow_t(fluid_id=self.air.id,T0=TMINCOMBAIR,T1=TMAXCOMBAIR,Q_t = QHXEqReq_t)
            Status.int.QHXEqReqChanged[cascadeIndex-1] = True

        # waste heat produced by present equipment
        Status.int.QWHj_Tt[cascadeIndex - 1] = QWHj_Tt
        Status.int.QWHj_T[cascadeIndex - 1] = Status.int.calcQ_T(QWHj_Tt)
        Status.int.QWHj_t[cascadeIndex - 1] = QWHj_Tt[Status.NT + 1].copy()
        Status.int.QWHjChanged[cascadeIndex - 1] = True
        Status.int.QWHEqChanged[cascadeIndex - 1] = True

        if fuel is None or QWHj == 0:
            Status.int.QWHEq_Tt[cascadeIndex - 1] = HeatFlow_t()
        else:
            Status.int.QWHEq_Tt[cascadeIndex - 1] = HeatFlow_t(fluid_id= -fuel.id, T0=TMinOffGas, T1=TExhaustGas, Q_t=QWHj_t,name="ExhaustGas")
        
        logTrack("ModuleBB (calculateEnergyFlows): Total energy supplied by equipment %s MWh" % (USHj * Status.EXTRAPOLATE_TO_YEAR))
        logTrack("ModuleBB (calculateEnergyFlows): QHXj %s QWHj %s [MWh]" % (QHXj * Status.EXTRAPOLATE_TO_YEAR,QWHj))

        Status.int.cascadeUpdateLevel = cascadeIndex

        # Global results (annual energy flows)
        USHj *= Status.EXTRAPOLATE_TO_YEAR
        QWHj *= Status.EXTRAPOLATE_TO_YEAR
        HPerYear *= Status.EXTRAPOLATE_TO_YEAR
        Status.int.USHj[cascadeIndex - 1] = USHj

        if fuel is None: # electrical heaters 
            if COPh_nom > 0:
                FETel_j = USHj / COPh_nom
                Status.int.FETel_j_t[j] = np.array(USHj_Tt[Status.NT+1])/COPh_nom
            else:
                FETel_j = 0.0
                showWarning("Strange boiler with COP = 0.0")

            FETFuel_j = 0.0
            
        else:
            if COPh_nom > 0:
                FETFuel_j = (USHj - QHXj)/ COPh_nom
            else:
                FETFuel_j = 0.0
                showWarning("Strange boiler with COP = 0.0")

            if PNom > 0:
                FETel_j = FETFuel_j * PEl / PNom
                if COPh_nom > 0:
                    Status.int.FETel_j_t[j] = (np.array((USHj_Tt[Status.NT+1] - QHXj_t)) / COPh_nom) * (PEl / PNom)
            else:
                FETel_j = 0

        Status.int.FETFuel_j[cascadeIndex - 1] = FETFuel_j
        Status.int.FETel_j[cascadeIndex - 1] = FETel_j
        Status.int.ElGen_j[cascadeIndex - 1] = 0.0
        Status.int.ElGen_t[cascadeIndex - 1] = Status.int.createQ_t()
        Status.int.HPerYearEq[cascadeIndex - 1] = HPerYear

        Status.int.QWHj[cascadeIndex - 1] = QWHj   # not considering the latent heat(condensing water)
        Status.int.QHXj[cascadeIndex - 1] = QHXj

        return (USHj_Tt, USHj)

    def sortBoiler (self):
        """Move all the boilers to the end of the cascade. 
        Sort boilers by temperature and by efficiency.
        
        """
        #first bring all your boilers in a listing for easier access
        boilerList = []
        for i in range(len(Status.int.cascade)):
            entry = Status.int.cascade[i]
            a = getEquipmentClass(entry["equipeType"])
            if a == "BB":
                eqID = entry["equipeID"]
                equipe = dbio.get("qgenerationhc", id=eqID)
                efficiency = equipe.HCGTEfficiency
                temperature = equipe.TMaxSupply
                boilerList.append({"equipeID":eqID, "efficiency":efficiency, "temperature":temperature})

        # then reorganize boilerList by efficiencies (maybe there are more intelligent ways to do this, but here's one ...:
        # if you want the most efficient on top [n-1] instead on bottom [0], just change the sign of the comparison from > to <
        for i in range(len(boilerList)):
            for j in range(i, len(boilerList)):
                if boilerList[j]["efficiency"] > boilerList[i]["efficiency"]:
                    bi = boilerList[i]
                    bj = boilerList[j]
                    boilerList[i] = bj
                    boilerList[j] = bi

        # then reorganize boilerList by temperature levels:
        # first level:
        for i in range(len(boilerList)):
            for j in range(i, len(boilerList)):
                if boilerList[i]["temperature"] > 80 and boilerList[j]["temperature"] <= 80:
                    bi = boilerList[i]
                    bj = boilerList[j]
                    boilerList[i] = bj
                    boilerList[j] = bi

        # second level:
        for i in range(len(boilerList)):
            for j in range(i, len(boilerList)):
                if  boilerList[i]["temperature"] > 140 and  boilerList[j]["temperature"] <= 140:
                    bi = boilerList[i]
                    bj = boilerList[j]
                    boilerList[i] = bj
                    boilerList[j] = bi

        # then move consecutively to the bottom of the cascade (move the one last, which you finally want to have at the bottom
        for i in range(len(boilerList)):
            eqID = boilerList[i]["equipeID"]
            equipe = dbio.get("qgenerationhc", id=eqID)
            cascadeIndex = equipe.CascadeIndex
            if cascadeIndex < Status.int.NEquipe:
                Status.mod.moduleHC.cascadeMoveToBottom(cascadeIndex)

        #identify last heating equipment in cascade
##        allEquipments = Status.prj.getEquipment(cascade=True)
        self.lastHeating = -1
        for j, eqObject in enumerate(Status.int.systemModel.eqObject):
            if eqObject.heating:
                self.lastHeating = j

        # find the position in the cascade of the first and last boiler of each T level
        n80 = 0
        n140 = 0
        
        for i in range(len(boilerList)):
            if  boilerList[i]["temperature"] > 80 :
                n80 += 1
            if  boilerList[i]["temperature"] > 140 :
                n140 += 1

        self.firstBB = len(Status.int.cascade) - len(boilerList)
        self.firstBB140 = len(Status.int.cascade) - n80
        
        if self.maxTemp > 140:
            self.firstBBmaxTemp = len(Status.int.cascade) - n140
        else:
            self.firstBBmaxTemp = len(Status.int.cascade) - n80
        logTrack("ModuleBB.sortBoiler: indexes BB %s BB140 %s BBmaxTemp %s" % (self.firstBB, self.firstBB140, self.firstBBmaxTemp))

    def automDeleteBoiler (self, minEfficencyAccepted=0.80):  #0.80 is a default value for minimum of efficiency
        """Delete unefficient boiler"""
        self.screenEquipments()
        for i in range (len (self.BBList)):
            eq = dbio.get("qgenerationhc", id=self.BBList[i]['equipeID'])
            eff = eq.HCGTEfficiency
            if eff < minEfficencyAccepted:
                # add the fuel criterion: if not biomass, biofuels?, 
                # gas methane ->delete ???
                self.dummyEqId = self.BBList[i]['equipeID']
                self.deleteEquipment(None)

    def findmaxTemp(self, QDa):
        """Find the maximum temperature level of the heat demand
        QDa is the annual heat demand by temperature.
        
        """
        QDcrit = 0.99 * QDa[Status.NT + 1]
        for i in range(1, Status.NT + 1):
            if QDa[i] > QDa[i - 1] and QDa[i - 1] < QDcrit:
                self.maxTemp = Status.TemperatureInterval * i #temperatureInterval is defined in status.py

    def redundancy(self):
        """When redundancy is required provides suitable boilers.
        N.B. the possibility of retriveing deleted boilers is not 
        implemented yet.
        N.B. The possibility that a boiler with nominal power > than biggerBB 
        is in the list has to be considered.
        
        """
        self.maxPow80 = 0
        self.maxPow140 = 0
        self.maxPowTmax = 0

        # looking for equipment with maximum power in each of the three temperature levels
        equipment = Status.prj.getEquipment()
        for eq in equipment:
            equipeType = eq.EquipType
            if getEquipmentClass(equipeType) == "BB":
                Pnom = eq.HCGPnom
                TMaxSupply = eq.TMaxSupply
                if TMaxSupply <= 90 and Pnom > self.maxPow80:     # TMaxSupply to be substituted with the operating temperature
                    self.maxPow80 = Pnom
                elif (90 < TMaxSupply and TMaxSupply <= 150) and Pnom > self.maxPow140:
                    self.maxPow140 = Pnom
                elif TMaxSupply > 150 and Pnom > self.maxPowTmax:
                    self.maxPowTmax = Pnom

        if self.maxPow80 > 0:
            self.selectAndAddBB(max(self.minPow, self.maxPow80), 80)
            
        if self.maxPow140 > 0:  #and self.maxTemp>140:
            self.selectAndAddBB(max(self.minPow, self.maxPow140), 140)
            
        if self.maxPowTmax > 0:
            self.selectAndAddBB(max(self.minPow, self.maxPowTmax), self.maxTemp)

    def findBiggerBB(self):
        """Find the maximum power of boilers available in the DB 
        (for each temperature level) and write them to self 
        properties.
        
        Each property is set to INFINITE if no fitting boiler is found.
        That should make the "found boiler" too large for any use. 
        
        """
        logTrack("moduleBB: Looking for a Boiler with high Temperature of %s" % 
                 (self.maxTemp)) 
        sqlQuery = "BoilerTemp >='%s' ORDER BY BBPnom DESC" % (self.maxTemp)
        search = dbio.selectByQuery("dbboiler", sqlQuery)
        self.biggermaxTemp = search[0].BBPnom
        sqlQuery = "140 <= BoilerTemp AND BoilerTemp < '%s' ORDER BY BBPnom DESC" % (self.maxTemp)
        search = dbio.selectByQuery("dbboiler", sqlQuery)
        try:
            self.bigger140 = search[0].BBPnom
        except IndexError:
            self.bigger140 = INFINITE # no boilers found, create a value to big to be ever chosen
        sqlQuery = "80 <= BoilerTemp AND BoilerTemp < 140 ORDER BY BBPnom DESC"
        search = dbio.selectByQuery("dbboiler", sqlQuery)
        try:
            self.bigger80 = search[0].BBPnom
        except IndexError:
            self.bigger80 = INFINITE # no boilers found, create a value to big to be ever chosen
        
    def selectBB(self, Pow, Top):
        """The query should consider the fuel too but at the moment the 
        fuel column doesn't exist in mySQL. We could give the 
        possibility to choose the boiler to the user (in the 
        interactive mode) in a 'selected' list. In this version we take
        the first element of the list.
        
        """
        Power = max(self.minPow, Pow)
        sqlQuery = "BoilerTemp >= '%s'AND BBPnom >= '%s' ORDER BY BBPnom ASC" % (Top, Power)
        selected = dbio.selectByQuery("dbboiler", sqlQuery)
##        for i in range(len(selected)):
##            for j in range(i, len(selected)):
##                if selected[j].BoilerTemp < selected[i].BoilerTemp:
##                    bi = selected[i]
##                    bj = selected[j]
##                    selected[i] = bj
##                    selected[j] = bi

        if len(selected) <= 0: return None
        
        i = 0
        goodEf = False
        while i < len(selected) and goodEf == False:
            if selected[i].BBEfficiency >= self.minEff:
                goodEf = True
            i = i + 1
        if goodEf == True:
            modelID = selected[(i - 1)].id
        else:
            modelID = selected[0].id
            showWarning(_("in the database no boiler with the desired efficiency has be found"))
        return modelID

    def selectAndAddBB(self, Pow, Top):
        modelID = self.selectBB(Pow, Top)
        if modelID is not None:
            equipe = self.addEquipmentDummy()
            self.setEquipmentFromDB(equipe, modelID) # assign model from DB to current equipment in equipment list
            model = dbio.get("dbboiler", id=modelID)
            addedPower = model.BBPnom
            return addedPower
        else:
            return 0
            
        
    def designBB80(self, demand):
        """Design a boiler system at 80°C"""
        iT80 = int(80 / Status.TemperatureInterval)
        iTmax = int(self.maxTemp / Status.TemperatureInterval)

        logTrack("moduleBB starting function 'designBB80'. QD80: %s QDmax: %s" % (demand[iT80], demand[iTmax]))

        added = 0
        if demand[iT80] >= 0.1 * demand[iTmax]: #we design boiler at this temperature level only if the demand is bigger than the 10% of the total demand

            Preq = self.QDh80[0] * self.securityMargin
            PminOp = self.QDh80[int((self.minOpTime) * Status.Nt / 8760)]

##            logTrack("ModuleBB.designBB80 TEST TEST Preq %s self.minPow %s 1.3*PminOp %s"%(Preq, self.minPow, 1.3*PminOp))
            if Preq >= self.minPow:
                if Preq >= 2 * self.minPow:
                    if Preq < PminOp * 1.3 or (Preq - PminOp) < self.minPow:
                        self.selectAndAddBB(Preq, 80) # select the right bb from the database.                        
                    else:
                        # first add boiler cascade for base load until 80 C
                        if PminOp > self.bigger80:
                            for i in range (int(PminOp / self.bigger80)):
                                added += self.selectAndAddBB(self.bigger80, 80)
                        else:
                            added += self.selectAndAddBB(PminOp, 80) # aggiungere condizione sul rendimento

                        if Preq - added >= 2 * self.minPow:
                            if Preq - added >= self.bigger80:
                                for i in range (int(Preq - added) / self.bigger80):
                                    added += self.selectAndAddBB(self.bigger80, 80)
                            else:
                                added += self.selectAndAddBB(Preq - added, 80) # sempre aggiungere anche il criterio di efficienza
                                if Preq - added >= 0:
                                    added += self.selectAndAddBB(self.minPow, 80) # sempre aggiungere anche il criterio di efficienza
                        else:
                            if added < Preq:                                                       #EF 26/10/2008
                                self.selectAndAddBB(min(Preq - added, self.minPow), 80)          #EF 26/10/2008
                else:   # required power is between MinPow and 2*MinPow
                    logTrack("moduleBB.designBB80: (MinPow < Preq < 2*MinPow. looking for a boiler with T>'%s' and Pow>'%s'" % (80, self.QDh80[0] * self.securityMargin))
                    self.selectAndAddBB(Preq, 80)

        self.sortBoiler()

    def designBB140(self, demand140, demand):
        """Design a boiler system at 140°C."""

        iT140 = int(140 / Status.TemperatureInterval)
        iTmax = int(self.maxTemp / Status.TemperatureInterval)

        logTrack("moduleBB starting function 'designBB140'. QD140: %s QDmax: %s" % (demand[iT140], demand[iTmax]))

        added = 0
        if demand140[iT140] >= 0.1 * demand[iTmax]:

            Preq = self.QDh140[0] * self.securityMargin
            PminOp = self.QDh140[int((self.minOpTime) * Status.Nt / 8760)]
            
            if Preq >= self.minPow:
                if Preq >= 2 * self.minPow:
                    if Preq < PminOp * 1.3 or \
                    Preq - PminOp < self.minPow:
                        self.selectAndAddBB((Preq), 140) # select the right bb from the database.
                    else:
                        if PminOp > self.bigger140:
                            for i in range (int(PminOp / self.bigger140)):
                                added += self.selectAndAddBB(self.bigger140, 140)
                        else:
                            added += self.selectAndAddBB(PminOp, 140)  # select the base load boiler from DB

                        if Preq - added >= 2 * self.minPow:
                            if Preq - added >= self.bigger140:
                                for i in range (int((Preq - added) / self.bigger140)):
                                    added += self.selectAndAddBB(self.bigger140, 140)

                                added += self.selectAndAddBB((Preq - added), 140)

                            else:
                                added += self.selectAndAddBB(((Preq - added)), 140) # sempre aggiungere anche il criterio di efficienza
                                if Preq - added >= 0:                                                         #EF 05/06/2009
                                    added += self.selectAndAddBB(((Preq - added)), 140)
                        else:
                            if added < Preq:                                                 #EF 26/10/2008     
                                self.selectAndAddBB(min((Preq - added), self.minPow), 140)    #EF 26/10/2008
                else:
                    self.selectAndAddBB(Preq, 140)

        self.sortBoiler()

    def designBBmaxTemp(self):
        """Design a boiler system at the maximum temperature of the 
        heat demand.
        
        """
        logTrack("moduleBB starting function 'designBBmaxTemp'")
        logTrack("ModuleBB.designBBmaxTemp: QDhmaxTemp[0] = %s" % self.QDhmaxTemp[0])

        iTmax = int(self.maxTemp / Status.TemperatureInterval)
        Preq = self.QDhmaxTemp[0] * self.securityMargin
        PminOp = self.QDhmaxTemp[int((self.minOpTime) * Status.Nt / 8760)]
        
        added = 0
        if Preq >= 2 * self.minPow:
            if Preq < PminOp * 1.3 or (Preq - PminOp) < self.minPow:
                added += self.selectAndAddBB(max((Preq), self.minPow), self.maxTemp) #HS....,...)  #select the right bb from the database.
            else:
                if PminOp > self.biggermaxTemp:
                    for i in range (int(PminOp / self.biggermaxTemp)):
                        added += self.selectAndAddBB(self.biggermaxTemp, self.maxTemp)
                else:
                    added += self.selectAndAddBB(max(PminOp, self.minPow), self.maxTemp)
                if Preq - added >= 2 * self.minPow:
                    if Preq - added >= self.biggermaxTemp:
                        for i in range (int((Preq - added) / self.biggermaxTemp)):
                            added += self.selectAndAddBB(self.biggermaxTemp, self.maxTemp)
                        self.selectAndAddBB(max((Preq - added), self.minPow), self.maxTemp)
                    else:
                        added += self.selectAndAddBB(max(((Preq - added) / 2), self.minPow), self.maxTemp)
                        if Preq - added >= 0:                                                     #EF 05/06/2009
                            added += self.selectAndAddBB(max(((Preq - added) / 2), self.minPow), self.maxTemp)
                else:
                    added += self.selectAndAddBB(max((Preq - added), self.minPow), self.maxTemp)
        else:
            self.selectAndAddBB(max(Preq, self.minPow), self.maxTemp)

        self.sortBoiler()

    def designAssistant(self):
        """auto-design of boiler cascade"""
        # getting configuration parameters of DA

        logTrack("ModuleBB: auto-design")
        DA = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        if DA is None:
            logTrack("ModuleBB (design assistant): WARNING - no DA configuration parameters available")
            return
        
        if DA.BBRedundancy == True:
            pass
        try:
            self.securityMargin = 1 + DA.BBSafety / 100
        except:
            self.securityMargin = 1.2
        if DA.BBMaintain != None:
            self.Maintain = DA.BBMaintain
        else:
            self.Maintain = 1
        if DA.BBRedundancy != None:
            self.redund = DA.BBRedundancy
        else:
            self.redund = 0
        if DA.BBPmin != None:
            self.minPow = DA.BBPmin
        else:
            self.minPow = 200
        if DA.BBHOp != None:
            self.minOpTime = DA.BBHOp
        else:
            self.minOpTime = 100
        if DEBUG in ("BASIC", "ALL"):
            print "ModuleBB (design assistant): securityMargin =", self.securityMargin
            print "ModuleBB (design assistant): maintain =", self.Maintain
            print "ModuleBB (design assistant): redundancy =", self.redund
            print "ModuleBB (design assistant): minPow =", self.minPow
            print "ModuleBB (design assistant): minOpTime =", self.minOpTime
        if DA.BBEff != None:
            self.minEff = DA.BBEff / 100
        else:
            self.minEff = 80
        if DA.BBFuelType != "Natural Gas":
            showWarning(_("In the present version only Natural Gas boilers has been collected in the DB"))
            return

        if self.Maintain == 0:
            for row in self.BBList:
                self.dummyEqId = row['equipeID']
                self.deleteEquipment(None)
        else:
            self.automDeleteBoiler(self.minEff)   # delete inefficient boiler

        logTrack("ModuleBB.designAssistant: %s equipments remaining after deleting inefficient or old ones" % self.NEquipe)

        # first do some sorting ...
        if self.cascadeIndex > 0:
##            self.findmaxTemp(calcQ_T(Status.mod.moduleEnergy.getUSHjRequired(self.cascadeIndex)))
            self.findmaxTemp(calcQ_T(Status.int.QD_Tt_mod[self.cascadeIndex-1]))       
        else:
            self.findmaxTemp(Status.int.USHTotal_T)

        logTrack("ModuleBB.designAssistant: maximum temperature = %s" % self.maxTemp)
        
        self.findBiggerBB()
        self.sortBoiler() # sort boilers by temperature (ascending) and by efficiency (descending)

        DAfirstBB = self.firstBB
        DAfirstBB140 = self.firstBB140
        DAfirstBBmaxTemp = self.firstBBmaxTemp
##        logTrack("ModuleBB.designAssistant TEST TEST BB80 %s BB140 %s BBmaxTemp %s"%(DAfirstBB,DAfirstBB140,DAfirstBBmaxTemp))

        # after shifting, the equipment cascade has to be updated, as the modified demand is used
        # in the following
        logTrack("MouleBB: DA after deleting old boilers cUL = %s NJ = %s" % (Status.int.cascadeUpdateLevel, self.NEquipe))
        if Status.int.cascadeUpdateLevel < self.NEquipe or Status.mod.moduleEnergy.outputLevel < 3:
            Status.mod.moduleEnergy.runSimulation(outputLevel=3, checkPeakPower=False) # check peak power = False -> avoid annoying warnings because all boilers may have been deleted
        elif self.NEquipe == 0:
            # the simulation will not have run but the heat recovery data may be stale
            if Status.schedules.equipmentSchedulesOutOfDate is True:
                if Status.SimTool == "Plus":
                    Status.mod.moduleEnergy.runSimulation(outputLevel=3, checkPeakPower=False) # check peak power = False -> avoid annoying warnings because all boilers may have been deleted
                else:
                    equipments.estimateEquipmentWasteHeat()
                    Status.mod.moduleHR.runHRModule()
                
            
        # low temperature boiler look-up
        exBP = 0 # power of the boiler in the cascade operating at 80°C

        for row in Status.int.cascade:
            equipTry = dbio.get("qgenerationhc", id=row["equipeID"])
            if equipTry is not None:
                if getEquipmentClass(row["equipeType"]) == "BB" and equipTry.TMaxSupply <= 80:
                    exBP += row['equipePnom']

        logTrack("ModuleBB: power of the boilers at 80°C present in the cascade at the moment %s" % exBP)
        
        if DAfirstBB < 0:
            logDebug("MoudleBB (DA): error in no. of firstBB: %s (NEquipe = %s)" % (DAfirstBB, self.NEquipe))

        iT80 = int(80 / Status.TemperatureInterval + 0.5)

        logTrack("ModuleBB: firstBB %s NEquipe %s lastHeating %s" % (DAfirstBB, self.NEquipe, self.lastHeating))
        if DAfirstBB >= self.NEquipe:    #no boiler remaining in cascade
            if self.lastHeating < 0:
                residualDemand = Status.int.USHTotal_Tt   # if no heating equipment at all, take total heat demand
                residualDemand_T = Status.int.USHTotal_T
            else:
                residualDemand = Status.int.QD_Tt_mod[self.lastHeating]
                residualDemand = residualDemand - Status.int.USHj_Tt[self.lastHeating]
                residualDemand_T = Status.int.calcQ_T(residualDemand)
        else:
            residualDemand = Status.int.QD_Tt_mod[DAfirstBB]
            residualDemand_T = Status.int.QD_T_mod[DAfirstBB]

        logTrack("residualDemand_T = %s" % residualDemand_T)

        QDmax = np.max(residualDemand[iT80])

        logTrack("ModuleBB: maximum power demand at 80°C %s" % QDmax)

        b = max((QDmax - exBP), 0)
        b1 = max(((QDmax * self.securityMargin) - exBP), 0) #   minimum power of the new boilers at 80°C
        c = residualDemand[iT80].copy()
        c = c.clip(max=b).tolist()
        
        self.QDh80 = c  # demand to be supplied by new boilers at 80°C
        self.QDh80.sort(reverse=True)
        
        level80 = False
        NEquipeOld = self.NEquipe
        logTrack('moduleBB, design assistant: QDmax %s b %s self.QDh80[0] %s exBP %s' % (QDmax, b, self.QDh80[0], exBP))
        if self.QDh80[0] > 0:
            self.designBB80(residualDemand_T)

        if self.NEquipe > NEquipeOld: level80 = True # detect wether an 80 °C level boiler has been added
        logTrack("ModuleBB: number of equipments after adding boilers for 80 C level %s" % self.NEquipe)
        
        # 140 °C boilers
        if self.maxTemp > 160:   # N.B. The difference between temperature levels is now setted in 20°C but in the future could be a parameter.
            intermediateLevel = True
            exBP = 0       #   power of the boiler in the cascade operating at 140°C
            for row in Status.int.cascade:
                equipTry = dbio.get("qgenerationhc", id=row["equipeID"])
                if equipTry is not None:
                    if getEquipmentClass(row["equipeType"]) == "BB" and 90 <= equipTry.TMaxSupply <= 140:
                        exBP += row['equipePnom']

            iT140 = int(140 / Status.TemperatureInterval + 0.5)

            if DAfirstBB140 >= self.NEquipe:
                residualDemand140 = residualDemand
                residualDemand140_T = residualDemand_T.copy()
                
                pass    # take same residual demand as for 80 C boiler
            else:
                residualDemand140 = Status.int.QD_Tt_mod[DAfirstBB140] - residualDemand
                residualDemand140_T = Status.int.QD_T_mod[DAfirstBB140] - residualDemand_T

            QDmax = np.max(residualDemand140[iT140])
            logTrack("ModuleBB: maximum power demand at 140°C %s" % QDmax)

            b = max((QDmax - exBP), 0) #   minimum power of the new boilers at 140°C
            c = []
            for it in range (Status.Nt):
                c.append (min (b, residualDemand[iT140][it]))

            self.QDh140 = c  # demand to be supplied by new boilers at 140°C
            self.QDh140.sort(reverse=True)

            if self.QDh140[0] > 0:
                self.designBB140(residualDemand140_T, residualDemand_T)

            residualDemand = residualDemand140
            residualDemand_T = residualDemand140_T
                                     
        else:
            intermediateLevel = False

        logTrack("ModuleBB: number of equipments after adding boilers for 140 C level %s" % self.NEquipe)
        
        iT_maxTemp = int((self.maxTemp / Status.TemperatureInterval) + 0.5)

        #TODO: check if this still works well in pipe-by-pipe modelling approach. what is self.cascadeIndex if (a) theres no boiler (b) theres no heating equipment at all
##        logTrack("ModuleBB: TEST TEST DAfirstBBmaxTemp = %s NEqupipe %s iT_maxTemp %s iT80 %s"%(DAfirstBBmaxTemp, self.NEquipe, iT_maxTemp, iT80))
        if DAfirstBBmaxTemp >= self.NEquipe:
            residualDemandmaxTemp = residualDemand.copy()
        else:
            residualDemandmaxTemp = Status.int.QD_Tt_mod[DAfirstBBmaxTemp]

        if DAfirstBBmaxTemp == DAfirstBB140 and intermediateLevel:
            logTrack("ModuleBB.DA: Residual power of max. temp level as diff. of max. temp and 140C level")
            self.QDhmaxTemp = (residualDemandmaxTemp[iT_maxTemp] - residualDemand140[iT140]).tolist()

        elif DAfirstBBmaxTemp == DAfirstBB and level80:
            logTrack("ModuleBB.DA: Residual power of max. temp level as diff. of max. temp and 80C level")
            self.QDhmaxTemp = (residualDemandmaxTemp[iT_maxTemp] - residualDemandmaxTemp[iT80]).tolist()
            
        else:
            logTrack("ModuleBB.DA: Residual power of max. temp level as demand of maxtemp boiler in original cascade")
            self.QDhmaxTemp = residualDemandmaxTemp[iT_maxTemp].tolist()

        self.QDhmaxTemp.sort(reverse=True)

        logTrack("ModuleBB: maximum power demand at max. temperature =  %s" % self.QDhmaxTemp[0])

        if self.QDhmaxTemp[0] > 0:
            self.designBBmaxTemp()

        logTrack("ModuleBB: number of equipments after adding boilers for max. temperature level %s" % self.NEquipe)
        
        self.sortBoiler()
        if self.redund == 1:
            self.redundancy()

        logTrack("ModuleBB: number of equipments after considering redundancy %s" % self.NEquipe)
        

    def estimateEconomicParameters(self, eq):
        """ Default values for economic parameters
        """

        if hasattr(eq, "eqType"): # is an eqObject        
            eqType = eq.eqType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OMFix
            OMVar = eq.OMVar

            Pnom = eq.Pnom
            
        else:   # kept for backward compatibility - still used in setEquipmentFromDB
            eqType = eq.EquipType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OandMfix
            OMVar = eq.OandMvar

            Pnom = eq.HCGPnom
            

        # security check if the reference power is defined
        Pref = Pnom 
        if Pref is None:
            return (None, None, None)

        if turnKey is None:
            if hasattr(eq, "Price") and eq.Price is not None:
                turnKey = max(2.*eq.Price,eq.Price+40*Pnom)

            elif eqType == "steam boiler":
                turnKey = 100.0*Pnom
            elif eqType == "condensing boiler":
                turnKey = 120.0*Pnom
            elif eqType in ["hot water boiler", "burner (direct heating)", "burner (indirect heating)"]:
                turnKey = 70.0*Pnom
            else:
                logWarning(_("Turnkey investment costs for equipment %s could not be calculated") % eq.name)
                turnKey = 0.0
            
        OMFix = noneCheck(OMFix, 0.005 * turnKey)
        OMVar = noneCheck(OMVar, 0.0)      

        return (turnKey, OMFix, OMVar)


if __name__ == "__main__":
    pass
