# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#    CompositeCurve
#            
#------------------------------------------------------------------------------
#            
#    Class HeatFlow: defines a heat flow Q_T(T) at a given time step
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from modules.heatFlow_Tit import CompositeCurve
from modules.fluids import Fluid
import numpy as np

EPS = 1.e-6   # accuracy for condensation temperature

def buildCC(streamGlb):
    """Builds a basic composite curve from a streamGlb-segment
    """
    cc = CompositeCurve()

    T0 = streamGlb.T0
    T1 = streamGlb.T1
    Q = streamGlb.Q

    if Q > 0:
        cc.T.append(T0)
        cc.Q.append(0)

        if streamGlb.fluidId is not None:
            fluid = Fluid(streamGlb.fluidId)
            Tc = fluid.TCond
            cp = fluid.cp
            hL = fluid.hL
            cpG = fluid.cpG
            condensation = (Tc >= T0 and Tc <= T1)
        else:
            condensation = False

        if condensation:  #phase change
            dQL = cp * (Tc - T0)
            dQC = hL                  #TODO: possible cases of partial evaporation / condensation are not considered
            dQV = cpG * (T1 - Tc)
            dq = dQL + dQC + dQV

            if dq > 0:      
                f = Q / dq
                dQL *= f
                dQC *= f
                dQV *= f
            else:       # security feature for strange case: (T0 = T1 = Tc or cp = cpV = 0) and hL = 0. check not necessary if this strange fluid properties are avoided at entry)
                dQL = 0
                dQC = dQ
                dQV = 0

            if dQC > 0:
                if Tc > T0:
                    cc.T.append(Tc)
                    cc.Q.append(dQL)

                if Tc < T1:
                    cc.T.append(Tc)
                    cc.Q.append(dQL + dQC)
                    
                cc.Tc.append(Tc)
        
        cc.T.append(T1)
        cc.Q.append(Q)

    return cc
       
        
   

    
