# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	BUILDINGS
#			
#------------------------------------------------------------------------------
#			
#	Tools for simulation and modeling of buildings
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
"""
:mod:`einstein.modules.buildings` -- Buildings
==============================================

   :synopsis: Functions and a class to get, create and manipulate buildings and
       their properties.
   :author:   Jan Ries <jan.ries@energyxperts.net>

"""
from GUI.dialogGauge import DialogGauge
from modules.fluids import Fluid, getDefaultAir

from math import modf, pi, cos
from datetime import date, datetime, timedelta
from database.dbio import check
import database.dbio as dbio
from GUI.status import *
from constants import BUILDINGPROCESSES, TRANSYESNO
from modules.messageLogger import *
from auxiliary.auxiliary import noneCheck
from modules.schedules.profiles import WeeklyProfile, WEEKDAYS
from modules.schedules.schedules import ProcessPeriod, deleteProcessPeriodsFromDB, PeriodSchedule, ScheduleNotFoundError
from modules.project import InconsistentDataBaseError
from model.objects.processStreams import InflowingStream, OutflowingStream, getStreams
from modules.calendar import Calendar, overlapCalendars, getGeneralCalendar
from lib.dateTime.dateTimeTools import DOYToDate

##PERIODREFERENCE = date(2007, 1, 1) # year 2007 starts and ends on Monday

class ProjectBuildingsNotFoundError(Exception):
    """No building for project with given id found in DB."""
    pass


class BuildingNotFoundError(Exception):
    """No building with given name for project found in DB."""
    pass


def getBuildingNamesFromDB(projectId, ANo=None):
    """Retrieve all known building names for given projectId from data base.
    
    :param projectId: Id of project with buildings to look for.
    :type  projectId: Integer
    :param ANo: number of the alternative proposal, 
        defaults to the current propsal if None is given.
    :type alternativePorposal: Integer 

    :returns: All building names stored in data base for projectId.
    :rtype:   List of strings
    
    """
    if ANo is None:
        ANo = Status.ANo
    return [b.BuildName for b in dbio.select("qbuildings", Questionnaire_id=projectId, \
            AlternativeProposalNo=ANo)]

def getBuildings(projectId, ANo=None):
    """Retrieve all buildings associated with the given projectId from 
    the database.
    
    :param projectId: Id of the process whose buildings are to retrieved.
    :type projectId: Integer
    :param ANo: number of the alternative proposal, 
        defaults to the current propsal if None is given.
    :type alternativePorposal: Integer 
    
    :returns: a list of all buildings in the project. 
    :rtype: a List of Building objects.
    
    """
    if ANo is None:
        ANo = Status.ANo
    names = getBuildingNamesFromDB(projectId, ANo)
    result = []
    for name in names:
        building = Building(name)
        building.loadFromDB(projectId, ANo)
        result.append(building)
    return result

class Building(object):
    """Represent a building in a project with all its properties."""
    def __init__(self, name):
        """Create a building, either by setting its properties
        explicitly or by calling loadFromDb() later. 
        
        """
        self.name = name
        self.dbId = None
        self.buildConstructSurface = None
        self.buildUsefulSurface = None
        self.buildUsage = None
        self.buildEnergyDemand = None
        self.buildMaxHP = None
        self.buildMaxCP = None
        self.buildAnnualHeating = None
        self.buildAnnualAirCond = None
        self.buildTHeating = None
        self.buildTAirCond = None
        self.buildDailyDHW = None
        self.buildHoursOccup = None
        self.buildDaysInUse = None
        self.buildHolidaysPeriodStart_1 = None
        self.buildHolidaysPeriodStart_2 = None
        self.buildHolidaysPeriodStart_3 = None
        self.buildHolidaysPeriodStop_1 = None
        self.buildHolidaysPeriodStop_2 = None
        self.buildHolidaysPeriodStop_3 = None
        self.buildHeatingPeriodStart = None
        self.buildHeatingPeriodStop = None
        self.buildAirCondPeriodStart = None
        self.buildAirCondPeriodStop = None
        self.enclBuildDraw = None
        self.airHandlingUnit = None
        self.airFlowRateH = None
        self.airFlowRateC = None
        self.TInflowRecH = None
        self.TInflowRecC = None
        self.TIndoorH = None
        self.TIndoorC = None
        self.processIDs = [None,None,None] #Process IDs for Heating, Cooling and HW

##        q = dbio.select("dbfluid", FluidName="Air")
##        if len(q) > 0:
##            self.fluidAirId = q[0]["DBFluid_ID"]
##            self.fluidAir = Fluid(self.fluidAirId)
##        else:
##            self.fluidAir = None
        self.fluidAir = getDefaultAir()

    def deleteFromDB(self, projectId=Status.PId, ANo=Status.ANo):
        """Delete the building with our name associated to the given 
        project.
        
        :param projectId: If None is given the projectId of the current 
            project is used as a default.
        :param ANo: number of the alternative proposal, 
            defaults to the current propsal if None is given.
        :type alternativePorposal: Integer 
      
        """
        # get the id of the building
        dbbuilding = dbio.select("qbuildings", Questionnaire_id=projectId, \
                        AlternativeProposalNo=ANo, \
                        BuildName=self.name)
        if len(dbbuilding) == 0:
            raise BuildingNotFoundError
        elif len(dbbuilding) > 1:
            raise InconsistentDataBaseError, "The buildingname %s is not unique\
            in the project %s." % (self.Name, projectId)
        else:
            # first remove all processes attached to the building
            buildingId = dbbuilding[0].QBuildings_ID
            dbio.delete("qprocessdata", QBuildings_ID=buildingId)
            # now remove the building itself
            dbbuilding[0].delete()

    def saveToDB(self, projectId, ANo=Status.ANo, omitProcesses=False):
        """Save the building as part of the project whose is given.
        
        :param projectId: the id of the project the building is part
            of. 
        :param ANo: the alternative proposal number, defaults to the
            current alternative proposal
        :param omitProcesses: if True then no the processes are 
            created, deleted or updated, which usually represent the 
            demands of the building.
        
        """
        dbBuilding = {'BuildName': check(self.name),
                      'BuildConstructSurface': check(self.buildConstructSurface),
                      'BuildUsefulSurface': check(self.buildUsefulSurface),
                      'BuildUsage': check(self.buildUsage),
                      'BuildEnergyDemand': check(self.buildEnergyDemand),
                      'BuildMaxHP': check(self.buildMaxHP),
                      'BuildMaxCP': check(self.buildMaxCP),
                      'BuildAnnualHeating': check(self.buildAnnualHeating),
                      'BuildAnnualAirCond': check(self.buildAnnualAirCond),
                      'BuildTHeating': check(self.buildTHeating),
                      'BuildTAirCond': check(self.buildTAirCond),
                      'AirHandlingUnit': check(self.airHandlingUnit),
                      'AirFlowRateH': check(self.airFlowRateH),
                      'AirFlowRateC': check(self.airFlowRateC),
                      'TInflowRecH': check(self.TInflowRecH),
                      'TInflowRecC': check(self.TInflowRecC),
                      'IndoorTHeating': check(self.TIndoorH),
                      'IndoorTAirCond': check(self.TIndoorC),
                      'BuildDailyDHW': check(self.buildDailyDHW),
                      'BuildHoursOccup': check(self.buildHoursOccup),
                      'BuildDaysInUse': check(self.buildDaysInUse),
                      'BuildHolidaysPeriodStart_1': check(self.buildHolidaysPeriodStart_1),
                      'BuildHolidaysPeriodStart_2': check(self.buildHolidaysPeriodStart_2),
                      'BuildHolidaysPeriodStart_3': check(self.buildHolidaysPeriodStart_3),
                      'BuildHolidaysPeriodStop_1': check(self.buildHolidaysPeriodStop_1),
                      'BuildHolidaysPeriodStop_2': check(self.buildHolidaysPeriodStop_2),
                      'BuildHolidaysPeriodStop_3': check(self.buildHolidaysPeriodStop_3),
                      'BuildHeatingPeriodStart': check(self.buildHeatingPeriodStart),
                      'BuildHeatingPeriodStop': check(self.buildHeatingPeriodStop),
                      'BuildAirCondPeriodStart': check(self.buildAirCondPeriodStart),
                      'BuildAirCondPeriodStop': check(self.buildAirCondPeriodStop),
                      'EnclBuildDraw': check(self.enclBuildDraw),
                      'AlternativeProposalNo': ANo,
                      'Questionnaire_id': projectId}
        try:
            buildingrow = dbio.select("qbuildings", Questionnaire_id=projectId, \
                         AlternativeProposalNo=ANo, \
                         BuildName=check(self.name))
            buildingrow[0].set(**dbBuilding)
            buildingId = buildingrow[0].QBuildings_ID
        except (LookupError, IndexError): # no building with our name found for given projectId
            buildingId = dbio.insertRow("qbuildings", **dbBuilding)
        dbio.commit()
        
        if not omitProcesses:
            self._updateProcesses(buildingId)
            dbio.commit()

    def _updateProcesses(self, buildingid):
        """Update the properties of the processes attached to the building.
        
        :param buildingid: The QBuildings_ID value from the DB

        :returns: the number of processes attached to the building.
        
        """
        dlg = DialogGauge(Status.main, _("EINSTEIN building module"), _("calculating"))
        building = dbio.get("qbuildings",id=buildingid)
        ANo = building.AlternativeProposalNo
        PId = building.Questionnaire_id
        processcount = 0
        unitOps = Status.prj.getUnitOpDict()
        heatingID = 0
        coolingID = 0
        hwID = 0

        # get building holiday period
        self.getHolidayPeriodList(building) # stores building holidays within self.holidayPeriods
        buildingCalendar = overlapCalendars(getGeneralCalendar(),Calendar(self.holidayPeriods))

        # default weekly occupation, overwritten if user specified
        self.nDaysPerWeek = 7
        self.fDaysPerWeek = 1.0
        
        if self.buildDaysInUse is not None and buildingCalendar.days > 0:
            rDaysPerWeek = self.buildDaysInUse * 7.0 / buildingCalendar.days
            if rDaysPerWeek > 7:
                logWarning(_("Inconsistency between holiday schedules and days of building occupation %s (max. %s)")%(self.buildDaysInUse, buildingCalendar.days))
                logWarning(_("Occupation of 7 days per week assumed"))
            
            else:
                self.nDaysPerWeek = 7 - int(7 - rDaysPerWeek)
                self.fDaysPerWeek = rDaysPerWeek / self.nDaysPerWeek
        
        for key, value in unitOps.items():
            if value == u"Heating of production halls":
                heatingID = key
            elif value == u"Cooling of production halls":
                coolingID = key
            elif value == u"Service water":
                hwID = key
        for n, pname in enumerate(self.getProcessNames()):
            if pname.endswith("heating"):
                if self.buildAnnualHeating == 0:
                    # no heating energy is used, delete the process
                    self._deleteProcess("heating", buildingid)
                else:
                    process = self._getProcess(pname, "heating", buildingid, PId, ANo)
                    process.Description = "This is the heating process for building %s." % (building.BuildName,)
                    process.DBUnitOperation_id = heatingID
                    if coolingID == 0:
                        logError("PanelQ8.OnButtonOK.updateProcesses: missing UnitOperation: Heating of production halls")

                    if self.TIndoorH is not None:
                        process.PT = float(self.TIndoorH)
                    else:
                        process.PT = BUILDINGPROCESSES["heating_temperature"]

                    if self.buildTHeating is not None:
                        process.DPSTmin = max(float(self.buildTHeating) - process.PT, 0)
                    else:
                        process.DPSTmin = 0
                    process.ExternalSupply = "no" # default = False, if any AHU, then set to True
                    
                    process.QOpProc = building.BuildMaxHP
                    process.UPH = building.BuildAnnualHeating
                    process.heating_cooling = "heating"

                    try:
                        self._buildHeatingSchedule(process.QProcessData_ID,
                                                   buildingid, PId, ANo)
                        processcount += 1
                        self.processIDs[0] = process.QProcessData_ID
                    except ValueError:
                        logWarning(_("Couldn't create the heating process. "\
                                      "The start and end date of the heating period is needed."))
                        self._deleteProcess("heating", buildingid)    

                    else:
                        AHU_heating = (self.airHandlingUnit > 0)
                        if AHU_heating:

                            process.ExternalSupply = "yes"

                            fluidAir = getDefaultAir()

                            (inStreams, outStreams) = getStreams(process.QProcessData_ID)
                            assert(len(inStreams) <= 1)
                            if len(inStreams) == 0:
                                stream = InflowingStream("air inflow")
                            else:
                                stream = inStreams[0]

                            TInAirH = self._getTInAHU(self.heatingStart, self.heatingEnd, 0, 24)
                            if TInAirH > process.PT:
                                logWarning("Average outdoor temperature during heating period %s °C higher than indoor temperature %s °C" % (TInAirH, process.PT))
                                TInAirH = process.PT
                            stream.PTInFlow = TInAirH

                            stream.HeatRecExist = False
                            stream.PTInFlowRec = stream.PTInFlow

                            if self.TInflowRecH is not None:
                                if self.TInflowRecH > process.PT or self.TInflowRecH < stream.PTInFlow:
                                    logWarning("Inconsistency in air inflow temperature after heat recovery: %s °C (Indoor: %s, Ambient: %s" % 
                                             (self.TInflowRecH, process.PT, stream.PTInFlow))
                                    
                                stream.PTInFlowRec = max(min(process.PT, self.TInflowRecH), stream.PTInFlow)
                                if stream.PTInFlowRec > stream.PTInFlow:
                                    stream.HeatRecExist = True

                            mInAirH = self.airFlowRateH
                            if mInAirH is None:
                                mInAirH = 0
                            else:
##                                mInAirH *= fluidAir.rho # convert volume flow rate from GUI into mass flow
                                mInAirH /= self.fAvgH          # calculate nominal (=maximum flow rate) from average of parabola: 2/3
                                try:
                                    mInAirH *= self.buildHoursOccup / 24.0 # correct air flow rate (reality: ventilation only during occupancy, model: 24h
                                except TypeError:
                                    pass # assume 24 hours of occupation.
                                
                            stream.mInFlowNom = mInAirH
                            stream.VInFlowCycle = None
                            stream.HInFlow      = None
                            stream.XInFlow      = None
                            stream.UPHc         = None
                            stream.QdotProc_c   = None
                            stream.Medium = fluidAir.name
                            stream.dbfluid_id = fluidAir.id
                            stream.saveToDB(process.QProcessData_ID)

                            if self.airHandlingUnit >= 2:
                                assert(len(outStreams) <= 1)
                                if len(outStreams) == 0:
                                    stream = OutflowingStream("air outflow")
                                else:
                                    stream = outStreams[0]

                                TOutAirH = process.PT
                                stream.PTOutFlowRec = TOutAirH
                                stream.PTOutFlow = None
                                stream.HeatRecExist = True
                                stream.PTFinal = 0
                                stream.PTFinalRequired = None
                                stream.HeatRecOK = True

                                mOutAirH = mInAirH
                                    
                                stream.mOutFlowNom = mOutAirH
                                stream.VOutFlowCycle = None
                                stream.HOutFlow = None
                                stream.XOutFlow = None
                                stream.UPHw = None
                                stream.QdotProc_w = None
                                stream.UPHwReq = None

                                stream.Medium = fluidAir.name
                                stream.dbfluid_id = fluidAir.id
                                stream.saveToDB(process.QProcessData_ID)
                            else:
                                if outStreams:
                                    outStreams[0].deleteFromDB()
                                

                        else: # no AHU at all
                            (inStreams, outStreams) = getStreams(process.QProcessData_ID)
                            if outStreams:
                                outStreams[0].deleteFromDB()
                            if inStreams:
                                inStreams[0].deleteFromDB()

            elif pname.endswith("cooling"):
                if self.buildAnnualAirCond == 0:
                    # no cooling energy is used, delete the process
                    self._deleteProcess("cooling", buildingid)
                else:
                    process = self._getProcess(pname, "cooling", buildingid, PId, ANo)
                    process.Description = "This is the cooling process for building %s." % (self.name,)
                    process.DBUnitOperation_id = coolingID
                    if coolingID == 0:
                        logError("PanelQ8.OnButtonOK.updateProcesses: missing UnitOperation: Cooling of production halls")

                    if self.TIndoorC is not None:
                        process.PT = float(self.TIndoorC)
                    else:
                        process.PT = BUILDINGPROCESSES["cooling_temperature"]

                    if self.buildTAirCond is not None:
                        process.DPSTmin = max(process.PT - float(self.buildTAirCond), 0)
                    else:
                        process.DPSTmin = 0
                    process.ExternalSupply = "no"

                    process.QOpProc = self.buildMaxCP
                    process.UPH = self.buildAnnualAirCond
                    process.heating_cooling = "cooling"

                    try:
                        self._buildCoolingSchedule(process.QProcessData_ID,
                                                   buildingid, PId, ANo)
                        processcount += 1
                        self.processIDs[1] = process.QProcessData_ID
                    except ValueError:
                        logWarning(_("Couldn't create the air condition process. "\
                                      "The start and end date of the air conditioning period is needed."))
                        self._deleteProcess("cooling", buildingid)

                    else:
                        #TODO: read in from building questionnaire
                        AHU_cooling = (self.airHandlingUnit > 0)
                        if AHU_cooling is True:

                            process.ExternalSupply = "yes"

                            fluidAir = getDefaultAir()

                            (inStreams, outStreams) = getStreams(process.QProcessData_ID)
                            assert(len(inStreams) <= 1)
                            if len(inStreams) == 0:
                                stream = InflowingStream("air inflow")
                            else:
                                stream = inStreams[0]

                            try:
                                    startHour = max(12.0 - self.buildHoursOccup / 2.0, 0)
                                    stopHour = min(startHour + self.buildHoursOccup, 24.0)
                            except TypeError:
                                startHour = 0
                                stopHour = 24
                            
                            TInAirC = self._getTInAHU(self.coolingStart, self.coolingEnd, startHour, stopHour)
                            if TInAirC < process.PT:
                                logWarning("Average outdoor temperature during air conditioning period %s °C lower than indoor temperature %s °C" % (TInAirC, process.PT))
                                TInAirC = process.PT
                            stream.PTInFlow = TInAirC

                            stream.HeatRecExist = False
                            stream.PTInFlowRec = stream.PTInFlow

                            if self.TInflowRecC is not None:
                                if self.TInflowRecC < process.PT or self.TInflowRecC > stream.PTInFlow:
                                    logWarning("Inconsistency in air inflow temperature after heat recovery: %s °C (Indoor: %s, Ambient: %s)" % 
                                             (self.TInflowRecC, process.PT, stream.PTInFlow))
                                    
                                stream.PTInFlowRec = min(max(process.PT, self.TInflowRecC), stream.PTInFlow)
                                if stream.PTInFlowRec < stream.PTInFlow:
                                    stream.HeatRecExist = True

                            mInAirC = self.airFlowRateC
                            if mInAirC is None:
                                mInAirC = 0
                            else:
##                                mInAirC *= fluidAir.rho # convert volume flow rate from GUI into mass flow
                                mInAirC /= self.fAvgC # calculate nominal (=maximum flow rate) from average of parabola: 2/3
                                
                            stream.mInFlowNom = mInAirC
                            stream.Medium = fluidAir.name
                            stream.dbfluid_id = fluidAir.id
                            stream.saveToDB(process.QProcessData_ID)

                            if self.airHandlingUnit >= 2:
                                assert(len(outStreams) <= 1)
                                if len(outStreams) == 0:
                                    stream = OutflowingStream("air outflow")
                                else:
                                    stream = outStreams[0]

                                TOutAirC = process.PT
                                stream.PTOutFlowRec = TOutAirC
                                stream.PTOutFlow = None
                                stream.HeatRecExist = True
                                stream.PTFinal = max(TInAirC, TOutAirC) #Avoid negative deltaT
                                stream.PTFinalRequired = None
                                stream.HeatRecOK = True

                                mOutAirC = mInAirC
                                    
                                stream.mOutFlowNom = mOutAirC
                                stream.Medium = fluidAir.name
                                stream.dbfluid_id = fluidAir.id

                                stream.saveToDB(process.QProcessData_ID)
                            else:
                                if outStreams:
                                    outStreams[0].deleteFromDB()
                        
                        else: # no AHU at all
                            (inStreams, outStreams) = getStreams(process.QProcessData_ID)
                            if outStreams:
                                outStreams[0].deleteFromDB()
                            if inStreams:
                                inStreams[0].deleteFromDB()
                        
            elif pname.endswith("HW"):
                if self.buildDailyDHW == 0:
                    # no domestic hot water is used, delete the process
                    self._deleteProcess("domestic_hot_water", buildingid)
                else:
                    process = self._getProcess(pname, "domestic_hot_water", buildingid, PId, ANo)
                    process.Description = "This is the domestic hot water process for building %s." % (self.name,)
                    process.DBUnitOperation_id = hwID
                    if hwID == 0:
                        logError("PanelQ8.OnButtonOK.updateProcesses: missing UnitOperation: Service water")
                    process.PT = BUILDINGPROCESSES["hotwater_final_temperature"]
                    process.PTInFlow = BUILDINGPROCESSES["hotwater_inlet_temperature"]
                    process.VInFlowCycle = self.buildDailyDHW
                    process.QOpProc = 0
                    process.heating_cooling = "heating"
                    self._buildDHWSchedule(process.QProcessData_ID, buildingid,
                                           PId, ANo)
                    self.processIDs[2] = process.QProcessData_ID
                    processcount += 1
                    (inStreams, dummy) = getStreams(process.QProcessData_ID)
                    assert(len(inStreams) <= 1)
                    if len(inStreams) == 0:
                        stream = InflowingStream("Domestic Hotwater Inflow")
                    else:
                        stream = inStreams[0]
                    stream.PTInFlow = BUILDINGPROCESSES["hotwater_inlet_temperature"]
                    stream.PTInFlowRec = stream.PTInFlow
                    stream.VInFlowCycle = self.buildDailyDHW
##                    stream.QdotProc_c = 0
                    stream.HeatRecExist = False
                    stream.Medium = Fluid(0).name
                    stream.dbfluid_id = 0

                    stream.saveToDB(process.QProcessData_ID)
            dlg.update((n + 1) * (100 / len(self.getProcessNames())))
        dlg.Destroy()
        dbio.commit()

        Status.int.systemModel.changed()
        Status.int.systemModel.build()
        return processcount


    def loadFromDB(self, projectId=None, ANo=None):
        """Load Building with our name associated with given projectId.
        If no alternative proposal number is given, then the current 
        one is the default.
        
        """
        projectId = projectId if projectId is not None else Status.PId
        ANo = ANo if ANo is not None else Status.ANo
        buildingrow = dbio.get("qbuildings", Questionnaire_id=projectId, \
                        AlternativeProposalNo=ANo, \
                        BuildName=check(self.name))
        if buildingrow is None:
            raise BuildingNotFoundError
        
        self.dbId = buildingrow.QBuildings_ID
        self.buildConstructSurface = buildingrow.BuildConstructSurface
        self.buildUsefulSurface = buildingrow.BuildUsefulSurface
        self.buildUsage = buildingrow.BuildUsage
        self.buildEnergyDemand = buildingrow.BuildEnergyDemand
        self.buildMaxHP = buildingrow.BuildMaxHP
        self.buildMaxCP = buildingrow.BuildMaxCP
        self.buildAnnualHeating = buildingrow.BuildAnnualHeating
        self.buildAnnualAirCond = buildingrow.BuildAnnualAirCond
        self.buildTHeating = buildingrow.BuildTHeating
        self.buildTAirCond = buildingrow.BuildTAirCond
        self.buildDailyDHW = buildingrow.BuildDailyDHW
        self.buildHoursOccup = buildingrow.BuildHoursOccup
        self.buildDaysInUse = buildingrow.BuildDaysInUse
        self.buildHolidaysPeriodStart_1 = buildingrow.BuildHolidaysPeriodStart_1
        self.buildHolidaysPeriodStart_2 = buildingrow.BuildHolidaysPeriodStart_2
        self.buildHolidaysPeriodStart_3 = buildingrow.BuildHolidaysPeriodStart_3
        self.buildHolidaysPeriodStop_1 = buildingrow.BuildHolidaysPeriodStop_1
        self.buildHolidaysPeriodStop_2 = buildingrow.BuildHolidaysPeriodStop_2
        self.buildHolidaysPeriodStop_3 = buildingrow.BuildHolidaysPeriodStop_3
        self.buildHeatingPeriodStart = buildingrow.BuildHeatingPeriodStart
        self.buildHeatingPeriodStop = buildingrow.BuildHeatingPeriodStop
        self.buildAirCondPeriodStart = buildingrow.BuildAirCondPeriodStart
        self.buildAirCondPeriodStop = buildingrow.BuildAirCondPeriodStop
        self.enclBuildDraw = buildingrow.EnclBuildDraw
        self.airHandlingUnit = noneCheck(buildingrow.AirHandlingUnit, 0)
        self.airFlowRateH = buildingrow.AirFlowRateH
        self.airFlowRateC = buildingrow.AirFlowRateC
        self.TInflowRecH = buildingrow.TInflowRecH
        self.TInflowRecC = buildingrow.TInflowRecC
        self.TIndoorH = buildingrow.IndoorTHeating
        self.TIndoorC = buildingrow.IndoorTAirCond
#        self.airFlowRateH = 1000
#        self.airFlowRateC = 1000
        #TODO: add required indoor set temperature in addition to heating / cooling system inlet temperature
#        self.TIndoorH = 20
#        self.TIndoorC = 25

#        self.airHandlingUnit = 2 #0 no AHU; 1: only air inlet; 2: in- and outlet

#        self.TInflowRecC = self.TIndoorC + 2.5
#        self.TInflowRecH = self.TIndoorH - 2.5

    def getProcessNames(self):
        """Return the names of the process that could be created if
        this building has the appropriate properties.
        """
        return [self.name + "_" + name for name in 
                ["heating", "cooling", "HW"]]
        
    def _deleteProcess(self, purpose, buildingID):
        """
        Remove the process with the given name attached to the building from the DB.
        
        If the process doesn't exist no error is raised. 
        
        :param purpose: the purpose of the process to be deleted, is 
            one of "heating", "cooling" or "domestic_hot_water".
        :param buildingID: The building id the processes is attached to.
        :returns: True if some process got deleted, False otherwise.
        
        """
        result = False
        q = dbio.select("qprocessdata", QBuildings_ID=buildingID, Building_purpose=purpose)
        if len(q) > 0:
            Status.prj.deleteProcess(q[0].id)
            result = True
            
        if purpose == "heating":
            self.processIDs[0] = None
        elif purpose == "cooling":
            self.processIDs[1] = None
        elif purpose == "domestic_hot_water":
            self.processIDs[2] = None
            
        return result

    def _getProcess(self, processName, purpose, buildingID, PId, ANo):
        """Get process data from DB and create new one if there was none.
        
        If a process is created it is filled with default values where
        sensible. 
        
        :param processName: The name of the process to be fetched.
        :param purpose: the purpose of the process to be fetched, is 
            one of "heating", "cooling" or "domestic_hot_water".
        :param buildingID: The building id the processes is attached to.
        :returns: a psql.Result object representing the process.
        
        """
        q = dbio.select("dbfluid", FluidName="Water")
        if len(q) > 0:
            fluidid = q[0].id
        else:
            fluidid = 0
            
        q = dbio.select("qprocessdata", QBuildings_ID=buildingID, \
            Building_purpose=purpose)
        if len(q) == 0:
            process = Status.prj.addProcessDummy(PId, ANo)
            process.Process = check(processName)
            process.QBuildings_ID = buildingID
            process.Building_purpose = purpose
            process.ProcMedDBFluid_id = fluidid
            process.ProcType = "continuous"
            process.mInFlowNom = 0
            process.VolProcMed = 0
            process.HeatRecOK = TRANSYESNO["no"]
            process.HeatRecExist = TRANSYESNO["no"]
            process.mOutFlowNom = 0
            process.NBatch = 1
        else:
            process = q[0]
        return process
    
    def _buildYearlyHCSchedule(self, firstDay, lastDay, amplitude=None):
        """Return a Schedule for the building heating or cooling process.
        
        With float scale factors from the interval [0,1].
        
        :param firstDay: first day in the year (from 1 to 365) with demand > 0
        :param lastDay: last day in the year (from 1 to 365) with demand > 0
        :param amplitude: amplitude of variation in demand (for full-year demand only)
                            +1 -> maximum in July/August
                            -1 -> maximum in January/February

        :returns:  a list of tuples ((firstday, lastday), scalefactor)
                 
        """
        if firstDay <= lastDay:
            if lastDay > 364 or lastDay is None:   #security check for avoiding infinite while - loops below
                logWarning(_("Erroneous parameter in building periods"))
                lastDay = 364

            t0 = int(firstDay / 7) * 7
            periods = [(firstDay, t0 + 6)]     #list of periods as tuple of (first,last) day in period ...
            tpstart = t0 + 7
            while tpstart <= lastDay:
                periods.append((tpstart, tpstart + 6))
                tpstart += 7
            periods[-1] = (periods[-1][0], lastDay)
        else:                               # split in two parts, from 1.1. to lastDay and from firstDay to 31.12.
            t0 = 1
            periods = [(0, 6)]
            tpstart = 7
            while tpstart <= lastDay:
                periods.append((tpstart, tpstart + 6))
                tpstart += 7
            periods[-1] = (periods[-1][0], lastDay)

            t0 = int(firstDay / 7) * 7
            periods.append((firstDay, t0 + 6))
            tpstart = t0 + 7
            while tpstart <= 364:
                periods.append((tpstart, tpstart + 6))
                tpstart += 7
            periods[-1] = (periods[-1][0], 364)
            
        #now create the parabola or cosine function
        fscale = []

        if firstDay == 0 and lastDay == 364:

            for period in periods:
                dmax = 183 + 45 # shift of 30 days with respect to mid of year
                omega = 2 * pi / 365.
                
                davg = 0.5 * (period[0] + period[1])
                f = 1. + amplitude * cos(omega * (davg - dmax))
                fscale.append(f)
        else:
            if lastDay >= firstDay:
                dTotal = lastDay - firstDay + 1 + 2   #+2 -> the 0's are the day before first and the day after last
            else:
                dTotal = lastDay + (365 - firstDay + 1) + 2

            dmax = firstDay + 0.5 * dTotal # day of maximum

            for period in periods:
                davg = 0.5 * (period[0] + period[1])
                if davg < firstDay: davg += 365
                f = 1. - pow((davg - dmax) / (0.5 * dTotal), 2)
                fscale.append(f)
                                           
        fmax = max(fscale)
        fsum = 0
        nsum = 0
        if fmax != 0:           # normalise to maximum = 1
            for i, f in enumerate(fscale):
                fscale[i] /= fmax
            if fscale[i] > 0:
                fsum += fscale[i]
                nsum += 1

        # TODO: here some variable parameter should be implemented in order to adjust
        #       ratio between maximum and average power (-> can be in daily profile)
        return zip(periods, fscale)
            
    def _getTInAHU(self, firstDay, lastDay, tstart, tstop):
        """Calculates the average inlet temperature :

        Tin_avg = INT (TInDoor - Text) over heating or cooling period
        
        With float scale factors from the interval [0,1].
        
        :param firstDay: first day in the year (from 1 to 365) with demand > 0
        :param lastDay: last day in the year (from 1 to 365) with demand > 0
        :param tstart: starting time 0.0 - 24.0 (float)
        :param tstop: stop time 0.0 - 24.0 (float)
        
        :returns:  average temperature to be used in ventilation calculations
                 
        """
        if firstDay <= lastDay:
            intervals = [(firstDay, lastDay)]
        else:
            intervals = [(1, lastDay), (firstDay, 365)]

        TAmb = Status.WeatherData.Tdb
        dTsum = 0
        nctr = 0
        
        for interval in intervals:
            for it in range((interval[0] - 1) * 24, interval[1]*24):
                itDay = it % 24
                if itDay >= tstart - 0.5 and itDay <= tstop - 0.5:
                    dTsum += TAmb[it]
                    nctr += 1

        if nctr > 0:
            TAmbAvgInPeriod = dTsum / nctr
        else:
            TAmbAvgInPeriod = Status.TAmbAvg

        logWarning("Please re-confirm building parameters after each change in weather data")

        return TAmbAvgInPeriod

    def _buildCoolingSchedule(self, QProcessData_ID, buildingid, PId, ANo):
        """ Calculates and builds a Schedule for the building cooling process.
        
        :param QProcessData_ID: The process id of the new building cooling process.
        :param buildingid: The building id the processes is attached to.
        :param PId: The project id.
        :param ANo: The current alternative Proposal.

        """
        building = dbio.get("qbuildings", id=buildingid)
        if building.BuildAirCondPeriodStart is None: 
            raise ValueError("Missing start date.")
        if building.BuildAirCondPeriodStop is None: 
            raise ValueError("Missing end date.")
        BeginningofYear = date(building.BuildAirCondPeriodStart.year, 1, 1)
    
        
        self.coolingStart = (building.BuildAirCondPeriodStart - BeginningofYear).days % 365
        self.coolingEnd = (building.BuildAirCondPeriodStop - BeginningofYear).days % 365

        try:
            amplitude = 0.5 * Status.WeatherData.latitude / 45.
        except:
            amplitude = 0

        self._saveSchedule(self._buildYearlyHCSchedule(self.coolingStart, self.coolingEnd, amplitude),
                                 QProcessData_ID,
                                 'cooling-days-%s-%s' % (self.coolingStart, self.coolingEnd),
                                 building,
                                 respectHolidays=True)

        if self.coolingStart == 0 and self.coolingEnd == 364:
            self.fAvgC = 1.
        else:
            self.fAvgC = 2./3.
       
    def _buildHeatingSchedule(self, QProcessData_ID, buildingid, PId, ANo):
        """ Calculates and builds a Schedule for the building heating process.
        
        :param QProcessData_ID: The process id of the new building heating process.
        :param buildingid: The building id the processes is attached to.
        :param PId: The project id.
        :param ANo: The current alternative Proposal.

        """
        building = dbio.get("qbuildings", id=buildingid)
        # If no Start or end of a period exists, schedules can't be build
        if building.BuildHeatingPeriodStart == None: 
            raise ValueError("Missing start date.")
        if building.BuildHeatingPeriodStop == None:
            raise ValueError("Missing end date.")
        BeginningofYear = date(building.BuildHeatingPeriodStart.year, 1, 1)
        
        self.heatingStart = (building.BuildHeatingPeriodStart - BeginningofYear).days % 365
        self.heatingEnd = (building.BuildHeatingPeriodStop - BeginningofYear).days % 365
        
        try:
            amplitude = -0.5 * Status.WeatherData.latitude / 45.
        except:
            amplitude = 0

        self._saveSchedule(self._buildYearlyHCSchedule(self.heatingStart, self.heatingEnd, amplitude),
                                 QProcessData_ID,
                                 'heating-days-%s-%s' % (self.heatingStart, self.heatingEnd),
                                 building,
                                 respectHolidays=False)

        if self.heatingStart == 0 and self.heatingEnd == 364:
            self.fAvgH = 1.
        else:
            self.fAvgH = 2./3.
    
    def _buildDHWSchedule(self, QProcessData_ID, buildingid, PId, ANo):
        """ Calculates and builds a Schedule for the building DHW process.
        The results are directly saved into the database.
        
        :param QProcessData_ID: The process id of the new building DHW process.
        :param buildingid: The building id the processes is attached to.
        :param PId: The project id.
        :param ANo: The current alternative Proposal.
        :returns: 0 if the Schedule was build without errors. -1 otherwise.
        
        """
        building = dbio.get("qbuildings", id=buildingid)
        if self.buildHoursOccup == None:
            hours_DHW = 12.0
        else:
            hours_DHW = self.buildHoursOccup

        HolidayPeriods = self.getHolidayPeriodList(building)
        
        addPeriods = []
        
        if len(HolidayPeriods) == 0: 
            schedule_tuples = [((0, 364), 1.0)]
        else:
            schedule_tuples = []
            periodStartDay = 0
            periodEndDay = 364
            for i in xrange(0, len(HolidayPeriods), 2):
                periodEndDay = (HolidayPeriods[i] - date(HolidayPeriods[i].year,1,1)).days - 1
                if periodEndDay >= periodStartDay:
                    schedule_tuples.append(((periodStartDay,periodEndDay), 1.0))
                periodStartDay = (HolidayPeriods[i+1] - date(HolidayPeriods[i].year,1,1)).days + 1
            if periodStartDay <= 364:
                schedule_tuples.append(((periodStartDay,364), 1.0))

        self._saveSchedule(schedule_tuples,
                                 QProcessData_ID,
                                 'HW_%s_hours' % (hours_DHW),
                                 building,
                                 respectHolidays=True)

        self.fAvgH = 1.

        return
    
    def getHolidayPeriodList(self, building):
        """Return ordered list of HolidayPeriods.
        
        :param building: building pSQL object
        :returns: list of datetime elements.
  
        """
        HolidayPeriods = []
        HolidayPeriodsStart = []
        HolidayPeriodsStop = []
        if building.BuildHolidaysPeriodStart_1 != None and building.BuildHolidaysPeriodStop_1 != None:
            HolidayPeriodsStart.append(building.BuildHolidaysPeriodStart_1)
            HolidayPeriodsStop.append(building.BuildHolidaysPeriodStop_1)
        if building.BuildHolidaysPeriodStart_2 != None and building.BuildHolidaysPeriodStop_2 != None:
            HolidayPeriodsStart.append(building.BuildHolidaysPeriodStart_2)
            HolidayPeriodsStop.append(building.BuildHolidaysPeriodStop_2)
        if building.BuildHolidaysPeriodStart_3 != None and building.BuildHolidaysPeriodStop_3 != None:
            HolidayPeriodsStart.append(building.BuildHolidaysPeriodStart_3)
            HolidayPeriodsStop.append(building.BuildHolidaysPeriodStop_3)
            
##        HolidayPeriodsStart.sort()
##        HolidayPeriodsStop.sort()

        self.holidayPeriods = []
        for i in xrange(len(HolidayPeriodsStart)):
            HolidayPeriods.append(HolidayPeriodsStart[i])
            HolidayPeriods.append(HolidayPeriodsStop[i])
            self.holidayPeriods.append((HolidayPeriodsStart[i], HolidayPeriodsStop[i]))
        return HolidayPeriods
                
    def _saveSchedule(self, schedule_tuples, QProcessData_ID,
                            profilename, building, respectHolidays=False):
        """Take the given schedule_tuples and create an according 
        process schedule in the db and link it to the given process.
        
        :param schedule_tuples: a list of ((startday, endday), scale) 
        tuples with startday and endday from 0 to 364 and scale a float 
        in [0,1]
        :param QProcessData_ID: the ID of the appropriate process
        :param profilename: the name of the profile, where a "Building_" 
            is prefixed by this function
        :param building: building pSQL object, only used if respectHolidays
            is true
        :param respectHolidays: if True don't heat/cool during building 
            holidays
        
        """
        id = QProcessData_ID
            
        deleteProcessPeriodsFromDB(id)
        
        try:    
            periodSchedule = PeriodSchedule(building.BuildName + "_" + profilename)
            periodSchedule.loadFromDB(id)
        except ScheduleNotFoundError:
            logDebug("New " + str(profilename) + " Schedule build")
    
        try:
            fullName = "Building_" + profilename
            if str(profilename).startswith('cooling') or str(profilename).startswith('HW'):
                if str(profilename).startswith('cooling'):
                    occupation = self.buildHoursOccup
                    if occupation is None:
                        occupation = 12.0
                    fullName += "_%2.1fh"%occupation
                if self.nDaysPerWeek <> 7:
                    fullName += "_%sd"%(self.nDaysPerWeek)
                if self.fDaysPerWeek <> 1.0:
                    fullName += "_x%.4f"%(self.fDaysPerWeek)
                    
            profile = WeeklyProfile(fullName)
            profile.loadFromDB()
            
        except:
            logDebug("New " + str(profilename) + " Profile build")

        if profile.id is None: # build and save profile only if not yet existing in data base
            profile.setAllWeekdays(True)
            profile.clearIntervals()
            
            if str(profilename).startswith('cooling') or str(profilename).startswith('HW'):
                for i in range(self.nDaysPerWeek, 7):
                    profile.weekday[i] = False
                fscale = 100*self.fDaysPerWeek

                if self.buildHoursOccup is None:
                    profile.addInterval(datetime(2010, 1, 1, 0, 0).time(), datetime(2010, 1, 1, 23, 59).time(), fscale)
                else:
                    tstart = 12 - (self.buildHoursOccup / 2) 
                    starth = int(tstart)
                    startm = int((tstart - starth) * 60)
                    tstop = 12 + (self.buildHoursOccup / 2)
                    stoph = int(tstop)
                    stopm = int((tstop - stoph) * 60)
                    stoph = stoph % 24
                    if stoph == 0: # catch 24h case and reformulate it like the None case
                        stoph = 23
                        stopm = 59
                    profile.addInterval(datetime(2010, 1, 1, starth, startm).time(), datetime(2010, 1, 1, stoph, stopm).time(), fscale)
            else:
                profile.addInterval(datetime(2010, 1, 1, 0, 0).time(), datetime(2010, 1, 2, 23, 59).time(), 100)

            profile.saveToDB()
        
        # reduce scaling in the holiday to 0 for full weeks and proportionate
        # for weeks that lay partially in holidays.

##        if respectHolidays:
##            schedule_tuples = [[(start, stop), scale] for ((start, stop), scale) in schedule_tuples]
##            HolidayPeriods = self.getHolidayPeriodList(building)
##            for i in xrange(0, len(HolidayPeriods) - 1, 2):
##                startyear = HolidayPeriods[i].year
##                endyear = HolidayPeriods[i + 1].year
##                startday = HolidayPeriods[i] - date(startyear, 1, 1) 
##                endday = HolidayPeriods[i + 1] - date(endyear, 1, 1)
##                
##                for index in xrange(len((schedule_tuples))):
##                    period = schedule_tuples[index][0]
##                    if period[0] >= startday.days and period[0] <= endday.days:
##                        # find overlap with holiday
##                        overlap = endday.days - period[1]
##                        # if overlap is negative it is only partial
##                        # otherwise the whole period can be set to 0
##                        if overlap >= 0:
##                            factor = 0
##                        else:
##                            factor = -overlap / float(period[1] - period[0] + 1)
##                        schedule_tuples[index][1] *=  factor
##                    elif period[1] <= endday.days and period[1] >= startday.days:
##                        overlap = startday.days - period[0]
##                        if overlap <= 0:
##                            factor = 0
##                        else:
##                            factor = overlap / float(period[1] - period[0] + 1)
##                        schedule_tuples[index][1] *=  factor 

        periodSchedule.holidayScale = 1*(not respectHolidays)
        
        hour = timedelta(hours=1)
        currentDate = datetime(2007, 1, 1, 0, 0)
        addPeriods = []
        
        for ((start, stop), scale) in schedule_tuples:
            try:
                assert(stop >= start)
            except:
                print "Buildings._saveSchedule. WARNING start %s stop %s"%(start, stop)
                
            pp = ProcessPeriod()
            pp.start = DOYToDate(start)
            pp.stop = DOYToDate(stop)
            pp.scale = scale*100
            pp.profiles = [profile]

            periodSchedule.periodIntervals.append(pp)
            
            
        periodSchedule.saveToDB(QProcessData_ID, True, saveAll=True)

def checkBuildings():
    """
    Helper fucntion for debugging
    Checks consistency of entries in qbuildings and qprocessdata
    """
    PId = Status.PId
    ANo = Status.ANo

    print "CheckBuildings: project %s.%s"%(PId, ANo)
    
    buildings = dbio.select("qbuildings", Questionnaire_id=PId, AlternativeProposalNo=ANo)
    processes = dbio.select("qprocessdata", Questionnaire_id=PId, AlternativeProposalNo=ANo)
    processes = [proc for proc in processes if proc.QBuildings_ID is not None]

    buildingIds = [bui.id for bui in buildings]
    
    for building in buildings:
        nonZeroDemand = 3 - 1*(building.BuildAnnualHeating==0) + \
                        1*(building.BuildAnnualAirCond==0) + \
                        1*(building.BuildDailyDHW==0)

        nProc = len([proc for proc in processes if proc.QBuildings_ID==building.id])
        print "checkBuildings: Building %s (%s) - number of processes %s (%s) -> %s"%(building.BuildName, building.id, nProc, nonZeroDemand, (nProc==nonZeroDemand))

    strangeProcesses = [proc for proc in processes if proc.QBuildings_ID not in buildingIds]

    if len(strangeProcesses) > 0:
        print "checkBuildings: strange Processes"
        for proc in strangeProcesses:
            print "Process %s (%s): buiId = %s [False]"%(proc.Process, proc.id, proc.QBuildings_ID)
                      
    
