# -*- coding: utf-8 -*-
#==============================================================================                 
#
#	 E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    CCheck (Consistency Check)
#            
#------------------------------------------------------------------------------
#            
#    Short description:
#    
#    Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================                 
from math import fabs, log, exp
from constants import *                
from modules.messageLogger import *
from GUI.GUITools import *     #needed for function check
from ccTracker import estimateTracker, screen, conflict, cycle
import numpy as np

global DEBUG
EPSILON = 1.e-10     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"
MINIMUM_VALUE = 1.e-10
DEFAULT_SQERR = 1.e-6 # default value for sqerr assigned to questionnaire values (until V2.1 default had been set to 1e-4)
NUMERIC_ERR = 1.e-10 # accuracy of numeric calculations
CONFLICT_ERR = 1.e-6
MAX_SQERR = 0.09     # critical square error for screening
CONFIDENCE = 4.0    # maximum relation between statistical error and abs.min/max
MAXTRACKENTRIES = 10 # maximum number of variables to be tracked
    
TEST = False
TESTCASE = 3


#------------------------------------------------------------------------------
class CCPar():
#------------------------------------------------------------------------------
#   Class for grouping value and attributes of a parameter
#------------------------------------------------------------------------------

    def __init__(self, name, parType=None, priority=3, unitType=None):
        """Create new CCPar (consistency check parameter).
        
        :param name: The (human readable) name of the parameter
        :param parType: Can be one of "T": a temperature of a heating
        process, "TC": temperature of a cooling process, "DT": a 
        positive difference, "DTmN": a difference that may be negative,
        "X": a fraction, or "S":  a signed float. Each type has its own
        maximal and minimal valid values. 
        :param priority: a hint for the viewer whether this parameter
        should be considered of high importance, e.g. for display of
        parameters with high deviation values. Priority is a integer 
        taken from 1 (high), 2 (medium) or 3 (low).
        
        """
        # TODO: what means priority
        self.name = name
        self.priority = priority
        self.parType = parType
        self.unitType = None     # no default unit type
        self.val = None
        self.userDefined = False # attribute originally set by setValue, but inherited if a direct function of other user defined values
        self.userInput = None # only set in setValue if a real user input, otherwise kept None
        self.sqerr = INFINITE
        self.sqdev = INFINITE
        self.dev = pow(INFINITE, 0.5)

        self.valMin = 0         #default: non-negative values !!!!
        if parType in ["T","TC","DTmN"]:    #check if there's really a need / distinction for three types ?
            self.valMax = MAXTEMP
            self.valMin = MINTEMP
            if parType == "DTmN":
                self.unitType = "TEMPERATUREDIFFERENCE"
            else:
                self.unitType = "TEMPERATURE"
        elif parType == "DT":
            self.valMax = (MAXTEMP - MINTEMP)
            self.unitType = "TEMPERATUREDIFFERENCE"
        elif parType == "X":
            self.valMax = 1.0
            self.unitType = "FRACTION"

        elif parType == "S":
            self.valMin = -INFINITE
            self.valMax = INFINITE
        else:
            self.valMax = INFINITE

        self.track = []             #new version: own name in track only if user input !!! (-> setValue)
        self.trackDetails = None
##        self.track = [self.name]  #old version of tracking. each parameter has its own name in track
        self.stack = []

        if unitType is not None:
            self.unitType = unitType

        self.screened = False

##        print "CCHECK TEST TEST - cycling paramteres initialised: %s"%self.name
##        self.dif = 0.0
##        self.cyclingIndicator = 1.0
##        self.cyclingLagged = 1.0

    def __repr__(self):
        """Print the current name, val, squerr, valMin, valMax and for
        each element on the stack.
        """
        err = pow(self.sqerr, 0.5)
        dev = pow(self.sqdev, 0.5)
        attr = "U" if self.userDefined else ""
        message =  "%s = %s (err: %s dev: %s) [%s, %s]%s"%(self.name, self.val, err, dev, self.valMin, self.valMax, attr)
        message += " p:%s uT:%s t:%s"%(self.parType, self.unitType, self.track)
        if self.stack != []:
            message += "\nstack:"
            for item in self.stack:
                message += "\n"+ item.__repr__()
            message += "---"
        return message
    
#------------------------------------------------------------------------------
    def calcDev(self):
#------------------------------------------------------------------------------
#   calculates sqDev from sqErr and value
#------------------------------------------------------------------------------
        if (self.val > NUMERIC_ERR or self.val < -NUMERIC_ERR) and self.val is not None:
            if self.sqerr == INFINITE:
                self.sqdev = INFINITE
            else:
                self.sqdev = self.sqerr * pow(self.val, 2.)
        elif self.sqerr == 0:
            self.sqdev = 0
        else:
            try: #note: try-except necessary if valMax / valMin are very large numbers so that pow(x,2) goes beyond max size of float numbers
                self.sqdev = pow(max(self.valMax - self.valMin, 0), 2.)
            except:
                self.sqdev = INFINITE
                
        self.dev = pow(self.sqdev, 0.5)
        return self.sqdev

#------------------------------------------------------------------------------
    def calcErr(self):
#------------------------------------------------------------------------------
#   calculates sqErr from sqDev and value
#------------------------------------------------------------------------------
        if (self.val > NUMERIC_ERR or self.val < -NUMERIC_ERR) and self.val is not None:
            if self.sqdev >= 1.e-20 * INFINITE:
                self.sqerr = INFINITE
            else:
                self.sqerr = self.sqdev / pow(self.val, 2.)
        elif self.sqdev == 0:
            self.sqerr = 0
        else:
            self.sqerr = INFINITE
        return self.sqerr

#------------------------------------------------------------------------------
    def invSqErr(self):
#------------------------------------------------------------------------------
#   calculates the square error of 1/x
#   sqerr(1/x) = sqerr(x) for sqerr << 1
#------------------------------------------------------------------------------
        if self.sqerr >= pow(CONFIDENCE, -2.0): # value 0.0 can not be excluded
            return INFINITE
        else:
            err = pow(self.sqerr, 0.5)
            return pow(err / (1 - err), 2.0)

    def update(self, new):
        """Set all numeric values to the ones given in new."""
        self.val = new.val
            
        self.sqerr = new.sqerr
        self.sqdev = new.sqdev
        self.dev = new.dev
        self.valMin = new.valMin
        self.valMax = new.valMax

        newtrack = new.track
        self.track = []
        self.track.extend(newtrack)
        if new.trackDetails is None:
            self.trackDetails = None
        else:
            self.trackDetails = new.trackDetails[:]

        self.userDefined = new.userDefined
        self.userInput = new.userInput

        self.parType = new.parType
        self.unitType = new.unitType

    def copy(self):
        "creates a copy of the current value (without stack !)"
        new = CCPar(self.name)
        new.update(self)
        return new

#------------------------------------------------------------------------------
    def cleanUp(self):
#------------------------------------------------------------------------------
#   cleans the track from double entries
#------------------------------------------------------------------------------

        cleanTrack = []
        for entry in self.track:            
            if (entry not in cleanTrack) or (entry in ["+","-","*","/"]):
                if len(cleanTrack) < MAXTRACKENTRIES:
                    cleanTrack.append(entry)
                elif (len(cleanTrack) == MAXTRACKENTRIES):
                    cleanTrack.append("etc.")
                    break                
        self.track = cleanTrack
        
    def simplifyTrack(self):
        """ sets the own name as track and deletes the history """

        if self.track is None or len(self.track) == 0:
            return

        # detect if its already a simplified track (contains only #-entries)
        isSimpleTrack = False
        for elem in self.track:
            if len(elem) > 0:
                if elem[0] == "#":
                    isSimpleTrack = True
                else:
                    isSimpleTrack = False # if only one entry is not a #-entry, then consider as a detailed track
                    break
                    
        # otherwise convert into a simplified one
        if not isSimpleTrack:
            self.trackDetails = self.track[:]
            self.track = ["#"+self.name]
        
        
    def setValue(self, val, err=DEFAULT_SQERR, default=None, setTrack = True):
        """Set the object with values (e.g. from data questionnaire). 
        If now error is given the error values are set to default values. 
        If val is None the given default is used. If it is None too
        the value in the middle of the scale is chosen with an error 
        large enough to span the whole scale.

        :params setTrack: True - fill track with own parameter name.
        
        """
        if val is None: # check for default value
            val = default
        self.val = val
        if val is None: # if val is still None, act accordingly
            self.sqerr = INFINITE

# special case EXACT ZERO
        elif val == 0:
            self.sqerr = 0
            self.userDefined = True
            self.userInput = 0

# special case for fractions (values between 0 and 1): -> set to EXACT 1
# is required in checkPipe for setting feed-up flow to EXACTLY 0

        elif self.parType == "X" and self.val > 1.0 - NUMERIC_ERR:
            self.val = 1.0
            self.sqerr = 0
            self.userDefined = True
            self.userInput = self.val
        elif self.parType in ("T", "TC") and err == DEFAULT_SQERR:
            self.sqdev = 0.0625
            self.calcErr()
            self.userDefined = True
            self.userInput = self.val
        else:
            self.sqerr = err
            if err <= DEFAULT_SQERR:
                self.userDefined = True
                self.userInput = self.val

        self.constrain(updateLimits=True)

        if val is not None:
            if setTrack:
                self.track = [self.name]    #new version of track: parameter gets its name into track if a user input !!!
            else:
                self.track = []
                
#------------------------------------------------------------------------------
    def constrain(self, updateLimits=False):
#------------------------------------------------------------------------------

        maxAbs = max(abs(self.valMin), abs(self.valMax))
        if maxAbs < NUMERIC_ERR:   # avoid errors due to special case where due to numeric errors valMin = valMax = 1.e-12 or something very small.
            self.valMin = 0
            self.valMax = 0
            self.val = 0
            self.sqerr = 0
        else:
            maxDif = maxAbs * NUMERIC_ERR 
            if (self.valMin - self.valMax > maxDif):

                print "======================================================"
                print "======================================================"
                print "%s.CONSTRAIN: SEVERE ERROR - MIN > MAX VALUE !!!! " % self.name
                self.show()
                print "======================================================"
                print "======================================================"

                logWarning("Minimum value %s > maximum value %s for parameter %s. Check your input data"%(self.valMin, self.valMax, self.name))

                newMax = self.valMin
                newMin = self.valMax
                self.valMin = newMin
                self.valMax = newMax
##                self.track.extend(["CONSTRAINED"])
##                self.cleanUp()

#.............................................................................
# general control: min/max values between [0,INFINITE] and min < max

        if self.parType not in ["S","DTmN","T","TC"]: self.valMin = max(self.valMin, 0)
        self.valMax = min(self.valMax, INFINITE)
            
        if self.val is not None:

            if (abs(self.val) > MINIMUM_VALUE):

#.............................................................................
# first set absolute constraint of the value to reasonable limits around its
# actual value-in function of the specified error margins

                if self.sqerr < 1:  #CHECK: new feature (2011-02-21), before this correction was done for all (INFINITE instead of 1). Test if this gives problems at other places.
                    errMax = CONFIDENCE * pow(self.sqerr, 0.5)

                    if self.val > 0:
                        newValMin = self.val * (1 - errMax)
                        newValMax = self.val * (1 + errMax)

                    else:
                        newValMin = self.val * (1 + errMax)
                        newValMax = self.val * (1 - errMax)

                    epsDif1 = max(abs(newValMin),abs(self.valMax))*NUMERIC_ERR
                    epsDif2 = max(abs(newValMax),abs(self.valMin))*NUMERIC_ERR
                    if (newValMin - self.valMax) > epsDif1 or (newValMax - self.valMin) < - epsDif2:
##                        self.track.extend(["CONSTRAINED"])
##                        self.cleanUp()
                        
                        print "======================================================"
                        print "======================================================"
                        print "%s.CONSTRAIN: SEVERE ERROR - VALUE OUT OF [MIN, MAX] BOUNDARIES: [%s, %s]!!!! " %(self.name, newValMin, newValMax)
                        self.show()
                        print "======================================================"
                        print "======================================================"

                    if updateLimits:
                        self.valMin = max(self.valMin, newValMin)
                        self.valMin = min(self.valMin, self.valMax) #new min can not be higher as max
                        
                        self.valMax = min(self.valMax, newValMax)
                        self.valMax = max(self.valMin, self.valMax) #new max can not be lower than min

#.............................................................................
# special case: exact zero

            elif self.val == 0 and self.sqerr == 0:
                self.valMin = 0.0
                self.valMax = 0.0

#.............................................................................
# special case: unallowed negative values

            else:
                self.val = 0.0
                self.sqerr = INFINITE

#.............................................................................
# and finally constrain the value itself within the allowed limits

            self.sqdev = self.calcDev() #maintain sqDev as determinant error margin
            self.val = min(self.val, self.valMax)
            self.val = max(self.val, self.valMin)
            self.calcErr()

#.............................................................................
# consider that the error can not be larger than [maxVal - minVal]

            if self.val != 0:
                
                try:    #helps to avoid crash for very large errors
##                    newErr = max(pow((self.valMax - self.val) / self.val, 2.0), \
##                                 pow((self.val - self.valMin) / self.val, 2.0))
                    #TODO: 13.11.2014: the previous version gives problems of convergence in Guide41PS
                    # with the more conservative estimate below the correct values of TExhaustGas are obtained
                    # Requires further testing and consistent use throughout Ccheck
                    newErr = pow((self.valMax - self.valMin) / self.val, 2.0)
                    self.sqerr = min(self.sqerr, newErr)  #XXX Security feature !!!
                except:
                    pass

                self.sqerr = min(self.sqerr, INFINITE)
                
            elif self.val == 0 and self.valMax == 0 and self.valMin == 0:    #set to exact zero
                self.sqerr = 0

#.............................................................................
# check for all values larger than 0 !!!
# if constrain in between min/max works, and min> 0, the following should be unnecessary

            if self.val < 0 and self.parType not in ["S","DTmN","T","TC"]:
                print "======================================================"
                print "======================================================"
                print "%s.CONSTRAIN: SEVERE ERROR - VALUE < 0 !!!! " % self.name
                self.show()
                print "======================================================"
                print "======================================================"

        elif self.valMax < 1.e-10 * INFINITE:
            self.val = 0.5 * (self.valMin + self.valMax)
            self.dev = pow((self.valMax - self.valMin), 2.0)
                
    def show(self, showAttributes=False):
        """Print the current name, val, squerr, valMin, valMax and for
        each element on the stack.
        """
        err = pow(self.sqerr, 0.5)
        dev = pow(self.sqdev, 0.5)
        attr = "U" if self.userDefined else ""
        message =  "%s = %s (err: %s dev: %s) [%s, %s]%s"%(self.name, self.val, err, dev, self.valMin, self.valMax, attr)
        if showAttributes: message += " p:%s uT:%s t:%s"%(self.parType, self.unitType, self.track)
        print message
        if self.stack != []:
            print "stack:"
            for item in self.stack:
                item.show(showAttributes=showAttributes)
            print "---"

#------------------------------------------------------------------------------
    def screen(self):
#------------------------------------------------------------------------------
#   creates a register entry if the parameter is None or with a high error
#------------------------------------------------------------------------------
        screen.nScreened += 1

        if (self.val == None) or (self.sqerr > MAX_SQERR):
            screen.screenList.append([self.name, self.val, self.sqerr, screen.dataGroup, self.priority, self.unitType, screen.objectGroup])
            self.screened = True
        else:
            self.screened = False
#        if (self.val == None):
#            CCScreen.screenList.append([self.name,"---","---",screen.dataGroup,self.priority])
#        elif (self.sqerr > MAX_SQERR):
#            if self.sqerr < 1.0:
#                err = pow(self.sqerr,0.5)
#                CCScreen.screenList.append([self.name,'%9.2f'%self.val,'%5.2f'%err,screen.dataGroup,self.priority])
#            CCScreen.screenList.append([self.name,self.val,err,screen.dataGroup,self.priority])
#            else:
#                CCScreen.screenList.append([self.name,'%9.2f'%self.val,'>100.00',screen.dataGroup,self.priority])
                
        if DEBUG in ["ALL", "BASIC"]:
            print "CCPar (screen): screening parameter"
            self.show()
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def setEstimate(self, val, newSqErr=MAX_SQERR, limits=None, record=True):
#------------------------------------------------------------------------------
#   creates a register entry if the parameter is None or with a high error
#------------------------------------------------------------------------------
        changed = False
        if (self.val is None) or (self.sqerr > MAX_SQERR):

            if DEBUG in ["ALL", "BASIC"]:
                if val < self.valMin or val > self.valMax:
                    self.show()
                    print "ERROR: estimate trying to set a value (%s) out of limits ([%s, %s])"%(val, self.valMin, self.valMax)  
            changed = changed or ((self.val is None) or abs(self.val - val) > EPSILON)
            self.val = val
            if limits is not None:
                (newMin, newMax) = limits

# update limits only if not in conflict with old limits !!!!

                if DEBUG in ["ALL", "BASIC"]:
                    if newMin > self.valMax or newMax < self.valMin:
                        self.show()
                        print "ERROR: estimate trying to set new limits ([%s, %s]) in conflict with current limits ([%s, %s])"%(newMin, newMax, self.valMin, self.valMax)
                        
                newMin = min(newMin, self.valMax)
                newMax = max(newMax, self.valMin)

                changed = changed or newMin > self.valMin or newMax < self.valMax
                
                self.valMin = max(newMin, self.valMin)
                self.valMax = min(newMax, self.valMax)
                
#            if newSqErr is not None: self.sqerr = min(self.sqerr,newSqErr)
# HS 20090611 -> default error margin only applied if no limits are given !!!
            elif newSqErr is not None:
                changed = changed or newSqerr < self.sqErr
                self.sqerr = min(self.sqerr, newSqErr)
                if abs(self.val) < EPSILON: # if estimate is set to 0 with relative error margin -> set exactly to 0
                    self.val = 0.0
                    self.valMin = 0.0
                    self.valMax = 0.0
                    
            self.constrain()

            if DEBUG in ["ALL", "BASIC"]:
                print "CCPar (setEstimate): paramter estimated to"
                self.show()

            if changed and record:
                logMessage("Parameter %s estimated to %s [%s,%s]" % 
                           (self.name + "[" + str(estimateTracker.dataGroup) + "]",
                            self.val, self.valMin, self.valMax))

                estimateTracker.add(self)
            self.track = [self.name]
            self.userDefined = True
            
    def setZero(self):
        """Assigns exact zero to the parameter
        
        """
        self.val = 0
        self.valMin = 0
        self.valMax = 0
        self.sqerr = 0
        self.sqdev = 0

    def push(self, new=None):
        """Add the new estimate to the stack. 
        
        If new is given, add the new CCPar object to the stack. If None
        is given add the numeric values of self to the stack.
        
        :param new: the new numeric values for this CCPar object
        
        """
        currentEstimate = CCPar(self.name + "[%2d]" % len(self.stack))
        if new is None:
            currentEstimate.update(self)
            self.stack.insert(0, currentEstimate) # move at first place of stack (needed for bestOfRow/meanOfRow)
        else:
            currentEstimate.update(new)
            if currentEstimate.unitType is None:
                currentEstimate.unitType = self.unitType
            self.stack.append(currentEstimate)
            


        ####DEBUG DEBUG DEBUG
##        if len(self.stack) > 10:
##            print "WARNING: stack size of %s > 10"%self.name


    def check(self, alpha=1.0):
##    def check(self):
        """Add current estimate to the stack and calculate a new mean
        which becomes the new current estimate.
        
        The stack is emptied after the calculation.
        
        """
        from ccCheckers import bestOfRow, meanOfRow, CHECKMODE
##        :params alpha: subrelaxation coefficient. 1: no subrelaxation < 1: subrelax
##                    (uses the value at row[0] as the old estimate

        self.push() # now everything is in the stack

        #TODO: 26.2.2014 -> limit sqerr < 1.0 introduced. TESTING required if ccheck continues being stable with this modification
        valList = [entry.val for entry in self.stack if entry.sqerr < 1.0]
        bal = cycle.checkBalance(valList,name=self.name, parType = self.parType, unitType = self.unitType) #TODO: document what that is doing
        
        if DEBUG in ["ALL", "MAIN", "CHECK"]:
            print "CCPar.check(before...)________________________________"
            for entry in self.stack:
                entry.show()

        y = bestOfRow(self.stack)
        
##        if CHECKMODE == "BEST":
##            y = bestOfRow(self.stack)
##        else:
##            y = meanOfRow(self.stack)

## SUBRELAXATION BLOCK. CURRENTLY NOT USED (V2.5ff)
##            if alpha < 1.0 and y.val is not None and self.val is not None and not y.userDefined: # do not subrelax for user defined values
##                y.val = alpha*y.val + (1.0 - alpha)*self.val
##                y.sqdev = alpha*y.sqdev + (1.0 - alpha)*self.sqdev
##                y.val = min(y.val, y.valMax)
##                y.val = max(y.val, y.valMin)
##                y.calcErr()



##        #TO DO: TEST TEST this is test code only for development purposes. Should be commented out in release version
##
##        if self.val is not None and y.val is not None:
##            dif = self.val - y.val
##        else:
##            dif = 1.0
##
##        if self.dif <> 0:
##            self.cyclingIndicator = dif / self.dif
##            self.cyclingLagged += 0.1*(self.cyclingIndicator - self.cyclingLagged)
##        else:
##            self.cyclingIndicator = 1.0
##            
##        self.dif = dif
##        print "TEST TEST cycling in ccheck: dif %s par %s self.name cycling %s cyclingLagged %s"%(self.dif,
##                                                                                                  self.name,
##                                                                                                  self.cyclingIndicator,
##                                                                                                  self.cyclingLagged)
        self.update(y)
        self.stack = []

        if DEBUG in ["ALL", "MAIN", "CHECK"]:
            print "CCPar.check(after...)________________________________"
            self.show()
            print "________________________________Mode = %s Balance = %s"%(CHECKMODE,bal)


def CCRow(name, m, parType=None, unitType = None):
    """Return a row of m CCPars objects."""
    row = [CCPar(name + "[" + str(i + 1) + "]", parType=parType, unitType=unitType) for i in range(m)]
    return row

def CCMatrix(name, ncol, nrow, unitType = None):

    matrix = []
    for j in range(nrow):
        matrix.append(CCRow(name + "[" + str(j + 1) + "]", ncol, unitType = unitType))
    return matrix


def CCOne():
    """Return a CCPar object with the precise Value of 1.0 and no 
    margin of error.
    
    """
    one = CCPar("1")
    one.val = 1.0
    one.sqerr = 0
    one.sqdev = 0
    one.dev = 0
    one.valMin = 1.0
    one.valMax = 1.0
    one.userDefined = True
    
    return one
