# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CCSchedules
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	CCSchedule class (for CheckSchedules)
#
#==============================================================================
#
#	(C) copyleft E4-Experts GmbH, Berlin, Germany
#	www.energyxperts.net / info@energyxperts.net
#
#============================================================================== 				

import numpy as np
from GUI.status import Status
from constants import EPSILON

global idCtr
idCtr = 0

class CCSchedule():
    """ Class defining an effective schedule with uncertainty

        s(t) = c*s_d(t) + (1-c)*umask*r

        where r is a random schedule with sum(r) = 1
        
    """

    def __init__(self, cc, schedule=None):
        self.cc = cc    # ccPar in ModuleCC which corresponds to this schedule
        self.setSchedule(schedule)

        self.err = False    # set to true, if there is a value, but with a high associated error (close to INFINITE)

        self.updateValue()

        global idCtr
        self.id = idCtr
        idCtr += 1

        self.changedInLoop = False

    def setSchedule(self, schedule):
        self.schedule = schedule if schedule is not None else np.zeros(Status.Nt)
        self.c = 1.0 if schedule is not None else 0.0          # fraction of the part determined by schedule
        
    def updateValue(self):
        """ updates the value from the associated cc-parameter
        """
        
        cc = self.cc
        if cc is None:
            self.val = 0.0
        else:
            if cc.val is None:
                self.val = 0.5 * (cc.valMin + cc.valMax)
            else:
                self.val = cc.val

        self.changed = True     # helper flag for tracking changes in iterative calculations

        # TODO: this (umask) is currently not used. thought for situations where the schedule is unknown, but
        #       where for certain time intervals f=0 is sure.
##        self.umask = np.zeros(Status.Nt) # mask of random schedule. where umask = 0 the schedule is 0.
##        if not schedule: self.umask += 1

    def __repr__(self):
        name = "---" if self.cc is None else self.cc.name
        maxval = self.val*max(self.schedule) if (self.schedule is not None and self.val is not None) else None
        hop = sum(self.schedule>EPSILON) if self.schedule is not None else None
        return "CCSchedule (%s): %s val %s (max=%s) hop %s c %s "%(self.id,
                                                                   name,
                                                              self.val,
                                                              maxval,
                                                              hop,
                                                              self.c)
        
    def copySchedule(self, origin):
        """ copies the schedule parameters from origin, leaving val unchanged
        """

        self.schedule = origin.schedule
        self.c = origin.c
        self.umask = origin.umask
        self.val = origin.val #TODO: eliminate this row. should be substituted by an import of the actual CC result

def scheduleRowSum(s, m, svect, onChange=False):
    """
    m: vector - if 0 -> exclude elements for this index
    """

    if onChange and not any([si.changed for si in svect]):
        return

    mask = [(m is None or mi is None or abs(mi) > EPSILON) for mi in m]
    w = np.zeros(len(svect))
    s.schedule = np.zeros(Status.Nt)
    s.c = 0
    s.val = sum([si.val for i, si in enumerate(svect) if mask[i]])

    for i, si in enumerate(svect):
        if mask[i]:
            w = si.val/s.val if s.val > 0 else 1./len(svect)
            s.schedule += w*si.schedule
            s.c += w*si.c

            s.err = any([si.err for si in svect])

    s.changedInLoop = True
    s.changed = True


def scheduleRowSumC(s, f, svect, onChange=False):

    if onChange and not any([si.changed for si in svect]):
        return

    try:
        sival = np.nan_to_num(np.array([(si.val if si.val is not None else 0.0 ) for si in svect ]))
    except:
        sival = np.zeros(len(svect))
        
    sval = np.dot(f, sival)
##    s.val = np.sum(sival)

    if sval <> 0:
        w = f*sival/sval
    else:
        w = f / np.sum(f) if np.sum(f) > 0 else f

##    s.schedule = np.dot(w, np.array([si.schedule if si.schedule is not None else np.zeros(8760) for si in svect ]))
    schedule = np.zeros(8760)
    c = 0
    for i, si in enumerate(svect):
        if w[i] <> 0:
            schedule += w[i]*si.schedule
            c += w[i]*si.c

    s.schedule = schedule
    s.val = sval
    s.c = c        
##    s.c = np.dot(w, np.array([si.c for si in svect]))
    s.err = any([si.err for si in svect])

    s.changedInLoop = True
    s.changed = True

def scheduleDiff(s, s1, s2, onChange=False):

    if onChange and not s1.changed and not s2.changed:
        return
    
    # calculate weight of schedules
    s.val = s1.val - s2.val

    w1 = s1.val/s.val if abs(s.val) > EPSILON else 1.0
    w2 = s2.val/s.val if abs(s.val) > EPSILON else 1.0

    sumWeight = w1*s1.c - w2*s2.c
    s.schedule = (w1*s1.c*s1.schedule - w2*s2.c*s2.schedule)/sumWeight if sumWeight > 0 else s1.schedule - s2.schedule

    # noise (1-c) sums up !
    s.c = max(1. - w1*(1.-s1.c) - w2*(1. - s2.c), 0.)

    s.err = s1.err or s2.err

    s.changedInLoop = True
    s.changed = True

    
def scheduleProd(s, a, s1, onChange=False):

    if onChange and not s1.changed:
        return
    
    s.val = a*s1.val
    s.schedule = s1.schedule
    s.c = s1.c

    s.err = s1.err

    s.changedInLoop = True
    s.changed = True




