# -*- coding: utf-8 -*-
#==============================================================================                 
#
#	 E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    CCheck (Consistency Check)
#            
#------------------------------------------------------------------------------
#            
#    Short description:
#    
#    Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================                 
from math import fabs, log, exp
from constants import *                
from modules.messageLogger import *
from GUI.GUITools import *     #needed for function check
import numpy as np

global DEBUG
EPSILON = 1.e-10     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"
MINIMUM_VALUE = 1.e-10
DEFAULT_SQERR = 1.e-6 # default value for sqerr assigned to questionnaire values (until V2.1 default had been set to 1e-4)
NUMERIC_ERR = 1.e-10 # accuracy of numeric calculations
CONFLICT_ERR = 1.e-6
MAX_SQERR = 0.09     # critical square error for screening
CONFIDENCE = 4.0    # maximum relation between statistical error and abs.min/max
CHECKMODE = "BEST"  # MEAN or BEST. MEAN recommened as BEST doesn't give consistent results
MAXTRACKENTRIES = 10 # maximum number of variables to be tracked

def setCheckMode(mode):
    global CHECKMODE
    CHECKMODE = mode
    
TEST = False
TESTCASE = 3


class CCEstimate():
    """
    module for screening of errors and missing data
    """

    estimateList = []
    nEstimated = 0
    
    def __init__(self):
        self.reset()

    def setDataGroup(self, name, no):
        self.dataGroup = str(no)

    def reset(self):
        CCEstimate.estimateList = []
        CCEstimate.nEstimated = 0

    def show(self):
        print "CCEstimate: %s parameters estimated" % len(CCEstimate.estimateList)
        for i in range(len(CCEstimate.estimateList)):
            print i + 1, CCEstimate.estimateList[i]

    def add(self, par):
        #TODO: avoid counting several estimates for the same parameters, keep only the last one !
        for i in range(len(CCEstimate.estimateList)):
            item = CCEstimate.estimateList[i]
            if item[0] == par.name and item[4] == self.dataGroup:
                CCEstimate.estimateList.remove(item)
                break
        CCEstimate.estimateList.append([par.name,par.val,par.valMin,par.valMax,self.dataGroup])
        CCEstimate.nEstimated = len(CCEstimate.estimateList)


#------------------------------------------------------------------------------
class CCScreen():
#------------------------------------------------------------------------------
#   module for screening of errors and missing data
#------------------------------------------------------------------------------

    screenList = []
    nScreened = 0
    
    def __init__(self):
        self.dataGroup = None
        self.reset()

    def setDataGroup(self, name, no):
        self.dataGroup = str(no)
        self.objectGroup = name

    def reset(self):
        CCScreen.screenList = []
        CCScreen.nScreened = 0

    def show(self):
        print "CCScreen: %s parameters screened" % len(CCScreen.screenList)
        for i in range(len(CCScreen.screenList)):
            print i + 1, CCScreen.screenList[i]

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
class CCConflict():
#------------------------------------------------------------------------------
#   module for screening of conflicts between data
#------------------------------------------------------------------------------

    conflictList = []
    nConflictChecked = 0
    nConflicts = 0
    
    def __init__(self):
        self.reset()
        self.dataGroup = None

    def reset(self):
        CCConflict.conflictList = []
        CCConflict.nConflictChecked = 0
        CCConflict.nConflicts = 0

    def setDataGroup(self, name, no):
        self.dataGroup = name + "[" + str(no) + "]"

    def show(self):
        print "CCConflict: %s parameters screened for possible conflicts" % len(CCConflict.conflictList)
        for i in range(len(CCConflict.conflictList)):
            print i + 1, CCConflict.conflictList[i]

    def screen(self, y1, y2, dif, maxDif):
        y1err = pow(y1.sqerr, 0.5)
        y2err = pow(y2.sqerr, 0.5)
        
        row = [dif, maxDif, y1.name, y1.track, y1.val, y1err, y2.name, y2.track, y2.val, y2err, self.dataGroup, y1.unitType]
        CCConflict.conflictList.append(row)
        CCConflict.nConflicts += 1

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
class Cycle():
#------------------------------------------------------------------------------
#   class for convergence testing in check cycles
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def __init__(self, itMax, maxDifference, minImprovement, itMaxImprove):
#------------------------------------------------------------------------------
        self.itMax = itMax                     #absolute maximum number of iterations
        self.maxDifference = maxDifference      #maximum difference in precision below which cycle stops
        self.minImprovement = minImprovement    #min. improvement required below which cycle stops
        self.itMaxImprove = itMaxImprove        #maximum number of iterations for trying if there's still
                                                #improvement possible

        self.balanceSum = 0
        self.balanceMax = 0
        self.nBalanceChecks = 0

        self.maxEnergy = 1000.  # default for energy reference 
        
#------------------------------------------------------------------------------
    def initCycle(self):
#------------------------------------------------------------------------------
        self.itCtr = 0
        self.itImproveCtr = 0
        self.lastDiff = INFINITE
        self.diff = 0.5 * INFINITE
        self.nChecks = 0

#------------------------------------------------------------------------------
    def count(self, diff):
#------------------------------------------------------------------------------
        self.diff += abs(diff)
        self.nChecks += 1
        
#------------------------------------------------------------------------------
    def checkBalance(self, values, name = None, parType = None, unitType = None):

        """ called from the check function of individual parameters
        """
        
        minVal = INFINITE
        maxVal = -INFINITE
        values = [value for value in values if value is not None]
        minVal = min(values + [minVal])
        maxVal = max(values + [maxVal])
        maxAbs = max(abs(minVal),abs(maxVal))

        if parType == "X":
            maxAbs = max(1.0,maxAbs) # set relative error = absolute error for fractions of 1. important for checking of steam fractions !

        if unitType == "ENERGY":
            maxAbs = max(0.1*self.maxEnergy,maxAbs) # do not count very small energy errors, even if they are big relatively speaking

        if len(values) == 0 or maxAbs < NUMERIC_ERR: #All nones are not a balance error ! -> balance = 0 if len(values) = 0
            balance = 0
        else:
            balance = min(abs(maxVal - minVal) / maxAbs, 1.0) #> 1 and Nones count the same ...

        if unitType == "ENERGY" and balance < 1.e-2:
            self.maxEnergy = max(self.maxEnergy,maxAbs) # auto-adaptation to maximum energy flow in the system
            
        self.balanceSum += balance
        if balance > self.balanceMax:
            self.blackSheep = name

        self.balanceMax = max(balance, self.balanceMax)
        self.nBalanceChecks += 1
        return balance

    def checkBalanceSchedules(self, residual):
        balance = min(residual, 1.0)
        name = "Schedules"
        
        self.balanceSum += balance
        if balance > self.balanceMax:
            self.blackSheep = name

        self.balanceMax = max(balance, self.balanceMax)
        self.nBalanceChecks += 1

    
#------------------------------------------------------------------------------
    def initCheckBalance(self):
#------------------------------------------------------------------------------
        self.balanceSum = 0
        self.balanceMax = 0
        self.nBalanceChecks = 0
        self.blackSheep = None

#------------------------------------------------------------------------------
    def getMeanBalance(self):
#------------------------------------------------------------------------------
        if self.nBalanceChecks > 0:
            return self.balanceSum / self.nBalanceChecks
        else:
            return 0

#------------------------------------------------------------------------------
    def getMaxBalance(self):
#------------------------------------------------------------------------------
        return self.balanceMax

        
#------------------------------------------------------------------------------
    def checkTotalBalance(self):
#------------------------------------------------------------------------------
        self.totalBalanceSum += self.getMeanBalance()
        balanceMax = self.getMaxBalance()
        
        if balanceMax > self.totalBalanceMax:
            groupName = conflict.dataGroup
            if groupName is None: groupName = "---"
            self.totalBlackSheep = groupName + "." + self.blackSheep

        self.totalBalanceMax = max(self.totalBalanceMax, balanceMax)
        self.nTotalBalanceChecks += 1

        if DEBUG in ["ALL","MAIN"]:
            print "checkTotalBalance - current maximum = %s (%s)"%(self.totalBalanceMax,self.totalBlackSheep)

#------------------------------------------------------------------------------
    def initTotalBalance(self):
#------------------------------------------------------------------------------
        self.totalBalanceSum = 0
        self.totalBalanceMax = 0
        self.nTotalBalanceChecks = 0
        self.totalBlackSheep = None

#------------------------------------------------------------------------------
    def getMeanTotalBalance(self):
#------------------------------------------------------------------------------
        if self.nTotalBalanceChecks > 0:
            return self.totalBalanceSum / self.nTotalBalanceChecks
        else:
            return 0

#------------------------------------------------------------------------------
    def getMaxTotalBalance(self):
#------------------------------------------------------------------------------
        return self.totalBalanceMax

        
#------------------------------------------------------------------------------
    def converged(self):
#------------------------------------------------------------------------------

        if (self.nChecks > 0):
            self.diff /= self.nChecks

        self.improvement = self.lastDiff - self.diff
        if self.diff < self.maxDifference or self.improvement <= 0:
            self.itImproveCtr += 1
        else:
            self.itImproveCtr = 0
        self.lastDiff = self.Diff

        if self.itImproveCtr >= self.itMaxImprove:
            return True
        else:
            return False
        

#creates a Cycle-instance for being used in cycle checking
cycle = Cycle(10, 1.e-3, 1.e-5, 3)     #basic cycle set-up for cycle class

#==============================================================================
conflict = CCConflict()     #instance for access to CCConflict functions
screen = CCScreen()
estimateTracker = CCEstimate()


