# -*- coding: utf-8 -*-
#==============================================================================                 
#
#	 E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    CCheck (Consistency Check)
#            
#------------------------------------------------------------------------------
#            
#    Short description:
#    
#    Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================                 
from math import fabs, log, exp
from constants import *                
from modules.messageLogger import *
from GUI.GUITools import *     #needed for function check
from ccTracker import estimateTracker, screen, conflict, cycle
from ccCheckers import *
from ccPar import CCPar
import numpy as np

global DEBUG
EPSILON = 1.e-10     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"
MINIMUM_VALUE = 1.e-10
DEFAULT_SQERR = 1.e-6 # default value for sqerr assigned to questionnaire values (until V2.1 default had been set to 1e-4)
NUMERIC_ERR = 1.e-10 # accuracy of numeric calculations
CONFLICT_ERR = 1.e-6
MAX_SQERR = 0.09     # critical square error for screening
CONFIDENCE = 4.0    # maximum relation between statistical error and abs.min/max
MAXTRACKENTRIES = 10 # maximum number of variables to be tracked

def debugPars(criteria, parvect, msg):
    if DEBUG in criteria:
        print msg
        for par in parvect:
            par.show()
    
    
TEST = False
TESTCASE = 3

def combinedErr(sqerr1, sqerr2):
    """ combination of errors in products, which gives reasonable values also for err>1
    """
    sqerr = sqerr1 + sqerr2 + sqerr1*sqerr2
    return min(sqerr, INFINITE)
    
#==============================================================================                 
# Global CCheck auxiliary functions
#==============================================================================                 

def setIdentity(x1, x2):
    """ first checks independently x1 and x2 and then
        puts into one """

    x1.check()
    x2.check()
    ccheck1(x1,x2)

# Block for min/max equations
def equationK_MinMax(y, a, x1):
    """ Equation Y = a * x1
        Calculation of Min/Max values
    """
    y.valMin = a * x1.valMin
    y.valMax = min(INFINITE, a * x1.valMax)
    y.constrain()

def equationProdC_MinMax(y, a, x1, x2):
    """ Equation Y = a * x1 * x2
        Calculation of Min/Max values
    """
    #TODO: blocks for calcProd commented out !
    if x1.parType not in ["S","DTmN","T"] and x2.parType not in ["S","DTmN","T"]:
##        y.valMin = a * x1.valMin * x2.valMin
##        y.valMax = min(INFINITE, a * x1.valMax * x2.valMax)
##
##
        if (a * x2.valMax > 0):
            newMin = min(y.valMin / (a * x2.valMax), INFINITE)
        elif (y.valMin > 0):
            newMin = INFINITE
        else:
            newMin = x1.valMin
    
        if (a * x2.valMin > 0):
            newMax = min(y.valMax / (a * x2.valMin), INFINITE)
        elif (y.valMax > 0):
            newMax = INFINITE
        else:
            newMax = x1.valMax
    
        if (a * x1.valMax > 0):
            x2.valMin = min(y.valMin / (a * x1.valMax), INFINITE)
        elif (y.valMin > 0):
            x2.valMin = INFINITE
    
        if (a * x1.valMin > 0):
            x2.valMax = min(y.valMax / (a * x1.valMin), INFINITE)
        elif (y.valMax > 0):
            x2.valMax = INFINITE
    
        x1.valMin = newMin
        x1.valMax = newMax

    else:
##        LL = x1.valMin * x2.valMin
##        LH = x1.valMin * x2.valMax
##        HL = x1.valMax * x2.valMin
##        HH = x1.valMax * x2.valMax
##        y.valMin = min(INFINITE,a*min(LL,LH,HL,HH))
##        y.valMax = min(INFINITE,a*max(LL,LH,HL,HH))
##
        if (a*x2.valMax) <> 0:
            LH1 = y.valMin/(x2.valMax*a)
            HH1 = y.valMax/(x2.valMax*a)
        else:
            LH1 = -INFINITE
            HH1 = INFINITE
        if (a*x2.valMin) <> 0:
            LL1 = y.valMin/(x2.valMin*a)
            HL1 = y.valMax/(x2.valMin*a)
        else:
            LL1 = -INFINITE
            HL1 = INFINITE
        if (a*x1.valMax) <> 0:
            LH2 = y.valMin/(x1.valMax*a)
            HH2 = y.valMax/(x1.valMax*a)
        else:
            LH2 = -INFINITE
            HH2 = INFINITE
        if (a*x1.valMin) <> 0:
            LL2 = y.valMin/(x1.valMin*a)
            HL2 = y.valMax/(x1.valMin*a)
        else:
            LL2 = -INFINITE
            HL2 = INFINITE

        if x2.valMin > 0 or x2.valMax < 0:
            newX1Min = min(INFINITE,min(LL1,LH1,HL1,HH1))
            newX1Max = min(INFINITE,max(LL1,LH1,HL1,HH1))
        else:
            newX1Min = -INFINITE
            newX1Max = INFINITE
            
        if x1.valMin > 0 or x1.valMax < 0:
            x2.valMin = min(INFINITE,min(LL2,LH2,HL2,HH2))
            x2.valMax = min(INFINITE,max(LL2,LH2,HL2,HH2))
        else:
            x2.valMin = -INFINITE
            x2.valMax = INFINITE

        x1.valMin = newX1Min
        x1.valMax = newX1Max

    y.constrain()
    x1.constrain()
    x2.constrain()
    

#------------------------------------------------------------------------------
def calcK(yname, a, x1, parType = None, unitType = None):
#------------------------------------------------------------------------------
#   Default function for calculating the product among a value and a constant
#------------------------------------------------------------------------------
    assert(a>=0)    #not valid for negative constants

    y = CCPar(yname, parType = parType, unitType = unitType)

    if (x1.val == None):
        y.val = None
        y.sqerr = INFINITE
    else:
        y.val = a * x1.val
        y.sqerr = x1.sqerr
        y.userDefined = x1.userDefined

    y.valMin = a * x1.valMin
    y.valMax = min(INFINITE, a * x1.valMax)
    y.constrain()

    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.cleanUp()

    debugPars(["ALL", "CALC"], [x1, y],"CalcK(%s*x1)_______________________" % a)

    return y

#------------------------------------------------------------------------------
def calcNegative(yname, x, parType = None, unitType = None):
    """ identical to calcK with a = -1 """
    
    y = CCPar(yname, parType = parType, unitType = unitType)
    y.update(x)

    if (x.val is not None):
        y.val = - x.val
        y.userDefined = x.userDefined #TODO -> not sure if this is done automatically in update. If yes, can be eliminated

    y.valMin = - x.valMax
    y.valMax = - x.valMin

    if DEBUG in ["ALL", "CALC"]:
        print "CalcNegative(x)_______________________"
        x.show()
        print "-x = "
        y.show()
        print "___________________________________________________"

    return y

#------------------------------------------------------------------------------
def calcLog(yname, x1):
    """
    Default function for calculating the natural logarithm of a variable
    """
    
    y = CCPar(yname)

    if (x1.val == None):
        y.val = None
        y.sqerr = INFINITE
    elif x1.val <= 0:
        y.val = -INFINITE
        y.sqerr = INFINITE
    else:
        y.val = log(x1.val)
        y.sqdev = x1.sqerr
        y.calcErr()
        y.userDefined = x1.userDefined

    if x1.valMin > 0:
        y.valMin = max(-INFINITE, log(x1.valMin))
    else:
        y.valMin = - INFINITE
        
    if x1.valMax > 0:
        y.valMax = min(INFINITE, log(x1.valMax))
    else:
        y.valMax = INFINITE
        
    y.constrain()

    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.cleanUp()

    if DEBUG in ["ALL", "CALC"]:
        print "CalcLog(x1)_______________________"
        x1.show()
        print "log(x1) = "
        y.show()
        print "___________________________________________________"

    return y

#------------------------------------------------------------------------------
def calcProd(yname, x1, x2, parType = None, unitType = None):
    """ Default function for calculating the product of two values
    """
    return calcProdC(yname, 1.0, x1, x2, parType, unitType)

def calcProdC(yname, a, x1, x2,parType=None, unitType=None):
    """ Default function for calculating the product among two values and a constant
    """
    y = CCPar(yname,parType=parType,unitType=unitType)
    
    if (x1.val == 0 and x1.sqerr == 0) or (x2.val == 0 and x2.sqerr == 0):
        y.val = 0
        y.sqerr = 0
        if x1.val == 0 and x1.sqerr == 0:
            y.userDefined = x1.userDefined
        if x2.val == 0 and x2.sqerr == 0:
            y.userDefined = x2.userDefined

    elif (x1.val == None or x2.val == None):
        y.val = None
        y.sqerr = INFINITE
    else:
        y.val = a * x1.val * x2.val
        y.sqerr = combinedErr(x1.sqerr, x2.sqerr)
        y.userDefined = x1.userDefined and x2.userDefined

    if x1.parType not in ["S","DTmN","T"] and x2.parType not in ["S","DTmN","T"]:
        y.valMin = a * x1.valMin * x2.valMin
        y.valMax = min(INFINITE, a * x1.valMax * x2.valMax)
    else:
        LL = x1.valMin * x2.valMin
        LH = x1.valMin * x2.valMax
        HL = x1.valMax * x2.valMin
        HH = x1.valMax * x2.valMax
        y.valMin = min(INFINITE,a*min(LL,LH,HL,HH))
        y.valMax = min(INFINITE,a*max(LL,LH,HL,HH))
    y.constrain()

    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.track.extend(x2.track)
        y.cleanUp()

    debugPars(["ALL", "CALC"], [x1, x2, y], "CalcProdC y=%s*x1*x2)___________________________________"%a)
    
    return y  

def calcSum(yname, x1, x2, parType=None):
    """
    Calculate the sum of two values.
    
    @param yname: the name for the resulting CCPar object
    @param x1: the one summand
    @param x2: the other summand
    @param parType: the parType for the resulting CCPar object 
    """
    try:
        assert(x1.unitType == x2.unitType)
    except:
        print "Assertion Error: different unitType in calcSum"
        x1.show(showAttributes=True)
        x2.show(showAttributes=True)
        
    y = CCPar(yname, parType=parType, unitType=x1.unitType) #inherit unit type from first element
    if (x1.val == None or x2.val == None):
        y.val = None
        y.sqerr = INFINITE
    else:
        y.val = x1.val + x2.val
        x1.calcDev()
        x2.calcDev()
        y.dev = x1.dev + x2.dev
        y.sqdev = pow(y.dev, 2.0)
        y.calcErr()
        y.userDefined = x1.userDefined and x2.userDefined

    y.valMin = min(INFINITE, x1.valMin + x2.valMin)
    y.valMax = min(INFINITE, x1.valMax + x2.valMax)

    if DEBUG in ["ALL", "CALC"]:
        print "CalcSum(x1,x2)_(before constrain)__________________"
        print "x1+x2 = "
        y.show()
        print "parType = ", y.parType
        print "___________________________________________________"

    y.constrain()

    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.track.extend(x2.track)
        y.cleanUp()

    if DEBUG in ["ALL", "CALC"]:
        print "CalcSum(x1,x2)_____________________________________"
        x1.show()
        x2.show()
        print "x1+x2 = "
        y.show()
        print "___________________________________________________"

    return y
    
#------------------------------------------------------------------------------
def calcSum3(yname, x1, x2, x3,parType=None): # To be replaced by the matrixsumcalc
#------------------------------------------------------------------------------
#   Default function for calculating the sum of 3 values
#------------------------------------------------------------------------------

    return calcRowSum(yname, [x1, x2, x3], parType)

def calcRowSum(name, row, parType = None):
    """Return the sum of a list of CCPar objects."""
    y = CCPar(name, parType=parType)
    if len(row) == 0:   #special case: empty list -> sum = 0
        y.setValue(0, setTrack=False)
        return y
    else:
        y.unitType = row[0].unitType #inherit type of first element
    
    (nones, dummy) = countNones(row)

    #   Case 1: all values are defined 
    if nones == 0:
        y.dev = 0
        y.valMin = 0
        y.valMax = 0
        y.val = 0
        y.userDefined = True
        for item in row:
            y.val += item.val
            item.calcDev()
            y.dev += item.dev
            y.valMin += item.valMin
            y.valMax += item.valMax
            y.valMax = min(y.valMax, INFINITE)
            y.userDefined = y.userDefined and item.userDefined

        y.sqdev = pow(y.dev, 2.0)
        y.calcErr()
    
    #   Case 2: any of the values is None 
    else:
        y.val = None
        y.sqerr = INFINITE

        for item in row:
            y.valMin += item.valMin
            y.valMax += item.valMax
            y.valMax = min(y.valMax, INFINITE)

    y.constrain()

    if y.val is not None:
        y.track = []
        for item in row:
            y.track.extend(item.track)

        y.cleanUp()

    if DEBUG in ["ALL", "CALC"]:
        print "CalcRowSum(y,row)__________________________________"
        for item in row:
            item.show()
        print "y = SUM x[i] = "
        y.show()
        print "___________________________________________________"

    return y   

def calcDiff(yname, x1, x2, parType=None):
    """Return the difference of two CCPar values. The result is always
    non negative. The result is lifted to 0 if necessary. 
    
    :param yname: name of the result
    :param x1: the minuend
    :param x2: the subtrahend
    :param parType: The parType of the result. See CCpar documentation
    for more details.
    
    :returns: a CCPar value containing the difference x1-x2

    """
    y = CCPar(yname, parType=parType, unitType = x1.unitType)
    if (x1.val is None or x2.val is None):
        y.val = None
        y.sqerr = INFINITE
    else:
        y.val = x1.val - x2.val
        y.userDefined = x1.userDefined and x2.userDefined
            
        if (y.val != 0):
            x1.calcDev()
            x2.calcDev()
            y.dev = x1.dev + x2.dev
            y.sqdev = pow(y.dev, 2.0)
            y.calcErr()
        elif iszero(x1) and iszero(x2):   # all the two are exactly zero
            y.sqerr = 0
        else:
            y.sqerr = INFINITE

    y.valMin = x1.valMin - x2.valMax
    y.valMax = x1.valMax - x2.valMin

    if y.parType not in ["S","DTmN","T"]:
        y.valMin = max(y.valMin, 0)
        y.valMax = max(y.valMax, 0)

    y.constrain()
    
    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.track.extend(x2.track)
        y.cleanUp()

    if DEBUG in ["ALL", "CALC"]:
        print "CalcDiff(x1,x2)____________________________________"
        x1.show()
        x2.show()
        print "x1-x2 = "
        y.show()
        print "___________________________________________________"

    return y

def calcFlow_stack(Qdotname, cp, m, T1, T0, DT,unitType=None):
    """
    Calculate the power required to increase the temperature of a mass
    flow m at a starting temperature T0 up to the final temperature T1.
    :returns: the required power
    """
    diff = calcDiff(DT.name, T1, T0)
    DT.push(diff)    
    DT.check()
    Qdot = calcProdC(Qdotname, cp, m, DT, unitType=unitType)
    if DEBUG in ["ALL", "CALC"]:
        print "CalcFlow-DT________________________________________"
        T1.show()
        T0.show()
        print "DT = T1 - T0 = "
        DT.show()
        print "___________________________________________________"
    if DEBUG in ["ALL", "CALC"]:
        print "CalcFlow-Qdot______________________________________"
        m.show()
        print "cp = ", cp
        print "Qdot = m*cp*DT = "
        Qdot.show()
        print "___________________________________________________"
    return Qdot

def enthalpy(cL, cV, h, T0, T, x):
#------------------------------------------------------------------------------
#   enthalpy calculation
#------------------------------------------------------------------------------

    if (T == None):
        return(None)
    
    elif h is None or T0 is None:
        return (cL * T)
    
    elif isequal(T, T0):
        if x is None:
            return (cL * T0 + 0.5 * h)
        else:
            return(cL * T0 + x * h)
    
    elif T < T0:
        return(cL * T)
    
    elif T > T0:
        return(cL * T0 + h + cV * (T - T0))

#------------------------------------------------------------------------------
def invEnthalpy(cL, cV, h, T0, y):
#------------------------------------------------------------------------------
#   enthalpy calculation
#------------------------------------------------------------------------------

    h0 = cL * T0
    h1 = h0 + h
    
    if (y == None):
        return(None, None)
    
    elif y < h0:     #liquid phase
        return (y / cL, 0.0)
    
    elif y < h1:
        if h > 0:   #avoid zero division
            x = (y - h0) / h
        else:
            x = 1.0
        return(T0, x)
    
    elif y >= h1:
        return(T0 + (y - h1) / cV, 1.0)
    
#------------------------------------------------------------------------------
def calcH(yname, cL, cV, h, T0, x1, x2):
#------------------------------------------------------------------------------
#   Default function for calculating the entalpy h
#------------------------------------------------------------------------------
    y = CCPar(yname,parType="S", unitType = "SPECIFICENTHALPY")

    y.val = enthalpy(cL, cV, h, T0, x1.val, x2.val)
    y.valMin = enthalpy(cL, cV, h, T0, x1.valMin, x2.valMin)
    y.valMax = enthalpy(cL, cV, h, T0, x1.valMax, x2.valMax)

    if x1.val is None:
        y.sqerr = INFINITE
    else:
        devT = pow(x1.calcDev(), 0.5)
        devh = pow(x2.calcDev(), 0.5)
#        TL = max(0.0, x1.val - devT)
        TL = x1.val - devT

        if x2.val is None:
            xL = 0
        else:
            xL = max(0.0, x2.val - devh)
        hL = enthalpy(cL, cV, h, T0, TL, xL)
        
        TH = x1.val + devT
        if x2.val is None:
            xH = 1
        else:
            xH = x2.val + devh
        hH = enthalpy(cL, cV, h, T0, TH, xH)

        y.sqdev = pow(max(hH - y.val, y.val - hL), 2.0)
        y.calcErr()

    if x1.valMin >= T0 or x1.valMax <= T0:  #if condensation temperature out of limits, h = f(T) only.
        y.userDefined = x1.userDefined
    else:
        y.userDefined = x1.userDefined and x2.userDefined
        
    y.constrain()

    if y.val is not None:
        y.track = []
        y.track.extend(x1.track)
        y.track.extend(x2.track)
        y.cleanUp()

    if DEBUG in ["ALL", "CALC"]:
        print "CalcH(x1,x2)____________________________________"
        x1.show()
        x2.show()
        print "h(%s,x1,x2) = " % cL
        y.show()
        print "___________________________________________________"

    return y


def adjustK_stack(y, a, x1):
    """ function added for symmetry of nomenclature
    -> see function adjustcalcK_stack
    """
    
    adjustcalcK_stack(y, a, x1)
    
#------------------------------------------------------------------------------
def adjustcalcK_stack(y, a, x1_old):
#------------------------------------------------------------------------------
#   Default function for calculating and adjusting the product between a value and a constant
#------------------------------------------------------------------------------
    x1 = CCPar(x1_old.name)
    x1.update(x1_old)

    if a == 0: return   # cant adjust anything in this case. If y <> 0 conflict will be detected in corresponding calc-function

    if y.val is not None:
        
        x1.val = y.val / a
        x1.sqerr = y.sqerr
        x1.userDefined = y.userDefined

    x1.valMin = min(y.valMin / a, INFINITE)
    x1.valMax = min(y.valMax / a, INFINITE)

    x1.constrain()

    if y.val is not None:
        x1.track = []
        x1.track.extend(y.track)

        x1.cleanUp()
    
    x1_old.push(x1)

#------------------------------------------------------------------------------
def adjustLog_stack(y, x1_old):
#------------------------------------------------------------------------------
#   Default function for calculating and adjusting the product between a value and a constant
#------------------------------------------------------------------------------
    x1 = CCPar(x1_old.name)
    x1.update(x1_old)

    if y.val is not None:
        
        x1.val = exp(y.val)
        y.calcDev()
        x1.sqerr = y.sqdev
        x1.userDefined = y.userDefined

    x1.valMin = min(exp(y.valMin), INFINITE)
    x1.valMax = min(exp(y.valMax), INFINITE)

    x1.constrain()

    if y.val is not None:
        x1.track = []
        x1.track.extend(y.track)

        x1.cleanUp()
    
    x1_old.push(x1)

#------------------------------------------------------------------------------
def adjustProd_stack(y, x1_old, x2_old):
    """
    Adjust two values of which the product is known.
    
    The new adjusted values are pushed on their respective stacks.
    If y, x1_old and x2_old are lists or tuples with common length 
    adjustProd_stack is applied for each triple of entries.
    
    :param y: known product
    :param x1_old: one factor
    :param x2_old: the other factor

    """
    return adjustProdC_stack(y, 1.0, x1_old, x2_old)

def adjustSumOfProds_stack(y, x1, x2):    
    """
    Adjust lists of corresponding factors, where the sum of the products is
    known.
    
    The Sum should be of the form:
    sum over n of (x1_n * x2_n)
    The new, adjusted values are pushed on their respective stacks.
    
    :param y: the known sum
    :param x1: the one list of factors
    :type x1: list of CCPar objects
    :param x2: the other list with the corresponding factors
    :type x2: list of CCPar objects
    """
    assert(len(x1) == len(x2))
    products = [calcProd("temp. Product %s" % (i,), x1[i], x2[i], unitType = x1[i].unitType) 
                for i in xrange(len(x1))]
    adjRowSum_stack(y, products, stack=False)
    for i in xrange(len(products)):
        adjustProd_stack(products[i], x1[i], x2[i])

def adjustSumOfProdsC_stack(y, a, x1, x2):    
    """
    Adjust two lists of corresponding variable factors and keep a list
    of constant factors as is, where the sum of the products is known.
    
    The Sum should be of the form:
    sum over n of (a_n * x1_n * x2_n)
    The new, adjusted values are pushed on their respective stacks.
    
    :param y: the known sum
    :param a: the list of constant factors 
    :param x1: the one list of factors
    :type x1: list of CCPar objects
    :param x2: the other list with the corresponding factors
    :type x2: list of CCPar objects
    """
    assert(len(x1) == len(x2) == len(a))
    products = [calcProdC("temp. Product %s" % (i,), a[i], x1[i], x2[i], unitType = x1[i].unitType) 
                for i in xrange(len(x1))]
    adjRowSum_stack(y, products, stack=False)
    for i in xrange(len(products)):
        adjustProdC_stack(products[i], a[i], x1[i], x2[i])


#------------------------------------------------------------------------------
def adjustProdC_stack(y, a, x1_old, x2_old):
    """
    Calculate and adjust the variable factors of a product among two values and
    a constant that has a known result.
    
    The new adjusted values are pushed on their respective stacks.
    If y, a, x1_old and x2_old are lists or tuples with common length 
    adjustProd_stack is applied for each quadruple of entries.
    
    :param y: the known product 
    :param a: the constant factor
    :param x1_old: the one variable factor
    :param x2_old: the other variable factor
    """
    # are the input parameters are list/tuples of compatible length? 
    if (all(isinstance(para, (type([]), type(()))) for para in (y, a, x1_old, x2_old))
         and len(y) == len(x1_old) == len(x2_old)):
        for i in xrange(len(y)):
            adjustProdC_stack(y[i], a[i], x1_old[i], x2_old[i])
    else: # no they are not:
        x1 = CCPar(x1_old.name)
        x2 = CCPar(x2_old.name)
        x1.update(x1_old)
        x2.update(x2_old)
    
        if DEBUG in ["ALL", "ADJUST"]:
            print "___________________________________________________"
            print "adjustProdC starting", x1.name, x2.name
            y.show()
            x1.show()
            x2.show()

        if a == 0: return   # cant adjust anything in this case. If y <> 0 conflict will be detected in corresponding calc-function
        
        x1updated = False
        x2updated = False
        
        if y.val is not None:
            if iszero(y):
                if x1.valMin*x1.valMax > 0: # x1 is different from zero
                    x2.setZero()
                    x2.userDefined = y.userDefined
                    x2updated = True
                elif x2.valMin*x2.valMax > 0:
                    x1.setZero()
                    x1.userDefined = y.userDefined
                    x1updated = True
            else:
                prefer_x1 = (x2.val is None or (x1.userDefined and not x2.userDefined)) # preference for one of both independent of sqerr-comparison
                prefer_x2 = (x1.val is None or (x2.userDefined and not x1.userDefined))
                if not (prefer_x1 or prefer_x2):
                    prefer_x1 = (x1.sqerr < x2.sqerr)

                if (x1.val is not None) and prefer_x1:
                    x2updated = True
                    if (x1.val == 0):
                        if y.val > 0:
                            x2.val = INFINITE
                            x2.sqerr = INFINITE
                        else:
                            x2updated = False
                    else:
                        x2.val = y.val / (a * x1.val)
                        x2.sqerr = combinedErr(y.sqerr, x1.invSqErr())
                        x2.calcDev()
                        x2.userDefined = y.userDefined and x1.userDefined
        
                elif (x2.val is not None):
                    x1updated = True
                    if (x2.val == 0):
                        if y.val > 0:
                            x1.val = INFINITE
                            x1.sqerr = INFINITE
                        else:
                            x1updated = False
                    else:
                        x1.val = y.val / (a * x2.val)
                        x1.sqerr = combinedErr(y.sqerr, x2.invSqErr())
                        x1.calcDev()
                        x1.userDefined = y.userDefined and x2.userDefined
    
        equationProdC_MinMax(y, a, x1, x2)
##        if x1.parType not in ["S","DTmN","T"] and x2.parType not in ["S","DTmN","T"]:
##            if (a * x2.valMax > 0):
##                newMin = min(y.valMin / (a * x2.valMax), INFINITE)
##            elif (y.valMin > 0):
##                newMin = INFINITE
##            else:
##                newMin = x1.valMin
##        
##            if (a * x2.valMin > 0):
##                newMax = min(y.valMax / (a * x2.valMin), INFINITE)
##            elif (y.valMax > 0):
##                newMax = INFINITE
##            else:
##                newMax = x1.valMax
##        
##            if (a * x1.valMax > 0):
##                x2.valMin = min(y.valMin / (a * x1.valMax), INFINITE)
##            elif (y.valMin > 0):
##                x2.valMin = INFINITE
##        
##            if (a * x1.valMin > 0):
##                x2.valMax = min(y.valMax / (a * x1.valMin), INFINITE)
##            elif (y.valMax > 0):
##                x2.valMax = INFINITE
##        
##            x1.valMin = newMin
##            x1.valMax = newMax
##    
##        else:
##            if (a*x2.valMax) <> 0:
##                LH1 = y.valMin/(x2.valMax*a)
##                HH1 = y.valMax/(x2.valMax*a)
##            else:
##                LH1 = -INFINITE
##                HH1 = INFINITE
##            if (a*x2.valMin) <> 0:
##                LL1 = y.valMin/(x2.valMin*a)
##                HL1 = y.valMax/(x2.valMin*a)
##            else:
##                LL1 = -INFINITE
##                HL1 = INFINITE
##            if (a*x1.valMax) <> 0:
##                LH2 = y.valMin/(x1.valMax*a)
##                HH2 = y.valMax/(x1.valMax*a)
##            else:
##                LH2 = -INFINITE
##                HH2 = INFINITE
##            if (a*x1.valMin) <> 0:
##                LL2 = y.valMin/(x1.valMin*a)
##                HL2 = y.valMax/(x1.valMin*a)
##            else:
##                LL2 = -INFINITE
##                HL2 = INFINITE
##
##            if x2.valMin > 0 or x2.valMax < 0:
##                newX1Min = min(INFINITE,min(LL1,LH1,HL1,HH1))
##                newX1Max = min(INFINITE,max(LL1,LH1,HL1,HH1))
##            else:
##                newX1Min = -INFINITE
##                newX1Max = INFINITE
##                
##            if x1.valMin > 0 or x1.valMax < 0:
##                x2.valMin = min(INFINITE,min(LL2,LH2,HL2,HH2))
##                x2.valMax = min(INFINITE,max(LL2,LH2,HL2,HH2))
##            else:
##                x2.valMin = -INFINITE
##                x2.valMax = INFINITE
##
##            x1.valMin = newX1Min
##            x1.valMax = newX1Max


        x1.constrain()
        x2.constrain()
    
        if y.val is not None:
            newtrackx1 = []
            newtrackx1.extend(y.track)
            if x2.val is not None:
                newtrackx1.extend(x2.track)
    
            newtrackx2 = []
            newtrackx2.extend(y.track)
            if x1.val is not None:
                newtrackx2.extend(x1.track)
    
            if x1updated: x1.track = newtrackx1
            if x2updated: x2.track = newtrackx2
    
            x1.cleanUp()
            x2.cleanUp()
    
        debugPars(["ALL", "ADJUST"], [y, x1, x2], "adjustProdC %s________________________________________"%a)

        x1_old.push(x1)
        x2_old.push(x2)

#------------------------------------------------------------------------------
def adjustSum_stack(y, x1_old, x2_old):    
#------------------------------------------------------------------------------
#   Default function for calculating the product of two values
#------------------------------------------------------------------------------
    x1 = CCPar(x1_old.name)
    x2 = CCPar(x2_old.name)
    x1.update(x1_old)
    x2.update(x2_old)

    x1updated = False
    x2updated = False

    if DEBUG in ["ALL", "ADJUST"]:
        print "___________________________________________________"
        print "adjustSum starting", x1.name, x2.name
        y.show()
        x1.show()
        x2.show()

    if iszero(y) and (x1.parType not in ["S","DTmN","T"]) and (x2.parType not in ["S","DTmN","T"]):   #special case SUM = 0, valid if all numbers are positive
        x1.val = 0.0
        x1.sqerr = 0.0
        x1.userDefined = y.userDefined
        x2.val = 0.0
        x2.sqerr = 0.0
        x2.userDefined = y.userDefined

        x1updated = True
        x2updated = True
        
    elif not (y.val == None):
        y.calcDev()
        x1.calcDev()
        x2.calcDev()
        
        if not (x1.val is None) and ((x1.sqdev < x2.sqdev) or x2.val is None):

            x2updated = True
            
            x2.val = y.val - x1.val
            x2.dev = y.dev + x1.dev
            x2.sqdev = pow(x2.dev, 2.0)
            x2.userDefined = y.userDefined and x1.userDefined

            if (x2.val != 0):
                x2.calcErr()
            else:
                x2.sqerr = INFINITE

        elif not (x2.val is None) and ((x2.sqdev <= x1.sqdev) or x1.val is None): # small assymetry for not considering the special case x1.sqdev = x2.sqdev

            x1updated = True
            
            x1.val = y.val - x2.val
            x1.dev = y.dev + x2.dev
            x1.sqdev = pow(x1.dev, 2.0)
            x1.userDefined = y.userDefined and x2.userDefined

            if (x1.val != 0):
                x1.calcErr()
            else:
                x1.sqerr = INFINITE

        elif x1.val is None and x2.val is None:
            sumMin = x1.valMin + x2.valMin
            sumMax = x1.valMax + x2.valMax
            if sumMax > sumMin:
                gamma = (y.val - sumMin) / (sumMax - sumMin)
            else:
                gamma = 0
                
            x1.val = x1.valMin + gamma * (x1.valMax - x1.valMin)
            x2.val = x2.valMin + gamma * (x2.valMax - x2.valMin)
            
            newDev1 = y.dev + x2.dev
            newDev2 = y.dev + x1.dev

            x1.sqdev = pow(x1.dev, 2.0)
            x2.sqdev = pow(x2.dev, 2.0)
            x1.calcErr()
            x2.calcErr()
    
    newMin1 = y.valMin - x2.valMax - abs(x2.valMax)*NUMERIC_ERR
    newMax1 = y.valMax - x2.valMin + abs(x2.valMin)*NUMERIC_ERR

    if abs(newMin1) < NUMERIC_ERR: newMin1 = 0
    if abs(newMax1) < NUMERIC_ERR: newMax1 = 0

    newMin2 = y.valMin - x1.valMax - abs(x1.valMax)*NUMERIC_ERR
    newMax2 = y.valMax - x1.valMin + abs(x1.valMin)*NUMERIC_ERR

    if abs(newMin2) < NUMERIC_ERR: newMin2 = 0
    if abs(newMax2) < NUMERIC_ERR: newMax2 = 0

    x1.valMin = newMin1
    x1.valMax = newMax1

    x2.valMin = newMin2
    x2.valMax = newMax2

    if x2.parType not in ["S","DTmN","T"]:
        x2.valMin = max(x2.valMin, 0.0)
        x2.valMax = max(x2.valMax, 0.0)

    if x1.parType not in ["S","DTmN","T"]:
        x1.valMin = max(x1.valMin, 0.0)
        x1.valMax = max(x1.valMax, 0.0)
        

    if DEBUG in ["ALL", "ADJUST"]:
        print "adjustSum before constraints", x1.name, x2.name
        y.show()
        x1.show()
        x2.show()
        print "___________________________________________________"

    x1.constrain()
    x2.constrain()

    if y.val is not None:
        newtrackx1 = []
        newtrackx1.extend(y.track)
        if x2.val is not None:
            newtrackx1.extend(x2.track)

        newtrackx2 = []
        newtrackx2.extend(y.track)
        if x1.val is not None:
            newtrackx2.extend(x1.track)

        if x1updated: x1.track = newtrackx1
        if x2updated: x2.track = newtrackx2

        x1.cleanUp()
        x2.cleanUp()

    if DEBUG in ["ALL", "ADJUST"]:
        print "adjustSum after adjustment", x1.name, x2.name
        y.show()
        x1.show()
        x2.show()
        print "___________________________________________________"
    
    x1_old.push(x1)
    x2_old.push(x2)
       
#------------------------------------------------------------------------------
def adjustSum3_stack(y, x1, x2, x3): #Adjusting any value not only the worst
#------------------------------------------------------------------------------
#   Default function for calculating and adjusting the components of a sum of 3 values
#------------------------------------------------------------------------------
    return adjRowSum_stack(y, [x1, x2, x3])

#------------------------------------------------------------------------------
def adjRowSum_minMax(y, row):
    """ auxiliary function for adjRowSum_stack

        calculates the min and max values for all row parameters

    :returns rowmin: vector of minimum values (numpy array)
    :returns rowmax: vector of maximum values
    """

    m = len(row)
    if m <= 0: return(None,None)
    
    rowmin = np.array([float(x.valMin) for x in row])
    rowmax = np.array([float(x.valMax) for x in row])

    newmin = np.array([float(y.valMin)]*m)
    newmax = np.array([float(y.valMax)]*m)

    #first substract all the mins/maxs for j < i
    minSum = 0
    maxSum = 0
    minSumAbs = 0
    maxSumAbs = 0
    
    for i in range(1,m):
        minSum += row[i-1].valMin
        maxSum += row[i-1].valMax

        minSumAbs += abs(row[i-1].valMin)
        maxSumAbs += abs(row[i-1].valMin)

        newmin[i] -= (maxSum + maxSumAbs*NUMERIC_ERR)
        newmax[i] -= (minSum - minSumAbs*NUMERIC_ERR)

    #then substract all the mins/maxs for j > i
    minSum = 0
    maxSum = 0
    for i in range(m-2,-1,-1):
        minSum += row[i+1].valMin
        maxSum += row[i+1].valMax

        minSumAbs += abs(row[i+1].valMin)
        maxSumAbs += abs(row[i+1].valMin)

        newmin[i] -= (maxSum + maxSumAbs*NUMERIC_ERR)
        newmax[i] -= (minSum - minSumAbs*NUMERIC_ERR)

    # try to avoid conflicts due to rounding errors

    newmin = np.maximum(rowmin,newmin)
    newmax = np.minimum(rowmax,newmax)

    return (newmin,newmax)
        

def adjRowSum_stack(y, row_old, m = None, stack=True):
    """
    Adjust given values and estimate the ones that are None such that
    their sum is as prescribed.
    
    The new values are pushed on their respective stacks, and the 
    absolute difference of the incoming summands and the given
    sum is returned.
    
    :param y: The known sum
    :param row_old: a list of CCPar summands
    :param m: the number of items in row_old that are to be adjusted.
      the first m items are adjusted.
    :param stack: default is True. Option stack=False only used in function adjSumOfProducts
    """
    assert(m == len(row_old) or m is None)
    m = len(row_old)
    
    if stack:
        row = []
        for i in range(len(row_old)):
            row.append(CCPar(row_old[i].name))
            row[i].update(row_old[i])
    else:
        row = row_old

    newMin, newMax = adjRowSum_minMax(y,row)
    for i in range(m):
        row[i].valMin = newMin[i]
        row[i].valMax = newMax[i]
    
    nNones = 0
    Sum = 0.0
    SumMin = 0.0
    SumMax = 0.0
    SumErr = 0.0
    SumSqDev = 0.0
    SumMinNones = 0.0
    SumMaxNones = 0.0
    updated = [False]*m

    y.calcDev()
    
    indexLastUndefined = -1
    nUndefined = 0
    unsigned = True
    for i in range(m):
        if row[i].parType in ["S","DTmN","T"]:
            unsigned = False
        if row[i].val is None:
            nNones += 1
            SumMinNones += row[i].valMin
            SumMaxNones += row[i].valMax
        else:
            Sum += row[i].val
            SumMin += row[i].valMin
            SumMax += row[i].valMax
            SumErr += row[i].sqerr
            SumSqDev += row[i].calcDev()
        if not row[i].userDefined:
            indexLastUndefined = i
            nUndefined += 1

            
    sMin = y.valMin - SumMax #sMin/sMax is the minimum and maximum value of the residual to be distributed to None-values
    sMax = y.valMax - SumMin
    if unsigned:
        sMin = max(sMin,0)
        sMax = max(sMax,0)

    if DEBUG in ["ALL", "ADJUST"]:
        print "AdjustRowSum(y,row)________________________________"
        y.show()
        for i in range(m):
            row[i].show()  
        print "___________________________________________________"
        print "adjRowSum: Sum of not Nones =", Sum, "SumErr=", SumErr, "SumSqDev=", SumSqDev, "NNones = ", nNones
        print "adjRowSum: sMin=", sMin, "sMax=", sMax

    # Special case 0. Sum is 0 and all x are unsigned -> set all to 0.
    if iszero(y) and unsigned:   #special case SUM = 0, valid if all numbers are positive
        for i in range(m):
            row[i].val = 0
            row[i].sqerr = 0
            row[i].sqdev = 0
            updated[i] = True

        diff = abs(Sum - y.val)

    # Case 1: some of the x are still None
    elif (nNones > 0):

        newSqDev = y.sqdev + pow(sMax, 2)
        if (y.val is not None):
            diffMinMaxNones = SumMaxNones - SumMinNones
            if diffMinMaxNones > 0 and diffMinMaxNones < 0.1*INFINITE:
                gamma = ((y.val - Sum) - SumMinNones) / diffMinMaxNones
            else:
                gamma = 0.0
##            sEst = float (y.val - Sum) / nNones
##            if DEBUG in ["ALL", "ADJUST"]:
##                print "adjustRowSum: (nNones=%s) sEst = " % nNones, sEst
##                
##            if (sEst > 0):
##                newSqerr = newSqDev / pow(sEst, 2)
##            else:
##                sEst = 0.0
##                newSqerr = INFINITE
        else:
##            sEst = None
            gamma = 0.0
##            newSqerr = INFINITE
       
        for i in range(m):
            if row[i].val is None:

                updated[i] = True

##                if (sEst is not None):
##                    row[i].val = min(sEst, row[i].valMax)
##                    row[i].val = max(row[i].val, row[i].valMin)
                if gamma > 0: #calculation time saver only (would also work without if)
                    row[i].val = row[i].valMin + gamma*(row[i].valMax - row[i].valMin)
                
##                if nNones == 1 or unsigned:
##                    row[i].valMax = min(row[i].valMax, sMax)
##                    row[i].valMin = max(row[i].valMin, sMin)
##                #also for nNones > 1 and not unsigned min/max values could be calculated, but this is done below anyway    

                    row[i].sqdev = newSqDev
                    row[i].calcErr()
                        
        diff = nNones

    # Case 2. exactly one value is not user defined -> calculate by difference
    elif (nUndefined == 1 and y.val is not None):         # special case: only one xi is not a user defined value ...

        sumDefined = 0
        sumDefinedMin = 0
        sumDefinedMax = 0
        sumDefinedSqDev = 0
        for i in range(m):
            if row[i].userDefined:
                sumDefined += row[i].val
                sumDefinedMin += row[i].valMin
                sumDefinedMax += row[i].valMax
                sumDefinedSqDev += row[i].sqdev
                
        row[indexLastUndefined].val = y.val - sumDefined
        row[indexLastUndefined].userDefined = y.userDefined
        row[indexLastUndefined].sqdev = y.sqdev + sumDefinedSqDev
        row[indexLastUndefined].calcErr()

        updated[indexLastUndefined] = True

        diff = nNones

    # Case 3. general case, all values are available: -> adjust values (no distinction between user defined or not
    else:
# first adjustment
        newDev = [0]*m
        
        if SumErr != 0:
            for i in range(m):
                
##                newValMin = y.valMin
##                newValMax = y.valMax
##                for j in range(m):
##                    if not j == i:
##                        newValMin -= row[j].valMax
##                        newValMax -= row[j].valMin
##              
##                newValMin = max(newValMin, row[i].valMin)
##                newValMax = min(newValMax, row[i].valMax)
##
##                row[i].valMax = newValMax
##                row[i].valMin = newValMin

                if y.val is not None:
                    newval = row[i].val + ((row[i].sqdev / SumSqDev) * (y.val - Sum))
                    newval = max(newval, row[i].valMin)
                    newval = min(newval, row[i].valMax)
                    newval = min(newval, y.val)  #no value can be bigger than the sum.
                    if DEBUG in ["ALL", "ADJUST"]:
                        print "In adjRowSum newval = ", newval, " substitutes "
                        row[i].show()
                        
                    shift = newval - row[i].val

                    if (abs(shift) > 1.e-2*row[i].dev): updated[i] = True # consider an update more than 1 % of current imprecision

                    if not isequal(row[i].val,newval):
                        row[i].userDefined = False
                    row[i].val = newval

                    newDev[i] = row[i].dev + abs(shift)
                                
            for i in range(m):
                if y.val is not None:
                    maxDev = pow(y.sqdev, 0.5)
                    for j in range(m):
                        if not j == i:
                            maxDev += newDev[j]

                    row[i].dev = min(newDev[i], maxDev)
                    row[i].sqdev = pow(row[i].dev, 2.0)
                    row[i].calcErr()

                    if DEBUG in ["ALL", "ADJUST"]:
                        print "... to new value:"
                        row[i].show()
                        
        elif not isequal(Sum, y.val) and y.val is not None:

            print "====================================================="
            print "WARNING"
            print "In adjRowSum sumErr=0 and Sum != y. Check consistency"
            print "SumErr = ", SumErr, " Sum=", Sum, " y = ", y.val
            y.show()
            for i in range(m):
                row[i].show()  
            print "====================================================="

        if y.val is not None:
            diff = abs(Sum - y.val)
        else:
            diff = 1.0

# second pass -> try to adjust the balance

        Sum = 0.0
        for i in range(m):
            if row[i].val is None:
                nNones += 1
            else:
                Sum += row[i].val

#        dummy = None                
#        if dummy is not None:
        if y.val is not None:
            newdiff = Sum - y.val
            if newdiff != 0:
                for i in range(m):
                    newval = row[i].val - newdiff
                    newval = max(newval, row[i].valMin)
                    newval = min(newval, row[i].valMax)
                    shift = row[i].val - newval
                    newdiff -= shift

                    if not isequal(row[i].val,newval):
                        row[i].userDefined = False
                    row[i].val = newval
                    row[i].dev += abs(shift)
                    row[i].sqdev = pow(row[i].dev, 2.0)
                    row[i].calcErr()

    for i in range(m):        
        row[i].constrain()

    if y.val is not None:
        newtrack = []
        for i in range(m):
            newtrack.append([])
            newtrack[i].extend(y.track)
            for j in range(m):
                if not (i == j) and row[j].val is not None:
                    newtrack[i].extend(row[j].track)
        for i in range(m):
            if updated[i]:
                row[i].track = []
                row[i].track.extend(newtrack[i])
                row[i].cleanUp()

    if DEBUG in ["ALL", "ADJUST"]:
        print "results:___________________________________________"
        y.show()
        for i in range(m):
            row[i].show()  
        print "___________________________________________________"
        
    if stack:
        for i in range(len(row_old)):
            row_old[i].push(row[i])
        
    return diff     #returns the value of the adjustment difference for iteration control

#------------------------------------------------------------------------------
def equationRowSum(y, row_old, m = None):
    """
    Like adjRowSum_stack, but adjusts both the row and the sum (y)
    
    :param y: The known sum
    :param row_old: a list of CCPar summands
    :param m: the number of items in row_old that are to be adjusted.
      the first m items are adjusted.
    """

    zero = CCPar("0")
    zero.setValue(0.0)

    ny = calcNegative("-"+y.name,y)

    newRow = [ny] + row_old

    adjRowSum_stack(zero, newRow, m+1)

    ny = newRow[0]
    y.push(calcNegative(y.name, ny))
    
#------------------------------------------------------------------------------
def adjustDiff_stack(y, x1_old, x2_old, GTZero=False):
    """
    Adjust two values of which the difference is known.
    
    The new adjusted values are pushed on their respective stacks.
    If y, x1_old and x2_old are lists or tuples with common length 
    adjustDiff_stack is applied for each triple of entries.
    
    :param y: known difference
    :param x1_old: the minuend
    :param x2_old: the subtrahend

    """
    # are the input parameters list/tuples of compatible length? 
    if (all(isinstance(para, (type([]), type(()))) for para in (y, x1_old, x2_old))
         and len(y) == len(x1_old) == len(x2_old)):
        for i in xrange(len(y)):
            adjustDiff_stack(y[i], x1_old[i], x2_old[i])
    else: # no they are not:
        x1 = CCPar(x1_old.name)
        x2 = CCPar(x2_old.name)
        x1.update(x1_old)
        x2.update(x2_old)
        
        if DEBUG in ["ALL", "ADJUST"]:
            print "___________________________________________________"
            print "adjustDiff starting", x1.name, x2.name
            y.show()
            x1.show()
            x2.show()
        
        if y.val is not None:
    
            y.calcDev()
            x1.calcDev()
            x2.calcDev()
    
            if iszero(y):   #special case SUM = 0
                        # in this case holds x1 = x2                    
                if GTZero == False:
                    ccheck1(x1, x2)
    
            elif (x1.val is not None) and ((x1.sqdev < x2.sqdev) or x2.val == None):
              
                x2.val = x1.val - y.val
                x2.userDefined = x1.userDefined and y.userDefined
    
                x2.dev = y.dev + x1.dev
                x2.sqdev = pow(x2.dev, 2.0)
                x2.calcErr()
                    
            elif x2.val is not None:
                
                x1.val = y.val + x2.val
                x1.userDefined = x2.userDefined and y.userDefined
                x1.dev = y.dev + x2.dev
                x1.sqdev = pow(x1.dev, 2.0)
                x1.calcErr()
                
            elif x1.val is None and x2.val is None:
                diffMin = x1.valMin - x2.valMax
                diffMax = x1.valMax - x2.valMin
                if diffMax > diffMin:
                    gamma = (y.val - diffMin) / (diffMax - diffMin)
                else:
                    gamma = 0
                    
                x1.val = x1.valMin + gamma * (x1.valMax - x1.valMin)
                x2.val = x2.valMax - gamma * (x2.valMax - x2.valMin)
                
                newDev1 = y.dev + x2.dev
                newDev2 = y.dev + x1.dev

                x1.sqdev = pow(x1.dev, 2.0)
                x2.sqdev = pow(x2.dev, 2.0)
                x1.calcErr()
                x2.calcErr()

    
        if DEBUG in ["ALL", "ADJUST"]:
            print "results before constrain and min/max adjust:___________________________________________"
            y.show()
            x1.show()
            x2.show()
            print "___________________________________________________"
    
        newMin = min(y.valMin + x2.valMin, INFINITE)
        newMax = min(y.valMax + x2.valMax, INFINITE)
    
        x2.valMin = x1.valMin - y.valMax
        x2.valMax = x1.valMax - y.valMin
    
        if x2.parType not in ["S","DTmN","T"]:
            x2.valMin = max(x2.valMin, 0)
            x2.valMax = max(x2.valMax, 0)
    
        x1.valMin = newMin
        x1.valMax = newMax
    
        x1.constrain()
        x2.constrain()
    
        if y.val is not None:
            newtrackx1 = []
            newtrackx1.extend(y.track)
            if x2.val is not None:
                newtrackx1.extend(x2.track)
    
            newtrackx2 = []
            newtrackx2.extend(y.track)
            if x1.val is not None:
                newtrackx2.extend(x1.track)
    
            x1.track = newtrackx1
            x2.track = newtrackx2
    
            x1.cleanUp()
            x2.cleanUp()
    
        if DEBUG in ["ALL", "ADJUST"]:
            print "results:___________________________________________"
            y.show()
            x1.show()
            x2.show()
            print "___________________________________________________"
    
        x1_old.push(x1)
        x2_old.push(x2)


def adjustFlow_stack(Qdot, cp, m, T1, T0, DT):
    """
    Adjust values of which the flow is known.
    
    The new adjusted values are pushed on their respective stacks.
    If all parameters are lists or tuples adjustFlow_stack is applied
    for each 6-tuple of entries.
    """
    # are the input parameters list/tuples of compatible length? 
    if all((isinstance(para, (type([]), type(()))) 
           for para in (Qdot, cp, m, T1, T0, DT))):
        for i in xrange(len(Qdot)):
            adjustFlow_stack(Qdot[i], cp[i], m[i], T1[i], T0[i],
                             DT[i])
    else: # no they are not:
        
        adjustDiff_stack(DT, T1, T0)
        adjustProdC_stack(Qdot, cp, m, DT)
    
        if DEBUG in ["ALL", "CALC"]:
            print "CalcFlow-Qdot______________________________________"
            m.show()
            print "cp = ", cp
            print "Qdot = m*cp*DT = "
            Qdot.show()
            print "___________________________________________________"
            print "adjustFlow-T1 y T0_________________________________"
            T1.show()
            T0.show()
            print "DT = T1 - T0 = "
            DT.show()
            print "___________________________________________________"
            
#------------------------------------------------------------------------------
def adjustH_stack(y, cL, cV, h, T0, x1_old, x2_old):
    """ Default function for adjusting the enthalpy h
    """
    x1 = CCPar(x1_old.name)
    x2 = CCPar(x2_old.name)
    x1.update(x1_old)
    x2.update(x2_old)

    if DEBUG in ["ALL", "ADJUST"]:
        print "___________________________________________________"
        print "adjustH starting", x1.name, x2.name
        y.show()
        print "CL= ", cL
        x1.show()
        x2.show()

    h0 = enthalpy(cL, cV, h, T0, T0, 0.0)
    h1 = enthalpy(cL, cV, h, T0, T0, 1.0)
    
    if not (y.val == None):

        dev = 2*pow(y.calcDev(), 0.5) #TODO: factor 2 added 13.11.2014 as a hack in order to overcome problems with recent CC-update
        yL = max(y.val - dev, y.valMin)
        yH = min(y.val + dev,y.valMax)
        
        (T, x) = invEnthalpy(cL, cV, h, T0, y.val)            
        (TL, xL) = invEnthalpy(cL, cV, h, T0, yL)            
        (TH, xH) = invEnthalpy(cL, cV, h, T0, yH)

        devT = max(T - TL, TH - T)
        devx = xH - xL

        x1.val = T
        x1.userDefined = y.userDefined
        x1.sqdev = pow(devT, 2.0)
        x1.calcErr()

        x2.val = x
        x2.userDefined = y.userDefined
        x2.sqdev = pow(devx, 2.0)
        x2.calcErr()

    (Tmin, xmin) = invEnthalpy(cL, cV, h, T0, y.valMin)
    (Tmax, xmax) = invEnthalpy(cL, cV, h, T0, y.valMax)

    x1.valMin = max(min(Tmin - NUMERIC_ERR, INFINITE), x1.valMin)
    x1.valMax = min(Tmax + NUMERIC_ERR, INFINITE, x1.valMax)
    
    x1.constrain()
    
    x2.valMin = max(min(xmin, INFINITE), 0.0)
    x2.valMax = min(xmax, 1.0)

    x2.constrain()

    if y.val is not None:
        x1.track = []
        x1.track.extend(y.track)

        x1.cleanUp()

    if DEBUG in ["ALL", "ADJUST"]:
        print "adjustH________________________________________"
        y.show()
        x1.show()
        x2.show()
        print "___________________________________________________"
        
    x1_old.push(x1)
    x2_old.push(x2)

#------------------------------------------------------------------------------
def calcConstraintLT(y, xlim):
#------------------------------------------------------------------------------
#   Changes y so that y <= xlim
#------------------------------------------------------------------------------

    # FIXME: check calc/adjConstraintLT/GT comments, directions 

    if DEBUG in ["ALL", "CALC"]:
        print "constrainLT(x1,x2)_________________________________"
        y.show()

    y.valMin = min(y.valMin, xlim.valMax)
    y.valMax = min(y.valMax, xlim.valMax)
    y.val = min(y.val, xlim.val)
    y.constrain()

    if DEBUG in ["ALL", "CALC"]:
        xlim.show()
        print "y (constrained) = "
        y.show()
        print "___________________________________________________"

    return y

#------------------------------------------------------------------------------
def adjConstraintLT(y, xlim):
#------------------------------------------------------------------------------
#   Changes xlim so that y <= xlim
#------------------------------------------------------------------------------
    
    calcConstraintGT(xlim, y)

#------------------------------------------------------------------------------
def calcConstraintGT(y, xlim):
#------------------------------------------------------------------------------
#   Changes y so that y >= xlim
#------------------------------------------------------------------------------

    if DEBUG in ["ALL", "CALC"]:
        print "constrainLT(x1,x2)_________________________________"
        y.show()

    y.valMin = max(y.valMin, xlim.valMin)
    y.valMax = max(y.valMax, xlim.valMin)
    y.val = max(y.val, xlim.val)
    y.constrain()

    if DEBUG in ["ALL", "CALC"]:
        xlim.show()
        print "y (constrained) = "
        y.show()
        print "___________________________________________________"

    return y

#------------------------------------------------------------------------------
def adjConstraintGT(y, xlim):
#------------------------------------------------------------------------------
#   Changes y so that y <= xlim
#------------------------------------------------------------------------------
    calcConstraintLT(xlim, y)

#------------------------------------------------------------------------------
def adjConstraintGT_stack(y, xlim_old):
#------------------------------------------------------------------------------
#   Changes y so that y <= xlim
#------------------------------------------------------------------------------
    xlim = CCPar(xlim_old.name)
    xlim.update(xlim)
    xlim_old.push(calcConstraintLT(xlim, y))

def isequal(a, b):
    """Return whether the absolute difference between a and b is smaller than 
    EPSILON.
    
    """
    try:
        if (fabs(a - b) < EPSILON):
            return True
        else:
            return False
    except TypeError:
        if None in (a, b):
            return False
        else:
            raise
        
def iszero(x):
#------------------------------------------------------------------------------
#   checks if a value is exactly zero
#------------------------------------------------------------------------------

    try:
        if x.val == 0 and x.sqerr == 0:
            return True
        else:
            return False

    except Exception:
        return False
        
def countNones(row):
    """Count the number of None object in a list.
    
    Return the absolute number as well as a list of booleans, where for
    every entry in row is a True for a None entry and a False else.
    
    :param row: list to inspect.
    :returns: a pair (number of Nones in the list, incidence list)

    """
    isnone = [item.val is None for item in row]
    noneCounter = 0
    for item in row:
        if (item.val is None):
            noneCounter += 1            
    return (noneCounter, isnone)
