# -*- coding: cp1252 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CCheck (Consistency Check)
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"

from math import *
from ccheckFunctions import *
from ccPar import CCPar, CCOne, CCRow
from modules.fluids import *
from modules.messageLogger import *
from modules.chiller.moduleCH import DEFAULT_COPEX_WATERCOOLED, DEFAULT_COPEX_AIRCOOLED, CHTDROP_COND_WATERCOOLED, CHTDROP_COND_AIRCOOLED
from modules.heatPump.moduleHP import DEFAULT_COPEX_HP, HPTDROP
from modules.chp.moduleCHP import DEFAULT_CPARTLOAD, DEFAULT_EEEMIN, DEFAULT_COPEX, COPCarnot
import database.dbio as dbio
from auxiliary.auxiliary import noneCheck

#libraries necessary for SQL access:
from GUI.status import *

#------------------------------------------------------------------------------
class CheckEq():
#------------------------------------------------------------------------------
#   Carries out consistency checking for equipe j
#------------------------------------------------------------------------------

    def __init__(self, j, eq=None, checkInPlace=False):
        """Setup the check with the data for equipment nr. j
        
        :param checkInPlace: If False take the data from ANo -1 and
        write it to ANo 0, otherwise work on the current alternative proposal. 
        
        """     
        if eq is None:
            eq = Status.int.systemModel.eqObject[j]

        j = eq.no
        self.eq = eq
        
        self.checkInPlace = checkInPlace
        self.exportToDB = not checkInPlace
        
        self.index = j
        self.dataGroup = "Equipment"

# assign a variable to all intermediate/calculated values needed

        self.HPerYearEq = CCPar("HPerYearEq", unitType = "TIME")
        self.HPerYearEq.valMax = 8760
        self.HPerYearEqNom = CCPar("HPerYearEqNom", unitType = "TIME")
        self.HPerYearEqNom.valMax = 8760
        
        self.USHConv = CCPar("USHConv",priority=2, unitType = "ENERGY")
        
        self.USHj = CCPar("USHj",priority=2, unitType = "ENERGY")
        self.dotUSHj = CCPar("dotUSHj", unitType = "POWER")
        self.dotQHXEq = CCPar("dotQHXEq", unitType = "POWER")

        self.dotFETj = CCPar("dotFETj", parType="S", unitType = "POWER")   # from/to the FET matrix
        self.dotFETel_j = CCPar("dotFETel_j",parType = "S", unitType = "POWER")


        self.ConvLoss = CCPar("ConvLoss")
        self.FETFuelRatio = CCPar("FETFuelRatio")
        self.FETelRatio = CCPar("FETelRatio")
        self.QConvLoss = CCPar("QConvLoss", unitType = "ENERGY")
        self.dotQConvLoss = CCPar("dotQConvLoss", unitType = "POWER")
        self.One = CCOne()
        
        self.FlowCombAir = CCPar("FlowCombAir", unitType = "MASSFLOW")#
        self.FlowExhaustGas = CCPar("FlowExhaustGas", unitType = "MASSFLOW")#
        
        self.DTExhaustGas = CCPar("DTExhaustGas",parType="DT")#
        self.DTOffGasUse = CCPar("DTOffGasUse",parType="DT")#
        self.OffGasUse = CCPar("OffGasUse",parType="X")
        self.dotQExhaustGas = CCPar("dotQExhaustGas", unitType = "POWER")#
        self.QExhaustGas = CCPar("QExhaustGas", unitType = "ENERGY")#

        
        self.QWHEqTot = CCPar("QWHEqTot", unitType = "ENERGY")#
        self.dotQWHEqTot = CCPar("dotQWHEqTot", unitType = "POWER")

        self.QLossEq = CCPar("QLossEq", unitType = "ENERGY")#
        self.dotQLossEq = CCPar("QdotLossEq", unitType = "POWER")
        self.QInputEq = CCPar("QInputEq", unitType = "ENERGY")#
        self.QOutEq = CCPar("QOutEq", unitType = "ENERGY")#

        self.dotQOutEq = CCPar("dotQOutEq", unitType = "POWER")#
        self.dotQInputEq = CCPar("dotQInputEq", unitType = "POWER")#

        self.PL1 = CCPar("PL1",parType="X")
        self.PartLoadFact = CCPar("PartLoadFact",parType = "X")
        self.PartLoadFact.valMin = 0.9

        self.TCorrFact = CCPar("TCorrFact")

        # exhaust gas ratios
        self.ExhaustFact = CCPar("ExhaustFact") # =: QExhaustGas / FETFuel
        self.CombAirExcess = CCPar("CombAirExcess") # =: CombAir*ExcessAirRatio
        self.OffGasRatio = CCPar("OffGasRatio") # = 1 + CombAirExcess
        
        self.QHXRatio = CCPar("QHXRatio")
        self.QHXRatio.valMax = 0.5 # TODO: an upper limit is needed unless there is a check whether a HX is connected or not
        
        self.USHRatio = CCPar("USHRatio")
        self.FETRatio = CCPar("FETRatio")

        self.LossFactFET = CCPar("LossFactFET")
        self.VesselLoss = CCPar("VesselLoss")

        
        #coefficients of equation PartLoadFact = 0.9 + 0.1* average PartLoad
        self.PLC0 = CCPar("PLC0", unitType="FRACTION")
        self.PLC0.setValue(0.9,err=0.0)
        self.PLC0.valMin = 0.9
        self.PLC0.valMax = 0.9
        
        self.PLC1 = 0.1 #constant used in calcK, no need for CCPar

        self.partLoadCorrection = False #by default, activated only for chillers
        self.temperatureCorrection = False #by default, activated only for chillers

        self.condensing = False # deactivates energy balance from ExhaustGas
        self.exhaustGas = False
          

        self.OpCostVar = CCPar("OpCostVar", unitType = "PRICE")
        self.OpCostTot = CCPar("OpCostTot", unitType = "PRICE")

        self.importData(j)

        if DEBUG in ["ALL","BASIC"]:
            self.showAll()

#------------------------------------------------------------------------------
    def importData(self, j): 
        """Import process data from SQL tables. 
        
        The data is taken from AlternativeProposalNo = -1 or the current one
        depending on whether to check in place or not. 
        """
        eq = self.eq
        self.EqNo = j+1
        if self.checkInPlace is True:
            ANo = Status.ANo
        else:
            ANo = -1

        # assign empty CCPar to all questionnaire parameters
        self.HCGPnom = CCPar("HCGPnom",priority=2, unitType = "POWER")
        self.ElectriProduction = CCPar("ElectriProduction",priority=2, unitType = "POWER")
        
        self.FuelConsum = CCPar("FuelConsum", unitType = "MASSFLOW")
        self.FuelConsumFL = CCPar("FuelConsumFL", unitType = "MASSFLOW")
        self.FuelConsumNom = CCPar("FuelConsumNom", unitType = "MASSFLOW")
        self.ElectriConsum = CCPar("ElectriConsum", unitType = "POWER")
        self.ElectriConsumFL = CCPar("ElectriConsumFL", unitType = "POWER")
        self.ElectriConsumNom = CCPar("ElectriConsumNom", unitType = "POWER")
        self.ThermalConsumNom = CCPar("ThermalConsumNom", unitType = "POWER")
        
        self.NDaysEq = CCPar("NDaysEq")
        self.HPerDayEq = CCPar("HPerDayEq", unitType = "TIME")
        self.PartLoad = CCPar("PartLoad",parType="X")
        
        self.HCGTEfficiency = CCPar("HCGTEfficiency",priority=2)
        self.HCGTEfficiencyFL = CCPar("HCGTEfficiencyFL")
        self.HCGTEfficiencyNom = CCPar("HCGTEfficiencyNom",priority=2)
        
        self.HCGEEfficiency = CCPar("HCGEEfficiency")


        self.TExhaustGas = CCPar("TExhaustGas", parType="T")#
        self.ExcessAirRatio = CCPar("ExcessAirRatio")#
        self.ExcessAirRatio.valMin = 1.0
        self.ExcessAirRatio.valMax = 2.0

        self.FETj = CCPar("FETj",priority=2, parType="S", unitType = "ENERGY")   # from/to the FET matrix
        self.FETel_c_j = CCPar("FETel_c_j", unitType = "ENERGY")
        self.FETel_j = CCPar("FETel_j",priority=2,parType="S", unitType = "ENERGY")
        self.ElGen_j = CCPar("ElGen_j",priority=2, unitType = "ENERGY")

        self.FETFuel_j = CCPar("FETFuel_j", unitType = "ENERGY")
        self.dotFETFuel_j = CCPar("dotFETFuel_j", unitType = "POWER")
        
        self.FETHeat_j = CCPar("FETHeat_j", unitType = "ENERGY")

        self.QDrive = CCPar("QDrive", unitType = "ENERGY")
        self.dotQDrive = CCPar("dotQDrive", unitType = "POWER")
        self.TDrive = CCPar("TDrive", parType="T")
        
        self.QHXEq = CCPar("QHXEq",priority=2, unitType = "ENERGY") #from/to the heat recovery matrix
        self.QWHEq = CCPar("QWHEq", unitType = "ENERGY")  #from/to the heat recovery matrix  #

        self.TEnvEq = CCPar("TEnvEq", parType="T")    #average ambient temp. Not defined into questionnaire as generic parameter but only into ST questionnaire
        self.LossFactEq = CCPar("LossFactEq",parType="X")  # %of losses with respect to USHj (not to USHConv)
        self.TMinOffGas = CCPar("TMinOffGas",parType = "T")   # percentage of OffGas that can be used as waste heat

        self.TSupply = CCPar("TSupply", parType="T")
        self.TemperatureReCooling = CCPar("TemperatureReCooling", parType="T")
        self.THeatSourceLT = CCPar("THeatSourceLT", parType="T")
        

        self.COPex = CCPar("COPex", parType="X")
        self.WaterConsum = CCPar("WaterConsum", unitType = "VOLUME")

        self.dryTower = False # default, parameter only relevant for water cooled chillers

        self.TurnKeyPrice = CCPar("TurnKeyPrice")
        
        self.OandMfix = CCPar("OandMfix", priority=2, unitType = "PRICE")
        self.OandMvar = CCPar("OandMvar", priority=2)
        
#..............................................................................
# reading data from eqObject
        
        self.equipmentType = self.eq.eqType
        self.equipmentClass = self.eq.eqClass
        
        if self.equipmentType in ["thermal chiller (water cooled)",
                                  "thermal chiller (air cooled)",
                                  "compression chiller (air cooled)",
                                  "compression chiller (water cooled)"]:
            self.partLoadCorrection = True
            self.temperatureCorrection = True

            if self.equipmentType in ["thermal chiller (water cooled)",
                                      "compression chiller (water cooled)"]:
                CTType = self.eq.DestinationWasteHeat
                if CTType == "cooling tower (dry)":
                    self.dryTower = True
                else:
                    self.dryTower = False

        elif self.equipmentType in ["compression heat pump",
                                    "mechanical vapor compression"]:
            self.PartLoadFact.setValue(1.0,err=0.0,setTrack = False)
            self.temperatureCorrection = True
        else:
            self.PartLoadFact.setValue(1.0,err=0.0,setTrack = False)
            self.TCorrFact.setValue(1.0,err=0.0, setTrack = False)

        self.heatSourceHT = False

        fuel_number = self.eq.DBFuel_id   #IMPORT from the fuelDB

        if fuel_number <= 0 or fuel_number is None or self.eq.mdotFuel == 0:
            if self.equipmentType in ["thermal chiller (water cooled)",
                                      "thermal chiller (air cooled)",
                                      "CHP ORC",
                                      "CHP TFC"]:
                self.mainSource = "Heat"
                self.heatSourceHT = (self.eq.HeatSourceHT is not None)
            else:
                self.mainSource = "Electricity"
            self.FuelLCV = 0.0
            self.OffgasHeatCapacity = 1.0 #
            self.CpToLCV = 0.0
            self.CombAir = 0.0 #
            self.FuelConsum.setValue(0.0, setTrack = False)
            self.FETFuel_j.setValue(0.0, setTrack = False)
            
            self.TExhaustGas.setValue(18.0, err=0.0)          #

            if self.equipmentType in ["CHP ORC",
                                      "CHP TFC"]:
                Pel = noneCheck(self.eq.Pel, 0)
                self.ElectriConsumNom.setValue(Pel)
            else:
                self.ElectriConsumNom.setValue(self.eq.Pel)

            self.ExcessAirRatio.val = 1.0
            self.TMinOffGas.setValue(18.0,err = 0.0)
            self.exhaustGas = False
        else:
            self.mainSource = "Fuel"
            eq_fuel = Fuel(fuel_number)
            self.FuelLCV = eq_fuel.LCV
            self.OffgasHeatCapacity = eq_fuel.OffgasHeatCapacity #
            try:
                self.CpToLCV = self.OffgasHeatCapacity/self.FuelLCV
            except:
                self.CpToLCV = 0.0
                
#                    print "heat capacity = ",self.OffgasHeatCapacity
            self.CombAir = eq_fuel.CombAir #

            self.FuelConsumNom.setValue(self.eq.mdotFuel)
                
            if self.eq.Pel is None:
                logWarning("CheckEq (importData): no electricity consumption specified for eq.no. %s. Zero assumed"%(j+1))
                self.ElectriConsumNom.setValue(0)
            else:
                self.ElectriConsumNom.setValue(self.eq.Pel)                        

            self.TExhaustGas.setValue(self.eq.TExhaustGas,err = 0.01) #high error. gives problems
            TCond = max(eq_fuel.TCondOffGas(),0)
            if self.equipmentType in ["burner (direct heating)", "burner (indirect heating)"]:
                self.TMinOffGas.setValue(18.0,err = 0.0)
            else:
                self.TMinOffGas.setValue(TCond,err = 0.0)
                
            if (self.eq.ExcessAirRatio is None) and (self.eq.TExhaustGas is None):
                self.ExcessAirRatio.setValue(1.50,err = 0.0)
            else:
                if self.eq.ExcessAirRatio is not None: #extend limits if necessary, for direct combustion burners !!!!
                    self.ExcessAirRatio.valMax = max(self.eq.ExcessAirRatio,self.ExcessAirRatio.valMax)
                self.ExcessAirRatio.setValue (self.eq.ExcessAirRatio)
            self.exhaustGas = True
                

        self.HCGPnom.setValue(self.eq.Pnom,err = 0.0)  #if specified, take as fixed

        if ANo == -1: # for local checking in alternatives those values are calculation results, not inputs
            self.NDaysEq.setValue(self.eq.NDays, err=0.0) #integer -> 0 error
            self.HPerDayEq.setValue(self.eq.HPerDay, err = 0.0) #if specified, take as fixed
            self.PartLoad.setValue(self.eq.PartLoad)
        
        self.TEnvEq.setValue(18.0, err=0.0)          #
        if self.equipmentType in ["burner (direct heating)",
                                  "burner (indirect heating)"]:
            self.LossFactEq.setValue(0.0)  #  heat losses of vessel supposed between 0 and 2 %
        elif self.equipmentClass in ["CHP", "CH", "ST"]:
            self.LossFactEq.setValue(0.0)
        else:
            self.LossFactEq.setValue(0.01, err=1.0)  #  heat losses of vessel supposed between 0 and 2 %
            self.LossFactEq.valMax = 0.02
            self.LossFactEq.userDefined = True # this is not done automatically by setValue if err=1.0

        # here equipment type specific checking of input data
        
        self.HCGTEfficiencyNom.setValue(self.eq.HCGTEfficiency)

        if self.equipmentClass == "CH":
            if self.equipmentType == "compression chiller (water cooled)":
                self.FETelChiller = CCPar("FETelChiller", unitType = "ENERGY")
                self.FETelCT = CCPar("FETelCT", unitType = "ENERGY")

                self.PelChiller = CCPar("PelChiller", unitType = "POWER")
                self.PelCT = CCPar("PelCT", unitType = "POWER")

                self.SystemEfficiency = CCPar("SystemEfficiency", unitType = "FRACTION")

##                self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, INFINITE)

            self.TSupply.setValue(self.eq.TSupply)
            self.TemperatureReCooling.setValue(self.eq.TemperatureReCooling)
            self.COPex.setValue(self.eq.COPex)

            if self.COPex.val is None and self.HCGTEfficiencyNom.val is None:
                if self.equipmentType == "compression chiller (water cooled)":
                    estimateTracker.setDataGroup("Equipment",self.EqNo)
                    self.COPex.setEstimate(DEFAULT_COPEX_WATERCOOLED)

                elif self.equipmentType == "compression chiller (air cooled)": # calculation of COP for air cooled compression chillers
                    estimateTracker.setDataGroup("Equipment",self.EqNo)
                    self.COPex.setValue(DEFAULT_COPEX_AIRCOOLED)

            self.userDefinedWorkingPoint = True
            if self.eq.TemperatureReCooling is None:
                estimateTracker.setDataGroup("Equipment",self.EqNo)
                self.TemperatureReCooling.setEstimate(35.0)
                self.userDefinedWorkingPoint = False

            if self.eq.TSupply is None:
                estimateTracker.setDataGroup("Equipment",self.EqNo)
                self.TSupply.setEstimate(0.0)
                self.userDefinedWorkingPoint = False

            if self.equipmentType == "compression chiller (water cooled)":
                TDropCond = CHTDROP_COND_WATERCOOLED
            else:
                TDropCond = CHTDROP_COND_AIRCOOLED
                
            self.EERCarnot = Status.mod.moduleCH.EERCarnot(self.TemperatureReCooling.val,
                                                           self.TSupply.val,
                                                           TLevels="external",
                                                           TDropCond=TDropCond)
            
            self.fRecool = Status.mod.moduleCH.getFRecool(self.equipmentType, self.dryTower)

            

        elif self.equipmentClass == "HP":
            self.TSupply.setValue(self.eq.TSupply)
            self.THeatSourceLT.setValue(self.eq.THeatSourceLT)
            self.COPex.setValue(self.eq.COPex)

            if self.COPex.val is None and self.HCGTEfficiencyNom.val is None:
                if self.equipmentType in  ["compression heat pump", "mechanical vapour recompression"]:
                    estimateTracker.setDataGroup("Equipment",self.EqNo)
                    self.COPex.setEstimate(DEFAULT_COPEX_HP)

            self.userDefinedWorkingPoint = True
            if self.eq.THeatSourceLT is None:
                estimateTracker.setDataGroup("Equipment",self.EqNo)
                self.THeatSourceLT.setEstimate(20.0)
                self.userDefinedWorkingPoint = False

            if self.eq.TSupply is None:
                estimateTracker.setDataGroup("Equipment",self.EqNo)
                self.TSupply.setEstimate(60.0)
                self.userDefinedWorkingPoint = False

            if self.TSupply.val is not None and self.THeatSourceLT.val is not None:
                TDropCond = CHTDROP_COND_AIRCOOLED
                TDrop = 0.0 if self.equipmentType=="mechanical vapour recompression" else HPTDROP
                Th = self.TSupply.val + TDrop
                Tc = self.THeatSourceLT.val - TDrop
                self.COPCarnot = Status.mod.moduleHP.calculateCOPh_Carnot(Th, Tc)
                self.EERCarnot = self.COPCarnot - 1.0
##                self.EERCarnot = Status.mod.moduleCH.EERCarnot(self.TSupply.val,
##                                                               self.THeatSourceLT.val,
##                                                               TLevels="external",
##                                                               TDropCond=TDropCond)
            else:
                self.COPCarnot = None
                self.EERCarnot = None
                
##                self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, INFINITE)
            pass
        elif self.equipmentClass in ["BB"]:
            if self.mainSource == "Electricity":    
##                    self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, INFINITE)  #TODO: this should be 1, but currently an electrical boiler with eta > 1 is often used as workaround for a heat pump.
                self.exhaustGas = False
                self.condensing = True # avoid calculations that are not valid for eff > 1 (-> allow for workaround of HP modeling as electrical boiler with COP > 1 !!)
            else:
                if self.equipmentType not in ["condensing boiler", "burner (direct heating)", "burner (indirect heating)"]:
                    self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, 1.0)
                    if self.HCGTEfficiencyNom.val is not None:
                        if self.HCGTEfficiencyNom.val > 0.97 and self.checkInPlace:
                            showWarning(_("The conversion efficiency entered is very high (> 0.97) and may be inconsistent with minimum losses by exhaust gas"))
                else:
                    if self.equipmentType == "condensing boiler":
                        self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, 1.2)    #TODO: should be at maximum ratio HCV/LCV of the specific fuel

                    self.condensing = True
                    
            if self.HCGTEfficiency.val >= 0.999:
                logTrack("CheckEq (importData): boiler efficiencies > 100 % are not supported yet")
                self.condensing = True
                self.exhaustGas = False     #dont calculate exhaust gas for condensing boilers and burners. theres probably nothing more to do
            else:
                self.exhaustGas = True
                
        elif self.equipmentClass in ["ST"]:
##                self.HCGTEfficiency.valMax = INFINITE
            self.exhaustGas = False
            self.condensing = True # avoid calculations that are not valid for eff > 1
        else:
            self.HCGTEfficiency.valMax = min(self.HCGTEfficiency.valMax, 1.0)


        if ANo == -1:  # for local checking in alternatives those values are calculation results, not inputs
            if self.equipmentType == "compression chiller (water cooled)":
                self.SystemEfficiency.setValue(self.eq.HCGTEffReal)
            else:
                self.HCGTEfficiency.setValue(self.eq.HCGTEffReal)

        if self.equipmentClass == "CHP":
            self.HCGEEfficiency.setValue(self.eq.HCGEEfficiency)
            self.ElectriProduction.setValue(self.eq.ElectriProduction)
            self.exhaustGas = False     #dont calculate exhaust gas for CHP
#                    self.ElGen_j.setValue(self.eq.STAreaFactor)
            if self.mainSource == "Heat":
                if self.heatSourceHT:
                    self.FETHeat_j.setValue(self.eq.FETHeat_j)
                else: # no external heat source connected
                    self.FETHeat_j.setValue(0.0)

            if self.equipmentType in ["CHP ORC", "CHP TFC"]:
                
                self.TDrive.setValue(self.eq.THeatSourceHT)
                ccheck1(self.TDrive, self.TSupply)
                self.COPex.setValue(self.eq.COPex)

                if self.COPex.val is None and self.HCGEEfficiency.val is None and self.HCGTEfficiencyNom.val is None:
                        estimateTracker.setDataGroup("Equipment",self.EqNo)
                        self.COPex.setEstimate(DEFAULT_COPEX)

                self.userDefinedWorkingPoint = True
                if self.eq.TemperatureReCooling is None:
                    estimateTracker.setDataGroup("Equipment",self.EqNo)
                    self.TemperatureReCooling.setEstimate(20.0)
                    self.userDefinedWorkingPoint = False
                else:
                    self.TemperatureReCooling.setValue(self.eq.TemperatureReCooling)

                if self.TDrive.val is None:
                    self.userDefinedWorkingPoint = False
                
                if self.userDefinedWorkingPoint:    
                    self.COPCarnot = COPCarnot(self.TDrive.val, self.TemperatureReCooling.val)
                else:
                    self.COPCarnot = None
            
                cPartLoad = self.eq.ST_C0
                if cPartLoad is None:
                    self.eq.ST_C0 = DEFAULT_CPARTLOAD
            else:
                EEEmin = self.eq.ST_C1
                if EEEmin is None:
                    self.eq.ST_C1 = DEFAULT_EEEMIN
        else:
            self.HCGEEfficiency.setValue(0.0, setTrack = False)
            self.ElectriProduction.setValue(0.0, setTrack = False)
            self.ElGen_j.setValue(0.0, setTrack = False)

            self.FETj.parType = None
            self.FETel_j.parType = None

        self.TurnKeyPrice.setValue(self.eq.TurnKeyPrice)
        if ANo == 0:
            self.TurnKeyPrice.setValue(0.0)   # present state equipment has investment cost zero
            self.eq.TurnKeyPrice = dbio.check(self.TurnKeyPrice.val)

        self.OandMfix.setValue(self.eq.OMFix)
        self.OandMvar.setValue(self.eq.OMVar)

##        else:
##            logTrack("CheckEq(importData): didn't find table entry in self.eq for EqNo = %s"%self.EqNo)
##                

    def exportData(self):  
        """Store corrected data in SQL.
        
        Under AlternativeProposalNo = 0 or the current one depending on whether 
        to check in place or not. 
        
        """
        if self.checkInPlace is True:
            ANo = max(Status.ANo, 0)
            if Status.ANo < 0: return
        else:
            ANo = 0

        eq = self.eq

        if ANo > 0: # for inplace checking directly estimate economic parameters before saving
            self.estimateEconomicParameters()
        
        # writing data into table "qgenerationhc"

        eq.set(Pnom = dbio.check(self.HCGPnom.val),
                    mdotFuel = dbio.check(self.FuelConsumNom.val),
                    Pel = dbio.check(self.ElectriConsumNom.val),
                    ThermalConsum = dbio.check(self.ThermalConsumNom.val),
                    HCGTEfficiency = dbio.check(self.HCGTEfficiencyNom.val),
                    ElectriProduction = dbio.check(self.ElectriProduction.val),
                    NDays = dbio.check(self.NDaysEq.val),
                    HPerDay = dbio.check(self.HPerDayEq.val),
                    PartLoad = dbio.check(self.PartLoad.val),                                     
                    HCGEEfficiency = dbio.check(self.HCGEEfficiency.val),
                    HPerYear = dbio.check(self.HPerYearEq.val),       
                    USHj = dbio.check(self.USHj.val),        
                    FETj = dbio.check(self.FETj.val),
                    FETFuel_j = dbio.check(self.FETFuel_j.val),
                    FETHeat_j = dbio.check(self.FETHeat_j.val),
                    FETel_j = dbio.check(self.FETel_j.val),
                    QHXEq = dbio.check(self.QHXEq.val),
                    QHXj = dbio.check(self.QHXEq.val),
                    QWHEq = dbio.check(self.QWHEq.val),
                    FlowExhaustGas = dbio.check(self.FlowExhaustGas.val),       
                    TSupply = dbio.check(self.TSupply.val),
                    TurnKeyPrice = dbio.check(self.TurnKeyPrice.val),
                    OMFix = dbio.check(self.OandMfix.val),
                    OMVar = dbio.check(self.OandMvar.val))
                        
        if self.equipmentClass == "CHP":
            eq.set(STAreaFactor = dbio.check(self.ElGen_j.val))
            if self.equipmentType in ["CHP TFC", "CHP ORC"]:
                eq.set(COPex = dbio.check(self.COPex.val))
                
            
        elif self.equipmentClass == "CH":
##                    "STAreaFactor = dbio.check(self.WaterConsum.val)  #saving of equipment specific water consumption not needed for the moment. only industry total
            eq.set(TemperatureReCooling = dbio.check(self.TemperatureReCooling.val),
                    COPex = dbio.check(self.COPex.val))

            if self.equipmentType == "compression chiller (water cooled)":
                eq.set(HCGTEffReal = dbio.check(self.SystemEfficiency.val))
            else:
                eq.set(HCGTEffReal = dbio.check(self.HCGTEfficiency.val))

        elif self.equipmentClass == "HP":
            eq.set(THeatSourceLT = dbio.check(self.THeatSourceLT.val),
                   COPex = dbio.check(self.COPex.val))

        if self.exhaustGas:
            eq.set(TExhaustGas = dbio.check(self.TExhaustGas.val),
                        ExcessAirRatio = dbio.check(self.ExcessAirRatio.val),
                        QExhaustGas = dbio.check(self.QExhaustGas.val))

        # direct calculation of OandM costs (not modeled as CCPar)
        USX = self.USHj.val
        if USX is not None and self.OandMfix.val is not None and self.OandMvar.val is not None:
            eq.set(OM = dbio.check(self.OandMfix.val + self.OandMvar.val*USX))
        else:
            eq.set(OM = dbio.check(None))

        # self.USHConv not into the qgenerationhc and not exported yet   

        # export to DB in old-fashioned CheckProc usage
        if self.exportToDB:
            eq.saveToDB(ANo=0, warn=False)
            dbio.commit()

                
    def showAll(self):
        
        print "====================="
        self.HCGPnom.show()
        self.FuelConsum.show()
        self.FuelConsumFL.show()
        self.FuelConsumNom.show()
        self.ElectriConsum.show()
        self.ElectriConsumFL.show()
        self.ElectriConsumNom.show()
        self.HPerYearEqNom.show()
        self.HPerYearEq.show()
        self.HPerDayEq.show()
        self.NDaysEq.show()
        self.PartLoad.show()
        self.PartLoadFact.show()
        self.PL1.show()
        self.PLC0.show()
        self.TCorrFact.show()
        
        self.HCGTEfficiency.show()
        self.HCGTEfficiencyFL.show()
        self.HCGTEfficiencyNom.show()
        self.COPex.show()
        
        self.ConvLoss.show()
        self.FETFuelRatio.show()
        self.FETelRatio.show()
        self.LossFactEq.show()
        self.USHj.show()
        self.dotUSHj.show()
        self.USHConv.show()
        self.QHXEq.show()
        self.dotQHXEq.show()

        self.FETj.show()#
        self.dotFETj.show()
        
        self.FETel_j.show()#
        self.dotFETel_j.show()

        self.FETel_c_j.show()#
        self.FETFuel_j.show()
        self.dotFETFuel_j.show()

        self.FETHeat_j.show()
        self.QDrive.show()
                
        self.ElectriProduction.show()
        self.ElGen_j.show()
        self.HCGEEfficiency.show()

        self.FlowCombAir.show()
        self.FlowExhaustGas.show()
        self.DTExhaustGas.show()
        self.DTOffGasUse.show()
        self.dotQExhaustGas.show()
        self.QExhaustGas.show()

        self.QConvLoss.show()
        self.dotQConvLoss.show()
        self.QWHEqTot.show()
        self.QWHEq.show()
        self.QLossEq.show()
        self.dotQLossEq.show()
        self.QInputEq.show()
        self.QOutEq.show()

        self.dotQInputEq.show()
        self.dotQOutEq.show()
        self.dotQWHEqTot.show()

        self.TExhaustGas.show()
        self.ExcessAirRatio.show()
        self.TMinOffGas.show()
        self.OffGasUse.show()

        self.ExhaustFact.show()
        self.CombAirExcess.show()
        self.OffGasRatio.show()
        
        self.QHXRatio.show()
        self.USHRatio.show()
        self.FETRatio.show()

        self.LossFactFET.show()
        self.VesselLoss.show()

        self.WaterConsum.show()

        self.OpCostTot.show()

        if self.equipmentType == "compression chiller (water cooled)":
            self.FETelChiller.show()
            self.FETelCT.show()

            self.PelChiller.show()
            self.PelCT.show()

            self.SystemEfficiency.show()
            print "fRecool: ", self.fRecool

        if self.equipmentType in ["compression chiller (water cooled)",
                                  "compression chiller (air cooled)",
                                  "compression heat pump",
                                  "mechanical vapour recompression",
                                  "CHP TFC",
                                  "CHP ORC"]:
            self.COPex.show()

        print "====================="
#------------------------------------------------------------------------------
    def screen(self):  
#------------------------------------------------------------------------------
#   screens all variables in the block
#------------------------------------------------------------------------------

########## Change of priority for parameters not needed

        if iszero(self.HCGPnom):
            self.FuelConsum.priority = 99
            
#................................................................................
        self.HCGPnom.screen()
        self.FuelConsum.screen()
        self.ElectriConsumNom.screen()

        if Status.ANo <= 0:
            self.NDaysEq.screen()
            self.HPerDayEq.screen()
            self.PartLoad.screen()
            self.HPerYearEq.screen()        

        self.HCGTEfficiency.screen()
        self.HCGTEfficiencyNom.screen()

        self.USHj.screen()
        self.USHConv.screen()
        self.FETj.screen()
        self.QHXEq.screen()

        self.QWHEq.screen()#
        if self.exhaustGas:
            self.QExhaustGas.screen()#
            self.TExhaustGas.screen()#
            self.ExcessAirRatio.screen()#
            self.FlowExhaustGas.screen ()#
            
        if self.equipmentClass == "CH":
            if self.equipmentType in ["compression chiller (air cooled)","compression chiller (water cooled)","fresh or ground water"]:
                self.TSupply.priority = 2
                self.TSupply.screen()
                
            if self.equipmentType in ["compression chiller (air cooled)","compression chiller (water cooled)"]:
                self.TemperatureReCooling.priority = 2
                self.TemperatureReCooling.screen()

                self.COPex.priority = 2
                self.COPex.screen()
                if self.equipmentType == "compression chiller (water cooled)":
                    self.SystemEfficiency.priority = 2
                    self.SystemEfficiency.screen()

        elif self.equipmentClass == "HP":
            if self.equipmentType in  ["compression heat pump", "mechanical vapour recompression"]:
                self.COPex.priority = 2
                self.COPex.screen()

        self.OandMfix.screen()
        self.OandMvar.screen()

             
#------------------------------------------------------------------------------
    def check(self):     #function that is called at the beginning when object is created
#------------------------------------------------------------------------------
#   main function carrying out the check of the block
#------------------------------------------------------------------------------
#        print "CheckEq - TESTPRINT TESTPRINT"
#        print "track of USHj"
#        self.USHj.show()
#        print self.USHj.track

        if DEBUG in ["ALL","MAIN"]:
            print "-------------------------------------------------"
            print " Equipment checking"
            print "-------------------------------------------------"

            self.showAll()


        for n in range(1):

            cycle.initCheckBalance()

            if DEBUG in ["ALL","MAIN"]:
                print "-------------------------------------------------"
                print "Ciclo %s"%n
                print "-------------------------------------------------"
        
# Step 1: Call all calculation routines in a given sequence

                print "Step 1: calculating from left to right (CALC)"
            

            if self.exhaustGas:
                self.FlowCombAir.push(calcProdC("FlowCombAir",self.CombAir,self.FuelConsum,self.ExcessAirRatio, unitType = "MASSFLOW")) # 
                self.FlowExhaustGas.push(calcSum("FlowExhaustGas",self.FlowCombAir,self.FuelConsum))#
                self.dotQExhaustGas.push(calcFlow_stack("dotQExhaustGas",self.OffgasHeatCapacity,self.FlowExhaustGas,self.TExhaustGas,
                                            self.TEnvEq,self.DTExhaustGas,unitType="ENERGY"))#
                self.dotQExhaustGas.push(calcDiff("dotQExhaustGas",self.dotQConvLoss,self.dotQLossEq))
                self.QExhaustGas.push(calcProd("QExhaustGas",self.dotQExhaustGas,self.HPerYearEqNom, unitType="ENERGY"))
            else:
                self.FlowExhaustGas.setValue(0.0, setTrack = False)   # security feature for communication with HX network calculations
                
            self.HPerYearEq.push(calcProd("HPerYearEq",self.HPerDayEq,self.NDaysEq,unitType="TIME"))
            self.HPerYearEqNom.push(calcProd("HPerYearEqNom",self.HPerYearEq,self.PartLoad, unitType="TIME"))
            
            self.USHConv.push(calcProd("USHConv",self.HCGPnom,self.HPerYearEqNom, unitType = "ENERGY"))
                
            self.FETel_c_j.push(calcProd("FETel_c_j",self.ElectriConsum,self.HPerYearEqNom, unitType = "ENERGY"))
            self.FETel_j.push(calcDiff("FETel_j",self.FETel_c_j,self.ElGen_j,parType="S"))
            self.FETel_j.push(calcProd("FETel_j",self.dotFETel_j,self.HPerYearEqNom,parType="S", unitType = "ENERGY"))

            self.dotFETel_j.push(calcDiff("dotFETel_j", self.ElectriConsum, self.ElectriProduction, parType="S"))

            self.FETFuel_j.push(calcProdC("FETFuel_j",self.FuelLCV,self.FuelConsum,self.HPerYearEqNom, unitType = "ENERGY"))
            self.FETFuel_j.push(calcProd("FETFuel_j",self.dotFETFuel_j,self.HPerYearEqNom, unitType = "ENERGY"))
            self.dotFETFuel_j.push(calcK("dotFETFuel_j",self.FuelLCV,self.FuelConsum, unitType = "POWER"))
            
            if self.mainSource == "Fuel":
#                print "CheckEq (calcs): fuel as main source"
                self.USHConv.push(calcProd("USHConv",self.FETFuel_j,self.HCGTEfficiency, unitType = "ENERGY"))
               
                self.HCGPnom.push(calcProdC("HCGPnom",self.FuelLCV,self.FuelConsum,self.HCGTEfficiency, unitType = "POWER"))

                if self.exhaustGas:
                    self.DTOffGasUse.push(calcProd("DTOffGasUse",self.OffGasUse,self.DTExhaustGas, unitType = "TEMPERATUREDIFFERENCE"))
                    self.DTOffGasUse.push(calcDiff("DTOffGasUse",self.TExhaustGas,self.TMinOffGas))
                    self.QWHEq.push(calcProd("QWHEq",self.OffGasUse,self.QWHEqTot, unitType = "ENERGY"))

                    self.QExhaustGas.push(calcProd("QExhaustGas", self.ExhaustFact, self.FETFuel_j, unitType="ENERGY"))
                    self.CombAirExcess.push(calcK("CombAirExcess", self.CombAir, self.ExcessAirRatio))
                    self.OffGasRatio.push(calcSum("OffGasRatio", self.One, self.CombAirExcess))

                    self.ExhaustFact.push(calcProdC("ExhaustFact",self.CpToLCV,self.OffGasRatio, self.DTExhaustGas))

                    self.QHXEq.push(calcProd("QHXEq", self.QHXRatio, self.USHConv, unitType="ENERGY"))
                    self.USHRatio.push(calcSum("USHRatio", self.One, self.QHXRatio))

                    self.FETj.push(calcProd("FETj", self.FETRatio, self.FETFuel_j, unitType="ENERGY"))
                    self.ConvLoss.push(calcDiff("ConvLoss", self.FETFuelRatio, self.HCGTEfficiency))

                    self.LossFactFET.push(calcProd("LossFactFET", self.LossFactEq, self.HCGTEfficiency))
                    self.VesselLoss.push(calcProd("VesselLoss", self.LossFactFET, self.USHRatio))

                    self.ExhaustFact.push(calcDiff("ExhaustFact", self.ConvLoss, self.VesselLoss))
                        
            elif self.mainSource == "Heat":
                self.QDrive.push(calcSum("QDrive", self.FETHeat_j, self.QHXEq))
                self.USHConv.push(calcProd("USHConv", self.QDrive, self.HCGTEfficiency, unitType = "ENERGY"))

                self.HCGPnom.push(calcProd("HCGPnom", self.ThermalConsumNom, self.HCGTEfficiencyNom, unitType = "POWER"))
                
                self.DTOffGasUse.setValue(0.0, setTrack = False)
                self.QWHEq.push(self.QWHEqTot)

            else:
                if self.equipmentType == "compression chiller (water cooled)":
                    self.USHConv.push(calcProd("USHConv",self.FETelChiller,self.HCGTEfficiency, unitType = "ENERGY"))
                    self.FETel_c_j.push(calcSum("FETel_c_j", self.FETelChiller, self.FETelCT))

                    self.HCGPnom.push(calcProd("HCGPnom",self.PelChiller,self.HCGTEfficiency, unitType = "POWER"))
                    self.ElectriConsum.push(calcSum("ElectriConsum",self.PelChiller, self.PelCT))

                    self.FETelCT.push(calcK("FETelCT", self.fRecool, self.QWHEqTot))

                    self.FETelChiller.push(calcProd("FETelChiller",self.PelChiller,self.HPerYearEqNom, unitType = "ENERGY"))
                    self.FETelCT.push(calcProd("FETelCT",self.PelCT,self.HPerYearEqNom, unitType = "ENERGY"))

                    self.USHConv.push(calcProd("USHConv",self.FETel_c_j, self.SystemEfficiency, unitType = "ENERGY"))
                    self.HCGPnom.push(calcProd("HCGPnom",self.ElectriConsum, self.SystemEfficiency, unitType = "POWER"))

                else:
                    self.USHConv.push(calcProd("USHConv",self.FETel_c_j,self.HCGTEfficiency, unitType = "ENERGY"))
                
                    self.HCGPnom.push(calcProd("HCGPnom",self.ElectriConsum,self.HCGTEfficiency, unitType = "POWER"))

                self.HCGPnom.push(calcProd("HCGPnom",self.ElectriConsumNom,self.HCGTEfficiencyNom, unitType = "POWER"))

                self.DTOffGasUse.setValue(0.0, setTrack = False)
                self.QWHEq.push(self.QWHEqTot)

                
            if self.mainSource=="Heat":
                self.USHj.push(calcDiff("USHj", self.USHConv, self.QWHEq))  # generated heat partially unused (-> dissipated as WH)
##                self.dotUSHj.push(calcDiff("dotUSHj",self.HCGPnom, self.dotQWHEq))
            else:
                self.USHj.push(calcSum("USHj",self.USHConv,self.QHXEq))
                self.dotUSHj.push(calcSum("dotUSHj",self.HCGPnom,self.dotQHXEq))
                                   
            self.USHj.push(calcProd("USHj",self.dotUSHj,self.HPerYearEqNom, unitType = "ENERGY"))
                           
            if self.mainSource=="Heat":
                self.FETj.push(calcSum("FETj", self.FETHeat_j, self.FETel_j, parType="S"))
                self.dotQInputEq.push(calcSum("dotQInputEq",self.ThermalConsumNom, self.dotFETel_j, parType="S"))
            else:
                self.FETj.push(calcSum("FETj", self.FETFuel_j, self.FETel_j, parType="S"))
                self.dotFETj.push(calcSum("dotFETj",self.dotFETFuel_j,self.dotFETel_j,parType="S"))

            self.FETj.push(calcProd("FETj",self.dotFETj,self.HPerYearEqNom,parType="S", unitType = "ENERGY"))

            self.QLossEq.push(calcProd("QLossEq",self.LossFactEq,self.USHj, unitType = "ENERGY"))
            self.dotQLossEq.push(calcProd("dotQLossEq",self.LossFactEq,self.dotUSHj, unitType = "POWER"))
            self.QOutEq.push(calcSum("QOutEq",self.QLossEq,self.USHj))
            self.dotQOutEq.push(calcSum("dotQOutEq", self.dotQLossEq, self.HCGPnom))

            self.QInputEq.push(calcSum("QInputEq",self.FETj,self.QHXEq))
            
            if self.equipmentClass in ["CH"]:  #temporary solution for chillers - equivalent heat model
                self.QWHEqTot.push(calcSum("QWHEqTot",self.FETj,self.USHj))
            elif not self.condensing:
                self.QWHEqTot.push(calcDiff("QWHEqTot",self.QInputEq,self.QOutEq))
                self.dotQWHEqTot.push(calcDiff("dotQWHEqTot", self.dotQInputEq,self.dotQOutEq))
            else:
                self.QWHEqTot.setValue(0.0, setTrack = False)
                self.QWHEq.setValue(0.0, setTrack = False)

            if self.equipmentClass in ["BB"] and not self.condensing:
#                print "checkEq in mode BB"
                self.QConvLoss.push(calcProd("QConvLoss",self.ConvLoss,self.FETFuel_j, unitType = "ENERGY"))
                self.dotQConvLoss.push(calcProd("dotQConvLoss",self.ConvLoss,self.dotFETFuel_j, unitType = "POWER"))
                self.QWHEqTot.push(calcDiff("QWHEqTot",self.QConvLoss,self.QLossEq))
                self.FETj.push(calcSum("FETj",self.QConvLoss,self.USHConv))
                self.FETj.push(calcProd("FETj", self.FETFuelRatio, self.FETFuel_j, unitType="ENERGY"))
                self.dotFETj.push(calcProd("dotFETj", self.FETFuelRatio, self.dotFETFuel_j, unitType="POWER"))
                self.FETFuelRatio.push(calcSum("FETFuelRatio",self.ConvLoss,self.HCGTEfficiency))

                self.dotFETel_j.push(calcProd("dotFETel_j", self.FETelRatio, self.dotFETFuel_j, unitType="POWER"))
                self.FETFuelRatio.push(calcSum("FETFuelRatio", self.One, self.FETelRatio))

            elif self.equipmentClass == "CHP":
                self.ElGen_j.push(calcProd("ElGen_j",self.ElectriProduction,self.HPerYearEqNom, unitType = "ENERGY"))
                self.dotQWHEqTot.setValue(0.0, setTrack=False)

                if self.mainSource=="Heat":
                    self.ElGen_j.push(calcProd("ElGen_j",self.QDrive, self.HCGEEfficiency, unitType = "ENERGY"))
                    self.ElectriProduction.push(calcProd("ElectriProduction", self.ThermalConsumNom, self.HCGEEfficiency, unitType="POWER")) 
                else:
                    self.ElGen_j.push(calcProd("ElGen_j",self.FETFuel_j,self.HCGEEfficiency, unitType = "ENERGY"))
                    self.ElectriProduction.push(calcProdC("ElectriProduction",self.FuelLCV,self.FuelConsum,self.HCGEEfficiency, unitType = "ENERGY"))


            if self.partLoadCorrection:
                self.PL1.push(calcK("PL1",self.PLC1,self.PartLoad))
                self.PartLoadFact.push(calcSum("PartLoadFact",self.PLC0,self.PL1))

            # relation between nominal and real full load parameters               
            self.HCGTEfficiencyFL.push(calcProd("HCGTEfficiencyFL",self.TCorrFact,self.HCGTEfficiencyNom))
            self.ElectriConsumNom.push(calcProd("ElectriConsumNom",self.TCorrFact,self.ElectriConsumFL, unitType = "POWER"))
            self.FuelConsumNom.push(calcProd("FuelConsumNom",self.TCorrFact,self.FuelConsumFL, unitType = "MASSFLOW"))

            # relation between full load and part load parameters                
            self.HCGTEfficiency.push(calcProd("HCGTEfficiency",self.HCGTEfficiencyFL,self.PartLoadFact))
            self.FuelConsumFL.push(calcProd("FuelConsumFL",self.FuelConsum,self.PartLoadFact, unitType = "MASSFLOW"))
            self.ElectriConsumFL.push(calcProd("ElectriConsumFL",self.ElectriConsum,self.PartLoadFact, unitType = "POWER"))

            if self.equipmentType in  ["compression chiller (air cooled)", "compression chiller (water cooled)"]:
                if self.EERCarnot is not None: self.HCGTEfficiencyNom.push(calcK("HCGTEfficiencyNom", self.EERCarnot, self.COPex))
                
            elif self.equipmentType in ["compression heat pump", "mechanical vapour recompression"]:
                if self.COPCarnot is not None: self.HCGTEfficiencyNom.push(calcK("HCGTEfficiencyNom", self.COPCarnot, self.COPex))
                
            elif self.equipmentType in ["CHP ORC", "CHP TFC"]:
                if self.COPCarnot is not None: self.HCGEEfficiency.push(calcK("HCGEEfficiency", self.COPCarnot, self.COPex))
                
            self.QHXEq.push(calcProd("QHXEq",self.dotQHXEq,self.HPerYearEqNom, unitType = "ENERGY"))

            self.OpCostVar.push(calcProd("OpCostVar", self.OandMvar, self.USHj, unitType = "PRICE"))
            self.OpCostTot.push(calcSum("OpCostTot", self.OandMfix, self.OpCostVar))
                   
            if DEBUG in ["ALL","MAIN"]:
                self.showAll()

# Step 2: Cross check the variables

                print "Step 2: cross checking"

            self.ccheckAll()

            if DEBUG in ["ALL","MAIN"]:
                self.showAll()

# Step 3: Adjust the variables (inverse of calculation routines)

                print "Step 3: calculating from right to left (ADJUST)"
            
            # relation between nominal and real full load parameters               
            adjustProd_stack(self.HCGTEfficiencyFL,self.TCorrFact,self.HCGTEfficiencyNom)
            adjustProd_stack(self.ElectriConsumNom,self.TCorrFact,self.ElectriConsumFL)
            adjustProd_stack(self.FuelConsumNom,self.TCorrFact,self.FuelConsumFL)

            if self.equipmentType in  ["compression chiller (air cooled)", "compression chiller (water cooled)"]:
                if self.EERCarnot is not None: adjustcalcK_stack(self.HCGTEfficiencyNom, self.EERCarnot, self.COPex)
                
            elif self.equipmentType in  ["compression heat pump", "mechanical vapour recompression"]:
                if self.COPCarnot is not None: adjustcalcK_stack(self.HCGTEfficiencyNom, self.COPCarnot, self.COPex)

            elif self.equipmentType in ["CHP ORC", "CHP TFC"]:
                if self.COPCarnot is not None: adjustcalcK_stack(self.HCGEEfficiency, self.COPCarnot, self.COPex)

            if self.partLoadCorrection:
                adjustcalcK_stack(self.PL1,self.PLC1,self.PartLoad)
                adjustSum_stack(self.PartLoadFact,self.PLC0,self.PL1)
                
            adjustProd_stack(self.HCGTEfficiency,self.HCGTEfficiencyFL,self.PartLoadFact)
            adjustProd_stack(self.FuelConsumFL,self.FuelConsum,self.PartLoadFact)
            adjustProd_stack(self.ElectriConsumFL,self.ElectriConsum,self.PartLoadFact)

            if self.equipmentClass in ["BB"] and not self.condensing:
                adjustSum_stack(self.FETj,self.QConvLoss,self.USHConv)
                adjustDiff_stack(self.QWHEqTot,self.QConvLoss,self.QLossEq)#
                adjustProd_stack(self.QConvLoss,self.ConvLoss,self.FETFuel_j)
                adjustProd_stack(self.dotQConvLoss,self.ConvLoss,self.dotFETFuel_j)
                adjustSum_stack(self.FETFuelRatio,self.HCGTEfficiency,self.ConvLoss)

                adjustProd_stack(self.FETj,self.FETFuelRatio,self.FETFuel_j)
                adjustProd_stack(self.dotFETj,self.FETFuelRatio,self.dotFETFuel_j)

                adjustProd_stack(self.dotFETel_j, self.FETelRatio, self.dotFETFuel_j)
                adjustSum_stack(self.FETFuelRatio, self.One, self.FETelRatio)

            if self.equipmentClass in ["CH"]:  #temporary solution for chillers - equivalent heat model
                adjustSum_stack(self.QWHEqTot,self.FETj,self.USHj)
            elif not self.condensing:
                adjustDiff_stack(self.QWHEqTot,self.QInputEq,self.QOutEq) #
                adjustDiff_stack(self.dotQWHEqTot, self.dotQInputEq,self.dotQOutEq)
                
            adjustSum_stack(self.QInputEq,self.FETj,self.QHXEq)#
            adjustSum_stack(self.QOutEq,self.QLossEq,self.USHj)#
            adjustSum_stack(self.dotQOutEq, self.dotQLossEq, self.HCGPnom)

            adjustProd_stack(self.QLossEq,self.LossFactEq,self.USHj)#
            adjustProd_stack(self.dotQLossEq,self.LossFactEq,self.dotUSHj)

            if self.mainSource=="Heat":
                adjustSum_stack(self.FETj,self.FETHeat_j,self.FETel_j)
                adjustSum_stack(self.dotQInputEq, self.ThermalConsumNom, self.dotFETel_j)
            else:
                adjustSum_stack(self.FETj,self.FETFuel_j,self.FETel_j)
                adjustSum_stack(self.dotFETj,self.dotFETFuel_j,self.dotFETel_j)
            adjustProd_stack(self.FETj,self.dotFETj,self.HPerYearEqNom)
                           
            if self.mainSource=="Heat":
                adjustDiff_stack(self.USHj, self.USHConv, self.QWHEq)
##                adjustDiff_stack(self.dotUSHj, self.HCGPnom, self.dotQWHEq)
            else:
                adjustSum_stack(self.USHj,self.USHConv,self.QHXEq)
                adjustSum_stack(self.dotUSHj,self.HCGPnom,self.dotQHXEq)

            adjustProd_stack(self.USHj,self.dotUSHj,self.HPerYearEqNom)

            adjustProd_stack(self.USHConv,self.HCGPnom,self.HPerYearEqNom)
                

            if self.mainSource == "Fuel":

                if self.exhaustGas:
                    adjustProd_stack(self.QWHEq,self.OffGasUse,self.QWHEqTot)
                    adjustProd_stack(self.DTOffGasUse,self.OffGasUse,self.DTExhaustGas)
                    adjustDiff_stack(self.DTOffGasUse,self.TExhaustGas,self.TMinOffGas)

                    adjustProd_stack(self.QExhaustGas, self.ExhaustFact, self.FETFuel_j)
                    adjustcalcK_stack(self.CombAirExcess, self.CombAir, self.ExcessAirRatio)
                    adjustSum_stack(self.OffGasRatio, self.One, self.CombAirExcess)

                    adjustProdC_stack(self.ExhaustFact,self.CpToLCV,self.OffGasRatio, self.DTExhaustGas)

                    adjustProd_stack(self.QHXEq, self.QHXRatio, self.USHConv)
                    adjustSum_stack(self.USHRatio, self.One, self.QHXRatio)

                    adjustProd_stack(self.FETj, self.FETRatio, self.FETFuel_j)
                    adjustDiff_stack(self.ConvLoss, self.FETFuelRatio, self.HCGTEfficiency)

                    adjustProd_stack(self.LossFactFET, self.LossFactEq, self.HCGTEfficiency)
                    adjustProd_stack(self.VesselLoss, self.LossFactFET, self.USHRatio)
                    adjustDiff_stack(self.ExhaustFact, self.ConvLoss, self.VesselLoss)

                adjustProd_stack(self.USHConv,self.FETFuel_j,self.HCGTEfficiency)
                adjustProdC_stack(self.HCGPnom,self.FuelLCV,self.FuelConsum,self.HCGTEfficiency)

            elif self.mainSource == "Heat":
                adjustSum_stack(self.QDrive, self.FETHeat_j, self.QHXEq)
                adjustProd_stack(self.USHConv, self.QDrive, self.HCGTEfficiency)
                adjustProd_stack(self.HCGPnom, self.ThermalConsumNom, self.HCGTEfficiencyNom)


                self.OffGasUse.setValue(0.0, setTrack = False)

            else:
                if self.equipmentType == "compression chiller (water cooled)":
                    adjustProd_stack(self.USHConv, self.FETelChiller, self.HCGTEfficiency)
                    adjustSum_stack(self.FETel_c_j, self.FETelChiller, self.FETelCT)

                    adjustProd_stack(self.HCGPnom, self.PelChiller, self.HCGTEfficiency)
                    adjustSum_stack(self.ElectriConsum ,self.PelChiller, self.PelCT)

                    adjustcalcK_stack(self.FETelCT, self.fRecool, self.QWHEqTot)

                    adjustProd_stack(self.FETelChiller,self.PelChiller,self.HPerYearEqNom)
                    adjustProd_stack(self.FETelCT, self.PelCT, self.HPerYearEqNom)

                    adjustProd_stack(self.USHConv, self.FETel_c_j, self.SystemEfficiency)
                    adjustProd_stack(self.HCGPnom, self.ElectriConsum, self.SystemEfficiency)

                else:
                    adjustProd_stack(self.USHConv,self.FETel_c_j,self.HCGTEfficiency)
                    adjustProd_stack(self.HCGPnom,self.ElectriConsum,self.HCGTEfficiency)

                adjustProd_stack(self.HCGPnom, self.ElectriConsumNom, self.HCGTEfficiencyNom)

                self.OffGasUse.setValue(0.0, setTrack = False)
                
            adjustDiff_stack(self.FETel_j,self.FETel_c_j,self.ElGen_j)
            adjustProd_stack(self.FETel_c_j,self.ElectriConsum,self.HPerYearEqNom)
            adjustProd_stack(self.FETel_j,self.dotFETel_j,self.HPerYearEqNom)
            adjustDiff_stack(self.dotFETel_j, self.ElectriConsum, self.ElectriProduction)
                           
            adjustProdC_stack(self.FETFuel_j,self.FuelLCV,self.FuelConsum,self.HPerYearEqNom)
            adjustProd_stack(self.FETFuel_j,self.dotFETFuel_j,self.HPerYearEqNom)
            adjustcalcK_stack(self.dotFETFuel_j,self.FuelLCV,self.FuelConsum)

            adjustProd_stack(self.HPerYearEqNom,self.HPerYearEq,self.PartLoad)
            adjustProd_stack(self.HPerYearEq,self.HPerDayEq,self.NDaysEq)

            if self.exhaustGas:
                adjustProd_stack(self.QExhaustGas,self.dotQExhaustGas,self.HPerYearEqNom)#
                adjustFlow_stack(self.dotQExhaustGas,self.OffgasHeatCapacity,self.FlowExhaustGas,self.TExhaustGas,
                                                self.TEnvEq,self.DTExhaustGas)#
                adjustDiff_stack(self.dotQExhaustGas,self.dotQConvLoss,self.dotQLossEq)
                adjustSum_stack(self.FlowExhaustGas,self.FlowCombAir,self.FuelConsum)#
                adjustProdC_stack(self.FlowCombAir,self.CombAir,self.FuelConsum,self.ExcessAirRatio)

            if self.equipmentClass == "CHP":
                adjustProd_stack(self.ElGen_j,self.ElectriProduction,self.HPerYearEqNom)

                if self.mainSource=="Heat":
                    adjustProd_stack(self.ElGen_j, self.QDrive, self.HCGEEfficiency)
                    adjustProd_stack(self.ElectriProduction, self.ThermalConsumNom, self.HCGEEfficiency) 
                else:
                    adjustProd_stack(self.ElGen_j,self.FETFuel_j,self.HCGEEfficiency)
                    adjustProdC_stack(self.ElectriProduction,self.FuelLCV,self.FuelConsum,self.HCGEEfficiency)
            

            adjustProd_stack(self.QHXEq,self.dotQHXEq,self.HPerYearEqNom)

            adjustProd_stack(self.OpCostVar, self.OandMvar, self.USHj)
            adjustSum_stack(self.OpCostTot, self.OandMfix, self.OpCostVar)


            if DEBUG in ["ALL","MAIN"]:
                self.showAll()
            
# Step 4: Cross check again the variables

                print "Step 4: second cross checking"

            self.ccheckAll()
            
            if DEBUG in ["ALL","MAIN"]:
                self.showAll()

# Final: calculation of derived variables

        if self.equipmentClass == "CH" and self.USHConv.val is not None:

            if self.FETj.val is not None:   # aproximate calculation for wate heat from chillers. currently only used in water cooled compression chiller
                                            # TO DO: should be improved, because FET includes fan power of CT, but does not contribute to water consumption
                QWH = self.USHConv.val + self.FETj.val
            else:
                QWH = self.USHConv.val
            
            waterConsumption = Status.mod.moduleCH.getWaterConsumption(self.equipmentType,self.USHConv.val,QWH, self.dryTower)
            self.WaterConsum.setValue(waterConsumption,err = self.USHConv.sqerr)
            
        else:
            self.WaterConsum.setValue(0.0, setTrack = False)

        if DEBUG in ["ALL","BASIC"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()
        
#------------------------------------------------------------------------------
    def ccheckAll(self):     #function that is called at the beginning when object is created
#------------------------------------------------------------------------------

            self.HCGEEfficiency.check()
            self.HCGTEfficiency.check()
            self.COPex.check()
            
            self.ConvLoss.check()
            self.FETFuelRatio.check()
            self.FETelRatio.check()
            
            self.FlowCombAir.check()
            self.ExcessAirRatio.check()
            self.FlowExhaustGas.check()

            self.dotQExhaustGas.check()
            self.TExhaustGas.check()
            self.TEnvEq.check()

            self.HPerYearEq.check()
            self.HPerDayEq.check()
            self.NDaysEq.check()
            
            self.HPerYearEqNom.check()
            self.HCGPnom.check()
            self.FETj.check()
            self.dotFETj.check()
            self.FETFuel_j.check()
            self.dotFETFuel_j.check()
            self.FETHeat_j.check()
            self.QDrive.check()
            self.dotQDrive.check()
            self.FETel_c_j.check()
            self.FETel_j.check()
            self.dotFETel_j.check()
            self.FuelConsum.check()
            self.ElectriConsum.check()
            self.USHConv.check()
            self.QHXEq.check()
            self.dotQHXEq.check()
            self.USHj.check()
            self.dotUSHj.check()
            self.QConvLoss.check()
            self.dotQConvLoss.check()
            self.LossFactEq.check()
            self.QLossEq.check()
            self.dotQLossEq.check()
            self.QInputEq.check()
            self.dotQInputEq.check()
            self.QOutEq.check()
            self.dotQOutEq.check()

            if self.equipmentClass == "CH":
                self.QWHEq.update(self.QWHEqTot)
            elif self.exhaustGas:
                if not self.condensing:
                    self.QExhaustGas.check()
                    self.QWHEqTot.push(self.QExhaustGas)
                    self.QExhaustGas.update(self.QWHEqTot)

                self.ExhaustFact.check()
                self.CombAirExcess.check()
                self.OffGasRatio.check()

                self.QHXRatio.check()
                self.USHRatio.check()
                self.FETRatio.check()

                self.LossFactFET.check()
                self.VesselLoss.check()

                self.DTOffGasUse.check()
                self.OffGasUse.check()
                self.TMinOffGas.check()
                self.DTExhaustGas.check()

            self.QWHEqTot.check()
            self.dotQWHEqTot.check()
            self.QWHEq.check()
            self.PartLoad.check()

            self.ElectriProduction.check()
            self.ElGen_j.check()

            if self.partLoadCorrection:
                self.PL1.check()
                self.PLC0.check()
                
            self.PartLoadFact.check()
            self.TCorrFact.check()

            self.HCGTEfficiencyFL.check()
            self.FuelConsumFL.check()
            self.ElectriConsumFL.check()

            self.HCGTEfficiencyNom.check()
            self.FuelConsumNom.check()
            self.ElectriConsumNom.check()
            self.ThermalConsumNom.check()

            if self.equipmentType == "compression chiller (water cooled)":
                self.FETelChiller.check()
                self.FETelCT.check()

                self.PelChiller.check()
                self.PelCT.check()

                self.SystemEfficiency.check()

            if self.equipmentType in ["compression chiller (water cooled)",
                                      "compression chiller (air cooled)"]:
                self.COPex.check()

            self.OandMfix.check()
            self.OandMvar.check()
            self.OpCostVar.check()
            self.OpCostTot.check()

            self.One.check()


#------------------------------------------------------------------------------
    def estimate(self):  
#------------------------------------------------------------------------------
#   estimates some of the data that are not sufficiently precise
#   should be a subset of the data that are within screen
#   (not necessarily ALL data have to be estimated)
#------------------------------------------------------------------------------

        if self.equipmentClass in ["BB"]:
            if self.mainSource == "Electricity":
                self.HCGTEfficiency.setEstimate(0.99,limits=(0.95,1.0))
            else:
                self.HCGTEfficiency.setEstimate(0.85,limits=(0.7,0.95))
                        
        elif self.equipmentClass in ["CH"]:
            defaultEER = Status.mod.moduleCH.getDefaultEER(self.equipmentType)
            self.HCGTEfficiency.setEstimate(defaultEER,limits=(0.5*defaultEER,2.0*defaultEER))
            
        elif self.equipmentClass in ["HP"]:
            self.HCGTEfficiency.setEstimate(5.0,limits=(2.0,8.0))
            
        elif self.equipmentClass in ["CHP"]:
            self.HCGTEfficiency.setEstimate(0.55,limits=(0.4,0.9))
            self.HCGEEfficiency.setEstimate(0.35,limits=(0.1,0.6))
            
        self.PartLoad.setEstimate(0.5,limits=(0.2,0.8))

        self.estimateEconomicParameters()

    def estimateEconomicParameters(self):
        """ estimates the economic parameters for the equipment if not given
        """
        
        if self.equipmentClass == "BB":
            (turnKey, OMFix, OMVar) = Status.mod.moduleBB.estimateEconomicParameters(self.eq)
        elif self.equipmentClass == "CH":
            (turnKey, OMFix, OMVar) = Status.mod.moduleCH.estimateEconomicParameters(self.eq)
        elif self.equipmentClass == "CHP":
            (turnKey, OMFix, OMVar) = Status.mod.moduleCHP.estimateEconomicParameters(self.eq)
        elif self.equipmentClass == "HP":
            (turnKey, OMFix, OMVar) = Status.mod.moduleHP.estimateEconomicParameters(self.eq)
        elif self.equipmentClass == "ST":
            (turnKey, OMFix, OMVar) = Status.mod.moduleST.estimateEconomicParameters(self.eq)
        else:
            (turnKey, OMFix, OMVar) = (None, None, None)

        estimateTracker.setDataGroup("Equipment",self.EqNo)
        if self.TurnKeyPrice.val is None:
            if Status.ANo <=0:
                self.TurnKeyPrice.setValue(0.0)
            elif turnKey is not None:
                self.TurnKeyPrice.setEstimate(turnKey)
                logWarning(_("Investment cost of equipment %s is not defined. Estimated to %s"%(self.eq.name, turnKey)))
                       
        if self.OandMfix.val is None and OMFix is not None:
            self.OandMfix.setEstimate(OMFix)
            logWarning(_("Parameters for O&M costs of equipment %s are not defined. %s assumed")%(self.eq.name, OMFix) \
                       + " " + _("for fix part"))
        if self.OandMvar.val is None and OMVar is not None:
            self.OandMvar.setEstimate(OMVar)
            logWarning(_("Parameters for O&M costs of equipment %s are not defined. %s assumed")%(self.eq.name, OMVar) \
                       + " " + _("for variable part"))

    def storeEconomicParameters(self):
        """ auxiliary function
            stores economic parameters in eqObject
        """

        self.eq.set(TurnKeyPrice = dbio.check(self.TurnKeyPrice.val),
                          OMFix = dbio.check(self.OandMfix.val),
                          OMVar = dbio.check(self.OandMvar.val))

#==============================================================================
if __name__ == "__main__":
    
    pass    
#==============================================================================

