# -*- coding: cp1252 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CCheck (Consistency Check)
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from auxiliary.auxiliary import toGettextAndUnicode as _U
from math import *
from ccheckFunctions import *
from ccPar import *
from ccTracker import estimateTracker
from GUI.status import *
import database.dbio as dbio

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"


#------------------------------------------------------------------------------
class CheckFETel():
#------------------------------------------------------------------------------
#   Carries out consistency checking 
#------------------------------------------------------------------------------

    def __init__(self, el=None, checkInPlace = False):     #function that is called at the beginning when object is created

        self.index = -1 # attribute needed for compatibility with localCheck - function

        if el is None: # keep this here temporarily for compatibility with other ccheck modules which still read/write from/to DB
            self.el = Status.int.systemModel.elObject[0]
        else:
            self.el = el
        
        self.checkInPlace = checkInPlace
        self.exportToDB = not checkInPlace
# assign a variable to all intermediate/calculated values needed
        
        self.ElectricityNet = CCPar("ElectricityNet", unitType = "ENERGY") 
        
        self.FECel_c = CCPar("FECel_c",priority=1, unitType = "ENERGY")
        self.FECel = CCPar("FECel",priority=1,parType="S", unitType = "ENERGY")
        self.FEOel = CCPar("FEOel", unitType = "ENERGY")
        
        self.FETel_c = CCPar("FETel_c",priority=1, unitType = "ENERGY")

        self.FETel = CCPar("FETel",priority=1,parType="S", unitType = "ENERGY")
        
        self.importData(0)

        if DEBUG in ["ALL","BASIC","MAIN"]:
            self.showAllFETel()


#------------------------------------------------------------------------------
    def importData(self,i):  
#------------------------------------------------------------------------------
#   imports data from SQL tables (from AlternativeProposalNo = -1)
#------------------------------------------------------------------------------

        if self.checkInPlace:
            ANo = Status.ANo
        else:
            ANo = -1
        
#..............................................................................
# assign empty CCPar to all questionnaire parameters

        self.ElectricityGen = CCPar("ElectricityGen", unitType = "ENERGY")
        self.ElectricitySales = CCPar("ElectricitySales", unitType = "ENERGY")
        self.ElectricityTotYear = CCPar("ElectricityTotYear", unitType = "ENERGY")

        self.ElectricityMotors = CCPar("ElectricityMotors", unitType = "ENERGY")
        self.ElectricityChem = CCPar("ElectricityChem", unitType = "ENERGY")
        self.ElectricityLight = CCPar("ElectricityLight", unitType = "ENERGY")
        self.ElectricityRef = CCPar("ElectricityRef", unitType = "ENERGY")
        self.ElectricityAC = CCPar("ElectricityAC", unitType = "ENERGY")
        self.ElectricityThOther = CCPar("ElectricityThOther", unitType = "ENERGY")

        self.ElTariffCTot = CCPar("ElTariffCTot", priority = 2, unitType = "TARIFF")
        self.ElTariffPowTot = CCPar("ElTariffPowTot", unitType = "TARIFF")
        self.ElTariffFeedIn = CCPar("ETariffFeedIn", priority = 2, unitType = "TARIFF")
        self.ElTariffCHP = CCPar("ElTariffCHP", priority = 2, unitType = "TARIFF")

        self.ElCostYearTot = CCPar("ElCostYearTot", priority = 1, unitType = "PRICE")
        self.ElSalesYearFeedIn = CCPar("ElSalesYearFeedIn", priority = 1, unitType = "PRICE")
        self.ElSalesYearCHP = CCPar("ElSalesYearCHP", priority = 1, unitType = "PRICE")

#..............................................................................
# reading data from table "self.el"

        self.ElectricityGen.setValue(self.el.ElGenera)
        self.ElectricitySales.setValue(self.el.ElSales)

        self.ElectricityTotYear.setValue(self.el.ElectricityTotYear)
        self.ElectricityMotors.setValue(self.el.ElectricityMotors)
        self.ElectricityChem.setValue(self.el.ElectricityChem)
        self.ElectricityLight.setValue(self.el.ElectricityLight)
        self.ElectricityRef.setValue(self.el.ElectricityRef)
        self.ElectricityAC.setValue(self.el.ElectricityAC)
        self.ElectricityThOther.setValue(self.el.ElectricityThOther)

        self.ElTariffCTot.setValue(self.el.ElTariffCTot)
        self.ElTariffPowTot.setValue(self.el.ElTariffPowTot)
        self.ElTariffFeedIn.setValue(self.el.ElTariffFeedIn)
        self.ElTariffCHP.setValue(self.el.ElTariffCHP)

        if self.el.DataFile is not None:
            Status.schedules.calculatePlantSchedule()
            self.el.estimateAverageTariffs()

            estimateTracker.setDataGroup("FETel", 0)
            self.ElTariffCTot.setEstimate(self.el.tariffCAvg, limits = (self.el.tariffCMin, self.el.tariffCMax))
            self.ElTariffCHP.setEstimate(self.el.tariffCHPAvg, limits = (self.el.tariffCHPMin, self.el.tariffCHPMax))

        self.ElCostYearTot.setValue(self.el.ElCostYearTot)
        self.ElSalesYearFeedIn.setValue(self.el.ElSalesYearFeedIn)
        self.ElSalesYearCHP.setValue(self.el.ElSalesYearCHP)

#..............................................................................
# check if calculation of FECel is principally possible

        if (self.ElectricityTotYear.val is None):
##           (self.ElectricityGen.val is None) or \
##           (self.ElectricitySales.val is None):
            if (self.ElectricityMotors.val is None) or \
               (self.ElectricityLight.val is None) or \
               (self.ElectricityChem.val is None):
                logWarning(_U("WARNING: No data available for electricity consumption for non-thermal uses. Set to 0 !!!"))
                if (self.ElectricityMotors.val is None): self.ElectricityMotors.setValue(0.0)
                if (self.ElectricityLight.val is None): self.ElectricityLight.setValue(0.0)
                if (self.ElectricityChem.val is None): self.ElectricityChem.setValue(0.0)
#                self.FEOel.setValue(0.0)

#------------------------------------------------------------------------------
    def exportData(self):  
#------------------------------------------------------------------------------
#   stores corrected data in SQL (under AlternativeProposalNo = 0)
#------------------------------------------------------------------------------

        if self.checkInPlace:
            ANo = max(Status.ANo, 0)
        else:
            ANo = 0
        
#..............................................................................
# writing data to table "cgeneraldata"

        if ANo == 0:
            self.el.set(FECel = dbio.check(self.FECel.val),
                        FEOel = dbio.check(self.FEOel.val),
                        FETel = dbio.check(self.FETel.val))
                
#..............................................................................
# writing data to table "self.el"

        self.el.set(ElectricityTotYear = dbio.check(self.ElectricityTotYear.val),
                         ElGenera = dbio.check(self.ElectricityGen.val),
                         ElSales = dbio.check(self.ElectricitySales.val),
                         ElectricityMotors = dbio.check(self.ElectricityMotors.val),
                         ElectricityChem = dbio.check(self.ElectricityChem.val),
                         ElectricityLight = dbio.check(self.ElectricityLight.val),
                         ElectricityRef = dbio.check(self.ElectricityRef.val),
                         ElectricityAC = dbio.check(self.ElectricityAC.val),
                         ElectricityThOther = dbio.check(self.ElectricityThOther.val),
                         ElTariffCTot = dbio.check(self.ElTariffCTot.val),
                         ElTariffPowTot = dbio.check(self.ElTariffPowTot.val),
                         ElTariffFeedIn = dbio.check(self.ElTariffFeedIn.val),
                         ElTariffCHP = dbio.check(self.ElTariffCHP.val),
                         ElCostYearTot = dbio.check(self.ElCostYearTot.val),
                         ElSalesYearFeedIn = dbio.check(self.ElSalesYearFeedIn.val),
                         ElSalesYearCHP = dbio.check(self.ElSalesYearCHP.val))

        # export to DB in old-fashioned CheckProc usage
        if self.exportToDB:
            self.el.saveToDB(ANo=0, warn=False)
            dbio.commit()
        
#------------------------------------------------------------------------------
    def showAllFETel(self):
#------------------------------------------------------------------------------
#   function for debug-plotting
#------------------------------------------------------------------------------

        print "====================="
        self.ElectricityNet.show()
        self.ElectricityGen.show()
        self.ElectricitySales.show()
        self.ElectricityTotYear.show()
        self.FECel.show()
        self.FECel_c.show()
        self.ElectricityMotors.show()
        self.ElectricityChem.show()
        self.ElectricityLight.show()
        self.ElectricityRef.show()
        self.ElectricityAC.show()
        self.ElectricityThOther.show()
        self.FEOel.show()
        self.FETel_c.show()
        self.FETel.show()
                  
        print "====================="
    
#------------------------------------------------------------------------------
    def screen(self):  
#------------------------------------------------------------------------------
#   screens all variables in the block
#------------------------------------------------------------------------------

        self.ElectricityNet.screen()
        self.ElectricityGen.screen()
        self.ElectricitySales.screen()
        self.ElectricityTotYear.screen()

        self.ElectricityMotors.screen()
        self.ElectricityChem.screen()
        self.ElectricityLight.screen()
        self.ElectricityRef.screen()
        self.ElectricityAC.screen()
        self.ElectricityThOther.screen()
        
        self.FEOel.screen()
        self.FETel.screen()
        self.FECel.screen()

        self.ElTariffCTot.screen()

        self.ElCostYearTot.screen()
        self.ElSalesYearFeedIn.screen()
        self.ElSalesYearCHP.screen()

#------------------------------------------------------------------------------
    def check(self):     #function that is called at the beginning when object is created
        """
        main function carrying out the check of the block

        FEC/FET/FEOel: net consumption of electrical energy (incl. self-generation)
        FEC/FETel_c:    gross consumption of electrical energy (independent of whether self-generated or purchased)
        -> FEC/FETel = FEC/FETel_c - ElGen

        FECel = ElectricityTotYear (purchase) + ElGen (self-generation) - ElSales (sales to grid) 
        """

        if DEBUG in ["ALL","MAIN","BASIC"]:
            print "-------------------------------------------------"
            print " Checking FETel"
            print "-------------------------------------------------"

        for n in range(1):

            cycle.initCheckBalance()

            if DEBUG in ["ALL","MAIN"]:
                print "-------------------------------------------------"
                print "Ciclo %s"%n
                print "-------------------------------------------------"
        
# Step 1: Call all calculation routines in a given sequence

            if DEBUG in ["ALL","MAIN"]:
                print "Step 1: calculating from left to right (CALC)"
            
            self.ElectricityNet.push(calcDiff("ElectricityNet1",self.ElectricityGen,self.ElectricitySales))
            self.FECel_c.push(calcSum("FECel_c1",self.ElectricityNet,self.ElectricityTotYear))
            self.FECel.push(calcDiff("FECel",self.FECel_c,self.ElectricityGen,parType="S"))

            self.FEOel.push(calcSum3("FEOel1",self.ElectricityMotors,self.ElectricityChem,self.ElectricityLight))
            self.FETel_c.push(calcDiff("FETel_c1",self.FECel_c,self.FEOel))
            self.FETel_c.push(calcSum3("FETel_c2",self.ElectricityRef,self.ElectricityAC,self.ElectricityThOther))
            self.FETel.push(calcDiff("FETel",self.FETel_c,self.ElectricityGen,parType="S"))

            self.ElCostYearTot.push(calcProd("ElCostYearTot1",self.ElTariffCTot,self.ElectricityTotYear, unitType = "PRICE"))
#### WARNING: power term not yet included in calculations !!!! ###
            
            self.ElSalesYearFeedIn.push(calcProd("ElSalesYearFeedIn",self.ElTariffFeedIn,self.ElectricityGen, unitType = "ENERGY"))
            self.ElSalesYearCHP.push(calcProd("ElSalesYearCHP",self.ElTariffCHP,self.ElectricitySales, unitType = "ENERGY"))

            if DEBUG in ["ALL","MAIN"]:
                self.showAllFETel()

# Step 2: Cross check the variables

                print "Step 2: cross checking"

            self.ccheckAll()

            if DEBUG in ["ALL","MAIN"]:
                self.showAllFETel()

# Step 3: Adjust the variables (inverse of calculation routines)

                print "Step 3: calculating from right to left (ADJUST)"

            adjustProd_stack(self.ElCostYearTot,self.ElTariffCTot,self.ElectricityTotYear)
            adjustProd_stack(self.ElSalesYearFeedIn,self.ElTariffFeedIn,self.ElectricityGen)
            adjustProd_stack(self.ElSalesYearCHP,self.ElTariffCHP,self.ElectricitySales)

            adjustDiff_stack(self.FETel,self.FETel_c,self.ElectricityGen)
            adjustSum3_stack(self.FETel_c,self.ElectricityRef,self.ElectricityAC,self.ElectricityThOther)
            adjustDiff_stack(self.FETel_c,self.FECel_c,self.FEOel)
            adjustSum3_stack(self.FEOel,self.ElectricityMotors,self.ElectricityChem,self.ElectricityLight)
            
            adjustDiff_stack(self.FECel,self.FECel_c,self.ElectricityGen)
            adjustSum_stack(self.FECel_c,self.ElectricityNet,self.ElectricityTotYear)
            adjustDiff_stack(self.ElectricityNet,self.ElectricityGen,self.ElectricitySales)
                        
            if DEBUG in ["ALL","MAIN"]:
                self.showAllFETel()

# Step 4: second cross check the variables

                print "Step 4: second cross checking"
               
            self.ccheckAll()
            
            if DEBUG in ["ALL","MAIN"]:
                self.showAllFETel()
        

# End of the cycle. Last print in DEBUG mode


        if DEBUG in ["ALL","BASIC","MAIN"]:
            self.showAllFETel()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()

#------------------------------------------------------------------------------
    def ccheckAll(self):     
#------------------------------------------------------------------------------
#   ccheck block
#------------------------------------------------------------------------------
        self.ElectricityNet.check()
        self.ElectricityTotYear.check()
        self.ElectricityGen.check()
        self.ElectricitySales.check()

        self.ElectricityMotors.check()
        self.ElectricityChem.check()
        self.ElectricityLight.check()
        self.ElectricityRef.check()
        self.ElectricityAC.check()
        self.ElectricityThOther.check()
        
        self.FECel_c.check()
        self.FECel.check()
        self.FEOel.check()
        self.FETel_c.check()
        self.FETel.check()

        self.ElTariffCTot.check()
        self.ElTariffFeedIn.check()
        self.ElTariffCHP.check()

        self.ElCostYearTot.check()
        self.ElSalesYearFeedIn.check()
        self.ElSalesYearCHP.check()

                       
#------------------------------------------------------------------------------
    def estimate(self):  
#------------------------------------------------------------------------------
#   estimates some of the data that are not sufficiently precise
#   should be a subset of the data that are within screen
#   (not necessarily ALL data have to be estimated)
#------------------------------------------------------------------------------

        if "CHP" not in Status.int.systemModel.eqTypes:
            self.ElectricityGen.setEstimate(0,limits=(0,0))
        else:
            self.ElTariffFeedIn.setEstimate(0.0)
            self.ElTariffCHP.setEstimate(self.ElTariffCTot.val,limits=(self.ElTariffCTot.valMin, self.ElTariffCTot.valMax))
            
        self.ElectricitySales.setEstimate(self.ElectricityGen.val,limits=(self.ElectricityGen.valMin,self.ElectricityGen.valMax))
        self.FEOel.setEstimate(0,limits=(0,0))
        
        self.ElTariffCTot.setEstimate(0.08,limits=(0.05,0.15))

# limits: optional and fix absolute minimum and maximum values
# sqerr: optional input that fixes the (stochastic) relative square error

#==============================================================================

if __name__ == "__main__":

# direct connecting to SQL database w/o GUI. for testing only
    stat = Status("testCheckFETel")
##    Status.SQL = MySQLdb.connect(user="root", db="einstein")
##    Status.DB = pSQL.pSQL(Status.SQL, "einstein")
    Status.PId = 41
    Status.ANo = -1
#..............................................................................
    
    screen = CCScreen()
    
    ccFETel = CheckFETel()       # creates an instance of class CCheck
    ccFETel.check()
    ccFETel.exportData()

    ccFETel.screen()
    screen.show()
    
#==============================================================================
