# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CheckHX
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Consistency check of heat exchanger data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from ccheckFunctions import *
from ccPar import *
from GUI.status import *
from constants import *
import math
import database.dbio as dbio

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"


class CheckHX():
#------------------------------------------------------------------------------
#   Carry out consistency checking for heat exchanger h
#------------------------------------------------------------------------------
    def __init__(self, h, checkInPlace = False):

        self.checkInPlace = checkInPlace
        self.index = h
        self.dataGroup = "HX"
        
        # assign a variable to all intermediate/calculated values needed
        self.QWH = CCPar("QWH", unitType = "ENERGY")

        self.DTSource = CCPar("DTSource", parType = "DT")
        self.DTSink = CCPar("DTSink", parType="DT")
        self.UpperDT = CCPar("UpperDT", parType="DT")
        self.UpperDT.valMin = 2.0
        
        self.LowerDT = CCPar("LowerDT", parType="DT")
        self.LowerDT.valMin = 2.0
        
        self.HPerYear = CCPar("HPerYear", unitType = "TIME")
        self.HPerYear.valMax = YEAR
        self.attachedProcessNumbers = []    # used to get the checkProc results for
                                            # HPerYearProc of the right processes

        self.OpCostVar = CCPar("OpCostVar", unitType = "PRICE")
        self.OpCostTot = CCPar("OpCostTot", unitType = "PRICE")

        self.importData(h)

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAll()

#------------------------------------------------------------------------------
    def importData(self, h):  
#------------------------------------------------------------------------------
#   imports data from SQL tables (from AlternativeProposalNo = -1)
#------------------------------------------------------------------------------
        self.HXNo = h + 1

        if self.checkInPlace:
            ANo = Status.ANo
        else:
            ANo = -1
        
        # assign empty CCPar to all questionnaire parameters
        self.QHX = CCPar("QHX", priority=2, unitType = "ENERGY")
        self.QdotHX = CCPar("QdotHX", unitType = "POWER")
        self.QdotHXNom = CCPar("QdotHXNom", unitType = "POWER")
        self.DQHX = CCPar("DQHX", unitType = "ENERGY")
        self.UA = CCPar("UA", unitType = "HEATTRANSFERCOEF")
        self.HXLMTD = CCPar("HXLMTD", parType = "DT")
        self.HXLMTDNom = CCPar("HXLMTDNom", parType = "DT")
        
        self.HXTSourceInlet = CCPar("HXTSourceInlet", parType = "T")
        self.HXTSourceOutlet = CCPar("HXTSourceOutlet", parType = "T")
        self.HXTSinkInlet = CCPar("HXTSinkInlet", parType = "T")
        self.HXTSinkOutlet = CCPar("HXTSinkOutlet", parType = "T")
        self.LMTDCounterCurrent = CCPar("LMTDCounterCurrent", parType = "DT")

        self.HXTSourceInletNom = CCPar("HXTSourceInletNom", parType = "T")
        self.HXTSourceOutletNom = CCPar("HXTSourceOutletNom", parType = "T")
        self.HXTSinkInletNom = CCPar("HXTSinkInletNom", parType = "T")
        self.HXTSinkOutletNom = CCPar("HXTSinkOutletNom", parType = "T")

        self.TurnKeyPrice = CCPar("TurnKeyPrice")        
        self.OMFix = CCPar("OMFix", priority=2, unitType = "PRICE")
        self.OMVar= CCPar("OMVar", priority=2)

        qheatexchanger = Status.int.systemModel.hx[h]
        if qheatexchanger is not None:
            self.name = qheatexchanger.HXName

            HXID = qheatexchanger.id
            
##            # get maximal running time by getting the running time
##            # of the attached sink/source processes
            
            hx = Status.int.hx[h]
                    
            self.sinkinlets = []
            self.sourceinlets = []
            if not self.checkInPlace:
                for conn in hx.sinkstreams:
                    dconn = conn.compactConnToUpstreamComponent()      # splitters are not shown explicitely
                    if dconn.objectType in ["ProcessH", "ProcessC"]:
                        k = dconn.objectNo
                        self.attachedProcessNumbers.append(k)
                        ccProc = Status.mod.moduleCC.ccProc[k]
                        if dconn.componentNo == 0:
                            self.sinkinlets.append(ccProc.PT)
                        elif dconn.componentNo == -1:
                            self.sinkinlets.append(ccProc.PTStartUp)
                        else:
                            if dconn.objectType == "ProcessH":
                                self.sinkinlets.append(ccProc.PTInFlow[dconn.componentNo-1])
                            else:
                                self.sinkinlets.append(ccProc.PTOutFlow[dconn.componentNo-1])                            

                for conn in hx.sourcestreams:
                    dconn = conn.compactConnToUpstreamComponent()      # splitters are not shown explicitely
                    if dconn.objectType in ["ProcessH", "ProcessC"]:
                        k = dconn.objectNo
                        self.attachedProcessNumbers.append(k)
                        ccProc = Status.mod.moduleCC.ccProc[k]
                        if dconn.componentNo == 0:
                            self.sourceinlets.append(ccProc.PT)
                        elif dconn.componentNo == -1:
                            self.sourceinlets.append(ccProc.PTStartUp)
                        else:
                            if dconn.objectType == "ProcessH":
                                self.sourceinlets.append(ccProc.PTOutFlow[dconn.componentNo-1])
                            else:
                                self.sourceinlets.append(ccProc.PTInFlow[dconn.componentNo-1])
                            
            
            self.HXTSourceInletNom.setValue(hx.ThiNom)
            self.HXTSourceOutletNom.setValue(hx.ThoNom)
            self.HXTSinkInletNom.setValue(hx.TciNom)
            self.HXTSinkOutletNom.setValue(hx.TcoNom)
            
##            try:
##                HXTSourceIn = max(sourceinlets)
##                HXTSourceOut = min(sourceoutlets)
##                HXTSinkIn = min(sinkinlets)
##                HXTSinkOut = max(sinkoutlets)
##            except:
##                logDebug("checkHX: some problem with reading temperatures of sources or sinks")
                
##            if sourceinlets != []:
##                self.HXTSourceInlet.setValue(max(sourceinlets))
##            if sourceoutlets != []:
##                self.HXTSourceOutlet.setValue(min(sourceoutlets))
##            if sinkinlets != []:
##                self.HXTSinkInlet.setValue(min(sinkinlets))
##            if sinkoutlets != []:
##                self.HXTSinkOutlet.setValue(max(sinkoutlets))
            
##                if self.HXTSinkInlet.val == None or self.HXTSinkOutlet.val == None or \
##                   self.HXTSourceInlet.val == None or self.HXTSourceOutlet.val == None:
##                    self.LMTDCounterCurrent.setValue(1)
##                else:
##                    HXT = (HXTSourceIn-HXTSinkOut)/(HXTSourceOut-HXTSinkIn)
##                    if HXT > 1:
##                        self.LMTDCounterCurrent.setValue(((HXTSourceIn-HXTSinkOut)-(HXTSourceOut-HXTSinkIn))/(math.log(HXT)))
##                    else:
##                        self.LMTDCounterCurrent.setValue(((HXTSourceIn-HXTSinkOut)-(HXTSourceOut-HXTSinkIn))/1)
            
##            self.QdotHX.setValue(qheatexchanger.QdotHX)
            self.QdotHXNom.setValue(qheatexchanger.QdotHXNom)

            UA = qheatexchanger.UA
            self.directX = UA is not None and UA < 0
            if not self.directX:
                self.UA.setValue(qheatexchanger.UA)

##            self.HXLMTD.setValue(qheatexchanger.HXLMTD)
            self.HXLMTDNom.setValue(qheatexchanger.HXLMTDNom)

            # one of both parameters QdotNom or LMTDNom HAS to be defined, otherwise LMTDnom = 10K is fixed as default
            if self.QdotHXNom.val is None and self.HXLMTDNom.sqerr >= 10.0:
                self.HXLMTDNom.setValue(10.0, err = 0.0)
            
            self.QHX.setValue(qheatexchanger.QHX)  

            self.TurnKeyPrice.setValue(qheatexchanger.TurnKeyPrice)
            if ANo <= 0:
                self.TurnKeyPrice.setValue(0)   # present state equipment has investment cost zero
                qheatexchanger.TurnKeyPrice = dbio.check(self.TurnKeyPrice.val)

            self.OMFix.setValue(qheatexchanger.OMFix)
            self.OMVar.setValue(qheatexchanger.OMVar)


    def updateHPerYearEstimate(self, seq):
        """Update the HPerYear estimation with the minimum of the 
        CCPar estimates in the list seq.
        
        """
        if seq != []:
            self.HPerYear.valMin = max(self.HPerYear.valMin, min([H.valMin for H in seq]))
            self.HPerYear.valMax = min(self.HPerYear.valMax, min([H.valMax for H in seq]))
            if self.HPerYear.val is None or self.HPerYear.val < self.HPerYear.valMin or self.HPerYear.val > self.HPerYear.valMax:
                self.HPerYear.val = 0.5 * (self.HPerYear.valMin + self.HPerYear.valMax)
                
            self.HPerYear.constrain()
            if DEBUG in ["ALL", "MAIN"]:
                print "-------------------------------------------------"
                print " Estimating operating hours for heat exchanger no. ", self.HXNo
                self.HPerYear.show()
                print "-------------------------------------------------"
        

    def exportData(self):  
#------------------------------------------------------------------------------
#   stores corrected data in SQL (under AlternativeProposalNo = 0)
#------------------------------------------------------------------------------

        if self.checkInPlace:
            ANo = max(Status.ANo, 0)
        else:
            ANo = 0
        
##        qheatexchangerTable = Status.DB.qheatexchanger.ProjectID[Status.PId].AlternativeProposalNo[ANo].HXNo[self.HXNo]
##        if len(qheatexchangerTable) > 0:
##            qheatexchanger = qheatexchangerTable[0]
        qheatexchanger = Status.prj.getHX(PId=Status.PId, ANo=ANo, no=self.HXNo)
        if qheatexchanger is not None:

            # calculate equipment type specific cost parameters

            if ANo > 0: # for inplace checking directly estimate economic parameters before saving
                self.estimateEconomicParameters()

            UA = -1.0 if self.directX else self.UA.val
            
            qheatexchanger.set(QHX = dbio.check(self.QHX.val),
                               QdotHX = dbio.check(self.QdotHX.val) if (self.QdotHX.sqerr < 0.1) else dbio.check(None),
                               QdotHXNom = dbio.check(self.QdotHXNom.val) if (self.QdotHXNom.sqerr < 0.1) else dbio.check(None),
                               UA = dbio.check(UA),
                               HXLMTD = dbio.check(self.HXLMTD.val) if (self.HXLMTD.sqerr < 0.1) else dbio.check(None),
                               HXLMTDNom = dbio.check(self.HXLMTDNom.val) if (self.HXLMTDNom.sqerr < 0.1) else dbio.check(None),
                               TurnKeyPrice = dbio.check(self.TurnKeyPrice.val),
                               OMFix = dbio.check(self.OMFix.val),
                               OMVar = dbio.check(self.OMVar.val))
            
            dbio.commit()

    def showAll(self):
#------------------------------------------------------------------------------
#   plotting of values for debugging
#------------------------------------------------------------------------------
        print "====================="
        self.QHX.show()
        self.QWH.show()
        self.QdotHX.show()
        self.HPerYear.show()

        self.HXLMTD.show()
        self.UA.show()

        self.DTSource.show()
        self.DTSink.show()
        self.UpperDT.show()
        self.LowerDT.show()
        self.DQHX.show()

        self.HXTSinkInlet.show()
        self.HXTSinkOutlet.show()
        self.HXTSourceInlet.show()
        self.HXTSourceOutlet.show()

        self.OpCostTot.show()
        print "====================="

    def screen(self):  
#------------------------------------------------------------------------------
#   screens all variables in the block
#------------------------------------------------------------------------------
        self.QHX.screen()
        self.DTSource.screen()
        self.DTSink.screen()
        self.UpperDT.screen()
        self.LowerDT.screen()
#        self.LMTD1.screen()
        self.DQHX.screen()
        self.QdotHX.screen()
        self.QdotHXNom.screen()
        self.HXLMTDNom.screen()

        self.OMFix.screen()
        self.OMVar.screen()

        
    def check(self):     
        """main function carrying out the check of the block"""
        if DEBUG in ["ALL", "MAIN"]:
            print "-------------------------------------------------"
            print " Checking heat exchanger no. ", self.HXNo
            print "-------------------------------------------------"

        for n in range(1):

            cycle.initCheckBalance()

            # Step 1: Call all calculation routines in a given sequence
            if DEBUG in ["ALL"]:
                print "-------------------------------------------------"
                print "Cycle %s" % n
                print "-------------------------------------------------"
                print "Step 1: calculating from left to right (CALC)"

            # Update maximum tempeatures by connected process
            if len(self.sourceinlets) > 0:
##                for source in self.sourceinlets: source.show()
                
                self.HXTSourceInlet.valMin = max(self.HXTSourceInlet.valMin, min([T.valMin for T in self.sourceinlets]))
                self.HXTSourceInlet.valMax = min(self.HXTSourceInlet.valMax, max([T.valMax for T in self.sourceinlets]))
                                            
            if len(self.sinkinlets) > 0:
            # Check Source Temperatures
##                for sink in self.sinkinlets: sink.show()
                
                self.HXTSinkInlet.valMin = max(self.HXTSinkInlet.valMin, min([T.valMin for T in self.sinkinlets]))
                self.HXTSinkInlet.valMax = min(self.HXTSinkInlet.valMax, max([T.valMax for T in self.sinkinlets]))
                                            
             # Check Source Temperatures
            self.DTSource.push(calcDiff("DTSource", self.HXTSourceInlet, self.HXTSourceOutlet, "DT"))

            # Check Sink Temperatures 
            self.DTSink.push(calcDiff("DTSink", self.HXTSinkOutlet, self.HXTSinkInlet, "DT"))
            
            # Check UpperDT
            self.UpperDT.push(calcDiff("UpperDT", self.HXTSourceInlet, self.HXTSinkOutlet, "DT"))
            
            # Check Lower DT
            self.LowerDT.push(calcDiff("LowerDT", self.HXTSourceOutlet, self.HXTSinkInlet, "DT"))
            
#            self.LMTD1.push(calcDiff("LMTD1", self.HXLMTD, self.LMTDCounterCurrent))
            self.HXLMTD.valMin = max(self.HXLMTD.valMin, min(self.UpperDT.valMin, self.LowerDT.valMin))
            self.HXLMTD.valMax = min(self.HXLMTD.valMax, max(self.UpperDT.valMax, self.LowerDT.valMax))
            
            # Check QHX
            self.QHX.push(calcProd("QHX", self.HPerYear, self.QdotHX, unitType = "ENERGY"))
            setIdentity(self.QWH, self.QHX)
            
            if not self.directX:
                self.QdotHX.push(calcProd("QdotHX", self.UA, self.HXLMTD, unitType = "POWER"))
                self.QdotHXNom.push(calcProd("QdotHXNom", self.UA, self.HXLMTDNom, unitType = "POWER"))
                    
            self.OpCostVar.push(calcProd("OpCostVar", self.OMVar, self.QHX, unitType = "PRICE"))
            self.OpCostTot.push(calcSum("OpCostTot", self.OMFix, self.OpCostVar))

            if DEBUG in ["ALL"]:
                self.showAll()

# Step 2: Cross check the variables

                print "Step 2: cross checking"

            self.checkAll()
            
            if DEBUG in ["ALL"]:
                self.showAll()

# Step 3: Adjust the variables (inverse of calculation routines)

                print "Step 3: calculating from right to left (ADJUST)"
            
            # Check Source Temperatures
            adjustDiff_stack(self.DTSource, self.HXTSourceInlet, self.HXTSourceOutlet)

            # Check Sink Temperatures 
            adjustDiff_stack(self.DTSink, self.HXTSinkOutlet, self.HXTSinkInlet)
            
            # Check UpperDT
            adjustDiff_stack(self.UpperDT, self.HXTSourceInlet, self.HXTSinkOutlet)
            
            # Check Lower DT
            adjustDiff_stack(self.LowerDT, self.HXTSourceOutlet, self.HXTSinkInlet)
        
#            adjustDiff_stack(self.LMTD1, self.HXLMTD, self.LMTDCounterCurrent)
            
            # effective yearly operating time - nominal power - yearly energy transferred
            adjustProd_stack(self.QHX, self.HPerYear, self.QdotHX)
            self.QHX.push(self.QWH)
            
            # Q = UA * LMTD
            if not self.directX:
                adjustProd_stack(self.QdotHX, self.UA, self.HXLMTD)
                adjustProd_stack(self.QdotHXNom, self.UA, self.HXLMTDNom)

            adjustProd_stack(self.OpCostVar, self.OMVar, self.QHX)
            adjustSum_stack(self.OpCostTot, self.OMFix, self.OpCostVar)

            if DEBUG in ["ALL"]:
                self.showAll()
            
# Step 4: Cross check again the variables

                print "Step 4: second cross checking"

            self.checkAll()

            if DEBUG in ["ALL"]:
                self.showAll()

        if DEBUG in ["ALL", "BASIC"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()

    def checkAll(self):
        """"Carry out the check of all variables."""
        self.DTSource.check()
        self.DTSink.check()
        self.UpperDT.check()
        self.LowerDT.check()
        self.DQHX.check()
#        self.LMTD1.check()
        self.QdotHX.check()
        self.QdotHXNom.check()
        self.QWH.check()
        self.QHX.check()
        self.HPerYear.check()

        self.HXLMTD.check()
        self.HXLMTDNom.check()
        
        if not self.directX:
            self.UA.check()
            
        self.HXTSinkInlet.check()
        self.HXTSinkOutlet.check()
        self.HXTSourceInlet.check()
        self.HXTSourceOutlet.check()
        self.HXTSinkInletNom.check()
        self.HXTSinkOutletNom.check()
        self.HXTSourceInletNom.check()
        self.HXTSourceOutletNom.check()

        self.OMFix.check()
        self.OMVar.check()
        self.OpCostVar.check()
        self.OpCostTot.check()
        
    def estimate(self):  
        """ estimates some of the parameters
        """

        self.UpperDT.valMin = 2.0
        self.LowerDT.valMin = 2.0
        
    def estimateEconomicParameters(self):
        """ estimates the economic parameters for the equipment if not given
        """
                
        if self.TurnKeyPrice.val is None:
            if self.UA.val is not None:
                turnKey = 1000.*self.UA.val
            else:
                turnKey = 0.0
                
            self.TurnKeyPrice.setEstimate(turnKey)
            logWarning(_("Investment cost of heat exchanger %s is not defined. Estimated to %s"%(self.name, turnKey)))
                       
        if self.OMFix.val is None:
            if self.UA.val is not None:
                OMFix = 10.0*self.UA.val
            else:
                OMFix = 0.0
                
            self.OMFix.setEstimate(OMFix)
            logWarning(_("Parameters for O&M costs of heat exchanger %s are not defined. %s assumed")%(self.name, OMFix) \
                       + " " + _("for fix part"))
        if self.OMVar.val is None:
            self.OMVar.setEstimate(0.0)
            logWarning(_("Parameters for O&M costs of heat exchanger %s are not defined. %s assumed")%(self.name, 0.0) \
                       + " " + _("for variable part"))


if __name__ == "__main__":
    import einstein.GUI.pSQL as pSQL
    import MySQLdb
# direct connecting to SQL database w/o GUI. for testing only
    stat = Status("testCheckProc")
    Status.SQL = MySQLdb.connect(user="root", db="einstein")
    Status.DB = pSQL.pSQL(Status.SQL, "einstein")
    Status.PId = 41
    Status.ANo = -1
    
    screen = CCScreen()
    
    ccHX = CheckHX(1)       # creates an instance of class CCheck
    ccHX.check()
    ccHX.exportData(1)

    ccHX.screen()
    screen.show()

