# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CheckMassFlow
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Consistency check of mass flows (raw materials, intermediate products
#       and final products
#
#==============================================================================
#	
#------------------------------------------------------------------------------		
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2017
#	www.energyxperts.net / info@energyxperts.net
#
#============================================================================== 				
from ccPar import CCPar, CCRow
from ccTracker import cycle
from ccCheckers import ccheck1
from ccheckFunctions import calcRowSum, adjRowSum_stack, calcProd, adjustProd_stack
from GUI.status import Status
from constants import DEBUG
import database.dbio as dbio
from modules.messageLogger import logError

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"

class Node():
    def __init__(self, id, objectType=None, objectId=None, name=_("New node"), autoLoad=True):
        """ use cases:
        id defined -> reads the data from qnode in DB
        id=None but objectType and objectId defined: -> if autoLoad: looks for an appropriate entry in qnode
            (should exist if node is connected) or alternatively creates a virtual Node (id=None)

        qin, qout: vector of class Flow objects (inputs, outputs)
        
        """
        
        self.qin = []       # flow-objects of in/outflows
        self.qout = []

        self.qinVal = []       # values of in/outflows
        self.qoutVal = []

        self.qinInfo = []       # flow objects of information inflows
        self.qinInfoVal = []
        self.id = id        # id of row in qnode-table
        self.objectType=objectType
        self.objectId = objectId  # id of corresponding object (product, process, ...)
        self.q = None
        self.k = None # process index. set only for PROC nodes

        # special parameters only for objectType INT
        self.name = name
        self.nodeType = None
        self.scaleFactor = None
        self.splitFraction = None

        self.equation=None

        self.connected = False

        self.clearPosition()

        qnode = None
        if id is None and autoLoad:  # try to load from database
            qnode = dbio.get("qnode", ObjectId=objectId, ObjectType=objectType)
            if qnode is not None:
                self.id = qnode.id
            

        else:
            qnode = dbio.get("qnode", id=self.id)
            if qnode is not None:
                self.objectType = qnode.ObjectType
                self.objectId = qnode.ObjectId
                self.equation = qnode.Equation
            
        if self.objectType == "PROC":
            proc = dbio.get("qprocessdata", id=self.objectId)
            k = proc.ProcNo-1
            self.k = k
            inStreams = Status.int.systemModel.processStreamsIn[k]
            outStreams = Status.int.systemModel.processStreamsOut[k]
            self.inputs = [stream.Name for stream in inStreams]
            self.outputs = [stream.Name for stream in outStreams]

            self.getMassFlows(proc, inStreams, outStreams)
            
        elif self.objectType=="PROD":
            qproduct = dbio.get("qproduct", id=self.objectId)
            self.q = qproduct.QProdYear*1000 if (qproduct is not None and qproduct.QProdYear is not None) else None  # assume tons as default unit and convert to kg
            # positioning at the right
            self.pos = 0
            self.inputs = ["->"]
            self.outputs = []

            self.qinVal = [self.q]
            
        elif self.objectType=="RAW":
            if self.equation is None:
                self.equation = "%s;"%self.name

            self.nodeType="SRC"

                
            pars = self.equation.split(";")
            # pars: NAME; val
            self.name = pars[0]

            val = pars[1]
            self.q = None if val == "" else float(val)
            self.splitFraction = [self.q/1000 if self.q is not None else None]
            self.scaleFactor = []
            
            self.inputs = []
            self.outputs = ["->"]
            # positioning at the left
            self.pos = -2

            self.qoutVal = [self.q]
            
        elif self.objectType=="INT":
            if self.equation is None:
                self.equation = "%s;MIX;1.;1.;1.;;;;"%self.name

            pars = self.equation.split(";")
            # pars: NAME; TYPE; f1; ... fn; in; out1; ... outn -> total = ncon + 2
            self.name = pars[0]
            self.nodeType = pars[1]
            ncon = (len(pars)-3)/2
            if self.nodeType=="MIX":
                self.scaleFactor = [None if par=="" else float(par) for par in pars[2:ncon+2]]
                self.qinVal = [None if par=="" else float(par) for par in pars[ncon+2:2*ncon+2]]
                par = pars[2*ncon+2]
                self.qoutVal = [None if par=="" else float(par)]
                self.splitFraction = [1.0]
            elif self.nodeType=="SPLIT":
                self.splitFraction = [None if par=="" else float(par) for par in pars[2:ncon+2]]
                self.scaleFactor = [1.0]
                par = pars[ncon+2]
                self.qinVal = [None if par=="" else float(par)]
                self.qoutVal = [None if par=="" else float(par) for par in pars[ncon+3:2*ncon+3]]
            self.updateNodeData()

        else:
            logError("Erroneous objectType in node id. %s"%(self.id))

        # space for connections (to be filled with Flow-objects)
        self.qin = [None for con in self.inputs]
        self.qinInfo = [[] for con in self.inputs]
        self.qout = [None for con in self.outputs]

    def __repr__(self):
        objectInfo = self.objectId if self.objectType != "INT" else "[%s in:out %s] %s"%(len(self.qin), len(self.qout), self.equation)
        string = "Node(%s): %s (%s-%s) at (%s, %s): q = %s"%(self.id, self.name, self.objectType, objectInfo, self.xpos, self.ypos, self.q)
        return string
        
    def clearPosition(self):
        # attributes for positioning in grid
        self.lastSrc = -1001
        self.firstSnk = 1
        self.lastSnk = -1001
        self.xpos = None    # horizontal position, integer value from 0 to ncol-1
        self.ypos = None    # vertical position, integer value from 0 to INF
        
    def add(self, flow, direction):
        if direction > 0:
            self.qin.append(flow)
        else:
            self.qout.append(flow)

    def saveToDB(self):
        row = {"ObjectId": self.objectId,
           "ObjectType": self.objectType,
           "Equation": self.equation}
        
        if self.id is None:     
            self.id = dbio.insertRow("qnode", **row)
        else:
            row = {"ObjectId": self.objectId,
                   "ObjectType": self.objectType,
                   "Equation": self.equation}
            qnode = dbio.get("qnode", id=self.id)
            qnode.set(**row)

        if self.objectType=="PROD":  # save value for qproduct nodes (is not stored elsewhere, as in the case of process nodes)
            qproduct = dbio.get("qproduct", id=self.objectId)
            if qproduct is not None:
                qproduct.QProdYear = float(self.q / 1000) if self.q is not None else None

    def deleteFromDB(self):
        dbio.delete("qnode", id=self.id)
        self.id = None

    def updateNodeData(self):
        """ special function for internal nodes
            updates the node parameters from NodeData pars
        """
        if self.objectType == "INT":   
            if len(self.scaleFactor) == 1:
                self.inputs = ["in"]
            else:
                self.inputs = ["in %s"%(n+1) for n in range(len(self.scaleFactor))]

            if len(self.splitFraction) == 1:
                self.outputs = ["out"]
            else:
                self.outputs = ["out %s"%(n+1) for n in range(len(self.splitFraction))]

            if self.nodeType == "MIX":
                parVect = self.scaleFactor
            else:
                parVect = self.splitFraction
            self.equation = self.name + ";" + self.nodeType
            for val in parVect:
                self.equation += ";" + ("" if val is None else "%s"%val)
            for val in self.qinVal:
                self.equation += ";" + ("" if val is None else "%s"%val)
            for val in self.qoutVal:
                self.equation += ";" + ("" if val is None else "%s"%val)

        elif self.objectType == "RAW":
            val = self.splitFraction[0]
            self.inputs = []
            self.outputs = ["->"]
            self.equation = self.name + ";" + ("" if val is None else "%s"%(val*1000))
            

    def getMassFlows(self, proc, inStreams, outStreams):
        """ auxiliary function for calculating yearly mass flow
            - not directly stored in streams - tables
        """

        hPerYearInFlow = proc.HPerYearInFlow
        hPerYearOutFlow = proc.HPerYearOutFlow

        self.qinVal = [None for stream in inStreams]
        self.qoutVal = [None for stream in outStreams]

        for s, stream in enumerate(inStreams):
            try:
                self.qinVal[s] = hPerYearInFlow*stream.mInFlowNom
            except:
                pass

        for s, stream in enumerate(outStreams):
            try:
                self.qoutVal[s] = hPerYearInFlow*stream.mOutFlowNom
            except:
                pass
        
class Flow():
    def __init__(self, fromNode, fromCon, toNode, toCon, flow=None, id=None, real=True):
        self.fromNode = fromNode
        self.toNode = toNode
        
        self.fromCon = fromCon
        self.toCon = toCon

        self.real = real    # False: information flow only
        
        self.flow = flow
        self.id = id

    def __repr__(self):
        msg = "Flow(%s): %s.%s -> %s.%s [q=%s]"%(self.id,
                                                 self.fromNode.id, self.fromCon,
                                                 self.toNode.id, self.toCon,
                                                 self.flow)
        return msg

    def exportToNode(self, delete=False):
        """ builds a link to itself at the connected nodes
        """

        if self.real:
            q = None if delete else self

            self.fromNode.qout[self.fromCon-1] = q
            self.toNode.qin[self.toCon-1] = q
        else:
            if self in self.toNode.qinInfo[self.toCon-1]:
                if delete:
                    self.toNode.qinInfo[self.toCon-1].remove(self)

            else:
                if not delete:
                    self.toNode.qinInfo[self.toCon-1].append(self)

    def saveToDB(self):
        row = {"ProjectID": Status.PId,
               "AlternativeProposalNo": Status.ANo,
               "FromId": self.fromNode.id,
               "FromCon": self.fromCon if self.real else -self.fromCon,
               "ToId": self.toNode.id,
               "ToCon": self.toCon if self.real else -self.toCon,
               "Flow": self.flow}
        if self.id is None:
            self.id = dbio.insertRow("qflow", **row)
        else:
            qflow = dbio.get("qflow", id=self.id)
            qflow.set(**row)

    def deleteFromDB(self):
        dbio.delete("qflow", id=self.id)

    def setMode(self, mode):
        real = (mode==0)
        if real != self.real:
            if self.real:           # change from real to virtual
                self.exportToNode(delete=True) # delete as 
            self.real = real
            self.exportToNode()
            return True
        else:
            return False
        


class MFSystem():
    """ builds the system of nodes and connections

    all information then is within
    node: list of nodes (class Node objects)
    flow: list of connections (class Flow objects)
    """

    def __init__(self, ANo=None):
        ANo = Status.ANo if ANo is None else ANo
        self.loadFromDB(ANo)
        self.deletedFlows = []
        self.deletedNodes = []
        
    def loadFromDB(self, ANo=None):
        ANo = Status.ANo if ANo is None else ANo
       
        qflows = dbio.select("qflow", ProjectID = Status.PId, AlternativeProposalNo=ANo)
        self.nf = len(qflows)
        connectedNodeIds = list(set([qflow.FromId for qflow in qflows] + [qflow.ToId for qflow in qflows]))
        self.nnc = len(connectedNodeIds)

        qnodes = [dbio.get("qnode", id=nodeId) for nodeId in connectedNodeIds]
        qnodes = [qnode for qnode in qnodes if qnode is not None]
        qproduct = dbio.select("qproduct", Questionnaire_id=Status.PId, AlternativeProposalNo=ANo)
        qprocessdata = dbio.select("qprocessdata", Questionnaire_id=Status.PId, AlternativeProposalNo=ANo)
        qrawMaterial = []   # TODO. add raw materials somehow somewhere ...

        
        self.processNode = [Node(None, objectId=proc.id, objectType="PROC", name=proc.Process) for proc in qprocessdata]
        self.productNode = [Node(None, objectId=product.id, objectType="PROD", name=product.Product) for product in qproduct]
        self.rawMaterialNode = [Node(qnode.id, objectType="RAW") for qnode in qnodes if qnode.ObjectType=="RAW"]
        self.internalNode = [Node(qnode.id, objectType="INT") for qnode in qnodes if qnode.ObjectType=="INT"]

        # enumerate products and raw materials
        for n, node in enumerate(self.productNode): node.k = n 
        for n, node in enumerate(self.rawMaterialNode): node.k = n 

        self.node = self.productNode + self.processNode + self.internalNode + self.rawMaterialNode
        nodeIds = [node.id for node in self.node]
        self.nn = len(self.node)

        
        self.flow = []
        self.deletedFlows = []
        self.deletedNodes = []

        # readin connection data from qflow table
        for q, qflow in enumerate(qflows):
            if qflow.FromId in connectedNodeIds and qflow.ToId in connectedNodeIds:            
                try: # try-except for the case of corrupt flow entries in the database
                    fromNo = nodeIds.index(qflow.FromId)
                    fromNode = self.node[fromNo]
                    fromCon = qflow.FromCon
                    
                    toNo = nodeIds.index(qflow.ToId)
                    toNode = self.node[toNo]
                    toCon = qflow.ToCon

                    flowVal = qflow.Flow

                    if fromCon > 0 and toCon > 0:
                        flow = Flow(fromNode, fromCon, toNode, toCon, flow=flowVal, id=qflow.id)
                    else:
                        flow = Flow(fromNode, -fromCon, toNode, -toCon, flow=flowVal, id=qflow.id, real=False)

                    flow.exportToNode()
                    self.flow.append(flow)

                except:
                    logError("Error loading flow from data base: %s"%qflow)
                    flow = Flow(0, 0, 0, 0, id=qflow.id)
                    flow.deleteFromDB()
                    dbio.commit()
            else:
                logError("Error loading flow from data base: %s"%qflow)
                flow = Flow(0, 0, 0, 0, id=qflow.id)
                flow.deleteFromDB()
                dbio.commit()
                

    def saveToDB(self):

        for node in self.node:
            node.connected = False

        for flow in self.flow:
            flow.fromNode.connected = True
            flow.toNode.connected = True

        for node in self.node:
            if node.connected or node.objectType=="RAW":
                node.saveToDB()
            else:
                node.deleteFromDB()

        for node in self.deletedNodes:
            node.deleteFromDB()
        self.deletedNodes = []

        for flow in self.flow:
            flow.saveToDB()

        for flow in self.deletedFlows:
            flow.deleteFromDB()
        self.deletedFlows = []

        dbio.commit()

class CCNode():
    """ complement to node for CC checking purposes
        processes the cons and establishes links between the
        CCPars in the ModuleCC structure

    """
    def __init__(self, node):
        self.origin = node

        self.qin = CCRow(self.origin.name+".qin", len(node.qin))  # space for connections. will be overwritten where connected
        self.qinInfo = [[] for q in node.qin]
        self.q = CCPar(self.origin.name+".q", unitType="MASS")
        self.q.setValue(self.origin.q)
        self.qinTarget = None
        self.checkBalance = False
        self.scaleFactor = None
        self.splitFraction = None

        if self.origin.objectType == "PROC":
            k = self.origin.k
            self.qinTarget = Status.mod.moduleCC.ccProc[k].mInFlowYear
            self.qout = Status.mod.moduleCC.ccProc[k].mOutFlowYear
            
        elif self.origin.objectType == "INT":
            self.qout = CCRow(self.origin.name+".qout", len(node.qout))
            self.qinTarget = CCRow(self.origin.name+".qin*f", len(node.qin))
            if self.origin.nodeType == "MIX":
                self.checkBalance = True
                if len(self.origin.scaleFactor)==1 and self.origin.scaleFactor[0]==1.0: # special case: 1:1 mixer
                    self.qout[0] = self.q
                else:
                    self.scaleFactor = CCRow(self.origin.name+".fscale", len(node.qin))
                    for i, par in enumerate(self.scaleFactor):
                        par.setValue(self.origin.scaleFactor[i])
                    
            elif self.origin.nodeType == "SPLIT":
                if len(self.origin.splitFraction)==1 and self.origin.splitFraction[0]==1.0: # special case: 1:1 splitter
                    self.qout[0] = self.q
                    
                else:
                    self.splitFraction = CCRow(self.origin.name+".fsplit", len(node.qout))
                    for i, par in enumerate(self.splitFraction):
                        par.setValue(self.origin.splitFraction[i])

        elif self.origin.objectType == "RAW":
            self.qout = [self.q]
            self.checkBalance = True

        elif self.origin.objectType == "PROD":
            self.qout = []            
            self.qinTarget = [self.q]

        else:
            logError("CCNode.__init__: erroneous object type %s"%self.origin.objectType)

    def __repr__(self):
        return "CCNode: origin %s"%self.origin
    
    def check(self):

        if self.qin is not None and self.qinTarget is not None and len(self.qin) > 0:

            if self.scaleFactor is None:
                for n, qin in enumerate(self.qin):
                    # fetch info flows if available
                    for qinInfo in self.qinInfo[n]:
                        ccheck1(self.qin[n], qinInfo)
                        
                    ccheck1(self.qin[n], self.qinTarget[n])
            else:
                # qinTarget[n] = fscale[n]*qin[n]
                for n, qin in enumerate(self.qin):

                    self.qinTarget[n].push(calcProd(self.qinTarget[n].name, self.scaleFactor[n], self.qin[n]))
                    self.qinTarget[n].check()
                    adjustProd_stack(self.qinTarget[n], self.scaleFactor[n], self.qin[n])
                    self.scaleFactor[n].check()
                    self.qin[n].check()
                     
        # internal equation: q = SUM(qin_Target)
        if len(self.qin) > 0:
            self.q.push(calcRowSum(self.q.name, self.qinTarget))
            self.q.check()

            adjRowSum_stack(self.q, self.qinTarget)
                     
            for q in self.qinTarget:
                q.check()

        if len(self.qout) > 0:
            # q = SUM(qout)
            if self.checkBalance:
                self.q.push(calcRowSum(self.q.name, self.qout))
                self.q.check()
                adjRowSum_stack(self.q, self.qout)
                for q in self.qout:
                     q.check()
            # qout[n] = fsplit[n] * q
            elif self.splitFraction is not None:
                for n, qout in enumerate(self.qout):
                    qout.push(calcProd(qout.name, self.splitFraction[n], self.q))
                    qout.check()
                    adjustProd_stack(qout, self.splitFraction[n], self.q)
                    self.splitFraction[n].check()
                self.q.check()                
        
    def show(self):
        print "CCNode.show: %s"%self.origin.name
        self.q.show()
        print "inputs:"
        for qin in self.qin:
            qin.show()
        print "outputs"
        for qout in self.qout:
            qout.show()

        
class CheckMassFlow():
    def __init__(self, checkInPlace = False):

        self.dataGroup = "MF"
        
        self.importData()

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAll()

    def importData(self):  
        """ imports data from SQL tables (from AlternativeProposalNo = -1)
        """

        self.sysMF = MFSystem()
        self.ccnode = [CCNode(node) for node in self.sysMF.node]

        for flow in self.sysMF.flow:
            fromNo = self.sysMF.node.index(flow.fromNode)
            fromCon = flow.fromCon
            toNo = self.sysMF.node.index(flow.toNode)
            toCon = flow.toCon

            if flow.real:
                self.ccnode[toNo].qin[toCon-1] = self.ccnode[fromNo].qout[fromCon-1]
            else:
                self.ccnode[toNo].qinInfo[toCon-1].append(self.ccnode[fromNo].qout[fromCon-1])

        if DEBUG in ["ALL", "MAIN", "BASIC"]:
            self.showAll()            

    def exportData(self):  
        """
        stores corrected data in SQL (under AlternativeProposalNo = 0)
        """
        
        ANo = 0          
        sysTmp = MFSystem(ANo)
        for n, checkedNode in enumerate(sysTmp.node):
            checkedNode.qinVal = [q.val for q in self.ccnode[n].qin]
            checkedNode.qoutVal = [q.val for q in self.ccnode[n].qout]
            checkedNode.q = self.ccnode[n].q.val

            if checkedNode.objectType=="RAW":
                q = checkedNode.qoutVal[0]
                checkedNode.splitFraction = [q if q is None else q/1000]
                
            checkedNode.updateNodeData()
            checkedNode.saveToDB()
            
        dbio.commit()

    def showAll(self):
        """
        plotting of values for debugging
        """
        
        print "====================="
        for node in self.ccnode:
            print "Node:------------"
            node.q.show()
            print "In:------------"
            for flow in node.qin:
                flow.show()
            print "InTarget:------------"
            if node.qinTarget:
                for flow in node.qinTarget:
                    flow.show()
                print "Out:------------"
            for flow in node.qout:
                flow.show()
            print "----------------"

        print "====================="

        
    def check(self):     
        """main function carrying out the check of the block"""
        if DEBUG in ["ALL", "MAIN"]:
            print "-------------------------------------------------"
            print " Checking yearly mass flows in the system"
            print "-------------------------------------------------"

        for n in range(1):

            cycle.initCheckBalance()

            # Step 1: Call all calculation routines in a given sequence
            if DEBUG in ["ALL", "MAIN"]:
                print "-------------------------------------------------"
                print "Cycle %s" % n
                print "-------------------------------------------------"


            for node in self.ccnode:
                node.check()

            if DEBUG in ["ALL", "MAIN"]:
                self.showAll()

        if DEBUG in ["BASIC"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()

        

if __name__ == "__main__":
    print "No unit-demo available"

