# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CCheck (Consistency Check)
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from ccheckFunctions import *
from ccPar import CCPar, CCOne, CCRow, MAX_SQERR
from math import log
from modules.fluids import Fluid
from modules.messageLogger import *
from GUI.status import *
import database.dbio as dbio
from checkProc import CheckProc
from checkEq import CheckEq
from auxiliary.auxiliary import almostEqual

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"

class CheckPipe():
    def __init__(self, m, pipe=None, checkInPlace = False):

        if pipe is None: # keep this here temporarily for compatibility with other ccheck modules which still read/write from/to DB
            pipe = Status.int.systemModel.pipeObject[m]
            
        m = pipe.no
        self.pipe = pipe

        self.checkInPlace = checkInPlace
        self.exportToDB = not checkInPlace

        self.index = m
        self.dataGroup = "Pipe/Duct"
        
        """Carry out consistency checking for pipeduct m."""
        self.DistribCircFlow = CCPar("DistribCircFlow", unitType = "MASSFLOW")
        self.PercentRecirc = CCPar("PercentRecirc", parType="X")
        self.TFeedUp = CCPar("TFeedUp", parType="T")
        self.TotLengthDistPipe = CCPar("TotLengthDistPipe") #one way
        self.DDistPipe = CCPar("DDistPipe", unitType = "LENGTH")
        self.DeltaDistPipe = CCPar("DeltaDistPipe", unitType = "LENGTH")

        self.TEnvPipe = CCPar("TEnvPipe", parType="T") 
        self.TrefPipe = CCPar("TrefPipe", parType="T")
        self.QHXPipe = CCPar("QHXPipe", unitType = "ENERGY")# It comes from the matrix (and from calculation from questionnaire??)
        self.UPHProcm = CCPar("UPHProcm", unitType = "ENERGY")# It comes from the matrix (and from calculation from questionnaire??)
        self.USHm = CCPar("USHm", priority=2, unitType = "ENERGY")# It comes from the matrix (and from calculation from questionnaire??)

        # assign a variable to all intermediate/calculated values needed
        self.xForwIn = CCPar("xForwIn", parType="X")
        self.xForwOut = CCPar("xForwOut", parType="X")
        self.xRetIn = CCPar("xRetIn", parType="X")
        self.xRetOut = CCPar("xRetOut", parType="X")
        self.xRetRec = CCPar("xRetRec", parType="X")
        self.xFeedUp = CCPar("xFeedUp", parType="X")

        self.hRetIn = CCPar("hRetIn",parType = "S", unitType = "SPECIFICENTHALPY")
        self.hRetRec = CCPar("hRetRec",parType = "S", unitType = "SPECIFICENTHALPY")
        self.hRetOut = CCPar("hRetOut",parType = "S", unitType = "SPECIFICENTHALPY")
        self.hForwIn = CCPar("hForwIn",parType = "S", unitType = "SPECIFICENTHALPY")        
        self.hForwOut = CCPar("hForwOut",parType = "S", unitType = "SPECIFICENTHALPY")
        self.hFeedUp = CCPar("hFeedUp",parType = "S", unitType = "SPECIFICENTHALPY")

        self.DhRetRecFeedup = CCPar("DhRetRecFeedup",parType = "S", unitType = "SPECIFICENTHALPY")
        self.DhRetRecFeedupPercent = CCPar("DhRetRecFeedupPercent",parType = "S", unitType = "SPECIFICENTHALPY")
        self.DhRetRecFeedupPercInv = CCPar("DhRetRecFeedupPercInv",parType = "S", unitType = "SPECIFICENTHALPY")
             
        self.DhForw = CCPar("DhForw",parType = "S", unitType = "SPECIFICENTHALPY")
        self.DhRet = CCPar("DhRet",parType = "S", unitType = "SPECIFICENTHALPY")            #auxiliary variable added for DTin = DTout + DTfwd + DTret
        self.DhIn = CCPar("DhIn", unitType = "SPECIFICENTHALPY")
        self.DhOut = CCPar("DhOut", unitType = "SPECIFICENTHALPY")
        self.DhRec = CCPar("DhRec",parType = "S", unitType = "SPECIFICENTHALPY")
        self.DhWH = CCPar("DhWH",parType = "S", unitType = "SPECIFICENTHALPY")
        self.DhWHEff = CCPar("DhWHEff",parType = "S", unitType = "SPECIFICENTHALPY")

        self.TForwOut = CCPar("TForwOut", parType="T")

        self.TRetIn = CCPar("TRetIn", parType="T")
        self.TRetOut = CCPar("TRetOut", parType="T")

        self.PercentFeedUp = CCPar("PercentFeedUp", parType="X")
        self.FeedUpFlow = CCPar("FeedUpFlow", unitType = "MASSFLOW")

        self.RecFlow = CCPar("RecFlow", unitType = "MASSFLOW")

        self.DTForwLoss = CCPar("DTForwLoss", parType="DTmN")
        self.DTRetLoss = CCPar("DTRetLoss", parType="DTmN")
        self.QdotLossForw = CCPar("QdotLossForw", parType="S", unitType = "POWER")
        self.QdotLossRet = CCPar("QdotLossRet", parType="S", unitType = "POWER")

        self.QWHdotPipe = CCPar("QWHdotPipe", unitType = "POWER")
        self.QWHPipe = CCPar("QWHPipe", priority=2, unitType = "ENERGY")

        self.QdotLossPipe = CCPar("QdotLossPipe", parType="S", unitType = "POWER")
        self.QLossPipe = CCPar("QLossPipe", parType="S", unitType = "ENERGY")

        self.USHdotPipe = CCPar("USHdotPipe", unitType = "POWER")
        self.USHPipe = CCPar("USHPipe", unitType = "ENERGY")

        self.QHXPipe = CCPar("QHXPipe", priority=2, unitType = "ENERGY")

        self.UPHdotProcm = CCPar("UPHdotProcm", unitType = "POWER")

        self.DoPipe = CCPar("DoPipe", unitType = "LENGTH")
        
        self.HPerYearPipe = CCPar("HPerYearPipe", unitType = "TIME")
        self.HPerYearPipe.valMax = 8760
        self.HDEffAvg = CCPar("HDEffAvg")
        
        self.QWHRatio = CCPar("QWHRatio", parType = "X")
        self.QWHRatioOut = CCPar("QWHRatioOut")
        
        self.UAPipe = CCPar("UAPipe", unitType = "HEATTRANSFERCOEF")
        self.dUAPipe = CCPar("dUAPipe", unitType = "CONDUCTIVITY")

        self.ONE = CCOne()
        self.isHeatingPipe = None
        
        self.importData(m)
        if DEBUG in ["ALL", "BASIC"]:
            self.showAllPipe()

    def importData(self, m):  
        """Import data from SQL tables (from AlternativeProposalNo = -1)."""
        pipe = self.pipe
        self.PipeDuctNo = pipe.no + 1

        if self.checkInPlace:
            ANo = Status.ANo
        else:
            ANo = 0

        # assign empty CCPar to all questionnaire parameters
        self.DistribCircFlow = CCPar("DistribCircFlow", unitType = "MASSFLOW")
        self.ToutDistrib = CCPar("ToutDistrib", parType="T")
        self.TreturnDistrib = CCPar("TreturnDistrib", parType="T")
        self.PercentRecirc = CCPar("PercentRecirc", parType="X")
        self.TFeedUp = CCPar("TFeedUp", parType="T")
        self.TotLengthDistPipe = CCPar("TotLengthDistPipe") #one way
        self.DDistPipe = CCPar("DDistPipe", unitType = "LENGTH")
        self.DeltaDistPipe = CCPar("DeltaDistPipe", unitType = "LENGTH")

        self.TEnvPipe = CCPar("TEnvPipe", parType="T") 
        self.TrefPipe = CCPar("TrefPipe", parType="T")
        
        self.UPHProcm = CCPar("UPHProcm", priority=2, unitType = "ENERGY")# It comes from the matrix (and from calculation from questionnaire??)
        self.USHm = CCPar("USHm", priority=2, unitType = "ENERGY")# It comes from the matrix (and from calculation from questionnaire??)

        # reading data from table "qdistributionhc"
##        qdistributionhcTable = Status.DB.qdistributionhc.Questionnaire_id[Status.PId].AlternativeProposalNo[ANo].PipeDuctNo[self.PipeDuctNo]
##    
##        if len(qdistributionhcTable) > 0:
##            qdistributionhc = qdistributionhcTable[0]

        if pipe is not None:
            pipe.pipeType
            
            if pipe.fluidId is None and self.checkInPlace:
                if pipe.pipeType != "direct":
                    showError(_("Indication of heat distribution medium is mandatory"))
                fluid = Fluid(0)
            else:   
                fluid = Fluid(int(pipe.fluidId))
                
            self.FluidCp = fluid.cp
            self.Fluid_hL = fluid.hL
            self.FluidTCond = fluid.TCond

            self.DistribCircFlow.setValue(pipe.DistribCircFlow)
            self.ToutDistrib.setValue(pipe.ToutDistrib)
            self.PercentRecirc.setValue(pipe.PercentRecirc)

            self.pipeType = pipe.pipeType
            if self.pipeType=="direct":
                self.hasFeedUp = False
                self.HDEffAvg.setValue(1.0)
            if self.pipeType=="open":
                self.hasFeedUp = True
                self.PercentFeedUp.setValue(1.0)
                self.PercentRecirc.setValue(0.0)
            elif self.pipeType=="closed":
                fRecirc = self.PercentRecirc.val
                self.hasFeedUp = (fRecirc is not None and fRecirc < 1.0 - EPSILON)

            if not self.hasFeedUp:
                self.PercentRecirc.setValue(1.0,err=0, setTrack = False)
                self.PercentFeedUp.setValue(0.0, setTrack = False)
                self.FeedUpFlow.setValue(0.0, setTrack = False)

            #TODO: this is a provisional hack in order to avoid QWH > 0 due to random temperature error for pipes with Treturn = Tfeedup
            TFeedUp = pipe.Tfeedup
            TRet = pipe.TreturnDistrib
            if almostEqual(TFeedUp, TRet, EPSILON):
                self.TFeedUp.setValue(pipe.Tfeedup,err=0)
                self.TreturnDistrib.setValue(pipe.TreturnDistrib,err=0)
            else:
                self.TFeedUp.setValue(pipe.Tfeedup)
                self.TreturnDistrib.setValue(pipe.TreturnDistrib)

            self.TotLengthDistPipe.setValue(pipe.TotLengthDistPipe)
            self.UAPipe.setValue(pipe.UAPipe)
            self.DDistPipe.setValue(pipe.DDistPipe)
            self.DeltaDistPipe.setValue(pipe.DeltaDistPipe)

            self.TEnvPipe.setValue(18, err=0.0)  #�C
            self.TrefPipe.setValue(0, err=0.0)   #�C, NOT USED
        else:
            self.FluidCp = 0.00116
            self.Fluid_hL = 0.0
            self.FluidTCond = INFINITE
            logTrack("CheckPipe(importData): error reading data from pipe in PipeNo: %s" % self.PipeDuctNo)

        # calculate dUAPipe from DDistPipe and DoPipe only if it cannot be calculated by dUA = UA/LPipe
        if (self.UAPipe.val is None) or (self.TotLengthDistPipe is None):
            if self.DDistPipe.val is not None and self.DeltaDistPipe.val is not None:
                self.DoPipe.setValue (self.DDistPipe.val + (2 * self.DeltaDistPipe.val))
                try:
                    self.dUAPipe.setValue(0.000314 / (log(self.DoPipe.val) - log(self.DDistPipe.val)))
                except:
                    logTrack("checkPipes: exception when calculating dUA from diameters for pipe no. %s" % self.PipeDuctNo)
                    self.dUAPipe.setValue(0.0)
            else:   # for testing. IMPORTANT THAT THERE's A VALUE
                estimateTracker.setDataGroup("Pipe/Duct",self.PipeDuctNo)
                self.dUAPipe.setEstimate(0.0005,limits=(0.00025,0.00075))   # in kW/mK !

        self.isHeatingPipe = self.pipe.heating 
        if self.isHeatingPipe is None and self.checkInPlace: # allow for local checking of unconnected pipes
            if (self.hasFeedUp and self.ToutDistrib.val > self.TFeedUp.val) or (self.pipeType != "open" and self.ToutDistrib.val > self.TreturnDistrib.val):
                self.isHeatingPipe = True
            else:
                self.isHeatingPipe = False

        if self.checkInPlace and ANo > 0:
            self.constrainOperatingHours()
            
        if self.DistribCircFlow.val is None and (ANo > 0) and self.TotLengthDistPipe.val != 0 and pipe.pipeType != "direct":
            showWarning(_("When modifying existing or adding new pipes indication of nominal mass flow rate is mandatory") + "\n" +\
                        _("Check calculated distribution efficiency for plausibility")) 

            
    def exportData(self):   
        """stores corrected data in SQL (under AlternativeProposalNo = 0)"""

        if self.checkInPlace:
            ANo = max(Status.ANo, 0)
            if Status.ANo <0: return 
        else:
            ANo = 0
        pipe = self.pipe

        # avoid exporting nonesense in ANo > 0
        UAPipe = None if (self.UAPipe.sqerr > MAX_SQERR and ANo > 0) else self.UAPipe.val
        HDEffAvg = None if ((self.HDEffAvg.sqerr > MAX_SQERR or self.HDEffAvg.val > 1.1) and ANo > 0) else self.HDEffAvg.val

        if iszero(self.PercentFeedUp):
            self.TFeedUp.val = None
        
        # writing data back into pipe object"
        pipe.set(DistribCircFlow = dbio.check(self.DistribCircFlow.val),
                 ToutDistrib = dbio.check(self.ToutDistrib.val),
                 TreturnDistrib = dbio.check(self.TreturnDistrib.val),
                 PercentRecirc = dbio.check(self.PercentRecirc.val),
                 Tfeedup = dbio.check(self.TFeedUp.val),
                 TotLengthDistPipe = dbio.check(self.TotLengthDistPipe.val),
                 DDistPipe = dbio.check(self.DDistPipe.val),
                 DeltaDistPipe = dbio.check(self.DeltaDistPipe.val),
                 HDEffAvg = dbio.check(HDEffAvg),
                 QHXPipe = dbio.check(self.QHXPipe.val),
                 QWHPipe = dbio.check(self.QWHPipe.val),
                 UAPipe = dbio.check(UAPipe),
                 UPHProcm = dbio.check(self.UPHProcm.val),
                 USHm = dbio.check(self.USHm.val),
                 USHPipe = dbio.check(self.USHPipe.val),
                 QLossPipe = dbio.check(self.QLossPipe.val),
                 TForwOut = dbio.check(self.TForwOut.val),
                 TRetIn = dbio.check(self.TRetIn.val),
                 TRetOut = dbio.check(self.TRetOut.val),
                 FlowRateReturn = dbio.check(self.RecFlow.val),
                 HPerYearPipe = dbio.check(self.HPerYearPipe.val),
                 TEnvPipe = dbio.check(self.TEnvPipe .val),
                 TrefPipe =  dbio.check(self.TrefPipe.val))

        if self.exportToDB:
            self.pipe.saveToDB(ANo=0, warn=False)
            dbio.commit()
                
    def showAllPipe(self):
        print "====================="
        print "isHeatingPipe ? ",self.isHeatingPipe

        self.xForwIn.show()
        self.xForwOut.show()
        self.xRetIn.show()
        self.xRetOut.show()
        self.xRetRec.show()
        self.xFeedUp.show()
        
        self.hForwOut.show()
        self.hRetIn.show()
        self.hRetRec.show()
        self.hRetOut.show()
        self.hForwIn.show()
        self.hFeedUp.show()
        
        self.DhForw.show()
        self.DhRec.show()
        self.DhIn.show()
        self.DhOut.show()
        self.DhWH.show()
        self.DhWHEff.show()

        self.ToutDistrib.show()
        self.TreturnDistrib.show()
        self.TForwOut.show()
        self.TRetIn.show()
        self.TRetOut.show()
        self.TFeedUp.show()

        self.PercentRecirc.show()
        self.PercentFeedUp.show()
        self.DistribCircFlow.show()
        self.RecFlow.show()
        self.FeedUpFlow.show()

        self.DTForwLoss.show()
        self.DTRetLoss.show()
        self.QdotLossForw.show()
        self.QdotLossRet.show()
        self.QWHdotPipe.show()
        self.QdotLossPipe.show()

        self.USHdotPipe.show()
        self.UPHdotProcm.show()

        self.DoPipe.show()
        self.UAPipe.show()
        self.dUAPipe.show()
        self.TotLengthDistPipe.show()
        self.DDistPipe.show()
        self.DeltaDistPipe.show()

        self.HPerYearPipe.show()
        self.HDEffAvg.show()
        self.QWHRatio.show()
        self.QWHRatioOut.show()
        
        self.QHXPipe.show()
        self.QWHPipe.show()
        self.USHm.show()
        self.USHPipe.show()
        self.QLossPipe.show()
        self.QWHPipe.show()
        self.UPHProcm.show()

        print "====================="

    def screen(self):  
        """Screen all variables in the block."""

        self.HDEffAvg.screen()
        self.HPerYearPipe.screen()
        self.QHXPipe.screen()
        self.QWHPipe.screen()
        self.QLossPipe.screen
        self.UPHProcm.screen()
        self.USHm.screen()
        self.UAPipe.screen()

        if self.pipeType != "direct":            

##            if iszero(self.PercentRecirc):
##                self.TreturnDistrib.priority = 99
       
            self.DistribCircFlow.screen()
            self.ToutDistrib.screen()
            self.TreturnDistrib.screen()
            self.PercentRecirc.screen()

            if iszero(self.PercentFeedUp):
                self.TFeedUp.priority = 99
            else:
                self.TFeedUp.screen()

            self.TotLengthDistPipe.screen()
            self.DDistPipe.screen()
            self.DeltaDistPipe.screen()

    
    def check(self):
        """Carry out the check of the block"""
        if DEBUG in ["ALL", "MAIN"]:
            print "-------------------------------------------------"
            print " PipeDucts checking"
            print "-------------------------------------------------"

        for n in range(1):
            cycle.initCheckBalance()

            if DEBUG in ["ALL", "MAIN"]:
                print "-------------------------------------------------"
                print "Ciclo %s" % n
                print "-------------------------------------------------"
        
            # Step 1: Call all calculation routines in a given sequence
            if DEBUG in ["ALL", "MAIN"]:
                print "-------------------------------------------------"
                print "Step 1: calculating from left to right (CALC)"
                print "-------------------------------------------------"
            
            # a) convert temperatures to enthalpies
            self.hForwIn.push(calcH("hForwIn",
                                  self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                                  self.ToutDistrib, self.xForwIn)) #ToutDistrib
            self.hForwOut.push(calcH("hForwOut", self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                                   self.TForwOut, self.xForwOut))#TForwOut
            self.hRetIn.push(calcH("hRetIn", self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                                 self.TRetIn, self.xRetIn))
            self.hRetRec.push(calcH("hRetRec", self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                                  self.TreturnDistrib, self.xRetIn))#TreturnDistrib
            self.hRetOut.push(calcH("hRetOut", self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                                  self.TRetOut, self.xRetOut))
            if self.hasFeedUp:
                self.hFeedUp.push(calcH("hFeedUp", self.FluidCp,
                                      self.FluidCp,
                                      self.Fluid_hL,
                                      self.FluidTCond,
                                      self.TFeedUp, self.xFeedUp))

            # b) calculate percentage of recirculation and flow rates
            

            if self.hasFeedUp:
                self.FeedUpFlow.push(calcProd ("FeedUpFlow", self.DistribCircFlow, self.PercentFeedUp)) # PercentRecirc
                self.DistribCircFlow.push(calcSum("DistribCircFlow", self.FeedUpFlow, self.RecFlow))  # DistribCircFlow, RecFlow
                self.PercentFeedUp.push(calcDiff("PercentFeedUp", self.ONE, self.PercentRecirc))#hFeedUp,hRetRec
                self.RecFlow.push(calcProd ("RecFlow", self.DistribCircFlow, self.PercentRecirc)) # PercentRecirc
            else:
                self.DistribCircFlow.push(calcK("DistribCircFlow", 1.0, self.RecFlow))

            # c) calculate enthalpy differences (inverted for cooling)
            
            # The following block is based on the equation:
            #
            #   hRetOut = %FeedUp*hFeedUp + %Recirc*hRetRec
            #           = hFeedUp + %Recirc*(hRetRec - hFeedUp) =: hFeedUp + DhRetRecFeedUp%
            #           = hRetRec + %FeedUp*(hFeedUp - hRedRec) = hRetRec - %FeedUp*(hRedRec - hFeedUp) =: hFeedUp - DhRetRecFeedUp%INV
            if self.hasFeedUp:
                if self.isHeatingPipe:
                    self.DhRetRecFeedup.push(calcDiff("DhRetRecFeedup", self.hRetRec, self.hFeedUp,parType = "S")) # hFeedUp,hRetRec
                else:
                    self.DhRetRecFeedup.push(calcDiff("DhRetRecFeedup", self.hFeedUp, self.hRetRec,parType = "S")) # hFeedUp,hRetRec

                self.DhRetRecFeedupPercent.push(calcProd("DhRetRecFeedupPercent", self.DhRetRecFeedup, self.PercentRecirc,parType = "S"))#PercentRecirc
                self.DhRetRecFeedupPercInv.push(calcProd("DhRetRecFeedupPercInv", self.DhRetRecFeedup, self.PercentFeedUp,parType = "S"))#PercentRecirc

                if self.isHeatingPipe:
                    self.hRetOut.push(calcSum("hRetOut", self.DhRetRecFeedupPercent, self.hFeedUp,parType="S")) 
                    self.hRetOut.push(calcDiff("hRetOut", self.hRetRec, self.DhRetRecFeedupPercInv,parType="S"))
                else:
                    self.hRetOut.push(calcDiff("hRetOut", self.hFeedUp, self.DhRetRecFeedupPercent,parType="S")) 
                    self.hRetOut.push(calcSum("hRetOut", self.hRetRec, self.DhRetRecFeedupPercInv,parType="S"))
            else:

                setIdentity(self.hRetOut, self.hRetRec)
                

            if self.isHeatingPipe:
                self.DhForw.push(calcDiff("DhForw", self.hForwIn, self.hForwOut, parType = "S")) # hForwIn, hForwOut
                self.DhRet.push(calcDiff("DhRet", self.hRetIn, self.hRetOut, parType = "S"))
                self.DhIn.push(calcDiff("DhIn", self.hForwIn, self.hRetOut)) # hForwIn
                self.DhOut.push(calcDiff("DhOut", self.hForwOut, self.hRetIn)) # hForwOut,hRetIn
                self.DhRec.push(calcDiff("DhRec", self.hRetIn, self.hRetRec, parType = "S")) # hRetIn, hRetRec
                if self.hasFeedUp:
                    self.DhWH.push(calcDiff("DhWH", self.hRetIn, self.hFeedUp, parType = "S")) # hFeedUp3,hRetIn3
            else:
                self.DhForw.push(calcDiff("DhForw", self.hForwOut, self.hForwIn, parType = "S")) # hForwIn, hForwOut
                self.DhRet.push(calcDiff("DhRet", self.hRetOut, self.hRetIn, parType = "S"))
                self.DhIn.push(calcDiff("DhIn", self.hRetOut, self.hForwIn)) # hForwIn
                self.DhOut.push(calcDiff("DhOut", self.hRetIn, self.hForwOut)) # hForwOut,hRetIn
                self.DhRec.push(calcDiff("DhRec", self.hRetRec, self.hRetIn, parType = "S")) # hRetIn, hRetRec
                if self.hasFeedUp:
                    self.DhWH.push(calcDiff("DhWH", self.hFeedUp, self.hRetIn, parType = "S")) # hFeedUp3,hRetIn3
            
            self.DhIn.push(calcSum3("DhIn", self.DhForw, self.DhOut, self.DhRet))

            if self.hasFeedUp:
                self.DhWHEff.push(calcProd("DhWHEff", self.DhWH, self.PercentFeedUp))                                  


            # d) thermal losses of pipe

            self.UAPipe.push(calcProdC("UAPipe", 2.0, self.dUAPipe, self.TotLengthDistPipe))    #TotLength is one way !!

            self.QdotLossForw.push(calcProdC("QdotLossForw", 0.5, self.UAPipe, self.DTForwLoss, parType = "S", unitType = "POWER")) # DTForwLoss
            if self.isHeatingPipe:
                self.DTForwLoss.push(calcDiff("DTForwLoss", self.ToutDistrib, self.TEnvPipe, parType = "DTmN")) # ToutDistrib
            else:
                self.DTForwLoss.push(calcDiff("DTForwLoss", self.TEnvPipe, self.ToutDistrib, parType = "DTmN")) # ToutDistrib

            if self.pipeType == "open":
                self.QdotLossRet.setValue(0.0)
            else:
                self.QdotLossRet.push(calcProdC("QdotLossRet", 0.5, self.UAPipe, self.DTRetLoss,parType = "S", unitType = "POWER")) # DTForwLoss
                if self.isHeatingPipe:
                    self.DTRetLoss.push(calcDiff("DTRetLoss", self.TreturnDistrib, self.TEnvPipe, parType = "DTmN")) # TreturnDistrib
                else:
                    self.DTRetLoss.push(calcDiff("DTRetLoss", self.TEnvPipe, self.TreturnDistrib, parType = "DTmN")) # TreturnDistrib

            self.QdotLossPipe.push(calcSum("QdotLossPipe", self.QdotLossForw, self.QdotLossRet,parType = "S"))

                
            # e) power = mass flow rate * dh

            if self.hasFeedUp:
                self.QWHdotPipe.push(calcProd("QWHdotPipe", self.DhWH, self.FeedUpFlow))
            else:
                self.QWHdotPipe.setValue(0.0)
                
            self.USHdotPipe.push(calcProd("USHdotPipe", self.DhIn, self.DistribCircFlow)) # DistribCircFlow
            self.UPHdotProcm.push(calcProd("UPHdotProcm", self.DhOut, self.DistribCircFlow)) # DistribCircFlow3
            self.QdotLossForw.push(calcProd("QdotLossForw", self.DhForw, self.DistribCircFlow,parType = "S")) # DistribCircFlow
            if self.pipeType != "open":
                self.QdotLossRet.push(calcProd("QdotLossRet", self.DhRec, self.RecFlow,parType = "S")) # calculated from DHRec, RecFlow

            # f) energy = power * time

            self.USHPipe.push(calcProd("USHPipe", self.USHdotPipe, self.HPerYearPipe)) # HPerYearPipe
            self.UPHProcm.push(calcProd("UPHProcm", self.UPHdotProcm, self.HPerYearPipe)) # HPerYearPipe, UPHdotProcm
            self.QLossPipe.push(calcProd("QLossPipe", self.QdotLossPipe, self.HPerYearPipe,parType = "S")) # HPerYearPipe, UPHdotProcm
            self.QWHPipe.push(calcProd("QWHPipe", self.QWHdotPipe, self.HPerYearPipe)) # HPerYearPipe, UPHdotProcm

            # g) energy balance: (input) USHPipe = (output) UPHProcm + QLossPipe + QWHPipe

            self.USHdotPipe.push(calcSum3("USHdotPipe", self.QdotLossPipe, self.UPHdotProcm, self.QWHdotPipe))#UPHdotProcm
            self.USHPipe.push(calcSum3("USHPipe", self.QLossPipe, self.UPHProcm, self.QWHPipe))#UPHdotProcm

            self.UPHProcm.push(calcProd("USHProcm", self.USHPipe, self.HDEffAvg))
                
            self.UPHdotProcm.push(calcProd("UPHdotProcm", self.USHdotPipe, self.HDEffAvg))
            self.DhOut.push(calcProd("dhOut", self.DhIn, self.HDEffAvg))

            self.QWHPipe.push(calcProd("QWHPipe", self.USHPipe, self.QWHRatio))
            self.QWHdotPipe.push(calcProd("QWHPipe", self.USHdotPipe, self.QWHRatio))
            self.DhWHEff.push(calcProd("dhOut", self.DhIn, self.QWHRatio))

            self.QWHPipe.push(calcProd("QWHPipe", self.UPHProcm, self.QWHRatioOut))
            self.QWHdotPipe.push(calcProd("QWHPipe", self.UPHdotProcm, self.QWHRatioOut))
            self.DhWHEff.push(calcProd("dhOut", self.DhOut, self.QWHRatioOut))
            
            self.USHPipe.push(calcSum("USHPipe", self.USHm, self.QHXPipe))
         
            # Step 2: Cross check the variables
            if DEBUG in ["ALL", "MAIN"]:
                self.showAllPipe()
                print "-------------------------------------------------"
                print "Step 2: cross checking"
                print "-------------------------------------------------"

            self.ccheckAll()                
            
            # Step 3: Adjust the variables (inverse of calculation routines)
            if DEBUG in ["ALL", "MAIN"]:
                self.showAllPipe()
                print "-------------------------------------------------"
                print "Step 3: calculating from right to left (ADJUST)"
                print "-------------------------------------------------"
  
            # g) energy balance: (input) USHPipe = (output) UPHProcm + QLossPipe + QWHPipe

            adjustSum3_stack(self.USHdotPipe, self.QdotLossPipe, self.UPHdotProcm, self.QWHdotPipe)
            adjustSum3_stack(self.USHPipe, self.QLossPipe, self.UPHProcm, self.QWHPipe)#UPHdotProcm

            adjustProd_stack(self.UPHProcm, self.USHPipe, self.HDEffAvg)

            adjustProd_stack(self.UPHdotProcm, self.USHdotPipe, self.HDEffAvg)
            adjustProd_stack(self.DhOut, self.DhIn, self.HDEffAvg)

            adjustProd_stack(self.QWHPipe, self.USHPipe, self.QWHRatio)
            adjustProd_stack(self.QWHdotPipe, self.USHdotPipe, self.QWHRatio)

            adjustProd_stack(self.DhWHEff, self.DhIn, self.QWHRatio)

            adjustProd_stack(self.QWHPipe, self.UPHProcm, self.QWHRatioOut)
            adjustProd_stack(self.QWHdotPipe, self.UPHdotProcm, self.QWHRatioOut)
            adjustProd_stack(self.DhWHEff, self.DhOut, self.QWHRatioOut)

            adjustSum_stack(self.USHPipe, self.USHm, self.QHXPipe)

            # f) energy = power * time

            adjustProd_stack(self.UPHProcm, self.UPHdotProcm, self.HPerYearPipe) 
            adjustProd_stack(self.USHPipe, self.USHdotPipe, self.HPerYearPipe)
            adjustProd_stack(self.QLossPipe, self.QdotLossPipe, self.HPerYearPipe) 
            adjustProd_stack(self.QWHPipe, self.QWHdotPipe, self.HPerYearPipe) 

            # e) power = mass flow rate * dh

            adjustProd_stack(self.UPHdotProcm, self.DhOut, self.DistribCircFlow)
            adjustProd_stack(self.USHdotPipe, self.DhIn, self.DistribCircFlow)

            if self.hasFeedUp:
                adjustProd_stack(self.QWHdotPipe, self.DhWH, self.FeedUpFlow)
            else:
                self.QWHdotPipe.setValue(0.0, setTrack = False)

            adjustProd_stack(self.QdotLossForw, self.DhForw, self.DistribCircFlow)
            if self.pipeType == "open":
                self.DhRec.setValue(0.0)
                self.RecFlow.setValue(0.0)
            else:
                adjustProd_stack(self.QdotLossRet, self.DhRec, self.RecFlow)

            # d) thermal losses of pipe

            adjustProdC_stack(self.UAPipe, 2.0, self.dUAPipe, self.TotLengthDistPipe)

            if self.pipeType != "open":
                adjustProdC_stack(self.QdotLossRet, 0.5, self.UAPipe, self.DTRetLoss)
                if self.isHeatingPipe:
                    adjustDiff_stack(self.DTRetLoss, self.TreturnDistrib, self.TEnvPipe, GTZero=True)
                else:
                    adjustDiff_stack(self.DTRetLoss, self.TEnvPipe, self.TreturnDistrib, GTZero=True)
                            
            adjustProdC_stack(self.QdotLossForw, 0.5, self.UAPipe, self.DTForwLoss)
            if self.isHeatingPipe:
                adjustDiff_stack(self.DTForwLoss, self.ToutDistrib, self.TEnvPipe, GTZero=True)
            else:
                adjustDiff_stack(self.DTForwLoss, self.TEnvPipe, self.ToutDistrib, GTZero=True)
                               
            adjustSum_stack(self.QdotLossPipe, self.QdotLossForw, self.QdotLossRet)

            # c) calculate enthalpy differences (inverted for cooling)
                
            if self.hasFeedUp:
                adjustProd_stack(self.DhWHEff, self.DhWH, self.PercentFeedUp)
                                  
            adjustSum3_stack(self.DhIn, self.DhForw, self.DhOut, self.DhRet)
                               
            if self.isHeatingPipe:
                if self.hasFeedUp:
                    adjustDiff_stack(self.DhWH, self.hRetIn, self.hFeedUp)
                                  
                adjustDiff_stack(self.DhRec, self.hRetIn, self.hRetRec)
                adjustDiff_stack(self.DhOut, self.hForwOut, self.hRetIn)
                adjustDiff_stack(self.DhIn, self.hForwIn, self.hRetOut)
                adjustDiff_stack(self.DhRet, self.hRetIn, self.hRetOut)
                adjustDiff_stack(self.DhForw, self.hForwIn, self.hForwOut)

                if self.hasFeedUp:
                    adjustDiff_stack(self.hRetOut, self.hRetRec, self.DhRetRecFeedupPercInv) # hFeedUp
                    adjustSum_stack(self.hRetOut, self.DhRetRecFeedupPercent, self.hFeedUp)
                else:
                    setIdentity(self.hRetOut,self.hRetRec)

            else:
                if self.hasFeedUp:
                    adjustDiff_stack(self.DhWH, self.hFeedUp, self.hRetIn)
                adjustDiff_stack(self.DhRec, self.hRetRec, self.hRetIn)
                adjustDiff_stack(self.DhOut, self.hRetIn, self.hForwOut)
                adjustDiff_stack(self.DhIn, self.hRetOut, self.hForwIn)
                adjustDiff_stack(self.DhRet, self.hRetOut, self.hRetIn)
                adjustDiff_stack(self.DhForw, self.hForwOut, self.hForwIn)
                
                if self.hasFeedUp:
                    adjustSum_stack(self.hRetOut, self.hRetRec, self.DhRetRecFeedupPercInv) # hFeedUp
                    adjustDiff_stack(self.hRetOut, self.hFeedUp, self.DhRetRecFeedupPercent)
                else:
                    ccheck1(self.hRetOut,self.hRetRec)

            if self.hasFeedUp:
                adjustProd_stack(self.DhRetRecFeedupPercInv, self.DhRetRecFeedup, self.PercentFeedUp) # PercentRecirc
                adjustProd_stack(self.DhRetRecFeedupPercent, self.DhRetRecFeedup, self.PercentRecirc)

                if self.isHeatingPipe:
                    adjustDiff_stack(self.DhRetRecFeedup, self.hRetRec, self.hFeedUp)
                else:
                    adjustDiff_stack(self.DhRetRecFeedup, self.hFeedUp, self.hRetRec)

            # b) calculate percentage of recirculation and flow rates
            if self.hasFeedUp:
                adjustSum_stack(self.DistribCircFlow, self.FeedUpFlow, self.RecFlow)
                adjustProd_stack(self.FeedUpFlow, self.DistribCircFlow, self.PercentFeedUp) # PercentRecirc
                adjustProd_stack(self.RecFlow, self.DistribCircFlow, self.PercentRecirc)

                adjustDiff_stack(self.PercentFeedUp, self.ONE, self.PercentRecirc) # hFeedUp,hRetRec
            else:
                setIdentity(self.RecFlow,self.DistribCircFlow)                  

            # a) convert temperatures to enthalpies

            if self.hasFeedUp:
                adjustH_stack(self.hFeedUp, self.FluidCp,
                                      self.FluidCp,
                                      self.Fluid_hL,
                                      self.FluidTCond,
                        self.TFeedUp, self.xFeedUp)
            
            adjustH_stack(self.hRetOut, self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                    self.TRetOut, self.xRetOut)
            
            adjustH_stack(self.hRetRec, self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                    self.TreturnDistrib, self.xRetIn)
            
            adjustH_stack(self.hRetIn, self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                    self.TRetIn, self.xRetIn)
            
            adjustH_stack(self.hForwOut, self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                    self.TForwOut, self.xForwOut)
            
            adjustH_stack(self.hForwIn, self.FluidCp,
                                  self.FluidCp,
                                  self.Fluid_hL,
                                  self.FluidTCond,
                    self.ToutDistrib, self.xForwIn) 
            
            # Step 4: Second cross check of the variables
            if DEBUG in ["ALL", "MAIN"]:
                self.showAllPipe()
                print "-------------------------------------------------"
                print "Step 4: cross checking"
                print "-------------------------------------------------"

            self.ccheckAll()                

            if DEBUG in ["ALL", "MAIN"]:
                self.showAllPipe()

          
        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAllPipe()
            print "-------------------------------------------------"
            print "Cycle balance: mean %s max %s " % (cycle.getMeanBalance(), cycle.getMaxBalance())
            print "-------------------------------------------------"


##        self.USHPipe.simplifyTrack()
##        self.UPHProcm.simplifyTrack()
##        self.QWHPipe.simplifyTrack()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()
    
    def ccheckAll(self):     
        """Check all given and calculated values in this block for consistency."""
        self.UAPipe.check()
        self.dUAPipe.check()
        self.TotLengthDistPipe.check()
        
        self.HPerYearPipe.check()
 
        self.TFeedUp.check()
        self.TForwOut.check()
        self.TRetOut.check()
        self.TRetIn.check()
        self.ToutDistrib.check()
        self.TreturnDistrib.check()
        self.TEnvPipe.check()

        self.DTForwLoss.check()
        self.DTRetLoss.check()

        self.hForwIn.check()
        self.hRetIn.check()
        self.hRetRec.check()
        self.hForwOut.check()
        self.hRetOut.check()
        if self.hasFeedUp:
            self.hFeedUp.check()


        self.xForwIn.check()
        self.xRetIn.check()
        self.xRetRec.check()
        self.xForwOut.check()
        self.xRetOut.check()

        if self.hasFeedUp:
            self.xFeedUp.check()
                              
            self.DhRetRecFeedup.check()
            self.DhRetRecFeedupPercent.check()
            self.DhRetRecFeedupPercInv.check()

        self.DhForw.check()
        self.DhRet.check()
        self.DhIn.check()
        self.DhOut.check()
        self.DhRec.check()
        self.DhWH.check()
        self.DhWHEff.check()

        self.QWHRatio.check()
        self.QWHRatioOut.check()

        self.PercentRecirc.check()
        self.PercentFeedUp.check()

        self.DistribCircFlow.check()
        self.RecFlow.check()
        self.FeedUpFlow.check()

        self.QdotLossForw.check()
        self.QdotLossRet.check()
        self.QdotLossPipe.check()
 
        self.QWHdotPipe.check()
        self.USHdotPipe.check()
        self.UPHdotProcm.check()

        self.USHm.check()
        self.QWHPipe.check()
        self.QLossPipe.check()
        self.UPHProcm.check()
        self.QHXPipe.check()
        self.USHPipe.check()

        self.HDEffAvg.check()

        self.ONE.check()    # necessary for clearing stack and check for conflicts

#------------------------------------------------------------------------------
    def estimate(self):  
#------------------------------------------------------------------------------
#   estimates some of the data that are not sufficiently precise
#   should be a subset of the data that are within screen
#   (not necessarily ALL data have to be estimated)
#------------------------------------------------------------------------------

#..............................................................................
# recirculation rate: suppose closed circuit

        self.PercentRecirc.setEstimate(1.0,limits=(1.0,1.0))
        
        if self.PercentRecirc.val == 1.0:
            self.hasFeedUp = False

#..............................................................................
# pipe length [TO DO: estimate could be finer adjusted to size of the company]

        self.TotLengthDistPipe.setEstimate(100.0,limits=(30.0,300.0))

##        self.dUAPipe.setEstimate(0.0005,limits=(0.0004,0.0006))   # in kW/mK !
            
#..............................................................................
# forward, return and feed-up temperature of the piping
#   [TO DO: estimates should take into account:
#       - process supply temperatures of processes connected to the pipe
#       - fluid type
#       - ... ]

        if self.hasFeedUp:
            self.TFeedUp.setEstimate(10.0,newSqErr = 0.1)
        
        if self.FluidTCond >= 999.0:
            if self.isHeatingPipe:
                self.ToutDistrib.setEstimate(100.0,limits = (50.0,200.0))
            else:
                self.ToutDistrib.setEstimate(0.0,limits = (-10.0,10.0))
                
        else:   # if fluid with phase change (e.g. steam) supppose temperature slightly above TCond
            if self.isHeatingPipe:
                self.ToutDistrib.setEstimate(self.FluidTCond + 5.0,limits = (self.FluidTCond + 0.1,self.FluidTCond + 50.0))
            else:
                self.ToutDistrib.setEstimate(self.FluidTCond - 5.0,limits = (self.FluidTCond -50.0,self.FluidTCond - 0.1))
                

        if self.PercentRecirc.val < 0.5 or self.PercentRecirc.val is None:  # very open pipe. None check should not be necessary, as in this case it would have been set to 1.0 above
            
            self.TreturnDistrib.setEstimate(self.TFeedUp.val,limits=(self.TFeedUp.valMin,self.TFeedUp.valMax))
            
        else:
            if self.isHeatingPipe:
                TRet = min(self.FluidTCond - 5.0,self.ToutDistrib.val - 20.0)
                TRetMin = min(self.FluidTCond - 10.0, self.ToutDistrib.valMin - 50.0)
                TRetMax = min(self.FluidTCond - 0.1, self.ToutDistrib.valMax - 5.0)
            else:
                TRet = self.ToutDistrib.val + 20.0
                TRetMin = self.ToutDistrib.valMin + 5.0
                TRetMax = self.ToutDistrib.valMax + 50.0
                if self.FluidTCond < 999.0:
                    TRet = max(TRet,self.FluidTCond+5.0)
                    TRetMin = max(TRetMin,self.FluidTCond+0.1)
                    TRetMin = max(TRetMax,self.FluidTCond+10)

            self.TreturnDistrib.setEstimate(TRet,limits = (TRetMin,TRetMax))


    def constrainOperatingHours(self):

        """
        constrains pipe operating hours from schedules of connected processes
        1)Check operating hours of pipes
        Suppositions:
        (a) minimum hour = minimum hours of connected process with maximum
                       duration
        (b) maximum hours = hours of industry operation / sum of process
        """
        m = self.index
        
        hopdaymax = 0
        ndaymax = 0
        hopmin = self.HPerYearPipe.valMin

        UPHProcMax = 0

        if self.checkInPlace:
            ccProc = [CheckProc(k, checkInPlace=True) for k in range(Status.int.systemModel.NK)]
            ccEq =   [CheckEq(j, checkInPlace=True) for j in range(Status.int.systemModel.NJ)]
        else:
            ccProc = Status.mod.moduleCC.ccProc
            ccEq   = Status.mod.moduleCC.ccEq
        
        for k, cc in enumerate(ccProc):
            if ((m, k) in Status.int.systemModel.UPHLink) or ((m, k) in Status.int.systemModel.UPCReqLink):
                hopmin = max(cc.HPerYearOp.valMin, hopmin)
                ndaymax = max(cc.NDaysProc.valMax, ndaymax)
                hopdaymax += cc.HPerDayOp.valMax

                if (m, k) in Status.int.systemModel.UPHLink:
                    UPHProcMax += cc.UPH.valMax
                else:
                    UPHProcMax += cc.UPHwReqOfProcess.valMax
                
        for j, cc in enumerate(ccEq):
            if (m, j) in Status.int.systemModel.FEThLink:
                hopmin = max(cc.HPerYearEq.valMin, hopmin)
                ndaymax = max(cc.NDaysEq.valMax, ndaymax)
                hopdaymax += cc.HPerDayEq.valMax

                UPHProcMax += cc.FETHeat_j.valMax
                
        hopdaymax = min(24.0, hopdaymax)
        ndaymax = min(365., ndaymax)
        hopmax = ndaymax * hopdaymax

        self.HPerYearPipe.valMin = max(self.HPerYearPipe.valMin, hopmin)
        self.HPerYearPipe.valMax = min(self.HPerYearPipe.valMax, hopmax)
        self.HPerYearPipe.val = 0.5 * (self.HPerYearPipe.valMin + self.HPerYearPipe.valMax)
        self.HPerYearPipe.constrain()

        self.UPHProcm.valMax = min(self.UPHProcm.valMax, UPHProcMax)
        #TODO: minimum cannot be given, as for this limits for QHXProc would have to be calculated !
        self.UPHProcm.constrain()
            
if __name__ == "__main__":
    #-----  Imports for standalone testing in any directory
    import os
    import sys
    import einstein.GUI.pSQL as pSQL, MySQLdb

    sys.path.append(os.path.abspath('../../..'))
    # direct connecting to SQL database w/o GUI. for testing only
    stat = Status("testCheckProc")
    Status.SQL = MySQLdb.connect(user="root", db="einstein")
    Status.DB = pSQL.pSQL(Status.SQL, "einstein")
    Status.PId = 41
    Status.ANo = -1
    
    m = 0
    conflict.setDataGroup("pipe", m + 1)

    ccPipe = CheckPipe(m)       # creates an instance of class CCheck
#    ccPipe.UPHProcm.setValue(5000000)
    ccPipe.check()
#    ccPipe.exportData(1)

    screen.setDataGroup("pipe", m + 1)
    ccPipe.screen()
    screen.show()
