# -*- coding: utf-8 -*-
#==============================================================================                 
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    CCheck (Consistency Check)
#            
#------------------------------------------------------------------------------
#            
#    Short description:
#    
#    Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================                 
from globals import logger

from ccheckFunctions import *
from ccPar import CCPar, CCRow
from checkSchedules import CCSchedule
from modules.fluids import Fluid
from GUI.status import *
from GUI.GUITools import check
from constants import *
from auxiliary.auxiliary import toGettextAndUnicode as _U
import database.dbio as dbio
import numpy as np
from modules.schedules.schedules import DEFAULTCHARGETIME

EPSILON = 1.e-3     # required accuracy for function "isequal"
INFINITE = 1.e99    # numerical value assigned to "infinite"
global DEBUG

class CheckProc():
    """
    Carries out a consistency check for a process object
    """

    def __init__(self, k, proc=None, checkInPlace=False):
        """Setup the check with the data for process nr. k
        
        :param checkInPlace: If False take the data from ANo -1 and
        write it to ANo 0, otherwise work on the current alternative proposal. 
        
        """
        if proc is None: # keep this here temporarily for compatibility with other ccheck modules which still read/write from/to DB
            proc = Status.int.systemModel.processObject[k]
            
        k = proc.no
        self.proc = proc
        
        self.checkInPlace = checkInPlace
        self.exportToDB = not checkInPlace
        
        # assign a variable to all intermediate/calculated values needed
        self.ProcNo = k + 1
        self.index = k
        self.dataGroup = "Process"
        self.name = "Process %s"%self.ProcNo    # initialising with default name. real name is read in later from database
        
        self.PTOutFlowRec = [CCPar("PTOutFlowRec", parType="T")]  # specific to individual out-flowing stream
        self.PTFinal = [CCPar("PTFinal", parType="T")]     # specific to individual out-flowing stream
        self.PTFinalRequired = [CCPar("PTFinalRequired", parType="T")] # specific to individual out-flowing stream
        self.HOutFlowRec = [CCPar("HOutFlowRec", parType="S")]            # specific to individual out-flowing stream
        self.HFinal = [CCPar("HFinal", parType="S", unitType = "SPECIFICENTHALPY")]                    # specific to individual out-flowing stream
        self.HFinalRequired = [CCPar("HFinalRequired", parType="S", unitType = "SPECIFICENTHALPY")]    # specific to individual out-flowing stream
        self.dHOutFlow = [CCPar("dHOutFlow", unitType = "SPECIFICENTHALPY")]              # specific to individual out-flowing stream
        self.dHOutFlowRes = [CCPar("dHOutFlowRes", unitType = "SPECIFICENTHALPY")]        # specific to individual out-flowing stream
        self.dHFinal = [CCPar("dHFinal", unitType = "SPECIFICENTHALPY")]                  # specific to individual out-flowing stream
        self.dHOutFlowRequired = [CCPar("dHOutFlowRequired", unitType = "SPECIFICENTHALPY")] # specific to individual out-flowing stream
        self.DPSTmin = CCPar("DPSTmin", parType="DT")       # one per processes
        self.DPSTminMCD = CCPar("DPSTminMCD", parType="DT") # one per processes

        self.XOutFlow = [CCPar("XOutFlow", parType="X")]    # specific to individual out-flowing stream
        self.XOutFlowRec = [CCPar("XOutFlowRec", parType="X")] # specific to individual out-flowing stream
        self.XFinal = [CCPar("XFinal", parType="X")]      # specific to individual out-flowing stream
        self.XFinalRequired = [CCPar("XFinalRequired", parType="X")] # specific to individual out-flowing stream

        self.HPerYearProc = CCPar("HPerYearProc", priority=2, unitType = "TIME")
        self.HPerYearOp = CCPar("HPerYearOp", unitType = "TIME")
        self.HPerYearOp.valMax = 8760


        self.VInFlowCycleEff = [CCPar("VInFlowCycleEff", unitType = "VOLUME")] # specific to individual in-flowing stream      
        self.VOutFlowCycleEff = [CCPar("VOutFlowCycleEff", unitType = "VOLUME")] # specific to individual in-flowing stream
        
        self.FluidCp_c = CCPar("") # specific to individual in-flowing stream 
        self.FluidDensity_c = CCPar("") # specific to individual in-flowing stream 
        self.FluidRhoCp_c = CCPar("") # specific to individual in-flowing stream 

        self.DTLoss = CCPar("DTLoss", parType="DTmN")
        self.dotQLossProc = CCPar("dotQLossProc", parType="S", unitType = "ENERGY")
        self.UPHm = CCPar("UPHm", unitType = "ENERGY")
        self.DTUPHcGross = [CCPar("DTUPHcGross", parType="DT")] # specific to individual in-flowing stream
        self.dUPHcGross = [CCPar("dUPHcGross", parType = "ENERGY")]             # specific to individual in-flowing stream
        
        self.DTQHXIn = [CCPar("DTQHXIn", parType="DT")]     # specific to individual in-flowing stream
        self.DTQHXOut = [CCPar("DTQHXOut", parType="DT")]    # specific to individual out-flowing stream
    
        self.DTCrossHXLT = CCPar("DTCrossHXLT", parType="DT")   # only for single in-flowing and single out-flowing stream
        self.DTCrossHXLT.valMin = 2.0   #engineering lower limit
        self.DTCrossHXHT = CCPar("DTCrossHXHT", parType="DT")
        self.DTCrossHXHT.valMin = 2.0   #engineering lower limit
        
        self.DTOutFlow = [CCPar("DTOutFlow", parType="DT")]   # specific to individual out-flowing stream
        self.dQHXProcIntIn = []        # specific to individual in-flowing stream
        self.dQHXProcIntOut = []       # specific to individual out-flowing stream
        self.dQHXProcInt = CCPar("dQHXProcInt", unitType = "ENERGY") # the sum of the stream QHX
        
        self.NCyclesPerYear = CCPar("NCyclesPerYear")
        
        self.DTUPHcNet = [CCPar("DTUPHcNet", parType="DT")]   # specific to individual in-flowing stream
        self.dUPHc = [CCPar("dUPHc", unitType = "ENERGY")]                  # specific to individual in-flowing stream
        self.UPHc = [CCPar("UPHc", unitType = "ENERGY")]                 # specific to individual in-flowing stream 
        self.UPHcOfProcess = CCPar("UPHc", unitType = "ENERGY") # individual in-flowing streams are summed
        
        self.DTUPHs = CCPar("DTUPHs", parType="DT")
        self.dUPHs = CCPar("dUPHs", unitType = "ENERGY")
        self.UPHs = CCPar("UPHs", unitType = "ENERGY")
        
        self.UPH = CCPar("UPH", unitType = "ENERGY")

        self.UPHwOfProcess = CCPar("UPHwOfProcess", unitType = "ENERGY") # sum of the out-flowing streams. It is the waste heat of the outflow. In general it is QWHProc but if we have a vessel it does not
        self.dUPHw = [CCPar("dUPHw", unitType = "ENERGY")]                # specific to individual out-flowing stream
        self.UPHw = [CCPar("UPHw", unitType = "ENERGY")]                 # specific to individual out-flowing stream
        self.UPHwReq_dot = [CCPar("UPHwReq_dot", unitType = "ENERGY")]   # specific to individual out-flowing stream
        self.UPHwReq = [CCPar("UPHwReq", unitType = "ENERGY")]                 # specific to individual out-flowing stream
        self.UPHwReqOfProcess = CCPar("UPHwReqOfProcess", unitType = "ENERGY") # sum of the out-flowing streams
        self.UPHwRes_dot = [CCPar("UPHwRes_dot", unitType = "ENERGY")]   # specific to individual out-flowing stream
        self.UPHwRes = [CCPar("UPHwRes", unitType = "ENERGY")]   # the remaining waste heat that need not be cooled down, specific to individual out-flowing stream
        self.UPHwResOfProcess = CCPar("UPHwResOfProcess", unitType = "ENERGY") # sum of the out-flowing streams
        
        self.UPHProc = CCPar("UPHProc", unitType = "ENERGY") # The heating demand of the process that is delivered by the pipe 
                                        # (excluding heat from heat recovery). Vice versa for cooling processes.
        self.UPCProc = CCPar("UPCProc", unitType = "ENERGY") # The cooling demand for mandatory cooling down that is delivered by the other pipe
                                        # (excluding heat from heat recovery). Senseless for cooling processes.
        self.UAProc = CCPar("UAProc", unitType = "HEATTRANSFERCOEF")
        self.dotQEvapProc = CCPar("dotQEvapProc", unitType = "ENERGY")
        
        self.UPHcGrossProd = [CCPar("UPHcGrossProd", unitType = "ENERGY")]           # specific to individual in-flowing stream
        self.UPHcGross = CCPar("UPHcGross", unitType = "ENERGY")                 # individual in-flowing streams are summed
        
        self.QHXProcInt = [CCPar("QHXProcInt", unitType = "ENERGY")]              # specific to individual in-flowing stream
        self.QHXProc = CCPar("QHXProc", unitType = "ENERGY") # It comes from the heat exchange matrix
        self.QWHProc = CCPar("QWHProc", unitType = "ENERGY") # It comes from the heat exchange matrix. In theory it is the sum UPHwOfProcess and UPHmass (not existing yet)
        self.QWHProcReq = CCPar("QWHProcReq", priority = 2, unitType = "ENERGY") # The part of the mandatory cooling down that is covered by the heat exchangers.
        self.QWHProcRes = CCPar("QWHProcRes", unitType = "ENERGY") # .
        self.QWHProcUsed = CCPar("QWHProcUsed", unitType = "ENERGY") # 
        self.QWHAmb = CCPar("QWHAmb", unitType = "ENERGY") # The part of the mandatory cooling down that is covered by the heat exchangers.
        
        self.mInFlowYear = [CCPar("mInFlowYear", unitType = "MASS")]             # specific to individual in-flowing stream
        self.mOutFlowYear = [CCPar("mOutFlowYear", unitType = "MASS")]            # specific to individual out-flowing stream
        self.mInFlowYearOfProcess = CCPar("mInFlowYearOfProcess", unitType = "MASS")    # should match the sum of the in-flowing streams
        self.mOutFlowYearOfProcess = CCPar("mOutFlowYearOfProcess", unitType = "MASS") # should match the sum of the out-flowing streams

        self.isHeatingProcess = None # Do we check a heating process? 

        self.importData()

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAll()

    def importStreamsData(self, dataName, streams, readIn=True):
        """Import the content of the attribute dataName of each element of
        streams into a list as the attribute dataName of self.
        
        :param dataName: the name of the attribute to import
        :param streams: the list of streams
        """
        # assume single template variable to determine parType
        if len(streams) > 0:
            data = getattr(self, dataName).pop()
        setattr(self, dataName, [])
        for i in range(len(streams)):
            newitem = CCPar('%s [%s]' % (dataName, streams[i].Name),
                            parType=data.parType, unitType=data.unitType)
            if readIn: newitem.setValue(getattr(streams[i], dataName)) 
            getattr(self, dataName).append(newitem)
            
    def exportStreamsData(self, dataName, streams):
        for i in range(len(streams)):
##            setattr(streams[i], dataName, check(getattr(self, dataName)[i].val))  ### check is already done in saveToDB. leads to error for PTFinalRequired -> None has a meaning !!!
            setattr(streams[i], dataName, getattr(self, dataName)[i].val)
            
    def importData(self): 
        """Import process data from process object Proc. 
        
        The data is taken from AlternativeProposalNo = -1 or the current one
        depending on whether to check in place or not. 
        """
        proc = self.proc
        self.ProcNo = proc.no + 1
        if self.checkInPlace is True:
            ANo = Status.ANo
        else:
            ANo = -1
        
        # assign empty CCPar to all questionnaire parameters
        self.PTInFlow = [CCPar("PTInFlow", parType="T")]    # specific to individual in-flowing stream
        self.PT = CCPar("PT", parType = "T", priority=2)
        self.PTOutFlow = [CCPar("PTOutFlow", parType="T")]   # specific to individual out-flowing stream
        self.PTInFlowRec = [CCPar("PTInFlowRec", parType="T")]  # specific to individual in-flowing stream
        self.PTStartUp = CCPar("PTStartUp", parType="T")
        self.DPSTmin = CCPar("DPSTmin", parType="DT")       # one per processes
        self.DPSTminMCD = CCPar("DPSTminMCD", parType="DT") # one per processes

        self.DtInFlow = CCPar("DtInFlow")
        self.DtOutFlow = CCPar("DtOutFlow")

        self.VInFlowCycle = [CCPar("VInFlowCycle", unitType = "VOLUME")]             # specific to individual in-flowing stream
        self.mInFlowNom = [CCPar("mInFlowNom", unitType = "MASSFLOW")]               # specific to individual in-flowing stream
        self.VOutFlowCycle = [CCPar("VOutFlowCycle", unitType = "VOLUME")]            # specific to individual out-flowing stream
        self.mOutFlowNom = [CCPar("mOutFlowNom", unitType = "MASSFLOW")]              # specific to individual out-flowing stream
        self.HOutFlow = [CCPar("HOutFlow", parType="S")]                 # specific to individual out-flowing stream
        self.VolProcMed = CCPar("VolProcMed", unitType = "VOLUME")
        self.ThermalMass = CCPar("ThermalMass", unitType = "THERMALMASS")
        self.NDaysProc = CCPar("NDaysProc")
        self.NDaysProc.valMax = 365
        self.HPerDayProc = CCPar("HPerDayProc", unitType = "TIME")
        self.HPerDayOp = CCPar("HPerDayProc", unitType = "TIME")
        self.HPerDayOp.valMax = 24 
        self.HPerYearInFlow = CCPar("HPerYearInFlow", unitType = "TIME")             # independent of streams, assigned from single process schedule
        self.HPerYearOutFlow = CCPar("HPerYearOutFlow", unitType = "TIME")            # independent of streams, assigned from single process schedule
        self.HPerYearInFlowNom = CCPar("HPerYearInFlowNom", unitType = "TIME")             # independent of streams, assigned from single process schedule
        self.HPerYearOutFlowNom = CCPar("HPerYearOutFlowNom", unitType = "TIME")            # independent of streams, assigned from single process schedule
        self.PartLoad = CCPar("PartLoad")
        self.HBatch = CCPar("HBatch", unitType = "TIME")
        self.NBatch = CCPar("NBatch")
        self.TEnvProc = CCPar("TEnvProc", parType="T")       # Now cannot be entered by questionnaire
        self.dotUPHm = CCPar("dotUPHm", unitType = "POWER")
        self.UPH = CCPar("UPH", unitType = "ENERGY")                        # Useful process heat asked into the questionnaire is UPH not UPHproc

        # read in data from table "qprocessdata"
        self.name = proc.name
        self.isHeatingProcess = proc.heating
        
        self.autoFit = proc.autoFit # autoFit of process schedule to monitoring data
        self.continuous = (proc.ProcType==0)

        # get all in-flowing and out-flowing streams
        self.nInflowingStreams = len(proc.inStreams)
        self.nOutflowingStreams = len(proc.outStreams)

        self.FluidCp_c = [None] * self.nInflowingStreams
        self.FluidDensity_c = [None] * self.nInflowingStreams
        self.FluidRhoCp_c = [None] * self.nInflowingStreams
        self.VInFlowCycleEff = [None] * self.nInflowingStreams
        self.mInFlowYear = [None] * self.nInflowingStreams
        self.UPHcGrossProd = [None] * self.nInflowingStreams
        self.dUPHcGross = [None] * self.nInflowingStreams
        self.DTUPHcGross = [None] * self.nInflowingStreams 
        self.DTQHXIn = [None] * self.nInflowingStreams
        self.dUPHc = [None] * self.nInflowingStreams
        self.dotUPHc = [None] * self.nInflowingStreams
        self.DTUPHcNet = [None] * self.nInflowingStreams
        self.internalHR = [None] * self.nInflowingStreams
        
        fluids = Status.prj.getFluidDict()
        for i in range(self.nInflowingStreams):
            stream = proc.inStreams[i]
            fluidId = stream.dbfluid_id
            if fluidId is None:
                logger.logWarning(_("No medium specified for inflow %s. Water (default) assumed")%stream.Name)
                fluidId = 0
                stream.dbfluid_id = fluidId
                
            stream_fluid = Fluid(stream.dbfluid_id)   #IMPORT from the FluidDB
            self.FluidCp_c[i] = stream_fluid.cp
            self.FluidDensity_c[i] = stream_fluid.rho
            self.FluidRhoCp_c[i] = self.FluidCp_c[i] * self.FluidDensity_c[i]
            self.VInFlowCycleEff[i] = CCPar("VInFlowCycleEff [%s]" % (stream.Name,))
            self.mInFlowYear[i] = CCPar("mInFlowYear [%s]" % (stream.Name,))
            self.UPHcGrossProd[i] = CCPar("UPHcGrossProd [%s]" % (stream.Name,))
            self.dUPHcGross[i] = CCPar("dUPHcGross [%s]" % (stream.Name,)) 
            self.DTUPHcGross[i] = CCPar("DTUPHcGross [%s]" % (stream.Name,),
                                        parType="DT") 
            self.DTQHXIn[i] = CCPar("DTQHXIn [%s]" % (stream.Name,), parType="DT")
            self.dUPHc[i] = CCPar("dUPHc [%s]" % (stream.Name,))
            self.dotUPHc[i] = CCPar("dUPHcGross [%s]" % (stream.Name,)) 
            self.DTUPHcNet[i] = CCPar("DTUPHcNet [%s]" % (stream.Name,), parType="DT")

        self.importStreamsData('PTInFlow', proc.inStreams)
        self.PT.setValue(proc.PT, err=0.0)   # if specified, take as fixed
        if self.isHeatingProcess:
            self.DPSTmin.setValue(proc.DPSTmin, default=10)
        else:
            self.DPSTmin.setValue(proc.DPSTmin, default=5)
        self.DPSTminMCD.setValue(proc.DPSTminMCD, default=5)
        
        self.importStreamsData('PTOutFlow', proc.outStreams)
        # if no temperature or enthalpy for Outflow is specified, 
        # start with PT as initial estimate
        # but leave error range from 0 to infinite !!!!
        self.importStreamsData('PTOutFlowRec', proc.outStreams)

        self.importStreamsData('HOutFlow', proc.outStreams)
        self.importStreamsData('XOutFlow', proc.outStreams)
        self.importStreamsData('PTInFlowRec', proc.inStreams)
        self.importStreamsData('mInFlowNom', proc.inStreams)
        self.importStreamsData('dotUPHc', proc.inStreams)
        self.importStreamsData('VInFlowCycle', proc.inStreams)
        self.importStreamsData('mOutFlowNom', proc.outStreams)
        self.importStreamsData('VOutFlowCycle', proc.outStreams)
        self.importStreamsData('UPHc', proc.inStreams, readIn=False)            # not imported, as this parameter is not user input
        self.importStreamsData('UPHw', proc.outStreams, readIn=False)           # not imported, as this parameter is not user input
        self.importStreamsData('UPHwReq', proc.outStreams, readIn=False)        # not imported, as this parameter is not user input
        self.importStreamsData('PTFinal', proc.outStreams)
        self.importStreamsData('PTFinalRequired', proc.outStreams)


        self.VOutFlowCycleEff = [None] * self.nOutflowingStreams
        self.dHOutFlow = [None] * self.nOutflowingStreams
        self.dHOutFlowRec = [None] * self.nOutflowingStreams
        self.dHOutFlowRes = [None] * self.nOutflowingStreams
        self.HOutFlowRec = [None] * self.nOutflowingStreams
        self.HFinal = [None] * self.nOutflowingStreams
        self.dHFinal = [None] * self.nOutflowingStreams
        self.dHOutFlowRequired = [None] * self.nOutflowingStreams
        self.dUPHw = [None] * self.nOutflowingStreams
        self.dotUPHw = [None] * self.nOutflowingStreams
        self.UPHwReq_dot = [None] * self.nOutflowingStreams
        self.UPHwRes_dot = [None] * self.nOutflowingStreams
        self.mOutFlowYear = [None] * self.nOutflowingStreams
        self.XOutFlowRec = [None] * self.nOutflowingStreams
        self.XFinal = [None] * self.nOutflowingStreams
        self.XFinalRequired = [None] * self.nOutflowingStreams
        self.UPHwRes = [None] * self.nOutflowingStreams
        self.HFinalRequired = [None] * self.nOutflowingStreams
        for i in range(self.nOutflowingStreams):
            stream = proc.outStreams[i]
            fluidId = stream.dbfluid_id
            if fluidId is None:
                logger.logWarning(_("No medium specified for outflow %s. Water (default) assumed")%stream.Name)
                fluidId = 0
                stream.dbfluid_id = fluidId

            if stream.PTOutFlow is None and stream.HOutFlow is None and stream.PTOutFlowRec is None:
                self.PTOutFlow[i].setValue(self.PT.val)
                self.PTOutFlowRec[i].setValue(self.PT.val)
                
            self.VOutFlowCycleEff[i] = CCPar("VOutFlowCycleEff [%s]" 
                                             % (stream.Name,))
            self.dHOutFlow[i] = CCPar("dHOutFlow [%s]" 
                                      % (stream.Name,))
            self.HOutFlowRec[i] = CCPar("HOutFlowRec [%s]"
                                         % (stream.Name,), parType="S")
            self.dHOutFlowRec[i] = CCPar("dHOutFlowRec [%s]" 
                                         % (stream.Name,))
            self.dHOutFlowRes[i] = CCPar("dHOutFlowRes [%s]" 
                                         % (stream.Name,))
            self.HFinal[i] = CCPar("HFinal [%s]" % (stream.Name,), parType="S")
            self.dHFinal[i] = CCPar("dHFinal [%s]" % (stream.Name,))

            if self.PTFinal[i].val is not None and self.PTFinalRequired[i].val is not None:
                if self.PTFinal[i].val == self.PTFinalRequired[i].val:
                    self.dHOutFlowRes[i].setValue(0.0)   # set to exactly 0 if input values are equal
                    
            if self.PTFinal[i].val is not None and self.PTOutFlow[i].val is not None:
                if isequal(self.PTFinal[i].val, self.PTOutFlow[i].val) and self.PTFinal[i].userDefined and self.PTOutFlow[i].userDefined:
                    self.dHFinal[i].setValue(0.0)   # set to exactly 0 if input values are equal
                    
            self.dHOutFlowRequired[i] = CCPar("dHOutFlowRequired [%s]" 
                                            % (stream.Name,))
            self.dUPHw[i] = CCPar("dUPHw [%s]" 
                                     % (stream.Name,)) 
            self.dotUPHw[i] = CCPar("dotUPHw [%s]" 
                                     % (stream.Name,)) 
            self.UPHwReq_dot[i] = CCPar("UPHwReq_dot [%s]" 
                                     % (stream.Name,)) 
            self.UPHwRes_dot[i] = CCPar("UPHwRes_dot [%s]" 
                                     % (stream.Name,)) 
            self.UPHwRes[i] = CCPar("UPHwRes [%s]" % (stream.Name, ), unitType="ENERGY")
            self.mOutFlowYear[i] = CCPar("mOutFlowYear [%s]"
                                         % (stream.Name,)) 
            self.XOutFlowRec[i] = CCPar("XOutFlowRec [%s]"
                                        % (stream.Name,)) 
            self.XFinal[i] = CCPar("XFinal [%s]" % (stream.Name,),
                                   parType="X")   
            self.XFinalRequired[i] = CCPar("XFinalRequired [%s]" 
                                           % (stream.Name,), parType="X")
            self.HFinalRequired[i] = CCPar("HFinalRequired [%s]" 
                                           % (stream.Name,), parType="S")

        self.VolProcMed.setValue(proc.VolProcMed)
        self.ThermalMass.setValue(proc.ThermalMass)
        if not (iszero(self.VolProcMed) or iszero(self.ThermalMass)):
            self.PTStartUp.setValue(proc.PTStartUp)
        if isequal(self.PTStartUp.val, self.PT.val) or iszero(self.VolProcMed) or iszero(self.ThermalMass):
            self.startUp = False
        else:
            self.startUp = True

        # import global Fluid data
        fluidId = proc.ProcMedDBFluid_id
        if proc.ProcMedDBFluid_id is None:
            fluidId = 0
            if self.startUp: 
                logger.logWarning(_("No process medium specified. Water (default) assumed"))
                stream.dbfluid_id = fluidId

        global_fluid = Fluid(int(fluidId))
        self.FluidCp = global_fluid.cp
        self.FluidDensity = global_fluid.rho
        try:
            self.FluidRhoCp = global_fluid.cp * global_fluid.rho
        except TypeError: # at least one of the operands is None
            self.FluidRhoCp = None
            
        self.NDaysProc.setValue(proc.NDays, err=0.0) # number -> exact value
        self.HPerDayProc.setValue(proc.HPerDay)
        self.HPerYearInFlow.setValue(proc.HPerYearInFlow)
        self.HPerYearOutFlow.setValue(proc.HPerYearOutFlow)

        self.DtInFlow.setValue(proc.InFlowDuration)
        self.DtOutFlow.setValue(proc.OutFlowDuration)

##        if proc.PartLoad is None:   # should be always set from GUI, but for being sure!
##            self.PartLoad.setValue(1.0, err=0.0, setTrack = False)
##        else:
        
        if proc.scheduleType==0: # for simple schedules PardLoad = 1 exactly.
            self.PartLoad.setValue(1.0, err=0.0)
        else:
            self.PartLoad.setValue(proc.PartLoad)

        # if 24.0 hours are specified, suppose that this value is exact.
        if self.HPerDayProc.val > 23.99:
            self.HPerDayProc.setValue(24.0, err=0.0)
            
        if proc.NCycles is None:
            if not self.autoFit: self.NBatch.setValue(1.0, err=0.0)   # if no number is specified, suppose 1!
        else:
            self.NBatch.setValue(proc.NCycles, err=0.0) #number -> exact value

        self.HBatch.setValue(proc.HCycle)

        self.HeatRecOK = [None] * self.nOutflowingStreams
        for i in range(self.nOutflowingStreams):
            if proc.outStreams[i].HeatRecOk is None:
                logWarning(_("Possibility of heat recovery for process no. %s (%s) stream %d (%s) is not specified.\nYES is assumed.") 
                           %  (self.ProcNo, proc.name, i, proc.outStreams[i].Name))
                self.HeatRecOK[i] = True
            else:
                self.HeatRecOK[i] = proc.outStreams[i].HeatRecOk
            if self.HeatRecOK[i] and proc.outStreams[i].PTFinal is None: #read in PT final only if heat recovery is possible!
                    if self.isHeatingProcess:
                        logMessage(_("No limit specified for cooling of outgoing stream %s (%s). 0 °C is assumed")
                                   %  (i, proc.outStreams[i].Name))
                        self.PTFinal[i].setValue(0.0)
                    else:
                        self.PTFinal[i].setValue(MAXTEMP)
        if proc.TEnvProc is None:
            self.TEnvProc.setValue(18.0)
        else:
            self.TEnvProc.setValue(proc.TEnvProc)
            
        self.dotUPHm.setValue(proc.dotUPHm)
        self.UAProc.setValue(0.01, err=INFINITE, setTrack = False)
        self.UPH.setValue(proc.UPH, err=0.0) # if specified, take exact value 
        self.dQHXProcIntIn = [None] * self.nInflowingStreams
        self.FluidCp_w = [None] * self.nOutflowingStreams
        self.FluidDensity_w = [None] * self.nOutflowingStreams
        self.FluidRhoCp_w = [None] * self.nOutflowingStreams
        self.Fluid_hL_w = [None] * self.nOutflowingStreams
        self.FluidTCond_w = [None] * self.nOutflowingStreams
        self.dQHXProcIntOut = [None] * self.nOutflowingStreams
        for i in range(self.nInflowingStreams):
##            self.internalHR[i] = not (isequal(self.PTInFlow[i].val, self.PTInFlowRec[i].val) 
##                                      or (self.PTInFlowRec[i].val is None))
            self.internalHR[i] = proc.inStreams[i].HeatRecExist
            self.dQHXProcIntIn[i] = CCPar("dQHXProcIntIn [%s]" % 
                                            proc.inStreams[i].Name)  
            stream = proc.inStreams[i]
            stream_fluid = Fluid(stream.dbfluid_id)  #IMPORT from the FluidDB
            condensation, Tc = checkForCondensation(stream_fluid, self.PTInFlow[i].val, self.PT.val)
            if condensation:
                if self.isHeatingProcess:   # a hack in order to exclude condensation (-> slightly above or below)
                    Tc += 0.1
                    Tc = min(Tc, self.PT.val)

                else:
                    Tc -= 0.1
                    Tc = max(Tc, self.PT.val)

                message = _("Phase change in incoming streams is not supported in current version")
                message += _("\nInlet temperature in stream %s corrected to T = %s"%(stream.Name, Tc))
                logWarning(message)

                    
                self.PTInFlow[i].valMin = Tc
                self.PTInFlow[i].valMax = Tc
                self.PTInFlow[i].setValue(Tc, err=0.0)
                if self.PTInFlowRec[i].val is not None:
                    self.PTInFlowRec[i].valMin = Tc
                    self.PTInFlowRec[i].valMax = Tc
                    self.PTInFlowRec[i].setValue(Tc, err=0.0)
                            
        for i in xrange(self.nOutflowingStreams):
            stream_fluid_w = Fluid(proc.outStreams[i].dbfluid_id)  #IMPORT from the FluidDB
            self.FluidDensity_w[i] = stream_fluid_w.rho

            self.dQHXProcIntOut[i] = CCPar("dQHXProcIntOut [%s]" % 
                                            proc.outStreams[i].Name) 
            if (self.HeatRecOK[i] or any(self.internalHR) or 
                any([stream.PTFinalRequired is not None for stream in proc.outStreams])):   # fluid pars only needed in this case
                self.FluidCp_w[i] = stream_fluid_w.cp
                self.FluidRhoCp_w[i] = self.FluidCp_w[i] * self.FluidDensity_w[i]
                self.Fluid_hL_w[i] = stream_fluid_w.hL
                self.FluidTCond_w[i] = stream_fluid_w.TCond

            if not any(self.internalHR): # could be None, which means No
                if self.PTOutFlowRec[i].val is None:
                    self.PTOutFlowRec[i].update(self.PTOutFlow[i])
                elif self.PTOutFlow[i].val is None:
                    self.PTOutFlow[i].update(self.PTOutFlowRec[i])
                self.dHOutFlow[i].setValue(0.0)
                    

        # objects for schedule checking
        if hasattr(Status.schedules, "procOpSchedules"):
            k = proc.no
            if k >= len(Status.schedules.procOpSchedules):
                print "CheckProc.importData: error in length of procOpSchedules: k = %s len(pOS) = %s"%(k, len(Status.schedules.procOpSchedules))
                self.ccsUPH = CCSchedule(self.UPH)
                self.ccsUPHc = [CCSchedule(q) for q in self.UPHc]
                self.ccsUPHw = [CCSchedule(q) for q in self.UPHw]

                self.ccsmInFlowYear = [CCSchedule(flow) for flow in self.mInFlowYear]
                self.ccsmOutFlowYear = [CCSchedule(flow) for flow in self.mOutFlowYear]
            else:                        
                self.ccsUPH = CCSchedule(self.UPH, Status.schedules.procOpSchedules[k].fav)
                self.ccsUPHc = [CCSchedule(q, Status.schedules.procInFlowSchedules[k].fav) for q in self.UPHc]
                self.ccsUPHw = [CCSchedule(q, Status.schedules.procOutFlowSchedules[k].fav) for q in self.UPHw]

                self.ccsmInFlowYear = [CCSchedule(flow, Status.schedules.procInFlowSchedules[k].fav) for flow in self.mInFlowYear]
                self.ccsmOutFlowYear = [CCSchedule(flow, Status.schedules.procOutFlowSchedules[k].fav) for flow in self.mOutFlowYear]
        
            if self.autoFit:
                self.ccsUPH.schedule = None
                for ccs in self.ccsUPHc + self.ccsUPHw: ccs.schedule = None
        
            self.ccsQHXProc = CCSchedule(self.QHXProc, None)

##                self.checkMassFlow = True
##                if self.checkMassFlow:
                
    def exportData(self):  
        """Store corrected data in SQL.
        
        Under AlternativeProposalNo = 0 or the current one depending on whether 
        to check in place or not. 
        
        """
        if self.checkInPlace is True:
            ANo = max(Status.ANo,0)
            if Status.ANo < 0: return
        else:
            ANo = 0
##            proc = self.proc.copy()

        proc = self.proc

        proc.set(PT = dbio.check(self.PT.val),
                         DPSTmin = dbio.check(self.DPSTmin.val),
                         DPSTminMCD = dbio.check(self.DPSTminMCD.val),
                         PTStartUp = dbio.check(self.PTStartUp.val),
                         VolProcMed = dbio.check(self.VolProcMed.val),
                         ThermalMass = dbio.check(self.ThermalMass.val),
                         NDays = dbio.check(self.NDaysProc.val),
                         HPerDay = dbio.check(self.HPerDayProc.val),
                         NCycles = dbio.check(self.NBatch.val),
                         HCycle = dbio.check(self.HBatch.val),
                         TEnvProc = dbio.check(self.TEnvProc.val),
                         dotUPHm = dbio.check(self.dotUPHm.val),
                         UPHProc = dbio.check(self.UPHProc.val),
                         HPerYear = dbio.check(self.HPerYearProc.val),
                         HPerYearInFlow = dbio.check(self.HPerYearInFlow.val),
                         HPerYearOutFlow = dbio.check(self.HPerYearOutFlow.val),
                         InFlowDuration = dbio.check(self.DtInFlow.val),
                         OutFlowDuration = dbio.check(self.DtOutFlow.val),
                         PartLoad = dbio.check(self.PartLoad.val),
                         UAProc = dbio.check(self.UAProc.val),
                         QEvapProc = dbio.check(self.dotQEvapProc.val),
                         UPHcGross = dbio.check(self.UPHcGross.val),
                         QHXProcInt = dbio.check(calcRowSum("QHXProcInt", self.QHXProcInt).val),
                         UPHm = dbio.check(self.UPHm.val),
                         UPHs = dbio.check(self.UPHs.val),
                         UPH = dbio.check(self.UPH.val),
                         UPHc = dbio.check(self.UPHcOfProcess.val),
                         UPHw = dbio.check(self.UPHwOfProcess.val),
                         UPHwReq = dbio.check(self.UPHwReqOfProcess.val),
                         QWHProc = dbio.check(self.QWHProcUsed.val) if self.isHeatingProcess else dbio.check(self.QHXProc.val),
                         QHXProc = dbio.check(self.QHXProc.val) if self.isHeatingProcess else dbio.check(self.QWHProcUsed.val))

        self.exportStreamsData("UPHc", proc.inStreams)
        self.exportStreamsData("dotUPHc", proc.inStreams)
        self.exportStreamsData("PTInFlow", proc.inStreams)
        self.exportStreamsData("PTInFlowRec", proc.inStreams)
        self.exportStreamsData("VInFlowCycle", proc.inStreams)
        self.exportStreamsData("mInFlowNom", proc.inStreams)

        self.exportStreamsData("UPHw", proc.outStreams)
        self.exportStreamsData("dotUPHw", proc.outStreams)
        self.exportStreamsData("PTOutFlow", proc.outStreams)
        self.exportStreamsData("PTOutFlowRec", proc.outStreams)
        self.exportStreamsData("HOutFlow", proc.outStreams)
        self.exportStreamsData("PTFinal", proc.outStreams)
        self.exportStreamsData("VOutFlowCycle", proc.outStreams)
        self.exportStreamsData("mOutFlowNom", proc.outStreams)
        self.exportStreamsData("UPHwReq", proc.outStreams)

        for i in range(self.nOutflowingStreams):
            proc.outStreams[i].HeatRecOk = self.HeatRecOK[i]
            # Keep undefined PTFinalRequired values, as they indicate
            # their nonexistence in the industry.
            if proc.outStreams[i].PTFinalRequired is None:
                self.PTFinalRequired[i].val = None
        self.exportStreamsData("PTFinalRequired", proc.outStreams)

        # export to DB in old-fashioned CheckProc usage
        if self.exportToDB:
            self.proc.saveToDB(ANo=0, warn=False)
            dbio.commit()

    def showAll(self):
        print "====================="
        for item in (
        self.UPH,
        self.UPHm,
        self.UPHs,
        self.UPHcOfProcess,
        self.dotUPHc,
        self.UPHc,
        self.UPHcGross,
        self.UPHwOfProcess,
        self.UPHw,
        self.UPHwReqOfProcess,
        self.UPHwResOfProcess,
        self.dUPHs,
        self.dUPHc,
        self.dUPHcGross, # For convention UPH and UPHcOfProcess are net!
        self.dUPHw,
        self.dotUPHw,
        self.UPHwReq_dot,
        self.UPHwReq,
        self.UPHwRes_dot,
        self.UPHwRes,
        self.QHXProcInt,
        self.DTUPHcGross,
        self.dotUPHm,
        self.VolProcMed,
        self.HPerYearProc,
        self.HPerYearInFlow,
        self.HPerYearOutFlow,
        self.HPerDayProc,
        self.NDaysProc,
        self.VInFlowCycleEff,
        self.VInFlowCycle,
        self.VOutFlowCycleEff,
        self.VOutFlowCycle,
        self.PartLoad,
        self.mInFlowYear,
        self.mInFlowNom,
        self.mOutFlowYear,
        self.mOutFlowNom,
        self.PTInFlow,
        self.PT,
        self.PTStartUp,
        self.PTFinal,
        self.HFinal,
        self.PTInFlowRec,
        self.PTOutFlow,
        self.HOutFlow,
        self.PTOutFlowRec,
        self.HOutFlowRec,
        self.dHOutFlow,
        self.dHFinal,
        self.NBatch,
        self.HBatch,
        self.DtInFlow,
        self.DtOutFlow,
        self.UAProc,
        self.dotQEvapProc,
        self.TEnvProc,
        self.DTLoss,
        self.dotQLossProc,
        self.DTUPHcGross,
        self.DTQHXIn,
        self.DTQHXOut,
        self.NCyclesPerYear,
        self.DTUPHcNet,
        self.DTUPHs,
        self.UPCProc,
        self.QWHProc,
        self.QWHProcReq,
        self.QWHProcRes,
        self.QWHProcUsed,
        self.QWHAmb,
        self.QHXProc,
        self.UPHProc):
            try:
                item.show()
            except AttributeError:
                for subitem in item:
                    subitem.show()
        print "HeatRecOK: ", self.HeatRecOK
        print "====================="


    def screen(self):  
        """screen all variables in the block"""
        # Change of priority for parameters not needed
        for i in range(self.nInflowingStreams):
            if iszero(self.VInFlowCycle[i]):
                self.PTInFlow[i].priority = 99
                self.PTInFlowRec[i].priority = 99
        for i in range(self.nOutflowingStreams):
            if iszero(self.VOutFlowCycle[i]) or (not self.HeatRecOK[i]):
                self.PTOutFlow[i].priority = 99
                self.PTOutFlowRec[i].priority = 99
        if iszero(self.VolProcMed):
            self.PTStartUp.priority = 99
        if iszero(self.dotUPHm):
            self.TEnvProc.priority = 99
            self.UAProc.priority = 99

        if self.UPHProc.valMax < 0.2*self.UPH.valMin:
            self.UPHProc.priority = 3

        if self.isHeatingProcess:
            if self.QHXProc.valMax < 0.2*self.UPH.valMin:
                self.QHXProc.priority = 3
        else:
            if self.QWHProcUsed.valMax < 0.2*self.UPH.valMin:
                self.QWHProcUsed.priority = 3

        self.UPH.screen()
        self.UPHcOfProcess.screen()
        self.UPHm.screen()
        self.UPHs.screen()
        self.UPHwOfProcess.screen()
        self.UPHwReqOfProcess.screen()
        self.QWHProc.screen()
        self.QWHProcReq.screen()
        self.UPHProc.screen()
        self.QHXProc.screen()
        self.QWHProcUsed.screen()
        self.UPHcGross.screen()
        for i in xrange(self.nInflowingStreams):
            self.QHXProcInt[i].screen()
        self.PT.screen()
        for i in range(self.nInflowingStreams):
            self.PTInFlow[i].screen()
            self.PTInFlowRec[i].screen()
        for i in range(self.nOutflowingStreams):
            self.PTOutFlow[i].screen()
            self.PTOutFlowRec[i].screen() 
            self.UPHwReq[i].screen()
        self.PTStartUp.screen()
        for i in range(self.nInflowingStreams):
            self.VInFlowCycle[i].screen()
            self.mInFlowNom[i].screen()
        for i in range(self.nOutflowingStreams):
            self.VOutFlowCycle[i].screen()
            self.mOutFlowNom[i].screen()
        self.VolProcMed.screen()
        self.HPerYearProc.screen()
        self.HPerYearInFlow.screen()
        self.HPerYearOutFlow.screen()
        self.NDaysProc.screen()
        self.HPerDayProc.screen()
        self.NBatch.screen()
        self.TEnvProc.screen()
        self.dotUPHm.screen()
        self.UAProc.screen()
        self.dotQEvapProc.screen()
        
    def definePriority(self, mainProcess):
        """Change the priority in function of the importance of the process."""
        if mainProcess == True:
            
            self.UPH.priority = 1
            self.UPHProc.priority = 1
            self.UPHs.priority = 1
            self.UPHcOfProcess.priority = 1
            self.UPHm.priority = 1
            self.UPHwOfProcess.priority = 1 
            self.PT.priority = 1
            self.dotUPHm.priority = 1

            for i in range(self.nInflowingStreams):
                if self.PTInFlow[i].priority < 99: self.PTInFlow[i].priority = 1
                if self.PTInFlowRec[i].priority < 99: self.PTInFlowRec[i].priority = 1
            if self.PTStartUp.priority < 99: self.PTStartUp.priority = 1

            # here some verifications are redundant but they do not affect the results
            for i in range(self.nOutflowingStreams):
                if self.PTOutFlow[i].priority < 99: self.PTOutFlow[i].priority = 2
                if self.PTOutFlowRec[i].priority < 99: self.PTOutFlowRec[i].priority = 2
                if self.VOutFlowCycle[i].priority < 99: self.VOutFlowCycle[i].priority = 2

            for i in range(self.nInflowingStreams):
                if self.VInFlowCycle[i].priority < 99: self.VInFlowCycle[i].priority = 2

            if self.VolProcMed.priority < 99: self.VolProcMed.priority = 2
            if self.QHXProc.priority < 99: self.QHXProc.priority = 2
            #if self.QWHProc.priority < 99: self.QWHProc.priority = 2 Now it coincides with UPHwOfProcess. Assigned priority 2 coherent with the QHX
            if self.HPerYearProc.priority < 99: self.HPerYearProc.priority = 2
            
            if self.NBatch.priority < 99: self.NBatch.priority = 3
            if self.NDaysProc.priority < 99: self.NDaysProc.priority = 3
            if self.HPerDayProc.priority < 99: self.HPerDayProc.priority = 3
            if self.UPHcGross.priority < 99: self.UPHcGross.priority = 3
            for i in xrange(self.nInflowingStreams):
                if self.QHXProcInt[i].priority < 99: self.QHXProcInt[i].priority = 3
            if self.TEnvProc.priority < 99: self.TEnvProc.priority = 3
            if self.UAProc.priority < 99: self.UAProc.priority = 3
            if self.dotQEvapProc.priority < 99: self.dotQEvapProc.priority = 3
            
            # sense of the if: -> if the value is not needed, because massflow = 0, then priority should remain 99 = not needed

        else:
            
            self.UPH.priority = 2
            self.UPHProc.priority = 2
            self.UPHs.priority = 2
            self.UPHcOfProcess.priority = 2
            self.UPHm.priority = 2 
            self.UPHwOfProcess.priority = 2 
            self.PT.priority = 2
            self.dotUPHm.priority = 2

            for i in range(self.nInflowingStreams):
                if self.PTInFlow[i].priority < 99: self.PTInFlow[i].priority = 2
                if self.PTInFlowRec[i].priority < 99: self.PTInFlowRec[i].priority = 2
            if self.PTStartUp.priority < 99: self.PTStartUp.priority = 2

            # here some verifications are redundant but they do not affect the results
            for i in range(self.nOutflowingStreams):
                if self.PTOutFlow[i].priority < 99: self.PTOutFlow[i].priority = 3
                if self.PTOutFlowRec[i].priority < 99: self.PTOutFlowRec[i].priority = 3
                if self.VOutFlowCycle[i].priority < 99: self.VOutFlowCycle[i].priority = 3
            for i in range(self.nInflowingStreams):
                if self.VInFlowCycle[i].priority < 99: self.VInFlowCycle[i].priority = 3
            if self.VolProcMed.priority < 99: self.VolProcMed.priority = 3
            if self.TEnvProc.priority < 99: self.TEnvProc.priority = 3
            if self.UAProc.priority < 99: self.UAProc.priority = 3
            if self.dotQEvapProc.priority < 99: self.dotQEvapProc.priority = 3
            if self.QHXProc.priority < 99: self.QHXProc.priority = 3
            #if self.QWHProc.priority < 99: self.QWHProc.priority = 3
            if self.HPerYearProc.priority < 99: self.HPerYearProc.priority = 3
            if self.NBatch.priority < 99: self.NBatch.priority = 3
            if self.NDaysProc.priority < 99: self.NDaysProc.priority = 3
            if self.HPerDayProc.priority < 99: self.HPerDayProc.priority = 3
            if self.UPHcGross.priority < 99: self.UPHcGross.priority = 3
            for i in range(self.nInflowingStreams):
                if self.QHXProcInt[i].priority < 99: self.QHXProcInt[i].priority = 3

    def check(self):     #function that is called at the beginning when object is created
        """main function, carry out the check of the block"""

        proc = self.proc
        
        if DEBUG in ["ALL"]:
            print "-------------------------------------------------"
            print " Process checking"
            print "-------------------------------------------------"
        # check/create basic consistency assumptions
        
        cycle.initCheckBalance()
        # Step 1: Call all calculation routines in a given sequence
        if DEBUG in ["ALL", "MAIN"]:
            print "-------------------------------------------------"
            print "Step 1: calculating from left to right (CALC)"
            print "-------------------------------------------------"

        if self.FluidRhoCp is not None:
            self.ThermalMass.push(calcK("ThermalMass", self.FluidRhoCp, self.VolProcMed, unitType="THERMALMASS"))
            
        self.HPerYearProc.push(calcProd("HPerYearProc", self.NDaysProc, self.HPerDayProc, unitType = "TIME"))
        self.HPerDayOp.push(calcProd("HPerDayOp", self.HBatch, self.NBatch, unitType = "TIME"))
        self.HPerDayProc.push(calcProd("HPerDayProc", self.PartLoad, self.HPerDayOp, unitType = "TIME"))

        if self.proc.scheduleType == 0 and not self.continuous:
            self.DtInFlow.push(calcK("DtInFlow", DEFAULTCHARGETIME, self.HBatch))
            self.DtOutFlow.push(calcK("DtOutFlow", DEFAULTCHARGETIME, self.HBatch))

        self.HPerYearInFlowNom.push(calcProd("HPerYearInFlowNom", self.NCyclesPerYear, self.DtInFlow, unitType="TIME"))
        self.HPerYearOutFlowNom.push(calcProd("HPerYearOutFlowNom", self.NCyclesPerYear, self.DtOutFlow, unitType="TIME"))

        self.HPerYearInFlow.push(calcProd("HPerYearInFlow", self.HPerYearInFlowNom, self.PartLoad, unitType="TIME"))
        self.HPerYearOutFlow.push(calcProd("HPerYearOutFlow", self.HPerYearOutFlowNom, self.PartLoad, unitType="TIME"))
        self.HPerYearProc.push(calcProd("HPerYearOutFlow", self.HPerYearOp, self.PartLoad, unitType="TIME"))
        for i in range(self.nInflowingStreams):
            self.VInFlowCycleEff[i].push(calcProd("VInFlowCycleEff", self.PartLoad, self.VInFlowCycle[i], unitType = "VOLUME"))
        for i in range(self.nOutflowingStreams):
            self.VOutFlowCycleEff[i].push(calcProd("VOutFlowCycleEff", self.PartLoad, self.VOutFlowCycle[i], unitType = "VOLUME"))

        # TODO: check whether UAProc !=0, and negative DTLoss throws conflict. (it should)
        if self.dotUPHm.val != 0:
            if self.isHeatingProcess:
                self.DTLoss.push(calcDiff("DTLoss", self.PT, self.TEnvProc, parType="DTmN"))
            else:
                self.DTLoss.push(calcDiff("DTLoss", self.TEnvProc, self.PT, parType="DTmN"))
        
        self.dotQLossProc.push(calcProd("dotQLossProc", self.UAProc, self.DTLoss, parType="S", unitType = "ENERGY"))

        # UA: add suggestion how to calculate
        self.dotUPHm.push(calcSum("dotUPHm", self.dotQLossProc, self.dotQEvapProc))
        self.UPHm.push(calcProd("UPHm", self.dotUPHm, self.HPerYearProc, unitType = "ENERGY"))
        for i in range(self.nInflowingStreams):
            if self.isHeatingProcess:
                self.dUPHcGross[i].push(calcFlow_stack("dUPHcGross", 
                                                   self.FluidRhoCp_c[i], 
                                                   self.VInFlowCycleEff[i],
                                                   self.PT, 
                                                   self.PTInFlow[i],
                                                   self.DTUPHcGross[i], unitType = "ENERGY"))
            else: #cooling
                self.dUPHcGross[i].push(calcFlow_stack("dUPHcGross", 
                                                   self.FluidRhoCp_c[i], 
                                                   self.VInFlowCycleEff[i],
                                                   self.PTInFlow[i],
                                                   self.PT, 
                                                   self.DTUPHcGross[i], unitType = "ENERGY"))
        if any(self.internalHR):
#                self.dQHXProcInt1 = calcFlow("dQHXProcInt1",self.FluidRhoCp_w,self.VOutFlowCycle,
#                                               self.PTOutFlowRec,self.PTOutFlow,
#                                                   self.DTQHXOut,self.DTQHXOut1)
            if self.isHeatingProcess:
                HOutFlowRec = self.HOutFlowRec
                HOutFlow = self.HOutFlow
                PTInFlowRec = self.PTInFlowRec
                PTInFlow = self.PTInFlow
            else: # cooling
                HOutFlowRec = self.HOutFlow
                HOutFlow = self.HOutFlowRec
                PTInFlowRec = self.PTInFlow
                PTInFlow = self.PTInFlowRec
            for i in range(self.nOutflowingStreams):
                self.dHOutFlow[i].push(calcDiff("dHOutFlow", HOutFlowRec[i], HOutFlow[i]))
                self.dQHXProcIntOut[i].push(calcProdC("dQHXProcInt", self.FluidDensity_w[i],
                                            self.dHOutFlow[i], self.VOutFlowCycleEff[i], unitType = "ENERGY"))
            for i in range(self.nInflowingStreams):
                self.dQHXProcIntIn[i].push(calcFlow_stack("dQHXProcInt", 
                                                            self.FluidRhoCp_c[i], 
                                                            self.VInFlowCycleEff[i],
                                                            PTInFlowRec[i], 
                                                            PTInFlow[i],
                                                            self.DTQHXIn[i], unitType = "ENERGY"))
            self.dQHXProcInt.push(calcRowSum("dQHXProcInt", self.dQHXProcIntIn))
            self.dQHXProcInt.push(calcRowSum("dQHXProcInt", self.dQHXProcIntOut))
            
        else:
            
            for item in self.dQHXProcIntIn:
                item.setValue(0)
            for item in self.dQHXProcIntOut:
                item.setValue(0)
        for i in range(self.nInflowingStreams):
            self.dUPHc[i].push(calcDiff("dUPHc", self.dUPHcGross[i], self.dQHXProcIntIn[i]))
            if self.isHeatingProcess:
                self.dUPHc[i].push(calcFlow_stack("dUPHc", 
                                                    self.FluidRhoCp_c[i], 
                                                    self.VInFlowCycleEff[i],
                                                    self.PT, 
                                                    self.PTInFlowRec[i],
                                                    self.DTUPHcNet[i], unitType = "ENERGY"))
                self.dotUPHc[i].push(calcFlow_stack("dotUPHc", 
                                                    self.FluidCp_c[i], 
                                                    self.mInFlowNom[i],
                                                    self.PT, 
                                                    self.PTInFlowRec[i],
                                                    self.DTUPHcNet[i], unitType = "POWER"))
            else: # cooling
                self.dUPHc[i].push(calcFlow_stack("dUPHc", 
                                                    self.FluidRhoCp_c[i], 
                                                    self.VInFlowCycleEff[i],
                                                    self.PTInFlowRec[i],
                                                    self.PT, 
                                                    self.DTUPHcNet[i], unitType = "ENERGY"))
                self.dotUPHc[i].push(calcFlow_stack("dotUPHc", 
                                                    self.FluidCp_c[i], 
                                                    self.mInFlowNom[i],
                                                    self.PTInFlowRec[i],
                                                    self.PT, 
                                                    self.DTUPHcNet[i], unitType = "POWER"))
            self.UPHc[i].push(calcProdC("UPHc Stream %s" 
                                        % proc.inStreams[i].Name, 
                                        self.FluidCp_c[i], 
                                        self.mInFlowYear[i], 
                                        self.DTUPHcNet[i], unitType = "ENERGY"))
            self.UPHc[i].push(calcProd("UPHc Stream %s" 
                                        % proc.inStreams[i].Name, 
                                        self.dotUPHc[i], 
                                        self.HPerYearInFlow, 
                                        unitType = "ENERGY"))
            
        self.UPHcOfProcess.push(calcRowSum("UPHcOfProcess", self.UPHc))

        sum_dUPHc = calcRowSum('UPHc_streamSum', self.dUPHc)
##        sum_dotUPHc = calcRowSum('dotUPHc_streamSum', self.dotUPHc)
        self.UPHcOfProcess.push(calcProd("UPHcOfProcess", sum_dUPHc, self.NCyclesPerYear, unitType = "ENERGY"))
##        self.UPHcOfProcess.push(calcProd("UPHc", sum_dotUPHc, self.HPerYearInFlow, unitType = "ENERGY"))

        UPHcStreams = [None] * self.nInflowingStreams
        for i in xrange(self.nInflowingStreams):
            stream = proc.inStreams[i]
            UPHcStreams[i] = calcProdC("UPHc [%s]" % (stream.Name,),
                                    self.FluidCp_c[i], self.mInFlowYear[i],
                                    self.DTUPHcNet[i], unitType = "ENERGY")
        self.UPHcOfProcess.push(calcRowSum("UPHcOfProcess", UPHcStreams))
        
        #Start-up heat demand UPHs
        if self.startUp:
            if self.isHeatingProcess:
                PTStartUp1 = self.PT
                PTStartUp0 = self.PTStartUp
            else:
                PTStartUp1 = self.PTStartUp
                PTStartUp0 = self.PT

            self.DTUPHs.push(calcDiff("DTUPHs", PTStartUp1, PTStartUp0))
                
        self.dUPHs.push(calcProd("dUPHs", self.ThermalMass, self.DTUPHs, unitType = "ENERGY"))

        if self.FluidRhoCp is not None:
            self.dUPHs.push(calcProdC("dUPHs", self.FluidRhoCp, self.VolProcMed, self.DTUPHs, unitType = "ENERGY"))
                
        self.NCyclesPerYear.push(calcProd("NCyclesPerYear", self.NDaysProc, self.NBatch))
        self.UPHs.push(calcProd("UPHs", self.dUPHs, self.NCyclesPerYear, unitType = "ENERGY"))

        #Total heat demand UPH = UPHm + UPHs + UPHc
        self.UPH.push(calcSum3("UPH", self.UPHm, self.UPHcOfProcess, self.UPHs))

        if self.isHeatingProcess:
            self.UPH.push(calcSum("UPH", self.UPHProc, self.QHXProc))
##            scheduleDiff(self.ccsUPHProc, self.ccsUPH, self.ccsQHXProc)
        else:
            self.UPH.push(calcSum("UPH", self.UPHProc, self.QWHProcUsed))
##            scheduleDiff(self.ccsUPHProc, self.ccsUPH, self.ccsQWHProcUsed)

        # only if single in-flowing and single out-flowing stream, else skip, see also adjust
        if (self.nInflowingStreams == self.nOutflowingStreams == 1) and any(self.internalHR):
            if self.isHeatingProcess:
                PTOutFlow = self.PTOutFlow
                PTInFlow = self.PTInFlow
                PTOutFlowRec = self.PTOutFlowRec
                PTInFlowRec = self.PTInFlowRec
            else:
                PTOutFlow = self.PTInFlow
                PTInFlow = self.PTOutFlow
                PTOutFlowRec = self.PTInFlowRec
                PTInFlowRec = self.PTOutFlowRec

            if proc.inStreams[0].HeatRecExist or proc.outStreams[0].HeatRecExist:
                self.DTCrossHXLT.push(calcDiff("DTCrossHXLT", PTOutFlow[0], 
                                               PTInFlow[0]))
                self.DTCrossHXHT.push(calcDiff("DTCrossHXHT", PTOutFlowRec[0],
                                               PTInFlowRec[0]))
        
        if self.isHeatingProcess:
            HOutFlow = self.HOutFlow
            HFinal = self.HFinal
            HFinalRequired = self.HFinalRequired
        else: # cooling
            HOutFlow = self.HFinal
            HFinal = self.HOutFlow
            # There is no required temperature for out streams that would need heating.
            HFinalRequired = self.HFinalRequired 
        recoverableIDs = []
        for i in range(self.nOutflowingStreams):
            # calculate mandatory cooling down if any exists
            if proc.outStreams[i].PTFinalRequired is not None:
                self.dHOutFlowRequired[i].push(calcDiff("dHOutFlowRequired",
                                                        HOutFlow[i],
                                                        HFinalRequired[i]))
                self.dHOutFlowRes[i].push(calcDiff("dhOutflowRes",
                                                   self.HFinalRequired[i],
                                                   self.HFinal[i]))
                self.UPHwReq[i].push(calcProd("UPHwRequired Stream: %s" 
                                              % proc.outStreams[i].Name,
                                              self.mOutFlowYear[i],
                                              self.dHOutFlowRequired[i], unitType = "ENERGY"))
                self.HFinalRequired[i].push(calcH("HFinalRequired", 
                                                  self.FluidCp_w[i],
                                                  self.FluidCp_w[i],
                                                  self.Fluid_hL_w[i],
                                                  self.FluidTCond_w[i],
                                                  self.PTFinalRequired[i], 
                                                  self.XFinalRequired[i]))
                self.UPHwReq_dot[i].push(calcProdC("UPHwReq_dot", 
                                                   self.FluidDensity_w[i],
                                                   self.dHOutFlowRequired[i], 
                                                   self.VOutFlowCycleEff[i], unitType = "ENERGY"))
                self.UPHwRes[i].push(calcProd("UPHwRes",
                                              self.mOutFlowYear[i],
                                              self.dHOutFlowRes[i], unitType = "ENERGY"))
                self.UPHwRes_dot[i].push(calcProdC("UPHwRes_dot", 
                                                   self.FluidDensity_w[i],
                                                   self.dHOutFlowRes[i], 
                                                   self.VOutFlowCycleEff[i], unitType = "ENERGY"))
            else:
                self.UPHwReq_dot[i].setValue(0.0)
                self.UPHwReq_dot[i].push()
                self.UPHwReq[i].setValue(0.0)
                self.UPHwReq[i].push()
                
            if self.HeatRecOK[i]:
#                   self.dUPHw1 = calcFlow("dUPHw1",self.FluidRhoCp_w,self.VOutFlowCycle,
#                                          self.PTOutFlow,self.PTFinal,
#                                          self.DTOutFlow,self.DTOutFlow1)
                recoverableIDs.append(i)
                self.dHFinal[i].push(calcDiff("dHFinal", 
                                              HOutFlow[i], 
                                              HFinal[i]))
                self.dUPHw[i].push(calcProdC("dUPHw", 
                                                self.FluidDensity_w[i],
                                                self.dHFinal[i], 
                                                self.VOutFlowCycleEff[i], unitType = "ENERGY"))
                self.UPHw[i].push(calcProd("UPHw [%s]" 
                                           % proc.outStreams[i].Name, 
                                           self.mOutFlowYear[i], 
                                           self.dHFinal[i], unitType = "ENERGY"))
                self.dotUPHw[i].push(calcProd("dotUPHw [%s]"% proc.outStreams[i].Name, 
                                              self.mOutFlowNom[i], 
                                              self.dHFinal[i], unitType = "POWER"))
                # TODO: find out whether h/c difference exists
                self.HOutFlowRec[i].push(calcH("HOutFlow", 
                                               self.FluidCp_w[i],
                                               self.FluidCp_w[i],
                                               self.Fluid_hL_w[i],
                                               self.FluidTCond_w[i],
                                               self.PTOutFlowRec[i], 
                                               self.XOutFlowRec[i]))
                self.HOutFlow[i].push(calcH("HOutFlow", 
                                            self.FluidCp_w[i],
                                            self.FluidCp_w[i],
                                            self.Fluid_hL_w[i],
                                            self.FluidTCond_w[i],
                                            self.PTOutFlow[i], 
                                            self.XOutFlow[i]))
                self.HFinal[i].push(calcH("HFinal", 
                                          self.FluidCp_w[i],
                                          self.FluidCp_w[i],
                                          self.Fluid_hL_w[i],
                                          self.FluidTCond_w[i],
                                          self.PTFinal[i], self.XFinal[i]))
            else:
                self.dUPHw[i].setValue(0.0, setTrack = False)
                self.dUPHw[i].push()
                self.UPHw[i].setValue(0.0, setTrack = False)
                self.UPHw[i].push()
                self.dotUPHw[i].setValue(0.0, setTrack = False)
                self.dotUPHw[i].push()

            self.mOutFlowYear[i].push(calcProd("mOutFlowYear", 
                                               self.mOutFlowNom[i], 
                                               self.HPerYearOutFlow, unitType = "MASS"))
            self.mOutFlowYear[i].push(calcProdC("mOutFlowYear", 
                                                self.FluidDensity_w[i], 
                                                self.VOutFlowCycleEff[i], 
                                                self.NCyclesPerYear, unitType = "MASS"))
                
            self.UPHw[i].push(calcSum("UPHw", self.UPHwReq[i],
                                          self.UPHwRes[i]))
        # Count the outflow of out streams only if they have
        # heat that may be recovered.
        yearlyHRableOutflow = [self.mOutFlowYear[i] 
                               for i in recoverableIDs]
        if yearlyHRableOutflow != []:
            self.mOutFlowYearOfProcess.push(calcRowSum("", 
                                                       yearlyHRableOutflow))

        UPHw_streamSum = calcRowSum("UPHw_streamSum", self.dUPHw)
        self.UPHwOfProcess.push(calcProd("UPHwOfProcess", UPHw_streamSum, self.NCyclesPerYear, unitType = "ENERGY"))

        UPHwReq_streamSum = calcRowSum("UPHwReq_streamSum", self.UPHwReq_dot)
        self.UPHwReqOfProcess.push(calcProd("UPHwReqOfProcess", 
                                            UPHwReq_streamSum, 
                                            self.NCyclesPerYear, unitType = "ENERGY"))

        UPHwRes_streamSum = calcRowSum("UPHwRes_streamSum", self.UPHwRes_dot)
        self.UPHwResOfProcess.push(calcProd("UPHwResOfProcess", 
                                            UPHwRes_streamSum, 
                                            self.NCyclesPerYear, unitType = "ENERGY"))

        
        self.UPHwOfProcess.push(calcSum("UPHwOfProcess", 
                                        self.UPHwReqOfProcess, 
                                        self.UPHwResOfProcess))
        self.UPHwOfProcess.push(calcSum("UPHwOfProcess",
                                        self.UPCProc,
                                        self.QWHProc))
        self.UPHwReqOfProcess.push(calcSum("UPHwReqOfProcess",
                                           self.UPCProc,
                                           self.QWHProcReq))

        self.UPHwResOfProcess.push(calcSum("UPHwResOfProcess",
                                           self.QWHAmb,
                                           self.QWHProcRes))

        if self.isHeatingProcess:

            self.QWHProc.push(calcSum("QWHProc",
                                               self.QWHAmb,
                                               self.QWHProcUsed))

            self.QWHProcUsed.push(calcSum("QWHProcUsed",
                                               self.QWHProcReq,
                                               self.QWHProcRes))
        else:

            self.QWHProc.push(calcSum("QWHProc",
                                               self.QWHAmb,
                                               self.QHXProc))
            self.QHXProc.push(calcSum("QWHProcUsed",
                                               self.QWHProcReq,
                                               self.QWHProcRes))

            
            

        for i in range(self.nInflowingStreams):
            self.mInFlowYear[i].push(calcProd("mInFlowYear", self.mInFlowNom[i], self.HPerYearInFlow, unitType = "MASS"))
            self.mInFlowYear[i].push(calcProdC("mInFlowYear", self.FluidDensity_c[i], self.VInFlowCycleEff[i], self.NCyclesPerYear, unitType = "MASS"))
###CHECK: sum of massflows should be added
# TODO: to check against what? The process mass inflows seem to be calculated nowhere else.


        if DEBUG in ["ALL", "MAIN"]:
            self.showAll()

# Step 2: Cross check the variables

            print "-------------------------------------------------"
            print "Step 2: cross checking"
            print "-------------------------------------------------"
            
        self.ccheckAll()            

        # Step 3: Adjust the variables (inverse of calculation routines)
        if DEBUG in ["ALL", "MAIN"]:
            self.showAll()
            print "-------------------------------------------------"
            print "Step 3: calculating from right to left (ADJUST)"
            print "-------------------------------------------------"

        if self.FluidRhoCp is not None:
            adjustK_stack(self.ThermalMass, self.FluidRhoCp, self.VolProcMed)

        for i in range(self.nInflowingStreams):
            adjustProd_stack(self.mInFlowYear[i], self.mInFlowNom[i], 
                             self.HPerYearInFlow)
            adjustProdC_stack(self.mInFlowYear[i], self.FluidDensity_c[i], 
                              self.VInFlowCycleEff[i], self.NCyclesPerYear)
        if yearlyHRableOutflow != []:
            adjRowSum_stack(self.mOutFlowYearOfProcess, yearlyHRableOutflow)
        if self.isHeatingProcess:
            HOutFlow = self.HOutFlow
            HFinal = self.HFinal
            HFinalRequired = self.HFinalRequired
        else:
            HOutFlow = self.HFinal
            HFinal = self.HOutFlow
            # There is no required temperature for out streams that would need heating.
            HFinalRequired = self.HFinalRequired
        for i in range(self.nOutflowingStreams):
            if self.HeatRecOK[i]:
                #                adjustFlow(self.dUPHw1,self.FluidRhoCp_w,self.VOutFlowCycleEff2,
#                                     self.PTOutFlow,self.PTFinal,self.DTOutFlow,self.DTOutFlow1)
                adjustH_stack(self.HOutFlow[i], self.FluidCp_w[i],
                              self.FluidCp_w[i],
                              self.Fluid_hL_w[i],
                              self.FluidTCond_w[i],
                              self.PTOutFlow[i], self.XOutFlow[i])
                # TODO: check whether h/c difference exists
                adjustH_stack(self.HOutFlowRec[i], self.FluidCp_w[i],
                              self.FluidCp_w[i],
                              self.Fluid_hL_w[i],
                              self.FluidTCond_w[i],
                              self.PTOutFlowRec[i], self.XOutFlowRec[i])
                adjustH_stack(self.HFinal[i], self.FluidCp_w[i],
                          self.FluidCp_w[i],
                          self.Fluid_hL_w[i],
                          self.FluidTCond_w[i],
                          self.PTFinal[i], self.XFinal[i])
                adjustDiff_stack(self.dHFinal[i], HOutFlow[i], 
                                 HFinal[i])
                adjustProdC_stack(self.dUPHw[i], self.FluidDensity_w[i], 
                                  self.dHFinal[i], self.VOutFlowCycleEff[i])
                adjustProd_stack(self.UPHw[i], self.mOutFlowYear[i], 
                                 self.dHFinal[i])
                adjustProd_stack(self.dotUPHw[i], self.mOutFlowNom[i], 
                                 self.dHFinal[i])


            else:
                self.PTFinal[i].update(self.PTOutFlow[i]) #for security (link with HR module): -> zero UPHwOfProcess !!!
                self.PTFinal[i].push()

            adjustProd_stack(self.mOutFlowYear[i], self.mOutFlowNom[i], self.HPerYearOutFlow)
            adjustProdC_stack(self.mOutFlowYear[i], self.FluidDensity_w[i], self.VOutFlowCycleEff[i], self.NCyclesPerYear)

            # mandatory cooling down demand is independent of HR  
            if proc.outStreams[i].PTFinalRequired is not None:
                adjustProd_stack(self.UPHwRes[i], self.mOutFlowYear[i],
                                 self.dHOutFlowRes[i])
                
                adjustProdC_stack(self.UPHwReq_dot[i], self.FluidDensity_w[i], 
                                  self.dHOutFlowRequired[i], self.VOutFlowCycleEff[i])
                adjustProdC_stack(self.UPHwRes_dot[i], self.FluidDensity_w[i], 
                                  self.dHOutFlowRes[i], self.VOutFlowCycleEff[i])
                adjustH_stack(self.HFinalRequired[i], self.FluidCp_w[i],
                              self.FluidCp_w[i],
                              self.Fluid_hL_w[i],
                              self.FluidTCond_w[i],
                              self.PTFinalRequired[i], 
                              self.XFinalRequired[i])
                adjustDiff_stack(self.dHOutFlowRequired[i], HOutFlow[i],
                                 HFinalRequired[i])
                adjustDiff_stack(self.dHOutFlowRes[i], 
                                 self.HFinalRequired[i],
                                 self.HFinal[i])
                adjustProd_stack(self.UPHwReq[i], self.mOutFlowYear[i], 
                                 self.dHOutFlowRequired[i])
            adjustSum_stack(self.UPHw[i], self.UPHwReq[i], self.UPHwRes[i])

        adjustSumOfProds_stack(self.UPHwReqOfProcess,
                               self.UPHwReq_dot, 
                               (self.NCyclesPerYear,) * self.nOutflowingStreams)

        adjustSumOfProds_stack(self.UPHwResOfProcess,
                               self.UPHwRes_dot, 
                               (self.NCyclesPerYear,) * self.nOutflowingStreams)

        adjustSumOfProds_stack(self.UPHwOfProcess, self.dUPHw, 
                               (self.NCyclesPerYear,) * self.nOutflowingStreams)
        
        adjustSum_stack(self.UPHwOfProcess, self.UPHwReqOfProcess, 
                        self.UPHwResOfProcess)
        adjustSum_stack(self.UPHwOfProcess, self.UPCProc, self.QWHProc)
        adjustSum_stack(self.UPHwReqOfProcess, self.UPCProc, 
                        self.QWHProcReq)
        adjustSum_stack(self.UPHwResOfProcess, self.QWHAmb, 
                        self.QWHProcRes)

        if self.isHeatingProcess:

            adjustSum_stack(self.QWHProc, self.QWHAmb, self.QWHProcUsed)
            adjustSum_stack(self.QWHProcUsed, self.QWHProcReq, self.QWHProcRes)
        else:
            adjustSum_stack(self.QWHProc, self.QWHAmb, self.QHXProc)
            adjustSum_stack(self.QHXProc, self.QWHProcReq, self.QWHProcRes)

        if any(self.internalHR):
            adjRowSum_stack(self.dQHXProcInt, self.dQHXProcIntIn, self.nInflowingStreams)
            adjRowSum_stack(self.dQHXProcInt, self.dQHXProcIntOut, self.nOutflowingStreams)
            
        if (self.nInflowingStreams == self.nOutflowingStreams == 1) and any(self.internalHR):
            adjustDiff_stack(self.DTCrossHXHT, PTOutFlowRec[0], PTInFlowRec[0])
            adjustDiff_stack(self.DTCrossHXLT, PTOutFlow[0], PTInFlow[0])
        # only if single in-flowing and single out-flowing stream, else skip, see also calc
        if (self.nInflowingStreams == self.nOutflowingStreams == 1) and any(self.internalHR):
            if self.isHeatingProcess:
                PTInFlowRec = self.PTInFlowRec
                PTInFlow = self.PTInFlow
            else: # cooling
                PTInFlowRec = self.PTInFlow
                PTInFlow = self.PTInFlowRec
            adjustFlow_stack(self.dQHXProcIntIn, self.FluidRhoCp_c, 
                                 self.VInFlowCycleEff, PTInFlowRec, 
                                 PTInFlow, self.DTQHXIn)
            adjustProdC_stack(self.dQHXProcIntIn, self.FluidDensity_w, 
                              self.dHOutFlow, self.VOutFlowCycleEff)
        if self.isHeatingProcess:
            adjustDiff_stack(self.dHOutFlow, self.HOutFlowRec, 
                             self.HOutFlow)
        else: # cooling
            adjustDiff_stack(self.dHOutFlow, self.HOutFlow, 
                             self.HOutFlowRec)
             
        if self.isHeatingProcess:
            adjustSum_stack(self.UPH, self.UPHProc, self.QHXProc)
        else:
            adjustSum_stack(self.UPH, self.UPHProc, self.QWHProcUsed)
            
        adjustSum3_stack(self.UPH, self.UPHm, self.UPHcOfProcess, self.UPHs)
        
        adjustProd_stack(self.UPHs, self.dUPHs, self.NCyclesPerYear)
        adjustProd_stack(self.NCyclesPerYear, self.NDaysProc, self.NBatch)

        if self.startUp:
            adjustDiff_stack(self.DTUPHs, PTStartUp1, PTStartUp0)
        adjustProd_stack(self.dUPHs, self.ThermalMass, self.DTUPHs)
        if self.FluidRhoCp is not None:
            adjustProdC_stack(self.dUPHs, self.FluidRhoCp, self.VolProcMed, self.DTUPHs)

        adjustSumOfProdsC_stack(self.UPHcOfProcess, self.FluidCp_c,
                               self.mInFlowYear, self.DTUPHcNet)
        adjustSumOfProds_stack(self.UPHcOfProcess, self.dUPHc, 
                              (self.NCyclesPerYear, ) * self.nInflowingStreams)
        if self.isHeatingProcess:
            adjustFlow_stack(self.dUPHc, 
                             self.FluidRhoCp_c, 
                             self.VInFlowCycleEff, 
                             (self.PT,) * self.nInflowingStreams, 
                             self.PTInFlowRec, 
                             self.DTUPHcNet)
            adjustFlow_stack(self.dotUPHc,
                             self.FluidCp_c,
                             self.mInFlowNom,
                             (self.PT,) * self.nInflowingStreams,
                             self.PTInFlowRec,
                             self.DTUPHcNet)
        else:
            adjustFlow_stack(self.dUPHc, 
                             self.FluidRhoCp_c, 
                             self.VInFlowCycleEff, 
                             self.PTInFlowRec,
                             (self.PT,) * self.nInflowingStreams, 
                             self.DTUPHcNet)
            adjustFlow_stack(self.dotUPHc,
                             self.FluidCp_c,
                             self.mInFlowNom,
                             self.PTInFlowRec,
                             (self.PT,) * self.nInflowingStreams,
                             self.DTUPHcNet)

        adjustDiff_stack(self.dUPHc, self.dUPHcGross, self.dQHXProcIntIn)
        if self.isHeatingProcess:
            adjustFlow_stack(self.dUPHcGross, 
                         self.FluidRhoCp_c, 
                         self.VInFlowCycleEff,
                         (self.PT, )* self.nInflowingStreams, 
                         self.PTInFlow, 
                         self.DTUPHcGross)
        else: #cooling
            adjustFlow_stack(self.dUPHcGross, 
                         self.FluidRhoCp_c, 
                         self.VInFlowCycleEff,
                         self.PTInFlow,
                         (self.PT, )* self.nInflowingStreams, 
                         self.DTUPHcGross)
        adjustProd_stack(self.UPHm, self.dotUPHm, self.HPerYearProc)

        adjustSum_stack(self.dotUPHm, self.dotQLossProc, self.dotQEvapProc)
        adjustProd_stack(self.dotQLossProc, self.UAProc, self.DTLoss)
        if self.dotUPHm.val != 0:
            if self.isHeatingProcess:
                adjustDiff_stack(self.DTLoss, self.PT, self.TEnvProc)
            else: 
                adjustDiff_stack(self.DTLoss, self.TEnvProc, self.PT)
            
        for i in range(self.nInflowingStreams):
            adjustProd_stack(self.VInFlowCycleEff[i], self.PartLoad, self.VInFlowCycle[i])
        for i in range(self.nOutflowingStreams):
            adjustProd_stack(self.VOutFlowCycleEff[i], self.PartLoad, self.VOutFlowCycle[i])

        if self.proc.scheduleType == 0 and not self.continuous:
            adjustK_stack(self.DtInFlow, DEFAULTCHARGETIME, self.HBatch)
            adjustK_stack(self.DtOutFlow, DEFAULTCHARGETIME, self.HBatch)

        adjustProd_stack(self.HPerYearInFlowNom, self.NCyclesPerYear, self.DtInFlow)
        adjustProd_stack(self.HPerYearOutFlowNom, self.NCyclesPerYear, self.DtOutFlow)

        adjustProd_stack(self.HPerYearInFlow, self.HPerYearInFlowNom, self.PartLoad)
        adjustProd_stack(self.HPerYearOutFlow, self.HPerYearOutFlowNom, self.PartLoad)
        adjustProd_stack(self.HPerYearProc, self.HPerYearOp, self.PartLoad)
        adjustProd_stack(self.HPerDayOp, self.HBatch, self.NBatch)
        adjustProd_stack(self.HPerDayProc, self.PartLoad, self.HPerDayOp)
                    
        adjustProd_stack(self.HPerYearProc, self.NDaysProc, self.HPerDayProc)
        
        adjustProd_stack(self.UPHm,self.dotUPHm,self.HPerYearProc)

        # FIXME: delete self.dotQLossProc.push()
        adjustSum_stack(self.dotUPHm,self.dotQLossProc,self.dotQEvapProc)

        if DEBUG in ["ALL", "MAIN"]:
            self.showAll()
       
# Step 4: Second cross check of the variables

            print "-------------------------------------------------"
            print "Step 4: cross checking"
            print "-------------------------------------------------"
            
        self.ccheckAll()
        
        if DEBUG in ["ALL", "MAIN"]:
            self.showAll()

# Arrived at the end

        for i in range(self.nInflowingStreams):
            self.UPHcGrossProd[i] = calcProd("UPHcGrossProd_%s" % (i,), self.dUPHcGross[i], self.NCyclesPerYear) # Not adjusted
        self.UPHcGross = calcRowSum("UPHcGross", self.UPHcGrossProd)
        self.QHXProcInt = [calcProd("QHXProcInt_%s" % (i,), self.dQHXProcIntIn[i], self.NCyclesPerYear)
                           for i in xrange(self.nInflowingStreams)] # Not adjusted 
        
    #adido este ultimo show all. sino no se ven los ultimos dos resultados ...
        if DEBUG in ["ALL", "BASIC"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()

#------------------------------------------------------------------------------
    def ccheckAll(self):    
#------------------------------------------------------------------------------
#   check block
#------------------------------------------------------------------------------

        for i in range(self.nInflowingStreams):
            self.mInFlowYear[i].check()
            self.mInFlowNom[i].check()
        self.mInFlowYearOfProcess.check()

        for i in range(self.nOutflowingStreams):
            self.mOutFlowYear[i].check()
            self.mOutFlowNom[i].check()
        
        self.mOutFlowYearOfProcess.check()

        self.PT.check()
        for i in range(self.nInflowingStreams):
            self.PTInFlow[i].check()
            self.PTInFlowRec[i].check()
        self.PTStartUp.check()
        
        self.VolProcMed.check()
        self.ThermalMass.check()
        
        for i in range(self.nInflowingStreams):
            self.VInFlowCycle[i].check()
        for i in range(self.nOutflowingStreams):
            self.VOutFlowCycle[i].check()
        self.PartLoad.check()
                          
        for i in range(self.nInflowingStreams):
            self.VInFlowCycleEff[i].check()
        for i in range(self.nOutflowingStreams):
            self.VOutFlowCycleEff[i].check()
            self.PTOutFlow[i].check()
            self.PTOutFlowRec[i].check()
            
        # only if single in-flowing and single out-flowing stream
        if self.nInflowingStreams == self.nOutflowingStreams == 1:
            self.DTCrossHXLT.check()
            self.DTCrossHXHT.check()

        for i in range(self.nInflowingStreams):
            self.DTUPHcNet[i].check()
        
        self.HPerYearProc.check()
        self.HPerYearOp.check()
        self.HPerDayProc.check()
        self.HPerDayOp.check()
        self.HPerYearOutFlow.check()
        self.HPerYearInFlow.check()
        self.HPerYearOutFlowNom.check()
        self.HPerYearInFlowNom.check()
        self.DtInFlow.check()
        self.DtOutFlow.check()

        if self.continuous:
            ccheck1(self.HPerYearInFlow, self.HPerYearOutFlow)
            ccheck1(self.HPerYearProc, self.HPerYearInFlow)
            ccheck1(self.HPerYearProc, self.HPerYearOutFlow)

        self.HBatch.check()
        self.NBatch.check()
                          
        self.DTLoss.check()
        self.dotQLossProc.check()
        self.dotUPHm.check()
        self.dotQEvapProc.check()
        self.UPHm.check()
        for item in (self.dUPHcGross,
                     self.dUPHc,
                     self.dotUPHc,
                     self.dQHXProcIntIn, 
                     self.UPHc,
                     self.DTUPHcGross):
            for streamdata in item:
                streamdata.check() 
        self.UPHcOfProcess.check()
        self.dUPHs.check()
        self.NDaysProc.check()
        self.NCyclesPerYear.check()
        self.UPHs.check()
        self.DTUPHs.check()
        self.UPHProc.check()
        self.QHXProc.check()
        self.UPH.check()
        self.UAProc.check()
        self.TEnvProc.check()

        for item in (self.dUPHw, 
                     self.DTOutFlow, 
                     self.PTFinal, 
                     self.PTFinalRequired, 
                     self.HOutFlow, 
                     self.HOutFlowRec, 
                     self.HFinal,
                     self.dHOutFlow, 
                     self.dHFinal, 
                     self.XOutFlow,
                     self.XOutFlowRec, 
                     self.XFinal, 
                     self.XFinalRequired, 
                     self.dQHXProcIntOut,
                     self.UPHw,
                     self.dotUPHw,
                     self.UPHwReq,
                     self.UPHwReq_dot,
                     self.HFinalRequired,
                     self.dHOutFlowRequired,
                     self.UPHwRes,
                     self.UPHwRes_dot,
                     self.dHOutFlowRes):
            for streamdata in item:
                streamdata.check() 
             
###CHECK: for obtaining real equality AFTER checking it would be better first check both, then push one to the stack of the other, then check again,
### and then do a direct assignment
        self.dQHXProcInt.check()
        self.UPHwOfProcess.check()
##        self.UPHwOfProcess.push(self.QWHProc) #At the moment they are the same. Change it next future when vessel heat recovery will be implemented
##        self.QWHProc.push(self.UPHwOfProcess)
##        self.UPHwOfProcess.check()
        self.UPHwReqOfProcess.check()
        self.UPHwResOfProcess.check()
        self.QWHProc.check()
        self.UPCProc.check()
        self.QWHProcReq.check()
        self.QWHProcRes.check()
        self.QWHProcUsed.check()
        self.QWHAmb.check()

        for i, stream in enumerate(self.proc.inStreams):
            if stream.HeatRecExist == False:
                # no HR means the temperatures have to be equal
##                self.PTInFlowRec[i].update(self.PTInFlow[i])
                ccheck1(self.PTInFlowRec[i],self.PTInFlow[i])

            
    def estimate(self):  
        """
        Estimate some of the data that are not sufficiently precise.
        This should be a subset of the data that are within screen.
        (not necessarily ALL data have to be estimated)
        """
        if (all(self.dQHXProcIntOut[i].val is None for i in xrange(self.nOutflowingStreams))
               or not any(self.internalHR)):
            for i in range(self.nOutflowingStreams):
                self.PTOutFlowRec[i].setEstimate(self.PT.val, limits=(self.PT.valMin, self.PT.valMax))
            
        if self.nInflowingStreams == self.nOutflowingStreams == 1:
            self.VOutFlowCycle[0].setEstimate(self.VInFlowCycle[0].val, limits=(self.VInFlowCycle[0].valMin, self.VInFlowCycle[0].valMax))
            for i in range(self.nInflowingStreams):
                self.PTInFlow[i].calcDev()
                if self.PTInFlow[i].dev > 15.0:
                    if self.PTInFlow[i].val is None or abs(self.PTInFlow[i].val - 15.0) < self.PTInFlow[i].dev:
                        self.PTInFlow[i].setEstimate(15.0, limits=(5.0, 35.0))

        if any(self.internalHR):
            if self.nInflowingStreams == self.nOutflowingStreams == 1:
                for i in range(self.nInflowingStreams):
                    if (self.PT.val - self.PTInFlow[i].val > 5.0):
            
                        self.DTCrossHXHT.setEstimate(10.0, limits=(5.0, 999.0))
                        self.DTCrossHXLT.setEstimate(10.0, limits=(5.0, 999.0))
            
                    else:
                        self.PTInFlowRec[i].update(self.PTInFlow[i])
                        self.internalHR[i] = False
        
        
        if self.NBatch.val > 0 and (self.NBatch.val is not None):
            vol1 = 0.0
            for i in range(self.nInflowingStreams):
                try:
                    vol1 += self.VInFlowCycle[i].val / self.NBatch.val
                except TypeError:
                    pass # self.VInFlowCycle[i].val had no valid value yet
        else:
            vol1 = 0.0
            
        if self.VolProcMed.val is not None:
            vol2 = self.VolProcMed.val
        else:
            vol2 = 0.0
            
        vol = max(vol1, vol2)

        if vol < INFINITE:
        
            surface = 10.0 * pow(vol, 2.0 / 3)
            UAmin = 0.0004 * surface * 0.1  #0.4 W/m2K well insulated vessel
            UA = 0.008 * surface
            UAmax = 0.002 * surface * 1.0   #2.0 W/m2K badly insulated equipment
            
##            self.UAProc.setEstimate(UA, limits=(UAmin, UAmax))

            # estimate of Evaporation losses

##        dotQLossProcMax = UAmax * self.DTLoss.valMax
##        
##        if self.dotUPHm.valMax <= dotQLossProcMax:    # maintenance can be fully explained by
##                                                # thermal losses
##            self.dotQEvapProc.setEstimate(0.0, limits=(0.0, 0.0))
            
# WARNING: this apriori assumption can give problems in cases with evaporation !!!!

        


    def getInStreamNo(self, name):

        for i, stream in enumerate(proc.inStreams):
            if stream.Name == name:
                return(i+1)
        return None

    def getOutStreamNo(self,name):
        
        for i, stream in enumerate(proc.outStreams):
            if stream.Name == name:
                return(i+1)
        return None

def checkForCondensation(fluid, T0, TTarget):
    """ checks whether condensation may occur in the given temperature range
        
    """
    Tc = fluid.TCond
    if Tc is None:
        return False, Tc
    elif (T0 >= Tc and TTarget <= Tc) or (TTarget >= Tc and T0 <= Tc):
        return True, Tc
    else:
        return False, Tc

        
#==============================================================================
if __name__ == "__main__":
    import einstein.GUI.pSQL as pSQL, MySQLdb
# direct connecting to SQL database w/o GUI. for testing only
    stat = Status("testCheckProc")
    Status.SQL = MySQLdb.connect(user="root", db="einstein")
    Status.DB = pSQL.pSQL(Status.SQL, "einstein")
    Status.PId = 41
    Status.ANo = -1
#..............................................................................
    
    screen = CCScreen()
    
    ccProc = CheckProc(1)       # creates an instance of class CCheck
    ccProc.check()
    ccProc.exportData()

    ccProc.screen()
    screen.show()
    
#==============================================================================
