# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CheckSchedules
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Consistency check of time schedules
#
#==============================================================================
#
#	(C) copyleft E4-Experts GmbH, Berlin, Germany
#	www.energyxperts.net / info@energyxperts.net
#
#============================================================================== 				

from GUI.status import Status
from constants import DEBUG, EPSILON, INFINITE
from modules.messageLogger import *
from auxiliary.auxiliary import normalize, toFloat, nanmean, noneCheck
import numpy as np
from numpy.linalg import matrix_rank
from lib.optimise.hessian import hessian, test_hessian
from scipy.optimize import leastsq
from scipy.linalg import lu, inv
import database.dbio as dbio
from ccPar import DEFAULT_SQERR
from dataSet import DataSet, DataFile, loadDataSets, readDataFile, objectCodes, makeSchedule
from ccSchedule import CCSchedule, scheduleRowSumC, scheduleRowSum, scheduleProd, scheduleDiff
from modules.schedules.schedules import globalsFromFav
from ccTracker import cycle

MINSCHEDULEWEIGHT = 0.01    # minimum weight for components of derived schedules for being stored

class CheckSchedules():

    def __init__(self, moduleCC):
        self.cc = moduleCC

        # create working space
        self.ccsUPH = [ccProc.ccsUPH for ccProc in self.cc.ccProc]
        self.ccsUPHc = [ccProc.ccsUPHc for ccProc in self.cc.ccProc]
        self.ccsUPHw = [ccProc.ccsUPHw for ccProc in self.cc.ccProc]
        self.ccsmInFlowYear = [ccProc.ccsmInFlowYear for ccProc in self.cc.ccProc]
        self.ccsmOutFlowYear = [ccProc.ccsmOutFlowYear for ccProc in self.cc.ccProc]

        # determine which of the UPH schedules have to be auto-fit and add the rest to the list of existing profiles
        self.autoFit = [k for k, ccs in enumerate(self.ccsUPH) if ccs.schedule is None]
        self.calcFromDS = []      # those process schedules of autoFit which can be calculated directly from data sets
        
        self.ccsQHXProc = [CCSchedule(ccProc.QHXProc) for ccProc in self.cc.ccProc]

        self.ccsUPHProc_k = [CCSchedule(ccProc.UPHProc) for ccProc in self.cc.ccProc]
        
        self.ccsUPHProc_m = [CCSchedule(ccPipe.UPHProcm) for ccPipe in self.cc.ccPipe]
        self.ccsUSHm = [CCSchedule(ccPipe.USHm) for ccPipe in self.cc.ccPipe]
        self.ccsUSHPipe = [CCSchedule(ccPipe.USHPipe) for ccPipe in self.cc.ccPipe]
        self.ccsQHXPipe = [CCSchedule(ccPipe.QHXPipe) for ccPipe in self.cc.ccPipe]
        self.HDEffAvg = [1.0 for ccPipe in self.cc.ccPipe]
        self.HDEffAvgErr = [True for ccPipe in self.cc.ccPipe]

        self.ccsUSHj =      [CCSchedule(ccEq.USHj) for ccEq in self.cc.ccEq]
        self.ccsUSHConv =      [CCSchedule(ccEq.USHConv) for ccEq in self.cc.ccEq]
        self.ccsQHXEq =      [CCSchedule(ccEq.QHXEq) for ccEq in self.cc.ccEq]
        self.ccsFETFuel_j =  [CCSchedule(ccEq.FETFuel_j) for ccEq in self.cc.ccEq]
        self.ccsFETel_j =  [CCSchedule(ccEq.FETel_j) for ccEq in self.cc.ccEq]
        self.f_fuel = [1./0.9 for ccEq in self.cc.ccEq]
        self.f_el = [0.0 for ccEq in self.cc.ccEq]

        self.ccsFETFuel_i =  [CCSchedule(ccFET.FETFuel) for ccFET in self.cc.ccFET]
        self.ccsFETel_c_i =  CCSchedule(self.cc.ccFETel.FETel_c)

        self.ccsFECFuel_i =  [CCSchedule(ccFET.FECFuel) for ccFET in self.cc.ccFET]
        self.ccsFECel_c_i =  CCSchedule(self.cc.ccFETel.FECel_c)

        self.ccsProduct = [CCSchedule(ccnode.q) for ccnode in self.cc.ccMassFlow.ccnode if ccnode.origin.objectType=="PROD"]
        self.ccsRawMat  = [CCSchedule(ccnode.q) for ccnode in self.cc.ccMassFlow.ccnode if ccnode.origin.objectType=="RAW"]

        self.xIndexes = [[]]*len(self.ccsUPH) # mapping of process indexes k to indexes i in massflow-x-vector
        self.xIndexesProd = [None]*len(self.ccsProduct)
        self.dsProd = [None]*len(self.ccsProduct)

        self.dsFits = None # data sets for fits of process schedules

        self.maxDevSC = 0


##        # load target data sets from DB
##        self.dataSets, dataFiles = loadDataSets()
        dataFiles = Status.int.systemModel.dataFiles
        self.dataSets = [ds for ds in Status.int.systemModel.dataSets \
                         if ds.parCode not in ["TDrive", "TReject", "PTInFlow", "PTOutFlow"]] # eliminate data sets not used for checking

        # read data into dataSets
        self.checkMassFlows = False      # TODO: check for possibilities of CPU-time optimisation depending on availability of data sets etc.
        

        # determine list of independent schedules to be used for fitting
        # - first use plant schedule
        Status.schedules.calculatePlantSchedule()
        scheduleList = [Status.schedules.plantSchedule.fav]
        self.scheduleListSrc = [(None, -1, None)]   # data set in the origin, type (-1: plant Schedule, 0: process, 1: data file), data set for hourly schedule base

        # - then add (user defined) process schedules
        for k, ccs in enumerate(self.ccsUPH):
            if ccs.schedule is None: continue

            newSchedule = ccs.schedule
            # test if linearly independent from the other ones
            test = np.array(scheduleList + [newSchedule])
            if matrix_rank(test, tol=1.e-4) > len(scheduleList):
                scheduleList.append(newSchedule)
                self.scheduleListSrc.append((k, 0, None))

        # - then add schedules from hourly data files
        dataSetsPeriods = []
        
        for df in dataFiles:
            pd = readDataFile(df.fileName)
            df.cols = pd.keys
            err = (pd.targets is None)  # set if read error from data file

            dataSets = self.dataSets[:]            
            for i, ds in enumerate(dataSets):
                if ds.parCode in ["Product", "RawMat"]:
                    self.checkMassFlows = True
                    
                if ds.df==df:
                    if err:
                        index = self.dataSets.index(ds)
                        self.dataSets.pop(index)
                    else:
                        ds.values = ds.scale*pd.targets[ds.col-1]
                        if pd.totals is not None: ds.total = ds.scale*pd.totals[ds.col-1]
                        ds.dataType = pd.dataType
                        if ds.dataType == 1: # period data
                            ds.periods = pd.periods
                            dataSetsPeriods.append(ds)

                        else:
                            newSchedule = makeSchedule(ds)

                            # test if linearly independent from the other ones
                            test = np.array(scheduleList + [newSchedule])
                            if matrix_rank(test, tol=1.e-4) > len(scheduleList):
                                scheduleList.append(newSchedule)
                                self.scheduleListSrc.append((ds, 1, None))

                        if DEBUG != "OFF" and len(scheduleList) > 0:
                            print "ScheduleChecker.init: data set %s hop = %s"%(ds, sum(scheduleList[-1]))


        # - and finally add schedules from period data files
        bsList = scheduleList[1:] if len(scheduleList) > 1 else scheduleList[:]     # use generic plant schedule only if no other schedules are available
        bsListSrc = self.scheduleListSrc[1:] if len(scheduleList) > 1 else self.scheduleListSrc[:]
##        bsList = scheduleList[:]      # TODO: would be desirable to include plant schedule always, but creates numeric convergence problems (TC-EMS-02t)
        
##        print "Len(BSList) = %s"%len(bsList)
##        print "Len(dataSetsPeriods) = %s"%len(dataSetsPeriods)
        
        for ds in dataSetsPeriods:
            for n, bs in enumerate(bsList):
                bsSrc = bsListSrc[n]
                
                newSchedule = makeSchedule(ds, baseSchedule=bs)
                
                # test if linearly independent from the other ones
                test = np.array(scheduleList + [newSchedule])
                if matrix_rank(test, tol=1.e-4) > len(scheduleList):
                    scheduleList.append(newSchedule)
                    self.scheduleListSrc.append((ds, 1, bsSrc))

        self.fit = []
        if self.checkMassFlows:
            self.calcMFNetWork()

        for k in self.autoFit:
            if len(self.xIndexes[k])==0 or not self.checkMassFlows:  # autoFit process has not yet been added to fit in calcMFNetWork
                self.fit.append(-k-1)

        self.nSchedules = len(scheduleList)
        self.scheduleCoefs = np.array([[1.0/(max(self.nSchedules, 1.0))]*self.nSchedules for index in self.fit])
        self.scheduleMatrix = np.array(scheduleList)

        for n, ds in enumerate(self.dsProd):
            i = self.xIndexesProd[n]
            if ds is not None and ds.dataType==0:
                val = np.sum(ds.values)
                self.ccsProduct[n].val = val
                if val > 0: self.ccsProduct[n].setSchedule(ds.values/val)
                
        if DEBUG != "OFF":
            print "Basic schedules for fitting:"
            for s, src in enumerate(self.scheduleListSrc):
                print "%s: src %s hop %s"%(s, src, np.sum(scheduleList[s]>0))

            if self.checkMassFlows:
                print "ccs table:"
                for i, ccs in enumerate(self.ccsX):
                    print "%3d: %s"%(i, ccs)

                for i, mf in enumerate(self.massFlows):
                    print "mf[%s]: %s"%(i, self.massFlows[i])
                
                print "xIndexes: %s"%self.xIndexes
                print "xIndexesProd: %s"%self.xIndexesProd
            print "fit %s"%self.fit
            print "autoFit %s"%self.autoFit
            

        
#============================================================================
# Main check function
#============================================================================

    def check(self):
        """ this check function is called within the iterative cycle of ModuleCC.basicCheck
        here the new values and boundaries should be imported and then the new estimates
        pushed and checked into the CCPar objects
        """

        if len(self.dataSets)==0:
            return
        
        self.vars = []
        self.xo = []
        self.xdesc = []  # description for debug prints
        # Step 1: import the values of independent parameters of schedule calculations
        for k, ccProc in enumerate(self.cc.ccProc):
            for key in ["UPH", "QHXProc"]:
                ccPar = getattr(ccProc, key)
                # set initial value of parameter

                val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

                if not(ccPar.sqerr <= 100*DEFAULT_SQERR or ccPar.userDefined):
                    self.vars.append(("ccs"+key, k))
                    self.xo.append(val)
                    self.xdesc.append(key)
                    
                ccs = getattr(self, "ccs"+key)
                ccs[k].val = val

                ccs[k].err = (ccPar.sqerr > 0.1*INFINITE)

            for key in ["UPHc", "UPHw"]:
                ccRow = getattr(ccProc, key)
                for i, ccPar in enumerate(ccRow):
                    val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

#TODO: check whether those parameters should be added to potential auto-fit mechanism ? 
##                    if not(ccPar.sqerr < EPSILON or ccPar.userDefined):
##                        self.vars.append(("ccs"+key+".%s"%i, k))
##                        self.xo.append(val)
##                        self.xdesc.append(key+".%s"%i)

                    ccs = getattr(self, "ccs"+key)
                    ccs[k][i].val = val
                    ccs[k][i].err = (ccPar.sqerr > 0.1*INFINITE)

            for key in ["mInFlowYear", "mOutFlowYear"]:
                ccRow = getattr(ccProc, key)
                for i, ccPar in enumerate(ccRow):
                    val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

                    if self.checkMassFlows:
                        if not(ccPar.sqerr <= 100*DEFAULT_SQERR or ccPar.userDefined):
                            self.vars.append(("ccs"+key+".%s"%i, k))
                            self.xo.append(val)
                            self.xdesc.append(key+".%s"%i)

                    if not self.checkMassFlows:
                    # TODO: this is no longer necessary (for massFlow-checking mode !, not clear whether necessary for the other case)
                        ccs = getattr(self, "ccs"+key)
                        ccs[k][i].val = val
                        ccs[k][i].err = (ccPar.sqerr > 0.1*INFINITE)
##                        print "CheckSchedules.check: %s[%s][%s]=%s; %s"%(key, k, i, val, ccs[k][i])

            if self.checkMassFlows:

                #TODO: SCALEFACTOR
                for n, ccnode in enumerate(self.ccnode):
                    for i, qin in enumerate(ccnode.qin):
                        index = self.baseQin[n]+i
                        self.ccsX[index].val = qin.val
                        self.ccsX[index].err = (qin.sqerr > 0.1*INFINITE)

                    for i, qout in enumerate(ccnode.qout):
                        index = self.baseQout[n]+i
                        self.ccsX[index].val = qout.val
                        self.ccsX[index].err = (qout.sqerr > 0.1*INFINITE)


        for m, ccPipe in enumerate(self.cc.ccPipe):
            for key in ["QHXPipe"]:
                ccPar = getattr(ccPipe, key)
                # set initial value of parameter
                val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

                if not(ccPar.sqerr < EPSILON or ccPar.userDefined):
                    self.vars.append(("ccs"+key, m))
                    self.xo.append(val)
                    self.xdesc.append(key)

                ccs = getattr(self, "ccs"+key)
                ccs[m].val = val

                ccs[m].err = (ccPar.sqerr > 0.1*INFINITE)

            key = "HDEffAvg"
            ccPar = getattr(ccPipe, key)
            # set initial value of parameter
            val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)
            self.HDEffAvg[m] = val
            self.HDEffAvgErr[m] = (ccPar.sqerr >= 1.0)
            
        for j, ccEq in enumerate(self.cc.ccEq):
            for key in ["QHXEq"]:
                ccPar = getattr(ccEq, key)
                # set initial value of parameter
                val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

                if not(ccPar.sqerr < EPSILON or ccPar.userDefined):
                    self.vars.append(("ccs"+key, j))
                    self.xo.append(val)
                    self.xdesc.append(key)

                ccs = getattr(self, "ccs"+key)
                ccs[j].val = val

                ccs[j].err = (ccPar.sqerr > 0.1*INFINITE)

            key = "HCGTEfficiency"
            ccPar = getattr(ccEq, key)
            # set initial value of parameter
            val = ccPar.val if ccPar.val is not None else 0.5*(ccPar.valMin+ccPar.valMax)

            eta = 1./val if val > 0 else 0.0
            if ccEq.mainSource=="Fuel":
                self.f_fuel[j] = eta
                self.f_el[j] = 0.0
            else:
                self.f_el[j] = eta
                self.f_fuel[j] = 0.0

            # TODO: some special cases are not yet correctly reflected here

        for k, ccsUPH in enumerate(self.ccsUPH):
            self.ccsUPH[k].updateValue()
            for q in self.ccsmInFlowYear[k] + self.ccsmOutFlowYear[k]:
                q.updateValue()
                    

        # append schedule coefficients to the estimate of the unknown

        if self.nSchedules > 0:
            for j, coefs in enumerate(self.scheduleCoefs):
                self.xo = np.concatenate((self.xo, coefs[:-1]))
                self.xdesc += ["coef(%s, %s)"%(j, i) for i in range(len(coefs)-1)]

        else:
            self.scheduleCoefs = np.array([[] for index in self.fit])    
            
        self.hasErr = False

        if len(self.xo) > 0:

            if DEBUG != "OFF":
                print "\n-----------------------------"
                print "SCHEDULES"
                print "-----------------------------"
                print "CheckSchedules: System variables subject to optimisation"
                for i, var in enumerate(self.vars):
                    print "%s [%s]"%(var, self.xo[i])
                print "initial estimate (xo): %s"%self.xo
            
            diag = [1.e-6 if i < len(self.vars) else 1.0 for i in range(len(self.xo))]
            output = leastsq(fv,
                             self.xo,
                             full_output=True,
                             maxfev=10*(len(self.xo)+1),
                             factor=1,
                             diag=diag,
                             ftol=1.0)  # ??? TODO - CHECK. This is a hack in order to achieve a more stable cov_f calculation
            self.x = output[0]

            cov_f = output[1]
            fvmin = output[2]["fvec"]
            fmin = np.sum(np.square(fvmin))
            msg = output[3]
            ier = output[4]

## DEVNOTES V3.0: calculation of dev based on return values from leastsq has been deactivated, as there were problems with the correct
## determination of the deviation in TC-EMS-02t.
## To be further analysed.
        
##        if cov_f is not None:
##            fo = min(fmin, 0.01)
##            print "TEST TEST CheckSchedules: calculation of devs: fo %s fmin %s msg %s ier %s"%(fo, fmin, msg, ier)
##            print [cov_f[i][i] for i in xrange(len(self.x))]
####            fo = fmin
##            sqdev = [2*fo*cov_f[i][i] for i in xrange(len(self.x))]
##            dev = [pow(val, 0.5) for val in sqdev]
##        else:
####            print "No information on accuracy of fit"
##            sqdev = None # cannot define accuracy of fit
##            dev = None

            res = pow(fmin, 0.5)

            cycle.initCheckBalance()
            cycle.checkBalanceSchedules(res)
            cycle.checkTotalBalance()

            n = len(self.x)
            der = np.zeros(n)
            dev = np.zeros(n)
            delta(self.x, der, dev, 0.001/np.array(diag))
            sqdev = np.square(dev)

            hess = hessian(f, self.x, np.array(diag))
            eigMin, sqdevH = test_hessian(hess)
##            eigMin, sqdevH = None, None if resTest is None else resTest
            eigMin = noneCheck(eigMin, -1.0)

            #TODO: sqdevH is currently not used, but should give a better approximation than the dev obtained from the delta function
            # -> to be expanded to length of x (best within function test_hessian) and to be divided by diag

            # write the x-values into the corresponding vectors
            if DEBUG != "OFF":
                print self.xdesc
                print "result of minimisation: %s"%self.x
                print "dev = %s"%dev
                print "sqdev = %s"%sqdev
                print "ier: %s msg: %s"%(ier, msg)
                print "der: %s"%der
                print "hessian: %s"%hess
                print "minimum eigen value: %s"%eigMin
                print "sqdevH = %s"%str(sqdevH)
                print "residual = %s\n"%res

            # store result for scheduleCoefs for the next iteration or reinitialise
            self.maxDevSC = 0
            if self.nSchedules > 0:    
                if res < 0.1:                                      
                    for i, coefs in enumerate(self.scheduleCoefs):
                        nvars = len(self.vars)
                        for j, index in enumerate(self.fit):
                            i0 = nvars + j*(self.nSchedules-1)
                            i1 = i0 + (self.nSchedules-1)
                            self.scheduleCoefs[j][:-1] = self.x[i0:i1]
                            self.scheduleCoefs[j][-1] = 1.0 - np.sum(self.x[i0:i1])

                            if i1 > i0:
                                try:        # exceptions may arise if dev = None or i0==i1
                                    maxDev = max(dev[i0:i1])
                                except:
                                    maxDev = 1.0
                                    
                                if DEBUG != "OFF":
                                    print "CheckSchedules: maxDev fit[%s] = %s"%(j, maxDev)
                                    
                                self.maxDevSC = max(self.maxDevSC, maxDev)

                if DEBUG != "OFF":
                    print "CheckSchedules: maxDevSC = %s"%self.maxDevSC

            ### TODO: find a robust criterion: limit of eigMin for TC-EMS-02 has to be lower than 2.e-5.    
            if sqdev is not None and not self.hasErr and res < 0.1 and eigMin > 1.e-6 and self.maxDevSC < 0.1:
                for i, var in enumerate(self.vars):
                    ccsKey, no = var
                    key = ccsKey[3:]

                    if key in ["UPH", "QHXProc"]:
                        if not np.isnan(dev[i]):
                            ccPar = getattr(self.cc.ccProc[no], key)

                            if dev[i] > abs(self.x[i]) and dev[i] > 1000.0: continue
                            
                            newPar = ccPar.copy()
                            newPar.val = self.x[i]
                            newPar.sqdev = max(sqdev[i] , pow(self.x[i], 2.0)*100*DEFAULT_SQERR)
                            newPar.calcErr()
                            newPar.constrain(updateLimits=True)

                            ccPar.push(newPar)
                            if DEBUG != "OFF":
                                print "Setting CCPar %s (Process %s) from schedule fit: "%(key, no+1),
                                ccPar.show()

                            ccPar.check()

                    


        # take over schedule globals for process schedules to be fitted
        for k in self.calcFromDS:
            globalData = globalsFromFav(self.ccsUPH[k].schedule)
            if DEBUG != "OFF":
                print "Global Schedule Data for Process %s: %s"%(k, globalData)

            ccProc = self.cc.ccProc[k]
            ccProc.HBatch.setValue(float(globalData["HBatch"]))
            ccProc.NBatch.setValue(float(globalData["NBatch"]))
            ccProc.NDaysProc.setValue(float(globalData["NDays"]))
            ccProc.NCyclesPerYear.setValue(float(globalData["NCycles"]))
            
        # check also the yearly totals if available
        self.checkTotals()

    def calculateScheduleGlobals(self):
        from ccPar import DEFAULT_SQERR

        schedulesSet = False

        if self.maxDevSC < 0.1:
            for k in self.autoFit:
                if k in self.calcFromDS or (self.ccsUPH[k].schedule is None):
                    continue

                schedulesSet = True
                
                globalData = globalsFromFav(self.ccsUPH[k].schedule, threshold=1.e-6)
                if DEBUG != "OFF":
                    print "CheckSchedules.calculateScheduleGlobals: Global Schedule Data for Process %s: %s"%(k, globalData)

                ccProc = self.cc.ccProc[k]

                ccProc.HPerYearProc.setEstimate(float(globalData["HPerYear"]), newSqErr=DEFAULT_SQERR, record=False)
                ccProc.NDaysProc.setEstimate(float(globalData["NDays"]), newSqErr=0.0, record=False)
                ccProc.NCyclesPerYear.setEstimate(float(globalData["NCycles"]), newSqErr=0.0, record=False)
                ccProc.HBatch.setEstimate(float(globalData["HBatch"]), newSqErr=DEFAULT_SQERR, record=False)
                ccProc.NBatch.setEstimate(float(globalData["NBatch"]), newSqErr=DEFAULT_SQERR, record=False)

                if ccProc.continuous:
                    ccProc.HPerYearInFlowNom.setEstimate(float(globalData["HPerYear"]), newSqErr=DEFAULT_SQERR, record=False)
                    ccProc.HPerYearOutFlowNom.setEstimate(float(globalData["HPerYear"]), newSqErr=DEFAULT_SQERR, record=False)
                else:
                    ccProc.HPerYearInFlowNom.setEstimate(0.2*float(globalData["HPerYear"]), newSqErr=DEFAULT_SQERR, record=False)
                    ccProc.HPerYearOutFlowNom.setEstimate(0.2*float(globalData["HPerYear"]), newSqErr=DEFAULT_SQERR, record=False)

        return schedulesSet

    def exportData(self):
        """ exports process schedules if applies """

        for k in self.autoFit:

            sources = []

            processes = Status.prj.getProcesses(PId = Status.PId, ANo = 0)
            proc = processes[k]
            procId = proc.id

            oldCombSchedules = dbio.select("qcombSchedules", ProcessId=procId)
            oldSources = [(entry.id, entry.ObjectId, entry.Type, entry.ComponentNo) for entry in oldCombSchedules]


            # get the sources for indirectly calculated processes and get the coefficients for directly fitted schedules
            if self.checkMassFlows and len(self.xIndexes[k]) > 0:
                i = self.xIndexes[k][0]
                coefs = np.zeros(len(self.scheduleListSrc))
            
                mf = self.massFlows[i]
                for elem in mf:
                    isrc = elem[1]
                    w = elem[0]

                    val = self.ccsX[isrc].val
                    if val is None: val = 0.0

                    srcType = -1
                    if isrc in self.fit:
                        
                        j = self.fit.index(isrc)
                        coefs += w * val * self.scheduleCoefs[j]

                    else:
                        for ksrc, indexes in enumerate(self.xIndexes):
                            if isrc in indexes:
                                srcProc = processes[ksrc]
                                objectId = srcProc.id
                                srcType = 0
                                break

                        if isrc in self.xIndexesProd:
                            prodNo = self.xIndexesProd.index(isrc)
                            ds = self.dsProd[prodNo]
                            objectId = ds.id
                            srcType = 1

                        if srcType in [0, 1]:
                            sources.append((w*val, objectId, srcType, None))
            else:
                val = self.ccsUPH[k].val
                if -k-1 in self.fit:
                    j = self.fit.index(-k-1)
##                    coefs = val * self.scheduleCoefs[j]
                    coefs = self.scheduleCoefs[j]
                else: 
                    assert(False) # this case should not occur - indirect calculation is only possible via mass flows
                    sources.append((val, procId, 0, None))


            for j, w in enumerate(coefs):
                source, srcType, bsSrc = self.scheduleListSrc[j]
                if srcType == 0:   # process
                    srcProc = processes[source]
                    objectId = srcProc.id
                elif srcType == 1: # data set
                    ds = source
                    objectId = ds.id
                elif srcType == -1: # plant schedule
                    objectId = None

                if bsSrc is not None:
                    assert(srcType==1) # base schedule is applied only to period data from data files
                    source, bsSrcType, dummy = bsSrc
                    if bsSrcType == 0:   # process
                        srcProc = processes[source]
                        bsSrcObjectId = srcProc.id
                    elif bsSrcType == 1: # data set
                        bsSrcObjectId = source.id
                    elif bsSrcType == -1: # plant schedule
                        bsSrcObjectId = None
                    
                else:
                    bsSrcType = -1
                    bsSrcObjectId = None

                if bsSrcType > 0: srcType += 10*bsSrcType  # hack in order to compact for storage
                    
                if abs(w) > EPSILON: sources.append((w, objectId, srcType, bsSrcObjectId))


            sumW = sum([elem[0] for elem in sources])
            sumWSignificant = sum([elem[0] for elem in sources if abs(elem[0]) > MINSCHEDULEWEIGHT*sumW])
            sumW = sumWSignificant
            if abs(sumW) < EPSILON: sumW = 1.0  # should not occur, but anyway avoid zero division later on

            for w, objectId, srcType, bsSrcObjectId in sources:
                if abs(w) > MINSCHEDULEWEIGHT*sumW:
                
                    combSchedule = dbio.get("qcombschedules", ProcessId=procId, ObjectId=objectId)
                    if combSchedule is None:
                        newId = dbio.insertRow("qcombschedules")
                        combSchedule = dbio.get("qcombschedules", id=newId)

                    tmp = {"ProcessId": procId,
                           "Weight": float(w / sumW),
                           "Type": srcType,
                           "ObjectId": objectId,
                           "ComponentNo": bsSrcObjectId}
                           
                    combSchedule.set(**tmp)
                    
                else:
                    dbio.delete("qcombschedules", ProcessId = procId, ObjectId=objectId, Type=srcType, ComponentNo=bsSrcObjectId)

            newSources = [(src[1], src[2], src[3]) for src in sources]
            for oldId, oldObjectId, oldSrcType, oldBsSrcObjectId in oldSources:
                if (oldObjectId, oldSrcType, oldBsSrcObjectId) not in newSources:
                    dbio.delete("qcombschedules", id=oldId)

        dbio.commit()
        
        
#============================================================================
# Methods for calculation of mass flow network
#============================================================================

    def calcMFNetWork(self):

        # build vector x of nodes
        self.ccnode = self.cc.ccMassFlow.ccnode      # use the ccnodes already existing from CheckMassFlows
        self.baseQin = [None]*len(self.ccnode)       # base index of input in x-vector
        self.baseQout = [None]*len(self.ccnode)

        self.xdesc = []
        nPtr = 0
        
        # mapping of ccnodes (node index n) into ccsX vector (index nPtr)
        for n, ccnode in enumerate(self.ccnode):
            self.baseQin[n] = nPtr
            nPtr += len(ccnode.qin)
            for i in xrange(len(ccnode.qin)):
                self.xdesc.append("%s-%s.qin[%s]"%(ccnode.origin.objectType, ccnode.origin.objectId, i+1))
            
            self.baseQout[n] = nPtr
            nPtr += len(ccnode.qout)

            for i in xrange(len(ccnode.qout)):
                self.xdesc.append("%s-%s.qout[%s]"%(ccnode.origin.objectType, ccnode.origin.objectId, i+1))

        self.nx = nPtr
        self.eqA = []
        self.eqb = []
        self.fit = []

        self.ccsX = [None]*self.nx

##        for i in xrange(self.nx):
##            print "x[%s]: %s"%(i, self.xdesc[i])
        
        # build node equations
        for n, ccnode in enumerate(self.ccnode):
            objectType = ccnode.origin.objectType
            
            # balance equation: sum(qin) = sum(qout)
            if objectType in ["INT"]:
                nodeType = ccnode.origin.nodeType
                eqA = np.zeros(self.nx)
                eqA[self.baseQin[n]:self.baseQin[n]+len(ccnode.qin)] = 1.0

                #NOTE: SCALEFACTOR not necessary for ccs - checking !
##                if nodeType == "MIX":
##                    eqA[self.baseQin[n]:self.baseQin[n]+len(ccnode.qin)] *= ccnode.origin.scaleFactor
                
                eqA[self.baseQout[n]:self.baseQout[n]+len(ccnode.qout)] = -1.0

                self.eqA.append(eqA)
                self.eqb.append([])

                for i in xrange(len(ccnode.qin)):
                    self.ccsX[self.baseQin[n]+i] = CCSchedule(ccnode.qin[i])   # link ccsX with the corresponding in/outflow-schedule
                for i in xrange(len(ccnode.qout)):
                    self.ccsX[self.baseQout[n]+i] = CCSchedule(ccnode.qout[i])
                
                
            # definition equation for known process schedules and type-t (hourly) data files
            if objectType in ["PROC"]:
                k = ccnode.origin.k
                if k not in self.autoFit:
                    for i, q in enumerate(ccnode.qin):
                        eqA = np.zeros(self.nx)
                        eqA[self.baseQin[n]+i] = 1.0
                        self.eqA.append(eqA)
                        self.eqb.append([[1.0, self.baseQin[n]+i]])
                            
                    for i, q in enumerate(ccnode.qout):
                        eqA = np.zeros(self.nx)
                        eqA[self.baseQout[n]+i] = 1.0
                        self.eqA.append(eqA)
                        self.eqb.append([[1.0, self.baseQout[n]+i]])

                self.xIndexes[k] = range(self.baseQin[n], self.baseQout[n]+len(ccnode.qout))

                for i in xrange(len(ccnode.qin)):
                    self.ccsX[self.baseQin[n]+i] = self.ccsmInFlowYear[k][i]   # link ccsX with the corresponding in/outflow-schedule
                for i in xrange(len(ccnode.qout)):
                    self.ccsX[self.baseQout[n]+i] = self.ccsmOutFlowYear[k][i]   
                        
            elif objectType == "PROD":
                objectId = ccnode.origin.objectId
                objectNames, objectIds, dummy = objectCodes("Product")
                if objectId in objectIds:
                    objectNo = objectIds.index(objectId)
                else:
                    objectNo = None

                for ds in self.dataSets:
                    if ds.parCode != "Product" or ds.objectNo != objectNo:
                        continue
                    
                    else:
                        if ds.dataType == 0: # hourly data
                            eqA = np.zeros(self.nx)
                            eqA[self.baseQin[n]] = 1.0
                            self.eqA.append(eqA)
                            self.eqb.append([[1.0, self.baseQin[n]]])

                        self.dsProd[objectNo] = ds

                index = self.baseQin[n]
                self.xIndexesProd[objectNo] = index
                self.ccsX[index] = self.ccsProduct[objectNo]

                                        
                
        # build connection equations
        for flow in self.cc.ccMassFlow.sysMF.flow:
            fromNo = self.cc.ccMassFlow.sysMF.node.index(flow.fromNode)
            fromCon = flow.fromCon
            toNo = self.cc.ccMassFlow.sysMF.node.index(flow.toNode)
            toCon = flow.toCon

            eqA = np.zeros(self.nx)
            eqA[self.baseQout[fromNo]+fromCon-1] = 1.0
            eqA[self.baseQin[toNo]+toCon-1] = -1.0
            self.eqA.append(eqA)
            self.eqb.append([])

        
        self.eqA = np.array(self.eqA)

        self.massFlows = self.solver_LU()

        # filter the required fits
        requiredFits = []
        # - 1. those required for autoFit of process UPHs
        requiredMF = []
        for k in self.autoFit:
            if len(self.xIndexes[k]) > 0:
                i = self.xIndexes[k][0]
                requiredMF += self.massFlows[i]

        # - 1. those required for fitting product data
        for n, ds in enumerate(self.dsProd):                
            i = self.xIndexesProd[n]
            if ds is not None:
                requiredMF += self.massFlows[i]
            
        for elem in requiredMF:
            j = elem[1]
            f = elem[0]
            if abs(f) > EPSILON:

                isDefined = False

                for k, indexes in enumerate(self.xIndexes):
                    if j in indexes and k not in self.autoFit:
                        isDefined = True
                        break

                for n, ds in enumerate(self.dsProd):
                    i = self.xIndexesProd[n]
                    if i==j and ds.dataType==0:
                        isDefined = True
                        break
                    
                if not isDefined: requiredFits.append(j)
                
        self.fit = list(set(requiredFits))

        self.calcFromDS = []       
        for k in self.autoFit:
            if len(self.xIndexes[k]) > 0:
                i = self.xIndexes[k][0]
                requiredMF = self.massFlows[i]

                needsFit = False
                for elem in requiredMF:
                    j = elem[1]
                    f = elem[0]
                    if j in self.fit:
                        needsFit = True
                        break

                if not needsFit:
                    self.calcFromDS.append(k)
                    
##        print "really required fits: ", self.fit

    def compress(self, x):
        """ auxiliary function for calcMFNetWork """

        tmp = np.zeros(self.nx)
        
        for elem in x:
            val = elem[0]
            i = elem[1]
            tmp[i] += val

        x = []
        for i in xrange(self.nx):
            if abs(tmp[i]) > EPSILON:
                x.append([tmp[i], i])

        return x
        
    def solver_LU(self):
        """ equation solver for calcMFNetWork """
        
        self.nEqs = len(self.eqb)
        
        P, L, U = lu(self.eqA)

        # check U and extend matrix until U is no longer underdetermined
        indexes = np.ones(self.nx)
        for i, row in enumerate(U):
            if abs(row[i]) > EPSILON:
                indexes[i] = 0.0

        if np.sum(indexes) > 0: # extend if necessary
##            print "U before extension: %s x %s\n"%U.shape, U
##            print "indexes %s"%indexes
            
            for i in xrange(self.nx):
                if indexes[i] > 0:
                    eqA = np.zeros(self.nx)
                    eqA[i] = 1.0
                    self.eqA = np.append(self.eqA, [eqA], axis=0)
                    self.eqb.append([[1.0, i]])
                    self.fit.append(i)

            P, L, U = lu(self.eqA)
##            print "Diimensions of equation system after completing: A %s b %s"%(self.eqA.shape, len(self.eqb))
            
        self.nEqs = len(self.eqb)
        
        if DEBUG == "ALL":
            print "Dimensions of equation system: A %s b %s"%(self.eqA.shape, len(self.eqb))
            print "A: %s"%self.eqA
            print "b: ", self.eqb
            
            print "P: %s x %s\n"%P.shape, P
            print "L: %s x %s\n"%L.shape, L
            print "U: %s x %s\n"%U.shape, U

        # solve L*y = Pb for y
        # - calculate Pb
        Pb = [None]* self.nEqs

        indexes = np.linspace(0, self.nEqs-1, self.nEqs)
        indexesT = np.dot(indexes, P)   # note: the return of the lu function gives P as P*L*U*x = b        
        Pb = [self.compress(self.eqb[int(i)][:]) for i in indexesT]
##        print "Pb: ", Pb

        # - iteratively calculate y
        #   y_j = (Pb_j - sum(k<j) L_jk*y_k) / L_jj
        y = [None]*self.nx
        for j, row in enumerate(L):
            if j >= self.nx:   # ignore last rows of the L-matrix (redundant equations)
                break
            y[j] = Pb[j][:]
            for elem in y[j]:
                elem[0] /= row[j]

            for k in xrange(j):
                for elem in y[k]:
                    if row[k]<> 0:
                        val = elem[:]
                        val[0] *= - row[k]/row[j]
                        y[j].append(val)
                
        y = [self.compress(val) for val in y]
##        print "y: ", y

        # - iteratively calculate x
        #   x_j = (y_j - sum(k>j) U_jk*x_k) / U_jj
        x = [None]*self.nx
        for i in xrange(self.nx-1, -1, -1):
            row = U[i]
            x[i] = y[i][:]
            for elem in x[i]:
                elem[0] /= row[i]

            for k in xrange(i+1, self.nx):
                for elem in x[k]:
                    val = elem[:]
                    if row[k] <> 0:
                        val[0] *= - row[k]/row[i]
                        x[i].append(val)
                
        x = [self.compress(val) for val in x]
##        print "x: ", x
##        print "fit: ", self.fit

        requiredFits = []
        for xi in x:
            for elem in xi:
                index = elem[1]
                if index in self.fit:
                    requiredFits.append(index)
        self.fit = list(set(requiredFits))

        if DEBUG != "OFF":
            print "required fits: ", self.fit
        
        return x
                
#============================================================================
# Methods for parameter fitting
#============================================================================

    def residual(self, x):
        """ function to be used in optimisation routine
        calculates the actual values for hourly energy flows in the system
        (hourly energy balances) and the residual comparing with targets
        """

##        print "\nResidual starting ==========================="
        NK = len(self.cc.ccProc)
        NM = len(self.cc.ccPipe)
        NJ = len(self.cc.ccEq)
        NI = len(self.cc.ccFET)
        
##        print "residual: vars = %s"%self.vars
        
        # write the x-values into the corresponding vectors
        self.hasMassFlow = False
        
        for i, var in enumerate(self.vars):
            keyString, no = var

            keyElements = keyString.split(".")
            key = keyElements[0]

            if key.startswith("ccsm"):
                self.hasMassFlow = True

            par = getattr(self, key)

            if len(keyElements) > 1:
                s = int(keyElements[1])
                par[no][s].val = x[i]
                par[no][s].changed = True
            else:
                par[no].val = x[i]
                par[no].changed = True

        nvars = len(self.vars)
        for j, index in enumerate(self.fit):
            i0 = nvars + j*(self.nSchedules-1)
            i1 = i0 + (self.nSchedules-1)
            self.scheduleCoefs[j][:-1] = x[i0:i1]
            self.scheduleCoefs[j][-1] = 1.0 - np.sum(x[i0:i1])

        resStack = []

        # step 0: for processes with auto-fit of schedules

        # - calculate fit schedules from fit coefficients
        for j, i in enumerate(self.fit):
            weights = self.scheduleCoefs[j]
            if np.sum(weights) == 0.0:
                newSchedule = np.zeros(8760)
            else:
                newSchedule = np.average(self.scheduleMatrix, 0, weights=weights)

            newSchedule -= min(np.min(newSchedule), 0.0)    # correct so that all entries are > 0
            
            if i >= 0:
                self.ccsX[i].schedule = newSchedule
                self.ccsX[i].c = 1.0
            else: # TODO. Check if this is ok here. Way to calculate operation 
                k=-i-1
                self.ccsUPH[k].schedule = newSchedule
                self.ccsUPH[k].c = 1.0

        # - calculate dependent process schedules
        for k in self.autoFit:
            if self.checkMassFlows:
                for i in self.xIndexes[k]:
                    mf = self.massFlows[i]
                    
                    f = np.zeros(self.nx)
                    for elem in mf:
                        j = elem[1]
                        f[j] = elem[0]

                    if f[i] == 0:
                        scheduleRowSumC(self.ccsX[i], f, self.ccsX)
                    else:
                        assert(abs(np.sum(f) - f[i]) < EPSILON) # build assignment to itself should be 1:1 without any other components

            if len(self.xIndexes[k]) > 0: #
                i = self.xIndexes[k][0]
                self.ccsUPH[k].schedule = self.ccsX[i].schedule
                self.ccsUPH[k].c = self.ccsX[i].c
                
        if len(self.autoFit) > 0: self.hasMassFlow = True
        
        # step 1: calculation of schedules of energy balances
        
        for k in xrange(NK):
            # TODO - time saver: recalculate only if changed
            scheduleDiff(self.ccsUPHProc_k[k],
                         self.ccsUPH[k],
                         self.ccsQHXProc[k])

        for m in xrange(NM):
            scheduleRowSum(self.ccsUPHProc_m[m], [self.cc.UPHCMatrix.FRow[k][m].val for k in xrange(NK)], self.ccsUPHProc_k)
            fDist = 1./self.HDEffAvg[m] if self.HDEffAvg[m] > 0 else INFINITE
            scheduleProd(self.ccsUSHPipe[m], fDist, self.ccsUPHProc_m[m])

            if self.HDEffAvgErr[m]: self.ccsUSHPipe[m].err = True
            scheduleDiff(self.ccsUSHm[m],
                         self.ccsUSHPipe[m],
                         self.ccsQHXPipe[m])

        for j in xrange(NJ):
            scheduleRowSum(self.ccsUSHj[j], [self.cc.USHMatrix.FRow[m][j].val for m in xrange(NM)], self.ccsUSHm)
            scheduleDiff(self.ccsUSHConv[j],
                         self.ccsUSHj[j],
                         self.ccsQHXEq[j])
            
            scheduleProd(self.ccsFETFuel_j[j], self.f_fuel[j], self.ccsUSHConv[j])
            scheduleProd(self.ccsFETel_j[j], self.f_el[j], self.ccsUSHConv[j])
            
        for i in xrange(NI):
            scheduleRowSum(self.ccsFETFuel_i[i], [self.cc.FETFuelMatrix.FRow[j][i].val for j in xrange(NJ)], self.ccsFETFuel_j)

        # step 2: calculation of schedules of products (mass flows)
        if self.checkMassFlows:
            for n, ds in enumerate(self.dsProd):                
                i = self.xIndexesProd[n]
                if ds is None: continue # calculate only those data which can be compared with data
                
                mf = self.massFlows[i]
                
                f = np.zeros(self.nx)
                for elem in mf:
                    j = elem[1]
                    f[j] = elem[0]

                scheduleRowSumC(self.ccsX[i], f, self.ccsX)
                
        return self.checkTargets()

    def checkTargets(self):
        """ auxiliary function for repetitive blocks in function residual
        """
        hasErr = False
        resStack = []

        nDataSets = 0
        for ds in self.dataSets:#
            if not self.hasMassFlow and ds.parCode.startswith("P"): # do not check mass flow data, if theres no appropriate fit parameter
                continue

            if ds.parCode=="Control": # currently not checked. only for rt-sim
                continue

            parCode = ds.parCode
            if parCode in ["UPHc", "UPHw", "mInFlowYear", "mOutFlowYear"]:
                ccs = self.ccsUPH[ds.objectNo]
                ccs_totals = getattr(self, "ccs"+parCode)[ds.objectNo][ds.cNo]
                parVal = ccs_totals.val
            else:
                ccs = getattr(self, "ccs"+parCode)[ds.objectNo]
                parVal = ccs.val

            target = ds.values

            actual = parVal*ccs.schedule
            hasErr = hasErr or ccs.err

            if DEBUG == "ALL":
                print "CheckTargets: parCode %s[%s] val %s s %s|%s"%(ds.parCode, ds.objectNo, parVal, sum(target>EPSILON), sum(ccs.schedule>EPSILON))

            if ds.dataType==1: # data periods
                actual_byP = np.array([np.sum(actual[start:stop]) for (start, stop) in ds.periods])
                res = actual_byP - target

            else:
                res = actual - target

            sigma = ((1. - ccs.c)*parVal)/target.shape[0]

            if np.sum(np.isnan(target))==0: # full year data
                avgDiff = (1. - ccs.c)*np.average(parVal - target)
            else:
                avgDiff = (1. - ccs.c)*min(np.average(parVal - np.nan_to_num(target)), 0.0)
            
            if ccs.c <> 0:
                res = ccs.c*res
                res += np.sign(res)*(abs(sigma)+abs(avgDiff))
            else:
                res = 0*res + abs(sigma) + abs(avgDiff)
                
            res = np.nan_to_num(res) # -> converts NaNs to 0 - used for the cases where target is None

            # relative error for testing purposes
            # normalize residual for matching
            # - scale with respect to targetAvg (or targetMaxAbs)
            targetAvg = nanmean(target)
            ds.relErr = pow(np.average(np.square(res)), 0.5)

            if abs(targetAvg) > EPSILON:
                nres = res / targetAvg
                ds.relErr /= targetAvg
                
            else:
                targetMaxAbs = np.max(np.abs(target))
                if targetMaxAbs > EPSILON:
                    nres = res / targetMaxAbs
                    ds.relErr /= targetMaxAbs
                else:
                    nres = res
                    
            # - scale period data to equivalent size as hourly data
            if len(nres) > 0:
                nres /= pow(len(nres), 0.5) # sqrt -> gives the same sum of squares !! 

            resStack.append(nres)
            nDataSets += 1

            if DEBUG == "ALL":
                if ds.dataType==1:
                    print "CheckSchedules.checkTargets: err %s actual %s c %s target %s res %s"%(ds.relErr, actual_byP, ccs.c, target, res)
                else:
                    print "CheckSchedules.checkTargets: err %s actual (max) %s c %s target %s res %s"%(ds.relErr, max(actual), ccs.c, max(target), max(np.abs(res)))
        
        self.hasErr = hasErr
        resGlb = np.concatenate(resStack) if len(resStack) > 0 else None
        resGlb /= pow(max(nDataSets, 1), 0.5) # normalize residual

        if DEBUG == "ALL":
            print "CCScheduleChecker.checkTargets: %s"%pow(np.sum(np.square(resGlb)), 0.5)

        return resGlb

    def checkTotals(self):
        """ carries out a standard ccheck for those target data sets where totals are given
        """
        if DEBUG != "OFF":
            print "-----------------------------"
            print "CCScheduleChecker.checkTotals"
            print "-----------------------------"

        for ds in self.dataSets:
            if ds.total is not None:
                if ds.parCode in ["UPHc", "UPHw", "mInFlowYear", "mOutFlowYear"]:
                    ccs = getattr(self, "ccs"+ds.parCode)[ds.objectNo][ds.cNo]
                else:
                    ccs = getattr(self, "ccs"+ds.parCode)[ds.objectNo]
                ccPar = ccs.cc
                newPar = ccPar.copy()
                newPar.setValue(ds.total)
                ccPar.push(newPar)
                if DEBUG != "OFF":
                    print "ScheduleChecker.checkTotals: assigning parameter from data set totals"
                    ccPar.show()
                ccPar.check()

            


    def screen(self):
            
        for ds in self.dataSets:
            if ds.relErr > 0.3:
                logWarning("Schedule checking: relative Error for data set %s[%s] > 30 %%: %.2f %%"%(ds.parCode, ds.objectNo, 100*ds.relErr))
            elif ds.relErr > 0.1:
                logWarning("Schedule checking: relative Error for data set %s[%s] > 10 %%: %.2f %%"%(ds.parCode, ds.objectNo, 100*ds.relErr))


        if self.maxDevSC > 0.1:
            logWarning("Schedule checking: error in schedule coefficients > 10 %%: %.2f %%"%(100*self.maxDevSC))
            



#================================================================================

def f(x):
    return np.sum(np.square(Status.mod.moduleCC.sc.residual(x)))

def fv(x):
    return Status.mod.moduleCC.sc.residual(x)


def delta(xo, der, dev, dx):
    
    fo = f(xo)
    for i in range(len(xo)):
        dxi = max(dx[i], 0.001*abs(xo[i]))
        x = 1.0*xo
        x[i] = xo[i] + dxi
        fp = f(x)
        x[i] = xo[i] - dxi
        fn = f(x)

        fo_constrained = max(min(fo, 0.01), 1.e-5)      # the minimum is a hack in order to assure reasonable calculation of errors also for the case where some data set does not converge
                                                        # the maximum is for considering some random noise / uncertainty in case of very exact matching

        der[i] = (fp - fn)/(2*dxi)
        cur = (fp + fn - 2*fo)/pow(dxi, 2)

        ### error definition: 0.5*cur*err2 = fo
        #   -> err2 = fo/0.5*cur

        if cur > 0 and fp > fo + EPSILON and fn > fo + EPSILON:
            sqdev = fo_constrained/(0.5*cur)
            dev[i] = pow(sqdev, 0.5)
##            print "DELTA(%s): fo %s cur %s sqdev(unc.) %s dev (constr.) %s"%(i, fo, cur, fo/(0.5*cur), dev[i])
        else:
            dev[i] = INFINITE
    


