# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	CCheckTotals (Consistency Check)
#			
#------------------------------------------------------------------------------
#	
#	Functions for consistency checking of data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from ccheckFunctions import *
from ccPar import CCPar, CCRow
from GUI.GUITools import check
from GUI.status import Status
from constants import DEBUG
from modules.messageLogger import logDebug
import database.dbio as dbio

MAXBALANCEERROR = 1.e-3
NMAXITERATIONS = 5
INFINITE = 1.e99    # numerical value assigned to "infinite"

class CheckTotals():
    """Carry out consistency checking for the totals of FEC/FET/PEC/PET/USH/UPH/USC/UPC."""
    def __init__(self, FECi, FETi, USHj, UPH, UPHwReq, WaterConsum, OMh, OMj,
                 checkInPlace=False, onlyProcessTotals=False):    
        """Setup the check with the data for process nr. k
        
        :param checkInPlace: If False take the data from ANo -1 and
        write it to ANo 0, otherwise work on the current alternative proposal. 
        :param onlyProcessTotals: If True only check the totals for UPH and UPC.
        
        """        
        self.checkInPlace = checkInPlace
        self.onlyProcessTotals = onlyProcessTotals
        self.dataGroup = "Totals"
        
        # assign a variable to all intermediate values needed
        if self.onlyProcessTotals is False:
            self.NFuels = len(FETi) - 1
            self.NEquipe = len(USHj)
        
        self.NThProc = len(UPH)
        
        self.FETi = FETi
        self.FECi = FECi
        self.USHj = USHj
        self.UPHk = UPH
        self.UPHwReqk = UPHwReq
        self.WaterConsum_j = WaterConsum
        self.OMh = OMh
        self.OMj = OMj
        
        # USHj and UPHk contain heating and cooling pipes therefore we have
        # to split the indices 
        if self.onlyProcessTotals is False:
            equipment = Status.int.systemModel.eqObject
            self.heatingUSHjIds = [eq.no for eq in equipment if eq.heating]
            self.coolingUSHjIds = [eq.no for eq in equipment if not eq.heating]
        
        processes = Status.prj.getProcesses()
        allUPHkIds = range(len(self.UPHk))
        self.heatingUPHkIds = [id for id in allUPHkIds
                               if processes[id].heating_cooling == "heating"]
        self.coolingUPHkIds = [id for id in allUPHkIds
                               if processes[id].heating_cooling == "cooling"]

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            if self.onlyProcessTotals is False:
                print "checkTotals: USHjIds - heating %s - cooling %s" % (self.heatingUSHjIds, self.coolingUSHjIds)
        
        self.FEC = CCPar("FEC", priority=2, unitType = "ENERGY")
        self.FEO = CCPar("FEO", priority=2, unitType = "ENERGY")
        self.FET = CCPar("FET", priority=1, unitType = "ENERGY")
        self.USH = CCPar("USH", priority=1, unitType = "ENERGY")
        self.UPH = CCPar("UPH", priority=1, unitType = "ENERGY")
        self.USC = CCPar("USC", priority=1, unitType = "ENERGY")
        self.UPC = CCPar("UPC", priority=1, unitType = "ENERGY")
        self.WaterConsum = CCPar("WaterConsum", unitType = "VOLUME")
        self.OMhTotal = CCPar("OMhTotal", unitType = "PRICE")
        self.OMjTotal = CCPar("OMjTotal", unitType = "PRICE")
        self.OMTotal = CCPar("OMTotal", unitType = "PRICE")

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAll()
        
    def exportData(self):  
        """Store corrected data in SQL (under AlternativeProposalNo = 0)."""
        if self.checkInPlace is True:
            ANo = Status.ANo
        else:
            ANo = 0
##        cgeneraldataTable = Status.DB.cgeneraldata.Questionnaire_id[Status.PId].AlternativeProposalNo[ANo]
##        if len(cgeneraldataTable) > 0:
##            logDebug("exporting data to cgeneraldata")
##            cgeneraldata = cgeneraldataTable[0]
        cgeneraldata = Status.prj.getGeneralData(PId=Status.PId, ANo=ANo)
        if cgeneraldata is not None:
            if self.onlyProcessTotals is False:
                cgeneraldata.set(FET = dbio.check(self.FET.val),
                                 FEO = dbio.check(self.FEO.val),
                                 FEC = dbio.check(self.FEC.val),
                                 USH = dbio.check(self.USH.val),
                                 USC = dbio.check(self.USC.val),
                                 WaterConsum = dbio.check(self.WaterConsum.val),
                                 UPH = dbio.check(self.UPH.val),
                                 UPC = dbio.check(self.UPC.val))
            dbio.commit()
        else:
            logDebug("CheckTotals (exportData): error writing data to cgeneraldata")
        
    def showAll(self):
        """Print all used and generated variables just for debugging."""
        print "====================="
        for item in (self.FEC, self.FET, self.FEO, self.WaterConsum,
                     self.USH, self.USC, self.UPH, self.UPC, self.OMjTotal, self.OMhTotal, self.OMTotal):
            item.show()                                 
        print "====================="

    def screen(self):  
        """Screen all variables in the block"""
        if self.onlyProcessTotals is False:
            self.FET.screen()
            self.FEC.screen()
            self.FEO.screen()
            self.USH.screen()
            self.USC.screen()
        self.UPH.screen()
        self.UPC.screen()

    def check(self):     
        """Carrying out the check of the block"""
        if DEBUG in ["ALL", "MAIN", "BASIC"]:
            print "-------------------------------------------------"
            print " Checking Totals"
            print "-------------------------------------------------"

        for n in range(1):  #while no adjust, 1 cycle is enough !!!!

            cycle.initCheckBalance()

            if DEBUG in ["ALL", "MAIN"]:
                print "-------------------------------------------------"
                print "Ciclo %s" % n
                print "-------------------------------------------------"
        
            # Step 1: Call all calculation routines in a given sequence
            if DEBUG in ["ALL","MAIN"]:
                print "Step 1: calculating from left to right (CALC)"
            UPHk = [self.UPHk[i] for i in self.heatingUPHkIds]
            UPCk = [self.UPHk[i] for i in self.coolingUPHkIds]+[self.UPHwReqk[i] for i in self.heatingUPHkIds]
            self.UPH.push(calcRowSum("UPH", UPHk))
            self.UPC.push(calcRowSum("UPC", UPCk))
            
            if self.onlyProcessTotals is False:
                self.FEC.push(calcRowSum("FEC", self.FECi))
                self.FET.push(calcRowSum("FET", self.FETi))
                self.FEO = calcDiff("FEO", self.FEC, self.FET) #TODO:  FEO not pushed and checked, gave problems for convergency
                self.WaterConsum.push(calcRowSum("WaterConsum",self.WaterConsum_j))
                USHj = [self.USHj[i] for i in self.heatingUSHjIds]
                USCj = [self.USHj[i] for i in self.coolingUSHjIds]
                self.USH.push(calcRowSum("USH", USHj))
                self.USC.push(calcRowSum("USC", USCj))

                self.OMhTotal.push(calcRowSum("OMhTotal", self.OMh))
                self.OMjTotal.push(calcRowSum("OMjTotal", self.OMj))
                self.OMTotal.push(calcSum("OMTotal", self.OMhTotal, self.OMjTotal))
                
            # Step 2: Cross check the variables
            if DEBUG in ["ALL","MAIN","BASIC"]:
                self.showAll()
                print "Step 2: cross checking"

            self.ccheckAll()
            # Step 3: Adjust the variables (inverse of calculation routines)
            if DEBUG in ["ALL"]:
                self.showAll()
                print "Step 3: calculating from right to left (ADJUST)"

            if self.onlyProcessTotals is False:
                adjRowSum_stack(self.OMhTotal, self.OMh)
                adjRowSum_stack(self.OMjTotal, self.OMj)
                adjustSum_stack(self.OMTotal, self.OMhTotal, self.OMjTotal)


            # Step 4: second cross check the variables
            if DEBUG in ["ALL"]:
                self.showAll()
                print "Step 4: second cross checking"

            if self.onlyProcessTotals is False:
                self.ccheckAll()    #not necessary while adjust is not implemented

            # End of the cycle. Last print in DEBUG mode
            if DEBUG in ["ALL"]:
                self.showAll()

        if DEBUG in ["ALL", "BASIC", "MAIN"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()
            
    def ccheckAll(self):     
        """Call the consistency checks for all recalculated values."""
        if self.onlyProcessTotals is False:
                self.FET.check()
##                self.FEO.check()
                self.FEC.check()
                self.WaterConsum.check()
                self.USH.check()
                self.USC.check()
                self.OMjTotal.check()
                self.OMhTotal.check()
                self.OMTotal.check()
        self.UPH.check()
        self.UPC.check()

if __name__ == "__main__":
    NI = 2
    NJ = 3
    
    FECi = CCRow("FECi", NI)
    FECi[0].val = 2
    FECi[0].sqerr = 0.
    FECi[0].valMin = 1.9
    FECi[0].valMax = 2.1
    
    FECi[1].val = 4
    FECi[1].sqerr = 0.
    FECi[1].valMin = 3.9
    FECi[1].valMax = 4.1

    FECj = CCRow("FECj", NJ)
    FECj[0].val = 3
    FECj[0].valMin = 2.9
    FECj[0].valMax = 3.1
    FECj[0].sqerr = 0.001
    FECj[1].val = 2.5
    FECj[1].sqerr = 0.001
    FECj[1].valMin = 2.4
    FECj[1].valMax = 2.6
    FECj[2].val = None
    FECj[2].sqerr = INFINITE
    FECj[2].valMin = 0
    FECj[2].valMax = INFINITE
    
    FECwReq = CCRow("FECwReq", NJ)
    FECwReq[0].val = 3
    FECwReq[0].valMin = 2.9
    FECwReq[0].valMax = 3.1
    FECwReq[0].sqerr = 0.001
    FECwReq[1].val = 2.5
    FECwReq[1].sqerr = 0.001
    FECwReq[1].valMin = 2.4
    FECwReq[1].valMax = 2.6
    FECwReq[2].val = None
    FECwReq[2].sqerr = INFINITE
    FECwReq[2].valMin = 0
    FECwReq[2].valMax = INFINITE

    CT = CheckTotals("FECi-FECj", FECi, FECj, FECLink, FECwReq)
    
    CT.check()
