# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	CheckWHEE
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Consistency check of waste heat of electrical equipment data
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 				
from GUI.status import *
from ccPar import CCPar
from constants import DEBUG
from ccheckFunctions import *
from GUI.GUITools import check
from modules.fluids import Fluid
import database.dbio as dbio


class CheckWHEE():
    """Carry out consistency checking for heat exchanger h"""
    def __init__(self, n, checkInPlace = False):
        """Setup the check with the data of the WHEE with WHEENo == n + 1""" 

        self.checkInPlace = checkInPlace
        self.index = n
        self.dataGroup = "WHEE"

        # assign a variable to all intermediate/calculated values needed
        self.QWHEEYear = CCPar("QWHEEYear", priority=2)
        self.HPerYear = CCPar("HPerYear")
        self.HPerDayWHEE = CCPar("HPerDayWHEE")
        self.NDaysWHEE = CCPar("NDaysWHEE")
        self.fluid = None

        self.DTWHEEMax = CCPar("DTWHEEMax", parType = "DT")
        self.TEnv = CCPar("TEnv", parType = "T")
        self.TEnv.setValue(18.0,err=0.0,setTrack=False)
        
        self.dotQWHEEMax = CCPar("dotQWHEEMax", unitType = "POWER")
        self.diffQWHEE = CCPar("diffQWHEE", unitType = "POWER")

        self.DTWHEE = CCPar("DTWHEE", parType = "DT")
        self.WHEETReturn = CCPar("WHEETreturn", parType = "T")
        self.WHEETReturn.valMin = self.TEnv.val
        
        self.importData(n)

        if DEBUG in ["ALL", "BASIC"]:
            self.showAll()

    def importData(self, n):
        """Import data from SQL tables (from AlternativeProposalNo = -1)
        for the WHEE with WHEENo == n + 1
        
        """

        if self.checkInPlace:
            ANo = Status.ANo
        else:
            ANo = -1

        self.WHEENo = n + 1
        
        # assign empty CCPar to all questionnaire parameters
        self.QWHEE = CCPar("QWHEE", unitType = "POWER") #TODO: change name consistently into dotQWHEE !!!
        self.HPerDayWHEE = CCPar("HPerDayWHEE", unitType = "TIME")
        self.HPerDayWHEE.valMax = 24.0
        self.NDaysWHEE = CCPar("NDaysWHEE")
        self.NDaysWHEE.valMax = 365.
        self.NBatchWHEE = CCPar("NBatchWHEE")
        self.HBatchWHEE = CCPar("HBatchWHEE", unitType = "TIME")
        self.HBatchWHEE.valMax = 24.0
        self.WHEEFlow = CCPar("WHEEFlow", unitType = "MASSFLOW")
        self.WHEETOutlet = CCPar("WHEETOutlet", parType = "T", priority=2)
        
        self.allParameters = (self.QWHEE, self.HPerDayWHEE, self.NDaysWHEE,
                              self.NBatchWHEE, self.HBatchWHEE, self.WHEEFlow, self.WHEETOutlet, self.WHEETReturn,
                              self.DTWHEE, self.DTWHEEMax, self.TEnv, self.dotQWHEEMax, self.diffQWHEE)

        # reading data from table "qwasteheatelequip"
##        qwheeTable = Status.DB.qwasteheatelequip.ProjectID[Status.PId].AlternativeProposalNo[ANo].WHEENo[self.WHEENo]
##
##        if len(qwheeTable) > 0:
##            qwhee = qwheeTable[0]
        qwhee = Status.prj.getWHEE(PId=Status.PId, ANo=ANo, no=self.WHEENo)
        if qwhee is not None:

            self.QWHEE.setValue(qwhee.QWHEE)
            self.WHEEFlow.setValue(qwhee.WHEEFlow)
            self.WHEETOutlet.setValue(qwhee.WHEETOutlet)

            # if from the input parameters no temperature levels can be calculated
            if self.WHEETOutlet.val is None:
                logWarning(_("Electrical equipment %s: no temperature level specified. %s °C assumed"%(self.index, 60.0)))
                self.WHEETOutlet.setValue(60.0)

            if self.WHEEFlow.val is None:
                    self.DTWHEE.setValue(10.0, err=16.0)
                
            if qwhee.NDaysWHEE is not None:
                self.NDaysWHEE.setValue(qwhee.NDaysWHEE, err=0.0)
                
            self.HPerDayWHEE.setValue(qwhee.HPerDayWHEE)                
            
            if qwhee.NBatchWHEE is not None:
                self.NBatchWHEE.setValue(qwhee.NBatchWHEE, err=0.0)

            if qwhee.HBatchWHEE is not None:
                self.HBatchWHEE.setValue(qwhee.HBatchWHEE, err=0.0)

            if qwhee.WHEEMedium is not None:
                self.fluid = Fluid(int(qwhee.WHEEMedium))

    def exportData(self):
        """Store corrected data in SQL (under AlternativeProposalNo = 0)"""
        if self.checkInPlace:
            ANo = max(Status.ANo, 0)
        else:
            ANo = 0
            
##        qwheeTable = Status.DB.qwasteheatelequip.ProjectID[Status.PId].AlternativeProposalNo[ANo].WHEENo[self.WHEENo]
##        if len(qwheeTable) > 0:
##            qwhee = qwheeTable[0]
        qwhee = Status.prj.getWHEE(PId=Status.PId, ANo=ANo, no=self.WHEENo)
        if qwhee is not None:

            qwhee.set(QWHEE = dbio.check(self.QWHEE.val),
                      NDaysWHEE = dbio.check(self.NDaysWHEE.val),
                      HPerDayWHEE = dbio.check(self.HPerDayWHEE.val),
                      NBatchWHEE = dbio.check(self.NBatchWHEE.val),
                      HBatchWHEE = dbio.check(self.HBatchWHEE.val),
                      WHEETOutlet = dbio.check(self.WHEETOutlet.val),
                      WHEEFlow = dbio.check(self.WHEEFlow.val))
            
            dbio.commit()
                
    def showAll(self):
        """Print all values for debugging"""
        print "====================="
        for param in self.allParameters:
            param.show()
        print "====================="
        
    def screen(self):  
        """Screen all variables in the block"""
        for param in self.allParameters:
            param.screen()

    def check(self):
        """main function, carry out the check of the block"""
        if DEBUG in ["ALL"]:
            print "-------------------------------------------------"
            print " Checking heat exchangers "
            print "-------------------------------------------------"

        for n in range(1):

            cycle.initCheckBalance()

            # Step 1: Call all calculation routines in a given sequence
            if DEBUG in ["ALL"]:
                print "-------------------------------------------------"
                print "Cycle %s" % n
                print "-------------------------------------------------"
                print "Step 1: calculating from left to right (CALC)"

            self.HPerYear.push(calcProd("HPerYearWHEE", self.NDaysWHEE, self.HPerDayWHEE))
            self.QWHEEYear.push(calcProd("QWHEEYear", self.QWHEE, self.HPerYear))
            self.HPerDayWHEE.push(calcProd("HPerDayWHEE", self.NBatchWHEE, self.HBatchWHEE))

            if self.fluid is not None:
                self.dotQWHEEMax.push(calcFlow_stack("dotQWHEEMax", 
                                                            self.fluid.rho*self.fluid.cp, 
                                                            self.WHEEFlow,
                                                            self.WHEETOutlet, 
                                                            self.TEnv,
                                                            self.DTWHEEMax,
                                                     unitType = "POWER"))
                self.QWHEE.push(calcFlow_stack("QWHEE", 
                                                            self.fluid.rho*self.fluid.cp, 
                                                            self.WHEEFlow,
                                                            self.WHEETOutlet, 
                                                            self.WHEETReturn,
                                                            self.DTWHEE,
                                                     unitType = "POWER"))
                
                self.diffQWHEE.push(calcDiff("diffQWHEE",self.dotQWHEEMax, self.QWHEE))

            pass     #here calcSum(self.QWHEE,sel.QWHEERec,self.QWHEEAmb)

            # Step 2: Cross check the variables
            if DEBUG in ["ALL"]:
                self.showAll()
                print "Step 2: cross checking"

            self.ccheckAll()
            
            # Step 3: Adjust the variables (inverse of calculation routines)
            if DEBUG in ["ALL"]:
                self.showAll()
                print "Step 3: calculating from right to left (ADJUST)"
            
            adjustProd_stack(self.HPerYear, self.NDaysWHEE, self.HPerDayWHEE)
            adjustProd_stack(self.QWHEEYear, self.QWHEE, self.HPerYear)
            adjustProd_stack(self.HPerDayWHEE, self.NBatchWHEE, self.HBatchWHEE)
            
            if self.fluid is not None:
                adjustFlow_stack(self.dotQWHEEMax, self.fluid.rho*self.fluid.cp, 
                                                            self.WHEEFlow,
                                                            self.WHEETOutlet, 
                                                            self.TEnv,
                                                            self.DTWHEEMax)
                adjustFlow_stack(self.QWHEE, self.fluid.rho*self.fluid.cp, 
                                                            self.WHEEFlow,
                                                            self.WHEETOutlet, 
                                                            self.WHEETReturn,
                                                            self.DTWHEE)
                adjustDiff_stack(self.diffQWHEE, self.dotQWHEEMax, self.QWHEE)

            # Step 4: Cross check again the variables
            if DEBUG in ["ALL"]:
                self.showAll()
                print "Step 4: second cross checking"

            self.ccheckAll()

            if DEBUG in ["ALL"]:
                self.showAll()

        if DEBUG in ["ALL", "BASIC"]:
            self.showAll()

        cycle.checkTotalBalance()
        return cycle.getMeanBalance()
        
    def ccheckAll(self):    
        """Adjust the approximations of each WHEE parameter, and check
        each each one for internal for consistency.
        
        """
        for param in self.allParameters:
            param.check()


if __name__ == "__main__":
    from modules.ccheck.checkHX import CheckHX
    from modules.ccheck.ccheckFunctions import CCScreen
    import GUI.pSQL as pSQL, MySQLdb

    # direct connecting to SQL database w/o GUI. for testing only
    stat = Status("testCheckProc")
    Status.SQL = MySQLdb.connect(user="root", db="einstein")
    Status.DB = pSQL.pSQL(Status.SQL, "einstein")
    Status.PId = 41
    Status.ANo = -1
    
    screen = CCScreen()
    
    ccHX = CheckHX(1)       # creates an instance of class CCheck
    ccHX.check()
    ccHX.exportData(1)

    ccHX.screen()
    screen.show()

