# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	DataSet
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Managament of (schedule) data sets (Modules QM and CheckSchedule)
#
#==============================================================================
#
#	(C) copyleft E4-Experts GmbH, Berlin, Germany
#	www.energyxperts.net / info@energyxperts.net
#
#============================================================================== 				

from GUI.status import Status
from modules.messageLogger import *
from auxiliary.auxiliary import normalize, toFloat
import numpy as np
from lib.fileIO.fileIO import readTableFromFile, writeTableToFile, readColumn, readHeaders  # reading in schedules from data files
from lib.dateTime.dateTimeTools import stringToDate3, dateToHOY
from lib.fileIO.config import Config
import database.dbio as dbio
from constants import APPDATA

def loadDataSets():
    """ loads all data sets for a given project
    """

    dataFiles = []
    dataSets = []
    
    dbfiles = dbio.select("qmdatafiles", ProjectID = Status.PId)
    for dbfile in dbfiles:
        df = DataFile(id=dbfile.id,
                      fileName=dbfile.FileName,
                      shortName=dbfile.ShortName)
        
        dataFiles.append(df)

        dbsets = dbio.select("qmdatasets", DataFileId=df.id)
        for dbset in dbsets:

            parCode = dbset.ParCode
            objectId = dbset.ObjectId
            
            objectNames, objectIds, objectIndex = objectCodes(parCode)
            if objectId in objectIds:
                index = objectIds.index(objectId)
                if objectIndex is None:
                    objectNo = index
                    cNo = 0
                else:
                    objectNo, cNo = objectIndex[index]
            else:
                logError("Error importing data set. Cannot find object for parameter %s with Id %s"%(parCode, objectId))
                objectNo = None
                cNo = None

            scale = dbset.Scale
            scale = scale if scale is not None else 1.0
            
            ds = DataSet(id=dbset.id,
                         parCode=parCode,
                         col=dbset.ColNo,
                         objectNo=objectNo,
                         cNo=cNo,
                         df=df,
                         scale=scale)
            dataSets.append(ds)

    return dataSets, dataFiles
            

class DataSet():
    def __init__(self, id=None, parCode=None, objectNo=None, cNo=0, fileNo=None, col=None, df=None, scale=1.0):
        self.id = id
        
        self.parCode = parCode
        self.objectNo = objectNo
        self.cNo = cNo
        self.col = col
        self.df = df
        self.scale = scale

        # attributes which have to be set when readin in data from datafile
        self.dataType = None
        self.periods = []   # list of tuples (startHour, stopHour), where stopHour is the first hour AFTER the period
                            # so that the sum can be calculated as sum(x[startHour:stopHour])
        self.values = None  # values of data
        self.total = None   # sum of valid data
        self.min = None     # max of valid data
        self.max = None     # min of valid data
##        self.totalYear = None # yearly total value
        
        self.relErr = None  # output. set in checkTargets

    def __repr__(self):
        shortName = self.df.shortName if self.df else None
        cNo = "[%s]"%self.cNo if self.parCode in ["PTInFlow", "PTOutFlow", "mInFlowYear", "mOutFlowYear", "UPHc", "UPHw"] else ""
        string = "DataSet(%s): %s[%s]%s from %s[%s]. Type %s Totals = %s"%(self.id, self.parCode, self.objectNo, cNo, shortName, self.col, self.dataType, self.total)
        return string

    def saveToDB(self):

        if any([(par is None) for par in [self.parCode, self.objectNo, self.cNo, self.col, self.df]]):
            logError(_("Cannot save incomplete data set"))
            return
               
        if self.id is None:
            self.id = dbio.insertRow("qmdatasets")


        dbentry = dbio.get("qmdatasets", id=self.id)

        objectNames, objectIds, objectIndex = objectCodes(self.parCode)
        if objectIndex is None:
            objectId = objectIds[self.objectNo]
        elif (self.objectNo, self.cNo) in objectIndex:
            index = objectIndex.index((self.objectNo, self.cNo))
            objectId = objectIds[index]
        else:
            objectId = None
        
        entry = {"DataFileId":  self.df.id,
                 "ColNo":       self.col,
                 "ObjectId":    objectId,
                 "ParCode":     self.parCode,
                 "Scale":       self.scale}

        dbentry.set(**entry)
        dbio.commit()

    def deleteFromDB(self):
        if self.id is not None:
            dbio.delete("qmdatasets", id=self.id)


    def getData(self):
        if self.df is None: return True
        
        df = self.df

        df.read()
        pd = df.pd

        if not df.err:
            self.values = self.scale*pd.targets[self.col-1]
            self.min = np.nanmin(self.values)
            self.max = np.nanmax(self.values)
                              
            if pd.totals is not None: self.total = self.scale*pd.totals[self.col-1]
            
            self.dataType = pd.dataType
            if self.dataType == 1: # period data
                self.periods = pd.periods

        return df.err  


#========================================================================================
# class DataFile and related functions
#========================================================================================

class DataFile():
    def __init__(self, id=None, fileName=None, shortName=None, cols=[], dataType=None):
        self.id = id

        self.fileName = fileName
        self.shortName = shortName if shortName is not None else fileName
        self.cols = cols
        self.dataType = dataType
        self.srcType = "dbcon" if fileName.endswith(".conf") else "file"

        # storage of data 
        self.pd = None
        self.err = None

    def __repr__(self):
        string = "DataFile(%s) %s: %s %s %s"%(self.id, self.shortName, self.fileName, self.cols, self.dataType)
        return string
        
    def saveToDB(self):
        if self.id is None:
            self.id = dbio.insertRow("qmdatafiles")

        dbentry = dbio.get("qmdatafiles", id=self.id)

        entry = {"ProjectID": Status.PId,
                 "Path":        "",
                 "FileName":    self.fileName,
                 "ShortName":   self.shortName}

        dbentry.set(**entry)
        dbio.commit()

    def deleteFromDB(self):
        if self.id is not None:
            dbio.delete("qmdatafiles", id=self.id)

    def read(self):
        """ reads in the data from the data file and stores within self.pd
        """
        if self.pd is None: # read only once
            self.pd = readDataFile(self.fileName)
            self.cols = self.pd.keys
            self.err = (self.pd.targets is None)  # set if read error from data file
            
        return self.err

def readDataFile(fileName):
    """
    reads data from a data file and stores them in the "ProcessedData"-class
    """

    pd = ProcessedData()
    dbcon = False
    
    if fileName.endswith(".conf"): # special case: data base connector configuration file
        pd.dataType = 0
        dbcon = True
        config = Config(APPDATA + "/" + fileName) # get configuration parameters before set-up
        if config.err>0:
            logError(_("Data file %s does not have a correct format")%fileName)
            return pd

        labels = config.get("sensors", "labels")
        ncol = len(labels)+1
        pd.keys = list(labels)
        
        pd.fullYear = False
        pd.targets = [np.zeros(8760, dtype=float) for col in xrange(1, ncol)]
        return pd
    table = readTableFromFile(fileName, separator="\t")

    if not table or len(table) <= 0:
        return pd

    # determine data type from headers
    headers = readHeaders(table)
    if len(headers) > 0:
        header = headers[0]
    else:
        return pd

    h1 = str(header[0]).upper()
    h2 = str(header[1]).upper()

    if h1.startswith("TIME"):
        pd.dataType = 0
    elif h1.startswith("FIRST") and h2.startswith("LAST"):
        pd.dataType = 1
    else:
        logError("Data file does not have the correct format (erroneous headers)")
    
    ncol = len(table[0])

    if pd.dataType==0: # hourly data
        time = readColumn(table, 0)
        # readin time intervals -> may be a subset of 8760
        err = 0
        
        ndata = len(time)
        if ndata==0:
            logError(_("Error in data file. No data rows"))
            return pd
        elif ndata > 8760:
            logWarning(_("Number of data rows in the data file is higher than 8760. Only one year will be read"))

        pd.targets = [np.array([None]*8760, dtype=float) for col in xrange(1, ncol)]
        rawTargets = [np.array([toFloat(val) for val in readColumn(table, col)], dtype=float) for col in xrange(1, ncol)]

        pd.fullYear = False
        
        for r, t in enumerate(time):
            if str(t).upper=="TOTAL":
                pd.fullYear = True
                pd.totals = [float(rawTargets[col-1][r]) for col in xrange(1, ncol)]
                
            if t >= 1:
                it = int(float(t)-0.5)%8760
                for col in xrange(1, ncol):
                    try:
                        pd.targets[col-1][it] = float(rawTargets[col-1][r])
                    except: # try except necessary for reading NaN values
                        pass

        if not pd.fullYear:
            if int(float(time[0]) + 0.5)==1 and len(time)==8760 and int(float(time[-1]) + 0.5)==8760:
                try:
                    pd.totals = [np.sum(rawTarget) for rawTarget in rawTargets]
                    pd.fullYear = True
                except:
                    pass
            
        pd.keys = table[0][1:]


    elif pd.dataType==1: # period data
        startDates = readColumn(table, 0)
        stopDates = readColumn(table, 1)

        pd.periods = []
        pd.fullYear = False
        iTotals = -1
        for i, startDateStr in enumerate(startDates):
            if startDateStr.upper().startswith("TOTAL"):
                fullYear = True
                iTotals = i
            else:
                stopDateStr = stopDates[i]
                
                startHour = dateToHOY(stringToDate3(startDateStr))
                stopHour = dateToHOY(stringToDate3(stopDateStr), startHour=False)

                pd.periods.append((startHour, stopHour))

        # check if the intervals cover the whole year (only non-overlapping intervals allowed !)
        if not pd.fullYear:
            covered = np.zeros(8760)
            for period in pd.periods:
                startHour, stopHour = period
                covered[startHour:stopHour] += 1

            if np.sum(covered - 1)==0:
                pd.fullYear = True
            
        rawTargets = [readColumn(table, col) for col in xrange(2, ncol)]
        pd.targets = [np.array([val for i, val in enumerate(rt) if i != iTotals], dtype=float) for rt in rawTargets]
        if iTotals > 0: pd.totals = np.array([rt[iTotals] for rt in rawTargets], dtype=float)
        
        pd.keys = table[0][2:]
        
    return pd

class ProcessedData():
    def __init__(self):
        self.keys = None
        self.targets = None
        self.dataType = None
        self.periods = None
        self.totals = None
        self.fullYear = None    # true if periods or f(t) covers the full year (SUM_i f_i = f_year)


def objectCodes(parCode):
    from modules.fluids import Fuel
    from modules.ccheck.checkMassFlow import MFSystem
    
    sysMF = MFSystem(ANo=-1)

    objectIndex = None
    
    if parCode in ["FECFuel_i", "FETFuel_i"]:
        fuels = Status.prj.getFuels(PId=Status.PId, ANo=-1)
        objectNames = [Fuel(fuel.DBFuel_id).name for fuel in fuels]
        objectIds = [fuel.id for fuel in fuels]
    elif parCode in ["FECel", "FETel"]:
        objectNames = []
        objectIds = []
    elif parCode in ["FETj", "USHj", "QHXEq", "FETFuel_j", "FECFuel_j", "FETel_j", "FECel_j", "TDrive", "TReject", "Control"]:
        equipment = Status.prj.getEquipment(PId=Status.PId, ANo=-1, cascade=True)
        objectNames = [eq.Equipment for eq in equipment]
        objectIds = [eq.id for eq in equipment]
    elif parCode in ["USHm", "UPHProc_m", "USHPipe", "QHXPipe"]:
        pipes = Status.prj.getPipes(PId=Status.PId, ANo=-1) 
        objectNames = [pipe.Pipeduct for pipe in pipes]
        objectIds = [pipe.id for pipe in pipes]
    elif parCode in ["UPH", "UPHProc_k", "QHXProc"]:
        processes = Status.prj.getProcesses(PId=Status.PId, ANo=-1) 
        objectNames = [proc.Process for proc in processes]
        objectIds = [proc.id for proc in processes]

    elif parCode in ["Product"]:
        objectType = "PROD"
        objectNames = [node.name for node in sysMF.node if node.objectType==objectType]
        objectIds = [node.objectId for node in sysMF.node if node.objectType==objectType]

    elif parCode in ["RawMat"]:
        objectType = "RAW"        
        objectNames = [node.name for node in sysMF.node if node.objectType==objectType]
        objectIds = [node.id for node in sysMF.node if node.objectType==objectType]

    elif parCode in ["PTInFlow", "PTOutFlow", "mInFlowYear", "mOutFlowYear", "UPHc", "UPHw"]:
        objectNames = []
        objectIndex = []
        objectIds = []
        processes = Status.prj.getProcesses(PId=Status.PId, ANo=-1) 
        for k, proc in enumerate(processes):
            procName = proc.Process
            procId = proc.id
            (inStreams, outStreams) = Status.prj.getStreams(procId)
            streams = inStreams if parCode in ["PTInFlow", "mInFlowYear", "UPHc"] else outStreams
            for s, stream in enumerate(streams):
                name = procName + ":" + stream.Name
                objectNames.append(name)
                objectIds.append(stream.id)
                objectIndex.append((k, s))
            
    else:
        objectNames = []
        objectIds = []

    if objectIndex is None:
        objectIndex = [(i, 0) for i in xrange(len(objectNames))]
        
    return objectNames, objectIds, objectIndex


def makeSchedule(ds, baseSchedule=None):
    """ helper function converting a period profile from a data set into a
    hourly schedule
    """

    if ds.dataType==0:
        return normalize(ds.values)
    
    else: # period profile
        schedule = np.zeros(8760)

        for p, period in enumerate(ds.periods):
            start, stop = period
            if stop > start:
                base = 1.0
                if baseSchedule is not None:
                    baseAvg = np.average(baseSchedule[start:stop])
                    if baseAvg > 0:
                        base = baseSchedule[start:stop]/baseAvg
                
                schedule[start:stop] = base*ds.values[p]/(stop-start)

    return normalize(schedule)


#===================================================================================================
def testProfile():
    """
    ad hoc created profile for testing only
    """

    tp = np.zeros(Status.Nt)
    for doy in range(365):
        weekDay = doy % 7
        dayProfile = np.zeros(24)
        dayProfile[6:18] += 1.0

        c = 3.0*182.5/182
        peak = 363.5

        if doy >= 180:
            f = max(1.0 - np.square((peak - doy)/91), 0)
        else:
            f = max(1.0 - np.square((365.0-peak + doy)/91), 0)

        dayProfile += c*f

        tp[doy*24:(doy+1)*24] = dayProfile

    if np.sum(tp) > 0:
        tp /= np.sum(tp)


    table = [[i, 1500000*tp[i]] for i in xrange(Status.Nt)]

    from lib.fileIO.fileIO import writeTableToFile
    writeTableToFile(table, "testProfile.csv")
    return tp


