# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleCH
#			
#------------------------------------------------------------------------------
#			
#	Module for calculation of chillers
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from modules.modulePipes import getAllCoolingPipes
import numpy as np
import copy
from constants import *
from GUI.status import Status
from modules.messageLogger import *
from auxiliary.auxiliary import findInListASC
from lib.auxiliary.auxiliary import noneCheck
from modules.heatFlow import HeatFlow_t

import database.dbio as dbio

CHTDROP_OLD = 7     # temperature difference between inlet to evaporator/condenser and refrigerant (used until V2.0).

CHTDROP_EVAP = 9.25     # temperature difference between inlet chilled water to evaporator and evaporation temperature.
CHTDROP_COND_AIRCOOLED = 13.5     # temperature difference between inlet cooling air to condenser and condensing temperature.
CHTDROP_COND_WATERCOOLED = 10     # temperature difference between inlet cooling water to condenser and condensing temperature.

CHDTIO = 5      # temperature lift in chilled water and heat rejection loops
assert(CHDTIO==Status.TemperatureInterval)
#TODO: WARNING Code currently valid only if CHDTIO = Temperature Interval of temperature scale. For other values adapt code of calculation of temperature distribution of waste heat

EPS = 1.e-10    # tolerance for numerical errors

DEFAULT_DTRECOOL_AIRCOOLED_OLD = 16.5   #BREF cooling, average value for indirect air cooling (water cooled chillers + dry cooling tower). Used until V2.0
DEFAULT_DTRECOOL_WATERCOOLED_OLD = 14.5 #Bref cooling: 7 - 14 + 3-5 K. Used until V2.0

DEFAULT_DTRECOOL_AIRCOOLED = 9.5   #applies for dry cooling towers: water outlet at Tower vs. Tdb
DEFAULT_DTRECOOL_WATERCOOLED = 7 #wet cooling towers (and water cooled chillers): water outlet at Tower vs. Twb

DEFAULT_TMINRECOOL = 20 - CHTDROP_EVAP

DEFAULT_COPEX_AIRCOOLED = 0.4983    #average chillers from manufacturer York
DEFAULT_COPEX_WATERCOOLED = 0.5194  #average chillers from manufacturer York

DEFAULT_DEGRAD_FACT = 0.9   # Degradation factor value, as indicated in prEN14825
DEFAULT_COP_CTDRY = 22.222  # COP considers circulation pump, cooling tower pump and cooling tower fans
DEFAULT_COP_CTWET = 33.333  # COP considers circulation pump, cooling tower pump and cooling tower fans
DEFAULT_COP_FRESHWATER = 100.0  # COP considers circulation pumps

WCRATIO_CT = 0.002          # specific water consumption in m3/kWh for wet cooling towers
WCRATIO_DIRECT = 3.6 / 4.19 / 25.      # specific water consumption for direct cooling (for 25 K of temperature rise)


class ModuleCH(object):

    def __init__(self, keys):
        self.keys = keys
        # up to which cascade level do we know of existing chillers 
        self.cascadeIndex = 0
        self.neweqs = 0 # counter of added new equipments
        self.NEquipe = None # counter of all equipments of the project

    def initPanel(self):
        logTrack("ModuleChiller (initPanel)")

    def updatePanel(self):
        """
        Carry out any calculations necessary previous to displaying the CH
        design assistant window
        """
        # get list of chillers in the system
        # and update the energetic calculation up to the level needed for representation
        (CHList, CHTableDataList) = self.screenEquipments()

        if Status.int.cascadeUpdateLevel < self.cascadeIndex:
            Status.mod.moduleEnergy.runSimulation(last=self.cascadeIndex)

        # 1. List of equipments
        data = np.array(CHTableDataList)
        Status.int.setGraphicsData('CH Table', data)

        # 2. Preparing data
        PowerSumTmax = 0
        for item in CHList:
            pNom = item["equipePnom"] if item["equipePnom"] != None else 0
            PowerSumTmax += pNom

        if len(CHList) == 0:
            # assures that index >= 0
            # even if there's NO equipment at all
            index = max(len(Status.prj.getEquipments()) + 1, 1)
        else:
            ch = dbio.get("qgenerationhc", id=CHList[0]["equipeID"])
            if ch is not None:
                index = ch.CascadeIndex

        QD_20C = Status.int.QD_Tt_mod[index - 1][int((20 - Status.MinimumTemperatureC) / Status.TemperatureIntervalC + 0.5)].tolist()
        QD_20C.sort(reverse=True)

        QD_05C = Status.int.QD_Tt_mod[index - 1][int((5 - Status.MinimumTemperatureC) / Status.TemperatureInterval)].tolist()
        QD_05C.sort(reverse=True)

        QDn10C = Status.int.QD_Tt_mod[index - 1][int((-10 - Status.MinimumTemperatureC) / Status.TemperatureInterval)].tolist()
        QDn10C.sort(reverse=True)
        
        QD_Total = Status.int.QD_Tt_mod[index - 1][0][-1]
        
        # 2. XY Plot
        TimeIntervals = [(Status.TimeStep * (it + 1) * Status.EXTRAPOLATE_TO_YEAR)
                         for it in xrange(Status.Nt)]
        try:

            Status.int.setGraphicsData('CH Plot', [TimeIntervals, QD_20C, QD_05C, QDn10C])
        except:
            logDebug("ModuleCH (updatePanel): problems sending data for CH Plot")

        # 3. Configuration design assistant
#        config = self.getUserDefinedPars()
        config = [1, 10, 0, "Natural Gas", 100, 500, 80]
        Status.int.setGraphicsData('CH Config', config)

        # 4. additional information (Info field right side of panel)
        info = []
        info.append(None)
        info.append(None)
        info.append(max(0, QD_20C[0]))
        info.append(max(0, (QD_05C[0] - QD_20C[0])))
        info.append(max(0, (QDn10C[0] - QD_05C[0])))
        info.append(max(0, (QD_Total - QDn10C[0])))
        Status.int.setGraphicsData('CH Info', info)

    def addEquipmentDummy(self):
        """Add a new dummy equipment to the equipment list and return
        the equipment object.
        
        """
        self.equipe = Status.prj.addEquipmentDummy()
        self.dummyEqId = self.equipe.QGenerationHC_ID

        self.neweqs += 1 #No of last equip added
        NewEquipmentName = "New chiller %s" % (self.neweqs)

        # connect the dummy to all heating pipes
        dummy, pipeIDs = getAllCoolingPipes()
        pipeIDs = [str(pipeId) for pipeId in pipeIDs ]
        IDString = ";".join(pipeIDs)
        
        equipeData = {"Equipment":NewEquipmentName, 
                      "EquipType":"compression chiller (air cooled)",
                      "PipeDuctEquip":IDString}
        
        self.equipe.set(**equipeData)

        self.screenEquipments()
        self.cascadeIndex = self.NEquipe

        return(self.equipe)

    def deleteEquipment(self, rowNo, automatic=False):
        """Remove the equipment in row number rowNO from the project."
        
        If automatic is True, rowNo is expected to be an equipmentId
        from the database.
        
        """
        if automatic == False:
            if rowNo == None:   # indicates to delete last added equipment dummy
                BBid = self.dummyEqId
            else:
                # delete CH from the equipment list under current alternative 
                # from C&QGenerationHC under ANo
                BBid = self.getEqId(rowNo)
        else:
            BBid = rowNo

        Status.prj.deleteEquipment(BBid)
        self.screenEquipments()

    def setEquipmentFromDB(self, equipe, modelID): #SD change 30/04.2008
        """ takes an equipment from the data base and stores it under a given Id in
            the equipment data base
        """
        # chiller takes equipments from HP database
        model = dbio.get("dbheatpump", id=modelID)
        if model is None:
            logDebug("Heat pump / chiller model with id %s not found"%modelID)
            return

        if model.HPManufacturer is not None: equipe.Manufact=model.HPManufacturer
        if model.HPYearUpdate is not None: equipe.YearManufact=model.HPYearUpdate
        if model.HPModel is not None: equipe.Model=model.HPModel

        if model.HPType in HPTYPES: # compression and thermal chillers are classified by HPTYPES and HPSUBTYPES !!!
            if model.HPType == "compression":
                if model.HPSubType in ["water-water" , "air-water"]:
                     equipe.EquipType = "compression chiller (water cooled)"
                     equipe.DestinationWasteHeat = "cooling tower (wet)"
                else:
                     equipe.EquipType = "compression chiller (air cooled)"
                     equipe.DestinationWasteHeat = "direct heat rejection"
                     
            elif model.HPType == "thermal":
                if model.HPSubType in ["water-air" , "air-air"]:
                     equipe.EquipType = "thermal chiller (water cooled)"
                     equipe.DestinationWasteHeat = "cooling tower (wet)"
                else:
                     equipe.EquipType = "thermal chiller (air cooled)"
                     equipe.DestinationWasteHeat = "direct heat rejection"

        if model.HPCoolCap is not None: equipe.HCGPnom=model.HPCoolCap
        if model.HPCoolCOP is not None: equipe.HCGTEfficiency=model.HPCoolCOP #changed from HPThHeatCOP,SD
        equipe.NumEquipUnits=1

        if model.HPElectConsum is not None: equipe.ElectriConsum=model.HPElectConsum

        if model.HPEvapTmin is not None: equipe.TMaxSupply=model.HPEvapTmin
        if model.HPEvapTinC is not None:
            TcoNom = model.HPEvapTinC - CHDTIO
            equipe.TSupply=TcoNom
            
        if model.HPCondTinC is not None:
            equipe.TemperatureReCooling=model.HPCondTinC
        
        if model.HPExCoolCOP is not None:
            equipe.HPExCoolCOP=model.HPExCoolCOP
        else:
            COP = model.HPCoolCOP
            COPth = model.HPThCoolCOP
            if COP is not None and COPth is not None:
                if COP <= COPth and COPth > 0:
                    COPex = COP / COPth
                    logMessage("Exergetic efficiency of chiller %s calculated to %s" % \
                               (equipe.Model, COPex))
                    equipe.HPExCoolCOP=COPex
                else:
                    logWarning("Error in efficiency data of the heat pump %s" % equipe.Model)
                
        # references to DB
        equipe.IsSelectedFromDB=1
        equipe.DatabaseNameSelection="DBHeatPump"
        if model.HPType is not None: equipe.EquipTypeFromDB=model.HPType
        equipe.EquipIDFromDB=model.DBHeatPump_ID

        # fuel consumption. usually not used
        if model.DBFuel_id is not None: equipe.DBFuel_id=model.DBFuel_id
        if model.HPFuelConsum is not None: equipe.FuelConsum=model.HPFuelConsum
        if model.HPUnitsFuelConsum is not None: equipe.UnitsFuelConsum=model.HPUnitsFuelConsum

        equipe.FlowExhaustGas = 0.0
        equipe.TExhaustGas = 0.0

        # economic parameters
        equipe.set(**{"TurnKeyPrice": model.HPTurnKeyPrice,
                       "OandMfix": model.HPOandMfix,
                       "OandMvar": model.HPOandMvar})
        (turnKey, OMFix, OMVar) = self.estimateEconomicParameters(equipe)
        equipe.set(**{"TurnKeyPrice": turnKey,
                       "OandMfix":OMFix,
                       "OandMvar":OMVar})

        #dataChecking
        if (model.HPExCoolCOP is None) and \
            ((model.HPCoolCOP is None) or (model.HPEvapTinC is None) or (model.HPCondTinC is None)):
            logWarning("Efficiency data for chiller %s %s are not complete\n"%(model.HPManufacturer, model.HPModel))
                     
        dbio.commit()

#------------------------------------------------------------------------------
    def getEqId(self, rowNo):
        """Return the EqId from the rowNo in the CHList"""
        return self.CHList[rowNo]["equipeID"]


    def screenEquipments(self, setIndex=True):
        """Update the internal databuffers and return already existing 
        chillers in the industry and their data. 
        
        The updates comprise of the self.NEquipe counter and CHList.
        
        """
        equipments = Status.prj.getEquipments()
        Status.int.getEquipmentCascade()
        self.NEquipe = len(equipments)
        self.CHList = []
        maxIndex = self.NEquipe # default to highest possible value
        index = 0
        for j, row in enumerate(Status.int.cascade):
            index += 1
            eqObject = Status.int.systemModel.eqObject[j]
            if not eqObject.heating:
                self.CHList.append(row)
                maxIndex = index

        if setIndex == True:
            self.cascadeIndex = maxIndex

        CHTableDataList = []
        for j, row in enumerate(Status.int.EquipTableDataList):
            eqObject = Status.int.systemModel.eqObject[j]
            if not eqObject.heating:
                CHTableDataList.append(row)

        #screen list and substitute None with "not available"
        for i in range(len(CHTableDataList)):
            for j in range(len(CHTableDataList[i])):
                if CHTableDataList[i][j] == None:
                    CHTableDataList[i][j] = "not available"
        return (self.CHList, CHTableDataList)

    def getTAmbDryBulb(self):
        return Status.WeatherData.Tdb.copy()
    
    def getTAmbWetBulb(self):
        return Status.WeatherData.Twb.copy()   

###@JAN: please check if correct with cooling temperature scale and ordering. 
        
    def getTMinD(self, Q_T):
        """Looks for the minimum temperature in cooling demand
        @param Q_T: temperature dependent cooling demand (descending)

        """

        for iT in xrange(Status.NT + 1):
            if Q_T[iT + 1] < Q_T[iT] - EPS:
                break
        return(Status.int.TC[iT])        #find the T corresponding to index

    def estimateCOPex(self, COP, Th, Tc, TLevels="internal", TDropCond=CHTDROP_COND_AIRCOOLED):
        """estimates the exergetic efficiency (COPex) for a chiller with
        a given COP at nominal conditions Th and Tc
        @param Th,Tc: hot and cold temperature in Celsius
        @param TLevels: None: internal temperatures; "external": inlet to condenser, outlet from evaporator
        """

        if COP is None:
            COPex = None
        else:
            EERCarnot = self.EERCarnot(Th, Tc, TLevels, TDropCond)
            if COP > EERCarnot:
                COPex = 1.0
            else:
                COPex = COP / EERCarnot
        
        return(COPex)

    def EERCarnot(self, Th, Tc, TLevels="internal", TDropCond=CHTDROP_COND_AIRCOOLED):
        """Looks for the minimum temperature in cooling demand
        @param ThC,TcC: hot and cold temperature in Celsius
        @param TLevels: None: internal temperatures; "external": inlet to condenser, outlet from evaporator

        """
        if Th is None:
            ThK = 30 + KELVIN
        else:
            ThK = Th + KELVIN
            
        if Tc is None:
            TcK = KELVIN
        else:
            TcK = Tc + KELVIN

        if TLevels != "internal":
            ThK += TDropCond
            TcK += (CHDTIO - CHTDROP_EVAP)

        if ThK > TcK:
            EER = TcK / (ThK - TcK)
        else:
            EER = INFINITE

#        logDebug("ModuleCH: EERCarnot estimated to %s at Th = %s and Tc = %s (%s)"%(EER,Th,Tc,TLevels))
            
        return EER

    def getFRecool(self, equipmentType, dryTower=False):
        """gets the fRecool value as a fucntion of the equipment type. For calls from outside the chiller module (ccheck)
        """
        
        if equipmentType == "compression chiller (water cooled)":
            if dryTower:
                return(1. / DEFAULT_COP_CTDRY)
            else:
                return(1. / DEFAULT_COP_CTWET)
        else:
            return(0)

    def getWaterConsumption(self, equipmentType, USCj, QWHj, dryTower=False):

        if equipmentType in ["compression chiller (water cooled)","thermal chiller (water cooled)"]:
            if dryTower:
                waterConsumption = 0
            else:
                waterConsumption = QWHj * WCRATIO_CT
        elif equipmentType == "cooling tower (wet)":
            waterConsumption = USCj * WCRATIO_CT
        elif equipmentType == "fresh or ground water":
            waterConsumption = USCj * WCRATIO_DIRECT
        else:
            waterConsumption = 0

        return waterConsumption

    def getDefaultEER(self, equipmentType):

        if equipmentType == "cooling tower (wet)":
            return(DEFAULT_COP_CTWET)
        elif equipmentType == "cooling tower (dry)":
            return(DEFAULT_COP_CTDRY)
        elif equipmentType == "fresh or ground water":
            return(DEFAULT_COP_FRESHWATER)
        else:
            return 3.0

        
    def calculateEnergyFlows(self, equipe, QDc_Tt, cascadeIndex):
        """Update the cascade by calculating of energy flows in the chillers.
        
        :param equipe: The equipment to calculate
        :param cascadeIndex: The cascade index of the equipment
        
        """
#        print "ModuleCH (cEF): cascadeIndex = %s"%cascadeIndex
#        print "Demand = %s\n"%sum(QDc_Tt[0]),QDc_Tt,"\n---"
        Status.int.extendCascadeArrays(cascadeIndex)

        NT = Status.NT
        DT = Status.TemperatureInterval
        Nt = Status.Nt
        Dt = Status.TimeStep

        j = cascadeIndex - 1
        equipmentType = equipe.EquipType
        eqObject = Status.int.systemModel.eqObject[cascadeIndex-1]

        PcNom = equipe.HCGPnom
        if PcNom is None:

            if equipmentType == "fresh or ground water":
                PcNom = INFINITE    # by default suppose infinite availability of water for cooling
            else:
                logTrack("ModuleChiller (cEF-dummy): no equipment capacity specified")
                PcNom = 0

        controlSchedule = Status.int.systemModel.eqObject[cascadeIndex-1].controlSchedule
        control = controlSchedule.schedule if controlSchedule else 1.0

        dryTower = False #default. True/relevant only for water-cooled compression chillers
        if equipmentType in ["compression chiller (water cooled)"]:
            CTType = equipe.DestinationWasteHeat
            if CTType == "cooling tower (dry)":
                dryTower = True
                  
                
        

        EERSystemNom = equipe.HCGTEfficiency
        if EERSystemNom is None:
            if equipmentType == "cooling tower (dry)":
                COPc_nom = DEFAULT_COP_CTDRY # COP considers circulation pump, cooling tower pump and cooling tower fans
            elif equipmentType == "cooling tower (wet)":
                COPc_nom = DEFAULT_COP_CTWET # COP considers circulation pump, cooling tower pump and cooling tower fans
            elif equipmentType == "fresh or ground water":
                COPc_nom = DEFAULT_COP_FRESHWATER # COP considers circulation pump, cooling tower pump and cooling tower fans
            else:
                COPc_nom = None
        else:
            if EERSystemNom > 0:
                fRecool = self.getFRecool(equipmentType, dryTower)
                fSystem = 1. / EERSystemNom
                if fSystem <= fRecool:
                    logError("Erroneous data in equipment efficiency for chiller (equipment no. %s: %s)" % (equipe.EqNo, equipe.Equipment))
                    COPc_nom = 0.0
                else:
                    COPc_nom = (1. + fRecool) / (fSystem - fRecool)

                logDebug("ModuleCH.calculateEnergyFlows: EERSystemNom %s fRecool %s COPc_nom %s"%(EERSystemNom,fRecool,COPc_nom))

            else:
                COPc_nom = 0.0
                

        #temperature of cold supply in current application
        TSupply = equipe.TSupply
        if equipe.TSupply is None:
            if equipmentType in ["compression chiller (air cooled)", "compression chiller (water cooled)"]:
                TSupply = -10.0
            elif equipmentType in ["thermal chiller (air cooled)", "thermal chiller (water cooled)"]:
                TSupply = 5.0
            elif equipmentType in ["cooling tower (dry)", "cooling tower (wet)"]:
                TSupply = 30.0
            elif equipmentType in ["fresh or ground water"]:
                TSupply = 10.0
        
        #minimum possible supply temperature for current equipment
        TMinSupply = equipe.TMaxSupply
        if TMinSupply is None:
            if equipmentType in ["compression chiller (air cooled)", "compression chiller (water cooled)"]:
                if TSupply is not None:
                    TMinSupply = max(-60, TSupply - 20)
                else:
                    TMinSupply = -60
            else:
                TMinSupply = TSupply
                
                    
        if (equipmentType in ["compression chiller (air cooled)", "thermal chiller (air cooled)"]):
            Trecool = self.getTAmbDryBulb()             # air cooled chillers are considered direct cooled by ambient air
            DTRecool = 0

        elif (equipmentType in ["cooling tower (dry)"]):
            DTRecool = DEFAULT_DTRECOOL_AIRCOOLED       #  
            Trecool = Status.WeatherData.Tdb.copy()

            Trecool += DTRecool

        elif (equipmentType in ["compression chiller (water cooled)", "thermal chiller (water cooled)"]):
            if dryTower:
                DTRecool = DEFAULT_DTRECOOL_AIRCOOLED       #  
                Trecool = Status.WeatherData.Tdb.copy()                
            else:
                DTRecool = DEFAULT_DTRECOOL_WATERCOOLED        # cooling down to dry bulb ambient temperature possible
                Trecool = Status.WeatherData.Twb.copy()
            Trecool += DTRecool

        elif (equipmentType in ["cooling tower (wet)"]):
            DTRecool = DEFAULT_DTRECOOL_WATERCOOLED        # cooling down to dry bulb ambient temperature possible

            Trecool = Status.WeatherData.Twb.copy()
            Trecool += DTRecool
        else:
            DTRecool = 0
            Trecool = Status.int.createQ_t()

        if (equipmentType in ["compression chiller (air cooled)", "compression chiller (water cooled)"]):
            Trecool.clip(min=DEFAULT_TMINRECOOL, out=Trecool)            # for compression chillers limit recooling temperature to minimum allowed

        if equipmentType in ["compression chiller (air cooled)","thermal chiller (air cooled)"]:
            TDropCond = CHTDROP_COND_AIRCOOLED
        else:
            TDropCond = CHTDROP_COND_WATERCOOLED

        TrecoolNom = equipe.TemperatureReCooling
        if TrecoolNom is None:
            TrecoolNom = np.mean(Trecool)

        COPex = equipe.HPExCoolCOP
        
        if COPex is None:  # set to default values if nothing is specified in equipment tables
            if equipmentType in ["compression chiller (air cooled)", "compression chiller (water cooled)"]: # calculation of COP for water cooled compression chillers
                COPex = self.estimateCOPex(COPc_nom, TrecoolNom, TSupply, TLevels="external", TDropCond=TDropCond)

                if COPex is None:  # still None -> estimate from user input was not possible or not realistic
                    if equipmentType == "compression chiller (water cooled)":
                        COPex = DEFAULT_COPEX_WATERCOOLED  # default value for efficiency based on water cooled chillers of York

                    elif equipmentType == "compression chiller (air cooled)": # calculation of COP for air cooled compression chillers
                        COPex = DEFAULT_COPEX_AIRCOOLED  # default value for efficiency based on air cooled chillers of York

                logMessage("ModuleCH: COPex of chiller %s estimated to %s based on COPc_nom = %s, TrecoolNom=%s and TSupply=%s" % (equipe.Equipment, COPex, COPc_nom, TrecoolNom, TSupply))

        fRecool = self.getFRecool(equipmentType, dryTower)
              
        # get demand data for CascadeIndex/EquipmentNo from Interfaces
        # and create arrays for storing heat flow in equipment
        
###        QDc_Tt = Status.int.QDc_Tt_mod[cascadeIndex - 1]

        USCj_Tt = Status.int.createQ_Tt()
        QWHj_Tt = Status.int.createQ_Tt()

        # Local vectors for storing time and temperature dependent results
        FETel = Status.int.createQ_t()
        QWH = Status.int.createQ_t()
        dUSCj_T = Status.int.createQ_T()
        FETh = Status.int.createQ_t()

        sumTc = 0
        sumTh = 0
        sumh = 0

        # start simulation
        if PcNom == 0 or COPex == 0:
            pass # nothing is actually done by the chiller
            
        elif equipe.EquipType in ["thermal chiller (water cooled)","thermal chiller (air cooled)"]:

            # outer iterative cycle: over all time steps
            iTmin = findInListASC(TMinSupply, Status.int.TC)
              
            for it in xrange(Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]

                if QDc_Tt[0][it] == 0 or fctrl==0:  # calculation time saver for 0 demand
                    USCj_Tt[:, it] = 0
                    QWHj_Tt[:, it] = 0
                    FETel[it] = 0
                    continue
                    
                # inner iterative cycle: temperature intervals              
                for iT in xrange(NT + 1, iTmin, -1):   #+1 = the > 100 °C case
                    USCj_Tt[iT][it] = min(QDc_Tt[iT][it], PcNom * Dt * fctrl) #just max(demand,nom.power)
                    dUSCj_T[iT] = USCj_Tt[iT][it]

                for iT in xrange(iTmin, -1, -1):
                    USCj_Tt[iT][it] = USCj_Tt[iT + 1][it]
                    dUSCj_T[iT] = USCj_Tt[iT][it]
                dUSCj = dUSCj_T[0]

                if dUSCj == 0:
                    FETel[it] = 0
                else:
                    EER_fullLoad = COPc_nom
                         
                    PLR = dUSCj / (PcNom * Dt)                             # Calculation of the Part Load Factor

                    #TODO: change part-load function back to the real one once consistent solution between ccheck and simulation results has been found
                    #for the moment substituted by simpler function (equal to that used in ccheck) that comes closer to what would 

                    EER_partLoad = EER_fullLoad * (0.9 + 0.1*PLR)
                    
##                    degrad_fact = 0.9                                               # Degradation factor value, as indicated in prEN14825
##                    EER_partLoad = EER_fullLoad * (PLR / ((degrad_fact * PLR) + (1 - degrad_fact)))

                    FETh[it] = dUSCj / COPc_nom
                    FETel[it] = fRecool * (dUSCj + FETh[it]) / (1. - fRecool)

##                    print "it %s USCj %s FETel %s EER %s Tc %s Th %s"%(it,dUSCj,FETel[it],EER_partLoad,self.getTMinD(dUSCj_T),Trecool[it])                    
                    sumTc += self.getTMinD(dUSCj_T)
                    sumTh += Trecool[it]
                    sumh += 1

                QWH[it] = FETel[it] + dUSCj + FETh[it]

                
                QWHj_Tt[0][it] = QWH[it]
                iTres = NT+2
                for iT in xrange(1,NT+2):
                    if Status.int.T[iT] <= Trecool[it]:
                        QWHj_Tt[iT][it] = QWH[it]
                    else:
                        iTres = iT+1
                        QWHj_Tt[iT][it] = QWH[it]*(Trecool[it] + CHDTIO - Status.int.T[iT])/CHDTIO
                        break
                
                for iT in xrange(iTres+1, NT+2):
                        QWHj_Tt[iT][it] = 0
                                    
        elif equipmentType in ["compression chiller (air cooled)","compression chiller (water cooled)"]:

            iTmin = findInListASC(TMinSupply,Status.int.TC)

            for it in xrange(Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]

                if QDc_Tt[0][it] == 0 or fctrl==0:  # calculation time saver for 0 demand
                    USCj_Tt[:, it] = 0
                    QWHj_Tt[:, it] = 0
                    FETel[it] = 0
                    FETh[it] = 0
                    continue
                    
                # inner iterative cycle: temperature intervals              
                for iT in xrange(NT + 1,iTmin,-1):   #+1 = the > 100 °C case
                    USCj_Tt[iT][it] = min(QDc_Tt[iT][it], PcNom * Dt *fctrl) #just max(demand,nom.power)
                    dUSCj_T[iT] = USCj_Tt[iT][it]

                for iT in xrange(iTmin,-1,-1):
                    USCj_Tt[iT][it] = USCj_Tt[iT+1][it]
                    dUSCj_T[iT] = USCj_Tt[iT][it]
                dUSCj = dUSCj_T[0]

                if dUSCj == 0:
                    FETel[it] = 0
                    FETh[it] = 0
                else:
                         
                    EER_fullLoad = COPex*self.EERCarnot(Trecool[it],self.getTMinD(dUSCj_T),"external",TDropCond)
                    PLR = dUSCj / (PcNom * Dt)                             # Calculation of the Part Load Factor

                    #TODO: change part-load function back to the real one once consistent solution between ccheck and simulation results has been found
                    #for the moment substituted by simpler function (equal to that used in ccheck) that comes closer to what would 

                    EER_partLoad = EER_fullLoad * (0.9 + 0.1 * PLR)
                    
##                    degrad_fact = 0.9                                               # Degradation factor value, as indicated in prEN14825
##                    EER_partLoad = EER_fullLoad * (PLR / ((degrad_fact * PLR) + (1 - degrad_fact)))

                    FETel[it] = dUSCj / EER_partLoad * (1 + (EER_partLoad + 1)* fRecool) / (1 - fRecool)

##                    print "it %s USCj %s FETel %s EER %s Tc %s Th %s"%(it,dUSCj,FETel[it],EER_partLoad,self.getTMinD(dUSCj_T),Trecool[it])
                    sumTc += self.getTMinD(dUSCj_T)
                    sumTh += Trecool[it]
                    sumh += 1
                    
                QWH[it] = FETel[it] + dUSCj

                QWHj_Tt[0][it] = QWH[it]
                iTres = NT+2
                for iT in xrange(1,NT + 2):
                    if Status.int.T[iT] <= Trecool[it]:
                        QWHj_Tt[iT][it] = QWH[it]
                    else:
                        iTres = iT+1
                        QWHj_Tt[iT][it] = QWH[it]*(Trecool[it] + CHDTIO - Status.int.T[iT])/CHDTIO
                        break
                
                for iT in xrange(iTres+1, NT+2):
                        QWHj_Tt[iT][it] = 0


        elif equipe.EquipType in ["cooling tower (dry)", "cooling tower (wet)"]:
              
            for it in xrange(Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]
                
                iTmin = findInListASC(Trecool[it], Status.int.TC)
                
                for iT in xrange(NT + 1, iTmin, -1):   #+1 = the > 100 °C case
                    USCj_Tt[iT][it] = min(QDc_Tt[iT][it], PcNom * Dt * fctrl) #just min(demand, nom.power)
                    dUSCj_T[iT] = USCj_Tt[iT][it]
                for iT in xrange(iTmin, -1, -1):
                    USCj_Tt[iT][it] = USCj_Tt[iT + 1][it]
                    dUSCj_T[iT] = USCj_Tt[iT][it]

                dUSCj = USCj_Tt[0][it]

                FETel[it] = dUSCj / COPc_nom

                if dUSCj > EPSILON:
                    sumTc += self.getTMinD(dUSCj_T)
                    sumTh += Trecool[it]
                    sumh += 1

        elif equipe.EquipType in ["fresh or ground water"]:
            iTmin = findInListASC(TMinSupply, Status.int.TC)
              
            for it in xrange(Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]

                for iT in xrange(NT + 1, iTmin, -1):   #+1 = the > 100 °C case
                    USCj_Tt[iT][it] = min(QDc_Tt[iT][it], PcNom * Dt * fctrl) #just min(demand, nom.power)
                    dUSCj_T[iT] = USCj_Tt[iT][it]
                for iT in xrange(iTmin, -1, -1):
                    USCj_Tt[iT][it] = USCj_Tt[iT + 1][it]
                    dUSCj_T[iT] = USCj_Tt[iT][it]

                dUSCj = USCj_Tt[0][it]

                FETel[it] = dUSCj / COPc_nom

                if dUSCj > EPSILON:
                    sumTc += self.getTMinD(dUSCj_T)
                    sumTh += Trecool[it]
                    sumh += 1
                                   
        if sumh > 0:
            eqObject.TcAvg = sumTc/sumh
            eqObject.ThAvg = sumTh/sumh
            logTrack("Average temperatures at chiller: Tc %s Th %s"%(sumTc/sumh,sumTh/sumh))

        # calculate the real running time
        nonnulls = [item for item in USCj_Tt[0] if item > 0]
        HPerYear = len(nonnulls) * Dt

        USCj = sum(USCj_Tt[0])
        QWHj = sum(QWHj_Tt[0])

##        print "FETel = ",FETel
        FETel_j = sum(FETel)
        FETHeat_j = sum(FETh)

        waterConsumption = self.getWaterConsumption(equipmentType, USCj, QWHj, dryTower)
        # end simulation. 

        # storing results in Interfaces
        # remaining heat/cooling demand and availability for next equipment in cascade
        QA_Tt = copy.deepcopy(Status.int.QA_Tt_mod[cascadeIndex - 1])
        Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
        Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)

        # cooling delivered by present equipment                            
        Status.int.USHj_Tt[cascadeIndex - 1] = USCj_Tt
        Status.int.USHj_T[cascadeIndex - 1] = Status.int.calcQ_T(USCj_Tt)
        Status.int.USHj_t[cascadeIndex - 1] = USCj_Tt[0].copy()
        Status.int.USHjChanged[cascadeIndex - 1] = True

        # waste heat produced by present equipment
              
###HS: @Jan: -> to be decided whether using HeatFlow_t format or old format for QWH of equipments
        Status.int.QWHj_Tt[cascadeIndex - 1] = QWHj_Tt
        Status.int.QWHj_T[cascadeIndex - 1] = Status.int.calcQ_T(QWHj_Tt)
        Status.int.QWHj_t[cascadeIndex - 1] = copy.deepcopy(QWHj_Tt[0])
        Status.int.QWHjChanged[cascadeIndex - 1] = True

        try:
            if equipmentType in ["compression chiller (water cooled)", "compression chiller (air cooled)",
                                 "thermal chiller (water cooled)", "thermal chiller (air cooled)"]:
                Status.int.QWHEq_Tt[cascadeIndex - 1] = HeatFlow_t(fluid_id=0, T0=Trecool, T1=Trecool + CHDTIO, Q_t=QWH)
            else:
                Status.int.QWHEq_Tt[cascadeIndex - 1] = HeatFlow_t()
        except:
            logError("ModuleCH.cEF: ERROR in waste heat calculation for chiller %s"%equipe.Equipment)
            pass

        # waste heat used by present equipment
        Status.int.QHXEqReq_Tt[cascadeIndex - 1] = HeatFlow_t()
        Status.int.QHXEqReqChanged[cascadeIndex - 1] = True

        # high temperature heat used by present equipment
        Status.int.FETHeat_j_Tt[cascadeIndex - 1] = HeatFlow_t(fluid_id=0,T0=75.0,T1=95,Q_t=FETh)
        Status.int.FETHeat_jChanged[cascadeIndex - 1] = True

        Status.int.FETel_j_t[j] = FETel

        Status.int.cascadeUpdateLevel = cascadeIndex

        # Global results (annual energy flows)
        Status.int.USHj[cascadeIndex - 1] = USCj * Status.EXTRAPOLATE_TO_YEAR
        Status.int.QWHj[cascadeIndex - 1] = QWHj * Status.EXTRAPOLATE_TO_YEAR
        Status.int.QHXj[cascadeIndex - 1] = 0

        Status.int.FETFuel_j[cascadeIndex - 1] = 0.0
        Status.int.FETel_j[cascadeIndex - 1] = FETel_j
        Status.int.ElGen_j[cascadeIndex - 1] = 0.0
        Status.int.ElGen_t[cascadeIndex - 1] = Status.int.createQ_t()
        Status.int.FETHeat_j[cascadeIndex-1] = FETHeat_j
        Status.int.waterConsumption_j[cascadeIndex - 1] = waterConsumption
        Status.int.HPerYearEq[cascadeIndex - 1] = HPerYear * Status.EXTRAPOLATE_TO_YEAR

        logTrack("ModuleCH: eq.no.:%s USC: %s FETFuel: %s FETel: %s FETh: %s HPerYear: %s [MWh]" % \
                   (equipe.EqNo, \
                    USCj * Status.EXTRAPOLATE_TO_YEAR / 1000.0, \
                    0.0, \
                    FETel_j * Status.EXTRAPOLATE_TO_YEAR / 1000.0, \
                    FETHeat_j, \
                    HPerYear))

        return (USCj_Tt, None)

    def estimateEconomicParameters(self, eq):
        """ Default values for economic parameters
        """

        if hasattr(eq, "eqType"): # is an eqObject        
            eqType = eq.eqType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OMFix
            OMVar = eq.OMVar

            Pnom = eq.Pnom
            
        else:   # kept for backward compatibility - still used in setEquipmentFromDB
            eqType = eq.EquipType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OandMfix
            OMVar = eq.OandMvar

            Pnom = eq.HCGPnom
            

        # security check if the reference power is defined
        Pref = Pnom 
        if Pref is None:
            return (None, None, None)

        if turnKey is None:
            if hasattr(eq, "Price") and eq.Price is not None:
                turnKey = max(2.*eq.Price,eq.Price+40*Pnom)

            else:
                turnKey = 300*Pnom
            
        OMFix = noneCheck(OMFix, 0.005 * turnKey)
        OMVar = noneCheck(OMVar, 0.04)      

        return (turnKey, OMFix, OMVar)
