# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleCHP (Combined Heat and Power)
#			
#------------------------------------------------------------------------------
#			
#	Module for calculation of CHP equipment
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
import math
from auxiliary.auxiliary import findInListASC, noneFilterNumber
from GUI.status import *
from modules.fluids import Fuel
from constants import *
from modules.modulePipes import getAllHeatingPipes
from modules.heatFlow import HeatFlow_t, diffQ_Tt
from modules.interfaces import createQ_t, createQ_T, createQ_Tt
from modules.hf.heatFlowQ_T import match_heatFlows
from modules.messageLogger import *
from lib.auxiliary.auxiliary import noneCheck
import database.dbio as dbio

CHPMAXTEMP = {           #temperature level of supply (circuit 1 and 2, + energy fraction circuit 1)
    "CHP engine": (95,200,0.7),
    "CHP gas turbine": (250.0,0,1.0),
    "CHP steam turbine": (140,0,1.0),
    "CHP fuel cell": (250,0,1.0)
    }

CHPDESIGNTEMP = {           #temperature level of demand used for auto-design
    "CHP engine": 95.0,
    "CHP gas turbine": 250.0,
    "CHP steam turbine": 140,
    "CHP fuel cell": 250
    }

CHPPRIORITYRATIO = {           # ratio of power at HOpMin considered as equivalent
    "CHP engine": 1.0,
    "CHP gas turbine": 0.5,
    "CHP steam turbine": 0.5,
    "CHP fuel cell": 0.1
    }

CHPFUELTYPES = {           # possible fuel types
    "CHP engine":           ['NG','LPG','HFO','BUT','PRO','GO','BIOG','BIOF'],
    "CHP gas turbine":      ['NG','LPG','BUT','PRO','BIOG'],
    "CHP steam turbine":    ['NG','LPG','HFO','BUT','PRO','GO','BIOS','BIOG','BIOF'],
    "CHP fuel cell":        ['NG','LPG','BUT','PRO','BIOG']
    }

DEFAULTSAUTODESIGNCHP =  {"CHPMaintain":1,
                             "CHPType":"any",
                             "CHPFuelType":"NG",
                             "CHPHOp":2000.0,
                             "CHPEff":0.55}

TAU_CHP = 4.0 #default 4.0 hours of storage used
DEFAULT_EEEMIN = 0.6  #minimum instantaneous equivalent electrical efficiency
ETAREF = 0.9  #reference conversion efficiency for calculation of EEE

DEFAULT_TDRIVE = 100
DEFAULT_DTDRIVE = 5.0
DEFAULT_TDROP_COND = 10.0   # ad hoc reasonable estimate
DEFAULT_CPARTLOAD = 0.2195  # Enogia turbine data from ITHERM-Project
TDRIVERET_TFC = 25.0
DEFAULT_COPEX = 0.5

#============================================================================== 
#============================================================================== 
class ModuleCHP(object):
#============================================================================== 
#============================================================================== 

    CHPList = []
    
    def __init__(self, keys):
        self.keys = keys # the key to the data is sent by the panel

        self.neweqs = 0 #new equips added
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def initPanel(self):
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

        if Status.int.cascadeUpdateLevel < 0:
            Status.int.initCascadeArrays(0)
    
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def updatePanel(self):
#------------------------------------------------------------------------------
#       Here all the information should be prepared so that it can be plotted on the panel
#------------------------------------------------------------------------------

        logTrack("ModuleCHP: updatePanel")
#............................................................................................
# get list of equipments and boilers in the system
# and update the energetic calculation up to the level needed for representation

        (CHPList,CHPTableDataList) = self.screenEquipments()

        if Status.int.cascadeUpdateLevel < self.cascadeIndex or Status.mod.moduleEnergy.outputLevel < 2:
            Status.mod.moduleEnergy.runSimulation(last=self.cascadeIndex, outputLevel = 2)

#............................................................................................
# 1. List of equipments

        matrix = []
        for row in CHPTableDataList:
            matrix.append(row)

        data = np.array(matrix)

        Status.int.setGraphicsData('CHP Table',data)

#............................................................................................
# 2a. Preparing data

        if Status.SimTool == "Plus":
            QD80C  = filter_and_sort(Status.int.USHTotal_Lt[1],TAU_CHP/2).tolist()       
            QD120C = filter_and_sort(Status.int.USHTotal_Lt[2],TAU_CHP/2).tolist()
            QD250C = filter_and_sort(Status.int.USHTotal_Lt[3],TAU_CHP/2).tolist()
            QDTot  = filter_and_sort(Status.int.USHTotal_Lt[0],TAU_CHP/2).tolist()

            if len(self.CHPList) > 0:
                QCHP = Status.int.USHj_t[self.cascadeIndex-1].tolist()
                QCHP.sort(reverse=True)
            else:
                QCHP = createQ_t()
        else:
            if len(self.CHPList) > 0:
                QCHP = Status.int.USHj_Tt[self.cascadeIndex-1][Status.NT+1].tolist()
                QCHP.sort(reverse=True)
            else:
                QCHP = createQ_t()

            QD80C  = filter_and_sort(Status.int.USHTotal_Tt[int(80/Status.TemperatureInterval+0.5)],TAU_CHP/2).tolist()       
            QD120C = filter_and_sort(Status.int.USHTotal_Tt[int(120/Status.TemperatureInterval)],TAU_CHP/2).tolist()
            QD250C = filter_and_sort(Status.int.USHTotal_Tt[int(250/Status.TemperatureInterval)],TAU_CHP/2).tolist()
            QDTot  = filter_and_sort(Status.int.USHTotal_Tt[Status.NT+1],TAU_CHP/2).tolist()

#............................................................................................
# 2b. XY Plot
        TimeIntervals=[Status.TimeStep*(it+1)*Status.EXTRAPOLATE_TO_YEAR for it in xrange(Status.Nt)]
        Status.int.setGraphicsData('CHP Plot',[TimeIntervals,
                                                        QCHP,
                                                        QD80C,
                                                        QD120C,
                                                        QD250C,
                                                        QDTot])

#............................................................................................
# 3. Configuration design assistant

        config = self.getUserDefinedPars()
        Status.int.setGraphicsData('CHP Config',config)
        
#............................................................................................
# 4. additional information (Info field right side of panel)

        info = []

        self.operatingHours = "-"
        self.effOperatingHours = "-"
        self.eee = "-"
        self.utilisationRatio = "-"

        if len(CHPList) > 0 and self.cascadeIndex is not None:
##            eq = Status.prj.getEquipe(cascadeIndex=self.cascadeIndex)
            eq = Status.int.systemModel.equipment[self.cascadeIndex-1]
            if eq is not None:
                self.operatingHours = eq.HPerYearEq
                self.eee = effectiveElEfficiency(eq.FETFuel_j,eq.FETel_j,eq.USHj)
                try:
                    self.effOperatingHours = eq.FETFuel_j*eq.HCGTEfficiency/eq.HCGPnom
                except:
                    self.effOperatingHours = 0.
                try:
                    self.utilisationRatio = eq.USHj / (eq.FETFuel_j*eq.HCGTEfficiency)
                except:
                    self.utilisationRatio = 0.

        self.minHOp = 2000.0
        
        info.append(self.effOperatingHours)  #first value to be displayed
        info.append(self.eee)

        indexMinHOp = (int) (self.minHOp / (Status.TimeStep*Status.EXTRAPOLATE_TO_YEAR))
        indexMinHOp = min(indexMinHOp,Status.Nt-1)
        
        info.append(max(0,QD80C[indexMinHOp]))
        info.append(max(0,QD120C[indexMinHOp]))
        info.append(max(0,QD250C[indexMinHOp]))
        info.append(max(0,QDTot[indexMinHOp]))

        info.append(self.utilisationRatio)

        Status.int.setGraphicsData('CHP Info',info)

        self.estimateTariff()

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def getUserDefinedPars(self):
#------------------------------------------------------------------------------
#   gets the user defined data from UHeatPump and stores it to interfaces to be shown in HP panel
#------------------------------------------------------------------------------

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)

            config = [1,"any",'NG',2000.0,0.55]            
            Status.int.setGraphicsData('CHP Config',config)

            self.setUserDefinedPars()

        else:
            u = urow
            config = [u.CHPMaintain,
                      u.CHPType,
                      u.CHPFuelType,
                      u.CHPHOp,
                      u.CHPEff]
            Status.int.setGraphicsData('CHP Config',config)

        self.CHPMaintain = config[0]
        self.CHPType = config[1]
        self.fuelType = config[2]
        self.minHOp = config[3]

        return config

#------------------------------------------------------------------------------
    def setUserDefinedPars(self):
#------------------------------------------------------------------------------

        config = Status.int.GData['CHP Config']

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        
        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
            
        u = urow

        u.CHPMaintain = dbio.check(config[0])
        u.CHPType = dbio.check(config[1])
        u.CHPFuelType = dbio.check(config[2])
        u.CHPHOp = dbio.check(config[3], string=False)
        u.CHPEff = dbio.check(config[4], string=False)

        self.minHOp = dbio.check(config[3])

        dbio.commit()

#------------------------------------------------------------------------------
    def screenEquipments(self,setIndex = True):
#------------------------------------------------------------------------------
#       screens existing equipment, whether there are already heat pumps
#------------------------------------------------------------------------------
        Status.int.systemModel.build()
        self.equipment = Status.int.systemModel.equipment
        Status.int.getEquipmentCascade()
        self.NEquipe = len(self.equipment)

        self.CHPList = []
        maxIndex = 0
        index = 0
        for row in Status.int.cascade:
            index += 1
            if getEquipmentClass(row["equipeType"]) == "CHP":
                self.CHPList.append(row)
                maxIndex = index

        if setIndex == True:
            self.cascadeIndex = maxIndex

        CHPTableDataList = []

        for jc,row in enumerate(Status.int.EquipTableDataList):
            
            if getEquipmentClass(row[3]) == "CHP":
                CHPTableDataList.append(row)
                jchp = len(CHPTableDataList)-1
                eq = self.equipment[jc]
                try:    #security checks for avoiding crash in case of corrupt equipment data
                    CHPTableDataList[jchp][1] = "%6.0f/%6.0f"%(eq.ElectriProduction,eq.HCGPnom,)
                except:
                    logTrack("Error: corrupt equipment data %s Pel %s Pth %s"%(eq.Equipment,eq.ElectriProduction,eq.HCGPnom))
                    CHPTableDataList[jchp][1] = "n/a / n/a"
                    
                try:
                    CHPTableDataList[jchp][2] = "%5.3f/%5.3f"%(eq.HCGEEfficiency,eq.HCGTEfficiency)  
                except:
                    logTrack("Error: corrupt equipment data %s Pel %s Pth %s"%(eq.Equipment,eq.ElectriProduction,eq.HCGPnom))
                    CHPTableDataList[jchp][1] = "n/a / n/a"

        #screen list and substitute None with "not available"
        for i, row in enumerate(CHPTableDataList):
            for j, item in enumerate(row):
                CHPTableDataList[i][j] = noneFilterNumber(item, _('not available'))
                
            
#        print"ModuleCHP (screenEquipments): the list of CHPs is:",self.CHPList
        return (self.CHPList,CHPTableDataList)
        

        
#------------------------------------------------------------------------------
    def getEqId(self,rowNo):
#------------------------------------------------------------------------------
#   gets the EqId from the rowNo in the CHPList
#------------------------------------------------------------------------------

        CHPId = self.CHPList[rowNo]["equipeID"]
        return CHPId

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def deleteEquipment(self,rowNo,automatic=False):
#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
        if automatic == False:
            if rowNo == None:   #indicates to delete last added equipment dummy
                CHPid = self.dummyEqId
            else:
            #--> delete CHP from the equipment list under current alternative #from C&QGenerationHC under ANo
                CHPid = self.getEqId(rowNo)
        else:
            CHPid= rowNo
        
        Status.prj.deleteEquipment(CHPid)
        self.screenEquipments()

#------------------------------------------------------------------------------
    def addEquipmentDummy(self):
#------------------------------------------------------------------------------
#       adds a new dummy equipment to the equipment list and returns the
#       position in the cascade 
#------------------------------------------------------------------------------
        self.equipe = Status.prj.addEquipmentDummy()
        self.dummyEqId = self.equipe.id

        if self.equipe.CascadeIndex > 1:
            Status.mod.moduleHC.cascadeMoveToTop(self.equipe.CascadeIndex) 
        
        self.equipe = Status.prj.getEquipe(self.dummyEqId) #after having moved, the equipe - table has to be updated !!!
        self.cascadeIndex = self.equipe.CascadeIndex

        self.neweqs += 1 #No of last equip added
        NewEquipmentName = "New CHP %s"%(self.neweqs)

        # connect the dummy to all heating pipes
        dummy, pipeIDs = getAllHeatingPipes()
        pipeIDs = [str(id) for id in pipeIDs ]
        IDString = ";".join(pipeIDs)
        
        equipeData = {"Equipment":NewEquipmentName, 
                      "EquipType": CHPTYPES[0],
                      "PipeDuctEquip":IDString}
        self.equipe.set(**equipeData)
        dbio.commit()
                
        return(self.equipe)

#------------------------------------------------------------------------------

    def setEquipmentFromDB(self,equipe,modelID):
        """
        takes an equipment model from the data base (dbchp) and stores it under a given Id in
        the equipment data base

        :params equipe: psql-dictionary corresponding to equipment dummy
        :params modelID: ID of equipment model in database dbchp
        """
        
        model = dbio.get("dbchp", id=modelID)

        if model.CHPPt is not None: equipe.HCGPnom = model.CHPPt
        if model.CHPPe is not None: equipe.ElectriProduction=model.CHPPe
        equipe.ElectriConsum=0.0
        if model.FuelConsum is not None: equipe.FuelConsum=model.FuelConsum
        
        if model.Eta_t is not None:
            if model.Eta_t > 1.0 and model.Eta_t < 100.0:
                logTrack("ModuleCHP: Efficiency data should be stored internally as fractions of 1")
                eff = model.Eta_t/100.0
                model.Eta_t=eff
            else:
                eff = model.Eta_t
            equipe.HCGTEfficiency=eff
            
        if model.Eta_e != None:
            if model.Eta_e > 1.0 and model.Eta_e < 100.0:
                logTrack("ModuleCHP: Efficiency data should be stored internally as fractions of 1")
                eff = model.Eta_e/100.0
                model.Eta_t=eff
            else:
                eff = model.Eta_e
            equipe.HCGEEfficiency=eff

        equipmentType = getEquipmentType("CHP",model.Type)
        
        if model.TSupply is not None:    
            equipe.TMaxSupply=model.TSupply
        else:
            if equipmentType in CHPMAXTEMP:
                equipe.TMaxSupply=CHPMAXTEMP[equipmentType][0]
            else:
                logError("Erroneous equipment type in CHP database: %s"%equipmentType)
                equipe.TMaxSupply=CHPMAXTEMP["CHP engine"][0]
                

        if model.TSupply2 is not None:    
            equipe.TMaxSupply2=model.TSupply2
        
        equipe.Manufact=model.Manufacturer
        equipe.EquipType=equipmentType
        equipe.NumEquipUnits=1
        equipe.EquipTypeFromDB=model.SubType
        if model.id != None: equipe.EquipIDFromDB=model.id
        equipe.DBFuel_id=1  #use Natural Gas as default -> should later on be adjusted to type of equipment


#HS 2008-10-25: equipment parameters that are set defined by default if not specified
        fuel_number = equipe.DBFuel_id   #IMPORT from the fuelDB
        eq_fuel = Fuel(fuel_number)

        try:
            equipe.FuelConsum = equipe.HCGPnom/(eq_fuel.LCV * equipe.HCGTEfficiency)
        except:
            pass

        try:
#            equipe.FlowExhaustGas = equipe.FuelConsum*(1.0 + eq_fuel.CombAir*equipe.ExcessAirRatio)
            equipe.FlowExhaustGas = 0.0
        except:
            pass

        try:
            equipe.ElectriConsum = 0.0*equipe.HCGPnom
        except:
            pass

        try:
#            LossFactEq = 0.01
#            TExhaustGas = eq_fuel.LCV*(1.0 - equipe.HCGTEfficiency - equipe.HCGEEfficiency - LossFactEq) \
#                          /((1.0 + eq_fuel.CombAir) * equipe.ExcessAirRatio * eq_fuel.OffgasHeatCapacity)
#            equipe.TExhaustGas = TExhaustGas
            equipe.TExhaustGas = 0.0
        except:
            pass


        equipe.set(**{"TurnKeyPrice": model.InvRate,
                       "OandMfix": model.OMRateFix,
                       "OandMvar": model.OMRateVar})
        (turnKey, OMFix, OMVar) = self.estimateEconomicParameters(equipe)
        equipe.set(**{"TurnKeyPrice": turnKey,
                       "OandMfix":OMFix,
                       "OandMvar":OMVar})

        dbio.commit()
        logTrack("moduleCHP (setEquipmentFromDB): CHP system added:'%s',type:'%s',Pow:'%s',T'%s'"%\
                 ("---",model.CHPequip,model.CHPPt,-1.0))

    def calculateEnergyFlows(self, equipe, demand_Tt, cascadeIndex):
        """Calculate the energy flows in the equipment identified by 
        "cascadeIndex" 
        
        :returns: a pair consisting of the time and temperature 
        dependent numpy array containing the covered demand. Also known
        as USHj_Tt of the given equipment and the same data aggregated 
        over time and temperature, USHj.

        """
        jc = cascadeIndex-1

        if Status.int.cascadeUpdateLevel < (cascadeIndex - 1):
            logDebug("ModuleCHP (calculateEnergyFlows): cannot calculate without previously updating the previous levels")
            Status.mod.moduleEnergy.runSimulation(last=(cascadeIndex-1))
        Status.int.extendCascadeArrays(cascadeIndex)

        if cascadeIndex > 0 and cascadeIndex <= Status.NEquipe:
            logTrack("ModuleCHP (calculateEnergyFlows): starting (cascade no: %s)"%cascadeIndex)
        else:
            logError("ModuleCHP (calculateEnergyFlows): cannot simulate index %s: out of cascade [%s]"%\
                     (cascadeIndex,Status.NEquipe))
            return (createQ_Tt(), 0)

        # get equipment data from equipment list in SQL
        CHPModel = equipe.Model
        CHPType = equipe.EquipType
        CHPSubType = getEquipmentSubClass(CHPType)
        EquipmentNo = equipe.EqNo
        j = cascadeIndex - 1
        controlSchedule = Status.int.systemModel.eqObject[cascadeIndex-1].controlSchedule
        control = controlSchedule.schedule if controlSchedule else 1.0

        controlByHeatDemand = (CHPType not in ["CHP ORC", "CHP TFC"])

        if CHPType in ["CHP ORC", "CHP TFC"]:
            PNom_el = equipe.ElectriProduction
            COPe = equipe.HCGEEfficiency
            COPh_nom = 1. - COPe
            TDriveNom = equipe.THeatSourceHT
            if TDriveNom is None: TDriveNom = DEFAULT_TDRIVE
            heatSourceHT = (equipe.HeatSourceHT is not None)

            TDrive = Status.int.systemModel.eqObject[jc].TDrive
            if TDrive is None: TDrive = np.ones(Status.Nt)*TDriveNom
            
            if COPe > 0:
                PNom = PNom_el * COPh_nom / COPe
            else:
                # This is a crash case. Error warning is emitted below
                PNom = 0

            Trecool = Status.int.systemModel.eqObject[jc].TReject
            
            if Trecool is None:
                Trecool = Status.WeatherData.Tdb.copy()

            TrecoolAvg = np.mean(Trecool)
            TrecoolNom = equipe.TemperatureReCooling
            if TrecoolNom is None:
                TrecoolNom = TrecoolAvg

            COPex = equipe.HPExHeatCOP
            
            if COPex is None:  # set to default values if nothing is specified in equipment tables
                COPex = estimateCOPex(COPe, TDriveNom, TrecoolNom)

                if COPex is None:  # still None -> estimate from user input was not possible or not realistic
                    COPex = DEFAULT_COPEX

            fcorr_T = np.zeros(Status.Nt)
            if COPe > 0:
                for it in xrange(Status.Nt):
                    COPe_t = COPex*COPCarnot(TDrive[it], Trecool[it])
                    fcorr_T[it] = min(max(0.5 + 0.5*COPe_t/COPe, 0.8), 1.2) # smooth and constrain correction

            cPartLoad = equipe.ST_C0
            if cPartLoad is None: cPartLoad = DEFAULT_CPARTLOAD

        else:
            PNom = equipe.HCGPnom
            COPh_nom = equipe.HCGTEfficiency
            COPe = equipe.HCGEEfficiency

        EEEmin = equipe.ST_C1
        if EEEmin is None: EEEmin = DEFAULT_EEEMIN
        
        TMax = equipe.TMaxSupply
        TMax2 = equipe.TMaxSupply2

        if TMax is None:
            if TMax2 is not None:
                TMax = TMax2
                TMax2 = None

        elif TMax2 is not None: # both temperatures are defined, order so that TMax < TMax2
            if TMax > TMax2:    
                TMaxNew = TMax2
                TMax2 = TMax
                TMax = TMaxNew
                

        if PNom is None:
            PNom = 0.0
            logWarning("ModuleCHP (calculateEnergyFlows): No nominal power specified for equipe no. %s"%\
                 (EquipmentNo))

        if COPe is None:
            COPe = 0.0
            logWarning("ModuleCHP (calculateEnergyFlows): No electrical efficiency specified for equipe no. %s"%\
                 (EquipmentNo))

        if COPh_nom > 0:
            PNom_fuel = PNom/COPh_nom
            PNom_el = PNom_fuel*COPe
            if COPe > EEEmin:
                fu_min = 0
            else:
                fu_min = ETAREF/COPh_nom*(1-COPe/EEEmin)

            if fu_min > 1:
                showWarning("ModuleCHP (calculateEnergyFlows): Global efficiency (electrical and thermal) of CHP equipment too low. Cannot fulfil minimum EEE = %s"%EEEmin)

        else:
            PNom_fuel = 0.0
            PNom_el = 0.0
            if COPe > EEEmin:
                fu_min = 0
            else:
                fu_min = INFINITE

        if TMax is None:
            if CHPSubType in CHPMAXTEMP.keys():
                TMax = CHPMAXTEMP[CHPSubType][0]
            else:
                TMax = 95

        if TMax2 is None:
            if CHPSubType in CHPMAXTEMP.keys():
                TMax2 = CHPMAXTEMP[CHPSubType][1]
            else:
                TMax2 = 95

        if CHPSubType in CHPMAXTEMP.keys():
            fCircuit1 = CHPMAXTEMP[CHPSubType][2]
        else:
            fCircuit1 = 1.0

            logDebug("ModuleCHP (calculateEnergyFlows): no Tmax specified for equipe no. %s"%EquipmentNo)
  
        logTrack("ModuleCHP (calculateEnergyFlows): Model = %s Type = %s PNom = %s"%\
                 (CHPModel,CHPType,PNom))

        # get demand data for CascadeIndex/EquipmentNo from Interfaces
        # and create arrays for storing heat flow in equipment
        QD_Tt = demand_Tt.copy()

        #get connected HX(es)

        if CHPType in ["CHP ORC", "CHP TFC"]:
            HXList = Status.QHXEqLink[j][:]

            connectedHX = []
            drivenByWH = False
            for h,hxCon in enumerate(HXList):
                if hxCon == 1:
                    connectedHX.append(h)
                    drivenByWH = True
        else:
            drivenByWH = False


        if drivenByWH:    #special block for waste heat driven equipment (ORC, ...)
            qin = HeatFlow_t()

            if Status.HRTool != HRMODES[2]:
                logWarning("Connection of heat exchangers as heat source is only possible in HR mode ActualHXNet V2")
                
            for h in connectedHX:
                try:
                    qin.add(Status.int.hx[h].qhx)
                except:
                    logDebug("ModuleHP: Could not add qhx of hx[%s]"%(h+1))

            QHXj_Tt = qin
##            QHXj_Tt = qin.Q_Tt()

        QA_Tt = Status.int.QA_Tt_mod[cascadeIndex-1].copy()

        USHj_Tt = Status.int.createQ_Tt()        
        QWHj_Tt = Status.int.createQ_Tt()
        ElGen_t = Status.int.createQ_t()

        # define temperature profile of heat that can be supplied, depending on the subtype
        dUSHMax_T = Status.int.createQ_T()

        if CHPSubType   == "CHP engine":
            TMin = TMax - 20
            TMin2 = 0
        elif CHPSubType == "CHP steam turbine":
            TMin = TMax - 0.1
            TMin2 = 0
        elif CHPSubType == "CHP gas turbine":
            TMin = 20
            TMin2= 0
        elif CHPSubType == "CHP fuel cell":
            TMin = TMax - 20
            TMin2 = 0
        else:
            TMin = TMax - 20
            TMin2 = 0

        tauStorage = TAU_CHP    #default: 4 hours of storage supposed
        
        DT = max(TMax-TMin,1.e-10)  #security: avoid division by zero
        DT2 = max(TMax2-TMin2,1.e-10)
        dQMax = PNom * Status.TimeStep
            
        for iT in range(Status.NT+2):
            T = iT*Status.TemperatureInterval
            if T < TMin:                    #contribution first circuit
                dUSHMax_T[iT] = 0.0
            elif T <= TMax:
                f = (T-TMin)/DT
                dUSHMax_T[iT] = f*dQMax*fCircuit1
            else:
                dUSHMax_T[iT] = dQMax*fCircuit1
                
            if T >= TMin2 and T<= TMax2:    #contribution second circuit
                f2 = (T - TMin2)/DT2
                dUSHMax_T[iT] += f2*dQMax*(1.-fCircuit1)
            elif T > TMax2:
                dUSHMax_T[iT] += dQMax*(1.-fCircuit1)

        if DEBUG != "OFF":
            logTrack("dUSHMax_T = %s"%dUSHMax_T)
        
        QStorageMax_T = dUSHMax_T * tauStorage / Status.TimeStep
        QStorage_T = 1 * QStorageMax_T
        QDStorage_T = createQ_T()
        
        # Start hourly loop
        USHj = 0
        FETel_j = 0
        FETFuel_j = 0
        FETHeat_j = 0
        HPerYear = 0
        QD = 0
        hoursConstrained = 0

        QDeff_T = Status.int.createQ_T()
        dUSHeff_T = Status.int.createQ_T()

        if controlByHeatDemand:
            for it in xrange(Status.Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]
                
            # Calculate heat delivered by the given equipment for each time interval
                QDeff_T = QD_Tt[:, it] + QDStorage_T
                
                if QDeff_T[Status.NT+1] == 0 or fctrl==0:
                    f = 0.0
                    QWH_T = Status.int.createQ_T()
                    dUSHact_T = Status.int.createQ_T()

                else:
                    f = 1.0        #assure that while loop does at least one cycle !   
                    fu = -1.
                    nits = 0
                    
                    while fu < fu_min and nits < 5:
                        
                        dUSHeff_T = f * fctrl * dUSHMax_T
                        shift = np.max(dUSHeff_T - QDeff_T) # determine pinch shift for present time step
                        shift = max(0.0, shift)

                        fup = fu
                        if dUSHeff_T[Status.NT+1] > 0:
                            fu = 1. - shift/dUSHeff_T[Status.NT+1]
                        else:
                            fu = 1.0
                            
                        if fu < fu_min:
                            if f <= 0.2:
                                f = 0
                                dUSHeff_T = Status.int.createQ_T()
                                break
                            
                            f *= fu/fu_min
                            
                            if f <= 0.25*0.2:   # if new estimate far below minimum, then don't even try
                                f = 0
                                dUSHeff_T = Status.int.createQ_T()
                                break
                            
                            else:
                                f = max(0.2,f)  # do not operate at less than 20% part load

                        if fu <= fup:
                            f = 0   # No improvement obtained by reducing partload factor
                            dUSHeff_T = Status.int.createQ_T()
                            break

                        nits += 1

                    if DEBUG != "OFF" and it < 48:
                        logTrack("ModuleCHP: it %s - CHP operating at part load f = %s with utilisation factor %s (%s)"%(it,f*fctrl,fu,fu_min))
                    if f == 0: hoursConstrained +=1

                    dUSHact_T,QWH_T = match_heatFlows(dUSHeff_T,QDeff_T)

                dUSHj_T, QStorage_T = match_heatFlows(dUSHact_T + QStorage_T, QD_Tt[:, it])
                QStorage_T = QStorage_T[0] - QStorage_T
                USHj_Tt[:,it] = dUSHj_T
                QDStorage_T = QStorageMax_T - QStorage_T
##                
##                USHj_Tt[:, it] = np.min((QD_Tt[:, it], dUSHact_T + QStorage_T), 0)
##                if it==4: print "ModuleCHP: dUSHj_T preliminary: %s"%[USHj_Tt[iT][it] for iT in xrange(Status.NT+2)]
##                for iT in xrange (Status.NT+2):
##                    if iT > 0:
##                        #1. check: gradient of dUSH cannot be higher as gradient of demand
##                        dQD = QD_Tt[iT][it] - QD_Tt[iT-1][it]
##                        USHj_Tt[iT][it] = min(USHj_Tt[iT][it],USHj_Tt[iT-1][it]+dQD)
##                    
##                    dQStorage = dUSHact_T[iT] - USHj_Tt[iT][it]             # production to storage
##                    if iT > 0:
##                        # thermodynamic limitation of storage discharge: gradient after discharge has still to be positive
##                        dQStorage = max(dUSHact_T[iT] - USHj_Tt[iT][it],QStorage_T[iT-1]-QStorage_T[iT]) 
##                        QStorage_T[iT] += dQStorage
##                        USHj_Tt[iT][it] = dUSHact_T[iT] - dQStorage
##
##                    QDStorage_T[iT] -= dQStorage
                
                QD_Tt[:, it] -= USHj_Tt[:, it]
                QWHj_Tt[:, it] = QWH_T.copy()

                    
                USHj += USHj_Tt[Status.NT+1][it]

                dElGen = f*PNom_el*Status.TimeStep
                ElGen_t[it] = dElGen
                Status.int.FETel_j_t[j][it] = -dElGen
                Status.int.ElGen_t[j][it] = dElGen
                
                FETel_j -= dElGen
                FETFuel_j += f*PNom_fuel*Status.TimeStep
                
                if dUSHact_T[Status.NT+1] > 0:
                    HPerYear += Status.TimeStep

                QD += QD_Tt[Status.NT+1][it]

            ElGen_j = -FETel_j

        else:   # special case CHP ORC / TFC controlled by available waste heat
            dQin_max = PNom_fuel*Status.TimeStep
            QDriveReq_t = np.array([dQin_max]*Status.Nt)*control
            QWHEq_t = Status.int.createQ_t()

            TDriveRet = TDrive-DEFAULT_DTDRIVE if CHPType=="CHP ORC" else np.ones(Status.Nt)*TDRIVERET_TFC

            if heatSourceHT:    # connected to an active external heat supply

                QDriveReq_Tt = HeatFlow_t(fluid_id=0,T0=TDriveRet,T1=TDrive,Q_t = QDriveReq_t)

                if drivenByWH:  # additionaly connected to HX as second heat source
                    FETHeat_Tt, res = diffQ_Tt(QDriveReq_Tt, QHXj_Tt)
                    QHXj = QHXj_Tt.Q

                else:
                    FETHeat_Tt = QDriveReq_Tt
                    QHXj = 0

                # part load correction                
                fth = control
                fPartLoad = (fth - cPartLoad) / (1. - cPartLoad)
                fPartLoad = np.maximum(fPartLoad, 0.0)
                dElGen = PNom_el*Status.TimeStep*fPartLoad*fcorr_T
                dQWH = dQin_max*control - dElGen
                ElGen_t = dElGen
                
##                if control==1:
##                    QWHEq_t = np.array([dQWH]*Status.Nt)
##                    
##                    FETel_j = - dElGen*Status.Nt
##                    FETHeat_j = FETHeat_Tt.Q
##                    HPerYear = Status.Nt
##
##                else:
                QWHEq_t = dQWH
                FETel_j = -np.sum(dElGen)
                FETHeat_j = FETHeat_Tt.Q
                HPerYear = int(np.sum((dElGen > 0)))

                Status.int.FETel_j_t[j] = -dElGen
                Status.int.ElGen_t[j] = dElGen
                

            elif drivenByWH:               # connected to HX as only heat source

                QHXj = 0

                
                Tin = QHXj_Tt.getT1()
                Qreq = Status.int.QHXEqReq_Tt[cascadeIndex-1].getQ()
            
                for it in xrange(Status.Nt):
                    fctrl = 1.0 if isinstance(control, float) else control[it]
                    
                    dQin = min(dQin_max*fctrl, QHXj_Tt.getQ(it))
                    
                    #TODO: introduce control of temperature level of available heat

                    f = dQin / dQin_max

                    # part load correction                
                    fth = f
                    fPartLoad = (fth - cPartLoad) / (1. - cPartLoad)
                    fPartLoad = np.maximum(fPartLoad, 0.0)
                    dElGen = fPartLoad*PNom_el*Status.TimeStep*fcorr_T[it]
                    dQWH = dQin - dElGen
                    QWHEq_t[it] = dQWH
                    
                    FETel_j -= dElGen
                    ElGen_t[it] =dElGen
                    QHXj += dQin
                    
                    if dElGen > 0:
                        HPerYear += Status.TimeStep

                    if Tin[it] < TDrive[it] and Qreq[it] > 0 and fctrl > 0: # adapt QDriveReq in order to achieve TDrive
                        QreqNew = max(Qreq[it]*(Tin[it] - TDriveRet[it])/(TDrive[it] - TDriveRet[it]), 0.0)
                        QDriveReq_t[it] = min(QDriveReq_t[it], QreqNew)
                        #TODO: this may lead to oscillations in the general non-linear case (e.g. phase change)

                    Status.int.ElGen_t[j][it] = dElGen
                    Status.int.FETel_j_t[j][it] = -dElGen

                    if it in SNAPSHOTS:
                        print "Equipment %s at it=%s:"%(j, it)
                        print "fcorr_T: %s"%fcorr_T[it]
                        print "fPartLoad: %s"%fPartLoad
                        print "Tin: %s"%Tin[it]
                        print "QDriveReq: %s"%QDriveReq_t[it]

                QDriveReq_Tt = HeatFlow_t(fluid_id=0,T0=TDriveRet,T1=TDrive,Q_t = QDriveReq_t)
                                            
            else:   # this case should not occur, but for avoiding crashes
                QWHEq_t = np.zeros(Status.Nt)

                dElGen = 0
                FETel_j = 0
                FETHeat_j = 0
                HPerYear = 0
                QHXj = 0
                

            ElGen_j = -FETel_j
            
        # End of year reached. Store results in interfaces
        # remaining heat demand and availability for next equipment in cascade
        Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
        Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)
        
        # heat delivered by present equipment
        Status.int.USHj_Tt[cascadeIndex-1] = USHj_Tt
        Status.int.USHj_T[cascadeIndex-1] = Status.int.calcQ_T(USHj_Tt)
        Status.int.USHj_t[cascadeIndex-1] = USHj_Tt[Status.NT+1].copy()
        Status.int.USHjChanged[cascadeIndex-1] = True

        # electricity generated
        Status.int.ElGen_t[cascadeIndex-1] = ElGen_t

        # waste heat absorbed by present equipment
        if drivenByWH:
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = QDriveReq_Tt
            Status.int.QHXEqReqChanged[cascadeIndex-1] = True
            
        else:
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = HeatFlow_t()   #empty HeatFlow_t object: no waste heat required
            Status.int.QHXEqReqChanged[cascadeIndex-1] = True


##        Status.int.QHXj_Tt[cascadeIndex-1] = QHXj_Tt
##        Status.int.QHXj_T[cascadeIndex-1] = Status.int.calcQ_T(QHXj_Tt)
##        Status.int.QHXj_t[cascadeIndex-1] = copy.deepcopy(QHXj_Tt[Status.NT+1])

        # waste heat generated by present equipment
        if CHPType in ["CHP ORC", "CHP TFC"]:
            Status.int.QWHEq_Tt[cascadeIndex-1] = HeatFlow_t(fluid_id=0, T0=TrecoolAvg, T1=TrecoolAvg+1.0, Q_t = QWHEq_t)
            QWHj_Tt = Status.int.QWHEq_Tt[cascadeIndex-1].Q_Tt(descending=True)
        else:  
            Status.int.QWHEq_Tt[cascadeIndex-1] = HeatFlow_t()
            #TODO: calculate real waste heat ... once this is used 

        Status.int.QWHj_Tt[cascadeIndex-1] = QWHj_Tt
        Status.int.QWHj_T[cascadeIndex-1] = Status.int.calcQ_T(QWHj_Tt)
        Status.int.QWHj_t[cascadeIndex-1] = QWHj_Tt[0].copy()
        Status.int.QWHjChanged[cascadeIndex-1] = True

        # high temperature heat used by present equipment
        if CHPType in ["CHP ORC", "CHP TFC"] and heatSourceHT:
            Status.int.FETHeat_j_Tt[cascadeIndex - 1] = FETHeat_Tt
            Status.int.FETHeat_jChanged[cascadeIndex - 1] = True

        Status.int.cascadeUpdateLevel = cascadeIndex

        #........................................................................
        # Global results (annual energy flows)
        
        ##        QD *= Status.EXTRAPOLATE_TO_YEAR
        ##        USHj *= Status.EXTRAPOLATE_TO_YEAR
        ##        HPerYear *= Status.EXTRAPOLATE_TO_YEAR
        
        Status.int.USHj[cascadeIndex-1] = USHj
        
        Status.int.FETFuel_j[cascadeIndex-1] = FETFuel_j
        Status.int.FETel_j[cascadeIndex-1] = FETel_j
        Status.int.ElGen_j[cascadeIndex-1] = ElGen_j
        Status.int.FETHeat_j[cascadeIndex-1] = FETHeat_j
        Status.int.HPerYearEq[cascadeIndex-1] = HPerYear
        
        Status.int.QWHj[cascadeIndex-1] = Status.int.QWHj_T[cascadeIndex-1][0]   # not considering the latent heat(condensing water)
        if drivenByWH:
            Status.int.QHXj[cascadeIndex-1] = QHXj
        else:
            Status.int.QHXj[cascadeIndex-1] = 0.0
            

        logTrack("CHP: eq.no.:%s energy flows [MWh] USH: %s FETFuel: %s FETel: %s QD: %s HPerYear: %s QWH %s hoursConstrained %s"%\
                   (equipe.EqNo,\
                    USHj/1000.0,\
                    FETFuel_j/1000.0,\
                    FETel_j/1000.0,\
                    QD/1000.0,\
                    HPerYear/1000.0,\
                    Status.int.QWHj[cascadeIndex-1]/1000.0, \
                    hoursConstrained))

        return (np.array(USHj_Tt), USHj)    

    def designAssistant1(self):
#------------------------------------------------------------------------------
#   auto-design of CHP plant - step 1: pre-selection
#------------------------------------------------------------------------------

#............................................................................................
# getting configuration parameters of DA

        DA = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        if DA is None:
            logTrack("ModuleCHP (design assistant): WARNING - no DA configuration parameters available")
            return

        if DA.CHPMaintain is not None:
            self.Maintain = DA.CHPMaintain
        else:
            self.Maintain = 1

        if DA.CHPHOp is not None:
            self.minHOp = DA.CHPHOp
        else:
            self.minHOp = 2000.0
        indexMinHOp = (int) (self.minHOp / (Status.TimeStep*Status.EXTRAPOLATE_TO_YEAR))
        indexMinHOp = min(indexMinHOp,Status.Nt-1)

        if DA.CHPEff is not None:
            self.minEff= DA.CHPEff
        else:
            self.minEff = 0.55

        if DA.CHPFuelType is not None:
            if DA.CHPFuelType == "any" or DA.CHPFuelType not in TRANSFUELTYPES.keys():
                self.fuelTypes = TRANSFUELTYPES.keys()
            else:
                self.fuelTypes = [DA.CHPFuelType]
        else:
            self.fuelTypes = TRANSFUELTYPES.keys()    # use any fuel
        
        if self.Maintain == 0:  #delete existing CHP equipments and design from scratch
            for row in self.CHPList:
                self.dummyEqId = row['equipeID']
                self.deleteEquipment(None)

#............................................................................................
# after shifting, the equipment cascade has to be updated, as the modified demand is used
# in the following

        if Status.int.cascadeUpdateLevel < self.NEquipe or Status.mod.moduleEnergy.outputLevel < 2:
            logTrack("ModuleCHP: running simulation before activating auto-design")
            Status.mod.moduleEnergy.runSimulation(outputLevel=2)

        QD = Status.int.USHTotal_Tt
###TODO: this works only for CHPMaintain = False. If there's already a CHP - equipment, the remaining demand after this equipment should be used as a base.
        
#............................................................................................
# get maximum temperature for heat demand
            
        self.CHPType = DA.CHPType
        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP (DA1): CHPType = %s FuelTypes = %s"%(self.CHPType,self.fuelTypes)

        if self.CHPType is None or self.CHPType == "any":    # autoselect appropriate CHP type if not pre-specified by user

            selectedType = None
            selectedQdotEquivalent = 0
            
            for tryCHPType in CHPDESIGNTEMP.keys():
                tryTmax = CHPDESIGNTEMP[tryCHPType]
                iTtry = max(min(int(tryTmax/Status.TemperatureInterval+0.5),Status.NT+1),0)

                if DEBUG in ("BASIC", "MAIN", "ALL"):
                    print "QD[iTtry] before sorting %s"%(max(QD[iTtry]))
                QD_try = filter_and_sort(QD[iTtry],TAU_CHP/2)
                QdotDmaxEquivalent = QD_try[indexMinHOp]/Status.TimeStep*CHPPRIORITYRATIO[tryCHPType]

                if DEBUG in ("BASIC", "MAIN", "ALL"):
                    print "ModuleCHP (DA): trying out type %s QdotMaxEq. = %s"%(tryCHPType,QdotDmaxEquivalent)
                    print "iTtry %s QD_try(0) %s QD_try(minHOp) %s"%(iTtry,QD_try[0],QD_try[indexMinHOp])

                fuelOK = False
                for fuelType in self.fuelTypes:
                    if fuelType in CHPFUELTYPES[tryCHPType]:
                        fuelOK = True
                        break
                if self.fuelTypes == ["any"]:
                    fuelOK = True
                
                if QdotDmaxEquivalent > selectedQdotEquivalent and fuelOK:
                    selectedQdotEquivalent = QdotDmaxEquivalent
                    selectedType = tryCHPType

            if selectedType is None:
                showMessage(_("no appropriate CHP system type found"))
                return ("CANCEL",[])
            else:
                self.CHPType = selectedType


        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "DA1: selected Type = %s",self.CHPType

        if self.CHPType in CHPDESIGNTEMP.keys():
            self.Tmax = CHPDESIGNTEMP[self.CHPType]
        else:
            logDebug("ModuleCHP: CHPType %s not found in CHPDESIGNTEMP"%self.CHPType)
            self.Tmax = 95.0

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP Type %s Tmax %s"%(self.CHPType,self.Tmax)
        
# Total demand: CHP by default as base load equipment TODO: if there are already existing CHP's, then better the remaining demand should be used
        iTmax = max(min(int(self.Tmax/Status.TemperatureInterval+0.5),Status.NT+1),0)
        QD_Tmax = filter_and_sort(QD[iTmax],TAU_CHP/2)
        QdotDmax = QD_Tmax[indexMinHOp]/Status.TimeStep

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "indexMinHOp %s QD %s"%(indexMinHOp,QdotDmax)

#............................................................................................
#   preselect list of equipment that fulfils the criteria

        self.preselection = []

#............................................................................................
# Initial selection: maximum reasonable power for a CHP Equipment
# From the annual Heat demand curve (QDa): calculate the necesary heating capacity (starting value)

        CHPList = self.getCHPList(self.CHPType, self.fuelTypes)

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP: found the following equipment for CHPType %s and FuelTypes %s\n"%(self.CHPType, self.fuelTypes)
            print "ModuleCHP: CHPList"
            for elem in CHPList:
                print elem
            print ""
        
        listIndexH = findInListASC(1.4*QdotDmax,CHPList[1])
        listIndexL = findInListASC(0.7*QdotDmax,CHPList[1])

        if listIndexH < 0:
            showWarning(_("There is no appropriate CHP available in DB for this application")+\
                        _("\n(thermal power approx) %s kW")%QdotDmax)
            self.preselection = []
            powers = []
            globalEffs = []
        else:
            self.preselection = CHPList[0][max(listIndexL,0):(listIndexH+1)]
            powers = CHPList[1][max(listIndexL,0):(listIndexH+1)]
            globalEffs = CHPList[2][max(listIndexL,0):(listIndexH+1)]

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP (DA): QdotMax = %s. Preselected equipment:\n"%QdotDmax
            print self.preselection

#............................................................................................
# Automatic final selection:
#............................................................................................

        nSelected = len(self.preselection)
        if nSelected <= 0: #no equipment could be selected
            return("CANCEL",[])
        
        else:   #several options possible -> manual or automatic final selection

            if Status.UserInteractionLevel == "interactive" or Status.UserInteractionLevel == "semi-automatic":

                    return("MANUAL",self.preselection)  #"MANUAL" indicates to panel to open the dialog

            else:
                listIndex = None
                maxPriority = 0
                for i, globalEff in enumerate(globalEffs):
                    priority = globalEff - 0.25*abs(powers[i]-QdotDmax)/max(QdotDmax,1.e-10) - 0.125*(powers[i]-QdotDmax)/max(QdotDmax,1.e-10)
                    if priority > maxPriority:
                        listIndex = i
                        maxPriority = priority

                if listIndex is None:
                    listIndex = findInListASC(QdotDmax,powers)
                    if listIndex < 0: listIndex = 0
                    elif listIndex >= nSelected: listIndex = nSelected-1
                
                finalSelection = self.preselection[listIndex]
                return ("AUTOMATIC",[finalSelection])
                                             
#------------------------------------------------------------------------------
    def designAssistant2(self,modelID):
#------------------------------------------------------------------------------
##        Design Assistant 2 - final selection
#------------------------------------------------------------------------------

        if (modelID == None or modelID < 0):
            logTrack("ModuleCHP (DA2): user cancelled the selection of the CHP equipment")

        else:                        
#............................................................................................
#   add a new equipment as space holder to the equipment list
#   and move it to the top of the list

            equipe = self.addEquipmentDummy()
            self.setEquipmentFromDB(equipe,modelID) #add selected equipment to the equipment list #SD change 30/04.2008
            logTrack("ModuleCHP (DA2): CHP equipment added. model no: %s"%modelID)

#------------------------------------------------------------------------------
    def getCHPList(self, CHPType, fuelTypes = []):
        """Return the lists (ascending order for CHPPt) that are 
        needed for the selection of a heat pump and the 
        calculateEnergyFlows()

        CHPList[0]: list of ID's
        CHPList[1]: list of nominal power (thermal)

        """
        CHPList = [[],[],[]]
        if CHPType is None or CHPType not in CHPTYPES:
            sqlQuery = "1=1 ORDER BY CHPPt ASC"
        else:
            sqlQuery = "Type LIKE '%s' ORDER BY CHPPt ASC"%(CHPType)
        subset = dbio.selectByQuery("dbchp", sqlQuery) #this is a list of rows (dictionaries) in DBCHP
        
        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP.getCHPList subset -> ",
            for elem in subset:
                print elem
        for i in range(len(subset)):
            if DEBUG in ("BASIC", "MAIN", "ALL"):
                print "Model %s name: %s FuelType %s"%(i+1,subset[i].CHPequip,subset[i].FuelType)

            if subset[i].FuelType in fuelTypes or fuelTypes == [] or subset[i].FuelType is None:
                CHPList[0].append(subset[i].id)
                CHPList[1].append(subset[i].CHPPt)
                try:
                    globalEfficiency = 2.5*subset[i].Eta_e + subset[i].Eta_t
                except:
                    globalEfficiency = 0
                CHPList[2].append(globalEfficiency)

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleCHP.getCHPList CHPList -> ",CHPList

        return CHPList


    def estimateEconomicParameters(self, eq):
        """ Default values for economic parameters
        """

        if hasattr(eq, "eqType"): # is an eqObject        
            eqType = eq.eqType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OMFix
            OMVar = eq.OMVar

            Pnom = eq.Pnom
            ElGen = eq.ElectriProduction
            
        else:   # kept for backward compatibility - still used in setEquipmentFromDB
            eqType = eq.EquipType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OandMfix
            OMVar = eq.OandMvar

            Pnom = eq.HCGPnom
            ElGen = eq.ElectriProduction
            

        # security check if the reference power is defined
        Pref = ElGen if eqType in ["CHP ORC", "CHP TFC"] else Pnom 
        if Pref is None:
            return (None, None, None)

        if eqType in ["CHP ORC", "CHP TFC"]:
            turnKey = noneCheck(turnKey, 2500 * ElGen)  # i-ThERM D2.2 (TATA): 1.427, D3.2: 2.200, 
            OMFix = noneCheck(OMFix, 0.02 * turnKey)
            OMVar = noneCheck(OMVar, 0.0)
        else:
            turnKey = noneCheck(turnKey, 700 * Pnom)
            OMFix = noneCheck(OMFix, 0.005 * turnKey)
            OMVar = noneCheck(OMVar, 0.04)      # estimation: 0.5 - 1 c€/kWhe = 0.025 - 0.05 c€/kWht
                                                # TODO: look for more accurate / documented values; independency of selected currency unit !!!

        return (turnKey, OMFix, OMVar)

    def estimateTariff(self):
        """ Sets default values for tariffs of self-generation and sale of electricity
            when not given by the user """
        
        el = Status.prj.getElectricity()
        eTariffCHP = el.ETariffCHP
        eTariffFeedIn = el.ElTariffFeedIn

        useDefaults = False
        message = ""
        
        if eTariffCHP is None:
            useDefaults = True
            message += _("No tariff for sales of self generated electricity given. Assumed equal to tariff on consumption\n")
            el.ETariffCHP = el.ElTariffCTot
            eTariffCHP = el.ElTariffCTot

        if eTariffFeedIn is None:
            useDefaults = True
            message += _("No feed-in tariff for self generated electricity given. Assumed equal to 0\n")
            el.ElTariffFeedIn = 0
            eTariffFeedIn = 0

        if useDefaults:
            showWarning(message)
            dbio.commit()

        return eTariffCHP, eTariffFeedIn

#------------------------------------------------------------------------------


def COPCarnot(Th, Tc):
    """Looks for the minimum temperature in cooling demand
    @param ThC,TcC: hot and cold temperature in Celsius
    @param TLevels: None: internal temperatures; "external": inlet to condenser, outlet from evaporator

    """

    if Th is None or Tc is None:
        return None
    
    ThK = Th + KELVIN
    TcK = Tc + KELVIN - DEFAULT_TDROP_COND

    if ThK > TcK:
        COP = (ThK - TcK) / ThK
    else:
        COP = 0

    return COP

def estimateCOPex(COP, Th, Tc):
    """estimates the exergetic efficiency (COPex) for a chiller with
    a given COP at nominal conditions Th and Tc
    @param Th,Tc: hot and cold temperature in Celsius
    @param TLevels: None: internal temperatures; "external": inlet to condenser, outlet from evaporator
    """

    if COP is None or Th is None or Tc is None:
        COPex = None
    else:
        COP_carnot = COPCarnot(Th, Tc)
        if COP > COP_carnot:
            COPex = 1.0
        else:
            COPex = COP / COP_carnot
    
    return(COPex)

#------------------------------------------------------------------------------
        
def filter_and_sort(QD,tau):
    """ auxiliary function that filters current demand by a time constant

    """
    QDf = Status.int.createQ_t()
    alpha = 1. - math.exp(-Status.TimeStep/tau)
    QDint = QD[0]
    for it in xrange(Status.Nt):
        QDint += alpha*(QD[it] - QDint)
        QDf[it] = QDint
    QDf = np.sort(QDf)[::-1]
        
    return QDf

#==============================================================================
        
def effectiveElEfficiency(FETFuel,FETel,USH):

    """
    auxiliary function that calculates the effective electrical efficiency
    
    :returns: effElEff
    :rtype:   double
    """

    if FETFuel is not None and USH is not None:
        FETeff = FETFuel - USH/0.9
        if FETel is not None and FETeff > 0:
            return - FETel / FETeff
        elif FETeff == 0:
            return 0
        else:
            return None
        
    else:
        return None


def checkAscending(Q_T,eps=1.e-6):
    ascending = True
    for iT in range(Status.NT+1):
        if Q_T[iT+1] < Q_T[iT] - eps:
            return False

    return True
    

    

if __name__ == "__main__":
    pass                    
