#!/usr/bin/env python
# -*- coding: utf-8 -*-
#==============================================================================
#
#   E I N S T E I N
#
#   Expert System for an Intelligent Supply of Thermal Energy in Industry
#   (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   CashFlow
#
#------------------------------------------------------------------------------
#
#   Calculation kernel of economics module
#   Carries out calculations based on a yearly cash flow
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import sys
from modules.messageLogger import *
from auxiliary.auxiliary import toGettextAndUnicode as _U
from GUI.status import Status
cu = Status.ProjectCurrency

class CashFlow:
    def __init__(self, maxYears, interestRate, energyPrizeRate):
        """
        Class for managing cash flow calculations
        Can be real or nominal cash flows corresponding to the appropriate values for interestRate and energyPriceRate

        Note: the functions addInvestment, addContingency ... suppose that for periodic payments all payments are equal
                = constan (real) cash flows (exception: energy prices which can vary at a rate different to the inflation rate)

        Definitions: (+) income (e.g. revenues);
                     (-) expenses (e.g. investment, energy costs, etc.)

        input:   maxYears - number of years to look in the future
        input:   interestRate - interest corrected by inflation
        input:   energyPrizeRate - prize development for energy, corrected by inflation
        returns: a object of type CashFlow
        """
        
        self.maxYear         = int(maxYears)
        self.rateInterest    = interestRate
        self.rateEnergyPrize = energyPrizeRate
        self.cashflow = [0.0] * self.maxYear
        
    def addInvestment(self, year, value):
        '''insert a singular payment or income

        input:   year - year in which payment is due
        input:   value - height of payment - expenses are negative, income is positive
        returns: nothing'''
        year  = int(year)
        value = float(value)
        if year > self.maxYear:
            logError("CashFlow/Investment: year %d exceeds range (0-%d)" % (year, self.maxYear))
        else:
            self.cashflow[year] -= value

    def addContingency(self, from_year, value):
        '''insert periodic payments from a ertain year onwards

        input:   from_year - year starting the payment
        input:   value - height of payment - expenses are negative, income is positive
        returns: nothing'''
        year  = int(from_year)
        value = float(value)
        if year > self.maxYear:
            logError(sys.stderr, "CashFlow/Contingency: year %d exceeds range (0-%d)" % (year, self.maxYear))
        else:
            for i in xrange(year, self.maxYear):
                self.cashflow[i] -= value
                
    def addOperating(self, value):
        '''operating costs (same as Contingency, always starting in year one)

        input:   value - height of payment - expenses are negative, income is positive
        returns: nothing'''
        value = float(value)
        self.addContingency(1, value)

    def addEnergy(self, value):
        '''Energie costs, similar to operating costs, but prize development is different
        and usuall bigger

        input:   value - height of payment - expenses are negative, income is positive
        returns: nothing'''
        value = float(value)
        for i in xrange(1, self.maxYear):
            self.cashflow[i] -= value * pow(1+self.rateEnergyPrize, i)

    def CF(self):
        '''cash flow as calculated till now
        
        returns: cash flow (array, on entry per year)'''
        return self.cashflow[:]
        
    def __repr__(self):
        '''Pretty print the cash flow as calculated till now
        returns: nothing'''
        result = "\n"
        for i in xrange(len(self.cashflow)):
            result += "%3d --> % 13.2f" % (i, self.cashflow[i])
        result += "\n"
        return result

def NPV(cf_new, cf_old, rate):
    '''NPV -- net present value
    http://en.wikipedia.org/wiki/Net_present_value
    
    input:   cf_new - cash flow of new process
    input:   cf_old - cash flow of old process
    input:   rate - net interest (scalar), in percent
    returns: npv - net present value (array, on entry per year)'''
    rate = rate
    sum = 0.0
    try:  
        assert len(cf_new) == len(cf_old), ValueError("Length of cashflows don't match.")
    except:
        return [0.0] * len(cf_new)
    npv = [0.0] * len(cf_new)
    for i in xrange(0,len(cf_new)):
        sum += (cf_new[i] - cf_old[i])*pow(1 + rate, -i)
        npv[i] = sum
        #print i, (cf_new[i] - cf_old[i]) * pow(1 + rate, -i),  sum
    return npv[:]

def NPVcf(cf, rate):
    #print "cf-------"
    #print cf
    '''NPV -- net present value
    http://en.wikipedia.org/wiki/Net_present_value
    
    input:   cf - net cash flow (old process - new process)
    input:   rate - net interest (scalar), in percent
    returns: npv - net present value (array, on entry per year)'''
    rate = rate
    sum = 0.0
    npv = [0.0] * (len(cf))
    for i in xrange(len(cf)):
        sum += cf[i] * pow(1 + rate, -i)
        npv[i] = sum
    return npv


def payback_period(npv):
    '''payback period
    
    input:   npv -- net present value (array, one entry for every year)
    returns: payback period (scalar)'''
    frac=0;
    i=0.0;
    # find interval where sign changes and interpolate linear    
    for i in xrange(0,len(npv)-1):
        if 0.0 >= npv[i]*npv[i+1]:
            try: frac = abs(npv[i])/abs(npv[i+1]-npv[i])
            except ZeroDivisionError: 
                frac=0
                logWarning(_U("Could not calculate payback period. NPV is zero."))
            break
    return i+frac


def MIRR(cf_new, cf_old, financeRate, reinvestRate):
    '''MIRR -- modified internal rate of return
    http://en.wikipedia.org/wiki/Modified_Internal_Rate_of_Return
    http://en.wikipedia.org/wiki/Internal_rate_of_return
    http://www.visitask.com/internal-rate-of-return.asp

    input:   cf_new - cost of old process (array, one entry for every year)
    input:   financeRate (scalar, percent)
    input:   reinvestRate (scalar, percent)
    returns: mirr (array, one entry for every year)'''
    #print "mirr calculation-----"
    #print cf_new
    #print cf_old
    #print financeRate
    #print reinvestRate
    try:
        assert len(cf_new) == len(cf_old)
    except:
        return [0.0]*len(cf_new)
    
    cf_pos = [0.0] * (len(cf_new))
    cf_neg = [0.0] * (len(cf_new))
    mirr   = [0.0] * (len(cf_new))
    fv1    = [0.0] * (len(cf_new))

    # separate positive and negative periods of cash flow
    for i in xrange(0, len(cf_new)):
        cFlow = cf_new[i] - cf_old[i] 
        if (cFlow >= 0):
            cf_pos[i] = cFlow
        else:
            cf_neg[i] = cFlow
    #print "CF possitive:  "
    #print cf_pos
    #print "CF negative:   "
    #print cf_neg
        
    # calculate their future value and net present valuse
    # and sum up with their respective interest rates
    for i in xrange(1, len(cf_new)):
        fv1[i] = 0.0
        for j in xrange (i):
            fv1[i]+= cf_pos[j+1]*pow(1.0+ reinvestRate,i-j-1)

            #print j, "cf possitive j:   ", cf_pos[j]
        #print i, "fv1:  ", fv1[i]

    npv2 = NPVcf(cf_neg, financeRate)
    #print i, "npv2:   ", npv2[i]

    for i in xrange(1, len(cf_new)):
        try:
            mirr[i] = pow(fv1[i]/(-npv2[i]),1.0/i) - 1.0
        except:
            mirr[i] = 0.0

    return mirr                 


    

def BCR(cf_new, cf_old, interest):
    """benefit cost ratio
    http://en.wikipedia.org/wiki/Benefit-cost_ratio
    
    input:   cf_new -- new cost (array, one entry for every year)
    input:   cf_old -- old cost (array, one entry for every year)
    input:   interest - discount rate (scalar)
    returns: bcr (array, one entry for every year)
    
    
    """
    try:
        assert len(cf_new) == len(cf_old)
    except:
        return [0.0]*len(cf_new)
    interest = interest
    
    sum_old = 0.0;
    sum_new = 0.0;
    bcr = [0.0] * len(cf_new)
    for i in xrange(len(cf_new)):
        sum_old += cf_old[i] * pow(1.0 + interest, -i)
        sum_new += cf_new[i] * pow(1.0 + interest, -i)
        try: bcr[i] = None #bcr[i] = -sum_old / sum_new
        except ZeroDivisionError: 
            bcr[i] = 1e300000
            logWarning(_U("Could not calculate BCR."))
    return bcr
        

def ANNUITY(I0,r,N):
    #Total Investment Capital
    #real interest rate of external financing
    #ProjectLifetime
    if (I0==0)or(N==0):
        return 0.0
    
    #print "ANNUITY I0=%s r=%s N=%s" %(I0,r,N)
    sum = 0
    try:
        for i in xrange(1,int(N+0.5)+1):
            sum+=1.0/pow(1+r,i)
        a = 1.0 / sum
        A = a * I0
    except ZeroDivisionError:
        logWarning(_U("Could not calculate annuity period."))
        return 0.0
    return A





#################
## MAIN - for testing only
#################

if __name__ == "__main__":
    from pylab import *

    a = CashFlow(30, 5, 3) # max_years, interest rate (inflation - net interest), energie prize development
    a.Investment(0,-180000*0.7)         # investment WT (<invcost>)  (- 30 % funding)
    a.Investment(0,-50000*0.8)          # investment biomass boiler (- 20 % funding)
    a.Investment(0,-400000*0.7+5000)   # investment solarthermal (- 30 % funding, 5000 funding)
    a.Investment(0, 50000*5/20*0.5)    # sale boiler (<residual_value>) AFA: 5 years out of 20 years depracation left
    a.Energy(-3200000*0.051) # energy price (eg. fossil gas) (<energy_demands>)
    a.Operating(-5000)       # operating costs, material maintance
    a.Operating(-800)        # operating costs, 
    a.Operating(-3000)       # operating costs,
    a.Operating(-1000)       # operating costs, 
    a.Investment(4,-30000)     # (<non_recurring_costs>) permits
    a.Contingency(5, 20000)  # zzz
    #a.Print()
    #arrayA = a.CF()          # get cash flow as an array of floats
    #print len(arrayA)

    # old CashFlow to compare with
    b = CashFlow(30, 5, 3)
    b.Energy(-5000000*0.051) # more energy
    b.Operating(-4000)       # operating costs, materials, maintainance and service
    b.Operating(-2000)       # operating costs, materials, maintainance and service
    b.Operating(-400)      # operating costs, wages, operating personal
    b.Investment(1,-4000)    # (<non_recurring_costs>) repair for energy equipment
    b.Investment(4,-50000)   # (<non_recurring_costs>) fines/penalties
    b.Contingency(5, -3000)  # xxx
    #b.Print()
    
    print "CF new:  ", a.CF()
    print "CF old:  ", b.CF()
    
    print
    npv = NPV(a.CF(), b.CF(), 5)       # call with <InterestRate>?
    print "NPV: ", npv
    print "PBP: ", payback_period(npv)
    plot(npv)
    xlabel('time / Y')
    ylabel('NPV / %s')%cu
    title('net present value')
    grid(True)
    show()

    print
    mirr = MIRR(a.CF(), b.CF(), 5, 6) # call with <InterestRate>, <basDiscountRate>
    print "MIRR:", mirr     
    figure(2)
    plot(mirr)
    xlabel('time / Y')
    ylabel('MIRR')
    title('modified internal rate of return')
    grid(True)
    show()

    bcr = BCR(a.CF(), b.CF(), 5) # call with <InterestRate>?
    print
    print "BCR: ", bcr
    figure(3)
    plot(bcr)
    xlabel('time / Y')
    ylabel('BCR')
    title('benefit cost ratio')
    grid(True)
    show()

    print
    raw_input('Press any key to continue.')
    close('all')
    
        
