# -*- coding: utf-8 -*-
#==============================================================================
#
#   E I N S T E I N
#
#   Expert System for an Intelligent Supply of Thermal Energy in Industry
#   (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   ModuleECO (Economic analysis)
#           
#------------------------------------------------------------------------------
#           
#   Cost analysis module
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import numpy as np
from GUI.status import *
from constants import *
from dataECO import DataECO
from cashFlow import CashFlow, NPV, MIRR, BCR, ANNUITY, payback_period
from itertools import chain

from modules.messageLogger import *
from auxiliary.auxiliary import toGettextAndUnicode as _U, transpose
from lib.auxiliary.auxiliary import noneCheck


class ModuleECO(object):

    def __init__(self, keys):
        self.keys = keys # the key to the data is sent by the panel 
        self.data = None    
        self.result = None
        self.change = False
        self.presentstate = None

    def updatePanel(self, checkStatusECO=True):
        """Load the costs data and set defaults when necessary.
        
        :param checkStatusECO: If False don't check whether the underlying data
        has changed since the last simulation run. Otherwise a call to the 
        simulation may happen to update the simulation results.
        
        """           
        #Create or Update data according to PID and ANo in Status    
        if Status.StatusECO < 1 and checkStatusECO:
            if Status.StatusEnergy < 1:
                Status.mod.moduleEnergy.runSimulation()
            if not self.presentstate: self.runPresentState()
            self.run()
            
        if (self.change == False):
            self.data = DataECO(Status.PId,Status.ANo)
            self.data.loadTCAData()        
        if (Status.PId != self.data.pid) or (Status.ANo != self.data.ano):
            self.data = DataECO(Status.PId,Status.ANo)
            self.data.loadTCAData()
        
    def reset(self):
        """Reset all data of all alternatives to the default values."""
        # As the first alternative Nr.0 is reset, and thus the general
        # economic data is also reset from any changes made from the TCA
        # dialogs.
        for aNo in xrange(Status.NoOfAlternatives + 1):  
            data = DataECO(Status.PId, aNo) 
            data.resetTCAData()
        
        self.data = DataECO(Status.PId,Status.ANo)                 
        self.data.loadTCAData()               
    
    def runModule(self, checkStatusECO=True):
        """Load the costs data and calculate the data that isn't there or has
        to be updated.

        Warning: this function carries out calculations for ALL alternatives !
        
        :param checkStatusECO: If False don't check whether the underlying data
        has changed since the last simulation run. Otherwise a call to the 
        simulation may happen to update the simulation results.
        
        """
        self.updatePanel(checkStatusECO)   
        self.result = []
                        
        self.runPresentState()
        
        ecopars = []
        ecopars.append([self.presentState.Inflation])
        ecopars.append([self.presentState.NIR])
        ecopars.append([self.presentState.DEP])
        ecopars.append([self.presentState.CSDR])
        ecopars.append([self.presentState.TimeFrame])

        Status.int.setGraphicsData("TCAPARS", ecopars)

        #get all proposal numbers and names
        alternatives = Status.prj.getAlternatives()

        #calculate result for each proposal in project
        for ANo, alt in enumerate(alternatives[1:]): # ano = [ Number, Name ] !!!
            name = alt.ShortName

            Status.prj.setActiveAlternative(ANo)
            
            data = self.run(ANo)
            self.result.append(data)

        self.__setDataForReport()
            
        
    def runPresentState(self):
        """ Loads the economic parameters for the present state and sets
        the global reference parameters for difference calculations.

        This function has to be called before calling the run-function
        """
        
        #get cashflow for the present state
        self.presentState = DataECO(Status.PId, 0)                
        self.presentState.loadTCAData()
                                  
        #set parameters
        self.interestRate = self.realRate(0.01*self.presentState.NIR, 0.01*self.presentState.Inflation)
        self.discountRate = self.realRate(0.01*self.presentState.CSDR, 0.01*self.presentState.Inflation)

        #PrimaryEnergyDemand = data.PETel + data.PETFuels
        self.projectLifetime = self.presentState.TimeFrame
          
        self.presentState = self.calculateCashFlow(self.presentState)
        self.presentState.storePresentStateInformationToCGeneral()
        return self.presentState
        
    def run(self, ANo = None):
        """Carries out economic calculations for one alternative
        
        :param checkStatusECO: If False don't check whether the underlying data
        has changed since the last simulation run. Otherwise a call to the 
        simulation may happen to update the simulation results.
        
        """
        ANo = noneCheck(ANo, Status.ANo)
        alt = Status.prj.getAlternative(ANo=ANo)
        name = alt.ShortName

        if alt.StatusECO < 1:
            # the current alternatives equipment has changed
            # force an update of the TCA data that
            # depends on equipment or energy demand data
            forceUpdate = True
        else:
            forceUpdate = False
        #load data for alternative
        data = DataECO(Status.PId, ANo)
        data.loadTCAData(forceUpdate)
        
        #calculate cashflow for alternative
        data = self.calculateCashFlow(data)
        current = data.cashflow

        #calculate the results  
        npv = NPV(current.CF(), self.presentState.cashflow.CF(), self.interestRate)
        mirr = MIRR(current.CF(), self.presentState.cashflow.CF(), self.interestRate, self.discountRate)
##        bcr = BCR(current.CF(), self.presentState.cashflow.CF(), self.interestRate)  
##        bcr = [None] * len(bcr)
        np.seterr(divide="ignore")
        bcr = np.array(npv)/-npv[0]
        np.seterr(divide="warn")
        annuity = ANNUITY(current.TotalInvestmentCapital,self.interestRate,self.projectLifetime)              
        pp = payback_period(npv)
       
        energycost0 =  self.presentState.getTotalEnergyCost()
        energycost     =  data.getTotalEnergyCost()
        
        opcost0     =  self.presentState.totalopcost
        opcost         =  data.totalopcost
        
        energydemand0 = self.presentState.getTotalEnergyDemand()
        energydemand = data.getTotalEnergyDemand()
        
        PEPresentState = self.presentState.getTotalPrimaryEnergy()
        PEAlternative = data.getTotalPrimaryEnergy()

        additionalcost = annuity+opcost+energycost-opcost0-energycost0         
        try: additionalcostpersavePE = additionalcost/(PEPresentState-PEAlternative)
        except: additionalcostpersavePE = 0.0

        #set the results in data, store data in result list
        display=True
        data.setResult(name, npv, mirr, bcr, annuity, pp, additionalcost, additionalcostpersavePE, display)                                
##        self.result.append(data)

        if ANo > 0: data.storeResultToCGeneralData()
        
        if alt.StatusEnergy > 0: # set ECO flag only, if results are based on converged energy simulation
            Status.prj.setStatus("StatusECO", 1, ANo=ANo)

        return data
                                      
        
    def calculateCashFlow(self,data):
        amortisationTime       = data.TimeFrame + 1 #to fix calculation

        # correct NIR and DEP by general inflation rate in order to calculate REAL cash flows in prices of year 0
        inflationRate = 0.01*data.Inflation if data.Inflation is not None else 0.0
        interestRate           = self.realRate(0.01*data.NIR, inflationRate)
        energyPriceDevelopment = self.realRate(0.01*data.DEP, inflationRate)

        cf = CashFlow(amortisationTime, interestRate, energyPriceDevelopment)
        
        cf.TotalInvestmentCapital = 0
        cf.EffectiveInvestmentCapital = 0
        cf.TotalFundings = 0
        for investment in data.investment:
            #Investment ------------------------------  
            value = investment[1]
            if value is None: continue
            fp    = noneCheck(investment[2], 0)
            ff    = noneCheck(investment[3], 0)   
            cf.TotalInvestmentCapital += value   
            cf.EffectiveInvestmentCapital += value
            cf.EffectiveInvestmentCapital -= value*(fp/100.0)+ff         
            cf.addInvestment(0, value)              

            #Funding ---------------------------------
            funding = value*(fp/100.0)+ff         
            cf.TotalFundings += funding
            cf.addInvestment(0, -funding)
             
        cf.EffectiveInvestmentCapital -= data.revenue 
        cf.addInvestment(0, -data.revenue)       #Revenue

        for energy in data.energycosts:    #Energy
            demand = energy[1]
            price  = energy[2]
            if demand is not None and price is not None: # security feature. Had become necessary in V2.4 for TC1.01.
                cf.addEnergy(demand*price)

        cf.addOperating(data.totalopcost)      #Operating
        
        for cont in data.contingencies:    #Contingencies
            year  = cont[2]
            value = cont[1]
            if year is not None and value is not None:
                cf.addContingency(year, value)  
        
        for non in data.nonrecurring: #Non reocuring costs
            value = non[1]
            year  = non[2]
            type  = non[3]
            if year is not None and value is not None:
                cf.addInvestment(year, value)
##            if (type == 'Cost'):
##                cf.addInvestment(year, value)
##            else:
##                cf.addInvestment(year, -value)
        
        data.cashflow = cf    
        
        return data 
    
    def __setDataForReport(self):
        logTrack("ModuleTCA.__setDataForReport: Final Alternative = %s"%(Status.FinalAlternative))
        try:
            self.__setReport52(self.keys[0], self.result[Status.FinalAlternative - 1])
        except TypeError:
            pass
        except IndexError:
            logTrack("ModuleTCA.__setDataForReport: An invalid FinalAlternative is given.")
                
    def __setReport52(self, key, alternative):
        data = []
        if (alternative==None):
            return
        
        cont = alternative.contingencies[:2]
        cont += [["no item", 0, 0, "costs"] for dummy in xrange(5 - len(alternative.contingencies))]
        cont = list(chain.from_iterable([item[:3] for item in cont[:5]]))
        nonroc = alternative.nonrecurring[:2]
        nonroc += [["no item", 0, 0, "costs"] for dummy in xrange(5 - len(alternative.nonrecurring))]
        nonroc = list(chain.from_iterable([item[:3] for item in nonroc[:5]]))

        data.append(alternative.TIC)
        data.append(alternative.totalfunding)   
        data.append(alternative.revenue)  
        data.append(alternative.annuity)
        data.append("no percentage known")
        data.append(alternative.energycost)
        data.append(alternative.totalopcost)
        data += cont
        data += nonroc
        data.append(alternative.npv[len(alternative.npv)-1])
        data.append(alternative.mirr[len(alternative.mirr)-1])
        data.append(alternative.PP)
        data.append(alternative.bcr[len(alternative.bcr)-1])

        # find feed in of energy if any
        feedinTariff = [item for item in alternative.energycosts if item[0] == "Feedin-Tariff"]
        if feedinTariff != []:
            data.append(feedinTariff[0][1] * feedinTariff[0][2]) # the amount of money for the feed in
            data.append(feedinTariff[0][2]) # the actual tariff of the feed in
        else:
            data.append("") 
            data.append("")

        data.append(alternative.additionalcostpersavePE)

        logTrack("ModuleTCA.__setReport52: data written to report %r"%data)
        Status.int.setGraphicsData(key, transpose([data]))

    
    def calculateTotalOpCostFromDetailedOpcost(self):
        if (self.data.ano<=0): #dont calculate opcost from detailed-opcost for current process
            return
        
        cost = 0
        for category in self.data.detailedopcost:
            for entry in category:
                cost+=entry[1]
                
        if (cost>=0):
            self.data.totalopcost = cost
       
    def storeData(self):
        """Store the TCA data in the database."""
        self.data.storeTCAData() 
    
    def resetData(self):
        self.data.resetTCAData()
        self.data.loadTCAData()

    def realRate(self, nominalRate, inflationRate):
        """
        helper function for converting nominal to real rates (of interest, etc.)
        """
        return (nominalRate - inflationRate)/(1.0 + inflationRate)
        
