# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ForeCast
#			
#------------------------------------------------------------------------------
#			
#	Module for calculating forecasts of boundary conditions
#
#==============================================================================#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 

import numpy as np
if __name__=="__main__":
    import sys
    import os
    sys.path.append(os.path.abspath("../.."))
    sys.path.append(os.path.abspath("../../.."))

from constants import EPSILON
from lib.statistics.correlation import autoCorrelation_t, checkAutoCorrelation

class ForeCast():
    """
    nomenclature:
    t: time of last data
    tp: time of forecast [from 1*dt to n*dt]
    u: = interval of forecast
    """

    def __init__(self, model="TREND", dt=1.0, n=24, par=[], state=[]):
        """ set-up of forecast
        dt: time step for forecasts
        n:  number of time steps to forecast
        """
        
        self.n = n
        self.dt = dt
        self.tp = np.linspace(self.dt, self.n*self.dt, self.n)
        
        self.model = model
        self.par = par

        self.t = None
        self.xt = None

        self.prepare_model(state)

        self.history = []


    def prepare_model(self, state):
        """ carries out initial calculations for a specific model
        """

        if self.model=="TREND":
            self.gamma_trend= self.dt/self.par[0]
            self.eta_trend  = self.dt/self.par[1]
            
            tau_inst = self.par[2]
            self.alpha_rand = self.par[3]
            self.tau_rand = self.par[4]
            
            if tau_inst > EPSILON:
                self.alpha_inst = np.exp(-self.tp/tau_inst)
            else:
                self.alpha_inst = np.zeros(self.n)

            if len(state) > 0:
                self.x_trend = state[0]
                self.g_trend = state[1]
            else:
                self.x_trend = None
                self.g_trend = None

            self.saveState()

        elif self.model=="CYCLE":
            self.TCycle = self.par[0]
            self.gamma_cycle = self.par[1]
            self.TSample = self.par[2]      # sample time for historic data
            self.nCycle = int(self.TCycle/self.dt + 0.5)
            self.u0 = 0
            self.du = 1.0/self.dt
            self.xc = np.zeros(self.nCycle)


    def saveState(self):
        if self.model=="TREND":
            self.state = [self.x_trend, self.g_trend]
        elif self.model=="CYCLE":
            self.state = [self.u0, self.du] + list(self.xc)
        elif self.model=="CONST":
            self.state = [self.xt]
        
    def get(self):
        """ gets a forecast for n equidistant timesteps
            t + dt ... t + n*dt
        """
        if self.model=="TREND":
            fc = self.f_trend()
        elif self.model=="CYCLE":
            fc = self.f_cycle()
        else: # CONST model"
            fc = self.xt*np.ones(self.n)
            
        return fc
            

    def update(self, x, t, info=None):
        """ updates the forecast with a new measurement at time t
        """

        self.lastDt = 0.0 if self.t is None else t - self.t
        while self.lastDt < 0:  # check for change of the year of HOY data
            self.lastDt += 8760.
            
        self.t = t

        xo = self.xt
        self.xt = x

        if self.model=="TREND":
            if self.gamma_trend is None or self.x_trend is None:
                self.x_trend = x
            else:
                self.x_trend += self.gamma_trend*(x - self.x_trend)*self.lastDt
                
            if abs(self.lastDt) > EPSILON:
                g = (x - xo)/self.lastDt
                if self.g_trend is None:
                    self.g_trend = g
                else:
                    self.g_trend += self.eta_trend*(g - self.g_trend)*self.lastDt

        elif self.model=="CYCLE":
            # append new measurement to history and delete very old ones
            self.history.append([t, x])
            self.history = [entry for entry in self.history if (t - entry[0])%8760 <= self.TSample]

            t = np.array([entry[0] for entry in self.history])
            f = np.array([entry[1] for entry in self.history])
            ac = autoCorrelation_t(f, t, tstep=self.dt)
            self.ac = ac
            self.TCycle = checkAutoCorrelation(np.array([entry[1] for entry in self.ac]))
            if self.TCycle is not None: self.TCycle *= self.dt
                        
            
##            delta_u = int(self.du*self.lastDt + 0.5)
##            self.u0 = (self.u0 + delta_u)%self.nCycle
##            xc = self.xc[self.u0]
##            self.xc[self.u0] += self.gamma_cycle*(x - xc)
##            print "Update u0: %s x %s xc %s"%(self.u0, x, xc)
            

        elif self.model=="CONST":
            """
            simplest possible forecast
            supposes current values to be constant
            """
            pass

        self.saveState()


    def f_trend(self):
        """ trend function which combines
        - average value x_trend
        - average growth rate g_trend
        - use of instantaneous value for tp's close to t
        - random component
        """
        g = 0.0 if self.g_trend is None else self.g_trend

        # calulate fxi, a random function between [-1:1] for tau_random = 0 and a larger brownian motion for tau_random > 0
        xi = 2.0*np.random.random(self.n) - 1.0

        if self.tau_rand > EPSILON:
            fxi = np.zeros(self.n)
            alpha = self.dt/self.tau_rand
            base = 0
            for i, xi in enumerate(xi):
                fxi[i] = alpha*xi + (1. - alpha)*base
                base = fxi[i]
            
        else:
            fxi = xi

        return (self.alpha_inst + self.alpha_rand*fxi)*self.xt + (1. - self.alpha_inst)*(self.x_trend + g*self.tp)

##    def f_cycle(self):
##        """ cycle function
##
##        returns fc(u) with
##
##        u = [u0 + du*tp]%nCycle with u in [0:nCycle]
##        """
##
##        u = np.mod(self.u0 + self.du*self.tp, self.nCycle)
##
##        i0 = np.floor(u + EPSILON)
##        w1 = u - i0
##        w0 = 1.0 - w1
##        i1 = np.mod(i0+1, self.nCycle)
##
####        print "u0", self.u0
####        print "u", u
####        print "i0", i0
####        print "i1", i1
####        print "w0", w0
####        print "w1", w1
####        a = w0[i]*self.xc[int(i0[i])]
####        b = 
##        f = [(w0[i]*self.xc[int(i0[i])] + w1[i]*self.xc[int(i1[i])]) for i in xrange(self.nCycle)]
##
##        return np.array(f)

    def f_cycle(self):
        if self.TCycle is None:
            return np.ones(self.n)*self.xt
        else:
            t_hist = np.array([entry[0] for entry in self.history])
            f_hist = np.array([entry[1] for entry in self.history])
            return np.interp(self.tp+self.t-self.TCycle, t_hist, f_hist)
        
##        for it, tp in enumerate(self.tp):
            

def unitTest():
    nt = 40
    T = 10
    t = np.linspace(0, T, T*nt+1)
    ftestA = np.sin(2.0*np.pi*t)    # sine function
    ftestB = 1.0*(ftestA>EPSILON)         # on-off function

    gt = 0.2
    ftestC = np.exp(gt) + ftestA    # exponential growth + sine function
    

    for f in [ftestA, ftestB, ftestC]:
##    for f in [ftestB]:
##        fc1 = ForeCast(model="TREND", dt=1./nt, n=nt, par=[0.1, 0.5, 0.2, 0.1, 0.1])
        fc1 = ForeCast(model="CYCLE", dt=1./nt, n=nt, par=[1.0, 0.5, 3.0])

    ##    print "t:\tA:\tB:\tC:"
        for i, tp in enumerate(t[:-nt]):
            fc1.update(f[i], tp)

        print "AC\n"
        for entry in fc1.ac:
            print "%10.2f %10.2f"%(entry[0], entry[1])
        print fc1.TCycle
        print fc1.history
        
        xfc = fc1.get()

        ctr = 0
        errAvg = 0.0
        errMax = 0.0
        
        print "t:\tx:\txfc:\terr:"
        for j, tp in enumerate(t[-nt:]):
            i = j + (T-1)*nt + 1
            err = xfc[j] - f[i]
            errAvg += err
            errMax = max(errMax, abs(err))
            ctr += 1
            print "%10.2f\t%10.2f\t%10.2f\t%10.2f"%(tp, f[i], xfc[j], err) 
            
        errAvg /= ctr

        print "Average error %10.2f (max: %10.2f)"%(errAvg, errMax)
    ##        print "%10.2f\t%10.2f\t%10.2f\t%10.2f"%(tp, ftestA[i], ftestB[i], ftestC[i]) 
        

if __name__=="__main__":

    import cProfile
    import pstats
    cProfile.run("unitTest()", "runstats.txt")
    s = pstats.Stats("runstats.txt")
    s.sort_stats("cumulative")
    s.print_stats(100)

