# -*- coding: cp1252 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEnergy
#			
#------------------------------------------------------------------------------
#			
#	Module for simulation of energy flows in the system
#
#==============================================================================
#
#	Version 2.4plus, October 2016
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 

import numpy as np
from constants import *
from GUI.status import Status
from modules.messageLogger import *
from GUI.dialogGauge import DialogGauge
from modules.energy.simulationPlus import SimulationPlus
from modules.energy.simulationNP import SimulationNP
from modules.fluids import Fuel
import database.dbio as dbio
from modules.interfaces import calcQ_T
from auxiliary.auxiliary import noneCheck

from GUI.GUITools import BLUE, RED

#============================================================================== 
#============================================================================== 
class ModuleEnergy(object):
    def __init__(self, keys):
        self.keys = keys
        # Keep for each equipment id the id of the pipes that 
        # depend on them more or less heavily.
        self.pipesWithPriority = []  
        self.pipesWithoutPriority = []

        if Status.SimTool == "Plus":
            self.outputLevel = 0
        else:
            self.outputLevel = 999 # higher than maximum possible

    def initPanel(self):
        logTrack("ModuleEnergy (initPanel)")
        # contain for each equipment the indices of the pipes that
        # have and that not have priority on the equipment.
        # They have priority, if there is no equipment later in the
        # cascade that supplies them.
        self.NM = 0 # number of pipes in project
        self.NJ = 0 # number of equipments in project
        
        
    def updatePanel(self):
        """Carry out any calculations necessary previous to displaying 
        the BB design assistant window
        
        """
        if Status.ANo <= 0:
            logWarning("Energy simulation in present state is not possible. Data not updated.")
            return
        # Plot global energy data
        equipment = Status.int.systemModel.eqObject
        NJ = len(equipment)
        self.NJ = NJ

        self.calculate()
            
        Status.mod.moduleEA.update()
        generalData = Status.prj.getGeneralData()
        
        USH = generalData.USH
        UPH = generalData.UPH

        USC = generalData.USC
        UPC = generalData.UPC
        
        PET = generalData.PET
        PETFuels = generalData.PETFuels
        PETel = generalData.PETel

        tableDataPanel = [[PET, PETFuels, PETel, USH, UPH, USC, UPC]]
            
        PET = PET / 1000.0 if PET is not None else "---"
        PETFuels = PETFuels  / 1000.0 if PETFuels is not None else "---"
        PETel = PETel / 1000.0 if PETel is not None else "---"
        USH = USH / 1000.0 if USH is not None else "---"
        UPH = UPH / 1000.0 if UPH is not None else "---"

        tableData = np.array([[PET, PETFuels, PETel, USH, UPH]])

        Status.int.setGraphicsData("ENERGY", tableData)

        data = {"Table": {"data": tableDataPanel}}

        # Plot weekly heat supply

        if Status.int.cascadeUpdateLevel < NJ:    #cascade not up to date -> show zero line
            weeks = [1.0 * i for i in range(53)]
            vals = [0.0] * 53
            dataList = [weeks, vals]
        else:
            yData = []
            labels = []
            colours = []

            for showHeating in [True, False]:
                # first filter for heating xor cooling equipment 
                USHj_weekly = [[eq.name] for eq in equipment 
                               if eq.heating == showHeating]
                # this list contains the global indices of the equipments
                equipmentNos = [eq.no for eq in equipment 
                               if eq.heating == showHeating]
                
                NJHC = len(equipmentNos) # number of heating or cooling equipment
                USHj_sum = [0.0] * NJHC # reset of the sum for each equipment
                weeks = ["week"]
                nweek = 0.0

                # now create the plot data
                for it in range(Status.Nt):
                    for local, jc in enumerate(equipmentNos):
                        USHj_sum[local] += Status.int.USHj_t[jc][it] / 1000.0

                    if ((it + 1) % 168) == 0:  #end of the week
                        USHj_week = 0
                        for local in xrange(NJHC):
                            if local > 0: USHj_sum[local] += USHj_sum[local-1]  #sum one above the other

                            USHj_weekly[local].append(USHj_sum[local] * Status.EXTRAPOLATE_TO_YEAR)  #append twice, for
                            USHj_weekly[local].append(USHj_sum[local] * Status.EXTRAPOLATE_TO_YEAR)  #start and stop time of the week
                        USHj_sum = [0.0] * NJHC # reset of the sum for each equipment

                        weeks.append(nweek)
                        nweek += 1.0
                        weeks.append(nweek)

                dataList = [weeks]
                xData = [np.array(weeks[1:], dtype=float)]
                for line in USHj_weekly:
                    dataList.append(line)
                    yData.append(np.array(line[1:], dtype=float))
                    labels.append(line[0])
                    colours.append(RED if showHeating else BLUE)
                                 
                plotData = np.array(dataList)
                Status.int.setGraphicsData("ENERGY Plot1", plotData)

            data.update({"Plot": {"xData": xData, "yData": yData, "labels": labels, "colours": colours}})
        
        return data

    def getEquipmentList(self):
        """Get the equipment list"""
        self.equipment = Status.int.systemModel.equipment
        
        self.NJ = len(self.equipment)
        logTrack("ModuleEnergy (getEquipmentList): %s utilities found" % self.NJ)

        Status.int.getEquipmentCascade()

    def calculateOM(self, eq, USH):
        """ Calculates current O&M costs based on USH of equipment """
        
        OMFix = eq.OMFix
        OMVar = eq.OMVar

        try:
            OM = OMFix + OMVar*USH
        except:
            logTrack(_("ModuleEnergy.cEEB: OM costs for equipment %s could not be calculated")%eq.name)
            OM = 0.0

        eq.OM = dbio.check(OM)

    def calculate(self, requireConverged = True):
        """
        Function designed for centralising the management
        of energy calculations
        As an output the updated energy calculation results should be available
        """
        # rebuild the systemModel if necessary
        Status.int.systemModel.build()
        modelGlb = Status.int.systemModel
        streams = modelGlb.streams

        if Status.ANo < 0 or Status.StatusCC==0: return
        
        if Status.processData.outOfDate:
            Status.processData.createAggregateDemand()
            
        if Status.SimTool == "NP":
            self.NJ = modelGlb.NJ
            if Status.schedules.equipmentSchedulesOutOfDate == True:
                equipments.estimateEquipmentWasteHeat()
            if Status.int.heatRecoveryOutOfDate:
                    
                Status.mod.moduleHR.runHRModule()

            Status.int.extendCascadeArrays(self.NJ) # if not yet done, create arrays
            if Status.StatusEnergy == 0 or Status.int.cascadeUpdateLevel < Status.NEquipe:
                if Status.ANo > 0:
                    if requireConverged:
                        self.runSimulation()
                    else:
                        logWarning(_("Take care. Results for USH/USC are not yet converged. Run calculation of energy performance"))

        else:
            if Status.ANo >= 0 and Status.int.cascadeUpdateLevel < modelGlb.NJ:
                self.runSimulation()

        # link HeatFlow_Tit objects from the simulation tool to the global stream list

        if Status.SimTool == "Plus":
            modelPlus = self.sim.systemModel
            for stream in streams:
                stream.hf = modelPlus.getStream(stream.origin)
                
                stream.copyAttributesFromHeatFlow_Tit()
        else:
            for stream in streams:
                stream.copyAttributesFromHeatFlow_t()

#------------------------------------------------------------------------------
    def runSimulation(self,first=1,last=None,loop=True, checkPeakPower=False, outputLevel=1):
        """
        :params outputLevel: 0: basic simulation, only global results are stored
                             1: level required for basic display options (EA panels, etc.)
                             2: USHTotal_Tt is stored
                             3: USJj's are stored
        """

        self.outputLevel = outputLevel
        
        #...................................................................
        # common checks and preparation phase independent of type of SimTool 

        if Status.ANo <= 0: #security feature against abuse of calls to runSimulation
            if Status.ANo < 0 or Status.SimTool != "Plus":
                logError("ModuleEnergy.runSimulation: call is not allowed for present state. Please report as a bug")
                return
        
        #...................................................................
        # run the simulation 

        if Status.SimTool == "Plus":
            self.sim = SimulationPlus(self)
            self.sim.runSimulation(first, last, loop, checkPeakPower, outputLevel)
        else:
            self.sim = SimulationNP(self)
            self.sim.runSimulation(first, last, loop, checkPeakPower)

#-------------------------------------------------------------------------------------------------------------------
            
    def calculateEquipmentEnergyBalances(self, calledFromRunSimulation = False):
        """Calculate the annual fuel balances and energy transferred
        by pipes from the results of system simulations and store 
        them within the database.
        
        """
        logTrack("ModuleEnergy (calculateEquipmentEnergyBalances)")
        
        projectData = Status.int.systemModel.projectData
        generalData = Status.int.systemModel.generalData
        equipment = Status.int.systemModel.eqObject
        pipes = Status.int.systemModel.pipeObject
        processes = Status.int.systemModel.processObject
        fuels = Status.prj.getQFuels()
        fuelList = Status.prj.getQFuelList("DBFuel_id")
        
        electricity = Status.prj.getElectricity()
        el = Status.int.systemModel.elObject[0]

        # Update of FETi/FECi values based on equipment fuel consumption FETj
        # Carried out only if ANo > 0. For ANo = 0 comes from consistency check
        if Status.ANo > 0:
            if Status.int.cascadeUpdateLevel < len(equipment) or Status.StatusEnergy == 0 and (not calledFromRunSimulation):     #previously used criterion cascadeUpdateLevel only is not sufficient.
                                                                                                #Does not necessarily mean that simulation has converged.
                logTrack("ModuleEA (calcEq.En.Bal.): calling runSimulation before updating energy balances")
                Status.mod.moduleEnergy.runSimulation()
                
            FETi = [0.0] * len(fuels)
            FETiUnKnown = 0.0
            FETel = 0.0
            ElGen = 0.0
            FETFuel = 0.0
            FETHeat = 0.0
            waterConsumption = 0.0
            FET = 0.0
            USH = 0.0
            USC = 0.0
            QHX = 0.0
            QHXReq = 0.0
            QWH = 0.0

            for eq in equipment:
                isHeating = eq.heating

                jc = eq.no
                dFETFuel = Status.int.FETFuel_j[jc]
                if dFETFuel is None:
                    logDebug("ModuleEA (cEEB): FETFuel of equipment %s (cascadeIndex = %s) is None" % \
                             (eq.EqNo, eq.CascadeIndex))
                    dFETFuel = 0.0
                dFETel = Status.int.FETel_j[jc]
                if dFETel is None:
                    logDebug("ModuleEA (cEEB): FETel of equipment %s (cascadeIndex = %s) is None" % \
                             (eq.EqNo, eq.CascadeIndex))
                    dFETel = 0.0
                if eq.HCGEEfficiency > 0:
                    dElGen = Status.int.ElGen_j[jc]
                else:
                    dElGen = 0.0
                if dElGen is None:
                    dElGen = 0.0
                dFETHeat = Status.int.FETHeat_j[jc]
                if dFETHeat is None:
                    logDebug("ModuleEA (cEEB): FETHeat of equipment %s (cascadeIndex = %s) is None"%\
                             (eq.EqNo,eq.CascadeIndex))
                    dFETHeat = 0.0
                    
                dUSH = Status.int.USHj[jc]
                dQHX = Status.int.QHXj[jc]
                dQHXReq = np.sum(Status.int.QHXEqReq_Tt[jc].Q)
                dQWH = Status.int.QWHj[jc]

                eq.FETFuel_j = dbio.check(dFETFuel)
                eq.FETel_j = dbio.check(dFETel)
                
                if eq.HCGEEfficiency > 0:
                    eq.STAreaFactor = dbio.check(dElGen)   #TODO XXXXXXtemporarily STAreaFactor used as storage for ElGen
                eq.FETHeat_j = dbio.check(dFETHeat)
                eq.FETj = dbio.check(dFETel + dFETFuel + dFETHeat) #simple sum of fuel + electricity. doesn't make very much sense, but ... 
                
                eq.USHj = dbio.check(dUSH)
                eq.QHXj = dbio.check(dQHX)       
                eq.QHXEq = dbio.check(dQHX)    #redundancy in nomenclature ... should be unified ... !!!
                eq.QWHj = dbio.check(dQWH)
                eq.QWHEq = dbio.check(dQWH)
                eq.QExhaustGas = dbio.check(dQWH)

                FET += dFETFuel + dFETel
                FETFuel += dFETFuel
                FETel += dFETel
                ElGen += dElGen
                FETHeat += dFETHeat
                waterConsumption += Status.int.waterConsumption_j[jc]

                if isHeating:
                    USH += dUSH
                else:
                    USC += dUSH
                QHX += dQHX
                QHXReq += dQHXReq
                QWH += dQWH

                fuelID = eq.DBFuel_id
                if fuelID in fuelList:
                    i = fuelList.index(fuelID)
                    FETi[i] += dFETFuel
                else:
                    FETiUnKnown += dFETFuel

                # TODO: here the warning should not be emitted later on, 
                # if we have an electrically driven equipment
                # e.g. electrical heater or electrically driven chiller.
                # see how this is handled in consistency check (checkEq).
                    if dFETFuel > 0:
                        logWarning("Fuel type for eq %s is not known" % eq.name)

                self.calculateOM(eq, dUSH)

            # calculate derived quantities and store in equipment table
            for j, eq in enumerate(equipment):

                # average thermal conversion efficiency
                if eq.FETFuel_j > EPSILON:
                    eq.HCGTEffReal = dbio.check(Status.int.USHj[j] / eq.FETFuel_j)
                elif eq.FETHeat_j > EPSILON:
                    eq.HCGTEffReal = dbio.check(Status.int.USHj[j] / eq.FETHeat_j)
                elif eq.FETel_j > EPSILON:
                    eq.HCGTEffReal = dbio.check(Status.int.USHj[j] / eq.FETel_j)
                else:
                    eq.HCGTEffReal = 0.0

                # real operating hours
                eq.HPerYear = dbio.check(Status.int.HPerYearEq[j])
                
                # part load factor
                if eq.Pnom is not None:
                    if eq.HPerYear is not None:
                        USHmax = eq.HPerYear * eq.Pnom 
                        if USHmax > 0:
                            eq.PartLoad = dbio.check(Status.int.USHj[j] / USHmax)
                        else:
                            eq.PartLoad = 0.0
                    else:
                        eq.PartLoad = 1.0
                        if eq.Pnom > 0:
                            eq.HPerYear = dbio.check(Status.int.USHj[j] / eq.Pnom)
                        else:
                            eq.HPerYear = 0.0
                else:
                    eq.PartLoad = 0.0
                
                # part load factor
                #HCGEEff -> only for CHP ??? or also for electrically driven chillers ???

            ##### TAKE CARE: here HCGTEfficiency is both an input and a result ... for equipments with
            ##### variable COP depending on temperature / ... this may lead to confusions.

            # now store fuel and electricity consumption in the corresponding tables
            FETFuels = 0.0
            FECFuels = 0.0
            for fuel in fuels:
                i = fuel.FuelNo - 1
                fuel.FETFuel = dbio.check(FETi[i])
                FETFuels += FETi[i]
                FECFuels += FETi[i]
                if fuel.FEOFuel is not None:
                    fuel.FECFuel = dbio.check(FETi[i] + fuel.FEOFuel)
                    FECFuels += fuel.FEOFuel
                else:
                    fuel.FECFuel = dbio.check(FETi[i])
                    fuel.FEOFuel = 0.0
                    logDebug("ModuleEA (calculateEqEnergyBalances): no entry found for FEOFuel of fuel no. %s" % (i + 1))

                if fuel.FECFuel is not None and fuel.DBFuel_id in fuelList:
                    fuelData = Fuel(fuel.DBFuel_id)
                    if fuelData.LCV is not None and fuelData.LCV > 0:
                        fuel.MFuelYear = dbio.check(fuel.FECFuel / fuelData.LCV)
                    if fuel.FuelTariff is not None:
                        fuel.FuelCostYear = dbio.check(fuel.FECFuel * fuel.FuelTariff)
                        
            # prepare handling of electricity consumption and self generation
            FEOel = generalData.FEOel   # load FEOel from data base. is not affected by calculations
            if FEOel is None:
                logWarning(_("WARNING: No data available for electricity consumption for non-thermal uses. Set to 0 !!!"))
                FEOel = 0.0
                generalData.FEOel = float(0.0)
            FETel = noneCheck(FETel, 0.0)
            generalData.FETel = float(FETel)
            generalData.FECel = float(FETel + FEOel)
            electricity.ElGenera = float(noneCheck(ElGen, 0.0))

            # update elObject
            el.FECel = generalData.FECel
            el.FETel = generalData.FETel
            el.FEOel = generalData.FEOel
            el.ElGenera = electricity.ElGenera

            if Status.ANo > 0:
                if Status.SimTool=="Plus":
                    modelPlus = self.sim.systemModel
                    FETel_t = modelPlus.FETel_t
                    ElGen_t = modelPlus.ElGen_t
                else:
                    FETel_t = np.sum(Status.int.FETel_j_t, axis=0)
                    ElGen_t = Status.int.ElGen_t

                el.aggregateData(FETel_t, FEOel, ElGen_t)
                el.distributeSelfGen()
                el.calculateAverageTariffs()

            # export total energy consumption
            FEO = generalData.FEO   # load FEO from data base. is not affected by calculations
            if FEO is None:
                logWarning(_("WARNING: No data available for energy consumption for non-thermal uses. Set to 0 !!!"))
                FEO = 0.0
                generalData.FEO = 0.0

            FET = noneCheck(FET, 0.0)
            
            generalData.FET = float(FET) # total FET as simple sum
            generalData.FEC = float(FET + FEO)

            generalData.USH = dbio.check(USH) # total USH
            generalData.USC = dbio.check(USC) # total USC


            # export other aggregates from equipment
            generalData.WaterConsum = dbio.check(waterConsumption)

            Status.int.USHTotal = USH
            Status.int.USCTotal = USC
            Status.int.QWHEqTotal = QWH
            Status.int.QHXEqTotal = QHX
            Status.int.QHXEqReqTotal = QHXReq
            Status.int.FETHeatTotal = FETHeat
            
            # update pipe data
            for m, pipe in enumerate(pipes):
                USHC = Status.int.USHCPipe[m]

                pipe.USHPipe = dbio.check(USHC)
                pipe.USHm = dbio.check(USHC)    # this ignores heat recovery on the pipe, 
                                        # but we have not better value for now
                pipe.UPHProcm = dbio.check(Status.int.UPHCProcm[m])
                pipe.QHXPipe = dbio.check(Status.int.QHXPipe_Tt[m].Q)

##            processes = Status.prj.getProcesses()
##            for k, proc in enumerate(processes):
            for k, proc in enumerate(processes):
##                procObj = Status.int.systemModel.processObject[k]
                UPX = noneCheck(Status.int.UPX[k], 0)
                UPXProc = noneCheck(Status.int.UPXProck[k], 0)
                QHXProc = max(UPX - UPXProc, 0) # hack: max had become necessary, because in some cases (e.g. TC-EMS-03c) some rounding errors make QHXProc negative
                proc.QHXProc = dbio.check(QHXProc)
                proc.UPHProc = dbio.check(UPXProc)
                
                
        # in case of present state:
        # some global derived quantities that are not calculated in the CCheck module
        else: 
            UPH = generalData.UPH
            USH = generalData.USH
            UPC = generalData.UPC
            USC = generalData.USC
            if (all(item is not None for item in (UPH, USH, UPC, USC)) and
                (USH + USC > 0)):
                generalData.HDEffAvg = dbio.check((UPH + UPC) / (USH + USC))
                if generalData.HDEffAvg < 0.5:
                    logWarning(_("ModuleEA (calcEq.En.Bal): check your input data.\ncalculated distribution efficiency is less than 50 %"))
            else:
                logTrack("ModuleEA (calcEq.En.Bal): Error in UPH/USH data. cannot calculate distribution efficiency")
                generalData.HDEffAvg = 0.0

            el.FETel = noneCheck(el.FETel, 0.0)
            el.FECel = noneCheck(el.FECel, 0.0)
            el.FEOel = noneCheck(el.FEOel, 0.0)
            generalData.FETel = float(el.FETel)
            generalData.FECel = float(el.FETel + el.FEOel)
            generalData.FEOel = float(el.FEOel)

        # from here on actions also for ANo = 0

        # conversion to primary energy and environmental impact parameters
        el.calculateENV()
        
        generalData.PETel = float(el.PETel)
        generalData.PECel = float(el.PECel)

        generalData.ProdCO2el = float(el.ProdCO2)
        generalData.ProdNoNukesEl = float(el.ProdNoNukes)

        # calculate environmental parameters for fuels
        PETFuels = 0.0
        PECFuels = 0.0
        ProdCO2Fuels = 0.0
        FECFuels = 0.0  # direct sum of fuel FECs -> used for SEC values only !!
        
        for fuel in fuels:
            i = fuel.FuelNo - 1
            f = Fuel(fuel.DBFuel_id)
            
            if fuel.FETFuel is None:
                logWarning(_("WARNING: No data available for fuel consumption for thermal uses (fuel no. %s). Set to 0 !!!")
                           % (i + 1))
                fuel.FETFuel = 0.0

            if fuel.FECFuel is None:
                logWarning(_("WARNING: No data available for total fuel consumption (fuel no. %s). Set to 0 !!!")
                           % (i + 1))
                fuel.FECFuel = 0.0

            dPET = fuel.FETFuel * f.PEConv
            dFEC = fuel.FECFuel
            dPEC = fuel.FECFuel * f.PEConv
            dCO2 = fuel.FECFuel * f.CO2Conv

            fuel.PETFuel = dbio.check(dPET)
            fuel.PECFuel = dbio.check(dPEC)
            fuel.ProdCO2Fuel = dbio.check(dCO2)

            PETFuels += dPET
            PECFuels += dPEC
            ProdCO2Fuels += dCO2
            FECFuels += dFEC

        generalData.PETFuels = dbio.check(PETFuels)
        generalData.PECFuels = dbio.check(PECFuels)
        generalData.ProdCO2Fuels = dbio.check(ProdCO2Fuels)

        generalData.PEC = dbio.check(generalData.PECFuels + generalData.PECel)
        generalData.PET = dbio.check(generalData.PETFuels + generalData.PETel)

        # energy intensity and specific energy consumption
        # Part 1: Energy Intensity
        Turnover = projectData.Turnover #in euro

        if (Turnover > 0) and Turnover is not None:    
            generalData.FUEL_INT = dbio.check(FECFuels / Turnover) 
            generalData.EL_INT = dbio.check(generalData.FECel / Turnover) 
            generalData.PE_INT = dbio.check(generalData.PEC / Turnover) 
        else:
            generalData.FUEL_INT = dbio.check(None)
            generalData.EL_INT = dbio.check(None)
            generalData.PE_INT = dbio.check(None)

        #Part 2: SEC by products
        products = Status.prj.getProducts()
        if Status.ANo >= 0: # distribution of energy consumption to products by proportion in turnover.
                    # could later on be improved taking into account the original distribution ...
            sumTurnoverProd = 0.0
            for product in products:
                if product.TurnoverProd is not None: sumTurnoverProd += product.TurnoverProd

            for product in products:
                if sumTurnoverProd > 0:
                    if product.TurnoverProd is not None:
                        fTurnover = product.TurnoverProd / sumTurnoverProd
                    else:
                        fTurnover = 0.0
                else:
                    fTurnover = 1. / len(products)

                quantity = product.QProdYear

                PEC = generalData.PEC
                if Status.ANo > 0 or product.PE_SEC is None:
                    if PEC is not None and quantity > 0 and quantity is not None:
                        product.PE_SEC = dbio.check(fTurnover * PEC / quantity)
                    else:
                        product.PE_SEC = 0.0

                FECel = generalData.FECel
                if Status.ANo > 0 or product.EL_SEC is None:
                    if FECel is not None and quantity > 0 and quantity is not None:
                        product.EL_SEC = dbio.check(fTurnover * FECel / quantity)
                    else:
                        product.EL_SEC = 0.0

                if Status.ANo > 0 or product.FUEL_SEC is None:
                    if FECFuels is not None and quantity > 0 and quantity is not None:
                        product.FUEL_SEC = dbio.check(fTurnover * FECFuels / quantity)
                    else:
                        product.FUEL_SEC = 0.0

        #Part 3: SEC by unit operations
        for proc in processes:
            UPH = proc.UPH
            try:
                quantity = dbio.check(proc.VInFlowDay * proc.NDays)
            except TypeError:
                quantity = 1.0
            try: 
                proc.UPH_SEC = dbio.check(UPH / quantity)
            except (TypeError, ZeroDivisionError): 
                proc.UPH_SEC = 0.0
            proc.PE_SEC = 0.0
            proc.EL_SEC = 0.0

        # set Status flag indicating that ANNUAL energy balances for the present alternative
        # are up to date
        if Status.ANo == 0: Status.prj.setStatus("Energy") #changed 25.07.2012 -> for ANo > 0 done in runSimulation only if simulation has converged !

##        print "ModuleEnergy.cEEB: ElGen_j(t) = %s"%Status.int.systemModel.ElGen[j][:12]
        if Status.ANo > 0:
            el.saveToDB()
            
            for eq in equipment:
                eq.saveToDB(warn=False, simResultsOnly=True)

            for pipe in pipes:
                pipe.saveToDB(warn=False)

        for proc in processes:
            proc.saveToDB(warn=False, simResultsOnly=True)
                
        dbio.commit()

#==============================================================================

    def getPEConvRatios(self):
        el = Status.int.systemModel.elObject[0]

        fPE_el = el.PEConv
        fPE_fuel = []
        for j, eq in enumerate(Status.int.systemModel.equipment):
            fuelId = eq.DBFuel_id
            if fuelId is not None:
                f = Fuel(fuelId)
                fPE_fuel.append(noneCheck(f.PEConv, 1.1))
            else:
                fPE_fuel.append(1.1)

        self.fPE_el = fPE_el
        self.fPE_fuel = fPE_fuel
        return fPE_el, fPE_fuel

    def calculatePET(self):
        """ minimalist version of cEEB for quick calculation (optimisation purposes)
        used in forecast simulations
        """
        
        PET = 0
        for j, eq in enumerate(self.sim.systemModel.equipment):
            FETFuel = Status.int.FETFuel_j[j]
            FETel = Status.int.FETel_j[j]
##            FETHeat = Satus.int.FETHeat_j[j]
##            ElGen = Status.int.ElGen_j[j] (already included in FETel = FETel_c - ElGen

            PET += self.fPE_el*FETel + self.fPE_fuel[j]*FETFuel

        for h, hx in enumerate(self.sim.systemModel.hx):
            E_fan = hx.E_fan
            PET += self.fPE_el*E_fan
            
        return PET



if __name__ == "__main__":
    print "Testing moduleEnergy"
    import einstein.GUI.pSQL as pSQL, MySQLdb
    from einstein.modules.myInterfaces import *
    sql = MySQLdb.connect(user="root", db="einstein")
    DB = pSQL.pSQL(sql, "einstein")
    stat = Status("testModuleEnergy")
    Status.DB = DB

    interf = Interfaces(80,8760)

    print "instantiating moduleEnergy"
    mod = ModuleEnergy()
    mod.runSimulation()
