# -*- coding: cp1252 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	RealTime
#			
#------------------------------------------------------------------------------
#			
#	Module for controlling real time simulations (batch mode)
#
#==============================================================================#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 

from GUI.status import Status
from modules.energy.simulationPlus import SimulationPlus
from lib.fileIO.fileIO import readTableFromFile, CSVTables
from constants import APPDATA, EPSILON, INFINITE
from globals import Globals
from time import sleep
from datetime import datetime, timedelta
from lib.dateTime.dateTimeTools import timeDeltaToHours, hoursToTimeDelta, dateTimeToString, stringToDateTime, dateTimeToH
from lib.auxiliary.float import toFloat
from modules.messageLogger import *
from math import ceil
from database.dataHub import hub
from auxiliary.auxiliary import noneCheck, transpose as transposeTable
import numpy as np
from lib.optimise.optimiser import Optimiser, readOptPars
from foreCast import ForeCast

from modules.energy.hx import NSTORAGE

CHECKUPINTERVAL = 5.0 # minutes for checking availability of new monitoring data

class RealTimeHandler():
    """ Simulation of systems in real time

    source:  source of data (csv-file or data base connection

    mode:    real-time: real-time simulation
             real-time-emulation: emulation of real-time simulation with speed-up
             ex-post: simulation ex-post simulation reading data from specified source

    Time scales used:

    t_data(_sim):     time stamp of monitoring data for boundary conditions
    t_ref_(sim):      time in the reference year (2007) corresponding to t_data(sim)

    t_(sim):          rt-simulation time in hours since simulation start (simulation starting - almost - by zero)

    
    """

    def __init__(self, source, mode=None, forecast=(0, 0, "CONST"), opt=None):

        logTrack("RealTimeHandler.init: source = %s mode = %s"%(source, mode))
        
        self.source = source

        self.sourceType = "dbcon" if  source.endswith(".conf") else "csv"   # csv: data file, dbcon: connection to data base

        if mode is None:
            self.mode = "real-time" if self.sourceType=="dbcon" else "real-time-emulation"
            
        elif mode in ["real-time", "real-time-emulation", "ex-post"]:
            self.mode = mode
            
        else:
            logError("Erroneous mode for real time handler: %s"%mode)
            self.mode = "real-time"


        ### parameters for virtual monitoring data
        self.vDataPtr = None    # only used for virtual mode. pointer for reading of virtual system performance data
        self.vDataTable = None
        self.dtMax = 0.25       # maximum time step allowed. if separation between monitoring data is bigger, data
                                # will be interpolated
        self.dtMin = 0.1        # minimum time step allowed. if monitoring data are available more frequently, some measured data
                                # will be jumped over


        try:
            tfc, dtfc, fcModeStr = forecast
        except:
            logWarning("Erroneous parameters for forecast: %s"%str(forecast))
            tfc, dtfc, fcModeStr = (0, 0, "CONST")

        if tfc > 0 and dtfc > 0:
            self.dtfc = max(0.05, min(dtfc, 1.0))   # limit dtfc and nfc to reasonable time intervals
            self.nfc = int(min((tfc/self.dtfc + 0.5), (168.0/self.dtfc)))
        else:           
            self.dtfc = 0.1         # 1h forecast in 0.1h intervals
            self.nfc = 0

        # translate fcModeStr into fcMode and fcPars
        try:
            index = fcModeStr.index(":")
            self.fcMode = fcModeStr[:index].strip()
            parStr = fcModeStr[index+1:].strip(" ")
            parStr = parStr.strip("[")
            parStr = parStr.strip("]")
        except:
            self.fcMode = "CONST" if fcModeStr is None else fcModeStr.strip()
            parStr = None

        if parStr is not None:
            parSplit = parStr.split(";")
            self.fcPars = [toFloat(split) for split in parSplit]
        else:
            self.fcPars = []
        
        self.foreCasts = {}

        self.restart = None

        self.eqStorage0 = None
        self.hxStorage0 = None


        self.lastData = None

        if mode != "ex-post":
            self.stateFile = CSVTables(APPDATA + "/" + source + ".state", separator="\t", comment="#")
            self.stateKeys = ["[time]", "[eqStorage]", "[hxStorage]", "[lastdata]"]

            state = self.stateFile.read()
            if state is None:
                logMessage("Starting real time simulation from scratch")
            else:
                keys, colTitles, rowTitles, array = state
                if keys == self.stateKeys:
                    tstart = rowTitles[0][0]
                    self.restart = stringToDateTime(tstart)
                    self.eqStorage0 = array[1].transpose()
                    self.hxStorage0 = array[2].transpose()

                    self.lastData = {}
                    for i, par in enumerate(colTitles[3]):
                        # TODO: this works only well if the last data interval between monitoringData and lastData has been fully completed
                        val = array[3][1][i]
                        self.lastData.update({par:[val]})
                    self.lastData.update({"time": [stringToDateTime(rowTitles[3][1])]})
                    logMessage("Restarting real time simulation at t=%s"%tstart)
        else:
            self.stateFile = None
            self.restart = None

        self.opt = None # default
        if opt:
            optData = readOptPars(APPDATA + "/" + opt)
            if optData is not None:
                self.opt, self.optPars, xmin, xmax = optData
##                self.optimiser = Optimiser(self.fopt, options={"maxiter": 20})
                self.optimiser = Optimiser(self.fopt)
                self.optimiser.setBounds((xmin, xmax))

    def __repr__(self):
        string = "RTHandler"
        return string


    def start(self, startTime=None, stopTime=None):
        """
        starts the loop of the real time handler
        """
        if startTime is None: startTime = datetime(2007, 1, 1)
        
        if self.restart is not None and self.restart > startTime:
            startTime = self.restart

        logTrack("RealTimeHandler.start: simulating from %s to %s"%(startTime, stopTime))

        mod = Status.mod.moduleEnergy
        mod.sim = SimulationPlus(mod)
        self.sim = mod.sim
        self.sim.startRTSimulation(eqStorage=self.eqStorage0, hxStorage=self.hxStorage0)

        mod.getPEConvRatios()   # necessary for analysis of simulation results (PET calculation)

##      TODO: commented out. current version is working always in the same model instance
##        if self.nfc > 0:
##            self.fc = SimulationPlus(mod)
##            self.fc.buildModel()
##            self.fcModel = self.fc.systemModel

        # structure for state-file table
        self.stateCols = [[],
                          ["EQ%3d"%(j+1) for j, eq in enumerate(self.sim.systemModel.equipment)],
                          ["HX%3d"%(h+1) for h, hx in enumerate(self.sim.systemModel.hx)]]
        self.stateRows = [["-"],
                          ["%2d"%(Status.TemperatureInterval*i) for i in xrange(Status.NT+2)],
                          ["%1d"%(i+1) for i in xrange(NSTORAGE)]]

##        if self.sourceType=="dbcon":
##            linkedDS = [ds for ds in Status.schedules.dataSets if ds.df.srcType=="dbcon"]
##        elif self.sourceType=="csv":
        linkedDS = [ds for ds in Status.schedules.dataSets if ds.df.fileName==self.source]

        if len(linkedDS)==0:
            logError("Data source for real-time simulation is not linked to any system parameter")
            fileNames = list(set([ds.df.fileName for ds in Status.schedules.dataSets]))
            logError("Datafiles used in the project: %s"%fileNames)
            return
                
        if self.restart:
            self.t0 = self.lastData["time"][0]
            self.last = dateTimeToH(self.t0)
        else:            
            self.last = 0       # end of previous time step in hours
            self.t0 = None      # end of first simulation time step in hours (corresponds to first row in data files)

        self.rawData = []
        self.dataBuffer = {}


##### CHANGE MANUALLY FOR EXPORTING SIMULATION DATA TO DATABASE
        exportData = False

        if self.sourceType=="dbcon":
            self.hub = hub
            hub.checkSource(self.source)
            err = hub.connect(lastRead = self.restart)
            if err > 0:
                logError("RealTime.start: Error in data base connection")
                return
            
            hub.getSensors()          # get sensors w/o input -> takes plantId from configuration file
##            self.rawData = hub.read() # reads all what is there up to now

        elif exportData:
            self.hub = hub
            hub.connStr = "mysql://user:pwd@host/db"
            hub.connect()
            data = hub.read()
            if len(data) >= 4*168:
                print "TEST TEST already too much data in the hub. first delete old stuff"
                exportData = False

        # build dictionary for mapping data into component connections

        self.sensorKeyDict = {}
        self.keyColDict = {}
        self.keyScaleDict = {}
        for ds in linkedDS:
            key = ds.parCode + "[%s]"%ds.objectNo
            if ds.parCode in ["PTInFlow", "PTOutFlow", "mInFlowYear", "mOutFlowYear", "UPHc", "UPHw"]: key += "[%s]"%ds.cNo 
            colNo = ds.col
            if self.sourceType=="dbcon":
                sensorNo = hub.sensors[colNo-1]
                self.sensorKeyDict.update({sensorNo:key})
            else:
                self.keyColDict.update({key:ds.col})

            self.keyScaleDict.update({key:ds.scale})
            

        t = 0
        ctr = 0

        if self.mode=="real-time-emulation":
            readUntil = startTime    # used only for real-time-emulation mode, otherwise by default read all data available on source
        else:
            readUntil = None    # used only for real-time-emulation mode, otherwise by default read all data available on source

        self.eof = False        
        while (self.lastData is None or stopTime is None or self.lastData["time"][0] < stopTime):
            ctr += 1

            if self.mode=="real-time-emulation":
                readUntil += hoursToTimeDelta(CHECKUPINTERVAL/60.0)
                
            # get monitoring data for current time step
            self.monitoringData = self.getMonitoringData(readUntil=readUntil)
            if self.eof: break # stop if end of file reached

            if self.monitoringData and "time" in self.monitoringData.keys():
                t_data = self.monitoringData["time"][0]
                if self.t0 is None:
                    self.t0 = t_data - hoursToTimeDelta(self.dtMax) ## simulate first measurement with time step dtMax
                    self.t0 = max(self.t0, datetime(2007, 1, 1))
                    logTrack("RealTimeHandler starting at t_data = %s"%t_data)
                    self.lastData = self.monitoringData.copy()
                    self.lastData["time"] = self.monitoringData["time"][:]  #  break link (deepcopy)
                    self.lastData["time"][0] = self.t0
                    
                    self.last = dateTimeToH(self.t0)

                  
                t = timeDeltaToHours(t_data - self.t0) + dateTimeToH(self.t0)
                if t > self.last:
                    dt = t - self.last

                    if dt < self.dtMin - EPSILON: # new measurement point too close in time -> skip
                                                  # EPSILON for the case that dtMin in coincidence with time step of measurements
                        continue
                    else:
                        nSteps = int(ceil((dt-EPSILON)/self.dtMax))
                        dt_sim = float(dt)/nSteps
                        dt_timedelta = hoursToTimeDelta(dt_sim)
                        dataSet = interpolateData(self.lastData, self.monitoringData, nSteps)
                        t_sim = self.last
                        t_data_sim = self.lastData["time"][0]
                        for n, cdata in enumerate(dataSet):
                            t_sim += dt_sim
                            t_data_sim += dt_timedelta

##                            print "RealTime.step: dt %s t_sim %s t_data_sim %s cdata %s"%(dt_sim, t_sim, t_data_sim, cdata)
                            self.step(dt_sim, t_data_sim, cdata)

                            if exportData: # this is a hack for writing test data
                                mdot_water = cdata["mInFlowYear[0][0]"][0]
                                Tin_water = cdata["PTInFlow[0][0]"][0]
                                mdot_exhaust = cdata["mOutFlowYear[1][0]"][0]
                                T_exhaust = cdata["PTOutFlow[1][0]"][0]

                                dataList = [(1, t_data_sim, mdot_water),
                                            (2, t_data_sim, Tin_water),
                                            (3, t_data_sim, mdot_exhaust),
                                            (4, t_data_sim, T_exhaust)]

                                hub.write(dataList)

                        if self.nfc > 0:
                            dt_scale = self.dtfc/dt_sim
                            self.foreCastBC(t_data, cdata, dt_scale)

                            if self.opt:
                                self.current = t_data
                                xmin = self.optimiser.getGlobalMin()
                                
##------------ helper for manual testing !!! -----------------------------------------------------------------------------
##                                for i in range(3):
##                                    print "\n---------\nRealTime.start: iteration no. %s"%(i+1)
##                                    TsMax = 30.0 + 20.0*i
##                                    qx_min = 0.0 + 0.3*i
##                                    print "TsMax = %s qx_min = %s"%(TsMax, qx_min)
##                                    self.sim.systemModel.hx[0].setControl(TsMax, qx_min)
####                                    self.sim.startRTSimulation(eqStorage=self.eqStorage0, hxStorage=self.hxStorage0)
##                                    self.simulateForeCast(t_data)
##------------ helper for manual testing !!! -----------------------------------------------------------------------------


                                #TODO: restore control parameters for real system simulation
                                #      currently implemented: direct take-over of optimisation result
                                self.setPars(xmin)

                                # simulate once again in order to produce output
                                self.simulateForeCast(t_data)

                            else:
                                self.simulateForeCast(t_data)

                self.last = t
##                self.lastData["time"] = t_data
                self.lastData = self.monitoringData

            elif readUntil is not None and stopTime is not None and readUntil >= stopTime:  # no monitoring data read and stop time superated
                break
            
            # different conditions for loop interruption, depending on mode
            if self.mode=="real-time": # infinite loop, no interrupt
                sleep(CHECKUPINTERVAL*60)   
            elif self.mode=="ex-post": # break when no more data are available
                if self.monitoringData is None: break

        self.sim.closeRTSimulation()
        if self.stateFile: self.stateFile.close()

    def setBC(self, t0, dtfc, n, cdata):
        """
        sets boundary conditions for step (it=0) or forecast(it=[0:nfc]

        dt: time interval of forecast simulation
        t0: START of first simulation time step (in HOY)
        cdata: values read / interpolated from monitoring
                (each parameter is a list of length n !)
        """

        # get index of discrete (1h -) time interval and load process demand
        it_ref = [int(t0 + it*dtfc + EPSILON)%8760 for it in xrange(n)]

        self.sim.systemModel.setWeatherData([Status.WeatherData.Tdb[it] for it in it_ref],
                                            [Status.WeatherData.Twb[it] for it in it_ref])
                                            
        
        keys = cdata.keys()

        processes = self.sim.systemModel.process
        systemModelGlb = Status.int.systemModel

        
        for k, proc in enumerate(processes):
            (inStreams, outStreams) = (systemModelGlb.processStreamsIn[k], systemModelGlb.processStreamsOut[k])
            PT = systemModelGlb.process[k].PT

            ### TODO CONTINUE HERE: if nfc -> variable time step (matching it vs. it_fc)
            q = Status.int.UPHm_Tt[k].getQ()*dtfc
            UPHm_sim = np.array([q[it] for it in it_ref])
            if "UPHm[%s]"%k in keys:
                UPHm_sim = cdata["UPHm[%s]"%k]
                
            q = Status.int.UPHs_Tt[k].getQ()*dtfc
            UPHs_sim = np.array([q[it] for it in it_ref])
            if "UPHs[%s]"%k in keys:
                UPHs_sim = cdata["UPHs[%s]"%k]

            UPHc_sim = []
            for s, stream in enumerate(inStreams):
                hf = Status.int.UPHc_Tt[k].getPart(s)
                q = hf.getQ()*dtfc
                UPHc_sim.append([q[it] for it in it_ref])

                if "UPHc[%s][%s]"%(k,s) in keys:
                    UPHc_sim[-1] = cdata["UPHc[%s][%s]"%(k,s)]
                else:
                    dm_t = cdata["mInFlowYear[%s][%s]"%(k,s)] if "mInFlowYear[%s][%s]"%(k,s) in keys else None
                    PTin_t = cdata["PTInFlow[%s][%s]"%(k,s)] if "PTInFlow[%s][%s]"%(k,s) in keys else None
                    if not (dm_t is None and PTin_t is None):
                        for it, dm in enumerate(dm_t):
                            PTin = PTin_t[it]
                            dm = noneCheck(dm, stream.mInFlowNom)
                            PTin = noneCheck(PTin, stream.PTInFlow)
                            T1 = max(PTin, PT)
                            T0 = min(PTin, PT)
                        
                            #OJO: dUPXc is used here only as working space. this should not make damage, but to be checked
                            q = self.sim.systemModel.process[k].dUPXc[s]
                            q.setValues_mT(T0, T1, dm)
                            dQ = q.Q
                        
                            UPHc_sim[-1][it] = dQ

            UPHw_sim = []
            for s, stream in enumerate(outStreams):
                hf = Status.int.UPHw_Tt_hf[k].getPart(s)
                q = hf.getQ()*dtfc
                UPHw_sim.append([q[it] for it in it_ref])
                if "UPHw[%s][%s]"%(k,s) in keys:
                    UPHw_sim[-1] = cdata["UPHw[%s][%s]"%(k,s)]
                else:
                    dm_t = cdata["mOutFlowYear[%s][%s]"%(k,s)] if "mOutFlowYear[%s][%s]"%(k,s) in keys else None
                    PTout_t = cdata["PTOutFlow[%s][%s]"%(k,s)] if "PTOutFlow[%s][%s]"%(k,s) in keys else None
                    if not (dm_t is None and PTout_t is None):
                        for it, dm in enumerate(dm_t):
                            PTout = PTout_t[it]
                            dm = noneCheck(dm, stream.mOutFlowNom)
                            PTout = noneCheck(PTout, stream.PTOutFlow)
                            PTfinal = noneCheck(stream.PTFinal, 0.0)
                            T1 = max(PTout, PTfinal)
                            T0 = min(PTout, PTfinal)

                            #OJO: dUPXc is used here only as working space. this should not make damage, but to be checked
                            q = self.sim.systemModel.process[k].dUPXw[s]
                            q.setValues_mT(T0, T1, dm)
                            dQ = q.Q
                            
                            UPHw_sim[-1][it] = dQ

            UPHwReq_sim = []
            for s in range(len(outStreams)):
                q = Status.int.UPHwReq_Tt_hf[k].getPart(s).getQ()*dtfc
                UPHwReq_sim.append([q[it] for it in it_ref])
                if "UPHwReq[%s][%s]"%(k,s) in keys:
                    UPHwReq_sim[-1] = cdata["UPHwReq[%s][%s]"%(k,s)]

            PTin_sim =  [(cdata["PTInFlow[%s][%s]"%(k, s)]  if  "PTInFlow[%s][%s]"%(k, s)  in keys else None) for s in range(len(inStreams))]
            PTout_sim = [(cdata["PTOutFlow[%s][%s]"%(k, s)] if  "PTOutFlow[%s][%s]"%(k, s) in keys else None) for s in range(len(outStreams))]

            self.sim.systemModel.process[k].setDemand(UPHm_sim, UPHs_sim,
                                                      UPHc_sim,
                                                      UPHw_sim,
                                                      UPHwReq_sim,
                                                      PTin_sim,
                                                      PTout_sim)

        equipment = self.sim.systemModel.equipment
        for j, eq in enumerate(equipment):

            hasData = False

            controlScheduleRef = systemModelGlb.eqObject[j].controlSchedule
            if controlScheduleRef:
                eq.setBoundaryConditions([controlScheduleRef[it] for it in it_ref])

            if Status.int.systemModel.eqObject[j].eqClass not in ["CHP", "CH"]:
                continue

            if Status.int.systemModel.eqObject[j].eqClass in ["CHP"]:

                TRejectRef = Status.int.systemModel.eqObject[j].TReject
                if TRejectRef is None:
                    TRejectRef = Status.WeatherData.Tdb

                TReject = [TRejectRef[it] for it in it_ref]

                TDriveRef = Status.int.systemModel.eqObject[j].TDrive
                if TDriveRef is not None:
                    TDrive = [TDriveRef[it] for it in it_ref]
                else:
                    TDrive = None

                hasData = True

            else:
                TDrive = None
                TReject = None

            if "TDrive[%s]"%j in keys:
                TDrive = cdata["TDrive[%s]"%j]
                hasData = True
            if "TReject[%s]"%j in keys:
                TReject = cdata["TReject[%s]"%j]
                hasData = True

            eq.setSpecificBC(TDrive, TReject)

            if "Control[%s]"%j in keys:
                fCtrl = cdata["Control[%s]"%j]
                hasData = True

                eq.setBoundaryConditions(controlSchedule=fCtrl, rt=True)


    def step(self, dt, t_data, cdata):
        """
        carries out a simulation step

        dt: time interval
        t, t_data: END of simulation time step in hours and as datetime
        cdata: values read / interpolated from monitoring
        """
        t0 = dateTimeToH(t_data) - dt
        self.setBC(t0, dt, 1, cdata)

        self.sim.stepRTSimulation(dt, t_data)

        if self.mode != "ex-post":

            tables = self.sim.systemModel.getState()
            tables = [transposeTable(table) for table in tables]
            self.stateRows[0] = [dateTimeToString(t_data)]
            state = [[[]]] + tables
            dataRows = [[dateTimeToString(self.lastData["time"][0]),
                         dateTimeToString(self.monitoringData["time"][0])]]
            dataKeys = [key for key in self.lastData.keys() if key != "time"]
            dataCols = [[key for key in dataKeys]]
            dataVals = [[[self.lastData[key][0] for key in dataKeys],
                         [self.monitoringData[key][0] for key in dataKeys]]]
                      
            self.stateFile.write(self.stateKeys, self.stateCols + dataCols, self.stateRows + dataRows, state + dataVals)
            
##            Globals.conf.set("rt", "time", t_data)
##            Globals.conf.write()


    def foreCastBC(self, t_data, cdata, dt_scale):
        """ forecasts boundary conditions for forecast simulations
        """

        ### dummy forecast: fc = N x current data
        keys = cdata.keys()
        fc = {}
        t0 = dateTimeToH(t_data)

        for key in keys:

            # initialise foreCast for key if not yet created

            if key not in self.foreCasts.keys():
                self.foreCasts.update({key: ForeCast(model=self.fcMode, par=self.fcPars, dt=self.dtfc, n=self.nfc)})
                
            t = t0
            if key.startswith("PT") or key.startswith("T"): # take identical value
                xt = cdata[key][0]
            else:
                xt = cdata[key][0]*dt_scale
            
            self.foreCasts[key].update(xt, t)

            data = self.foreCasts[key].get()
            fc.update({key:data})

        self.setBC(t0, self.dtfc, self.nfc, fc)

    def simulateForeCast(self, t_data, output=True):


        t0 = dateTimeToH(t_data)

        # store current state of the system model before forecast
        state = self.sim.systemModel.getState()

        self.sim.systemModel.simulate(self.dtfc, self.nfc, t0)

        self.sim.systemModel.storeResults()
        mod = Status.mod.moduleEnergy
        PET = mod.calculatePET()

        # produce an output analogous to the DEBUG output of runSimulation
        if output:
            for it in xrange(self.nfc):
                t0n = t0 + it*self.dtfc
                self.sim.writeResultToTable(it, rt="tfc %6.2f"%t0n)



        # reset system model to state before forecast
        eqStorage = state[0]
        hxStorage = state[1]
        
        for j, eq in enumerate(self.sim.systemModel.equipment):
            eq.setInitialValues(eqStorage[j])

        for h, hx in enumerate(self.sim.systemModel.hx):
            hx.setInitialValues(hxStorage[h])

##        print "RealTime.simulateForeCast: t=%s PET=%s"%(t_data, PET)

        if self.opt=="PET":
            return PET
        else:
            return None


    def getMonitoringData(self, readUntil=None):
        """
        readUntil: reads data only until a given datetime (used for real-time emulation)
                    for csv-files as source: reads only one line !!!
        """
        self.eof = False
        if self.sourceType == "csv":
            if self.vDataPtr is None:
                
                fileName = self.source
                self.vDataTable = readTableFromFile(APPDATA + "/" + fileName, separator="\t", comment="#")

                if self.restart is None:
                    self.vDataPtr = 0
                else:
                    self.vDataPtr = len(self.vDataTable)
                    for i in range(1, len(self.vDataTable)-1):
                        current = self.vDataTable[i]
                        time = datetime(2007, 1, 1) + timedelta(hours=toFloat(current[0]))
                        if time > self.restart:
                            self.vDataPtr = i-1
                            break

            if readUntil is None:
                # read in the next line
                self.vDataPtr += 1
                if self.vDataPtr < len(self.vDataTable):
                    current = self.vDataTable[self.vDataPtr]
                else:
                    self.eof = True
                    return None
                time = datetime(2007, 1, 1) + timedelta(hours=toFloat(current[0]))
            else:
                if self.vDataPtr+1 > len(self.vDataTable)-1:   # theres nothing more to read
                    self.eof = True
                    return None
                
                last = None
                for i in range(self.vDataPtr+1, len(self.vDataTable)-1):
                    current = self.vDataTable[i]
                    time = datetime(2007, 1, 1) + timedelta(hours=toFloat(current[0]))
                    if time <= readUntil:
                        last = i
                    else:
                        break

                if last is None:
                    return None
                
                self.vDataPtr = last
                current = self.vDataTable[last]
                time = datetime(2007, 1, 1) + timedelta(hours=toFloat(current[0]))

            data = {"time": [time]}
            for key in self.keyColDict.keys():
                col = self.keyColDict[key]
                val = toFloat(current[col])
                val = None if val is None else self.keyScaleDict[key]*val
                data.update({key: [val]})

            return data

        else:

            keys =  self.sensorKeyDict.values()
            sensors = self.sensorKeyDict.keys()

            self.rawData += (self.hub.read(readUntil=readUntil))

            data = self.dataBuffer
            self.dataBuffer = {}
                
            while not all([(key in data.keys()) for key in keys]):
                if len(self.rawData)==0:
                    break
                
                row = self.rawData.pop(0)

                if len(row) > 0:
                    s, t, val = row
                    if s in sensors:
                        key = self.sensorKeyDict[s]
                        val = toFloat(val)
                        val = None if val is None else self.keyScaleDict[key]*val
                        data.update({key: [val],
                                     "time": [t]})

            if all([(key in data.keys()) for key in keys]):
                return data
            else:
                self.dataBuffer.update(data)
                return None


    def setPars(self, x):
        """ helper function for optimisation
        reads the vector of independent parameters x and sets the corresponding
        parameters in the system model
        """

        for i, par in enumerate(self.optPars):
##            print par, "=", x[i]

            if par.startswith("HX["):
                try:
                    n = par.index("]")
                    hxNo = int(par[3:n])
                    h = hxNo-1
                    par = par[n+2:]
                    if par.startswith("qx_min"):
                        self.sim.systemModel.hx[h].setControl(INFINITE, x[i])
                    else:
                        assert(False)
                except:
                    logError("Could not understand optPar %s"%par)
                    pass
        
    def fopt(self, x):
        """ optimiser function """

        self.setPars(x)
        ret = self.simulateForeCast(self.current, output=False)

        return ret
        
                
def interpolateData(lastData, currentData, nSteps):
    """
    Current solution adopted for the interpolation problem:
    - linear interpolation for temperature data
    - no interpolation for flow data
    - flows in data are assumed to be cumulative values (energy or mass
      instead of (avg) power and (avg) mass flows
      -> therefore divided by nSteps
    """
    keys = currentData.keys()
    keys.remove("time")

    dataSet = []
    for n in xrange(0, nSteps):
        f = (n+0.5)/nSteps # for each interval choose Tavg = 0.5*(T_n + T_n-1)
        row = {}
        for key in keys:
            if key.startswith("PT") or key.startswith("T"): # linear interpolation of temperature data
                row.update({key: [f*currentData[key][0] + (1-f)*(lastData[key][0])]})
##            elif key.startswith("m"):
##                row.update({key: currentData[key]})
##                
            else: # step change and divide by nsteps for flow data
                row.update({key: [currentData[key][0]/nSteps]})

        dataSet.append(row)
        
    return dataSet


if __name__ == "__main__":
    pass
