# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	simulationNP
#			
#------------------------------------------------------------------------------
#			
#	system simulation module based on NumPy - arrays
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
import copy
import wx

from constants import *
from GUI.status import *
from modules.messageLogger import *
from GUI.dialogGauge import DialogGauge
from modules.equipments import getConnectedPipes
from modules import equipments
from modules.fluids import Fluid, Fuel
from modules.interfaces import calcQ_T
from auxiliary.auxiliary import toGettextAndUnicode as _U
from GUI.GUITools import check


class SimulationNP(object):
    def __init__(self, moduleEnergy):
        self.mod = moduleEnergy
        # Keep for each equipment id the id of the pipes that 
        # depend on them more or less heavily.
        self.pipesWithPriority = []  
        self.pipesWithoutPriority = []

    def calculateEnergyFlows(self, equipe, cascadeIndex):
        """Dummy function for calculation of energy flows in any equipment"""
        
        Status.int.extendCascadeArrays(cascadeIndex)
        PNom = equipe.HCGPnom
        if PNom is None:
            logTrack("ModuleEnergy (cEF-dummy): no equipment capacity specified")
            PNom = 0

        COPh_nom = equipe.HCGTEfficiency
        if COPh_nom is None:
            logTrack("ModuleEnergy (cEF-dummy): no equipment COP specified")
            COPh_nom = 0.90

        NT = Status.NT
        Nt = Status.Nt
        Dt = Status.TimeStep

        # get demand data for CascadeIndex/EquipmentNo from Interfaces
        # and create arrays for storing heat flow in equipment
        QD_Tt = Status.int.QD_Tt_mod[cascadeIndex - 1].copy()
        QA_Tt = Status.int.QA_Tt_mod[cascadeIndex - 1].copy()

        USHj_Tt = Status.int.USHj_Tt[cascadeIndex - 1]
        QHXj_Tt = Status.int.QHXj_Tt[cascadeIndex - 1]
        QWHj_Tt = Status.int.QWHj_Tt[cascadeIndex - 1]

        # start simulation
        USHj = 0
        QHXj = 0
        QWHj = 0
        HPerYear = 0

        # outer iterative cycle: over all time steps
        for it in range(Nt):
            # inner iterative cycle: temperature intervals
            for iT in range(NT + 2):   #+1 = the > 400 ?C case
                USHj_Tt[iT][it] = min(QD_Tt[iT][it], PNom * Dt) #just max(demand,nom.power) 
                QHXj_Tt[iT][it] = 0
                QWHj_Tt[iT][it] = 0

            USHj += USHj_Tt[NT + 1][it]
            QHXj += QHXj_Tt[NT + 1][it]
            QWHj += QWHj_Tt[NT + 1][it]
            if USHj_Tt[NT + 1][it] > 0:
                HPerYear += Dt
        # end simulation. 

        # storing results in Interfaces
        # remaining heat demand and availability for next equipment in cascade
        Status.int.QD_Tt_mod[cascadeIndex] = QD_Tt
        Status.int.QD_T_mod[cascadeIndex] = Status.int.calcQ_T(QD_Tt)
        Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
        Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)

        # heat delivered by present equipment                            
        Status.int.USHj_Tt[cascadeIndex - 1] = USHj_Tt
        Status.int.USHj_T[cascadeIndex - 1] = Status.int.calcQ_T(USHj_Tt)
        Status.int.USHj_t[cascadeIndex - 1] = USHj_Tt[Status.NT + 1].copy()

        # waste heat absorbed by present equipment                            
        Status.int.QHXj_Tt[cascadeIndex - 1] = QHXj_Tt
        Status.int.QHXj_T[cascadeIndex - 1] = Status.int.calcQ_T(QHXj_Tt)
        Status.int.QHXj_t[cascadeIndex - 1] = QHXj_Tt[Status.NT + 1].copy()

        # waste heat produced by present equipment
        Status.int.QWHj_Tt[cascadeIndex - 1] = QWHj_Tt
        Status.int.QWHj_T[cascadeIndex - 1] = Status.int.calcQ_T(QWHj_Tt)
        Status.int.QWHj_t[cascadeIndex - 1] = QWHj_Tt[Status.NT + 1].copy()

#        Status.int.cascadeUpdateLevel = cascadeIndex

        # Global results (annual energy flows)
        Status.int.USHj[cascadeIndex - 1] = USHj * Status.EXTRAPOLATE_TO_YEAR
        Status.int.QWHj[cascadeIndex - 1] = QWHj * Status.EXTRAPOLATE_TO_YEAR
        Status.int.QHXj[cascadeIndex - 1] = QHXj * Status.EXTRAPOLATE_TO_YEAR

        if COPh_nom > 0:
            FETFuel_j = USHj / COPh_nom
        else:
            FETFuel_j = 0.0
            logWarning("Strange equipment with COP = 0.0")
        FETel_j = 0.0

        Status.int.FETFuel_j[cascadeIndex - 1] = FETFuel_j
        Status.int.FETel_j[cascadeIndex - 1] = 0.0
        Status.int.HPerYearEq[cascadeIndex - 1] = HPerYear * Status.EXTRAPOLATE_TO_YEAR

        logMessage("Dummy: eq.no.:%s USH: %s FETFuel: %s FETel: %s HPerYear: %s [MWh]" % \
                   (equipe.EqNo, \
                    USHj * Status.EXTRAPOLATE_TO_YEAR / 1000.0, \
                    0.0, \
                    FETel_j * Status.EXTRAPOLATE_TO_YEAR / 1000.0, \
                    HPerYear * Status.EXTRAPOLATE_TO_YEAR / 1000.0))

        self.calculateOM(equipe, USHj * Status.EXTRAPOLATE_TO_YEAR)
        
        return (USHj_Tt.copy(), USHj)    
    
    def _calculatePipePriorities(self):
        """Calculate for each pipe the last cascade level at which it
        can cover its demand, and save these for the appropriate level 
        as a pipe with priority.
        
        The results are kept in self.pipesWithPriority and 
        self.pipesWithoutPriority.
        
        """
        NM = len(Status.int.USHLink)
        NJ = self.NJ
        mcimax = [0] * NM
        for m in range(NM):
            for j in range(NJ):
                if Status.int.USHLink[m][j]==1:
                    mcimax[m] = max(self.equipments[j].CascadeIndex, mcimax[m])

        self.pipesWithPriority = [[] for dummy in range(NJ)]
        self.pipesWithoutPriority = [[] for dummy in range(NJ)]
        for j in range(NJ):
            ci = self.equipments[j].CascadeIndex
            for m in range(NM):
                if mcimax[m] == ci:
                    self.pipesWithPriority[j].append(m)
                elif Status.int.USHLink[m][j]==1:
                    self.pipesWithoutPriority[j].append(m)
    
    def _distribute(self, demandm_Tt, coveredDemand_Tt, equipmentIndex):
        """Distribute the coveredDemand between the pipes that are 
        connected to the equipment with given index. 
        
        :param demandm_Tt: The demand of all pipes that is still to be
        covered.
        :type demandm_Tt: A three dimensional numpy array with the 
        dimensions [pipeindex][temp][time].
        :param coveredDemand_Tt: The demand that could be covered in
        this cascade.
        :type coveredDemand_Tt: A two dimensional numpy array with the
        dimensions [temp][time] 
        :param equipmentIndex: The index of the equipment that delivers
        the energy that is to be distributed.
        :type equipmentIndex: An integer, between 0 and number of 
        equipments - 1.
        
        :returns: The remaining demand in the same format as demandm_Tt.
        
        """
        remainingDemandChanged = Status.int.USHjChanged[equipmentIndex]
        
        USHAvailable_Tt = coveredDemand_Tt.copy()
        USHDemandRemainingm_Tt = [item.copy() for item in demandm_Tt]
        
        priorityPipes = self.pipesWithPriority[equipmentIndex]
        nonPriorityPipes = self.pipesWithoutPriority[equipmentIndex]

        # sum up the demand of the priority pipes
        USHReqPriority_Tt = [demandm_Tt[m] for m in priorityPipes]
        USHReqPriority_Tt = np.sum(USHReqPriority_Tt, 0)

        for m in (priorityPipes + nonPriorityPipes):
            remainingDemandChanged = remainingDemandChanged or Status.int.USHm_remainingChanged[m]
            Status.int.USHm_remainingChanged[m] = remainingDemandChanged 
        
        np.seterr(divide="ignore", invalid="ignore")
        # Calculate the ratios of covered demand and demand, 
        # with a maximum of 1.
        f = USHAvailable_Tt/USHReqPriority_Tt
        # Default an entry to 0 if the ratio NaN due to 0/0.
        f[np.isnan(f)] = 0
        f[np.isinf(f)] = 0
        f = f.clip(max=1)
        np.seterr(divide="warn", invalid="warn")
        
        # decrease available and and remaining demand accordingly 
        USHAvailable_Tt -= f*USHReqPriority_Tt
        for m in priorityPipes:
            USHDemandRemainingm_Tt[m] *= (1-f)
        
        # sum up the demand of the non priority pipes
        USHReqNonPrio_Tt = [demandm_Tt[m] for m in nonPriorityPipes]
        USHReqNonPrio_Tt = np.sum(USHReqNonPrio_Tt, 0)
        
        # Calculate the ratios of covered demand and demand, 
        # with a maximum of 1.
        np.seterr(divide="ignore", invalid="ignore")
        f = USHAvailable_Tt/USHReqNonPrio_Tt
        np.seterr(divide="warn", invalid="warn")
        # Default an entry to 0 if ratio is NaN due to 0/0.
        f[np.isnan(f)] = 0
        f[np.isinf(f)] = 0
        f = f.clip(max=1)
        
        # decrease available and and remaining demand accordingly 
        USHAvailable_Tt -= f*USHReqNonPrio_Tt

        for m in nonPriorityPipes:
            USHDemandRemainingm_Tt[m] *= (1-f)
        
        return USHDemandRemainingm_Tt

    def runSimulation(self, first=1, last=None, loop=True, checkPeakPower=True):
        """Update the energy flows for equipment cascade from first to last.
        
        If no last is given all equipments are simulated.
        If loop is True the simulation is run in turns with the heat
        recovery until the numbers stabilise.

        :params checkPeakPower: if False, check whether current equipment capacity is sufficient is not carried out
                                this is thought for use in simulations in intermediate stages of equipment cascade design (e.g. auto-design boilers)
        """
        logTrack("ModuleEnergy (runSimulation): starting")

        try:
            QWHEqConnected = (np.max(Status.QWHEqLink)>0)
        except:
            QWHEqConnected = False #exception occurs if there is no heat exchanger (one dimension of array with length zero)
        
        if Status.processData.outOfDate is True:
            Status.processData.createAggregateDemand()
            
        elif Status.int.heatRecoveryOutOfDate == True:
            # process data is fine but there was a change in the equipment
            # waste heat or in the heat recovery setup
            if Status.schedules.equipmentSchedulesOutOfDate is True:
                equipments.estimateEquipmentWasteHeat()
            Status.mod.moduleHR.runHRModule()   # update waste heat calculations
            
        self.mod.getEquipmentList()
        self.NJ = self.mod.NJ
        self.equipments = self.mod.equipment
        
        self.NPipe = len(Status.int.USHLink)
        self._calculatePipePriorities()

        self.NH = Status.int.systemModel.NH
        
        # initializing storage space for energy flows in cascade
        first = max((first, 1, Status.int.cascadeUpdateLevel + 1))    #avoid unnecessary calculations

        if last == None or last > self.NJ:
            last = self.NJ

        Status.int.extendCascadeArrays(self.NJ)
        Status.int.USHjRequiredChanged = [True] * self.NJ

        logTrack("SimulationNP.runSimulation: simulating from %s to %s" % (first, last))

        # now calculate the cascade
        # call the calculation modules for each equipment
        dlg = DialogGauge(Status.main, _("EINSTEIN system simulation"), _("calculating energy flows"))

        # get the correct initial demands of the first pipes 
        # use the former state otherwise
        self.updateUSHm_remaining(min(first, self.NJ))

        stepsDone = 0
        for cascadeIndex in range(first, last + 1):
            jc = cascadeIndex - 1
            eq = Status.int.systemModel.equipment[jc]
            equipeClass = getEquipmentClass(eq.EquipType)

            logTrack("=================================\nModuleEnergy (runSimulation): %s - %s [%s: %s]" % \
                     (cascadeIndex, eq.Equipment, equipeClass, eq.EquipType))
            
            # Get demand for the current equipment, by
            # summing up the remaining demand of the connected pipes: USHj_Required[j] = sum(all connected m's) USHm_remaining[m]

            USHjRequired_Tt = self._calculateCummulativeDemand(cascadeIndex, eq, Status)
            
            logTrack("Cumulative Demand h/c = %s/%s.\n--> calling calculateEnergyFlows"%(sum(USHjRequired_Tt[Status.NT+1]),sum(USHjRequired_Tt[0])))            

            # Reset optional output values which are written only by some of the equipment types
            Status.int.waterConsumption_j[jc] = 0   # only different from 0 for chillers

            Status.int.USHjChanged[jc] = False
            Status.int.QWHjChanged[jc] = False
            Status.int.QHXEqReqChanged[jc] = False

            if Status.int.USHjRequiredChanged[jc] or Status.int.QHXEqChanged[jc] or loop: #loop = True: in first pass of global iteration simulate all equipments
                #TODO: the if-criteria are not sufficient for heat pumps with generic waste heat!
                # Run the appropriate simulation.
                if equipeClass == "HP":
                    (USHj_Tt, dummy) = Status.mod.moduleHP.calculateEnergyFlows(eq, 
                                                                                USHjRequired_Tt, 
                                                                                cascadeIndex)
                elif equipeClass == "BB":
                    (USHj_Tt, dummy) = Status.mod.moduleBB.calculateEnergyFlows(eq, 
                                                                             USHjRequired_Tt, 
                                                                             cascadeIndex)
                elif equipeClass == "ST":
                    (USHj_Tt, dummy) = Status.mod.moduleST.calculateEnergyFlows(eq,
                                                                                USHjRequired_Tt,
                                                                                cascadeIndex)
                elif equipeClass == "CHP":
                    (USHj_Tt, dummy) = Status.mod.moduleCHP.calculateEnergyFlows(eq, 
                                                                                 USHjRequired_Tt, 
                                                                                 cascadeIndex)
                elif equipeClass == "CH":
                    (USHj_Tt, dummy) = Status.mod.moduleCH.calculateEnergyFlows(eq, 
                                                                                USHjRequired_Tt, 
                                                                                cascadeIndex)

                else:
                    logWarning("WARNING: unknown equipment type %s system"\
                               " simulation module will run a calculateEnergyFlows-dummy" % (eq.EquipType, ))
                    (USHj_Tt, dummy) = self.calculateEnergyFlows(eq, cascadeIndex)

                Status.int.QWHEq_Tt[jc].exportGlb(Status.int.systemModel.eqObject[jc].QWHGlb)
                Status.int.QHXEqReq_Tt[jc].exportGlb(Status.int.systemModel.eqObject[jc].QHXGlb)
            else:
                logTrack("simulation of equipment with cascadeIndex %s not necessary. already converged"%(jc+1))
                USHj_Tt = Status.int.USHj_Tt[jc]
                Status.int.cascadeUpdateLevel = jc+1 #this is done otherwise in the calculate energy flows - functions
                
            # Distribute the covered demand over the pipes.
            Status.int.USHm_remaining_Tt = self._distribute(Status.int.USHm_remaining_Tt, 
                                                            USHj_Tt, 
                                                            cascadeIndex - 1)

            Status.int.cascadeDistributionLevel = cascadeIndex

            logTrack("ModuleEnergy (runSimulation): end simulation=====================================")

            dlg.update(100.0 * (cascadeIndex - first + 1) / (last - first + 1))

        # update the pointer to the last calculated cascade
        Status.int.cascadeUpdateLevel = last
        
        # if a full cascade has been calculated, calculate the balances
        logTrack("ModuleEnergy (runSimulation): arrived at the end. %s [%s]" % \
                 (last, self.NJ))

        stepsDone += 1
        dlg.Destroy()

        if last == self.NJ and first <= self.NJ:  # check results only if something has to be simulated ! security feature, gave problems in runSimulation called from solar thermal DA.
            #test if equipment is sufficient for covering demands at any time t, else give a warning.

            assert (Status.int.cascadeDistributionLevel == last)
            
            uncoveredDemand = 0
            totalDemand = 0
            uncoveredDemandC = 0
            totalDemandC = 0
            for m in range(len(Status.int.USHm_remaining_Tt)):
                if Status.int.systemModel.pipeObject[m].heating:
                    iTmax = Status.NT+1
                    uncoveredDemand += sum(Status.int.USHm_remaining_Tt[m][iTmax])
                    totalDemand +=  sum(Status.int.USHCm_Tt[m][iTmax])
                else:
                    iTmax = 0
                    uncoveredDemandC += sum(Status.int.USHm_remaining_Tt[m][iTmax])
                    totalDemandC +=  sum(Status.int.USHCm_Tt[m][iTmax])
            
            logTrack("ModuleEnergy (runSimulation): updating Energy balances")
            if loop == True:

                for i in range(3):
                    USH0 = Status.int.USHTotal
                    USC0 = Status.int.USCTotal
                    QWHEq0 = Status.int.QWHEqTotal
                    QHXEq0 = Status.int.QHXEqTotal
                    QHXEqReq0 = Status.int.QHXEqReqTotal
                    FETHeat0 = Status.int.FETHeatTotal
                    self.mod.calculateEquipmentEnergyBalances(calledFromRunSimulation=True)
                    self.getEquipmentTotals()

                    USH = Status.int.USHTotal
                    USC = Status.int.USCTotal
                    QWHEq = Status.int.QWHEqTotal
                    QHXEq = Status.int.QHXEqTotal
                    QHXEqReq = Status.int.QHXEqReqTotal
                    FETHeat = Status.int.FETHeatTotal

                    logTrack("ModuleEnergy - iterative cycle i %s (USH): %s -> %s (QWH): %s -> %s (FETh): %s -> %s  (QHX): %s -> %s" % \
                                          (i, USH0, USH, QWHEq0, QWHEq, FETHeat0, FETHeat, QHXEq0, QHXEq))

                    USXref = 0.0001*max((USH+USH0,USC+USC0))

                    if abs(QWHEq - QWHEq0) < USXref and abs(FETHeat - FETHeat0) < USXref and abs(QHXEq - QHXEq0) < USXref and abs(QHXEqReq - QHXEqReq0) < USXref and not (QWHEqConnected and i==0):
                        logTrack(">>>>>>>>>>>>>> Convergence reached")
                        break
                    else:
                        Status.mod.moduleHR.runHRModule()   # update waste heat calculations
                        #to do: update of pipes / USH values should be done explicitely (calculateUSHCm -> get out of runHRModule
                        self.runSimulation(1, self.NJ, loop=False, checkPeakPower=checkPeakPower)

                # Use the (mostly) converged data to calculate the costs anew.
                if checkPeakPower:
                    if uncoveredDemand > 0.005 * totalDemand:
                        showWarning("Revise your design.\nCurrent equipment capacity is not sufficient for covering the %s demand\n"\
                                    "Remaining %s demand: %s [MWh]" % ("heat","heat", uncoveredDemand / 1000.0))

                    if uncoveredDemandC > 0.005 * totalDemandC:
                        showWarning("Revise your design.\nCurrent equipment capacity is not sufficient for covering the %s demand\n"\
                                    "Remaining %s demand: %s [MWh]" % ("cooling","cooling",uncoveredDemandC / 1000.0))

                Status.prj.setStatus("Energy")  #probably redundant, already done in cEEB

                if DEBUG != "OFF" or (hasattr(Status, "runningTest") and Status.runningTest is not None):
                    from lib.fileIO.fileIO import writeTableToFile
                    table = []

                    nh_results = 0 if Status.HRTool=="estimate" else self.NH
                    table.append(["time"] +
                                 [("USHj[%s]"%(j+1)) for j in xrange(self.NJ)] +
                                 [("ElGen[%s]"%(j+1)) for j in xrange(self.NJ) if Status.int.systemModel.eqObject[j].eqClass=="CHP"] +
                                 [("QHX[%s]"%(h+1)) for h in xrange(nh_results)] +
                                 [("QWXUsed[%s]"%(h+1)) for h in xrange(nh_results)])
                    
                    for it in xrange(Status.Nt):
                        table.append([it] +
                                     [str(Status.int.USHj_t[j][it]).replace(".",",") for j in xrange(self.NJ)] +
                                     [str(Status.int.ElGen_t[j][it]).replace(".",",") for j in xrange(self.NJ) if Status.int.systemModel.eqObject[j].eqClass=="CHP"] +
                                     [str(Status.int.hx[h].qhx.getQ(it)).replace(".",",") for h in xrange(nh_results)] +
                                     [str(Status.int.hx[h].qsrcUsed.getQ(it)).replace(".",",") for h in xrange(nh_results)])

                    writeTableToFile(table, "simulationResults.csv")

                for it in SNAPSHOTS:
                    self.printSnapShot(it)
        
    def _calculateCummulativeDemand(self, cascadeIndex, equipe, Status):
        """Return the demand (sum of remaining demand of connected pipes) for 
        the equipment at cascadeIndex out of USHC_remaining_Tt
        
        :param cascadeIndex (from 1 to NJ)
        :type cascadeIndex: Integer
         
        :returns: The demand to be covered by the equipment as demand_Tt.
        
        """
        assert(Status.int.cascadeDistributionLevel == cascadeIndex-1)        
        assert(cascadeIndex == equipe.CascadeIndex)
        
        connectedPipesIndices = getConnectedPipes(cascadeIndex - 1, Status)

        logTrack("getCumulativeDemand: cascade index : %s found pipes: %s"%(cascadeIndex, connectedPipesIndices))
        cummulativeDemand = np.array([Status.int.USHm_remaining_Tt[m] 
                                      for m in connectedPipesIndices])
        demandChanged = False
        for m in connectedPipesIndices: demandChanged = demandChanged or Status.int.USHm_remainingChanged[m]
        Status.int.USHjRequiredChanged[cascadeIndex-1] = demandChanged
        
        if len(cummulativeDemand)>0:
            cummulativeDemand = np.sum(cummulativeDemand, 0)
        else:
            cummulativeDemand = np.zeros(np.shape(Status.int.USHm_remaining_Tt[0]))

        jc = cascadeIndex-1

        Status.int.QD_Tt_mod[jc] = cummulativeDemand
        Status.int.QD_T_mod[jc] = Status.int.calcQ_T(cummulativeDemand)

        return cummulativeDemand


    def updateUSHm_remaining(self, cascadeIndex):
        """Update the USHm_remaining for a given cascadeIndex (from 1 to NJ)
        so that runSimulation(first = cascadeIndex) works well.
        after this function cascadeDistributionLevel = cascadeIndex - 1
        
        :param cascadeIndex (from 1 to NJ)
        :type cascadeIndex: Integer
         
        :returns: USHm_remaining_Tt 
        
        """
        logTrack("updateUSHm_remaining updating to cascadeIndex %s. cascadeUpdateLevel = %s cascadeDistributionLevel = %s" %
                 (cascadeIndex, Status.int.cascadeUpdateLevel,Status.int.cascadeDistributionLevel))

##        try:
##            for m, elem in enumerate(Status.int.USHm_remainingChanged):
##                logTrack("TEST TEST m %s USHm_remainingChanged %s"%(m,Status.int.USHm_remainingChanged[m]))
##        except:
##            logTrack("TEST TEST updateUSHm_remaining - Changed-array not yet defined")

        equipment = Status.int.systemModel.equipment
        
        if Status.int.cascadeDistributionLevel >= cascadeIndex:
            Status.int.cascadeDistributionLevel = -1

        if Status.int.cascadeDistributionLevel == - 1:

            Status.int.USHm_remainingChanged = copy.deepcopy(Status.int.USHCmChanged)
            Status.int.USHm_remaining_Tt = [item.copy() for item in Status.int.USHCm_Tt]

            Status.int.cascadeDistributionLevel = 0
            if len(equipment) > 0:
                
                self._calculateCummulativeDemand(1, equipment[0], Status)
            else:
                Status.int.QD_Tt_mod[0] = Status.int.createQ_Tt()
                Status.int.QD_T_mod[0] = Status.int.createQ_T()
                            
        for ci_previous in range(Status.int.cascadeDistributionLevel+1, cascadeIndex):
            if Status.int.cascadeUpdateLevel < ci_previous:
                logTrack("ModuleEnergy.updateUSHm_remaining: ERROR - cannot update to cascadeIndex %s. cascadeUpdateLevel = %s"
                         % (cascadeIndex, Status.int.cascadeUpdateLevel))
                break

            jp = ci_previous-1
            
            Status.int.USHm_remaining_Tt = self._distribute(Status.int.USHm_remaining_Tt, 
                                                            Status.int.USHj_Tt[jp], 
                                                            jp)

            Status.int.cascadeDistributionLevel = ci_previous

            # calculate and store in QD_Tt_mod the demand for the next equipment
            self._calculateCummulativeDemand(ci_previous+1, equipment[jp+1], Status)
                        
##        for m, elem in enumerate(Status.int.USHm_remainingChanged):
##            logTrack("TEST TEST m %s USHm_remainingChanged %s"%(m,Status.int.USHm_remainingChanged[m]))
            
        return Status.int.USHm_remaining_Tt

    def getUSHjRequired(self, cascadeIndex):
        """gets the demand (sum of remaining demand of connected pipes) for equipment at cascadeIndex
            function for getting demand of equipment j from outside of ModuleEnergy
        
        :param cascadeIndex (from 1 to NJ)
        :type cascadeIndex: Integer
         
        :returns: The demand to be covered by the equipment as demandm_Tt.
        
        """

        if Status.int.cascadeUpdateLevel < cascadeIndex - 1:    # this case should not happen, if both levels are reset simultaneously in "changeInCascade", but ...
            logTrack("getUSHjRequired: ERROR cannot calculate demand for equipment at level %s in cascade, as cascade is updated only to level %s"%(cascadeIndex,Status.int.cascadeUpdateLevel))
            return (Status.int.createQ_Tt())

        if Status.int.cascadeDistributionLevel < cascadeIndex - 1: # this can happen, if some calculateEnergyFlows is called independently of "runSimulation". Then the QD_Tt_mods are not updated.
                                                                   # then has to be reconstructed from USHCm and redistributing the USHjs of previous levels in the cascade
            self.updateUSHm_remaining(cascadeIndex-1)

        return Status.int.QD_Tt_mod[cascadeIndex-1]

    def calculateOM(self, equipe, USH):
        import database.dbio as dbio
        OMFix = equipe.OandMfix
        OMVar = equipe.OandMvar
        try:
            OM = OMFix + OMVar * USH
        except:
            logTrack(_("OM costs for equipment %s could not be calculated") % equipe.Equipment)
            OM = 0.0

        equipe.OandM = dbio.check(OM)

        dbio.commit()

    def getEquipmentTotals(self):
        # estimate equipment waste heat from annual QWHj and equipment schedules
        Status.int.QWHEqTotal_Tt = Status.int.createQ_Tt()
        for QWHj_Tt in Status.int.QWHj_Tt:
            Status.int.QWHEqTotal_Tt += QWHj_Tt
                    
        Status.int.QWHEqTotal_t = Status.int.QWHEqTotal_Tt[0].copy()
        Status.int.QWHEqTotal_T = Status.int.calcQ_T(Status.int.QWHEqTotal_Tt)
        Status.int.QWHEqTotal = Status.int.QWHEqTotal_T[0]

#================================================================================
# Helper function for debugging
#================================================================================
        
    def printSnapShot(self, it):
######        cdef int h, i, j, k, m
######        cdef Process proc
######        cdef Equipment eq
######        cdef Pipe pipe
######        cdef HX hx
######        cdef int NT2 = Status.NT+2, iT, iL, it, Nt = Status.Nt
######        cdef int NSTORAGE = 10            

        print "\nsimModel snapshot for it = %s"%it
######        print "Tamb = %s"%Status.Weather.Tamb[it]

        systemModel = Status.int.systemModel
        
        for k, proc in enumerate(systemModel.processObject):
            index = Status.NT+1 if proc.heating else 0
            print "\nProcess %s"%k
            print "dUPX: %s"%Status.int.UPH_Tt[k][index][it] 
            print "dUPXm: %s"%Status.int.UPHm_Tt[k].getQ(it) 
            print "dUPXs: %s"%Status.int.UPHs_Tt[k].getQ(it)
            print "dUPXc: %s"%Status.int.UPHc_Tt[k].getQ(it) 
            print "dUPXProc: %s"%Status.int.UPHProc_Tt[k][index][it]
######            print "dUPC: %s"%proc.dUPC
            print "dUPXw: %s"%Status.int.UPHw_Tt[k][index][it]
            
        for h, hx in enumerate(Status.int.hx):
            print "\nHX %s"%h
            if hx.qsrc is not None: print "qsrc: %s"%hx.qsrc.getQ(it)
            if hx.qsnk is not None: print "qsnk: %s"%hx.qsnk.getQ(it)
            if hx.qhx is not None: print "qhx:  %s"%hx.qhx.getQ(it)
            if hx.qsrcUsed is not None: print "qsrcUsed: %s"%hx.qsrcUsed.getQ(it)
            if hx.qsrcRes is not None: print "qsrcRes: %s"%hx.qsrcRes.getQ(it)
            
######            if hx.MCStorage > 0:
######                print "TStorage:",
######                for n in xrange(NSTORAGE):
######                    print " %7.2f"%hx.Ts[n],
######                print ""
            
        for m, pipe in enumerate(systemModel.pipe):
            pipeObj = systemModel.pipeObject[m]
            index = Status.NT+1 if pipeObj.heating else 0
            print "\nPipe %s"%m
            print "dUSXPipe: %s"%Status.int.USHCPipe_Tt[m][index][it]

            print "dUSXPipe_T / dUSXm_T:"
            for iT in range(Status.NT+2):
                print "%2d: %10.4f %10.4f"%(iT, Status.int.USHCPipe_Tt[m][iT][it], Status.int.USHCm_Tt[m][iT][it])
            
        for j, eq in enumerate(systemModel.equipment):
            print "\nEquipment %s"%j

            print "USXj %s"%Status.int.USHj_t[j][it]                
######            print "QWXj %s"%eq.dQWXj                
######            print "QHXj %s"%eq.dQHXj                
######            print "FETHeat %s"%eq.dFETHeat
######            print "QHXEqReq %s"%eq.dQHXEqReq
######
            print "FETel_j %s"%Status.int.FETel_j_t[j][it]
            print "ElGen_j %s"%Status.int.ElGen_t[j][it]
######            print "FETfuel_j %s"%eq.dFETfuel_j
######            print "FETHeat %s"%eq.dFETHeat_j
######
######            if hasattr(eq, "printState"):
######                eq.printState()

            print "dUSXjRequired_T:"
            for iT in range(Status.NT+2):
                print "%2d: %s"%(iT, Status.int.QD_Tt_mod[j][iT][it])


if __name__ == "__main__":
    pass
