# -*- coding: cp1252 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	ModuleEnergy
#			
#------------------------------------------------------------------------------
#			
#	Module for simulation of energy flows in the system
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 

from sys import *
from math import *
from numpy import *
import copy
import wx

from constants import *
from GUI.status import *
from modules.messageLogger import *
from GUI.dialogGauge import DialogGauge
from modules.energy.simModel import Model
from modules.energy.equipment import Equipment
from modules.energy.pipe import Pipe
from modules.energy.process import Process
from modules.energy.boiler import Boiler
from modules.heatRecovery.hxConnection import buildHXNet
#from modules.energy.hx import HX
from modules.energy.estimateHR import HRRATIO_DEFAULT, HRDELTAT_DEFAULT

from modules.fluids import Fuel, Fluid

from modules.equipments import getConnectedPipes
from model.systemConnections import getAllProcessesConnectedTo, getAllMandCoolDownProcessesConnectedTo, getAllEquipmentsConnectedTo
from lib.dateTime.dateTimeTools import dateTimeToHOY, dateTimeToString

#============================================================================== 
#============================================================================== 
class SimulationPlus(object):
    def __init__(self, moduleEnergy):
        self.mod = moduleEnergy
        # Keep for each equipment id the id of the pipes that 
        # depend on them more or less heavily.
        self.pipesWithPriority = []  
        self.pipesWithoutPriority = []

        self.outputLevel = self.mod.outputLevel

        self.resultTable = None

    def buildModel(self, outputLevel=0):

        """ sets up the system model for step by step simulation
            has to be called only once after the last change in the system
        """

        estimateMode = (Status.HRTool == "estimate") and (Status.ANo > 0) # In present state, no estimate mode calculations are carried out
        self.systemModel = Model(estimateMode, outputLevel=outputLevel)
        systemModelGlb = Status.int.systemModel
        systemModelGlb.build()
            

        processes = systemModelGlb.process
        self.NK = len(processes)

        for k, proc in enumerate(processes):
            (inStreams, outStreams) = (systemModelGlb.processStreamsIn[k], systemModelGlb.processStreamsOut[k])

            self.systemModel.addProcess(proc, inStreams, outStreams)

            #TODO at present takes the demand from processes.createAggregateDemand. could be internalized into model building process later on.

            #TODO: -> this information should already be stored in createAGGREGATEDemand
            UPHcQ = []
            for s in range(len(inStreams)):
                UPHcQ.append(Status.int.UPHc_Tt[k].getPart(s).getQ())

            UPHwQ = []
            for s in range(len(outStreams)):
                UPHwQ.append(Status.int.UPHw_Tt_hf[k].getPart(s).getQ())

            UPHwReqQ = []
            for s in range(len(outStreams)):
                UPHwReqQ.append(Status.int.UPHwReq_Tt_hf[k].getPart(s).getQ())

            self.systemModel.process[k].setDemand(Status.int.UPHm_Tt[k].getQ(), Status.int.UPHs_Tt[k].getQ(), UPHcQ, UPHwQ, UPHwReqQ)

        needsSun = False
##        equipment = Status.prj.getEquipments(cascade=True)
        equipment = systemModelGlb.eqObject
        
        self.NJ = len(equipment)
        for j, eq in enumerate(equipment):
            eqSQL = systemModelGlb.equipment[j]
            self.systemModel.addEquipment(eqSQL, eq.eqClass)
            pipeIndices = getConnectedPipes(eq.no, Status)
            self.systemModel.equipment[j].setPipeCon(pipeIndices)

            if eq.eqClass == "ST": needsSun = True

        if needsSun: self.systemModel.addSun(Status.WeatherData)



        pipes = systemModelGlb.pipe        
        self.NM = len(pipes)

        for m, pipe in enumerate(pipes):
            # security feature TODO -> these checks should be filtered at ccheck-stage
            if pipe.HeatDistMedium is None: pipe.HeatDistMedium = str(0)
            if pipe.Tfeedup is None: pipe.Tfeedup = pipe.TreturnDistrib
            if pipe.TRetIn is None: pipe.TRetIn = pipe.TreturnDistrib

            heating = systemModelGlb.pipeObject[m].heating
            self.systemModel.addPipe(pipe, heating)
            pipeModel = self.systemModel.pipe[m]

#TODO -> to be included in model !!!
            processIndices = getAllProcessesConnectedTo(m)
            processIndicesMCD = getAllMandCoolDownProcessesConnectedTo(m)
            equipmentIndices = getAllEquipmentsConnectedTo(m) #equipment connected as SINK (!) -> FETHeat_j

            pipeModel.setCon(processIndices, processIndicesMCD, equipmentIndices)

        # defining priority pipes for each equipment
        NM = self.NM
        NJ = self.NJ
        mcimax = [0] * NM
        for m in range(NM):
            for j in range(NJ):
                if Status.int.USHLink[m][j]==1:
                    mcimax[m] = max(equipment[j].no, mcimax[m])

        for j in range(NJ):
            pipesWithPriority = []
            pipesWithoutPriority = []

            ci = equipment[j].no
            for m in range(NM):
                if mcimax[m] == ci:
                    pipesWithPriority.append(m)
                elif Status.int.USHLink[m][j]==1:
                    pipesWithoutPriority.append(m)

            self.systemModel.equipment[j].setPipePriorities(pipesWithPriority, pipesWithoutPriority)

        whees = Status.int.systemModel.whee
        for n, whee in enumerate(whees):
            T1 = whee.WHEETOutlet
            if T1 is None: T1 = 60.0 # this should not occur and already be blocked by ccheck, but better double security

            # TODO: all this calculation stuff should be stored with the WHEE data and calculations done only once at CCheck stage
            T0 = T1 - 10
            mdot = whee.WHEEFlow
            Pnom = whee.QWHEE

            if mdot is not None and Pnom is not None:
                if whee.WHEEMedium is not None:
                    fluid = Fluid(int(whee.WHEEMedium))
                else:
                    fluid = Fluid(0)

                if (mdot*fluid.cp) > 0:
                    T0 = max(T1 - Pnom/(mdot*fluid.cp), 18.0)
            
            T1 = 60.0
            #TODO: calculate real temperatures of WHEE
            self.systemModel.addWHEE(whee, T0, T1)
      
            self.systemModel.whee[n].setWasteHeat(Status.int.QWHEE_Tt_hf[n].getQ())

        if not estimateMode:
##            buildHXNet(split=True)
            
            hxes = Status.int.hx
            self.NH = len(hxes)
            for h, hx in enumerate(hxes):
                hxTable = {"QHeatExchanger_ID": hx.id,
                           "HXName": hx.name,
                           "HXNo": hx.HXNo,
                           "UA": hx.UA,
                           "MCStorage": hx.MCStorage,
                           "FluidIDSource": hx.fluidIdSource,
                           "FluidIDSink": hx.fluidIdSink,
                           "Kc0": hx.Kc0,
                           "Kc1": hx.Kc1,
                           "Kh0": hx.Kh0,
                           "Kh1": hx.Kh1,
                           "Ts_max": hx.Ts_max}
                           
                self.systemModel.addHX(hxTable)

            for n, splitter in enumerate(Status.int.sourceSplitters):
                self.systemModel.addSourceSplitter(splitter.fractions, splitter.fluidId, splitter.residual)

            for splitter in Status.int.sinkSplitters:
                self.systemModel.addSinkSplitter(splitter.fractions, splitter.fluidId, splitter.residual)

            for nits in xrange(5):  # TODO: this is a hack in order to avoid problems with nested splitters.
                                    # -> better and more robust solution = an algorithm determining the appropriate order of splitters

                for h, hx in enumerate(hxes):

                    if len(hx.sourcestreams) == 0:
                        logError("Heat exchanger %s is not connected to any source"%hx.name)
                        return False
                    if len(hx.sinkstreams) == 0:
                        logError("Heat exchanger %s is not connected to any sink"%hx.name)
                        return False
                    
                    src = hx.sourcestreams[0]
                    snk = hx.sinkstreams[0]

                    srcComponentNo = -1 if src.componentNo is None else src.componentNo
                    snkComponentNo = -1 if snk.componentNo is None else snk.componentNo

    ##                #TODO: this step is necessary, as the value stored in "componentNo" is not reliable !
    ##                if src.objectType in ["ProcessC", "ProcessH"]:
    ##                    srcComponentNo = src.getIndexFromStreamName()

    ##
    ##                if snk.objectType in ["ProcessC", "ProcessH"]:
    ##                    snkComponentNo = snk.getIndexFromStreamName()
                    
    ##                print "ModuleEnergy.buildModel: connecting hx %s"%hx.name
    ##                print "source type %s no %s compNo %s fraction %s"%(src.objectType, src.objectNo, src.componentNo, src.fraction)
    ##                print "sink type %s no %s compNo %s fraction %s"%(snk.objectType, snk.objectNo, snk.componentNo, snk.fraction)

                    if any([par is None for par in [src.objectType, snk.objectType, src.objectNo, snk.objectNo]]):
                        logError("Error in connections of HX %s"%hx.name)
                        return False
                           
                    self.systemModel.connectHX(h,
                                               src.objectType, src.objectNo, srcComponentNo,
                                               snk.objectType, snk.objectNo, snkComponentNo)
                    

                    srcRD = hx.srcRD

        ##            print "ModuleEnergy.buildModel: connecting srcRD of hx %s"%hx.name
                    if srcRD is not None:
                        srcRDComponentNo = -1 if srcRD.componentNo is None else srcRD.componentNo

    ##                    #TODO: this step is necessary, as the value stored in "componentNo" is not reliable !
    ##                    if srcRD.objectType in ["ProcessC", "ProcessH"]:
    ##                        srcRDComponentNo = srcRD.getIndexFromStreamName()

        ##                print "source type %s no %s compNo %s fraction %s"%(srcRD.objectType, srcRD.objectNo, srcRD.componentNo, srcRD.fraction)
                        
                        self.systemModel.connectHXsrcRD(h, srcRD.objectType, srcRD.objectNo, srcRDComponentNo)

                    snkRD = hx.snkRD

    ##                print "ModuleEnergy.buildModel: connecting snkRD of hx %s"%hx.name
    ##                print "SnkRD = %r"%snkRD, snkRD
                    
                    if snkRD is not None:

                        snkRDComponentNo = -1 if snkRD.componentNo is None else snkRD.componentNo

    ##                    if snkRD.objectType in ["ProcessC", "ProcessH"]:
    ##                        snkRDComponentNo = snkRD.getIndexFromStreamName()
                        
        ##                print "sink type %s no %s compNo %s fraction %s"%(snkRD.objectType, snkRD.objectNo, snkRD.componentNo, snkRD.fraction)
                        
                        self.systemModel.connectHXsnkRD(h, snkRD.objectType, snkRD.objectNo, snkRDComponentNo)


##            for nits in xrange(5):  # TODO: this is a hack in order to avoid problems with nested splitters.
##                                    # -> better and more robust solution = an algorithm determining the appropriate order of splitters
                for n, splitter in enumerate(Status.int.sourceSplitters):
                    src = splitter.mainStream.compactConnToUpstreamComponent()
                    srcComponentNo = -1 if src.componentNo is None else src.componentNo

                    srcRD = splitter.mainStreamRD
                    srcRDComponentNo = -1 if srcRD.componentNo is None else srcRD.componentNo

    ##                if srcRD.objectType in ["ProcessC", "ProcessH"]:
    ##                    srcRDComponentNo = srcRD.getIndexFromStreamName()

                    self.systemModel.connectSourceSplitter(n,
                                                           src.objectType, src.objectNo, srcComponentNo, src.fraction,
                                                           srcRD.objectType, srcRD.objectNo, srcRDComponentNo)

                for n, splitter in enumerate(Status.int.sinkSplitters):
                    snk= splitter.mainStream.compactConnToUpstreamComponent()
                    snkComponentNo = -1 if snk.componentNo is None else snk.componentNo
    ##                if snk.objectType in ["ProcessC", "ProcessH"]:
    ##                    snkComponentNo = snk.getIndexFromStreamName()


                    snkRD = splitter.mainStreamRD
                    snkRDComponentNo = -1 if snkRD.componentNo is None else snkRD.componentNo

    ##                if snkRD.objectType in ["ProcessC", "ProcessH"]:
    ##                    snkRDComponentNo = snkRD.getIndexFromStreamName()

        ##            print "ModuleEnergy.buildModel: connecting sinkSplitter"
        ##            print "sink type %s no %s compNo %s fraction %s"%(snk.objectType, snk.objectNo, snk.componentNo, snk.fraction)


                    self.systemModel.connectSinkSplitter(n,
                                                         snk.objectType, snk.objectNo, snkComponentNo, snk.fraction,
                                                         snkRD.objectType, snkRD.objectNo, snkRDComponentNo)


        else: #here specific actions for estimate HR mode
#..............................................................................
# settings of the conversion

            self.NH = 0
            
            if Status.ANo > 0:
                fHR = HRRATIO_DEFAULT   # fraction of potential heat recovery that is really recovered
            else:
                fHR = 0.0
                
            if Status.NHX > 0: logWarning("ESTIMATIVE MODE FOR HEAT RECOVERY CALCULATIONS. REAL HX network is not taken into consideration !!!!")
            
            delta_iT = int (HRDELTAT_DEFAULT / Status.TemperatureInterval + 0.5)
            assert(delta_iT <= Status.NT + 1)

            self.systemModel.setEstimateModePars(fHR = fHR, DTmin = HRDELTAT_DEFAULT)

     
        #finally connect all residual waste heat streams into QWHAmb and build up the stream list
        self.systemModel.linkQWHAmb()
        self.systemModel.buildStreamList()

        return True

    def setBoundaryConditions(self):
        """ Here all the boundary conditions of the model are set:
            - demand + schedules for processes
            - control schedules for equipment
            - temperatures and solar radiation level
        """

        self.systemModel.setWeatherData(Status.WeatherData.Tdb, Status.WeatherData.Twb)

        systemModelGlb = Status.int.systemModel
        processes = systemModelGlb.process
        self.NK = len(processes)
        for k, proc in enumerate(processes):
            (inStreams, outStreams) = (systemModelGlb.processStreamsIn[k], systemModelGlb.processStreamsOut[k])

            #TODO at present takes the demand from processes.createAggregateDemand. could be internalized into model building process later on.

            #TODO: -> this information should already be stored in createAGGREGATEDemand
            UPHcQ = []
            for s in range(len(inStreams)):
                UPHcQ.append(Status.int.UPHc_Tt[k].getPart(s).getQ())

            UPHwQ = []
            for s in range(len(outStreams)):
                UPHwQ.append(Status.int.UPHw_Tt_hf[k].getPart(s).getQ())

            UPHwReqQ = []
            for s in range(len(outStreams)):
                UPHwReqQ.append(Status.int.UPHwReq_Tt_hf[k].getPart(s).getQ())

            self.systemModel.process[k].setDemand(Status.int.UPHm_Tt[k].getQ(), Status.int.UPHs_Tt[k].getQ(), UPHcQ, UPHwQ, UPHwReqQ)

        eqObjects = systemModelGlb.eqObject
        needsSun = False
        for j, eqObject in enumerate(eqObjects):
            schedule = eqObject.controlSchedule.schedule if eqObject.controlSchedule else None
            self.systemModel.equipment[j].setBoundaryConditions(schedule)

            if eqObject.eqClass == "CHP" and eqObject.eqType in ["CHP ORC", "CHP TFC"]:   # special parameters for heat driven equipment
                self.systemModel.equipment[j].setSpecificBC(eqObject.TDrive, eqObject.TReject)
            elif eqObject.eqClass == "CH":
                self.systemModel.equipment[j].setSpecificBC(None, None)

            if eqObject.eqClass == "ST": needsSun = True
        
    def setInitialValues(self, eqStorage, hxStorage):
        """ sets initial values for state variables. used in rt-simulation only """

        if eqStorage is not None:
            for j, eq in enumerate(self.systemModel.equipment):
                eq.setInitialValues(eqStorage[j])

        if hxStorage is not None:
            for h, hx in enumerate(self.systemModel.hx):
                hx.setInitialValues(hxStorage[h])

                
#------------------------------------------------------------------------------
    def runSimulation(self,first=None,last=None,loop=True, checkPeakPower=False, outputLevel=1):
        """
        :params outputLevel: 0: basic simulation, only global results are stored
                             1: level required for basic display options (EA panels, etc.)
                                 - temperature distribution of yearly totals (_T)
                                 - time distribution of USH for 5 temperature levels
                             2: USHTotal_Tt is stored
                             3: USHj_Tt's are stored for all equipments
        """

        self.outputLevel = outputLevel
        
        if Status.processData.outOfDate is True:
            Status.processData.createAggregateDemand()  # this is necessary before building the model

        if outputLevel > 0: # create space for depositing simulation results
            Status.int.USHTotal_Tt = Status.int.createQ_Tt()
            # TO DO: -> maybe create in another place ... ?

#..............................................................................
#..............................................................................
# now create model and intermediate storage space

        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runSimulation: Building Model")

        ok = self.buildModel(outputLevel)
    
        if not ok:
            logError("Error when building model for simulation")
            return

        self.setBoundaryConditions()
        
#..............................................................................
#..............................................................................
       
##        self.mod.getEquipmentList()
        
        # initializing storage space for energy flows in cascade
        first = max((first, 1, Status.int.cascadeUpdateLevel + 1))    #avoid unnecessary calculations

        if last == None or last > self.NJ:
            last = self.NJ

        Status.int.extendCascadeArrays(self.NJ)
##        Status.int.USHjRequiredChanged = [True] * self.NJ

#..............................................................................
#..............................................................................
            
##        print "-----------------------------------"
##        print "Starting Simulation"
##        print "-----------------------------------"
        dlg = DialogGauge(Status.main,_("EINSTEIN system simulation"),_("calculating energy flows"))

        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runSimulation: Starting simulation")

        self.systemModel.simulate(Status.TimeStep, Status.Nt, t0=0.0, dlg=dlg)
    
        dlg.Destroy()
##        print "-----------------------------------"
    
#..............................................................................
# update the pointer to the last calculated cascade

                          
        Status.int.cascadeUpdateLevel = self.NJ
        Status.int.heatRecoveryOutOfDate = False

        # retreive global simulation results from the model

        # TODO: check why this is necessary
        Status.int.USHCPipe = [0.0]*self.NM
        Status.int.UPHCProcm = [0.0]*self.NM
        Status.int.QHXPipe = [0.0]*self.NM

        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runSimulation: Storing results")
        uncoveredDemand, uncoveredDemandC = self.systemModel.storeResults()
        
        for j in range(self.NJ):
            Status.int.USHj_t[j] = self.systemModel.USXj_t[j]

        # Break here for "pseudo-simulations" in ANo = 0. Simulation results should not be stored in SQL
        if Status.ANo <= 0:
            return

##        for it in xrange(8760):
##            print "USHj_t[%s]: %s | %s"%(it, Status.int.USHj_t[0][it], Status.int.USHj_t[1][it])

        for h in range(self.NH):
            Status.int.hx[h].QHX = self.systemModel.hx[h].QHX                                                                                                     
            Status.int.hx[h].saveSimulationResultsToDB()

        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runSimulation: arrived at the end. %s [%s]"%\
                     (last,self.NJ))

            logTrack("SimulationPlus.runSimulation: updating Energy balances")
        self.mod.calculateEquipmentEnergyBalances(calledFromRunSimulation=True)

        # Check if supply = demand balance is hold. Otherwise results are not valid.       
        totalDemand = Status.int.USHTotal
        totalDemandC = Status.int.USCTotal

        peakPowerOK = True
        if checkPeakPower:
            if uncoveredDemand > 0.005 * totalDemand:
                showWarning("Revise your design.\nCurrent equipment capacity is not sufficient for covering the %s demand\n"\
                            "Remaining %s demand: %10.2f [MWh]" % ("heat","heat", uncoveredDemand / 1000.0))
                peakPowerOK = False

            if uncoveredDemandC > 0.005 * totalDemandC:
                showWarning("Revise your design.\nCurrent equipment capacity is not sufficient for covering the %s demand\n"\
                            "Remaining %s demand: %10.2f [MWh]" % ("cooling","cooling",uncoveredDemandC / 1000.0))
                peakPowerOK = False

        if peakPowerOK: Status.prj.setStatus("Energy")

        if DEBUG != "OFF" or (hasattr(Status, "runningTest") and Status.runningTest is not None):
            for it in xrange(Status.Nt):
                self.writeResultToTable(it)

            self.saveTable()            
                    
    def startRTSimulation(self, loop=True, checkPeakPower=False, outputLevel=1, eqStorage=None, hxStorage=None):
        """
        :params outputLevel: 0: basic simulation, only global results are stored
                             1: level required for basic display options (EA panels, etc.)
                                 - temperature distribution of yearly totals (_T)
                                 - time distribution of USH for 5 temperature levels
                             2: USHTotal_Tt is stored
                             3: USHj_Tt's are stored for all equipments

        :params eqStorage, hxStorage:  initial values for equipment and hx storage (array[j][iT])
        """

        self.outputLevel = outputLevel
        
        if Status.processData.outOfDate is True:
            Status.processData.createAggregateDemand()  # this is necessary before building the model

        if outputLevel > 0: # create space for depositing simulation results
            Status.int.USHTotal_Tt = Status.int.createQ_Tt()
            # TO DO: -> maybe create in another place ... ?

#..............................................................................
#..............................................................................
# now create model and intermediate storage space

        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runRealTime: Building Model")

        ok = self.buildModel(outputLevel)
        ### boundary conditions are set in RealTime !!!
    
        if not ok:
            logError("Error when building model for simulation")
            return

        self.setInitialValues(eqStorage, hxStorage)

#..............................................................................
#..............................................................................
       
        Status.int.extendCascadeArrays(self.NJ)

#..............................................................................
#..............................................................................
            
        if DEBUG in ["ALL"]:
            logTrack("SimulationPlus.runSimulation: Starting simulation")

    def stepRTSimulation(self, dt, timeStamp=None):

        # convert time stamp in hour of the year of the reference time interval
        t1 = dateTimeToHOY(timeStamp)
        t0 = t1 - dt
        if t0 < 0:  t0 += 8760
        self.systemModel.rtStep(dt, t0)
##        self.systemModel.simulate(dt, 1, t0) # this is a hack, as rtStep does not reset self.it in the model !

        # produce an output analogous to the DEBUG output of runSimulation
        it = self.systemModel.it
        self.writeResultToTable(0, rt=timeStamp)

    def closeRTSimulation(self):
        self.saveTable()


#---------------------------------------------------------------------------------
# functions for saving global simulation results

    def writeResultToTable(self, it, rt=None):
        systemModelGlb = Status.int.systemModel

        if self.resultTable is None:
            tableHeader = (["time"] +
                             [("USHj[%s]"%(j+1)) for j in xrange(self.NJ)] +
                             ["FETel"] +
                             [("ElGen[%s]"%(j+1)) for j in xrange(self.NJ) if systemModelGlb.eqObject[j].eqClass=="CHP"] +
                             [("QHX[%s]"%(h+1)) for h in xrange(self.NH)] +
                             [("QWXUsed[%s]"%(h+1)) for h in xrange(self.NH)])

            if hasattr(self.mod, "optPars"):
                tableHeader += self.mod.optPars
                
            self.resultTable = [tableHeader]

        if isinstance(rt, str):
            timeStr = rt
        else:
            if rt is not None:
                try:
                    timeStr = dateTimeToString(rt)
                except:
                    timeStr = str(rt)
            else:
                timeStr = str(it)
            
        result = ([timeStr] +
                  [str(self.systemModel.USXj_t[j][it]).replace(".",",") for j in xrange(self.NJ)] +
                  [str(self.systemModel.FETel_t[it]).replace(".",",")] +
                  [str(self.systemModel.ElGen_t[j][it]).replace(".",",") for j in xrange(self.NJ) if systemModelGlb.eqObject[j].eqClass=="CHP"] +
                  [str(self.systemModel.QHX_t[h][it]).replace(".",",") for h in xrange(self.NH)] +
                  [str(self.systemModel.QWXUsed_t[h][it]).replace(".",",") for h in xrange(self.NH)])

        if hasattr(self.mod, "optVals"):
            result += [str(val).replace(".",",") for val in self.mod.optVals]
        
        self.resultTable.append(result)

    def saveTable(self):
        from lib.fileIO.fileIO import writeTableToFile

        if self.resultTable is not None:
            writeTableToFile(self.resultTable, "simulationResults.csv")
        
        
        
                    
#==============================================================================



if __name__ == "__main__":
    pass
