# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEA3- Final energy by equipment- Yearly data
#			
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
from auxiliary.auxiliary import transpose
from GUI.status import *
from constants import getEquipmentClass
from lib.text.textConversion import toGettextAndUnicode as _U, toUnicode as U
import database.dbio as dbio
from modules.fluids import Fuel

class ModuleEA3(object):
    def __init__(self, keys, showHeating):
        self.keys = keys # two grids, so a list of (2) keys
        self._showHeating = showHeating
        self.initModule()

    def initModule(self):
        """module initialization"""
        Status.mod.moduleEA.update()
        
        allEquipments = Status.prj.getEquipments()
        eqObjects = Status.int.systemModel.eqObject

        filteredEquipments = [eq for j, eq in enumerate(allEquipments) if eqObjects[j].heating==self._showHeating]
        
        self.dbfuel = dbio.select("dbfuel")
        
        # Final energy consumption by equipments 
        allEquipName = []
        filteredEquipName = []
        FETj = []
        USHj = []
      
        for equipe in allEquipments:
            name = U(equipe.Equipment)
            allEquipName.append(name)
            if equipe.FETj is None:
                FETj.append(0.0)
            else:
                FETj.append(equipe.FETj / 1000.0)
        totalFETj = sum(FETj)
        allFuelType = [U(entry) for entry in self._getFuelTypes(allEquipments)]

        for equipe in filteredEquipments:
            name = U(equipe.Equipment)
            filteredEquipName.append(name)
            if equipe.USHj is None:
                USHj.append(0.0)
            else:
                USHj.append(equipe.USHj / 1000.0)
        totalUSHj = sum(USHj)
        
        if totalFETj > 0.0: # avoid division by zero
            FETjPercentage = [item * 100.0 / totalFETj for item in FETj]
        else:
            FETjPercentage = [0.0] * len(FETj)
        if totalUSHj > 0.0:
            USHjPercentage = [item * 100.0 / totalUSHj for item in USHj]
        else:
            USHjPercentage = [0.0] * len(USHj)

        #finish the table columns, add total, percentage total
        allEquipName.append('Total')
        allFuelType.append('')
        filteredEquipName.append('Total')
        FETj.append(totalFETj)
        USHj.append(totalUSHj)
        FETjPercentage.append(sum(FETjPercentage))
        USHjPercentage.append(sum(USHjPercentage))

        # upper grid FET by equipment
        TableColumnList1 = [allEquipName, allFuelType, FETj, FETjPercentage]
        #screen list and substitute None with 0
        for i in range(len(TableColumnList1)):
            for j in range(len(TableColumnList1[i])):
                if TableColumnList1[i][j] == None:
                    TableColumnList1[i][j] = 0           
        
        matrix1 = transpose(TableColumnList1)
        data1 = np.array(matrix1)     
                          
        Status.int.setGraphicsData(self.keys[0], data1)
        #
        # lower grid USH by equipment
        #
        TableColumnList2 = [filteredEquipName, USHj, USHjPercentage]

        #screen list and substitute None with 0
        for i in range(len(TableColumnList2)):
            for j in range(len(TableColumnList2[i])):
                if TableColumnList2[i][j] == None:
                    TableColumnList2[i][j] = 0
                    
        matrix2 = transpose(TableColumnList2)
        data2 = np.array(matrix2)

        Status.int.setGraphicsData(self.keys[1], data2)

        #----------------------------------------------------------------------
        reportMatrix1 = []
        for i in range(len(matrix1) - 1):
            if i < Status.Report_NJMax:
                reportMatrix1.append(matrix1[i])
        for i in range(len(matrix1) - 1, Status.Report_NJMax):
            reportMatrix1.append([" ", " ", " ", " "])
        reportMatrix1.append(matrix1[len(matrix1) - 1])
        reportData1 = np.array(reportMatrix1)

        if Status.ANo == 0:
            if self._showHeating:
                key = "EA3_FET_REPORT"
            else: 
                key = "EA3_FET_REPORT_C"
            Status.int.setGraphicsData(key, reportData1)
        elif Status.ANo == Status.FinalAlternative:
            if self._showHeating:
                key = "EA3_FET_REPORT_F"
            else: 
                key = "EA3_FET_REPORT_F_C"
            Status.int.setGraphicsData(key, reportData1)

        reportMatrix2 = []
        for i in range(len(matrix2) - 1):
            if i < 10: #Part USH, where heating and cooling equipment is listed separately, currently is still fixed to max. 10, NOT to Report_NJMax
                reportMatrix2.append(matrix2[i])
        for i in range(len(matrix2) - 1, 10):
            reportMatrix2.append([" ", " ", " "])
        reportMatrix2.append(matrix2[len(matrix2) - 1])
        reportData2 = np.array(reportMatrix2)

        if Status.ANo == 0:
            if self._showHeating:
                key = "EA3_USH_REPORT"
            else:
                key = "EA3_USH_REPORT_C"
            Status.int.setGraphicsData(key, reportData2)
        elif Status.ANo == Status.FinalAlternative:
            if self._showHeating:
                key = "EA3_USH_REPORT_F"
            else:
                key = "EA3_USH_REPORT_F_C"
            Status.int.setGraphicsData(key, reportData2)
    
    def _getFuelTypes(self, listOfEquipments):
        """Return a list of strings of the fuels used by the Equipments
        in the listOfEquipments.
        
        """
        result = []
        for equipe in listOfEquipments:
            DBFuel_id = equipe.DBFuel_id
            if DBFuel_id is None or len(self.dbfuel) == 0: 
                FuelName = _('Electricity')
            else:
                fuel = Fuel(DBFuel_id)
                FuelName = 'not available' if fuel is None else fuel.name              
                    
            equipmentClass = getEquipmentClass(equipe.EquipType)
            if equipmentClass == "CHP":
                FuelName += "(- gen.elect.)"   
            result.append(FuelName)
        return result 
    
    def updatePanel(self, showHeating):
        """Update the data to be shown by the panel.
        
        :showHeating: Set whether only heating (or only cooling) entities should be shown.
        
        """
        self._showHeating = showHeating
        self.initModule()
