# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEA4- Process heat- Yearly data
#			
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from math import floor, ceil
import numpy as np
from auxiliary.auxiliary import noneFilter, lastNonZero, transpose
from auxiliary.auxiliary import toGettextAndUnicode as _U
from constants import *
from GUI.status import *
from GUI.numCtrl import convertDoubleToString
from modules import equipments
from modules.messageLogger import logWarning

class ModuleEA4(object):
    def __init__(self, keys):
        """initialisation of the module (at Tool start-up)

        :param keys: a list of keys to the GData dict in Status
        
        """
        self.keys = keys
        self._qprocessdata = []
    
    def initPanel(self):
        """initial calculations at entry to the panel"""
        pass
    
    def updatePanel(self, heating=True, requireConverged = False):
        """Calculate the data displayed by panel EA4.
        
        Only take processes into account that heat OR cool depending on the
        heating flag
        Results are put into the GData Dict in einstein.modules.Interfaces.
        
        :param heating_cooling: If true the data of the heating 
        processes is prepared, if false the cooling processes'
        data is prepared.
        :type heating_cooling: Boolean
         
        :param requireConverged: True -> activates system simulation if converged final values of UPXProc / USX are required
                                 used for automatic activation -> calls from auto-pilot or report generator
        """
        # First get yearly totals and write to Panel EA4a
        Status.int.systemModel.build()

        # here data are prepared for the GUI and the report

        if Status.StatusEnergy == 0 or Status.int.cascadeUpdateLevel < len(Status.int.USHj):
            #TODO: here it could be made an exception, if simulation mode = ACTUAL HX and No. of HXes from / to Equipments = 0
            simulationConverged = False
            if requireConverged:
                Status.mod.moduleEA.update()
                simulationConverged = True
        else:
            simulationConverged = True

        if simulationConverged == False and self.keys[0]!="EA4a_Table":
            Status.mod.moduleEnergy.calculate(requireConverged=False)

        # Data for Panel EA4a: annual values
        if self.keys[0] == "EA4a_Table":
            return self.update4a(heating)

        # Data for Panel EA4b: temperature dependent plots
        elif self.keys[0] == "EA4b_Table":
            
            if Status.SimTool == "Plus":
                Status.processData.createYearlyDemand()     # Is not automatically generated by createAggDem in Plus mode
                
            return self.update4b()

        # Data for Panel EA4c: cumulative heat demand curve
        elif self.keys[0] == "EA4c_Table":
            return self.update4c()

    def update4a(self, heating):
        allprocesses = Status.int.systemModel.process
        if heating:
            processes = [p for p in allprocesses if p.heating_cooling == "heating"]
            processesMCD = []
        else:
            processes = [p for p in allprocesses if p.heating_cooling != "heating"]
            processesMCD = [p for p in allprocesses if (p.heating_cooling == "heating") and (p.UPHwReq > 0)]
            
        Processnames = []
        PT = []
        PST = []
        UPH = []
        UPHc = []
        UPHm = []
        UPHs = []
        UPHw = []

        PTi = []
        PTir = []

        TotalUPH = 0.0
        for process in processes:
            Processnames.append(unicode(process.Process, "utf-8"))

            PT.append(process.PT)
            PST.append(process.TSupply)
            
            if process.UPH is not None: UPH.append(process.UPH / 1000.)
            else: UPH.append(0.0)
            if process.UPHc is not None: UPHc.append(process.UPHc / 1000.)
            else: UPHc.append(0.0)
            if process.UPHm is not None: UPHm.append(process.UPHm / 1000.)
            else: UPHm.append(0.0)
            if process.UPHs is not None: UPHs.append(process.UPHs / 1000.)
            else: UPHs.append(0.0)
            if process.UPHw is not None: UPHw.append(process.UPHw / 1000.)
            else: UPHw.append(0.0)
            
            if process.UPH is None:
                TotalUPH += 0.0
            else:
                TotalUPH += process.UPH / 1000.0  #conversion to MWh !!!!

            inStreams,outStreams = Status.prj.getStreams(process.QProcessData_ID)
            pTi = process.PT
            if pTi is None: pTi = 0 #just for avoiding crash
            
            pTir = pTi
            
            for stream in inStreams:
                sTi = stream.PTInFlow
                if sTi is not None:
                    if heating:
                        pTi = min(pTi,sTi)
                    else:
                        pTi = max(pTi,sTi)
                    
                sTir = stream.PTInFlowRec
                if sTir is not None:
                    if heating:
                        pTir = min(pTir,sTir)
                    else:
                        pTir = max(pTir,sTir)
                        
                elif sTi is not None:       # if Tir = None but Ti is defined, suppose Tir = Ti
                    if heating:
                        pTir = min(pTir,sTi)
                    else:
                        pTir = max(pTir,sTi)
                        
            PTi.append(pTi)
            PTir.append(pTir)

        for process in processesMCD:

            Processnames.append(unicode(process.Process, "utf-8"))

            ###TO DO: -> minimum temperature PTFinalRequired of different mandatory cool-downs should be calclated and displayed
            PT.append("n/a")
            PST.append("n/a")
            PTi.append("n/a")
            PTir.append("n/a")
            
            if process.UPHwReq is not None:
                UPH.append(process.UPHwReq / 1000.)
                UPHc.append(process.UPHwReq / 1000.)
            else:
                UPH.append(0.0)
                UPHc.append(0.0)

            # mandatory cool-down considered as a process with circulation only                            
            UPHm.append(0.0)
            UPHs.append(0.0)
            UPHw.append(0.0)
            
            if process.UPHwReq is None:
                TotalUPH += 0.0
            else:
                TotalUPH += process.UPHwReq / 1000.0  #conversion to MWh !!!!

        PT = noneFilter(PT)
        PST = noneFilter(PST)
        UPH = noneFilter(UPH)
        UPHc = noneFilter(UPHc)
        UPHs = noneFilter(UPHs)
        UPHm = noneFilter(UPHm)
        UPHw = noneFilter(UPHw)
        
        UPHPercentage = []
        if TotalUPH > 0.0: 
            for k in xrange(len(UPH)):
                UPHPercentage.append(UPH[k] * 100.0 / TotalUPH)
        else:
            for k in xrange(len(UPH)):
                UPHPercentage.append(0.0)

        # finish the table columns, add total, percentage total
        Processnames.append(_U('Total'))
        PT.append(' ')
        PTi.append(' ')
        PTir.append(' ')
        
        PST.append(' ')
        UPH.append(TotalUPH)
        UPHc.append(' ')
        UPHs.append(' ')
        UPHm.append(' ')
        UPHw.append(' ')
        UPHPercentage.append(100.0)

        # table in EA4a: UPH by process
        TableColumnList1 = [Processnames, UPH, UPHPercentage, UPHc, UPHm, UPHs, PT, PST]

        matrix1 = transpose(TableColumnList1)
        data1 = matrix1

        Status.int.setGraphicsData("EA4a_Table", data1)

        # table in EA4a: for report
        reportMatrix1 = []
        reportMatrixAnnex = []
        
        for i in xrange(len(matrix1) - 1):
            if i < 20:
                reportMatrix1.append(matrix1[i])
                reportMatrixAnnex.append([PTi[i],PTir[i]])
        for i in xrange(len(matrix1) - 1, 20):
            reportMatrix1.append([" ", " ", " ", " ", " ", " ", " ", " "])
            reportMatrixAnnex.append([" "," "])
        reportMatrix1.append(matrix1[len(matrix1) - 1])
        reportMatrixAnnex.append([PTi[len(matrix1) - 1],PTi[len(matrix1) - 1]])

        reportData1 = reportMatrix1
        reportDataAnnex = reportMatrixAnnex

        if Status.ANo == 0:
            if heating:
                key = "EA4a_REPORT"
            else:
                key = "EA4a_REPORT_C"
            Status.int.setGraphicsData(key, reportData1)
            Status.int.setGraphicsData(key+"ANNEX", reportDataAnnex)
        elif Status.ANo == Status.FinalAlternative:
            if heating:
                key = "EA4a_REPORT_F"
            else:
                key = "EA4a_REPORT_F_C"
            Status.int.setGraphicsData(key, reportData1)
            Status.int.setGraphicsData(key+"ANNEX", reportDataAnnex)

        return None

    def update4b(self):

        panelData = {}
        
        for heating in [True, False]:
            if heating:
                UPHTotal_T = Status.int.UPHTotal_T
                UPHProcTotal_T = Status.int.UPHProcTotal_T
                USHTotal_T = Status.int.USHTotal_T
                maximumPoint = Status.NT + 1
                TLevels = [60.0, 80.0, 100.0, 120.0, 140.0, 180.0, 220.0, 300.0, 400.0, 10000.0]
                Titles = ["    <  60 °C",
                          " 60 -  80 °C",
                          " 80 - 100 °C",
                          "100 - 120 °C",
                          "120 - 140 °C",
                          "140 - 180 °C",
                          "180 - 220 °C",
                          "220 - 300 °C",
                          "300 - 400 °C",
                          "    > 400 °C",
                          _("Total")]
                
            else:
                UPHTotal_T = Status.int.UPCTotal_T
                UPHProcTotal_T = Status.int.UPCProcTotal_T
                USHTotal_T = Status.int.USCTotal_T
                maximumPoint = 0
                TLevels = [-20.0,-10, 5.0, 15.0, 25.0, 40.0, 60.0, 80.0, 100.0, 10000]
                Titles = ["    < -20 °C",
                          "-20 - -10 °C",
                          "-10 -   5 °C",
                          "  5 -  15 °C",
                          " 15 -  25 °C",
                          " 25 -  40 °C",
                          " 40 -  60 °C",
                          " 60 -  80 °C",
                          " 80 - 100 °C",
                          "    > 100 °C",
                          _("Total")]

            UPHTotal = UPHTotal_T[maximumPoint]
            UPHProcTotal = UPHProcTotal_T[maximumPoint]
            USHTotal = USHTotal_T[maximumPoint]
           
            # prepare data for tables
            if heating:
                TIndex = [min(int(floor(T / Status.TemperatureInterval + 0.5)), Status.NT+1) for i, T in enumerate(TLevels)]
            else:
                TIndex = [min(int(floor((T - Status.MinimumTemperatureC)/ Status.TemperatureIntervalC + 0.5)), Status.NT+1) for i, T in enumerate(TLevels)]

            UPH = np.array([UPHTotal_T[iT] for iT in TIndex])
            UPH, dUPH, UPHPerc, UPHPercCum = decompose(UPH, UPHTotal, ascending=heating)

            UPHProc = np.array([UPHProcTotal_T[iT] for iT in TIndex])
            UPHProc, dUPHProc, UPHProcPerc, UPHProcPercCum = decompose(UPHProc, UPHProcTotal, ascending=heating)
                
            USH = np.array([USHTotal_T[iT] for iT in TIndex])
            USH, dUSH, USHPerc, USHPercCum = decompose(USH, USHTotal, ascending=heating)

            # EA4b table in panel


            dataPanel = transpose([Titles, list(dUPH), list(UPHPerc), list(UPHPercCum), list(dUSH), list(USHPerc), list(USHPercCum)])
            
            panelDataKey = "DataH" if heating else "DataC"
            panelData.update({panelDataKey: {"data": dataPanel}})

            # EA4b table in report
            for i in xrange(len(dUPH)):
                dUPH[i] = convertDoubleToString(dUPH[i]/1000.0, nDecimals=2)
                UPHPerc[i] = convertDoubleToString(100.0*UPHPerc[i], nDecimals=2)
                UPHPercCum[i] = convertDoubleToString(100.0*UPHPercCum[i], nDecimals=2)
                dUSH[i] = convertDoubleToString(dUSH[i]/1000.0, nDecimals=2)
                USHPerc[i] = convertDoubleToString(100.0*USHPerc[i], nDecimals=2)
                USHPercCum[i] = convertDoubleToString(100.0*USHPercCum[i], nDecimals=2)

            data1 = transpose([Titles, dUPH, UPHPerc, UPHPercCum, dUSH, USHPerc, USHPercCum])

            gDataKey = "EA4b_Table" if heating else "EA4b_Table_C"
            Status.int.setGraphicsData(gDataKey, data1)

            if Status.ANo == 0:
                if heating:
                    key = "EA4b_REPORT"
                else:
                    key = "EA4b_REPORT_C"
                Status.int.setGraphicsData(key, data1)
            elif Status.ANo == Status.FinalAlternative:
                if heating:
                    key = "EA4b_REPORT_F"
                else:
                    key = "EA4b_REPORT_F_C"
                Status.int.setGraphicsData(key, data1)
                                  
            # data for EA4b plot
            # first determine maximum temperature level up to which supply is
            # still varying (supply is the highest T-level !!!)
            iTmax = Status.NT
            iTmin = 0
            if heating:
                for iT in xrange(Status.NT + 1):
                    if USHTotal_T[iT] / 1000.0 >= USHTotal * 0.9999:
                        iTmax = min(iT + 8, Status.NT)
                        break
                iTmax = max(20, iTmax)     # minimum plot up to 100 °C
            else:
                for iT in xrange(Status.NT , -1, -1):
                    if USHTotal_T[iT] / 1000.0 >= USHTotal * 0.9999:
                        iTmin = max(iT - 20, 0)
                        break
                iTmin = min(10, iTmin)     # minimum plot down to -40 °C
            UPH_plot = UPHTotal_T[iTmin:iTmax+1] / 1000.0
            UPHproc_plot = UPHProcTotal_T[iTmin:iTmax+1] / 1000.0
            USH_plot = USHTotal_T[iTmin:iTmax+1] / 1000.0

            
            TScala = Status.int.T[iTmin:iTmax+1] if heating else Status.int.TC[iTmin:iTmax+1]
##            Status.int.setGraphicsData(self.keys[1], [TScala,
##                                                     UPH_plot,
##                                                     UPHproc_plot,
##                                                     USH_plot])

            panelDataKey = "PlotH" if heating else "PlotC"
            xData = [np.array(TScala)]
            yData = [np.array(UPH_plot), np.array(UPHproc_plot), np.array(USH_plot)]
            labels = ["UPH", "UPH_proc", "USH"] if heating else ["UPC", "UPC_proc", "USC"] 
            panelData.update({panelDataKey: {"xData": xData, "yData": yData, "labels": labels}})


            TT = list(TScala)
            UPH_rep = list(UPH_plot)
            UPHproc_rep = list(UPHproc_plot)
            USH_rep = list(USH_plot)
            # Add identical values as dummies to the plot. 
            # A fixed array size needed for the report.
            for iT in xrange(iTmax+1, (Status.NT + 1)):
                TT.append(TScala[iTmax])
                UPH_rep.append(UPHTotal_T[iTmax] / 1000.0)
                UPHproc_rep.append(UPHProcTotal_T[iTmax] / 1000.0)
                USH_rep.append(USHTotal_T[iTmax] / 1000.0)
            for iT in xrange(0, iTmin):
                TT.insert(0, TScala[iTmin])
                UPH_rep.insert(0, UPHTotal_T[iTmin] / 1000.0)
                UPHproc_rep.insert(0, UPHProcTotal_T[iTmin] / 1000.0)
                USH_rep.insert(0, USHTotal_T[iTmin] / 1000.0)

            dataList = [TT, UPH_rep, UPHproc_rep, USH_rep]

            if Status.ANo == 0:
                if heating:
                    key = "EA4b_PLOT_REPORT"
                else:
                    key = "EA4b_PLOT_REPORT_C"
                Status.int.setGraphicsData(key, transpose(dataList))
            elif Status.ANo == Status.FinalAlternative:
                if heating:
                    key = "EA4b_PLOT_REPORT_F"
                else:
                    key = "EA4b_PLOT_REPORT_F_C"
                Status.int.setGraphicsData(key, transpose(dataList))
                    

        return panelData

    def update4c(self):                

        panelData = {}

        for heating in [True, False]:
            if heating:
                UPHTotal_T = Status.int.UPHTotal_T
                UPHProcTotal_T = Status.int.UPHProcTotal_T
                USHTotal_T = Status.int.USHTotal_T
                USHTotal_Tt = Status.int.USHTotal_Tt
                maximumPoint = Status.NT + 1
                TLevels = [10000.0, 80.0, 120.0, 250.0, 400.0]
                Titles = ["  Total  ",
                          " <  80 °C",
                          " < 120 °C",
                          " < 250 °C",
                          " < 400 °C"]
            else:
                UPHTotal_T = Status.int.UPCTotal_T
                UPHProcTotal_T = Status.int.UPCProcTotal_T
                USHTotal_T = Status.int.USCTotal_T
                USHTotal_Tt = Status.int.USCTotal_Tt
                maximumPoint = 0
                TLevels = [-100.0, -10.0, 5.0, 15.0, 30.0]
                Titles = ["  Total  ",
                          " > -10 °C",
                          " > 5 °C",
                          " > 15 °C",
                          " > 30 °C"]

            tLevels = [4000.0, 2000.0, 0]
            it_baseLoad = int(Status.Nt * tLevels[0] / YEAR)
            it_mediumLoad = int(Status.Nt * tLevels[1] / YEAR)
            it_peakLoad = int(Status.Nt * tLevels[2] / YEAR)
            USH = []
            baseLoad = []
            mediumLoad = []
            peakLoad = []
            energy_baseLoad = []
            energy_mediumLoad = []
            energy_peakLoad = []
            
            for i in xrange(len(TLevels)):
                T = TLevels[i]
                if heating:
                    iT = int(floor(T / Status.TemperatureInterval + 0.5))
                    iT = min(iT, Status.NT + 1)
                else:
                    iT = int(round((T / Status.TemperatureIntervalC) + 60 / Status.TemperatureIntervalC)) 
                    iT = max(iT, 0)
                
                if Status.SimTool == "Plus":
                    if heating:
                        USH.append(Status.int.USHTotal_Lt[i].tolist())
                    else:
                        USH.append(Status.int.USCTotal_Lt[i].tolist())
                else:
                    USH.append(USHTotal_Tt[iT].tolist()) # copy, in order not to sort the original list
                USH[i].sort()
                USH[i].reverse()

                baseLoad.append(USH[i][it_baseLoad])
                mediumLoad.append(USH[i][it_mediumLoad])
                peakLoad.append(USH[i][it_peakLoad])

                energy_baseLoad.append(0.0)
                energy_mediumLoad.append(0.0)
                energy_peakLoad.append(0.0)

                for it in xrange(Status.Nt):
                    energy_baseLoad[i] += min(baseLoad[i], USH[i][it])
                    energy_mediumLoad[i] += min(mediumLoad[i], USH[i][it])
                    energy_peakLoad[i] += min(peakLoad[i], USH[i][it])

                energy_peakLoad[i] -= energy_mediumLoad[i]
                energy_mediumLoad[i] -= energy_baseLoad[i]
                
                baseLoad[i] /= Status.TimeStep      #convert to power
                mediumLoad[i] /= Status.TimeStep
                peakLoad[i] /= Status.TimeStep

            
            # now determine maximum number of operating hours
            itMax = lastNonZero(USH[0])
            tMax = itMax * Status.TimeStep
            # TODO: What's this rounding for?
            itMax = int((500.0 / Status.TimeStep) * ceil(tMax / 500.0) + 0.5)  #round
            itMax = max(min(itMax, Status.Nt - 1),1)

            dataListPanel = [Titles,
                             baseLoad, energy_baseLoad,
                             mediumLoad, energy_mediumLoad,
                             peakLoad, energy_peakLoad]
            dataPanel = transpose(dataListPanel)

            panelDataKey = "DataH" if heating else "DataC"
            panelData.update({panelDataKey: {"data": dataPanel}})
            
            # data for EA4c table
            for i in xrange(len(TLevels)):
                energy_baseLoad[i] *= Status.EXTRAPOLATE_TO_YEAR / 1000.0           #convert to MWh
                energy_mediumLoad[i] *= Status.EXTRAPOLATE_TO_YEAR / 1000.0
                energy_peakLoad[i] *= Status.EXTRAPOLATE_TO_YEAR / 1000.0 

            dataList = [Titles, baseLoad, energy_baseLoad, mediumLoad,
                        energy_mediumLoad, peakLoad, energy_peakLoad]
            data = transpose(dataList)
            gDataKey = "EA4c_Table" if heating else "EA4c_Table_C"
            Status.int.setGraphicsData(gDataKey, data)

            if Status.ANo == 0:
                if heating:
                    key = "EA4c_REPORT"
                else: 
                    key = "EA4c_REPORT_C"
                Status.int.setGraphicsData(key, data)
            elif Status.ANo == Status.FinalAlternative:
                if heating:
                    key = "EA4c_REPORT_F"
                else: 
                    key = "EA4c_REPORT_F_C"
                Status.int.setGraphicsData(key, data)


            # data for EA4c plot
            nLevels = len(TLevels)
            time = ["t"]
            for it in xrange(itMax + 1):
                time.append(it * YEAR / Status.Nt)
                
            dataList = [time[0:itMax + 1]]
            for i in xrange(nLevels):
                row = [Titles[i]]
                row.extend(USH[i][0:itMax])
                dataList.append(row)
            data = dataList
            Status.int.setGraphicsData(self.keys[1], data)

            panelDataKey = "PlotH" if heating else "PlotC"
            xData = [np.linspace(1.0, 1.0*itMax, itMax)]
            yData = [np.array(USH[i][:itMax]) for i in xrange(nLevels)]
            labels = [Titles[i] for i in xrange(nLevels)]
            panelData.update({panelDataKey: {"xData": xData, "yData": yData, "labels": labels}})

            # data for EA4c plot in report
            timeReport = ["t"]
            DATAPOINTS = 100
            step = max(1.0, 1.0 * itMax / DATAPOINTS)
            for itr in xrange(DATAPOINTS):
                it = int(step * itr)
                timeReport.append(it * YEAR / Status.Nt)

            dataListReport = [timeReport]
            for i in xrange(len(TLevels)):
                row = [Titles[i]]
                for itr in range(DATAPOINTS):
                    it = int(step * itr)
                    USH[i][itr] = USH[i][it]
                row.extend(USH[i][0:DATAPOINTS])
                dataListReport.append(row)
            dataReport = transpose(dataListReport)

            if Status.ANo == 0:
                if heating:
                    key = "EA4c_PLOT_REPORT"
                else:
                    key = "EA4c_PLOT_REPORT_C"
                Status.int.setGraphicsData(key, dataReport)
            elif Status.ANo == Status.FinalAlternative:
                if heating:
                    key = "EA4c_PLOT_REPORT_F"
                else:
                    key = "EA4c_PLOT_REPORT_F_C"
                Status.int.setGraphicsData(key, dataReport)

        return panelData


def decompose(q, qTotal=None, ascending=True, addTotals=True):
    """
    helper function for decomposing a cumulative function q(T) into
    the increments and fractions (cumulative and incremental)
    """
    # inverto into ascending if cooling
    qqTotal = (qq[-1] if ascending else qq[0]) if qTotal is None else qTotal

    qq = q if ascending else qqTotal - q
    qqFractionCum = q / max(qqTotal, 0.001)
    dqq = qq[1:] - qq[:-1]
    dqq = np.insert(dqq, 0, qq[0])
    qqFraction = dqq / max(qqTotal, 0.001)

    # add last row for totals (if different from the first/last value in the q vector)
    if addTotals:
        qq = np.append(qq, qqTotal)
        dqq = np.append(dqq, qqTotal)
        qqFraction = np.append(qqFraction, 1.0)
        qqFractionCum = np.append(qqFractionCum, 1.0)

    return qq, dqq, qqFraction, qqFractionCum
