# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEM1- Energy performance - Monthly data
#			
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
from auxiliary.auxiliary import transpose
from GUI.status import *
from constants import MONTHS, MONTHSTARTDAY, TRANSMONTHS
import copy
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U

if Status.SimTool == "Plus":
    from modules.energy.process import Process



class ModuleEM1(object):
    def __init__(self, keys, showHeating):
        self.keys = keys
        self._showHeating = showHeating
        self.initModule()

    def sumMonthlyDemand(self, process):
        """Return the monthly sums of demand for heating of the
        _heating_ process nr. k, or the monthly sums of the cooling 
        demand of the process nr. k. 
        
        In the latter case the mandatory cool down demand of heating
        processes or the cooling demand of of cooling processes is
        taken into account. 
        
        :param process: the process' pSQL dataset object
        
        :returns list: UPHMonthlyProc which is the 12 monthly
            and the total demand of the process 
            
        """
        allprocesses = Status.prj.getProcesses()
        mandatoryCooldown = False
        if self._showHeating is True:
            maximumPoint = Status.NT + 1
        else: # sum up cooling demand
            maximumPoint = 0
            if process.heating_cooling == "heating":
                mandatoryCooldown = True
                
        UPH_TtIndex = allprocesses.index(process)

        if Status.SimTool == "Plus":
            if Status.int.cascadeUpdateLevel <= 0:
                Status.mod.moduleEnergy.runSimulation()
                
            UPX_t = np.zeros(Status.Nt)
            UPXwReq_t = np.zeros(Status.Nt)

            k = UPH_TtIndex

            Status.mod.moduleEnergy.sim.systemModel.process[k].getDemand(UPX_t, UPXwReq_t)
            if self._showHeating and process.heating_cooling == "heating":
                demand_t = UPX_t
            elif not self._showHeating:
                if process.heating_cooling == "heating":
                    demand_t = UPXwReq_t
                else:
                    demand_t = UPX_t
            else:
                demand_t = np.zeros(Status.Nt)

        else:
            if mandatoryCooldown is False:
                demand_Tt = Status.int.UPH_Tt[UPH_TtIndex]
            else:
                demand_Tt = Status.int.UPHwReq_Tt_hf[UPH_TtIndex].Q_Tt(mode="cooling")
            demand_t = demand_Tt[maximumPoint]
            
        UPHMonthlyProc = [0.0] * 13 #UPH single process
        month = 1
        for it in range(Status.Nt):
            time = Status.TimeStep * it * Status.EXTRAPOLATE_TO_YEAR
            if time >= 24.0 * MONTHSTARTDAY[month]:
                month += 1
            UPHMonthlyProc[month] += demand_t[it] / 1000 #converted to [MWh]
        
        UPHMonthlyProc[0] = sum(UPHMonthlyProc[1:]) #Yearly total of the process !!!
        return UPHMonthlyProc

    def initModule(self):
        # In this grid the nr. of cols is variable, so we generate the
        # column headings dynamically here
        Status.int.systemModel.build()  # assure that the system model is up to date
        if Status.int.cascadeUpdateLevel < 0:
            Status.int.initCascadeArrays(0)     #creates the demand profiles
        # get all processes and remove unwanted ones
        allprocesses = Status.prj.getProcesses()
        if self._showHeating:
            processes = [p for p in allprocesses if p.heating_cooling == "heating"]
            LabelRow1 = [_U('Process heat\ndemand')]
        else:
            # heating processes may have cool down demand
            processes = [p for p in allprocesses if ((p.heating_cooling != "heating") or
                                                     (p.UPHwReq > 0))]
            LabelRow1 = [_U('Process cooling\ndemand')]

        LabelRowReport = ["UPH/C"]
            
        # filling the label row (row with 1 + NThProc + 1 entries)
        # for the graphic data array
        for process in processes:
            LabelRow1.append(unicode(process.Process, "utf-8") + '\n[MWh]')
            LabelRowReport.append(process.Process)
        # for the table data array
        LabelRow = LabelRow1[:]
        LabelRow.append(_U('TOTAL\n[MWh]'))
        LabelRowReport.append(_("TOTAL"))

        # calculating the monthly data start
        UPHMonthly = [] #UPH all processes (matrix)
        LabelColumn = [_U('Total')]
        TMONTHS = [U(TRANSMONTHS[key]) for key in MONTHS]
        LabelColumn.extend(TMONTHS)
        UPHMonthly.append(LabelColumn) #1 row: Total + 12 months labels

        # start of calculation loop
        UPHMonthlyTotal = [0.0] * 13    #Monthly total of ALL PROCESSES
        
        for process in processes:
            UPHMonthlyProc = self.sumMonthlyDemand(process)
            # calculating the total for process: UPHMonthlyProc[0]
            for month in range(1, 13):
                UPHMonthlyTotal[month] += UPHMonthlyProc[month] # Monthly total of all processes
            UPHMonthly.append(UPHMonthlyProc) #add 1 row for each process

        # calculating monthly totals
        GraphUPHMonthly = transpose(copy.deepcopy(UPHMonthly))
        GraphUPHMonthly[0] = LabelRow1

        UPHMonthlyTotal[0] = sum(UPHMonthlyTotal[1:])
        UPHMonthly.append(UPHMonthlyTotal) # Add as last column the total of all processes

        TableUPHMonthly = transpose(UPHMonthly) #1st column: Total + 12 months labels
        TableUPHMonthly.insert(0, LabelRow)     #Add 1st row: process names + total


        #UPHMonthly format after transpose and insert:
        # Proc.Heat Dem.    Proc1 Proc2 ... ProcN   TOTAL
        # Total
        # January
        # ...
        # December
        
        # end of the calculation loop
        data1 = np.array(GraphUPHMonthly)

        # creating the data array for the table
        data = np.array(TableUPHMonthly)
        
        # create the data array for the report
        data2 = transpose(UPHMonthly)
        data2.insert(0,LabelRowReport)
        data2 = transpose(data2)
        data2 = data2[:-1] + [[" "] * 14] * (22 - len(data2)) + data2[-1:] #fill up with empty lines

        # writing the arrays in interfaces
        Status.int.setGraphicsData(self.keys[0], data)
        Status.int.setGraphicsData(self.keys[1], data1)
        if Status.ANo == 0:
            if self._showHeating is True:
                Status.int.setGraphicsData("EM1_REPORT", data2)
            else:
                Status.int.setGraphicsData("EM1_REPORT_C", data2)
        elif Status.ANo == Status.FinalAlternative:
            if self._showHeating is True:
                Status.int.setGraphicsData("EM1_REPORT_F", data2)
            else:
                Status.int.setGraphicsData("EM1_REPORT_F_C", data2)
            

    def updatePanel(self, showHeating):
        """Update the cached data according to showHeating"""
        self._showHeating = showHeating
        self.initModule()
