# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEM2- Heat supply - Monthly data
#			
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
import wx
from constants import *
from GUI.status import *
import copy
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, transpose
from modules.messageLogger import logTrack, logMessage, logDebug
from modules.interfaces import Interfaces

class ModuleEM2(object):
    def __init__(self, keys, showHeating):
        self.keys = keys
        self._showHeating = showHeating
        self.initModule()

    def update(self):
        """call energy simulations, if you're not up to date"""
        Status.prj.getEquipments()  # make sure that Status.NEquipe is up to date
                                    # maybe can deleted later on, once all functions
                                    # that modify the equipment list do this automatically

        if Status.StatusEnergy == 0 or Status.int.cascadeUpdateLevel < Status.NEquipe:
            logTrack("ModuleEM2 (update): StatusEnergy %s" % Status.StatusEnergy)
            if Status.StatusCC > 0:
                Status.processData.createYearlyDemand()
            
            if Status.ANo > 0:
                logMessage(_("EINSTEIN running system simulation for updating energy balances"))
                logMessage(_("This may take a while. please be patient ..."))
                wx.SafeYield()
                Status.mod.moduleEnergy.runSimulation()
            logMessage(_("EINSTEIN now updating annual energy balances"))
            Status.mod.moduleEnergy.calculateEquipmentEnergyBalances()

    def initModule(self):
        """module initialization"""
        allequipments = Status.prj.getEquipments(cascade=True)
        eqObjects = Status.int.systemModel.eqObject
        if self._showHeating:
            equipments = [e for j, e in enumerate(allequipments) if eqObjects[j].heating]
            LabelRow1 = [_U('Process heat\nsupply')]
            LabelRow2 = [_('Process heat supply')]
        else:
            equipments = [e for j, e in enumerate(allequipments) if not eqObjects[j].heating]
            LabelRow1 = [_U('Process cooling\nsupply')]
            LabelRow2 = [_('Process cooling supply')]
        # filling the label row
        # for the graphic data array
        for equip in equipments:
            LabelRow1.append(unicode(equip.Equipment, "utf-8") + '\n[MWh]')
            LabelRow2.append(unicode(equip.Equipment, "utf-8"))
        # for the table data array
        LabelRow = copy.deepcopy(LabelRow1)
        LabelRow.append(_U('TOTAL\n[MWh]'))
        LabelRow2.append(_('TOTAL'))
        
        # taking the USH values from interfaces
        self.update()       # makes sure, that USH arrays are up to date
        
        USH_Tt = Status.int.USHj_Tt
        USH_t = Status.int.USHj_t
        
        # calculating the monthly data start
        USHMonthly = [] #USH all equipments (matrix)
        USHTotal = np.array([0.0] * 13)
        labelColumn = [_U('Total')]
        TMONTHS = [U(TRANSMONTHS[key]) for key in MONTHS]
        labelColumn.extend(TMONTHS)
        USHMonthly.append(labelColumn)
        # start of calculation loop
        if self._showHeating:
            maximumPoint = Status.NT + 1
        else:
            maximumPoint = 0
        for localJ in xrange(len(equipments)):
            USHMonthlyEquip = [0.0] * 13 #USH single equipment
            #Annual simulation. Hours month assumed = 730 (8760/12)
            month = 1
            j = allequipments.index(equipments[localJ])
            for it in range(Status.Nt):
                time = Status.TimeStep * it * Status.EXTRAPOLATE_TO_YEAR
                if time >= 24.0 * MONTHSTARTDAY[month]:
                    month += 1
                if Status.SimTool == "Plus":
                    USHMonthlyEquip[month] += USH_t[j][it] / 1000 #converted to [MWh]                    
                else:
                    USHMonthlyEquip[month] += USH_Tt[j][maximumPoint][it] / 1000 #converted to [MWh]
            #calculating the total for process: USHMonthlyEquip[0]
            USHMonthlyEquip[0] = sum(USHMonthlyEquip[1:13])
            USHMonthly.append(USHMonthlyEquip)

        # transposing the matrix: order as shown in table
        USHMonthly = transpose(USHMonthly)
        # creating the data array for the graph
        GraphUSHMonthly = copy.deepcopy(USHMonthly)
        GraphUSHMonthly[0] = LabelRow1
        # don't plot the yearly total 
        graphdata = [GraphUSHMonthly[0]] + GraphUSHMonthly[1:]
        data1 = np.array(graphdata)

        # calculating monthly totals
        for i in range(len(USHMonthly)):
            USHMonthlyTotal = sum(USHMonthly[i][1:])
            USHMonthly[i].append(USHMonthlyTotal)

        # creating the data array for the table
        TableUSHMonthly = copy.deepcopy(USHMonthly)
        TableUSHMonthly.insert(0, LabelRow)
        tabledata = TableUSHMonthly
        data = np.array(tabledata)

        # writing the arrays in interfaces
        Status.int.setGraphicsData(self.keys[0], data)
        Status.int.setGraphicsData(self.keys[1], data1)

        dataR = tabledata[:]
        dataR[0] = LabelRow2
        dataR = transpose(dataR)
        dataR = dataR[:-1] + [[" "] * 14] * (12 - len(dataR)) + dataR[-1:] 

        if Status.ANo == 0:
            if self._showHeating is True:
                Status.int.setGraphicsData("EM2_REPORT", dataR)
            else:
                Status.int.setGraphicsData("EM2_REPORT_C", dataR)
        elif Status.ANo == Status.FinalAlternative:
            if self._showHeating is True:
                Status.int.setGraphicsData("EM2_REPORT_F", dataR)
            else:
                Status.int.setGraphicsData("EM2_REPORT_F_C", dataR)
    
    def updatePanel(self, showHeating):
        """Update the cached data according to showHeating"""
        self._showHeating = showHeating
        self.initModule()

    def exitModule(self, exit_option):
        """Carry out any calculations necessary previous to 
        displaying the HP design assistant window.
        
        """
        if exit_option == "save":
            logDebug("exitModule: here I should save the current configuration")
        elif exit_option == "cancel":
            logDebug("exitModule: here I should retrieve the previous configuration")
        else:
            logDebug("exitModule: function not yet defined")
        return "ok"
