# -*- coding: utf-8 -*-
#==============================================================================#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	ModuleEditCon - Calculations for connection editor
#			
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.status import Status
from globals import logger
from modules.energyStats.modulePA1 import sortStreams
import numpy as np


class ModuleEditHXNet(object):

    def __init__(self, key="EDITHXNET"):
        self.key=key

    def updatePanel(self):

        if not Status.int.systemModel or (Status.PId is None):
            return None

        Status.int.systemModel.build()

        srcNet = Status.int.systemModel.srcNet      # fetch the network from the system model
        snkNet = Status.int.systemModel.snkNet
            
        sourceStreams, sinkStreams = sortStreams()  # returns sorted streams from systemModel

        nodeListSrc, rdNodeListSrc = self.getNodeList(srcNet, sourceStreams)
        posMaxSrc = srcNet.buildFlowChart(nodeListSrc)

        nodeListSnk, rdNodeListSnk = self.getNodeList(snkNet, sinkStreams)
        posMaxSnk = snkNet.buildFlowChart(nodeListSnk)


        hxTable = []
        for h, hx in enumerate(Status.int.hx):
            Thi = hx.ThiAvg
            Tho = hx.ThoAvg
            Tci = hx.TciAvg
            Tco = hx.TcoAvg

            srcNode = srcNet.findNode(objectCode=("HX", h, 0, False))
            snkNode = snkNet.findNode(objectCode=("HX", h, 0, True))
            if not (srcNode is None or snkNode is None):
                hxTable.append([srcNode, snkNode, Thi, Tho, Tci, Tco, hx.name])
                srcNode.T1 = Thi
                srcNode.T0 = Tho
                snkNode.T1 = Tco
                snkNode.T0 = Tci
        
        data = {}
        data["hx"] = hxTable
        
##        data["srcSplitter"] = sourceSplitterTable
##        data["snkSplitter"] = sinkSplitterTable
        data["net"] = (srcNet, snkNet)
        
        posMax = np.maximum(posMaxSrc, posMaxSnk)

        for node in srcNet.node:
            if node.objectCode[0]=="SPLIT":
                up = node.up()
                node.T0 = up[0].T0
                node.T1 = up[0].T0
        for node in srcNet.node:
            if not hasattr(node, "T0"):
                up = node.up()
                node.T1 = up[0].T0
                node.T0 = 0

        for node in snkNet.node:
            if node.objectCode[0]=="SPLIT":
                up = node.up()
                node.T0 = up[0].T1
                node.T1 = up[0].T1
        for node in snkNet.node:
            if not hasattr(node, "T0"):
                up = node.up()
                node.T0 = up[0].T0
                node.T1 = 20.0
        # TODO: assign correct end temperatures for RD-nodes


        self.arrangeNodes(srcNet, snkNet, posMax, posMaxSrc, posMaxSnk, hxTable)
        data["grid"] = (posMaxSrc, posMaxSnk, posMax)

        if len(srcNet.node) + len(snkNet.node) > 0:
            Tmin = np.min([node.T0 for node in srcNet.node+snkNet.node])
            Tmax = np.max([node.T0 for node in srcNet.node+snkNet.node])
        else:
            Tmin = 0
            Tmax = 400
        data["TScale"] = (Tmin, Tmax)

        Status.int.GData[self.key] = data

        self.data = data

        return data

    def getNodeList(self, net, streams):
        """
        Helper function: getting the nodes associated to the streams
        """
        nodeList = []
        rdNodeList = []
        
        for stream in streams:
            code = stream.origin.compact()
            node = net.findNode(objectCode=code)
            if node is None:
                logger.logError("ModuleEditHXNet.getNodeList: no node found for code %s"%str(code))
                continue
            node.name = "%s: %s"%(stream.origin.objectName, stream.origin.streamName)
            T = stream.T0 if stream.heating else stream.T1
            node.T0 = T
            node.T1 = T
            nodeList.append(node)

            rdCode = list(code)
            rdCode[0] = "RD." + code[0]
            rdCode = tuple(rdCode)
            rdNodes = net.findNode(objectCode=rdCode, findAll=True)
            T = stream.T1 if stream.heating else stream.T0
            for rdNode in rdNodes:
                rdNode.T0 = T
                rdNode.T1 = T
                rdNode.name = "RD %s: %s"%(stream.origin.objectName, stream.origin.streamName)
                rdNodeList.append(node)

        return nodeList, rdNodeList


    def arrangeNodes(self, srcNet, snkNet, posMax, posMaxSrc, posMaxSnk, hxTable):
        """
        arranges the i-position of the node within the drawing
        in order to minimise the shift for each of the HXes
        """

        # step 1: assign ipos
        # ipos = pos for snk streams
        # ipos = n - pos for src streams, where n = max(Src, nSnk)
        # ipos ranges from 0 to n-1

        imax = posMax[0]
        for node in srcNet.node:
            setattr(node, "ipos", (imax - node.pos[0]) if hasattr(node, "pos") and node.pos[0] >= 0 else 0)

        for node in snkNet.node:
            setattr(node, "ipos", node.pos[0] if hasattr(node, "pos") and node.pos[0] >= 0 else imax)

        # step 2: move the nodes corresponding to each heat exchanger so that the
        # shift is minimised
        for ctr in range(3):
            for h, hx in enumerate(hxTable):
                srcNode = hx[0]
                snkNode = hx[1]

                shift = srcNode.ipos - snkNode.ipos
                if shift > 0: # try to move srcNode to the left or snkNode to the right
                    srcDown = srcNode.down()
                    iMinSrc = max([node.ipos for node in srcDown]) + 1 if len(srcDown)>0  else 0
                        
                    snkDown = snkNode.down()
                    iMaxSnk = min([node.ipos for node in snkDown]) - 1 if len(snkDown)>0  else imax


                    if iMinSrc <= imax - iMaxSnk: # preference to shift source
                        shiftSrc = min(shift, srcNode.ipos - iMinSrc)
                        shiftSnk = min(shift - shiftSrc, iMaxSnk - snkNode.ipos)
                    else:
                        shiftSnk = min(shift, iMaxSnk - snkNode.ipos)
                        shiftSrc = min(shift-shiftSnk, srcNode.ipos - iMinSrc)
                        
                    srcNode.ipos -= shiftSrc
                    snkNode.ipos += shiftSnk
                    
                elif shift < 0: # try to move srcNode to the right or snkNode to the left
                    shift = -shift
                    snkUp = snkNode.up()
                    iMinSnk = max([node.ipos for node in snkUp]) + 1 if len(snkUp)>0  else 0
                        
                    srcUp = srcNode.up()
                    iMaxSrc = min([node.ipos for node in srcUp]) - 1 if len(srcUp)>0 else imax

                    if iMinSnk <= imax - iMaxSrc: # preference to shift sink
                        shiftSnk = min(shift, snkNode.ipos - iMinSnk)
                        shiftSrc = min(shift - shiftSnk, iMaxSrc - srcNode.ipos)
                    else:
                        shiftSrc = min(shift, iMaxSrc - srcNode.ipos)
                        shiftSnk = min(shift - shiftSrc, snkNode.ipos - iMinSnk)
                        
                    srcNode.ipos += shiftSrc
                    snkNode.ipos -= shiftSnk
                else:
                    shiftSrc = 0
                    shiftSnk = 0


        # optional step: expand the grid where there are (too) many hx

        # final step: assign the properties to the flow (really necessary ?)

        



        
        
