# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#    equipments
#           
#------------------------------------------------------------------------------
#
#    convenience functions for access to equipment data
#           
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------     
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from GUI.status import *
from modules.interfaces import createQ_Tt, createQ_T
from modules.messageLogger import logDebug, logWarning, logTrack
from modules.fluids import Fuel, getDefaultAir
from modules.heatFlow import HeatFlow_t
from constants import getEquipmentClass
from modules.boiler.combustion import getCombustionParameters, TMINCOMBAIR, TMAXCOMBAIR
import numpy as np

def getConnectedPipes(equipmentIndex, status):
    """Return the array indices of pipes that are connected to the
    given equipment.
    
    :param equipmentIndex: the array address of the equipment. Should
    be the equipment number - 1. 
    
    :returns: a list of the array indices of the pipes that are 
    connected to the pipe.
    """
    result = [m for m in range(len(status.int.USHLink)) if (status.int.USHLink[m][equipmentIndex] == 1)]
    
    return result

def estimateEquipmentWasteHeat():
    """Estimate the waste heat of all equipments, based on their
    operation hours.
    
    That fills Status.int.QWHEqTotal_t and Status.int.QWHEq_Tt.
    
    """
    if Status.schedules.equipmentSchedulesOutOfDate == True:
        Status.schedules.calculateEquipmentSchedules()
    #   estimate equipment waste heat from annual QWHj and equipment schedules
    equipments = Status.prj.getEquipments(cascade=True)
    NJ = len(equipments)

    Status.int.QWHEqTotal_Tt = createQ_Tt()
    Status.int.QWHEqTotal_T = createQ_T()
    Status.int.QWHEqTotal_t = createQ_T()
    Status.int.QWHEqTotal = 0
    Status.int.QWHEq_Tt = []
    Status.int.QHXEqReq_Tt = []
    
    for j, equipe in enumerate(equipments):


        equipmentType = equipe.EquipType
        equipmentClass = getEquipmentClass(equipmentType)

        schedule = Status.schedules.equipmentSchedules[j]

        # defaults. to be overwritten by specific equipment types / classes
        f_combAir = 0

        if equipmentClass in ["CH"]:

            # temperature distribution of waste heat. assumed as fix
            
            QWHj = equipe.QWHEq
            if QWHj is None:
                QWHj = 0.0
            Status.int.QWHEqTotal += QWHj

            Treject = 30.
            
            QWHj_T = createQ_T()
            for iT in range(Status.NT + 2):
                temp = Status.int.T[iT]
                if temp <= Treject:
                    QWHj_T[iT] = QWHj
                Status.int.QWHEqTotal_T[iT] += QWHj_T[iT]
                

            QWHj_Tt = createQ_Tt()
            f_t = schedule.fav
            QWHj_Tt += QWHj_T.reshape((Status.NT + 2, 1)) # broadcast into Tt shape
            QWHj_Tt *= f_t # multiply with the schedule 
                
            Status.int.QWHEqTotal_Tt += QWHj_Tt

            if QWHj > 0:
                Status.int.QWHEq_Tt.append(HeatFlow_t(fluid_id=0,
                                           T0=Treject-5,
                                           T1=Treject,
                                           Q_t = QWHj*schedule.fav,name="HeatRejection"))
            else:
                Status.int.QWHEq_Tt.append(HeatFlow_t())

        elif equipmentClass in ["BB","CHP"]:
            
            PNom = equipe.HCGPnom
            PEl = equipe.ElectriConsum
            COPh_nom = equipe.HCGTEfficiency
            if PNom is None:
                PNom = 0.0
                logWarning("Equipments.estimateEquipmentWasteHeat: No nominal power specified for equipe no. %s" % \
                     (EquipmentNo))

            QWHj = equipe.QWHEq
            if QWHj is None:
                QWHj = 0.0
            Status.int.QWHEqTotal += QWHj

            if equipmentType in ["CHP ORC", "CHP TFC"]:
                PEl = 0.0
                fuel = None
                
                T0 = 20.0
                dTmin = 0.1
                
                dT_T = np.linspace(0, Status.NT + 1, Status.NT + 2)
                dT_T = dT_T * Status.TemperatureInterval - T0
                dT_T = dT_T.clip(min=0)
                f_QWHu_T = (1.0 - dT_T / dTmin).clip(min=0)
                
            else:
                if PEl is None:
                        PEl = 0.01 * PNom
                        logWarning("Equipments.estimateEquipmentWasteHeat: No electrical consumption specified for equipe no. %s" % \
                             (EquipmentNo))

                air = getDefaultAir()
                fuel, f_QWHu_T, fQWH_HX_u_T, f_combAir, TExhaustGas, TMinOffGas = getCombustionParameters(equipe, PNom, PEl, COPh_nom, air)
            

            QWHj_T = QWHj * f_QWHu_T
            Status.int.QWHEqTotal_T += QWHj_T
                
            QWHj_Tt = createQ_Tt()
            f_t = schedule.fav
            QWHj_Tt += QWHj_T.reshape((Status.NT + 2, 1)) # broadcast into Tt shape
            QWHj_Tt *= f_t # multiply with the schedule          
            Status.int.QWHEqTotal_Tt += QWHj_Tt

            if QWHj > 0:
                if fuel is None:
                    Status.int.QWHEq_Tt.append(HeatFlow_t(fluid_id=0,
                                               T0=99.1,
                                               T1=99.2,
                                               Q_t = QWHj*schedule.fav,name="WasteHeat"))
                else:
                    Status.int.QWHEq_Tt.append(HeatFlow_t(fluid_id=-fuel.id,
                                               T0=min(TMinOffGas + 0.001,TExhaustGas),
                                               T1=TExhaustGas,
                                               Q_t = QWHj*schedule.fav,name="ExhaustGas"))
            else:
                Status.int.QWHEq_Tt.append(HeatFlow_t())
        else:
            Status.int.QWHEq_Tt.append(HeatFlow_t())

        Status.int.QWHEq_Tt[j].exportGlb(Status.int.systemModel.eqObject[j].QWHGlb)

        if f_combAir == 0:
            Status.int.QHXEqReq_Tt.append(HeatFlow_t())

        elif equipmentType in ["CHP ORC", "CHP TFC"]:
            try:
                Qin = QWHj / COPh_nom
            except:
                Qin = 0.0
                
            QHXEqReq_t = Qin*Status.schedule.fav

            from modules.chp.moduleCHP import DEFAULT_TDRIVE, DEFAULT_DTDRIVE
            TDrive = equipe.THeatSourceHT
            if TDrive is None: TDrive = DEFAULT_TDRIVE

            Status.int.QHXEqReq_Tt.append(HeatFlow_t(fluid_id=air.id,T0=TDrive - DEFAULT_DTDRIVE,T1=TDrive,Q_t = QHXEqReq_t))            
            
        else:
            USHj = equipe.USHj if equipe.USHj is not None else 0.0
            QHXEqReq_t = USHj * f_combAir * schedule.fav
            Status.int.QHXEqReq_Tt.append(HeatFlow_t(fluid_id=air.id,T0=TMINCOMBAIR + 0.001,T1=TMAXCOMBAIR,Q_t = QHXEqReq_t))            
        
    # buffer up if there are more cascade levels than equipments
    try:
        diffsize = Status.int.cascadeSize - Status.NEquipe
    except:
        diffsize = 0
    
    Status.int.QWHEq_Tt += [HeatFlow_t() for dummy in xrange(diffsize)]
    Status.int.QHXEqReq_Tt += [HeatFlow_t() for dummy in xrange(diffsize)]
    Status.int.QHXEqReqChanged = [True for dummy in Status.int.QHXEqReq_Tt]
    
    Status.int.QWHEqTotal_t = Status.int.QWHEqTotal_Tt[0].copy()

    Status.int.QWHjChanged = [True for dummy in range(NJ + diffsize)]

    #Allocate empty space for simulation results
    Status.int.QWHEqUsed_Tt = [HeatFlow_t() for dummy in range(NJ + diffsize)]
##    Status.int.QHXEqReq_Tt = [HeatFlow_t() for dummy in range(NJ + diffsize)]
##    Status.int.QHXEqReqChanged = [True for dummy in range(NJ + diffsize)]
    Status.int.QHXEqTotal = 0
    Status.int.QHXEqReqTotal = 0
 
    #Start with 0 as first estimate for FETHeat and QHX
    Status.int.FETHeatTotal = 0
    Status.int.FETHeat_jChanged = [True for dummy in range(NJ + diffsize)]
   
class Equipment(object):
    """Represent an equipment of an industry. E.g. a boiler."""
    pass
        
