# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   moduleQ3
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel Q3 (process data)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#                   
#------------------------------------------------------------------------------     
#============================================================================== 
from globals import Globals, logger
from constants import EPSILON

# imports for equipment specific functions
from constants import HPTYPES, HPSUBTYPES, TRANSHPSUBTYPES, PRODUCTCODES, TRANSPRODUCTCODES

from GUI.status import Status
from GUI.conflictFrame import conflictFrame

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck, findKey
from auxiliary.dataTransfer import DataTransfer
from lib.auxiliary.float import toFloat

from model.objects.genericObject import GenericObject

from modules.fluids import Fluid
from modules.schedules.schedules import DEFAULTCHARGETIME

# keys for parameters which are 1:1 passed to data
# structure: key_in_procObject: (key_in_Panel (or None, if identical), dictionary)


class ModuleDB(object):
    """

    Attributes:

    
    """
    def __init__(self, panel, table, colKeys, sqlKeys=None, dataKeys=None, preselection=None, selectedId=None):
        """
        panel:      link to panel for feed-back loops
        colKeys:    keys of data to be represented in the overview table (ObjectList)
        sqlKeys:    keys of all data to be read from SQL [currently set later on directly from specific panel]
        dataKeys:   keys of all data to be exported to GUI
        """
        self.panel = panel
        self.panel.selectedId = selectedId    # return selectedId for EndModal
        self.selectedId = selectedId
        
        self.table = table
        self.colKeys = colKeys
        self.sqlKeys = sqlKeys
        self.dataKeys = dataKeys
        self.preselection = preselection
        
        self.selectedRow = preselection[0] if (preselection is not None and len(preselection)==1) else None
                                    # selected row (out of data currently on display)
        self.new = False

        self.sort_col = None        # memory used for sorting
        self.sort_reverse = False

        self.newObject = False         # indication for the panel if a new object has been created and shall be edited
        self.activeObject = None    # active object to be edited

        self.editor = None          # indicates if there is an edit-panel open

        # dictionary for dynamic translations
        self.dynamicTrans = {}
        
    def initPanel(self, filters):
        """  functions that have to be carried out only once
        when the panel is created
        """
        fluidDict = Status.prj.getFluidDict(includeZeroWater=True)
        fluidDictItems = [(fluidId, name) for fluidId, name in fluidDict.items()]
        fluidDictItems =  sorted(fluidDictItems, key=lambda item: item[1])
        self.fluidChoices = [item[1] for item in fluidDictItems]
        self.fluidIds = [item[0] for item in fluidDictItems]
        self.dynamicTrans["Fluid"]  = fluidDict
        self.dynamicTrans["FluidId"] = dict([(item, i) for i, item in enumerate(self.fluidIds)])

        unitOpDict = Status.prj.getUnitOpDict(code=True)
        unitOpItems = [(code, text) for code, text in unitOpDict.items()]
        sortedItems =  sorted(unitOpItems, key=lambda item: item[0])
        self.dynamicTrans["UnitOp"] = dict(sortedItems)
        
##        unitOpList = [text for (code, text) in sortedItems]
##        unitOpCodes = [code for (code, text) in sortedItems]

        self.loadObjects()

        # append information for filter choice entries
        data = {"Type": None, "SubType": None, "SubType2": None}

        self.filter = {"Type": None, "SubType": None, "SubType2": None}
        self.filterCol = {"Type": None, "SubType": None, "SubType2": None}
        for key in ["Type", "SubType", "SubType2"]:
            if key in filters.keys():
                f = filters[key]
                if f is not None:
                    choices = self.filterChoices(f[0], None)
                    choiceDict = f[3]
                    if choiceDict is not None:
                        choices = [choiceDict[choice] if choice in choiceDict.keys() else choice for choice in choices]
                    data[key] = {"values": choices, "label": f[1], "tip": f[2]}
                    self.filterCol[key] = None if f is None else f[0]

        return data
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """
        entries = self.getObjectList()

        # use selectedId only once at start-up
        if self.selectedRow is None and self.selectedId is not None:
            objectIds = [obj.id for obj in self.filteredObjectList]
            if self.selectedId in objectIds:
                self.selectedRow = objectIds.index(self.selectedId)
        self.selectedId is None
            
        selection = [] if self.selectedRow is None else [self.selectedRow]

        data = {}
        data["ObjectList"] = {"data": entries, "choices": None, "selection": selection}

        return data

    def typeChanged(self, index):
        choices = self.filterChoices(self.filterCol["Type"], None)
        newFilter = choices[index] if index > 0 else None
        changed = (newFilter != self.filter["Type"])
        self.filter["Type"] = newFilter
        self.filter["SubType"] = None

        if changed:
            return self.filterChoices(self.filterCol["SubType"], None)
        else:
            return None

    def subTypeChanged(self, index):
        print "ModuleDB.subTypeChanged: index = %s"%index
        choices = self.filterChoices(self.filterCol["SubType"], None)
        newFilter = choices[index] if index > 0 else None
        changed = (newFilter != self.filter["SubType"])
        self.filter["SubType"] = newFilter
        self.filter["SubType2"] = None

        if changed:
            return self.filterChoices(self.filterCol["SubType2"], None)
        else:
            return None
    
    def subType2Changed(self, index):
        choices = self.filterChoices(self.filterCol["SubType2"], None)
        newFilter = choices[index] if index > 0 else None
        changed = (newFilter != self.filter["SubType2"])
        self.filter["SubType2"] = newFilter

    
    def changeDataObjectList(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False
        nRows = len(self.filteredObjectList)

        # check validity of mode and index combination
        if (index > nRows or index < -1) and mode==3:
            return 1, None, None

        if mode == 0:     # view
            self.editObject = True
            
        elif mode == 1:     # add
            self.createNewObject()
            self.editObject = True
            
        elif mode == 2:   # delete
            self.selectRow(index)
            self.deleteObject()
            self.editObject = False

        data = self.getObjectList()

        choices = None
        
        return 0, data, choices

    def loadObjects(self):
        """
        helper function: loads all objects from the data base into
        a list of local objects (which can be locally modified without
        direct read/write to the data base
        """
        sqlTable = dbio.selectByList(self.table, "id", self.preselection) if self.preselection is not None else dbio.select(self.table) 

        self.mandatoryData = []
        self.optionalData = []

        self.objects = [GenericObject(self,
                                         id = sqlData.id,
                                         no = i,
                                         sqlData=sqlData,
                                         sqlKeys=self.sqlKeys,
                                         dynamicTrans=self.dynamicTrans,
                                         mandatoryData=self.mandatoryData,
                                         optionalData=self.optionalData,
                                         sqlTableName=self.table,
                                         keyPId=None,
                                         colObjectNo=None) \
                           for i, sqlData in enumerate(sqlTable)]

        # scale those values which are stored in kJ/kg(K) instead of kWh/kg(K) within the SQL data base
        # used only for dbfluid data base
        for obj in self.objects:
            for key in ["FluidCp", "FluidCpG", "LatentHeat", "SensibleHeat", "OffgasHeatCapacity"]:
                if hasattr(obj, key):
                    val = getattr(obj, key)
                    if val is not None:
                        setattr(obj, key, val / 3600.0)

            # special treatment -> extend dbbenchmark entries to 2 digits in order to match with Code+Subcode from DBNACECode
            if self.table=="dbbenchmark":
                if hasattr(obj, "NACECode"):
                    val = getattr(obj, "NACECode")
                    changeInNACE = False
                    while val is not None and len(val) < 7:
                        if not ("." in val):
                            val += "."
                        val = val + "0"
                        changeInNACE = True

                    if changeInNACE:
                        setattr(obj, "NACECode", val)

        self.deletedObjects = []
        self.changedObjects = []

    def getObjectList(self):
        """
        translates the data in self.proc into a data dicitionary as
        required by the panel
        """
        #TODO: provisional workaround, while not all colKeys are explicitely
        # specified in all DB-panels
        if self.colKeys == []:
            self.colKeys = self.colLabels
            
##                if subtype == "ProductCode":
##                    equipe_subtype = equipe_subtype.split(':')[0]
##                selectionCriteria.update({subtype: equipe_subtype})
##
##            equipe_subtype = '\'%s\''%equipe_subtype if equipe_subtype != 'NULL' else equipe_subtype
##            equipe_subtype2 = '\'%s\''%equipe_subtype2 if equipe_subtype2 != 'NULL' else equipe_subtype2

        filtered = self.objects
        for key in ["Type", "SubType", "SubType2"]:
            filterVal = self.filter[key]
            filterCol = self.filterCol[key]
            
            if filterVal is not None:
                filtered = [obj for obj in filtered if str(getattr(obj, filterCol))==filterVal]
              
        sort_by_col = 1 # by default sort by the object name
        sortedResults = self.sortResults(filtered, sort_by_col)

        self.filteredObjectList = []        
        for result in sortedResults:
            id = result.id
            if self.preselection is not None:
                if id not in self.preselection:
                    continue
                
##            self.appendResultToGrid(result)
            self.filteredObjectList.append(result)

        data = []
        for r, row in enumerate(self.filteredObjectList):
            guiRow = self.getObjectData(r)      # carried out translation where necessary using DATAKEYS
            data.append([guiRow[key] for key in self.colKeys])

        return data


######    def appendResultToGridBenchmark(self, result):
######        if result is not None:
######            self.grid.AppendRows(1, True)
######            for i in range(len(self.colLabels)):
######                colKey = self.colKeys[i] if i > 0 else "id"
######                value = getattr(result, colKey)
######                
######                if self.colLabels[i] == "UnitOp":
######                    try:
######                        self.grid.SetCellValue(self.grid.GetNumberRows() - 1, i, str(value) + ": " + self.unitOpDict[int(value)])
######                    except:
######                        self.grid.SetCellValue(self.grid.GetNumberRows() - 1, i, str(value))
######                else:
######                    self.grid.SetCellValue(self.grid.GetNumberRows() - 1, i, str(value))


    def sortResults(self, results, sort_by_col):

        if sort_by_col is not None:
            cols = range(1, len(self.colKeys))    # order by columns, move sort_by col first
            cols.remove(sort_by_col)
            cols = [sort_by_col] + cols

        for col in reversed(cols):
            results = sorted(results, key = lambda b: getattr(b, self.colKeys[col]), reverse = self.sort_reverse)

        return results


######    def getSecondaryUnits(self):
######        """ collects the conversion factors for secondary units depending on the fluids seleced
######        """
######        data = {"startUp": self.proc.fluid.rho if self.proc.fluid is not None else 1.0,
######                "inFlow": [stream.fluid.rho  if stream.fluid is not None else 1.0 for stream in self.proc.inStreams],
######                "outFlow": [stream.fluid.rho if stream.fluid is not None else 1.0 for stream in self.proc.outStreams]}
######
######        return dat

    def getObjectData(self, row=None):
        """
        collects the data for the object to be plotted on the DBEntry panel
        """
        q = self.activeObject if row is None else self.filteredObjectList[row]

        transfer = DataTransfer(self.dynamicTrans, self.dataKeys)
        data = transfer.write(q)

        return data

    def setObjectData(self, data):
        if self.activeObject is None:
            return

        transfer = DataTransfer(self.dynamicTrans, self.dataKeys)
        transfer.read(data, self.activeObject)
        
    def selectRow(self, row):
        self.selectedRow = row
        self.activeObject = self.filteredObjectList[row]
        self.panel.selectedId = self.activeObject.id

    def createNewObject(self):
        new = GenericObject(self,
                                         id = None,
                                         no = len(self.objects),
                                         sqlData=None,
                                         sqlKeys=self.sqlKeys,
                                         dynamicTrans=self.dynamicTrans,
                                         mandatoryData=self.mandatoryData,
                                         optionalData=self.optionalData,
                                         sqlTableName=self.table,
                                         keyPId=None,
                                         colObjectNo=None)

        self.activeObject = new
        self.newObject = True
                                                                                   
    def addObject(self):
        if not self.newObject: return
        self.objects.append(self.activeObject)
        self.changedObjects.append(self.activeObject)                 # keep memory in order to save to data base later on

        self.panel.unsavedChanges = True
        
        self.getObjectList()
        
        if self.activeObject in self.filteredObjectList:
            self.selectedRow = self.filteredObjectList.index(self.activeObject)
        else:
            self.selectedRow = None
            self.activeObject = None
            
        self.newObject = False

        
    def deleteObject(self):
        if self.selectedRow is None: return

        if self.activeObject is not None and self.activeObject in self.objects:
            index = self.objects.index(self.activeObject)
            self.deletedObjects.append(self.objects.pop(index))

            self.panel.unsavedChanges = True

    
    def saveToDB(self, data):
        for obj in self.changedObjects:
            if obj.id is not None:
                name = obj.name if (hasattr(obj, "name") and obj.name is not None) else _("this object")
                ret = logger.askConfirmation(_("Modifiying data for %s may affect all projects where these data are used\n")%name + \
                                               _("Do you want to continue ?"))
            else:
                ret = True

            if ret:
                # scale back from kWh to kJ before saving
                for key in ["FluidCp", "FluidCpG", "LatentHeat", "SensibleHeat", "OffgasHeatCapacity"]:
                    if hasattr(obj, key):
                        val = getattr(obj, key)
                        if val is not None:
                            setattr(obj, key, val * 3600.0)

                obj.saveToDB()

        for obj in self.deletedObjects:
            if obj.id is not None:
                name = obj.name if (hasattr(obj, "name") and obj.name is not None) else _("this object")
                ret = logger.askConfirmation(_("Deleting data for %s may affect all projects where these data are used\n")%name + \
                                               _("Do you want to continue ?"))
                if ret:
                    obj.deleteFromDB()
    
    def cancel(self):
        self.new = False


#============================================================
# functions for filling lists of choices
#============================================================

    def filterChoices(self, typeKey, typeDict, base=None):
        """
        gets the available types
        """
##        try:
        entries = self.objects if base is None else base
        
        typeList = []
        for entry in entries:
            result = getattr(entry, typeKey) if (typeKey is not None and hasattr(entry, typeKey)) else None
            if typeKey == "ProductCode": # dbbenchmark
                try:
                    result = TRANSPRODUCTCODES[str(result)]
                except:
                    result = str(result)
            if typeKey == "UnitOp": # dbbenchmark
                try:
                    result = str(result) + ": " + self.unitOpDict[int(result)]
                except:
                    result = str(result)

            if typeDict is not None: # translate, if dictionary available
                if str(result) in typeDict.keys():
                    result = typeDict[str(result)]
                
            if str(result) not in typeList and result is not None:
                typeList.append(str(result))

        typeList.sort()
        typeList = [_("All")] + typeList
##            self.addNoneFront(typeList)
        if type == "NACECode": # dbbenchmark
            typeList.extend(self.getNACECodeBranchList())

        return typeList

    def getNACECodeandNACESubCodeList(self, addDescription=True):
        naceTable = dbio.select("dbnacecode")
        naceList = []
        for entry in naceTable:
            naceCode = entry.CodeNACE
            naceSubCode = entry.CodeNACEsub
            naceSubName = entry.NameNACEsub
            naceCode = naceCode + "." + naceSubCode
##            if len(str(naceSubName).strip()) > 0:
            if naceSubName is not None and addDescription:
                strippedName = str(naceSubName).strip()
                if len(strippedName) > 0:
                    naceCode += ": " + strippedName
                    
            naceList.append(str(naceCode))
        naceList.sort()
        return naceList

    def getNACECodeBranchList(self):
        naceTable = dbio.select("dbnacecode")
        naceList = []
        for entry in naceTable:
            naceCode = entry.CodeNACE
            if naceCode is not None:
                if str(naceCode) + ".*" not in naceList:
                    naceList.append(str(naceCode) + ".*")
        naceList.sort()
        return naceList


    def getFuelUnitList(self):
        fuelUnitsList = []
        for entry in UNITS["MASSORVOLUME"]:
            fuelUnitsList.append(str(entry))
        fuelUnitsList.sort()
        return fuelUnitsList

    def getUnitOpCodeList(self):
        unitOpTable = dbio.select("dbunitoperation")
        unitOpList = []
        for entry in unitOpTable:
            unitOpCode = entry.UnitOperationCode
            unitOpName = entry.UnitOperation
            unitOpList.append(unitOpCode + ": " + unitOpName)
        unitOpList.sort()
        return unitOpList

    def getEUnitList(self):
        eUnitTable = dbio.select("dbbenchmark")
        eUnitList = []
        for entry in eUnitTable:
            eUnit = entry.E_Unit
            if str(eUnit) not in eUnitList and len(str(eUnit)) > 0:
                eUnitList.append(str(eUnit))
        return eUnitList

    def getHUnitList(self):
        hUnitTable = dbio.select("dbbenchmark")
        hUnitList = []
        for entry in hUnitTable:
            hUnit = entry.H_Unit
            if str(hUnit) not in hUnitList and len(str(hUnit)) > 0:
                hUnitList.append(str(hUnit))
        return hUnitList

    def getTUnitList(self):
        tUnitTable = dbio.select("dbbenchmark")
        tUnitList = []
        for entry in tUnitTable:
            tUnit = entry.T_Unit
            if str(tUnit) not in tUnitList and len(str(tUnit)) > 0:
                tUnitList.append(str(tUnit))
        return tUnitList

    def getSourceSinkList(self):
        sourceSinkList = HPSOURCESINK.values()
        sourceSinkList.sort()
        return sourceSinkList

    def getFluidNameList(self):
        fluidTable = dbio.select("dbfluid")
        fluidNameList = []
        for entry in fluidTable:
            fluidName = entry.FluidName
            if str(fluidName) not in fluidNameList and len(str(fluidName)) > 0:
                fluidNameList.append(str(fluidName))
        fluidNameList.sort()
        return fluidNameList

    def getHPSubTypeList(self, HPType=None):
        if HPType in HPSUBTYPES.keys():
            hpSubTypeList = HPSUBTYPES[HPType]
        else:
            hpSubTypeList = []
        
        return hpSubTypeList

    def getCHPSubTypeList(self):
        chpSubTypeList = []
        for list in CHPSUBTYPES.values():
            listEntry = ""
            for e in list:
                listEntry += e + "-"
            chpSubTypeList.append(listEntry.rstrip('-'))
        chpSubTypeList.sort()
        return chpSubTypeList

    def getAuditorName(self):
        stool = dbio.get("stool")
        auditor = dbio.get("auditor", id=stool.Auditor_ID)
        if auditor is None:
            return None
        else:
            return auditor.Name
        
    def getFluidIdOfName(self, name):
        fluid = None
        if name is not None and len(str(name)) > 0:
            fluid = dbio.get("dbfluid", FluidName=str(name))
        if fluid is None:
            return 0
        else:
            return fluid.id

    def getFluidNameOfId(self, id):
        result = None
        if id is not None:
            fluid = dbio.get("dbfluid", id=id)
            if fluid is not None:
                result = fluid.FluidName
        if result is None: result = "None"
        return str(result)

    def fillChoiceOfNaceCode(self, entry):
        naceList = self.getNACECodeandNACESubCodeList()
        self.addNoneFront(naceList)
        fillChoice(entry, naceList)


    def fillChoiceOfDBFuel(self, entry):
        fuelDict = FUELTYPES
        fuelList = fuelDict.values()
        fuelList.sort()
        self.addNoneFront(fuelList)
        fillChoice(entry, fuelList)

    def fillChoiceOfDBFuelType(self, entry):
        fuelDict = FUELTYPES
        fuelList = fuelDict.keys()
        fuelList.sort()
        self.addNoneFront(fuelList)
        fillChoice(entry, fuelList)

    def fillChoiceOfDBFuelUnits(self, entry):
        fuelUnitList = self.getFuelUnitList()
        self.addNoneFront(fuelUnitList)
        fillChoice(entry, fuelUnitList)

    def fillChoiceOfDBUnitOpCodes(self, entry):
        unitOpList = self.getUnitOpCodeList()
        self.addNoneFront(unitOpList)
        fillChoice(entry, unitOpList)

    def fillChoiceOfEUnit(self, entry):
        productUnitList = self.getEUnitList()
        self.addNoneFront(productUnitList)
        fillChoice(entry, productUnitList)

    def fillChoiceOfHUnit(self, entry):
        productUnitList = self.getHUnitList()
        self.addNoneFront(productUnitList)
        fillChoice(entry, productUnitList)

    def fillChoiceOfTUnit(self, entry):
        productUnitList = self.getTUnitList()
        self.addNoneFront(productUnitList)
        fillChoice(entry, productUnitList)

    def fillChoiceOfHPSourceSink(self, entry):
        sourceSinkList = self.getSourceSinkList()
##        self.addNoneFront(sourceSinkList)
        fillChoice(entry, sourceSinkList)

    def fillChoiceOfHPAbsHeatMed(self, entry):
        absHeatMedList = self.getFluidNameList()
        self.addNoneFront(absHeatMedList)
        fillChoice(entry, absHeatMedList)

    def fillChoiceOfBoilerType(self, entry):
        boilerTypeList = TRANSBBTYPES.values()
        boilerTypeList.sort()
##        self.addNoneFront(boilerTypeList)
        fillChoice(entry, boilerTypeList)

    def fillChoiceOfHPSubType(self, entry, HPType=None):
        hpSubTypeList = self.getHPSubTypeList(HPType=HPType)
##        self.addNoneFront(hpSubTypeList)
        fillChoice(entry, hpSubTypeList)


    def fillChoiceOfCHPSubType(self, entry):
        chpSubTypeList = self.getCHPSubTypeList()
##        self.addNoneFront(chpSubTypeList)
        fillChoice(entry, chpSubTypeList)




class ModuleDBEntry(object):
    """ auxiliary module for entryPanel
    """

    def __init__(self, prnt, panel, objectType=None):
        """
        prnt:       parent module = ModuleDB
        panel:      link to panel for feed-back loops
        objectType: (optional) has to be set if object-specific functions are to be used
                    this is the case for HP
        """
        self.panel = panel
        self.prnt = prnt

        self.objectType = objectType
        
    def initPanel(self):
        """  functions that have to be carried out only once
        when the panel is created
        """
        # dictionary for dynamic translations
        self.dynamicTrans = {}
        self.dynamicTrans["Fluid"]  = Status.prj.getFluidDict(includeZeroWater=True)

        unitOpDict = Status.prj.getUnitOpDict(code=True)
        unitOpItems = [(code, text) for code, text in unitOpDict.items()]
        sortedItems =  sorted(unitOpItems, key=lambda item: item[0])
        
        unitOpList = [text for (code, text) in sortedItems]
        unitOpCodes = [code for (code, text) in sortedItems]
        self.dynamicTrans["UnitOp"] = dict([(code, i) for i, code in enumerate(unitOpCodes)])

        naceCodeList = self.prnt.getNACECodeandNACESubCodeList(addDescription=False)
        naceCodeListWithDescription = self.prnt.getNACECodeandNACESubCodeList()

        self.prnt.dataKeys["NACECode"] = (None, dict([(code, i) for i, code in enumerate(naceCodeList)]))
        self.prnt.dataKeys["ProductCode"] = (None, dict([(code, i) for i, code in enumerate(PRODUCTCODES)]))
        
        return {"ChoicesUnitOp": unitOpList,
                "ChoicesNACECode": naceCodeListWithDescription,
                "ChoicesProductCode": [TRANSPRODUCTCODES[code] for code in PRODUCTCODES]}
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        data = self.prnt.getObjectData()

        if self.objectType=="HP":
            subTypeChoices = self.typeChange(data["HPType"])
            data.update({"ChoicesHPSubType": subTypeChoices})

        elif self.objectType=="ST":

            fluid = Fluid(0)    # use default water for convert volume to mass flows
            funit2 = fluid.rho
            data.update({"FUnit2STMassFlow": funit2})
            
            row = [_("Turn key price for solar system per unit area"),
                       ("float", "UNITPRICE", 2),
                       data["STUnitTurnKeyPrice30kW"],
                       data["STUnitTurnKeyPrice300kW"],
                       data["STUnitTurnKeyPrice3000kW"]]
            data.update({"TurnKeyPriceTable": {"data": [row]}})

        elif self.objectType=="BM":
            colKeys = ["_MIN", "_TARG", "_MAX"]
            rowKeys = [("E", _("Electricity")), ("H", _("Fuel")) , ("T", _("Total"))]

            tips = {"EnergyInt":_("Energetic intensity (energetic consumption  with respect to an economic value: (a) expressed as a production cost and (b) expressed as a turnover"),
                    "SEC": _("Specific energetic consumption per pruduct unit (generic ratios) or processed medium unit (specific ratios for determined unitary operation)"),
                    "Unit": _("Measuring unit for the quantity of product or the quantity of processed medium")}         

            table = []
            toolTips = []
            table.append([_("Energy intensity (production cost)"), None, ("text", None, None)] + [""]*3)
            toolTips.append([tips["EnergyInt"]]*6)
            for key, title in rowKeys:
                table.append([title, None, ("float", "ENERGYPERCU", 2)] + [data[key+"_EnergyInt"+colKey+"_PC"] for colKey in colKeys])
                toolTips.append([tips["EnergyInt"]]*6)
            
            table.append([_("Energy intensity (turnover)"), None, ("text", None, None)] + [""]*3)
            toolTips.append([tips["EnergyInt"]]*6)
            for key, title in rowKeys:
                table.append([title, None, ("float", "ENERGYPERCU", 2)] + [data[key+"_EnergyInt"+colKey+"_T"] for colKey in colKeys])
                toolTips.append([tips["EnergyInt"]]*6)
            
            table.append([_("Specific energy consumption"), None, ("text", None, None)] + [""]*3)
            toolTips.append([tips["SEC"]]*6)
            for key, title in rowKeys:
                pu = data[key+"_Unit"]
                if pu is not None: pu = pu.strip("kWh/")
                table.append([title, pu, ("float", "ENERGYPERPU", 2)] + [data[key+"_SEC"+colKey] for colKey in colKeys])
                toolTips.append([tips["SEC"]]*6)
            
            data.update({"DataTable": {"data": table, "toolTips": toolTips}})
                       
        return data

    def saveToDB(self, data):
        self.prnt.setObjectData(data)

        # special action for some panels (ST, BM): read out data table into the object attributes
        obj = self.prnt.activeObject
        if self.objectType=="ST":
            row = data["TurnKeyPriceTable"][0]
            for c, key in enumerate(["STUnitTurnKeyPrice30kW", "STUnitTurnKeyPrice300kW", "STUnitTurnKeyPrice3000kW"]):
                setattr(obj, key, row[c+2])

        elif self.objectType=="BM":    
            table = data["DataTable"]

            colKeys = ["_MIN", "_TARG", "_MAX"]
            rowKeys = [("E", _("Electricity")), ("H", _("Fuel")) , ("T", _("Total"))]
            
            for i, rowKey in enumerate(rowKeys):
                key, title = rowKey
                for j, colKey in enumerate(colKeys):
                    parCodePC  = key+"_EnergyInt"+colKey+"_PC"
                    setattr(obj, parCodePC, table[1+i][3+j])
                            
                    parCodeT   = key+"_EnergyInt"+colKey+"_T"
                    setattr(obj, parCodeT, table[5+i][3+j])

                    parCodeSEC = key+"_SEC"+colKey
                    setattr(obj, parCodeSEC, table[9+i][3+j])
                    
                parCodePU  = key+"_Unit"
                setattr(obj, parCodePU, table[9+i][1])          
            
        if self.prnt.newObject: self.prnt.addObject()
        if self.panel.getUserChange():
            self.prnt.panel.unsavedChanges = True
            if self.prnt.activeObject not in self.prnt.changedObjects:
                self.prnt.changedObjects.append(self.prnt.activeObject)

    def typeChange(self, iType):
        subTypeChoices = []

        if iType is not None:
            if self.objectType=="HP":
                HPType = HPTYPES[iType]
                subTypes = HPSUBTYPES[HPType]
                subTypeChoices = [TRANSHPSUBTYPES[stype] for stype in subTypes]

        return subTypeChoices
        
    def cancel(self):
        self.prnt.newObject = False
        self.prnt.activeObject = None
        
