# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   moduleA
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel A (alternative proposal table)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#   Created by:     Hans Schweiger
#
#------------------------------------------------------------------------------     
#   (C) copyleft energyXperts.BCN (E4-Experts SL), Barcelona, Spain 2014
#   www.energyxperts.net / info@energyxperts.net
#
#   This program is free software: you can redistribute it or modify it under
#   the terms of the GNU general public license as published by the Free
#   Software Foundation (www.gnu.org).
#
#============================================================================== 
from GUI.status import Status
from globals import logger, Globals
from auxiliary.auxiliary import noneCheck
import database.dbio as dbio
import numpy as np

from lib.auxiliary.float import toFloat

class ModuleA(object):
    """

    Attributes:

    
    """
    def __init__(self, panel):
        self.keys = ["A Table"]
        self.panel = panel
        
        self.activeProject = None
        self.projectData = []

        self.selection = []

        self.initPanel()

        
    def initPanel(self):
        """  functions that have to be carried out only once
        when the panel is created
        """
        self.updatePanel()
    
#------------------------------------------------------------------------------
    def updatePanel(self):
#------------------------------------------------------------------------------
#       Here all the information should be prepared so that it can be plotted on the panel
#------------------------------------------------------------------------------

        self.table, self.selection = self.getTable()

        energy = np.nan_to_num(np.array([row[4] for i, row in enumerate(self.table[1:]) if "E€" in row[3] or i==0], dtype=float))/1000.0
        cost = np.nan_to_num(np.array([row[5] for i, row in enumerate(self.table[1:]) if "E€" in row[3] or i==0], dtype=float))/1000.0
        labels = [row[1] for i, row in enumerate(self.table[1:]) if "E€" in row[3] or i==0]

        #TODO: this is a layout issue and should be moved from here to the panel or some global settings !!!
        from GUI.GUITools import ORANGECASCADE, BLACK, MARKERTYPES
        colours = [BLACK] + ORANGECASCADE
        while len(colours) < len(cost):
            colours += ORANGECASCADE

        lineTypes = [MARKERTYPES[0]] + [MARKERTYPES[i%len(MARKERTYPES)] for i, entry in enumerate(cost)]
 
        eMin = np.min(energy)
        eMax = np.max(energy)

        cMin = np.min(cost)
        cMax = np.max(cost)

        n = len(cost)
        
##        energy = np.linspace(2.0, 5.0, 14)
##        cost = np.linspace(1.5, 4.5, 14)
        data = {"Table": {"data": self.table, "selection": self.selection},
                "Plot": {"xData": list(cost.reshape((n,1))),
                         "yData": list(energy.reshape((n,1))),
                         "labels": labels,
                         "lineTypes": lineTypes,
                         "colours": colours}}

        return data

    def changeData(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False
        nDataSets = len(self.table)

        if index > nDataSets or index < -1: # invalid index
            return 1, None, None
        elif index <=1 and mode not in [0]: # present state cannot be modified, deleted or selected
            if mode==3 and index in [0, 1]:
                logger.showWarning("Cannot delete the present state")
                
            return 1, None, None

        baseANo = max(index - 1, 0)
        shortName = self.table[index][1] if index >= 0 else None

        if mode == 0:     # new
            shortName = _("New Proposal %s")%(Status.NoOfAlternatives+1)
            shortName = shortName[-45:]
            description = _("based on present state")
            
            Status.prj.createNewAlternative(0, shortName, description)
            Status.int.clearHXData()
            
        elif mode == 1:   # copy
            shortName += _(' (mod.)')
            description = self.table[index][2] + _('(modified alternative based on ')+ self.table[index][1] + ')'
            shortName = shortName[-45:]

            Status.prj.createNewAlternative(baseANo, shortName, description)
            Status.int.clearHXData()

        elif mode == 2:   # select
            Status.prj.setFinalAlternative(baseANo)
            logger.showMessage(_("Alternative %s: %s selected as final proposal") % (baseANo, Status.FinalAlternativeName))
            Status.int.clearHXData()

        elif mode == 3:   # delete
            ret =  logger.askConfirmation(_("do you really want to delete alternative no. %s (%s) ?")%(baseANo, shortName))
            if ret:
                if baseANo > 0:
                    Status.prj.deleteAlternative(baseANo)
                    Status.int.clearHXData()

        elif mode == -1:  # data changed
            row = index
            newName = data[1]
            newDescription = data[2]
            Status.prj.setAlternativeDescription(newName, newDescription)

        data, selection = self.getTable()
        self.table = data
        choices = None

        Globals.GUI.panelInfo.update()
        Globals.GUI.showMainMenuAlternatives()        

        return 0, data, choices, selection

    def getTable(self):
        """
        Gets the project table to be displayed
        """

        alternativeList = Status.prj.getAlternativeList()            
        Status.int.setGraphicsData(self.keys[0], np.array(alternativeList))

        table = []
        for alternative in alternativeList:
            row = alternative[:]
            row[4] = toFloat(row[4])
            row[5] = toFloat(row[5])
            if row[0]==-1: row[0] = "-"
                    
            table.append(row)

        plotdata = np.array(table)
        Status.int.setGraphicsData("A Plot", plotdata)

        tableReport = []
        for i in range(10):
            if i < (len(alternativeList)-2):
                tableReport.append(alternativeList[i+2][1:3])
            else:
                tableReport.append([" "," "])
        dataReport = np.array(tableReport)
        Status.int.setGraphicsData("A_REPORT", dataReport)

        self.selection = [Status.ANo + 1]

        return table, self.selection

##    def readLine(self, data):
##        """ reads a line of raw data from the table in the GUI
##        and translates it into internal code
##        (inverse of getTable)
##        """
##        pass


    def setSelection(self, row):
        ANo = row - 1
        if ANo==Status.ANo: return False # nothing to do

        Status.prj.setActiveAlternative(ANo)
        
        if Status.ANo == -1 and Status.StatusCC != 0:
            ret = logger.askConfirmation(_("Should in addition to switching to the original state it also be unblocked?\n")+
                                            _("The unblocking would give you the possibility to modify it again, but ")+
                                            _("it deletes _all_ alternative proposals."), alwaysAsk=True)
            if ret:
                Status.prj.copyQuestionnaire()

        return True




                        
        

if __name__ == "__main__":
    print "Hello World";
