# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   modulePRJ
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel PRJ (project table)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 2.5
#   Created by:     Hans Schweiger
#
#   Update No. 001
#
#                   
#------------------------------------------------------------------------------     
#   (C) copyleft energyXperts.BCN (E4-Experts SL), Barcelona, Spain 2014
#   www.energyxperts.net / info@energyxperts.net
#
#   This program is free software: you can redistribute it or modify it under
#   the terms of the GNU general public license as published by the Free
#   Software Foundation (www.gnu.org).
#
#============================================================================== 
from GUI.status import Status
from globals import logger, Globals
from GUI.dialogPRJ import DialogPRJ
from auxiliary.auxiliary import noneCheck
import database.dbio as dbio

class ModulePRJ(object):
    """

    Attributes:

    
    """
    def __init__(self, panel):
        self.panel = panel
        
        self.activeProject = None
        self.projectData = []

        self.selection = []

        self.initPanel()

        
    def initPanel(self):
        """  functions that have to be carried out only once
        when the panel is created
        """
        self.updatePanel()
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """
        pass
        # create system model -> contains list of heat exchangers
        # to do: shift to module !!!
        logger.logTrack("ModulePRJ.updatePanel")

        self.getTable()
        projectIds = [row[0] for row in self.projectData]
        if Status.PId in projectIds:
            self.pId = Status.PId
            self.selection = [projectIds.index(Status.PId)]
        else:
            self.selection = []

    def changeData(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False
        nDataSets = len(self.projectData)

        if index > nDataSets or index < -1:
            return 1, None, None
        elif index==-1 and mode not in [1, 4]:
            return 1, None, None

##        po = self.schedule.periodIntervals[index] if (index < nPeriods and index >=0) else ProcessPeriod()

        if index < nDataSets:
            pId = self.projectData[index][0]
        else:
            pId = None

        if mode == 0:     # copy
            self.activeProject = index
            self.copyProject(pId)
            
        elif mode in [1]:   # delete
            self.deleteProject(pId)

        self.activeProject = index-1 if index > 0 else (0 if len(self.projectData) > 0 else None)
            

        data, selection = self.getTable()
        choices = None
        return 0, data, choices, selection

    def getTable(self):
        """
        Gets the project table to be displayed
        """
        table = []

        sproject = dbio.select("sproject")

        self.projectData = []
        for s in sproject:
            q = dbio.get("questionnaire", id=s.ProjectID)

            if q is not None:
                self.projectData.append([s.ProjectID, s.ProjectName, q.DescripIndustry, q.Branch])
            else:
                print ("Corrupt DB entry for ProjectID %s: no corresponding entry in table questionnaire."%s.ProjectID)
                
        self.projectData = sorted(self.projectData, key=lambda row: row[1]) 

        table = [[str(noneCheck(elem, "")) for elem in row[1:]] for row in self.projectData]
            
        return table, self.selection

    def setSelection(self, selection):
        self.selection = selection
        
    def openProject(self, r):
        """ opens the project corresponding to row r in the table
        """
        pId = self.projectData[r][0]
        Status.prj.setActiveProject(pId)

        Globals.GUI.panelInfo.update()
        Globals.GUI.showMainMenuAlternatives()
        Globals.GUI.updateDrawings()
        
    def copyProject(self, pId):
        """ opens the project corresponding to row r in the table
        """
        r = self.getIndex(pId)
        if r is None: return

        projectName = noneCheck(self.projectData[r][1], _("no name"))
        description = noneCheck(self.projectData[r][2], "")

        logger.logTrack("PanelPRJ: Copying project %s"%projectName)

        shortName = projectName +_("(copy)")
        shortName = shortName[-80:]
        
        description += "\n" + _("new project copied from ")+projectName
        description = description[-200:]

       
        self.shortName = shortName
        self.description = description
        pu1 =  DialogPRJ(Globals.GUI, self)
        
        import wx
        if pu1.ShowModal() == wx.ID_OK:
            logger.logTrack("PanelP - OK %r %r"%(self.shortName, self.description))

            Status.prj.createNewProject(pId, self.shortName, self.description)
            self.panel.display()
            Globals.GUI.panelInfo.update()
            Globals.GUI.showMainMenuAlternatives()
            Globals.GUI.updateDrawings()
        
    def deleteProject(self, pId):
        """ opens the project corresponding to row r in the table
        """

        r = self.getIndex(pId)
        if r is None: return
        projectName = self.projectData[r][1]
        
        if logger.askConfirmation(_('Delete project %s ?') % (projectName,), alwaysAsk=True):
            logger.logTrack("PanelPRJ (delete): deleting %s"%projectName)

            Status.prj.deleteProject(pId)
            self.panel.display()
            Globals.GUI.panelInfo.update()
            Globals.GUI.showMainMenuAlternatives()
            Globals.GUI.updateDrawings()

    def getIndex(self, pId):
        """
        Helper function: gets the index of the project in the project table
        from the ProjectID
        """
        projectIds = [row[0] for row in self.projectData]
        if pId in projectIds:
            return projectIds.index(pId)
        else:
            return None

    def readLine(self, data):
        """ reads a line of raw data from the table in the GUI
        and translates it into internal code
        (inverse of getTable)
        """
        pass


    def saveToDB(self, data=None):
        """
        called from OK button in PRJ-panel
        -> same action as on Dclick
        """
        if len(self.selection)==1:
            r = self.selection[0]
            self.openProject(r)
            return True
        else:
            return False




                        
        

if __name__ == "__main__":
    print "Hello World";
