# -*- coding: utf-8 -*-
#==============================================================================
#
#      E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#      ModuleScheduleEditor
#            
#------------------------------------------------------------------------------
#            
#      Data table for data entry
#
#==============================================================================
#       (C) energyXperts.NET (E4-Experts GmbH, Berlin, Germany)
#============================================================================== 

from globals import logger

from datetime import date, timedelta
from modules.schedules.profiles import WeeklyProfile, getAllProfilesFromDB
from moduleScheduleViewer import periodView, scheduleView
from lib.dateTime.dateTimeTools import timeToString, stringToTime
from modules.messageLogger import logError
from moduleTableEntry import ModuleTableEntry
from constants import WEEKDAYS

def readInterval(interval, data):
    try:
        start = stringToTime(data[0])
        stop = stringToTime(data[1])
        scale = float(data[2])
        return (start, stop, scale)
    except:
        logError("Your data entry is not valid")
        return None

def defaultInterval():
    return (stringToTime("00:00"), stringToTime("24:00"), 100.0)

def insertInterval(upper, lower):
    if upper is not None and lower is not None:
        new = (upper[1], lower[0], 0.5*(upper[2]+lower[2]))
    elif upper is None and lower is None:
        new = defaultInterval()
    elif upper is not None:
        new = (upper[1], upper[1], upper[2])
    elif lower is not None:
        new = (lower[0], lower[0], lower[2])
    return new

def splitInterval(base):
    if base is None: base = defaultInterval()
    upper = (base[0], base[0], base[2])
    lower = base
    return upper, lower


    
    
#==============================================================================
class ModuleProfileSelector():
#==============================================================================
#==============================================================================

    def __init__(self, viewer, index=None):
        """
        :titles     list of titles for each column (-> defines the number of columns)
        """

        self.viewer = viewer

        self.getProfilesFromDB()
        self.profiles = self.dbProfiles[:]  # copy into internal list
        self.deletedProfiles = []
        
        self.getProfileTable()
        self.activeProfile = None if index is None else self.dbProfiles[index]
        
        self.defaultPlotRange = (0, 24) # default interval (in hours) to be plotted in ScheduleView)

        self.moduleIntervalTable = ModuleTableEntry(insert=insertInterval, split=splitInterval, read=readInterval)
        if self.activeProfile is not None: self.moduleIntervalTable.setTable(self.activeProfile.intervals)

##        self.moduleWeekDayTable = ModuleTableEntry()
##        self.fweek = [0]*7
##        self.moduleWeekDayTable.setTable(self.fweek)


    def getProfilesFromDB(self):
        """
        Loads the profiles currently stored within the DB
        """
        
        self.dbProfiles = sorted(getAllProfilesFromDB(), key = lambda profile: profile.name)
        self.dbProfileIDs = [profile.id for profile in self.dbProfiles]
        self.dbNames = [profile.name for profile in self.dbProfiles]
            

    def changeDataIntervalTable(self, index, mode, data=None):
        """
        modifies the profile interval table and calculates
        default data when adding / deleting a new line
        """

        err, activeItem = self.moduleIntervalTable.changeData(index, mode, data)

        data = self.getActiveProfileData()
        self.viewer.dialog.display()
        
        return 0, data["interval"]

    def changeDataWeekDayTable(self, index, mode, data=None):

        changedRow = data
        activeItem = index

        if activeItem >=0 and activeItem < 7:
            val = (changedRow[1]==_("on"))
            self.activeProfile.weekday[activeItem] = val

        data = self.getActiveProfileData()
        self.viewer.dialog.display()

        return 0, data["weekDay"]


        
#-----------------------------------------------------------------------------------------
# auxiliary functions creating information for representation on the GUI
#-----------------------------------------------------------------------------------------

    def getActiveProfileData(self):
        """ transforms the information contained in the profiles into a table of strings for
            representation on the GUI
        """
        intervalTable = []
        for interval in self.activeProfile.intervals:
            if interval is None:
                intervalTable.append(["00:00", "24:00", "100.00"])
            else:
                start, stop, scale = interval
                startString = timeToString(start)
                stopString = timeToString(stop)
                scaleString = "%6.2f"%(scale)

                intervalTable.append([startString, stopString, scaleString])

        weekDayTable = [[day, _("on") if self.activeProfile.weekday[d] else _("off")] for d, day in enumerate(WEEKDAYS)]

        profile = self.activeProfile
        profile.calculateGlobals()
        name = profile.name
        daysPerWeek = "%1d"%profile.nDays
        hPerDay = "%6.2f"%profile.hPerDay
        hPerDayNom = "%6.2f"%(profile.hPerDayNom)
        fAvg = "%6.4f"%profile.fAvg
        fMinfMax = "%6.4f/%6.4f"%(profile.fMin, profile.fMax)
        TPeriod = "%4d"%profile.TPeriod

        summaryTable = [[_("days per\nweek/period"), "-", daysPerWeek],
                        [_("operation hours per day"), "h", hPerDay],
                        [_("avg. ratio"), "-", fAvg],
                        [_("nominal operation hours per day"), "h", hPerDayNom]]

        data = {"name": profile.name,
                "interval": intervalTable,
                "weekDay": weekDayTable,
                "summary": summaryTable}
        return data


    def saveProfile(self, name):
        if self.activeProfile is not None:
            oldName = self.activeProfile.name

            if name != self.activeProfile.name:
                if name in self.dbNames:
                    logger.showWarning(_("The profile name %s already exists. Please enter a different one"%name))
                    return False
                else:
                    self.activeProfile.name = name
            ret = self.activeProfile.saveToDB()
            if not ret:
                self.activeProfile.name = oldName
            
            return True

        return False
            
    def getScheduleView(self):
        start, stop = self.defaultPlotRange
        SVf, SVt = scheduleView([self.activeProfile.intervals], start, stop)
        return [x/100.0 for x in SVf], SVt
        

    def getProfileTable(self):
        """
        function called from panel
        """
        
        profileTable = []
        for profile in self.profiles:
            name = profile.name
            daysPerWeek = "%1d"%profile.nDays
            hPerDay = "%6.2f"%profile.hPerDay
            fAvg = "%6.4f"%profile.fAvg
            fMinfMax = "%6.4f/%6.4f"%(profile.fMin, profile.fMax)
            TPeriod = "%4d"%profile.TPeriod

            profileTable.append([name, daysPerWeek, hPerDay, fAvg, fMinfMax, TPeriod])

        return profileTable

    def editProfile(self, index):
        """
        selects a profile for editing
        """
        if index is not None and index < len(self.profiles):
            self.activeProfile = self.profiles[index]
            self.moduleIntervalTable.setTable(self.activeProfile.intervals)
            
    def newProfile(self, index=None):
        """
        creates a new (local) profile
        has to be added later on to the profiles list (with addProfile)
        """
        if index is None:
            new = WeeklyProfile(_("new profile"))
        else:
            new = self.profiles[index].copy()
            
##        self.profiles.append(new)
        self.activeProfile = new
        self.moduleIntervalTable.setTable(self.activeProfile.intervals)
        return None

    def addProfile(self):
        """ adds the active profile to the profiles list
        """
        if self.activeProfile not in self.profiles:
            self.profiles.append(self.activeProfile)
            
    def deleteProfile(self, index):
        deleted = self.profiles[index] 
        ret = logger.askConfirmation(_("Do you want to delete the profile %s ?"%deleted.name))
        if ret:
            if deleted.deleteFromDB():
                self.profiles.pop(index)
        
###-----------------------------------------------------------------------------------------
### exectution of GUI actions
###-----------------------------------------------------------------------------------------
##
##    def selectProfiles(self, selection):
##        """ called from Profile-Selector Pop-Up
##        Stores the values selected on the GUI
##        """
##        # first eliminate non-existing indices from the GUI selection
##        checked = [index for index in selection if (index < len(self.dbProfiles) and index >= 0)]
####        print "TEST TEST ModuleScheduleEditor.selectProfiles: selection = %s\nchecked = %s"%(selection, checked)
##
##        self.schedule.periodIntervals[self.activePeriod].profiles = [self.dbProfiles[index] for index in checked]
##        self.schedule.build()
##


