# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   moduleQ3
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel Q3 (process data)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#                   
#------------------------------------------------------------------------------     
#============================================================================== 
from globals import Globals, logger
from constants import EPSILON
from GUI.status import Status
from GUI.conflictFrame import conflictFrame

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck, findKey
from auxiliary.dataTransfer import DataTransfer
from lib.auxiliary.float import toFloat, toInt

from modules.fluids import Fluid
from modules.schedules.schedules import DEFAULTCHARGETIME

# keys for parameters which are 1:1 passed to data
# structure: key_in_procObject: (key_in_Panel (or None, if identical), dictionary)

DATAKEYS = {"name": (None, None),
           "Description": (None, None),
           "ProcType": (None, None),
           "UnitOp": (None, "UnitOp"),
           "heating": (None, None),
           "PT": (None, None),
           "DPSTmin": (None, None),
           "DPSTminMCD": (None, None),
           "ExternalSupply": (None, None),
           "UPHs": (None, None),
           "UPHm": (None, None),
           "HPerDay": (None, None),
           "NCycles": (None, None),
           "HCycle": (None, None),
           "NDays": (None, None),
           "HPerYear": (None, None),
           "HPerYearInFlow": (None, None),
           "HPerYearOutFlow": (None, None),
           "scheduleType": (None, None)
           }        

from model.objects.processStreams import KEYS_INFLOW, KEYS_OUTFLOW

DATAKEYS_INSTREAM = {}
for key in KEYS_INFLOW:
    DATAKEYS_INSTREAM.update({key: (None, None)})

DATAKEYS_OUTSTREAM = {}
for key in KEYS_OUTFLOW:
    DATAKEYS_OUTSTREAM.update({key: (None, None)})

DATAKEYS_STARTUP = {"VolProcMed": (None, None),
                    "ThermalMass": (None, None)}


class ModuleQ3(object):
    """

    Attributes:

    
    """
    def __init__(self, panel):
        self.panel = panel
        
        self.hasMaintenance = 0         # flags (0 or 1) indicating whether start-up or maintenance are displayed in component table or not
        self.hasStartUp = 0

        # number of rows in the display table
        self.nRowsDemand = 0                    
        self.nRowsResidual = 0

        self.modTSEditor = None            # to be set from the panel

        self.deletedSchedules = []        # buffer for schedules to be deleted from DB

        self.ccProc = None

        self.new = False
        
    def initPanel(self, no):
        """  functions that have to be carried out only once
        when the panel is created
        """
        # dictionary for dynamic translations
        self.dynamicTrans = {}
        self.dynamicTrans["Fluid"]  = Status.prj.getFluidDict(includeZeroWater=True)

        unitOpDict = Status.prj.getUnitOpDict(code=True)
        unitOpItems = [(code, text) for code, text in unitOpDict.items()]
        sortedItems =  sorted(unitOpItems, key=lambda item: item[0])
        
        unitOpList = [text for (code, text) in sortedItems]
        unitOpCodes = [code for (code, text) in sortedItems]
        self.dynamicTrans["UnitOp"] = dict([(code, i) for i, code in enumerate(unitOpCodes)])

        if no >= 0:
            self.proc = Status.int.systemModel.processObject[no]
        else:
            self.proc = Status.int.systemModel.addObject("PROC")
            self.new = True
        return {"unitOps": unitOpList,
                "BUI": (self.proc.buildingId is not None)}
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        if self.proc is None: return

        data = self.getProcessDisplayData()            

        demandTable, demandChoices = self.getUPXTable(data)
        self.nRowsDemand = len(demandTable)

        # pointers give a reference to lines in table where data can be found
        pointers = {"maintenance": 0 if self.hasMaintenance else None,
                    "startUp": 1*self.hasMaintenance if self.hasStartUp else None,
                    "inFlow": 1*self.hasMaintenance + 1*self.hasStartUp,
                    "total": self.nRowsDemand-1
                    }
        data["UPXTable"] = {"data": demandTable, "choices": demandChoices, "pointers": self.getPointers()}

        residualTable, residualChoices = self.getQWXTable(data)
        self.nRowsResidual = len(residualTable)
        data["QWXTable"] = {"data": residualTable, "choices": residualChoices}
        data["anyMCD"] = self.anyMCD
        
        data["funit2"] = self.getSecondaryUnits()
        
        # try to load detailed schedule
        self.modTSEditor.load(self.proc.id, self.proc.name)
        self.periodSchedule = self.modTSEditor.schedule
                
        detailedSchedule = (self.periodSchedule is not None)
        
        if detailedSchedule:
            data["DTStartUp"] = self.periodSchedule.startup
            data["DTInFlow"]  = self.periodSchedule.inflow
            data["DTOutFlow"] = self.periodSchedule.outflow

        # export parameters for PanelCC
        k = self.proc.no
        dataGroup = ("Process[" + str(k+1) + "]")
        data["cc"] = (self.ccProc, dataGroup)

        return data
        
    def selectProcess(self):
        
        k = self.selectedProcessNo
        proc = Status.int.systemModel.process[k]

        processData = {"Questionnaire_id":Status.PId,
                       "AlternativeProposalNo":Status.ANo}

    def saveToDB(self, data={}):
        """ saves the data received from the panel to the data base
        """
        if not Status.prj.editAllowed():
            return False

        proc = self.proc
##        if proc.buildingId is not None:
##            logger.showWarning(_("This process belongs to a building.\n"\
##                                 "Its properties can be edited in the 'Buildings' panel."))
##            return
        
        # collects here all the parameters from the panel (exclunding that in tables) which can be stored directly or via translation dictionary
        # note: parameters displayed in tables have already been updated in proc directly when changed

        for key in DATAKEYS.keys():
            if key in data.keys():
                val = data[key]

                trans = DATAKEYS[key][1]
                trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
                if trans is None:
                    setattr(proc, key, dbio.check(val))
                else:
                    setattr(proc, key, dbio.check(findKey(trans, val)))

        # read schedule parameters from panel        
        if self.periodSchedule:
            for key in ["DTStartUp", "DTInFlow", "DTOutFlow"]:
                if hasattr(data, key):
                    setattr(self.periodSchedule, key, toFloat(data[key]))

            self.proc.HPerYearInFlow = self.periodSchedule.HPerYearInflow
            self.proc.HPerYearOutFlow = self.periodSchedule.HPerYearOutflow
 
            err = self.periodSchedule.check()
            if err:
                logError(err)
                return False
        else:
            self.proc.PartLoad = 1.0 if self.proc.scheduleType==0 else None

            try:
                self.proc.HPerYearInFlow = self.proc.HCycle * self.proc.NCycles * self.proc.NDays
                if self.proc.ProcType == 1:
                    self.proc.HPerYearInFlow *= DEFAULTCHARGETIME
            except TypeError: # one of the operands is None == NULL -> Store NULL
                self.proc.HPerYearInFlow = None
            self.proc.HPerYearOutFlow = self.proc.HPerYearInFlow

            if hasattr(self.proc, "schedule"):
                self.proc.schedule.HBatch = self.proc.HCycle
                self.proc.schedule.NBatch = toInt(self.proc.NCycles)
                self.proc.schedule.NDaysProc = toInt(self.proc.NDays)
                self.proc.schedule.setDefault()

        k = proc.no
        conflicts = Status.mod.moduleCC.procCheck(proc)
        Status.mod.moduleCC.initPanel()
        if conflicts > 0:
            cf = conflictFrame(Globals.GUI.editPanel)
            cf.Show()
            return False

        # auto complete some mandatory data
        if proc.DPSTmin is None:
            proc.DPSTmin = 10.0 if proc.heating else 5.0

        if proc.DPSTminMCD is None and self.anyMCD:
            proc.DPSTminMCD = 5.0

        if Status.ANo >= 0 and not proc.checkCompleteness():
            return False
        

        proc.saveToDB()

        #TODO: schedule should be an attribute of processObject, not of moduleQ3. Saving should be included in proc.saveToDB.
        if self.periodSchedule:
            self.periodSchedule.saveToDB(self.proc.id, withHolidays=True, saveAll=True)
            
        for schedule in self.deletedSchedules:
            schedule.deleteFromDB()

        Status.prj.projectChanged()
        Status.int.changeInCascade(0, keepModel=True)   # indicate that energy calculations are out of date

        self.new = False
        return True


    def changeDataUPXTable(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False
        nRows = self.nRowsDemand

        # check validity of mode and index combination
        if (index > nRows or index < -1) and mode==3:
            return 1, None, None

        if mode == 0:     # add maintenance
            if self.hasMaintenance:
                logger.showWarning(_("Maintenance heat is already considered for this process"))
            else:
                self.proc.dotUPHm = None
                self.hasMaintenance = True
                
            self.activeDataSet = nRows
            
        elif mode == 1:   # add start-up
            if self.hasStartUp:
                logger.showWarning(_("Start-up heat demand is already considered for this process"))
            else:
                self.proc.UPHs = None
                self.proc.VolProcMed = None
                self.proc.ThermalMass = None
                self.proc.ProcMedDBFluid_id = None
                self.proc.PTStartUp = None
                self.hasStartUp = True
                
            nRows += 1

            self.activeDataSet = nRows
            
        elif mode == 2:    # add inflow
            self.proc.addInFlow()
            self.activeDataSet = nRows
            
        elif mode == 3:    # delete
            componentNo = self.getComponentNo(index)

            if componentNo==-1:
                self.proc.UPHs = 0.0
                self.proc.ThermalMass = 0.0
                self.proc.VolProcMed = None
                self.proc.ProcMedDBFluid_id = None
                self.proc.PTStartUp = None
                
            elif componentNo==0:
                self.proc.dotUPHm = 0.0
                
            else:
                self.proc.deleteInFlow(componentNo-1)
            
            self.activeDataSet = None

        elif mode == -1:   # change of data in a dataSet

            row = data
            componentNo = self.getComponentNo(index)            
                    
            newVals = self.readLineUPXTable(row, componentNo)
            if newVals:
                name, mediumId, mdot, T0, dotQ, Q = newVals

                try:
                    fluid = Fluid(mediumId)
                except:
                    fluid = None
                
                if componentNo > 0:
                    if componentNo <= len(self.proc.inStreams):
                        stream = self.proc.inStreams[componentNo-1]
                        stream.Name = name
                        stream.dbfluid_id = mediumId
                        stream.fluid = fluid
                        stream.PTInFlowRec = T0
                        #TO DO: check between PTInFlow and PTInFlowRec
                        stream.mInFlowNom = mdot
                        stream.dotUPHc = dotQ

                    else:
                        self.proc.UPH = Q

                elif componentNo == -1:
                    self.proc.ProcMedDBFluid_id = mediumId
                    self.proc.PTStartUp = T0
                    self.proc.fluid = fluid

                elif componentNo == 0:
                    self.proc.dotUPHm = dotQ
                 
                self.panel.setSecondaryUnits()
            else:
                logger.logError(_("Invalid data set. Please try again"))
                err = True
    
            self.activeDataSet = index

        self.panel.setPointers(self.getPointers())
                                   
        data = self.getProcessDisplayData()
        tableData, choices = self.getUPXTable(data)
        
        return 0, tableData, choices

    def getComponentNo(self, index):
        """
        helper function
        calculates the componentNo (as used in StreamGlbs) from the display table index and the flags for start-up and maintenance
        """
        if index < 0: return None # click to unit or label row
        
        s = index - self.hasStartUp - self.hasMaintenance # index of the stream. -1 or -2 for maintenance and start-up components
        if s==-2 or (s==-1 and not self.hasStartUp) is None: # maintenance data row
            componentNo = 0
        elif s==-1 and self.hasStartUp:
            componentNo = -1
        else: # edit stream data
            componentNo = s+1

        return componentNo

    def getUPXTable(self, data):
        """
        transforms the information contained in the process (received as data) into a table of strings for
        representation on the GUI
        """
        q = self.proc
        # collect data for demand table

        tableData = {}
        demandTable = []

        # maintenance

        tableData["dotUPHm"] = q.dotUPHm
        self.hasMaintenance = q.dotUPHm > EPSILON or q.dotUPHm is None
                
        demandTable.append([_("maintenance"), "-", "-", str(data["PT"]), 
                            tableData["dotUPHm"], str(data["UPHm"]), str(data["HPerYear"]), None])

        # start-up
        tableData["ProcMed"] = self.dynamicTrans["Fluid"].get(q.ProcMedDBFluid_id)
        tableData["VolProcMed"] = str(q.VolProcMed)

        try:
            q.fluid = Fluid(q.ProcMedDBFluid_id)
        except:
            q.fluid = None

        tableData["ThermalMass"] = q.ThermalMass

        tableData["PTStartUp"] = str(q.PTStartUp)
        additionalInfo = _("view/edit") if (q.VolProcMed is not None) else _("add")

        demandTable.append([_("start-up"), tableData["ProcMed"], "-", tableData["PTStartUp"], 
                            "?", str(data["UPHs"]), "?", additionalInfo])

        self.hasStartUp = data["UPHs"] > EPSILON or \
                          (data["UPHs"] is None and (q.ThermalMass > EPSILON or q.ThermalMass is None))
                          

        # in-flows
        inStreams = self.proc.inStreams

##            if q.HeatRecExist is not None and q.HeatRecExist.lower() in TRANSYESNO:

        for stream in inStreams:
            T0_display = stream.PTInFlowRec
            additionalInfo = _("view/edit") if (stream.HeatRecExist != False or stream.VInFlowCycle is not None) else _("add")
            fluidName =  self.dynamicTrans["Fluid"].get(stream.dbfluid_id)
            try:
                stream.fluid = Fluid(stream.dbfluid_id)
            except:
                stream.fluid = None
            
            demandTable.append([stream.Name,
                                fluidName,
                                str(stream.mInFlowNom),
                                str(T0_display),
                                str(stream.dotUPHc),
                                str(stream.UPHc),
                                str(q.HPerYearInFlow),
                                additionalInfo])

        # total
        tableData["UPH"] = str(q.UPH)
        demandTable.append([_("TOTAL") , None, None, None, None, tableData["UPH"], None, None])

        # filter table: eliminate maintenance and start-up if Q=0

        baseIndexInFlows = self.hasMaintenance + self.hasStartUp
        demandTableFiltered = ([demandTable[0]] if self.hasMaintenance else []) + \
                              ([demandTable[1]] if self.hasStartUp else []) + \
                              demandTable[2:]

        fluidList = self.dynamicTrans["Fluid"].values()
        fluidList.sort()
        choices = [[None, fluidList, None, None,
                     None, None, None, None] for row in range(len(demandTableFiltered))]
        return demandTableFiltered, choices


    def readLineUPXTable(self, row, componentNo):
        """ reads a line of raw data from the table in the GUI
        and translates it into internal code
        (inverse of getTable)
        """ 
        if len(row) > 8:
            return None
        
        name, medium, mdot, T0, dotQ, Q, hop, addInfo = row

        name = name.strip() if name is not None else None

        mediumId = findKey(self.dynamicTrans["Fluid"], medium)        
            
        mdot = toFloat(mdot)
        T0 =   toFloat(T0)
        dotQ = toFloat(dotQ)
        Q = toFloat(Q)
        
        return (name, mediumId, mdot, T0, dotQ, Q)


    def changeDataQWXTable(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """


        err = False
        nRows = self.nRowsResidual

        # check validity of mode and index combination
        if (index > nRows or index < -1) and mode==1:
            return 1, None, None

        if mode == 0:    # add outflow
            self.proc.addOutFlow()
            self.activeDataSet = nRows
            
        elif mode == 1:    # delete
            self.proc.deleteOutFlow(index)
            self.activeDataSet = None

        elif mode == -1:   # change of data in a dataSet

            row = data
            componentNo = index+1            
                    
            newVals = self.readLineQWXTable(row, componentNo)
            if newVals:
                name, mediumId, mdot, T0, T1, dotQ = newVals
                
                try:
                    fluid = Fluid(mediumId)
                except:
                    fluid = None

                if componentNo > 0:
                    stream = self.proc.outStreams[componentNo-1]
                    stream.Name = name
                    if stream.dbfluid_id != mediumId:   # clear enthalpies on fluid change in order to avoid conflicts
                        stream.HOutFlow = None
                        stream.XOutFlow = None
                    stream.dbfluid_id = mediumId
                    stream.fluid = fluid
                    stream.PTOutFlowRec = T0
                    stream.PTFinal = T1
                    stream.mOutFlowNom = mdot 

                    self.panel.setSecondaryUnits()

            else:
                logger.logError(_("Invalid data set. Please try again"))
                err = True
    
            self.activeDataSet = index

        data = self.getProcessDisplayData()
        tableData, choices = self.getQWXTable(data)
        self.nRowsResidual = len(tableData)
        
        return 0, tableData, choices

    def getQWXTable(self, data):
        """
        transforms the information contained in the process (received as data) into a table of strings for
        representation on the GUI
        """
        q = self.proc
        # collect data for demand table

        tableData = {}
        table = []
        # outflows
        outStreams = self.proc.outStreams

        self.anyMCD = False
        for stream in outStreams:
            self.anyMCD = self.anyMCD or stream.MCD # set True if any of the streams is true
            T0_display = stream.PTOutFlow if stream.PTOutFlowRec is None else stream.PTOutFlowRec
            
            additionalInfo = _("view/edit") if ((stream.HeatRecExist is not False) or \
                                                (stream.MCD is not False) or \
                                                (stream.VOutFlowCycle is not None) or \
                                                (stream.HOutFlow is not None) or \
                                                (stream.XOutFlow is not None)) else _("add")

            fluidName =  self.dynamicTrans["Fluid"].get(stream.dbfluid_id)
            try:
                stream.fluid = Fluid(stream.dbfluid_id)
            except:
                stream.fluid = None
            
            table.append([stream.Name,
                                fluidName,
                                str(stream.mOutFlowNom),
                                str(T0_display),
                                str(stream.PTFinal),
                                str(stream.dotUPHw),
                                str(stream.UPHw),
                                str(q.HPerYearOutFlow),
                                additionalInfo])

        fluidList = self.dynamicTrans["Fluid"].values()
        fluidList.sort()
        choices = [[None, fluidList, None, None,
                     None, None, None, None, None] for row in range(len(table))]

        return table, choices


    def readLineQWXTable(self, row, componentNo):
        """ reads a line of raw data from the table in the GUI
        and translates it into internal code
        (inverse of getTable)
        """ 
        if len(row) > 9:
            return None
        
        name, medium, mdot, T0, T1, dotQ, Q, hop, addInfo = row

        name = name.strip() if name is not None else None

        mediumId = findKey(self.dynamicTrans["Fluid"], medium)
        mdot = toFloat(mdot)
        T0 =   toFloat(T0)
        T1 =    toFloat(T1)
        dotQ = toFloat(dotQ)
        
       # read parameters name, medium, mdot (or dm) and T0
        if name is None or len(name.strip())==0:
            return None

        return (name, mediumId, mdot, T0, T1, dotQ)

    def getProcessDisplayData(self):
        """
        translates the data in self.proc into a data dicitionary as
        required by the panel
        """
        
        # read in all data from processObject translate where necessary
        q = self.proc
        
        data = {}
        for dataKey in DATAKEYS:
            val = getattr(q, dataKey)
            key = noneCheck(DATAKEYS[dataKey][0], dataKey)
            trans = DATAKEYS[dataKey][1]
            trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
            if trans is None:
                data[key] = val
            else:
                data[key] = trans.get(val)

        return data


    def getPointers(self):
        """
        Helper function: creats a list of indexes for the entries in the UPXTable
        """
        ptrTotal = 1*self.hasMaintenance + 1*self.hasStartUp + len(self.proc.inStreams)
        # pointers give a reference to lines in table where data can be found
        pointers = {"maintenance": 0 if self.hasMaintenance else None,
                    "startUp": 1*self.hasMaintenance if self.hasStartUp else None,
                    "inFlow": 1*self.hasMaintenance + 1*self.hasStartUp,
                    "total": ptrTotal
                    }
        return pointers

    def getSecondaryUnits(self):
        """ collects the conversion factors for secondary units depending on the fluids seleced
        """
        data = {"startUp": self.proc.fluid.rho if self.proc.fluid is not None else 1.0,
                "inFlow": [stream.fluid.rho  if stream.fluid is not None else 1.0 for stream in self.proc.inStreams],
                "outFlow": [stream.fluid.rho if stream.fluid is not None else 1.0 for stream in self.proc.outStreams]}

        return data
        
    def getOutStreamData(self, s):
        """ collects the data for the outstream to be plotted on the "additional info" panel
        """
        q = self.proc.outStreams[s]        
        data = {}
        for dataKey in DATAKEYS_OUTSTREAM:
            val = getattr(q, dataKey)
            key = noneCheck(DATAKEYS_OUTSTREAM[dataKey][0], dataKey)
            trans = DATAKEYS_OUTSTREAM[dataKey][1]
            trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
            if trans is None:
                data[key] = val
            else:
                data[key] = trans.get(val)

        return data

    def setOutStreamData(self, s):
        data = self.buffer
        q = self.proc.outStreams[s]        
        
        for key in DATAKEYS_OUTSTREAM.keys():
            if key in data.keys():
                val = data[key]

                trans = DATAKEYS_OUTSTREAM[key][1]
                trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
                if trans is None:
                    setattr(q, key, dbio.check(val))
                else:
                    setattr(q, key, dbio.check(findKey(trans, val)))


    def getMCD(self):
        return any([stream.MCD for stream in self.proc.outStreams])
        
 
    def getDemandData(self, cNo):
        """ collects the data for the outstream to be plotted on the "additional info" panel
        """

        if cNo > 0:
            s = cNo -1
            q = self.proc.inStreams[s]

            transfer = DataTransfer(self.dynamicTrans, DATAKEYS_INSTREAM)
            data = transfer.write(q)

        elif cNo == -1:
            q = self.proc   

            transfer = DataTransfer(self.dynamicTrans, DATAKEYS_STARTUP)
            data = transfer.write(q)

        return data

    def setDemandData(self, cNo):
        """
        translates data received from the GUI-PopUp (written by fetchData into self.buffer)
        into attributes of the corresponding object (inStream or process)
        """

        data = self.buffer
        if cNo > 0:
            s = cNo - 1
            q = self.proc.inStreams[s]
            
            transfer = DataTransfer(self.dynamicTrans, DATAKEYS_INSTREAM)
            transfer.read(data, q)
            
        elif cNo == -1:

            q = self.proc
            transfer = DataTransfer(self.dynamicTrans, DATAKEYS_STARTUP)
            transfer.read(data, q)
 

    def fetchData(self, data):
        """
        Helper function to be called from dialogs in panel
        Used for reading out panel data
        """
##        for key in DATAKEYS_READONLY:
##            if key in data.keys():
##                data.pop(key)
                
        self.buffer = data

    def setScheduleType(self, scheduleType):
        """
        To be called from the event handler in the GUI if a detailed schedule is activated
        """
        self.proc.scheduleType = scheduleType
        
        newData = None
        if scheduleType==0:
            if self.periodSchedule: self.deletedSchedules.append(self.periodSchedule)
            self.periodSchedule = None
            self.proc.autoFit = False
            
        elif scheduleType==1:
            self.proc.autoFit = False
        
            if self.periodSchedule is None: # no detailed schedule yet -> create new empty one
                
                self.modTSEditor.viewer.load(self.proc.id, self.proc.name, createNew=True)
                self.periodSchedule = self.modTSEditor.schedule

            newData = self.updateOperationTimes()
                
        elif scheduleType==2:
            if self.periodSchedule: self.deletedSchedules.append(self.periodSchedule)
            self.periodSchedule = None
            self.proc.autoFit = True

            newData = {}
            for key in ["HPerDay", "HCycle", "NCycles", "NDays"]:
                newData.update({key: None})
        return newData

    def updateOperationTimes(self):
        """Calculate the operation times from the detailed schedule and
        diplay them.
        
        """
        if self.periodSchedule is None:
            return None
        
        self.proc.HPerDay = self.periodSchedule.HPerDay
        self.proc.HCycle = self.periodSchedule.HBatch
        self.proc.NCycles = self.periodSchedule.NBatch
        self.proc.NDays = self.periodSchedule.NDays
        self.proc.PartLoad = self.periodSchedule.PartLoad
        self.proc.HPerYearInFlow = self.periodSchedule.HPerYearInflow
        self.proc.HPerYearOutFlow = self.periodSchedule.HPerYearOutflow

        data = {}
        for key in ["HPerDay", "HCycle", "NCycles", "NDays"]:
            data.update({key: getattr(self.proc, key)})

        return data

    def cancel(self):
        if self.new:
            Status.int.systemModel.processObject.pop(-1)
        self.new = False

