# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   moduleQ4
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel Q4 (equipment data)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#                   
#------------------------------------------------------------------------------     
#============================================================================== 
from globals import Globals, logger
from constants import EPSILON, EQUIPTYPE, AMBIENTSINK, TRANSAMBIENTSINK, EQUIPMENT
from constants import EQCLASSES, TRANSEQCLASSES
from GUI.status import Status
from GUI.conflictFrame import conflictFrame

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck, findKey
from auxiliary.dataTransfer import DataTransfer
from lib.auxiliary.float import toFloat

from modules.fluids import Fluid
from modules.schedules.schedules import DEFAULTCHARGETIME
import numpy as np
from modules.chp.moduleCHP import COPCarnot, DEFAULT_CPARTLOAD
from GUI.GUITools import ORANGE, BLACK, DARKGREY, MIDDLEGREY

# keys for parameters which are 1:1 passed to data
# structure: key_in_procObject: (key_in_Panel (or None, if identical), dictionary)

eqClassDict = dict([(code, i) for i, code in enumerate(EQCLASSES)])

DATAKEYS = {"name": (None, None),
            "Manufact": (None, None),
            "YearManufact": (None, None),
            "Model": (None, None),
            "eqClass": (None, eqClassDict),
            "eqSubClass": (None, "subClass"),
            "DBFuel_id": (None, None),
            "Pnom": (None, None),
            "mdotFuel": (None, None),
            "Pel": (None, None),
            "ThermalConsum": (None, None),
            "ElectriProduction": (None, None),
            "HCGTEfficiency": (None, None),
            "HCGTEffReal": (None, None),
            "HCGEEfficiency": (None, None),
            "COPex": (None, None),
            "CPartLoad": (None, None),
            "ControlPar1": (None, None),
            "TExhaustGas": (None, None),
            "ExcessAirRatio": (None, None),
            "TSupply": (None, None),
            "TMaxSupply": (None, None),
            "TMaxSupply2": (None, None),
            "THeatSourceHT": (None, None),
            "THeatSourceLT": (None, None),
            "TemperatureReCooling": (None, None),
            "PartLoad": (None, None),
            "HPerDay": (None, None),
            "NDays": (None, None),
            "HPerYear": (None, None),
            "HPerYearNom": (None, None),
            "TurnKeyPrice": (None, None),
            "OMFix": (None, None),
            "OMVar": (None, None),
##            "controlSchedule": (None, None),
           }        # sqlkeys of data from SQL tables to be read; in case of different names for

# list of tool tips for parameters in table
TIPS = {"name":     _("Give some brief name of the equipments to identify them in the reports"),
        "Model":    _("Model according manufacturer nomenclature"),
        "Pnom":     _("Power at manufacturer nominal or actual working conditions"),
        "mdotFuel": _("Nominal fuel consumption in units of mass or volume flow"),
        "Pel":      _("Electrical power, incl. auxiliary components, such as water pumps, control,...") + \
                    _("Power at nominal conditions. ") + _("Not including consumption of cooling tower for water-cooled chillers"),
        "HCGTEfficiency":   _("Efficiency of boiler or EER(COP) for cold generation at full load and nominal working conditions") + \
                            _("Not including consumption of cooling tower for water-cooled chillers"),
        "HCGTEffReal":  _("Average yearly conversion efficiency under real conditions") + \
                        _("Including consumption of cooling tower for water-cooled chillers"),        
        "TSupply":      _("(Average) temperature at the outlet of the equipment."),
        "TMaxSupply":   _("The maximum temperature that can be delivered by current equipment (minimum for cooling)"),
        "TMaxSupply2":  _("The maximum temperature that can be delivered by current equipment (minimum for cooling)")+"\n"+\
                        _("2nd heat or cooling supply circuit. E.g. exhaust gas in CHP engines"),
        "THeatSourceLT":    _("Temperature of the medium entering the evaporator"),
        "ThermalConsum":    _("Power applied to the generator of a thermal heat pump or chiller"),
        "THeatSourceHT":    _("Temperature of heat supply fluid entering the generator"),
        "DestinationWasteHeat": _("Specify type of cooling tower.") + u'\n' + _("If applies, specify heat exchanger where waste heat is used"),
        "TemperatureReCooling": _("Inlet temperature of cooling water or hot air stream to the chiller"),
        "HPerDay":      _("Specify representative mean values"),
        "NDays":        _("Specify representative mean values"),
        "HPerYear":     _("Real operating hours per year"),
        "HPerYearNom":  _("Nominal operating hours per year (equivalent hours at full load)"),
        "PartLoad":     _("Specify the mean supplied power of the boiler/cooler/etc. with respect to its nominal power"),
}        


# TODO: parameters from old Q4 panel - to be checked if still necessary
##            self.tc30.SetValue(q.HeatSourceLT)
##            self.tc32.SetValue(pipeDict[int(q.HeatSourceHT)])
##             self.tc35.SetValue(TRANSAMBIENTSINK[q.DestinationWasteHeat])
##            if q.Refrigerant in fluidDict.keys():
##                self.tc102.SetValue(fluidDict[q.Refrigerant])

class ModuleQ4(object):
    """

    Attributes:

    
    """
    def __init__(self, panel):
        self.panel = panel
        
        # number of rows in the display table
        self.nRowsTechData = 0                    

        self.modTSEditor = None            # to be set from the panel

        self.ccEq = None

        self.deletedSchedules = []          # buffer for managing delete of temporarily used schedules

        self.new = False
        
    def initPanel(self, no):
        """  functions that have to be carried out only once
        when the panel is created
        """
        # dictionary for dynamic translations
        self.dynamicTrans = {}
        self.dynamicTrans["Fluid"]  = Status.prj.getFluidDict(includeZeroWater=True)
        self.dynamicTrans["Fuel"]  = Status.prj.getFuelDict()

        #TODO:
        
##        self.dynamicTrans["EquipType"] = EQUIPTYPES

##        unitOpDict = Status.prj.getUnitOpDict(code=True)
##        unitOpItems = [(code, text) for code, text in unitOpDict.items()]
##        sortedItems =  sorted(unitOpItems, key=lambda item: item[0])
##        
##        unitOpList = [text for (code, text) in sortedItems]
##        unitOpCodes = [code for (code, text) in sortedItems]
##        self.dynamicTrans["UnitOp"] = dict([(code, i) for i, code in enumerate(unitOpCodes)])

        if no >= 0:
            self.eq = Status.int.systemModel.eqObject[no]
        else:
            Status.int.systemModel.addEquipment()
            Status.int.systemModel.build()
            self.eq = Status.int.systemModel.eqObject[-1]
            self.new = True

        data = {}

        eqClass = self.eq.eqClass
        self.subClasses = [entry[2] for entry in EQUIPMENT[eqClass]]
        self.subClassesTrans = [entry[3] for entry in EQUIPMENT[eqClass]]
        data["eqSubClassChoices"] = self.subClassesTrans

        self.dynamicTrans["subClass"] = dict([(code, i) for i, code in enumerate(self.subClasses)])

        return data

    def setEqClass(self, index):
        """
        Helper function for setting equipment class from the GUI
        """
        if index is None: index = 0
        
        eqClass = EQCLASSES[index]
        changed = (self.eq.eqClass != eqClass)
        self.eq.setEqClass(eqClass)
        
        self.subClasses = [entry[2] for entry in EQUIPMENT[eqClass]]
        self.subClassesTrans = [entry[3] for entry in EQUIPMENT[eqClass]]

        self.dynamicTrans["subClass"] = dict([(code, i) for i, code in enumerate(self.subClasses)])
        
        return self.subClassesTrans, changed
    
##        return {"unitOps": unitOpList,
##                "BUI": (self.proc.buildingId is not None)}

    def setEqSubClass(self, index):
        """ here appropriate entries should be set to True or False
        """
        subClass = self.subClasses[index]
        if self.eq.eqSubClass != subClass:
            self.eq.setEqSubClass(subClass)
            return True

        return False
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        if self.eq is None: return

        data = self.getEquipmentDisplayData()            

        techDataTable, techDataChoices = self.getDataTable(data)
        self.nRowsTechData = len(techDataTable)

        data["DataTable"] = {"data": techDataTable, "choices": techDataChoices, "pointers": self.pointers,
                             "parCodes": self.parCodesDataTable,
                             "toolTips": self.toolTipsDataTable}
        data["funit2"] = self.funit2Info
        
        # try to load detailed schedule
        self.modTSEditor.load(-self.eq.id if self.eq.id is not None else None, self.eq.name)
        self.controlSchedule = (self.modTSEditor.schedule is not None)

        self.eq.controlSchedule = self.modTSEditor.schedule
        data["controlSchedule"] = 1*self.controlSchedule

        # export parameters for PanelCC
        j = self.eq.no
        dataGroup = ("Equipment[" + str(j+1) + "]")
        data["cc"] = (self.ccEq, dataGroup)

        if self.eq.eqType in ["CHP ORC", "CHP TFC"]:
            data.update(self.buildEfficiencyCurves())
        else:
            data.update({"EffPlots": False})
            
        return data
        
    def selectProcess(self):
        
        k = self.selectedProcessNo
        proc = Status.int.systemModel.process[k]

        processData = {"Questionnaire_id":Status.PId,
                       "AlternativeProposalNo":Status.ANo}

    def saveToDB(self, data={}):
        """ saves the data received from the panel to the data base
        """
        if not Status.prj.editAllowed():
            self.errMsgSave = _("Editing of this alternative is not allowed")
            return False

        eq = self.eq
      
        # collects here all the parameters from the panel (exclunding that in tables) which can be stored directly or via translation dictionary
        # note: parameters displayed in tables have already been updated in proc directly when changed

        for key in DATAKEYS.keys():
            if key in data.keys():
                val = data[key]

                trans = DATAKEYS[key][1]
                trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
                if trans is None:
                    setattr(eq, key, dbio.check(val))
                else:
                    setattr(eq, key, dbio.check(findKey(trans, val)))

        # read schedule parameters from panel        
        if self.eq.controlSchedule:
            err = self.eq.controlSchedule.check()
            if err:
                logError(err)
                self.errMsgSave = _("Erroneous data in control schedule")
                return False

        j = eq.no
        conflicts = Status.mod.moduleCC.eqCheck(eq)
        Status.mod.moduleCC.initPanel()
        if conflicts > 0:
            cf = conflictFrame(Globals.GUI.editPanel)
            cf.Show()
            return False

        # auto complete some mandatory data
        if Status.ANo >= 0 and not eq.checkCompleteness():
            self.errMsgSave = _("Some mandatory data are missing")
            return False

        eq.saveToDB()

        Status.prj.projectChanged()
        Status.int.changeInCascade(j+1, keepModel=True)
        Status.int.systemModel.build()

        self.errMsgSave = None
        self.new = False
        return True


    def changeDataDataTable(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False

##        print "ModuleQ4.changeDataDataTable: index %s mode %s data %s"%(index, mode, data)
        
        if mode == -1:   # change of data in a dataSet

            row = data
            editable = self.editableDataTable[index]
            parCodes = self.parCodesDataTable[index]

            newVals = self.readLineDataTable(row)
            if newVals:
                for i in xrange(2):
                    if editable[i] and (parCodes[i] is not None):
                        if parCodes[i]=="DestinationWasteHeat":
                            newVals[i] = AMBIENTSINK[int(newVals[i])] if newVals[i] is not None else None
                        setattr(self.eq, parCodes[i], newVals[i])
                 
##                self.panel.setSecondaryUnits()
            else:
                logger.logError(_("Invalid data set. Please try again"))
                err = True
    
            self.activeDataSet = index

##        self.panel.setPointers(self.getPointers())
##                                   
        data = self.getEquipmentDisplayData()
        tableData, choices = self.getDataTable(data)
        
        return 0, tableData, choices

    def getDataTable(self, data):
        """
        transforms the information contained in the equipment (received as data) into a table of strings for
        representation on the GUI
        """
        eq = self.eq
        # collect data for demand table

        tableData = {}
        techDataTable = []
        choices = []

        # flags controlling specific entries for different equipment types
        heating = eq.eqClass != "CH"
        hasFuel = (eq.fuel is not None)

        hasDrivingHeat = eq.eqType in ["thermal chiller (air cooled)", "thermal chiller (water cooled)",
                                          "CHP ORC", "CHP TFC"]
        
        hasDissipation = eq.eqType in ["compression chiller (air cooled)", "compression chiller (water cooled)",
                                          "thermal chiller (air cooled)", "thermal chiller (water cooled)",
                                          "CHP ORC", "CHP TFC"]

        hasLTHeat = (eq.eqClass == "HP")
        hasCircuit2 = (eq.eqClass=="CHP" and eq.eqSubClass in ["CHP engine"])
        
        hc = _("heat") if heating else _("cooling")
        minMax = _("Maximum") if heating else _("Minimum")

        # formatting information for table
        subTitles = []
        editable = []
        funit2Info = []
        pointers = {}
        parCodes = []
        
        # FET + QHX_HT
        subTitles.append(len(techDataTable))
        pointers.update({"FET": len(techDataTable)})
        techDataTable.append([_("Final energy consumption"), ("text", None, None), None, None])
        choices.append([None]*4)
        editable.append([False, False])
        parCodes.append([None, None])
        if hasFuel:
            techDataTable.append([_("Power input (fuel)"), ("float", "POWER", 1), "-", eq.PfuelAvg])
            choices.append([None]*4)
            editable.append([False, False])
            parCodes.append([None, None])
            
            funit2Info = [(len(techDataTable), eq.fuel.rho)]
            techDataTable.append([_("Fuel consumption (nominal)"), ("float", "MASSORVOLUMEFLOW", 0), eq.mdotFuel, None])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["mdotFuel", None])

            techDataTable.append([_("Auxiliary electricity consumption"), ("float", "POWER", 1), eq.Pel, eq.PelAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["Pel", None])

        elif hasDrivingHeat:
            techDataTable.append([_("Power input (heat)"), ("float", "POWER", 1), eq.ThermalConsum, eq.PheatAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ThermalConsum", None])

            techDataTable.append([_("Driving temperature"), ("float", "TEMPERATURE", 2), eq.THeatSourceHT, None])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["THeatSourceHT", None])

            techDataTable.append([_("Auxiliary electricity consumption"), ("float", "POWER", 1), eq.Pel, eq.PelAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["Pel", None])

        elif self.eq.eqClass=="ST":
            techDataTable.append([_("Auxiliary electricity consumption"), ("float", "POWER", 1), eq.Pel, eq.PelAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["Pel", None])
            
        else:
            techDataTable.append([_("Power input (electricity)"), ("float", "POWER", 1), eq.Pel, eq.PelAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["Pel", None])


        # QHX_LT
        if hasLTHeat:
            pointers.update({"QHX": len(techDataTable)})
            subTitles.append(len(techDataTable))
            techDataTable.append([_("Low temperature heat source"), ("text", None, None), None, None])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append([None, None])

            techDataTable.append([_("Temperature of low temp. heat source"), ("float", "TEMPERATURE", 2), eq.THeatSourceLT, None])
            choices.append([None]*4)
            editable.append([True, False])           
            parCodes.append(["THeatSourceLT", None])

        # output
        pointers.update({"USH": len(techDataTable)})
        subTitles.append(len(techDataTable))
        techDataTable.append([_("Useful supply %s"%hc), ("text", None, None), None, None])
        choices.append([None]*4)
        editable.append([True, False])
        parCodes.append([None, None])

        techDataTable.append([_("Thermal power output"), ("float", "POWER", 1), eq.Pnom, eq.Pavg])
        choices.append([None]*4)
        editable.append([True, False])
        parCodes.append(["Pnom", None])

        if self.eq.eqClass != "ST":
            techDataTable.append([_("Thermal conversion efficiency"), ("float", "FRACTION", 3), eq.HCGTEfficiency, eq.HCGTEffReal])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["HCGTEfficiency", "HCGTEffReal"])

            techDataTable.append([_("Temperature of %s supplied"%hc), ("float", "TEMPERATURE", 2), eq.TSupply, "-"])
            choices.append([None]*4)
            editable.append([self.TCorr, False])
            parCodes.append(["TSupply", None])
            
        techDataTable.append([_("%s temp. of %s supplied")%(minMax, hc), ("float", "TEMPERATURE", 2), eq.TMaxSupply, "-"])
        choices.append([None]*4)
        editable.append([True, False])
        parCodes.append(["TMaxSupply", None])

        if hasCircuit2:
            techDataTable.append([_("Maximum temp. of heat supplied - 2nd circuit"), ("float", "TEMPERATURE", 2), eq.TMaxSupply2, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["TMaxSupply2", None])

        if self.eq.eqClass == "CHP":
            subTitles.append(len(techDataTable))
            techDataTable.append([_("Electricity generation"), ("text", None, None), None, None])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append([None, None])

            techDataTable.append([_("Electrical power output"), ("float", "POWER", 1), eq.ElectriProduction, eq.PelGenAvg])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ElectriProduction", None])

            techDataTable.append([_("Electrical conversion efficiency"), ("float", "FRACTION", 3), eq.HCGEEfficiency, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["HCGEEfficiency", None])


        if self.eq.eqClass == "ST":
            subTitles.append(len(techDataTable))
            techDataTable.append([_("Solar thermal system"), ("text", None, None), None, None])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append([None, None])

            techDataTable.append([_("C0"), ("float", "FRACTION", 3), eq.ST_C0, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_C0", None])

            techDataTable.append([_("C1"), ("float", "HEATLOSSCOEFF", 3), eq.ST_C1, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_C1", None])

            techDataTable.append([_("C2"), ("float", "HEATLOSSCOEFF2", 3), eq.ST_C2, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_C2", None])

            techDataTable.append([_("K50L"), ("float", "FRACTION", 3), eq.ST_K50L, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_K50L", None])

            techDataTable.append([_("K50T"), ("float", "FRACTION", 3), eq.ST_K50T, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_K50T", None])

            techDataTable.append([_("Storage Volume"), ("float", "VOLUME", 2), eq.ST_Volume, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_Volume", None])

            techDataTable.append([_("System Efficiency"), ("float", "FRACTION", 3), eq.ST_SysEff, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ST_SysEff", None])

            
        if hasDissipation:
            pointers.update({"QWH": len(techDataTable)})
            subTitles.append(len(techDataTable))
            techDataTable.append([_("Heat dissipation"), ("text", None, None), None, None])
            choices.append([None]*4)
            editable.append([False, False])
            parCodes.append([None, None])

            ambientSink = eq.DestinationWasteHeat
            iSink = AMBIENTSINK.index(ambientSink) if ambientSink in AMBIENTSINK else None
            choicesSink = [TRANSAMBIENTSINK[sink] for sink in AMBIENTSINK]

            if eq.eqType in ["compression chiller (air cooled)", "thermal chiller (air cooled)"]:
                techDataTable.append([_("Destination of waste heat"), ("choice", None, None), 2, None])
                choices.append([None, None, choicesSink, choicesSink])
                editable.append([False, False])                
                parCodes.append(["DestinationWasteHeat", None])
            else:
                techDataTable.append([_("Destination of waste heat"), ("choice", None, None), iSink, None])
                choices.append([None, None, choicesSink[:2], choicesSink[:2]])
                editable.append([True, False])
                parCodes.append(["DestinationWasteHeat", None])

            techDataTable.append([_("Dissipated power (heat rejection)"), ("float", "POWER", 1), "-", "-"])
            choices.append([None]*4)
            editable.append([False, False])
            parCodes.append([None, None])

            techDataTable.append([_("Temperature (heat rejection)"), ("float", "TEMPERATURE", 2), eq.TemperatureReCooling, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["TemperatureReCooling", None])

        # exhaust gas parameters
        if hasFuel and self.eq.eqClass == "BB":
            subTitles.append(len(techDataTable))
            techDataTable.append([_("Exhaust gas parameters"), ("text", None, None), None, None])
            choices.append([None]*4)
            editable.append([False, False])
            parCodes.append([None, None])

            techDataTable.append([_("Excess Air Ratio"), ("float", "FRACTION", 3), eq.ExcessAirRatio, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["ExcessAirRatio", None])

            techDataTable.append([_("Exhaust Gas Temperature"), ("float", "TEMPERATURE", 2), eq.TExhaustGas, "-"])
            choices.append([None]*4)
            editable.append([True, False])
            parCodes.append(["TExhaustGas", None])

        toolTips = []
        for row in parCodes:
            toolTips.append([None, None]+[TIPS[par] if par in TIPS.keys() else None for par in row])

        # export formatting information
        self.subTitlesDataTable = subTitles
        self.editableDataTable = editable
        self.parCodesDataTable = parCodes
        self.toolTipsDataTable = toolTips

        self.funit2Info = funit2Info
        self.pointers = pointers
        
        return techDataTable, choices



######        """Enable and disable the entry fields that are only sensible
######        for a certain class of equipment, according to the chosen 
######        equipment.
######        
######        :param event: Is ignored and only there to fullfill the API of
######        wx eventhandlers.
######        
######        """
######        allEntries = set([self.tc8, self.tc9, self.tc12,
######                          self.tc13b, self.tc16, self.tc16_2, self.tc15, self.tc14, 
######                          self.tc102, self.COPex, self.tc34, self.tc33, self.tc32, 
######                          self.tc31, self.tc30, self.tc35, self.tc36,
######                          self.TMaxSupply,self.TMaxSupply2, self.entryControlPar1])
######        
######        eType = findKey(TRANSEQUIPTYPE, self.tc5.GetValue(text=True))
######        eClass = getEquipmentClass(eType)
######        
######        if eClass == "HP":
######            if eType in ["thermal chiller (air cooled)", "thermal chiller (water cooled)"]: # for thermally driven heat pumps. not yet existing
######                sensibleEntries = set([self.tc12, self.tc13b, self.tc102, self.COPex, self.tc34, self.tc33, self.tc32, self.tc31,
######                                       self.tc30, self.TMaxSupply])
######            else:
######                sensibleEntries = set([self.tc12, self.tc13b, self.tc102, self.COPex, self.tc34, self.tc33, self.tc32, self.tc31,
######                                       self.tc30, self.TMaxSupply])
######        elif eClass == "BB":
######            sensibleEntries = set([self.tc8, self.tc9, self.tc12, self.tc16, self.tc16_2, self.TMaxSupply])
######        elif eClass == "CH":
######            if eType in ["thermal chiller (air cooled)", "thermal chiller (water cooled)"]: # thermally driven chillers
######                sensibleEntries = set([self.tc12, self.tc13b, self.tc102, self.tc34, self.tc33, self.tc32,
######                                       self.tc35, self.tc36, self.TMaxSupply])
######            elif eType in ["compression chiller (air cooled)", "compression chiller (water cooled)"]: # compression chillers
######                sensibleEntries = set([self.tc12, self.tc13b, self.tc102,
######                                       self.tc35, self.tc36, self.TMaxSupply, self.COPex])
######            else:
######                sensibleEntries = set()
######                
######        elif eClass == "ST":
######            sensibleEntries = set([self.tc12])
######        elif eClass == "CHP":
######            if eType in ["CHP ORC", "CHP TFC"]:
######                sensibleEntries = set([self.tc15, self.tc14, self.TMaxSupply, self.TMaxSupply2,
######                                       self.tc34, self.tc33, self.tc32, self.tc36, self.COPex])
######            else:
######                sensibleEntries = set([self.tc8, self.tc9, self.tc13b, self.tc15, self.tc14, self.TMaxSupply, self.TMaxSupply2,
######                                       self.entryControlPar1])
######        else:
######            logDebug("PanelQ4: Unknown equipment class %s." % (eClass,))
######            sensibleEntries = set()
######        for entry in sensibleEntries:
######            entry.Enable()
######            entry.Show()
######        for entry in allEntries - sensibleEntries:
######            entry.Disable()
######            entry.Hide()
######            try:
######                entry.SetChoice(0) # reset choices
######            except AttributeError:
######                entry.Clear() # empty input fields
######
######        self.entryTurnKeyPrice.Enable(Status.ANo > 0)

    def readLineDataTable(self, row):
        """ reads a line of raw data from the table in the GUI
        and translates it into internal code
        (inverse of getTable)
        """ 
        return [toFloat(entry) for entry in row[2:]]



    def getEquipmentDisplayData(self):
        """
        translates the data in self.eq into a data dicitionary as
        required by the panel
        """
        
        # read in all data from eqObject translate where necessary
        q = self.eq
        
        data = {}
        for dataKey in DATAKEYS:
            val = getattr(q, dataKey)
            key = noneCheck(DATAKEYS[dataKey][0], dataKey)
            trans = DATAKEYS[dataKey][1]
            trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
            if trans is None:
                data[key] = val
            else:
                data[key] = trans.get(val)

        data["PartLoadCorr"] = self.eq.eqClass in ["CH"] or self.eq.eqSubClass in ["CHP ORC", "CHP TFC"]
        data["TCorr"] = self.eq.eqClass in ["HP"] or \
                        self.eq.eqSubClass in ["CHP ORC", "CHP TFC"] or \
                        self.eq.eqType in ["compression chiller (air cooled)",
                                              "compression chiller (water cooled)"]
        
        data["NControlPars"] = 1*(self.eq.eqType in ["CHP engine",
                                                        "CHP gas turbine",
                                                        "CHP steam turbine",
                                                        "CHP fuel cell"])

        self.TCorr = data["TCorr"]

        return data
        

    def setScheduleType(self, scheduleType):
        """
        To be called from the event handler in the GUI if a detailed schedule is activated
        """

        newData = None
        if scheduleType is None or scheduleType==0:
            if self.eq.controlSchedule:
                self.eq.deletedSchedules.append(self.eq.controlSchedule)
            self.eq.controlSchedule = None
            
        elif scheduleType==1:
            if self.eq.controlSchedule is None: # no detailed schedule yet -> create new empty one        
                self.modTSEditor.viewer.load(-self.eq.id if self.eq.id is not None else None, self.eq.name, createNew=True)
                self.eq.controlSchedule = self.modTSEditor.schedule


        return newData

    def cancel(self):
        if self.new:
            Status.int.systemModel.eqObject.pop(-1)
        self.new = False

    def buildEfficiencyCurves(self):
        """
        specific function currently only used for HTP equipment
        """
        # export plot data for efficiency curves
        # nominal data
        TDriveNom = noneCheck(self.eq.THeatSourceHT, 100)
        TRejectNom = noneCheck(self.eq.TemperatureReCooling, 20)
        COPex = self.eq.COPex
        COPe = self.eq.HCGEEfficiency
        ElGenNom = self.eq.ElectriProduction

        # flags for checking if calculations are possible
        nominalEffOK = not any([val is None for val in [TDriveNom, TRejectNom, COPe]])
        effCurvesOK = COPex is not None and COPe is not None
        nominalPLOK = not any([val is None for val in [TDriveNom, TRejectNom, COPe, COPex]])
        PLCurveOK = True

        if nominalEffOK:
            xDataEff = [np.array([TDriveNom])]
            yDataEff = [np.array([COPe])]
            labelsEff = [_("nominal")]

            lineTypesEff = ['o']
            coloursEff = [BLACK]
            lineWidthsEff = [10]
            markerSizesEff = [10]

        else:
            xDataEff = []
            yDataEff = []
            labelsEff = []                              

            lineTypesEff = []
            coloursEff = []
            lineWidthsEff = []
            markerSizesEff = []

        TDriveNom = noneCheck(TDriveNom, 100)
        TRejectNom = noneCheck(TRejectNom, 20)
        Tmin = TDriveNom - 20
        Tmax = TDriveNom + 20

        if effCurvesOK:
            xvect = np.linspace(Tmin, Tmax, 101)       
            Tbase = 1.0*int((TRejectNom + 0.5))
            for dT in range(-20, 30, 10):
                TReject = Tbase  + dT

                COPe_vect = np.array([COPex*COPCarnot(T, TReject) for T in xvect])
                fcorr_T_vect = np.minimum(np.maximum(0.5 + 0.5*COPe_vect/COPe, 0.8), 1.2) # smooth and constrain correction

                xDataEff.append(xvect)
                yDataEff.append(COPe*fcorr_T_vect)
                labelsEff.append("T_rej  % 3d °C"%int(TReject+0.5))

            lineTypesEff  += ['-.', ':', '-', '--', '-']
            coloursEff    += [DARKGREY, MIDDLEGREY, ORANGE, MIDDLEGREY, DARKGREY]
            lineWidthsEff += [1, 1, 2, 1, 1]
            markerSizesEff += [1]*5

        # part load correction function
        if nominalPLOK:
            fth = np.linspace(0, 1, 101)
            cPartLoad = noneCheck(self.eq.ST_C0, DEFAULT_CPARTLOAD)

            fPartLoad = np.maximum((fth - cPartLoad) / (1. - cPartLoad), 0.0)
            COPe_partLoad = COPe*fPartLoad/np.maximum(1.e-10, fth)

            xDataPL = [fPartLoad]
            yDataPL = [COPe_partLoad]
            labelsPL = [_("nominal")]

            lineTypesPL = ['-']
            coloursPL = [ORANGE]
            lineWidthsPL = [1]
        else:
            xDataPL = []
            yDataPL = []
            labelsPL = []

            lineTypesPL = []
            coloursPL = []
            lineWidthsPL = []
                              
            
        # actual data (only ANo > 1):
        systemModel = Status.int.systemModel
        if Status.ANo > 0 and systemModel.upToDate:
##        systemModel.update()  # assure that the system model is up to date

            TDrive = TDriveNom*np.ones(Status.Nt) if self.eq.TDrive is None else self.eq.TDrive
            j = self.eq.no
            TReject = Status.WeatherData.Tdb if self.eq.TReject is None else self.eq.TReject

            j = self.eq.no
            ElGen = Status.int.ElGen_t[j] if Status.SimTool=="NP" else Status.mod.moduleEnergy.sim.systemModel.ElGen_t[j]
            
            xact = []
            yact = []

            xactPL = []
            yactPL = []
            
            for i, T in enumerate(TReject):
                if ElGen[i] > EPSILON and COPex is not None:
                    xact.append(TDrive[i])
                    COPe_actFL = COPex*COPCarnot(TDrive[i], TReject[i])

                    fcorr_T = min(max(0.5 + 0.5*COPe_actFL/COPe, 0.8), 1.2) # smooth and constrain correction
                    yact.append(COPe*fcorr_T)

                    fe = ElGen[i]/ElGenNom/fcorr_T
                    ft = cPartLoad + fe * (1. - cPartLoad)
                    COPe_act = yact[i]*fe/np.maximum(1.e-10, ft)
                    xactPL.append(fe)
                    yactPL.append(COPe_act)

            if len(xact) > 0:
                xDataEff.append(np.array(xact))
                yDataEff.append(np.array(yact))
                labelsEff.append(_("actual"))
                lineTypesEff.append(".")
                coloursEff.append(DARKGREY)
                lineWidthsEff.append(1)
                markerSizesEff.append(2)

                xDataPL.append(np.array(xactPL))
                yDataPL.append(np.array(yactPL))
                labelsPL.append(_("actual"))
                lineTypesPL.append(",")
                coloursPL.append(DARKGREY)
                lineWidthsPL.append(1)


        data = {}                    
        data.update({"EfficiencyCurves": {"xData": xDataEff, "yData": yDataEff, "labels": labelsEff,
                                          "lineTypes": lineTypesEff, "colours": coloursEff, "lineWidths": lineWidthsEff, "markerSizes": markerSizesEff}})

        data.update({"PartLoadBehaviour": {"xData": xDataPL, "yData": yDataPL, "labels": labelsPL,
                                          "lineTypes": lineTypesPL, "colours": coloursPL, "lineWidths": lineWidthsPL}})

        data.update({"EffPlots": True})

        return data


