# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   moduleQ5
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel Q5 (equipment data)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 3.0
#                   
#------------------------------------------------------------------------------     
#============================================================================== 
from globals import Globals, logger
from constants import EPSILON, EQUIPTYPE, AMBIENTSINK, TRANSAMBIENTSINK, EQUIPMENT
from constants import PIPETYPES, TRANSPIPETYPES
from GUI.status import Status
from GUI.conflictFrame import conflictFrame

import database.dbio as dbio
from auxiliary.auxiliary import noneCheck, findKey
from auxiliary.dataTransfer import DataTransfer
from lib.auxiliary.float import toFloat

from modules.fluids import Fluid
from modules.schedules.schedules import DEFAULTCHARGETIME

# keys for parameters which are 1:1 passed to data
# structure: key_in_procObject: (key_in_Panel (or None, if identical), dictionary)

pipeTypeDict = dict([(code, i) for i, code in enumerate(PIPETYPES)])

DATAKEYS = {"name" : (None, None),
            "pipeType": (None, pipeTypeDict),
            "fluidId" : (None, "Fluid"),
            "DistribCircFlow" : (None, None),
            "ToutDistrib" : (None, None),
            "TreturnDistrib" : (None, None),
            "PercentRecirc" : (None, None),
            "Tfeedup" : (None, None),
            "PressDistMedium" : (None, None),
            "TotLengthDistPipe" : (None, None),
            "UAPipe" : (None, None),
            "DDistPipe" : (None, None),
            "DeltaDistPipe" : (None, None),
            "HDEffAvg" : (None, None),
            "QHXPipe" : (None, None),
            "QWHPipe" : (None, None),
            "USHPipe" : (None, None),
            "USHm" : (None, None),
            "UPHProcm" : (None, None),
            "HPerYearPipe" : (None, None)}        # sqlkeys of data from SQL tables to be read; in case of different names for

class ModuleQ5(object):
    """

    Attributes:

    
    """
    def __init__(self, panel):
        self.panel = panel

        self.ccPipe = None
        self.pipeTypePrevious = None
        self.new = False
        
    def initPanel(self, no):
        """  functions that have to be carried out only once
        when the panel is created
        """
        # dictionary for dynamic translations
        self.dynamicTrans = {}

        fluidIDDict = Status.prj.getFluidDict(includeZeroWater=True)
        fluidChoices = fluidIDDict.values()
        fluidChoices.sort()
        fluidTrans = dict([(findKey(fluidIDDict, name), i) for i, name in enumerate(fluidChoices)])
        
        self.dynamicTrans["Fluid"]  = fluidTrans

        if no >= 0:
            self.pipe = Status.int.systemModel.pipeObject[no]
        else:
            Status.int.systemModel.addObject("PIPE")
            self.pipe = Status.int.systemModel.pipeObject[-1]
            self.new = True

        data = {"FluidChoices": fluidChoices,
                "pipeTypeChoices": [TRANSPIPETYPES[pipeType] for pipeType in PIPETYPES]}

##        eqClass = self.eq.eqClass
##        self.subClasses = [entry[2] for entry in EQUIPMENT[eqClass]]
##        self.subClassesTrans = [entry[3] for entry in EQUIPMENT[eqClass]]
##        data["eqSubClassChoices"] = self.subClassesTrans
##
##        self.dynamicTrans["subClass"] = dict([(code, i) for i, code in enumerate(self.subClasses)])

        self.pipeTypePrevious = self.pipe.pipeType
        return data

####    def setEqClass(self, index):
####        """
####        Helper function for setting equipment class from the GUI
####        """
####        if index is None: index = 0
####        
####        eqClass = EQCLASSES[index]
####        changed = (self.eq.eqClass != eqClass)
####        self.eq.setEqClass(eqClass)
####        
####        self.subClasses = [entry[2] for entry in EQUIPMENT[eqClass]]
####        self.subClassesTrans = [entry[3] for entry in EQUIPMENT[eqClass]]
####
####        self.dynamicTrans["subClass"] = dict([(code, i) for i, code in enumerate(self.subClasses)])
####        
####        return self.subClassesTrans, changed
####    
######        return {"unitOps": unitOpList,
######                "BUI": (self.proc.buildingId is not None)}
####
####    def setEqSubClass(self, index):
####        """ here appropriate entries should be set to True or False
####        """
####        subClass = self.subClasses[index]
####        self.eq.setEqSubClass(subClass)
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        if self.pipe is None: return

        data = self.getPipeDisplayData()       # fetches data to be directly displayed in entries     

        data["funit2"] = self.pipe.fluid.rho if self.pipe.fluid is not None else 1.0
                

        # export parameters for PanelCC
        dataGroup = ("Pipe[" + str(self.pipe.no+1) + "]")
        data["cc"] = (self.ccPipe, dataGroup)

        return data

    def setPipeType(self, index):
        """
        index: index within PIPETYPES
        """
        self.pipeTypePrevious = self.pipe.pipeType
        
        if index >= 0 and index < len(PIPETYPES):
            self.pipe.setPipeType(PIPETYPES[index])

    def changeFluidId(self, index):
        fluidId = findKey(self.dynamicTrans["Fluid"], index)
        self.pipe.fluidId = fluidId
        self.pipe.updateFluid()
        funit2 = self.pipe.fluid.rho if self.pipe.fluid is not None else 1.0
        return funit2

    def readPanel(self, data={}):

        # collects here all the parameters from the panel (exclunding that in tables) which can be stored directly or via translation dictionary
        # note: parameters displayed in tables have already been updated in proc directly when changed

        for key in DATAKEYS.keys():
            if key in data.keys():
                val = data[key]

                trans = DATAKEYS[key][1]
                trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
                if trans is None:
                    setattr(self.pipe, key, dbio.check(val))
                else:
                    setattr(self.pipe, key, dbio.check(findKey(trans, val)))

        # change pipe type if currently selected type = closed and recirculation rate has been set to 0
        self.pipe.setPipeType(previous = self.pipeTypePrevious)
        self.pipeTypePrevious = self.pipe.pipeType
        
    def saveToDB(self, data={}):
        """ saves the data received from the panel to the data base
        """
        if not Status.prj.editAllowed():
            self.errMsgSave = _("Editing of this alternative is not allowed")
            return False
        
        self.readPanel(data)

        # read schedule parameters from panel        
        conflicts = Status.mod.moduleCC.pipeCheck(self.pipe)
        self.pipe.cc = Status.mod.moduleCC.ccLocal
        Status.mod.moduleCC.initPanel()
        if conflicts > 0:
            cf = conflictFrame(Globals.GUI.editPanel)
            cf.Show()
            return False

        # auto complete some mandatory data
        if Status.ANo >= 0 and not self.pipe.checkCompleteness():
            self.errMsgSave = _("Some mandatory data are missing")
            return False

        self.pipe.saveToDB()

        Status.prj.projectChanged()
        Status.int.systemModel.build()

        self.errMsgSave = None
        self.new = False
        return True


    def getPipeDisplayData(self):
        """
        translates the data in self.pipe into a data dicitionary as
        required by the panel
        """
        
        # read in all data from eqObject translate where necessary
        q = self.pipe
        
        data = {}
        for dataKey in DATAKEYS:
            val = getattr(q, dataKey)
            key = noneCheck(DATAKEYS[dataKey][0], dataKey)
            trans = DATAKEYS[dataKey][1]
            trans = self.dynamicTrans[trans] if isinstance(trans, str) else trans
            if trans is None:
                data[key] = val
            else:
                data[key] = trans.get(val)

##        data["PartLoadCorr"] = self.eq.eqClass in ["CH"] or self.eq.eqSubClass in ["CHP ORC", "CHP TFC"]
##        data["TCorr"] = self.eq.eqClass in ["HP"] or \
##                        self.eq.eqSubClass in ["CHP ORC", "CHP TFC"] or \
##                        self.eq.eqType in ["compression chiller (air cooled)",
##                                              "compression chiller (water cooled)"]
##        
##        data["NControlPars"] = 1*(self.eq.eqType in ["CHP engine",
##                                                        "CHP gas turbine",
##                                                        "CHP steam turbine",
##                                                        "CHP fuel cell"])
##
##        self.TCorr = data["TCorr"]

        return data

    def cancel(self):
        if self.new:
            Status.int.systemModel.pipeObject.pop(-1)
        self.new = False
        
