# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   moduleQ6
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel Q6 (heat recovery)
#           
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------     
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.status import Status
from globals import logger
from modules.messageLogger import logTrack, logDebug, logWarning, showWarning
from modules.heatRecovery.hxConnection import buildHXNet, HXConnection, StreamGlobals, HeatExchanger
from constants import OBJECTTYPES_HX, TRANSOBJECTTYPES_HX

from auxiliary.auxiliary import toGettextAndUnicode as _U
from auxiliary.auxiliary import toUnicode as U
from auxiliary.auxiliary import toUTF8

from copy import copy


SHOWENERGY = True           #Controls wether energy or power is shown in pinch graphics


class ModuleQ6(object):
    """

    Attributes:

    hx:  selected heat exchanger being edited / displayed
    activeConn  connection object being edited in the connection editor (EditConnection)
    
    """
    def __init__(self, keys, no):
        self.keys = keys

        self.hx = None
        self.HXNames = []

        self.initPanel(no)

        self.new = False


    def initPanel(self, no):
        """  functions that have to be carried out only once
        when the panel is created
        """
        self.selectedHXId = None
        self.selectedSourceId = None
        self.selectedSinkId = None
        self.selectedSource = None
        self.selectedSink = None
        self.selectedSourceNo = None
        self.selectedSinkNo = None

        # current and backup of previous source stream. used for modification of HX connections from drawings
        # for UNDO function -> restore previous connection 
        self.sinkstreamDS = None
        self.sinkstreamDS_bak = None
        self.sourcestreamDS = None
        self.sourcestreamDS_bak = None
        self.hxDS = []     # storage for downstream HX modified by changeConnection

        if no >= 0:
            self.selectHX(Status.int.hx[no])
        else:
            self.addHX()
            self.new = True
            
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        # create system model -> contains list of heat exchangers
        # to do: shift to module !!!
        logTrack("ModuleQ6.updatePanel")

        Status.int.systemModel.build()
        self.HXNames = [hx.name for hx in Status.int.hx]


    def selectHX(self, hx):
        """ imports selected hx from Panel
        """
        #TODO: change panel so that the selected HX is the one of the module
        self.hx = hx
        
        self.selectSink(0)
        self.selectSource(0)

    def addHX(self):
        """ adds an empty template for a new heat exchanger
        """

        name = _("New HX")

        i = 0
        while name in self.HXNames:
            i +=1
            name = _("New HX") + " (%2d)"%i
            
##        self.HXNames.append(name)
        self.hx = HeatExchanger(100, None, name)
        self.hx.HXNo = len(Status.int.hx)+1
        
        self.selectHX(self.hx)

        return self.hx

    def deleteHX(self):
##        print "ModuleQ6.deleteHX: deleting hx %s"%self.hx
        if self.hx is not None:
            self.hx.deleteFromDB()
            Status.int.changeInHeatRecovery()

        self.hx = None
        return self.hx


#------------------------------------------------------------------------------------
# functions for event handlers connection panel

    def selectSink(self, index):
        if self.hx is None: # security check for avoiding crashes
            self.selectedSource = None
            self.selectedSourceNo = None
            return
        
        elif index < len(self.hx.sinkstreams) and index >= 0:
            self.selectedSink = self.hx.sinkstreams[index]
            self.selectedSinkNo = index
        else:
            self.selectedSink = None
            self.selectedSinkNo = None

    def selectSource(self, index):
        if self.hx is None: # security check for avoiding crashes
            self.selectedSource = None
            self.selectedSourceNo = None
            return
        
        elif index < len(self.hx.sourcestreams) and index >= 0:
            self.selectedSource = self.hx.sourcestreams[index]
            self.selectedSourceNo = index
        else:
            self.selectedSource = None
            self.selectedSourceNo = None

#------------------------------------------------------------------------------------
# functions for event handlers connection editor

    def editConnection(self, heating):
        """ before call to the connection editor for editig and existingconnection
        """

        if heating:
            self.activeConn = self.selectedSink
        else:
            self.activeConn = self.selectedSource

        self.activeConnBackup = copy(self.activeConn)            
        self.newConnection = False

        if self.activeConn is None:
            self.addConnection(heating=heating)

##        print "ModuleQ6.editConnection: conn = %s"%self.activeConn

    def addConnection(self, heating=True):
        """ before call to the connection editor for adding a new connection
        """
        
        if self.hx is not None:
            self.activeConn = HXConnection(HXId = self.hx.id, heating=heating)
            self.newConnection = True

        else:
            self.activeConn = None
            self.newConnection = False
            
##        print "\nModuleQ6.addConnection: newConnection = %s\n"%self.newConnection
        
    def getObjectDescriptor(self, conn):
        """ Gets the descriptive text of an object
        """

        if conn is None:
            return None, None, None
        
        dconn = conn.compactConnToUpstreamComponent()

##        print "ModuleQ6.getObjectDescriptor for conn: objectName %s streamName %s\n%s\ndconn: objectName %s streamName %s\n%s"%(conn.objectName, conn.streamName, conn,
##                                                                                                                                dconn.objectName, dconn.streamName, dconn)
        noStream = False

        if dconn.objectType in ["ProcessH", "ProcessC"]:
            oType = _U("Process")
            
        elif dconn.objectType in ["PipeH", "PipeC"]:
            oType = _U("Distribution Line")
            
        elif dconn.objectType in ["Equipment"]:
            oType = _U("Equipment")
            
        elif dconn.objectType in ["HX"]:
            oType = _U("Heat exchanger")
            
        elif dconn.objectType in ["WHEE"]:
            oType = _U("Electrical equipment")

        elif dconn.objectType is None:
            oType = None
            
        else:
            oType = _U("Unknown object type")
            
        if dconn.objectType is None:
            oName = None
            streamName = None
        else:
            oName = U(dconn.objectName)
            if dconn.objectNo is None:
                streamName = None
            else:
                streamName = U(dconn.streamName)

##        print "ModuleQ6.getObjectDescriptor oType %s oName %s streamName %s"%(oType, oName, streamName)
        return oType, oName, streamName

    def getObjectList(self):
        """
        Gets a list of objects corresponding to the objectType in self.activeConn
        """
        
        dconn = self.activeConn.compactConnToUpstreamComponent()

        if dconn.objectType == "HX":

            self.objectList = [(hx.HXNo-1, "HX", hx.name) for hx in Status.int.hx if hx.id != self.hx.id]

        else:

            if dconn.objectType in ["ProcessH", "ProcessC"]:
                allowedTypes = ["ProcessH", "ProcessC"]
                           
            elif dconn.objectType in ["PipeH", "PipeC"]:
                allowedTypes = ["PipeH", "PipeC"]
            else:
                allowedTypes = [dconn.objectType]
                           
            self.objectList = [(streamGlb.origin.objectNo, streamGlb.origin.objectType, streamGlb.origin.objectName) \
                               for streamGlb in Status.int.systemModel.streams if (streamGlb.origin.objectType in allowedTypes) and
                                                                                            (streamGlb.origin.heating == dconn.heating)]
            self.objectList = list(set(self.objectList))
            self.objectList =  sorted(self.objectList, key=lambda obj: obj[0])

        self.objectNoList = [entry[0] for entry in self.objectList]
        self.objectTypeList = [entry[1] for entry in self.objectList]

        if dconn.objectNo in self.objectNoList:
            index = self.objectNoList.index(dconn.objectNo)
        else:
            logDebug("ModuleQ6.getObjectList: ERROR. objectNo of active connection not in object list")
##            print "ModuleQ6.getObjectList: error building object list. dconn.objectNo %s objectNolist %s" %(dconn.objectNo, self.objectNoList)
            index = None
        
##        print "ModuleQ6.getObjectList: list after sorting %s"%self.objectNoList

        objectNames = [entry[2] for entry in self.objectList]

        return objectNames, index

    def getStreamList(self):
        """ returns a list of stream names

        keeps internally a list of the full objects self.streamGlbList for look-up
        """
        
        dconn = self.activeConn.compactConnToUpstreamComponent()

        if dconn.objectType == "HX":
            if self.activeConn.heating:
                streamNames = [_("cold side out")]
            else:
                streamNames = [_("hot side out")]

            glbName = streamNames[0]
            
            hx = Status.int.hx[dconn.objectNo]
            self.streamList = [StreamGlobals(origin=HXConnection(objectName = hx.name,
                                                                      objectType="HX",
                                                                      objectNo=hx.HXNo-1,
                                                                      componentNo=0,
                                                                      heating=dconn.heating),
                                             name = glbName)]

        else:

            self.streamList = [stream for stream in Status.int.systemModel.streams \
                               if (stream.origin.compressObjectType() == dconn.compressObjectType()) and
                               (stream.origin.objectNo == dconn.objectNo) and
                               (stream.origin.heating == dconn.heating)]
            
            streamNames = [stream.origin.streamName for stream in self.streamList]

        index = None
        if dconn.componentNo is not None:
            for i, stream in enumerate(self.streamList):
                if stream.origin.componentNo == dconn.componentNo:
                    index = i
                    break
        if index is None:
            logDebug("ModuleQ6.getStreamList: ERROR. componentNo of active connection not in stream list")
##            print "ModuleQ6.getStreamList: error building stream list dconn.objectNo %s streamlist %s" %(dconn.componentNo, streamNames)

        return streamNames, index

    def setActiveConn(self, index):
        """ for call from GUI
            sets the active connection being edited

        :params index: index in the currently displayed list self.streamList
        """
        
##        print "ModuleQ6.setActiveConn: index %s list %s"%(index, self.streamList)
        if index is not None and index >= 0 and index < len(self.streamList):
            stream = self.streamList[index]
            self.activeConn.componentNo = stream.origin.componentNo
            self.activeConn.pinchStreamName = stream.origin.pinchStreamName
##            self.activeConn.connId = stream.origin.connId
##            self.activeConn.pinchStreamId = stream.origin.pinchStreamId
            #TODO: check of correct flow of information for overwriting existing connections where possible
        else:
            self.activeConn.componentNo = None
            if index is not None:
                logDebug("ModuleQ6.setActiveConn: ERROR - called with erroneous index: %s"%index)
                logDebug("ModuleQ6.setActiveConn: streamGlbList %r"%self.streamList)
        

    def setObjectType(self, selection, no=None):
        """ for call from GUI
            sets the active object Type within the active connection
            
        :params objectType: index in the currently displayed list self.streamGlbList
        """
        if selection is None:
            logger.showWarning(_("Please select an object"))
            return False
        
        previousType = self.activeConn.objectType
        
        if selection == 0: # corresponding to the entries in OBJECTTYPES_HX
            if no is not None:
                proc = Status.int.systemModel.processObject[no]
                self.activeConn.objectType = "ProcessH" if proc.heating else "ProcessC"
            else:
                self.activeConn.objectType = "ProcessH"
        elif selection == 1:
            if no is not None:
                pipe = Status.int.systemModel.pipeObject[no]
                self.activeConn.objectType = "PipeH" if pipe.heating else "PipeC"
            else:
                self.activeConn.objectType = "PipeH"
                
        elif selection == 2:
            self.activeConn.objectType = "Equipment"
        elif selection == 3:
            self.activeConn.objectType = "WHEE"
        elif selection == 4:
            self.activeConn.objectType = "HX"
        else:
            logWarning("ModuleQ6.setObjectType: erroneous object type %s"%objectType)

##        print "ModuleQ6.setObjectType: previous %s current %s"%(previousType, self.activeConn.objectType)
        if self.activeConn.objectType != previousType:
            self.activeConn.objectNo = None
            self.activeConn.componentNo = None
            self.activeConn.fraction = 1.0
            return True
        else:
            return False

    def selectObject(self, index):
        """ for call from GUI
            sets the active object within the active connection
            
        :params index: index in the currently displayed list self.objectList
        """

        previousNo = self.activeConn.objectNo

        if index is None:
##            self.activeConn.objectType = None
            self.activeConn.objectNo = None

        else:
            objectNo = self.objectNoList[index]
            objectType = self.objectTypeList[index]
            
            self.activeConn.objectNo = objectNo
            self.activeConn.objectType = objectType # This is necessary in order to select appropriately between "ProcessC/H" or "PipeC/H"

        if self.activeConn.objectNo != previousNo:
            self.activeConn.componentNo = None
            self.activeConn.fraction = 1.0
            return True
        else:
            return False

    def storeConnection(self, fraction):
        """ for call from GUI
            stores the connection in self.activeConn
            
        :params index: index in the currently displayed list self.objectList
        """

##        print "ModuleQ6.storeConnection: Storing connection %s\n"%self.activeConn
        
        if self.activeConn.objectType is not None and \
           self.activeConn.objectNo is not None and \
           self.activeConn.componentNo is not None:


            objectNames, index = self.getObjectList()
            self.activeConn.objectName = objectNames[index]
            
            streamNames, index = self.getStreamList()
            self.activeConn.streamName = streamNames[index]

            if fraction > 0.0 and fraction <= 1.0:
                if self.activeConn.objectType == "SPLIT":
                    if self.activeConn.heating:
                        splitter = Status.int.sinkSplitters[self.activeConn.objectNo]
                        splitter.fractions[self.activeConn.componentNo] = fraction
                    else:
                        splitter = Status.int.sourceSplitters[self.activeConn.objectNo]
                        splitter.fractions[self.activeConn.componentNo] = fraction

                    splitter.checkFractions()
                else:
                    self.activeConn.fraction = fraction
            else:
                showWarning(_("Erroneous fraction of mass flow. Please correct your data entry"))
                return False
        
            if self.activeConn.heating:
                if self.newConnection:
                    self.selectedSinkNo = len(self.hx.sinkstreams)
                    self.hx.sinkstreams.append(self.activeConn)
                    self.selectedSink = self.activeConn
                else:
                    self.hx.sinkstreams[self.selectedSinkNo] = self.activeConn
            else:
##                print "ModuleQ6.storeConnection: len(srcStreams) %s"%len(self.hx.sourcestreams)
                if self.newConnection:
                    self.selectedSourceNo = len(self.hx.sourcestreams)
                    self.hx.sourcestreams.append(self.activeConn)
                    self.selectedSource = self.activeConn
                else:
                    self.hx.sourcestreams[self.selectedSourceNo] = self.activeConn

##            print "ModuleQ6.storeConnection: connection stored %s. Returning\n"%self.activeConn
            return True

        else:
            showWarning(_("Specification not complete. Please correct your data entry"))
            return False

    def changeConnection(self, oTypeDS, oNoDS, heating, hxDS=None):
        """
        changes connections of downStream object for current HX
        oType can be a downstream HX or a downstream splitter

        the previous connection is in backup, so that the previous setting can be restored in case the heat exchanger is not stored,
        """
        if oTypeDS=="HX":
            hxDS = Status.int.hx[oNoDS]
            if heating:
                conn = hxDS.sinkstreams[0]
                self.sinkstreamDS = conn
                self.sinkstreamDS_bak = conn.copy()
            else:
                conn = hxDS.sourcestreams[0]
                self.sourcestreamDS = conn
                self.sourcestreamDS_bak = conn.copy()
            conn.objectType="HX"
            conn.objectNo=self.hx.HXNo-1
            conn.componentNo=0
            conn.heating = heating
            conn.fraction = 1.0
            self.hxDS = [hxDS]
            
        elif oTypeDS=="SPLIT":
            splitter = Status.int.sinkSplitters[oNoDS] if heating else Status.int.sourceSplitters[oNoDS]
            conn = splitter.mainStream
            conn.objectType="HX"
            conn.objectNo=self.hx.HXNo-1
            conn.componentNo=0
            conn.heating = heating
            conn.fraction = 1.0
            self.hxDS = [Status.int.hx[code[1]] for code in hxDS]
        else:
            logger.logError("erroneous object (%s) for downstream connection"%oTypeDS)

    def restoreConnections(self):
        """
        undo-function for changeConnection
        """
        if self.sinkstreamDS_bak is not None:
            self.sinkstreamDS.reconnect(self.sinkstreamDS_bak)
            self.sinkstreamDS = None
            
        if self.sourcestreamDS_bak is not None:
            self.sourcestreamDS.reconnect(self.sourcestreamDS_bak)
            self.sourcestreamDS = None

    def cancel(self):
        if self.new:
            Status.int.hx.pop(-1)
        self.new = False


if __name__ == "__main__":
    print "Hello World";
