# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   moduleQM
#           
#------------------------------------------------------------------------------
#
#   Background module for data entry panel QM (monitoring data)
#           
#==============================================================================
#
#   EINSTEIN Version No.: 2.5
#   Created by:     Hans Schweiger
#
#   Update No. 001
#
#                   
#------------------------------------------------------------------------------     
#   (C) copyleft energyXperts.BCN (E4-Experts SL), Barcelona, Spain 2014
#   www.energyxperts.net / info@energyxperts.net
#
#   This program is free software: you can redistribute it or modify it under
#   the terms of the GNU general public license as published by the Free
#   Software Foundation (www.gnu.org).
#
#============================================================================== 
from GUI.status import Status
from modules.messageLogger import *
from auxiliary.auxiliary import noneCheck
from lib.GUI.tools.filePicker import pickFile
from modules.ccheck.dataSet import readDataFile, DataSet, DataFile, objectCodes, loadDataSets
from constants import PROCESSPARS, PIPEPARS, EQPARS, FECPARS, MFPARS, PARCODES, DATATYPES
from lib.auxiliary.float import FloatIO

PARCHOICES = [PARCODES[key] for key in FECPARS + EQPARS + PROCESSPARS + MFPARS + PIPEPARS]

class ModuleQM(object):
    """

    Attributes:

    
    """
    def __init__(self):
        self.initPanel()
        
    def initPanel(self):
        """  functions that have to be carried out only once
        when the panel is created
        """
        Status.int.systemModel.build()  # necessary in order to get the right order of the objects
        
        # read data sets and datafile names
        self.dataSets, self.dataFiles = loadDataSets()

        # read those parameters from datafiles which are not stored in the DB
        for df in self.dataFiles:
            pd = readDataFile(df.fileName)
##            keys, targets, dataType, periods = readDataFile(df.fileName)
            df.cols = pd.keys
            if pd.targets is not None:
                df.dataType=pd.dataType

        self.deletedDataSets = []   # intermediate storage before deleting from DB

##        print "ModuleQM.initPanel: dataSets = %s"%self.dataSets
##        print "ModuleQM.initPanel: dataFiles = %s"%self.dataFiles
    
    def updatePanel(self):
        """Functions to be carried out each time when information
            to be displayed on the panel has changed
        """

        # create system model -> contains list of heat exchangers
        # to do: shift to module !!!
        logTrack("ModuleQM.updatePanel")

        pass

    def changeData(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        or modifying data in the table

        this fucntion is called from the TableEntry event handler
        """

        err = False
        nDataSets = len(self.dataSets)

        index = -1 if index is None else index

        if index > nDataSets or index < -1:
            return 1, None, None
        elif index==-1 and mode not in [1, 4]:
            return 1, None, None

##        po = self.schedule.periodIntervals[index] if (index < nPeriods and index >=0) else ProcessPeriod()


        if mode == 0:     # insert above
            self.dataSets.insert(index, DataSet())
            nDataSets += 1

            self.activeDataSet = index
            
        elif mode in [1, 2]:   # insert below
            if nDataSets == 0:
                index = -1
                
            self.dataSets.insert(index+1, DataSet())
            nDataSets += 1

            self.activeDataSet = index+1
            
        elif mode == 3:    # delete period
            ds = self.dataSets.pop(index)
            self.deletedDataSets.append(ds)

            self.activeDataSet = index-1 if index > 0 else (0 if len(self.dataSets) > 0 else None)
            
        elif mode == 4:    # load a data file, can also be done if no data row is selected yet
            fileName = pickFile()

            pd = readDataFile(fileName)

            if pd.targets is not None:
                self.dataFiles.append(DataFile(fileName=fileName,
                                               cols=pd.keys,
                                               dataType=pd.dataType))

                if nDataSets == 0:
                    index = -1

                if index==-1: # automatically insert a new data set if no data set is selected         
                    self.dataSets.insert(index+1, DataSet(df=self.dataFiles[-1], col=1))
                    nDataSets += 1
                else: # assign the new file to the selected data set
                    self.dataSets[index].df = self.dataFiles[-1]
                    self.dataSets[index].col = 1

                self.activeDataSet = index+1

        elif mode == -1:   # change of data in a dataSet
            newVals = self.readLine(data)
            if newVals:
                ds = self.dataSets[index]
                ds.parCode, ds.objectNo, ds.cNo, ds.df, ds.col, ds.scale = newVals
            else:
                logError(_("Invalid data set. Please try again"))
                err = True
    
            self.activeDataSet = index

        data, choices = self.getTable()
        return 0, data, choices

    def getTable(self):
        """ transforms the information contained in the dataSets into a table of strings for
            representation on the GUI
        """
        table = []
        choices = []
        for ds in self.dataSets:
            df = ds.df
            fileName = df.fileName if df else ""
            colName = df.cols[ds.col-1] if (df and (ds.col is not None) and df.cols is not None) else ""
            dataType = DATATYPES[df.dataType] if (df and df.dataType is not None) else ""

##            print "ModuleQM.getTable: fileName = %s col = %s dataSets %s dataCols %s"%(fileName, ds.col, self.dataSets, df.cols if df else [])
            
            objectNames, objectIds, objectIndex = objectCodes(ds.parCode)
            if objectIndex is None and (ds.objectNo is not None and ds.objectNo >= 0 and ds.objectNo < len(objectNames)):
                objectName = objectNames[ds.objectNo]
            elif objectIndex is not None and (ds.objectNo, ds.cNo) in objectIndex:
                index = objectIndex.index((ds.objectNo, ds.cNo))
                objectName = objectNames[index]
            else:
                objectName = ""

            # determine min/max valued. TODO: try-except, because in some cases is a cause for crash -> to be checked why
            try:
                ds.getData()    # needed for minMax only
                f = FloatIO(decimals = None)
                smin = f.write(ds.min)
                smax = f.write(ds.max)
                
                minMax = "%s / %s"%(smin, smax)
            except:
                minMax = "-- / --"
                
            parDescription = PARCODES[ds.parCode] if ds.parCode in PARCODES.keys() else ""
#            table.append([noneCheck(ds.parCode, ""),
            table.append([parDescription,
                          objectName,
                          fileName,
                          colName,
                          "%10.4f"%ds.scale,
                          dataType,
                          minMax])

            fileChoices = [df.fileName for df in self.dataFiles]
            colChoices = df.cols if df else []
            choices.append([PARCHOICES, objectNames, fileChoices, colChoices, None, None, None])

        return table, choices


    def readLine(self, data):
        """ reads a line of raw data from the table in the GUI
        and translates it into internal code
        (inverse of getTable)
        """
        parDescription, objectName, fileName, colName, scale, dummy1, dummy2 = data

        try:
            scale = float(scale)
        except:
            scale = 1.0

        parCode = None
        for key in PARCODES.keys():
            if PARCODES[key]==parDescription:
                parCode=key
                break
        objectNames, objectIds, objectIndex = objectCodes(parCode)
        index = objectNames.index(objectName) if objectName in objectNames else None
        if index is not None:
            objectNo, cNo = objectIndex[index]
        else:
            objectNo = None
            cNo = None

        fileNames = [df.fileName for df in self.dataFiles]
        fileNo = fileNames.index(fileName) if fileName in fileNames else None
        df = self.dataFiles[fileNo] if fileNo is not None else None
        colNames = df.cols if df else []
        col = colNames.index(colName)+1 if colName in colNames else None
        return parCode, objectNo, cNo, df, col, scale


    def saveToDB(self, data=None):
        """ saves the current data sets to the data base
        """

        # first save all data files used in the project
        usedFiles = set([ds.df for ds in (self.dataSets)])

        for df in self.dataFiles:  
            if df in usedFiles:
                df.saveToDB()
            else:
                df.deleteFromDB()

        for ds in self.dataSets: 
            ds.saveToDB()

        for ds in self.deletedDataSets:
            print "ModuleQM.saveToDB: deleting data set %s"%ds
            ds.deleteFromDB()
        self.deletedDataSets = []
        logMessage(_("%s data sets saved, %s deleted")%(len(self.dataSets), len(self.deletedDataSets)))


                        
        

if __name__ == "__main__":
    print "Hello World";
