# -*- coding: utf-8 -*-
#==============================================================================
#
#      E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#      ModuleScheduleEditor
#            
#------------------------------------------------------------------------------
#            
#      Data table for data entry
#
#==============================================================================
#       (C) energyXperts.NET (E4-Experts GmbH, Berlin, Germany)
#============================================================================== 

from datetime import date, timedelta
from modules.schedules.profiles import WeeklyProfile, getAllProfilesFromDB
from moduleScheduleViewer import periodView, scheduleView
from lib.dateTime.dateTimeTools import dateToString, stringToDate, dateToDOY, REFERENCEDATE, MONTHLYPERIODS
from modules.messageLogger import logError
from modules.schedules.periodSchedule import PeriodSchedule, ProcessPeriod, ScheduleNotFoundError
from moduleProfileSelector import ModuleProfileSelector

#==============================================================================
class ModuleScheduleEditor():
#==============================================================================
#==============================================================================

    def __init__(self, viewer, masterModule=None):
        """
        :titles     list of titles for each column (-> defines the number of columns)
        """

        self.viewer = viewer
        self.masterModule = masterModule

        self.schedule = None
        self.scheduleType = "simple"
        self.scheduleId = None

        self.moduleProfileSelector = ModuleProfileSelector(None)
        self.dbProfileTable = self.moduleProfileSelector.getProfileTable()
        self.profileSelection = []  # list of indexes of selected profiles (for representation in ProfileSelector)

        self.activePeriod = None    # index of period currently being edited
        self.defaultPlotRange = (0, 180) # default interval (in hours) to be plotted in ScheduleView)
        
    def load(self, objectId, objectName, createNew=False, defaultProfile=None):
        """ loads a schedule from the database
        """
        if objectId is None or objectName is None: # no object selected or new object -> default to simple schedule
            self.schedule = None
            self.scheduleType = "simple"
            return
        
        ### Schedule settings
        try:
            self.schedule = PeriodSchedule(objectName)
            if not createNew:                                    
                self.schedule.loadFromDB(objectId, withDetails=True)
            else:
                self.setMonthlyPeriodsTemplate(defaultProfile)
                self.schedule.loadHolidays()

        except ScheduleNotFoundError:
            self.schedule = None
            self.scheduleType = "simple"
        else:
            self.schedule.calculateGlobals()
            self.scheduleType = "detailed"
            self.scheduleId = objectId

        return self.schedule
            

        
    def changeData(self, index, mode, data=None):
        """
        modifies the period table and calculates
        default data when adding / deleting a new line
        """

        err = False
        nPeriods = len(self.schedule.periodIntervals)
##        print "ModuleScheduleEditor.changeData index %s of [%s], mode=%s"%(index, nPeriods, mode)

        if index > nPeriods or index < -1:
            return 1, None
        elif index==-1 and mode<>1:
            return 1, None
        po = self.schedule.periodIntervals[index] if (index < nPeriods and index >=0) else ProcessPeriod()


        if mode == 0:     # insert above
            self.schedule.periodIntervals.insert(index, po.copy())
            nPeriods += 1
            if index+1 < nPeriods:
                self.schedule.periodIntervals[index].profiles = self.schedule.periodIntervals[index+1].profiles

            # assign default values
                self.schedule.periodIntervals[index].stop = self.schedule.periodIntervals[index+1].start - timedelta(days=1)
            if index > 0:
                self.schedule.periodIntervals[index].start = self.schedule.periodIntervals[index-1].stop + timedelta(days=1)

            self.activePeriod = index
            
        elif mode == 1:   # insert below
            if nPeriods == 0:
                index = -1
                
            self.schedule.periodIntervals.insert(index+1, po.copy())
            nPeriods += 1
            if index >= 0:
                self.schedule.periodIntervals[index+1].profiles = self.schedule.periodIntervals[index].profiles

            # assign default values
                self.schedule.periodIntervals[index+1].start = self.schedule.periodIntervals[index].stop + timedelta(days=1)
            if index + 2 < len(self.schedule.periodIntervals):
                self.schedule.periodIntervals[index+1].stop = self.schedule.periodIntervals[index+2].start - timedelta(days=1)

            self.activePeriod = index+1
            
        elif mode == 2:   # period split
            if po is None: return
            pnew = po.copy()
            pnew.stop = po.start
            self.schedule.periodIntervals.insert(index, pnew)

            self.schedule.periodIntervals[index+1].start = pnew.stop + timedelta(days=1)

            self.activePeriod = index

        elif mode == 3:    # delete period
            self.schedule.deletePeriodInterval(index)

            self.activePeriod = index-1 if index > 0 else (0 if len(self.schedule.periodIntervals) > 0 else None)

        elif mode == -1:   # change of data in a period
            newVals = readLine(data)
            if newVals:
                start, stop, scale = newVals
                self.schedule.periodIntervals[index].start = start
                self.schedule.periodIntervals[index].stop = stop
                self.schedule.periodIntervals[index].scale = scale*100
            else:
                logError(_("Invalid data in period. Please try again"))
                err = True
    

            self.activePeriod = index

        self.setActivePeriod(self.activePeriod)  # to be called here in order to update the default plot range
        
        # call display function of viewer in order to update the GUI
        self.schedule.build()
        self.schedule.calculateGlobals()
        self.viewer.display()
        self.exportChanges()

        choices = None
        
        return 0, self.getTable(), choices


    def save(self, processId = None):
       
##        # save detailed schedule parameters
####        self.periodSchedule.startup      = self.page2.StartupFloatEntry.GetValue()
####        self.periodSchedule.inflow       = self.page2.InflowFloatEntry.GetValue()
####        self.periodSchedule.outflow      = self.page2.OutflowFloatEntry.GetValue()
####        self.periodSchedule.tolerance    = self.page2.ToleranceFloatEntry.GetValue()
##        
##        self.schedule.holidayScale = self.page2.HolidaySpinCtrl.GetValue()
##        
####        self.periodSchedule.updateToleranceOffset()

        err = self.schedule.check()
        if not err:
            if processId:
                self.scheduleId = processId
            self.schedule.saveToDB(self.scheduleId, withHolidays=True, saveAll=True)
            return True
        else:
            logError(err)
            return False

    def setMonthlyPeriodsTemplate(self, defaultProfile):
        bakActivePeriod = self.activePeriod
        for m, month in enumerate(MONTHLYPERIODS):
            start, stop = month
            po = ProcessPeriod(start=start, stop=stop)
            self.schedule.periodIntervals.append(po)

            if defaultProfile is not None:  
                self.setActivePeriod(m)
                self.selectProfiles(defaultProfile)
                
        self.schedule.build()
        self.activePeriod = bakActivePeriod  # restore previous state
        
        
#-----------------------------------------------------------------------------------------
# auxiliary functions creating information for representation on the GUI
#-----------------------------------------------------------------------------------------

    def getTable(self):
        """ transforms the information contained in the periods into a table of strings for
            representation on the GUI
        """
        table = []
        for p in self.schedule.periodIntervals:
            startString = dateToString(p.start)
            stopString = dateToString(p.stop)
            scaleString = "%6.4f"%(p.scale/100)
            profileString = ";".join([profile.name for profile in p.profiles])

            table.append([startString, stopString, scaleString, profileString])
        return table

    def getPeriodView(self):
        PVf, PVt = periodView(self.schedule.periodIntervals, self.schedule.holidayIntervals, self.schedule.holidayScale)
        return PVf, PVt

    def getScheduleView(self):
        start, stop = self.defaultPlotRange
        SVf, SVt = scheduleView(self.schedule.scaleIntervals, start, stop)
        return SVf, SVt
        
    def getProfileSelection(self):
        if self.activePeriod >= 0 and self.activePeriod < len(self.schedule.periodIntervals):
            profiles = self.schedule.periodIntervals[self.activePeriod].profiles
            profileIDs = [profile.id for profile in profiles]

            selection = [self.moduleProfileSelector.dbProfiles.index(profile) for profile in \
                         self.moduleProfileSelector.dbProfiles if profile.id in profileIDs]
##            print "TEST TEST ModuleScheduleEditor.getProfileSelection: selection = %s"%selection
            return selection
                
        else:
##            print "TEST TEST ModuleScheduleEditor.getProfileSelection: no Period active"
            return []

#-----------------------------------------------------------------------------------------
# exectution of GUI actions
#-----------------------------------------------------------------------------------------

    def selectProfiles(self, selection):
        """ called from Profile-Selector Pop-Up
        Stores the values selected on the GUI
        """
        # first eliminate non-existing indices from the GUI selection
        checked = [index for index in selection if (index < len(self.moduleProfileSelector.dbProfiles) and index >= 0)]
##        print "TEST TEST ModuleScheduleEditor.selectProfiles: selection = %s\nchecked = %s"%(selection, checked)

        self.schedule.periodIntervals[self.activePeriod].profiles = [self.moduleProfileSelector.dbProfiles[index] for index in checked]
        self.schedule.build()
        self.exportChanges()

    def setActivePeriod(self, index):
        if index >=0 and index < len(self.schedule.periodIntervals):
            self.activePeriod = index
##            print "TEST TEST ModuleScheduleEditor.setActivePeriod: set to %s"%(index)

            # determine default plot interval
            period = self.schedule.periodIntervals[index]
            startDay = dateToDOY(period.start)
            stopDay = dateToDOY(period.stop)
##            print "ModuleScheduleEditor.setActivePeriod: period %s startDay %s stopDay %s"%(period, startDay, stopDay)
            weekStart = int(startDay/7)
            weekStop = int(stopDay/7)

            startDayPlot = (weekStart)*7
            if weekStop > weekStart and startDay > startDayPlot and stopDay >= startDayPlot + 7 + 6: # shift to an entire week if possible
                startDayPlot += 7
            else:
                startDayPlot = startDay
            startHourPlot = startDayPlot*24
            
            start = max(startHourPlot - 12, 0)
            stop = min(startHourPlot + 168 + 12, 8760)

            self.defaultPlotRange = (start, stop)

            return True
        else:
            return False

    def exportChanges(self):
        """ exports changes to the master module (ModuleQ3) """
        if self.viewer.changeData:
            self.schedule.calculateGlobals()
            print "ModuleScheduleEditor.exportChanges: HPerDay = %s"%self.schedule.HPerDay
            self.viewer.changeData()
##        if self.masterModule:
##            self.masterModule.changeInSchedule(self.schedule)

def checkPeriod(start, stop, scale):
    """ checks an individual period for consistency """
    if start is None or stop is None or scale is None or \
       stop < start or \
       start < date(2007, 1, 1) or \
       stop > date(2007, 12, 31) or \
       scale < 0:
            return False
    else:
        return True


class PeriodData():
    """ simplified period class for mapping from
        real period data
    """

    def __init__(self):
        self.start = REFERENCEDATE
        self.stop = REFERENCEDATE + timedelta(364)
        self.scale = 1.0
        self.profiles = ["myProfile"]

    def copy(self):
        new = PeriodData()
        new.start = self.start
        new.stop = self.stop
        new.scale = self.scale
        new.profiles = self.profiles[:]
        return new


def readLine(s):
    """ reads in a list of strings
    note: profile names in s[3] are not read !!!
    """
    start = stringToDate(s[0])
    stop = stringToDate(s[1])
    scale = float(s[2])

    if checkPeriod(start, stop, scale):
        return(start, stop, scale)
    else:
        return None





