# Helper functions

def copyObject(obj):
    """ creates a copy of an object which has either a copy function
    or is a list or a tuple
    """

    if obj is None: return None
    
    try:
        return obj.copy()
    except:
        return obj[:]

def insertDefault(upper, lower):
    return copyObject(lower) if upper is None else copyObject(upper)

def splitDefault(row):
    return row, copyObject(row)

def readDefault(obj, data):
    return None if data is None else data[:]

# Helper functions

class ModuleTableEntry():
    """
    Provides a generic changeData - helper function which allows to manage default table
    actions (insert above/below, split, delete rows)
    
    split: function obj0, obj1 = split(obj) to be called when a row is split
    insert: function obj = insert(obj0, obj1) to be called when a row is inserted
    read: function obj = read(obj, row) to modify an object depending on data read from table
    """
    def __init__(self, insert=insertDefault, split=splitDefault, read=readDefault, units=None):
        self.insert = insert
        self.split = split
        self.read = read

        # properties for tables with unit selection/conversion
        self.units = units
        self.cells = []

        self.table = []

    def setTable(self, table):
        """
        sets the original table (list of objects) in the background of the data
        displayed on the table

        self.table can be linked to any external 
        """
        self.table = table

    def changeData(self, index, mode, data=None):

        """
        generic helper function for data change in a table
        
        objects in the table should have a copy()-function or alternatively a list-like behaviour
        (copy with [:]-assignment)

        Note: the return values of this helper do NOT correspond to the list of return values
        as required for a changeData-function in TableEntry.
        The additional arguments (data, choices, selection) should be calculated in the
        customised change data function.
        """

        print "ModuleTableEntry: index %s mode %s data %s"%(index, mode, data)
        
        err = False
        nTable = len(self.table)

        if index is None and mode in [0, 1] and len(self.table)==0:
            new = self.insert(None, None)
            self.table.insert(0, new)
            print "ModuleTableEntry: self.table after inserting data %s"%self.table
            return 0, 0
        elif index > nTable or index < -1:
            return 1, None
        elif index==-1 and mode<>1:
            return 1, None
        
        elem = self.table[index] if (index < nTable and index >=0) else None 
        
        if mode == 0:     # insert above
            lower = elem
            upper = self.table[index-1] if index > 0 else None
            
            new = self.insert(upper, lower)
            self.table.insert(index, new)

            nTable += 1
            activeElem = index
            
        elif mode == 1:   # insert below
            lower = self.table[index+1] if index < nTable-2 else None
            upper = elem

            new = self.insert(upper, lower)

            self.table.insert(index+1, new)
            nTable += 1

            activeElem = index+1
            
        elif mode == 2:   # period split
            if elem is None: return
            upper, lower = self.split(elem)
            
            self.table.insert(index, upper)
            self.table[index+1] = lower
            nTable +=1

            activeElem = index

        elif mode == 3:    # delete period
            self.table.pop(index)
            nTable -=1

            activeElem = index-1 if index > 0 else (0 if len(self.table) > 0 else None)

        elif mode == -1:   # change of data in a period
            new = self.read(elem, data)
            if new is not None:
                self.table[index] = new
            else:
                logError(_("Invalid data in table. Please try again"))
                err = True

            activeElem = index

        
        return err, activeElem


