# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    HeatFlow
#            
#------------------------------------------------------------------------------
#            
#    Class HeatFlow: defines a heat flow Q_T(T) at a given time step
#       Class HeatFlow_t: defines a vector (8760 h) of heat flows Q_Tt(T,t)
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from fluids import Fluid
from GUI.status import *
import numpy as np
from constants import INFINITE, TemperatureInterval, TemperatureIntervalC, MinimumTemperatureC, EPSILON
from auxiliary.auxiliary import findFirstGE
from modules.messageLogger import logTrack, logWarning, logDebug
from warnings import warn
from math import ceil, floor
from modules.hf.heatFlowQ_T import getTIndex

EPS = 1.e-6   # accuracy for condensation temperature

def floorWithLimits(tempIndex):
    """ python floor function
        extended by a tolerance EPS
        and checking of limits
    """
    result = int(floor(tempIndex + EPS))
    result = min(result, Status.NT + 1)
    result = max(result, 0)
    return result

def ceilWithLimits(tempIndex):
    """ python ceil function
        extended by a tolerance EPS
        and checking of limits
    """
    result = int(ceil(tempIndex - EPS))
    result = min(result, Status.NT + 1)
    result = max(result, 0)
    return result


def createHeatFlow_tDict(HeatFlow, schedule, HeatFlow_t, ProcessID):
    """
    Function that is to be called from anywhere in the tool, 
    whenever the data any process of the present alternative is changed.
    HeatFlow can be UPHm, UPHs, QWHEE, QWH
    """
    # CommentsHS:
    # what is this "hours_nominal=0.1" ??? avoiding 0 divisions should not be at the price of introducing (even small) numeric errors !!!
    if HeatFlow is not None:
        hours_nominal=0.1
        for elem in schedule:
            hours_nominal += elem
            
        list = []
        sum=0
        
        if type(HeatFlow) == type([]):
            for hf in HeatFlow:
                l = []
                sum = 0
                for elem in schedule:
                    l.append((elem*hf)/hours_nominal)
                    sum+=(elem*hf/(hours_nominal))
                list.append(l)
        else:
            for elem in schedule:
    #            try:
                list.append(elem*HeatFlow/(hours_nominal))
                sum +=(elem*HeatFlow/(hours_nominal))
    #            except: list.append(-1)
        if HeatFlow_t == None:
            return list
        HeatFlow_t[ProcessID] = list

##class HeatFlow(object):
##    """Offer the ..."""
##    def __init__(self, Q_Tt=None, fluid_id=None, T0=0, T1=0):
##        if Q_Tt is not None and fluid_id is not None:
##            # TODO: ... here I should convert Q_Tt into HeatFlow
##            pass
##        else:
##            self.id = None
##            self.Q = 0
##            self.T0 = T0
##            self.T1 = T1
##            self.C = 0
##            self.part = []  # list of heat flow components
##            self.Q_T = Status.int.createQ_T() # Q_T representation of heat flow
##
##    def QT(self):
##        # convert HeatFlow object to QT array
##        pass
##
##    def add(self, Qnew):
##        # adds a new component to a heat flow
##
##        if self.part == [] and self.Q > 0:
##            self.part.append(self)
##        self.part.append(Qnew)
##
##        self.Q += Qnew.Q
##        self.T0 = min(self.T0, Qnew.T0)
##        self.T1 = max(self.T1, Qnew.T1)
##        self.C = None
##        self.id = None
##
##    def clear(self):
##        self.Q = 0
##        self.C = 0
##        self.Q_T = Status.int.createQ_T()


class HeatFlow_t(object):
    """Temporarily used for Step 2 simulation algorithm, as an alternative
    for the 8760 x 82 size Q_Tt arrays.
    Use the Q_Tt function to get the full Q_Tt array back, or as an
    alternative use the T or t functions to get only the distributions
    of a specific time or temperature index, or the combined Tt function
    for a single entry.
    
    """

    def __init__(self, fluid_id=None, T0=None, T1=None, Q_t=None,name="", objectNo=None, componentNo=None, shift=0):
        """Set the properties of the heatflow object. These properties
        fixed at this time and not open to changes afterwards. 
        
        If the sum of Q_t equals 0 the heatflow is initialized as a 
        constant 0 function without any other properties. Q_t, T0 and 
        T1 are then kept as None and any output will be 0 or a 0 array
        where sensible. 
        
        :param fluid_id: The database id of the used medium.
        :type fluid_id: Integer
        :param T0: The lower of the two temperatures.
        :type T0: Either a float for a constant temperature or a Q_t 
        array with the temperature over time.
        :param T1: The higher of the two temperatures.
        :type T1: Either a float for a constant temperature or a Q_t 
        array with the temperature over time.
        :param Q_t: The time array that signifies the activity of the
        :type Q_t: Numpy 1d-array heatflow over the time. 
        
        """
        self.name = name

        # optional markers that help to associate a HeatFlow-object to its origin.
        self.objectNo = objectNo
        self.componentNo = componentNo
        self.shift = shift # used for treatment of DT in externally heated streams
        
        self.singlePartName = None # used only as intermediate storage for special case where parts are sequentially added to a zero HF
        if fluid_id is None:
            self.fluid = Fluid(0)
            self.id = 0
        else:
            self.fluid = Fluid(fluid_id)
            self.id = fluid_id
        
        if all([param is not None for param in (Q_t, T0, T1)]):
            self.Q_t = Q_t.copy()
            self.Q = sum(Q_t)

            self.part = [] # a list to keep multiple heat flows if necessary

            if self.Q == 0:
                self.clear()
            else:
                if np.isnan(T0).any() or np.isinf(T0).any():
                    warn("T0 may not be NaN or Inf but is: %s. Defaulting these entries to 0." % T0.tolist(), stacklevel=2)
                    logWarning("nan- or inf-Error for heatFlow object %s"%self.name)
                    T0[np.isnan(T0)] = 0
                    T0[np.isinf(T0)] = 0
                if np.isnan(T1).any() or np.isinf(T1).any():
                    warn("T1 may not be NaN or Inf but is: %s. Defaulting these entries to 0." % T0.tolist(), stacklevel=2)
                    logWarning("nan- or inf-Error for heatFlow object %s"%self.name)
                    T1[np.isnan(T1)] = 0
                    T1[np.isinf(T1)] = 0
                self.T0 = T0
                self.T1 = T1

                self.state = "C"

                if not hasattr(T0, "__iter__"):  # TODO: should be controlled also for varying temperature ... !!!
                    if T0 == self.fluid.TCond + self.shift:
                        logTrack("WARNING: HeatFlow_t object ignores enthalpy at condensing temperature)")
                    elif T0 > self.fluid.TCond + self.shift:
                        self.state = "G"
                        
                else:
                    if min(self.T0) > self.fluid.TCond + self.shift:
                        self.state = "G"
                    

                if not hasattr(T1, "__iter__"):  # TODO: should be controlled also for varying temperature ... !!!
                    if T1 == self.fluid.TCond + self.shift:
                        logTrack("WARNING: HeatFlow_t object ignores enthalpy at condensing temperature)")
                    elif T1 < self.fluid.TCond + self.shift:
                        self.state = "L"

                else:
                    if max(self.T1) < self.fluid.TCond + self.shift:
                        self.state = "L"

            
            if np.any(self.getT0() > (self.getT1() + EPSILON)):
                warning_message = "HeatFlow_t (init): error in T0/T1 data entry - T0 > T1 at it = %s"\
                     "\n Heat flow object name: %s"\
                     "\n The complete values are: T0:%s, T1:%s"\
                     "\n A zero heatflow is created instead." \
                     % (np.where(self.getT0() > self.getT1()),self.name,self.T0, self.T1)
                warn(warning_message, stacklevel=2)
                logTrack("WARNING: "+warning_message)
                
                self.clear()
                return

                
            self.part = [] # a list to keep multiple heat flows if necessary
        elif all([param is None for param in (Q_t, T0, T1, fluid_id)]):
            self.clear()
            self.id = None
        else:
            nones = []
            others = []
            if Q_t is None:
                nones.append("Q_t")
            else:
                others.append("Q_t")
            if T0 is None:
                nones.append("T0")
            else:
                others.append("T0")
            if T1 is None:
                nones.append("T1")
            else:
                others.append("T1")
            if fluid_id is None:
                nones.append("fluid_id")
            else:
                others.append("fluid_id")
            nones = ", ".join(nones)
            others = "," .join(others)
            raise ValueError("Inconsistent parameters, couldn't create the object. %s are None while %s are not" % (nones, others))

    def Q_Tt(self, mode="heating", descending=None):
        """Return the heat flow as an Q_Tt array.

        :param mode: If "heating" return the heatflow over the 
        Interfaces.T scale, otherwise over the Interfaces.TC scale. 

        """
        Q_Tt = Status.int.createQ_Tt()
        if self.Q == 0:
            pass
        else:
            for it in xrange(Status.Nt):
                Q_Tt[:, it] = self.T(it, mode, descending)
        return Q_Tt

        #TODO for speed-up: explicit formulation of this function may avoid calculating same things 8760 times.
    def __repr__(self):
        """Return a human readable representation of the object."""
        if isinstance(self.T0, np.ndarray):
            T0 = "<%s>"%np.average(self.T0, weights=self.Q_t)
        else:
            T0 = self.T0
        if isinstance(self.T1, np.ndarray):
            T1 = "<%s>"%np.average(self.T1, weights=self.Q_t)
        else:
            T1 = self.T1
        result = "HeatFlow_t %s[%s] Q: %s [%s->%s] " % (self.name, self.componentNo, self.Q, T0, T1)
        for part in self.part:
            if part is not self:
                result += "\n--\t %s" % part
        if self.singlePartName is not None:
            result += "--%s" % self.singlePartName
            

        return result
        
    def T(self, it, mode="heating", descending=None):
        """Return a list of the temperature dependent heat distribution 
        at time index it.
        
        :param mode: If "heating" calculate over the interfaces.T heat
        scale and aggregate with ascending temperature. Otherwise use 
        the interfaces.TC heat scale and aggregate with descending
        temperature.
        :param descending: Use a descending aggregation even if the
        heating scale is used. Without effect for the non heating
        case.
        
        """
        heating = (mode=="heating")
        if heating:
            T = Status.int.T                # pointer to temperature list
        else:
            T = Status.int.TC
            
        Q_T = Status.int.createQ_T()    # space for the result

        if self.Q == 0:
            return Q_T
        elif self.Q_t[it] == 0:
            return Q_T
        elif self.part == []:
            Tc = self.fluid.TCond + self.shift
            T0 = self.getT0(it)
            T1 = self.getT1(it)
            Q = self.Q_t[it]
            
            iT0 = floorWithLimits(getTIndex(T0+EPS, heating))
            iT1 = ceilWithLimits(getTIndex(T1-EPS, heating))

            if Tc < (T0 + EPS) or Tc > (T1 - EPS):

                if T1 > T0 + EPS:
                    C = Q / (T1 - T0)
                    Q_T[iT0:iT1] = C * (T[iT0:iT1] - T0).clip(min=0)
                else:
                    Q_T[iT0] = 0.0
                    
                Q_T[iT1:] = Q
                                        
                if iT0 == iT1 and descending:
                    Q_T[iT0] = 0
                                        
            else:
                iTcH = ceilWithLimits(getTIndex(Tc-EPS, heating)) #min: security for avoiding iterations out of range in for loops below.
                iTcL = floorWithLimits(getTIndex(Tc+EPS, heating))
                iTc = iTcH
                QL = self.fluid.cp * (Tc - T0)
                QC = self.fluid.hL                  #TODO: possible cases of partial evaporation / condensation are not considered
                QV = self.fluid.cpG * (T1 - Tc)
                q = QL + QC + QV

                if q > 0:      
                    f = Q / q
                    QL *= f
                    QC *= f
                    QV *= f
                else:       # security feature for strange case: (T0 = T1 = Tc or cp = cpV = 0) and hL = 0. check not necessary if this strange fluid properties are avoided at entry)
                    QL = 0
                    QC = Q
                    QV = 0
                
# commented out (time saver): not necessary, as array has been created as 0 array.
##                for iT in xrange(0, iT0): 
##                    Q_T[iT] = 0.0
                CL = QL / (Tc - T0)
                Q_T[iT0:iTc] = CL * (T[iT0:iTc] - T0).clip(min=0)
                
                CV = QV / (T1 - Tc)
                Q_T[iTc:iT1] = QL + QC + CV * (T[iTc:iT1] - Tc)

                Q_T[iT1:] = Q
                    
                if iTcL == iTcH and descending:
                    Q_T[iTcL] = max(CL * (T[iTcL] - T0), 0.0)

            if mode != "heating" or descending == True: # convert to descending aggregation
                Q_T = Q - Q_T                        

        else: # here for multi-component heat flows
            if self in self.part:
                raise ValueError("Recursion detected.")
            for part in self.part:
                dQ_T = part.T(it, mode=mode, descending=descending)
                Q_T += dQ_T
            
        return Q_T

                    
    def t(self, iT):
        """Return a list of the time dependent heat distribution at
        temperature index t.
        
        """
        pass
    
    def Tt(self, iT, it):
        """Return the heat flow at temperature index iT and time index it."""
        pass

    def add(self, Qnew, forceAddZeros=False):
        """Add the Heatflow_t Qnew to the heat flow.
        
        A reference on Qnew is kept and used. Therefore if Qnew 
        changes after this add call, this HeatFlow_t changes also.
        
        """
        for containedHF in self.part + [self]:
            for newHF in [Qnew] + Qnew.part: 
                if newHF is containedHF:
                    raise ValueError("The heatflow to be added is already part of "\
                                     "this heatflow. Recursion is not supported.")
        if Qnew.Q == 0 and not forceAddZeros:
            return
        # If self would be the constant 0 we can drop the data.
        elif self.Q == 0 and not forceAddZeros:
            # copying something to 0
            self.Q = Qnew.Q
            if Qnew.Q_t is not None:
                self.Q_t = Qnew.Q_t.copy()
            else:
                self.Q_t = None
                
            if hasattr(Qnew.T0, "__iter__"):
                if isinstance(Qnew.T0, np.ndarray):
                    self.T0 = Qnew.T0.copy()
                else:
                    self.T0 = Qnew.T0[:]
            else:
                self.T0 = Qnew.T0
            if hasattr(Qnew.T1, "__iter__"):
                if isinstance(Qnew.T0, np.ndarray):
                    self.T1 = Qnew.T1.copy()
                else:
                    self.T1 = Qnew.T1[:]
            else:
                self.T1 = Qnew.T1
                
            self.fluid = Qnew.fluid
            self.id = Qnew.id
            self.state = Qnew.state
            self.singlePartName = Qnew.name
            self.componentNo = Qnew.componentNo

            self.part = []
            for item in Qnew.part:
                self.part.append(item.copy())
            
            if self in self.part:
                raise ValueError("Recursion detected.")
            return
                
        elif self.part == [] and self.Q > 0:
            # But here we have to keep it.
            self.part.append(self.copy())
            self.part[-1].name = self.singlePartName
            self.singlePartName = None
            self.componentNo = None
            
        self.part.append(Qnew)
        
        if Qnew.Q > 0:
            if self.Q > 0:
                self.Q_t += Qnew.Q_t
            else:
                self.Q_t = Qnew.Q_t.copy()
            self.Q += Qnew.Q

            if (hasattr(self.T0, "__iter__") or hasattr(self.T1, "__iter__") or 
                hasattr(Qnew.T0, "__iter__") or hasattr(Qnew.T1, "__iter__")):
                self.T0 = np.min((self.getT0(), Qnew.getT0()), 0)
                self.T1 = np.max((self.getT1(), Qnew.getT1()), 0)
                
            else:
                self.T0 = min(self.T0, Qnew.T0)
                self.T1 = max(self.T1, Qnew.T1)
                                                     
        self.id = None  # indicates that no specific fluid associated -> mix of several!
        self.fluid = None

        if self in self.part:
            raise ValueError("Recursion detected.")


    def clear(self):
        """Set the heatflow to a constant 0 heatflow."""
        self.Q = 0
        self.T0 = None
        self.T1 = None
        self.part = []
        self.state = None
        self.Q_t = None

    def scale(self, fraction, createCopy = True):
        """Creates a copy of the heat flow applying a scale factor

        :params fraction: scale factor (fraction of 1)
        :type fraction: double

        :params createCopy: if False -> changes and overwrites current HeatFlow_t object
        
        A reference on Qnew is kept and used. Therefore if Qnew 
        changes after this add call, this HeatFlow_t changes also.
        
        """

        if fraction == 1:
            #TODO: CHECK - no copy created in the case of f = 1, as the original object is not modified.
            return self
        else:
            if createCopy:
                newQ = self.copy()
                if newQ.Q_t is not None: newQ.Q_t *= fraction
                newQ.Q *= fraction
                for part in newQ.part:
                    part.scale(fraction,createCopy)
                return newQ
            else:
                if self.Q_t is not None: self.Q_t *= fraction
                self.Q *= fraction
                for part in self.part:
                    part.scale(fraction,createCopy)
                return self
        
        return newQ
        
    def getQ(self, it=None):
        """Return the total heatflow at a certain interval it (or the 
        whole vector if it == None)
        
        """
        if it is None:
            if self.Q == 0:
                return Status.int.createQ_t()
            else:
                return self.Q_t.copy()
        else:
            if self.Q == 0:
                return (0)
            else:
                return self.Q_t[it]

    def trim(self, it, T_trim, x=0):
        """
        returns the fractions of demand below T_trim
        at a given time step
        x: the fraction of condensing heat to be returned if T_trim = Tc
        """
        T0 = self.getT0(it)
        T1 = self.getT1(it)

        if T_trim < T0 - EPSILON:
            return 0
        elif T_trim > T1 + EPSILON:
            return 1.0

        Tc = self.fluid.TCond
        if Tc >= T0 - EPSILON and Tc <= T1 + EPSILON: # condensing case
            fL, fC, fG = self.getFractions(T0, T1, Tc)
            if T_trim < Tc + EPSILON:
                if fL > 0:
                    f0 =  fL * (T_trim - T0)/(Tc - T0)
                else:
                    f0 = 0
            elif T_trim > Tc + EPSILON:
                f0 = fL + fC
                if fG > 0:
                    f0 += fG*(T_trim - Tc)/(T1 - Tc)
            elif abs(T_trim-Tc) >= EPSILON:
                f0 = fL + fC*x
                    
        elif T1 > T0:   # simple linear case, no condensation
            f0 = (T_trim - T0)/(T1 - T0)
        else: # T1 <= T0 - should not happen
            return None

        return f0
            

    def getQ_t(self):
        return self.getQ()
            
    @property
    def dQavg(self):
        """Average energy per time step over all time steps with dQ > 0
        """
        
        if self.Q == 0:
            return 0
        else:     
            w = (1*(self.Q_t > 0))
            if np.sum(w) == 0:
                return 0.0
            else:
                return np.average(self.Q_t,weights=w)

    @property
    def hop(self):
        """ real operating hours
        (note: hourly fav-vector is used as base for calculation !)
        """
        
        if self.Q == 0:
            return 0
        else:     
            w = (1*(self.Q_t > 0))
            return sum(w)*Status.TimeStep

    @property
    def T0avg(self):
        """Average energy per time step over all time steps with dQ > 0
        """
        
        if hasattr(self.T0, "__iter__"):
            if self.Q > 0:
                return np.average(self.T0, weights=self.Q_t)
            else:
                return np.mean(self.T0)
        else:
            return self.T0

    @property
    def T1avg(self):
        """Average energy per time step over all time steps with dQ > 0
        """
        
        if hasattr(self.T1, "__iter__"):
            if self.Q > 0:
                return np.average(self.T1, weights=self.Q_t)
            else:
                return np.mean(self.T1)
        else:
            return self.T1

    def getC(self, it=None, cond = False):
        """Returns a vector of C = Q(DT) or the corresponding value at 
        time interval it.
        
        """        
        if cond == False:
            if self.Q == 0: return None
            if it is None:
                result = self.Q_t/(self.T1 - self.T0)
            else:
                result = self.Q_t[it]/(self.getT1(it) - self.getT0(it))
        elif it is None:
            if self.Q == 0: return Status.int.createQ_t(), Status.int.createQ_t(), Status.int.createQ_t()
            
##            print "TEST TEST getC in cond mode, state = ",self.state
            if self.state == "L":
                result =  (None, None, self.Q_t/(self.T1 - self.T0))
            elif self.state == "G":
                result =  (self.Q_t/(self.T1 - self.T0), None, None)
            else:
                Tc = self.fluid.TCond + self.shift
                CL = Status.int.createQ_t()
                QC = Status.int.createQ_t()
                CG = Status.int.createQ_t()
                
                for it in xrange(Status.Nt):
                    if self.Q_t[it] == 0:
                        CL[it] = 0
                        QC[it] = 0
                        CG[it] = 0
                    else:
                        qL = self.fluid.cp * (Tc - self.getT0(it))
                        qC = self.fluid.hL                  #TODO: possible cases of partial evaporation / condensation are not considered
                        qG = self.fluid.cpG * (self.getT1(it) - Tc)
                        q = qL + qC + qG

                        if qL + qG > 0:      
                            f = self.Q_t[it] / q
                            qL *= f
                            qC *= f
                            qG *= f
                        else:       # security feature for strange case: (T0 = T1 = Tc or cp = cpV = 0) and hL = 0. check not necessary if this strange fluid properties are avoided at entry)
                            qL = 0
                            qC = self.Q_t[it]
                            qG = 0
                            
                        if (qL > 0): CL[it] = qL/(Tc - self.getT0(it))
                        else: CL[it] = 1.
                        QC[it] = qC
                        if (qG > 0): CG[it] = qG/(self.getT1(it) - Tc)
                        else: CG[it] = 1.
                        #TODO: check if this CG / CL = 1 gives problems
                        #TODO: check special case partial condensation / evaporation
                        
                    if it < 24:
                        logDebug("TEST TEST getC it %s qt %s CG %s QC %s CL %s" % 
                                 (it, self.Q_t[it], CG[it], QC[it], CL[it]))                               
                    
                result = (CG, QC, CL)
                 
        else:
            logWarning("getC currently not working for one time step if condensation is specified")
            result = None
            

        return result

    def getFractions(self, T0, T1, Tc):
        """
        returns the fractions L-C-G (liquid, condensation, gas) 
        gives valid results only if T0 < Tc < T1 (to be checked externally)
        """
        qL = self.fluid.cp * (Tc - T0)
        qC = self.fluid.hL                  #TODO: possible cases of partial evaporation / condensation are not considered
        qG = self.fluid.cpG * (T1 - Tc)
        q = qL + qC + qG

        if q > 0:
            return qL/q, qC/q, qG/q
        else:
            return 0, 0, 0

    def getT0(self, it=None):
        """Return the temperature at a certain interval it (or the 
        whole array if it == None).
        
        """
        if hasattr(self.T0, "__iter__"):
            if it is None:
                return self.T0
            else:
                return self.T0[it]
        elif self.T0 is None:
            if it is None:
                return Status.int.createQ_t()
            else:
                return 0
        else:
            if it is None:
                return Status.int.createQ_t() + self.T0
            else:
                return self.T0
        
    def getT1(self, it=None):
        """Return the temperature at a certain interval it (or the 
        whole array if it == None).
        
        """
        if hasattr(self.T1, "__iter__"):
            if it is None:
                return self.T1
            else:
                return self.T1[it]
        elif self.T1 is None:
            if it is None:
                return Status.int.createQ_t() 
            else:
                return 0
        else:
            if it is None:
                return Status.int.createQ_t() + self.T1
            else:
                return self.T1

    def getPart(self, no = None, name = None):
        """Returns a part of the heat flow object
            Numbering from no = 0 to len(part)-1
        """
        if no is None:
            if name is not None:
                for part in self.part:
                    if part.name == name: return part

                if self.name == name: return self
                
                logTrack("HeatFlow_t: part with name %s not found in HeatFlow_t object %s"%(name,self.name))
                return HeatFlow_t()
            else:
                return None
        elif no >= len(self.part) or no < 0: return self
        else: return self.part[no]

    def copy(self):
        """Return a heatflow_t object with the same properties as this one."""
        Q = HeatFlow_t()
        Q.Q = self.Q
        if self.Q_t is not None:
            Q.Q_t = self.Q_t.copy()
        else:
            Q.Q_t = None
            
        if hasattr(self.T0, "__iter__"):
            if isinstance(self.T0, np.ndarray):
                Q.T0 = self.T0.copy()
            else:
                Q.T0 = self.T0[:]
        else:
            Q.T0 = self.T0
        if hasattr(self.T0, "__iter__"):
            if isinstance(self.T0, np.ndarray):
                Q.T1 = self.T1.copy()
            else:
                Q.T1 = self.T1[:]
        else:
            Q.T1 = self.T1
            
        Q.fluid = self.fluid
        Q.id = self.id
        Q.state = self.state
        Q.name = self.name
        Q.componentNo = self.componentNo

        for item in self.part:
            Q.part.append(item.copy())
        
        return Q

    def split(self, TSplit):
        """Divides a heat-flow object into two parts below/above TSplit"""

        if len(self.part) > 0:
            logWarning("HeatFlow_t.split: not yet implemented for combined heat flows")
            return (None, None)

        qL = self.copy()
        qH = self.copy()

        if self.Q > 0:
            q = self.getQ()
            T0 = self.getT0()
            T1 = self.getT1()
            
            qL.T1 = Status.int.createQ_t()
            qH.T0 = Status.int.createQ_t()

            Tc = self.fluid.TCond + self.shift

            for it in xrange(Status.Nt):

                if q[it] > 0:
                    TSplit_it = max(min(TSplit,T1[it]), T0[it]) # constrained split temperature within limits of T0, T1
                    if (Tc < (T0[it] + EPS) or Tc > (T1[it] - EPS)) and (T0[it] < T1[it]): # no condensation + security check T0 < T1

                        C = q[it] / (T1[it] - T0[it])
                                        
                        qL.Q_t[it] = C * (TSplit_it - T0[it])
                        qH.Q_t[it] = C * (T1[it] - TSplit_it)

                        qL.T1[it] = TSplit_it
                        qH.T0[it] = TSplit_it
                            
                    else:
                        CL = self.fluid.cp
                        QC = self.fluid.hL                  #TODO: possible cases of partial evaporation / condensation are not considered
                        CV = self.fluid.cpG * (T1[it] - Tc)
                        
                        if TSplit > Tc: #condensation part goes to qL
                            qL.Q_t[it] = CL * (Tc - T0[it]) + QC + CV * (TSplit_it - Tc)
                            qH.Q_t[it] = CV * (T1[it] - Tc)                                
                        else: # condensation part goes to qH
                            qL.Q_t[it] = CL * (TSplit_it - T0[it])
                            qH.Q_t[it] = CL * (Tc - TSplit_it) + QC + CV * (T1[it] - Tc)

                    qL.T1[it] = TSplit_it
                    qH.T0[it] = TSplit_it
                                               
                else:
                    qL.T1[it] = TSplit
                    qH.T0[it] = TSplit

            qL.Q = np.sum(qL.Q_t)
            qH.Q = np.sum(qH.Q_t)

        return (qL, qH)
            

    def buildCC(self, it):
        """ Return the composite curve of heat demand
        as a list of tuples (Ti,Qi) where
        Ti[0] = Tmin, Ti[last] = Tmax
        
        """
        cc = compositeCurve()    #each cc starts at absolute 0
        if self.part == [] and self.Q > 0:
            T0 = self.getT0(it)
            T1 = self.getT1(it)
            Tc = self.fluid.TCond + self.shift
            Q = self.getQ(it)

            cc.T.add(T0)
            cc.Q.add(0)
            if Tc < T0 or Tc > T1:  #no phase change
                cc.T.add(T1)
                cc.Q.add(self.getQ(it))
            else:                   #get decomposition in liquid, latent, vapour
                QL = self.fluid.cp * (Tc - T0)
                QC = self.fluid.hL                  #TODO: possible cases of partial evaporation / condensation are not considered
                QV = self.fluid.cpG * (T1 - Tc)
                q = QL + QC + QV

                if q > 0:      
                    f = Q / q
                    QL *= f
                    QC *= f
                    QV *= f
                else:       # security feature for strange case: (T0 = T1 = Tc or cp = cpV = 0) and hL = 0. check not necessary if this strange fluid properties are avoided at entry)
                    QL = 0
                    QC = Q
                    QV = 0

                cc.T.add(Tc)
                cc.Q.add(QL)
                cc.T.add(Tc)
                cc.Q.add(QL + QC)
                cc.Tc.add(Tc)
                
                cc.T.add(T1)
                cc.T.add(Q)
            
            
        else:
            for part in self.part:
                ccpart = part.buildCC()
                addCC(cc, ccpart)

        return cc

    def exportGlb(self, streamGlb):
        """ Stores global parameters of the heat flow object into a StreamGlb object
        
        """
##        print "HeatFlow.exportGlb:\nHF %s:\n%s\nsGLB: %s\n"%(self.name, self, streamGlb)
        if isinstance(streamGlb, list):
            if len(self.part) > 0:
                for s, part in enumerate(self.part):
                    part.exportGlb(streamGlb[s])
            elif len(streamGlb)==1:
                self.exportGlb(streamGlb[0])
            else:
                assert(self.Q==0)
                
        elif streamGlb is None:
            return
        else:
            streamGlb.Q = self.Q
            streamGlb.T0 = self.T0avg
            streamGlb.T1 = self.T1avg
            streamGlb.hours = self.hop
            streamGlb.dotQavg = self.dQavg
            streamGlb.shift = self.shift

def diffQ_Tt(Q1_Tt, Q2_Tt):

        """Subtracts two heat flows so that for the differentials at each temperature interval holds
        (example: Q1: available waste heat, Q2: used waste heat (at source temperature)
        => diffQ: remaining waste heat (part of Q1 that is not used)
        => resQ:  used waste heat that came from another source (could not be allocated from Q1)

        d_diffQ(T) = max(dQ1_Tt - dQ2_Tt,0)
        d_resQ(T) = max(dQ2_Tt - dQ1_Tt,0)
        :returns: a 2-tuple of Heatflow_t objects (d_diffQ, d_resQ).

        """

        if Q1_Tt.part == []:
            if Q2_Tt.part == []:
                if Q1_Tt.componentNo != Q2_Tt.componentNo or Q1_Tt.componentNo is None or Q2_Tt.componentNo is None:
                    diffQ = Q1_Tt.copy()
                    resQ = Q2_Tt.copy()
                elif Q2_Tt.Q == 0:
                    diffQ = Q1_Tt.copy()
                    resQ = HeatFlow_t()
                elif Q1_Tt.Q == 0:
                    diffQ = HeatFlow_t()
                    resQ = Q2_Tt.copy()
                elif Q1_Tt.id == Q2_Tt.id:
                    fluid_id = Q1_Tt.id
                                                   
                    QHdiff_t = Status.int.createQ_t()
                    QHres_t = Status.int.createQ_t()

                    QXdiff_t = Status.int.createQ_t()
                    QXres_t = Status.int.createQ_t()
                                                   
                    QLdiff_t = Status.int.createQ_t()
                    QLres_t = Status.int.createQ_t()
                                                   
                    if (hasattr(Q1_Tt.T0, "__iter__") or hasattr(Q1_Tt.T1, "__iter__") or 
                        hasattr(Q2_Tt.T0, "__iter__") or hasattr(Q2_Tt.T1, "__iter__")):

                        T1H_t = Status.int.createQ_t()
                        T1X_t = Status.int.createQ_t()
                        T0X_t = Status.int.createQ_t()
                        T0L_t = Status.int.createQ_t()
                                                               
                        for it in xrange(Status.Nt):
                            dQ = calcDiffQ(Q1_Tt.fluid, Q1_Tt.getT0(it), Q1_Tt.getT1(it), Q2_Tt.getT0(it), Q2_Tt.getT1(it))
                            (q, diff, res) = dQ.calculate(Q1_Tt.Q_t[it], Q2_Tt.Q_t[it])
                                                               
                            QHdiff_t[it] = diff[0]
                            QHres_t[it] = res[0]
                            T1H_t[it] = dQ.T1H
                                                                   
                            QXdiff_t[it] = diff[1]
                            QXres_t[it] = res[1]
                            T0X_t[it] = dQ.T0x
                            T1X_t[it] = dQ.T1x
                                                                   
                            QLdiff_t[it] = diff[2]
                            QLres_t[it] = res[2]
                            T0L_t[it] = dQ.T0L

                        Qdiff_Tt = HeatFlow_t(fluid_id=fluid_id, Q_t=QHdiff_t, T0=T1X_t, T1=T1H_t, componentNo=Q1_Tt.componentNo)
                        Qdiff_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QXdiff_t, T0=T0X_t, T1=T1X_t, componentNo=Q1_Tt.componentNo))
                        Qdiff_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QLdiff_t, T0=T0L_t, T1=T0X_t, componentNo=Q1_Tt.componentNo))
                                                               
                        Qres_Tt = HeatFlow_t(fluid_id=fluid_id, Q_t=QHres_t, T0=T1X_t, T1=T1H_t, componentNo=Q1_Tt.componentNo)
                        Qres_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QXres_t, T0=T0X_t, T1=T1X_t, componentNo=Q1_Tt.componentNo))
                        Qres_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QLres_t, T0=T0L_t, T1=T0X_t, componentNo=Q1_Tt.componentNo))
                        
                        diffQ = Qdiff_Tt
                        resQ =  Qres_Tt                                    
                    else:
                        dQ = calcDiffQ(Q1_Tt.fluid, Q1_Tt.T0, Q1_Tt.T1, Q2_Tt.T0, Q2_Tt.T1)

                        for it in xrange(Status.Nt):
                            (q, diff, res) = dQ.calculate(Q1_Tt.Q_t, Q2_Tt.Q_t)
                                                               
                            QHdiff_t[it] = diff[0][0]
                            QHres_t[it] = res[0][0]
                            T1H_t[it] = diff[0][2]
                                                                   
                            QXdiff_t[it] = diff[1][0]
                            QXres_t[it] = res[1][0]
                            T0X_t[it] = diff[1][1]
                            T1X_t[it] = diff[1][2]
                                                                   
                            QLdiff_t[it] = diff[2][0]
                            QLres_t[it] = res[2][0]
                            T0L_t[it] = diff[2][1]

                        Qdiff_Tt = HeatFlow_t(fluid_id=fluid_id, Q_t=QHdiff_t, T0=dQ.T1X, T1=dQ.T1H, componentNo=Q1_Tt.componentNo)
                        Qdiff_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QXdiff_t, T0=dQ.T0X, T1=dQ.T1X, componentNo=Q1_Tt.componentNo))
                        Qdiff_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QLdiff_t, T0=dQ.T0L, T1=dQ.T0X, componentNo=Q1_Tt.componentNo))
                                                               
                        Qres_Tt = HeatFlow_t(fluid_id=fluid_id, Q_t=QHres_t, T0=dQ.T1X, T1=dQ.T1H, componentNo=Q1_Tt.componentNo)
                        Qres_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QXres_t, T0=dQ.T0X, T1=dQ.T1X, componentNo=Q1_Tt.componentNo))
                        Qres_Tt.add(HeatFlow_t(fluid_id=fluid_id, Q_t=QLres_t, T0=dQ.T0L, T1=dQ.T0X, componentNo=Q1_Tt.componentNo))
                        
                        diffQ = Qdiff_Tt
                        resQ =  Qres_Tt
                else:
                    print "Sorry, diffQ_Tt for non identical fluids does not yet work!"
                    return (Q1_Tt.copy(), Q2_Tt.copy())
                    
            else:
                diffQ = Q1_Tt.copy()
                resQ = HeatFlow_t()
                for part in Q2_Tt.part:
                    (diffQ, res) = diffQ_Tt(diffQ, part)
                    resQ.add(res)
                            
        else:
            diffQ = HeatFlow_t()
            resQ = Q2_Tt.copy()

            for part in Q1_Tt.part:
                (diff, resQ) = diffQ_Tt(part, resQ)
                diffQ.add(diff)
        
        return (diffQ, resQ)
                                                                   
class calcDiffQ():

    """" calculates the part of heat transferred between Q_Tsrc and Q_Tsnk
    as a function of Qsrc and Qsnk
    gives excess heat Qsrc-Qsnk in temp. ranges H,X,L as a list of tuples
    a,b,T0,T1 where a = coefficient for Qsrc and b = coefficient for Qsnk
    positive heat flows: -> diff, negative heat flows -> res
        
    """
    def __init__(self, fluid, T0src, T1src, T0snk, T1snk):
        """Obtain coefficients depending on temperatures."""
        T0x = max(T0src, T0snk)
        T1x = min(T1src, T1snk)

        if T0x >= T1x:   #no overlap of temperature ranges -> no exchange
            if T0src < T0snk:
                self.coefs = [(0, 1), (0, 0), (1, 0)]
            else:
                self.coefs = [(1, 0), (0, 0), (0, 1)]

            self.T0L = min(T0src, T0snk)
            self.T0x = T1x
            self.T1x = T0x
            self.T1H = max(T1src, T1snk)
            
        else:
            hsrc0 = fluid.enthalpy(T0src, x=0)
            hsrc1 = fluid.enthalpy(T1src, x=1)
                    
            hsnk0 = fluid.enthalpy(T0snk, x=0)
            hsnk1 = fluid.enthalpy(T1snk, x=1)
            h0x = fluid.enthalpy(T0x, x=0)
            h1x = fluid.enthalpy(T1x, x=1)
       
            fsrc = 1. / (hsrc1 - hsrc0)
            fsnk = 1. / (hsnk1 - hsnk0)

            self.coefs = []

            #excess source or sink in high temperature range
            if T1src > T1x:
                c = fsrc * (hsrc1 - h1x)
                self.coefs.append((c, 0))
                self.T1H = T1src
            elif T1snk > T1x:
                c = fsnk * (hsnk1 - h1x)
                self.coefs.append((0, c))
                self.T1H = T1snk
            else:
                self.coefs.append((0,0))
                self.T1H = T1x

            #excess source or sink in range of temperature overlap
            csrc = fsrc * (h1x - h0x)
            csnk = fsnk * (h1x - h0x)
            self.coefs.append((csrc, csnk))

            self.T0x = T0x
            self.T1x = T1x

            #excess source or sink in low temperature range
            if T0src < T0x:
                c = fsrc * (h0x - hsrc0)
                self.coefs.append((c, 0))
                self.T0L = T0src
                                          
            elif T0snk < T0x:
                c = fsnk * (h0x - hsnk0)
                self.coefs.append((0, c))
                self.T0L = T0snk
                
            else:
                self.coefs.append((0,0))
                self.T0L = T0x

                                      
    def calculate(self, qsrc, qsnk):
        """" returns tuples (diff,res), each containing
        (Q,T0,T1) for given values of qsrc and qsnk

        """
        diff = []
        res = []
        q = 0                      
        if len(self.coefs) == 0:
            diff = [qsrc, 0, 0]
            res = [0, 0, qsnk]
        else:
            for coef in self.coefs:
                q = coef[0] * qsrc - coef[1] * qsnk
                if q > 0:
                    diff.append(q)
                    res.append(0)
                else:
                    diff.append(0)
                    res.append(-q)
        return (q, diff, res)
         

##class compositeCurve():
##    """ composite curve is an alternative description of a heat flow
##    based on a temperature vector and a corresponding cumulative energy vector
##    in addition a Tc vector is given specifying the points of condensation / phase change
##
##    """
##    def __init__(self):
##        self.T = []
##        self.Q = []
##        self.Tc = []
##
##    def getQ(self, T, x=0):
##
##        if T > self.T[-1]:
##            return self.Q[-1]
##        elif T < self.T[0]:
##            return 0
##        else:
##            iT = findFirstGE(T, self.T)
##            if iT == 0:
##                print "HeatFlow_t - compositeCurve - getQ: here something should be improved"
##                pass            #TODO check if T0 is in Tcs ...
##            elif T == self.T[iT]:   # equal sign holds
##                if T in self.Tc:
##                    if x == 1:
##                        return self.Q[iT + 1]
##                    else:
##                        return self.Q[iT]
##                else:
##                    return self.Q[iT]
##            else:                   # T < self.T[iT]
##                TL = self.T[iT - 1]
##                TH = self.T[iT]
##                f = (T - TL) / (TH - TL)
##                Q = self.Q[iT - 1] * f + self.Q[iT] * (1 - f)                      
##
##def addCC(cc1, cc2):
##    """Return the sum of two composite curves as built by buildCC"""
##    cc = compositeCurve()
##    
##    #build composite list of condensation temperatures
##    cc.Tc = cc1[:]
##    for Tc in cc2.Tc:            
##        if Tc not in cc.Tc:
##            cc.Tc.append(Tc)
##    cc.Tc.sort()
##
##    #build composite temperature scale
##    cc.T = cc1.T[:]
##    for T in cc2.T:
##        if T not in cc1.T:
##            cc.T.append(T)
##    cc.T.sort()
##
##    #now sum up energy values
##    x = 0
##    for T in cc.T:
##        cc.Q = cc1.getQ(T, x) + cc2.getQ(T, x)
##        if T in cc.Tc:
##            x = 1 - x
##
##    return cc
##        
##def pinchCC(hcc, ccc, dTmin):
##    """ shifts hcc so that dTmin criterion is fulfilled"""
##    cc = compositeCurve()   #shifted hcc - curve
##    
##    #build composite list of condensation temperatures
##    cc.Tc = ccc[:]
##    for Tc in hcc.Tc:            
##        if (Tc - dTmin) not in ccc.Tc:
##            cc.Tc.append(Tc - dTmin)
##    cc.Tc.sort()
##
##    #build composite temperature scale (in ccc temperatures, not Tavg !!!)
##    cc.T = ccc.T[:]
##    for T in hcc.T:
##        if (T - dTmin) not in ccc.T:
##            cc.T.append(T - dTmin)
##    cc.T.sort()
##
##    #now calculate difference of energy values
##    x = 0
##    shift = 0
##    TPinch = INFINITE
##    
##    for iT in range(len(cc.T)):
##        T = cc.T[iT]
##        cc.Q.append(ccc.getQ(T, x) - hcc.getQ(T + dTmin, x))
##        if cc.Q[iT] < shift:
##            shift = cc.Q[iT]
##            TPinch = T
##            
##        if T in cc.Tc:
##            x = 1 - x
##
##    for iT in range(len(cc.T)):
##        cc.Q[iT] += shift
##    
##    return (cc, TPinch)
##
def convertTCtoTScale(Q_Tt):
    """ converts cooling demand (TC scale,descending) into waste 
    heat (T scale, descending)
    
    """
    Qconv_Tt = Q_Tt.copy()

    iTmax = int(Status.MaximumTemperatureC / Status.TemperatureInterval)
    for iT in range(1, iTmax + 1):
        T = Status.int.T[iT]
        iTC = int((T - Status.MinimumTemperatureC) / Status.TemperatureIntervalC)
        Qconv_Tt[iT] = Q_Tt[iTC]
    for iT in range(iTmax + 1, Status.NT + 1):
        Qconv_Tt[iT] = [0] * Status.Nt

    return Qconv_Tt
        
def convertTtoTCScale(Q_Tt):
    """ converts waste heat used (T scale,descending) into cooling 
    demand supplied (TC scale, descending)
    
    """
    Qconv_Tt = Q_Tt.copy()

    iTC0 = int(-Status.MinimumTemperatureC / Status.TemperatureIntervalC)
    for iTC in range(1, iTC0):
        Qconv_Tt[iTC] = Qconv_Tt[0][:]
        
    for iTC in range(iTC0, Status.NT + 1):
        TC = Status.int.TC[iTC]
        iT = int(TC / Status.TemperatureInterval)
        Qconv_Tt[iTC] = Q_Tt[iT][:]

    Qconv_Tt[Status.NT + 1] = Q_Tt[Status.NT + 1]

    return Qconv_Tt
        
        
def shiftQ_Tt(Q_Tt, DT, mode):
    """ shifts a Q_Tt - array by DT in temperature

    :param mode: heating or cooling (defines temperature scale and 
        ascending/descending

    """
    if DT == 0:
        return Q_Tt
    else:
        if mode == "heating":
            if DT < 0:
                logWarning("shiftQ_Tt: negative temperature shift for heating demands not possible")
                return Q_Tt
            else:
                diT = int(DT/TemperatureInterval)
                xiT = DT/TemperatureInterval - diT #indicates the fraction to be added at diT + 1

                Qnew0_Tt = Status.int.createQ_Tt()
                Qnew0_Tt[:diT] = Q_Tt[0]
                Qnew0_Tt[diT:-1] = Q_Tt[:-diT - 1]
                Qnew0_Tt[-1] = Q_Tt[-1]
                
                if xiT == 0:
                    return Qnew0_Tt
                else:    
                    Qnew1_Tt = Status.int.createQ_Tt()
                    Qnew1_Tt[:diT + 1] = Q_Tt[0]
                    Qnew1_Tt[diT + 1:-1] = Q_Tt[:-diT - 2]
                    Qnew1_Tt[-1] = Q_Tt[-1]
                    
                    return xiT*Qnew1_Tt + (1-xiT)*Qnew0_Tt

        else:
            if DT > 0:
                logWarning("shiftQ_Tt: positive temperature shift for cooling demands not possible")
                return Q_Tt
            else:
                diT = int(-DT/TemperatureIntervalC)
                xiT = -DT/TemperatureIntervalC - diT

                Qnew0_Tt = Status.int.createQ_Tt()
                Qnew0_Tt[0] = Q_Tt[0]
                if diT > 0:
                    Qnew0_Tt[1:-diT] = Q_Tt[diT + 1:]
                    Qnew0_Tt[-diT:] = Q_Tt[-1]
                else:
                    Qnew0_Tt[1:] = Q_Tt[1:]
                
                if xiT == 0:
                    return Qnew0_Tt
                else:    
                    Qnew1_Tt = Status.int.createQ_Tt()
                    Qnew1_Tt[0] = Q_Tt[0]
                    Qnew1_Tt[1:-diT-1] = Q_Tt[diT+2:]
                    for iT in range(diT+1):
                        Qnew1_Tt[-iT-1] = Q_Tt[-1]
                    return xiT*Qnew1_Tt + (1-xiT)*Qnew0_Tt

                             
def sharesQ_T(T0,T1,fluid,x0=0,x1=1):
    """ calculates the energy shares below, during and above condensation

    :returns a tuple of (xL,xC,xH) - fractions of energy
    
    """

    fL = max(fluid.TCond - T0,0)*fluid.cp
    fH = max(T1 - fluid.TCond,0)*fluid.cpG
    if T0 < fluid.TCond and T1 > fluid.TCond:
        fC = fluid.hL
    else:
        fC = (x1 - x0)*fluid.hL

    fsum = fL + fH + fC
    if fsum > 0:
        return (fL/fsum,fC/fsum,fH/fsum)
    else:
        return (0,1,0)

def detectStrangeValues(Q_Tt, name = "Q_Tt"):
    """ detects NaN's, None's and INF's in Q_Tt arrays

    :returns a tuple of (iT,it,value)
    
    """

    for iT in range(Status.NT+2):
        for it in range(Status.Nt):
            x = Q_Tt[iT][it]
            if np.isnan(x) or np.isinf(x) or (x is None):
                logTrack("Strange Value in %s[%s][%s]: %s"%(name,iT,it,x))
                return (iT,it,x)

    return None
    

    
