# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	HEAT PUMP - Module for heat pump selection and calculation
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Module for heat pump selection and calculation
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from math import fabs
import numpy as np

from globals import logger
from GUI.status import *
from constants import *
from GUI.GUITools import check
from modules.equipments import getConnectedPipes
from modules.modulePipes import getAllHeatingPipes
from modules.heatFlow import HeatFlow_t
from modules.interfaces import createQ_Tt
from auxiliary.auxiliary import (firstNonZero, lastNonZero,
                                          interpolateList, findInListASC, 
                                          interpolateTable)
from lib.auxiliary.auxiliary import noneCheck
from lib.physics.thermo import carnotEfficiency, carnotCOPh, carnotCOPc
from modules.messageLogger import (logTrack, logDebug, logWarning,
                                            logMessage, logError, showWarning)
from GUI.dialogGauge import DialogGauge
import database.dbio as dbio

EPS_COP = 1.e-4
EPS_TEMP = 1.e-4
EPS = 1.e-10
HPTDROP = 7
HPDTIO = 5

POWERRANGE = 0.8    #   range of power for heat pump preselection: 1 = only maximum possible power
COPCARNOTMAX = 50

DEFAULTSAUTODESIGNHP = { "UHPMaintain":1,
                         "UHPType":"compression",
                         "UHPMinHop":1500.0,
                         "UHPDTMax":60.0,
                         "UHPTgenIn":100.0,
                         "UHPmaxT":100.0,
                         "UHPminT":-10.0}

DEFAULT_COPEX_HP = 0.3

class ModuleHP():

    equipments = None
    cascadeIndex = None
    
#------------------------------------------------------------------------------
    def __init__(self,keys):
#------------------------------------------------------------------------------
#..............................................................................
# getting list of equipment in SQL

        self.keys = keys

        self.setupid = Status.SetUpId

        self.neweqs = 0 #new equips added

#------------------------------------------------------------------------------
    def initPanel(self):
#------------------------------------------------------------------------------
#       initialisation of the panel
#------------------------------------------------------------------------------

        logTrack("ModuleHP (initPanel): starting")
        self.getUserDefinedParamHP() #returns to the GUI the default user-defined data to be shown in HP panel

        if Status.int.cascadeUpdateLevel < 0:

            logTrack("ModuleHP: initCascadeArrays")
            Status.int.initCascadeArrays(0)

            if DEBUG in ("ALL", "BASIC"):
                print "QA",Status.int.QA_T
                print "QAmod",Status.int.QA_T_mod

    
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def updatePanel(self):
        """
        Here all the information is prepared so that it can be plotted on the panel
        """               
#............................................................................................
# 0. Prepare the system information necessary for starting work

        (HPList,HPTableDataList) = self.screenEquipments()

        logTrack("ModuleHP (updatePanel): cascade index %s update level %s"%\
                 (self.cascadeIndex,Status.int.cascadeUpdateLevel))

        logTrack("ModuleHP.updatePanel: before simulation")

        if DEBUG in ("ALL", "BASIC"):
            print "QA",Status.int.QA_T
            print "QAmod",Status.int.QA_T_mod

        if Status.int.cascadeUpdateLevel < self.cascadeIndex or Status.mod.moduleEnergy.outputLevel < 3:
            logTrack("ModuleHP: calling run simulation")
            Status.mod.moduleEnergy.runSimulation(last=self.cascadeIndex, outputLevel=3)

#............................................................................................
# 1. List of equipments

        matrix = []
        for row in HPTableDataList:
            matrix.append(row)

        data = np.array(matrix)

        Status.int.setGraphicsData('HP Table',data)

#............................................................................................
# 2. XY Plot


        if len(HPList) > 0: # there is already some heat pump
            if Status.SimTool == "Plus":
                QD_T = Status.int.calcQ_T(Status.int.QD_Tt_mod[self.cascadeIndex-1])
            else:
                QD_T = Status.int.QD_T_mod[self.cascadeIndex-1]
                
            self.QD_T = QD_T
            if Status.SimTool == "Plus":
                Status.int.USHj_T[self.cascadeIndex-1] = Status.int.calcQ_T(Status.int.USHj_Tt[self.cascadeIndex-1])
                Status.int.QHXj_T[self.cascadeIndex-1] = Status.int.calcQ_T(Status.int.QHXj_Tt[self.cascadeIndex-1])
                
            QDres_T = QD_T - Status.int.USHj_T[self.cascadeIndex-1]
            if Status.SimTool == "Plus":
                QA_T = Status.int.calcQ_T(Status.int.QA_Tt_mod[self.cascadeIndex-1])
                QAres_T = Status.int.calcQ_T(Status.int.QA_Tt_mod[self.cascadeIndex])
            else:   
                QA_T = Status.int.QA_T_mod[self.cascadeIndex-1]
                QAres_T = QA_T - Status.int.QHXj_T[self.cascadeIndex-1]

        else:
            QD_T = Status.int.USHTotal_T
            self.QD_T = QD_T
            QDres_T = QD_T

            if Status.SimTool == "Plus":
                QA_T = Status.int.calcQ_T(Status.int.QA_Tt_mod[self.cascadeIndex-1])
            else:
                QA_T = Status.int.QA_T
            QAres_T = QA_T
        
        iTmax = Status.NT/2
        QD_plot = []
        QDres_plot = []
        QA_plot = []
        QAres_plot = []
        TScale = Status.int.T[0:iTmax].tolist()

        for i in range(iTmax):
            QD_plot.append(QD_T[i]/1000.0)
            QA_plot.append(QA_T[i]/1000.0)
            QDres_plot.append(QDres_T[i]/1000.0)
            QAres_plot.append(QAres_T[i]/1000.0)

        Status.int.setGraphicsData('HP Plot',
                                   [TScale,
                                    QD_plot,
                                    QA_plot,
                                    QDres_plot,
                                    QAres_plot])

##        except:
##            logDebug("ModuleHP (updatePanel): error trying to send graphics data")

#............................................................................................
# 3. Configuration design assistant

        self.getUserDefinedParamHP() #returns to the GUI the default user-defined data to be shown in HP panel

#............................................................................................
# 4. additional information
        (TPinch,TGap) = self.calcTPinchAndTGap()
        info = []
        info.append(TPinch)  #first value to be displayed
        info.append(TGap)  #second value to be displayed

        Status.int.setGraphicsData('HP Info',info)

#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
    def getUserDefinedParamHP(self):
#------------------------------------------------------------------------------
#   gets the user defined data from UHeatPump and stores it to interfaces to be shown in HP panel
#------------------------------------------------------------------------------

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)

            config = [True, 'compression',1500.0,60.0,100.0,-10.0,100.0]
            Status.int.setGraphicsData('HP Config', config)            
            self.setUserDefinedParamHP()

        else:
            uHP = urow
            #returns to the GUI the default user-defined data to be shown in HP panel
            if uHP.UHPMaintain == 1:
                maintainExisting = True
            else:
                maintainExisting = False
            config = [maintainExisting,
                      uHP.UHPType,
                      uHP.UHPMinHop,
                      uHP.UHPDTMax,
                      uHP.UHPmaxT,
                      uHP.UHPminT,
                      uHP.UHPTgenIn]
            Status.int.setGraphicsData('HP Config', config)
            
        return config

#------------------------------------------------------------------------------
    def setUserDefinedParamHP(self):
#------------------------------------------------------------------------------

        UDList = Status.int.GData['HP Config']

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        
        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)
            urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
            
        row = urow

        if UDList[0] == True:
            maintainExisting = 1
        else:
            maintainExisting = 0
            
        row.UHPMaintain = dbio.check(maintainExisting) # to add in UHeatPump
        row.UHPType = dbio.check(UDList[1])
        row.UHPMinHop = dbio.check(UDList[2], string=False)
        row.UHPDTMax = dbio.check(UDList[3], string=False)
        row.UHPmaxT = dbio.check(UDList[4], string=False)
        row.UHPminT = dbio.check(UDList[5], string=False)
        row.UHPTgenIn = dbio.check(UDList[6], string=False)

        dbio.commit()

#------------------------------------------------------------------------------
    def initUserDefinedParamHP(self):
#------------------------------------------------------------------------------
#   gets the default user defined data for HP from PSetUpData table. Stores them in UHeatPump.
#   to be executed only onece when new alternative is created
#------------------------------------------------------------------------------
        default = dbio.get("psetupdata", id=Status.SetUpId)
        uheatpump = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        uheatpump.UHPType = default.UHPType
        uheatpump.UHPMinHop = default.UHPMinHop
        uheatpump.UHPDTMax = default.UHPDTMax
        uheatpump.UHPmaxT = default.UHPmaxT
        uheatpump.UHPminT = default.UHPminT
        uheatpump.UHPTgenIn = default.UHPTgenIn

        dbio.commit()
        
##XXX idea aparte:
##en vez de duplicar todas las columnas de la tabla en UHeatPump en PSetUp, no ser�a mejor crear una convenci�n tipo
##PId = -1 en la tabla UHeatPump mismo = espacio reservado para los default values ????
### La definicion de la sql ahora no permute entrar PId negativo, SD, 28/03/2008, se tiene que cambiar


   
#------------------------------------------------------------------------------
    def screenEquipments(self):
#------------------------------------------------------------------------------
#       screens existing equipment, whether there are already heat pumps
#       XXX to be implemented
#------------------------------------------------------------------------------

#............................................................................................
# 1. get equipment list and cascade tables
                 
        self.equipments = Status.prj.getEquipments()
        self.NEquipe = len(self.equipments)
        Status.int.getEquipmentCascade()

#............................................................................................
# 2. screen for existing heat pumps
                 
        self.HPList = []

        i = 0
        idx = 0
        for row in Status.int.cascade:
            i+=1
            if getEquipmentClass(row["equipeType"]) == "HP":
                self.HPList.append(row)
                idx = i
        self.cascadeIndex = idx #set to 0, if nothing is found !!!
        
        HPTableDataList = []
        for row in Status.int.EquipTableDataList:
#            if row[2] == 'Heat pump' or row[2] == 'HeatPump' or row[2] == 'Heat Pump' or row[2] == 'HP COMP':
            if getEquipmentClass(row[3]) == "HP":
                HPTableDataList.append(row)

        #screen list and substitute None with "not available"
        for i in range(len(HPTableDataList)):
            for j in range(len(HPTableDataList[i])):
                if HPTableDataList[i][j] == None:
                    HPTableDataList[i][j] = 'not available'        


        # determine first heating equipment in cascade (required for plus mode, if first equipment is cooling)
        jc = self.cascadeIndex - 1
        self.jFirstHeating = -1
        for j, eq in enumerate(Status.int.systemModel.equipment):
            if getEquipmentClass(eq.EquipType) not in ["CH"]:
                self.jFirstHeating = j
                break
        
        return (self.HPList,HPTableDataList)
        
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def calcTPinchAndTGap(self):
#------------------------------------------------------------------------------
#   calculates the temperature gap and Pinch temperature for the demand
#   seen by this module
#------------------------------------------------------------------------------

        index = max(self.cascadeIndex-1,0)
        iD = firstNonZero(self.QD_T)
        iA = lastNonZero(Status.int.QA_T_mod[index])

        TminD = (iD-1)*Status.TemperatureInterval
        TmaxA = (iA+1)*Status.TemperatureInterval

        TPinch = 0.5*(TminD + TmaxA)
        TGap = TminD - TmaxA

        return(TPinch,TGap)

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#   AUXILIARY FUNCTIONS FOR MANAGING EQUIPMENT ENTRIES IN SQL
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def getEqId(self,rowNo):
#------------------------------------------------------------------------------
#   gets the EqId from the rowNo in the HPList
#------------------------------------------------------------------------------

        HPId = self.HPList[rowNo]["equipeID"]
        return HPId

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def deleteEquipment(self,rowNo):
#------------------------------------------------------------------------------
#        deletes the selected heat pump in the current alternative
#------------------------------------------------------------------------------

        if rowNo == None:   #indicates to delete last added equipment dummy
            HPid = self.dummyEqId
        else:
        #--> delete HP from the equipment list under current alternative #from C&QGenerationHC under ANo
            HPid = self.getEqId(rowNo)
                 
        logTrack("Module HP (delete): id to be deleted = %s"%HPid)

        Status.prj.deleteEquipment(HPid)        

##        self.deleteFromCascade(Status.int.cascade, HPid) #actuallise the cascade list

        self.NEquipe -= 1
        Status.int.systemModel.changed()

    def addEquipmentDummy(self,onTop = True):
        """Add a new dummy equipment to the equipment list and return 
        the position in the cascade.
        
        """
        self.cascadeIndex = self.NEquipe + 1
        
        self.neweqs += 1 #No of last equip added
        NewEquipmentName = "New heat pump %s"%(self.neweqs)
        
        logDebug("ModuleHP (addEquipmentDummy): take care. present version works well only for compression heat pumps !!!")
        
        self.equipe = Status.prj.addEquipmentDummy()
        self.dummyEqId = self.equipe.QGenerationHC_ID
        
        # connect the dummy to all heating pipes
        dummy, pipeIDs = getAllHeatingPipes()
        pipeIDs = [str(id) for id in pipeIDs ]
        IDString = ";".join(pipeIDs)
        
        equipeData = {"Equipment":NewEquipmentName, 
                      "EquipType":"compression heat pump",
                      "HCGPnom": 0.0,
                      "TMaxSupply": 0.0,
                      "PipeDuctEquip":IDString}
        self.equipe.set(**equipeData)
        dbio.commit()
            

        if onTop:
            equipeID = self.equipe.QGenerationHC_ID  #SD change 30/04.2008
            logTrack("ModuleHP (DA1): equipe before moving - %s %s %s"%\
                  (self.equipe.Equipment,self.equipe.EqNo,self.equipe.CascadeIndex))
            
            Status.mod.moduleHC.cascadeMoveToTop(self.equipe.CascadeIndex)

            self.equipe = Status.prj.getEquipe(equipeID)
            logTrack("ModuleHP (DA1): equipe after moving - %s %s %s"%\
                  (self.equipe.Equipment,self.equipe.EqNo,self.equipe.CascadeIndex))
            self.cascadeIndex = self.equipe.CascadeIndex


        # update the interconnection tables
        Status.int.systemModel.build()

        Status.int.getEquipmentCascade()
        Status.int.addCascadeArrays()

        self.equipments = Status.prj.getEquipments()
        return(self.equipe) 

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def setEquipmentFromDB(self,equipe,modelID): #SD change 30/04.2008
#------------------------------------------------------------------------------
#   takes an equipment from the data base and stores it under a given Id in
#   the equipment data base
#------------------------------------------------------------------------------

        model = dbio.get("dbheatpump", id=modelID)
        if model is None:
            logDebug("ModuleHP.setEquipmentFromDB: heatpump model with ID %s not in data base"%modelID)
            return

#        print "ModuleHP(setEquipmentFromDB): updating parameters in SQL"        
        if model.HPHeatCap != None:
            equipe.HCGPnom = model.HPHeatCap
            Pnom = model.HPHeatCap
        else:
            equipe.HCGPnom=0.0
            logWarning("Heat pump with 0 kW capacity loaded from database. Check your heat pump database")
            Pnom = 0

        eqType=getEquipmentType("HP",model.HPType)
        equipe.EquipType=eqType
            
        equipe.HCGTEfficiency=model.HPHeatCOP
        if model.HPManufacturer != None: equipe.Manufact=model.HPManufacturer
        if model.HPYearUpdate != None: equipe.YearManufact=model.HPYearUpdate
        if model.HPModel != None: equipe.Model=model.HPModel
        equipe.NumEquipUnits=1
        if model.DBFuel_id != None: equipe.DBFuel_id=model.DBFuel_id
        if model.HPFuelConsum != None: equipe.FuelConsummodel.HPFuelConsum
        if model.HPUnitsFuelConsum != None: equipe.UnitsFuelConsum=model.HPUnitsFuelConsum
        if model.HPElectConsum != None: equipe.ElectriConsum=model.HPElectConsum

        equipe.TSupply = model.HPCondTinH
        equipe.THeatSourceLT = model.HPEvapTinH
        COP = model.HPHeatCOP
        COPex = model.HPExHeatCOP

        COPex_calculated = self.calculateCOPex(COP, model.HPCondTinH, model.HPEvapTinH, eqType)
        if COPex is None:
            COPex = COPex_calculated
        elif COPex_calculated is not None:
            diff = abs(COPex - COPex_calculated)
            if diff > 1.e-3:
                logger.logError("Erroneous value COPex = %s in heat pump model %s. Calculated value: %s"%(COPex,                                                                                   model.HPModel,
                                                                                                          COPex_calculated))
            COPex = COPex_calculated

        equipe.HPExHeatCOP = COPex

##        COPth = model.HPThHeatCOP
##        if COP is not None:
##            if COPth is not None and model.HPThHeatCOP is not None:
##                if COP <= COPth and COPth > 0:
##                    COPex = COP/COPth
##                    logMessage("Exergetic efficiency of HP %s calculated to %s"% \
##                               (equipe.Model,COPex))
##                    equipe.HPExHeatCOP = dbio.check(COPex)
##                else:
##                    logWarning("Error in efficiency data of the heat pump %s"%equipe.Model)

                
        equipe.IsSelectedFromDB=1
        equipe.DatabaseNameSelection="DBHeatPump"
        if model.HPType != None: equipe.EquipTypeFromDB=model.HPType
        equipe.EquipIDFromDB=model.id

        if model.HPCondTmax is not None: equipe.TMaxSupply=model.HPCondTmax
        if model.HPEvapTmin is not None: equipe.ST_C0 = model.HPEvapTmin
        if model.HPLimDT is not None: equipe.ST_C1 = model.HPLimDT
        
#HS 2008-10-25: equipment parameters that are set defined by default if not specified
        
        equipe.FlowExhaustGas = 0.0
        equipe.TExhaustGas = 0.0

        equipe.set(**{"TurnKeyPrice": model.HPTurnKeyPrice,
                       "OandMfix": model.HPOandMfix,
                       "OandMvar": model.HPOandMvar})
        (turnKey, OMFix, OMVar) = self.estimateEconomicParameters(equipe)
        equipe.set(**{"TurnKeyPrice": turnKey,
                       "OandMfix":OMFix,
                       "OandMvar":OMVar})
        dbio.commit()
                   
        Status.int.systemModel.updateEquipmentData(equipe.CascadeIndex-1, equipe)

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

    def getHPList(self, HPTypeUD, TMinD, DTMin):
        """
       Create the lists (ascending order for HPHeatCap) which are needed for
       the selection of a heat pump and the calculateEnergyFlows()
        """
        HPList = []
        HPListPNom = []
        sqlQuery = "HPType LIKE '%s' ORDER BY HPHeatCap ASC"%(HPTypeUD)
        subset = dbio.selectByQuery("dbheatpump", sqlQuery) #this is a list of rows (dictionaries) in DBHeatPump
        DTMax = 0
        TCondMax = 0
        for i, model in enumerate(subset):
            if model.HPHeatCap is not None and model.HPHeatCap > 0: # eliminate the possibility that chillers are introduced !!!

                HPList.append(model.id)
                HPListPNom.append(model.HPHeatCap)
                DTMax = max(DTMax,model.HPLimDT)
                TCondMax = max(TCondMax,model.HPCondTmax)        

        return (HPList,HPListPNom,DTMax,TCondMax)
        
    def calculateCOPh_Carnot(self, Th, Tc,Tg = None):
        """
        Calculates the theoretical Carnot COP.
        """
        if Tg is None:
            COPh_Carnot = carnotCOPh(Th ,Tc)                
        else:                                       # absorption heat pumps
            COPh_Carnot = carnotCOPc(Th, Tc) * carnotEfficiency(Tg, Th) + 1.0

        return min(COPh_Carnot, COPCARNOTMAX)
        
#------------------------------------------------------------------------------
    def getTMinD(self,Q_T):
#------------------------------------------------------------------------------
#   Looks for the minimum temperature in heat demand
#------------------------------------------------------------------------------
        iT = max(firstNonZero(Q_T) - 1,0)      #find the index, SD: -1: assume linear change in the last interval,06/05/2008
        return(Status.int.T[iT])        #find the T corresponding to index
        
#------------------------------------------------------------------------------
    def getTMaxA(self,Q_T):
#------------------------------------------------------------------------------
#   Looks for the minimum temperature in heat demand
#------------------------------------------------------------------------------
        iT = min(lastNonZero(Q_T) + 1,Status.NT+1)       #find the index, SD: +1: assume linear change in the last interval,06/05/2008
        return(Status.int.T[iT])        #find the T corresponding to index
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def designAssistant1(self):
#------------------------------------------------------------------------------
#
#        Design Assistant 1 - Function for selection of heat pump and calculation of the heat flows in it (annual, hourly) on base of
#        temperature and time dependent heat demand and availability curves. 
#------------------------------------------------------------------------------
            
        if self.jFirstHeating < 0:
            logWarning(_("Cannot run auto design without an existing heating equipment"))
            return("CANCEL",[])

        logTrack("ModuleHP (designAssistant1): starting")

        Status.mod.moduleEnergy.runSimulation()
        
#............................................................................................
#   add a new equipment as space holder to the equipment list
#   and move it to the top of the list

        equipe = self.addEquipmentDummy()

        if Status.SimTool == "Plus":
            # An update of the equipment type in the simulation model is necessary
            Status.mod.moduleEnergy.sim.buildModel(outputLevel=3)
 	else:
            # these lines are required for the case that the first equipment in the case (before adding HP) is a cooling equipment
            USHjRequired = Status.int.USHTotal_Tt
            QD_T = Status.int.calcQ_T(USHjRequired)
            Status.int.QD_Tt_mod[0] = USHjRequired
            Status.int.QD_T_mod[0] = QD_T


#............................................................................................
# get configuration of design assistant and do basic checking

        DA = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        if DA is not None:
            DA_UHPType = DA.UHPType
            if DA_UHPType is None:
                DA_UHPType = HPTYPES[0]

            DA_UHPMinHop = DA.UHPMinHop
            if DA_UHPMinHop <= 0.0:
                DA_UHPMinHop = 1500.0

            DA_UHPmaxT = DA.UHPmaxT
            if DA_UHPmaxT <= 0.0:
                DA_UHPmaxT = 100.0
        else:
            logDebug("ModuleHP (DA1): no u-Table found")
            return("CANCEL",[])
        
        if DA.UHPMinHop > YEAR:        #check of user input
            logWarning('ModuleHP (DA1): Required working hours are greater than the hours of the year!')               
            DA_UHPType = HPTYPES[0]
            DA_UHPMinHop = 1500.0
            DA_UHPmaxT = 100.0
                       

#............................................................................................
#   analyze heat demand for previous checks
            
        if Status.SimTool == "Plus":
            Status.int.QD_T_mod[self.jFirstHeating] = Status.int.calcQ_T(Status.int.QD_Tt_mod[self.jFirstHeating])
            Status.int.QA_T_mod[self.jFirstHeating] = Status.int.calcQ_T(Status.int.QA_Tt_mod[self.jFirstHeating])
        
        cummulativeDemand = Status.int.QD_Tt_mod[self.jFirstHeating]

        TMinD = self.getTMinD(self.QD_T)
        TMaxA = self.getTMaxA(Status.int.QA_T_mod[self.jFirstHeating])

        TDrop = 0.0 if DA_UHPType=="mechanical vapour recompression" else HPTDROP
        DTMin = TMinD - TMaxA  + 2.0*TDrop #minimum necessary temperature lift

##        print "ModuleHP.DA1: TminD = %s TmaxA = %s DT = %s"%(TMinD, TMaxA, DTMin)

#............................................................................................
#   preselect list of equipment that fulfils the criteria

        (HPList,HPListPNom,DTMaxInList,TCondMaxInList) = self.getHPList(DA_UHPType, TMinD, DTMin)  #get the sorted list of available heat pumps

        self.preselection = []

        if DTMin > DTMaxInList:
            logWarning("ModuleHP (DA1): Heat pump application impossible. "+\
                     "Temperature lift from QD-QA %s greater than limit value %s!"%(DTMin,DTMaxInList))
            bestHOp = 0

        else:  #in the case heat pump application is possible: calculate all this...

            if DA_UHPmaxT > TCondMaxInList:
                logTrack('ModuleHP (designAssistant1): UHPmaxT = %s is higher than TCondMaxInList = %s'%\
                         (DA_UHPmaxT,TCondMaxInList))
                logTrack('ModuleHP (designAssistant1): The value of TCondMaxInList = %s will be used as initial estimate of Th0'%\
                         TCondMaxInList)
            
            #Start temperature for calculation Th0 = to the user-defined temperature
            Th0 = min(DA_UHPmaxT, TCondMaxInList)

            if Th0 is None:
                logDebug("ModuleHP (designAssistant1): WARNING -> Th0 = None was obtained")
                Th0 = 100.0
            
#............................................................................................
# Initial selection: maximum reasonable power for a heat pump
# From the annual Heat demand curve (QDa): calculate the necesary heating capacity (starting value)

            Qh0 = interpolateList(Th0/Status.TemperatureInterval,self.QD_T) #calculates the annual energy demand for the Th_o from QDa
            dotQh0 = Qh0/DA_UHPMinHop #the initial heat capacity of the heat pump is obtained dividing by the hours of year #SD: *10 deleted,06/05/2008
            
            listIndex = findInListASC(dotQh0,HPListPNom)

            if listIndex < 0:
                logError('Warning: There is no heat pump available in DB for this application')

#............................................................................................
# Check appropriateness for several heat pumps (preselection)

            PNomMax = 0                 #maximum possible heat pump capacity for desired operating hours
            
            Status.parametricStudy = []
            
            bestHOp = 0 # longest HOp found
            bestUSH = 0
            bestPnom = 0
            while listIndex >= 0:       #decrease gradually until HOp >= MinHOp

                modelID = HPList[listIndex] #a row in the DBHeatPump corresponding to j
                                    
                self.setEquipmentFromDB(equipe, modelID)   #assign model from DB to current equipment in equipment list #SD change 30/04.2008
                Status.int.systemModel.updateEquipmentData(equipe.CascadeIndex-1, equipe)                

##                print "ModuleHP (DA1): equipe ID %s PNom %s QD %s"%\
##                      (modelID,equipe.HCGPnom, cummulativeDemand[Status.NT+1])

                (dummy, USHj) = self.calculateEnergyFlows(equipe,
                                                          cummulativeDemand,
                                                          self.cascadeIndex)
                HOp = USHj/equipe.HCGPnom
##                print "ModuleHP (DA1): USH %s HOp %s"%(USHj,HOp)

                Status.parametricStudy.append(([modelID, equipe.Model, equipe.HCGPnom, equipe.TMaxSupply],[USHj, HOp, (HOp >= DA_UHPMinHop)]))

                eligible = (HOp >= DA_UHPMinHop)

                if eligible:                    
                    if USHj >= bestUSH - EPSILON:
                        bestPnom = equipe.HCGPnom
                        
                    bestHOp = max(bestHOp, HOp)
                    bestUSH = max(bestUSH, USHj)
                
                if USHj < POWERRANGE*bestUSH and equipe.HCGPnom < 0.5*bestPnom: # the second criterion is necessary in order to allow for HPHTs in default data base
                    break
                
                listIndex -= 1  #Select next smaller heat pump

##        print "bestHop: %s bestUSH %s"%(bestHOp, bestUSH)

        eligibleEquipment = [elem for elem in Status.parametricStudy if elem[1][2]]  
        results = [elem[1] for elem in eligibleEquipment]

        bestHOpConstrained = 0.0
        for index, result in enumerate(results):
            USHj = result[0]
            HOp = result[1]
            if USHj >= POWERRANGE*bestUSH:
                bestHOpConstrained = max(bestHOpConstrained, HOp)
                
        for index, result in enumerate(results):
            USHj = result[0]
            HOp = result[1]
            eligible = result[2]
            if eligible and USHj >= POWERRANGE*bestUSH and HOp >= POWERRANGE*bestHOpConstrained:
                self.preselection.append(eligibleEquipment[index][0][0])

        logTrack("ModuleHP (designAssistant1): preselected equipment: %s"%str(self.preselection))

#............................................................................................
# Automatic final selection:
#............................................................................................

        nSelected = len(self.preselection)
#        print "ModuleHP (designAssistant1): return to GUI for manual selection "

        if nSelected <= 0: #no equipment could be selected
            if bestHOp < DA_UHPMinHop:
                showWarning(_("The longest operation time found for the given"\
                               " parameters was %s hours.") % bestHOp)
            return("CANCEL",[])
        
        else:   #several options possible -> manual or automatic final selection

            if Status.UserInteractionLevel == "interactive" or Status.UserInteractionLevel == "semi-automatic":
                if DEBUG != "OFF":
                    print "ModuleHP.DA1: preselection (manual) = %s"%self.preselection

                return("MANUAL",self.preselection)  #"MANUAL" indicates to panel to open the dialog

            else:   
                maxCOP = 0
                for modelID in self.preselection:
                    model = dbio.get("dbheatpump", id=modelID)
                    if (model.HPHeatCOP > maxCOP):
                        bestModelID = modelID
                        maxCOP = model.HPHeatCOP
                self.preselection = [bestModelID]
                if DEBUG != "OFF":
                    print "ModuleHP.DA1: preselection (auto: [bestModel]) = %s"%self.preselection

                return ("AUTOMATIC",self.preselection)
                                             
#------------------------------------------------------------------------------
    def designAssistant2(self, modelID):
#------------------------------------------------------------------------------
#
#        Design Assistant 2 - 
#------------------------------------------------------------------------------

        if DEBUG != "OFF":
            print "ModuleHP.DA2: selected modelID = %s"%modelID

        if (modelID == None or modelID < 0):
            logTrack("ModuleHP (DA2): user cancelled the selection of the heat pump")
            self.deleteEquipment(None)   

        else:                        
            self.setEquipmentFromDB(self.equipe, modelID) #add selected equipment to the equipment list #SD change 30/04.2008
            logTrack("ModuleHP (DA2): heat pump added. model no: %s"%modelID)
            Status.int.changeInCascade(self.equipe.CascadeIndex)
            

        Status.int.systemModel.changed()
        Status.int.systemModel.build()

        if Status.SimTool != "Plus": Status.mod.moduleEnergy.runSimulation()
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def calculateEnergyFlows(self,equipe,demand_Tt,cascadeIndex): #SD change 30/04.2008
        """Calculate the energy flows in the heat pump identified by 
        "cascadeIndex" and update the corresponding global data.
        
        :returns: a pair consisting of the time and temperature 
        dependent numpy array containing the covered demand. Also known
        as USHj_Tt of the given equipment and the same data aggregated 
        over time and temperature USHj.
        
        """
        if Status.int.cascadeUpdateLevel < (cascadeIndex - 1) or Status.mod.moduleEnergy.outputLevel < 3:
            logDebug("ModuleHP (calculateEnergyFlows): cannot calculate without previously updating the previous levels")
            Status.mod.moduleEnergy.runSimulation(last=(cascadeIndex-1), outputLevel=3)
        Status.int.extendCascadeArrays(cascadeIndex)

        jc = cascadeIndex-1
##        j = Status.int.equipmentCascade[jc]
        HXList = Status.QHXEqLink[jc][:]

        connectedHX = []
        genericWH = True
        for h,hxCon in enumerate(HXList):
            if hxCon == 1:
                connectedHX.append(h)
                genericWH = False

        if Status.SimTool == "Plus":
            USHj_Tt = Status.int.createQ_Tt()

            if genericWH:
                QA_Tt = Status.int.QA_Tt_mod[cascadeIndex-1]
                QA_T = Status.int.calcQ_T(QA_Tt)
            else:
                QA_Tt = None

            USHj = Status.mod.moduleEnergy.sim.systemModel.simulateEquipment(USHj_Tt, demand_Tt, cascadeIndex-1, QWHAmbRes_Tt=QA_Tt)

##            print "QD %s QA %s USHj %s"%(sum(demand_Tt[NT+1]), sum(QA_Tt[0]), USHj)
                
            return USHj_Tt, USHj

        # Remaining part from here on only for old simulation mode (!= Plus)

        if cascadeIndex > 0 and cascadeIndex <= Status.NEquipe:
            logTrack("ModuleHP (calculateEnergyFlows): starting (cascade no: %s)"%cascadeIndex)
        else:
            logError("ModuleHP (calculateEnergyFlows): cannot simulate index %s: out of cascade [%s]"%\
                     (cascadeIndex,Status.NEquipe))
            return (createQ_Tt(), 0)

#..............................................................................
# get equipment data from equipment list in SQL
        PNom = equipe.HCGPnom
        TMax = equipe.TMaxSupply
        if TMax is None:
            TMax = 90.0
            logWarning("ModuleHP (calculateEnergyFlows): no TMax defined in HP data for HP %s"%equipe.Equipment)
        DTMax = noneCheck(equipe.ST_C1, 60.0)
        self.TMin = noneCheck(equipe.ST_C0, 0.0)  # minimum temperature of heat source used at equipment 
### HERE HPLimDT and HPEvapTmin should be copied from the heat pump table
        
        if PNom is None:
            logError("ModuleHP (calculateEnergyFlows): nominal power of heat pump %s not defined"%equipe.Equipment)
            PNom = 0.0
            
        if equipe.HPExHeatCOP is None or equipe.HPExHeatCOP == 'NULL': 
            COPex = DEFAULT_COPEX_HP
            logTrack("ModuleHP (calculateEnergyFlows): default COPex assumed for heat pump %s"%equipe.Equipment)
        else:
            COPex = equipe.HPExHeatCOP

        Tg = None

        controlSchedule = Status.int.systemModel.eqObject[cascadeIndex-1].controlSchedule
        control = controlSchedule.schedule if controlSchedule else 1.0

        if genericWH:
            invUA = 0.0
            name = "-"
        else:
            if len(connectedHX) > 1:
                logWarning("Warning: currently only one heat exchanger is allowed as low temperature heat source for heat pumps")

            h = connectedHX[0]
            hxes = Status.prj.getHXes()
            hx = hxes[h]
            UA = hx.UA
            name = hx.HXName

            if UA is None or UA == 0:
                invUA = 0
            else:
                invUA = 1./UA

##        print "ModuleHP - genericWH %s HX %s invUA %s"%(genericWH, name, invUA)
                
#..............................................................................
# get demand data for CascadeIndex/EquipmentNo from Interfaces
# and create arrays for storing heat flow in equipment

        QD_Tt = demand_Tt.copy() # copy the given demand array for internal use

        if genericWH:
            QA_Tt = Status.int.QA_Tt_mod[cascadeIndex-1].copy()
        else:
            qsrc = HeatFlow_t()

            if Status.HRTool != HRMODES[2]:
                logWarning("Connection of heat exchangers as low temperature heat source is only possible in HR mode calculatio of actual HX network")
                
            for h in connectedHX:
                try:
                    qsrc.add(Status.int.hx[h].qsrc)
                except:
                    logDebug("ModuleHP: Could not add qsrc of hx[%s]"%(h+1))
                    
            QA_Tt = qsrc.Q_Tt(descending=True)

        if DEBUG != "OFF":
            QA_T = Status.int.calcQ_T(QA_Tt)
            logTrack("QA_T: %r"%QA_T)

        USHj_Tt = Status.int.createQ_Tt()
        USHj_t = Status.int.createQ_t()

        QHXj_Tt = Status.int.createQ_Tt()
        QHXj_t = Status.int.createQ_t()
            
        QWHj_Tt = Status.int.createQ_Tt()

        Tc_t = Status.int.createQ_t()

#..............................................................................
# Start hourly loop

        USHj = 0
        QHXj = 0
        QWHj = 0

        HPerYear = 0

        FETel_j = 0

        if equipe.HCGTEfficiency is not None:
            COPMin = 0.8 * equipe.HCGTEfficiency  #Regulate HP so that operates at least at 80% of nominal efficiency
        else:   
            COPMin = 3.0    # This is the objective for design of HP power
                            # -> not MAXIMUM technically possible energy,
                            # but maximum energy at a reasonable COP
                            # criterion could be optimised ... e.g. if 80% of the energy
                            # delivered rises the COP to 4.0, better use this ?
                            # => minimise  E = Q_HP/COP_HP + (Q_total - QHP)/COP_backup_system
                            # -> dE/dQ_HP = (1/COP_HP - 1/COP_backup) - (dCOP/dQ_HP) * Q_HP/COP_HP**2

        dlg = DialogGauge(Status.main, _("EINSTEIN system simulation"), _("simulating %s") % (_("heat pump")))
        if equipe.EquipType == "mechanical vapour recompression":
            TDrop = 0
        else:
            TDrop = HPTDROP

        for it in range(Status.Nt):
            self.it = it
            fctrl = 1.0 if isinstance(control, float) else control[it]

            if it in SNAPSHOTS:
                print "\n---------------\nSnapShot HeatPump %s (%s kW) - Starting time step %s"%(equipe.Equipment, PNom, it)

            if  QD_Tt[Status.NT+1][it] < EPS or QA_Tt[0][it] < EPS: #no demand or no availability
                fop = 0
                if it in SNAPSHOTS:
                    print "HeatPump: fop = 0 - no demand or availability"
            else:
                # create demand and availability profile for present time step
                QD_T = QD_Tt[:, it].copy()
                QA_T = QA_Tt[:, it].copy()

                if it in SNAPSHOTS:
                    print "HeatPump: QD and QA"
                    print "QD: ", QD_T
                    print "QA: ", QA_T
                dQMax = PNom*Status.TimeStep*fctrl
                
                # previous calculations: maximum possible power
                dQh_max = min(interpolateTable(TMax, Status.int.T, QD_T), PNom*Status.TimeStep)
##                dQc_max = interpolateTable(TMin, Status.int.T, QA_T))
##                QA_T = np.minimum(QA_T, dQc_max)

                if dQh_max < 0.01*PNom:
                    fop = 0
                    if it in SNAPSHOTS:
                        print "HeatPump: fop = 0 - no demand below TMax %s"%TMax
                        print "QD: ", QD_T

                else:
                
##                    COPMax = 1.5*COPMin
                    COPMax = 30.0

                    self.matchInit(None, QD_T, QA_T, TDrop, invUA, COPex, COPMax)

                    err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i = self.matchForQh(dQh_max, QD_T, QA_T, TDrop, invUA, COPex, COPMax)

                    if it in SNAPSHOTS:
                        print "MatchForQh: err %s COPh %s dQh %s dQc %s Th %s Tc %s"%(err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i)

                    if COPh_i < (COPMin - EPSILON) or err:
                        err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i = self.matchForCOP(COPMin, QD_T, QA_T, TDrop, invUA, COPex, COPMax)

                        if it in SNAPSHOTS:
                            print "MatchForCOP: err %s COPh %s dQh %s dQc %s Th %s Tc %s"%(err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i)

                    if (Th_i - Tc_i) > (DTMax + EPSILON) or err:
                        err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i = self.matchForDT(DTMax, QD_T, QA_T, TDrop, invUA, COPex, COPMax)

                        if it in SNAPSHOTS:
                            print "MatchForDT: err %s COPh %s dQh %s dQc %s Th %s Tc %s"%(err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i)
                        
    ## limitations of old HP-Module. Have been partially substituted by equivalent COPMax
    ##                        COPh_i = min(COPh_i,30.0)               #practical limit for COP at small temp. lift
    ##                        COPh_i = max(min(COPh_i,1.2*COPh0_i),0.8*COPh0_i)     #limit variations in COP to max 5%

                    fop = (dQh_i > EPSILON)
                    
                    if DEBUG != "OFF" and it < 48:
                        logTrack("ModuleHP.cEF: Tc_i %s Th_i %s COPh_i %s dotQh_i %s"%(Tc_i,Th_i,COPh_i,dQh_i))

#..............................................................................

            if fop==0:
                Tc_i = 0.0
                Th_i = 0.0
                COPh_i  = 0.0 
                dQh_i = 0.0
                dQc_i = 0.0

                Tc_t[it] = np.nan

            else:
                Tc_t[it] = Tc_i
                HPerYear += Status.TimeStep
            
            USHj_t[it] = dQh_i
            QHXj_t[it] = dQc_i
            dFETel = dQh_i - dQc_i
            FETel_j += dFETel
            Status.int.FETel_j_t[j][it] = dFETel
                
            USHj += USHj_t[it]
            QHXj += QHXj_t[it]
             
#..............................................................................
# temperature resolution of delivered and consumed heat
# => at the moment this is done by supplying heat to the lowest temperature
# level where there is a demand, and consuming waste heat at the highest level
# this is optimum for the single equipment, but not for the system !!!
            
            for iT in range(Status.NT+2):   #+1 = the > 400 �C case
                USHj_Tt[iT][it] = min(QD_Tt[iT][it],USHj_t[it])     #from low to high T
                QD_Tt[iT][it] -= USHj_Tt[iT][it]

            if genericWH:
                for iT in range((Status.NT+1),-1,-1):
                    QHXj_Tt[iT][it] = min(QA_Tt[iT][it],QHXj_t[it])
                    QA_Tt[iT][it] -= QHXj_Tt[iT][it]
            else:
                for iT in range((Status.NT+1),-1,-1):
                    QA_Tt[iT][it] -= QHXj_Tt[iT][it]
                
            if it % 200 == 0:
                dlg.update(100.0 * it / Status.Nt)
        
        dlg.Destroy()
#..............................................................................
# End of year reached. Store results in interfaces

        Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
        Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)
        
        Status.int.USHj_Tt[cascadeIndex-1] = USHj_Tt
        Status.int.USHj_T[cascadeIndex-1] = Status.int.calcQ_T(USHj_Tt)
        Status.int.USHj_t[cascadeIndex-1] = USHj_Tt[Status.NT+1].copy()
        Status.int.USHjChanged[cascadeIndex-1] = True

        Status.int.QHXj_Tt[cascadeIndex-1] = QHXj_Tt
        Status.int.QHXj_T[cascadeIndex-1] = Status.int.calcQ_T(QHXj_Tt)
        Status.int.QHXj_t[cascadeIndex-1] = QHXj_Tt[Status.NT+1].copy()
        Status.int.QHXEqChanged[cascadeIndex-1] = True

        if genericWH:
## TODO: here NP and plus models behave different if Tc_i is a function of time
##            TQAMax = self.Tc_i_t + HPTDROP                
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = HeatFlow_t()
            Status.int.QHXEqReqChanged[cascadeIndex-1] = False
        else:
##            TQAMax = np.min(self.Tc_i_t+HPTDROP, TQAMax - self.invUA * dotQc_i)

            #TODO: change back to np.nanmean once a newer Python / Numpy - Version is implemented
            Tc_avg = nanmean(Tc_t) if HPerYear > 0 else 0.0
            Status.int.QHXEqReq_Tt[cascadeIndex-1] = HeatFlow_t(fluid_id=0, T0=Tc_avg, T1=Tc_avg+HPDTIO, Q_t = QHXj_t)
            Status.int.QHXEqReqChanged[cascadeIndex-1] = True

# waste heat produced by present equipment

        Status.int.QWHj_Tt[cascadeIndex-1] = QWHj_Tt
        Status.int.QWHj_T[cascadeIndex-1] = Status.int.calcQ_T(QWHj_Tt)
        Status.int.QWHj_t[cascadeIndex-1] = QWHj_Tt[Status.NT+1].copy()
        Status.int.QWHjChanged[cascadeIndex-1] = True

        Status.int.QWHEq_t[cascadeIndex-1] = HeatFlow_t()

#........................................................................

        Status.int.cascadeUpdateLevel = cascadeIndex

#........................................................................
# Global results (annual energy flows)

        Status.int.USHj[cascadeIndex-1] = USHj
        Status.int.QWHj[cascadeIndex-1] = QWHj   # not considering the latent heat(condensing water)
        Status.int.QHXj[cascadeIndex-1] = QHXj
        
        Status.int.FETFuel_j[cascadeIndex-1] = 0.0
        Status.int.FETel_j[cascadeIndex-1] = FETel_j

        Status.int.ElGen_j[cascadeIndex - 1] = 0.0
        Status.int.ElGen_t[cascadeIndex - 1] = Status.int.createQ_t()

        Status.int.HPerYearEq[cascadeIndex-1] = HPerYear

        logTrack("ModuleHP: eq.no.:%s USH: %s [MWh] FETFuel: %s [MWh] FETel: %s [MWh] QHX: %s HPerYear: %s [h]"%\
                   (equipe.EqNo,\
                    USHj/1000.0,\
                    0.0,\
                    FETel_j/1000.0,\
                    QHXj/1000.0, \
                    HPerYear))
                   
        return (np.array(USHj_Tt), USHj)   
#------------------------------------------------------------------------------

    def estimateEconomicParameters(self, eq):
        """ Default values for economic parameters
        """

        if hasattr(eq, "eqType"): # is an eqObject        
            eqType = eq.eqType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OMFix
            OMVar = eq.OMVar

            Pnom = eq.Pnom
            
        else:   # kept for backward compatibility - still used in setEquipmentFromDB
            eqType = eq.EquipType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OandMfix
            OMVar = eq.OandMvar

            Pnom = eq.HCGPnom
            

        # security check if the reference power is defined
        Pref = Pnom 
        if Pref is None:
            return (None, None, None)

        if turnKey is None:
            if hasattr(eq, "Price") and eq.Price is not None:
                turnKey = max(2.*eq.Price,eq.Price+40*Pnom)

            elif eqType == "compression heat pump":
                turnKey = 200*Pnom
            elif eqType == "thermal heat pump":
                turnKey = 250*Pnom
            elif eqType == "mechanical vapour recompression":
                turnKey = 200*Pnom
            else:
                logWarning(_("Turnkey investment costs for equipment %s could not be calculated") % eq.name)
                turnKey = 0.0
            
        OMFix = noneCheck(OMFix, 0.005 * turnKey)
        OMVar = noneCheck(OMVar, 0.004)      

        return (turnKey, OMFix, OMVar)

    def calculateCOPex(self, COPh, Th, Tc, eqType=None):
        """
        Helper function for calculating COPex from nominal operation conditions
        """
        if any([val is None for val in (COPh, Th, Tc)]):
            return None
        else:
            TDrop = 0 if eqType == "mechanical vapour recompression" else HPTDROP
            COPh_Carnot = self.calculateCOPh_Carnot(Th + TDrop, Tc - TDrop)
            COPex = COPh/COPh_Carnot if COPh_Carnot > EPSILON else 0.0
            return COPex

    def matchInit(self, dQh_i, QD_T, QA_T, TDrop, invUA, COPex, COPMax):
        """
        initial calculations for a given time step
        prepares the match functions
        """
        self.dQh = np.array(QD_T)
        self.dQc = np.array(QA_T)

        self.TScale = np.array(Status.int.T)
        self.Tc_cp = self.TScale - np.maximum(invUA * np.array(self.dQc)/ Status.TimeStep, HPTDROP)      # Temperature scale for Tc as a function of Tcp
        
    def matchForQh(self, dQh_i, QD_T, QA_T, TDrop, invUA, COPex, COPMax):
        """
        Looks the match of the curves QA(Tc) and the required low temperature heat
        dQc_req(Tc) calculated on the basis of dQh, Th and COP(Tc)

        Tcp: temperature scale at source side of QA-heat exchanger

        subindex _i: values at match
        """
        Th_i = interpolateTable(dQh_i, self.dQh, self.TScale)

        Tc = self.Tc_cp
       
        COP = COPex*(Th_i + TDrop + KELVIN)/np.maximum((Th_i - Tc + 2*TDrop), 1.e-10)
        COP = np.minimum(COP, COPMax)

        dQc_req = dQh_i * (COP-1)/COP

        dif = self.dQc - dQc_req

        Tcp_i = interpolateTable(0, dif, self.TScale)
        dQc_i = interpolateList(Tcp_i/Status.TemperatureInterval, self.dQc)        

        Tc_i = Tcp_i - max(invUA * dQc_i/Status.TimeStep, HPTDROP)
        COPh_i = dQh_i / (dQh_i - dQc_i)

        err = 0
        dQh_i_new = dQh_i
        if dQc_i < EPSILON or Tc_i < self.TMin - EPSILON - Status.TemperatureInterval: # Status.TemperatureInterval to account for discretisation errors
                                                                                       # E.g. TC1.01 TestREF - ANo 6 requires a tolerance of 3.25 K
            dQh_i_new = 0.0
            COPh_i = 0.0
            Tc_i = 0.0
            Th_i = 0.0
            err = 1

        return err, COPh_i, dQh_i_new, dQc_i, Th_i,  Tc_i
        
    def matchForCOP(self, COP, QD_T, QA_T, TDrop, invUA, COPex, COPMax):
        """
        Looks the match of the curves QA(Tc(Th)) and the required low temperature heat
        dQc_req(Th) = dQh(Th)*COP/(COP-1) calculated on the basis of COP/COPCarnot

        where Tc(Th) is the solution of that COPCarnot(Th, Tc) = COPCarnot
        """
        Th = self.TScale
        DeltaT_int = (Th+TDrop+KELVIN)*(COPex/COP) 
        DeltaT_ext = DeltaT_int - 2*TDrop
        Tc = Th - DeltaT_ext

        dQh = self.dQh
        QA_req = dQh * (COP-1)/COP      # QA_req as a function of Th
        dif = QA_req.copy()

        for iT, Tc_i in enumerate(Tc):
            Tcp_i = Tc_i + max(invUA*QA_req[iT]/Status.TimeStep, HPTDROP)
            dQA = interpolateList(Tcp_i/Status.TemperatureInterval, self.dQc)       #
            dif[iT] -= dQA
        
        dQh_i = interpolateTable(0, dif, self.dQh)

        err = 0
        if dQh_i < EPSILON:
            dQh_i = 0.0
            COPh_i = 0.0
            Tc_i = 0.0
            Th_i = 0.0
            err = 1

        else:
            err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i = self.matchForQh(dQh_i, QD_T, QA_T, TDrop, invUA, COPex, COPMax)
                
        return err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i



    def matchForDT(self, DT, QD_T, QA_T, TDrop, invUA, COPex, COPMax):
        """
        Looks the match of the curves QA(Tc(Th)) and the required low temperature heat
        dQc_req(Th) = dQh(Th)*COP/(COP-1) calculated on the basis of COP/COPCarnot

        where Tc(Th) is the solution of that COPCarnot(Th, Tc) = COPCarnot
        """
        # vectors as a function of Th
        Th = self.TScale

        Tc = np.maximum(Th - DT, self.TMin)
        DTeff = np.maximum(Th - Tc + 2.0*TDrop, 1.e-10) 
        COP = np.minimum(COPex*(Th + TDrop + KELVIN) / (DTeff), COPMax)
        dQc_req = np.array(QD_T)*(COP-1)/COP     # dQc_req in Th-Scale

        Tcp = Tc + np.maximum(invUA * dQc_req/Status.TimeStep, HPTDROP)  # Tcp vs. Th
        
        dif = dQc_req

        for iT, Tcp_i in enumerate(Tcp):
            dQc_i = interpolateList(Tcp_i/Status.TemperatureInterval, self.dQc)
            dif[iT] -= dQc_i

        dQh_i = interpolateTable(0, dif, self.dQh)

        err = 0
        if dQh_i < EPSILON:
            dQh_i = 0.0
            COPh_i = 0.0
            Tc_i = 0.0
            Th_i = 0.0
            err = 1

        else:
            err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i = self.matchForQh(dQh_i, QD_T, QA_T, TDrop, invUA, COPex, COPMax)

        return err, COPh_i, dQh_i, dQc_i, Th_i, Tc_i
        

def nanmean(x):
    """ has been written here, as the function np.nanmean did not exist yet in Python 2.6 still used
    """
    n = np.sum(~np.isnan(x))
    if n>0:
        return np.sum(np.nan_to_num(x))/n
    else:
        return 0.0
                
