# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   DesignHXNet
#   Package for auto-design of heat exchanger networks
#            
#------------------------------------------------------------------------------
#            
#   Class DesignHXNet
#   (New Class for SimTool Plus mode. Equivalent to "HXProposal"
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from pinchAnalysis import PinchAnalysis
from heatExchanger import HeatExchanger
from hxConnection import HXConnection, StreamGlobals
from modules.heatFlow_Tit import correlateHF, optimiseExergy, optimiseSize

from GUI.status import *
##from einstein.modules.energystreams.StreamConstants import *
from constants import HRMODES, DEBUG, EPSILON, HXDESIGNSTRATEGIES, INFINITE
from GUI.GUITools import check
from math import log
import numpy as np
from modules.messageLogger import *

MAXSTORAGETIME = 6
MINFRACTION = 0.005     # minimum split fraction
MINENERGYFRACTION = 0.02    # minimum amount of energy for a heat exchanger (fraction of qhxMax from pinch analysis)

class DesignHXNet():
    """
    The HXProposal Class automatically proposes and generates new Heat-exchangers,
    for unassigned and already assigned Streams, that are created from the programs
    processes, distribution lines, e.g.
    """
    streams = []
##    streams_bpC = []
##    streams_bpH = []
##    streams_apC = []
##    streams_apH = []
    
    # Matches
##    match_bpC = []
##    match_bpH = []
##    match_apC = []
##    match_apH = []
    
    consider_existing_hx = False
    pinch_temperature_lower = -994
    pinch_temperature_upper = -999
    dT = 5
    res = 0
    
    def __init__(self, pa, consider_existing_hx = False, strategy = HXDESIGNSTRATEGIES[0]):
        """
        params: pa  PinchAnalysis object with the results of the PA carried out previously
        """
        self.streams_bpC = []
        self.streams_bpH = []
        self.streams_apC = []
        self.streams_apH = []

        self.match_bpC = []
        self.match_bpH = []
        self.match_apC = []
        self.match_apH = []            

        self.consider_existing_hx = consider_existing_hx

        self.DTmin = pa.DTmin
        
        self.pinch_temperature_lower = pa.TPinchC
        self.pinch_temperature_upper = pa.TPinchH

        self.pa = pa # store whole pinch analysis object (used in matchStreamsEXERGYBPNEW)

        self.streamGlb = [stream for stream in pa.pinchStreams if stream.Q > 0]

        if strategy in HXDESIGNSTRATEGIES:
            self.optimisationStrategy = strategy
        else:
            self.optimisationStrategy = None
            logError("BUG: Erroneous strategy for optimisation of HX design: %s"%strategy)
            
    def calculate(self):
        """
        Main function for HX network design
        Gives the flowchart of the procedure
        """        
        
        if self.consider_existing_hx == False:
            if len(Status.int.hx) > 0:
                for hx in Status.int.hx: 
                    hx.deleteFromDB()
                Status.int.changeInHeatRecovery()
                Status.int.systemModel.changed()
                Status.int.systemModel.build()
        
        self.__splitStreamsIntoSectors()

        self.__splitAboveBelowPinch()
        
        # Sort by M*CP
        self.sortStreams()

        if DEBUG != "OFF":
            self.printToSplit()

            
        if self.optimisationStrategy in ["EXERGY", "MINSIZE"]:
            # calculate those parts of the HCC/CCC to be used for heat exchange (QHX/QWHUsed)
            if self.optimisationStrategy == "EXERGY":
                self.hccu, self.cccu = optimiseExergy(self.pa.hcc_o, self.pa.ccc_o, self.pa.gcc,
                                                      self.pinch_temperature_upper, self.DTmin)
            elif self.optimisationStrategy == "MINSIZE":
                self.hccu, self.cccu = optimiseSize(self.pa.hcc_o, self.pa.ccc_o, self.pa.qhxMax)

            csuList = self.selectColdStreamsAP(self.streams_apC)            
            self.matchStreamsGenericAP(self.streams_apH, csuList, self.match_apH, self.match_apC)

            hsuList = self.selectHotStreamsBP(self.streams_bpH)
            self.matchStreamsGenericBP(hsuList, self.streams_bpC, self.match_bpH, self.match_bpC)
            
        else: # mode FLOWRATE
            self.matchStreams(self.streams_apH, self.streams_apC, self.match_apH, self.match_apC)
            self.streams_bpH = [s for s in self.streams_bpH if s.required] + \
                               [s for s in self.streams_bpH if not s.required]
            self.matchStreams(self.streams_bpH, self.streams_bpC, self.match_bpH, self.match_bpC)
            
        if DEBUG != "OFF":
            self.printMatches()

        self.convertMatchesToHX()
    
    def __findUsedStreamSectors(self, stream):
        """
        Find Streams that are assigned to a specific HeatExchanger,
        and return all Sectors the stream is used in.
        
        :param stream: Element of the StreamGlobals Class.
        :returns: list of Elements of the HXConnection Class.
        
        """
        sectors = []
        for hx in Status.int.hx:

            if stream.heating:
                for conn in hx.sinkstreams:
                    if stream.hasSameSource(conn):
                        # TODO: check if this works for already existing serial connected HX, or if primary source has to be
                        #       looked-up
                        sector = stream.copy()
                        sector.T0 = hx.TciAvg
                        sector.T1 = hx.TcoAvg
                        sector.fraction = conn.primaryFraction()
                        sector.upstream = [stream]
                        sector.hx = hx
                        sectors.append(sector)
            else:
                for conn in hx.sourcestreams:
                    if stream.hasSameSource(conn):
                        sector = stream.copy()
                        sector.T0 = hx.ThoAvg
                        sector.T1 = hx.ThiAvg
                        sector.fraction = conn.primaryFraction()
                        sector.upstream = [stream]
                        sector.hx = hx
                        sectors.append(sector)                


        returnList =  sorted(sectors, key=lambda sector: sector.fraction, reverse=True)

##        print "======================================================"
##        print "__findUsedSectors: StreamSectors after Calculation"
##
##        for sector in returnList:
##            print sector

        return returnList

##        return sorted(sectors, key=lambda sector: sector.fraction, reverse=True)
##        return sectors

    def __splitStreamsIntoSectors(self):
        """
        If existing Heat Exchangers are considered,
        SplittStreamSectors splits the Streams that are assigned to HX
        to get the remaining Stream-parts.
        The normal StreamList is replaced be the new StreamSectors List. 
        
        """
        self.freeSectors = []
        self.usedSectors = []
        
        for stream in self.streamGlb:
            usedSectors = self.__findUsedStreamSectors(stream)
            unUsedSectors = [stream.copy()]
            for used in usedSectors:

                ##################################################
                #   unUsed_L        unUsed (*)    unUsed_H_byPass
                # =============>--------------------------------->
                #               |                
                #               ---------------->---------------->
                #                   used             unUsed_H
                #
                # (*): at the end of processing
                # flow direction inverted for hot streams (cooling)
                ##################################################
                
                n = 0
                addToUnUsed = []
                for n, unUsed in enumerate(unUsedSectors):
                    if used.fraction > unUsed.fraction + EPSILON:   #subtract used sectors only from residual sectors with higher fraction
                        continue
                    

                    # first extract the part at the lower end upstream for heating demands, downstream for cooling demands
                    if unUsed.T0 < used.T0 and unUsed.T1 >= used.T0:  
                        unUsed_L = unUsed.copy()
                        unUsed_L.T1 = used.T0

                        downstream = not stream.heating
                        
                        if not (used.fraction < unUsed.fraction and downstream): # leave unUsed unchanged if there is a split and downstram
                            unUsed.T0 = unUsed_L.T1

                        addToUnUsed.append(unUsed_L)

                        if downstream: # unUsed_L is upstream
                            unUsed_L.upstream = [used]
                        else:
                            used.upstream = [unUsed_L]
                    else:
                        unUsed_L = None

                        
                    # then extract the part at the lower end downstream for heating demands, upstream for cooling demands
                    if unUsed.T1 > used.T1 and unUsed.T0 <= used.T1:
                        unUsed_H = unUsed.copy()
                        unUsed_H.T0 = used.T1

                        downstream = stream.heating
                            
                        if not(used.fraction < unUsed.fraction and downstream): # split 
                            unUsed.T1 = unUsed_H.T0

                        addToUnUsed.append(unUsed_H)

                        
                        if downstream: # unUsed_H is downstream
                            unUsed_H.upstream = [used]
                        else:
                            used.upstream = [unUsed_H]
                    else:
                        unUsed_H = None

                    unUsed.fraction -= used.fraction
                    if unUsed.fraction > 0:
                        if stream.heating and unUsed_H is not None:
                            unUsed_H.upstream.append(unUsed)
                        elif unUsed_L is not None:
                            unUsed_L.upstream.append(unUsed)
                        

                unUsedSectors += addToUnUsed

            self.freeSectors.append(unUsedSectors)    # list of unused sectors for each stream
            self.usedSectors.append(usedSectors)        # list of used sectors for each stream
                
##        print "======================================================"
##        print "__splitStreamsSectors: StreamSectors after Calculation"
##        for i, sectorList in enumerate(self.freeSectors):
##            print "\nfree sectors of stream \n%s"%self.streamGlb[i]
##            for sector in sectorList:
##                print sector
##                print "upstream sectors: \n%s"%sector.upstream
            
    def __splitAboveBelowPinch(self):
        """
        Generates a copy of the streams in the Stream-format.

        Then splits the Streams to Streams that are above and below the pinch-temperature, 
        and creates new Lists for these Splits

        """
        
        self.streams_bpC = []
        self.streams_bpH = []
        self.streams_apC = []
        self.streams_apH = []
            
#        print "Calculated Streams/Streamparts:"
#        for elem in self.streams:
#            print elem
            
        for sectorList in self.freeSectors:
            for s, sector in enumerate(sectorList):
    #            if self.consider_existing_hx == True:
    #                self.matchStreamToHX(stream, i)

                if sector.fraction < EPSILON: continue # do not consider streams without energy

    ##            print "designHXNet.__splitAboveBelowPinch: checking stream %s T0Avg %s T1Avg %s Q %s"%(stream, stream.T0avg, stream.T1avg, stream.Qsum)
                
                if sector.heating: # COLD STREAMS
                    up, down = sector.divide(self.pinch_temperature_lower)
                    if up is not None: self.streams_bpC.append(up)
                    if down is not None: self.streams_apC.append(down)
                                        
                else: # HOT STREAMS
                    up, down = sector.divide(self.pinch_temperature_upper)
                    if down is not None: self.streams_bpH.append(down)
                    if up is not None: self.streams_apH.append(up)

    def sortStreams(self):
        """
        All Stream-lists are sorted.
        """
        self.streams_apC = self.sortStreamWithMCPHours(self.streams_apC)
        self.streams_apH  = self.sortStreamWithMCPHours(self.streams_apH)
        self.streams_bpC = self.sortStreamWithMCPHours(self.streams_bpC)
        self.streams_bpH  = self.sortStreamWithMCPHours(self.streams_bpH)
        

    def matchStreams(self, hsList, csList, hsMatches, csMatches):
        """
        Streams Above Pinch Temperature are matched.
        Streams are matched until No Stream remains on hot or cold side, or MAXRUN length is reached.
        """
#        print "MATCH STREAMS:"
        # ABOVE PINCH
        MAXRUN = len(hsList)*10
        run = 0

        while len(csList) > 0 and len(hsList) > 0:
            if run >= MAXRUN:
                break
            run +=1
             
            # Select Hot Stream with maximal m*cp*hours
            hs = hsList[0]
            mcp_hot = hs.mcp * hs.fraction
            dT_hot=abs(hs.T1-hs.T0)
            if DEBUG != "OFF":
                print "DesignHXNet.matchStreams: finding match for hot stream %s"%hs

            # Find cold stream with closest m*cp value
            dT_cold_best=0
            mcp_cold_best=0

            diff = INFINITE*2 # *2 necessary in order to accept phase change streams
            qBest = INFINITE
            
            dT_cold_index = None
            csList = sorted(csList, key=lambda stream: (stream.T0))
            minStartTemp = min([cs.T0 for cs in csList])

            if hs.T1<(minStartTemp+self.DTmin):                # eliminate hot stream if too cold for exchanging heat with any of the cold streams
                if DEBUG != "OFF":
                    print "DesignHXNet.matchStreams: hot stream too cold (T1 = %s)"%hs.T1
                del hsList[0]
            else:    
                cs_best = None

                for i, cs in enumerate(csList):
                    if DEBUG != "OFF":
                        print "DesignHXNet.matchStreams: trying cs = %s"%cs

                    mcp_cold = cs.mcp * cs.fraction
                    if abs(mcp_cold) < EPSILON:
                        del csList[i]
                        continue

                    if abs(mcp_hot - mcp_cold) < diff - EPSILON:
##                    
                        Tci = cs.T0
                        Tco = min(cs.T1, hs.T1 - self.DTmin)
                        Thi = hs.T1
                        Tho = max(cs.T0 + self.DTmin, hs.T0)
                        
                        dT_cold = Tco - Tci
                        dT_hot = Thi - Tho

                        qc = mcp_cold*dT_cold if mcp_cold < 0.1*INFINITE else cs.Q*cs.fraction
                        qh = mcp_hot*dT_hot if mcp_hot < 0.1*INFINITE else hs.Q*hs.fraction
                        
                        if DEBUG != "OFF":
                            print "DesignHXNet.matchStreams: DTHot %s - DTCold %s"%(dT_hot, dT_cold)
                        
##                        print "qc %s"%qc
                        if qc > EPSILON and qh > EPSILON:

                            if abs(qc-qh) < abs(qBest-qh):
                                index=i
                                cs_best=cs
                                diff=abs(mcp_hot-mcp_cold)

                                qBest=qc
                                

                if DEBUG != "OFF":
                    print "DesignHXNet.matchStreams: best matching cold stream = %s"%cs_best
                    
                if cs_best is None:
                    del hsList[0]  # no appropriate cold stream found
                    continue

        # Check: mcp_hot < mcp_cold

                elif (qh) <= (qBest + EPSILON):
                    csMatches.append(cs_best)
                    hsMatches.append(hsList[0])

                    if DEBUG != "OFF":
                        print "DesignHXNet.matchStreams: BINGO hs/cs_best added to matches"
                        
                    del csList[index]
                    del hsList[0]
                    
                else:
                    
                    f = (qBest)/(qh)
                    hs, hsp = hs.split(f)
                            
                    if hs.fraction < MINFRACTION:
                        if DEBUG != "OFF":
                            print "DesignHXNet.matchStreams: fraction of hs too low (%s)"%hs.fraction
                            
                        del hsList[0]
                    else:
                        csMatches.append(cs_best)
                        hsMatches.append(hsList[0])
                        if DEBUG != "OFF":
                            print "DesignHXNet.matchStreams: BINGO hs-split/cs_best added to matches"
                            
                        del csList[index]
                        del hsList[0]

                    if hsp.fraction >= MINFRACTION:
                        hsList.append(hsp)
                        hsList = self.sortStreamWithMCPHours(hsList)
            

    def selectColdStreamsAP(self, csList):
        """ selects the cold streams for heat exchange above pinch from cccu """

        
        cccu = self.cccu
        if DEBUG != "OFF":
            print "=============================================="
            print "DesignHXNet.selectColdStreamsAP"
            print "=============================================="
            print "cccu = %s"%cccu
        
        # select stream components out of the cold streams in order to match with cccu
        csuList = []

        for cs in csList:

            if DEBUG == "ALL":
                print "allocating cs %s"%cs
            
            skip, rest = cs.divide(cccu.T[0], keep=1)
            for i, T1 in enumerate(cccu.T[1:]):
                T0 = cccu.T[i]

                if rest is None:
                    break
                
                elif rest.T0 > T0 + 0.001: # no energy in this interval, try the next
                    continue
                
                elif rest.T1 >= T0 + 0.001 or rest.T1 > T1 - EPSILON:   # rest is still useful (second case for condensing)
                    if rest.T1 < T1 + EPSILON:
                        part = rest
                        rest = None
                    else:
                        part, rest = rest.divide(T1)
                        if part is None:
                            continue
                    
                    # split part if necessary
                    qmaxPart = cccu.Q[i+1] - cccu.Q[i]
                    qPart = part.mcp * part.fraction * (part.T1 - part.T0)
                    if part.mcp > 0.1*INFINITE:
                        qPart = part.Q * part.fraction

                    if qPart > 0:
                        fSplit = min(qmaxPart/qPart, 1.0)
                        if fSplit*part.fraction < MINFRACTION:
                            continue
                        elif abs(1. - fSplit) >= MINFRACTION:
                            part.fraction *= fSplit
                            qPart *= fSplit

                    if DEBUG == "ALL":
                        print "qPart = %s (%s) part.fraction = %s"%(qPart, qmaxPart, part.fraction)                        

                    # add part to hsuList and increment hcca
                    if part.fraction >= MINFRACTION and qPart > EPSILON:
                        
                        if DEBUG == "ALL":
                            print "appending %s to csuList"%part
                        csuList.append(part)
                        cccu.addIncrement(i+1, -qPart)
                    
                else:
                    if DEBUG == "ALL":
                        print "breaking with rest = %s"%rest
                        
                    break

        if DEBUG != "OFF":
            print "csuList: %s"%csuList

            for cs in csuList:
                print "cold stream %s id %s up %s down %s"%(cs.name, cs.id,
                                                            cs.upstream[0].id if len(cs.upstream) > 0 else None,
                                                            cs.downstream[0].id if len(cs.downstream) > 0 else None)
            for cs in csuList:
                print "cs: %s %s->%s f %s fmcp %s Q %s"%(cs.name, cs.T0, cs.T1, cs.fraction,
                                                         cs.fraction*cs.mcp,
                                                         cs.fraction*cs.mcp*(cs.T1 - cs.T0))

        return csuList
##        self.matchStreamsGenericAP(hsList, csuList, hsMatches, csMatches)

        
    def matchStreamsGenericAP(self, hsList, csuList, hsMatches, csMatches):
        
        MAXRUN = len(csuList)*10
        run = 0

        while len(csuList) > 0 and len(hsList) > 0 and run < MAXRUN:
            if DEBUG != "OFF":
                print "\nDesignHXNet.matchStreamsGenericAP: loop with %s cold and %s hot streams"%(len(csuList), len(hsList))
                
            run +=1
             
            mcpfMax = max([mcpEff(hs) for hs in hsList ] + [1.0])
            
            hsList =  sorted(hsList, key=lambda hs: (hs.T0 - 1.0*mcpEff(hs)/mcpfMax), reverse=True) # sort and select the remaining hot stream with maximum temperature
                                                                                                # within the same temperature, those with high mass flow first
##            print "----------------------------------------------->"
##            print "DesignHXNet.matchStreamsExergyAP. sorted cs List [len %s]"%len(hsList)
##            print hsList

            # Select hot stream with lowest T0 and maximal m*cp
            hs = hsList[-1]
            i_hs = len(hsList)-1
            

            mcp_hot = hs.mcp * hs.fraction

            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsGenericAP: \n\n>>> finding match for hs %s"%hs

            csuList = sorted(csuList, key=lambda cs: (abs(cs.mcp*cs.fraction-mcp_hot) - cs.T1)) # look for best flow rate match - highest T1 first
##            print "DesignHXNet.matchStreamsGenericAP. sorted csu List [len %s]"%len(csuList)
##            print csuList
            
            cs = None
            i_cs = None
            for i, cs in enumerate(csuList):
                if hs.T1 > cs.T0 + self.DTmin + 0.001 or hs.T1 >= cs.T1 + self.DTmin - EPSILON:
                    i_cs = i
                    break
##                else:
##                    del hsList[i]
                
            if i_cs is not None:
                cs = csuList[i_cs]
            else:
                del hsList[-1]  # there can nothing be done with the current hot stream. try the next

                if DEBUG != "OFF":
                    print "DesignHXNet.matchStreamsExergyAP. deleting hot stream"
                    
                continue

            mcp_cold = cs.mcp * cs.fraction

            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsExergyAP: cold stream %s"%cs

            # truncate cs and hs and limit to the useful temperature range
            self.truncate(hs, cs, hsList, csuList)
                
            # use only the minimum required fraction of the hot and cold stream
            qSnk = mcp_cold*(cs.T1 - cs.T0) if cs.mcp < 0.1*INFINITE else cs.Q*cs.fraction
            qSrc = mcp_hot *(hs.T1 - hs.T0) if hs.mcp < 0.1*INFINITE else hs.Q*hs.fraction

            if qSrc < EPSILON or qSnk < EPSILON:
                continue

            elif qSrc > qSnk + EPSILON:
                T1Match = hs.T0 + (hs.T1 - hs.T0)*qSnk/qSrc

                if DEBUG == "ALL":
                    print "T1Match = %s hs.T1 = %s"%(T1Match, cs.T1)

                if cs.T1 <= T1Match - self.DTmin + EPSILON:        # use the lower temperature part of the hot stream if possible
                    hs_H, hs = hs.divide(T1Match, keep=1)
                    if hs_H is not None: hsList.append(hs_H)

                    if DEBUG == "ALL":
                        print "dividing hot stream hs_H %s h %s"%(hs_H, hs)

                else:                                   # ... or otherwise split the hot stream
                    f = qSnk / qSrc
                    if f < MINFRACTION:
                        continue
                    elif abs(1.0 - f) > MINFRACTION:
                        hs, hs_P = hs.split(f, lists = [hsList, self.streams_bpH])

                        if DEBUG == "ALL":
                            print "splitting hot stream with f = %s [id %s up %s||idP %s upP %s]"%(f,
                                                                                                   hs.id,
                                                                                                   hs.upstream[0].id if len(hs.upstream) > 0 else None,
                                                                                                   hs_P.id,
                                                                                                   hs_P.upstream[0].id if len(hs_P.upstream) > 0 else None)

            elif qSnk > qSrc + EPSILON:
                T0Match = cs.T1 - (cs.T1 - cs.T0)*qSrc/qSnk

                if DEBUG == "ALL":
                    print "T0Match = %s hs.T0 = %s"%(T0Match, hs.T0)
                    
                if hs.T0 >= T0Match + self.DTmin - EPSILON:        # supply to the higher temperature part of the cold stream if possible
                    cs_L, cs = cs.divide(T0Match, 1)
                    if cs_L is not None: csuList.append(cs_L)

                else:                                   # ... or otherwise split the cold stream
                    f = qSrc / qSnk
                    if f < MINFRACTION:
                        continue
                    elif abs(1.0 - f) >= MINFRACTION:
                        cs, cs_P = cs.split(f)
                        csuList.append(cs_P)

                        if DEBUG == "ALL":
                            print "splitting cold stream with f = %s\ncs: %s\ncs_P%s"%(f, cs, cs_P)
                        

            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsGenericAP. csMatch %s"%cs
                print "DesignHXNet.matchStreamsGenericAP. hsMatch %s"%hs
                print ""

            csMatches.append(cs)
            hsMatches.append(hs)
            
            del csuList[i_cs]
            del hsList[i_hs]


    def selectHotStreamsBP(self, hsList):

        hccu = self.hccu

        if DEBUG != "OFF":
            print "=============================================="
            print "DesignHXNet.selectHotStreamsBP"
            print "=============================================="

            print "TEST TEST hccu = %s"%hccu
        
        # select stream components out of the hot streams in order to match with hccu
        hsuList = []

        hsMCD = [hs for hs in hsList if hs.required]
        hsMCD =  sorted(hsMCD, key=lambda hs: hs.T0) # sort and give priority to hot streams with low T0

        for hs in hsMCD:
            if DEBUG != "OFF":
                print "allocating hs (MCD) %s"%hs
                
            rest, skip = hs.divide(hccu.T[0]) # skip the part of hs below T0 of hccu
            
            for i, T1 in enumerate(hccu.T[1:]):
                if rest is None: break
                T0 = hccu.T[i]
                if rest.T1 >= T0 + 0.001 or rest.T1 > T1 - EPSILON:   # rest is still useful
                    T1rest = rest.T1
                    if rest.T1 < T1 + EPSILON:
                        part = rest
                        rest = None
                    else:
                        rest, part = rest.divide(min(T1, rest.T1), keep=1)
                        
                    if part is None: continue


                    # split part if necessary
                    qmaxPart = hccu.Q[i+1] - hccu.Q[i]
                    qPart = part.mcp * part.fraction * (part.T1 - part.T0)
                    if part.mcp > 0.1*INFINITE: qPart = part.Q*part.fraction
                        
                    if qPart > 0:
                        if qPart > qmaxPart:
                            T1MatchEnergy = part.T0 + (T1rest - part.T0) * qmaxPart / qPart
##                            print "T1 Match = %s (T1: %s)"%(T1MatchEnergy, T1)

                            if T1MatchEnergy < T1:
                                fSplit = min(qmaxPart/qPart, 1.0)
                                if abs(1. - fSplit) >= MINFRACTION:
                                    part.fraction *= fSplit
                                    qPart *= fSplit
                                
                            else:
                                part.T1 = T1MatchEnergy
                                if rest is not None: rest.T0 = T1MatchEnergy
                                qPart = qmaxPart
                                
                        # add part to hsuList and increment hcca
                        if part.fraction >= MINFRACTION and qPart > EPSILON:
                            hsuList.append(part)
                            hccu.addIncrement(i+1, -qPart)
                    
                else:
                    break

        if DEBUG != "OFF":
            print "hccu after allocating MCD: %s"%hccu
        
        for hs in [hs for hs in hsList if not hs.required]:

            if DEBUG != "OFF":
                print "allocating hs %s"%hs
            
            rest, skip = hs.divide(hccu.T[0])
            
            for i, T1 in enumerate(hccu.T[1:]):
                if rest is None: break
                T0 = hccu.T[i]
                if rest.T0 > T0 + 0.001: # no energy in this interval, try the next
                    continue
                
                elif rest.T1 >= T0 + 0.001 or rest.T1 > T1 - EPSILON:   # rest is still useful
                    if rest.T1 < T1 + EPSILON:
                        part = rest
                        rest = None
                    else:
                        rest, part = rest.divide(min(T1, rest.T1), keep=1)
                        
                    if part is None: continue

                    # split part if necessary
                    qmaxPart = hccu.Q[i+1] - hccu.Q[i]
                    qPart = part.mcp * part.fraction * (part.T1 - part.T0)
                    if part.mcp > 0.1*INFINITE: qPart = part.Q*part.fraction

                    if qPart > 0:
                        fSplit = min(qmaxPart/qPart, 1.0)
                        if abs(1. - fSplit) >= MINFRACTION:
                            part.fraction *= fSplit
                            qPart *= fSplit

                    # add part to hsuList and increment hcca
                    if part.fraction > MINFRACTION:
                        hsuList.append(part)
                        hccu.addIncrement(i+1, -qPart)
                    
                else:
                    break

        if DEBUG != "OFF":
            print "hsuList: %s"%hsuList

        return hsuList        
##        self.matchStreamsGenericBP(hsuList, csList, hsMatches, csMatches)


    def matchStreamsGenericBP(self, hsuList, csList, hsMatches, csMatches):
        
        MAXRUN = len(hsuList)*10
        run = 0

        while len(csList) > 0 and len(hsuList) > 0 and run < MAXRUN:

            if DEBUG != "OFF":
                print "\nDesignHXNet.matchStreamsGenericBP: loop with %s cold and %s hot streams"%(len(csList), len(hsuList))

            run +=1
             
            mcpfMax = max([mcpEff(cs) for cs in csList] + [1.0])
            
            csList =  sorted(csList, key=lambda cs: (cs.T1 + 1.0*mcpEff(cs)/mcpfMax)) # sort and select the remaining hot stream with maximum temperature
                                                                                                # within the same temperature, those with high mass flow first
##            print "----------------------------------------------->"
##            print "DesignHXNet.matchStreamsExergyBP. sorted cs List [len %s]"%len(csList)
##            print csList

            # Select cold stream with highest T1 and maximal m*cp
            cs = csList[-1]
            i_cs = len(csList)-1
            
            mcp_cold = cs.mcp * cs.fraction

            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsGenericBP. searching match for cold stream %s"%cs

            hsuList = sorted(hsuList, key=lambda hs: (abs(hs.mcp*hs.fraction-mcp_cold) + hs.T1)) # look for best flow rate match, lowest T1 first
##            print "DesignHXNet.matchStreamsExergyBP. sorted hsu List [len %s]"%len(hsuList)
##            print hsuList
            
            hs = None
            i_hs = None
            for i, hs in enumerate(hsuList):
                if hs.T1 > cs.T0 + self.DTmin + EPSILON or hs.T1 > cs.T1 + self.DTmin - EPSILON:
                    i_hs = i
                    break
##                else:
##                    del hsList[i]
                
            if i_hs is not None:
                hs = hsuList[i_hs]
            else:
                del csList[-1]  # there can nothing be done with the current hot stream. try the next
                if DEBUG != "OFF": print "DesignHXNet.matchStreamsGenericBP. deleting cold stream"
                continue

            mcp_hot = hs.mcp * hs.fraction

            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsGenericBP: hot stream %s "%hs

            # truncate cs and hs and limit to the useful temperature range
            self.truncate(hs, cs, hsuList, csList)

            # use only the minimum required fraction of the hot stream
            qSnk = mcp_cold*(cs.T1 - cs.T0) if cs.mcp < 0.1*INFINITE else cs.Q*cs.fraction
            qSrc = mcp_hot *(hs.T1 - hs.T0) if hs.mcp < 0.1*INFINITE else hs.Q*hs.fraction
             
            if qSrc > qSnk + EPSILON:
                f = qSnk / qSrc
                if f < MINFRACTION:
                    continue
                elif abs(1. - f) >= MINFRACTION:
                    hs, hs_P = hs.split(f, lists=[hsuList])

                    if DEBUG != "OFF":
                        print "splitting hot stream with f = %s"%f
                    

            elif qSnk > qSrc + EPSILON:

                T0Match = cs.T1 - (cs.T1 - cs.T0)*qSrc/qSnk

                if DEBUG != "OFF":
                    print "T0Match = %s hs.T0 = %s"%(T0Match, hs.T0)
                    
                if hs.T0 >= T0Match + self.DTmin - EPSILON:        # supply to the higher temperature part of the cold stream if possible
                    cs_L, cs = cs.divide(T0Match, keep=1)
                    if cs_L is not None: csList.append(cs_L)
                    
                else:                                   # ... or otherwise split the cold stream
                    f = qSrc / qSnk
                    if f < MINFRACTION:
                        continue
                    elif abs(1. -f) >= MINFRACTION:
                        cs, cs_P = cs.split(f, lists=[csList, self.streams_apC])
                        #TODO: if a cs is split AFTER having been processed in the upstream, this update is not taken into account
                        #      -> not clear if this can happen. for the moment an error message has been included in LinkHX, is f is too high.
                        #      -> should be made more robust

                        if DEBUG != "OFF":
                            print "splitting cold stream with f = %s"%f           
                    
            if DEBUG != "OFF":
                print "DesignHXNet.matchStreamsExergyBP. csMatch %s"%cs
                print "DesignHXNet.matchStreamsExergyBP. hsMatch %s"%hs

            csMatches.append(cs)
            hsMatches.append(hs)
            
            del csList[i_cs]
            del hsuList[i_hs]
        
    def truncate(self, hs, cs, hsList, csList):
        """ auxiliary function used in matchStreamsExergyAP/BP """

        # truncate cs and limit to the useful temperature range
        T1max = hs.T1 - self.DTmin

        if cs.T1 > T1max + EPSILON:

            cs, cs_H = cs.divide(T1max)
            if cs_H is not None: csList.append(cs_H)

            if DEBUG != "OFF":
                print "Truncate: dividing cold stream at T=%s"%T1max
                print "Truncate: cs %s cs_H %s"%(cs, cs_H)

        # truncate hs and limit to the useful temperature range
        T0min = cs.T0 + self.DTmin

        if hs.T0 < T0min - EPSILON:
            hs, hs_L = hs.divide(T0min)
            if hs_L is not None: hsList.append(hs_L)

            if DEBUG != "OFF":
                print "Truncate: dividing hot stream at T=%s"%T0min
                print "Truncate: hs %s hs_L %s"%(hs, hs_L)
    

    def convertMatchesToHX(self):
        """
        Matched Streams are converted into Heatexchangers,
        and written into the database.
    
        """

        matchesH = self.match_apH + self.match_bpH
        matchesC = self.match_apC + self.match_bpC
        upsH = [ i for i, match in enumerate(matchesH) if len(match.upstream) > 0]
        upsC = [ i for i, match in enumerate(matchesC) if len(match.upstream) > 0]

        mergeHX = []
        for up in upsH:
            pu = matchesH.index(matchesH[up].upstream[0]) if matchesH[up].upstream[0] in matchesH else None
            if pu in upsC: # possible
                uppu = matchesC.index(matchesC[pu].upstream[0]) if matchesC[pu].upstream[0] in matchesC else None
                if up == uppu:              
                    if abs(matchesH[up].fraction - matchesH[pu].fraction) < 0.001 and \
                       abs(matchesC[up].fraction - matchesC[pu].fraction) < 0.001:
                        mergeHX.append((up, pu))

        popList = []
        
        while len(mergeHX) > 0:
            merge = mergeHX.pop()
            
            sHd = matchesH[merge[0]]
            sHu = matchesH[merge[1]]
            sCd = matchesC[merge[1]]
            sCu = matchesC[merge[0]]

            sHd.T1 = sHu.T1
            sHd.upstream = sHu.upstream
            sCu.T1 = sCd.T1
            sCu.downstream = sCd.downstream

            popList.append(merge[1])

            # substitute entries for merge[1] by merge[0]
            for i, mrg in enumerate(mergeHX):
                if mrg[0] == merge[1]:
                    mergeHX[i] = (merge[0], mrg[1])
                elif mrg[1] == merge[1]:
                    mergeHX[i] = (mrg[0], merge[0])

        matchesH = [match for i, match in enumerate(matchesH) if i not in popList]
        matchesC = [match for i, match in enumerate(matchesC) if i not in popList]

        if DEBUG != "OFF":
            print "ConvertMatchesToHX: number of matches after merging = %s"%len(matchesH)
            print "Hot Streams"
            for hs in matchesH:
                print hs

            print "Cold Streams"
            for cs in matchesC:
                print cs

            
        # now build heat exchangers for each match
        
        self.hxList = []
        
        NHXexisting = len(Status.int.hx)

        no = 0
        for i in xrange(len(matchesC)):
            hs = matchesH[i]
            cs = matchesC[i]
            q = cs.mcp*cs.fraction*(cs.T1-cs.T0) if cs.mcp < 0.1*INFINITE else cs.Q*cs.fraction

            if q >= MINENERGYFRACTION*self.pa.qhxMax:
                
                no += 1
                hx = self.convertConnectionsToHX(hs, cs, no, "")
                cs.hx = hx
                hs.hx = hx
            else:
                if DEBUG != "OFF":
                    print "match %s not converted to HX. energy too low"%(i+1)



        # link serial connected heat exchangers
        #TODO: check - this is probably also necessary, if streams within ap or bp are split !!

        if DEBUG != "OFF":
            print "LINKING BLOCK CHECKING SERIAL CONNECTIONS ====================================="
            print "\nHot streams"

        for match in matchesH:
            if match.hx is None: continue
            upstreamHX = match.upstreamHX()
            if upstreamHX is not None:
                if DEBUG != "OFF":
                    print "Upstream HX detected on hot side: %s (%s) ===> %s (%s)"%(upstreamHX.hx.name, upstreamHX.id,
                                                                          match.hx.name, match.id)
                f = match.fraction/upstreamHX.fraction if upstreamHX.fraction > 0 else 0.0
                self.linkHX(upstreamHX.hx, match.hx, "hot", f)

        if DEBUG != "OFF":
            print "\nCold streams"
        
        for match in matchesC:
            if match.hx is None: continue
            upstreamHX = match.upstreamHX()
            if upstreamHX is not None:
                if DEBUG != "OFF":
                    print "Upstream HX detected on cold side: %s (%s) ===> %s (%s)"%(upstreamHX.hx.name, upstreamHX.id,
                                                                           match.hx.name, match.id)
                f = match.fraction/upstreamHX.fraction if upstreamHX.fraction > 0 else 0.0
                self.linkHX(upstreamHX.hx, match.hx, "cold", f)

##        print "\nExisting HX"
        for sectorList in self.usedSectors:
            for used in sectorList:
                if used.hx is None: continue
                upstreamHX = used.upstreamHX()
                if upstreamHX is not None:
##                    print "Upstream HX detected on hot side: %s ===> %s"%(upstreamHX.hx.name, used.hx.name)

                    if used.heating:
                        side = "cold"
                    else:
                        side = "hot"
                    self.linkHX(upstreamHX.hx, used.hx, side)
        #TODO -> the search for existing HX probably does not work yet ---> CHECK
        


    def convertConnectionsToHX(self, hs, cs, number = 0, label="ap"):
        """
        Matched Streams are converted into Heatexchangers,
        and written into the database.
    
        """

        if DEBUG != "OFF":
            print "\nconvertConnectionsToHX: hs %s"%hs
            print "convertConnectionsToHX: cs %s\n"%cs
        
    ##            def __init__(self,HXId,ObjectType,ObjectNo,ComponentNo,fraction=1.0,name=""):

        hconn = HXConnection(objectType=hs.origin.objectType,
                             objectNo=hs.origin.objectNo,
                             componentNo=hs.origin.componentNo,
                             fraction=hs.fraction,
                             objectName = hs.origin.objectName,
                             streamName = hs.origin.streamName,
                             heating = False)

        cconn = HXConnection(objectType=cs.origin.objectType,
                             objectNo=cs.origin.objectNo,
                             componentNo=cs.origin.componentNo,
                             fraction=cs.fraction,
                             objectName = cs.origin.objectName,
                             streamName = cs.origin.streamName,
                             heating = True)
        
##        print "\nconvertConnectionsToHX: hconn %s"%hconn
##        print "convertConnectionsToHX: cconn %s\n"%cconn

        #estimate required UA-Value
        Thi = hs.T1
        Tho = min(max(hs.T0,cs.T0 + self.DTmin), Thi)
        Tci = cs.T0
        Tco = max(min(cs.T1, hs.T1 - self.DTmin), Tci)
        dT1 = Thi - Tco
        dT2 = Tho - Tci
        if dT1 == dT2:
            lmtd = dT1
        else:
            lmtd = (dT1 - dT2)/log(dT1/dT2)
            
        if DEBUG != "OFF":
            print "convertConnectionsToHX: Tci/Tco %s/%s Thi/Tho %s/%s lmtd %s"%(Tci,Tco,Thi,Tho,lmtd)

        # storage dimensioning -> draft based on power fractions fh/fc of Q_t (valid for constant temperature streams)
        if (hs.T1 - hs.T0) > 0:
            fh = hs.fraction*(hs.T1 - Tho) / (hs.T1 - hs.T0)
        else:
            fh = hs.fraction

        if (cs.T1 - cs.T0) > 0:
            fc = cs.fraction*(Tco - cs.T0) / (cs.T1 - cs.T0)
        else:
            fc = cs.fraction

        qhs = hs.mcp*(hs.T1-hs.T0) if hs.mcp < 0.1*INFINITE else hs.Q
        hs.dotQavg = qhs/hs.hours if hs.hours > 0 else 0.0        
        dotQmax_h = hs.dotQavg*fh

##                print "H dotQ %s mdot %s cp %s perc %s DT %s"%(dotQmax_h, hconn.stream.MassFlowAvg,hconn.stream.SpecHeatCap,hconn.percentHeatFlow,(Thi - Tho))
        qcs = cs.mcp*(cs.T1-cs.T0) if cs.mcp < 0.1*INFINITE else cs.Q
        cs.dotQavg = qcs/cs.hours if cs.hours > 0 else 0.0        
        dotQmax_c = cs.dotQavg*fc
##                print "C dotQ %s mdot %s cp %s perc %s DT %s"%(dotQmax_c, cconn.stream.MassFlowAvg,cconn.stream.SpecHeatCap,cconn.percentHeatFlow,(Tco - Tci))
        dotQmax = min(dotQmax_h, dotQmax_c)

##        print "convertConnectionsToHX: hs.mcp %s hs.Q %s cs.mcp %s cs.Q %s"%(hs.mcp, hs.mcp*(hs.T1-hs.T0), cs.mcp, cs.mcp*(cs.T1-cs.T0))
##        
##        print "convertConnectionsToHX: dotQmax_h %s hs.Q %s hs.f %s dotQmax_c %s cs.Q %s cs.f %s"%(dotQmax_h, hs.dotQavg, hs.fraction,
##                                                                                   dotQmax_c, cs.dotQavg, cs.fraction)
        dTStorage = Tco-Tci
        
        lmtd = max(lmtd, self.DTmin)
        dTStorage = max(dTStorage, self.DTmin)
        
        if lmtd > 0 and dTStorage > 0:
            UA = dotQmax/lmtd
            MC1h = dotQmax/dTStorage
        else:
            UA = 0
            MC1h = 0

##        if DEBUG != "OFF":
##            print "StreamNames before setting storage size"
##            print "hot stream %s %s"%(hs.name, hs.name)
##            print "hot stream hf",hs.hf.Q
##            print "cold stream %s %s"%(cs.name, cs.name)
##            print "cold stream hf",cs.hf.Q
            
##                vol = self.__setStorageSize(fh*hconn.stream.hf.getQ(),fc*cconn.stream.hf.getQ(), dTStorage, volMax=volMax, volMin=volMin)
        #TODO: reactivate storage-dimensioning (requires storage of q(t)-vector of heat flows !!!
        if Status.SimTool == "Plus":

            corr = correlateHF(hs.hf, cs.hf)
            
##            itStorage = 0
##            corrStorage = corr[0]
##
            it_corr = min(len(corr), MAXSTORAGETIME)
            for itShift in xrange(1, it_corr):
                corr[itShift] += corr[itShift-1]
##
##                dcorr = corr[itShift] - corrStorage
##                dit = itShift - itStorage
##                                  
##                if corr[itShift] > self.__corrTarget(itShift)and dcorr/dit > 0.5/MAXSTORAGETIME:
##                    itStorage = itShift
##                    corrStorage = corr[itShift]                  
##                    
##            print "DesignHXNet.convertConnectionsToHX: cumulative cross correlation %s"%corr
##            print "DesignHXNet.convertConnectionsToHX: itStorage %s corrStorage %s"%(itStorage, corrStorage)

            if corr[0] > 0.9:
                mc = 0 # high correlation at shift = 0 -> no storage necessary
            else:
                fcorr = corr[it_corr-1]

                
                mcMax = MC1h * MAXSTORAGETIME
                mcMin = MC1h * 0.5

                qh = fcorr*fh*hs.hf.getQ_t()
                qc = fcorr*fc*cs.hf.getQ_t()
                
                mc = self.__setStorageSize(qh, qc, dTStorage, mcMax, mcMin)
        else:
            mc = 0


##        if UA > 0: #TODO: if UA = 0, the heat exchanger should NOT even be created. for the moment commented out
                                # as this may give upstream problems

        #TODO: once consistent connection handling is implemented, this will be obsolete. Works only for one object connected to each HX
        # names should be checked-up and stored by writeToDB
        
        HXSource = hs.origin.objectName
        HXSink = cs.origin.objectName


        hx = HeatExchanger(name="HX%s_%02d"%(label, number), UA = UA, MCStorage = mc) #create local object
        hx.estimateCosts()
        
        logTrack("DesignHXNet: adding new HX %s"%([hs.name,cs.name,label]))

        hx.sinkstreams.append(cconn)
        hx.sourcestreams.append(hconn)
        
        hx.TciNom = Tci
        hx.TcoNom = Tco
        hx.ThiNom = Thi
        hx.ThoNom = Tho

        hx.dotQNom = dotQmax
        hx.lmtdNom = lmtd

        hx.sourceObjectName = HXSource
        hx.sinkObjectName = HXSink

        h = len(Status.int.hx)
        hx.HXNo = h+1

        HXID = hx.writeToDB()

        Status.int.hx.append(hx)
        Status.int.changeInHeatRecovery()

        return hx
    
        
    def linkHX(self, hxu, hxd, side = "hot", f=1.0):
        """Creates a serial connection between the heat exchangers associated
        with the streams hxu (upstream) and hsd (downstream, flow direction from hxu to hxd)

        :input hxu: upstream heat exchanger
        :input hxd: downstream heat exchanger
        :input side: side at which HXes are connected
        :input f: fraction of connection

        PRECONDITIONS FOR CORRECT OPERATION (not checked):
        hxd has to be previously connected and have exactly one connection at SIDE
        hxu has to have the correct hx-number assigned already in attribute HXNo
        """

        if DEBUG != "OFF":
            print "DesignHXNet.linkHX: linking %s side of %s to %s"%(side, hxu.name, hxd.name)
            
        if side == "hot":
            conn = hxd.sourcestreams[0]
        else:
            conn = hxd.sinkstreams[0]
            
        conn.objectType = "HX"
        conn.objectNo = hxu.HXNo-1
        conn.outletHX = hxu.id

        if f > 1.0 + EPSILON:
            logWarning("LinkHX: erroneous mass flow fraction for heat exchanger %s"%hxd.name)
            #TODO: this might arise from an erroneous management of successive splits
            f = 1.0
                                        
        conn.fraction = f

        hxd.writeToDB()
                
    def __setStorageSize(self, qh, qc, dT, mcMax=None, mcMin = 0):
        """
        Looks for an optimum storage dimension
        
        :param stream: Element of the Stream Class.
        :returns: List of Elements of the PinchStream-Connection Class.
        
        """

        dq = qh - qc
        if np.max(dq) <= EPSILON or np.min(dq) >= -EPSILON: # always positive or negative
            return 0

        dqMax = mcMax*dT

        dqcum = np.zeros(Status.Nt)
        
        for nits in xrange(2):
            dqcum[0] = dq[0] + dqcum[Status.Nt-1]
            dqcum[0] = min(max(dqcum[0], -dqMax), dqMax)
                
            for it in xrange(1, Status.Nt):
                dqcum[it] = dqcum[it-1] + dq[it]
                dqcum[it] = min(max(dq[it], -dqMax), dqMax)

##        for it in xrange(1, Status.Nt):
##            dqlin = dq[0] + dq[Status.Nt-1]*it/(Status.Nt-1)
##            dq[it] -= dqlin
        dqmax = np.max(dqcum)
        dqmin = np.min(dqcum)

        mc = (dqmax - dqmin)/ dT

        if np.isnan(mc):
            mc = 0 #security feature for avoiding crash

        if mcMax is not None:
            mc = min(mc, mcMax)
            
        if mc < mcMin:
            mc = 0

        return mc

    def __corrTarget(self, itShift):
        target = 0.8*pow(1.0*itShift/MAXSTORAGETIME, 0.5)
        return target


#==============================================================================
# auxiliary functions

    def sortStreamWithMCP(self, streamlist):
##        try:
        return sorted(streamlist, key=lambda stream: (stream.mcp*stream.fraction))
##        except:
##            return []
    

    def sortStreamWithMCPHours(self, streamlist):
        return sorted(streamlist, key=lambda stream: (stream.mcp*stream.fraction*stream.hours/8760.), reverse=True)


    def printToSplit(self):
        """ auxiliary function for debugging
        prints the streams after decomposition hot/cold & above/below
        """
        print "--------------------- STREAMS TO SPLIT -------------------"
        print "Above pinch cold"
        for elem in self.streams_apC:
            print elem
        print "Above pinch hot"
        for elem in self.streams_apH:
            print elem
        print "Below pinch cold"
        for elem in self.streams_bpC:
            print elem
        print "Below pinch hot"
        for elem in self.streams_bpH:
            print elem
        
        print "----------------------------------------------------------"
        
    def printMatches(self):
        print "------------Matches before merging---------------------"
        print "Above: "
        print "Cold:"
        for elem in self.match_apC:
            print elem
        print "Hot:"
        for elem in self.match_apH:
            print elem
        print "Below:"
        print "Cold:"
        for elem in self.match_bpC:
            print elem
        print "Hot:"
        for elem in self.match_bpH:
            print elem
        print "-------------------------------------------------------"


def mcpEff(s):
    return min(s.mcp, s.Q/5.0)*s.fraction
   
if __name__ == "__main__":
    print "HX Proposal Module";
