# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   estimateHR (Heat Recovery)
#           
#------------------------------------------------------------------------------
#           
#   Module for heat recovery calculations
#   Estimative mode based on total of heat flows
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------     
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
from constants import *
from modules.messageLogger import *
from modules.heatFlow import convertTCtoTScale, convertTtoTCScale
from GUI.status import *

HRRATIO_DEFAULT = 0.8
HRDELTAT_DEFAULT = 10

def estimateHR():
    """ estimates the heat recovery potential without need for explicitely defining heat exchanger network
       starting point: USHTotal_Tt -> process heat demand at process entry
                       USHwTotal_Tt -> waste heat demand

       QHX = min(fHR*USHw_Tt,QHXmax)

    """
    assert(Status.processData.outOfDate == False) # raise exception if called from the wrong context
    
    changeInSourceOrSink = False
    
    UPHTotal_Tt = Status.int.UPHTotal_Tt
    UPCTotal_Tt = Status.int.UPCTotal_Tt #UPCTotal contains cooling demands of cooling processes
                                                            # and mandatory cool-down of heating processes
    UPHwResTotal_Tt = Status.int.UPHwResTotal_Tt
    
    QWH_Tt = UPHwResTotal_Tt.Q_Tt(descending=True)      # initial value = UPHw (QWHProc) + QWHEq. will be reduced by QHX
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        QWH_T = Status.int.calcQ_T(QWH_Tt)
        print "UPHwResTotal", QWH_T

    QWH_Tt[:,:] += (Status.int.QWHEqTotal_Tt[:,:] + 
                    Status.int.QWHEE_Tt[:,:])

    for elem in Status.int.QWHjChanged:
        changeInSourceOrSink = changeInSourceOrSink or elem
        
    if not changeInSourceOrSink:
        logTrack("estimateHR: no change in source or sink. no need for updating results")
        return
    
    QHXProcTotal_Tt = Status.int.createQ_Tt()    # heat recovered for process heating
    UPHProcTotal_Tt = None    # heat supplied externally to processes

    QWHAmb_Tt = QWH_Tt.copy()      # initial value = UPHw (QWHProc) + QWHEq. will be reduced by QHX

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "estimateHR:"
        QWH_T = Status.int.calcQ_T(QWH_Tt)
        print "QWH before conversion", QWH_T

        UPCTotal_T = Status.int.calcQ_T(UPCTotal_Tt)
        print "UPCTotal before conversion", UPCTotal_T

    QWH_Tt[:, :] += convertTCtoTScale(UPCTotal_Tt)[:, :]

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        QWH_T = Status.int.calcQ_T(QWH_Tt)
        print "QWH after adding UPCTotal", QWH_T

#..............................................................................
# settings of the conversion

    if Status.ANo > 0:
        fHR = HRRATIO_DEFAULT   # fraction of potential heat recovery that is really recovered
    else:
        fHR = 0.0
        
    if Status.NHX > 0: logWarning("ESTIMATIVE MODE FOR HEAT RECOVERY CALCULATIONS. REAL HX network is not taken into consideration !!!!")
    
    delta_iT = int (HRDELTAT_DEFAULT / Status.TemperatureInterval + 0.5)
    assert(delta_iT <= Status.NT + 1)
#..............................................................................
#..............................................................................
#..............................................................................
    QWH_max_t = QWH_Tt[delta_iT]
    #maximum available energy
    if delta_iT > 0:
        QHXProcTotal_Tt[:-delta_iT, :] = fHR * (QWH_max_t - QWH_Tt[delta_iT:, :])
        if DEBUG in ("BASIC", "MAIN", "ALL"):
                checkSlope(QHXProcTotal_Tt, range(Status.NT + 2 - delta_iT))
        # includes shift by DTmin
        QHXProcTotal_Tt[-delta_iT:-1, :] = QHXProcTotal_Tt[Status.NT + 1 - delta_iT, :]
        QHXProcTotal_Tt[-1, :] = fHR * QWH_max_t 
    else:
        QHXProcTotal_Tt[:, :] = fHR * (QWH_max_t - QWH_Tt[:, :])
        

    QHXProcTotal_T = Status.int.calcQ_T(QHXProcTotal_Tt)
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "QHXProcTotal_T as available waste heat potential, before pinch shifting", QHXProcTotal_T

    # then shift so that QHXProcTotal always < UPH -> = Hot Composite Curve
    shift_t = np.max(QHXProcTotal_Tt - UPHTotal_Tt, 0)
    shift_t = shift_t.clip(min=0)
    QHXProcTotal_Tt = QHXProcTotal_Tt - shift_t
    QHXProcTotal_Tt = QHXProcTotal_Tt.clip(min=0)

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        for it in xrange(Status.Nt):
            for iT in xrange(1, Status.NT + 2):
                if QHXProcTotal_Tt[iT][it] <= QHXProcTotal_Tt[iT - 1][it] - 0.001:
                    print ("error in slope of QHXProcTotal[%s][%s] after shift" % (iT, it))
        
        QHXProcTotal_T = Status.int.calcQ_T(QHXProcTotal_Tt)
        print "QHXProcTotal_T after pinch shifting", QHXProcTotal_T
        checkSlope(QHXProcTotal_Tt, xrange(Status.NT + 2))
   
    # subtract recovered heat from demand -> UPHProcTotal
    UPHProcTotal_Tt = UPHTotal_Tt - QHXProcTotal_Tt

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        UPHProcTotal_T = Status.int.calcQ_T(UPHProcTotal_Tt)
        print "UPHProcTotal_T before cutting noses", UPHProcTotal_T

    # cut "noses" in UPHProcTotal
    for iT in xrange(Status.NT, -1, -1):
        UPHProcTotal_Tt[iT] = np.min((UPHProcTotal_Tt[iT + 1], UPHProcTotal_Tt[iT]), 0)

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        UPHProcTotal_T = Status.int.calcQ_T(UPHProcTotal_Tt)
        print "UPHProcTotal_T after cutting noses", UPHProcTotal_T

    QHXProcTotal_Tt = UPHTotal_Tt - UPHProcTotal_Tt

    UPHProcTotal_T = Status.int.calcQ_T(UPHProcTotal_Tt)
    UPHTotal_T = Status.int.calcQ_T(UPHTotal_Tt)
    QHXProcTotal_T = Status.int.calcQ_T(QHXProcTotal_Tt)
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "final results of heating part"
        print "UPHProcTotal_T", UPHProcTotal_T
        print "UPHTotal_T", UPHTotal_T
        print "QHXProcTotal_T", QHXProcTotal_T


    #first subtract QHX from UPC -> UPCProc
    UPCProcTotal_Tt = Status.int.createQ_Tt()
    QCXProcTotal_Tt = None
    QHXProcAmb_Tt = None

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "START COOLING CALCULATIONS"

    QHXProcTotalTC_Tt = convertTtoTCScale(QHXProcTotal_Tt)
    UPCTotal_Tt = Status.int.UPCTotal_Tt

    #..........................................
    UPCTotal_T = Status.int.calcQ_T(UPCTotal_Tt)
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "UPCTotal_T", UPCTotal_T

    QHXProcTotalTC_T = Status.int.calcQ_T(QHXProcTotalTC_Tt)
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "QHXProcTotalTC_T", QHXProcTotalTC_T
    #..........................................

    UPCTotal_max_t = UPCTotal_Tt[0]
#    UPCProcTotal_Tt[delta_iT:] = (UPCTotal_max_t - UPCTotal_Tt[delta_iT:]) - QHXProcTotal_Tt[:-delta_iT]
    delta_iTC = delta_iT*2
    UPCProcTotal_Tt[:delta_iTC, :] = UPCTotal_max_t - UPCTotal_Tt[:delta_iTC, :]           
    UPCProcTotal_Tt[delta_iTC: Status.NT + 2, :] = UPCTotal_max_t - UPCTotal_Tt[delta_iTC: Status.NT + 2, :] - QHXProcTotalTC_Tt[: Status.NT + 2 - delta_iTC, :]

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        UPCProcTotal_T = Status.int.calcQ_T(UPCProcTotal_Tt)
        print "UPCProcTotal_T before cutting noses", UPCProcTotal_T
    
    for iT in range(1, Status.NT + 2):         #cut noses -> force to ascending order !!!
        UPCProcTotal_Tt[iT] = np.max((UPCProcTotal_Tt[iT], UPCProcTotal_Tt[iT - 1]), 0)

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        UPCProcTotal_T = Status.int.calcQ_T(UPCProcTotal_Tt)
        print "UPCProcTotal_T after cutting noses", UPCProcTotal_T
        
    UPCProcTotal_Tt = UPCProcTotal_Tt[Status.NT + 1] - UPCProcTotal_Tt #now reconvert to descending ...

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        UPCProcTotal_T = Status.int.calcQ_T(UPCProcTotal_Tt)
        print "UPCProcTotal_T) after reverting order", UPCProcTotal_T
    
        UPCTotal_T = Status.int.calcQ_T(UPCTotal_Tt)
        print "UPCTotal_T before calculating UPC - UPCProc", UPCTotal_T
    
    QCXProcTotal_Tt = UPCTotal_Tt - UPCProcTotal_Tt

    # before conversion of QCXProc use for calculation of residual ...
    QCXProcTotal_max_t = QCXProcTotal_Tt[0]
    QHXProcAmb_Tt = QHXProcTotal_Tt - (QCXProcTotal_max_t - QCXProcTotal_Tt)
    for iT in range(1, Status.NT + 2):         #cut noses -> force to ascending order !!!
        QHXProcAmb_Tt[iT] = np.max((QHXProcAmb_Tt[iT], QHXProcAmb_Tt[iT - 1]), 0)
    
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        QCXProcTotal_T = Status.int.calcQ_T(QCXProcTotal_Tt)
        print "QCXProcTotal_T before conversion", QCXProcTotal_T

#    QCXProcTotal_Tt = convertTtoTCScale(QCXProcTotal_Tt)
    
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        QCXProcTotal_T = Status.int.calcQ_T(QCXProcTotal_Tt)
        print "QCXProcTotal_T after conversion", QCXProcTotal_T

    # subtract recovered heat from total available waste heat -> QHWAmb
    QHXProcAmb_max_t = QHXProcAmb_Tt[Status.NT + 1]
        
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        QWHAmb_T = Status.int.calcQ_T(QWHAmb_Tt)
        print "QWHAmb_T before subtracting QHXProcAmb\n", QWHAmb_T
    
        QHXProcAmb_T = Status.int.calcQ_T(QHXProcAmb_Tt)
        print "QHXProcAmb_T", QHXProcAmb_T
    
        print "QHXProcAmb_max = ", sum(QHXProcAmb_max_t)
    
    QWHAmb_Tt[0:delta_iT] = QWHAmb_Tt[0:delta_iT] - QHXProcAmb_max_t
    QWHAmb_Tt[delta_iT:-1] = QWHAmb_Tt[delta_iT:-1] - QHXProcAmb_max_t + QHXProcAmb_Tt[:-(delta_iT+1)]   #The last row (Status.NT+1) of QWHAmb remains unchanged.

    # assure for monotony of the result
    for iT in xrange(Status.NT+1):
        QWHAmb_Tt[iT+1] = np.min((QWHAmb_Tt[iT + 1], QWHAmb_Tt[iT]), 0)

    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "ESTIMATIVE MODE: results 1 (Totals)"
        print "UPCProcTotal: %s %s" % (UPCProcTotal_Tt[0][12], UPCProcTotal_Tt[Status.NT + 1][12])
        print "QCXProcTotal: %s %s" % (QCXProcTotal_Tt[0][12], QCXProcTotal_Tt[Status.NT + 1][12])

        UPCProcTotal_T = Status.int.calcQ_T(UPCProcTotal_Tt)
        print "UPCProcTotal_T finally ...", UPCProcTotal_T
    
        QCXProcTotal_T = Status.int.calcQ_T(QCXProcTotal_Tt)
        print "QCXProcTotal_T", QCXProcTotal_T
        
        QWHAmb_T = Status.int.calcQ_T(QWHAmb_Tt)
        print "QWHAmb_T", QWHAmb_T

#..............................................................................
# now distribute cooling generated by HR
    
# subtract the distributed heat exchanger heat
# from the energy that has to be delivered to the process
# from the equipment via the pipes

    processes = Status.prj.getProcesses()
    NK = len(Status.int.UPH_Tt)
    QHXProc_Tt = Status.int.QHXProc_Tt
    UPCProc_Tt = Status.int.UPCProc_Tt
    UPH_Tt = Status.int.UPH_Tt

    np.seterr(divide="ignore")
    for k in range(NK):
        if processes[k].heating_cooling == "heating":
            UPHwReq_Tt = Status.int.UPHwReq_Tt_hf[k].Q_Tt(mode="cooling")
            QHXProc_Tt[k] = UPH_Tt[k] * (QHXProcTotal_Tt / UPHTotal_Tt)
            UPCProc_Tt[k] = UPHwReq_Tt * (1-(QCXProcTotal_Tt / UPCTotal_Tt))
            QHXProc_Tt[k][np.isnan(QHXProc_Tt[k])] = 0 # division by 0 results 
            QHXProc_Tt[k][np.isinf(QHXProc_Tt[k])] = 0
            QHXProc_Tt[k].clip(min=0, out=QHXProc_Tt[k])
            UPCProc_Tt[k][np.isnan(UPCProc_Tt[k])] = 0
            UPCProc_Tt[k][np.isinf(UPCProc_Tt[k])] = 0
            UPCProc_Tt[k].clip(min=0, out=UPCProc_Tt[k])
        else:
            QHXProc_Tt[k] = UPH_Tt[k] * (QCXProcTotal_Tt / UPCTotal_Tt)
            QHXProc_Tt[k][np.isnan(QHXProc_Tt[k])] = 0 # division by 0 results 
            QHXProc_Tt[k][np.isinf(QHXProc_Tt[k])] = 0
            QHXProc_Tt[k].clip(min=0, out=QHXProc_Tt[k])
            
            UPCTotal_T = Status.int.calcQ_T(UPCTotal_Tt)
            QHXProc_T = Status.int.calcQ_T(QHXProc_Tt[k])
            
            if DEBUG in ("BASIC", "MAIN", "ALL"):
                print "process %s" % k
                print "UPCTotal: %s %s" % (UPCTotal_Tt[0][12], UPCTotal_Tt[Status.NT + 1][12])
                print "QHXProc : %s %s" % (QHXProc_Tt[k][0][12], QHXProc_Tt[k][Status.NT + 1][12])
                print "UPCTotal_T", UPCTotal_T
                print "QHXProc_T", QHXProc_T
    np.seterr(divide="warn")   
    if DEBUG in ("BASIC", "MAIN", "ALL"):
        print "END ESTIMATE-HR================================"

    Status.int.QHXProc_Tt = QHXProc_Tt
    Status.int.QHXProcChanged = [changeInSourceOrSink]*NK

    Status.int.UPCProc_Tt = UPCProc_Tt #results for cooling processes directly calculated and stored 
    Status.int.UPHwUsedChanged = [changeInSourceOrSink]*NK
    Status.int.QHXProcTotal_Tt = QHXProcTotal_Tt

    Status.int.QWHAmb_Tt = QWHAmb_Tt

def estimateHR_hf():

    """ carries out an estimative calculation of the heat recovery potential
        based on aggregate demand and waste heat availability data
       starting point: USHTotal_Tt -> process heat demand at process entry
                       USHwTotal_Tt -> waste heat demand

       QHX = min(fHR*USHw_Tt,QHXmax)

       version in HeatFlow_t format

    """

    if Status.processData.outOfDate == True:
        Status.processData.createAggregateDemand()
        return  # createAggregateDemand calls the heat recovery calculation 
                # itself.
        
    QHXProcTotal_Tt = Status.int.createQ_Tt()    # heat recovered for process heating
    UPHProcTotal_Tt = Status.int.createQ_Tt()    # heat supplied externally to processes

    UPHTotal_Tt_hf = Status.int.UPHTotal_Tt_hf
    UPHwTotal_Tt_hf = Status.int.UPHwTotal_Tt_hf

    QWH_Tt = UPHwTotal_Tt_hf.copy()      # initial value = UPHw (QWHProc) + QWHEq. will be reduced by QHX
    QWH_Tt.add(Status.int.QWHEqTotal_Tt_hf)
    QWH_Tt.add(Status.int.QHWEE_Tt_hf)
    
    QWHAmb_Tt = QWH_Tt.copy()      # initial value = UPHw (QWHProc) + QWHEq. will be reduced by QHX

#..............................................................................
# settings of the conversion

    if Status.ANo > 0:
        fHR = HRRATIO_DEFAULT   # fraction of potential heat recovery that is really recovered
    else:
        fHR = 0.0
    logWarning("ESTIMATIVE MODE FOR HEAT RECOVERY CALCULATIONS. REAL HX network is not taken into consideration !!!!")
    
#..............................................................................
#..............................................................................
#..............................................................................
###    QWH_max_t = QWH_Tt[delta_iT]
    #maximum available energy
    for it in range(Status.Nt):
        hcc = QWH_Tt.buildCC(it).scale(fHR)  #build hot composite curve
        ccc = UPHTotal_Tt_hf.buildCC(it)

        qhxproc, TPinch = pinchCC(hcc, ccc, HRDELTAT_DEFAULT)
        qhxproc.cutNegative()

        uphproc = diffCC(ccc, qhxproc)
        uphproc.cutNoses()

        qhxproc = diffCC(ccc, uphproc)
        
        UPHProcTotal_Tt_hf = uphproc.buildHF()    
        QHXProcTotal_Tt_hf = qhxproc.buildHF()

    #substract recovered heat from total available waste heat -> QHWAmb
    QHXProcTotal_max_t = QHXProcTotal_Tt[Status.NT + 1]
    QWHAmb_Tt[0:delta_iT] = QWH_Tt[0]
    QWHAmb_Tt[delta_iT:] = QWH_Tt[delta_iT:] - (QHXProcTotal_max_t - QHXProcTotal_Tt[iT - delta_iT])
#..............................................................................

    neededEnergyFactor = (1 - (QHXProcTotal_Tt / UPHTotal_Tt))
    # remove 0/0 errors leading to NaN values or something/0=infinity
    neededEnergyFactor[np.isnan(neededEnergyFactor)] = 1
    neededEnergyFactor.clip(min=0, out=neededEnergyFactor)

# subtract the distributed heat exchanger heat
# from the energy that has to be delivered to the process
# from the equipment via the pipes

    processes = Status.prj.getProcesses()
    QHXProc_Tt = [None] * NK
    NK = len(Status.int.UPH_Tt)

    for k in range(NK):
        if processes[k].heating_cooling == "heating":
##            print "pipe %s process %s"%(m,k)

##            UPHProcTotal_Tt[k] += UPHTotal_Tt[k] * neededEnergyFactor
##HS -> seems this numpy operation doesn't work well ??? substitute by explicit loop
##TODO: rewrite in more efficient way
       
            for iT in xrange(Status.NT + 2):
                for it in xrange(Status.Nt):
                    if UPHTotal_Tt[iT][it] > 0:
                        QHXProc_Tt[k][iT][it] = UPH_Tt[k][iT][it] * (QHXProcTotal_Tt[iT][it] / UPHTotal_Tt[iT][it])
                    else:
                        QHXProc_Tt[k][iT][it] = 0
        else:
            QHXProc_Tt[k] = Status.int.createQ_Tt()
        
#..............................................................................                

    Status.int.QHXProc_Tt = QHXProc_Tt
    Status.int.QHXProcTotal_Tt = QHXProcTotal_Tt

    Status.int.QWHAmb_Tt = QWHAmb_Tt


def checkSlope(Q_Tt, checkRange, descending=False):
    """ security check whether Q_T array is monotonous
    """
    
    count = 0
    for it in xrange(Status.Nt):
        for iT in checkRange:
            if iT > 0:
                if not descending:
                    if Q_Tt[iT][it] <= Q_Tt[iT - 1][it] - 0.001:
                        logDebug("error in slope of QWH[%s][%s]s" % (iT, it))
                        count += 1
                else:
                    if Q_Tt[iT][it] >= Q_Tt[iT - 1][it] + 0.001:
                        logDebug("error in slope of QWH[%s][%s]" % (iT, it))
                        count +=1
    logTrack("checkSlope found %s errors"%count)
                    
                          
