# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#   HeatExchanger
#   (ActualHXNet V2)
#   Package for simulation of heat exchanger network
#            
#------------------------------------------------------------------------------
#            
#   Class HeatExchanger: carries out calculations for heat exchangers
#   Class HXSegment: calculation of one heat exchanger segment
#   Function buildHXNet
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.status import *
from modules.messageLogger import logTrack,logError,logWarning,logDebug
from modules.heatFlow import HeatFlow_t
from database.dbio import check
import database.dbio as dbio
from constants import INFINITE, DEBUG, EPSILON, SNAPSHOTS
import math
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, noneCheck
import numpy as np
from copy import copy

from heatStorage import Storage

MINDT = 1.0   # Minimum temperature difference for control of charge/discharge/direct mode
TENV  = 18  # Temperature of environment for heat storage; = initial temperature of storage tanks
NSTORAGE = 10    # Number of nodes for storage tank model
DELTATINF = 0.001 # finite non-zero temperature difference assumed for circuits with Cc/Ch = INF

MAXITSDEBUG = 48 # Maximum number of time steps for which a detail output is created for DEBUG != OFF

CONVERGENCESERIALHXLOOP = 1.e-4 #max. relative variation between successive iterations (sum(abs(dqx))

class HeatExchanger(object):
    """ class representation of a heat exchanger"""
    def __init__(self, UA = None, id = None,name = "",HXNo=None,MCStorage = 0,writeAllowed=True):
        """
        :param self.MCStorage: Thermal mass of heat storage on cold side
        :type self.MCStorage: double

        """
        self.id = id
        self.HXNo = HXNo # numbering from 1 to NH !
        self.name = name
        
        self.UA = UA
        if (UA is None or UA <= 0) and Status.SimTool != "Plus": # in plus-mode UA < 0 means direct exchange w/o HX
            logWarning("HeatExchanger.__init__: heat exchanger %s has invalid UA value %s"%(name,UA))
            self.UA = EPSILON #avoid zero division error ! TODO: should be solved mor elegant with an if -> qhx = 0

        self.qsrc = None
        self.qsnk = None

        self.qx = Status.int.createQ_t()
        self.qSrcRes0 = 0.0
        self.qSnkRes0 = 0.0
        self.converged = False

        self.Qx = 0.0   # yearly energy exchanged at heat exchanger
        self.hop = 0
        self.QHX = None # yearly energy delivered to sink
        
        self.sourceNo = None
        self.sourceType = None
        self.sinkNo = None
        self.sinkType = None

        self.sourceObjectName = None
        self.sinkObjectName = None

        self.fluidIdSink = 0
        self.fluidIdSource = 0

        self.TciAvg = None
        self.ThiAvg = None
        self.TcoAvg = None
        self.ThoAvg = None

        self.TciNom = None
        self.ThiNom = None
        self.TcoNom = None
        self.ThoNom = None

        self.dotQNom = None
        self.dotQAvg = None
        self.lmtdNom = None
        self.lmtdAvg = None

        self.area = None
        self.type = None

        self.firstRun = True #is reset to false after first call to calculate()

        self.writeResultsToDBAllowed = writeAllowed
        
        self.sourcestreams = []
        self.sinkstreams = []
        self.deletedSource = []
        self.deletedSink = []

        self.turnKeyPrice = None
        self.OMFix = None
        self.OMVar = None

        self.MCStorage = MCStorage
        self.storage = Storage(NSTORAGE, self.MCStorage, TENV) if self.MCStorage > 0 else None
        
        self.Kh0 = 0.0
        self.Kh1 = 0.0
        self.Kc0 = 0.0
        self.Kc1 = 0.0

        self.Ts_max = 9999.9

    def __repr__(self):
        return "HX id %s name %s no. %s UA %s" % (self.id,self.name,self.HXNo,self.UA)
            
    def calculate(self):
        """Calculates the heat flows over a given heat exchanger

        :param self.qsrc: Heat flow of source
        :type  self.qsrc: HeatFlow_t
        :param self.qsnk: Heat flow of sink
        :type  self.qsnk: HeatFlow_t
        :param self.MCStorage: Thermal mass of heat storage on cold side
        :type self.MCStorage: double
            
        :returns: self.qhx: heat supplied to sink
        :type   self.qhx: HeatFlow_t
        :returns: self.qsrcUsed: heat extracted from source
        :type   self.qsrcUsed: HeatFlow_t

        """
        if DEBUG != "OFF":
            logTrack("-----------------------")
            logTrack("calculating HX[%s] %s"%(self.HXNo,self.name))

        changeInSourceOrSink = self.getSourceAndSinkHeatFlows()
        self.changed = (changeInSourceOrSink or self.firstRun)
        if not (changeInSourceOrSink or self.firstRun):
            logTrack("HeatExchanger.calculate HX %s already converged"%self.name)
            self.storeResults()
            return 0
        
        if self.qsrc is None or self.qsnk is None:
            logError("HeatExchanger.calculate no source or sink specified. cannot calculate HX %s"%self.name)
            return -1
        
        qh = self.qsrc
        qc = self.qsnk

        if DEBUG != "OFF":
            logTrack("source %s (%s MWh) sink %s (%s MWh)"%(qh.name,qh.Q,qc.name,qc.Q))

        Thi = qh.getT1()
        Tho_min = qh.getT0()
        
        Tci = qc.getT0()
        Tco_max = qc.getT1()

        fluid_h = self.qsrc.id
        fluid_c = self.qsnk.id

        TCondH = self.qsrc.fluid.TCond
        TCondC = self.qsnk.fluid.TCond
        
        y = Thi - Tci
        #### here control y > 0 missing ?
         
        Ch1, QhLat, Ch0 = qh.getC(cond=True)
        Cc1, QcLat, Cc0 = qc.getC(cond=True)

        #TODO: for the moment security feature as QhLat is used in some general places also for non-condensing streams.
        #to be improved ...
        if QhLat is None: QhLat = Status.int.createQ_t()
        if QcLat is None: QcLat = Status.int.createQ_t()

        qx = Status.int.createQ_t()
        qxC = Status.int.createQ_t()
        qxH = Status.int.createQ_t()
        Tho = Status.int.createQ_t()
        Tco = Status.int.createQ_t()

        lmtd = Status.int.createQ_t()

        self.Thi_lastNonZero = None
        self.Tci_lastNonZero = None
        self.Cc_lastNonZero = 0
        self.Ch_lastNonZero = 0

        #init value for storage temperatures
        if self.storage:
            self.storage.reset()
            Qs_o = self.storage.q
            
        qLoss = 0.0
        dqLoss = 0

        qh_t = qh.getQ()
        qc_t = qc.getQ()

        chargePeriods = []
        dischargePeriods = []

        wTc = Status.int.createQ_t()    #weighting factors for calculation of T-avgs
        wTh = Status.int.createQ_t()
        
        for it in xrange(Status.Nt):
            if it in SNAPSHOTS:
                print "\nit: %s HX %s ------------------"%(it, self.name)
            # storage variation due to mixing and thermal losses
            if self.MCStorage > 0:
                dqLoss = self.storage.stepLosses(Status.TimeStep)
                qLoss += dqLoss
                    
                if DEBUG != "OFF":
                    if it < MAXITSDEBUG: logTrack("Storage temperatures: %s"%self.storage.T)

                    if it in SNAPSHOTS:
                        print "it: %s - storage state"%it
                        self.storage.printState()

            #set direct mode parameters as default. Modified below for other 
            qxmax = min(qh_t[it], qc_t[it])
            
            if Thi[it] < TCondH:
                Ch = Ch0[it]
            else:
                Ch = Ch1[it]
            ThiHX = Thi[it]
            if np.isinf(Ch):
                Ch = qh_t[it]/DELTATINF

            if Tci[it] > TCondC:
                Cc = Cc1[it]
            else:
                Cc = Cc0[it]
            if np.isinf(Cc):
                Cc = qc_t[it]/DELTATINF
            TciHX = Tci[it]

            # keep full flow rates of source and sink
            ChSrc = Ch
            CcSnk = Cc

            
            #================================================================================================
            # controller 1. select operation mode
            #================================================================================================
            #select mode of operation: off/direct/charge/discharge
            if qh_t[it] == 0 and qc_t[it] == 0:
                mode = "off"
            elif self.MCStorage > 0:

                if qh_t[it] == 0:
                    if Tci[it] < self.storage.Ttop - MINDT:
                    #TODO: in reality here the criterio is Tso_average for a given massflow
                        mode = "discharge"
                    else:
                        mode = "off"
                elif qc_t[it] == 0:
                    if Thi[it] > self.storage.Tav + MINDT: #25.04.2012: changed from Ts[-1] to Tavg, as Ts[-1] gave problems with very stratified stores
                        mode = "charge"
                    else:
                        mode = "off"
                else:                       # both source and sink available
                    qsrcMax = qh_t[it]*(1 - qh.trim(it, Tci[it]+MINDT))
                    qsnkMax = qc_t[it]*qc.trim(it, Thi[it]-MINDT)

                    if qsrcMax >= qsnkMax and Thi[it] > self.storage.Ttop + MINDT:   # src > snk and Ts ok -> charge
                        mode = "directAndCharge"
                    elif qsrcMax <= qsnkMax and self.storage.Ttop > Tci[it]:         # src < snk and Ts ok -> discharge
                        mode = "directAndDischarge"
                    else:
                        mode = "direct"

                    if it in SNAPSHOTS:
                        print "it: %s (ctrl1) qsrcMax %s qsnkMax %s"%(it, qsrcMax, qsnkMax)
                                                    
            elif qh_t[it] == 0 or qc_t[it] == 0:
                mode = "off"
            else:
                mode = "direct"

            if y[it] <=0:   # negative temperature difference between Thi and Tci -> no direct heat exchange possible
                if mode == "directAndCharge":
                    mode = "charge"
                elif mode == "directAndDischarge":
                    mode = "discharge"
                elif mode == "direct":
                    mode = "off"

            if DEBUG != "OFF":
                if it < MAXITSDEBUG:
                    logTrack("it: %s HX - mode: %s qh %s Thi %s qc %s Tci %s Ts_top %s"%(it,mode,qh_t[it],Thi[it],qc_t[it],Tci[it],
                                                                                         self.storage.Ttop if self.storage else "-"))
                    logTrack("it: %s Tco_max %s Tho_min %s"%(it,Tco_max[it],Tho_min[it]))
                        

            if it in SNAPSHOTS:
                print("it: %s (ctrl1) HX - mode: %s qh %s Thi %s qc %s Tci %s Ts_top %s"%(it,mode,qh_t[it],Thi[it],qc_t[it],Tci[it],
                                                                                          self.storage.Ttop if self.storage else "-"))
                print("it: %s (ctrl1) Tco_max %s Tho_min %s"%(it,Tco_max[it],Tho_min[it]))

            #================================================================================================
            # controller 2. determine flow rates and inlet temperatures in heat exchanger
            #================================================================================================

            #prepare input for heat exchanger calculation depending on mode
            if mode == "charge":
                qxmax = qh_t[it]
                
                if Thi[it] < TCondH:
                    Cc = min(Ch,self.MCStorage)  #only sensible heat extracted at hot side
                else:
                    if self.storage.Tbot + MINDT >= TCondH or QhLat[it]==0:  #no condensation -> only sensible heat
                        Cc = min(Ch,self.MCStorage)
                    else:   #condensation possible
                        Cc = QhLat[it]/(TCondH - self.storage.Tbot) #obtain maximum temperature approach to condensing temperature
                        Cc = min(Cc,self.MCStorage)

                TciHX = self.storage.getTo(Cc)
                    
            elif mode == "discharge":
                qxmax = 0
                Ch = 0
                Cc = 0 

            elif mode == "directAndDischarge":
                CcSnk = Cc
                dqStorageMax = self.MCStorage*np.average(np.maximum((self.storage.T - Tci[it]), 0))
                                  
                qxmax = min(qsrcMax, qsnkMax)

                CcStorage = (min(max(qsnkMax - qsrcMax, 0), dqStorageMax)/dqStorageMax)*self.MCStorage if dqStorageMax > 0 else 0.0
                CcStorage = min(CcStorage, dqStorageMax/(qxmax+dqStorageMax)*CcSnk if (qxmax+dqStorageMax > 0) else 0.0)  # constrain if low amount of energy stored
                if qsnkMax > 0: # require a minimum flow across the heat exchanger
                    CcStorage = min(CcStorage, CcSnk*(1.0 - qxmax/qsnkMax))
                Cc = max(CcSnk - CcStorage, 0)
                CcSnkUsed = Cc + CcStorage
                
                if DEBUG != "OFF" and it < MAXITSDEBUG:
                    logTrack("directAndDischarge-Mode: Cc %s CcSnkUsed %s Ch %s"%(Cc,CcSnkUsed,Ch))

                if it in SNAPSHOTS:
                    print "it: %s (ctrl2) mode %s qsrcMax %s qsnkMax %s qStorageMax %s"%(it, mode, qsrcMax, qsnkMax, dqStorageMax)
                    print "it: %s (ctrl2) mode %s Cc %s CcStorage %s CcSnkUsed %s"%(it, mode, Cc, CcStorage, CcSnkUsed)

            elif mode == "direct":
                pass #default mode. parameters calculated above

            elif mode == "directAndCharge":
                # first calculate direct exchange
                TciHX = Tci[it]
                Cc = CcSnk
                ThoHX, TcoHX, qx[it], lmtd[it] = self.calculateHX(it, Ch, Cc, ThiHX, TciHX, INFINITE, TCondH, QhLat[it], None if Ch0 is None else Ch0[it])

                qsnkMax = min(qsnkMax, qx[it])  # reduce estimation for available heat for storage charge calculations
                
                if it in SNAPSHOTS:
                    print "directAndCharge (inputs): Ch %s Cc %s Thi %s Tci %s qxmax %s TCondH %s QhLat %s"%(Ch, Cc, ThiHX, TciHX, INFINITE, TCondH, QhLat[it])
                    print "it: %s directAndCharge - checking direct exchange. qx-direct %s qxmax %s ThoHX %s Tci %s"%(it, qx[it], INFINITE, ThoHX, Tci[it])

                # check if exchange is constrained by heat exchanger size. in this case switch to direct exchange only
                if (TcoHX < self.storage.Tbot):
                    mode = "direct"
                    CcStorage = 0
                    CcSnkUsed = CcSnk
                    
                    if it in SNAPSHOTS:
                        print "Switching mode from directAndCharge to direct !"
                    
                else:
                    # maximum possibilities of exchange for determination of flow rates
                    dqStorageMax = max(self.MCStorage*(Thi[it] - self.storage.Tav - MINDT), 0)
                                      
                    qxmax = min(qsrcMax, qsnkMax+dqStorageMax)
                    dqStorageMax = max(min(dqStorageMax, qxmax-qsnkMax), 0)

                    TcoHXTarget = min(Tco_max[it], ThoHX)  # priority to cover 100% of sink heat demand if possible
                    TsTarget = self.storage.Ttop

                    if dqStorageMax > 0:
                        TcoHXTarget = max(TcoHXTarget, TsTarget)

                    CcSnkUsed = CcSnk

                    # maximum flow rate CcSnkUsed in order to reach at least Ttop
    ##                if (TcoHXTarget - Tci[it]) > EPSILON:

                    CcSnkMax = qsnkMax / max((TcoHXTarget - Tci[it]), 0.1)
                    CcSnkUsed = min(CcSnkUsed, CcSnkMax)
                        
                    CcStorage = min(self.MCStorage, Ch*dqStorageMax/qxmax) if qxmax > 0 else 0
##                    CcStorage = min(max(Ch-CcSnkUsed, 0), self.MCStorage)     
                    if Thi[it] < TCondH:
                        pass
                    else:
                        if self.storage.Tbot + MINDT >= TCondH or QhLat[it]==0:  #no condensation -> only sensible heat
                            pass
                        else:   #condensation possible
                            CcStorage = min(QhLat[it]/(TCondH - self.storage.Tbot), self.MCStorage) #obtain maximum temperature approach to condensing temperature
             
                    #security feature for avoiding negative heat contribution to cold stream
                    if self.storage.Tbot < TcoHXTarget:
                        CcStorageMax = dqStorageMax/(TcoHXTarget - self.storage.Tbot)
                        CcStorage = min(CcStorage, CcStorageMax)

    ##                    if Tci[it] < TCondH - MINDT and Thi[it] >= TCondH: 
    ##                        CcStorageMax = max((Ch*(Thi[it] - TCondH) + QhLat[it])/(Tci[it]-self.storage.Tbot),0)
    ##                    else:
    ##                        CcStorageMax = max(Ch*(Thi[it] - Tci[it] - MINDT)/(Tci[it]-self.storage.Tbot),0)
    ##                    CcStorageMax = min(CcStorageMax, dqStorageMax/(Tci[it]-self.storage.Tbot))

                    Cc = CcSnkUsed + CcStorage

                    ToStorage = self.storage.getTo(CcStorage)                
                    TciHX = (CcSnkUsed*Tci[it]+CcStorage*ToStorage)/Cc


                    if it in SNAPSHOTS:
                        print "it: %s (ctrl2) mode %s qsrcMax %s qsnkMax %s qStorageMax %s"%(it, mode, qsrcMax, qsnkMax, dqStorageMax)
                        print "it: %s (ctrl2) mode %s Cc %s CcStorage %s CcSnkUsed %s"%(it, mode, Cc, CcStorage, CcSnkUsed)
                    
            else:               ###this else is only needed for print command !
                Cc = None
                Ch = None
                TciHX = None
                ThiHX = None
                
            if DEBUG != "OFF":
                if it < MAXITSDEBUG:
                    logTrack("it: %s mode %s Cc %s TciHX %s Ch %s ThiHX %s"%(it,mode,Cc,TciHX,Ch,ThiHX))

    
            qx[it] = 0 #default values
            lmtd[it] = 0

            #================================================================================================
            # calculate heat exchanger
            #================================================================================================

            
            if mode in ["direct","charge","directAndCharge","directAndDischarge"]: #in other modes no HX Calculation is necessary

                if it in SNAPSHOTS:
                    print "HX calculations (inputs): Ch %s Cc %s Thi %s Tci %s qxmax %s TCondH %s QhLat %s"%(Ch, Cc, ThiHX, TciHX, qxmax, TCondH, QhLat[it])

                ThoHX, TcoHX, qx[it], lmtd[it] = self.calculateHX(it, Ch, Cc, ThiHX, TciHX, qxmax, TCondH, QhLat[it], None if Ch0 is None else Ch0[it])

            #================================================================================================
            # calculate storage and mixing at cold side
            #================================================================================================

            if mode == "direct":
                Tho[it] = ThoHX
                qxH[it] = qx[it]
                Tco[it] = TcoHX
                qxC[it] = qx[it]
                CcSnkUsed = Cc
                ChSrcUsed = Ch
                                                                                    
            elif mode in ["discharge","directAndDischarge"]:

                if mode=="directAndDischarge":                                                    
                    Tho[it] = ThoHX
                    qxH[it] = qx[it]
                    ChSrcUsed = Ch
                else:
                    Tho[it] = Thi[it]
                    qxH[it] = 0
                    ChSrcUsed = 0
                    Cc = 0

                if it in SNAPSHOTS:
                    print "storage before charging"
                    self.storage.printState()

                qStorageMax = max(qc_t[it] - qxH[it], 0)
                CcStorage, qStorage = self.storage.get_CdQ(Cc - CcSnk, qStorageMax, Tci[it])

                if it in SNAPSHOTS:
                    print "storage before charging"
                    self.storage.printState()
                    print "result of getCdQ(%s, %s, %s)=%s, %s"%(Cc - CcSnk, qStorageMax, Tci[it], CcStorage, qStorage)

                CcSnkUsed = Cc + CcStorage

                qxC[it] = qxH[it] + qStorage
                Tco[it] = Tci[it] + qxC[it]/CcSnk

                TcoStorage = Tci[it] + qStorage / self.MCStorage

                dqStorage = self.storage.charge(-CcStorage, Tci[it])

                if it in SNAPSHOTS:
                    print "dqStorage = %s"%dqStorage
                    print "storage after charging"
                    self.storage.printState()
                    print "directAndDischargeMode: it: %s qStorageMax = %s qStorage = %s"%(it, qStorageMax, qStorage)
                    print "directAndDischargeMode: it: %s Cc %s CcStorage %s CcSnkUsed %s"%(it, Cc, CcStorage, CcSnk)
                    print "directAndDischargeMode: it: %s Tco %s TcoHX %s TcoStorage %s"%(it,
                                                                                          Tco[it],
                                                                                          TcoHX if mode=="directAndDischarge" else "-",
                                                                                          TcoStorage)


            elif mode == "charge":
                #TODO: limitation of temperature at cool-down not yet programmed
                Tho[it] = ThoHX
                qxH[it] = qx[it]
                               
                Tco[it] = Tci[it]
                qxC[it] = 0

                CcSnkUsed = 0
                ChSrcUsed = Ch

                dqStorage = self.storage.charge(Cc, TcoHX)

            elif mode == "directAndCharge":
                #TODO: limitation of temperature is not yet exactly correct. may lead to underestimation of heat stored.

                Tho[it] = ThoHX

                if TcoHX < Tci[it] - EPSILON: # heat exchanger size not sufficient to cover both storage and demand
                    CcSnkUsed = 0

                    # calculate dqStorage at really obtained TcoHX                    
                    CcStorage, dqStorage = self.storage.get_CdQ(CcStorage, qx[it], TcoHX)
                    dqStorage = self.storage.charge(CcStorage, TcoHX)

                    if dqStorage > qx[it]:
                        if DEBUG != "OFF": print "WARNING: dqStorage > qx at it %s"%it
                        dqStorage = qx[it]

                    ChSrcUsed = Ch*dqStorage/qx[it] if qx[it] > 0 else 0.0
                    qxH[it] = dqStorage
                    qx[it] = dqStorage
                    qxC[it] = 0.0
                    Tco[it] = Tci[it]
                
                else:
                        
                    qxH[it] = qx[it]
                    ChSrcUsed = Ch
                    CcSnkUsed = Cc - CcStorage
                    
                    # calculate dqStorage at really obtained TcoHX                    
                    dqStorage = self.storage.charge(CcStorage, TcoHX)

                    if it in SNAPSHOTS:
                        print "it: %s Cc %s CcStorage %s CcSnkUsed %s dQStorage %s"%(it, Cc, CcStorage, CcSnkUsed, dqStorage)

                    qxC[it] = qxH[it] - dqStorage
                    if qxC[it] > qsnkMax: # security feature - necessary, as from the previous Tco < Tco_max is not guaranteed
                                          # TODO: not clear if qxC < 0 has to be controlled, too.
                                          
                        dqerr = qxC[it] - qsnkMax
                        qx[it] -= dqerr
                        qxH[it] -= dqerr
                        qxC[it] = qsnkMax

                    Tco[it] = Tci[it] + qxC[it]/CcSnk

                Tho[it] = Thi[it] - qxH[it]/ChSrc if ChSrc > 0 else Thi[it]
                
                if Thi[it] > TCondH and Tho[it] < TCondH and QhLat[it] > 0: # condensing case
                    dq0 = max(qxH[it] - QhLat[it] - Ch*(Thi[it] - TCondH), 0.0)
                    Tho[it] = TCondH - dq0/Ch0[it] if Ch0 is not None and Ch0[it] is not None and Ch0[it] > 0 else TCondH
                
            else:
                Tho[it] = Thi[it]
                qxH[it] = 0
                
                Tco[it] = Tci[it]
                qxC[it] = 0

                CcSnkUsed = 0
                ChSrcUsed = 0

            #================================================================================================
            # energy balances for the time step
            #================================================================================================

            if self.storage:
                dQs = self.storage.q-Qs_o
                Qs_o = self.storage.q
            else:
                dQs = 0.0

            dqerr = qxH[it] - qxC[it] - dQs - dqLoss #energy balance error for current time step

            if DEBUG != "OFF":
                if it < MAXITSDEBUG:
                    logTrack("it: %s lmtd %s Tco %s Tci %s Thi %s Tho %s qxH %s qxC %s dQs %s dQLoss %s dQerr %s"%
                             (it,lmtd[it],Tco[it],Tci[it],Thi[it],Tho[it],qxH[it],qxC[it],dQs,dqLoss,dqerr))
                if abs(dqerr) > 1.e-4:
                    logWarning("it %s Energy Balance Error in HX %s = %s kWh"%(it,self.name,dqerr))

            if it in SNAPSHOTS:
                print("it: %s lmtd %s Tco %s Tci %s Thi %s Tho %s qxH %s qxC %s dQs %s dQLoss %s dQerr %s"%
                         (it,lmtd[it],Tco[it],Tci[it],Thi[it],Tho[it],qxH[it],qxC[it],dQs,dqLoss,dqerr))

            if abs(dqerr) > 1.e-4:
                print("it %s: Energy Balance Error in HX %s = %s kWh - mode %s"%(it, self.name, dqerr, mode))

##            if qxC[it] == 0:
##                wTc[it] = 0
##            else:
##                wTc[it] = CcSnkUsed
##
##            if qxH[it] == 0:
##                wTh[it] = 0
##            else:
##                wTh[it] = ChSrcUsed
            wTc[it] = 1.0 * (qxC[it] > EPSILON)
            wTh[it] = 1.0 * (qxH[it] > EPSILON)


        #================================================================================================
        # final checking, book-keeping and storage of results
        #================================================================================================

        difTco = (Tco - Tco_max)*qxC
        if np.max(difTco) > 0:
            logDebug("WARNING: Tco > Tco_max in heat exchanger %s "%self.name)
            Tco = np.min((Tco, Tco_max), 0)

        difTho = (Tho_min - Tho)*qxH
        if np.max(difTho) > 0:
            logDebug("WARNING: Tho < Tho_min in heat exchanger %s "%self.name)
            Tho = np.max((Tho, Tho_min), 0)


        # security feature: avoid Tco < Tci and Tho > Thi
        criterium = (Tco < Tci - EPSILON)
        if np.any(criterium) and DEBUG != "OFF":
            print "HeatExchanger %s: WARNING - Tco < Tci at it = %s"%(self.name, np.where(criterium))
        Tco = np.maximum(Tco, Tci)

        criterium = (Tho > Thi + EPSILON)
        if np.any(criterium) and DEBUG != "OFF":
            print "HeatExchanger %s: WARNING - Tho > Thi at it = %s"%(self.name, np.where(criterium))
        Tho = np.minimum(Tho, Thi)

##        print "chargePeriods:"
##        for period in chargePeriods:
##            print "it %s dQs %s"%period
##
##        print "dischargePeriods:"
##        for period in dischargePeriods:
##            print "it %s qxC %s"%period

        # calculation of variation between iterative calls
        qdif = qx - self.qx
        dif = np.sum(np.abs(qdif))
        qxsum = np.sum(qx)
        if  qxsum > 0:
            rdif = dif/qxsum
        elif dif > EPSILON:
            rdif = 1.0
        else:
            rdif = 0
        
        self.qx = qx
        self.lmtd = lmtd
        
        self.qhx = HeatFlow_t(fluid_id=fluid_c,T0=Tci,T1=Tco,Q_t=qxC, componentNo=self.qsnk.componentNo)
        self.qsrcUsed = HeatFlow_t(fluid_id=fluid_h,T0=Tho,T1=Thi,Q_t=qxH, componentNo=self.qsrc.componentNo)

        self.qsrcRes = HeatFlow_t(fluid_id=fluid_h,T0=Tho_min,T1=Tho,Q_t=qh_t-qxH, componentNo=self.qsrc.componentNo)
        self.qsnkRes = HeatFlow_t(fluid_id=fluid_c,T0=Tco,T1=Tco_max,Q_t=qc_t-qxC, componentNo=self.qsnk.componentNo)

        difSrcSnk = max(abs(self.qsrcRes.Q - self.qSrcRes0), abs(self.qsnkRes.Q - self.qSnkRes0))
        absSrcSnk = max(self.qsrcRes.Q, self.qSrcRes0, self.qsnkRes.Q, self.qSnkRes0)
        rdifSrcSnk = difSrcSnk / absSrcSnk if absSrcSnk > 0 else 0.0

        self.qSrcRes0 = self.qsrcRes.Q
        self.qSnkRes0 = self.qsnkRes.Q

        self.converged = (max(rdif, rdifSrcSnk) < CONVERGENCESERIALHXLOOP)

        dQS = (self.storage.Tav - TENV)*self.MCStorage if self.storage else 0.0
                                                                                           
        dQerr = self.qsrcUsed.Q - self.qhx.Q - dQS - qLoss
        if abs(dQerr) > 1.0:
            logWarning("Energy Balance Error in HX %s = %s kWh"%(self.name,dQerr))

        self.storeResults() #writes results into Status.int.QHXProc/Eq/... and QWHProc/Eq/...Used - Arrays

        wTcSum = max(np.sum(wTc),1.e-10) # max: avoid zero division
        self.TciAvg = np.sum(Tci*wTc)/wTcSum
        self.TcoAvg = np.sum(Tco*wTc)/wTcSum

        wThSum = max(np.sum(wTh),1.e-10) # max: avoid zero division
        self.ThiAvg = np.sum(Thi*wTh)/wThSum
        self.ThoAvg = np.sum(Tho*wTh)/wThSum

        self.firstRun = False

        if DEBUG != "OFF":
            logTrack("=============================================")
            logTrack("results HX[%s] %s"%(self.HXNo,self.name))
            logTrack("QHX.Q = %s (convergence abs: %s / rel: %s)"%(self.qhx.Q, dif, rdif))
            logTrack("QWHUsed.Q = %s"%self.qsrcUsed.Q)
            logTrack("QWHRes.Q  = %s"%self.qsrcRes.Q)
            logTrack("QSnkRes.Q = %s"%self.qsnkRes.Q)
            logTrack("dQStorage = %s dQerr = %s"%(dQS,dQerr))
            logTrack("HXdata: lmtd %s Tco %s Tci %s Thi %s Tho %s"%(max(lmtd),self.TcoAvg,self.TciAvg,self.ThiAvg,self.ThoAvg))

            for it in range(36):
                logTrack("it: %s lmtd %s Tco %s Tci %s Thi %s Tho %s"%(it,lmtd[it],Tco[it],Tci[it],Thi[it],Tho[it]))

        self.Qx = self.qsrcUsed.Q
        self.hop = sum(self.qx>0)

        return max(rdif, rdifSrcSnk)
    
    def calculateHX(self, it, Ch, Cc, ThiHX, TciHX, qxmax, TCondH, QhLat, Ch0):
        """
        Calculates the heat exchanger kernel for one time step

        Calculation is based on the maximum possible heat exchange for Ch, Cc ,Thi and Tci, but
        constrained by qxmax.

        Input parameters TCondH, QhLat and Ch0 required for the case of condensation on the hot side
        """

        #check for special case zero flow
        if Ch == 0 or Cc == 0:
            qx = 0
            lmtd = 0

            if Ch == 0: ThoHX = TciHX
            else: ThoHX = ThiHX

            if Cc == 0: TcoHX = ThiHX
            else: TcoHX = TciHX

        #check for special case: same set of input parameters already calculated
        #-> speed-up of calculations for on/off time schedules !
        elif TciHX == self.Tci_lastNonZero and ThiHX==self.Thi_lastNonZero \
             and Ch == self.Ch_lastNonZero and Cc == self.Cc_lastNonZero:
            qx = self.qx_lastNonZero
            lmtd = self.lmtd_lastNonZero

            TcoHX = self.Tco_lastNonZero
            ThoHX = self.Tho_lastNonZero
#                continue

        else:    
            if DEBUG != "OFF":
                if it < MAXITSDEBUG:
                    logTrack("it %s input to HXSegment dHX_H Ch %s Cc %s UA %s"%(it,Ch,Cc,self.UA))

            #now start heat exchanger calculations
            dHX_H = HXSegment(Ch,Cc,ThiHX,TciHX,self.UA)
            dHX_H.calculate(qxmax=qxmax)

        #TODO -> properties for condensing HX
            if dHX_H.Tho <= TCondH and dHX_H.Thi >= TCondH:
                condensation_in_H = True
                qx_H = Ch*(dHX_H.Thi - TCondH)

                if qxmax < QhLat + qx_H:
                    full_condensation_in_H = False
                    UA_H = self.UA
                else:
                    dHX_C = HXSegment(None,Cc,TCondH,TciHX,self.UA)
##                            logTrack("TEST TEST calling adjustUA with Q = %s"%QhLat[it])
                    UA_C = dHX_C.adjustUA(Q=QhLat)
                    if UA_C is None: UA_C = self.UA
##                            print "UA_C = %s, UA = %s"%(UA_C,self.UA)

##                            logTrack("TEST TEST Call1 TCondH = %s dHX_C.Tco = %s,self.UA = %s"%(TCondH,dHX_C.Tco,self.UA))
                    dHX_H = HXSegment(Ch,Cc,ThiHX,dHX_C.Tco,self.UA)
                    UA_H = dHX_H.adjustUA(Tho = TCondH)
                    if UA_H is None: UA_H = self.UA
                    
                    if UA_C >= self.UA or UA_C is None:
                        full_condensation_in_H = False
                    elif UA_H > self.UA - UA_C or UA_H is None:
                        full_condensation_in_H = False
                    else:
                        full_condensation_in_H = True

##                            print "TEST TEST UA_H = %s FullCondensation = %s"%(UA_H,full_condensation_in_H)

                if full_condensation_in_H == True:
                    EPS = 1.e-3
                    err = EPS+1
                    err0 = err
                    alpha = 0.8
                    nits = 0
                    UA_L = 0
                    
                    while err > EPS and nits < 100:
                        
                        UA_L_new = max(self.UA - UA_C - UA_H,0)
                        UA_L += alpha * (UA_L_new - UA_L)
                        
                        dHX_L = HXSegment(Ch0,Cc,TCondH,TciHX,UA_L)
                        dHX_L.calculate(qxmax = max(qxmax - dHX_C.qx - qx_H,0)) # partial condensation only is not 100% excluded by check above -> additional security defining qxmax

                        if DEBUG != "OFF" and it < MAXITSDEBUG:
                            logTrack("HX_L: Tci %s Tco %s Thi %s Tho %s qx %s"%(dHX_L.Tci,dHX_L.Tco,dHX_L.Thi,dHX_L.Tho, dHX_L.qx))

                        dHX_C = HXSegment(None,Cc,TCondH,dHX_L.Tco,self.UA)
                        UA_C = dHX_C.adjustUA(Q=min(QhLat,qxmax-qx_H)) # partial condensation only is not 100% excluded by check above -> additional security
                        if UA_C is None: UA_C = self.UA

                        if DEBUG != "OFF" and it < MAXITSDEBUG:
                            logTrack("HX_C: Tci %s Tco %s Thi %s Tho %s qx %s"%(dHX_C.Tci,dHX_C.Tco,dHX_C.Thi,dHX_C.Tho, dHX_C.qx))

                        dHX_H = HXSegment(Ch,Cc,ThiHX,dHX_C.Tco,self.UA)
##                                logTrack("TEST TEST Call2 TCondH = %s dHX_C.Tco = %s,self.UA = %s"%(TCondH,dHX_C.Tco,self.UA))
                        UA_H = dHX_H.adjustUA(Tho = TCondH)
                        if UA_H is None: UA_H = self.UA

                        if DEBUG != "OFF" and it < MAXITSDEBUG:
                            logTrack("HX_H: Tci %s Tco %s Thi %s Tho %s qx %s"%(dHX_H.Tci,dHX_H.Tco,dHX_H.Thi,dHX_H.Tho, dHX_H.qx))

                        UATotal = UA_C + UA_L + UA_H
                        err = abs(UATotal/self.UA - 1.)
                        nits += 1   #security feature: avoid infinite loops
                        
                        # iterative adjustment self adjusting alpha in case of divergence
                        difErr = err/err0
                        if (difErr > 1 or difErr < -0.2) and alpha > 0.125: # reduce alpha if diverging or oscillating
                            alpha *= 0.8
                        err0 = err

                        if DEBUG != "OFF" and it < MAXITSDEBUG:
                            logTrack("heatExchanger err = %s UAL/C/H %s %s %s"%(err,UA_L,UA_C,UA_H))

                    qx = dHX_H.qx + dHX_C.qx + dHX_L.qx
                    lmtd = qx/self.UA

                    TcoHX = dHX_H.Tco
                    ThoHX = dHX_L.Tho

                else:
                    EPS = 1.e-3
                    err = EPS+1
                    err0 = err
                    alpha = 0.8
                    nits = 0
                    UA_C = 0
                    
                    while err > EPS and nits < 100:
                        UA_C_new = max(self.UA - UA_H,0)
                        UA_C += alpha * (UA_C_new - UA_C)

                        dHX_C = HXSegment(None,Cc,TCondH,TciHX,UA_C)
                        dHX_C.calculate(qxmax = qxmax - qx_H)

                        dHX_H = HXSegment(Ch,Cc,ThiHX,dHX_C.Tco,self.UA)
                        UA_H = dHX_H.adjustUA(Tho = TCondH)
                        if UA_H is None: UA_H = self.UA

                        UATotal = UA_C + UA_H
                        err = abs(UATotal/self.UA - 1.)
                        nits += 1   #security feature: avoid infinite loops

                        # iterative adjustment self adjusting alpha in case of divergence
                        difErr = err/err0
                        if (difErr > 1 or difErr < -0.2) and alpha > 0.125: # reduce alpha if diverging or oscillating
                            alpha *= 0.8
                        err0 = err

                    if DEBUG != "OFF" and it < MAXITSDEBUG: logTrack("heatExchanger err = %s UAL/C/H %s %s %s"%(err,0,UA_C,UA_H))

                    qx = dHX_H.qx + dHX_C.qx
                    lmtd = qx/self.UA
                    TcoHX = dHX_H.Tco
                    ThoHX = dHX_C.Tho
                

            else:
                condensation_in_H = False
                qx = dHX_H.qx
                lmtd = dHX_H.lmtd
                TcoHX = dHX_H.Tco
                ThoHX = dHX_H.Tho

            if DEBUG != "OFF":
                if it < MAXITSDEBUG:
                    logTrack("it: %s HX qx %s TciHX %s TcoHX %s ThiHX %s ThoHX %s Ch %s Cc %s"%(it,qx,TciHX,TcoHX,ThiHX,ThoHX,Ch,Cc))


            if it in SNAPSHOTS:
                print("it: %s HX qx %s TciHX %s TcoHX %s ThiHX %s ThoHX %s Ch %s Cc %s"%(it,qx,TciHX,TcoHX,ThiHX,ThoHX,Ch,Cc))
                print "HXKernel finished ............"

            #store results of HX calculations (for speed-up option)
            self.Tci_lastNonZero = TciHX
            self.Thi_lastNonZero = ThiHX
            self.Tco_lastNonZero = TcoHX
            self.Tho_lastNonZero = ThoHX
            self.qx_lastNonZero = qx
            self.lmtd_lastNonZero = lmtd
            self.Ch_lastNonZero = Ch
            self.Cc_lastNonZero = Cc

        return ThoHX, TcoHX, qx, lmtd
    
    def loadFromDB(self, dict=None, loadConnections=True):
        """Looks for heat exchanger with given id in the SQL database and
        loads data.

        :params dict: data row (dictionary) from SQL table qheatexchanger
                otherwise data are looked-up in SQL by id

        """

        # First try to obtain data dictionary from dict or from SQL via self.id
        if dict is not None:
            hx = dict     
        elif self.id is not None:
            hx = Status.prj.getHX(self.id)

        else:
            self.HXNo = None
            return False

        if hx is not None:
            self.UA = hx.UA
            self.area = hx.Area
            self.QHX = hx.QHX
            self.dotQNom = hx.QdotHXNom
            self.dotQAvg = hx.QdotHX
            self.lmtdNom = hx.HXLMTDNom
            self.lmtdAvg = hx.HXLMTD
            if hx.StorageSize is not None:
                self.MCStorage = hx.StorageSize*4.19/3.6
            else:
                self.MCStorage = 0.0
                
            self.ThiAvg = hx.HXTSourceInlet
            self.ThoAvg = hx.HXTSourceOutlet
            self.TciAvg = hx.HXTSinkInlet
            self.TcoAvg = hx.HXTSinkOutlet
            
            self.sourceObjectName = hx.HXSource
            self.sinkObjectName = hx.HXSink
            self.HXNo = hx.HXNo


            self.turnKeyPrice = hx.TurnKeyPrice
            self.OMFix = hx.OMFix
            self.OMVar = hx.OMVar


            if loadConnections:
                self.getConnectionsFromDB()

            if hx.id is not None:
                hxCtrl = dbio.get("qhxcontrol", HXId=hx.id)
            else:
                hxCtrl = None

            if hxCtrl is not None:
                self.Kc0 = noneCheck(hxCtrl.Kc0, 0.0)
                self.Kc1 = noneCheck(hxCtrl.Kc1, 0.0)
                self.Kh0 = noneCheck(hxCtrl.Kh0, 0.0)
                self.Kh1 = noneCheck(hxCtrl.Kh1, 0.0)
            else:
                self.Kc0 = 0.0
                self.Kc1 = 0.0
                self.Kh0 = 0.0
                self.Kh1 = 0.0

            return True
        
        else:
            self.HXNo = None
            return False

    def estimateCosts(self):
        """ effectuates a cost estimate based on UA
        """
        if self.UA is None:
            return False
        else:
            self.turnKeyPrice = self.UA*1000.0
            self.OMFix = self.UA*10.0
            self.OMVar = 0.0
            return True

    def writeToDB(self):
        """ updates the heat exchanger data - if HX exists - or creates a new heat exchanger in the project

        """
        from modules.heatRecovery.hxConnection import getPrimarySourceOfConn, getPrimarySinkOfConn
        #TODO: these functions could be ATTRIBUTES of the connection class
        
        if not self.writeResultsToDBAllowed:
            logTrack("HeatExchanger.writeToDB: attempt to write blocked")
            return

        if self.id is not None:
            hx = Status.prj.getHX(self.id)
        else:
            hx = None

        if hx is None:

            if self.UA == 0:
                return None
            
            data = {"ProjectID":Status.PId,
                   "AlternativeProposalNo":Status.ANo,
                   "HXName":check(self.name)
                   }
            
            if DEBUG != "OFF":
                print "HeatExchanger.writeToDB: Inserting HX %s"%(data)
                
            self.id = dbio.insertRow("qheatexchanger", **data)
            dbio.commit()
                
            hx = Status.prj.getHX(self.id)
            
        if self.id is not None:

            self.sourceObjectName = None
            for conn in self.sourcestreams:
                conn.HXId = self.id
                # conserve the important parameters before looking up primarySource
                connId = conn.connId
                pinchStreamId = conn.pinchStreamId

                if connId is None and len(self.deletedSource) > 0: #re-use Id of previously deleted source
                    deleted = self.deletedSource.pop()
                    connId = deleted.connId
                    pinchStreamId = deleted.pinchStreamId

##                print "HeatExchanger.writeToDB: HX %s stream %s source fraction %s"%(self.name, conn.pinchStreamName, conn.fraction)
                
                pconn = getPrimarySourceOfConn(conn) # this conversion to primary source/sink is necessary for backward compatibility.
                #TODO: test treatment of fractions in serial connections !!!

                if pconn is None:
                    logError("HeatExchanger.writeToDB: HX %s has a corrupt source connection %s"%(self.name, conn.name))
                else:
                    pconn.connId = connId
                    pconn.pinchStreamId = pinchStreamId
                    pconn.HXId = self.id

##                    print "HeatExchanger.writeToDB: primary stream %s source fraction %s"%(conn.pinchStreamName, conn.fraction)
                    # determine outletHX based on conn parameters
                    upStreamHX = conn.upStreamHX()
                    pconn.outletHX = None if upStreamHX is None else upStreamHX.id 
##                    if conn.objectType=="HX":
##                        pconn.outletHX = Status.int.hx[conn.objectNo].id
##                    else:
##                        pconn.outletHX = None


                    pconn.writeToDB(TNom = (self.ThiNom, self.ThoNom), commit=False)

                    if self.sourceObjectName is None: #check only the first connection
                        self.sourceObjectName = pconn.objectName

            for deleted in self.deletedSource:
                deleted.deleteFromDB()
            self.deletedSource = []

            self.sinkObjectName = None
            for conn in self.sinkstreams:
                conn.HXId = self.id
                connId = conn.connId
                pinchStreamId = conn.pinchStreamId
                
                if connId is None and len(self.deletedSink) > 0: #re-use Id of previously deleted sink
                    deleted = self.deletedSink.pop()
                    connId = deleted.connId
                    pinchStreamId = deleted.pinchStreamId

##                print "HeatExchanger.writeToDB: HX %s stream %s sink fraction %s"%(self.name, conn.pinchStreamName, conn.fraction)
                pconn = getPrimarySinkOfConn(conn)
                if pconn is None:
                    logError("HeatExchanger.writeToDB: HX %s has a corrupt sink connection %s"%(self.name, conn.name))
                else:
                    pconn.HXId = self.id
                    pconn.connId = connId
                    pconn.pinchStreamId = pinchStreamId
                    # determine outletHX based on conn parameters
                    upStreamHX = conn.upStreamHX()
                    pconn.outletHX = None if upStreamHX is None else upStreamHX.id 
##                    if conn.objectType=="HX":
##                        pconn.outletHX = Status.int.hx[conn.objectNo].id
##                    else:
##                        pconn.outletHX = None

##                    print "HeatExchanger.writeToDB: primary stream %s sink fraction %s"%(conn.pinchStreamName, conn.fraction)
                    pconn.writeToDB(TNom = (self.TciNom, self.TcoNom), commit=False)

                    if self.sinkObjectName is None: #check only the first connection
##                        self.sinkObjectName = pconn.getObjectName()
                        self.sinkObjectName = pconn.objectName

            for deleted in self.deletedSink:
                deleted.deleteFromDB()
            self.deletedSink = []

##                print "HeatExchanger.writeToDB: sink object name %s"%self.sinkObjectName

                    

            if self.MCStorage is not None:
                storageSize = self.MCStorage/(4.19/3.6) # conversion from equivalent m3 of standard water to kWh/K
            else:
                storageSize = 0.0
                
            data = {"HXName": check(self.name),
                    "QHX":check(self.QHX),
                    "UA": check(self.UA),
                    "Area": check(self.area),
                    "StorageSize":check(storageSize),
                    "QdotHXNom":check(self.dotQNom),
                    "QdotHX": check(self.dotQAvg),
                    "HXLMTDNom":check(self.lmtdNom),
                    "HXLMTD": check(self.lmtdAvg),
                    "HXTSourceInlet":check(self.ThiAvg),
                    "HXTSourceOutlet":check(self.ThoAvg),
                    "HXTSinkInlet":check(self.TciAvg),
                    "HXTSinkOutlet":check(self.TcoAvg),
                    "HXSource":check(self.sourceObjectName),
                    "HXSink":check(self.sinkObjectName),
                    "HXNo":check(self.HXNo),
                    "TurnKeyPrice":check(self.turnKeyPrice),
                    "OMFix":check(self.OMFix),
                    "OMVar":check(self.OMVar),
                    }
            hx.set(**data)

            hxCtrl = dbio.get("qhxcontrol", HXId = self.id)
            if any([getattr(self, par) > EPSILON for par in ["Kc0", "Kc1", "Kh0", "Kh1"]]):
                if hxCtrl is None:
                    hxCtrl = dbio.insertRow("qhxcontrol", {"HXId": self.id})

                hxCtrl.set(**{"Kc0":check(self.Kc0),
                              "Kc1":check(self.Kc1),
                              "Kh0":check(self.Kh0),
                              "Kh1":check(self.Kh1)})
            elif hxCtrl is not None:
                dbio.delete("qxhcontrol", HXId = self.id)
                
            dbio.commit()

##            print "Writing HX data to DB %s"%data
            
        else:
            logWarning("HeatExchanger.writeToDB: could not update data of HX %s with id %s"%(self.name,self.id))

    def deleteFromDB(self):
        if self.id is None:
            logWarning(_("Cannot delete heat exchanger with ID None"))
            return

        #reconnect other heat exchangers that might have used the output of this one

        sourceSplittersChanged = []
        for splitter in Status.int.sourceSplitters:
            source = splitter.mainStream
            if source.objectType == "HX" and source.objectNo == self.HXNo - 1:
                if len(self.sourcestreams) > 0:
                    splitter.mainStream.reconnect(self.sourcestreams[0])
                    sourceSplittersChanged.append(splitter.no)

        sinkSplittersChanged = []
        for splitter in Status.int.sinkSplitters:
            sink = splitter.mainStream
            if sink.objectType == "HX" and sink.objectNo == self.HXNo - 1:
                if len(self.sinkstreams) > 0:
                    splitter.mainStream.reconnect(self.sinkstreams[0])
                    sinkSplittersChanged.append(splitter.no)
        #TODO: this gets more complicated once mixing is allowed
                    
        for hx in Status.int.hx:
            changed = False
            for source in hx.sourcestreams:
                if source.objectType == "HX" and source.objectNo == self.HXNo - 1:
                    hx.sourcestreams[0].reconnect(self.sourcestreams[0]) 
                    changed = True
                    break
                elif source.objectType == "SPLIT" and source.objectNo in sourceSplittersChanged:
                    changed = True
                    #TODO: stops at the first source. this does not work for the future with multi-connections
            for sink in hx.sinkstreams:
                if sink.objectType == "HX" and sink.objectNo == self.HXNo - 1:
                    hx.sinkstreams[0].reconnect(self.sinkstreams[0])
                    changed = True
                    break
                    #TODO: stops at the first source. this does not work for the future with multi-connections
                elif sink.objectType == "SPLIT" and sink.objectNo in sinkSplittersChanged:
                    changed = True

            if changed:
##                print "HeatExchanger.deleteFromDB: reconnecting source and/or sink from hx %s"%hx.name
##                if len(self.sourcestreams) > 0: print "HeatExchanger.deleteFromDB: source of deleted HX %s OutletHXId %s"%(self.sourcestreams, self.sourcestreams[0].outletHX)
##                if len(self.sinkstreams) > 0: print "HeatExchanger.deleteFromDB: sink of deleted HX %s OutletHXId %s"%(self.sinkstreams, self.sinkstreams[0].outletHX)
##
##
##                print "HeatExchanger.deleteFromDB: new source %s"%(hx.sourcestreams)
##                print "HeatExchanger.deleteFromDB: new sink %s"%(hx.sinkstreams)
                hx.writeToDB()
                    
##        print "HeatExchanger.deleteFromDB: now deleting HX with id %s from database"
        psIds = [hx_ps.pinchstream_id for hx_ps in dbio.select("heatexchanger_pinchstream", qheatexchanger_QHeatExchanger_Id=self.id)]
        dbio.deleteByList("pinchstream", "id", psIds)
        dbio.delete("heatexchanger_pinchstream", qheatexchanger_QHeatExchanger_Id=self.id)
        dbio.delete("qheatexchanger", id=self.id)
        dbio.commit()

        Status.prj.reorderHX()
        
        Status.int.systemModel.changed()        
        Status.int.systemModel.build()


    def saveSimulationResultsToDB(self):
        """Saves simulation results for heat exchanger in DB under given id"""
        if not self.writeResultsToDBAllowed:
            logTrack("HeatExchanger.saveSimulationResultsToDB: attempt to write blocked")
            return

        if self.hop > 0:
            self.dotQAvg = self.Qx/self.hop
            self.lmtdAvg = self.dotQAvg/self.UA
        else:
            self.lmtdAvg = 0
            self.dotQAvg = 0

        
        if Status.SimTool != "Plus":
            self.QHX = self.qhx.Q

        if DEBUG != "OFF": logTrack("HeatExchanger.saveSimulationResultsToDB: QdotHX %s HXLMTD %s QHX %s"%(self.dotQAvg,self.lmtdAvg,self.QHX))
        
        hx = Status.prj.getHX(self.id)
        if hx is not None:
            data = {
                "QdotHX":check(self.dotQAvg), 
                "HXLMTD":check(self.lmtdAvg), 
                "QHX":check(self.QHX),
                "HXTSourceInlet":check(self.ThiAvg),
                "HXTSourceOutlet":check(self.ThoAvg),
                "HXTSinkInlet":check(self.TciAvg),
                "HXTSinkOutlet":check(self.TcoAvg)
                }

            if Status.StatusEnergy > 0:
                logTrack("HeatExchanger.saveSimulationResultsToDB: results not stored, as simulation has already been concluded")
                logTrack("HeatExchanger.saveSimulationResultsToDB: QHX from DB %s; QHX from current simulation: %s"%(hx.QHX,data["QHX"]))
            else:                
                hx.set(**data)
                dbio.commit()
        else:
            logWarning("HeatExchanger.saveSimulationResultsToDB: could not save HX %s with id %s"%(self.name,self.id))




    def getSourceAndSinkHeatFlows(self):
        """Reads in the HeatFlow_t - objects for sources and sinks
        including pointer to space where to deposit calculation results
        
        """
        changeInSourceOrSink = False
        if DEBUG != "OFF": logTrack("HeatExchanger.getSourceAndSinkFlows: - number of sourcestreams %s"%len(self.sourcestreams))
        if len(self.sourcestreams) > 0:
            source = self.sourcestreams[0].compactConnToUpstreamComponent()

            if DEBUG != "OFF": logTrack("HeatExchanger.getSourceAndSinkFlows - source oType %s oNo %s cNo %s scale %s"%(source.objectType,source.objectNo,source.componentNo,source.fraction))
            if source.objectType == "ProcessC":                    
                if source.componentNo == None or source.componentNo==-1:
                    self.qsrc = Status.int.UPHs_Tt[source.objectNo].scale(source.fraction)
#                    self.qsrcUsed = Status.int.QCXProc_Tt[source.objectNo]
                elif source.componentNo == 0:
                    self.qsrc = Status.int.UPHm_Tt[source.objectNo].scale(source.fraction)
#                    self.qsrcUsed = Status.int.QCXProc_Tt[source.objectNo]
                else:                    
                    self.qsrc = Status.int.UPHc_Tt[source.objectNo].getPart(source.componentNo-1).scale(source.fraction)
#                    self.qsrcUsed = Status.int.QCXProc_Tt[source.objectNo]
            elif source.objectType == "ProcessH":
                if source.componentNo == None:
                    logWarning("No component number specified for UPHw component in heat exchanger %s. First component assumed"%self.name)
                    self.qsrc = Status.int.UPHw_Tt_hf[source.objectNo].getPart(0).scale(source.fraction)
                else:                       
                    self.qsrc = Status.int.UPHw_Tt_hf[source.objectNo].getPart(source.componentNo-1).scale(source.fraction)
#                self.qsrcUsed = Status.int.UPHwUsed_Tt_hf[source.objectNo].getPart(source.componentNo-1)
            elif source.objectType == "PipeH":
                self.qsrc = Status.int.QWHPipe_Tt[source.objectNo].scale(source.fraction)
                changeInSourceOrSink = Status.int.QWHPipeChanged[source.objectNo] or changeInSourceOrSink 
            elif source.objectType == "PipeC":
                self.qsrc = Status.int.USHCPipe_Tt_hf[source.objectNo].scale(source.fraction)
                changeInSourceOrSink = Status.int.USHCPipeChanged[source.objectNo] or changeInSourceOrSink 
            elif source.objectType == "Equipment":
                changeInSourceOrSink = Status.int.QWHjChanged[source.objectNo] or changeInSourceOrSink 
                self.qsrc = Status.int.QWHEq_Tt[source.objectNo].scale(source.fraction)
#                self.qsrcUsed = Status.int.QWHEqUsed_Tt[source.objectNo]
            elif source.objectType == "WHEE":
                self.qsrc = Status.int.QWHEE_Tt_hf[source.objectNo].scale(source.fraction)
#                self.qsrcUsed = Status.int.QWHEEUsed_Tt[source.objectNo]
            elif source.objectType == "HX":
##                h = Status.int.hxIDdict[source.objectNo]
                h = source.objectNo
                self.qsrc = Status.int.QWHRes_Tt[h].scale(source.fraction)
                changeInSourceOrSink = Status.int.QHXChanged[h] or changeInSourceOrSink
            else:
                logError("Severe ERROR in source connections of heat exchanger %s"%self.name)
                self.qsrc = HeatFlow_t()

            if DEBUG != "OFF":
                logTrack("HeatExchanger.getSourceAndSinkFlows qsrc: %s %s kWh"%(self.qsrc.name, self.qsrc.Q))

        if DEBUG != "OFF": logTrack("HeatExchanger.getSourceAndSinkFlows: - number of sinkstreams %s"%len(self.sinkstreams))
                
        if len(self.sinkstreams) > 0:
            sink = self.sinkstreams[0].compactConnToUpstreamComponent()
            if DEBUG != "OFF": logTrack("HeatExchanger.getSourceAndSinkFlows - sink oType %s oNo %s cNo %s scale %s"%(sink.objectType,sink.objectNo,sink.componentNo,sink.fraction))
            if sink.objectType == "ProcessH":
                if sink.componentNo == None or sink.componentNo==-1:
                    self.qsnk = Status.int.UPHs_Tt[sink.objectNo].scale(sink.fraction)
#                    self.qsnkSup = Status.int.QHXProc_Tt[sink.objectNo]
                elif sink.componentNo == 0:
                    self.qsnk = Status.int.UPHm_Tt[sink.objectNo].scale(sink.fraction)
#                    self.qsnkSup = Status.int.QHXProc_Tt[sink.objectNo]
                else:                    
                    self.qsnk = Status.int.UPHc_Tt[sink.objectNo].getPart(sink.componentNo-1).scale(sink.fraction)
#                    self.qsnkSup = Status.int.QHXProc_Tt[sink.objectNo]
            elif sink.objectType == "ProcessC":
                if sink.componentNo == None:
                    logWarning("No component number specified for UPHw component in heat exchanger %s. First component assumed"%self.name)
                    self.qsnk = Status.int.UPHw_Tt_hf[sink.objectNo].getPart(0).scale(sink.fraction)
                else:
                    self.qsnk = Status.int.UPHw_Tt_hf[sink.objectNo].getPart(sink.componentNo-1).scale(sink.fraction)
#                self.qsnkSup = Status.int.UPHwUsed_Tt_hf[sink.objectNo]
            elif sink.objectType == "PipeH":
                self.qsnk = Status.int.USHCPipe_Tt_hf[sink.objectNo].scale(sink.fraction)
                changeInSourceOrSink = Status.int.USHCPipeChanged[sink.objectNo] or changeInSourceOrSink 
            elif sink.objectType == "PipeC":
                self.qsnk = Status.int.QWHPipe_Tt[sink.objectNo].scale(sink.fraction)
                changeInSourceOrSink = Status.int.QWHPipeChanged[sink.objectNo] or changeInSourceOrSink 
#                self.qsnkSup = Status.int.QHXPipe_Tt[sink.objectNo]
            elif sink.objectType == "Equipment":
                self.qsnk = Status.int.QHXEqReq_Tt[sink.objectNo].scale(sink.fraction)
                changeInSourceOrSink = Status.int.QHXEqReqChanged[sink.objectNo] or changeInSourceOrSink 
#                self.qsnkSup = Status.int.QHXPipe_Tt[sink.objectNo]
            elif sink.objectType == "HX":
##                h = Status.int.hxIDdict[sink.objectNo]
                h = sink.objectNo
                self.qsnk = Status.int.QHXRes_Tt[h].scale(sink.fraction)
                changeInSourceOrSink = Status.int.QHXChanged[h] or changeInSourceOrSink
            else:
                logError("Severe ERROR in sink connections of heat exchanger %s"%self.name)
                self.qsnk = HeatFlow_t()

            if DEBUG != "OFF":
                logTrack("HeatExchanger.getSourceAndSinkFlows qsnk: %s %s kWh"%(self.qsnk.name,self.qsnk.Q))

        return changeInSourceOrSink
                
    def storeResults(self):
        """Writes the results in self.qsrcUsed and self.qsnkSup to the corresponding arrays"""

        from modules.heatRecovery.hxConnection import getPrimarySource, getPrimarySink

        if len(self.sourcestreams) > 0:
            source = self.sourcestreams[0].compactConnToUpstreamComponent()
            if source.objectType == "HX":
##                logTrack("TEST TEST: getting primary sink of HX no %s"%sink.objectNo)
                primarySource = getPrimarySource(source.objectNo)
                if primarySource is not None:
##                    logTrack("TEST TEST: primary sink is oT %s oNo %s"%(primarySink.objectType,primarySink.objectNo))
                    objectType = primarySource.objectType
                    objectNo = primarySource.objectNo
                else:
                    objectType = None
            else:
                objectType = source.objectType
                objectNo = source.objectNo

            if DEBUG != "OFF": logTrack("storeResults: HX %s objectType %s componentNo %s "%(self.name,source.objectType,source.objectNo))
            if objectType == "ProcessC":
                Status.int.QHXProcChanged[objectNo] = True
                if source.componentNo == None:
                    Status.int.QHXProc_Tt[objectNo].add(self.qsrcUsed)
                elif source.componentNo == 0:
                    Status.int.QHXProc_Tt[objectNo].add(self.qsrcUsed)
                else:                    
                    Status.int.QHXProc_Tt[objectNo].add(self.qsrcUsed)
            elif objectType == "ProcessH":
                Status.int.UPHwUsed_Tt[objectNo].add(self.qsrcUsed)
                Status.int.UPHwUsedChanged[objectNo] = True
            elif objectType == "PipeH":
                Status.int.QWHPipeUsed_Tt[objectNo].add(self.qsrcUsed)
                Status.int.QWHPipeUsedChanged[objectNo] = True
            elif objectType == "PipeC":
                Status.int.QHXPipe_Tt[objectNo].add(self.qsrcUsed)
                Status.int.QHXPipeChanged[objectNo] = True
            elif objectType == "Equipment":
                Status.int.QWHEqUsed_Tt[objectNo].add(self.qsrcUsed)
                Status.int.QWHEqUsedChanged[objectNo] = True
            elif objectType == "WHEE":
                Status.int.QWHEEUsed_Tt[objectNo].add(self.qsrcUsed)
                Status.int.QWHEEUsedChanged[objectNo] = True
            else:
                logError("Severe ERROR in source connections of heat exchanger %s"%self.name)
        
        if len(self.sinkstreams) > 0:
            sink = self.sinkstreams[0].compactConnToUpstreamComponent()
            if sink.objectType == "HX":
##                logTrack("TEST TEST: getting primary sink of HX no %s"%sink.objectNo)
                primarySink = getPrimarySink(sink.objectNo)
                if primarySink is not None:
##                    logTrack("TEST TEST: primary sink is oT %s oNo %s"%(primarySink.objectType,primarySink.objectNo))
                    objectType = primarySink.objectType
                    objectNo = primarySink.objectNo
                else:
                    objectType = None
            else:
                objectType = sink.objectType
                objectNo = sink.objectNo
                
            if objectType == "ProcessH":
##                logTrack("TEST TEST adding QHX %s to sink comp.No %s objNo %s"%(self.qhx.Q,sink.componentNo, objectNo))
                if sink.componentNo == None:
                    Status.int.QHXProc_Tt[objectNo].add(self.qhx)
                elif sink.componentNo == 0:
                    Status.int.QHXProc_Tt[objectNo].add(self.qhx)
                else:                    
                    Status.int.QHXProc_Tt[objectNo].add(self.qhx)
                Status.int.QHXProcChanged[objectNo] = True
##                logTrack("TEST TEST: QHXProc = %s"%Status.int.QHXProc_Tt[objectNo].Q)
            elif objectType == "ProcessC":
                Status.int.UPHwUsed_Tt[objectNo].add(self.qhx)
                Status.int.UPHwUsedChanged[objectNo] = True
            elif objectType == "PipeH":
                Status.int.QHXPipe_Tt[objectNo].add(self.qhx)
                Status.int.QHXPipeChanged[objectNo] = True
            elif objectType == "PipeC":
                Status.int.QWHPipeUsed_Tt[objectNo].add(self.qhx)
                Status.int.QWHPipeUsedChanged[objectNo] = True
            elif objectType == "Equipment":
                Status.int.QHXEq_Tt[objectNo].add(self.qhx)
                Status.int.QHXEqChanged[objectNo] = True
            else:
                logError("Severe ERROR in sink connections of heat exchanger %s"%self.name)

        # now store remaining source / sink (needed if HX is used as input for another serially connected one
        h = self.HXNo - 1
        Status.int.QHXRes_Tt[h] = self.qsnkRes
        Status.int.QWHRes_Tt[h] = self.qsrcRes
        Status.int.QHX_Tt[h] = self.qhx

        Status.int.QHXChanged[h] = (not self.converged or self.firstRun)

        if self.writeResultsToDBAllowed: self.saveSimulationResultsToDB()
        #TODO: should later on be shifted to another place
        #no need to do this in each iteration of the simulation cycle
                
    def getConnectionsFromDB(self):
        """ gets HX connections from database entries in table qheatexchanger (source/sink) and
        pinchstream (no. of stream in the case of processes
    
        """
        from modules.heatRecovery.hxConnection import HXConnection
        
        if DEBUG != "OFF": logTrack("HeatExchanger.getConnectionsFromDB HX = %s ID %s"%(self.name, self.id))

        self.sinkstreams = []
        self.sourcestreams = []


        # now load connections (-> determine componentNo, currently from pinchStreamName ...)
        lastId = -1 #streamNo can be None or from 0 to +INF
        lastHeating = None
        
        connections = Status.prj.getHXConnections(self.id)
        for connection in connections:
            conn = HXConnection(connData=connection)
            conn = conn.check()
            if conn is None: continue

            # this is a hack for avoiding importing three times for condensing streams (3 entries in pinchstream ???)
            new =  conn.componentId != lastId or conn.heating != lastHeating or conn.objectType not in ["ProcessH", "ProcessC"]
            lastId = conn.componentId
            lastSourceOrSink = conn.heating

            if new:
                if conn.heating:
                    self.sinkstreams.append(conn)
                else:
                    self.sourcestreams.append(conn)
                    
        #TODO: currently mixing of streams is not supported -> only one sink and one source stream allowed !!!
        if len(self.sinkstreams) != 1:
            logWarning("Erroneous number (%s) of sink streams specified for heat exchanger %s."%(len(self.sinkstreams),self.name))
        if len(self.sourcestreams) != 1:
            logWarning("Erroneous number (%s) of source streams specified for heat exchanger %s."%(len(self.sourcestreams),self.name))

class HXSegment():
    """ describes a segment of a heat exchanger for a given time step

    a heat exchanger can be composed of up to three segments in the
    case of condenstation (currently only allowed on hot side):
    above TCond / Condensation / below TCond

    in the case of sensible heat exchange only one heat exchanger
    has only one segment.

    """

    def __init__(self,Ch,Cc,Thi,Tci,dUA):
        """ 
        :param Ch: m*cp hot side (source); None -> condensing
        :param Cc: m*cp cold side (sink); None -> condensing
        :param Thi: inlet temperature hot side for a given time step it
        :param Tci: inlet temperature cold side for a given time step it
        :param dUA: UA-Value of heat exchanger segment

        """
        self.Ch = Ch
        self.Cc = Cc
        self.Thi = Thi
        self.Tci = Tci
        self.y = Thi - Tci
        self.dUA = dUA
#        print "TEST TEST HXSegment.__init__: Ch %s Cc %s y %s dUA %s"%(self.Ch,self.Cc,self.y,self.dUA)

    def calculate(self,qxmax=INFINITE):
        """ carries out calculations of a heat exchanger segment for one timestep

        :returns self.Tco: outlet temperature cold side
        :returns self.Tho: outlet temperature hot side
        :returns self.qx:  energy exchanged during time step

        """
        if self.Ch == 0 or self.Cc == 0 or self.y <= 0:#special case - no flow - no exchange
            self.qx = 0
            self.Tco = max(self.Thi, self.Tci)
            self.Tho = min(self.Tci,self.Thi)
            self.lmtd = 0

        elif self.dUA == 0: #special case: zero heat exchanger area
            self.qx = 0
            self.Tco = self.Tci
            self.Tho = self.Thi
            self.lmtd = self.y

        elif self.Ch is None:    #none indicates condensation
            if self.Cc is None: # condensation on both sides
                self.lmtd = self.y
                self.qx = self.dUA*self.lmtd
                self.Tho = self.Thi
                self.Tco = self.Tci
            else:                       #condensation on hot side only
                gamma = math.exp(-self.dUA/self.Cc)
                self.qx = self.Cc*self.y*(1 - gamma)
                self.Tho = self.Thi
                if self.qx > qxmax:
                    self.Tco = self.Tci + qxmax/self.Cc
                    self.qx = qxmax
                else:
                    self.Tco = self.Thi - gamma*self.y
                self.lmtd = self.qx/self.dUA
                
        elif self.Cc is None:   #condensation on cold side only; self.Ch is None is excluded already from previoous if !
            gamma = math.exp(-self.dUA/self.Ch)
            self.qx = self.Ch*self.y*(1 - gamma)
            self.Tco = self.Tci
            self.Tho = self.Tci + gamma*self.y
            if self.qx > qxmax:
                self.Tho = self.Thi - qxmax/self.Ch
                self.qx = qxmax
            self.lmtd = self.qx/self.dUA

        elif abs(self.Ch - self.Cc) < EPSILON: #special case matched flow
            
            ntu = self.dUA/self.Ch
            self.lmtd = self.y / (1 + ntu)
            #TODO - continue here - limitation Tcomin/max

            self.qx = self.dUA * self.lmtd
            if self.qx > qxmax:
                self.qx = qxmax
                self.Tco = self.Tci + qxmax/self.Cc
                self.Tho = self.Thi - qxmax/self.Ch
                self.lmtd = self.qx/self.dUA
            else:   
                self.Tco = self.Thi - self.lmtd
                self.Tho = self.Tci + self.lmtd
#            print "TEST TEST HXElement.calculate (Ch=Cc): ntu %s lmtd = %s qx = %s"%(ntu,self.lmtd,self.qx)
            
        else:
            dC = self.Ch - self.Cc
            dCe = 1/(1/self.Cc - 1/self.Ch)
            logGamma = self.dUA/dCe
#            print "TEST TEST HXElement.calculate logGamma %s dC = %s dCe = %s"%(logGamma,dC,dCe)
            if logGamma > 100.: #special cases for avoiding overflow of exp-function
                self.lmtd = self.y*dC/(self.Ch*logGamma)
                self.qx = self.dUA * self.lmtd
                self.Tco = self.Tci + self.qx/self.Cc
                self.Tho = self.Thi - self.qx/self.Ch
            elif logGamma < -100.:
                self.lmtd = self.y*dC/(self.Cc*logGamma)                
                self.qx = self.dUA * self.lmtd
                self.Tco = self.Tci + self.qx/self.Cc
                self.Tho = self.Thi - self.qx/self.Ch
            else:
                gamma = math.exp(logGamma)
                xi0 = self.y*dC/(gamma*self.Ch - self.Cc)
    #            print "TEST TEST gamma = %s dC = %s dCe = %s xi0 %s dUA %s"%(gamma,dC,dCe,xi0,self.dUA)
                self.lmtd = xi0 * (gamma-1)/logGamma
                self.qx = self.dUA * self.lmtd
                self.Tco = self.Thi - xi0
                self.Tho = self.Tci + xi0*gamma

            if self.qx > qxmax:
                self.qx = qxmax
                self.lmtd = self.qx/self.dUA
                self.Tco = self.Tci + self.qx/self.Cc
                self.Tho = self.Thi - self.qx/self.Ch
                

    def adjustUA(self,Q=None,Tho=None,Tco=None):
        """ Takes a given HX element with Tci and Thi fixed, and adjusts UA (if possible)
        so that a given criterion (depending on input parameter specified) is fulfilled

        :params Q: objective for amount of energy to be exchanged in current time step (Caution: currently only works for fluids
            condensing on hot side).
        :params Tho: objective for outlet temperature on hot side (Caution: currently only works for non-condensing fluids)

        :returns UAMin / self.UA: minimum heat transfer coefficient required (self.UA is automatically
            set to this value after execution of adjustUA)
 
        """
        if Q is not None:
##            print "TEST TEST adjustUA Q is not None"
            self.qx = Q
            if self.Ch is None and self.Cc > 0:    #none indicates condensation
                gamma = 1. - self.qx/(self.Cc*self.y)
                if gamma <= 0:
                    self.calculate()
                    return None
                else:
                    UAMin = -math.log(gamma)*self.Cc
                    self.UA = UAMin
                    self.Tho = self.Thi
                    self.Tco = self.Tci + Q/self.Cc

                    if self.UA > 0: self.lmtd = self.qx/self.UA
                    else: self.lmtd = 0
                    
                    return UAMin
                    
            else:
                logTrack("HeatExchanger.HXSegment.calculateUAMin: function adjustUA(Q) only prepared for fluids condensing on hot side")
                self.qx = 0
                self.lmtd = 0
                self.Tco = self.Tci
                self.Tho = self.Thi
                return None
            
        elif Tho is not None and Tho > self.Tci:
##            print "TEST TEST adjustUA Tho > self.Tci"
            
            if self.Ch is None or self.Cc is None:
                logTrack("HeatExchanger.HXSegment.calculateUAMin: function adjustUA(Tho) only prepared for non-condensing fluids")
                self.qx = 0
                self.lmtd = 0
                self.Tco = self.Tci
                self.Tho = self.Thi
                return None
            
            self.qx = self.Ch*(self.Thi - Tho)

            if self.Cc == self.Ch:
                self.lmtd = Tho - self.Tci
                if self.lmtd > 0:
                    UAMin = self.qx/self.lmtd
                    self.UA = UAMin
                    self.Tho = Tho
                    self.Tco = self.Tci + self.qx/self.Cc
                    return UAMin
                else:
                    logTrack("HeatExchanger.HXSegment.calculateUAMin: desired hot side outlet temperature lower than cold side inlet")
                    self.qx = 0
                    self.Tco = self.Tci
                    self.Tho = self.Thi

                    return None
            else:
                dC = self.Ch - self.Cc
                dCe = 1/(1/self.Cc - 1/self.Ch)
                xi0 = (Tho - self.Tci)
                gamma = 1 - self.qx/(xi0*dCe)
                if gamma > 0:
                    UAMin = -math.log(gamma)*dCe
                    self.UA = UAMin
                    self.lmtd = xi0 * (gamma-1)/math.log(gamma) if gamma <> 1.0 else xi0
                    self.Tco = self.Tci + self.qx/self.Cc
                    self.Tho = Tho
                    self.qx = self.UA*self.lmtd
                    return UAMin
                else:
                    self.calculate()
                    return None
        else:
##            print "TEST TEST adjustUA crash case"
            self.qx = 0
            self.lmtd = 0
            self.Tco = self.Tci
            self.Tho = self.Thi
            return None

def testHXKernel(UA=10.0,
           MCStorage=0.0,
           Ch = 15.4496194941,
           Ch0 = 15.4496194941,
           Cc = 1000000000.0,
           ThiHX = 26.7293072378,
           TciHX = 25.0,
           qxmax = 4.54939392135,
           TCondH = 1e+99,
           QhLat = 0.0):
    """ function for unit testing of the HX Kernel for different boundary conditions """

    hxTable = {"QHeatExchanger_ID": -1,
               "HXName": "test",
               "HXNo": 1,
               "UA": UA,
               "MCStorage": MCStorage,
               "FluidIDSource": 0,
               "FluidIDSink": 0,
               "Kc0": 0,
               "Kc1": 0,
               "Kh0": 0,
               "Kh1": 0,
               "Ts_max": 9999.9}

    hx = HeatExchanger(UA=UA, MCStorage=MCStorage)
    hx.Thi_lastNonZero = None
    hx.Tci_lastNonZero = None
    hx.Cc_lastNonZero = 0
    hx.Ch_lastNonZero = 0
    
    ThoHX, TcoHX, qx, lmtd = hx.calculateHX(0, Ch, Cc, ThiHX, TciHX, qxmax, TCondH, QhLat, Ch0)

    return {"qx": qx, "lmtd": lmtd, "Tho": ThoHX, "Tco": TcoHX}
