﻿import numpy as np
from constants import EPSILON

class Storage():
    """
    Model for a stratified storage with number of nodes N
    (mixing only in case of temperature inversion)
    """

    def __init__(self, n, mc, Tenv, Tinit=None, kLoss=None):
        """
        parameters:
        n: number of nodes
        mc: storage capacity in [kW/K]
        Tenv: temperature of environment (used in stepLosses()
        kLoss: inverse of the time constant 1/tau = UA/MC 
        Tinit: initial temperature - either constant value or array of length n

        state variables:
        q: energy stored (reference: 0 °C)
        """
        
        self.n = n
        self.mc = mc
        self.Tenv = Tenv

        # default correlation for storage losses
        if kLoss is None:
            self.kLoss = 0.00904*pow(self.mc,(-1./3.))   #TAU = pow(MCStorage,(1./3.)/(0.005*10*pow((4.19/3.6),-(2./3.)))
                                                          # Water storage with surface/volume ratio = 10m2/m3, U-Value = 0.5 W/m2.
        else:
            self.kLoss = kLoss

        self.reset(Tinit=Tinit)

    def reset(self, Tinit=None):
        """
        intialises storage temperatures and state variables
        """
        self.T = np.zeros(self.n) + self.Tenv if Tinit is None else Tinit
        self.aggregates()
    
    def charge(self, cx, Ti):
        """
        calculates charge (cx > 0) or discharge (cx<0)
        returns the outlet temperature and the 
        """
        if abs(cx) < EPSILON:
            return 0
        
        elif cx >= self.mc or cx <= - self.mc:
            self.T = Ti*np.ones(self.n)
            dq = self.mc*(Ti - self.Tav)
            To = Ti - dq/cx
        
        else:
            fCharge = (abs(cx)/self.mc)*self.n
            nCharge = int(fCharge)
            xCharge = fCharge - nCharge

            if cx > 0:
                To = (np.sum(self.T[:nCharge]) + xCharge*self.T[nCharge])/fCharge
                
                self.T[:-nCharge-1] = (1-xCharge)*self.T[nCharge:-1] + xCharge*self.T[nCharge+1:] 
                self.T[self.n-nCharge-1] = (1-xCharge)*self.T[-1] + xCharge*Ti
                self.T[self.n-nCharge:] = Ti
            else:
                To = (np.sum(self.T[self.n-nCharge:]) + xCharge*self.T[self.n-nCharge-1])/fCharge
                
                self.T[nCharge+1:] = (1-xCharge)*self.T[1:self.n-nCharge] + xCharge*self.T[:-nCharge-1] 
                self.T[nCharge] = (1-xCharge)*self.T[0] + xCharge*Ti
                self.T[:nCharge] = Ti

            dq = cx*(Ti - To)

        self.aggregates()
        return dq

    def getTo(self, cx, Ti=None):
        """
        calculates the outlet temperature for a charge (cx > 0) or discharge (cx<0) ratio
        but does not actually change the storage temperature
        returns the outlet temperature

        note:   for abs(cx) > self.mc the outlet temperature will depend on the inlet temperature
                if Ti is not given, Tav will be returned
        """
        if cx >= self.mc or cx <= - self.mc:
            return self.Tav if Ti is None else (self.Tav + (abs(cx)-1)*Ti)/(abs(cx))
        else:
            fCharge = (abs(cx)/self.mc)*self.n
            nCharge = int(fCharge)
            xCharge = fCharge - nCharge

            if cx > 0:
                To = (np.sum(self.T[:nCharge]) + xCharge*self.T[nCharge])/fCharge
            else:
                To = (np.sum(self.T[self.n-nCharge-1]) + xCharge*self.T[self.n-nCharge])/fCharge
            return To

    def get_CdQ(self, cxmax, dqmax, Ti):
        """
        calculates the charge / discharge flow rate and energy
        under the combined constraint of cxmax and dqmax

        dqmax:  maximum energy exchange (always > 0 !)
        cxmax:  maximum flow rate (>0: charge, <0: discharge
        Ti: inlet temperature for charge / discharge

        :returns:

        C:  flow rate
        dq: energy change
        """
        if abs(cxmax) < EPSILON:
            return 0, 0
        
        Cmax = min(abs(cxmax), self.mc)
        dTNodes = Ti - self.T

        charge = (cxmax >= 0)
        if charge:
            nodeSequence = range(self.n)
        else:
            dTNodes = -dTNodes
            nodeSequence = range(self.n-1, -1, -1)   # calculation order of nodes depends on charge or discharge

        dq = 0.0
        C = 0
        dCnode = self.mc/self.n

        for n in nodeSequence:
            if C >= Cmax: # nothing more to add
                break

            dCn = min(Cmax - C, dCnode)
            dTn = dTNodes[n]

            if dTn <= 0: break

            dq += dTn*dCn                        
            C += dCn

            if dq >= dqmax:
                diff = dq - dqmax
                C -= diff/dTn
                dq = dqmax
                break            

        return C, dq


    def stepLosses(self, dt, Tenv=None):
        """
        updates the Temperatures for heat losses
        over a time step dt
        Tenv: optional input if variable
        """
        if Tenv is not None: self.Tenv = Tenv

        dT = (self.T - self.Tenv)*min(self.kLoss*dt, 1.0)
        dTav = np.average(dT)
        dqLoss = self.mc*dTav

        self.T -= dT

        self.aggregates()
        
        return dqLoss

    def aggregates(self):
        """
        calculates aggregates
        """
        self.Tav = np.average(self.T)
        self.q = self.mc*self.Tav

    @property
    def Ttop(self):
        return self.T[-1]

    @property
    def Tbot(self):
        return self.T[0]

    def printState(self):        
        print "Storage state: TsAvg: %s Q: %s"%(self.Tav, self.q)
        for n in xrange(self.n-1, -1, -1):
            print "%2d: %10.4f"%(n+1, self.T[n])
        print "----------"

        
