# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   hxConnection
#   (ActualHXNet V2)
#   Package for definition of heat exchanger network
#            
#------------------------------------------------------------------------------
#            
#   Class HXConnection: connection between elements in a HX network
#   Class StreamGlobals: contains the connection plus the global data (temp-
#                           perature levels, etc.)
#   Class Splitter: virtual object for defining connections in HX network
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.status import *
from modules.messageLogger import logTrack,logError,logWarning,logDebug
from modules.heatFlow import HeatFlow_t
from database.dbio import check
import database.dbio as dbio
from constants import INFINITE, DEBUG, EPSILON
import math
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, noneCheck
import numpy as np
from copy import copy
from heatExchanger import HeatExchanger
from lib.network.genericNet import GenericNet, Node
from globals import logger


class HXConnection():
    """ Class indicating the connection (origin) of a stream

    :param objectType: type of the object: ProcessH, ProcessC, PipeH, PipeC, Equipment, WHEE, HX, SPLIT
    :param objectNo: number of the object k, j, m, n, h
    :param componentNo: component of the inputs / outputs of an object
    :param splitterNo: number of splitter for objects resulting from a split of streams
    :param splitterBranch: number of branch of the splitter output
    
    """
    def __init__(self, connData=None, HXId=None,
                 objectType=None, objectNo=None, componentNo=None, heating=True, fraction=1.0,
                 objectName=None, streamName=None,
                 connId = None, pinchStreamId=None):
            
        # main parameters defining a connection between objects
        self.objectType = objectType
        self.objectNo = objectNo
        self.componentNo = componentNo
        self.heating = heating

        # descriptive data for easier access
        self.objectName = objectName if objectName is not None else "N.N."
        self.streamName = streamName

        self.h = None # heat exchanger index

        # splitter parameters
        self.fraction = fraction
        self.splitterNo = None
        self.splitterBranch = 0
        self.branch = None

        if connData is not None: # signal for loading connection data from DB with connId = id !!!
            self.loadFromDB(connData)

        else:
            # parameters for storage in SQL tables
            self.objectId = None        # id of the linked object
            self.componentId = None     # only for process in- and outstreams

            self.HXId = HXId            # id of heat exchanger linked (relevant columns: HXSource, HXSink)
            self.connId = connId        # id in table heatexchanger_pinchstream (linking to table pinchstream)
            self.pinchStreamId = pinchStreamId # id in table pinchstream (relevant columns: pinchstreamname)
            self.outletHX = None

        self.pinchStreamName = self.getPinchStreamName() # composed name for compatibility with version2.2 and previous

    @property
    def name(self):
        name = self.objectName         # display name of the connection
        if (self.streamName is not None and self.streamName.strip != ""):
            name += ": " + self.streamName
        return name

##        print "HXConnection created: name %s heating %s"%(self.pinchStreamName, self.heating)

    @property
    def flowDir(self):
        """ returns +1 for inFlow, -1 for outFlow and 0 for other """
        inFlow = (self.objectType=="ProcessH" and self.heating) or (self.objectType=="ProcessC" and not self.heating)
        outFlow = (self.objectType=="ProcessC" and self.heating) or (self.objectType=="ProcessH" and not self.heating)
        return 1*inFlow - 1*outFlow

        
    def check(self):
        """ helper function
        returns self if a valid connection, else None
        """
        return None if self.objectType is None else self
    
    def __repr__(self):

        sType = "cs" if self.heating else "hs"
        result = "HXConn: %s(%s.%s)-%s f %s [%s:%s] Id: %s "%(self.objectType,
                                                                                       self.objectNo,
                                                                                       self.componentNo,
                                                                                       sType,
                                                                                       self.fraction,
                                                                                       self.objectName,
                                                                                       self.streamName,
                                                                                       self.HXId)

        return result

    def loadFromDB(self, connData):
        """ loads connection info from connection tables in DB: heatexchanger_pinchstream and pinchstream
        """
        # read in data from connection tables
        self.connId = connData.id                               # id in table heatexchanger_pinchstream
        self.pinchStreamId = connData.pinchstream_id            # id in table pinchstream
        self.HXId = connData.qheatexchanger_QHeatExchanger_Id   # id in table qheatechanger
        
        self.outletHX = connData.outletOfHX_id         # marker that the connection is outlet of another upstream heat exchanger

        fraction = noneCheck(connData.HeatFlowPercent, 100.0)
        if fraction is None or fraction < 0 or fraction > 100.0:
            logWarning(_("Erroneous mass flow fraction (%s) specified for stream with id %s. 100 percent assumed.")%(fraction, connData.id))
            self.fraction = 1.0
        else:
            self.fraction = 0.01*fraction #convert to fraction of one for further internal processing
        
        connectedStream = dbio.get("pinchstream", id=self.pinchStreamId)
        if connectedStream is None:
            logError("HXConnection.loadData. Corrupt stream id in HX %s: %s"%(self.name, self.pinchStreamId))
            self.objectType = None
            return
            

        if self.outletHX:
            self.objectType = "HX"
            if int(self.outletHX) in Status.int.hxIDdict.keys():
                h = Status.int.hxIDdict[int(self.outletHX)]
                self.objectNo = h
            else:
                logError("Broken serial connection. HXId %s HX source stream %s"%(self.outletHX, self))
                self.objectType = None
                return
                         
            self.objectName = Status.int.hx[h].name

        else:
            self.objectType      = connectedStream.source_type
            self.objectId        = connectedStream.source_id

            systemModel = Status.int.systemModel
            obj = systemModel.getObject(self.compressObjectType(), self.objectId)
            if obj is None:
                logError("Source or sink object not found: object %s(%s)"%(self.compressObjectType(), self.objectId))
                self.objectType = None
                return
            self.objectName = obj.name                
            self.objectNo = obj.no


        self.componentId     = connectedStream.pID
        sourceOrSink    = connectedStream.Hot_Cold
        self.heating = True if sourceOrSink in ['Cold', 'Sink'] else False

        if self.objectType=="HX":
            streamName = _("cold side out") if self.heating else _("hot side out")
            self.componentNo = 0

        elif self.objectType in ["ProcessH", "ProcessC"]:
            if self.componentId is None:    # this option here for backward compatibility with V2.5 and before
                                            # in upgrade function PID is set to None
                self.pinchStreamName = connectedStream.name
                streamNo, streamId, streamName = self.getIndexFromStreamName()
                self.componentId = streamId

                # save to DB at first reading
                connectedStream.pID = streamId
                dbio.commit()
            else:
                if self.componentId <= 0:
                    streamNo = self.componentId
                    streamName = _("maintenance") if streamNo==0 else _("start-up")
                else:
                    k = self.objectNo
                    if self.flowDir > 0:
                        procStreams = Status.int.systemModel.processStreamsIn[k]
                    else:
                        procStreams = Status.int.systemModel.processStreamsOut[k]

                    streamNo = None
                    for s, stream in enumerate(procStreams):
                        if stream.id == self.componentId:
                            streamNo =  s+1
                            streamName = stream.Name
                            break

            if streamNo is None:
                logger.logError("HXConnection: could not find stream for Process %s with id %s"%(self.objectNo+1, self.componentId))
                self.objectType = None
                return
                    
            self.componentNo = streamNo
                
        else:
            self.componentNo = 0 if self.componentId is None else self.componentId # use 0 as default in order to avoid None's
            streamName, objectName = self.getDefaultStreamName()
            streamName = noneCheck(streamName, "---")

        self.streamName=streamName

        if DEBUG != "OFF":
            logTrack("HeatExchanger.getConnectionsFromDB HXconnection \n%s"%(connData))


    def writeToDB(self, TNom = (None, None), commit=True):
        
        if self.objectType not in ["ProcessH", "ProcessC", "PipeH", "PipeC", "Equipment", "WHEE"]:
            logError("BUG: Cannot save stream %s.%s-%s%s. Erroneous object type"%(self.objectType, self.objectNo, self.componentNo, ('H' if self.heating else 'C')))
            return

        systemModel = Status.int.systemModel
        self.objectId = systemModel.objectId(self.compressObjectType(), self.objectNo)
        self.componentId = self.componentNo         # for most objectTypes it is sufficient to store the component number, except for process streams
        if self.objectType in ["ProcessH", "ProcessC"] and self.componentNo > 0:
            s = self.componentNo - 1
            k = self.objectNo
            if self.flowDir > 0:
                stream = systemModel.processStreamsIn[k][s]
            else:
                stream = systemModel.processStreamsOut[k][s]
            self.componentId = stream.id

        streamType = "Sink" if self.heating else "Source"

        self.pinchStreamName = self.getPinchStreamName()

        # collect data for table pinchstream        
        stream = {'source_type':    check(self.objectType),
                  'source_id' :     self.objectId,              # new in V3 - storage of objectID
                  'pID':            check(self.componentId),    # new in V3 - for in- or outstreams streamId is stored, still work in progress (see above)
                  'Hot_Cold':       check(streamType),
                  'name':           check(self.pinchStreamName), # key parameter until V2.5 for identification and linking of pinchstreams
                  'ProjectID':      check(Status.PId),
                  'AlternativeProposalNo': check(Status.ANo)}

        if self.pinchStreamId is None: self.pinchStreamId = dbio.insertRow("pinchstream", **stream)
        pinchStream = dbio.get("pinchstream", id=self.pinchStreamId)      
        pinchStream.set(**stream)

        percent = self.fraction*100.0
        hxconn = {'qheatexchanger_QHeatExchanger_Id': check(self.HXId),
                  'pinchstream_id': check(self.pinchStreamId),
                  'inletTemp':      check(TNom[0]),
                  'outletTemp':     check(TNom[1]),
                  'outletOfHX_id':  check(self.outletHX),
                  'HeatFlowPercent':check(percent),
                  'ProjectID':      check(Status.PId),
                  'AlternativeProposalNo':check(Status.ANo)}

        if self.connId is None:
            self.connId = dbio.insertRow("heatexchanger_pinchstream", **hxconn)

        row = dbio.get("heatexchanger_pinchstream", id=self.connId)
        if row is not None:
            row.set(**hxconn)

        if commit: dbio.commit()

##        print "TEST TEST HXConn.writeToDB: %s"%self
##        print stream
##        print hxconn
        
        return self.connId, self.pinchStreamId

    def reconnect(self, conn):
        """ Used for reconnecting heat exchangers before deleting
            Takes over the link from conn, without taking over id's
        """
        self.objectType = conn.objectType
        self.objectNo = conn.objectNo
        self.componentNo = conn.componentNo
        self.outletHX = conn.outletHX

        self.pinchStreamName = conn.pinchStreamName

        self.splitterNo = conn.splitterNo
        self.splitterBranch = conn.splitterBranch

        if Status.SimTool == "NP":
            self.fraction *= conn.fraction

    def copy(self):
        conn = HXConnection()
        conn.reconnect(self)
        return conn
    
    def deleteFromDB(self):
##        print "HXConnection.deleteFromDB: connId %s pinchStreamId %s"%(self.connId, self.pinchStreamId)
        if self.connId is not None:
            dbio.delete("heatexchanger_pinchstream", id=self.connId)
            dbio.commit()

        if self.pinchStreamId is not None:
            dbio.delete("pinchstream", id=self.pinchStreamId)
            dbio.commit()
        
    def primaryFraction(self):
        
        if self.objectType == "SPLIT":            
            if self.heating:
                splitter = Status.int.sinkSplitters[self.objectNo]
            else:
                splitter = Status.int.sourceSplitters[self.objectNo]
            
            return splitter.mainStream.primaryFraction() * splitter.fractions[self.componentNo]
        elif self.objectType == "HX":
            try:
                hx = Status.int.hx[self.objectNo]
                if self.heating:
                    conn = hx.sinkstreams[0]
                else:
                    conn = hx.sourcestreams[0]
                return self.fraction * conn.primaryFraction()
            except: # security for avoiding crash if HX is not connected
                return -1.0
            
        else:
            return self.fraction

    def compactConnToUpstreamComponent(self):
        """ function added for compatibility with V2.2 module where
            no explicit modelling of splitters has been done

            returns self, if objectType <> "SPLIT", and the splitter - mainstream else
        """

        if self.objectType == "SPLIT":
##            print "HXConnection.compact(...): compacting splitter %s"%self.objectNo
            if self.heating:
                splitter = Status.int.sinkSplitters[self.objectNo]
            else:
                splitter = Status.int.sourceSplitters[self.objectNo]
            conn = copy(splitter.mainStream)
            conn.fraction = splitter.fractions[self.componentNo]
            return conn
        else:
            return self

    def upStreamHX(self):
        """
        Auxiliary function checking if there is some HX upstream
        """
        if self.objectType=="HX":
            return Status.int.hx[self.objectNo]
        elif self.objectType=="SPLIT":
            splitter = Status.int.sinkSplitters[self.objectNo] if self.heating else Status.int.sourceSplitters[self.objectNo] 
            return splitter.mainStream.upStreamHX()
        else:
            return None

    def hasSameSource(self, conn):
        #NOTE: in the equivalent streamGlb-function a special treatment for splitters is foreseen. not (yet ?) here
        
        ok = (self.objectType == conn.objectType) and \
         (self.objectNo == conn.objectNo) and \
         (self.componentNo == conn.componentNo) and \
         (self.heating == conn.heating)
        return ok

    def getDefaultStreamName(self):
        if self.objectType=="HX":
            h = self.objectNo
            objectName = Status.int.hx[h].name
            streamName = _("cold side out") if self.heating else _("hot side out")
            return streamName, objectName 
        else:
            matches = [stream for stream in Status.int.systemModel.streams if self.hasSameSource(stream.origin)]
            if len(matches) > 0:
                match = matches[0]
                return match.name, match.origin.objectName
            else:
                return None, None       

    def compressObjectType(self):
        """
        Helper function
        Translates the codes used in HXConnection into systemModel object codes
        """
        
        if self.objectType in ["ProcessC", "ProcessH"]:
            return "PROC"
        elif self.objectType in ["PipeC", "PipeH"]:
            return "PIPE"
        elif self.objectType in ["Equipment"]:
            return "EQ"
        else:
            return self.objectType
        
    def compact(self):
        return (self.objectType, self.objectNo, self.componentNo, self.heating)

#===============================================================================
# Helper functions for translating internal HXConn information
# defined by oType, oNo, cNo and heating/cooling
# into the current storage structure, where connections are stored by
# a) the objectName (-> defining oType and oNo)
# b) the pinchStreamName (-> defining cNo and heating / cooling)
#-------------------------------------------------------------------------------
# objective is eliminating the use of pinchStreamName completely from all
# internal processing
#===============================================================================

    def getIndexFromStreamName(self):
        """associates the name of the stream in the table pinchstream with 
        the name of the process streams

        needed for backward compatibility with V2.2, as in this version
        the component-no is not stored
        
        """
        pinchStreamName = self.pinchStreamName
        k = self.objectNo
##        logTrack("HXConnection.getIndexFromStreamName %s"%(self.pinchStreamName))
        
        streamNo = None
        streamId = None
        
        if pinchStreamName is None:
            logError("HXConnection.getIndexFromStreamname: erroneous pinchStreamName %s"%pinchStreamName)
            return None

        if "_Maintenance" in pinchStreamName or "Maintainance" in pinchStreamName:
            streamNo = 0
            streamId = 0
            streamName = _("start-up")
        elif "_StartUp" in pinchStreamName:
            streamNo = -1
            streamId = -1
            streamName = _("maintenance")
        elif "_Circulation" in pinchStreamName:
            inStreams = Status.int.systemModel.processStreamsIn[k]
            
            for s, stream in enumerate(inStreams):
                if stream.Name in pinchStreamName:
                    streamNo =  s+1
                    streamId = stream.id
                    streamName = stream.Name

        elif "_WasteHeat" in pinchStreamName:           
            outStreams = Status.int.systemModel.processStreamsOut[k]
            for s,stream in enumerate(outStreams):
                if stream.Name in pinchStreamName:
                    streamNo = s+1
                    streamId = stream.id
                    streamName = stream.Name
            
        else:
            logError("HXConnection.getIndexFromStreamname: erroneous pinchStreamName %s"%pinchStreamName)
            return None, None, None

        if streamNo is None:
            logError("HXConnection.getIndexFromStreamname: could not associate %s to any stream"%pinchStreamName)
            return None, None, None
            
        if streamNo != self.componentNo:
            logDebug("WARNING: conflict in association to streams")
            logDebug("componentNo %s streamNo %s streamName %s"%(self.componentNo, streamNo, self.pinchStreamName))
##            self.componentNo = streamNo

        return streamNo, streamId, streamName

    def getPinchStreamName(self):
        """
        inverse function of getIndexFromStreamName
        composes the pinchStreamName as used in V2.5 and previous from the variables objectName, streamName,
        and - in the case of processes - heating or cooling
        """
        if self.streamName is None: self.streamName = ""
        psName = self.objectName
        if self.objectType in ["ProcessH", "ProcessC"]:
            oType = self.objectType
            cNo = self.componentNo
            heating = self.heating
            
            if self.flowDir > 0:
                if cNo==0:
                    psName += "_Maintenance"
                elif cNo==-1:
                    psName += "_StartUp"
                elif cNo>0:
                    psName += "_" + self.streamName + "_Circulation"
            elif self.flowDir < 0:
                psName += "_WasteHeat" + "_" + self.streamName
        elif self.streamName.strip() != "":
            psName += "_" + self.streamName

        return psName


def genericStreamName(oType, cNo, heating, eqClass=None, fuelId=None):
    """
    centralises the definition of generic stream names
    for purpose of displaying
    """

    if oType in ["ProcessH", "ProcessC"]:
        streamNames=[_("maintenance"), _("start-up")]
    elif oType == "PipeH":
        if heating:
            streamNames = [_("input")]
        else:
            streamNames = [_("waste heat")]
    elif oType == "PipeC":
        if heating:
            streamNames = [_("waste cooling")]
        else:
            streamNames = [_("input")]
    elif oType in ["WHEE"]:
        streamNames = [_("waste heat")]
    elif oType in ["Equipment"]:
        streamNames = []
        # heat input to equipment
        if heating:
            if eqClass == "BB":
                streamNames = [_("combustion air")]
            elif eqClass == "HP":
                streamNames = [_("Low temperature heat source"), _("driving heat")]
            elif eqClass in ["CH", "CHP"]:
                streamNames = [_("driving heat")]

        # waste heat streams of equipment                    
        elif eqClass not in ["ST"]:
            if eqClass == "BB" and fuelId is not None:
                streamNames = [_("exhaust gas")]
            else:
                streamNames = [_("waste heat")]

    else:
        streamNames = [] 

    index = abs(cNo)    # converts -1 from start-up into +1
    return streamNames[index] if index >= 0 and index<len(streamNames) else None

global idCtr
idCtr = 0

class StreamGlobals():
    """ Contains the origin (class HXConnection) and the global data of a stream

    :params Q: energy content corresponding to the full heat flow (w/o considering fraction)
    :params dotQavg: average power during operating hours corresponding to the full heat flow
    :params mcp: parameters of full flow (not reduced by fraction)
    :params pinchMask: indicates whether stream is considered in pinch analysis or not. Default: True
    :params required: True: indicates that this is a demand that has to be covered. False indicates waste heat / cooling that can be used optionally. Default: True
    """

    def __init__(self, name=None, T0=None, T1=None, Q=None, dotQavg=None, hours=None, heating=None, origin=None, pinchMask=True, fluidId = None, required=True):
        self.origin = origin
        self.name = name
        self.T0 = T0
        self.T1 = T1
        self.Q = Q
        self.dotQavg = dotQavg
        self.hours = hours
        self.mcp = None
        self.hf = None
        self.pinchMask = pinchMask # 
        self.required = required # 

        if (heating is None) and (origin is not None):
            self.heating = origin.heating
        else:
            self.heating = heating

        if self.name is None and origin is not None:
            self.name = origin.pinchStreamName
            
        self.fraction = 1.0

        # special attributes to be used in designHXnet (interconnection of stream-segments)
        self.upstream = [] # upstream stream-segment. Empty list for original sources
        self.downstream = []
        self.hx = None # stream segments corresponding to ("occupied by") a heat exchanger, otherwise = None
        self.fluidId = fluidId #NOTE: fluidId is now used also in buildHXNet for determining the fluidId of the HX from its sources
        self.shift = 0 # used only in NP version of BuildCC. Overwritten by HeatFlow.exportGlb

        global idCtr
        idCtr += 1
        self.id = idCtr

    def __repr__(self):

        up = None if len(self.upstream)==0 else self.upstream[0].id
        down = None if len(self.downstream)==0 else self.downstream[0].id
        
        if self.T0 is None or self.T1 is None:
            q = self.Q
            mcp = self.mcp if self.mcp is not None else -1.0
            result = "%40s (%10s) [        ->        ] mcp %8.0f f%5.3f [%s->%s->%s]"%(self.name, q, mcp, self.fraction, up, self.id, down)
        else:
            try:
                mcp = self.mcp*self.fraction
            except:
                mcp = -1.0
            try:
                q = mcp*(self.T1 - self.T0)
            except:
                q = self.Q
            q = noneCheck(q, -1.0)

            result = "%40s (%10s) [%8.2f->%8.2f] mcp %8.0f f%5.3f [%s->%s->%s]"%(self.name, q, self.T0, self.T1, mcp, self.fraction, up, self.id, down)
        return result
        
    def copyAttributesFromHeatFlow_Tit(self, hf=None):
        """ Used in SimTool Plus mode. Imports information
            from HeatFlow_Tit objects
        """
        
        if hf is None:
            hf = self.hf
        else:
            self.hf = hf

        if self.hf.T0avg == self.hf.T1avg:
            self.mcp = INFINITE
        else:
            self.mcp = hf.Qsum / (hf.T1avg - hf.T0avg)

        #TODO: this averages out the cp variations over the temp - interval in condensing cases
        # substitute by more accurate treatment
        self.hours = hf.hours
  
        if self.hours == 0:
            self.dotQavg = 0
        else:
            self.dotQavg = hf.Qsum / hf.hours

        self.heating = hf.heating

        self.T0 = hf.T0avg + hf.DTShift
        self.T1 = hf.T1avg + hf.DTShift

        self.Q = hf.Qsum
        
    def copyAttributesFromHeatFlow_t(self):
        """ Equivalent to similar function for SimTool Plus mode.
            Used in Pinch analysis
            Imports information from HeatFlow_t objects in NP mode
        """
        
        origin = self.origin
        
        if origin.objectType in ["ProcessH", "ProcessC"]:

            k = origin.objectNo
            if (origin.objectType == "ProcessH" and origin.heating) or \
               (origin.objectType == "ProcessC" and not origin.heating):
                if origin.componentNo == 0:
                    hf = Status.int.UPHm_Tt[k]
                elif origin.componentNo == -1:
                    hf = Status.int.UPHs_Tt[k]
                else:
                    hf = Status.int.UPHc_Tt[k].getPart(origin.componentNo-1)
            else:
                hf = Status.int.UPHw_Tt_hf[k].getPart(origin.componentNo-1)
                    
        elif origin.objectType in ["PipeH", "PipeC"]:
            m = origin.objectNo
            pipeIsHeating = Status.int.systemModel.pipeObject[m].heating
            if pipeIsHeating == origin.heating:
                hf = Status.int.USHCPipe_Tt_hf[m]
            else:
                hf = Status.int.QWHPipe_Tt[m]
                                    
        elif origin.objectType in ["WHEE"]:
            n = origin.objectNo
            hf = Status.int.QWHEE_Tt_hf[n]
                                    
        elif origin.objectType in ["Equipment"]:
            j = origin.objectNo
            if origin.heating:
                if origin.componentNo == 0:
                    hf = Status.int.QHXEqReq_Tt[j]
                else:
                    hf = Status.int.FETHeat_j[j]
            else:
                hf = Status.int.QWHEq_Tt[j]


        if hf.T0avg == hf.T1avg:
            self.mcp = INFINITE
        else:
            self.mcp = hf.Q / (hf.T1avg - hf.T0avg)

        #TODO: this averages out the cp variations over the temp - interval in condensing cases
        # substitute by more accurate treatment
        self.hours = hf.hop
  
        if self.hours == 0:
            self.dotQavg = 0
        else:
            self.dotQavg = hf.Q / hf.hop

##        self.heating = hf.heating

        self.T0 = hf.T0avg
        self.T1 = hf.T1avg
        self.Q = hf.Q
        self.hf = hf

    def copyAttributesFromSource(self):
        """ Equivalent to similar function for SimTool Plus mode.
            Used in Pinch analysis
            Imports information from HeatFlow_t objects in NP mode
        """
        
        origin = self.origin
        
        if origin.objectType in ["ProcessH", "ProcessC"]:

            k = origin.objectNo
            proc = Status.int.systemModel.process[k]
            
            if (origin.objectType == "ProcessH" and origin.heating) or \
               (origin.objectType == "ProcessC" and not origin.heating):
                if origin.componentNo == 0:
                    self.Q = proc.UPHm
                    self.T0 = proc.PT
                    self.T1 = proc.PT
                elif origin.componentNo == -1:
                    self.Q = proc.UPHs
                    self.T0 = proc.PTStartUp
                    self.T1 = proc.PT
                else:
                    s = origin.componentNo-1
                    stream = Status.int.systemModel.processStreamsIn[k][s]
                    self.Q = stream.UPHc
                    self.T0 = stream.PTInFlow
                    self.T1 = proc.PT
                        
            else:
                s = origin.componentNo-1
                stream = Status.int.systemModel.processStreamsOut[k][s]
                self.Q = stream.UPHw
                self.T0 = stream.PTOutFlow
                self.T1 = proc.PT
                    
        elif origin.objectType in ["PipeH", "PipeC"]:
            m = origin.objectNo
            pipe = Status.int.systemModel.pipe[m]
            pipeIsHeating = Status.int.systemModel.pipeObject[m].heating
            if pipeIsHeating == origin.heating:
                self.Q = pipe.USHPipe
            else:
                self.Q = pipe.QWHPipe
                                    
        elif origin.objectType in ["WHEE"]:
            n = origin.objectNo
            whee = Status.int.systemModel.whee[n]
            self.Q = whee.QWHEE
                                    
        elif origin.objectType in ["Equipment"]:
            j = origin.objectNo
            eq = Status.int.systemModel.equipment[j]
            if origin.heating:
                if origin.componentNo == 0:
                    self.Q = eq.QHXEq
                else:
                    self.Q = eq.FETHeat_j
            else:
                self.Q = eq.QWHEq


##        if hf.T0avg == hf.T1avg:
##            self.mcp = INFINITE
##        else:
##            self.mcp = hf.Q / (hf.T1avg - hf.T0avg)

        #TODO: this averages out the cp variations over the temp - interval in condensing cases
        # substitute by more accurate treatment
##        self.hours = hf.hop
##  
##        if self.hours == 0:
##            self.dotQavg = 0
##        else:
##            self.dotQavg = hf.Q / hf.hop

        
    def copyOrigin(self, streamGlb):

##        print "StreamGlobals.copyOrigin: copying %s"%streamGlb
        if self.origin is None:
            self.origin = HXConnection()

        self.origin.pinchStreamName = streamGlb.origin.pinchStreamName #full composed name        
        self.origin.objectType = streamGlb.origin.objectType
        self.origin.objectNo = streamGlb.origin.objectNo
        self.origin.objectName = streamGlb.origin.objectName
        self.origin.streamName = streamGlb.origin.streamName
        self.origin.componentNo = streamGlb.origin.componentNo
        self.origin.fraction = streamGlb.origin.fraction
        self.origin.heating = streamGlb.origin.heating
        self.heating = self.origin.heating
##        print "StreamGlobals.copyOrigin: result %s"%self

        if self.name is None:
            self.name = self.origin.pinchStreamName

    def copy(self):
        new = StreamGlobals(name = self.name,
                            T0 = self.T0,
                            T1 = self.T1,
                            Q = self.Q,
                            dotQavg = self.dotQavg,
                            hours = self.hours,
                            heating = self.heating)
        
        new.mcp = self.mcp
        new.hf = self.hf
        new.fraction = self.fraction

        # copying special attributes for designHXNet
        new.upstream = self.upstream
        new.downstream = self.downstream
        new.hx = self.hx
        new.required = self.required
        
        new.copyOrigin(self)
        return new

    def divide(self, TSplit, keep=0):
        """ divides a stream (self) into self=up and a new stream down

            [upstream] -> [up]||[down] -> [downstream]

        keep: 0 -> the source stream (self) is converted to the upstream
              1 -> ...                                   to the downstream
              other -> both up and down are copies of self
              (in the case of real splitting only)

        """

        if self.T0 >= TSplit - EPSILON:
            if self.heating:
                return None, self
            else:
                return self, None
        
        elif self.T1 <= TSplit + EPSILON:
            if self.heating:
                return self, None
            else:
                return None, self
        
        else:
            if keep==0:
                up = self
                down = self.copy()
            elif keep==1:
                up = self.copy()
                down = self
            else:
                up = self.copy()
                down = self.copy()

            up.downstream = [down]
            down.upstream = [up]

            # change upstream entry in downstreams of self.
            for s in down.downstream:
                if up in s.upstream:
                    i = s.upstream.index(up)
                    s.upstream[i] = down
                else:
                    s.upstream.append(down)
            
            if self.heating:    # cold stream / heating demand
                down.T0 = TSplit
                up.T1 = TSplit
            else: # hot stream / cooling demand
                down.T1 = TSplit
                up.T0 = TSplit

            return up, down

    def hasSameSource(self, conn):

        if conn is None: return False # security for avoiding crashes
        
        if conn.objectType == "SPLIT":
            if self.heating:
                splitter = Status.int.sinkSplitters[conn.objectNo]
            else:
                splitter = Status.int.sourceSplitters[conn.objectNo]
                
            return self.hasSameSource(splitter.mainStream)
            

        ok = (self.origin.objectType == conn.objectType) and \
             (self.origin.objectNo == conn.objectNo) and \
             (self.origin.componentNo == conn.componentNo) and \
             (self.origin.heating == conn.heating)
        return ok

    def upstreamHX(self):
        if self.upstream == []:
            return None
        else:
            upstream = self.upstream[0]
            hx = upstream.hx
            
            #TODO: up to now only one upstream sector is considered.
            if hx is None:
                return self.upstream[0].upstreamHX() # recursive search for upstream of upstream
            else:
                return upstream

    def split(self, f, lists = []):

        s = self
        sp = self.copy()

        # attach the splits to the lists containing self
        for lst in lists:
            if s in lst: lst.append(sp)
        
        s.fraction *= f
        sp.fraction *= (1-f)

        # split also the downstream if necessary
        downstream = []
        sp.downstream = []
        
        # first try to directly allocate a downstream to a split
        downstreams = self.downstream
        allocated = []
        free = s.fraction
        freep = sp.fraction
        for i, ds in enumerate(downstreams):
            if abs(ds.fraction - s.fraction) < 0.001:
                downstream.append(ds)
                ds.upstream = [s]
                allocated.append(ds)
                free = 0
            elif abs(ds.fraction - sp.fraction) < 0.001:
                sp.downstream.append(ds)
                ds.upstream = [sp]
                allocated.append(ds)
                freep = 0

##        for i, ds in enumerate(downstreams):

        downstreams = [ds for ds in self.downstream if ds not in allocated]
        i = 0
        while i < len(downstreams) and max(free, freep) >= 0.001:
            downstreams = sorted(downstreams, key = lambda ds: ds.fraction, reverse=True)           
            ds = downstreams[i]

            if ds.fraction <= min(free, freep) + 0.001:
                if free <= freep:
                    downstream.append(ds)
                    ds.upstream = [s]
                    free -= ds.fraction
                else:
                    sp.downstream.append(ds)
                    ds.upstream =  [sp]
                    freep -= ds.fraction
            elif ds.fraction <= free + 0.001:
                downstream.append(ds)
                ds.upstream = [s]
                free -= ds.fraction
            elif ds.fraction <= freep + 0.001:
                sp.downstream.append(ds)
                ds.upstream =  [sp]
                freep -= ds.fraction
            else:
                fmax = max(free, freep)/ds.fraction
                ds, dsp = ds.split(fmax, lists=lists)
                downstreams.append(dsp)
                continue
            
            i += 1

        s.downstream = downstream
        return s, sp
        

def buildHXNet(split=False, loadFullDataSet=False):
    """ Loads heat exchanger data from SQL and builds list of 
    HeatExchanger-class objects for existing heat exchanger network

    NOTE: it should be assured beforehand that the systemModel is created
    and connections are built. This is done automatically, if this
    function is called via systemModel.build().

    Add-On EINSTEIN V3: builds also the network model
    
    """

    systemModel = Status.int.systemModel        

    # set-up basic network without hx and splitters
    srcNet = GenericNet()
    snkNet = GenericNet()
    
    for s, stream in enumerate(systemModel.streams):
        net = snkNet if stream.heating else srcNet

        # add stream twice, once as start and once as end
        code = nodeCode(stream.origin)
        net.addNode(Node(0, 1, obj=stream.origin, objectCode=code))
        
        rdCode = list(code)
        rdCode[0] = "RD."+rdCode[0]
        rdCode = tuple(rdCode)
        net.addNode(Node(1, 0, obj=stream.origin, objectCode=rdCode))
        
        n = net.nNodes
        net.connectNodes(n-2, 0, n-1, 0)

    hxes = systemModel.hx
    logTrack("HeatExchanger.buildHXNet no. of HXes = %s"%len(hxes))

    Status.int.hx = []
    Status.int.hxIDdict = {}
    Status.int.sourceSplitters = []
    Status.int.sinkSplitters = []
    
    for h, hx in enumerate(hxes):
        
        hxId = hx.id
        name = hx.HXName
        if hx.UA is not None:
            UA = hx.UA 
        else:
            UA = INFINITE
            logWarning(_U("Heat exchanger %s has no UA value given, Infinity is assumed.") % U(name))
        HXNo = hx.HXNo
        
        MCStorage = hx.StorageSize
        if MCStorage is not None:
            MCStorage *= (4.19/3.6) # conversion from equivalent m3 of standard water to kWh/K
        else:
            MCStorage = 0
            logWarning(_U("No storage size specified for heat exchanger %s. Zero size assumed") % U(name))

        newHX = HeatExchanger(name=name,id = hxId,UA=UA,HXNo=HXNo,MCStorage=MCStorage,writeAllowed = (Status.ANo <> 0))
                
        Status.int.hx.append(newHX)
        Status.int.hxIDdict.update({hxId:h})

    for h, hxData in enumerate(hxes):
        hx = Status.int.hx[h]
        hx.loadFromDB(dict=hxData, loadConnections=False)
        
    for hx in Status.int.hx:
        hx.getConnectionsFromDB()

    # determine fluidId of hx from connections (-> info stored in pinchstream-table and read-in in getConnectionsFromDB is not reliable !)
    # TODO: this will no longer be necessary once it is assured that the parameters stored with the HX are correct
    #       could be maintained anyway as a double-check for security
    
    for hx in Status.int.hx:
        if len(hx.sourcestreams) > 0:
            primarySource = getPrimarySourceOfConn(hx.sourcestreams[0])
            stream = Status.int.systemModel.getStream(primarySource)
            if stream is not None:
                hx.fluidIdSource = stream.fluidId
            else:
                hx.fluidIdSource = 0 # should not occur, but for security
        else:
            hx.fluidIdSource = 0 # assign something different from None for avoiding crash
            
        if len(hx.sinkstreams) > 0:
            primarySink = getPrimarySinkOfConn(hx.sinkstreams[0])
            stream = Status.int.systemModel.getStream(primarySink)
            if stream is not None:
                hx.fluidIdSink = stream.fluidId
            else:
                hx.fluidIdSource = 0 # should not occur, but for security
                
        else:
            hx.fluidIdSink = 0

    if Status.SimTool == "Plus": # explicit modelling of splitter objects only in SimTool-Plus mode
        for h, hx in enumerate(Status.int.hx):
        
            if len(hx.sourcestreams) > 0:
                source = hx.sourcestreams[0]
                if source.fraction < 1.0:
##                    print "BuildHXNet: HX %s adding sourceSplit %s"%(hx.name, source)
                    splitterNo, branch = addSourceSplit(source, hx.fluidIdSource)
                    if split:
                        hx.sourcestreams = [HXConnection(HXId=-1,
                                                         objectType="SPLIT",
                                                         objectNo=splitterNo,
                                                         componentNo=branch,
                                                         fraction=1.0,
                                                         heating=False,
                                                         connId = source.connId,
                                                         pinchStreamId = source.pinchStreamId)]
            if len(hx.sinkstreams) > 0:
                sink = hx.sinkstreams[0]
                if sink.fraction < 1.0:
                    splitterNo, branch = addSinkSplit(sink, hx.fluidIdSink)
                    if split:
                        hx.sinkstreams = [HXConnection(HXId=-1,
                                                         objectType="SPLIT",
                                                         objectNo=splitterNo,
                                                         componentNo=branch,
                                                         fraction=1.0,
                                                         heating=True,
                                                         connId = sink.connId,
                                                       pinchStreamId = sink.pinchStreamId)]
                    

        
        for i, splitters in enumerate([Status.int.sourceSplitters, Status.int.sinkSplitters]):
            net = srcNet if i==0 else snkNet
            splitterBuffer = splitters
                
            nIts = 0
            while len(splitterBuffer) > 0 and nIts < 10:
                nIts += 1
                unresolved = []
                for splitter in splitterBuffer:
                    conn = splitter.getMainStreamRD()   # jump over heat exchangers, as those are not yet within the network
                                                        # will be added later on

                    upstreamNode = net.findNode(objectCode=nodeCode(conn))
                    if upstreamNode is None:
                        unresolved.append(splitter)
                        continue
                    branch = conn.componentNo

                    if upstreamNode.nodeType=="SPLIT":
                        flow = upstreamNode.qout[branch]
                    else:
                        flow = upstreamNode.qout[0]
                    net.splitFlow(flow, splitter.fractions, obj=splitter, objectCode=("SPLIT", splitter.no, -1, conn.heating))
                splitterBuffer = unresolved

        # first set primary source/sink as destination of HX residuals
        for h, hx in enumerate(Status.int.hx):
            hx.srcRD = getPrimarySource(h, split=True)
            hx.snkRD = getPrimarySink(h, split=True)

        # then eliminate destinations for upstream HXes
        for h, hx in enumerate(Status.int.hx):
    ##        print "TEST TEST buildHXNet - eliminating upstreamHX destinations for hx %s (%s)"%(h, hx.name)
            if len(hx.sourcestreams) > 0:
                source = hx.sourcestreams[0]
                if source.objectType == "HX":
                    objectNo = source.objectNo
    ##                hsrc = Status.int.hxIDdict[objectNo]
                    hsrc = objectNo
                    hxsrc = Status.int.hx[hsrc]
                    hxsrc.srcRD = None #eliminate destinations for upstream HXes

            if len(hx.sinkstreams) > 0:
                sink = hx.sinkstreams[0]
                if sink.objectType == "HX":
                    objectNo = sink.objectNo
    ##                hsnk = Status.int.hxIDdict[objectNo]
                    hsnk = objectNo
                    hxsnk = Status.int.hx[hsnk]
                    hxsnk.snkRD = None #eliminate destinations for upstream HXes

            if len(hx.sourcestreams)==0 or len(hx.sinkstreams)==0:
                logError("BuildHXNet: some connection missing for HX %s"%hx.name)
                continue

    loop = True
    nhx = len (Status.int.hx)
    hxOK = np.zeros((3, nhx))    # indicates whether connections are ok (false in the case that definition of some upstream HX is still missing)

    nitsMax = 100
    nits = 0
    while loop and nits < nitsMax:
        nits += 1
        loop = False
        for h, hx in enumerate(Status.int.hx):

            if len(hx.sourcestreams)==0 or len(hx.sinkstreams)==0:
                logError("BuildHXNet: missing source or sink stream in HX %s"%h)
                continue
            src = hx.sourcestreams[0]
            snk = hx.sinkstreams[0]
            
            for i, conn in enumerate([src, snk]):
                net = srcNet if i==0 else snkNet
                
                if hxOK[i][h]==1: continue     # connection has already been processed
                code = nodeCode(conn)
                hconn = code[1]
                if code[0]=="HX":
                    if hconn >=0 and hconn < nhx:   # security check in order to avoid crash in case of erroneous data
                        if hxOK[i][hconn]==0:
                            loop = True
                            continue                # cannot process yet. wait for the next round until upstream hx node is created

                node = net.findNode(objectCode=code)
                if node is None:
                    logError("BuildHXNet: could not find node for conn %s"%str(code))
                else:
                    if node.nodeType=="SPLIT":
                        flow = node.qout[conn.componentNo]
                    else:
                        flow = node.qout[0]
                    net.divideFlow(flow, obj=hx, objectCode=("HX", h, 0, conn.heating))

                hxOK[i][h] = 1 

    Status.int.systemModel.srcNet = srcNet
    Status.int.systemModel.snkNet = snkNet

    if DEBUG != "OFF":
        print "Source net: %s"%srcNet
        print "Sink net: %s"%snkNet
        printHXNet()

    if Status.SimTool != "Plus":
        clearHXResults()

def printHXNet():
    """ function for debugging only
        prints the relevant information on the heat exchanger network and connections
    """
    #TODO: might be improved for better readability
    
    print "=====================================\nprintHXNet: Heat exchanger network"
    for hx in Status.int.hx:
        print "HX %s id %s"%(hx.name, hx.id)
        if len(hx.sourcestreams) > 0: print "BuildHXNet: source %s outletHX %s"%(hx.sourcestreams, hx.sourcestreams[0].outletHX)
        if len(hx.sinkstreams) > 0: print "BuildHXNet: sink %s outletHX %s"%(hx.sinkstreams, hx.sinkstreams[0].outletHX)
        print

    for s, splitter in enumerate(Status.int.sourceSplitters):
        print "Source-Splitter %s: %s"%(s, splitter)
        print "mainstream %s"%splitter.mainStream

    for s, splitter in enumerate(Status.int.sinkSplitters):
        print "Sink-Splitter %s: %s"%(s, splitter)
        print "mainstream %s"%splitter.mainStream

    if Status.SimTool == "Plus":
        for h, hx in enumerate(Status.int.hx):
            print "heatExchanger.buildHXNet HX %s srcRD %r snkRD %r"%(h+1, hx.srcRD, hx.snkRD)
    print "=====================================\n"

def clearHXResults(updateFlags = True):
    """ Clears the results of the HX Network
    (needed before a new iteration is activated, as the function storeResults
    in the heatExchanger object ADDS to those arrays
    
    """
    # general call to clearHXResults outside of buildHXNet, when only the arrays should be emptied again (-> HX.calculate writes with "add")
    Status.int.QHXProc_Tt = [HeatFlow_t() for dummy in range(Status.NThProc)]
    Status.int.UPHwUsed_Tt = [HeatFlow_t() for dummy in range(Status.NThProc)]
    Status.int.QHXEq_Tt = [HeatFlow_t() for dummy in range(Status.NEquipe)]
    Status.int.QWHEqUsed_Tt = [HeatFlow_t() for dummy in range(Status.NEquipe)]
    Status.int.QHXPipe_Tt = [HeatFlow_t() for dummy in range(Status.NPipeDuct)]
    Status.int.QWHPipeUsed_Tt = [HeatFlow_t() for dummy in range(Status.NPipeDuct)]
    Status.int.QWHEEUsed_Tt = [HeatFlow_t() for dummy in range(Status.NWHEE)]
    Status.int.QHX_Tt = [HeatFlow_t() for dummy in range(Status.NHX)]

    if updateFlags: # specific functions when entering a new iteration of the outer cycle (after recalculation of other objects)
        Status.int.QHXProcChanged = [False for dummy in Status.int.QHXProc_Tt]
        Status.int.UPHwUsedChanged = [False for dummy in Status.int.UPHwUsed_Tt]
        Status.int.QHXEqChanged = [False]*Status.NEquipe
        Status.int.QWHEqUsedChanged = [False]*Status.NEquipe
        Status.int.QHXPipeChanged = [False for dummy in Status.int.QHXPipe_Tt]
        Status.int.QWHPipeUsedChanged = [False for dummy in Status.int.QWHPipeUsed_Tt]
        Status.int.QWHEEUsedChanged = [False for dummy in Status.int.QWHEEUsed_Tt]
        Status.int.QHXChanged = [False] * Status.NHX

        #Note: This in reality is only needed only once at build-up of the project in order to allocate the arrays
        Status.int.QHXRes_Tt = [HeatFlow_t() for dummy in range(Status.NHX)]
        Status.int.QWHRes_Tt = [HeatFlow_t() for dummy in range(Status.NHX)]


def getPrimarySink(objectNo, split=False):
    """ Gets the primary sink of the energy in case of serial connections

    split = True: the splitter is returned instead of the primary object
    
    """

    try:
##        h = Status.int.hxIDdict[objectNo]
        h = objectNo
        hx = Status.int.hx[h]
    except:
        logError("heatExchanger.getPrimarySink: HX with no %s does not exist"%(objectNo))
        return None
    
    if len(hx.sinkstreams) > 0:
        sink = hx.sinkstreams[0]
        if split and sink.splitterNo is not None:
            splitSink = copy(sink)
            splitSink.objectType = "SPLIT"
            splitSink.objectNo = sink.splitterNo
            splitSink.componentNo = sink.branch
            splitSink.fraction = 1.0
            return splitSink
        elif sink.objectType == "HX":
            primarySink = getPrimarySink(sink.objectNo, split) #recursive call of function
            if primarySink is not None:
                primarySink.outletHX = Status.int.hx[sink.objectNo].id
            return primarySink
        elif sink.objectType == "SPLIT":
            if split:
                return copy(sink)
            else:
                splitter = Status.int.sinkSplitters[sink.objectNo]
                primarySink = copy(getPrimarySinkOfConn(splitter.mainStream, split))
                return primarySink
        else:
            return copy(sink)
    else:
        return None

def getPrimarySinkOfConn(conn, split=False):
    """ Gets the primary sink of the energy in case of serial connections

    :params:

    conn: HCConnection object

    split = True: the splitter is returned instead of the primary object
    
    """

    if split and conn.splitterNo is not None:
        splitSink = copy(conn)
        splitSink.objectType = "SPLIT"
        splitSink.objectNo = conn.splitterNo
        splitSink.componentNo = conn.branch
        splitSink.fraction = 1.0
        return splitSink
    elif conn.objectType == "HX":
        primarySink = copy(getPrimarySink(conn.objectNo, split))
        if primarySink is not None: # can happen in the case of unconnected upstream HXes
            primarySink.outletHX = Status.int.hx[conn.objectNo].id
            primarySink.fraction = conn.fraction
        return primarySink #recursive call of function
    elif conn.objectType == "SPLIT":
        if split:
            return conn
        else:
            splitter = Status.int.sinkSplitters[conn.objectNo]
            primarySink = copy(getPrimarySinkOfConn(splitter.mainStream, split))
            primarySink.fraction = splitter.fractions[conn.componentNo]*primarySink.fraction*conn.fraction
            return primarySink
    else:
        return conn
    
def getPrimarySource(objectNo, split=False):
    """ Gets the primary source of a heat exchanger in case of serial connections
    :params objectNo: the number of the heat exchanger (= objectNo in connections to HX)

    split=True: breaks if a splitter is detected
    split=False: looks for the original source
    """

    try:
        h = objectNo
##        h = Status.int.hxIDdict[objectNo]
        hx = Status.int.hx[h]
    except:
        logError("heatExchanger.getPrimarySource: HX with ID %s does not exist"%(objectNo))
        return None
    
    if len(hx.sourcestreams) > 0:
        source = hx.sourcestreams[0]
        if split and source.splitterNo is not None:
            splitSource = copy(source)
            splitSource.objectType = "SPLIT"
            splitSource.objectNo = source.splitterNo
            splitSource.componentNo = source.branch
            splitSource.fraction = 1.0
            return splitSource
        elif source.objectType == "HX":
            primarySource = getPrimarySource(source.objectNo, split)
            if primarySource is not None:
                primarySource.outletHX = Status.int.hx[source.objectNo].id
            return primarySource #recursive call of function
        elif source.objectType == "SPLIT":
            if split:
                return copy(source)
            else:
                splitter = Status.int.sourceSplitters[source.objectNo]
                primarySource = copy(getPrimarySourceOfConn(splitter.mainStream, split))
            return primarySource
        else:
            return copy(source)
    else:
        return None

def getPrimarySourceOfConn(conn, split=False):
    """ Gets the primary source of the energy in case of serial connections """

    if split and conn.splitterNo is not None:
        splitSource = copy(conn)
        splitSource.objectType = "SPLIT"
        splitSource.objectNo = conn.splitterNo
        splitSource.componentNo = conn.branch
        splitSource.fraction = 1.0
        return splitSource
    elif conn.objectType == "HX":
        primarySource = copy(getPrimarySource(conn.objectNo, split))
        if primarySource is not None: # can happen in the case of unconnected upstream HXes
            primarySource.outletHX = Status.int.hx[conn.objectNo].id
            primarySource.fraction = conn.fraction
        return primarySource #recursive call of function
    elif conn.objectType == "SPLIT":
        if split:
            return conn
        else:
            splitter = Status.int.sourceSplitters[conn.objectNo]
            primarySource = copy(getPrimarySourceOfConn(splitter.mainStream, split))
            primarySource.fraction = splitter.fractions[conn.componentNo]*primarySource.fraction*conn.fraction
        return primarySource
    else:
        return conn
    
def addSourceSplit(conn, fluidId):

    thisSplitter = None
    for s, splitter in enumerate(Status.int.sourceSplitters):
        if ((splitter.mainStream.objectType == conn.objectType) and
            (splitter.mainStream.objectNo == conn.objectNo) and
            (splitter.mainStream.componentNo == conn.componentNo)):

            thisSplitter = splitter
            splitterNo = s
            break

    if thisSplitter is None: #create a new one
        splitterNo = len(Status.int.sourceSplitters)
        thisSplitter = Splitter(conn, splitterNo, sourceOrSink = "Source", fluidId = fluidId)
        Status.int.sourceSplitters.append(thisSplitter)
        branch = 1
        
    else:
        branch = thisSplitter.addBranch(conn)

##    print "addSourceSplit %s: %s%s-%s"%(splitterNo, conn.objectType, conn.objectNo, conn.componentNo)

    return splitterNo, branch
        

def addSinkSplit(conn, fluidId):

    thisSplitter = None
    for s, splitter in enumerate(Status.int.sinkSplitters):
        if ((splitter.mainStream.objectType == conn.objectType) and
            (splitter.mainStream.objectNo == conn.objectNo) and
            (splitter.mainStream.componentNo == conn.componentNo)):

            thisSplitter = splitter
            splitterNo = s
            break

    if thisSplitter is None: #create a new one
        splitterNo = len(Status.int.sinkSplitters)
        thisSplitter = Splitter(conn, splitterNo, sourceOrSink = "Sink", fluidId = fluidId)
        Status.int.sinkSplitters.append(thisSplitter)
        branch = 1
        
    else:
        branch = thisSplitter.addBranch(conn)

    return splitterNo, branch

class Splitter():

    def __init__(self, conn, no = None, sourceOrSink = "Source", fluidId = 0):

##        print "Splitter.__init__: creating splitter no %s from conn %s (outletHX %s)"%(no, conn, conn.outletHX)
        self.no = no

        if no is not None:
            self.name = "Splitter %2d"%no
        else:
            self.name = "Splitter --"

        self.mainStream = copy(conn)
        self.mainStream.HXId = None
        self.mainStream.fraction = 1.0
        self.fractions = [1.0]
        self.residual = True        # indicates whether there is a residual fraction which has not been explicitly allocated by addBranch
                                    # in this case branch 0 corresponds to the residual
        self.hxes = [0]
        self.fluidId = fluidId

        self.sourceOrSink = sourceOrSink

        self.addBranch(conn)

    def getMainStreamRD(self):
        conn = self.mainStream      #TODO: -> probably everything works fine just taking self.mainStream. But for double security for the moment the original is stored
        if self.sourceOrSink == "Source":
            self.mainStreamRD = getPrimarySourceOfConn(conn, split=True)
        else:
            self.mainStreamRD = getPrimarySinkOfConn(conn, split=True)
        return self.mainStreamRD

        

    def addBranch(self, conn):
        if self.fractions[0] < conn.fraction + EPSILON: # if the fraction of the new branch corresponds to the residual, then use the 0 branch instead of creating a new one
            branch = 0
            self.residual = False
        else:
            self.fractions[0] -= conn.fraction
            self.fractions.append(conn.fraction)
            self.N = len(self.fractions)
            branch = self.N-1
            
        self.hxes.append(conn.HXId)

        conn.splitterNo = self.no
        conn.branch = branch

##        print "Splitter.addBranch: %r"%self

        return branch

    def checkFractions(self, autoCorrect=True):
        """ checks that the sum of fractions of the splitter is 1
        """

        dif = sum(self.fractions) - 1.0
        for i, f in enumerate(self.fractions):
            if abs(dif) > EPSILON:
                if not autoCorrect:
                    return False
                else:
                    dif_i = min(dif, self.fractions[i])
                    self.fractions[i] - dif_i
                    dif -= dif_i
            else:
                break
        

        
    def __repr__(self):
        string = "Splitter no: %s \n"%self.no   
        string += "[oType: %s oNo %s cNo %s]\n"%(self.mainStream.objectType, self.mainStream.objectNo, self.mainStream.componentNo)
        string += "[fractions: %s]\n"%(self.fractions)
        return string
        

def nodeCode(conn):
    """
    helper function for eliminating branch no. from the node-code of splitters
    """
    if conn is None:
        return None
    objectCode = conn.compact()
    if objectCode[0]=="SPLIT":
        code = list(objectCode)
        code[2] = -1
        code = tuple(code)
        return code
    else:
        return objectCode
    
    
