# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   ModuleHR (Heat Recovery)
#           
#------------------------------------------------------------------------------
#           
#   Module for heat recovery calculations
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.dialogGauge import DialogGauge
from GUI.status import *
from auxiliary.auxiliary import noneFilter
from constants import *
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U

from modules.interfaces import calcQ_T, createQ_Tt
from modules.messageLogger import logDebug, logTrack, logError, logWarning
from modules.modulePipes import calculateUSHCPipe
from modules.heatFlow import HeatFlow_t
from hxConnection import buildHXNet, clearHXResults, getPrimarySourceOfConn, getPrimarySinkOfConn
from heatExchanger import CONVERGENCESERIALHXLOOP
from GUI.GUITools import check
import numpy as np
from designHXNet import DesignHXNet
from pinchAnalysis import PinchAnalysis
from modules.energyStats.modulePA2 import ModulePA2
from constants import HXDESIGNSTRATEGIES
import database.dbio as dbio

MAXITSSERIALHXLOOP = 20
DEFAULT_DTMIN = 5.0

class ModuleHR(object):

    def __init__(self, keys):
        self.keys = keys # the key to the data is sent by the panel
        self.ConCondensation = False
        self.HiddenHX = []
        self.data = None
        self.iscalculated = 0
        self.curvecalc = None
        self.HXPFlag = False

        self.pa = None
        self.DTmin = DEFAULT_DTMIN
        self.qhxTarget = None

        self.strategy = HXDESIGNSTRATEGIES[1]
                     
#----------------------------------------------------------------------------------------------------
# GUI Interaction
#----------------------------------------------------------------------------------------------------
    def initPanel(self):                      
        pass
    
    def updatePanel(self):
        """Update all data used in the GUI PanelHR."""

        logTrack("ModuleHR.updatePanel")

        Status.int.systemModel.build()
        Status.int.systemModel.update(requireConverged=False)               

        hxTable = []
        for hx in Status.int.hx:
            VStorage = hx.MCStorage
            if VStorage is not None:
                VStorage /= (4.19/3.6)
            else:
                VStorage = 0.0
                
            row = [U(hx.name),
                   hx.dotQNom,
                   VStorage]
            if len(hx.sourcestreams) > 0:
                source = hx.sourcestreams[0]
##                    source = source.compactConnToUpstreamComponent()
                source = getPrimarySourceOfConn(source)
                row += [U(source.pinchStreamName),
                        hx.ThoAvg,
                        hx.ThiAvg]
            else:
                row += ["---", "", ""]

            hxTable.append(row)

            qhx = hx.QHX
            if qhx is not None:
                qhx_MWh = qhx/1000
            else:
                qhx_MWh = "-"
                
            row = ["",qhx_MWh,""]
            
            if len(hx.sinkstreams) > 0:
                sink = hx.sinkstreams[0]
##                    sink = sink.compactConnToUpstreamComponent()
                sink = getPrimarySinkOfConn(sink)
                row += [U(sink.pinchStreamName),
                        hx.TciAvg,
                        hx.TcoAvg]
            else:
                row += ["---", "", ""]
            hxTable.append(row)


        Status.int.setGraphicsData(self.keys[0], hxTable)

        qhxTotal = 0
        for hx in Status.int.hx:
            if hx.QHX is not None:
                qhxTotal += hx.QHX
        self.qhxTotal = qhxTotal

        try:
            effHXNet = self.qhxTotal / self.qhxTarget
        except:
            effHXNet = None
        
        info = {"Strategy": self.strategy,
                "DTmin": self.DTmin,
                "QHX": self.qhxTotal,
                "QHXTarget": self.qhxTarget,
                "EffHXNet": effHXNet,
                "dQHXdQdot": None,
                "fQHX": None}

        Status.int.setGraphicsData(self.keys[1], info)
                
                            
                       
    def getDAConfig(self):
        """ reads in the design assistant parameters from the GData array
        """
        try:
            data = Status.int.GData[self.keys[1]]
        except:
            return

        self.strategy = data["Strategy"]
        newDTmin = data["DTmin"]
        if newDTmin > 0 and newDTmin < 1000.:
            self.DTmin = newDTmin
        else:
            logError(_("Erroneous value for minimum temperature difference"))
            
        self.fQHX = data["fQHX"]
        self.dQHXdQdot = data["dQHXdQdot"]


    def updateReportData(self, includePinch = True):
        """Collect the data needed for the master report and write it
        into Status.int.GData under the appropriate keys.

        :params includePinch: if True -> print also yearly data
        
        """
        if includePinch and Status.ANo > 0:
            Status.mod.moduleEnergy.runSimulation(outputLevel=3)    # outputLevel 3 required in plus version for storing QA_T
        Status.mod.moduleEA.update()     # assures that yearly values are up to date       
        
        self.initPanel() # trigger curve data generation for plots
        
        hexers = Status.prj.getHXes()
        qTotal = 0.0
        qdotTotal = 0.0
        for hx in hexers:
            q = hx.QHX
            if q is not None:
                qTotal += q

            qdot = hx.QdotHX
            if qdot is not None:
                qdotTotal += qdot

        dataListReport = []  
        index = 0
        
        for hx in hexers:
            
            q = hx.QHX
            if qTotal > 0 and q is not None: qhxperc = 100.0 * q / qTotal
            else: qhxperc = "---"

            if q is not None:
                qhx_MWh = q / 1000.
            else:
                qhx_MWh = 0.0
            row = [hx.HXName, hx.QdotHX, hx.HXSource, hx.HXSink, qhx_MWh, qhxperc]
            if index < 20:
                dataListReport.append(noneFilter(row))
            elif index == 20:
                logDebug("More than 20 HX in the system. Do not fit into the report")
            index += 1

        for i in range(index, 20):
            row = [" ", " ", " ", " ", " ", " "]
            dataListReport.append(row)

        row = [_("Total"), qdotTotal, " ", " ", qTotal / 1000.0, 100.0]
        dataListReport.append(row)
        
        dataReport = np.array(dataListReport)

        key = "HX%02d" % Status.ANo
#            print "%s\n"%key,dataReport
        Status.int.setGraphicsData(key, dataReport)
        if Status.ANo == Status.FinalAlternative:
            Status.int.setGraphicsData("HX_F", dataReport)
            Status.int.setGraphicsData("HX_SECOND_F", dataReport)

        if includePinch:

            #---------------------------------------------------------------------------
            #REPORT GRAPHS
            
            key = "HX%02d_PLOT_REPORT" % Status.ANo
            dataReport2 = [["CCC", "T-CCC", "HCC", "T-HCC", "QD", "QA"]]
            # the data transfer to OpenOffice spreadsheets doesn't like arrays with
            # numbers only

            CCC_X_index = 0
            CCC_Y_index = 1
            HCC_X_index = 2
            HCC_Y_index = 3
            QD_index = 4
            QA_index = 5
            MAXROWS = 81
            
            # create empty space
            for i in range(0, MAXROWS):
                row = ["", "", "", "", "", ""]
                dataReport2.append(row)

            QD_T =  Status.int.QD_T if Status.SimTool=="NP" else Status.int.USHTotal_T

            if Status.SimTool=="NP":
                QA_T = Status.int.QA_T
            else:
                QA_T = [sum(row) for row in Status.int.QA_Tt_mod[0]]

            keys = ['HR Table','HR Curves', "PA2"]

            modPA2 = ModulePA2(keys)
            modPA2.updatePanel()
            data = Status.int.GData["PA2"]
            ccc = data[0]
            hcc = data[1]
                
            if data is not None:
                # TODO: The string conversion is necessary here for TC001_Report
                #       or in general for very small or very large numbers in scientific representation
                
                entryCount = len(ccc.X)
                ncc = min(MAXROWS, entryCount)
                for i in range(ncc):            
                    dataReport2[i + 1][CCC_X_index] = "%s"%ccc.X[i]
                    dataReport2[i + 1][CCC_Y_index] = "%s"%ccc.Y[i]
    
                for i in range(ncc, MAXROWS):
                    dataReport2[i + 1][CCC_X_index] = "%s"%ccc.X[ncc - 1]
                    dataReport2[i + 1][CCC_Y_index] = "%s"%ccc.Y[ncc - 1]
                   
                entryCount = len(hcc.X)
                ncc = min(MAXROWS, entryCount)
                for i in range(ncc):            
                    dataReport2[i + 1][HCC_X_index] = "%s"%hcc.X[i]
                    dataReport2[i + 1][HCC_Y_index] = "%s"%hcc.Y[i]
    
                for i in range(ncc, MAXROWS):
                    dataReport2[i + 1][HCC_X_index] = "%s"%hcc.X[ncc - 1]
                    dataReport2[i + 1][HCC_Y_index] = "%s"%hcc.Y[ncc - 1]

                for i in range(MAXROWS):
                    dataReport2[i + 1][QD_index] = "%s"%(QD_T[i] / 1000.0)
                    dataReport2[i + 1][QA_index] = "%s"%(QA_T[i] / 1000.0)

                Status.int.setGraphicsData(key, np.array(dataReport2))

            else:
                logWarning("ModuleHR.updateReportData: cannot generate pinch curves. No data available")

        
#----------------------------------------------------------------------------------------------------
# Public Methods
#----------------------------------------------------------------------------------------------------   
                         
    def runHRDesign(self, exhx=True):

        # this is a hack in order to continue working with test results for TC1.01.
        try:
            if Status.skipHRDesign==True: return
        except:
            pass
        # -------------------------------------------------------------------------
        
        if Status.StatusEnergy == 0 or Status.int.cascadeUpdateLevel < len(Status.int.USHj):
            Status.mod.moduleEnergy.runSimulation() #require fully converge results

        self.getDAConfig()

        if DEBUG != "OFF":
            print "ModuleHR.runHRDesign: carrying out auto-design with DT = %s / Strategy = %s"%(self.DTmin, self.strategy)

        logTrack("ModuleHR.runHRDesign: carrying out auto-design with DT = %s / Strategy = %s"%(self.DTmin, self.strategy))
            
        self.pa = PinchAnalysis()
        self.pa.calculate(self.DTmin)
        self.pa.exportData()

        dlg = DialogGauge(Status.main, _("EINSTEIN HX auto-design"), _("calculating"))
        designAssistant = DesignHXNet(self.pa, strategy=self.strategy)
        designAssistant.calculate()
        Status.int.changeInHeatRecovery()
        Status.int.systemModel.build()
        dlg.Destroy()


    def runHRModule(self, exhx=True, HXPFlag = False):

        """Main function that carries out simulation of heat exchanger network

        :param exhx: ??? (only for "calculate heat exchanger" - mode)
        :param HXPFlag: ??? (only for "calculate heat exchanger" - mode)

        :returns QHXProc_Tt,QHXEq_Tt,QHXPipe_Tt,UPHwUsed_Tt,QWHPipeUsed_Tt,QWHEqUsed_Tt
        """
        dlg = DialogGauge(Status.main, _U("EINSTEIN heat recovery module"), _U("calculating"))
        if Status.processData.outOfDate == True:
            dlg.Destroy()
            Status.processData.createAggregateDemand()
            return  # createAggregateDemand calls the heat recovery calculation 
                    # itself.

        clearHXResults()
               
        if Status.HRTool == HRMODES[1]:
            estimateHR()
            
        elif Status.HRTool == HRMODES[2]:

            if Status.int.heatRecoveryOutOfDate == True:
                buildHXNet()    #vBuild HX network again, if some change has been introduced

            for i in range(MAXITSSERIALHXLOOP):
                clearHXResults(updateFlags=False)
                maxDif = 0
                for h,hx in enumerate(Status.int.hx):
                    dif = hx.calculate()
                    maxDif = max(dif,maxDif)
                    if i ==0: dlg.update(80*(h+1)/len(Status.int.hx))
                    logTrack("ModuleHR.runHRModule max of dif in HX loop %s"%maxDif)

                dlg.update(80 + 10.0*(i+1)/MAXITSSERIALHXLOOP)
                    
                if maxDif < CONVERGENCESERIALHXLOOP:
                    dlg.update(90)
                    break
                        
        dlg.update(90.0)
        
        calculateUSHCPipe()
        if Status.HRTool != HRMODES[1]: self.calculateResidualQWHAmb()
        
        dlg.update(99.0)
        
        self.__storeResults()
        dlg.Destroy()

    def calculateResidualQWHAmb(self):

        """Post processing of HR results: calculates residual waste heat not used in HR system
            not used in estimate mode, as calculations there are done directly.

        :param Status.int.UPHwResTotal_Tt:  Total available UPHw (residual part not including MCD)
        :type  Status.int.UPHwResTotal_Tt:  HeatFlow_t

        :param Status.int.UPHwUsedResTotal_Tt:  Residual part of UPHw (total) not including MCD, used in heat recovery
        :type  Status.int.UPHwUsedResTotal_Tt:  HeatFlow_t

        :param Status.int.UPHwUsed_Tt:      UPHw of a process used in heat recovery
        :type  Status.int.UPHwUsed_Tt:      List of HeatFlow_t

        :param Status.int.QWHEqTotal_Tt:    Total available waste heat from equipments
        :type  Status.int.QWHEqTotal_Tt:    Q_Tt - array, descending

        :param Status.int.QWHEqUsed_Tt:     Waste heat of equipment used in heat recovery
        :type  Status.int.QWHEqUsed_Tt:     List of HeatFlow_t

        :param Status.int.QWHEE_Tt:         Total available waste heat from non-thermal equipments
        :type  Status.int.QWHEE_Tt:         Q_Tt - array, descending

        :param Status.int.QWHEEUsed_Tt:     Available waste heat from non-thermal equipments used in heat recovery
        :type  Status.int.QWHEEUsed_Tt:     List of HeatFlow_t

        :returns Status.int.QWHAmb_Tt:      Remaining waste heat dissipated to ambient (not used in heat recovery system)
        :type  Status.int.QWHAmb_tt:        Q_Tt - array, descending
        
        """

        UPHwResTotal_Tt = Status.int.UPHwResTotal_Tt
        QWH_Tt = UPHwResTotal_Tt.Q_Tt(descending=True)      
        QWH_Tt[:,:] += (Status.int.QWHEqTotal_Tt[:,:] + 
                        Status.int.QWHEE_Tt[:,:])

        QWHUsed_hf = HeatFlow_t()
        for k, item in enumerate(Status.int.UPHwUsed_Tt):
##            if Status.int.UPHwReq_Tt_hf[k].Q ==0:
            QWHUsed_hf.add(item)
        QWHUsed_hf.add(Status.int.UPHwUsedResTotal_Tt)  # Adding residual part of UPHwUsed for processes with MCD

        for m,item in enumerate(Status.int.QWHPipeUsed_Tt):
            QWHUsed_hf.add(item)

        for j,item in enumerate(Status.int.QWHEqUsed_Tt):
            QWHUsed_hf.add(item)

        for n,item in enumerate(Status.int.QWHEEUsed_Tt):
            QWHUsed_hf.add(item)

        QWHUsed_Tt = QWHUsed_hf.Q_Tt(descending=True)

        QWH_T = Status.int.calcQ_T(QWH_Tt)
        if DEBUG != "OFF":
            print "Total waste heat QWH_T",QWH_T

        QWHAmb_Tt = QWH_Tt - QWHUsed_Tt

        QWHAmb_T = Status.int.calcQ_T(QWHAmb_Tt)
        if DEBUG != "OFF":
            print "Residual waste heat QWHAmb_T",QWHAmb_T

        for iT in xrange(Status.NT+1):
            QWHAmb_Tt[iT+1] = np.min((QWHAmb_Tt[iT + 1], QWHAmb_Tt[iT]), 0)

        Status.int.QWHAmb_Tt = QWHAmb_Tt.clip(min=0) # remove calculation errors
                
            
    def __storeResults(self,QHXProc_Tt = None,QWHAmb_Tt = None):
        """Store the results of the heat recovery estimation in the 
        appropriate Status variables and calculate missing ones where
        necessary.
        
        """
        heatingPipeIndices = [m for m, pipe in enumerate(Status.int.systemModel.pipeObject) if pipe.heating]
        coolingPipeIndices = [m for m, pipe in enumerate(Status.int.systemModel.pipeObject) if not pipe.heating]
        
        USH_Tt = createQ_Tt() # cumulative heat demand at entry of pipes
        USC_Tt = createQ_Tt() # cumulative cooling demand at entry of pipes
        for m in heatingPipeIndices:
            USH_Tt += Status.int.USHCPipe_Tt[m]
        for m in coolingPipeIndices:
            USC_Tt += Status.int.USHCPipe_Tt[m]
    
        Status.int.USHTotal_Tt = USH_Tt
        Status.int.USCTotal_Tt = USC_Tt
        Status.int.USHTotal_T = calcQ_T(USH_Tt)
        Status.int.USCTotal_T = calcQ_T(USC_Tt)
        
        Status.int.USHTotal = Status.int.USHTotal_T[Status.NT + 1]
        Status.int.USCTotal = Status.int.USCTotal_T[0]
    
        Status.int.QD_Tt = USH_Tt.copy()
        Status.int.QDc_Tt = USC_Tt.copy()
        Status.int.QD_T = calcQ_T(Status.int.QD_Tt)
        Status.int.QDc_T = calcQ_T(Status.int.QDc_Tt)
        
        # settings of the conversion UPH -> USH
        if QWHAmb_Tt is not None:
            Status.int.QWHAmb_Tt = QWHAmb_Tt
            Status.int.QWHAmb_T = calcQ_T(QWHAmb_Tt)
            # copying USH/QWHAmb to QD/QA copied into runHRModule

        try:
            Status.int.QA_Tt = Status.int.QWHAmb_Tt.copy()
        except:
            Status.int.QA_Tt = Status.int.createQ_Tt()
            logWarning("ModuleHR._storeResults: ERROR: could not create QA in Status.int. Zero arrays created in order to continue processing")
        Status.int.QA_T = calcQ_T(Status.int.QA_Tt)
        
        try:
            Status.int.QA_Tt_mod[0] = Status.int.QA_Tt.copy()
            Status.int.QA_T_mod[0] = Status.int.QA_T.copy()
        except IndexError:
            pass # no cascade initialised yet

        if QHXProc_Tt is not None:
            Status.int.QHXProc_Tt = QHXProc_Tt

        #TO DO: somewhere the QHXProc_Ts etc. should be calculated for completeness !                   

        if DEBUG in ("BASIC", "MAIN", "ALL"):
            print "ModuleHR (__storeResults) at the end"
            print "QA_T",Status.int.QA_T
        processes = Status.prj.getProcesses()
        try:
            for k, process in enumerate(processes):
                process.QHXProc = dbio.check(Status.int.QHXProc_Tt[k].Q)
        except: # in estimate mode QHXProc comes as old-format Q_Tt array
            for k, process in enumerate(processes):
                index = (Status.NT+1)*(process.heating_cooling == "heating")
                process.QHXProc = float(np.sum(Status.int.QHXProc_Tt[k][index]))
            
        Status.int.heatRecoveryOutOfDate = False
  
