# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#    HeatFlowQ_T
#            
#------------------------------------------------------------------------------
#            
#   auxiliary functions for managing Q_T tables
#
#==============================================================================
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from GUI.status import *
import numpy as np
from modules.messageLogger import logTrack, logWarning, logDebug
from constants import TemperatureInterval, TemperatureIntervalC, MinimumTemperatureC

def getTIndex(temp, heating=True):
    if heating:
        return temp/TemperatureInterval
    else:
        return (temp - MinimumTemperatureC)/TemperatureIntervalC

def createQ_T():
    return np.zeros(Status.NT+2)

def shiftQ_T(Q_T, DT, mode):
    """ shifts a Q_T - array by DT in temperature

    :param mode: heating or cooling (defines temperature scale and 
        ascending/descending

    """
    if DT == 0:
        return Q_T
    else:
        if mode == "heating":
            if DT < 0:
                logWarning("shiftQ_T: negative temperature shift for heating demands not possible")
                return Q_T
            else:
                diT = int(DT/TemperatureInterval)
                xiT = DT/TemperatureInterval - diT #indicates the fraction to be added at diT + 1

                Qnew0_T = Status.int.createQ_T()
                Qnew0_T[:diT] = Q_T[0]
                Qnew0_T[diT:-1] = Q_T[:-diT - 1]
                Qnew0_T[-1] = Q_T[-1]
                
                if xiT == 0:
                    return Qnew0_T
                else:    
                    Qnew1_T = Status.int.createQ_T()
                    Qnew1_T[:diT + 1] = Q_T[0]
                    Qnew1_T[diT + 1:-1] = Q_T[:-diT - 2]
                    Qnew1_T[-1] = Q_T[-1]
                    
                    return xiT*Qnew1_T + (1-xiT)*Qnew0_T

        else:
            if DT > 0:
                logWarning("shiftQ_T: positive temperature shift for cooling demands not possible")
                return Q_T
            else:
                diT = int(-DT/TemperatureIntervalC)
                xiT = -DT/TemperatureIntervalC - diT

                Qnew0_T = Status.int.createQ_T()
                Qnew0_T[0] = Q_T[0]
                if diT > 0:
                    Qnew0_T[1:-diT] = Q_T[diT + 1:]
                    Qnew0_T[-diT:] = Q_T[-1]
                else:
                    Qnew0_T[1:] = Q_T[1:]
                
                if xiT == 0:
                    return Qnew0_T
                else:    
                    Qnew1_T = Status.int.createQ_T()
                    Qnew1_T[0] = Q_T[0]
                    Qnew1_T[1:-diT-1] = Q_T[diT+2:]
                    for iT in range(diT+1):
                        Qnew1_T[-iT-1] = Q_T[-1]
                    return xiT*Qnew1_T + (1-xiT)*Qnew0_T

def match_heatFlows(Qsrc, Qsnk):

    """
    auxiliary function distributing an available heat supply (QS) to a given demand (QD)
    looking for maximum use of available exergy

    :param Qsrc: Heat source
    :type  Qsrc: List of size NT+2
    :param Qsnk: Heat sink
    :type  Qsnk: List of size NT+2
    
    :returns: QS: heat supplied to sink; QA: remaining heat available from source
    :rtype:   Tuple - 2x list of size NT+2

    the following equations hold:
    - source temperature level:
            Qsrc(Tsrc) = QA(Tsrc) + QSsrc(Tsrc)
    - sink temperature level
            Qsnk(Tsnk) = QS(Tsnk) + Qsnk,res(Tsnk)

    where QSsrc(T) <= QS(T) for all T, so that Tsrc(dQ) >= Tsnk(dQ) for all dQ
    """
    
    QS = Status.int.createQ_T()
    QA = Status.int.createQ_T()

    if Qsrc[Status.NT+1]==0:    # no exchange possible
        return (QS,QA)
    
    dQS = QS   # use of same space, no identity !
    dQsrc = QA # use of same space, no identity !
    dQsnk = QS # use of same space, no identity !

    dQ = 0

    for iT in xrange(Status.NT+1,0,-1):
        dQsrc[iT] = Qsrc[iT] - Qsrc[iT-1]
        dQ += dQsrc[iT]                 # remaining heat available at temperatures higher or equal to the current interval
        dQsnk = Qsnk[iT] - Qsnk[iT-1]   # heat demanded at current temperature interval
        dQS[iT] = min(dQsnk,dQ)         # heat supplied at current interval
        dQ -= dQS[iT]                   # reduce amount of remaining heat
        

    dQ = 0
    QA[0] = 0
    for iT in xrange(1,Status.NT+2):
        dQ += dQS[iT]                   # heat supplied at temperatures lower or equal than current interval
        dX = min(dQsrc[iT],dQ)          # heat extracted at current interval
        dQ -= dX                        # remaining heat supplied to be covered from higher temperature intervals
        QA[iT] = QA[iT-1] + dQsrc[iT] - dX
        
        QS[iT] = QS[iT-1] + dQS[iT]

    for iT in xrange(Status.NT+2):      # QA as descending array
        QA[iT] = QA[Status.NT+1]-QA[iT]

    return (QS,QA)

