# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	INTERFACES
#			
#------------------------------------------------------------------------------
#			
#	Definition of interfaces in between modules
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import numpy as np
from GUI.status import Status
import GUI.HelperClass as HelperClass
from constants import *
from modules.messageLogger import *
import copy
from modules.heatFlow import HeatFlow_t
import database.dbio as dbio

QUERY = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s' ORDER BY IndexNo ASC"

def createQ_Tt():
    """Return a numpy 2-d array for temperature and time dependent 
    energy flows. The array is filled with 0.0 entries.
    
    """
    return np.zeros((Status.NT + 2, Status.Nt))
    
def createQ_t():
    """Return a numpy 1-d array for time dependent energy flows.
    The array is filled with 0.0 entries.
    
    """
    return np.zeros((Status.Nt))
    
def createQ_T():
    """ Return a numpy 1-d array for temperature dependent energy 
    flows. The array is filled with 0.0 entries.
    
    """
    return np.zeros((Status.NT + 2))
    
def calcQ_T(Q_Tt):
    """Return the the annual integral of Q_Tt as a 1-d numpy array."""
    return np.sum(Q_Tt[:, 0:Status.Nt], 1)

class Interfaces(object):

    systemModel = None #TODO: place for system model should be out of the Interfaces object
    pa = None

# DATA BLOCK 1: energy flows in the system
    # temperature steps for heating
    T = np.array([iT * Status.TemperatureInterval for iT in xrange(Status.NT + 1)] + [999.0])
    # temperature steps for cooling
    TC = np.array([Status.MinimumTemperatureC + (i * Status.TemperatureIntervalC) for i in xrange(Status.NT+1)] + [999.0])

    # total heat/cold demand and availability in the system
    QD_T = createQ_T()   # aggregate heat demand as seen by the heat supply
    QDc_T = createQ_T()  # aggregate cooling demand as seen by the heat supply
    QA_T = createQ_T()   # equipment. similar (identical ?) to USHm_T
    QD_Tt = createQ_Tt()
    QDc_Tt = createQ_Tt()
    QA_Tt = createQ_Tt()

    # intermediate heat/cold demand / availability within equipment cascade
    QD_T_mod = []
    QA_T_mod = []
    QD_Tt_mod = []
    QA_Tt_mod = []

    # UPH by processes
    # Temperature dependent demand is calculated in createYearlyDemand
    UPX = []
    UPXProck = []
    QHXProc = []
    UPC = []
    UPCProck = []
    QCXProc = []
    
    UPH_T = []  # calculated in createYearlyDemand  
    UPHc_T = []
    UPHm_T = []
    UPHs_T = []
    UPHw_T = []
    UPHwReq_T = []
    UPH_Tt = []
    UPHw_Tt = []
    UPHwReq_Tt = []

    UPHTotal_T = []
    UPHTotal_Tt = []
    UPCTotal_Tt = []
    UPH_s_t = {}        #dictionaries for HR Module. TODO: convert HR module for use of  conventional arrays
    UPH_c_t = {}
    UPH_m_t = {}
    UPH_w_t = {}

    UPHs_Tt = []        # HeatFlow_t format
    UPHm_Tt = []        # HeatFlow_t format
    UPHc_Tt = []        # HeatFlow_t format
    UPHw_Tt_hf = []     #once old format UPHw is eliminated, extension _hf should be eliminated
    UPHwReq_Tt_hf = []
    UPHwUsed_Tt_hf = []

    UPHwTotal_T = []
    UPHwTotal_Tt = []
    UPCwTotal_Tt = []

    UPHProc_Tt = []
    UPHProcTotal_Tt = []    # heat supplied externally to processes (= UPH - QHXProc)
    UPHProcTotal_T = createQ_T()
    UPCProcTotal_Tt = []
    UPCProcTotal_T = createQ_T()

    QHXProc_Tt = []         # heat recovered and fed directly to processes
    QHXProcTotal_Tt = []    
    QHXProcTotal_T = []
    QHXProcTotal = 0

    QWHAmb_Tt = []  # waste heat dissipated to ambient (not used)
    QWHAmb_T = []

    # Equipment energy flows
    # heat and cooling supplied by each equipment
    USHj_Tt = []
    USHj_T = []
    USHj_t = []
    USHj = []
    USHjChanged = []

    USHTotal_Tt = []
    USHTotal_T = createQ_T()
    USHTotal = 0
    USCTotal_Tt = []
    USCTotal_T = createQ_T()
    USCTotal = 0

    #special arrays only used in SimTool = Plus
    USHTotal_Lt = np.zeros((5, Status.Nt))
    USCTotal_Lt = np.zeros((5, Status.Nt))
    
    # waste heat absorbed (QHXj) and generated (QWHj) in each equipment
    QHXj_Tt = []
    QHXj_T = []
    QHXj_t = []
    QHXj = []

    QHXEq_Tt = []      #actual heat recovery to equipments (HeatFlow_t format)
    QHXEqReq_Tt = []   #required (maximum) heat recovery to equipments (HeatFlow_t format)
    QHXEqReqChanged = []
    QHXEqTotal = 0

    #TODO: eliminate doubling-up into QWHj and QWHEq !!!
    QWHj_Tt = []
    QWHj_T = []
    QWHj_t = []
    QWHj = []
    QWHjChanged = []

    # Waste Heat Flow per Equipment
    QWHEq_t = {}    #dictionary for HR Module only. TODO: -> convert HR Module so that QWHEq is read.
    QWHEq_Tt = []   #waste heat from equipments (HeatFlow_t format)
    QWHEqUsed_Tt = []   #waste heat from equipments (HeatFlow_t format) actually used in heat recovery (HeatFlow_t format)
    QWHEqChanged = []
    QWHEqUsedChanged = []

    QWHEqTotal_Tt = []    # heat recovered and fed directly to processes
    QWHEqTotal_T = []
    QWHEqTotal = 0

# Pipe energy flows

    # heat and cooling delivered by each pipe (without the added recovered heat/cold)
    USHCPipe_Tt = np.array([])
    USHCPipe_Tt_hf = []     #TODO: renamed once old format version is no longer used
    USHCPipe = []
    UPHCProcm = []

    # keep the remaining demand of the pipes during energy flow simulation, per pipe 
    USHm_remaining_Tt = []

    QHXPipe_Tt = [] #waste heat recovered - fed into pipes (HeatFlow_t format)
    QHXPipe = []
    QWHPipe_Tt = [] #waste heat from equipments (HeatFlow_t format)
    QWHPipeUsed_Tt = [] #waste heat from equipments (HeatFlow_t format)

# WHEE energy flows

    QWHEE_Tt = []
    QWHEE_T = []
    QWHEE_t = []
    QWHEE_Tt_hf = []    # one heatflow per electrical equipment
                        #TODO: rename once old format QWHEE is no longer used !!!
    QWHEEUsed_Tt = []

# Waste Heat Flow per WHEE
#    QWHEEEq_t = []     #TODO: eliminate this. correct namee is QWHEE !!!

# other required outputs from system simulation (calculateEnergyFlows)

    # other required outputs from system simulation (calculateEnergyFlows)
    FETFuel_j = []
    FETel_j = []
    ElGen_j = []
    ElGen_t = []
    FETHeat_j = []
    FETHeat_jChanged = []
    FETHeat_j_Tt = []
    HPerYearEq = []
    waterConsumption_j = []

    # dictionary of the HC supply cascade. entries "equipeID" and "equipeNo"
    NEquipe = None
    cascadeSize = None
    cascadeUpdateLevel = -1 #indicates cascadeIndex (from 1 ... N) for which cascade is
                            #updated
                            #-1 indicates that not even demand of first cascade level is available
    cascadeDistributionLevel = -1   #indicates level of equipmment after which USHm_remaining contains the HC demand
    
                                    # equipments from 1 .. NJ
                                    # 0: USHm_remaining = USHCPipe
                                    # -1: USHm_remaining not yet initialised
                                    # intermediate levels of USHm_remaining aggregated by equipment index are stored in QD_Tt_mod
    cascade = []
    EquipTableDataList = []

    USHLink = np.array([[]])    # An 2-d array containing for each pipe m and equipment j 
                                # a 1 in [m][j] if they are connected and a 0 otherwise.
                                # Where j is the equipment number - 1.

## UPHLink no longer used since V3.0.04. Substituted by systemModel.UPHLink
##    UPHLink =  np.array([[]])   # An 2-d array containing for each pipe m and process k
##                                # a 1 in [k][m] if they are connected _except_ for mandatory
##                                # cooling down, and a 0 otherwise and a 0 otherwise.
##    UPCReqLink =  np.array([[]])    # An 2-d array containing for each pipe m and process k
##                                    # a 1 in [k][m] if they are connected for mandatory
##                                    # cooling down, and a 0 otherwise 
                                                                
    heatRecoveryOutOfDate = True    # Flag should be False after a run of the heat recovery
                                    # calculations. And be set to True via 
                                    # changeInHeatRecovery() call.
    
    HXPinchConnection = []
    StreamGen = None
    hrdata = None
    
    processUPHSinkDict = {}
    processUPHSourceDict = {}
    

    QHXProc_Tt = []
    UPHwUsed_Tt = []

    QHXPipe_Tt = []
    QWHPipeUsed_Tt = []

    QHXEq_Tt = []
    QWHEqUsed_Tt = []

    QWHEE_Tt = []
    QWHEEUsed_Tt = []
    
    dT_HXNetwork = 5
    TotHeatNom = 200
    TotEnDem = 1
    
    UPH_c_t_reload = {}
    UPH_w_t_reload = {}
#..........................................................................
# DATA BLOCK 2: graphics data dictionary for graphics on panels

# DATA BLOCK 2: graphics data dictionary for graphics on panels
    GData = {}

    def __init__(self):
        pass

    def createQ_Tt(self):
        """Return a numpy 2-d array for temperature and time dependent 
        energy flows. The array is filled with 0.0 entries.
        
        """
        return createQ_Tt()
        
    def createQ_t(self):
        """Return a numpy 1-d array for time dependent energy flows.
        The array is filled with 0.0 entries.
        
        """
        return createQ_t()
        
    def createQ_T(self):
        """ Return a numpy 1-d array for temperature dependent energy 
        flows. The array is filled with 0.0 entries.
        
        """
        return createQ_T()
        
    def calcQ_T(self, Q_Tt):
        """Return the the annual integral of Q_Tt as a 1-d numpy array."""
        return calcQ_T(Q_Tt)
        

    def initCascadeArrays(self, cascadeSize):
        """Initialize storage space for energy flows in a cascade
        assigning total heat demand and availability to the first 
        row in cascade.
        
        Warning: this function changes already existing arrays.
        :param cascadeSize: number of levels the cascade shall have
        """
        logTrack("Interfaces (initCascadeArrays): creating cascade of size %s" % cascadeSize)
        if self.cascadeUpdateLevel < 0:
            logDebug("Interfaces (initCascadeArrays): demand profile not yet created")
            Status.processData.createAggregateDemand()

        self.QD_Tt_mod = [Status.int.createQ_Tt()]
        self.QD_T_mod = [Status.int.createQ_T()]    # None as list element would here be sufficient, but zero-arrays help to avoid some crashes ...
##        Status.mod.moduleEnergy.updateUSHm_remaining(0)   # -> generates QD_Tt_mod[0] from USHCPipe. copy necessary, as USHm_remaining changes dynamically
##        self.QD_T_mod[0] = self.calcQ_T(self.QD_Tt_mod[0])

        self.QDc_Tt_mod = self.QD_Tt_mod    # QDc no longer necessary, but maintained until all output models are changed using QD instead of QDc
        self.QDc_T_mod = self.QD_T_mod
        
        self.QA_Tt_mod = [self.QA_Tt.copy()]
        self.QA_T_mod = [self.QA_T.copy()]

        self.USHj_Tt = []
        self.USHj_T = []
        self.USHj_t = []

        self.QHXj_Tt = []
        self.QHXj_T = []
        self.QHXj_t = []

        self.QHXEq_Tt   = []
        self.QHXEqChanged = []
        
        self.QHXEqReq_Tt   = []

        self.QWHj_Tt = []
        self.QWHj_T = []
        self.QWHj_t = []

        self.QWHEqUsed_Tt = []
        self.QWHEq_Tt = []

        self.FETHeat_j_Tt = []
        self.FETHeat_jChanged = []

        if Status.SimTool=="NP":
            self.FETel_j_t = []
        
        # lists of annual main results
        self.USHj = []
        self.QWHj = []
        self.QHXj = []
        self.FETFuel_j = []
        self.FETel_j = []
        self.ElGen_j = []
        self.FETHeat_j = []
        self.HPerYearEq = []
        self.waterConsumption_j = []

        self.cascadeSize = 0
        self.cascadeUpdateLevel = 0
        self.cascadeDistributionLevel = -1

        self.extendCascadeArrays(cascadeSize)

    def extendCascadeArrays(self, cascadeSize):
        """
        Create storage space for energy flows in cascade
        
        This function is similar to initCascadeArrays, but does not
        change the content of already existing arrays.
 
        :param cascadeSize: minimum number of levels the cascade shall have
        
        """
##        oldSize = self.cascadeSize
        if self.cascadeSize is None:
            self.initCascadeArrays(0)
        self.addCascadeArrays(cascadeSize - self.cascadeSize)

        # special treatment for arrays that are HX-outputs and not equipment outputs
        # necessary, as those arrays are sometimes set-back to really needed length by HX-functions
        # and then not increased again when addCascadeArrays is called
        # TODO: should be solved in a "cleaner" way ...
        
        if len(self.QHXEqChanged) < cascadeSize:
            n = cascadeSize - len(self.QHXEqChanged)
            self.QHXEqChanged += [False for dummy in range(n)]
        if len(self.QHXEq_Tt) < cascadeSize:
            n = cascadeSize - len(self.QHXEq_Tt)
            self.QHXEq_Tt   += [HeatFlow_t() for dummy in range(n)]
            
##        logTrack("TEST TEST Cascade extended from size %s to size %s [check self.cs %s]. Len QHXEqChanged = %s"%(oldSize, cascadeSize, self.cascadeSize, len(self.QHXEqChanged)))
####        assert(len(self.QHXEqChanged) >= self.cascadeSize)
##        assert(cascadeSize <= self.cascadeSize)

    def addCascadeArrays(self, number=1):
        self.QD_Tt_mod += [self.createQ_Tt() for dummy in range(number)]
        self.QD_T_mod += [self.createQ_T() for dummy in range(number)]
##        self.QDc_Tt_mod += [self.createQ_Tt() for dummy in range(number)] outdated. now only a pointer to QD
##        self.QDc_T_mod += [self.createQ_T() for dummy in range(number)]
        self.QA_Tt_mod += [self.createQ_Tt() for dummy in range(number)]
        self.QA_T_mod += [self.createQ_T() for dummy in range(number)]

        self.USHj_Tt += [self.createQ_Tt() for dummy in range(number)]
        self.USHj_T += [self.createQ_T() for dummy in range(number)]
        self.USHj_t += [self.createQ_t() for dummy in range(number)]
        self.USHjChanged += [False for dummy in range(number)]

        self.QHXj_Tt += [self.createQ_Tt() for dummy in range(number)]
        self.QHXj_T += [self.createQ_T() for dummy in range(number)]
        self.QHXj_t += [self.createQ_t() for dummy in range(number)]
        self.QHXEqChanged += [False for dummy in range(number)]
        self.QHXEq_Tt   += [HeatFlow_t() for dummy in range(number)]
        self.QHXEqReq_Tt   += [HeatFlow_t() for dummy in range(number)]
        self.QHXEqReqChanged += [False for dummy in range(number)]

        self.QWHj_Tt += [self.createQ_Tt() for dummy in range(number)]
        self.QWHj_T += [self.createQ_T() for dummy in range(number)]
        self.QWHj_t += [self.createQ_t() for dummy in range(number)]
        self.QWHjChanged += [False for dummy in range(number)]

        self.QWHEq_Tt += [HeatFlow_t() for dummy in range(number)]
        self.QWHEqUsed_Tt += [HeatFlow_t() for dummy in range(number)]
        self.QWHEqChanged += [False for dummy in range(number)]

        self.FETHeat_j_Tt += [HeatFlow_t() for dummy in range(number)]
        self.FETHeat_jChanged += [False for dummy in range(number)]

        # lists of annual main results
        self.USHj += [0.0 for dummy in range(number)]
        self.QWHj += [0.0 for dummy in range(number)]
        self.QHXj += [0.0 for dummy in range(number)]
        self.FETFuel_j += [0.0 for dummy in range(number)]
        self.FETel_j += [0.0 for dummy in range(number)]
        self.ElGen_j += [0.0 for dummy in range(number)]
        self.ElGen_t += [self.createQ_t() for dummy in range(number)]
        if Status.SimTool=="NP":
            self.FETel_j_t += [self.createQ_t() for dummy in range(number)]
        self.FETHeat_j += [0.0 for dummy in range(number)]

        self.HPerYearEq += [0.0 for dummy in range(number)]
        self.waterConsumption_j += [0.0 for dummy in range(number)]

        self.cascadeSize = len(self.USHj_Tt)

    def deleteCascadeArrays(self, NEquipe):
        self.QD_Tt_mod.pop(NEquipe)
        self.QD_T_mod.pop(NEquipe)
##        self.QDc_Tt_mod.pop(NEquipe)
##        self.QDc_T_mod.pop(NEquipe)
        self.QA_Tt_mod.pop(NEquipe)
        self.QA_T_mod.pop(NEquipe)

        self.USHj_Tt.pop(NEquipe - 1)
        self.USHj_T.pop(NEquipe - 1)
        self.USHj_t.pop(NEquipe - 1)
        self.USHjChanged.pop(NEquipe - 1)

        self.QHXj_Tt.pop(NEquipe - 1)
        self.QHXj_T.pop(NEquipe - 1)
        self.QHXj_t.pop(NEquipe - 1)
        self.QHXEqChanged.pop(NEquipe - 1)

        self.QHXEq_Tt.pop(NEquipe - 1)
        self.QHXEqReq_Tt.pop(NEquipe - 1)
        self.QHXEqReqChanged.pop(NEquipe - 1)

        self.QWHj_Tt.pop(NEquipe - 1)
        self.QWHj_T.pop(NEquipe - 1)
        self.QWHj_t.pop(NEquipe - 1)
        self.QWHjChanged.pop(NEquipe - 1)

        self.QWHEq_Tt.pop(NEquipe - 1)
        self.QWHEqChanged.pop(NEquipe - 1)
        self.QWHEqUsed_Tt.pop(NEquipe - 1)

        self.FETHeat_j_Tt.pop(NEquipe - 1)
        self.FETHeat_jChanged.pop(NEquipe - 1)

        # lists of annual main results
        self.USHj.pop(NEquipe - 1)
        self.QWHj.pop(NEquipe - 1)
        self.QHXj.pop(NEquipe - 1)
        self.FETFuel_j.pop(NEquipe - 1)
        self.FETel_j.pop(NEquipe - 1)
        self.ElGen_j.pop(NEquipe - 1)
        self.FETHeat_j.pop(NEquipe - 1)
        self.HPerYearEq.pop(NEquipe - 1)
        self.waterConsumption_j.pop(NEquipe - 1)

        self.cascadeSize -= 1
##        logTrack("TEST TEST CascadeArrays deleted. remaining size = %s len USHjTt array %s"%(self.cascadeSize, len(self.USHj_Tt)))
##        assert(self.cascadeSize == len(self.USHj_Tt))

    def printCascade(self):
        NT = Status.NT
        print "Heat Demand"
        print "CascadeIndex - QD_total - QD_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            print i, \
            "%10.4f" % self.QD_T_mod[i][NT + 1], \
            self.QD_Tt_mod[i][NT + 1][0:23]
        print "Cooling Demand"
        print "CascadeIndex - QDc_total - QDc_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            print i, \
            "%10.4f" % self.QDc_T_mod[i][NT + 1], \
            self.QDc_Tt_mod[i][NT + 1][0:23]
        print "Heat Availability"
        print "CascadeIndex - QA_total - QA_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            print i, \
            "%10.4f" % self.QA_T_mod[i][0], \
            self.QA_Tt_mod[i][0][0:23]

    def printUSH(self):
        NT = Status.NT
        print "USH"
        print "CascadeIndex - USHj_total - USHj_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            print i, \
            "%10.4f" % self.USHj_T[i][NT], \
            self.USHj_Tt[i][NT][0:23]

    def printCascade_mod(self, cascade):
        NT = Status.NT
        print "Heat Demand"
        print "CascadeIndex - QD_total - QD_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            if i == cascade:
                print i, \
                "%10.4f" % self.QD_T_mod[i][NT + 1], \
                self.QD_Tt_mod[i][NT + 1][0:23]
            else:
                pass
        print "Cooling Demand"
        print "CascadeIndex - QDc_total - QDc_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            if i == cascade:
                print i, \
                "%10.4f" % self.QDc_T_mod[i][NT + 1], \
                self.QDc_Tt_mod[i][NT + 1][0:23]
            else:
                pass
        print "Heat Availability"
        print "CascadeIndex - QA_total - QA_Tt(first day)"
        for i in xrange(self.NEquipe + 1):
            if i == cascade:
                print i, \
                "%10.4f" % self.QA_T_mod[i][0], \
                self.QA_Tt_mod[i][0][0:23]
            else:
                pass

    def getEquipmentCascade(self):
        """
        Fetch the equipment list into self.EquipTableDataList.
        """
        self.systemModel.build()
        self.equipments = self.systemModel.equipment
        self.NEquipe = len(self.equipments) #SD change 30/04.2008

        self.cascade = []
        for j in xrange(self.NEquipe):
            self.cascade.append({"equipeID":self.equipments[j].id, \
                                 "equipeNo":self.equipments[j].EqNo, \
                                 "equipeType":self.equipments[j].EquipType, \
                                 "equipePnom":self.equipments[j].HCGPnom})
            if (self.equipments[j].CascadeIndex != j + 1):
                logError("Interfaces.getEquipmentCascade): error in SQL data - cascade index %s corrected to new index %s" % \
                      (self.equipments[j].CascadeIndex, j + 1))
                self.equipments[j].CascadeIndex = j + 1
                dbio.commit()

        self.EquipTableDataList = []
        for j in xrange(self.NEquipe):
            self.EquipTableDataList.append([unicode(self.equipments[j].Equipment, "utf-8"),
                                            self.equipments[j].HCGPnom,
                                            self.equipments[j].HCGTEfficiency, \
                                            self.equipments[j].EquipType,
                                            self.equipments[j].HPerYearEq,
                                            self.equipments[j].YearManufact]) #SD change 30/04.2008

        if Status.SimTool != "Plus":
            self.extendCascadeArrays(self.NEquipe)  #security feature: assure that for all equipes there's enough space                                       

    def changeInCascade(self, index, keepEnergyData=False, keepModel=False):
        """Lower the cascadeUpdateLevel as changes have occurred in 
        level index (index counting from 1 to NJ, and 0 corresponds to 
        the basic energy demand calculations).
        
        The cascade level is then lowered by 1 and a full simulation 
        is necessary as partial simulation is not possible as 
        memory constraints forbid to keep the intermediate results.
        
        If the global energy data is not changed by this change,
        e.g. if a alternative is copied, set keepEnergyData to True
        to avoid unnecessary invalidation.
        
        """
        if index >= 0:
            if not keepEnergyData:
                    Status.prj.setStatus("Energy", 0)
                    Status.prj.setStatus("StatusECO", 0)
            if index <= self.cascadeUpdateLevel:
                self.cascadeUpdateLevel = index-1
                if self.cascadeDistributionLevel > self.cascadeUpdateLevel:
                    self.cascadeDistributionLevel = -1 # if cascadeUpdateLevel has been lowered below cascadeDistributionLevel, USHm_remaining_Tt is not useful anymore
#                Status.schedules.equipmentSchedulesOutOfDate = True
                logTrack("Interfaces (changeInCascade): StatusEnergy %s cascadeUpdateLevel: %s cascadeDistributionLevel: %s " %
                         (Status.StatusEnergy, self.cascadeUpdateLevel, self.cascadeDistributionLevel))
            else:
                pass
##                logDebug("Interfaces (changeInCascade): StatusEnergy %s kept the cascadeUpdateLevel: %s cascadeDistributionLevel: %s" %
##                         (Status.StatusEnergy, self.cascadeUpdateLevel, self.cascadeDistributionLevel))


            if index > 0 and any([(eqo.TReject is not None or eqo.TDrive is not None) for eqo in Status.int.systemModel.eqObject]):
                showWarning("Adding/deleting/reordering of equipment in combination with monitoring data for temperatures (TDrive, TReject) is not allowed" +
                            "\n" + "Please undo your changes")
        else:
            logDebug("Interfaces (changeInCascade): cannot change a cascade level " +
                     "that does not exist [-> level %s]" % index)

        if (Status.int.systemModel is not None) and (not keepModel): # necessary in order to call changeInCascade BEFORE building model in setActiveAlternative
            Status.int.systemModel.changed()
        
    def changeInHeatRecovery(self):
        """To be called if a heatexchanger has been changes and thus 
        the heat recovery calculations are invalidated.
        
        """
        self.changeInCascade(0)
        self.heatRecoveryOutOfDate = True
        
    def setGraphicsData(self, key, data):
        """
        Store graphics data.
        A (deep) copy of the data is stored in the dictionary GData 
        under the key 'key'.
        
        """
        self.GData[key] = copy.deepcopy(data)


    def clearHXData(self):
        """
        clears HX Data.
        used at open Project, change Proposal
        """
        
        self.HXPinchConnection = []
        self.StreamGen = None
        self.hrdata = None
    
    def clearHX(self):
        self.HXPinchConnection = []
        self.StreamGen = None

if __name__ == "__main__":
    # for testing purposes only
    # should be invoked: python interfaces.py
    #
    pass
