# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   ModulePipes (Heat Recovery)
#           
#------------------------------------------------------------------------------
#           
#   Module for calculation of pipe energy demands
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------     
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import numpy as np
from modules.messageLogger import *
from modules.heatFlow import HeatFlow_t, diffQ_Tt, shiftQ_Tt
from modules.interfaces import createQ_Tt, calcQ_T
from GUI.status import *
from constants import DEBUG
from auxiliary.auxiliary import noneCheck
from model.systemConnections import getAllProcessesConnectedTo, getAllEquipmentsConnectedTo, getAllMandCoolDownProcessesConnectedTo

def getDT_USHProc():
    """Return the list of temperature differences by which UPHProc/UPCProc
    has to be shifted before being added to USHCPipe
    
    """
    processes = Status.int.systemModel.process
    DT_USHProc = []
    DT_USHProcMCD = []

    for process in processes:
        heating = (process.heating_cooling == "heating")
        ExternalSupply = process.ExternalSupply == "yes"
        DPSTmin = process.DPSTmin                
        DPSTminMCD = process.DPSTminMCD

        if ExternalSupply is True:
            if heating:
                DT = DPSTmin
                DT_MCD = -DPSTminMCD
            else:
                DT = -DPSTmin
                DT_MCD = 0
            DT_USHProc.append(DT)
            DT_USHProcMCD.append(DT_MCD)
        else:
            DT_USHProc.append(0)
            DT_USHProcMCD.append(0)
            
    return DT_USHProc, DT_USHProcMCD

def getAllHeatingPipes():
    """Return the list of indices of pipes connected to heating
    processes and the list of database IDs (QDistributionHC_ID) 
    of these pipes.
    
    :returns: A tuple (list of indices, list of databaseIDs)
    
    """
    indices = []
    dbIds = []
    pipes = Status.prj.getPipes()
    for m, pipe in enumerate(pipes):
        pipeObj = Status.int.systemModel.pipeObject[m]
        if pipeObj.heating:
            indices.append(pipe.PipeDuctNo - 1)
            dbIds.append(pipe.QDistributionHC_ID)
    return (indices, dbIds)

def getAllCoolingPipes():
    """Return the list of indices of pipes connected to cooling
    processes and the list of database IDs (QDistributionHC_ID) 
    of these pipes.
    
    :returns: A tuple (list of indices, list of databaseIDs)
    
    """
    indices = []
    dbIds = []
    pipes = Status.prj.getPipes()
    for m, pipe in enumerate(pipes):
        pipeObj = Status.int.systemModel.pipeObject[m]
        if not pipeObj.heating:
            indices.append(pipe.PipeDuctNo - 1)
            dbIds.append(pipe.QDistributionHC_ID)
    return (indices, dbIds)

def calculateUSHCPipe():
    """Calculate for all pipes in the current alternative the heat/cold
    demand USH_Tt, USC_Tt, USHCPipe_Tt and the aggregated values.
    """
    if DEBUG != "OFF":
        logTrack("--------------------------------------")
        logTrack("ModulePipes.calculateUSHCPipe starting")
    # Reset stale data to conserve memory during the calculations
    Status.int.USHCPipe_Tt = [None for dummy in range(Status.NPipeDuct)]
    Status.int.USHCPipe = [0]*Status.NPipeDuct
    Status.int.USHCPipe_Tt_hf = [None for dummy in range(Status.NPipeDuct)]
    Status.int.UPHProc_Tt = [None for dummy in Status.int.UPH_Tt]
    Status.int.UPHCProcm = [0]*Status.NPipeDuct
    
    USHCPipe_Tt = [createQ_Tt() for dummy in range(Status.NPipeDuct)]
    Status.int.USHCm_Tt = [createQ_Tt() for dummy in range(Status.NPipeDuct)]
    USHCm_Tt = Status.int.USHCm_Tt
    QHXPipe_Tt = Status.int.QHXPipe_Tt
    
    Status.int.QWHPipe_Tt = [HeatFlow_t() for dummy in range(Status.NPipeDuct)]
    # TODO: for the moment dummy array. later on waste heat from pipes 
    # should be calculated.
    
    heatingPipeIndices = [m for m, pipe in enumerate(Status.int.systemModel.pipeObject) if pipe.heating]
    coolingPipeIndices = [m for m, pipe in enumerate(Status.int.systemModel.pipeObject) if not pipe.heating]

    DT_USHProc, DT_USHProcMCD = getDT_USHProc()
    
    pipes = Status.int.systemModel.pipe
    Status.int.USHCPipeChanged = [False] * len(pipes)
    Status.int.USHCmChanged = [False] * len(pipes)
    Status.int.QWHPipeChanged = [False] * len(pipes)

    #outputs to be generated (inputs from pipes to processes): UP[HC]Proc
    UPHCProcm = [0] * Status.NPipeDuct # the output per pipe
    UPHProc_Tt = [None for dummy in Status.int.UPH_Tt] # the outputs per process
    UPHProcTotal_Tt = createQ_Tt()
    UPCProc_Tt = [None for dummy in Status.int.UPH_Tt] #space for additional cooling of heating process
    UPCProcTotal_Tt = createQ_Tt()

    UPHwUsedResTotal_Tt = HeatFlow_t()
    
    # TODO: WARNING - here direct iteration over process indices would be better
    #       current version works only while one process - one pipe limit is maintained!
    for m in heatingPipeIndices:
        processIndices = getAllProcessesConnectedTo(m)
        for k in processIndices:
            if isinstance(Status.int.QHXProc_Tt[k], HeatFlow_t):
                QHXProc_Tt = Status.int.QHXProc_Tt[k].Q_Tt()
            else:
                QHXProc_Tt = Status.int.QHXProc_Tt[k]

            if DEBUG != "OFF": logTrack("ModulePipes: QHXProc[%s] = %s UPH = %s"%(k,sum(QHXProc_Tt[Status.NT+1]),sum(Status.int.UPH_Tt[k][Status.NT+1])))
            
            UPHProc_Tt[k] = Status.int.UPH_Tt[k] - QHXProc_Tt
            Status.int.UPHProcChanged[k] = Status.int.QHXProcChanged[k] or Status.int.UPHChanged[k]
            if Status.HRTool != "estimate":     # security item. necessary only in actualHXNet mode probably due to conversion errors from/to Q_Tt
                for iT in xrange(Status.NT,-1,-1):
                    UPHProc_Tt[k][iT] = np.min((UPHProc_Tt[k][iT + 1], UPHProc_Tt[k][iT]), 0)
                UPHProc_Tt[k] = UPHProc_Tt[k].clip(min=0)
            
            if DEBUG != "OFF":
                logTrack("ModulePipes: Pipe %s UPH %s QHXProc %s UPHProc %s"%(m, sum(Status.int.UPH_Tt[k][Status.NT+1]), sum(QHXProc_Tt[Status.NT+1]), sum(UPHProc_Tt[k][Status.NT+1])))

            UPHProcTotal_Tt += UPHProc_Tt[k]
            Status.int.UPXProck[k] = np.sum(UPHProc_Tt[k][Status.NT+1])
           
            UPC = Status.int.UPHwReq_Tt_hf[k]
            if UPC.Q > 0:
                if Status.HRTool == "estimate":
                    UPCProc_Tt[k] = Status.int.UPCProc_Tt[k] #direct outcome of estimate mode
                else:
                    QWHUsed = Status.int.UPHwUsed_Tt[k]
                    if DEBUG != "OFF":
                        print "ModulePipes: allocating QWHUsed (%s kWh)of heating process with MCD to process %s"%(QWHUsed.Q,k+1)
                    (UPCProc, QWHUsedRes) = diffQ_Tt(UPC,QWHUsed)
                    if DEBUG != "OFF":
                        print "ModulePipes: UPC %s QWHUsed %s UPCProc = %s QWHUsedRes %s"%(UPC.Q, QWHUsed.Q, UPCProc.Q, QWHUsedRes.Q)

                    UPHwUsedResTotal_Tt.add(QWHUsedRes)

                    UPCProc_Tt[k] = UPCProc.Q_Tt(mode="cooling")
                UPCProcTotal_Tt += UPCProc_Tt[k]
                Status.int.UPCProcChanged[k] = Status.int.UPHwUsedChanged[k] or Status.int.UPCChanged[k]
            else:
                if DEBUG != "OFF":
                    print "ModulePipes: UPCProc = 0. Process %s without MCD"%(k+1)
                UPCProc_Tt[k] = Status.int.createQ_Tt()
                Status.int.UPCProcChanged[k] = False
                
    # reduce cold exchange by heat extracted from processes
    for m in coolingPipeIndices:
        processIndices = getAllProcessesConnectedTo(m)
        for k in processIndices:
            if isinstance(Status.int.QHXProc_Tt[k], HeatFlow_t):
                QHXProc_Tt = Status.int.QHXProc_Tt[k].Q_Tt(mode="cooling")
            else:
                QHXProc_Tt = Status.int.QHXProc_Tt[k]

            UPHProc_Tt[k] = Status.int.UPH_Tt[k] - QHXProc_Tt
            Status.int.UPXProck[k] = np.sum(UPHProc_Tt[k][0])
            Status.int.UPHProcChanged[k] = Status.int.QHXProcChanged[k] or Status.int.UPHChanged[k]

            if Status.HRTool != "estimate":     # security item. necessary only in actualHXNet mode probably due to conversion errors from/to Q_Tt
                for iT in xrange(Status.NT):
                    UPHProc_Tt[k][iT+1] = np.min((UPHProc_Tt[k][iT + 1], UPHProc_Tt[k][iT]), 0)
                UPHProc_Tt[k] = UPHProc_Tt[k].clip(min=0)
                    
            if DEBUG != "OFF":
                logTrack("ModulePipes: Pipe %s UPH %s QHXProc %s UPHProc %s"%(m, sum(Status.int.UPH_Tt[k][0]), sum(QHXProc_Tt[0]), sum(UPHProc_Tt[k][0])))

            UPCProcTotal_Tt += UPHProc_Tt[k]
    # saving results at pipe entry (UPHProc, ...)
    
    Status.int.UPHProc_Tt = UPHProc_Tt
    Status.int.UPHProcTotal_Tt = UPHProcTotal_Tt
    Status.int.UPCProc_Tt = UPCProc_Tt
    Status.int.UPCProcTotal_Tt = UPCProcTotal_Tt

    Status.int.UPHwUsedResTotal_Tt = UPHwUsedResTotal_Tt

    Status.int.UPHProcTotal_T = calcQ_T(Status.int.UPHProcTotal_Tt)
    Status.int.UPCProcTotal_T = calcQ_T(Status.int.UPCProcTotal_Tt)

    (dummy, generalData) = Status.prj.getProjectData()
    if generalData.HDEffAvg is not None:
        distributionEfficiency = generalData.HDEffAvg
    else:
        logDebug("SimulateHR: error reading distribution efficiency from cgeneraldata")
        distributionEfficiency = 0.9
    if distributionEfficiency < 0.3:
        logWarning("Very low distribution efficiency: %s. Revise your data" % distributionEfficiency)
    fDist = 1. / max(distributionEfficiency, 0.1)  #distribution efficiency < 10% doesn't make much sense
    # TODO: get the getAllProcessesConnected logic right perhaps by changing the
    # loops to a per process basis. (Look above for the same thought.)
    for m in heatingPipeIndices:
        HDPipe = pipes[m].HDEffAvg
        if HDPipe is None or HDPipe <= 0:
            fDistPipe = max(fDist,1.02)   # on the long run fDist (the general one) should be eliminated or correctly calculated
            logWarning("ModulePipes: Distribution efficiency of pipe %s has been undefined. Set to default (1/1.02)")
            pipes[m].HDEffAvg = 1./fDistPipe
        else:
            fDistPipe = 1./HDPipe
            
        # find the indices of the processes that are fed by pipe m
        processIndices = getAllProcessesConnectedTo(m)
        for k in processIndices:
            # add the demand of the process 
            USHCPipe_Tt[m][:] += shiftQ_Tt(UPHProc_Tt[k][:],DT_USHProc[k],"heating") * fDistPipe
            UPHCProcm[m] += calcQ_T(UPHProc_Tt[k])[-1]
            Status.int.USHCPipeChanged[m] = Status.int.USHCPipeChanged[m] or Status.int.UPHProcChanged[k]
            
        if DEBUG != "OFF": logTrack("ModulePipes: USHCPipe[%s] = %s (before FETHeat)"%(m,sum(USHCPipe_Tt[m][Status.NT+1])))

        equipmentIndices = getAllEquipmentsConnectedTo(m)
        for j in equipmentIndices:
            try:    # necessary, because at first iteration FETHeat_j does not yet exist
                FETHeat_j = Status.int.FETHeat_j_Tt[j].Q
                USHCPipe_Tt[m] += Status.int.FETHeat_j_Tt[j].Q_Tt() * fDistPipe
                UPHCProcm[m] += FETHeat_j
                
                Status.int.USHCPipeChanged[m] = Status.int.USHCPipeChanged[m] or Status.int.FETHeat_jChanged[j]
            except:
                pass

        if DEBUG != "OFF":
            logTrack("ModulePipes: Pipe %s USHPipe %s UPHProc %s"%(m, sum(USHCPipe_Tt[m][Status.NT+1]), UPHCProcm[m]))


    for m in coolingPipeIndices:
        HDPipe = pipes[m].HDEffAvg
        if HDPipe is None or HDPipe <= 0:
            fDistPipe = max(fDist,1.02)   # on the long run fDist (the general one) should be eliminated or correctly calculated
            logWarning("ModulePipes: Distribution efficiency of pipe %s has been undefined. Set to default (1/1.02)")
            pipes[m].HDEffAvg = 1./fDistPipe
        else:
            fDistPipe = 1./HDPipe

        # for cooling shift by -5 K and use the same distribution efficiency
        processIndices = getAllProcessesConnectedTo(m)
        for k in processIndices:
            USHCPipe_Tt[m][:] += shiftQ_Tt(UPHProc_Tt[k][:],DT_USHProc[k],"cooling") * fDistPipe
            UPHCProcm[m] += calcQ_T(UPHProc_Tt[k])[0]
            Status.int.USHCPipeChanged[m] = Status.int.USHCPipeChanged[m] or Status.int.UPHProcChanged[k]
            
        processIndices = getAllMandCoolDownProcessesConnectedTo(m)
        for k in processIndices:
            USHCPipe_Tt[m][:] += shiftQ_Tt(UPCProc_Tt[k][:],DT_USHProcMCD[k],"cooling") * fDistPipe
            UPHCProcm[m] += calcQ_T(UPCProc_Tt[k])[0]
            Status.int.USHCPipeChanged[m] = Status.int.USHCPipeChanged[m] or Status.int.UPCProcChanged[k]
            
        if DEBUG != "OFF":
            logTrack("ModulePipes: Pipe %s USCPipe %s UPCProc %s"%(m, sum(USHCPipe_Tt[m][0]), UPHCProcm[m]))

    NK = len(Status.int.UPHChanged)
    Status.int.UPHChanged = [False] * NK
    Status.int.UPCChanged = [False] * NK
    NJ = len(Status.int.FETHeat_jChanged)
    Status.int.FETHeat_jChanged = [False] * NJ
    # make the results globally accessible
    Status.int.USHCPipe_Tt = USHCPipe_Tt

    if DEBUG != "OFF": 
        for m, pipe in enumerate(USHCPipe_Tt):
            USHC = sum(pipe[Status.NT+1][:])
            logTrack("ModulePipes: Pipe %s: USHC = %s"%(m+1,USHC))
        
    for m in heatingPipeIndices:
        fluidId = int(noneCheck(pipes[m].HeatDistMedium, 0))
        T0 = pipes[m].TreturnDistrib
        T1 = pipes[m].ToutDistrib
        Status.int.USHCPipe_Tt_hf[m] = HeatFlow_t(fluid_id=fluidId, T0=T0, T1=T1, Q_t = USHCPipe_Tt[m][Status.NT+1])
        Status.int.USHCPipe_Tt_hf[m].exportGlb(Status.int.systemModel.pipeObject[m].USHGlb)
        Status.int.USHCPipe[m] = Status.int.USHCPipe_Tt_hf[m].Q
        Status.int.QWHPipe_Tt[m].exportGlb(Status.int.systemModel.pipeObject[m].QWHGlb)
        
    for m in coolingPipeIndices:
        fluidId = int(noneCheck(pipes[m].HeatDistMedium, 0))
        T1 = pipes[m].TreturnDistrib
        T0 = pipes[m].ToutDistrib
        Status.int.USHCPipe_Tt_hf[m] = HeatFlow_t(fluid_id=fluidId, T0=T0, T1=T1, Q_t = USHCPipe_Tt[m][0])
        Status.int.USHCPipe_Tt_hf[m].exportGlb(Status.int.systemModel.pipeObject[m].USHGlb)
        Status.int.USHCPipe[m] = Status.int.USHCPipe_Tt_hf[m].Q
        Status.int.QWHPipe_Tt[m].exportGlb(Status.int.systemModel.pipeObject[m].QWHGlb)
        
    # subtract heat recovery
    for m in heatingPipeIndices:
        USHCm_Tt[m] = USHCPipe_Tt[m] - QHXPipe_Tt[m].Q_Tt()
        for iT in xrange(Status.NT,-1,-1):
            USHCm_Tt[m][iT] = np.min((USHCm_Tt[m][iT + 1], USHCm_Tt[m][iT]), 0)
        Status.int.USHCmChanged[m] =  Status.int.USHCPipeChanged[m] or Status.int.QHXPipeChanged[m]         
        USHCm_Tt[m] = USHCm_Tt[m].clip(min=0)
        if DEBUG != "OFF":
            logTrack("ModulePipes: m %s USHCm %s Changed ? %s USHCPipe %s QHXPipe %s"%
                     (m,sum(USHCm_Tt[m][Status.NT+1]),Status.int.USHCmChanged[m],sum(USHCPipe_Tt[m][Status.NT+1]),sum(QHXPipe_Tt[m].Q_Tt()[Status.NT+1])))

    for m in coolingPipeIndices:
        USHCm_Tt[m] = USHCPipe_Tt[m] - QHXPipe_Tt[m].Q_Tt(mode="cooling")
        for iT in xrange(Status.NT+1):
            USHCm_Tt[m][iT+1] = np.min((USHCm_Tt[m][iT + 1], USHCm_Tt[m][iT]), 0)
        Status.int.USHCmChanged[m] =  Status.int.USHCPipeChanged[m] or Status.int.QHXPipeChanged[m]         
        USHCm_Tt[m] = USHCm_Tt[m].clip(min=0)
        if DEBUG != "OFF":
            logTrack("ModulePipes: m %s USHCm %s Changed ? %s USHCPipe %s QHXPipe %s"%
                     (m,sum(USHCm_Tt[m][0]),Status.int.USHCmChanged[m],sum(USHCPipe_Tt[m][0]),sum(QHXPipe_Tt[m].Q_Tt(mode="cooling")[0])))

    Status.int.UPHCProcm = UPHCProcm
    Status.int.USHCPipe = [hf.Q for hf in Status.int.USHCPipe_Tt_hf]
    
    Status.int.cascadeUpdateLevel = 0 #indicates that demand profile is created
    Status.int.cascadeDistributionLevel = -1
    
    if DEBUG != "OFF":
        logTrack("END ModulePipes.calculateUSHCPipe")
        logTrack("--------------------------------------")
    
    
