# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#    PROCESSES
#            
#------------------------------------------------------------------------------
#            
#    Functions for calculation and temperature decomposition of
#       process heat demands
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import numpy as np
from constants import *
from GUI.status import *
from modules.messageLogger import logDebug, logWarning, logTrack
from modules.fluids import Fluid, Fuel
from GUI.dialogGauge import DialogGauge
from modules.heatFlow import HeatFlow_t, createHeatFlow_tDict
from modules.interfaces import createQ_T, createQ_Tt
from copy import deepcopy
from modules import equipments
from auxiliary.auxiliary import checkLimits, cutInterval, noneCheck
from modules.modulePipes import calculateUSHCPipe
import database.dbio as dbio


class Processes(object):
    """Module that handles all project schedules"""
    def __init__(self):
        self.outOfDate = True
        self.outOfDateYearly = True
       
    def createYearlyDemand(self):
        """
        Is created automatically in an alternative after cchecking or
        after selecting a new alternative that is already cchecked.
        
        """
        if not self.outOfDateYearly: return

        logTrack("Processes (createYearlyDemand): starting")
##        processes = Status.prj.getProcesses()
        Status.int.systemModel.build() # assure that model building is completed
        processes = Status.int.systemModel.process

        NK = len(processes)
        Status.int.UPX = [0]*NK
        Status.int.UPXProck = [0]*NK
        Status.int.QHXProc = [0]*NK
        Status.int.UPC = [0]*NK
        Status.int.UPCProck = [0]*NK
        Status.int.QCXProc = [0]*NK
        
        UPH_T = [createQ_T() for dummy in processes]
        UPHs_T = [createQ_T() for dummy in processes]
        UPHm_T = [createQ_T() for dummy in processes]
        UPHc_T = [createQ_T() for dummy in processes]
        UPHw_T = [createQ_T() for dummy in processes]
        UPHwReq_T = [createQ_T() for dummy in processes]
        UPHTotal_T = createQ_T()
        UPHwTotal_T = createQ_T()
        UPCTotal_T = createQ_T()
        UPCwTotal_T = createQ_T()
        NT = Status.NT
        
        self.UPHcperStream = []
        self.UPHwperStream = []
#        Status.int.UPH_c_t = {}
#        Status.int.UPH_w_t = {}

        self.UPHTotal = 0
        self.UPCTotal = 0
        
        for k, process in enumerate(processes):
            (inStreams, outStreams) = (Status.int.systemModel.processStreamsIn[k], Status.int.systemModel.processStreamsOut[k]) 
            
            UPHc_per_stream_T = np.zeros((len(inStreams), NT + 2), float)
            UPHw_per_stream_T = np.zeros((len(outStreams), NT + 2), float)
            UPHwReq_per_stream_T = np.zeros((len(outStreams), NT + 2), float)
            k = process.ProcNo - 1
            UPX = process.UPH
            
            Status.int.UPX[k] = UPX
            Status.int.UPXProck[k] = UPX
            Status.int.QHXProc[k] = 0

##            print "Processes.createYearlyDemand: UPX = %s"%UPX

            heating = (process.heating_cooling == "heating")
                
            DPSTmin = noneCheck(process.DPSTmin, 0)
            DPSTminMCD = noneCheck(process.DPSTminMCD, 0)

            ExternalSupply = process.ExternalSupply == "yes"
            if ExternalSupply is True:
                DT_UPH = 0
                DT_MCD = 0
            else:
                if heating is True:
                    DT_UPH = DPSTmin
                    DT_MCD = -DPSTminMCD
                else:
                    DT_UPH = -DPSTmin
                    DT_MCD = 0
                    
            if process.PT is not None:
                PSTmin = process.PT + DT_UPH
            else:
                PSTmin = None

            if process.PTStartUp is not None:
                PSTminStartUp = process.PTStartUp + DT_UPH
            else:
                PSTminStartUp = None

            if inStreams == []: # no inStreams
                UPHc_per_stream = np.zeros((1, 1), float)
                distInStreams = np.zeros((1, NT + 2), float)
            else:
                UPHc_per_stream = np.array([checkLimits(stream.UPHc, 0.0, INFINITE, 0.0) 
                                        for stream in inStreams])
                listTDist = []    
                for stream in inStreams:
                    T0 = stream.PTInFlowRec + DT_UPH
                    T1 = process.PT + DT_UPH
                    listTDist.append(self._createTempDist(T1, T0=T0, mode=process.heating_cooling))
                distInStreams = np.array(listTDist)
                
            UPHm = checkLimits(process.UPHm, 0.0, INFINITE, 0.0)
            distUPHm = self._createTempDist(PSTmin, mode=process.heating_cooling)

            UPHs = checkLimits(process.UPHs, 0.0, INFINITE, 0.0)
            distUPHs = self._createTempDist(PSTmin, T0=PSTminStartUp,
                                           mode=process.heating_cooling)

            UPHw_per_stream = np.array([checkLimits(stream.UPHw, 0.0, INFINITE, 0.0) 
                                        for stream in outStreams])
            UPHwReq_per_stream = np.array([checkLimits(stream.UPHwReq, 0.0, INFINITE, 0.0) 
                                           for stream in outStreams])
            tempDistOutStreams = [] 
            for index in xrange(len(outStreams)):
                if (UPHw_per_stream[index] <= 0):
                    tempDistOutStreams.append(createQ_T())
                else:
                    stream = outStreams[index]

                    T0 = stream.PTFinal
                    if T0 is not None: T0 += DT_MCD
                    T1 = stream.PTOutFlow
                    if T1 is not None: T1 += DT_MCD

                    if DT_MCD <> 0:
                        tempDistOutStreams.append(self.createInvTempDistLat(T1,
                                                                        None,
                                                                        0,
                                                                        T0=T0,
                                                                        mode=process.heating_cooling))
                    else:
                        tempDistOutStreams.append(self.createInvTempDistLat(T1,
                                                                        stream.HOutFlow,
                                                                        stream.dbfluid_id,
                                                                        T0=T0,
                                                                        mode=process.heating_cooling))
            tempDistCoolingDown = []
            for index in xrange(len(outStreams)):
                if process.heating_cooling == 'cooling':
                    # there is no "mandatory heating up" for now.
                    tempDistCoolingDown.append(createQ_T())
                else:
                    if UPHwReq_per_stream[index] <= 0:
                        tempDistCoolingDown.append(createQ_T())
                    else:
                        stream = outStreams[index]
                        if stream.PTFinalRequired is None:
                            # There is no mandatory cooling down for this stream.
                            tempDistCoolingDown.append(createQ_T())
                        else:
                            # The cooling down demand is the same as a cooling 
                            # process demand.

                            T1 = stream.PTFinalRequired
                            if T1 is not None: T1 += DT_MCD
                            T0 = stream.PTOutFlow
                            if T0 is not None: T0 += DT_MCD

                            if DT_MCD <> 0:
                                tempDistCoolingDown.append(self.createInvTempDistLat(T1,
                                                                                     None,
                                                                                     0,
                                                                                     T0=T0,
                                                                                     mode="cooling"))
                            else:

                                tempDistCoolingDown.append(self.createInvTempDistLat(stream.PTFinalRequired,
                                                                                     stream.HOutFlow,
                                                                                     stream.dbfluid_id,
                                                                                     T0=stream.PTOutFlow,
                                                                                     mode="cooling"))
           
            if len(outStreams) == 0: # no outStreams
                UPHw_per_stream = np.zeros((1, 1), float)
                tempDistOutStreams = np.zeros((1, Status.NT + 2), float)
                UPHwReq_per_stream = np.zeros((1, 1), float)
                tempDistCoolingDown = np.zeros((1, Status.NT + 2), float)
            else:
                tempDistOutStreams = np.array(tempDistOutStreams)
                tempDistCoolingDown = np.array(tempDistCoolingDown)
         
            UPHc_per_stream_T = (UPHc_per_stream * distInStreams.transpose()).transpose()
            UPHw_per_stream_T = (UPHw_per_stream * tempDistOutStreams.transpose()).transpose()
            UPHwReq_per_stream_T = (UPHwReq_per_stream * tempDistCoolingDown.transpose()).transpose()
            
            UPHc_T[k] = UPHc_per_stream_T.sum(0)
            UPHw_T[k] = UPHw_per_stream_T.sum(0)
            UPHwReq_T[k] = UPHwReq_per_stream_T.sum(0)
            UPHs_T[k] = (UPHs * distUPHs)
            UPHm_T[k] = (UPHm * distUPHm)
            # actually: UPH_T[k] = UPHc_T[k] + UPHm_T[k] + UPHs_T[k] but no arrays :-(
            UPH_T[k] = (UPHc_per_stream_T.sum(0) + (UPHm * distUPHm) 
                        + (UPHs * distUPHs))
            # TODO: deduce any recovered heat from UPHwReq before adding it to
            # UPCTotal_T as a cooling demand.
            if process.heating_cooling == "heating":
                UPXTotal_T = UPHTotal_T
                UPXwTotal_T = UPHwTotal_T
                UPXwReqTotal_T = UPCTotal_T

                self.UPHTotal += UPH_T[k][Status.NT+1]
                self.UPCTotal += UPHwReq_T[k][0]
                Status.int.UPC[k] = UPHwReq_T[k][0]
                Status.int.UPCProck[k] = UPHwReq_T[k][0]
                Status.int.QCXProc[k] = 0.0

            else:
                UPXTotal_T = UPCTotal_T
                UPXwTotal_T = UPCwTotal_T
                UPXwReqTotal_T = createQ_T()

                self.UPCTotal += UPH_T[k][0]

            UPXTotal_T += UPH_T[k]
            UPXwTotal_T += UPHw_T[k]
            UPXwReqTotal_T += UPHwReq_T[k]
            

        Status.int.UPH_T = UPH_T   
        Status.int.UPHc_T = UPHc_T   
        Status.int.UPHm_T = UPHm_T   
        Status.int.UPHs_T = UPHs_T   
        Status.int.UPHw_T = UPHw_T
        Status.int.UPHwReq_T = UPHwReq_T
        Status.int.UPHTotal_T = UPHTotal_T
        Status.int.UPHwTotal_T = UPHwTotal_T
        Status.int.UPCTotal_T = UPCTotal_T
        Status.int.UPCwTotal_T = UPCwTotal_T
        logTrack("Processes (createYearlyDemand): yearly heat demand = %s yearly waste heat availability = %s" % \
              (Status.int.UPHTotal_T[NT + 1], Status.int.UPHwTotal_T[0]))

        self.outOfDateYearly = False

        self.storeProcessTotals()

    def storeProcessTotals(self):
        """
        Writes the yearly totals of UPH and UPC into the database (cgeneraldata)
        """

        generalData = Status.int.systemModel.generalData

        generalData.UPH = dbio.check(self.UPHTotal)
        generalData.UPC = dbio.check(self.UPCTotal)
        dbio.commit()
        
    def createAggregateDemand(self):
        """Calculate the totals of the per process demands.
        
        This includes a call of the heat recovery calculations.
        
        """
        from modules.heatRecovery.hxConnection import buildHXNet, clearHXResults
        
        logTrack("Processes (createAggregateDemand): Creating time and temperature dependent heat demand")
        if Status.SimTool != "Plus":
            dlg = DialogGauge(Status.main, _("Time dependent aggregate heat demand"), _("generate demand profile"))
        
        projectData = Status.int.systemModel.projectData
        Status.HPerDayInd = projectData.HPerDayInd

        if Status.processData.outOfDate == False:
            logTrack("Processes (createAggregateDemand): WARNING - someone wants to create demand profile which is already up to date")
            return

        if Status.processData.outOfDateYearly == True:
            logTrack("Processes (createAggregateDemand): creating yearly demand UPHk(T)")
            self.createYearlyDemand()

        if Status.schedules.outOfDate == True:
            logTrack("Processes (createAggregateDemand): creating process schedules")
            Status.schedules.create()

##        processes = Status.prj.getProcesses()
        processes = Status.int.systemModel.process
        
        if Status.SimTool != "Plus":
            UPH_Tt = [createQ_Tt() for dummy in processes]
            UPHw_Tt = [createQ_Tt() for dummy in processes]
            UPHwReq_Tt = [createQ_Tt() for dummy in processes]
        
        UPHs_Tt = [HeatFlow_t(name="UPHs[%s]"%proc.Process) for proc in processes]
        UPHm_Tt = [HeatFlow_t(name="UPHm[%s]"%proc.Process) for proc in processes]
        UPHc_Tt = [HeatFlow_t(name="UPHc[%s]"%proc.Process) for proc in processes]
        UPHw_Tt_hf = [HeatFlow_t(name="UPHw[%s]"%proc.Process) for proc in processes]
        UPHwReq_Tt_hf = [HeatFlow_t(name="UPHwReq[%s]"%proc.Process) for proc in processes]
        UPHwResTotal_Tt = HeatFlow_t(name="UPHwResTotal")
        UPHwUsed_Tt = [HeatFlow_t(name="UPHwUsed[%s]"%proc.Process) for proc in processes]

        Status.int.UPHs_Tt = UPHs_Tt
        Status.int.UPHm_Tt = UPHm_Tt
        Status.int.UPHc_Tt = UPHc_Tt
        Status.int.UPHw_Tt_hf = UPHw_Tt_hf
        Status.int.UPHwReq_Tt_hf = UPHwReq_Tt_hf
        Status.int.UPHwResTotal_Tt = UPHwResTotal_Tt
        Status.int.UPHwUsed_Tt = UPHwUsed_Tt
    
#        Status.int.UPH_c_t_reload = {}
#        Status.int.UPH_w_t_reload = {}

        if Status.SimTool != "Plus":
            UPHTotal_Tt = createQ_Tt()
            UPHwTotal_Tt = createQ_Tt()
            UPCTotal_Tt = createQ_Tt()
            UPCwTotal_Tt = createQ_Tt()
            
            UPHc_T = Status.int.UPHc_T # Is used read only.
            UPHm_T = Status.int.UPHm_T # Is used read only.
            UPHs_T = Status.int.UPHs_T # Is used read only.
            UPHw_T = Status.int.UPHw_T # Is used read only.
            UPHwReq_T = Status.int.UPHwReq_T # Is used read only

        NK = len(processes)

        for k, process in enumerate(processes):
            scheduleC = Status.schedules.procInFlowSchedules[k]
            scheduleM = Status.schedules.procOpSchedules[k]
            scheduleS = Status.schedules.procStartUpSchedules[k]
            scheduleW = Status.schedules.procOutFlowSchedules[k]
            
            Nt = Status.Nt

            if Status.SimTool != "Plus":
                if process.heating_cooling == "heating":
                    UPXTotal_Tt = UPHTotal_Tt
                    UPXwTotal_Tt = UPHwTotal_Tt
                    UPXwReqTotal_Tt = UPCTotal_Tt
                else:
                    UPXTotal_Tt = UPCTotal_Tt
                    UPXwTotal_Tt = UPCwTotal_Tt
                    UPXwReqTotal_Tt = createQ_Tt()
                
                fC = scheduleC.fav
                fM = scheduleM.fav
                fS = scheduleS.fav
                fW = scheduleW.fav
                
                if Status.SimTool != "Plus":
                    UPH_Tt[k] = (UPHc_T[k].reshape((Status.NT + 2,1)) * fC + 
                                 UPHm_T[k].reshape((Status.NT + 2,1)) * fM + 
                                 UPHs_T[k].reshape((Status.NT + 2,1)) * fS)
                    UPHw_Tt[k] = UPHw_T[k].reshape((Status.NT + 2,1)) * fW
                    UPHwReq_Tt[k] = UPHwReq_T[k].reshape((82,1)) * fW

##            if DEBUG != "OFF":
##                print "TEST TEST (processes.py): scheduleC.fav(%s) = %s fM %s"%(process.Process,fC[:36],fM[:36])
            # TODO: Reduce any recovered heat before adding up UPHwReq_Tt as 
            # cooling demand.
                    UPXTotal_Tt += UPH_Tt[k]
                    UPXwTotal_Tt += UPHw_Tt[k]
                    UPXwReqTotal_Tt += UPHwReq_Tt[k]

            PT = process.PT
            PTStartUp = process.PTStartUp
            processMediumId = process.ProcMedDBFluid_id
            (inStreams, outStreams) = Status.prj.getStreams(process.QProcessData_ID) 
            
            heating = (process.heating_cooling == "heating")

            DPSTmin = noneCheck(process.DPSTmin, 0)
            DPSTminMCD = noneCheck(process.DPSTminMCD, 0)

            ExternalSupply = process.ExternalSupply == "yes"
            if ExternalSupply is True:
                DT_UPH = 0
                DT_MCD = 0
            else:
                if heating is True:
                    DT_UPH = DPSTmin
                    DT_MCD = -DPSTminMCD
                else:
                    DT_UPH = -DPSTmin
                    DT_MCD = 0

            if DT_UPH == 0:
                fluidProcId = processMediumId
            else:
                fluidProcId = 0

            
            T1 = PT + DT_UPH
            if PTStartUp is not None:
                T0 = PTStartUp + DT_UPH
            else:
                T0 = T1
                
            UPHm = checkLimits(process.UPHm, 0.0, INFINITE, 0.0)
            UPHm_Tt[k] = HeatFlow_t(fluid_id=0, T0=T1, T1=T1, Q_t = UPHm*scheduleM.fav, shift = DT_UPH)
            UPHm_Tt[k].exportGlb(Status.int.systemModel.processObject[k].UPHmGlb)

            if not heating:
                T0, T1 = T1, T0
            
            UPHs = checkLimits(process.UPHs, 0.0, INFINITE, 0.0)
            if UPHs > 0:
                UPHs_Tt[k] = HeatFlow_t(fluid_id=fluidProcId, T0=T0, T1=T1, Q_t = UPHs*scheduleS.fav, shift = DT_UPH)
            else:
                UPHs_Tt[k] = HeatFlow_t()
            UPHs_Tt[k].exportGlb(Status.int.systemModel.processObject[k].UPHsGlb)

            UPHc_Tt[k] = HeatFlow_t()
            
            forceAddZeros = (Status.SimTool=="Plus")
            for stream in inStreams:

                UPHc = checkLimits(stream.UPHc, 0.0, INFINITE, 0.0)

                T0 = stream.PTInFlowRec + DT_UPH
                T1 = PT + DT_UPH
                if not heating:
                    T0, T1 = T1, T0

                if DT_UPH == 0:
                    fluidStreamId = stream.dbfluid_id
                else:
                    fluidStreamId = 0
                    
                UPHc_Tt[k].add(HeatFlow_t(fluid_id=fluidStreamId, T0=T0, T1=T1, Q_t = UPHc*scheduleC.fav,name=stream.Name, shift=DT_UPH),
                               forceAddZeros=True)
            
            UPHc_Tt[k].exportGlb(Status.int.systemModel.processObject[k].UPHcGlb)
            
            for s, stream in enumerate(outStreams):

                fluidId = stream.dbfluid_id
                if DT_MCD == 0:
                    fluidStreamId = stream.dbfluid_id
                else:
                    fluidStreamId = 0

                UPHw = checkLimits(stream.UPHw, 0.0, INFINITE, 0.0)
                try:
                    UPHwReq = stream.UPHwReq if stream.UPHwReq is not None else 0
                except:
                    UPHwReq = 0
                    logTrack("database update missing -> UPHwReq cannot be imported !!! set to 0")
                    
                if heating:
                    T0 = noneCheck(stream.PTFinal, -999.9) + DT_MCD                 # noneChecks are a hack in order to avoid crash in case of corrupt data
                    T1 = noneCheck(stream.PTOutFlow, -999.9) + DT_MCD
                    if stream.PTFinalRequired is None:
                        T0Req = noneCheck(stream.PTOutFlow, -999.9) + DT_MCD
                    else:
                        T0Req = noneCheck(stream.PTFinalRequired, -999.9) + DT_MCD
                    T1Req = noneCheck(stream.PTOutFlow, -999.9) + DT_MCD
                    T0Res = T0
                    T1Res = T0Req
                    UPHwRes = UPHw - UPHwReq


#                    print "createAggDemand: adding stream to UPHwResTotal with T0/T1 = ",T0Res,T1Res,UPHwRes
                    if UPHw == 0:
                        UPHwResTotal_Tt.add(HeatFlow_t())
                    else:
                        UPHwResTotal_Tt.add(HeatFlow_t(fluid_id=fluidStreamId, T0=T0Res, T1=T1, Q_t = UPHw*scheduleW.fav, shift = DT_MCD))
##                        UPHwResTotal_Tt.add(HeatFlow_t(fluid_id=fluidStreamId, T0=T0Res, T1=T1Res, Q_t = UPHwRes*scheduleW.fav, shift = DT_MCD))

                else:
                    UPHwReq = 0   # -> for the moment UPHwReq = 0 for cooling processes !!!
                    T0 = stream.PTOutFlow + DT_MCD
                    T1 = stream.PTFinal + DT_MCD
                    T0Req = stream.PTOutFlow + DT_MCD
                    if stream.PTFinalRequired is None:
                        T1Req = stream.PTOutFlow + DT_MCD
                    else:
                        T1Req = stream.PTFinalRequired + DT_MCD
                    T0Res = T1Req
                    T1Res = T1
                    
                UPHw_Tt_hf[k].add(HeatFlow_t(fluid_id=fluidStreamId, T0=T0, T1=T1, Q_t = UPHw*scheduleW.fav,name=stream.Name, componentNo=s+1, shift=DT_MCD),
                                  forceAddZeros=True)
                UPHwReq_Tt_hf[k].add(HeatFlow_t(fluid_id=fluidStreamId, T0=T0Req, T1=T1Req, Q_t = UPHwReq*scheduleW.fav,name=stream.Name, componentNo=s+1, shift =DT_MCD),
                                     forceAddZeros=True)


            UPHw_Tt_hf[k].exportGlb(Status.int.systemModel.processObject[k].UPHwGlb)


            UPHwUsed_Tt = [HeatFlow_t() for dummy in processes]    #create just a dummy 0 array. no longer necessary once ModuleHR creates this.
                        
            if Status.SimTool != "Plus":
                dlg.update(90.0 * (k+1) / NK)
            
        
        if Status.SimTool != "Plus":
            Status.int.UPH_Tt = UPH_Tt
            Status.int.UPHw_Tt = UPHw_Tt
            Status.int.UPHwReq_Tt = UPHwReq_Tt

            Status.int.UPHTotal_Tt = UPHTotal_Tt   
            Status.int.UPHwTotal_Tt = UPHwTotal_Tt
            Status.int.UPHwResTotal_Tt = UPHwResTotal_Tt
            Status.int.UPCTotal_Tt = UPCTotal_Tt
            Status.int.UPCwTotal_Tt = UPCwTotal_Tt

            #create space for receiving results of heat recovery calculations
            Status.int.QHXProc_Tt = [Status.int.createQ_Tt() for dummy in processes]
            Status.int.UPCProc_Tt = [Status.int.createQ_Tt() for dummy in processes]
            Status.int.UPHProc_Tt = [Status.int.createQ_Tt() for dummy in processes]        
                           
        # set status-flag (required BEFORE call to runHRModule !!!)
        Status.int.cascadeUpdateLevel = 0 #indicates that demand profile is created !!!
        Status.processData.outOfDate = False
        
        # estimate pipe heat flows (with 0 heat recovery to processes)

        pipes = Status.prj.getPipes()
        Status.int.QWHPipeUsed_Tt = [HeatFlow_t() for pipe in pipes]            
        Status.int.QHXPipe_Tt = [HeatFlow_t() for pipe in pipes]

        Status.int.UPHChanged = [True] * NK
        Status.int.UPCChanged = [True] * NK
        Status.int.UPHProcChanged =  [True] * NK
        Status.int.UPCProcChanged =  [True] * NK

        if Status.SimTool != "Plus":
            clearHXResults()    #initial settings necessary as input for calculateUSHCPipe
            calculateUSHCPipe()
            equipments.estimateEquipmentWasteHeat()
                
        #   get waste heat from WHEEs

        whees = Status.int.systemModel.whee
        if Status.SimTool != "Plus":
            Status.int.QWHEE_Tt = createQ_Tt()
            Status.int.QWHEE_T = createQ_T()
            Status.int.QWHEE_t = createQ_T()
        Status.int.QWHEE = 0
        
        Status.int.QWHEE_Tt_hf = []

        for whee in whees:
            n = whee.WHEENo - 1
            schedule = Status.schedules.WHEESchedules[n]
           
            # temperature distribution of waste heat. assumed as fix
            Tout = whee.WHEETOutlet
            if Tout == None:
                logWarning("Processes: WHEE outlet temperature not specified. 70 C assumed")
                Tout = 70
            Tret = max(Tout - 20, 0.0) # by default Delta_T = 20 assumed
            dTtot = max(Tout - Tret, 1.e-10)

            try:
                QWHEE = whee.QWHEE * whee.HPerDayWHEE * whee.NDaysWHEE
            except TypeError:
                QWHEE = 0.0
                
            Status.int.QWHEE += QWHEE

            if Status.SimTool != "Plus":
                QWHEE_T = createQ_T()
                for iT in range(Status.NT + 2):
                    temp = Status.int.T[iT]
                    dT = max(temp - Tret, 0.)
                    QWHEE_T[iT] = QWHEE * max(0.0, 1.0 - dT / dTtot)
                
                Status.int.QWHEE_T += QWHEE_T
                
                f_Tt = np.array([schedule.fav] * (NT + 2)) #NT + 1 + 1 -> additional value for T > Tmax
                QWHEE_Tt = QWHEE_T.reshape((Status.NT + 2, 1)) * f_Tt
                Status.int.QWHEE_Tt += QWHEE_Tt

            if QWHEE > 0:
                Status.int.QWHEE_Tt_hf.append(HeatFlow_t(fluid_id=0, T0=Tret, T1=Tout, Q_t = QWHEE*schedule.fav))
            else:
                Status.int.QWHEE_Tt_hf.append(HeatFlow_t())
            Status.int.QWHEE_Tt_hf[n].exportGlb(Status.int.systemModel.wheeObject[n].QWHEEGlb)
##        
##        Status.int.QWHEE_t = Status.int.QWHEE_Tt[0].copy()
##        Status.int.QWHEE_T = Status.int.calcQ_T(Status.int.QWHEE_Tt)
##
##        # set flags to True a second time is necessary, as previously has been reset by calculateUSHPipe
##        Status.int.UPHChanged = [True] * NK
##        Status.int.UPCChanged = [True] * NK
##        Status.int.UPHProcChanged =  [True] * NK
##        Status.int.UPCProcChanged =  [True] * NK
##

        if Status.SimTool != "Plus":

#TODO: check - commented out during V3-works on 14.9.18. should no longer be necessary. To be tested !!!
##            buildHXNet()

            dlg.Destroy()
            
            #   now run HR module for calculating heat recovery and effective demand at pipe entry
            Status.mod.moduleHR.runHRModule()

            logTrack("Aggregate demand = %s" % str(Status.int.QD_T))

        
        
 
#------------------------------------------------------------------------------		
    def changeInProcess(self):
#------------------------------------------------------------------------------		
#       function that is called from anywhere in the tool, whenever process data
#       of the present alternative are changed
#------------------------------------------------------------------------------		
        self.outOfDate = True
        self.outOfDateYearly = True
        Status.schedules.outOfDate = True
        Status.int.changeInCascade(0)
        Status.int.heatRecoveryOutOfDate = True
        logTrack("Processes (changeInProcess): process data changed")
        
    def createTempDistHX(self, PT, T0, mode):
        return self._createTempDist(PT, T0, mode)
        
    def _createTempDist(self, PT, T0=None, mode="heating"):
        """
        Create a default temperature distribution for a given temperature PT. 
        
        Create a linear ASCENDING temperature distribution between T0 and T,
        or a step distribution, if T0 = None for heating processes and the
        equivalent DESCENDING temperature distribution for cooling processes.
        
        :param PT: Target temperature 
        :param T0: starting temperature
        :param mode: either "heating" or "cooling"
        :returns: a temperature distribution numpy array with elements 
        as formatted in modules.interfaces.Interfaces.T in the "heating"
        case or modules.interfaces.Interfaces.TC in the "cooling" case.
        
        """
        NT = Status.NT
        dist = createQ_T()
        if mode == "heating":
            T = Status.int.T
            if T0 == None or T0 >= PT:
                for iT in range(NT + 1):
                    if PT > T[iT]:
                        dist[iT] = 0.0
                    else:
                        dist[iT] = 1.0
                dist[-1] = 1.0    #last entry for T > Tmax
            else:
                for iT in range(NT + 1):
                    dist[iT] = cutInterval(T[iT], T0, PT)
                dist[-1] = 1.0    #last entry for T > Tmax
        elif mode == "cooling":
            T = Status.int.TC
            if T0 == None or T0 <= PT:
                dist[0] = 1.0    #First entry for T < Tmin
                for iT in range(1, NT + 2):
                    if PT < T[iT]:
                        dist[iT] = 0.0
                    else:
                        dist[iT] = 1.0
            else:
                dist[0] = 1.0    #First entry for T < Tmin
                for iT in range(1, NT + 1):
                    dist[iT] = 1 - cutInterval(T[iT], T0, PT)
        else:
            raise ValueError("Either 'heating' or 'cooling'.")
        return dist

    def createInvTempDistLat(self, PT, h, fluidID, T0=None, mode="heating"):
        """Create a piecewise linear DESCENDING temperature distribution
        starting with 1 at PT and ending with 0 at T0. If the 
        condensation point of the fluid is between T0 and PT the 
        distribution makes the sensible jump. An equivalent distribution
        with ASCENDING temperature is created if it is not heat that is lost, 
        but the medium is cooled for the process, 'cold loss' so to speak.
        If T0 is not given, the distribution simply jumps at PT.
           
        :param PT: lower of the two temperatures
        :param h: specific enthalpy
        :param fluidID: the id of the used medium
        :param T0: the higher temperature 
        :param recovering: whether "heating" or "cooling" is lost.
        :returns: a temperature distribution numpy array with elements
        as formatted in modules.interfaces.Interfaces.T in the "heating" 
        case or modules.interfaces.Interfaces.TC in the "cooling" case.
        
        """
        fluid = Fluid(fluidID)
        x = fluid.steamFraction(h, T=PT)
        x0 = fluid.steamFraction(None, T=T0)

        if x is None:
            x = 0

        if x0 is None:
            x0 = x

        if x > x0:
            fX = (x - x0) * fluid.hL
        else:
            fX = 0

        if PT > fluid.TCond:
            fV = fluid.cpG * (PT - fluid.TCond)
        else:
            fV = 0

        if x0 > 0:
            fL = 0
        else:
            if T0 is None:
                fL = 0
            else:
                fL = (min(PT, fluid.TCond) - T0) * fluid.cp

        f = fL + fX + fV
        if f <= 0 or f is None:
            f = 1.0
            fL = 1.0
            logDebug("Processes (createInvTempDistLat): f=0 error with PT %s T0 %s h %s x %s x0 %s" % \
                     (PT, T0, h, x, x0))

        distV = self._createTempDist(PT, T0=fluid.TCond, mode=mode)
        distX = self._createTempDist(fluid.TCond, T0=None, mode=mode)
        distL = self._createTempDist(min(PT, fluid.TCond), T0=T0, mode=mode)
        if mode=="heating":
            dist = (1.0 - (fV * distV + fX * distX + fL * distL) / f)
        else:
            dist = ((fV * distV + fX * distX + fL * distL) / f)
        return dist

