# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	PROJECT
#			
#------------------------------------------------------------------------------
#			
#	Definition of objects defining and handling a project (industry)
#       - projects (create, delete, ...)
#       - alternatives (create, delete, ...)
#       - access to SQL data tables (fuels, equipments, pipes, processes, ...)
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import database.dbio as dbio
from datetime import date
from modules.messageLogger import *
from globals import logger, Globals
from wx import ID_YES
from constants import *
from GUI.status import Status
from modules.schedules.schedules import Schedules
from modules.schedules.periodSchedule import deleteProcessPeriodsFromDB
from modules.processes import Processes
import modules.ccheck.ccTracker as ccTracker
##from einstein.GUI.GUITools import check
from database.dbio import check

from modules.weatherdata import WeatherData
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUnicode as U, toUTF8, reorder, noneCheck
from warnings import warn
from modules.heatPump.moduleHP import DEFAULTSAUTODESIGNHP
from modules.chp.moduleCHP import DEFAULTSAUTODESIGNCHP
from modules.boiler.moduleBB import DEFAULTSAUTODESIGNBB
from modules.solar.moduleST import DEFAULTSAUTODESIGNST
from model.systemModel import SystemModel

from collections import defaultdict

class InconsistentDataBaseError(Exception):
    """BUG: Data base structure is inconsistent."""
    pass

#------------------------------------------------------------------------------		
def copySQLRows(table, query, keyID=None, keyPar=None, valPar=None, sDict=None):
    """
    Helper function for copying full rows in SQL Tables.
    
    :param table: name of the table
    :param query: the part after the WHERE in the SQL clause
    :param keyID: name of the primary key in the table [OBSOLETE. NOW ALWAYS ASSUMED TO "id"]
    :param keyPar: name of a column with entries that are to be changed in the
    copy
    :param valPar: new value for column keyPar
    :returns: A dict with oldId: newID pairs
    """
    rows = dbio.selectByQuery(table, query)

    IDDict = {}
    newID = None
        
    for row in rows:
        oldID = row.id
        newID = dbio.copyRow(table, row)
        IDDict[oldID] = newID

#..............................................................................
# changes one parameter within the row (usually AlternativeNo or ProjectID

        if valPar is not None:
            newRow = dbio.get(table, id=newID)
            newRow.set(**{keyPar: valPar})

    # automatically collect information on rows copied
    if sDict is not None:
        if table in sDict.keys():
            sDict[table].update(IDDict)
        else:
            sDict.update({table: IDDict})
        
    return IDDict

def deleteSQLRows(table, query):
    """ auxiliary function for copying full rows in SQL Tables
    """
    dbio.deleteMany(table, query)
    
def cleanUpSQLRows(table, query, maxANo=None, mainField=None):
    """ auxiliary function for copying full rows in SQL Tables
       eliminates rows with erroneous ANO
    """	

##    checked = False
##    ctr = 0
####    rows = table.sql_select(query)
####    rows = dbio.selectMany(table, query)
####    maxCtr = len(rows)
##        
##
##    while (checked == False and ctr < maxCtr):
##        ctr += 1
##        
##        checked = True
        
##        rows = table.sql_select(query)
##        rows = dbio.selectMany(table, query)
##        
##        n = len(rows)
##        for i in range(n):

# check here if there are rows with an AlternativeProposalNo out of range [-1:maxANo]

##            if maxANo is not None:
##                if rows[i].AlternativeProposalNo > maxANo or rows[i].AlternativeProposalNo < -1:
##                    logDebug(_("Project (cleanUpSQLRows): entry with ANo %s > maxANo %s deleted -> \n%s") % \
##                             (rows[i].AlternativeProposalNo, maxANo, rows[i]))
##                    rows[i].delete()    #delete from 0 and update rows -> strange solution. see comment in deleteSQLRows
##                    checked = False
##                    Status.SQL.commit()
##                    break
    if maxANo is not None:
        queryANo = query + " AND (AlternativeProposalNo > %s OR AlternativeProposalNo < -1)"%maxANo
        dbio.deleteMany(table, queryANo)
                
# check here if the mandatory main entry field of the row is defined (avoid crashes due to nones in name-fields, etc.)

    if mainField is not None:
        queryMainField = query + "AND %s = 'NULL'"%mainField
        dbio.deleteMany(table, queryMainField)
            
##                if rows[i][mainField] is None:
##                
##                    logDebug(_("Project (cleanUpSQLRows): corrupt entry with ANo %s deleted\n%s") % \
##                             (rows[i].AlternativeProposalNo, rows[i]))
##                    rows[i].delete()    #delete from 0 and update rows -> strange solution. see comment in deleteSQLRows
##                    checked = False
##                    Status.SQL.commit()
##                    break                    

#------------------------------------------------------------------------------		
def shiftANoInSQLRows(table, query, shift):
#------------------------------------------------------------------------------		
#       auxiliary function for changing a value in rows
#------------------------------------------------------------------------------		
    rows = dbio.selectByQuery(table, query)

    for row in rows:
        newval = row.AlternativeProposalNo + shift
        row.AlternativeProposalNo = newval
                
#==============================================================================
#       PROJECT CLASS
#==============================================================================

class Project(object):

#------------------------------------------------------------------------------		
    def __init__(self):
        """
        initialisation. does the initialisation of basic parameters in Status
        """

        Status.prj = self
        
        # get last tool settings from table STOOLS (last project opened, etc.)

        self.getLastToolSettings()

        # create instance of Schedules and Processes (does no calculations)

        Status.schedules = Schedules()
        Status.processData = Processes()

        # set active project to the last one opened
##        try:
        self.setActiveProject(Status.PId)
##        except:
##            showWarning(_U("Couldn't open last project [%s], sorry."))

        
    def editAllowed(self, warn=True):
        """Return True if the data may be modified in the current state, 
        inform the user and return False if not.
        
        """
        if Status.ANo == 0:
            if warn:
                showWarning(_U("You cannot modify the checked data set for the present state of the project\n") + 
                            _U("These data are generated automatically"))
            return False
        elif (Status.ANo < 0 and Status.StatusCC > 0):
            if warn:
                showWarning(_U("You cannot modify the raw data set for the present state of the project after the data has been checked.\n") + 
                            _U("Reset the project state to unchecked first"))
            return False
        else:   
            return True
        
#------------------------------------------------------------------------------		
# HANDLING OF ALTERNATIVE PROPOSALS (= REPEATED PART OF A PROJECT)
#------------------------------------------------------------------------------		

    def createNewAlternative(self, originalANo, shortName, description):
        """Create a new alternative by copying all entries of the 
        original ANo and set it as the current alternative.
              
        """
        if Status.PId < 0: return
        
        Status.NoOfAlternatives += 1
        ANo = Status.NoOfAlternatives

        sproject = self.getProjectStatus()
        sproject.NoOfAlternatives = Status.NoOfAlternatives
        self.projectChanged()

        logTrack("Project (createNewAlternative) - project %s, copying from %s %s" % (Status.PId, originalANo, ANo))

#..............................................................................
# copying Q- and corresponding C-Tables

        sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, originalANo)

        #additional query necessary for old tables who still have Questionnaire_id instead of ProjectID
        sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, originalANo)

        sDict = self.copySQLData(sqlQuery, sqlQueryQ, ANo=ANo, originalANo=originalANo)
        self.restoreLinks(Status.PId, sDict)
# rename alternative

##        sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, ANo)
##        aa = DB.salternatives.sql_select(sqlQuery)
        a = self.getAlternative(ANo=ANo)
        if a is not None:
            a.ShortName = check(shortName)
            if description is None:
                description = "---"
            a.Description = check(description)

            dbio.commit()
            self.setActiveAlternative(ANo)
        else:
            logError(_("ERROR: possibly corrupt project data\n cannot copy alternative (PId: %s ANo: %s)") % (Status.PId, ANo))

# reset calculation flags for energy and economic performance in case of copying from present state
        if originalANo <= 0:
            self.setStatus("Energy", 0, ANo=ANo)
            self.setStatus("ECO", 0, ANo=ANo)
            if Status.HRTool == HRMODES[1]:
                logWarning(_("You are in ESTIMATE-Mode for heat recovery calculations.\nFor all alternatives except present state default heat recovery will be assumed !"))
#------------------------------------------------------------------------------

    def copySQLData(self, sqlQuery, sqlQueryQ, PId=None, ANo=None, originalANo=None):
        """
        common function to be used in both copying projects and alternatives
        ANo: NEW number of new alternative; None for copying projects
        PId: NEW project id; None for copying alternatives only
        """
        assert(ANo is None or PId is None)
        
        if PId is None:  # copying an alternative
            newProject = False
            PId = Status.PId
            cKey = "AlternativeProposalNo"
            cKeyQ = "AlternativeProposalNo"
            cNo = ANo
            originalANo = noneCheck(originalANo, Status.ANo)
        else:           # copying a project
            newProject = True
            cKey = "ProjectID"
            cKeyQ = "Questionnaire_id"
            cNo = PId
        
        sDict = defaultdict(lambda: {})     # default collection returning empty dictionary for a non-existing key

        # 1. copy rows from all tables which can be identified directly by PId/ANo            
        if newProject:
            copySQLRows("sproject", sqlQuery, "SProject_ID", cKey, cNo, sDict=sDict)
            copySQLRows("qrenewables", sqlQueryQ, "QRenewables_ID", cKeyQ, cNo, sDict=sDict)
            copySQLRows("qsurfarea", sqlQuery, "id", cKey, cNo, sDict=sDict)
            copySQLRows("qmdatafiles", sqlQuery, "id", cKey, cNo, sDict=sDict)


        copySQLRows("salternatives", sqlQuery, "SAlternative_ID", cKey, cNo, sDict=sDict)    
        copySQLRows("cgeneraldata", sqlQueryQ, "CGeneralData_ID", cKeyQ, cNo, sDict=sDict)
        
        copySQLRows("qelectricity", sqlQueryQ, "QElectricity_ID", cKeyQ, cNo, sDict=sDict)
        copySQLRows("qfuel", sqlQueryQ, "QFuel_ID", cKeyQ, cNo, sDict)
        copySQLRows("qgenerationhc", sqlQueryQ, "QGenerationHC_ID", cKeyQ, cNo, sDict=sDict)
        copySQLRows("qdistributionhc", sqlQueryQ, "QDistributionHC_ID", cKeyQ, cNo, sDict)
        copySQLRows("qprocessdata", sqlQueryQ, "QProcessData_ID", cKeyQ, cNo, sDict=sDict)
        copySQLRows("qheatexchanger", sqlQuery, "QHeatExchanger_ID", cKey, cNo, sDict=sDict)
        copySQLRows("qwasteheatelequip", sqlQuery, "QWasteHeatElEquip_ID", cKey, cNo, sDict=sDict)

        copySQLRows("qproduct", sqlQueryQ, "QProduct_ID", cKeyQ, cNo, sDict=sDict)
        copySQLRows("qbuildings", sqlQueryQ,"QBuildings_ID", cKeyQ, cNo, sDict=sDict)

        copySQLRows("heatexchanger_pinchstream", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("pinchstream", sqlQuery, "id", cKey, cNo, sDict=sDict)

        copySQLRows("qflow", sqlQuery, "id", cKey, cNo, sDict=sDict)

        copySQLRows("tcacontingencies", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("tcadetailedopcost", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("tcadetailedrevenue", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("tcaenergy", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("tcainvestments", sqlQuery, "id", cKey, cNo, sDict=sDict)
        copySQLRows("tcanonreoccuringcosts", sqlQuery, cKey, cNo, sDict=sDict)
        copySQLRows("tcageneraldata", sqlQuery, "IDTca", cKey, cNo, sDict=sDict)



        copySQLRows("uheatpump", sqlQueryQ, "UHeatPump_ID", cKeyQ, cNo, sDict=sDict)
            


        # 2. now copy rows which are references indirectly

        # copying process in- and outstreams (-> qprocessdata)
        inStreamIds = self._getStreamsForProcesses("process_streams_in",
                                                   sDict["qprocessdata"].keys(),
                                                   "streams_in_id")
        self._copyStreams(inStreamIds, sDict["qprocessdata"],
                          "process_streams_in", "streams_in_id",
                          "streams_in", PId, ANo=ANo, sDict=sDict)
        outStreamIds = self._getStreamsForProcesses("process_streams_out",
                                                    sDict["qprocessdata"].keys(),
                                                    "streams_out_id")
        self._copyStreams(outStreamIds, sDict["qprocessdata"],
                          "process_streams_out", "streams_out_id",
                          "streams_out", PId, ANo=ANo, sDict=sDict)

        # copying nodes (-> flows)
        sDict["qnode"] = self.copyNodes(PId, originalANo=originalANo)
        
        # copying data sets (-> qmdatafiles)
        for oldDataFileId in sDict["qmdatafiles"].keys():
            copySQLRows("qmdatasets", "DataFileId = %d" % oldDataFileId, "id", sDict=sDict)
        # copying project detailed schedule tables (-> qprocessdata)
        for oldQProcessdataId in sDict["qprocessdata"].keys():
            copiedProcessPeriods = copySQLRows("process_periods", "qprocessdata_QProcessData_Id = %d" % oldQProcessdataId, "id", sDict=sDict)
            for oldProcessPeriodsId in copiedProcessPeriods.keys():
                copySQLRows("process_period_profiles", "process_periods_id = %d" % oldProcessPeriodsId, "id", sDict=sDict)

            copySQLRows("qcombschedules", "ProcessId = %d"%oldQProcessdataId, "id", sDict=sDict)
        for oldId in sDict["qgenerationhc"].keys():
            copiedEquipmentPeriods = copySQLRows("process_periods", "qprocessdata_QProcessData_Id = %d" % (-oldId), "id", sDict=sDict)
            for oldProcessPeriodsId in copiedEquipmentPeriods.keys():
                copySQLRows("process_period_profiles", "process_periods_id = %d" % oldProcessPeriodsId, "id", sDict=sDict)

        # copying HX control table
        for oldHXId in sDict["qheatexchanger"].keys():
            copySQLRows("qhxontrol", "HXId = %d" % oldHXId, sDict=sDict)

        dbio.commit()
        
        return sDict
    
#------------------------------------------------------------------------------
    def deleteAlternative(self, ANo):
#------------------------------------------------------------------------------
#   deletes all entries for the original ANo
#------------------------------------------------------------------------------

        if ANo > Status.NoOfAlternatives or ANo == -1:
            
            logWarning(_U("Project (deleteAlternative) - project %s, cannot delete alternative %s") % (Status.PId, ANo))
            return - 1
        
        finalAlternative = Status.FinalAlternative
        if finalAlternative is not None:
            if finalAlternative > ANo:
                finalAlternative -= 1
            elif finalAlternative == ANo:
                finalAlternative = None
            if finalAlternative != Status.FinalAlternative: self.setFinalAlternative(finalAlternative)
        
        logTrack("Project (deleteAlternative) - project %s, deleting alternative %s" % (Status.PId, ANo))

#..............................................................................
# deleting Q- and corresponding C-Tables

        sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, ANo)

        #additional query necessary for old tables who still have Questionnaire_id instead of ProjectID
        sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, ANo)
        
        processes = self.getProcesses(PId=Status.PId, ANo=ANo)
        processIds = [p.QProcessData_ID for p in processes]
        for pId in processIds:
            deleteProcessPeriodsFromDB(pId)
            dbio.delete("qcombschedules", ProcessId = pId)
            self.deleteProcess(pId)

        projectEquipment = dbio.select("qgenerationhc", Questionnaire_id = Status.PId, AlternativeProposalNo=ANo)
        for eq in projectEquipment:
            deleteProcessPeriodsFromDB(-eq.id)

        nodeIds = self.getNodeIDs(Status.PId, ANo=ANo)
        for nodeId in nodeIds:
            dbio.delete("qnode", id=nodeId)
        deleteSQLRows("qflow", sqlQuery)

        hxes = self.getHXes(PId=Status.PId, ANo=ANo)
        for hx in hxes:
            dbio.delete("qhxcontrol", HXId = hx.id)
        
        deleteSQLRows("salternatives", sqlQuery)
        deleteSQLRows("cgeneraldata", sqlQueryQ)
        deleteSQLRows("ebenchmarkgeneraldata", sqlQueryQ)
        deleteSQLRows("ebenchmarkprocessdata", sqlQueryQ)
        deleteSQLRows("ebenchmarkproductdata", sqlQueryQ)
        deleteSQLRows("qbuildings", sqlQueryQ)
        deleteSQLRows("qdistributionhc", sqlQueryQ)
        deleteSQLRows("qelectricity", sqlQueryQ)
        deleteSQLRows("qfuel", sqlQueryQ)
        deleteSQLRows("qgenerationhc", sqlQueryQ)
        deleteSQLRows("qheatexchanger", sqlQuery)
        deleteSQLRows("qproduct", sqlQueryQ)
#        deleteSQLRows("qrenewables,sqlQueryQ)
        deleteSQLRows("qwasteheatelequip", sqlQuery)
        deleteSQLRows("uheatpump", sqlQueryQ)
        deleteSQLRows("heatexchanger_pinchstream", sqlQuery)
        deleteSQLRows("pinchstream", sqlQuery)
        deleteSQLRows("tcageneraldata", sqlQuery)
        deleteSQLRows("tcacontingencies", sqlQuery)
        deleteSQLRows("tcadetailedopcost", sqlQuery)
        deleteSQLRows("tcadetailedrevenue", sqlQuery)
        deleteSQLRows("tcaenergy", sqlQuery)
        deleteSQLRows("tcainvestments", sqlQuery)
        deleteSQLRows("tcanonreoccuringcosts", sqlQuery)

        
        dbio.commit()

#..............................................................................
# changing ANos in all rows with ANo higher than the deleted one

        for i in range(ANo + 1, Status.NoOfAlternatives + 1):
            sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, i)

        #additional query necessary for old tables who still have Questionnaire_id instead of ProjectID
            sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s'" % (Status.PId, i)

            shiftANoInSQLRows("salternatives", sqlQuery, -1)
            shiftANoInSQLRows("cgeneraldata", sqlQueryQ, -1)
            
            shiftANoInSQLRows("ebenchmarkgeneraldata", sqlQueryQ, -1)
            shiftANoInSQLRows("ebenchmarkprocessdata", sqlQueryQ, -1)
            shiftANoInSQLRows("ebenchmarkproductdata", sqlQueryQ, -1)
            shiftANoInSQLRows("qbuildings", sqlQueryQ, -1)
            shiftANoInSQLRows("qdistributionhc", sqlQueryQ, -1)
            shiftANoInSQLRows("qelectricity", sqlQueryQ, -1)
            shiftANoInSQLRows("qflow", sqlQuery, -1)
            shiftANoInSQLRows("qfuel", sqlQueryQ, -1)
            shiftANoInSQLRows("qgenerationhc", sqlQueryQ, -1)
            shiftANoInSQLRows("qheatexchanger", sqlQuery, -1)
            shiftANoInSQLRows("qprocessdata", sqlQueryQ, -1)
            shiftANoInSQLRows("streams_in", sqlQuery, -1)
            shiftANoInSQLRows("streams_out", sqlQuery, -1)
            shiftANoInSQLRows("process_streams_in", sqlQuery, -1)
            shiftANoInSQLRows("process_streams_out", sqlQuery, -1)

            shiftANoInSQLRows("qproduct", sqlQueryQ, -1)
#            shiftANoInSQLRows("qrenewables,sqlQueryQ,-1)
            shiftANoInSQLRows("qwasteheatelequip", sqlQuery, -1)
            shiftANoInSQLRows("uheatpump", sqlQueryQ, -1)
            shiftANoInSQLRows("heatexchanger_pinchstream", sqlQuery, -1)
            shiftANoInSQLRows("pinchstream", sqlQuery, -1)
            shiftANoInSQLRows("tcageneraldata", sqlQuery, -1)
            shiftANoInSQLRows("tcacontingencies", sqlQuery, -1)
            shiftANoInSQLRows("tcadetailedopcost", sqlQuery, -1)
            shiftANoInSQLRows("tcadetailedrevenue", sqlQuery, -1)
            shiftANoInSQLRows("tcaenergy", sqlQuery, -1)
            shiftANoInSQLRows("tcainvestments", sqlQuery, -1)
            shiftANoInSQLRows("tcanonreoccuringcosts", sqlQuery, -1)


#..............................................................................
# book-keeping: reduce number of alernatives in the system

        Status.NoOfAlternatives -= 1
        ANo = Status.NoOfAlternatives
        self.setActiveAlternative(ANo)

        sproject = self.getProjectStatus()
        sproject.NoOfAlternatives = Status.NoOfAlternatives

        self.projectChanged()

        dbio.commit()

#------------------------------------------------------------------------------
    def getAlternativeList(self):
#------------------------------------------------------------------------------

        defaultList = [[-1, _("Present State (original)"),
                            _("original data as delivered in questionnaire"),
                            "--", "---", "---"],
                            [0, _("Present State (checked)"),
                            _("complete data set for present state after\ncross-checking and data estimation"),
                            u"E€", "---", "---"]]

        alternativeList = []
        if Status.PId is None or Status.PId <= 0:
            return alternativeList
        
        aa = self.getAlternatives()
##        for ANo in range(-1, Status.NoOfAlternatives + 1):
##            if DBSERVER == "SQLITE":
##                aa = list(dbio.salternatives.selectBy(ProjectID=Status.PId, AlternativeProposalNo=ANo))
##            else:
##                aa = Status.DB.salternatives.ProjectID[Status.PId].AlternativeProposalNo[ANo]
##            if len(aa) > 0:
##                a = aa[0]
        for a in aa:
            ANo = a.AlternativeProposalNo
                
            stat = ("E" if a.StatusEnergy==1 else "-") + ("€" if a.StatusECO==1 else "-")
            if ANo == Status.FinalAlternative:
                stat += "\nselected"

#..............................................................................
# get main energy data if available

            if a.StatusEnergy > 0:
##                cc = Status.DB.cgeneraldata.Questionnaire_id[Status.PId].AlternativeProposalNo[ANo]
                cgeneraldata = self.getGeneralData(PId=Status.PId, ANo=ANo)
                if cgeneraldata is not None:
                    try: PEC = float(cgeneraldata.PEC)
                    except:
                        logTrack("Project (getAlternativeList): no PEC available")
                        PEC = 0.0

                    try: EnergySystemCost = float(cgeneraldata.EnergySystemCost)
                    except:
                        logTrack("Project (getAlternativeList): no EnergySystemCost available")
                        EnergySystemCost = 0.0
                    
                    if a.Description is None: a.Description = "---"
                    alternativeList.append([a.AlternativeProposalNo,
                                            a.ShortName,
                                            a.Description,
                                            stat, PEC, EnergySystemCost])
                else:
                    logError(_("Corrupt data in data base. no entry in cgeneraldata for ANo = %s") % ANo)
                    if ANo in [-1, 0]:
                        alternativeList.append(defaultList[ANo + 1])
            else:
                if a.Description is None: a.Description = "---"
                alternativeList.append([a.AlternativeProposalNo,
                                        a.ShortName,
                                        a.Description,
                                        stat, '---', '---'])
##        else:
##            logError(_("Corrupt data in data base. no entry in salternative for ANo = %s") % ANo)
##            if ANo in [-1, 0]:
##                alternativeList.append(defaultList[ANo + 1])

        return alternativeList
            
#------------------------------------------------------------------------------
    def setActiveAlternative(self, n, checked=False):
#------------------------------------------------------------------------------
        if (n < -1) or n > Status.NoOfAlternatives:
            logTrack("Project (setActiveAlternative): alternative number out of range [%s,%s]" % (-1, Status.NoOfAlternatives))
            return

        changeAlternative = (n <> Status.ANo)
        if changeAlternative:
            
            Status.int.clearHXData()    # needed for ... (please add a comment)
            Status.int.systemModel = None
        

            sproject = self.getProjectStatus()
            if sproject is not None:
                sproject.ActiveAlternative = n
                dbio.commit()
                Status.ANo = n
            else:
                logTrack("Project (setActiveAlternative): error writing active alternative no. to project table PId %s" % \
                         Status.PId)

        salternative = self.getAlternative(ANo=n)
        
        if salternative is not None:
            Status.ActiveAlternativeName = salternative.ShortName
            self.getStatus()
            logTrack("Project (setActiveAlternative): PId = %s ANo = %s StatusCC = %s" % \
              (Status.PId, Status.ANo, Status.StatusCC))

            if checked == True:
                Status.prj.setStatus("CC")                    
                
            if not Status.int.systemModel: Status.int.systemModel = SystemModel()
            Status.int.systemModel.build()
##            except Exception, e:
##                logWarning("Cannot build system model when loading the project")
##                logTrack("Project.setActiveAlternative. Error building Model for project %s\n%s"%(Status.PId, e))

            Status.schedules.outOfDate = True
            Status.processData.outOfDate = True
            Status.processData.outOfDateYearly = True
            Status.int.changeInCascade(0, keepEnergyData=True)

            if Status.StatusCC > 0 and n > -1:
                Status.processData.createYearlyDemand()
            
            Status.prj.setStatus("StatusECO", salternative.StatusECO)
##            Status.int.StreamGen = None # reset the basic hx simulation data

            try:
                Status.mod.moduleCC.reset()
            except:
                pass

            # if there is a GUI, show new alternative
            if Globals.GUI: Globals.GUI.updateDrawings()
##            try:
##                Status.main.panelinfo.update()
##            except:
##                logDebug("problems updating panelinfo")
##
##            try:
##                Status.main.showMainMenuAlternatives()
##            except:
##                logDebug("problems updating main menu alternatives")

##            try:
##            except:
##                logDebug("problems resetting moduleCC")
        else:
            logTrack("Project (setActiveAlternative): error trying to set alternative to %s" % n)
            Status.StatusCC = None
            Status.StatusQ = None
            Status.StatusECO = None
            Status.int.systemModel = None
            
    def setAlternativeDescription(self, newName=None, newDescription=None):
        """
        updates the name and the description of an alternative
        checks in order to avoid duplicate names

        :params unewName: new alternative name (unicode)
        :parmas unewDescription: new description (unicode)

        if None is input, attributes are unchanged
              
        """
        alternatives = self.getAlternatives()
            
        index = Status.ANo + 1

        nameList = [a.ShortName for a in alternatives]
        nameList.pop(index)

        newName = noneCheck(newName, _("Alternative %s"%Status.ANo))
        newDescription = noneCheck(newDescription, "")

        a = alternatives[index]
        if a is not None:
            baseName = newName[:40]
            newName = baseName
            i = 1
            while newName in nameList and i < 100:
                newName = baseName + " (%2d)"%i
                i += 1

            a.ShortName = newName[:45]

            Status.ActiveAlternativeName = newName[:45]
            if Status.ANo==Status.FinalAlternative:
                Status.FinalAlternativeName = Status.ActiveAlternativeName

            a.Description = newDescription[:1000]

            dbio.commit()

    def getAId(self, PId=None, ANo=None):
        PId = Status.PId if PId is None else PId
        if ANo is None: ANo = Status.ANo
        salt = dbio.get("salternatives", ProjectID=PId, AlternativeProposalNo=ANo)
        Status.AId = None if salt is None else salt.id
                    
                    
        return Status.AId

#------------------------------------------------------------------------------
    def copyQuestionnaire(self):
#------------------------------------------------------------------------------
#   Creates a 1:1 copy from alternative -1 to alternative 0
#------------------------------------------------------------------------------

#first remove everything with ANo > -1

        self.cleanUpProject(Status.PId) #use the opportunity to clean-up erroneous entries with ANo > NoOfAlternatives
        
        n = Status.NoOfAlternatives
        for ANo in range(n, -1, -1):
            self.deleteAlternative(ANo)
        self.setStatus("Q")
        self.setStatus("CC", 0)
        self.createNewAlternative(-1, _U("Present State (checked)"), \
                                  _U("complete data set for present state after\n cross-checking and data estimation"))
        self.setActiveAlternative(-1)
        
#------------------------------------------------------------------------------		
#------------------------------------------------------------------------------
    def setFinalAlternative(self, ANo):
#------------------------------------------------------------------------------
        finalANo = ANo
        if finalANo is None:
            self.setStatus("CS", 0)
            Status.FinalAlternativeName = "---"
        else:
            if ANo < 0:
                finalANo = 0     
            elif finalANo > Status.NoOfAlternatives:
                logDebug("Project (setFinalAlternative): error in no. of selected alternative [%s]" % finalANo)
                return

            Status.FinalAlternativeName = "---"

            self.setStatus("CS")

            a = self.getAlternative(ANo=finalANo)
##            aa = Status.DB.salternatives.ProjectID[Status.PId].\
##                 AlternativeProposalNo[finalANo]
##            if len(aa) > 0:
            if a is not None:
                name = a.ShortName
                if name is not None:
                    Status.FinalAlternativeName = unicode(name, "utf-8")
                else:
                    logDebug("Project (setFinalAlternative): no name specified for alternative no. [%s]" % finalANo)
                    Status.FinalAlternativeName = "---"
            else:
                logDebug("Project (setFinalAlternative): no entry in table salternative for ANo [%s]" % finalANo)
                Status.FinalAlternativeName = "---"

##        sprojects = Status.DB.sproject.ProjectID[Status.PId]
        sproject = dbio.get("sproject", ProjectID=Status.PId)
        if sproject is not None:
            if finalANo != sproject.FinalAlternative:
                sproject.FinalAlternative = check(finalANo)
    ##            Status.SQL.commit()
                dbio.commit()
            
        Status.FinalAlternative = finalANo
            
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------		
# HANDLING OF PROJECTS (CREATE,COPY,DELETE,INIT ...)
#------------------------------------------------------------------------------		

    def getProjectList(self):
        """
        returns a list with the names of the projects. used in panelQ0
        """

        projectList = []
        projectTables = dbio.select("sproject")
        for n in projectTables:
            if n.ProjectName is None:
                n.ProjectName = "(no name)"
            projectList.append(n.ProjectName)

        return projectList

    def getProjectsWithName(self, name):
        return dbio.select("sproject", ProjectName = name)
    
    def getQuestionnaire(self, PId = None):
        """ gets the table questionnaire """
        if PId is None: PId = Status.PId
        return dbio.get("questionnaire", id=PId)
        
    def getGeneralData(self, PId = None, ANo = None):
        """ delivers access to the cgeneraldata - Table of a given alternative
        For ANo = None the currently active alternative is returned
        
        """
        if PId is None: PId = Status.PId
        if ANo is None: ANo = Status.ANo

        return dbio.get("cgeneraldata", Questionnaire_id=PId, AlternativeProposalNo=ANo)
        
    def getProjectData(self, PId = None):
        """ returns the data in tables questionnaire and cgenerationhc
        """
        
        if PId is None:
            PId = Status.PId
            ANo = Status.ANo
        else:
            ANo = -1 #return by default original state data. Call with PId != None is only used for access to questionnaire table
        
        projectData = self.getQuestionnaire(PId)
        generalData = self.getGeneralData(PId=PId, ANo=ANo)

        return (projectData, generalData)

    def getProjectStatus(self, PId=None):
        """ gets data contained in table sproject
        """
        if PId is None: PId = Status.PId
        return dbio.get("sproject", ProjectID=PId)
        
    def getAlternatives(self, PId=None):
        """ gets data contained in table salternatives
        """
        if PId is None: PId = Status.PId
        return sorted(dbio.select("salternatives", ProjectID=PId), key=lambda s: s.AlternativeProposalNo)
        
    def getAlternative(self, PId=None, ANo=None):
        """ gets one Alternative
        """
        if PId is None: PId = Status.PId
        if ANo is None: ANo = Status.ANo
        return dbio.get("salternatives", ProjectID=PId, AlternativeProposalNo=ANo)
        
    def getProjectID(self, name):
        """
        returns the ID of a project as a function of the name
        """
        project = dbio.get("sproject", ProjectName = check(name))
        if project is not None:
            return project.ProjectID   
        else:
            logTrack("Project (getProjectID): project name %s unknown" % name)
            return None
        
    def getProjectName(self, PId):
        """Return the name of the project with the given PId, None if unknown"""
        project = dbio.get("sproject", ProjectID=PId)
        
        if project is not None:
            return project.ProjectName
        else:
            logTrack("Project (getProjectID): project id %s unknown" % PId)
            return None

#------------------------------------------------------------------------------
    def setActiveProject(self, PId, name=None, cleanUp=True):
        """
        sets the presently active project no.
        :params PId: If PId <= 0 and no name is given, then all projects are deselected
        :params name: project name
        """
        
        if name is not None:    #if a name is given, overwrite ID
            PId = self.getProjectID(name)

        if (PId <= 0 or (PId is None)):
            Status.PId = -1
            Status.ActiveProjectName = '---'
            Status.ActiveCompanyName = '---'
            Status.ActiveProjectDescription = '---'
            self.setFinalAlternative(None)
            Status.NoOfAlternatives = -1
            Status.WeatherData = None
            
            logTrack("Project (setActiveProject): no project selected")

        else:
            sproject = self.getProjectStatus(PId=PId)
            questionnaire = self.getQuestionnaire(PId=PId)

            if sproject is not None and questionnaire is not None:
                Status.ActiveProjectName = sproject.ProjectName
                
                if questionnaire.Name is None: questionnaire.Name = ""  #security item for avoiding crashes if questionnaire.Name is not set
                Status.ActiveCompanyName = questionnaire.Name

                if questionnaire.DescripIndustry is not None:
                    Status.ActiveProjectDescription = questionnaire.DescripIndustry
                else:
                    Status.ActiveProjectDescription = ""

                if cleanUp: self.cleanUpProject(PId)  

                Status.NoOfAlternatives = sproject.NoOfAlternatives
                Status.ANo = sproject.ActiveAlternative
                Status.FinalAlternative = sproject.FinalAlternative
                
                logTrack("Project (setActiveProject): number of alternatives in the project %s" % Status.NoOfAlternatives)
                logTrack("Project (setActiveProject): active alternative %s" % Status.ANo)

                Status.PId = PId

                self.getProjectUnits(sproject.UnitsReport)    
 
            else:
                Status.ActiveProjectName = '---'
                Status.ActiveProjectDescription = ''
                
                Status.NoOfAlternatives = -1
                Status.ANo = -1
                Status.FinalAlternative = None
                logTrack("Project (setActiveProject): could not open project no. %s" % PId)

                Status.PId = None

                
            ccTracker.conflict.reset()
            ccTracker.screen.reset()
            

            try:
                stool = dbio.get("stool")
                stool.ActiveProject = PId

                logTrack("Project (setActiveProject): active project set to %s" % PId)
            except:
                logTrack("Project (setActiveProject): error writing new active project ID to stool")                

            #Generate weather data.
            if PId is None:
                Status.WeatherData = None
            else:
                Status.WeatherData = WeatherData(PId)

            try:
                questionnaire = self.getQuestionnaire()
                Status.HPerDayInd = questionnaire.HPerDayInd # establish current data
                Status.NShifts = questionnaire.NShifts
                Status.NDaysInd = questionnaire.NDaysInd
            except:
                Status.HPerDayInd = 24
                Status.NShifts = 1
                Status.NDaysInd = 261

            self.setActiveAlternative(Status.ANo)
            self.setFinalAlternative(Status.FinalAlternative)

##        Status.SQL.commit()
        dbio.commit()

        try:                                #try, because at start-up there's not yet a moduleST
            Status.mod.moduleST.resetST()   #resets the calculator of surface areas
        except:
            pass



    def createNewProject(self, originalPId, shortName, description, originalName=None):
        """Either create a new project from the scratch or copy an 
        existing one.
        
        
        :param originalPId: Either the ID of the original project or 0 
        if the new project is not a copy of an existing one.
        :param shortName: The short name of the project.
        :param description: The description of the project.
        :param originalName: The name of the original project to copy,
        defaults to None.
        
        """
        questionnaires = dbio.select("questionnaire")
        sprojects = dbio.select("sproject")
        salternatives = dbio.select("salternatives")
        cgeneraldata = dbio.select("cgeneraldata")
        qelectricity = dbio.select("qelectricity")
        uheatpump = dbio.select("uheatpump")
                        
        if (originalPId <= 0) and (originalName is None):

            isNewProject = True
#..............................................................................
# start a new project from scratch (creates basic project tables)

            newProject = {"Name": "", "DescripIndustry":check(description)}

##            newID = questionnaires.insert(newProject)
            newID = dbio.insertRow("questionnaire", **newProject)

            logTrack("Project (createNewProject): new project inserted with ID %s " % newID)

            newSProject = {"ProjectID":newID,
                           "ProjectName": check(shortName, length=45),
                           "NoOfAlternatives":-1,
                           "ActiveAlternative":-1,
                           "WriteProtected":0,
                           "StatusQ":EINSTEIN_NOTOK,
                           "StatusCC":EINSTEIN_NOTOK,
                           "StatusCA":EINSTEIN_NOTOK,
                           "StatusR":EINSTEIN_NOTOK,
                           "LanguageReport":Status.LanguageTool,
                           "UnitsReport":Status.UnitsTool+"|"+Status.ProjectCurrency}
##            sprojects.insert(newSProject)
            dbio.insertRow("sproject", **newSProject)

            newAlternative = {"ProjectID":newID,
                              "AlternativeProposalNo":-1,
                              "ShortName":_("present state (original)"),
                              "Description":_("original data as submitted by industry")}
##            salternatives.insert(newAlternative)
            dbio.insertRow("salternatives", **newAlternative)
            
            newGeneralData = {"Questionnaire_id":newID,
                              "AlternativeProposalNo":-1,
                              "NFuels":0,
                              "NEquipe":0,
                              "NPipeDuct":0,
                              "NThProc":0,
                              "NProducts":0,
                              "NHX":0,
                              "NWHEE":0}
##            cgeneraldata.insert(newGeneralData)
            dbio.insertRow("cgeneraldata", **newGeneralData)

            newElectricity = {"Questionnaire_id":newID,
                              "AlternativeProposalNo":-1}
##            qelectricity.insert(newElectricity)
            dbio.insertRow("qelectricity", **newElectricity)

            # fill default values for design assistants (name uheatpump is for historical reasons. contains user-defined auto-design configuration settings for all types of equipment)
            newUHeatPump =  {"Questionnaire_id":newID,      
                             "AlternativeProposalNo":-1}

            newUHeatPump.update(DEFAULTSAUTODESIGNCHP)
            newUHeatPump.update(DEFAULTSAUTODESIGNHP)
            newUHeatPump.update(DEFAULTSAUTODESIGNBB)
            newUHeatPump.update(DEFAULTSAUTODESIGNST)

            
##            uheatpump.insert(newUHeatPump)
            dbio.insertRow("uheatpump", **newUHeatPump)

            Status.NoOfAlternatives = -1
                        
        else:

            isNewProject = False

            if originalName is not None:    #if a name is given, overwrite ID
                originalPId = self.getProjectID(originalName)
                if originalPId is None:
                    logError(_("Could not find project with name %s"%originalName))
                    return None

            logTrack("Project (createNewProject): copying from %s [%r]" % (originalPId, originalName))

#..............................................................................
# copy a project
# 1. create a copy of the entry in the main table (questionnaire) with different ID and get the new ID

            sqlQuery = "ProjectID = '%s'" % originalPId
            #additional query necessary for old tables who still have Questionnaire_id instead of ProjectID
            sqlQueryQ = "Questionnaire_id = '%s'" % originalPId

            pkey_questionnaire = "Questionnaire_ID" if DBSERVER == "MYSQL" else "id"
            sqlQueryQ_ID = "%s = '%s'" % (pkey_questionnaire, originalPId)

            newIDdict = copySQLRows("questionnaire", sqlQueryQ_ID, pkey_questionnaire, None, None)
            newID = newIDdict[originalPId]
#..............................................................................
# copying project data tables

            sDict = self.copySQLData(sqlQuery, sqlQueryQ, PId=newID)
            self.restoreLinks(newID, sDict)


#..............................................................................
# rename project

            q = dbio.get("questionnaire", id=newID)
            q.DescripIndustry = dbio.check(description)

            p = dbio.get("sproject", ProjectID=newID)
            p.ProjectName = dbio.check(shortName)


        dbio.commit()
        logger.logMessage("New project created: %s (id=%s)"%(shortName, newID))

        self.setActiveProject(newID, cleanUp=False)

        if isNewProject == True:
            self.setStatus("Q", 0)
            self.getDefaultSetUp()
            self.projectChanged(new=True)

        return newID

    def _copyStreams(self, streamIds, processIDsDict, connectionTable,
                     streamIdColumn, streamTable, newPID, ANo=None, sDict=None):
        """Copy all streams in streamTable that are referred to by Ids in 
        streamIDs. Create a corresponding entry in connectionTable.
        
        :param streamIds: a dict of pairs of IDs stream ID: process ID.
        :param processIDsDict: a dict with oldid: newid of copied processdata.
        :param connectionTable: usually process_streams_in or ...out.
        :param streamIdColumn: a String that is the name of the column that holds
        the stream IDs to set. 
        :param streamTable: the source and target table for the stream copies.
        :param newPID: the new Project ID.
        :param ANo: the number of the target alternative proposal if a new
        alternative is created
        :returns: a dict with oldid: newid of the copied rows
        
        """
        result = {}
        for sId, pId in streamIds.items():
            if ANo is not None: # This is just a copy of a Proposal
                streamIdsDict = copySQLRows(streamTable, "id = %s" % (sId,) , "id",
                                            "AlternativeProposalNo", ANo, sDict=sDict)
            else:
                streamIdsDict = copySQLRows(streamTable, "id = %s" % (sId,) , "id",
                                            "ProjectID", newPID, sDict=sDict)
            oldId, newId = streamIdsDict.items()[0]
            stream = dbio.get(streamTable, id=newId)
            if ANo is None:
                thisANo = stream.AlternativeProposalNo
            else:
                thisANo = ANo
            newEntry = {"qprocessdata_QProcessData_Id"  : processIDsDict[pId],
                        streamIdColumn                  : newId,
                        "ProjectID"                     : newPID,
                        "AlternativeProposalNo"         : thisANo 
                        }
            dbio.insertRow(connectionTable, **newEntry)
            result[oldId] = newId
        return result
            
    def _getStreamsForProcesses(self, connectiontable, processIds, keycolumn):
        """Return a dict of IDs of streams that belong to the processesIds 
        
        :param connectiontable: usually process_streams_in or ...out.
        :param processIds: a list of IDs of processes.
        :param keycolumn: a String that is the name of the column that holds
        the stream IDs that are searched for.
        :returns: a dict of stream IDs: process IDs.
        
        """
        result = {}
        for pId in processIds:
            streams = dbio.select(connectiontable, qprocessdata_QProcessData_Id=pId)
            sIds = [getattr(stream, keycolumn) for stream in streams]
            for sId in sIds:
                result[sId] = pId
        return result
    
    def _getProcessesForBuilding(self, buildingID):
        """Return the list of IDs of processes that belong to the
        building with the database ID buildingId.
        
        """
##        processes = Status.DB.qprocessdata["QBuildings_ID"][buildingID]
        processes = dbio.select("qprocessdata", QBuildings_ID=buildingID)
##        return [process.QProcessData_ID for process in processes]
        return [process.id for process in processes]
    
    def copyNodes(self, PId, originalANo=None):
        nodeIds = self.getNodeIDs(PId, ANo=originalANo)
        copyDict = {}
        for nodeId in nodeIds:
            qnode = dbio.get("qnode", id=nodeId)
            newId = dbio.copyRow("qnode", qnode)
            copyDict.update({nodeId: newId})
        return copyDict
            
#------------------------------------------------------------------------------
    def cleanUpProject(self, PId):
#------------------------------------------------------------------------------
#   deletes all tables for a given project with AlternativeProposalNo different
#   than ActiveAlternatives in sproject
#------------------------------------------------------------------------------

        logTrack("Project (cleanUpProject): cleaning project no. %s" % PId)
#..............................................................................
        sqlQuery = "ProjectID = '%s'" % PId
        sqlQueryQ = "Questionnaire_id = '%s'" % PId

        sproject = self.getProjectStatus(PId)
        if sproject is not None:
            maxANo = sproject.NoOfAlternatives

#..............................................................................
# repair corrupt AlternativeProposalNos in tables streams_in, streams_out, process_streams_in and process_streams_out
# may have occured due to a bug in V2.0beta (until beta10) that has been active from about April/May 2011 until 16.06.2011

            for ANo in range(1,maxANo+1):           # corrections are necessary only for ANo > 0 (hopefully)
                processes = self.getProcesses(PId,ANo)

                for process in processes:
                    ProcessID = process.QProcessData_ID
                    
##                    inStreamsTable  = DB.process_streams_in.qprocessdata_QProcessData_Id[ProcessID]
##                    outStreamsTable = DB.process_streams_out.qprocessdata_QProcessData_Id[ProcessID]
                    inStreamsTable  = dbio.select("process_streams_in", qprocessdata_QProcessData_Id=ProcessID)
                    outStreamsTable = dbio.select("process_streams_out", qprocessdata_QProcessData_Id=ProcessID)

                    for entry in inStreamsTable:
                        entry.AlternativeProposalNo = ANo
                        streams_in_id = entry.streams_in_id
                        inStreams = dbio.select("streams_in", id=streams_in_id)
                        for inStream in inStreams:
                            if inStream.AlternativeProposalNo != ANo:
                                inStream.AlternativeProposalNo = ANo
                                logWarning("Wrong alternative proposal number for an instream.")

                    for entry in outStreamsTable:
                        entry.AlternativeProposalNo = ANo
                        streams_out_id = entry.streams_out_id
                        outStreams = dbio.select("streams_out", id=streams_out_id)
                        for outStream in outStreams:
                            if outStream.AlternativeProposalNo != ANo:
                                outStream.AlternativeProposalNo = ANo
                                logWarning("Wrong alternative proposal number for an outstream.")
            
            # remove left over heat exchanger-pinchstream connections and
            # pinchstreams. They have been left over until 2.1beta1. 
            for ANo in range(-1, maxANo + 1):
                allHXids = self.getHXList("QHeatExchanger_ID", PId, ANo)
                allPinchIds = self.getPinchList("id", PId, ANo)
                # get connections without a valid heat exchanger
                # or without a valid pinch stream
                deleteConnections = set()
                connections = dbio.select("heatexchanger_pinchstream", ProjectID=PId, AlternativeProposalNo=ANo)
                for connection in connections:
                    if not connection.qheatexchanger_QHeatExchanger_Id in allHXids:
                        print("found dangling connection for missing HX %s" %
                                 connection.qheatexchanger_QHeatExchanger_Id)
                        deleteConnections.add(connection.id)
                    elif not connection.pinchstream_id in allPinchIds:
                        print("found dangling connection for missing pinchstream %s" %
                                 connection.pinchstream_id)
                        deleteConnections.add(connection.id)
                        
                for connectionId in deleteConnections:
                    dbio.delete("heatexchanger_pinchstream", id=connectionId)
                    
                self.deleteDanglingPinchstreams(PId, ANo)
                
##            Status.SQL.commit()
            dbio.commit()
                
#..............................................................................
# cleanUp of project data tables: checks ANo and Nones in main entry field (name)

            cleanUpSQLRows("salternatives", sqlQuery, maxANo, "ShortName")
        
            cleanUpSQLRows("cgeneraldata", sqlQueryQ, maxANo)

            cleanUpSQLRows("qbuildings", sqlQueryQ, maxANo, "BuildName")
            cleanUpSQLRows("qdistributionhc", sqlQueryQ, maxANo, "Pipeduct")
            cleanUpSQLRows("qelectricity", sqlQueryQ, maxANo)
            cleanUpSQLRows("qfuel", sqlQueryQ, maxANo, "FuelNo")
            cleanUpSQLRows("qgenerationhc", sqlQueryQ, maxANo, "Equipment")
            cleanUpSQLRows("qheatexchanger", sqlQuery, maxANo, "HXName")
            cleanUpSQLRows("qprocessdata", sqlQueryQ, maxANo, "Process")
            cleanUpSQLRows("streams_in", sqlQuery)
            cleanUpSQLRows("streams_out", sqlQuery)
            cleanUpSQLRows("process_streams_in", sqlQuery)
            cleanUpSQLRows("process_streams_out", sqlQuery)
            cleanUpSQLRows("qflow", sqlQuery, maxANo)
            cleanUpSQLRows("qproduct", sqlQueryQ, maxANo, "Product")
#            cleanUpSQLRows(DB.qrenewables,sqlQueryQ,maxANo)
            cleanUpSQLRows("qsurfarea", sqlQuery, mainField="SurfAreaName")  #surface areas have no ANo. fixed for the project !!!!
            cleanUpSQLRows("qwasteheatelequip", sqlQuery, maxANo, "WHEEName")
            cleanUpSQLRows("uheatpump", sqlQueryQ, maxANo)
            cleanUpSQLRows("heatexchanger_pinchstream", sqlQuery, maxANo)
            cleanUpSQLRows("pinchstream", sqlQuery, maxANo)
            cleanUpSQLRows("tcageneraldata", sqlQuery, maxANo)
            cleanUpSQLRows("tcacontingencies", sqlQuery, maxANo)
            cleanUpSQLRows("tcadetailedopcost", sqlQuery, maxANo)
            cleanUpSQLRows("tcadetailedrevenue", sqlQuery, maxANo)
            cleanUpSQLRows("tcaenergy", sqlQuery, maxANo)
            cleanUpSQLRows("tcainvestments", sqlQuery, maxANo)
            cleanUpSQLRows("tcanonreoccuringcosts", sqlQuery, maxANo)

            dbio.commit()

    def cleanUpDB(self, auto=False):
        """ function introduced in Sept. 2016
        checks the database entries for the following inconsistencies:
        - process_period_profiles - entries not linked to an existing process or equipment
        - process_periods - entries not linked to an existing process or equipment
        """

        errCount = 0
        # first clean-up all entries for projects which are no longer existing
        projectIds = [q.id for q in dbio.select("questionnaire")] + [0] # the 0 is for special entries

        tables = ["salternatives",
                  "qheatexchanger",
                  "streams_in",
                  "streams_out",
                  "process_streams_in",
                  "process_streams_out",
                  "heatexchanger_pinchstream",
                  "pinchstream",
                  "qflow",
                  "qsurfarea",
                  "qwasteheatelequip",
                  "tcageneraldata",
                  "tcacontingencies",
                  "tcadetailedopcost",
                  "tcadetailedrevenue",
                  "tcaenergy",
                  "tcainvestments",
                  "tcanonreoccuringcosts"]

        tablesQ = ["cgeneraldata",
                   "qbuildings",
                   "qdistributionhc",
                   "qelectricity",
                   "qfuel",
                   "qgenerationhc",
                   "qprocessdata",
                   "qproduct",
                   "uheatpump"]

        corruptTables = []
        for table in tables+tablesQ:
            if table in tables:
                PIds = [q.ProjectID for q in dbio.select(table)]
            else:
                PIds = [q.Questionnaire_id for q in dbio.select(table)]                
            
            corruptPIds = [PId for PId in PIds if PId not in projectIds]
            if len(corruptPIds) > 0:
                corruptTables.append([table, corruptPIds[:]])
                logError("Orphan entries for table %s with not existing ProjectIDs %s"%(table, corruptPIds))

        errCount += len(corruptTables)
        if len(corruptTables) > 0:
            ret = askConfirmation("Do you want to delete the corrupt entries ?")
            if ret:
                for table, PIds in corruptTables:
                    key = "ProjectID" if table in tables else "Questionnaire_id"
                    dbio.deleteByList(table, key, PIds)
                
                dbio.commit()
        else:
            logMessage("No orphan entries without ProjectID found")

        ret = askConfirmation("Do you want to check correct alternative proposal numbers ?")
        if ret:
            for PId in projectIds:
                self.cleanUpProject(PId)

        # clean-up in- and out-streams and corrupt link-tables
        linkTable = dbio.select("process_streams_in")

        corruptLinksIn = []
        for link in linkTable:
            linkPId = link.ProjectID
            linkANo = link.AlternativeProposalNo
            processId = link.qprocessdata_QProcessData_Id
            streamId = link.streams_in_id
            proc = dbio.get("qprocessdata", id=processId)
            if proc is None:
                corruptLinksIn.append(link.id)
                continue
            else:
                procPId = proc.Questionnaire_id
                procANo = proc.AlternativeProposalNo
                if (procPId != linkPId or procANo != linkANo):
                    corruptLinksIn.append(link.id)
                    continue

            stream = dbio.get("streams_in", id=streamId)
            if stream is None:
                corruptLinksIn.append(link.id)
            else:
                streamPId = stream.ProjectID
                streamANo = stream.AlternativeProposalNo
                if (streamPId != linkPId or streamANo != linkANo):
                    corruptLinksIn.append(link.id)
                    continue

        errCount += len(corruptLinksIn)

        linkTable = dbio.select("process_streams_out")
        corruptLinksOut = []
        for link in linkTable:
            linkPId = link.ProjectID
            linkANo = link.AlternativeProposalNo
            processId = link.qprocessdata_QProcessData_Id
            streamId = link.streams_out_id
            proc = dbio.get("qprocessdata", id=processId)
            if proc is None:
                corruptLinksOut.append(link.id)
                continue
            else:
                procPId = proc.Questionnaire_id
                procANo = proc.AlternativeProposalNo
                if (procPId != linkPId or procANo != linkANo):
                    corruptLinksOut.append(link.id)
                    continue

            stream = dbio.get("streams_out", id=streamId)
            if stream is None:
                corruptLinksOut.append(link.id)
            else:
                streamPId = stream.ProjectID
                streamANo = stream.AlternativeProposalNo
                if (streamPId != linkPId or streamANo != linkANo):
                    corruptLinksOut.append(link.id)
                    continue

        errCount += len(corruptLinksOut)
        
        if (len(corruptLinksIn) + len(corruptLinksOut)) > 0:
            ret = askConfirmation("Do you want to delete the corrupt process in/out stream links ?")
            if ret:
                dbio.deleteByList("process_streams_in", "id", corruptLinksIn)
                dbio.deleteByList("process_streams_out", "id", corruptLinksOut)
                dbio.commit()
        else:
            logMessage("No corrupt process in/out stream links found")

        # clean-up schedule entries which are not connected with any process or equipment schedule
                
        processes = dbio.select("qprocessdata")
        processIds = [proc.id for proc in processes]
        
        equipment = dbio.select("qgenerationhc")
        equipmentIds = [-eq.id for eq in equipment]

        allowedIds = processIds+equipmentIds

        pps = dbio.select("process_periods")
        allowedPPIds = []
        corruptPPIds = []
        for pp in pps:
            if pp.qprocessdata_QProcessData_ID not in allowedIds:
                logError("Project.cleanUpDB: corrupt entry for table process_periods id %s QProcessData_ID %s"%(pp.id, pp.qprocessdata_QProcessData_ID))
                corruptPPIds.append(pp.id)      
            else:
                allowedPPIds.append(pp.id)      

        corruptPPPIds = []
        ppps = dbio.select("process_period_profiles")
        for ppp in ppps:
            if ppp.process_periods_id not in allowedPPIds:
                logError("Project.cleanUpDB: corrupt entry for table process_period_profiles id %s QProcessData_ID %s"%(ppp.id, ppp.process_periods_id))
                corruptPPPIds.append(ppp.id)

        errCount += len(corruptPPIds) + len(corruptPPPIds)
        if len(corruptPPIds) > 0 or len(corruptPPPIds) > 0:
            ret = askConfirmation("Do you want to delete the corrupt entries ?")
            if ret:
                for id in corruptPPIds:
                    dbio.delete("process_periods", id=id)
                for id in corruptPPPIds:
                    dbio.delete("process_period_profiles", id=id)

                dbio.commit()
        else:
            logMessage("No corrupt schedule entries found")

        # clean-up node and flow system
        nodeIds = [q.id for q in dbio.select("qnode")]
        productIds = [q.id for q in dbio.select("qproduct")]

        fromNodeIds = [q.FromId for q in dbio.select("qflow")]
        corruptFromNodeIds = [nodeId for nodeId in fromNodeIds if nodeId not in nodeIds]
        if len(corruptFromNodeIds) > 0:
            logError("Corrupt FromIds %s referenced in qflow "%(corruptFromNodeIds))

        toNodeIds = [q.ToId for q in dbio.select("qflow")]
        corruptToNodeIds = [nodeId for nodeId in toNodeIds if nodeId not in nodeIds]
        if len(corruptToNodeIds) > 0:
            logError("Corrupt ToIds %s referenced in qflow "%(corruptToNodeIds))

        qnodeList = dbio.select("qnode")
##        qnodesInt = [(qnode.id, qnode.ObjectId, qnode.ObjectType) for qnode in qnodeList if qnode.ObjectType=="INT"]
        qnodesProc = [(qnode.id, qnode.ObjectId, qnode.ObjectType) for qnode in qnodeList if qnode.ObjectType=="PROC"]
        qnodesProd = [(qnode.id, qnode.ObjectId, qnode.ObjectType) for qnode in qnodeList if qnode.ObjectType=="PROD"]

        corruptNodes = [nodeId for nodeId, objectId, objectType in qnodesProc if objectId not in processIds] + \
                       [nodeId for nodeId, objectId, objectType in qnodesProd if objectId not in productIds]

        #TODO: INT-nodes not referenced in any flow do not make sense (or do they ?). CleanUp.                           
        if len(corruptNodes) > 0:
            logError("Corrupt objectId referenced in qnodes %s"%(corruptNodes))

        errCount += len(corruptToNodeIds) + len(corruptFromNodeIds) + len(corruptNodes)
        if len(corruptToNodeIds) > 0 or len(corruptFromNodeIds) > 0 or len(corruptNodes) > 0:
            ret = askConfirmation("Do you want to delete the corrupt node and flow entries ?")
            if ret:
                dbio.deleteByList("qflow", FromId, corruptFromNodeIds)
                dbio.deleteByList("qflow", ToId, corruptToNodeIds)
                dbio.deleteByList("qnode", id, corruptNodes)

                dbio.commit()
        else:
            logMessage("No corrupt node or flow entries found")

        if errCount == 0:
            showMessage("Database is ok")
        else:
            showError("%s Corrupt entries found in database"%errCount)
        
#------------------------------------------------------------------------------
    def getDefaultSetUp(self):
#------------------------------------------------------------------------------
#   charges the default set-up parameters for a NEW project
#------------------------------------------------------------------------------

        (dummy, generalData) = self.getProjectData()
        setups = dbio.select("psetupdata")
        if (len(setups) > 0):
            setup = setups[0]
            eMixID = setup.ElectricityMix

        else:
            logTrack("Project (getDefaultSetUp): no setup table found in DB. Default mix ID = 1")
##            setups.insert({"ElectricityMix": 1})
            dbio.insertRow("psetupdata", ElectricityMix=1)
            eMixID = 1

        emix = dbio.get("dbelectricitymix", id=eMixID)
        if emix is not None:
            generalData.PEConvEl = emix.PE2ConvEl
            generalData.CO2ConvEl = emix.CO2ConvEl
            generalData.NoNukesConvEl = emix.NoNukesConvEl

        else:
            generalData.PEConvEl = 1. / 0.34
            generalData.CO2ConvEl = 0.5
            generalData.NoNukesConvEl = 99.99
            logTrack("Project (getDefaultSetUp) Electricity Mix %s specified in Set-Up was not found" % eMixID)
           
                     
#------------------------------------------------------------------------------
    def deleteProject(self, PId, name=None):
#------------------------------------------------------------------------------
#   deletes all entries for a given project
#------------------------------------------------------------------------------

        if name is not None:    #if a name is given, overwrite ID
            PId = self.getProjectID(name)

        logMessage(_("Project (deleteProject) -deleting project %r (%r)") % (PId, name))

#..............................................................................
# deleting Q- and corresponding C-Tables

##        DB = Status.DB
        sqlQuery = "ProjectID = '%s'" % PId

        #additional query necessary for old tables who still have Questionnaire_id instead of ProjectID
        sqlQueryQ = "Questionnaire_id = '%s'" % PId
        pkey_questionnaire = "Questionnaire_ID" if DBSERVER == "MYSQL" else "id"
        sqlQueryID = "%s = '%s'" % (pkey_questionnaire, PId)

        nodeIds = self.getNodeIDs(PId)
        for nodeId in nodeIds:
            dbio.delete("qnode", id=nodeId)

        deleteSQLRows("questionnaire", sqlQueryID)
        deleteSQLRows("sproject", sqlQuery)
        deleteSQLRows("salternatives", sqlQuery)
                   
        deleteSQLRows("cgeneraldata", sqlQueryQ)

        deleteSQLRows("qbuildings", sqlQueryQ)
        deleteSQLRows("qdistributionhc", sqlQueryQ)
        deleteSQLRows("qelectricity", sqlQueryQ)
        deleteSQLRows("qfuel", sqlQueryQ)

        hxes = dbio.select("qheatexchanger", ProjectID=PId)
        for hx in hxes:
            dbio.delete("qhxcontrol", HXId = hx.id)
            
        deleteSQLRows("qheatexchanger", sqlQuery)
        deleteSQLRows("streams_in", sqlQuery)
        deleteSQLRows("streams_out", sqlQuery)
        deleteSQLRows("process_streams_in", sqlQuery)
        deleteSQLRows("process_streams_out", sqlQuery)

##        # delete detailed schedules and periods for processes from project before processes themselves
##        sql = Status.SQL.cursor(MySQLdb.cursors.DictCursor)
##        # wipe schedules for processes from all alternatives to be sure
##        sql.execute('SELECT QProcessData_ID FROM qprocessdata WHERE %s' % sqlQueryQ)
##        projectProcessIds = [r['QProcessData_ID'] for r in sql.fetchall()]
##        sql.close()
        projectProcesses = dbio.select("qprocessdata", Questionnaire_id = PId)
        for process in projectProcesses:
            deleteProcessPeriodsFromDB(process.id)
        projectEquipment = dbio.select("qgenerationhc", Questionnaire_id = PId)
        for eq in projectEquipment:
            deleteProcessPeriodsFromDB(-eq.id)

        for process in projectProcesses:
            dbio.delete("qcombschedules", ProcessId=process.id)

        deleteSQLRows("qgenerationhc", sqlQueryQ)
        deleteSQLRows("qprocessdata", sqlQueryQ)
        deleteSQLRows("qproduct", sqlQueryQ)
        deleteSQLRows("qrenewables", sqlQueryQ)
        deleteSQLRows("qsurfarea", sqlQuery)
        deleteSQLRows("qwasteheatelequip", sqlQuery)
        deleteSQLRows("uheatpump", sqlQueryQ)
        deleteSQLRows("heatexchanger_pinchstream", sqlQuery)
        deleteSQLRows("pinchstream", sqlQuery)
        deleteSQLRows("tcageneraldata", sqlQuery)
        deleteSQLRows("tcacontingencies", sqlQuery)
        deleteSQLRows("tcadetailedopcost", sqlQuery)
        deleteSQLRows("tcadetailedrevenue", sqlQuery)
        deleteSQLRows("tcaenergy", sqlQuery)
        deleteSQLRows("tcainvestments", sqlQuery)
        deleteSQLRows("tcanonreoccuringcosts", sqlQuery)

        dataFiles = dbio.select("qmdatafiles", ProjectID=PId)
        for dataFile in dataFiles:
            dbio.delete("qmdatasets", DataFileId=dataFile.id)
        deleteSQLRows("qmdatafiles", sqlQuery)
                                            
        deleteSQLRows("qflow", sqlQuery)

#..............................................................................
# book-keeping: reduce number of alernatives in the system

        self.setActiveProject(-1)

        Status.int.systemModel = None

        dbio.commit()


#------------------------------------------------------------------------------		

#------------------------------------------------------------------------------		
# RE-ESTABLISH LINKS AFTER COPYING PROJECTS OR ALTERNATIVES
#------------------------------------------------------------------------------		
    def reconnectEquipesToPipes(self, equipeIDDict, pipeIDDict):

        for equipeID in equipeIDDict.values():
            equipe = dbio.get("qgenerationhc", id=equipeID)
            if equipe is not None:
                pipeLink = equipe.PipeDuctEquip
                oldPipeLink = pipeLink
#                print "Project (reconnectEquipesToPipes): equipe ",equipe.Equipment,pipeLink
                pipeIDs = []
                if pipeLink is not None:
                    pipeLinkSplit = pipeLink.split(';')
                    for i in range(len(pipeLinkSplit)):
                        try:
                            pipeIDs.append(long(pipeLinkSplit[i]))
                        except:
                            pass
                newPipeIDs = []
                for pipeID in pipeIDs:
                    if pipeID in pipeIDDict.keys():
                        newPipeIDs.append("%d" % pipeIDDict[pipeID])
                    else:
                        newPipeIDs.append("%d" % pipeID)
                        logDebug(_("Project (reconnect equipes): pipe ID %s was not in old pipe table") % pipeID)
#                print "Project (reconnectEquipesToPipes): pipeIDs ",newPipeIDs
       
                pipeLink = ';'.join(newPipeIDs)
#                print "Project (reconnectEquipesToPipes): new pipeLink ",pipeLink
                equipe.PipeDuctEquip = pipeLink
                logTrack("Project (reconnect...): link of equipe %s updated from %s to %s" % (equipeID, oldPipeLink, pipeLink))
                                                    
    def reconnectProcessesToBuildings(self, processesIDDict, buildingsIDDict):
        """Helper function to relink copied processes to the new building ids.
        
        Actually only needed if the process is actually attached to a building,
        but harmless otherwise.
        """
        for oldid, newid in processesIDDict.items():
            newprocess = self.getProcess(newid)
            oldprocess = self.getProcess(oldid)
            if oldprocess.QBuildings_ID is not None:
                newprocess.QBuildings_ID = buildingsIDDict[oldprocess.QBuildings_ID] 
      
#------------------------------------------------------------------------------		
# OTHER SET-UP FUNCTIONS
#------------------------------------------------------------------------------		

#------------------------------------------------------------------------------
    def getLastToolSettings(self):
#------------------------------------------------------------------------------

        stool = dbio.get("stool")
        Status.PId = stool.ActiveProject
        logTrack(_("New session of tool starting with ProjectID %s") % Status.PId)

        Status.UserType = stool.UserType
        Status.AuditorId = stool.Auditor_ID
        Status.UserInteractionLevel = stool.UserInteractionLevel
        Status.LanguageTool = stool.LanguageTool
        Status.UnitsTool = stool.UnitsTool
        
#------------------------------------------------------------------------------
    def setUserInteractionLevel(self, level):
#------------------------------------------------------------------------------
        
        if level in INTERACTIONLEVELS:
            Status.UserInteractionLevel = level
            stool = dbio.get("stool")
            stool.UserInteractionLevel = level
            dbio.commit()
            logTrack("Project (setUserInteractionLevel): %s" % level)
        else:
            logDebug("Project (setUserInteractionLevel): ERROR in level %s" % level)

#------------------------------------------------------------------------------
    def setStatus(self, key, value=1, ANo=None):
#------------------------------------------------------------------------------
#   sets a status flag
#------------------------------------------------------------------------------
        if ANo is None:
            ANo = Status.ANo
        sproject = dbio.get("sproject",ProjectID=Status.PId)

        if sproject is not None:
            salternative = dbio.get("salternatives", ProjectID=Status.PId, AlternativeProposalNo=ANo)
            
            if key == "Q":
                if ANo==Status.ANo: Status.StatusQ = value
                sproject.StatusQ = value
            elif key == "CC":
                if ANo==Status.ANo: Status.StatusCC = value
                sproject.StatusCC = value
            elif key == "CS":
                if ANo==Status.ANo: Status.StatusCS = value
    #XXXXXXXX here name confusion between CA and CS -> to be unified to CS once SQL is changed !!!
                sproject.StatusCA = value
            elif key == "Energy" or key == "ENERGY":
                if ANo==Status.ANo: Status.StatusEnergy = value
                if salternative is not None:
                    salternative.StatusEnergy = value
            elif key == "StatusECO" or key == "ECO":
                if ANo==Status.ANo: Status.StatusECO = value
                if salternative is not None: salternative.StatusECO = value
            else:
                logDebug("Project (setStatus): status key %s unknown" % key)

            dbio.commit()

        self.setTreePermissions()
        
    def setAuditor(self, auditorId):
        """Set the auditor for the current session, and the new default
        auditor. 
        
        :param auditorId: The auditorId from the database table "auditor"
        
        """
        from modules.checkLicense import checkAuditor, checkDemoVersion
            
        Status.AuditorId = auditorId
        stool = dbio.get("stool")
        stool.Auditor_ID = auditorId
        dbio.commit()

        # check license for new auditor and store result
        Status.License.check = checkAuditor(auditorId)
        
    def projectChanged(self, PId=None, new=False):
        """
        Mark the current project as changed by current auditor

        :params PId -> marks the project with Id=PId. If None -> mark active project
        :params new     if True, use current date also as creation date
        
        
        """
        projectStatus = Status.prj.getProjectStatus(PId)
        projectStatus.LastModified = date.today()
        projectStatus.ModifiedBy = Status.AuditorId
        if new:
            projectStatus.CreationDate = date.today()
            projectStatus.CreatedBy = Status.AuditorId

    def setHRMode(self, hrMode):
        """
        changes the parameter Status.HRTool and the corresponding flags
        """

        previous = Status.HRTool
        if hrMode != previous and hrMode in HRMODES:
            Status.HRTool = hrMode
            
            Status.int.changeInHeatRecovery() # the algorithm changes, so does the modelling 
            Status.int.changeInCascade(0) # the HR results may/should change, so should the simulation
        
            sproject = Status.prj.getProjectStatus()
            if sproject is not None:
                sproject.HRTool = dbio.check(Status.HRTool)
                dbio.commit()

            for ANo in xrange(1, Status.NoOfAlternatives):
                self.setStatus("Energy", 0, ANo)

            return 1
        
        else:
            return 0
        
        
#------------------------------------------------------------------------------
    def setTreePermissions(self):
#------------------------------------------------------------------------------
#   sets the tree permissions ...
#------------------------------------------------------------------------------

        if not Status.main: return
        
        tmp = {}
        allow = (True, 0, "")
        
        tmp.update({_("Edit Industry Data"):allow})   #access to this window is always allowed

#..............................................................................
# StatusPId: actions allowed once an industry is selected

        if (Status.PId > 0):
            permit = allow           
        else:
            permit = (False, 0, _("Cannot access this function. First select a project"))

        tmp.update({_("General data"):permit})
        
#..............................................................................
# StatusQ: actions allowed once general data of the industry have been set and
#   and confirmed

# 1. manipulation of questionnaire

        if (Status.StatusQ > 0):
            permit = allow           
        else:
            permit = (False, 0, _("Cannot access this function. First open or define a new project and enter general data"))

        tmp.update({_("Energy consumption"):permit})
        tmp.update({_("Processes data"):permit})
        tmp.update({_("Generation of heat and cold"):permit})
        tmp.update({_("Distribution of heat and cold"):permit})
        tmp.update({_("Heat recovery"):permit})
        tmp.update({_("Renewable energies"):permit})
        tmp.update({_("Buildings"):permit})
        tmp.update({_("Economic parameters"):permit})

# 2. access to consistency check window

        if (Status.StatusQ > 0):
            permit = allow
        else:
            permit = (False, 0, _("Cannot access this function. First open or define a new project"))

        tmp.update({_("Consistency Check"):permit})
        
#..............................................................................
# StatusCC: actions allowed once consistency check is concluded
        if (Status.StatusCC > 0):
            permit = allow
        else:
            permit = (False, 0, _("Cannot access this function. First check the data for consistency"))
        tmp.update({_("Alternative proposals"):permit})

#..............................................................................
# StatusCC: actions allowed once consistency check is concluded and the current
# alternative is the right one.
        if (Status.StatusCC > 0):
            if Status.ANo > -1:
                permit = allow
            else:
                permit = (False, 0, _("Cannot display statistics for raw data. change view to checked version"))
        else:
            permit = (False, 0, _("Cannot access this function. First check the data for consistency"))
                                      
        tmp.update({_("Energy statistics"):permit})
        tmp.update({_("Annual data"):permit})
        tmp.update({_("Monthly data"):permit})
        tmp.update({_("Hourly performance\ndata"):permit})
                
        tmp.update({_("Primary energy"):permit})
        tmp.update({_("Final energy by fuels"):permit})
        tmp.update({_("Final energy by equipment"):permit})
        
        tmp.update({_("H&C demand (proc.)"):permit})
        tmp.update({_("H&C demand (temp.)"):permit})
        tmp.update({_("H&C demand (time)"):permit})
        
        tmp.update({_("Process heat"):permit})
        tmp.update({_("Energy intensity"):permit})
        tmp.update({_("Environmental Impact"):permit})

        tmp.update({_("Monthly demand"):permit})
        tmp.update({_("Monthly supply"):permit})
                   
        tmp.update({_("Hourly demand"):permit})
        tmp.update({_("Hourly supply"):permit})

        tmp.update({_("Pinch analysis"):permit})
        tmp.update({_("List of streams"):permit})
        tmp.update({_("Composite curves"):permit})

        tmp.update({_("Benchmark check"):permit})
        tmp.update({_("Global energy intensity"):permit})
        tmp.update({_("SEC by product"):permit})
        tmp.update({_("SEC by process"):permit})
        
#..............................................................................
# ActiveAlternative: actions allowed only for ANo > 0

        if (Status.ANo > 0):
            permit = allow
        else:
            permit = (False, 0, _("Cannot access this function for the present state of the industry.\nFirst define a new alternative !"))
                          
        tmp.update({_("Design"):permit})

        tmp.update({_("Process optimisation"):permit})
        tmp.update({_("Process optimisation interface 1"):permit})
        tmp.update({_("Process optimisation interface 2"):permit})

        tmp.update({_("HX network"):permit})
        tmp.update({_("H&C Supply"):permit})
        tmp.update({_("H&C Storage"):permit})
        tmp.update({_("CHP"):permit})
        tmp.update({_("Solar Thermal"):permit})
        tmp.update({_("Heat Pumps"):permit})
        tmp.update({_("Biomass"):permit})
        tmp.update({_("Chillers"):permit})
        tmp.update({_("Boilers & burners"):permit})

        tmp.update({_("H&C Distribution"):permit})

        tmp.update({_("Energy performance"):permit})
        tmp.update({_("Economic analysis"):permit})
        tmp.update({_("Economics 1"):permit})
        tmp.update({_("Economics 2"):permit})

#..............................................................................
# StatusCC: actions allowed once consistency check is concluded

        if (Status.StatusCC > 0):
            permit = allow
        else:
            permit = (False, 0, _("Cannot access this function. First check the data for consistency"))
            
        tmp.update({_("Comparative study"):permit})

        tmp.update({_("Report"):permit})
        tmp.update({_("Report generation"):permit})

        Status.main.treePermissions = tmp
        
#------------------------------------------------------------------------------
    def getStatus(self):
#------------------------------------------------------------------------------
#   sets a status flag
#------------------------------------------------------------------------------

        sproject = self.getProjectStatus()

        if sproject is not None:
            if sproject.StatusQ is not None: Status.StatusQ = sproject.StatusQ
            else:
                sproject.StatusQ = EINSTEIN_NOTOK
                Status.StatusQ = EINSTEIN_NOTOK

            if sproject.StatusCC is not None: Status.StatusCC = sproject.StatusCC
            else:
                sproject.StatusCC = EINSTEIN_NOTOK
                Status.StatusCC = EINSTEIN_NOTOK

            if sproject.HRTool in HRMODES[1:]: # old HRMODE 0 no longer supported
                Status.HRTool = sproject.HRTool
            else: 
                Status.HRTool = HRMODES[2]
            
        else:
            logTrack("Project (getStatus): could not find project table of last opened project")
#XXX -> getStatus is called BEFORE GUI is built -> logError not yet available !!!
            Status.StatusCC = EINSTEIN_NOTOK
            Status.StatusQ = EINSTEIN_NOTOK


        a = self.getAlternative()
        if a is not None:
            if a.StatusEnergy is not None:
                Status.StatusEnergy = a.StatusEnergy
            else:
                a.StatusEnergy = EINSTEIN_NOTOK
                Status.StatusEnergy = EINSTEIN_NOTOK

        self.setTreePermissions()

    def isEnergyStatusUpdated(self):
        """
        returns true if StatusEnergy Flag is calculated for all proposals,
        false otherwise
        """
        
##        sqlQuery = "ProjectID = '%s'" % (Status.PId)
##        salternatives = Status.DB.salternatives.sql_select(sqlQuery)
        salternatives = self.getAlternatives()
        for a in salternatives:
            try:
                ANo = a.AlternativeProposalNo
                if ANo != -1 and ANo != Status.ANo:
                    if a.StatusEnergy == 0:
                        return False 
            except:
                return False
        return True
        
    def getEnergyStatusNotUpdated(self):
        salternatives = self.getAlternatives()
        StatusEnergy = []
        
        for a in salternatives:
            try:
                ANo = a.AlternativeProposalNo
                if ANo != -1  and ANo != Status.ANo:
                    if a.StatusEnergy == 0:
                        StatusEnergy.append(a.ShortName)
            except:
                return []
        return StatusEnergy


    def getRenewables(self, PId=None):
        """ returns the table of renewables
        """
        if PId is None: PId = Status.PId
        return dbio.get("qrenewables", Questionnaire_id=PId)


    def deleteObject(self, objectCode, objectNo):
        """ GUI function added in EINSTEIN 3 for editing system """
                
        objectId = Status.int.systemModel.objectId(objectCode, objectNo)
        if objectCode=="FUEL":
            self.deleteFuel(objectId)
            
        elif objectCode=="PROC":
            self.deleteProcess(objectId)
            
        elif objectCode=="PIPE":
            self.deletePipe(objectId)

        elif objectCode=="EQ":
            self.deleteEquipment(objectId)

        elif objectCode=="HX":
            hx = Status.int.hx[objectNo]
            hx.deleteFromDB()

        elif objectCode=="WHEE":
            self.deleteWHEE(objectId)

        else:
            logger.logWarning("SystemModel.deleteObject: object %s.%s not found"%(objectCode, objectNo))
            return

        Status.int.systemModel.changed()
    
#------------------------------------------------------------------------------
    def deleteProduct(self, productName):
#------------------------------------------------------------------------------
#   deletes all entries for the original ANo
#------------------------------------------------------------------------------

        logTrack("Project (deleteProduct) - project %s, alternative %s, deleting product %s " % \
                 (Status.PId, Status.ANo, productName))

        sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s' AND Product = '%s'"\
                    % (Status.PId, Status.ANo, check(productName))

        qproducts = dbio.selectByQuery("qproduct", sqlQueryQ)
        productIds = [q.id for q in qproducts]

        # if a product is deleted, delete also the associated nodes and flows
        for productId in productIds:
            nodeIds = [node.id for node in dbio.select("qnode", ObjectId=productId, ObjectType="PROD")]
            for nodeId in nodeIds:
                dbio.delete("qflow", FromId = nodeId)
                dbio.delete("qflow", ToId = nodeId)
                
            dbio.delete("qnode", ObjectId=productId, ObjectType="PROD")

        deleteSQLRows("qproduct", sqlQueryQ)

    def getProducts(self, PId=None, ANo=None, name=None):
        """ returns a table of existing equipment
        """

        if ANo is not None:
            selectANo = ANo
        else:
            selectANo = Status.ANo
            
        #TODO: use of AlternativeProposalNo is not clearly specified and maybe not consistent. Unify criteria throughout the project.py-functions
        if PId is None:
            if name:
                products = dbio.select("qproduct", Questionnaire_id=Status.PId, AlternativeProposalNo=selectANo, Product=name)
            else:
                products = dbio.select("qproduct", Questionnaire_id=Status.PId, AlternativeProposalNo=selectANo)
        else:
            if name:
                products = dbio.select("qproduct", Questionnaire_id=PId, Product=name)
            elif ANo is None:
                products = dbio.select("qproduct", Questionnaire_id=PId)
            else:
                 products = dbio.select("qproduct", Questionnaire_id=PId, AlternativeProposalNo=selectANo)
       
        return products

    def getProductList(self, key, PId=None):
        """ returns a table of existing equipment
        """

        products = self.getProducts(PId)
        
        productList = []
        for product in products:
            pkey = getattr(product, key)
            if isinstance(pkey, str) or isinstance(pkey, unicode):
                productList.append(unicode(pkey, "utf-8"))
            else:
                productList.append(pkey)

        return productList

#------------------------------------------------------------------------------
    def addEquipmentDummy(self):
#------------------------------------------------------------------------------
#   enters a dummy of a new equiment
#------------------------------------------------------------------------------

#..............................................................................
# deleting Q- and corresponding C-Tables

        equipment = self.getEquipment()
        NEquipe = len(equipment)
        tmp = {
            "Questionnaire_id":Status.PId,
            "AlternativeProposalNo":Status.ANo,
            "EqNo": NEquipe + 1,
            "CascadeIndex": NEquipe + 1,
            "Equipment":"dummy"
            }          
        newID = dbio.insertRow("qgenerationhc", **tmp)
        newEquipe = dbio.get("qgenerationhc", id=newID)

        NEquipe += 1

        generaldata = self.getGeneralData()
        generaldata.NEquipe = NEquipe
        Status.NEquipe = NEquipe

        dbio.commit()

        Status.int.changeInCascade(NEquipe)     # updates Status-flags
        
        return newEquipe

#------------------------------------------------------------------------------
    def deleteEquipment(self, eqID):
        """ Deletes all entries for a given equipment
        Restores the correct equipment- and cascade numbering for the remaining
        equipment.

        Automatically deletes heat exchangers connected to this equipment
        """

        if eqID is None: return
        
        # deleting Q- and corresponding C-Tables

        logTrack("Project (deleteEquipment) %s" % (eqID,))

        deleteProcessPeriodsFromDB(-eqID)
        
        eq = dbio.get("qgenerationhc", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo, id=eqID)
        if eq is None:
            equipmentName = None
            logDebug("Project (deleteEquipment): cannont delete equipment with ID %s. No corresponding entry in database" % eqID)
            return

        # delete heat exchangers connected to the deleted equipment
        for h, hx in enumerate(Status.int.hx):
            for conn in hx.sourcestreams + hx.sinkstreams:
                if conn.objectType=="Equipment" and conn.objectNo==eq.CascadeIndex-1:
                    hx.deleteFromDB()
                    continue

        equipmentName = eq.Equipment
        deletedIndex = eq.CascadeIndex
        dbio.delete("qgenerationhc", id=eqID)
        dbio.commit()

        # re-build equipment and cascade numbering
        equipment = sorted(dbio.select("qgenerationhc", Questionnaire_id = Status.PId, AlternativeProposalNo=Status.ANo), key = lambda eq: eq.EqNo)
        for i, eq in enumerate(equipment): #assign new EqNo in QGenerationHC table
            eq.EqNo = i + 1

        equipment = sorted(dbio.select("qgenerationhc", Questionnaire_id = Status.PId, AlternativeProposalNo=Status.ANo), key = lambda eq: eq.CascadeIndex)
        for i, eq in enumerate(equipment): #assign new EqNo in QGenerationHC table
            eq.CascadeIndex = i + 1


        dbio.commit()

##        Status.int.cascadeUpdateLevel = min(Status.int.cascadeUpdateLevel, deletedIndex - 1)
        Status.int.changeInCascade(deletedIndex)
        
        generalData = self.getGeneralData()
        generalData.NEquipe = len(equipment)
        Status.NEquipe = len(equipment)

        dbio.commit()

        Status.int.systemModel.build() #rebuild the model after deleting some equipment

    def getEquipment(self, PId=None, ANo=None, cascade=False):
        """Return a pSQL.Result of all equipment used in the current 
        project.
        
        If the PId is given all equipment of all alternative proposals 
        is returned, otherwise only of the current one. Per default 
        the results are ordered ascending by the equipment number.
        
        :param PId: which Questinonnaire_id is requested if not the current one
        which is the default.
        :param cascade: order ascending by cascade order if true.
        
        """
        if PId is None:
            equipment = dbio.select("qgenerationhc", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        elif ANo is None:
            equipment = dbio.select("qgenerationhc", Questionnaire_id=PId)
        else:
            equipment = dbio.select("qgenerationhc", Questionnaire_id=PId, AlternativeProposalNo=ANo)
            
        if cascade == False:
            equipment = sorted(equipment, key = lambda eq: eq.EqNo)
        else:
            equipment = sorted(equipment, key = lambda eq: eq.CascadeIndex)

        if PId is None:
            Status.NEquipe = len(equipment)
        
        return equipment

    def getEquipments(self, PId=None, cascade=False):
        """ this function with orthographic error maintained for backward compatibility
            add-on of ANo is not included here, as this did not yet exist in the older
            versions.
            DEPRECATED. Should be eliminated once no longer needed.
        """
        return self.getEquipment(PId=PId, cascade=cascade)
    
    def getEquipmentCascade(self,):

        equipments = self.getEquipments(cascade=True)
        equipmentCascade = []
        cascadeIndices = [None]*len(equipments)
        
        for jc,eq in enumerate(equipments):
            j = eq.EqNo - 1
            equipmentCascade.append(j)
            cascadeIndices[j] = jc

        return(equipmentCascade,cascadeIndices)
            
            
    def getEquipe(self, id = None, cascadeIndex=None, no=None, PId=None, ANo=None):
        """Return the SQL entry (qgenerationhc) of an equipment
        :param id: the id in the database
        :

        """
        if PId is None: PId = Status.PId
        if ANo is None: ANo = Status.ANo

        if id is not None and id > 0:
            eq = dbio.get("qgenerationhc", id=id)
            return eq
        elif cascadeIndex is not None:
            eq = dbio.get("qgenerationhc", CascadeIndex=cascadeIndex, Questionnaire_id=PId, AlternativeProposalNo=ANo)
        elif no is not None:
            eq = dbio.get("qgenerationhc", EqNo=no, Questionnaire_id=PId, AlternativeProposalNo=ANo)
        else:
            return None
        
        return eq
        
    def getEquipmentList(self, key, PId=None, cascade=False, returnUnicode=True):
        """Return a list of the content of a single column of all the 
        equipments of the current project and alternative proposal.
        
        If a specific PId is given the equipments of all alternative
        proposals of the project are searched through. Per default 
        the results are ordered ascending by the equipment number.
        
        :param key: the column name
        :param PId: the ProcessID if not the current project is wanted.
        :param cascade: order ascending by cascade order if true.
        :returns: A list of column values which may be of integer, 
        String or what ever.
        
        """
        eqs = self.getEquipments(PId, cascade)
        eqList = []
        for eq in eqs:
            attr = getattr(eq, key)
            if isinstance(attr, (str, unicode)) and returnUnicode:
                eqList.append(unicode(attr, "utf-8"))
            else:
                eqList.append(attr)
        return eqList

    def getStreams(self, qProcessData_ID):
        """Return a 2-tuple of all streams associated with the given 
        qProcessData_ID.
        
        :param qProcessData_ID: Numeric database ID of the process for which
        the streams are requested
        :returns: A tuple of lists (inStreams, outStreams).
        
        """
        inStreamIDs = dbio.select("process_streams_in", qprocessdata_QProcessData_Id=qProcessData_ID)
        outStreamIDs = dbio.select("process_streams_out", qprocessdata_QProcessData_Id=qProcessData_ID)
        
        inStreams = [dbio.get("streams_in", id=link.streams_in_id) for link in inStreamIDs]
        outStreams = [dbio.get("streams_out", id=link.streams_out_id) for link in outStreamIDs]

        inStreams = [stream for stream in inStreams if stream is not None]
        outStreams = [stream for stream in outStreams if stream is not None]
        return (inStreams, outStreams)
        
    def addFuelDummy(self):
        """adds an empty fuel entry to the qfuel table"""
        
        fuels = self.getFuels()
        NFuels = len(fuels)
        tmp = {
            "Questionnaire_id":Status.PId,
            "AlternativeProposalNo":Status.ANo,
            "FuelNo": NFuels + 1
            }          
        newID = dbio.insertRow("qfuel", **tmp)

        NFuels += 1

        generaldata = self.getGeneralData()
        generaldata.NFuels = NFuels

        dbio.commit()

        return newID

    def deleteFuel(self, fuelId):
        """ deletes all entries for a given fuel type
        """

#..............................................................................
# deleting Q- and corresponding C-Tables

        dbio.delete("qfuel", id=fuelId)
        dbio.commit()

        fuels = self.getFuels()

        for i, fuel in enumerate(fuels): #assign new FuelNo in QFuel table
            fuels[i].FuelNo = i + 1

        
        generaldata = self.getGeneralData()
        generaldata.NFuels = len(fuels)

        dbio.commit()

    def addProcessDummy(self, projectId=None, ANo=None):
        """Create an empty process entry in the database and return a
        pSQL result, that represents the entry. If no projectId is
        given then take the current project, if no alternative proposal
        number is given then take the current proposal.
         
        """
        if projectId is None:
            projectId = Status.PId
        if ANo is None:
            ANo = Status.ANo

        processes = dbio.select("qprocessdata", Questionnaire_id=projectId, AlternativeProposalNo=ANo)
        NThProc = len(processes)
        # The highest process number is correct and
        # there are only distinct process numbers?
        if NThProc > 0:         #necessary, otherwise max function below gives crash !
            if (NThProc != max(process.ProcNo for process in processes) or
                NThProc != len(set(process.ProcNo for process in processes))):
                warn("Inconsistent process numbering found.")
        
        tmp = {
               "Questionnaire_id": projectId,
               "AlternativeProposalNo": ANo,
               "ProcNo": NThProc + 1
               }          
        newID = dbio.insertRow("qprocessdata", **tmp)
        newProcess = dbio.get("qprocessdata", id=newID)

        dbio.commit()
        return newProcess

    def deleteProcess(self, processID):
        """Delete the process with the given processID and all of it 
        streams.
        
        """
        if processID is None: return # security feature
        
        process = dbio.get("qprocessdata", id=processID)
        if process is not None:
            PId = process.Questionnaire_id
            ANo = process.AlternativeProposalNo

        dbio.delete("qprocessdata", id=processID)

        # delete detailed schedule entries
        # NOTE: since schedules and periods are only saved for the process in original state (ANo = -1),
        #       this only deletes anything for processID of the original process. Deleting processes in
        #       other alternatives keeps schedules for original and remaining alternatives.
        deleteProcessPeriodsFromDB(processID)
        #assign new EqNo in QGenerationHC table
        processes = self.getProcesses()
        for i in range(len(processes)): 
            processes[i].ProcNo = i + 1
        
        generaldata = dbio.get("cgeneraldata", Questionnaire_id=PId, AlternativeProposalNo=ANo) 
        try:
            generaldata.NThProc = len(processes)
        except IndexError:
            logTrack("A processes of a project without generaldata entry has been found and deleted.")
        
        inStreamIds = self._getStreamsForProcesses("process_streams_in",
                                                   [processID],
                                                   "streams_in_id")
        outStreamIds = self._getStreamsForProcesses("process_streams_out",
                                                   [processID],
                                                   "streams_out_id")
        for id in inStreamIds:
            self.deleteStream(id, "streams_in",
                              "process_streams_in", "streams_in_id")
        for id in outStreamIds:
            self.deleteStream(id, "streams_out",
                              "process_streams_out", "streams_out_id")
        
        nodeIds = [node.id for node in dbio.select("qnode", ObjectId=processID, ObjectType="PROC")]
        for nodeId in nodeIds:
            dbio.delete("qflow", FromId = nodeId)
            dbio.delete("qflow", ToId = nodeId)
            
        dbio.delete("qnode", ObjectId=processID, ObjectType="PROC")

        dbio.commit()

    def reorderProcesses(self, k0, k1):
        """reorganizes the processes moving process k0 to new position k1

        :params k0, k1 - process numbers in numbering from 0 to NK-1
        """

        processes = self.getProcesses()

        newNo, oldNo = reorder(k0, k1, len(processes))
        
        for k, process in enumerate(processes):
            process.ProcNo = newNo[k]+1 # +1, because ProcNo ranges from 1 - NK

    def deleteStream(self, id, streamTable, connectionTable, referenceColumn):
        """Delete the stream with ID id from table streamTable, and 
        remove the corresponding entry from the connectionTable.
        
        :param id: stream id in streamTable
        :param streamTable: usually streams_in or ...out.
        :param connectionTable: usually process_streams_in or ...out. 
        :param referenceColumn: a String, usually "streams_in_id" or 
        "streams_out_id"
        
        """
        sqlQueryQ = "id = '%s' " % (id, )
        deleteSQLRows(streamTable, sqlQueryQ)
        sqlQueryQ = "%s = '%s' " % (referenceColumn, id)
        deleteSQLRows(connectionTable, sqlQueryQ)

    def addPipeDummy(self):
        """ adds a new empty pipe field in table qdistributionhc
        """

#..............................................................................
# deleting Q- and corresponding C-Tables

        pipes = self.getPipes()
        NPipes = len(pipes)
        tmp = {
            "Questionnaire_id":Status.PId,
            "AlternativeProposalNo":Status.ANo,
            "PipeDuctNo": NPipes + 1
            }          
        newID = dbio.insertRow("qdistributionhc", **tmp)
        newPipe = dbio.get("qdistributionhc", id=newID)

        NPipes += 1

        generaldata = self.getGeneralData()
        generaldata.NPipeDuct = NPipes

        dbio.commit()

        return newPipe

    def deletePipe(self, pipeID):
        """ deletes all entries for a given fuel type
        """

#..............................................................................
# deleting Q- and corresponding C-Tables

##        DB = Status.DB
##        sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s' AND QDistributionHC_ID = '%s'"\
##                    % (Status.PId, Status.ANo, pipeID)  #query is redundant, but maintained as is for security
##
##        deleteSQLRows(DB.qdistributionhc, sqlQueryQ)
        dbio.delete("qdistributionhc", id=pipeID, Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        pipes = self.getPipes()
        for i in range(len(pipes)): #assign new EqNo in QGenerationHC table
            pipes[i].PipeDuctNo = i + 1
        
        generaldata = self.getGeneralData()
        generaldata.NPipeDuct = len(pipes)

        dbio.commit()

#------------------------------------------------------------------------------
    def getPipes(self, PId=None, ANo=None):
        """Return a list of the database entries of all pipes."""
        if PId is None:
            pipes = dbio.select("qdistributionhc", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        elif ANo is None:
            pipes = dbio.select("qdistributionhc", Questionnaire_id=PId)
        else:
            pipes = dbio.select("qdistributionhc", Questionnaire_id=PId, AlternativeProposalNo=ANo)
            
        pipes = sorted(pipes, key = lambda pipe: pipe.PipeDuctNo)
        
        return pipes

    def getPipe(self, id=None, PId=None, ANo = None, no=None):
        """Return a process specified by ProcId or ProcNo
        
        
        :param PId: If the Questionaire_id is not given explicitly, the current project (Status.PId is selected)
        
        :returns: The specified process, or None in case that no process has been found 
        """
        if id is None:

            if no is None:
                logWarning("Project.getProcess: either ProcId or ProcNo has to be specified")
                return None
            else:
            
                if PId is None:
                    pipe = dbio.get("qdistributionhc", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo, PipeDuctNo=no)
                elif ANo is None:
                    pipe = dbio.get("qdistributionhc", Questionnaire_id = PId, AlternativeProposalNo = Status.ANo, PipeDuctNo=no)
                else:
                    pipe = dbio.get("qdistributionhc", Questionnaire_id = PId, AlternativeProposalNo = ANo, PipeDuctNo=no)

        else:
            proc = dbio.get("qdistributionhc", id=id)
        
        return pipe

    def getPipeList(self, key, PId=None):
        """
        returns a list of existing pipes
        """

        pipes = self.getPipes(PId)
        
        pipeList = []
        for pipe in pipes:
##            if isinstance(pipe[key], str) or isinstance(pipe[key], unicode):
##                pipeList.append(unicode(pipe[key], "utf-8"))
##            else:
            pipeList.append(getattr(pipe, key))

        return pipeList
    
#------------------------------------------------------------------------------
    def getPipeDict(self):
        """Return a dict of the pipes in the current project with the 
        database ids as the keys and the names as the values.
        
        """
        pipes = self.getPipes()     
        pipeDict = {}
        for pipe in pipes:
            pipeName = pipe.Pipeduct
            pipeID = pipe.QDistributionHC_ID
            pipeDict.update({pipeID:pipeName})

        return pipeDict

    def deleteBuilding(self, buildingID):
        """Remove the building with its processes from the current 
        alternative proposal.
        
        """
##        DB = Status.DB
##        sqlQueryQ = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s' AND QBuildings_ID = '%s'"\
##                    % (Status.PId, Status.ANo, buildingID)  #query is redundant, but maintained as is for security
##        deleteSQLRows(DB.qbuildings, sqlQueryQ)
        dbio.delete("qbuildings", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo, id=buildingID)

        for processID in self._getProcessesForBuilding(buildingID):
            self.deleteProcess(processID)
            
        buildings = dbio.select("qbuildings", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
        for i in range(len(buildings)): #assign new EqNo in QGenerationHC table
            buildings[i].BuildingNo = i+1
        
        questionnaire = self.getQuestionnaire()
        questionnaire.NBuild = len(buildings)
        dbio.commit()
    
    def getHolidays(self,datetime=False):
        """Return a list of the holiday periods of the current project.
        
        :returns: a List of 2-tuples with the first and the last day of
            each holiday. Where 1 is the 1st of January and 365 is
            the 31st of December.
            
        """
        result = []
        questionnaire, dummy = self.getProjectData()
        if questionnaire is None:   # check added for security reasons
            return result
        for start, stop in ((questionnaire.NoProdStart_1, questionnaire.NoProdStop_1), 
                            (questionnaire.NoProdStart_2, questionnaire.NoProdStop_2), 
                            (questionnaire.NoProdStart_3, questionnaire.NoProdStop_3)):
            if None not in (start, stop):
                if datetime:
                    result.append((start,stop))
                else:
                    result.append((start.timetuple()[7], stop.timetuple()[7]))
                
        return result
            
    def addHXDummy(self):
        """ adds a new empty HX field in table qheatexchanger
        """
        hxes = self.getHXes()
        NHX = len(hxes)
        tmp = {
            "ProjectID":Status.PId,
            "AlternativeProposalNo":Status.ANo,
            "HXNo": NHX + 1
            }          
        newID = dbio.insertRow("qheatexchanger", **tmp)
        newhx = dbio.get("qheatexchanger", id=newID)

        NHX += 1

        generaldata = self.getGeneralData()
        generaldata.NHX = NHX

        dbio.commit()

        return newhx

    def reorderHX(self):
        """
        reorders the numbering of the heat exchangers in SQL
        """
        hxes = self.getHXes(commit=False)
        for i in range(len(hxes)): #assign new EqNo in QGenerationHC table
            hxes[i].HXNo = i + 1
        
        generaldata = self.getGeneralData()
        generaldata.NHX = len(hxes)

        self.deleteDanglingPinchstreams(commit=False)
        dbio.commit()

    def deleteDanglingPinchstreams(self, PId=None, ANo=None, commit=True):
        """Remove any entry in the pinchstream table, that is not
        referred to by at least one entry in the 
        heatexchanger_pinchstream table.
        
        """
        PId = PId if PId is not None else Status.PId
        ANo = ANo if ANo is not None else Status.ANo
##        connectionEntries = Status.DB.heatexchanger_pinchstream.ProjectID[PId].\
##            AlternativeProposalNo[ANo]
        connectionEntries = dbio.select("heatexchanger_pinchstream", ProjectID=PId, AlternativeProposalNo=ANo)
        usedPinchstreams = [str(conn.pinchstream_id) for conn in connectionEntries]
        if usedPinchstreams:
            sqlQuery = ("ProjectID = %s AND AlternativeProposalNo = %s AND id not in (%s)" % 
                        (PId, ANo, ", ".join(usedPinchstreams), ))
        else:
            sqlQuery = "ProjectID = %s AND AlternativeProposalNo = %s" % (PId, ANo)
##        deleteSQLRows(Status.DB.pinchstream, sqlQuery)
        deleteSQLRows("pinchstream", sqlQuery)
        if commit: dbio.commit()
        
    def getHXes(self, PId=None, ANo=None, commit=True):
        """ returns a list of existing heat exchangers
        """
        PId = PId if PId is not None else Status.PId
        ANo = ANo if ANo is not None else Status.ANo
        
##        sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s' ORDER BY HXNo ASC" % (PId, ANo)
##        hxes = Status.DB.qheatexchanger.sql_select(sqlQuery)
        hxes = sorted(dbio.select("qheatexchanger", ProjectID=PId, AlternativeProposalNo=ANo), key=lambda hx: hx.HXNo)

        changes = False
        for h,hx in enumerate(hxes):
            if hx.HXNo != h+1:
                logTrack("Project.getHXes: HX %s No. changed from %s to %s"%(hx.HXName,hx.HXNo,h+1))
                hx.HXNo = h+1
                changes = True


        if changes and commit: dbio.commit()
        
        return hxes

    def getHXList(self, key, PId=None, ANo=None, returnUnicode=True):
        """Return the list of values for key key from the list of 
        heatexchangers.
        
        """
        PId = PId if PId is not None else Status.PId
        ANo = ANo if ANo is not None else Status.ANo
        
        hxes = self.getHXes(PId, ANo)
        
        HXList = []
        for hx in hxes:
            attr = getattr(hx, key)
            if (isinstance(attr, str) or isinstance(attr, unicode)) and returnUnicode:
                HXList.append(unicode(attr, "utf-8"))
            else:
                HXList.append(attr)

        return HXList
    
    def getPinchList(self, key, PId=None, ANo=None):
        """Return the list of values for key key from the list of 
        pinch streams.
        
        """
        PId = PId if PId is not None else Status.PId
        ANo = ANo if ANo is not None else Status.ANo
##        pinches = Status.DB.pinchstream.ProjectID[PId].AlternativeProposalNo[ANo]
        pinches = dbio.select("pinchstream", ProjectID=PId, AlternativeProposalNo=ANo)
        
        pinchList = []
        for pinch in pinches:
            attr = getattr(pinch, key)
            if isinstance(attr, str) or isinstance(attr, unicode):
                pinchList.append(U(attr))
            else:
                pinchList.append(attr)

        return pinchList

    def getHX(self, HXID=None, PId=None, ANo=None, no=None):
        """
        Return the pSQL.Result object for the HX with the given HXID.

        :Parameters:
        :param processID: The QProcessData_ID to look for.
        :Returns: The corresponding pSQL.Result, None if no fitting process has been found.
        """
        if HXID is not None:
            return dbio.get("qheatexchanger", id=HXID)
        elif no is not None:
            ANo = noneCheck(ANo, Status.ANo)
            PId = noneCheck(PId, Status.PId)
            return dbio.get("qheatexchanger", HXNo=no, ProjectID=PId, AlternativeProposalNo=ANo)
        else:
            return None
            

    def getHXs(self, PId=None):
        """Return a pSQL.Result of all heat exchangers in the current 
        project.
        
        If the PId is given all heat exchangers of all alternative proposals 
        are returned, otherwise only of the current one. Per default 
        the results are ordered by the heat exchanger number.
        
        :param PId: which Questinonnaire_id is requested if not the current one
        which is the default.
        
        """
        return getHXes(PId=PId)
    
    def getHXEntryByID(self, ID, key):
        """Return the value for key key for a single heat exchanger, 
        identified by the primary key.
        
        :param id: Primary key to identify the heatexchanger.
        :param key: columnname in the heatexchanger database.
        :returns: Single String Database value
        
        """
        #TODO: check if this function is really needed
        
        hx = self.getHX(ID)
        return getattr(hx, key)

    def getHXConnections(self,HXId):

        """returns list of connections from table heatexchanger_pinchstream"""

        connections = dbio.select("heatexchanger_pinchstream", qheatexchanger_QHeatExchanger_Id=HXId)
        
        if len(connections) < 1:
            logDebug("Project.getHXConnections: Erroneous no. of results in getHXConnections[%s]: %s "%(HXId,len(connections)))

        return connections
        
    def getHXStream(self,streamId):
        """returns a source or sink stream of a heat exchanger from a given Id """

        stream = dbio.get("pinchstream", id=streamId)
        if stream is None:
            logDebug("Project.getStream: erroneous no. of results for stream Id = %s: %s"%(streamId,0))
        return stream
        
    def addWHEEDummy(self):
        """ adds a new empty WHEE field in table qwasteheatelequip
        """

        whees = self.getWHEEs()
        NWHEE = len(whees)
        tmp = {
            "ProjectID":Status.PId,
            "AlternativeProposalNo":Status.ANo,
            "WHEENo": NWHEE + 1
            }          
        newID = dbio.insertRow("qwasteheatelequip", **tmp)
        newwhee = dbio.get("qwasteheatelequip", id=newID)

        NWHEE += 1

        generaldata = self.getGeneralData()
        generaldata.NWHEE = NWHEE

        dbio.commit()

        return newwhee

    def deleteWHEE(self, WHEEID):
        """ deletes all entries for a given heat exchanger
        """

#..............................................................................
# deleting Q- and corresponding C-Tables

        dbio.delete("qwasteheatelequip", ProjectID=Status.PId, AlternativeProposalNo=Status.ANo, id=WHEEID)

        sqlQuery = "ProjectID = '%s' AND AlternativeProposalNo = '%s' ORDER BY WHEENo ASC" % (Status.PId, Status.ANo)
        whees = self.getWHEEs()

        for i in range(len(whees)): #assign new EqNo in QGenerationHC table
            whees[i].WHEENo = i + 1
            pass

        generaldata = self.getGeneralData()
        generaldata.NWHEE = len(whees)

        dbio.commit()

    def getWHEEs(self, PId=None):
        """ returns a list of existing heat exchangers
        """

        if PId is None:
            whees=dbio.select("qwasteheatelequip",ProjectID=Status.PId, AlternativeProposalNo = Status.ANo)
        else:
            whees=dbio.select("qwasteheatelequip", ProjectID=PId)

        whees = sorted(whees, key = lambda whee: whee.WHEENo)
        
        return whees

    def getWHEE(self, id=None, PId=None, ANo=None, no=None, name=None):
        """ returns a list of existing heat exchangers
        """

        if id is not None:
            return dbio.get("qwasteheatelequip",id=id)

        elif no is not None:
            ANo = noneCheck(ANo, Status.ANo)
            PId = noneCheck(PId, Status.PId)
            return dbio.get("qwasteheatelequip", ProjectID=PId, AlternativeProposalNo=ANo, WHEENo=no)
        
        elif name is not None:
            ANo = noneCheck(ANo, Status.ANo)
            PId = noneCheck(PId, Status.PId)
            return dbio.get("qwasteheatelequip", ProjectID=PId, AlternativeProposalNo=ANo, WHEEName=name)
        
        else:
            return None

    def getWHEEList(self, key, PId=None, returnUnicode=True):
        """ returns a list of existing heat exchangers
        """

        whees = self.getWHEEs(PId)
        
        WHEEList = []
        for whee in whees:
            item = getattr(whee, key)
            if isinstance(item, str) or isinstance(item, unicode) and returnUnicode:
                WHEEList.append(unicode(item, "utf-8"))
            else:
                WHEEList.append(item)

        return WHEEList

    def addSurfaceDummy(self):
        """ adds a new empty Surface to qsurfarea
        """

        surfaces = self.getSurfaces()
        NSurfaces = len(surfaces)
        tmp = {
            "ProjectID":Status.PId,
            "SurfAreaNo":(NSurfaces + 1),
            "SurfAreaName":"dummy"
            }          
        newID = dbio.insertRow("qsurfarea", **tmp)
        newSurface = dbio.get("qsurfarea", id=newID)

        dbio.commit()

        return newSurface

    def deleteSurface(self, surfaceID):
        """ deletes all entries for a given heat exchanger
        """

        dbio.delete("qsurfarea", ProjectID=Status.PId, id=surfaceID)

        surfaces = self.getSurfaces()

        for i in range(len(surfaces)): #assign new EqNo in QGenerationHC table
            surfaces[i].SurfAreaNo = i + 1

        dbio.commit()

    def getSurfaces(self, PId=None):
        """ returns a list of existing heat exchangers
        """

        if PId is None:
            surfaces = dbio.select("qsurfarea", ProjectID=Status.PId)
        else:
            surfaces = dbio.select("qsurfarea", ProjectID=PId)

        surfaces = sorted(surfaces, key = lambda s: s.SurfAreaNo)           
        if PId is None:
            self.surfaces = surfaces
        
        return surfaces

    def getSurfaceList(self, key, PId=None):
        """ returns a list of existing heat exchangers
        """

        surfaces = self.getSurfaces(PId)
        
        surfaceList = []
        for surface in surfaces:
            item = getattr(surface, key)
            if isinstance(item, str) or isinstance(item, unicode):
                surfaceList.append(unicode(item, "utf-8"))
            else:
                surfaceList.append(item)

        return surfaceList

    def getFuels(self, PId=None, ANo = None):
        """Return a table of all fuels used in the project in the
        current alternative proposal.
        
        :param PId: The project Id if all used fluids of all 
        alternative proposals of a certain project are wanted.
        
        """
        if PId is None:
            fuels = dbio.select("qfuel", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo)
        elif ANo is None:
            fuels = dbio.select("qfuel", Questionnaire_id = PId)
        else:
            fuels = dbio.select("qfuel", Questionnaire_id = PId, AlternativeProposalNo = ANo)
            
        fuels = sorted(fuels, key = lambda fuel: fuel.FuelNo)
        
        return fuels

    def getFuel(self, id=None, PId=None, ANo=None, no=None):
        """ returns a fuel identified either by id or by the fuel number no
        """

        if id is not None:
            qfuel = dbio.get("qfuel", id=id)
        elif no is not None:
            ANo = noneCheck(ANo, Status.ANo)
            PId = noneCheck(PId, Status.PId)
            qfuel = dbio.get("qfuel", Questionnaire_id=PId, AlternativeProposalNo=ANo, FuelNo=no)
        else:
            qfuel = None
        return qfuel
            
        
    def getFuelList(self, key, PId=None):
        """ returns a table of existing fuels
        """

        fuels = self.getFuels(PId)
        
        fuelList = []
        for fuel in fuels:
            item = getattr(fuel, key)
            if isinstance(item, str) or isinstance(item, unicode):
                fuelList.append(unicode(item, "utf-8"))
            else:
                fuelList.append(item)

        return fuelList

    def getProcesses(self, PId=None, ANo = None):
        """Return a pSQL.result list of all processes for the current 
        alternative proposal. 
        
        The list is ordered ascending by the same process number which 
        is also used in the lists in the global Status object.
        
        :param PId: If the Questionaire_id is given explicitly, all processes
        of all proposals are returned.
        :returns: A list of pSQL.result objects for all processes of the current
        alternative proposal.
        
        """
        if PId is None:
            processes = dbio.select("qprocessdata", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo)
        elif ANo is None:
            processes = dbio.select("qprocessdata", Questionnaire_id = PId)
        else:
            processes = dbio.select("qprocessdata", Questionnaire_id = PId, AlternativeProposalNo = ANo)

        self.processes = sorted(processes, key = lambda proc: proc.ProcNo)
        
        return self.processes

    def getProcess(self, ProcId=None, PId=None, ANo = None, ProcNo=None):
        """Return a process specified by ProcId or ProcNo
        
        
        :param PId: If the Questionaire_id is not given explicitly, the current project (Status.PId is selected)
        
        :returns: The specified process, or None in case that no process has been found 
        """
        if ProcId is None:

            if ProcNo is None:
                logWarning("Project.getProcess: either ProcId or ProcNo has to be specified")
                return None
            else:
            
                if PId is None:
                    proc = dbio.get("qprocessdata", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo, ProcNo=ProcNo)
                elif ANo is None:
                    proc = dbio.get("qprocessdata", Questionnaire_id = PId, AlternativeProposalNo = Status.ANo, ProcNo=ProcNo)
                else:
                    proc = dbio.get("qprocessdata", Questionnaire_id = PId, AlternativeProposalNo = ANo, ProcNo=ProcNo)

        else:
            proc = dbio.get("qprocessdata", id=ProcId)
        
        return proc

    def getProcessList(self, key, PId=None, returnUnicode=True):
        """Return a list of the content of a single column of all the 
        processes of the current project and alternative proposal.
        
        If a specific PId is given the processes of all alternative
        proposals of the project are searched through.
        
        :param key: the column name
        :param PId: the ProcessID if not the current project is wanted.
        :returns: A list of column values which may be of Int, String 
        or what ever.
        
        """
        #TODO: unicode should be completely eliminated from project py, for the moment
        # this has been defined as optional
        processes = self.getProcesses(PId)
        
        processList = []
        for process in processes:
            item = getattr(process, key)
            if isinstance(item, (str, unicode)) and returnUnicode==True:
                processList.append(unicode(item, "utf-8"))
            else:
                processList.append(item)

        return processList
    
    def getOutStreamList(self, key, PId=None):
        """Return a list of the content of a single column of all the 
        outstreams of the current project and alternative proposal.
        
        If a specific PId is given the processes of all alternative
        proposals of the project are searched through.
        
        :param key: the column name
        :param PId: the ProcessID if not the current project is wanted.
        :returns: A list of column values which may be of Int, String 
        or what ever.
        
        """

        streams = self.getOutStreams(PId)
        
        streamList = []
        for stream in streams:
            item = getattr(stream, key)
            if isinstance(item, (str, unicode)):
                streamList.append(unicode(item, "utf-8"))
            else:
                streamList.append(item)

        return streamList
    
    def getOutStreams(self, PId=None):
        if PId is None:
            outStreams = dbio.select("streams_out", ProjectID = Status.PId, AlternativeProposalNo = Status.ANo)
        else:
            outStreams = dbio.select("streams_out", ProjectID = PId)

        return outStreams
        
    def getInStreamList(self, key, PId=None):
        """Return a list of the content of a single column of all the 
        instreams of the current project and alternative proposal.
        
        If a specific PId is given the processes of all alternative
        proposals of the project are searched through.
        
        :param key: the column name
        :param PId: the ProcessID if not the current project is wanted.
        :returns: A list of column values which may be of Int, String 
        or what ever.
        
        """
        streams = self.getInStreams(PId)
        
        streamList = []
        for stream in streams:
            item = getattr(stream, key)
            if isinstance(item, (str, unicode)):
                streamList.append(unicode(item, "utf-8"))
            else:
                streamList.append(item)

        return streamList
    
    def getInStreams(self, PId=None):
        if PId is None:
            inStreams = dbio.select("streams_in", ProjectID = Status.PId, AlternativeProposalNo = Status.ANo)
        else:
            inStreams = dbio.select("streams_in", ProjectID = PId)

        return inStreams            
        
    def getProcessNo(self, processID):
        """
        Return the process number for the process with the given processID.

        :Parameters:
        :param processID: The QProcessData_ID to look for.
        :Returns: The corresponding pSQL.Result, None if no fitting process has been found.
        """
        process = self.getProcess(processID)
        if process is None:
            return None
        else:
            return process.ProcNo
    
    def getProcessWithName(self, processName):
        """
        Return the pSQL.Result object for the process with the given 
        name in the current project and alternative proposal.

        :Parameters:
        :param processName: The Process to look for.
        :Returns: The corresponding pSQL.Result, None if no fitting process has been found.
        
        """
        process = dbio.get("qprocessdata", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo, Process=processName)
        return process

    def getElectricity(self, PId=None, ANo=None):
        """ returns the electricity mix used in the project / alternative
        """

        if PId is None:
            q = dbio.get("qelectricity", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo)
        elif ANo is None:
            q = dbio.select("qelectricity", Questionnaire_id = PId)
        else:
            q = dbio.get("qelectricity", Questionnaire_id = PId, AlternativeProposalNo = ANo)
                
        return q

    def getQFuelList(self, key, PId=None):
        """ returns a list of existing fuels
        """
        
        fuels = self.getQFuels(PId)
        
        fuelList = []
        for fuel in fuels:
            item = getattr(fuel, key)
            if isinstance(item, str) or isinstance(item, unicode):
                fuelList.append(unicode(item, "utf-8"))
            else:
                fuelList.append(item)

        return fuelList

    def getQFuels(self,PId = None):
        """Return a list of fluids used in the project"""
        self.qfuel = self.getFuels(PId=PId)        
        return self.qfuel

    def getFluidDict(self, includeZeroWater=False):
        """Return a Dict of database IDs and names of the fluids that
        are already defined in the database.
        
        :param includeZeroWater: If yes the Fluid "Water" is added with the
        key 0. This may be necessary for copatibility with the Fluid class. 
         
        """
        fluids = dbio.select("dbfluid")       
        fluidDict = {}
        for fluid in fluids:
            if fluid.FluidName is None: # security check
                continue
            
##            fluidName = unicode(fluid.FluidName, "utf-8")
            fluidName = fluid.FluidName
            fluidID = int(fluid.id)
            fluidDict.update({fluidID:fluidName})
        if includeZeroWater is True:
            fluidDict[0] = "Water (default)"
        return fluidDict

    def getFuelDict(self):
        """Return a Dict of database IDs and names of the fuels that
        are already defined in the database.
        
        """
        fuels = dbio.select("dbfuel")       
        fuelDict = {}
        for fuel in fuels:
            fuelName = unicode(fuel.FuelName, "utf-8")
            fuelID = fuel.id
            fuelDict.update({fuelID:fuelName})

        return fuelDict

    def getUnitOpDict(self, code=False):
        """ returns a list of unit operations in DB
        """

        unitOperations = dbio.select("dbunitoperation")      
        unitOpDict = {}
        for unitOperation in unitOperations:
            unitOperationName = unitOperation.UnitOperation
            if code:
                key = int(unitOperation.UnitOperationCode)
                text = "[%4d] "%int(key) + unitOperationName
            else:
                key = unitOperation.id
                text = unitOperationName
            unitOpDict.update({key: text})

        return unitOpDict

    def getNACEDict(self, branch=None):
        """ returns a list of nace codes in DB
        """

        naceTable = dbio.select("dbnacecode")       
        naceDict = {}
        naceSubDict = {}
        for entry in naceTable:
            naceCode = entry.CodeNACE
            naceSubCode = naceCode + "." + entry.CodeNACEsub
            naceName = entry.NameNACE
            naceSubName = entry.NameNACEsub
            naceDict.update({naceCode:naceName})
            if branch == naceName:
                naceSubDict.update({naceSubCode:naceSubName})

#        print "Project (getNACEDict): branch = ",branch
#        print "---"
#        print "naceDict = ",naceDict
#        print "---"
#        print "naceSubDict = ",naceSubDict
#        print "------------------------------------------------"
            
        return (naceDict, naceSubDict)

    def getElectricityMixList(self, PId=None):
        mixes =  self.getElectricity(PId)
        mixIds = set([mix.dbelectricitymix_id for mix in mixes])
        if None in mixIds:
            mixIds.remove(None)
        return [int(id) for id in mixIds]

    def getDataSetIDs(self, PId=None):
        """ gets all the ids of data sets used in a given project
        """
        dbfiles = dbio.select("qmdatafiles", ProjectID=PId)
        idList = []
        for dbfile in dbfiles:
            dbsets = dbio.select("qmdatasets", DataFileId=dbfile.id)
            idList += [int(dbset.id) for dbset in dbsets]
        return idList

    def getBuildings(self, PId=None, ANo = None):
        """Return a pSQL.result list of all processes for the current 
        alternative proposal. 
        
        The list is ordered ascending by the same process number which 
        is also used in the lists in the global Status object.
        
        :param PId: If the Questionaire_id is given explicitly, all processes
        of all proposals are returned.
        :returns: A list of pSQL.result objects for all processes of the current
        alternative proposal.
        
        """
        if PId is None:
            buildings = dbio.select("qbuildings", Questionnaire_id = Status.PId, AlternativeProposalNo = Status.ANo)
        elif ANo is None:
            buildings = dbio.select("qbuildings", Questionnaire_id = PId)
        else:
            buildings = dbio.select("qbuildings", Questionnaire_id = PId, AlternativeProposalNo = ANo)
            sqlQuery = "Questionnaire_id = '%s' AND AlternativeProposalNo = '%s' ORDER BY BuildingNo ASC" % (PId, ANo)

        self.buildings = sorted(buildings, key = lambda bui: bui.BuildingNo)
        
        return self.buildings

    def getBuilding(self, PId=None, ANo=None, no=None):
        """ returns a list of existing heat exchangers
        """

        ANo = noneCheck(ANo, Status.ANo)
        PId = noneCheck(PId, Status.PId)
        return dbio.get("qbuildings", Questionnaire_id=PId, AlternativeProposalNo=ANo, BuildingNo=no)
        

    def getFluidAndFuelList(self, PId=None):
        """ returns a list of all fluids and fuels used in a project
        if PId = None: returns for some alternative in the active project
        if PId specified: returns for the total project (all alternatives)
        """

        fuelList = self.getQFuelList("DBFuel_id", PId)
        fuelIDs = set() # avoid double counting
        for fuelID in fuelList:
            if fuelID is not None:
                fuelIDs.add(int(fuelID))

        fluidList = []
        fluidList.extend(Status.prj.getEquipmentList("Refrigerant", PId))
        fluidList.extend(Status.prj.getPipeList("HeatDistMedium", PId))
        fluidList.extend(Status.prj.getProcessList("ProcMedDBFluid_id", PId))
        fluidList.extend(Status.prj.getProcessList("ProcMedOut", PId))
        fluidList.extend(Status.prj.getProcessList("SupplyMedDBFluid_id", PId))
        fluidList.extend(Status.prj.getWHEEList("WHEEMedium", PId))
        fluidList.extend(Status.prj.getOutStreamList("dbfluid_id", PId))
        fluidList.extend(Status.prj.getInStreamList("dbfluid_id", PId))
        
        fluidIDs = set() # avoid double counting
        for fluidID in fluidList:
            if fluidID is not None:
                fluidIDs.add(int(fluidID))
           
        return (list(fluidIDs), list(fuelIDs))
        

    def substituteForeignKey(self, table, primaryKey, primaryKeySelection, foreignKey, foreignKeySubstitutions, idAsString=False, specialValues=[]):
        """
        Helper function for -> restoreLinks
        substitutes oldKey by newKey for some column "foreignKey", but ONLY if primaryKey is within primaryKeySelection
        idAsString: if true, works for some old tables where id's are stored as string instead of int
        specialKeys: keys in this list are NOT substituted (applied e.g. for fluidId 0)
        """
        if len(foreignKeySubstitutions)==0 or len(primaryKeySelection)==0: return # nothing has been copied (happens e.g. for db-data when copying within the same database)
##        rows = dbio.select(table)
        rows = dbio.selectByList(table, primaryKey, primaryKeySelection)
        for row in rows:
##            rowId = getattr(row, primaryKey)
##            if rowId in primaryKeySelection:
            fKey = getattr(row, foreignKey)
            if fKey is not None and fKey not in specialValues:
                try:
                    if idAsString: # special case for old-style tables: convert from/to string
                        setattr(row, foreignKey, str(foreignKeySubstitutions[int(fKey)]))          
                    else:   # default / normal behaviour
                            setattr(row, foreignKey, foreignKeySubstitutions[fKey])
                except:
                    logError("Cannot update link for table %s: %s=%s. substitutions = %s"%(table, foreignKey, fKey, foreignKeySubstitutions))
                    setattr(row, foreignKey, None)
      
    def restoreLinks(self, PId, substitutions):
        """
        restores the links in an imported project
        PId = None indicates that only current alternative is processed
        """

        sDict = defaultdict(lambda: {})     # default collection returning empty dictionary for a non-existing key
        for key in substitutions.keys():
            sDict.update({key:dict(substitutions[key])})    # convert lists of 2-tuples into dict
               
        objectIdSubstitutions = {}
        objectIdSubstitutions.update(sDict["qprocessdata"])
        for entry in sDict["qgenerationhc"]:        # equipment is referenced with negative value of QGenerationHC_ID in process_periods
            objectIdSubstitutions.update({-entry:-sDict["qgenerationhc"][entry]})

        newProcessPeriodsIDList = sDict['process_periods'].values()
        self.substituteForeignKey('process_periods', 'id', newProcessPeriodsIDList, 'periods_id', sDict["periods"])
        self.substituteForeignKey('process_periods', 'id', newProcessPeriodsIDList, 'qprocessdata_QProcessData_ID', objectIdSubstitutions)

        newProcessPeriodProfilesIDList = sDict["process_period_profiles"].values()
        self.substituteForeignKey('process_period_profiles', 'id', newProcessPeriodProfilesIDList, 'process_periods_id', sDict["process_periods"])
        self.substituteForeignKey('process_period_profiles', 'id', newProcessPeriodProfilesIDList, 'profiles_id', sDict["profiles"])
            
        newProfileIntervalsIDList = sDict['profile_intervals'].values()
        self.substituteForeignKey('profile_intervals', 'id', newProfileIntervalsIDList, 'profiles_id', sDict["profiles"])
        self.substituteForeignKey('profile_intervals', 'id', newProfileIntervalsIDList, 'intervals_id', sDict["intervals"])

        # substitute links between nodes and flows (and products and processes)
        newQFlowIDList =  sDict['qflow'].values()
        self.substituteForeignKey('qflow', 'id', newQFlowIDList, 'ToId', sDict["qnode"])
        self.substituteForeignKey('qflow', 'id', newQFlowIDList, 'FromId', sDict["qnode"])

        newQNodeIDList =  sDict["qnode"].values()
        newQNodeList = dbio.selectByList("qnode", "id", newQNodeIDList)
        qnodesInt = [qnode.id for qnode in newQNodeList if qnode.ObjectType=="INT"]
        qnodesProc = [qnode.id for qnode in newQNodeList if qnode.ObjectType=="PROC"]
        qnodesProd = [qnode.id for qnode in newQNodeList if qnode.ObjectType=="PROD"]

        self.substituteForeignKey('qnode', 'id', qnodesInt, 'ObjectId', sDict["qnode"])
        self.substituteForeignKey('qnode', 'id', qnodesProc, 'ObjectId', sDict["qprocessdata"])
        self.substituteForeignKey('qnode', 'id', qnodesProd, 'ObjectId', sDict["qproduct"])

        # substitute links in data sets to objects
        newDataSetsIDList = sDict["qmdatasets"].values()
        newDataSetsList = dbio.selectByList("qmdatasets", "id", sDict["qmdatasets"].values())
                                
        self.substituteForeignKey('qmdatasets', 'id', newDataSetsIDList, 'DataFileId', sDict["qmdatafiles"])

        dsFUEL = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in FUELPARS]
        dsEQ = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in EQPARS]
        dsPIPE = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in PIPEPARS]
        dsPROCMAIN = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in PROCMAINPARS]
        dsPROCIN = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in PROCINPARS]
        dsPROCOUT = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in PROCOUTPARS]
        dsPROD = [dsentry.id for dsentry in newDataSetsList if dsentry.ParCode in ["Product"]]

        self.substituteForeignKey('qmdatasets', 'id', dsFUEL, 'ObjectId', sDict["qfuel"])
        self.substituteForeignKey('qmdatasets', 'id', dsEQ, 'ObjectId', sDict["qgenerationhc"])
        self.substituteForeignKey('qmdatasets', 'id', dsPIPE, 'ObjectId', sDict["qdistributionhc"])
        self.substituteForeignKey('qmdatasets', 'id', dsPROCMAIN, 'ObjectId', sDict["qprocessdata"])
        self.substituteForeignKey('qmdatasets', 'id', dsPROCIN, 'ObjectId', sDict["streams_in"])
        self.substituteForeignKey('qmdatasets', 'id', dsPROCOUT, 'ObjectId', sDict["streams_out"])
        self.substituteForeignKey('qmdatasets', 'id', dsPROD, 'ObjectId', sDict["qproduct"])

        # restablish links in combined schedules        
        qCombSchedulesIDList = sDict["qcombschedules"].values()
        qCombSchedules = dbio.selectByList("qcombschedules", "id", qCombSchedulesIDList)

        procSchedules = [q.id for q in qCombSchedules if q.Type == 0]
        dataSchedules = [q.id for q in qCombSchedules if q.Type in [1, 10, 11]]
        dataProcSchedules = [q.id for q in qCombSchedules if q.Type == 10]
        dataDataSchedules = [q.id for q in qCombSchedules if q.Type == 11]
##        plantSchedules = [q.id for q in qCombSchedules if q.Type == -1]

        self.substituteForeignKey('qcombschedules', 'id', qCombSchedulesIDList, 'ProcessId', sDict["qprocessdata"])
        self.substituteForeignKey('qcombschedules', 'id', procSchedules, 'ObjectId', sDict["qprocessdata"])
        self.substituteForeignKey('qcombschedules', 'id', dataSchedules, 'ObjectId', sDict["qmdatasets"])

        self.substituteForeignKey('qcombschedules', 'id', dataProcSchedules, 'ComponentNo', sDict["qprocessdata"])
        self.substituteForeignKey('qcombschedules', 'id', dataDataSchedules, 'ComponentNo', sDict["qmdatasets"])
            
        # substitute objectIds (and streamIds for processes)
        psIds = sDict["pinchstream"].values()
        psList = dbio.selectByList("pinchstream", "id", psIds)
        
        psProc = [ps for ps in psList if ps.source_type in ["ProcessH", "ProcessC"]]
        psEq =  [ps for ps in psList if ps.source_type in ["Equipment"]]
        psPipe =   [ps for ps in psList if ps.source_type in ["PipeH", "PipeC"]]
        psWHEE =    [ps for ps in psList if ps.source_type in ["WHEE"]]
        psHX =     [ps for ps in psList if ps.source_type in ["HX"]]
        
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psProc], 'source_id', sDict["qprocessdata"])
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psEq], 'source_id', sDict["qgenerationhc"])
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psPipe], 'source_id', sDict["qdistributionhc"])
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psWHEE], 'source_id', sDict["qwasteheatelequip"])
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psHX], 'source_id', sDict["qheatexchanger"])

        psProcIn = [ps for ps in psProc if (((ps.Hot_Cold in ["Sink", "Cold"] and ps.source_type=="ProcessH") or \
                                            (ps.Hot_Cold in ["Source", "Hot"] and ps.source_type=="ProcessC")) and \
                                            ps.pID > 0)]

        psProcOut = [ps for ps in psProc if ((ps.Hot_Cold in ["Source", "Hot"] and ps.source_type=="ProcessH") or \
                                            (ps.Hot_Cold in ["Sink", "Cold"] and ps.source_type=="ProcessC"))]
                    
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psProcIn], 'pID', sDict["streams_in"])
        self.substituteForeignKey('pinchstream', 'id', [ps.id for ps in psProcOut], 'pID', sDict["streams_out"])
            

        # replace fluidIDs everywhere
        self.substituteForeignKey('qgenerationhc', 'id', sDict["qgenerationhc"].values(), 'Refrigerant', sDict["dbfluid"], specialValues=[0])

        self.substituteForeignKey("qprocessdata", 'id', sDict["qprocessdata"].values(), 'ProcMedDBFluid_id', sDict["dbfluid"], specialValues=[0])
        self.substituteForeignKey("qprocessdata", 'id', sDict["qprocessdata"].values(), 'ProcMedOut', sDict["dbfluid"], specialValues=[0])
        self.substituteForeignKey("qprocessdata", 'id', sDict["qprocessdata"].values(), 'SupplyMedDBFluid_id', sDict["dbfluid"], specialValues=[0])

        self.substituteForeignKey('streams_in', 'id', sDict["streams_in"].values(), 'dbfluid_id', sDict["dbfluid"], specialValues=[0])
        self.substituteForeignKey('streams_out', 'id', sDict["streams_out"].values(), 'dbfluid_id', sDict["dbfluid"], specialValues=[0])

        self.substituteForeignKey('qheatexchanger', 'id', sDict["qheatexchanger"].values(), 'FluidIDSource', sDict["dbfluid"], specialValues=[0])
        self.substituteForeignKey('qheatexchanger', 'id', sDict["qheatexchanger"].values(), 'FluidIDSink', sDict["dbfluid"], specialValues=[0])

        self.substituteForeignKey('qdistributionhc', 'id', sDict["qdistributionhc"].values(), 'HeatDistMedium', sDict["dbfluid"], idAsString=True, specialValues=[0])
        self.substituteForeignKey('qwasteheatelequip', 'id', sDict["qwasteheatelequip"].values(), 'WHEEMedium', sDict["dbfluid"], idAsString=True, specialValues=[0])

        self.substituteForeignKey('qfuel', 'id', sDict["qfuel"].values(), 'DBFuel_id', sDict["dbfuel"])
        self.substituteForeignKey('qgenerationhc', 'id', sDict["qgenerationhc"].values(), 'DBFuel_id', sDict["dbfuel"])

        self.substituteForeignKey('qelectricity', 'id', sDict["qelectricity"].values(), 'dbelectricitymix_id', sDict["dbelectricitymix"])

        eqIds = sDict["qgenerationhc"].values()
        pipeDict = sDict["qdistributionhc"]
        self.substituteForeignKey('qgenerationhc', 'id', eqIds, 'HeatSourceHT', pipeDict)
        self.reconnectEquipesToPipes(sDict["qgenerationhc"], pipeDict)    # special function, as pipeId's in equipment currently
                                                                                # are stored as string

        self.substituteForeignKey('sproject', 'id', sDict["sproject"].values(), 'Auditor_ID', sDict["auditor"])
        self.substituteForeignKey('sproject', 'id', sDict["sproject"].values(), 'CreatedBy', sDict["auditor"])
        self.substituteForeignKey('sproject', 'id', sDict["sproject"].values(), 'ModifiedBy', sDict["auditor"])

        self.substituteForeignKey('process_streams_in', 'id', sDict["process_streams_in"].values(), 'streams_in_id', sDict["streams_in"])
        self.substituteForeignKey('process_streams_in', 'id', sDict["process_streams_in"].values(), 'qprocessdata_QProcessData_Id', sDict["qprocessdata"])

        self.substituteForeignKey('process_streams_out', 'id', sDict["process_streams_out"].values(), 'streams_out_id', sDict["streams_out"])
        self.substituteForeignKey('process_streams_out', 'id', sDict["process_streams_out"].values(), 'qprocessdata_QProcessData_Id', sDict["qprocessdata"])

        self.substituteForeignKey('qprocessdata', 'id', sDict["qprocessdata"].values(), 'QBuildings_ID', sDict["qbuildings"])
                
        hx_ps_ids = sDict["heatexchanger_pinchstream"].values()
        self.substituteForeignKey('heatexchanger_pinchstream', 'id',  hx_ps_ids, 'qheatexchanger_QHeatExchanger_Id', sDict["qheatexchanger"])                   
        self.substituteForeignKey('heatexchanger_pinchstream', 'id',  hx_ps_ids, 'outletOfHX_id', sDict["qheatexchanger"])                   
        self.substituteForeignKey('heatexchanger_pinchstream', 'id',  hx_ps_ids, 'inletOfHX_id' , sDict["qheatexchanger"])                   
        self.substituteForeignKey('heatexchanger_pinchstream', 'id',  hx_ps_ids, 'pinchstream_id' , sDict["pinchstream"])                   

        self.substituteForeignKey('qhxcontrol', 'id', sDict["qhxcontrol"].values(), 'HXId', sDict["qheatexchanger"])    

        # economic analysis data                    
        ecotables = ["tcacontingencies",
                     "tcadetailedopcost",
                     "tcadetailedrevenue",
                     "tcaenergy",
                     "tcainvestments",
                     "tcanonreoccuringcosts"]

        for table in ecotables:
            self.substituteForeignKey(table, 'id', sDict[table].values(), 'TcaID', sDict["tcageneraldata"])    

        dbio.commit()

    def getNodeIDs(self, PId, ANo = None):
        """ returns all nodes corresponding to the active alternative or project
        """

        if ANo is None:
            qflows = dbio.select("qflow", ProjectID=PId)
        else:
            qflows = dbio.select("qflow", ProjectID=PId, AlternativeProposalNo = ANo)
            
        productIds = [entry.id for entry in self.getProducts(PId=PId, ANo=ANo)]
        processIds = [entry.id for entry in self.getProcesses(PId=PId, ANo=ANo)]
        qrawMaterial = []   # TODO. add raw materials somehow somewhere ...
        if ANo is None:
            alternativeIds = [salt.id for salt in dbio.select("salternatives", ProjectID=PId)]
        else:
            AId = self.getAId(PId=PId, ANo=ANo)
            alternativeIds = [] if AId is None else [AId]

        nodeIds = [entry.id for entry in dbio.selectByList("qnode", "ObjectId", productIds) if entry.ObjectType=="PROD"] + \
                  [entry.id for entry in dbio.selectByList("qnode", "ObjectId", processIds) if entry.ObjectType=="PROC"] + \
                  [entry.id for entry in dbio.selectByList("qnode", "ObjectId", alternativeIds) if entry.ObjectType=="INT"]
            
        connectedNodeIds = list(set([qflow.FromId for qflow in qflows] + [qflow.ToId for qflow in qflows]))
                               
        nodeIds = list(set(nodeIds + connectedNodeIds))
        return nodeIds 
        
    def getAuditorID(self, all=False):
        """
        returns the ID of the responsible auditor for the present project
        """

        sproject = self.getProjectStatus()
        if sproject is None:
            logWarning(_U("Project (getAuditorID): Corrupt entry for project no. %s: table sproject not found") % Status.PId)
            if all:
                return (None, None, None)
            else:
                return None

        if all:
            return (sproject.Auditor_ID, sproject.CreatedBy, sproject.ModifiedBy)
        else:
            return sproject.Auditor_ID

    def getAuditorName(self, auditorId):
        """
        id: looks for the data of the auditor with id
        """

        auditor = dbio.get("auditor", id=auditorId)
        if auditor is not None:
            return auditor.Name
        else:
            logTrack(_("Project (getAuditorName): Cannot find auditor with ID = %s") % auditorId)
            return "---"

    def commit(self):
        """ bridge function for avoiding direct calls to dbio.commit from other modules
        """
        
        dbio.commit()

############# functions for conversion of pipe ID's in equipment connection into string, list etc.

    def getPipeNames(self, IDString):
        if IDString is not None:
            pipeIDsSQL = IDString.split(';')
        else:
            pipeIDsSQL = []

        pipeDict = Status.prj.getPipeDict()
            
        pipeIds = []
        pipeNames = []
        for i in pipeIDsSQL:
            try:
                pipeID = int(i)
            except:
                logDebug("Project (getPipeNames): erroneous pipeID-string in SQL: [%s|%s]"%\
                             (i,pipeIDsSQL))
                pipeID = None

            if pipeID in pipeDict.keys():
                pipeNames.append(pipeDict[pipeID])
                pipeIds.append(pipeID)

        return pipeNames, pipeIds
            
    def getPipeIDString(self, nameList, idList=None):
#        print "PanelQ4 (getPipeIDString): Getting pipeIDs from :",nameList

        if idList is None:  # get IDs from names
            pipeDict = Status.prj.getPipeDict()
            pipeIDs = []
            for name in nameList:
                pipeID = findKey(pipeDict, name)
                pipeIDs.append("%d"%pipeID)

        else:
            pipeIDs = ["%d"%pipeID for pipeID in idList]

            
        IDString = ";".join(pipeIDs)
        return IDString
            
    def getProjectUnits(self, string):
        """
        helper function. splits the string in sproject.UnitsReport into
        the units and the currency string
        """
        if string is not None:
            units = "SI-kWh"
            currency = "EUR"
            try:
                info=string.split("|")
                units = info[0]
                if len(info)>=2:
                    currency = info[1]
            except:
                pass
            Status.ProjectCurrency = currency
            Globals.unitConversion.setCurrency(currency)

            return units, currency
        else:
            return None, None        

