# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	DETAILEDSCHEDULES
#			
#------------------------------------------------------------------------------
#			
#	Functions for management of time schedules
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
"""
:mod:`einstein.modules.schedules` -- Detailed schedules of periodic operation
=============================================================================

   :synopsis: Periodic yearly load profiles of process operation
   :author:   Hans Schweiger <hans.schweiger@energyxperts.net>
   :author:   David Baehrens <david.baehrens@energyxperts.net>
"""
from profiles import WeeklyProfile
from datetime import date, timedelta, time
from math import fabs, fmod, floor, ceil
import numpy as np
from random import Random
import copy
from auxiliary.auxiliary import checkLimits, findFirstGE, normalize, noneCheck
from auxiliary.auxiliary import toGettextAndUnicode as _U
from GUI.status import Status
from modules.interfaces import createQ_t
from constants import *
from modules.messageLogger import *
from GUI.dialogGauge import DialogGauge
from lib.dateTime.dateTimeTools import timeToHour, REFERENCEDATE, FIRSTOFYEAR, LASTOFYEAR, dateToString, MIDNIGHT, MIDNIGHT_AS_ZERO
from scheduleTools import BasicSchedule
import database.dbio as dbio

### Constants
periodReference = REFERENCEDATE # year 2007 starts and ends on Monday
DEFAULTSCHEDULES = ["operation", "batchCharge", "batchDischarge"]
DEFAULTCHARGETIME = 0.2 #20% of batch duration
TOLERANCE_GAP = 0.5
"""Minimum time between consecutive cycles after random tolerance shift (hours)"""
def getProcessPeriods(processId):
            #lookup of an entry in process_periods table exists for current processId
    return dbio.select("process_periods", qprocessdata_QProcessData_ID=processId)

### Errors
class ScheduleNoProfilesError(Exception):
    """No weekly profiles selected for period schedule."""
    pass

class ScheduleNotFoundError(Exception):
    """No detailed schedule for process found in DB."""
    pass

class ProcessDataNotFoundError(Exception):
    """No data on process found in DB (table qprocessdata)."""
    pass

class InconsistentDataBaseError(Exception):
    """BUG: Data base structure is inconsistent."""
    pass

def dateToFirstYearHourOfDay(d):
    return (date(periodReference.year, d.month, d.day) - periodReference).days * int(DAY)

def dateToLastYearHourOfDay(d):
    return (((date(periodReference.year, d.month , d.day) - periodReference).days + 1) * int(DAY) - 1)


######################################################################################################
#   CLASS PERIODSCHEDULE
######################################################################################################


class PeriodSchedule(object):
    """Detailed schedules of process load by hour of year with periodic differences.
    
    :param name: Descriptive name of the schedule
    :type  name: String
    :param startup: Duration of process start up phase in hours
    :type  startup: Float
    :param inflow:  Duration of process charge time in hours
    :type  inflow:  Float
    :param outflow: Duration of process discharge time in hours
    :type  outflow: Float
    :param tolerance: Duration of tolerated back-/forward-shift of schedule times in hours
    :type  tolerance: Float
    :param holidayScale: Global load scale factor [0,1] of process during industry holidays
    :type  holidayScale: Integer
    :param holidays: Starting and stopping hours of holiday periods (hour of year [0,8759], inclusive)
    :type holidays:  List of Integer 2-Tuple
    :param schedule: Load [0,1] during each hour of year [0,8759]
    :type schedule:  List of Float

    :returns: Period schedule instance
    :rtype: einstein.modules.schedules.PeriodSchedule

    The schedule is fully represented in the following array:
    - self.periodIntervals:     list of PeriodProfile objects

    From this the function build generates the derived arrays:
    
    - self.intervals:           list of intervals (startHour, stopHour, scale) for each of the days
                                intervals is generated by "addPeriodIntervals"
    - self.scaleIntervals:      = self.intervals x holidays = THE FINAL EXACT SCHEDULE
                                [function getYearlyIntervals]
    - self.cycleAverageIntervals: obtained by fusing adjacent intervals
    - self.schedule:            fav-Vector: hourly average scale factor for 8760 hours of the year
                                [function getFavFromIntervals]


    Holiday treatment:

    Attributes:
    - self.holidays / isHolidays interval of holiday periods in hours
    - self.holidayIntervals / isHoliday     same as before, but in days
    - self.holidayScale

    """

    def __init__(self, name, startup=None, inflow=None, outflow=None, tolerance=None, holidayScale=None, holidays=None, schedule=None):
        self.name = name
        self.startup = startup
        self.inflow = inflow
        self.outflow = outflow
        if tolerance is not None:
            self.tolerance = tolerance
        else:
            self.tolerance = 0.
        if holidayScale is not None:
            self.holidayScale = holidayScale
        else:
            self.holidayScale = 0
        if holidays is not None:
            self.holidays = holidays
        else:
            self.clearHolidays()
        if schedule is not None:
            self.schedule = schedule
        else:
            self.clearSchedule()
        self.clearIntervals()
##        self.updateToleranceOffset()
        self.holidayIntervals = []
        self.periodIntervals = []

        self.continuous = True
        self.HPerYearInflow = None
        self.HPerYearOutflow = None

        self.deletedIntervals = []

    def __repr__(self):
        info = "PSchedule %s: "%self.name
        for period in self.periodIntervals:
            info += "\n\t[%s]"%period
            
        return info
        
    def calculateGlobals(self):
        """ calculates the global parameters of the schedule and stores them"""
        
        self.scaleIntervals = self.getYearlyIntervals() # is recurrently required for calculation of batchCharge/batchDischarge - profiles
        self.cycleAverageIntervals = fuseIntervals(self.scaleIntervals)
        NCyclesPerYear = len(self.cycleAverageIntervals)
        
        self.HPerYearNom = self.getOperationHoursPerYear()

##        self.yearlyIntervals = self.getYearlyProfileOnOffMeanIntervals(withHolidays=True, withTolerance=True)

        #discretised profile: fav-vector 
        operationProfile = self.getYearlyProfile()
        self.HPerYear = np.sum(operationProfile)
        self.operationProfile = normalize(operationProfile)

        self.NDays = self.getOperationDays()            

        if self.NDays > 0:
            self.HPerDay = self.HPerYear / self.NDays
            self.NBatch = max(float(NCyclesPerYear) / self.NDays, 1.0)
        else:
            self.HPerDay = 0
            self.NBatch = 1
       
        if self.NDays > 0 and self.NBatch > 0:
            self.HBatch = self.HPerYearNom / (self.NDays * self.NBatch)
        else:
            self.HBatch = 0
            
        if self.HPerYearNom > 0:
            self.PartLoad = self.HPerYear / self.HPerYearNom
        else:
            self.PartLoad = 0

        defaultChargeTime = DEFAULTCHARGETIME * self.HBatch
        self.startupTime = noneCheck(self.startup, defaultChargeTime)
        self.inflowTime = noneCheck(self.inflow, defaultChargeTime)
        self.outflowTime = noneCheck(self.outflow, defaultChargeTime)

        if self.continuous:
            self.HPerYearInflow = self.HPerYear
            self.HPerYearOutflow = self.HPerYear
            self.HPerYearInflowNom = self.HPerYearNom
            self.HPerYearOutflowNom = self.HPerYearNom
        else:
            self.inflowProfile, self.HPerYearInflow, self.HPerYearInflowNom = self.getYearlyBatchChargeProfile(self.inflowTime)
            self.outflowProfile, self.HPerYearOutflow, self.HPerYearOutflowNom = self.getYearlyBatchOutflowProfile()
        
#--------------------------------------------------------------------------------------------------------
#   functions for managing user input (build schedule from periods and weekly/dayly profiles)
#--------------------------------------------------------------------------------------------------------

    def build(self):
        """ Builds the schedule from self.periodIntervals and self.holidayScale which have to be defined previously
        """

        err = None

        self.clearIntervals()

        periods = sorted(self.periodIntervals, key=lambda period: period.start)
        lastStop = REFERENCEDATE - timedelta(1)
        for period in periods:
            if self.checkIfValidPeriod(period, lastStop):
                lastStop = period.stop
                for profile in period.profiles:
                    item = (period.start, period.stop, period.step, period.scale, profile)
                    self.addPeriodIntervals(*item)
            else:
                err = _("Error in period -> start: %s stop: %s scale %s")%(dateToString(period.start), dateToString(period.stop), period.scale)
    
        self.scaleIntervals = self.getYearlyIntervals() # is recurrently required for calculation of batchCharge/batchDischarge - profiles

    # get discrete operation schedule
        self.schedule = self.getFavFromIntervals(self.scaleIntervals)

        return err

    def checkIfValidPeriod(self, period, lastStop):
        """ checks the validity of a period with respect to the previous one
        """
        if period.start > lastStop and \
           period.stop >= period.start and \
           period.stop <= date(2007, 12, 31):
            return True
        else:
            return False

    def check(self):
        """ checks the validity of all periods in the schedule builder
        """
        err = None
        periods = sorted(self.periodIntervals, key=lambda period: period.start)
        lastStop = REFERENCEDATE - timedelta(1)
        for period in periods:
            if not self.checkIfValidPeriod(period, lastStop):
                err = _("Error in period -> start: %s stop: %s scale %s")%(dateToString(period.start), dateToString(period.stop), period.scale)
                break
            lastStop = period.stop
        return err

    def addPeriodProfile(self, start, stop, step, scale, profile):
        """Set an operation time profile in a specific period of the year.
        
        Repeated calls are cumulative. Holiday periods are handled separately by :meth:`einstein.modules.schedules.addHolidays`.
        Also adds the intervals thourgh :meth:`einstein.modules.schedules.addPeriodIntervals`.
        
        :param start: starting date of period (year does not matter, inclusive)
        :type  start: date
        :param  stop: stopping date of period (year does not matter, inclusive)
        :type   stop: date
        :param  step: daily stepping through the period (1=every day, 2=every second day, etc.)
        :type   step: Integer
        :param scale: Global load scaling [0,inf[ during the period
        :type  scale: Float
        :param profile: A weekly profile on selected weekdays during the period.
        :type  profile: einstein.modules.profiles.WeeklyProfile
        
        :returns: nothing
        """

        logTrack("Schedules.PeriodSchedule.addPeriodProfile start %s stop %s step %s scale %s profile %s"%(start, stop, step, scale, profile))
        self.addPeriodIntervals(start, stop, step, scale, profile)
        
    # Block of functions operating with "PeriodIntervals": -> start/stop/scale with exact start and stop time
    def addPeriodIntervals(self, start, stop, step, scale, profile):
        """Set an operation interval profile in a specific period of the year.
        
        Repeated calls are cumulative. Holiday periods are handled separately by :meth:`einstein.modules.schedules.addHolidayIntervals`.
        
        :param start: starting date of period (year does not matter, inclusive)
        :type  start: date
        :param  stop: stopping date of period (year does not matter, inclusive)
        :type   stop: date
        :param  step: daily stepping through the period (1=every day, 2=every second day, etc.)
        :type   step: Integer
        :param scale: Global load scaling [0,inf[ during the period
        :type  scale: Float
        :param profile: A weekly profile on selected weekdays during the period.
        :type  profile: einstein.modules.profiles.WeeklyProfile
        
        :returns: nothing
        """
        startDay = (date(periodReference.year, start.month, start.day) - periodReference).days
        stopDay  = (date(periodReference.year,  stop.month,  stop.day) - periodReference).days
        weekdays = profile.weekday
        for day in range(startDay, stopDay + 1, step):
            if not weekdays[day % len(weekdays)]:
                continue
            for (intervalStart, intervalStop, intervalScale) in profile.intervals:
                self.intervals[day].append((intervalStart, intervalStop, (scale * .01) * (intervalScale * .01)))
                
    def getWeeklyProfile(self, start):
        """Retrieve a weekly load profile without industry holidays starting on Monday not before a given start date.
        
        :param start: Date for the weekly profile to start not before (year does not matter)   
        :type  start: date
        
        :returns: Load [0,1] for week hour [0,167]
        :rtype:   List of Float
        """
        start = date(periodReference.year, start.month, start.day)
        if start.weekday() != 0: # start weekly profile on Monday
            start = start + timedelta(days=abs(start.weekday() - 7))
        startYearHour = dateToFirstYearHourOfDay(start)
        return self.schedule[startYearHour:startYearHour + int(WEEK)].copy()

    def clearHolidays(self):
        """Remove all holiday periods.
        
        :returns: nothing
        """
        self.holidays = []
        
    def addHolidays(self, start, stop):
        """Define a holiday period by two dates.
        
        Also adds the daily interval through :meth:`einstein.modules.schedules.addHolidayInterval`.
        
        :param start: starting date (year does not matter, inclusive)
        :type  start: date
        :param  stop: stopping date (year does not matter, inclusive)
        :type   stop: date
        
        :returns: nothing
        """
        holidayStartHour = dateToFirstYearHourOfDay(start)
        holidayStopHour = dateToLastYearHourOfDay(stop)
        self.holidays.append((holidayStartHour, holidayStopHour))
        self.addHolidayInterval(start, stop)
        
    def addHolidayInterval(self, start, stop):
        """Define a holiday daily interval by two dates.
        
        :param start: starting date (year does not matter, inclusive)
        :type  start: date
        :param  stop: stopping date (year does not matter, inclusive)
        :type   stop: date
        
        :returns: nothing
        """
        startDay = (date(periodReference.year, start.month, start.day) - periodReference).days
        stopDay  = (date(periodReference.year,  stop.month,  stop.day) - periodReference).days
        self.holidayIntervals.append((startDay, stopDay))
        
    def isHolidays(self, hour):
        """Determine whether a hour of year [0,8759] is during a holiday period.
        
        :param hour: 0-based hour of year [0,8759]
        :type  hour: Integer
        
        :returns: Is hour during holidays?
        :rtype:   Boolean
        """
        for (start, stop) in self.holidays:
            if start <= hour and hour <= stop:
                return True
        return False
    
    def isHoliday(self, day):
        """Determine whether a day of year [0,364] is during a holiday period.
        
        :param hour: 0-based day of year [0,364]
        :type  hour: Integer
        
        :returns: Is day during holidays?
        :rtype:   Boolean
        """
        for (start, stop) in self.holidayIntervals:
            if start <= day and day <= stop:
                return True
        return False


        ##Block of functions operating with fav-vector: -> start/stop/scale with exact start and stop time

    def updateToleranceOffset(self, toleranceOffset=None):
        """Draw a random tolerance offset for each cycle during the year.
        
        :param toleranceOffset: Offsets for each cycle, default: draw random offsets in interval +/- PeriodSchedule.tolerance 
        :type  toleranceOffset: List of float
        
        :returns: nothing
        """
        if toleranceOffset is not None:
            self.toleranceOffset = toleranceOffset
        else:
            self.toleranceOffset = []
##            random = Random()
##            random.seed()
##            for cycle in range(self.getNumberOfBatchesPerYear(withHolidays=False, withTolerance=False)):
##                self.toleranceOffset.append(random.uniform(-self.tolerance, self.tolerance))
    
#--------------------------------------------------------------------------------------------------------
#   functions managing exact schedule (intervals)
#--------------------------------------------------------------------------------------------------------

    def getYearlyIntervals(self):
        """Retrieve the yearly interval profile optionally with industry holidays.
        
        :param withHolidays: Include the scaling during the industry holiday periods?
        :type  withHolidays: Boolean
        :param fused: Combine overlapping and gapless intervalls into one interval?
        :type fused: Boolean
        
        :returns: List of 3-tuples (start, stop, scale) for every day of year [0,364]
        :rtype:   List of List of (date, data, float)
        
        """
        intervals = [[] for dummy in xrange(NDAYS)]
        for day in range(NDAYS):
            for (start, stop, scale) in self.intervals[day]:
                if self.isHoliday(day):
                    if self.holidayScale > 0: #avoid adding and counting 0 scale intervals.
                        intervals[day].append((start, stop, scale * self.holidayScale))
                else:
                    intervals[day].append((start, stop, scale))
        result = intervals
                
        return result
                
    def getOperationHoursPerYear(self):
        """Calulcate the mean nominal duration hours of a single interval 
        optionally with holidays.
        
        :returns: Mean duration hours of one batch
        :rtype:   Float
        """
        intervals = self.scaleIntervals
        yearlyDuration = 0.
        for day in range(NDAYS):
            for (start, stop, scale) in intervals[day]:
                stopdelta = timedelta(hours=stop.hour, minutes=stop.minute)
                # Check for stoptime 0:00 and treat it as 24:00 == end of day
##                if stopdelta == timedelta(0):
                if stop==MIDNIGHT or stop==MIDNIGHT_AS_ZERO:
                    stopdelta = timedelta(1)
                duration = stopdelta - timedelta(hours=start.hour, minutes=start.minute)
                if duration.days == 1:
                    duration = 24
                else:
                    duration = duration.seconds / 3600.
                yearlyDuration += duration
        return yearlyDuration        

    def getOperationDays(self):
        """Calculate the number of days non-zero intervals optionally with industry holidays.
        
        :param withHolidays: Ingnore days scaled to zero during the industry holiday periods?
        :type  withHolidays: Boolean
        
        :returns: The number of operation days during the year
        :rtype:   Integer
        """
        intervals = self.scaleIntervals
        nDays = 0
        for day in range(NDAYS):
            fractiondays = [scale for (start, stop, scale) in intervals[day] if (stop != start) or (stop == start == time(0, 0))]
            if len(intervals[day]) > 0 and sum(fractiondays) > 0.0:
                nDays += 1
        return nDays

#--------------------------------------------------------------------------------------------------------
#   functions for building discrete schedule (fav-vector)
#--------------------------------------------------------------------------------------------------------

    def clearSchedule(self):
        """Define an all-zero yearly profile.
        
        :returns: nothing
        """
        self.schedule  = createQ_t()
        
    def clearIntervals(self):
        """Define empty yearly intervals.
        
        :returns: nothing
        """
        self.intervals = [[] for dummy in xrange(NDAYS)]
    

    def getFavFromIntervals(self, intervals):
        """Converts the exact profile (in intervals) into hourly averages (fav-vector)
        
        :param withHolidays: Include the scaling during the industry holiday periods?
        :type  withHolidays: Boolean
        
        :returns: Load [0,1] for year hour [0,8759]
        :rtype:   Numpy n-d array of Float 
        
        """
        fav = createQ_t()

        yearlyIntervals = []
        
        for day, dailyIntervals in enumerate(intervals):
            for (startDate,stopDate,scale) in dailyIntervals:
                start = timedelta(hours=startDate.hour, minutes=startDate.minute).seconds / 3600.0 + day*24.0
                if stopDate==MIDNIGHT or stopDate==MIDNIGHT_AS_ZERO:
                    stop = (day+1)*24.0
                else:
                    stop = timedelta(hours=stopDate.hour, minutes=stopDate.minute).seconds / 3600.0 + day*24

                yearlyIntervals.append((start, stop, scale))
                    
                istart = int(start/Status.TimeStep)
                istop = int(stop/Status.TimeStep)
                if istart==istop: # special case - start time and stop time within same hour
                    fav[istart] += scale*(stop-start)
                else:
                    fav[istart] += scale * (Status.TimeStep*(istart+1) - start)
                    for i in range(istart+1,istop):
                        fav[i] += scale*Status.TimeStep

                    if istop < Status.Nt:
                        fav[istop] += scale * (stop - Status.TimeStep*istop)

##        fav /= sum(fav) # normalisation to one

        self.bs = BasicSchedule()
        self.bs.getPeriods(yearlyIntervals, tmax = 8760.)
        return fav
                    
    def getYearlyProfile(self):
        """Retrieve the yearly load profile optionally with industry holidays.
        
        :param withHolidays: Include the scaling during the industry holiday periods?
        :type  withHolidays: Boolean
        
        :returns: Load [0,1] for year hour [0,8759]
        :rtype:   Numpy n-d array of Float 
        
        """
        schedule = self.schedule.copy()

        for hour, scale in enumerate(schedule):
            if self.isHolidays(hour):
                schedule[hour] = scale * self.holidayScale


        return schedule
    
#--------------------------------------------------------------------------------------------------------
#   functions for derived schedules (batch inflow/outflow, startup)
#--------------------------------------------------------------------------------------------------------

    def getYearlyBatchChargeProfile(self, chargeTime):
        """Retrieve the yearly charge load profile optionally with industry holidays.
        
        :param withHolidays: Include the scaling during the industry holiday periods?
        :type  withHolidays: Boolean
        
        :returns: Load [0,1] for year hour [0,8759]
        :rtype:   List of Float
        """
        yearlyProfile = createQ_t()
        intervals = self.cycleAverageIntervals

        hPerYearNom = 0
        hPerYear = 0

        for (start, stop, scale) in intervals:
            # get precise start and stop time
            dt = chargeTime
            if chargeTime > (stop-start) and (stop-start) > 0:
                scale *= chargeTime/(stop-start)
                dt = (stop-start)
            hPerYear += dt*scale
            hPerYearNom += dt

            addIntervalToDiscreteSchedule((start, start+dt, scale), yearlyProfile)
                     
        return normalize(yearlyProfile), hPerYear, hPerYearNom
    
    
    def getYearlyBatchOutflowProfile(self):
        """Retrieve the yearly discharge load profile optionally with industry holidays.
        
        :param withHolidays: Include the scaling during the industry holiday periods?
        :type  withHolidays: Boolean
        
        :returns: Load [0,1] for year hour [0,8759]
        :rtype:   List of Float
        """
        yearlyProfile = createQ_t()
        intervals = self.cycleAverageIntervals
        outflow = self.outflowTime

        hPerYearNom = 0
        hPerYear = 0

        for n, interval in enumerate(intervals):
            nextInterval = findNextInterval(n, intervals)
            (start, stop, scale) = interval

            if stop >= YEAR:
                stop -= YEAR
            
            (nextStart, nextStop, dummy) = nextInterval
            if nextStart < stop: nextStart += YEAR

            dt = self.outflowTime
            if self.outflowTime > (nextStart - stop) and (nextStart-stop) > 0:
                scale *= self.outflowTime/(nextStart-stop)
                dt = nextStart-stop
            hPerYear += dt*scale
            hPerYearNom += dt

            
            addIntervalToDiscreteSchedule((stop, stop+dt, scale), yearlyProfile)

        return normalize(yearlyProfile), hPerYear, hPerYearNom
    
##def intervalsOnOffUnit(profile):
##        """Retrieve on/off intervals from load profile.
##        
##        Intervals start and stop when going through zero. The load is the unit load of 1.0.
##        
##        :returns: Uniform running intervals (load > 0) of inclusive starting hour [0,|profile|-1], inclusive stopping hour [0,|profile|-1], and load 1.0
##        :rtype:   List of 3-Tuples
##        """
##        profileOnOffIntervals = []
##        off = True
##        start = None
##        for (hour, scale) in enumerate(profile):
##            if off and (scale > 0.):
##                start = hour
##                scaleSum = scale
##                off = False
##            elif not(off) and (scale == 0.):
##                profileOnOffIntervals.append((start, hour - 1, 1.0))
##                off = True
##                start = None
##            elif not(off) and scale > 0.:
##                scaleSum += scale
##        if (start != None) and not(off):
##            profileOnOffIntervals.append((start, len(profile) - 1, 1.0))
##        return profileOnOffIntervals
                 
#--------------------------------------------------------------------------------------------------------
#   functions for I/O to database
#--------------------------------------------------------------------------------------------------------

    def loadFromDB(self, processId, withDetails=True):
        """Retrieve yearly profile, process schedule parameters, tolerance offsets, and industry holidays from data base for a given process.
        
        To load a detailed schedule for a process from the DB, construct an empty period schedule instance first.
        
        :param withDetails: If True load the real activity intervals from the
            DB. If you want to rebuild a PeriodSchedule from scratch, this may 
            be unwanted.
        
        :returns: nothing
        :exception einstein.modules.schedules.ProcessDataNotFoundError: No data on process found in DB (table qprocessdata).
        :exception einstein.modules.schedules.InconsistentDataBaseError: Data base structure is inconsistent. This would be a bug.
        :exception einstein.modules.schedules.ScheduleNotFoundError: No detailed schedule for process found in DB.
        """

        #lookup of an entry in process_periods table exists for current processId
        process_periods = getProcessPeriods(processId)
        if len(process_periods) <= 0:
            raise ScheduleNotFoundError

        self.dbId = processId

        if processId >= 0:          # < 0 -> equipment control schedule
            # set process schedule parameters
            processDataRow = dbio.get("qprocessdata", id=processId)
            if processDataRow is None:
                raise ProcessDataNotFoundError


            self.continuous = (processDataRow.ProcType=="continuous")
            
            self.startup = processDataRow.StartUpDuration
            self.inflow = processDataRow.InFlowDuration
            self.outflow = processDataRow.OutFlowDuration
            if processDataRow.ScheduleTolerance != None:
                self.tolerance = processDataRow.ScheduleTolerance
            if processDataRow.HolidayScale != None:
                self.holidayScale = float(processDataRow.HolidayScale) / 100

        else:
            equipmentId = -processId
            equipmentDataRow = dbio.get("qgenerationhc", id=equipmentId)
            if equipmentDataRow is None:
                raise ProcessDataNotFoundError
            q = Status.prj.getQuestionnaire(equipmentDataRow.Questionnaire_id)

        self.loadHolidays()
            
            
        # set intervals and build schedule details
        self.clearIntervals()
        if withDetails is True:
            self.periodIntervals = getIntervalsFromDB(processId)
            self.build()

        if processId >= 0:
            processDataRow = dbio.get("qprocessdata", id=processId)
            if processDataRow is None:
                raise ProcessDataNotFoundError
            self.build()
            self.calculateGlobals()
            processDataRow.set(**{"StartUpDuration"   : self.startup,
                                   "InFlowDuration"    : self.inflow,
                                   "OutFlowDuration"   : self.outflow,
                                   "ScheduleTolerance" : self.tolerance,
                                   "HolidayScale"      : int(self.holidayScale*100),
                                   "HPerDayProc"       : float(self.HPerDay),
                                   "HBatch"            : float(self.HBatch),
                                   "NBatch"            : self.NBatch,
                                   "NDaysProc"         : self.NDays,
                                   "PartLoad"          : float(self.PartLoad),
                                   "HPerYearInFlow"     : float(self.HPerYearInflow),
                                   "HPerYearOutFlow"    : float(self.HPerYearOutflow)})
            
        #TODO: is this really necessary here to SAVE the schedule within the LOAD-function ???
##    def getScheduleFromIntervals(self, withHolidays=False):
##        """ transforms the schedule intervals into the fav (hourly averages) vector """
##
##        self.schedule = self.getFavFromIntervals(withHolidays=withHolidays)

    def loadHolidays(self):
        """ loads the general holidays into the schedule """
        generalHolidays = Status.prj.getHolidays(datetime=True)
        for holidayPeriod in generalHolidays:
            holidayStart, holidayStop = holidayPeriod
            if holidayStart and holidayStop:
                self.addHolidays(holidayStart, holidayStop)


        
    def saveToDB(self, processId, withHolidays=True, realProcessId=None, saveAll=False):
        """Store yearly load profile intervals and process schedule 
        parameters (startup, inflow, outflow, tolerance, and holiday 
        scale) to data base for a given process.
        
        Also stores process effective operation hour parameters and 
        tolerance offsets. 
        
        Does not store the holidays, since they are global to the industry.
        
        :param processId: Data base ID of process to store the schedule for
        :type  processId: Integer
        :param withHolidays: Whether holiday periods are honored, only 
            for operation hour parameters. Intervals are always stored ignoring 
            holidays. 
        :type  withHolidays: Boolean
        :param realProcessId: If the process in question is not the one in the
            alternative proposal -1, then this additional id is needed 
            to update the per process parameters like for example PartLoad.
        :type realProcessId: Integer
        :param saveAll: If True, saves everything. If False, saves only the schedule
            attributes, but not the periods
        :type saveAll: Boolean
        
        :returns: nothing
        :exception einstein.modules.schedules.ProcessDataNotFoundError: 
            No data on process found in DB (table qprocessdata).
        
        """
        # store process schedule parameters
        self.build()
        self.calculateGlobals()

        realProcessId = processId if realProcessId is None else realProcessId
        if realProcessId >= 0:
            processDataRow = dbio.get("qprocessdata", id=realProcessId)
            if processDataRow is None:
                raise ProcessDataNotFoundError



            processDataRow.set(**{"StartUpDuration"   : self.startup,
                                   "InFlowDuration"    : self.inflow,
                                   "OutFlowDuration"   : self.outflow,
                                   "ScheduleTolerance" : self.tolerance,
                                   "HolidayScale"      : int(self.holidayScale*100),
                                   "HPerDayProc"       : float(self.HPerDay),
                                   "HBatch"            : float(self.HBatch),
                                   "NBatch"            : self.NBatch,
                                   "NDaysProc"         : self.NDays,
                                   "PartLoad"          : float(self.PartLoad),
                                   "HPerYearInFlow"     : float(self.HPerYearInflow),
                                   "HPerYearOutFlow"    : float(self.HPerYearOutflow)})

        if saveAll:
            err = self.build()
            if err:
                logError("Cannot save schedule %s. %s"%(self.name, err))
                return
                        
            for period in self.periodIntervals:
                period.saveToDB(realProcessId, commit=False) # commit = False: time saver -> commit is called at the end
                
            for period in self.deletedIntervals:
                period.deleteFromDB()
                        
            self.deletedIntervals = []
                
        dbio.commit()
        
        logTrack("Schedules.saveToDB: OperationHoursPerYear = %s"%self.HPerYear)


    def deletePeriodInterval(self, index):
        """ deletes an interval from self.periodIntervals and adds it to the listing
            self.deletedIntervals, in order to be deleted also in the DB with the next
            call to saveToDB
        """
        self.deletedIntervals.append(self.periodIntervals.pop(index))

    def deleteFromDB(self):
        for period in self.periodIntervals:
            period.deleteFromDB()
        self.periodIntervals = []
        dbio.commit()

#--------------------------------------------------------------------------------------------------------
#   auxiliary functions
#--------------------------------------------------------------------------------------------------------

def fuseIntervals(intervals):
    """ combines intervals of self.scaleIntervals into cycles
        whenever intervals are adjacent
        in this case, an average scale factor is calculated.
    """
    conFusion = []
    lastStop = -1.
    ptr = -1
    for day, dailyIntervals in enumerate(intervals):
        for n, interval in enumerate(dailyIntervals):
            (start, stop, scale) = interval
            tstart = day*24 + timeToHour(start)
            tstop = day*24 + timeToHour(stop)
            if stop==MIDNIGHT or stop==MIDNIGHT_AS_ZERO:
                tstop = (day+1)*24.0
                
            if tstart > lastStop + EPSILON:  # start a new interval
                conFusion.append((tstart, tstop, scale))
                ptr = len(conFusion) - 1
                
            else:   # add to an existing interval
                previous = conFusion[ptr]
                dt_previous = previous[1] - previous[0]
                scale_previous = previous[2]
                newStart = previous[0]
                newStop = tstop
                dt = (tstop-tstart)
                if dt+dt_previous > 0:
                    newScale = (scale*dt + scale_previous*dt_previous) / (dt + dt_previous)
                else:
                    newScale = 0
                conFusion[ptr] = (newStart, newStop, newScale)

            lastStop = tstop
                
    #finally check if first and last interval of the year are attached
    if len(conFusion) > 1:
        if (conFusion[0][0] == 0) and (conFusion[-1][1] == YEAR):
            newStart = conFusion[-1][0]
            newStop = conFusion[0][1] + YEAR
            dt1 = conFusion[0][1] - conFusion[0][0]
            dt2 = conFusion[-1][1] - conFusion[-1][0]
            s1 = conFusion[0][2]
            s2 = conFusion[-1][2]
            if dt1+dt2 > 0:
                newScale = (s1*dt1 + s2*dt2) / (dt1 + dt2)
            else:
                newScale = 0
                                         
            conFusion[-1] = (newStart, newStop, newScale)
            conFusion.pop(0)

    return conFusion
                    
def weeklyMeanOfYearlyProfile(yearlyProfile):
    """Retrieve a weekly load profile with mean load during the year.
        
    :returns: Load [0,1] for week hour [0,167]
    :rtype:   List of Float
    """
    weeklyProfile = np.zeros((int(WEEK)))
    for weekHour in xrange(int(WEEK)):
        yearlyWeekHours = yearlyProfile[weekHour:int(YEAR):int(WEEK)].copy()
        weeklyProfile[weekHour] = yearlyWeekHours.sum() / len(yearlyWeekHours)
    return weeklyProfile
        
          
def getIntervalsFromDB(processId):
    """Return a list of the scheduled process activity
      
    :param processId: the database id of the process
    
    :returns list: a list of  5-tuples (start, stop, step, scale, profile)
        where start and stop are date objects, step an Integer, scale a Float
        and profile a einstein.modules.profiles.WeeklyProfile.
        
    """
    pps = dbio.select("process_periods", qprocessdata_QProcessData_ID = processId)
    periods = []
    for pp in pps:
        period = ProcessPeriod(pid=pp.periods_id, ppid=pp.id, step=pp.step, scale=pp.scale)
        periods.append(period)

    periods = sorted(periods, key=lambda period: period.start)
    return periods


def addIntervalToDiscreteSchedule(interval, schedule):
    """ adds an interval (start, stop scale) to
        a discrete time schedule (fav-vector)
    """
    (start, stop, scale) = interval
    iFirst = int(start)
    if stop >= YEAR:
        n = stop%YEAR
        stop -= n*YEAR
    iLast = int(stop)

    if iFirst == iLast and stop >= start:
        schedule[iFirst] += (stop - start)*scale
    else:
        firstFrac = (iFirst+1) - start
        lastFrac = stop - iLast
        schedule[iFirst] += firstFrac*scale
        schedule[iLast] += lastFrac*scale
        if iLast > iFirst:
            schedule[iFirst+1:iLast] += scale
        else:
            schedule[iFirst+1:] += scale
            schedule[0:iLast] += scale    

def findNextInterval(n, intervals):
    "auxiliary function looking for the next interval"
    nyear = len(intervals)
    if n+1 < nyear:
        return intervals[n+1]
    else:
        return intervals[0]

######################################################################################################
######################################################################################################

class ProcessPeriod():
    """Class combining the information contained in tables periods and process_periods
    """
    
    def __init__(self, pid=None, ppid=None, processId=None, start=FIRSTOFYEAR, stop=LASTOFYEAR, step=True, scale=100.0):
        self.processId = processId #associated process
        
        self.profiles = []
        self.step = step
        self.scale = scale
        self.ppid = ppid

        self.start = start
        self.stop = stop
        self.pid = pid

        if ppid is not None:
            self.loadFromDB()

    def __repr__(self):
        info = "PP(ppid %s) procId %s pid %s [%s:%s] f=%s"%(self.ppid, self.processId, self.pid, self.start, self.stop, self.scale)
        for profile in self.profiles:
            info += "\n\t[%s]"%profile

        return info
    
    def loadFromDB(self):
        pp = dbio.get("process_periods", id = self.ppid)
        if pp is not None:
            self.step = pp.step
            self.scale = pp.scale
            self.processId = pp.qprocessdata_QProcessData_ID
            self.pid = pp.periods_id
        
            # read in start and stop from periods(pid)
            periodTable = dbio.get("periods", id=self.pid)
            if periodTable is not None:
                self.start = periodTable.start
                self.stop = periodTable.stop

            # read in profiles-ids from link table process_period_profiles and load profiles
            ppps = dbio.select("process_period_profiles", process_periods_id = self.ppid)
            for ppp in ppps:
                profileId = ppp.profiles_id
                if profileId not in Status.profilesInUse.keys():
                    profile = WeeklyProfile("-")
                    profile.loadFromDB(id=ppp.profiles_id)
                    Status.profilesInUse.update({profileId:profile})
                else:
                    profile = Status.profilesInUse[profileId]
                self.profiles.append(profile)

    def addProfileName(self, profileName):
        profile = dbio.get("profiles", name=profileName)
        profileId = profile.id
        if profileId is None:
            logWarning("Profile with name %s not existing in database"%profileName)
            return

        profile = WeeklyProfile("-")
        profile.id=profileId
        self.profiles.append(profile)
        #WARNING: this adds a profile with uncomplete information, just for ids!!!

    def saveToDB(self, processId=None, commit=True):
        """NOTE: Saves only the link to profiles, not the profiles themselves
        """
        if processId:
            self.processId = processId
            
        elif self.processId is None:
            logWarning("ProcessPeriod.saveToDB: cannot save period[%s:%s]. Period not assigned o any process"%(self.start, self.stop))
            return
            
        period = dbio.get("periods", start=self.start, stop=self.stop) # check if a period with same start/stop exists, otherwise create a new one
        if not period:
            self.pid = dbio.insertRow("periods", start=self.start, stop=self.stop)
        else:
            self.pid = period.id

        if self.ppid is None:            
            # save process period to DB
            self.ppid = dbio.insertRow("process_periods",
                                    qprocessdata_QProcessData_ID=self.processId,
                                    periods_id=self.pid,
                                    step=self.step,
                                    scale=self.scale)
        else:
            pp = dbio.get("process_periods", id=self.ppid)
            pp.set(step=self.step, scale=self.scale, periods_id=self.pid)

        # delete link to profiles which are no longer used
        # now save link to profiles and delete those ones no longer used

        toBeSaved = [profile.id for profile in self.profiles]
        ppps = dbio.select("process_period_profiles", process_periods_id=self.ppid)
        for ppp in ppps:    # overwrite the already 
            profileId = ppp.profiles_id
            if profileId in toBeSaved:  # nothing to be done, link is already stored
                index = toBeSaved.index(profileId)
                toBeSaved.pop(index)
            else:           # a profile had been associated, but was deleted in the meanwhile
                dbio.delete("process_period_profiles", process_periods_id=self.ppid, profiles_id=profileId)
            
        for profileId in toBeSaved:     # generate new entry for the remaining
            dbio.insertRow("process_period_profiles", process_periods_id=self.ppid, profiles_id=profile.id)

        if commit: dbio.commit()

    def deleteFromDB(self):

        # delete entries in link table process_period_profiles
        dbio.delete("process_period_profiles", process_periods_id=self.ppid)
        dbio.delete("process_periods", id=self.ppid)

    def copy(self):
        """ copies a PeriodProfile, associated to the same object (process) and that later on
        can be stored independently """
        pnew = ProcessPeriod(None, None, self.processId, self.step, self.scale)    # generated with pid=None in order to avoid re-loading from DB
                                                                                   # link ppid to process_periods is broken, in order to allow for modifying step and scale 
        pnew.start = self.start
        pnew.stop = self.stop
        pnew.profiles = self.profiles[:]

        return pnew

def deleteProcessPeriodsFromDB(processId):
    """Remove all profile periods defined for a given process from data base.
        
    :param processId: Data base ID of process from which to remove the periods
    :type  processId: Integer
        
    :returns: nothing
    """
    dbio.deleteMany("process_period_profiles", 'process_periods_id IN (SELECT id FROM process_periods WHERE qprocessdata_QProcessData_ID = %d)' % processId)
    dbio.delete("process_periods", qprocessdata_QProcessData_ID=processId)

def exportProcessPeriodsFromDB(processId):
    """Export all defined process periods from data base.
    
    The periods are sorted by the start hour.
    
    :returns: [{"start" : inclusive start hour of year, "stop" : inclusive stop hour of year, "step": stepping through period in hours, "scale" : load [0,1], "profiles" : [profile names]}]
    :rtype: List of Dictionaries
    """
    processPeriods = []
    for processPeriod in dbio.select("process_periods", qprocessdata_QProcessData_ID=processId):
        processPeriodProfileNames = []
        for processPeriodProfile in dbio.select("process_period_profiles", process_periods_id=processPeriod.id):
            profile = dbio.get("profiles", id=processPeriodProfile.profiles_id)
            processPeriodProfileNames.append(profile.name)
        period = dbio.get("periods", id=processPeriod.periods_id)
        startDate = period.start
        startHour = round(dateToFirstYearHourOfDay(startDate), 2)
        stopDate = period.stop
        stopHour = round(dateToLastYearHourOfDay(stopDate) + 1, 2) # stopHour is point in time 
        step = float(processPeriod.step) * DAY
        scale = float(processPeriod.scale) * .01
        processPeriods.append({"start"    : startHour,
                               "stop"     : stopHour,
                               "step"     : step,
                               "scale"    : scale,
                               "profiles" : processPeriodProfileNames})
    processPeriods.sort(key=lambda x: x["start"])
    return processPeriods
            
    
    
