# -*- coding: utf-8 -*-
#==============================================================================
#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   EINSTEIN profiles
#
#------------------------------------------------------------------------------
#
#   Run time profiles for processes
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
"""
:mod:`einstein.modules.profiles` -- Operation time of processes
===============================================================

   :synopsis: Running time profiles for processes
   :author:   David Baehrens <david.baehrens@energyxperts.net>
"""
from constants import DAY
from GUI.status import Status
from globals import logger
import database.dbio as dbio
from datetime import datetime, time
from modules.schedules.scheduleTools import BasicSchedule
from lib.dateTime.dateTimeTools import timeToHour, MIDNIGHT

### Constants

WEEKDAYS = (
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
            )
"""Weekday names in order."""

weekdayShort = (
                "Mon",
                "Tue",
                "Wed",
                "Thu",
                "Fri",
                "Sat",
                "Sun",
                )
"""Weekday short names in order."""

### Errors

class ProfileNotFoundError(Exception):
    """No profile found with given name."""
    pass

class ProfileWithoutIntervalsError(Exception):
    """Profile has no running time intervals."""
    pass


### Functions

def getAllProfilesFromDB():
    """Convenience function to get all known profiles at once.
    """
    profileIds = [q.id for q in dbio.select("profiles")]
    return [WeeklyProfile("---", id=profileId) for profileId in profileIds]

### Classes

class WeeklyProfile(object):
    """Daily intervals of running time during one week on selected weekdays.
    
    :param name: Unique name used for identification
    :type  name: String
    :param intervals: Initial daily intervals, default: empty
    :type  intervals: List of 3-tuples (start time, stop time, scale integer)
    :param weekday: Initial selection of weekdays, default: all False
    :type  weekday: Dictionary {<names from :data:`einstein.modules.profiles.WEEKDAYS`> : <Booleans>}
    :param id: The database id of the profile.
    :type id: Integer
    :returns: Weekly profile instance
    :rtype: einstein.modules.profiles.WeeklyProfile
    
    """
    
    def __init__(self, name, intervals=None, weekday=None, id=None):
        self.id   = id      # database id
        self.name = "weekly profile (generic)"

        if id is not None:      # load from DB or create a new one
            self.loadFromDB()
        else:
            self.name = name
            if intervals:
                self.intervals = intervals
            else:
                self.clearIntervals()
                self.intervals.append([time(0, 0), MIDNIGHT, 100.0])

            if weekday:
                self.weekday = weekday[:]
            else:
                self.weekday = [True]*len(WEEKDAYS)

            self.calculateGlobals()
            self.editAllowed = True

    def __repr__(self):
        string = "Profile(%s): %s [%s][%s]"%(self.id, self.name, self.intervals, self.weekday)
        return string

    def copy(self):
        new = WeeklyProfile(self.name + " ("+ _("copy") + ")",
                      intervals = self.intervals[:],
                      weekday = self.weekday[:])

        return new
        
    def loadFromDB(self, id = None):
        """Retrieve daily intervals and selected weekdays from data base.
    
        :returns: nothing
        :exception einstein.modules.profiles.ProfileNotFoundError:         No profile with our name found in data base.
        :exception einstein.modules.profiles.ProfileWithoutIntervalsError: Profile with our name has no intervals in data base.
        """
        # table profiles
        if id is not None:
            self.id = id

        if self.id is None:
            dbProfile = dbio.get("profiles", name=self.name)
            self.id = dbProfile.id
        else:
            dbProfile = dbio.get("profiles", id=self.id)
            self.name = dbProfile.name

        try:
            self.weekday = [getattr(dbProfile, w) != 0 for w in WEEKDAYS]
        except IndexError:
            raise ProfileNotFoundError
        
        self.intervals = []
        try:
            intervalIds = [p_i.intervals_id for p_i in dbio.select("profile_intervals", profiles_id = dbProfile.id)]
            for intervalId in intervalIds:
                interval = dbio.get("intervals", id=intervalId)
                self.addInterval(interval.start, interval.stop, interval.scale)
        except LookupError:
            raise ProfileWithoutIntervalsError

        self.calculateGlobals()
        
        self.checkIfEditAllowed()

    def saveToDB(self):
        """Store profile to data base.
        
        :returns: nothing
        """
        # table profiles
        name = self.name.encode("utf-8") if isinstance(self.name, unicode) else self.name
        profileRow = {'name' : name}
        profileRow.update(dict([ (WEEKDAYS[k],v and 1 or 0) for k,v in enumerate(self.weekday)]))
        dbProfile = dbio.get("profiles", id=self.id)

        if not self.editAllowed:
            logger.showWarning(_("Cannot save profile \'%s\'. The profile is used also in other projects")%name)
            return False
            
        if dbProfile is not None:
            dbProfile.set(**profileRow)
        else:
            self.id = dbio.insertRow("profiles", **profileRow)

        dbProfile = dbio.get("profiles", id=self.id)

        
        # wipe profile_intervals for this profile
        dbio.delete("profile_intervals", profiles_id=self.id)

        for interval in self.intervals:
            # table intervals
            dbInterval  = dbio.get("intervals", start=interval[0], stop=interval[1], scale=int(interval[2]))
            if dbInterval is None:
                intervalRow   = dict(zip(('start', 'stop', 'scale'), interval))
                intervalRow["scale"] = float(intervalRow["scale"]) # conversion from numpy.float to float necessary for dbio
                intervalRowId = dbio.insertRow("intervals", **intervalRow)
            else:
                intervalRowId = dbInterval.id
            # table profile_intervals
            dbio.insertRow("profile_intervals", **{'profiles_id':self.id,'intervals_id':intervalRowId})

        dbio.commit()

        self.checkIfEditAllowed()

        return True
        
    def deleteFromDB(self):
        """Remove profile from data base.
        
        :returns: nothing
        :exception einstein.modules.profiles.ProfileNotFoundError: No profile with our name found in data base.
        """
        if self.id is None:
            return True     # nothing to do

        if not self.editAllowed:
            logger.logWarning(_("Cannot delete profile %s. The profile is used also in other projects"))
            return False
        
        # table profiles
        dbProfile = dbio.get("profiles", id=self.id)
        if dbProfile is None:
            raise ProfileNotFoundError
        else:
            dbProfileRowId = dbProfile.id
            dbio.delete("profiles", id=self.id)
        
        # wipe profile_intervals for this profile
        dbio.delete("profile_intervals", profiles_id=dbProfileRowId)
        dbio.commit()

        return True
            
    def checkIfEditAllowed(self):

        ppIds = [ppp.process_periods_id for ppp in dbio.select("process_period_profiles", profiles_id=self.id)]
        ppIds = list(set(ppIds))

        objectIds = [pp.qprocessdata_QProcessData_ID for pp in dbio.selectByList("process_periods", "id", ppIds)]
        objectIds = list(set(objectIds))

        processIds = [id for id in objectIds if id > 0]
        equipmentIds = [-id for id in objectIds if id < 0]
        
        processes = dbio.selectByList("qprocessdata", "id", processIds)
        equipment = dbio.selectByList("qgenerationhc", "id", equipmentIds)
        usedIn = [obj.Questionnaire_id for obj in (processes + equipment)]
        usedIn = list(set(usedIn))

        self.usedIn = usedIn

        if (len(usedIn)==1 and usedIn[0]==Status.PId) or len(usedIn)==0:  # used only in current project, edit allowed
            self.editAllowed = True
        else:
            self.editAllowed = False
           
    def clearIntervals(self):
        """Remove all running time intervals from profile instance.
        
        :returns: nothing
        """
        self.intervals = []
        
    def addInterval(self, start, stop, scale):
        """Add an running time interval to profile instance.
                
        :param start: start time (inclusive)
        :type  start: time
        :param stop:  stop time (inclusive)
        :type  stop:  time
        :param scale: percentage of operation relative to nominal load
        :type  scale: Integer
        :returns:     nothing
        """

        firstLaterStart = 0
        while firstLaterStart < len(self.intervals):
            if start < self.intervals[firstLaterStart][0]:
                break
            else:
                firstLaterStart += 1
        self.intervals.insert(firstLaterStart, (start, stop, scale))

    def calculateGlobals(self):
        intervalsHour = [(timeToHour(start), timeToHour(stop), scale/100) for (start, stop, scale) in self.intervals]

##        weekDays = [self.weekday[w] for w in WEEKDAYS]
        weekDays = self.weekday
        nDays = weekDays.count(True)

        self.TPeriod = 168   # fixed value for the moment: one week
        self.nDays = nDays

        if nDays > 0 and len(intervalsHour) > 0:
            s = BasicSchedule()
            s.getPeriods(intervalsHour)
            hop, hopNom, fAvg, fMin, fMax = s.getGlobals()

            self.hPerDay = hop
            self.hPerDayNom = hopNom
            self.fAvg = fAvg
            self.fMin = fMin
            self.fMax = fMax
        else:
            self.hPerDay = 0
            self.hPerDayNom = 0
            self.fAvg = 0
            self.fMin = 0
            self.fMax = 0

    def setAllWeekdays(self, val):
        self.weekday = [val]*7
        
def processExists(processId):
    if processId >=0:
        return (dbio.get("qprocessdata", id=processId) is not None)
    else:
        return (dbio.get("qgenerationhc", id=-processId) is not None)



