if __name__=="__main__":
    import sys
    import os
    sys.path.append(os.path.abspath("../../.."))
    print sys.path
                    
from auxiliary.auxiliary import findFirstGE, findFirstGT
from constants import EPSILON

##################################################################################
#
# Package for managing generic schedule types:
#   - Basic Schedule: defined by f,t pairs 
#   - Generic Schedule: defined by (start, stop, scale) values 
#
##################################################################################

class BasicSchedule():
    """ Defines and provides tools for operating with (time-)schedules formed
        by pairs of values (f, t):

        f(t) = f_n

        for t_n <= t < t_(n+1)

        For t < t_0 and t > t_(n-1) f(t) = constant is supposed
        Boundaries tmin and tmax are specified implicitely by the first
        and last entry of the t-vector.

        Parameters:
        f: list of float values for function values
        t: list of float values for time
        step: True  -> step change of f(t) at each time t_n
              False -> gradual change of f(t) from t_n to t_n+1 (NOT YET IMPLEMENTED)
    """

    def __init__(self, t=None, f=None, step=True, tmin=0, tmax=None):
        self.step = step
        if f is not None:
            self.f = list(f)
            if t is not None:
                self.t = list(t)
            else:
                self.t = [float(val) for val in range(len(self.f))] # default (no t-vector defined): enumerate t from 0 to n-1

            if self.t[-1] < tmax and len(f) > 0:
                self.t.append(tmax)
                self.f.append(f[-1])
            if self.t[0] > tmin and len(f) > 0:
                self.t.insert(0, 0)
                self.f.insert(0, f[0])
        else:
            self.t = []
            self.f = []

            

    def __repr__(self):
        string = "BasicSchedule\n"
        for i, t in enumerate(self.t):
            string += "%10s %10s\n"%(t, self.f[i])
        string += "---\n"
        return string

    def multiply(self, s):
        """ multiplies two basic schedules:
        fp(t) = fo(t) * fs(t)
        """
        
        # combine the time intervals
        tp = sorted(list(set(self.t+s.t)))

        i = 0
        j = 0
        fo = self.f[0]
        fs = s.f[0]
        
        for i, t in enumerate(tp):
            if t in self.t:
                fo = self.f[i]
            else:
                self.t.insert(i, t)
                self.f.insert(i, fo)

            if t in s.t:
                fs = s.f[j]
                j += 1
                
            self.f[i] *= fs
                    

        return self

    def add(self, s, scale=1.0):
        """ adds two basic schedules:
        fp(t) = fo(t) + scale*fs(t)
        """
        # combine the time intervals
        tp = sorted(list(set(self.t+s.t)))

        i = 0
        j = 0
        fo = self.f[0] if len(self.f) > 0 else 0.0
        fs = s.f[0] if len(s.f) > 0 else 0.0
        
        for i, t in enumerate(tp):
            if t in self.t:
                fo = self.f[i]
        
            else:
                self.t.insert(i, t)
                self.f.insert(i, fo)

            if t in s.t:
                fs = s.f[j]
                j += 1
                
            self.f[i] += scale*fs


        self.f = [val if abs(val)>EPSILON else 0.0 for val in self.f]

    def getPeriods(self, periods, default=0.0, tmin=0.0, tmax=None):
        """ periods = sorted list of periods (start, stop, scale)
        f(t) = scale for start <= t <= stop

        :params default: default value for times t not included in any period

        """
        pp =  sorted(periods, key=lambda period: period[0])

        t_previous = tmin
        self.f = []
        self.t = []
        self.step = True
        
        for (start, stop, scale) in pp:
            if start > t_previous:  # current interval is not adjacent to the previous one -> fill with zero interval
                self.f.append(default)
                self.t.append(t_previous)

            self.f.append(scale)
            self.t.append(start)
            t_previous = stop

        if not tmax or t_previous < tmax:
            self.f.append(default)
            self.t.append(t_previous)
            if tmax is not None and t_previous < tmax:
                self.f.append(default)
                self.t.append(tmax)
        else:
            self.f.append(scale)
            self.t.append(tmax)

    def toPeriods(self):
        """ inverse of getPeriods """
        gs = GenericSchedule(step=self.step, default=0.0)
        gs.tmin = self.t[0]
        gs.tmax = self.t[-1]

        for i, f in enumerate(self.f[:-1]):
            if f != 0.0: 
                gs.intervals.append((self.t[i], self.t[i+1], f))

        return gs
        
    def copy(self):
        fc = BasicSchedule()
        fc.f = self.f[:]
        fc.t = self.t[:]
        return fc
    
    def plot(self):
        """ for step=True only: extends the data to a plottable curve
        """
        pc = self.copy()
        if self.step:
            for i in range(len(self.t)-1):
                pc.t.insert(2*i+1, self.t[i+1])
                pc.f.insert(2*i+1, self.f[i])

        return pc

    def getGlobals(self):
        if len(self.f)==0: return 0, 0, 0, 0, 0
        hop = 0
        hopNom = 0
        for i, f in enumerate(self.f[:-1]):
            if f > 0:
                dt = self.t[i+1] - self.t[i]
                hop += dt
                hopNom += dt*f

        fAvg = hopNom / hop if hop > 0 else 0
        fMax = max(self.f)
        fMin = 0 if fMax < EPSILON else min([f for f in self.f if f > EPSILON])
        return hop, hopNom, fAvg, fMin, fMax

    def periodicExtension(self, tminp, tmaxp, shift=0):
        """ extends the BS periodically and shifts it in time
        so that fp(t+n*T+shift) = f(t)
        """

        T = self.t[-1] - self.t[0]
        dn_min = int((tminp - self.t[0] - shift)/T)

        tCycle = [t + shift + dn_min*T for t in self.t]
        fCycle = self.f[:]

        i0 = findFirstGT(tminp, tCycle)-1
        
        self.t = tCycle[i0:]
        self.f = fCycle[i0:]
        self.t[0] = tminp
            
        while self.t[-1] < tmaxp:
            tCycle = [t+T for t in tCycle]
            self.t += tCycle
            self.f += fCycle[:]

        i1 = findFirstGE(tmaxp, self.t)
        self.t = self.t[:i1+1]
        self.f = self.f[:i1+1]
        if len(self.f) > 1 and tmaxp < self.t[-1]: self.f[-1] = self.f[-2]
        self.t[-1] = tmaxp

        self.compact()


    def compact(self):
        """ eliminates redundant entries t*:
        for step: when f(t*) == f(t-1)
        for non-step: when f(t-1) == f(t*) == f(t+1)
        """

        if len(self.t)==0: return # security check
        
        fp = 0
        fnew = []
        tnew = []
        n = len(self.t)

        if self.step:
            for i, f in enumerate(self.f):
                if i==0 or i==n-1 or abs(f - fp) > EPSILON:
                    fnew.append(f)
                    tnew.append(self.t[i])
                    fp = f

        else:
            for i, f in enumerate(self.f):
                if i==0 or i==n-1 or abs(f - fp) > EPSILON or abs(f - self.f[i+1]) > EPSILON:
                    fnew.append(f)
                    tnew.append(self.t[i])
                    fp = f
            
        self.f = fnew
        self.t = tnew

class GenericSchedule():
    """ Defines and provides tools for operating with (time-)schedules formed
        by a list of intervals (start, stop, scale)

        f(t) = scale for start <= t < stop
        f(t) = default if out of any interval

        Parameters:
        intervals: list of start/stop/scale intervals (float values)
        step: True  -> step change of f(t) at each start time t_n
              False -> gradual change of f(t) from scale_n to scale_n+1 (NOT YET IMPLEMENTED)
    """

    def __init__(self, intervals=None, step=True, default=0.0, tmin = 0.0, tmax = None):
        self.intervals = sorted(intervals, key=lambda interval: interval[0]) if intervals else []
        self.step = step
        self.default = default
        self.tmin = tmin
        self.tmax = tmax

    def __repr__(self):
        string = "GenericSchedule\n"
        for i, interval in enumerate(self.intervals):
            string += "%10s - %10s: %10s\n"%(interval)
        string += "---\n"
        return string

    def copy(self):
        fc = GenericSchedule()
        fc.intervals = self.intervals[:]
        fc.step = self.step
        fc.default = self.default
        fc.tmin = tmin
        fc.tmax = tmax
        
        return fc
    
    def plot(self):
        """ for step=True only: extends the data to a plottable curve
        """
        bs = BasicSchedule(step=self.step)
        bs.getPeriods(self.intervals, default=self.default, tmin=self.tmin, tmax=self.tmax)
        return bs.plot()

    def getGlobals(self):
        hop = 0
        hopNom = 0
        fMax = 0
        for (start, stop, f) in self.intervals:
            dt = stop - start
            if f > 0:
                hop += dt
                hopNom += dt*f
                fMax = max(f, fMax)

        if self.default > 0:
            hop_def += (self.tmax - self.tmin - hop)
            hop += hop_def
            hopNom += hop_def*self.default
            if hop_def > 0: fMax = max(f, self.default)

        fAvg = hopNom / hop if hop > 0 else 0
        return hop, hopNom, fAvg, fMax



if __name__ =="__main__":
    x = BasicSchedule()
    x.t = [0.0, 2.0, 3.0, 5.0, 8.0, 10.0, 13.0, 15.0]
    x.f = [0.5, 1.0, 0.5, 2.0, 0.5,  3.0,  0.5,  0.5]

    print "x", x

    y = BasicSchedule()
    y.t = [0.0, 1.0, 4.0, 6.0, 7.0, 9.0, 11.0, 12.0, 14.0, 15.0]
    y.f = [0.0, 1.0, 0.0, 2.0, 0.0, 3.0,  0.0,  4.0,  0.0,  0.0]

    print "y", y

    z = x.multiply(y)
    print "z", z

    fz_target = [0.0, 0.5,  1.0, 0.5, 0.0, 0.0, 4.0, 0.0, 0.0, 1.5,
                 9.0, 0.0, 12.0, 2.0, 0.0, 0.0]

    print "TESTRESULT (True=ok): %s"%(z.f==fz_target)

    a = x.copy()
    a.periodicExtension(16.5, 50.5, 2)
    print "Periodic Extension of a: \n"%a
    
