# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	SCHEDULES
#			
#------------------------------------------------------------------------------
#			
#	Functions for management of time schedules
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
"""
:mod:`einstein.modules.schedules` -- Detailed schedules of periodic operation
=============================================================================

   :synopsis: Periodic yearly load profiles of process operation
   :author:   Hans Schweiger <hans.schweiger@energyxperts.net>
   :author:   David Baehrens <david.baehrens@energyxperts.net>
"""
from profiles import WeeklyProfile
from periodSchedule import *
from scheduleTools import BasicSchedule
from modules.ccheck.dataSet import loadDataSets, readDataFile, makeSchedule

from datetime import date, timedelta, time
from math import fabs, fmod, floor, ceil
import numpy as np
import copy
from auxiliary.auxiliary import checkLimits, findFirstGE, normalize
from auxiliary.auxiliary import toGettextAndUnicode as _U
from GUI.status import Status
from modules.interfaces import createQ_t
from constants import *
from modules.messageLogger import *
from GUI.dialogGauge import DialogGauge

### Constants
periodReference = date(2007, 1, 1) # year 2007 starts and ends on Monday
DEFAULTSCHEDULES = ["operation", "batchCharge", "batchDischarge"]
DEFAULTCHARGETIME = 0.2 #20% of batch duration
TOLERANCE_GAP = 0.5
"""Minimum time between consecutive cycles after random tolerance shift (hours)"""

class Schedule():
    """
    class that defines EINSTEIN schedules

    :param self.holidays: list of tuples of holiday (start,stop) periods (both inclusive, period across the end of the
                          year is supported).

    :param fav: hourly scale factor

    Specific parameters for simplified schedules built as follows

    f(t) = fmonth x fweek x fholiday

    :param self.weekly: list of tuples (start,stop) that describe the weekly profile (on/off)
    :param self.monthly: list of 12 monthly scale factors
    :param self.daily: intermediate list for daily profile, used for building self.weekly
    
    """
    def __init__(self, name):     #by default assigns a constant profile throughout the year
        self.daily = [[(0.0, 24.0)]]
        self.weekly = [(0.0, 120.0)]
        self.monthly = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.holidays = [(365, 365)]
        self.NDays = 260
        self.HPerDay = 24.      #operating period for the present schedule
        self.NBatch = 1
        self.HBatch = 24.
        self.PartLoad = 1.0     # for effective hour scaling in detailed schedules
        self.name = name
        self.HPerYear = self.NDays * self.HPerDay
        self.hop = 1.0          #yearly operation hours from normalisation calculations
                                #should be finally identical with self.HPerYear

        self.fav = None         #fav is stored as a vector, as calculation seems
                                #quite time-consuming (-> calc_fav)

        self.periodScheduleDelete = []  # buffer for period schedules to be deleted
        self.scheduleType = None        # SIMPLE / PERIOD / COMBINED
        self.periodSchedule = None
        
        # specific parameters for process schedules
        self.subType = "operation"      # operation / batchCharge / batchDischarge
        self.component = None           # inflow / outflow / startup / maintenance

        # canceled out for speed-up. TODO: assure that Status variable is appropriately set whenever some change is carried out
##        projectData, dummy = Status.prj.getProjectData()
##        Status.HPerDayInd = projectData.HPerDayInd # establish current data

    def __repr__(self):
        string = "Schedule %s [%s;%s]: HPerYear %s HPerYearNom %s NDays %s HPerDay %s NBatch %s HBatch %s"%(self.name,
                                                                                                            self.scheduleType,
                                                                                                            self.subType,
                                                                                                            self.HPerYear,
                                                                                                            self.HPerYearNom,
                                                                                                            self.NDays,
                                                                                                            self.HPerDay,
                                                                                                            self.NBatch,
                                                                                                            self.HBatch)
        return string
                                                                                                        
                                                                                                        
    def zero(self):
        """sets the schedule to zero without calculations
        """

        self.daily = [[]]
        self.weekly = []
        self.monthly = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.holidays = []
        self.NDays = 0
        self.HPerDay = 0.      #operating period for the present schedule
        self.NBatch = 1
        self.HBatch = 0.
        self.PartLoad = 0.0     # for effective hour scaling in detailed schedules
        self.subType = "operation"
##        self.name = name
        self.HPerYear = self.NDays * self.HPerDay
        self.HPerYearNom = None
        self.hop = 0.0          #yearly operation hours from normalisation calculations
                                #should be finally identical with self.HPerYear

        self.fav = createQ_t()         #fav is stored as a vector, as calculation seems
        
    @property
    def NHolidays(self):
        result = 0
        for start, stop in self.holidays:
            result += (stop - start)%365 + 1
        return result
    
#------------------------------------------------------------------------------		
    def setPars(self, subType, NDays, HPerDay, NBatch, HBatch, holidays, periodSchedule=None, component="operation"):
#------------------------------------------------------------------------------		
#   sets the global parameters of this schedule
#------------------------------------------------------------------------------		
        if subType not in DEFAULTSCHEDULES: subType = DEFAULTSCHEDULES[0]
        self.subType = subType
        self.component = component
        
        self.NDays = int(checkLimits(NDays, 0, 365, default=260))
        self.NBatch = int(checkLimits(NBatch, 1, 100, default=1))
        self.holidays = list(holidays)

        if periodSchedule:
            self.periodSchedule = periodSchedule
            self.setParsDetailed()
            self.scheduleType = "PERIOD"
##        elif combSchedule:
##            self.combSchedule = combSchedule
##            self.setParsCombined()
##            self.scheduleType = "COMBINED"
        else:
            self.HPerDay = HPerDay
            try:
                self.HBatch = checkLimits(HBatch, 0.0, 24.0 / self.NBatch, default=Status.HPerDayInd / self.NBatch)
            except TypeError:
                self.HBatch = checkLimits(HBatch, 0.0, 24.0 / self.NBatch, default=24.0 / self.NBatch)

            self.setDefault() 
            self.build_fav()
            self.scheduleType = "SIMPLE"

        self.hop = self.HPerYear
        
    def setParsDetailed(self):
        """ functions for setpars specific for detailed schedules
        """
        if self.subType == "operation":
            self.fav = self.periodSchedule.operationProfile
            self.HPerYear = self.periodSchedule.HPerYear
            self.HPerYearNom = self.periodSchedule.HPerYearNom

        elif self.subType == "batchCharge":
            if self.component == "inflow":
                dt = self.periodSchedule.inflowTime
                self.fav = self.periodSchedule.inflowProfile
                self.HPerYear = self.periodSchedule.HPerYearInflow
                self.HPerYearNom = self.periodSchedule.HPerYearInflowNom
            else:
                dt = self.periodSchedule.startupTime
                self.fav, self.HPerYear, self.HPerYearNom      = self.periodSchedule.getYearlyBatchChargeProfile(chargeTime=dt)
                    
        elif self.subType == "batchDischarge":
##            self.fav, self.HPerYear, self.HPerYearNom      = self.periodSchedule.getYearlyBatchOutflowProfile()
            self.fav = self.periodSchedule.outflowProfile
            self.HPerYear = self.periodSchedule.HPerYearOutflow
            self.HPerYearNom = self.periodSchedule.HPerYearOutflowNom

        self.HPerDay = 0
        self.HBatch = 0
        if self.NDays > 0:
            self.HPerDay = self.HPerYear / self.NDays

            if self.HPerYearNom > 0 and self.NBatch > 0:
                self.HBatch = self.HPerYearNom / (self.NDays*self.NBatch)

        self.bs = self.periodSchedule.bs

    def setParsFav(self, fav):
        """ functions for setpars specific for schedules from hourly data
        """
        self.fav = fav

        self.HPerYear = np.sum(self.fav > EPSILON)
        self.HPerYearNom = self.HPerYear        # nominal operating hours cannot be determined from normalised fav-vector. PartLoad=1 supposed.
        self.NDays = 0
        nCyclesPerYear = 0
        for d in xrange(NDAYS):
            fd = self.fav[24*d:24*(d+1)]
            if np.max(fd) > EPSILON:
                self.NDays += 1

            fdc = fd > EPSILON
            fdc[1:24] = fdc[1:24] - fdc[0:23]
            nCycles = np.sum(fdc == 1) # count the jumps from 0 to 1
            nCyclesPerYear += nCycles

        self.HPerDay = 0
        self.HBatch = 0
        if self.NDays > 0:
            self.HPerDay = self.HPerYear / self.NDays
            self.NBatch = nCyclesPerYear / self.NDays

            if self.HPerYearNom > 0 and self.NBatch > 0:
                self.HBatch = self.HPerYearNom / (self.NDays*self.NBatch)

    def setBSFromFav(self):
        self.bs = BasicSchedule(f = self.fav, tmax = 8760)
        self.bs.compact()
        self.HPerYear = 1.0

#------------------------------------------------------------------------------		
    def f(self, time):
#------------------------------------------------------------------------------		
#   calculates the instantaneous value of the schedule at time 
#------------------------------------------------------------------------------		
        if time < 0.0 or time > YEAR: return 0.
        day = int(floor(time / DAY)) + 1

        if self.isHoliday(day):return 0.

        weekTime = time % WEEK
        fweek = 0.0
        for period in self.weekly:
            start, stop = period
            if weekTime >= start and weekTime <= stop:
                fweek = 1.
                break
            
        month = findFirstGE(day, MONTHSTARTDAY)
        fmonth = self.monthly[month - 1]
        
        return fweek * fmonth

#------------------------------------------------------------------------------		
    def calc_fav(self, time):
#------------------------------------------------------------------------------		
#   calculates the average value within the interval [time,time+Dt] 
#------------------------------------------------------------------------------		
        if time < 0.0 or time >= YEAR: return 0.
        day = int(time / DAY) + 1

        # First check whether we can respect the holidays and then
        # whether there is one.

        if self.NDays + self.NHolidays <= 365 and self.isHoliday(day):
            return 0.

        weekTime = time % WEEK
        fweek = 0.0

        for period in self.weekly:
            start, stop = period
            if weekTime < stop:
                weekTime1 = weekTime + Status.TimeStep
                if weekTime1 > start:
                    if weekTime >= start and weekTime1 <= stop:     #time interval fully within period
                        fweek += 1.
                        break
                    elif weekTime >= start and weekTime1 > stop:    #time interval starts within period, but ends afterwards
                        fweek += (stop - weekTime) / Status.TimeStep
                    elif weekTime < start and weekTime1 <= stop:    #time interval starts before period, but ends within
                        fweek += (weekTime1 - start) / Status.TimeStep
                    elif weekTime < start and weekTime1 > stop:
                        fweek += (stop - start) / Status.TimeStep
                else:
                    break
            
        month = findFirstGE(day, MONTHSTARTDAY)
        fmonth = self.monthly[month - 1]

        return fweek * fmonth
        
#------------------------------------------------------------------------------		
    def calc_weekly(self):
        """ calculates a weekly fav-vector that then is extended to a year


        input: self.weekly
        output: fweek
        """

        fweek = createQ_t()
        

        for period in self.weekly:
            start, stop = period
            it_start = int(start / Status.TimeStep )
            it_stop = int(stop / Status.TimeStep) + 1
            for it in xrange(it_start, it_stop):
                weekTime = it * Status.TimeStep
                if weekTime < stop:
                    weekTime1 = weekTime + Status.TimeStep
                    if weekTime1 > start:
                        if weekTime >= start and weekTime1 <= stop:     #time interval fully within period
                            fweek[it] += 1.
                        elif weekTime >= start and weekTime1 > stop:    #time interval starts within period, but ends afterwards
                            fweek[it] += (stop - weekTime) / Status.TimeStep
                        elif weekTime < start and weekTime1 <= stop:    #time interval starts before period, but ends within
                            fweek[it] += (weekTime1 - start) / Status.TimeStep
                        elif weekTime < start and weekTime1 > stop:
                            fweek[it] += (stop - start) / Status.TimeStep

        # repeat standard week 52 times
        for w in xrange(1,52):
            its_week = int(168*Status.TimeStep + 0.5)
            it_start = w*its_week
            it_stop = it_start + its_week
            fweek[it_start:it_stop] = fweek[0:its_week]

        # and add the 31st of December
        its_day = int(24*Status.TimeStep + 0.5)
        it_start = 52*its_week
        it_stop = it_start + its_day
        assert(it_stop == Status.Nt)
        fweek[it_start:it_stop] = fweek[0:its_day]
        
        return fweek
        
#------------------------------------------------------------------------------		
    def isHoliday(self, day):
#------------------------------------------------------------------------------		
        for period in self.holidays:
            start, stop = period
            if stop >= start:
                if day <= stop and day >= start: return True
            else:
                if day <= stop or day >= start: return True
            
        return False

#------------------------------------------------------------------------------		
    def normalize(self):
#------------------------------------------------------------------------------		
        ftot = Status.TimeStep * Status.Nt
        fsum = self.fav.sum()
        fsum *= YEAR / ftot
        self.hop = fsum

        if self.hop > 0:
            self.fav /= self.hop
        else:
            logDebug("Schedule (normalize): WARNING - schedule is 0 in all time steps")

        if fabs(self.hop - self.HPerYear) > 1.0:
            logDebug("Schedule (normalize): WARNING - normalized operating hours (%s) different from specified in HPerYear (%s)"\
                     % (self.hop, self.HPerYear))

        return fsum
    
#------------------------------------------------------------------------------		
    def build_fav(self):
#------------------------------------------------------------------------------		
#   calculates the vector with average values for all timesteps
#------------------------------------------------------------------------------		
##        self.fav = Status.int.createQ_t()
        
        # first readin monthly factor
##        for month in xrange(12):
##            it_start = int(MONTHSTARTDAY[month]*24*Status.TimeStep+0.5)
##            it_end = int(MONTHSTARTDAY[month+1]*24*Status.TimeStep+0.5)
##            self.fav[startHour:endHour] = self.monthly[month]

        # first readin monthly factor
        fav_holidays = Status.int.createQ_t() + 1.0
        for start, stop in self.holidays:
            
                it_start = (start-1)*24
                it_end = min((stop)*24, 8760)
                fav_holidays[it_start:it_end] = 0

                # add-on for holiday periods crossing the end of the year
                if stop < start or stop > 365:
                    it_end = (stop%365)*24
                    fav_holidays[0:it_end] = 0

        self.fav = fav_holidays*self.calc_weekly()
##            
##        for it in range(Status.Nt):
##            time = Status.TimeStep * it
##            self.fav[it]=self.calc_fav(time)

        self.normalize()

        self.build_bs()

    def build_bs(self):
        """ builds the yearly basic schedule for a simple schedule """
        bs_holidays = BasicSchedule()

        hsPeriods = [((start-1)*24,  min((stop)*24, 8760), 0.0) for (start, stop) in self.holidays]

        bs_holidays.getPeriods(hsPeriods, tmin=0, tmax=8760, default=1.0)

        self.bs = BasicSchedule()
        self.bs.getPeriods([(start, stop, 1.0) for start, stop in self.weekly], tmin=0, tmax=168)
        self.bs.periodicExtension(0, YEAR)

        self.bs.multiply(bs_holidays)


    def setDefault(self):
        """Based on the basic schedule parameters (basic Q) assign a 
        detailed default schedule
        
        """
        if self.subType == "operation":
            if self.NBatch > 0 and self.HBatch * self.NBatch <= Status.HPerDayInd:
                TPeriod = Status.HPerDayInd / self.NBatch
                tStartDay = 12.0 - 0.5 * Status.HPerDayInd
            else:
                TPeriod = 24. / self.NBatch
                tStartDay = 0
##                logMessage("%s: The combined cycle durations are larger than the "\
##                           "industry operating time" % (self.name, ))

            self.daily = [[]]
            for i in range(self.NBatch):
                start = tStartDay + TPeriod * i
                stop = start + self.HBatch
                self.daily[0].append((start, stop))

            self.HPerDay = self.NBatch * self.HBatch
        
        # Charge of batch process: first DEFAULTCHARGETIME % of process duration
        elif self.subType == "batchCharge":
            if self.NBatch > 0 and self.HBatch * self.NBatch <= Status.HPerDayInd:
                TPeriod = Status.HPerDayInd / self.NBatch
                tStartDay = 12.0 - 0.5 * Status.HPerDayInd
            else:
                TPeriod = 24. / self.NBatch
                tStartDay = 0
                
            self.daily = [[]]
            for i in range(self.NBatch):
                start = tStartDay + TPeriod * i
                stop = start + DEFAULTCHARGETIME * self.HBatch
                self.daily[0].append((start, stop))

            self.HPerDay = self.NBatch * self.HBatch * DEFAULTCHARGETIME

        # Discharge of batch process: first DEFAULTCHARGETIME % of process duration after process stop
        elif self.subType == "batchDischarge":
            if self.NBatch > 0 and self.HBatch * self.NBatch <= Status.HPerDayInd:
                TPeriod = Status.HPerDayInd / self.NBatch
                tStartDay = 12.0 - 0.5 * Status.HPerDayInd
            else:
                TPeriod = 24. / self.NBatch
                tStartDay = 0

            self.daily = [[]]
            for i in range(self.NBatch):
                start = (tStartDay + TPeriod * i + self.HBatch) % DAY
                stop = (start + DEFAULTCHARGETIME * self.HBatch) % DAY

                #### take care: the %DAY controls to avoid that processes discharge AFTER 24:00 h of a day
                #### this works well only for 7 days operation without holidays
                #### as this is a quite strange special case, should not give problems for the moment
                #### but should be improved !!!
                
                self.daily[0].append((start, stop))

            self.HPerDay = self.NBatch * self.HBatch * DEFAULTCHARGETIME

        # Now extend daily to weekly profile
        if self.NDays + self.NHolidays > 365: # the time demand wouldn't fit
            self.NDaysPerWeek = (self.NDays - 1.0) / 52.0
        else:
            self.NDaysPerWeek = (self.NDays + self.NHolidays - 1.0) / 52.0
            
        self.HPerYear = self.HPerDay * self.NDays
        self.HPerYearNom = self.HPerYear

        self.weekly = []

        nIntervals = self.NDaysPerWeek * len(self.daily[0])
        fullIntervals = int(nIntervals)
        residual = nIntervals - fullIntervals
        
        intCtr = 0
        for i in range(7):
            
            for dayinterval in self.daily[0]:
                (start, stop) = dayinterval
                start += 24.0 * i
                stop += 24.0 * i
                if intCtr < fullIntervals:
                    self.weekly.append((start, stop))
                    intCtr += 1
                elif intCtr == fullIntervals:
                    stop = start + (stop-start)*residual
                    if residual > EPSILON:
                        self.weekly.append((start, stop))
                        intCtr += 1
                else:
                    break

##        logTrack("Schedule (setDefault): weekly profile created (%s): \n%s" % (scheduleType, self.weekly))

    def combineSchedules(self, scheduleTable):
        """ combines the Schedules from a list of (weight, schedule) - pairs
        """

        #TODO: would be better build this based on (start, stop, scale) - intervals than on fav

        self.bs = BasicSchedule()
        for weight, schedule in scheduleTable:
            self.bs.add(schedule.bs, scale=weight/schedule.HPerYear)

        self.fav = getFavFromBS(self.bs)

    def saveToDB(self):
        """
        save function for generic schedule
        if periodSchedule saves that one, else delete the periodSchedule if exists
        """
        # save detailed schedule
        if self.detailedSchedule:
            self.periodSchedule.saveToDB()
        else:
            if self.periodSchedule:
                self.periodSchedule.deleteFromDB()
    
    
#------------------------------------------------------------------------------		
class Schedules(object):
#------------------------------------------------------------------------------		
#   Module that handles all project schedules
#------------------------------------------------------------------------------		
    
#------------------------------------------------------------------------------		
    def __init__(self):
#------------------------------------------------------------------------------		
        self.outOfDate = True
        self.equipmentSchedulesOutOfDate = True
        Status.profilesInUse = {}   #CPU-time saver: profiles that have already been loaded are stored
       
#------------------------------------------------------------------------------		
#------------------------------------------------------------------------------		
    def create(self):
#------------------------------------------------------------------------------		

        projectData = Status.int.systemModel.projectData
        generalData = Status.int.systemModel.generalData
        
        Status.HPerDayInd = projectData.HPerDayInd
        if Status.HPerDayInd is None:
            logWarning(_U("Industry operating hours are not defined !\n12 hours per day assumed"))
            Status.HPerDayInd = 12.0
        

        if Status.SimTool != "Plus":
            dlg = DialogGauge(Status.main, _("Schedules of operation"), _("generating schedules"))

        Status.profilesInUse = {}   #CPU-time saver: profiles that have already been loaded are stored

        self.calculatePlantSchedule()
        
        self.calculateProcessSchedules()
##        dlg.update(40)
        
        if Status.SimTool != "Plus":
            self.calculateEquipmentSchedules()
##        dlg.update(80)
        
        self.calculateWHEESchedules()

        if Status.SimTool != "Plus":
            dlg.Destroy()

        self.outOfDate = False

    def calculatePlantSchedule(self):

        self.plantSchedule = Schedule("Plant schedule")

        periodSchedule = None
        scheduleType = "operation"
        holidays = Status.prj.getHolidays()
        NDays = Status.NDaysInd
        HPerDay = Status.HPerDayInd
        NBatch = 1
        HBatch = HPerDay

        self.plantSchedule.setPars(scheduleType,
                            NDays,
                            HPerDay,
                            NBatch,
                            HBatch,
                            holidays,
                            periodSchedule)
        
    def calculateProcessSchedules(self):
        """Calculate the Schedules of the processes."""
##        logTrack("Schedules (calcProcS): running")
        
        processes = Status.int.systemModel.process
        processObjects = Status.int.systemModel.processObject

        # first create the working space (dummy schedules)
        self.procOpSchedules = []
        self.procStartUpSchedules = []
        self.procInFlowSchedules = []
        self.procOutFlowSchedules = []
        for k, process in enumerate(processes):
            scheduleNameCommon = "Process No.%s (%s): " % (process.ProcNo, process.Process)

            operationSchedule = Schedule(scheduleNameCommon + "Operation")
            startupSchedule = Schedule(scheduleNameCommon + "Operation")
            inflowSchedule = Schedule(scheduleNameCommon + "Operation")
            outflowSchedule = Schedule(scheduleNameCommon + "Operation")
            
    
            self.procOpSchedules.append(operationSchedule)
            self.procStartUpSchedules.append(startupSchedule)
            self.procInFlowSchedules.append(inflowSchedule)
            self.procOutFlowSchedules.append(outflowSchedule)

            processObject = processObjects[k]
            if processObject.combSchedule: continue             # calculate the schedules (except the combined ones)

            detailedSchedule = processObject.detailedSchedule

            if detailedSchedule:
                periodSchedule = PeriodSchedule(process.Process)
                periodSchedule.loadFromDB(process.QProcessData_ID)
                periodSchedule.calculateGlobals()

                NDays = periodSchedule.NDays
                HPerDay = periodSchedule.HPerDay
                NBatch = periodSchedule.NBatch
                HBatch = periodSchedule.HBatch

                processObject.HPerYear = periodSchedule.HPerYear
                processObject.HPerYearNom = periodSchedule.HPerYearNom
                processObject.PartLoad = periodSchedule.PartLoad
                
                processObject.HPerYearInFlow = periodSchedule.HPerYearInflow
                processObject.HPerYearInFlowNom = periodSchedule.HPerYearInflowNom

                processObject.HPerYearOutFlow = periodSchedule.HPerYearOutflow
                processObject.HPerYearOutFlowNom = periodSchedule.HPerYearOutflowNom

            elif not processObject.combSchedule:
                # simplified schedules
                periodSchedule = None

                NDays = process.NDaysProc
                HPerDay = process.HPerDayProc
                NBatch = process.NBatch
                HBatch = process.HBatch

            # now load global schedule parameters into the schedule class
            holidays = Status.prj.getHolidays()

            # schedule for process operation

            if process.UPHm > 0 or process.UPHm is None:
                if process.ProcType == "continuous":
                    subType = "operation"
                else:
                    subType = "operation"
                operationSchedule.setPars(subType,
                                    NDays,
                                    HPerDay,
                                    NBatch,
                                    HBatch,
                                    holidays,
                                    periodSchedule)
            else:
                operationSchedule.zero()
                
            # schedule for process start-up

            if process.UPHs > 0 or process.UPHs is None:
            
                if process.ProcType == "continuous":
                    subType = "batchCharge"
                else:
                    subType = "batchCharge"
                    
                startupSchedule.setPars(subType,
                                    NDays,
                                    HPerDay,
                                    NBatch,
                                    HBatch,
                                    holidays,
                                    periodSchedule,
                                    component="startup")
            else:
                startupSchedule.zero()
                
            # schedule for process in-flows (for the moment only ONE !!!)

            if process.UPHc > 0 or process.UPHc is None:
            
                if process.ProcType == "continuous":
                    subType = "operation"
                else:
                    subType = "batchCharge"
                    
                inflowSchedule.setPars(subType,
                                    NDays,
                                    HPerDay,
                                    NBatch,
                                    HBatch,
                                    holidays,
                                    periodSchedule,
                                    component="inflow")
            else:
                inflowSchedule.zero()
                
            if process.UPHw > 0 or process.UPHw is None:
                if process.ProcType == "continuous":
                    subType = "operation"
                else:
                    subType = "batchDischarge"
                outflowSchedule.setPars(subType,
                                    NDays,
                                    HPerDay,
                                    NBatch,
                                    HBatch,
                                    holidays,
                                    periodSchedule,
                                    component="outflow")
            else:
                outflowSchedule.zero()

        # build link table for combined schedules
        processIds = [p.id for p in processObjects]

        # load data sets from DB
        self.dataSets, dataFiles = loadDataSets()
        dataSetIds = [ds.id for ds in self.dataSets]
        
        self.dataSchedules = [None]*len(self.dataSets)
        for df in dataFiles:
            pd = readDataFile(df.fileName)
            df.cols = pd.keys
            err = (pd.targets is None)  # set if read error from data file
            if not err:
                dataSetsInFile = [(i, ds) for i, ds in enumerate(self.dataSets) if ds.df==df]       
                for i, ds in dataSetsInFile:
                    ds.values = pd.targets[ds.col-1]
                    if pd.totals is not None: ds.total = pd.totals[ds.col-1]
                    ds.dataType = pd.dataType
                    if ds.dataType == 1: # period data
                        ds.periods = pd.periods

                    
                    schedule = Schedule("DataSet No. %s"%(i+1))
                    schedule.setParsFav(makeSchedule(ds, baseSchedule=self.plantSchedule.fav))
                    schedule.setBSFromFav()
                    self.dataSchedules[i] = schedule

        for k, process in enumerate(processes):
            scheduleTable = []
            processObject = processObjects[k]
            combSchedule = processObject.combSchedule
            if combSchedule:
                for component in processObject.combSchedules:
                    if component.Type >= 10:
                        srcType = 1
                        bsSrcType = component.Type - 10
                    else:
                        srcType = component.Type
                        bsSrcType = None
                        
                    if srcType == 0: # process
                        if component.ObjectId in processIds:
                            ksrc = processIds.index(component.ObjectId)
                            scheduleTable.append((component.Weight, self.procOpSchedules[ksrc]))
                        else:
                            logError("Erroneous ID in combined schedule: processId=%s"%component.ObjectId)
                            continue
                    elif srcType == 1:
                        if component.ObjectId in dataSetIds:
                            isrc = dataSetIds.index(component.ObjectId)
                            if bsSrcType is None:
                                scheduleTable.append((component.Weight, self.dataSchedules[isrc]))
                            else:
                                ds = self.dataSets[isrc]

                                if bsSrcType==1:
                                    iBsSrc = dataSetIds.index(component.ComponentNo)
                                    base = self.dataSchedules[iBsSrc].fav
                                elif bsSrcType==0:
                                    ksrc = processIds.index(component.ComponentNo)
                                    base = self.procOpSchedules[ksrc].fav
                                schedule = Schedule("DataSet No. %s + BS"%(isrc))
                                schedule.setParsFav(makeSchedule(ds, baseSchedule=base))
                                schedule.setBSFromFav()
                                scheduleTable.append((component.Weight, schedule))

                        else:
                            logError("Erroneous ID in combined schedule: dataSetId=%s"%component.ObjectId)
                            continue
                    elif srcType == -1: # plant schedule
                        scheduleTable.append((component.Weight, self.plantSchedule))


                self.procOpSchedules[k].combineSchedules(scheduleTable)
                self.procInFlowSchedules[k] = self.procOpSchedules[k]
                self.procOutFlowSchedules[k] = self.procOpSchedules[k]
                self.procStartUpSchedules[k] = self.procOpSchedules[k]
                    
                    
            else:
                continue

        
                
           
            logDebug("Process No.%s (%s):" % (process.ProcNo, process.Process))
            
        for k, processObject in enumerate(processObjects):
            processObject.schedule = self.procOpSchedules[k]
##            processObject[k].scheduleIn = self.procInFlowSchedules[k]
##            processObject[k].scheduleOut = self.procOutFlowSchedules[k]
##            processObject[k].scheduleStartUp = self.procStartUpSchedules[k]
            
            
    def calculateEquipmentSchedules(self):
        """Generate the Schedules of the equipments."""
##        logTrack("Schedules (calcEquipeSchedules): running")
        
        equipments = Status.int.systemModel.equipment
        holidays = Status.prj.getHolidays()
        self.equipmentSchedules = []
        
        for equipe in equipments:
            # schedule for equipment operation
            newSchedule = Schedule("Equipe No.%s (%s)" % (equipe.EqNo, equipe.Equipment))

            newSchedule.setPars("operation",
                                equipe.NDaysEq,
                                equipe.HPerDayEq,
                                1,
                                equipe.HPerDayEq,
                                holidays)
            self.equipmentSchedules.append(newSchedule)
        
        self.equipmentSchedulesOutOfDate = False

    def calculateWHEESchedules(self):
        """Generate the Schedules of the electrical equipment w. waste heat"""
##        logTrack("Schedules (calcWHEESchedules): running")
        
        whees = Status.prj.getWHEEs()
        holidays = Status.prj.getHolidays()
        self.WHEESchedules = []
        
        for whee in whees:
            # schedule for equipment operation
            newSchedule = Schedule("WHEE No.%s (%s)" % (whee.WHEENo, whee.WHEEName))

            newSchedule.setPars("operation",
                                whee.NDaysWHEE,
                                whee.HPerDayWHEE,
                                whee.NBatchWHEE,
                                whee.HBatchWHEE,
                                holidays)
            self.WHEESchedules.append(newSchedule)
            


def getFavFromBS(bs):
    """ Otains a fav-vector from a basic schedule """
    fav = np.zeros(int(YEAR))

    itc = 0      # index of the active interval (partial values in df)
    df = 0       # fraction f*dt within the current interval
    tdf = 0      # time of the last
    fPrev = 0    # time of previous interval

    
    for i, t in enumerate(bs.t):
        f = bs.f[i]
        it0 = int(t)
                                                 
        tmax = min(t, itc+1)    # maximum time for calculation of df for current interval                                            
        df += fPrev *(tmax - tdf)
        
        if t >= itc+1: 
            fav[itc] = df        # store df into the current interval
            fav[itc+1: it0] = fPrev  # store f into full intervals
            itc = it0
            df = fPrev*(t - it0)     # store overlap in current interval into df

        fPrev = f
        tdf = t

    if itc < YEAR-1:
        tmax = min(t, itc+1)    # maximum time for calculation of df for current interval                                            
        df += fPrev *(tmax - tdf)
        fav[itc] = df
        fav[itc+1:-1] = fPrev
                                                 
    return fav

def globalsFromFav(fav, threshold = EPSILON):
    """ calculates the global values from a fav-vector
    """
    fav = normalize(fav)
    HPerYear = np.sum(fav > threshold)
    
    NDays = 0
    nCyclesPerYear = 0
    for d in xrange(NDAYS):
        fd = fav[24*d:24*(d+1)]
        if np.max(fd) > threshold:
            NDays += 1

            fdc = 1*(fd > threshold)
                
            fdc[1:24] = fdc[1:24] - fdc[0:23]
            nCycles = np.sum(fdc == 1) # count the jumps from 0 to 1
            nCyclesPerYear += nCycles

    if nCyclesPerYear > 0:
        HBatch = float(HPerYear) / nCyclesPerYear
    else:
        HBatch = 0

    
    if NDays > 0:
        HPerDay = float(HPerYear) / NDays
        NBatch = float(nCyclesPerYear) / NDays
    else:
        HPerDay = 0
        NBatch = 1.0

    result = {"HPerYear": HPerYear,
              "NCycles": nCyclesPerYear,
              "NDays": NDays,
              "HPerDay": HPerDay,
              "HBatch": HBatch,
              "NBatch": NBatch}
    
    return result
                
def getProcessCombSchedules(processId):
    return dbio.select("qcombschedules", ProcessId = processId)
    
            
    
