# -*- coding: utf-8 -*-
#==============================================================================#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   ModuleST (Solar thermal systems)
#         
#------------------------------------------------------------------------------
#         
#   Module for calculation of solar thermal systems
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
from math import sin, cos, tan, atan, exp, log, fabs, ceil, floor, pi
import numpy as np
import copy
from auxiliary.auxiliary import noneFilter, toFloat
from auxiliary.auxiliary import toUnicode as U
from lib.auxiliary.auxiliary import noneCheck
from GUI.status import *
from constants import *
from sunday import (Intensity, Angles, sun_hourly,
                                           sun_hourly_WDATA, Location, HOUR,
                                           prep_sun)
from modules.messageLogger import (logError, logMessage, logWarning,
                                            logTrack, showMessage, showWarning,
                                            logDebug)
from GUI.dialogGauge import DialogGauge
from modules.modulePipes import getAllHeatingPipes
from modules.heatFlow import HeatFlow_t
from modules.interfaces import createQ_Tt
import database.dbio as dbio

PARASITIC_ELECTRICITY = 0.01
DEFAULT_SYSTEM_EFFICIENCY = 0.95    # default value for solar system efficiency
MIN_SYSTEM_EFFICIENCY = 0.50        # minimum allowed value for solar system efficiency

DEFAULTSAUTODESIGNST = {"STSolFra":0.5,         # default configuration of auto-design for new projects
                        "STCollType":"any",
                        "STMinYield":300.0}

#default parameters for surface calculations
MININCLINATION_FPC = 15 #   Minimum inclination for flat plate collectors
SURFACEFACTOR = 1.1     #   Additional roof area occupied for piping, separation
                        #   Separation between collectors in rows, etc.

MINSURFACE = 20           #   The minimum area (net area of solar collectors) considered is 20 m^2
DISTANCECONSTRAINT = 1500 #   Absolute limit for distance of collector field
MININCLINATION = 20       #   Minimum inclination of roofs for in-roof-mounting
MAXINCLINATION = 90     #   Maximum inclination
MINTOTNETCOLLAREA = 20  #   Minimum collector area

MINSTATICROOF = 25      #   Minimum static roof capacity

M2TOKW = 0.7            #   Conversion from m2 to nominal thermal power

sunOnH = Intensity()
sunOnT = Intensity()
sunAnglesOnSurface = Angles()
dirSunOnT = [0.0, 0.0, 0.0]

def getDailyRadiation(I, day, lat):
    """Return some (more or less) "reasonable" distribution of yearly solar
    radiation
    very very basic
    
    @param I:
    @param day:
    @param lat:
    """
#       ***GETTING_STARTED***:
#   The SUNDAY functions are written radically in SI-units, this is J, W, s; no
#   hours, no kilos, no Megas ... !!!! => unit conversion from EINSTEIN-units
#   to Sunday (SI-) units is necessary

    kWh_to_J = 3.6e+6
    Hmean = I / 365 * kWh_to_J
    amplitude = max(min(lat / 67.0, 1), -1.)
    H = Hmean * (1.0 - amplitude * cos(2.0 * pi * (day + 10) / 365))

    if DEBUG in ["ALL"]:
        print "ModuleST.getDailyRadiation: H = %s Hmean %s amplitude %s lat %s, day = %s"%(H, Hmean, amplitude, lat, day)
        
    return H
                 
def collectorEfficiency(GT, dT, STc0, STc1, STc2):
    """Return the collector efficiency at a given working point
    
    Return 0.0 if values mismatch or are missing. 
    
    GT: Solar radiation on tilted surface (SI - units: W/m2)
    c1/c2 are in W/m2K, W/m2K2!

    """
    if GT <= 0:
        result = 0.0
    else:
        try:
            result = STc0 - (STc1 + STc2 * dT) * dT / (GT)
        except TypeError:
            logError(_("Couldn't calculate the collector efficiency, "\
                        "a value missing"))
            result = 0.0
    return result
                 
def collectorEfficiencyBiaxial(GbT, GdT, tanThetaL, tanThetaT, dT, STc0, STc1, STc2, K50L, K50T):
    """Return the collector efficiency at a given working point
 
    Return 0.0 if values mismatch or are missing. 
 
    GT: Solar radiation on tilted surface (SI - units: W/m2)
    c1/c2 are in W/m2K, W/m2K2! 
 
    """
    if (GbT + GdT) <= 0:
        return 0.0
    else:
        try:
            bT = -log(K50T) / pow(tan(50.0 * pi / 180.0), 2) 
            bL = -log(K50L) / pow(tan(50.0 * pi / 180.0), 2)
             
            if bT <= 0 and tanThetaT > tan(50.0 * pi / 180.0):
                KT = K50T * exp (1. - tanThetaT / tan(50.0 * pi / 180.0))
            else:
                KT = exp(-bT * pow(tanThetaT, 2)) #transversal IAM
    
            if bL <= 0 and tanThetaL > tan(50.0 * pi / 180.0):
                KL = K50L * exp (1. - tanThetaL / tan(50.0 * pi / 180.0))
            else:
                KL = exp(-bL * pow(tanThetaL, 2)) #longitudinal IAM
    
            c0_b = STc0 * KT * KL
            c0_d = STc0 * (0.5 + 0.5 * min(K50T, 1) * min(K50L, 1))     # 50� IAM supposed for diffuse radiation
            c0 = (c0_b * GbT + c0_d * GdT) / (GbT + GdT)    #weighted average between beam and diffuse
    
            result = c0 - (STc1 + STc2 * dT) * dT / (GbT + GdT)
        except ValueError:
            logError(_("Couldn't calculate the collector efficiency "\
                        "for a biaxial collector, a value is missing."))
            result = 0.0
        return result
    
def collectorEfficiencyFresnel(GbT, GdT, tanThetaL, tanThetaT, dT, STc0, STc1, STc2, K50L, K50T):
    """Return the collector efficiency at a given working point
 
    Return 0.0 if values mismatch or are missing. 
 
    GT: Solar radiation on tilted surface (SI - units: W/m2)
    c1/c2 are in W/m2K, W/m2K2! 
 
    """
    if (GbT + GdT) <= 0:
        return 0.0
    else:
        try:
            bT1 = 0.5 * (K50T - 1.) / 50.0  # constants 0.8 and 0.2 obtained from adjustment for PSE IAM
            bT2 = 0.5 * (K50T - 1.) / 2500.
            
            bL = -log(K50L) / pow(tan(50.0 * pi / 180.0), 2)
             
            thetaT = abs(atan(tanThetaT)) * 180. / pi
            
            KT = 1. + (bT1 + bT2 * thetaT) * thetaT
    
            if bL <= 0 and tanThetaL > tan(50.0 * pi / 180.0):
                KL = K50L * exp (1. - tanThetaL / tan(50.0 * pi / 180.0))
            else:
                KL = exp(-bL * pow(tanThetaL, 2)) #longitudinal IAM
    
            c0_b = STc0 * KT * KL
            c0_d = STc0 * (0.5 + 0.5 * min(K50T, 1) * min(K50L, 1))     # 50� IAM supposed for diffuse radiation
            c0 = (c0_b * GbT + c0_d * GdT) / (GbT + GdT)    #weighted average between beam and diffuse

            result = c0 - (STc1 + STc2 * dT) * dT / (GbT + GdT)
        except TypeError:
            logError(_("Couldn't calculate the collector efficiency "\
                        "for a Fresnel collector, a value is missing."))
            result = 0.0
        return result
                 
class SolarCostFunction():
    """ auxiliary function
        converts the cost function in tabular form (for 30, 300 and 3000 kW) into
        the parameters Pref, FixCost, AddCost of an exponential function

        C_turnKey(Pnom) = Pnom * (self.FixCost + self.AddCost * exp(-Pnom / self.Pref))

        Once the parameters are generated, the function calculate returns
        the corresponding cost parameters as a function of Pnom
    """
        
    def __init__(self, c30, c300, c3000, unitPrice300, OM):
        self.calculateCostCoefs(c30, c300, c3000)
        try:
            self.costRatio = float(unitPrice300) / c300
        except TypeError:
            self.costRatio = 0.5
        self.OM = float(OM)

    def calculate(self, Pnom):
        C_turnKey = Pnom * (self.FixCost + self.AddCost * exp(-Pnom / self.Pref))
        C_col = C_turnKey * self.costRatio
        C_OM = Pnom * self.OM
        return (C_turnKey, C_col, C_OM)
        
    def calculateCostCoefs(self, c30, c300, c3000):
        """calculate iteratively the coefficients of the cost function
        C = C0 + C1*exp(-Pnom/Pref)
  
        """
        # Previous checks if curve is reasonable:
        if c30 < c300 or c300 < c3000:
            logTrack("ModuleST: cost figures not reasonable")
            return (c300, 0, 300.0)

        if c3000 / c300 < c300 / c30:
            logTrack("ModuleST: warning, unrealistic case, not in consistence with correlation")
            c300 = pow(c30 * c3000, 0.5)

    #..............................................................................
    #   Initial estimate:

        C0 = 0.9 * c3000

    #..............................................................................
    #   xi = exp(-300/PRef)
    #   RSMALL = (C30 - C0)/(C300 - C0) = (xi**1/10/xi) = xi**-0.9)
    #   RLARGE = (C3000 - C0)/(C300 - C0) = (xi**9)
    #   -> xi_S = rsmall**(-1/0.9)
    #   -> xi_L = rlarge**(1/9.0)

        alpha = 0.5
        
        for i in range(1000):

            if (c300 > C0):
                rsmall = (c30 - C0) / (c300 - C0)
                xi = pow(rsmall, 1.0 / (0.1 - 1.0))
            else:
                xi = 1.0

            C1 = (c30 - c300) / (pow(xi, 0.1) - xi)

            C0p = c3000 - C1 * pow(xi, 10.0)
            C0p = min(C0p, c3000)

            diff = (C0p - C0) / C0
            C0 += alpha * (C0p - C0)

#            print "C0: %s C1: %s xi: %s diff: %s "%(C0,C1,xi,diff)
            if diff < 1.e-6:
                break

        Pref = -300.0 / log(xi)
#        print "C0: %s C1: %s Pref: %s "%(C0,C1,Pref)

#        print "C30 = %s"%(C0+C1*exp(-30.0/Pref))
#        print "C300 = %s"%(C0+C1*exp(-300.0/Pref))
#        print "C3000 = %s"%(C0+C1*exp(-3000.0/Pref))

        self.FixCost = C0
        self.AddCost = C1
        self.Pref = Pref
        
        return (C0, C1, Pref)

def equipmentDefaults():
    """
    default parameters set by manual entry of equipment (PanelQ4)
    """
    
    return {"STAreaFactor": 1.1}
    
#============================================================================== 
class ModuleST(object):
#============================================================================== 
#============================================================================== 
#============================================================================== 

    STList = []
    
#------------------------------------------------------------------------------
    def __init__(self, keys=["ST Table"]):
#------------------------------------------------------------------------------
#   init is called only once at the very beginning of the EINSTEIN session
#   at this moment no project data are available yet ...
#------------------------------------------------------------------------------
        self.keys = keys # the key to the data is sent by the panel
        self.atmosfericFlag = 0
        self.surfaceFlag = 0
        self.avAreaFactor = 1.1
        self.collType = None

        self.eq = None
        self.eqSQL = None
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def initPanel(self):
#------------------------------------------------------------------------------
#   initPanel is called when the user opens the DA panel
#   all the calculations should be carried out here that are necessary only
#   the first time. the rest should be done in "updatePanel"
#------------------------------------------------------------------------------
        self.manualChoosenColl = None
        self.selectedCollID = None
        self.collType = None
        self.avAreaFactor = 1.1
##        self.generalData()
#        self.screenSurfaces()
#        self.calcSurfArea()
#        (self.inclination,self.azimuth,self.shLoss)=self.defineCalcParams()
        self.guiSysPars = [0, DEFAULT_SYSTEM_EFFICIENCY, 0]
        self.TavCollMean = 0
        if Status.int.cascadeUpdateLevel < 0:
            Status.int.initCascadeArrays(0)
        self.updatePanel()
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def updatePanel(self):
#------------------------------------------------------------------------------
#   Here all the information should be prepared so that it can be plotted on the panel
#   This function will be called from the GUI every time some information to
#   be displayed might have changed
#------------------------------------------------------------------------------

        (STList, collectorList) = self.screenEquipments()

        self.generalData()
        self.screenSurfaces()
        self.calcSurfArea()
        (self.inclination, self.azimuth, self.shLoss) = self.defineCalcParams()
        
        if Status.int.cascadeUpdateLevel < self.cascadeIndex or Status.mod.moduleEnergy.outputLevel < 3:
            Status.mod.moduleEnergy.runSimulation(last=self.cascadeIndex, outputLevel=3)
#............................................................................................
# 1. List of equipments

        if len(collectorList) == 0 and self.selectedCollID is not None:
            data = np.array([])
            if self.selectedCollID is not None:
                model = dbio.get("dbsolarthermal", id=self.selectedCollID)
                if model is not None:
                    data = np.array(noneFilter([[model.STModel,
                                             model.STType,
                                             model.STc0,
                                             model.STc1,
                                             model.STc2,
                                             model.K50L,
                                             model.K50T]]))
                    
        else:
            data = np.array(collectorList)

        Status.int.setGraphicsData('ST Table', data)

#............................................................................................
# 2. Preparing data

        Tmax = 200.0
        iTmax = int(Tmax / Status.TemperatureInterval + 0.5)
        index = self.findIndex()
        
        if index > 0:
            USHRequired_T = Status.int.calcQ_T(Status.int.QD_Tt_mod[index])
        else:
            USHRequired_T = Status.int.USHTotal_T

        QD_plot = USHRequired_T[0:iTmax] / 1000.0

        USH_plot = []
                           
        if len(STList) > 0:
            if Status.SimTool == "Plus":
##                USHj_Tt, USHj = self.calculateEnergyFlows(None, Status.int.QD_Tt_mod[index], index)
                USHj_T = Status.int.calcQ_T(Status.int.USHj_Tt[index])
            else:
                USHj_T = Status.int.USHj_T[index]
                
            USH_plot = USHj_T[0:iTmax] / 1000.0
                
            Status.int.setGraphicsData('ST Plot', [Status.int.T[0:iTmax],
                                                  QD_plot.tolist(),
                                                  USH_plot.tolist()])

            self.eq = Status.int.systemModel.eqObject[0]
            self.eqSQL = Status.int.systemModel.equipment[0]

############ IDAE PLOT
##            dataList = [Status.int.T,Status.int.USHTotal_T,Status.int.USHj_T[index]]
##            printMatrix(dataList,titles=["T","USH","USH_solar"])

        else:                

            Status.int.setGraphicsData('ST Plot', [Status.int.T[0:iTmax],
                                                  QD_plot.tolist()])

            self.eq = None
            self.eqSQL = None
#............................................................................................
# 3. Configuration design assistant

        config = noneFilter(self.getUserDefinedPars())
        Status.int.setGraphicsData('ST Config', config)

    
#............................................................................................
# 4. additional information

        maxCapacity = self.netCollArea * M2TOKW                    ####E.F. 28/07
        if len(STList) > 0:
            yi = USHj_T[iTmax] / (STList[0][0])

            if USHRequired_T[iTmax] > 0:
                sf = USHj_T[iTmax] * 100 / USHRequired_T[iTmax]
            else:
                sf = 0
                
#            print "ModuleST; updatePanel: the heat supplyed in a year by the solar system is:",Status.int.USHj_T[index][iTmax]
            try: 
                SEff = USHj_T[iTmax] * 100 / (collectorList[0][8] * STList[0][0] / 0.7)
                if np.isnan(SEff) or np.isinf(SEff):
                    SEff = "--"    # system efficiency
            except TypeError:
                SEff = "--"
#            print "ModuleST; updatePanel: collectorList[0][8] is:",collectorList[0][8]
            # set division by zero results to sensible values
            if np.isnan(yi) or np.isinf(yi):
                yi = 0
        else:
            yi = 0
            sf = 0
            SEff = "--"

        if self.TavCollMean == 0:
            TC = "--"
        else:
            if len(STList) > 0:
                TC = self.TavCollMean
            else:
                TC = "--"
        
        info = [self.usableSurfacesArea, maxCapacity, sf, yi, SEff, TC, 0]
        
        if self.NSurfaces == 0:
            info[0] = "---"
            info[1] = "---"
            
        Status.int.setGraphicsData('ST Info', info)

#............................................................................................
# 5. System parameters to GUI


        if len(STList) > 0:
            sysPars = noneFilter(STList[0])
        else:
            sysPars = self.guiSysPars
            
        Status.int.setGraphicsData('ST SysPars', sysPars)

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def getUserDefinedPars(self):
        """Get the user defined data from the (misnamed) uheatpump table
        and return it or sensible default data instead if there is none.
        
        :returns: A list [STSolFra, STCollType, STMinYield]
        
        """
        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        if urow is None:
            dummy = {"Questionnaire_id":Status.PId,"AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)

            config = [50, STTYPES[0], 400.0]            
            Status.int.setGraphicsData('ST Config', config)

            self.setUserDefinedPars()

        else:
            u = urow

            # repair corrupt entries in uheatpump (None-settings)
            if u.STSolFra is None: u.STSolFra = DEFAULTSAUTODESIGNST["STSolFra"]
            if u.STCollType is None: u.STCollType = DEFAULTSAUTODESIGNST["STCollType"]
            if u.STMinYield is None: u.STMinYield = DEFAULTSAUTODESIGNST["STMinYield"]
            dbio.commit()
            
            config = noneFilter([u.STSolFra, #H.S. 01/08
                      u.STCollType, #H.S. 01/08
                      u.STMinYield])        #H.S. 01/08
            config[0] *= 100.0          #conversion to %        #H.S. 01/08

        return config

#------------------------------------------------------------------------------
    def setUserDefinedPars(self):
#------------------------------------------------------------------------------

        config = Status.int.GData['ST Config']

        urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)

        if urow is None:
            logTrack("ModuleST(setUserDefinedParamHP): corrupt data base - no entry for uheatpump under current ANo")
            
            dummy = {"Questionnaire_id":Status.PId, "AlternativeProposalNo":Status.ANo} 
            dbio.insertRow("uheatpump", **dummy)
            urow = dbio.get("uheatpump", Questionnaire_id=Status.PId, AlternativeProposalNo=Status.ANo)
            
        u = urow

        try:
            u.STSolFra = float(config[0]) / 100.0
        except:
            u.STSolFra = 0.5
            
        u.STCollType = config[1]
        u.STMinYield = toFloat(config[2])

        dbio.commit()


#------------------------------------------------------------------------------
    def screenEquipments(self):
#------------------------------------------------------------------------------
#       screens existing equipment, whether there are already heat pumps
#------------------------------------------------------------------------------

        Status.int.systemModel.build()
        equipments = Status.int.systemModel.equipment
        
        self.NEquipe = len(equipments)
        STList = []
        collectorList = []

        self.equipeIDs = []
        self.cascadeIndex = 0
        for equipe in equipments:
#            print "Equipment: ",equipe.EquipType,getEquipmentClass(equipe.EquipType)
            if getEquipmentClass(equipe.EquipType) == "ST":
                collectorTable = [equipe.Model,
                                  self.collType,
                                  equipe.ST_C0,
                                  equipe.ST_C1,
                                  equipe.ST_C2,
                                  equipe.ST_K50L,
                                  equipe.ST_K50T,
                                  equipe.STAreaFactor, ##E.F. 28/07
                                  equipe.HTilted]  ##E.F. 28/07
                systemTable = [equipe.HCGPnom,
                               equipe.ST_SysEff,
                               equipe.ST_Volume]
                self.equipeIDs.append(equipe.QGenerationHC_ID)
                if equipe.EquipIDFromDB is not None:
                    self.selectedCollID = equipe.EquipIDFromDB # none-check avoids overwriting by addEquipmentDummy !!

                self.collType = getEquipmentSubClass(equipe.EquipType)
                self.avAreaFactor = equipe.STAreaFactor
                
                STList.append(systemTable)
                collectorList.append(collectorTable)
                self.cascadeIndex = equipe.CascadeIndex
        if len(STList) > 1:
            logMessage(_("pay attenction more than one solar system"))
        return (STList, collectorList)
        

        
#------------------------------------------------------------------------------
    def getEqId(self, rowNo):
#------------------------------------------------------------------------------
#   gets the EqId from the rowNo in the STList
#------------------------------------------------------------------------------

        STId = self.equipeIDs[rowNo]
        return STId

#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
    def deleteEquipment(self, rowNo, automatic=False):
#------------------------------------------------------------------------------
# deletes the selected boiler in the current alternative
#------------------------------------------------------------------------------

        if automatic == False:
            if rowNo == None:   #indicates to delete last added equipment dummy
                if self.dummyEqId is None:
                    return
                else:
                    STid = self.dummyEqId
            else:
                STid = self.getEqId(rowNo)
                logTrack("Module ST (delete): id to be deleted = %s" % STid)
        else:
            STid = rowNo
            logTrack("Module ST (delete automaticly): id to be deleted = %s" % STid)
        
        Status.prj.deleteEquipment(STid)
        Status.prj.setStatus("energy", 0)   # the accumulated energy 
                                            # statistics are out of date
        self.screenEquipments()
        self.guiSysPars = [0, 0, 0]
        self.updatePanel()
        self.avAreaFactor = 1.1
        
        self.manualChoosenColl = None   ####E.F. 28/07
        self.selectedCollID = None
        self.collType = None

        Status.int.systemModel.changed()
        Status.int.systemModel.build()

        self.eq = None
        self.eqSQL = None
        

    def addEquipmentDummy(self):
        """Add a new dummy equipment to the equipment list and return 
        the position in the cascade.
        
        """
        (STList, collectorList) = self.screenEquipments()
        if len(STList) > 0:
            self.dummyEqId = None
            eq = dbio.get("qgenerationhc",
                          Questionnaire_id=Status.PId,
                          AlternativeProposalNo=Status.ANo,
                          CascadeIndex=self.cascadeIndex)
            
            if eq is not None:
                self.equipe = eq
                showWarning(_("You are changing the current solar system."))
                return self.equipe

            else:
                logDebug("ModuleST (addEquipmentDummy): fatal error in STList!!!!!")

        self.equipe = Status.prj.addEquipmentDummy()
        self.dummyEqId = self.equipe.QGenerationHC_ID
        if self.equipe.CascadeIndex > 1:
            Status.mod.moduleHC.cascadeMoveToTop(self.equipe.CascadeIndex) 
        
        self.equipe = Status.prj.getEquipe(self.dummyEqId) #after having moved, the equipe - table has to be updated !!!
        self.cascadeIndex = self.equipe.CascadeIndex

        NewEquipmentName = "Solar thermal system"
        # connect the dummy to all heating pipes
        dummy, pipeIDs = getAllHeatingPipes()
        pipeIDs = [str(id) for id in pipeIDs ]
        IDString = ";".join(pipeIDs)
        
        equipeData = {"Equipment":NewEquipmentName,
                      "EquipType":"solar thermal (flat-plate)", # by default assign any valid equipe type.
                                                                # should be updated in setEquipmentFromDB
                      "PipeDuctEquip":IDString}
        self.equipe.set(**equipeData)
        dbio.commit()
        
        self.screenEquipments()

        Status.int.systemModel.changed()
        Status.int.systemModel.build()

        self.eq = Status.int.systemModel.eqObject[0]
        self.eqSQL = self.equipe

        return(self.equipe)

#------------------------------------------------------------------------------
    def setEquipmentFromDB(self, equipe, modelID):
#------------------------------------------------------------------------------
#   takes an equipment from the data base and stores it under a given Id in
#   the equipment data base
#------------------------------------------------------------------------------

        model = dbio.get("dbsolarthermal", id=modelID)
        if model is not None:
            self.selectedCollID = modelID
            self.collType = model.STType
        else:
            logTrack("ModuleST: selected model ID %s not in database" % modelID)
            return
            
        if model.STManufacturer != None: equipe.Manufact=model.STManufacturer
        if model.STModel != None: equipe.Model=model.STModel
        logTrack("ModuleST: setting equipment type from %s" % model.STType)
        equipe.EquipType=getEquipmentType("ST", model.STType)
        equipe.NumEquipUnits=1
        
        if model.id != None: equipe.EquipIDFromDB=model.id
        if model.STType != None: equipe.EquipTypeFromDB=str(model.STType)
        
        if model.STc0 != None: equipe.ST_C0=model.STc0
        if model.STc1 != None: equipe.ST_C1=model.STc1
        if model.STc2 != None: equipe.ST_C2=model.STc2

        if model.K50L != None: equipe.ST_K50L=model.K50L
        else:
            logTrack("No longitudinal incidence angle modifier defined for collector ID %s. 1.1 assumed" % modelID)
            equipe.ST_K50L=0.95
            
        if model.K50T != None: equipe.ST_K50T=model.K50T
        else:
            logTrack("No transversal incidence angle modifier defined for collector ID %s. 1.1 assumed" % modelID)
            equipe.ST_K50T=0.95

        if model.STAreaFactor != None: equipe.STAreaFactor=model.STAreaFactor
        else:
            logTrack("No surface area factor defined for collector ID %s. 1.1 assumed" % modelID)
            equipe.STAreaFactor=1.1

        equipe.FuelConsum = 0.0
        equipe.FlowExhaustGas = 0.0
        equipe.TExhaustGas = 0.0

        if model.STOMUnitFix is not None and equipe.HCGPnom is not None:
            OMFix = model.STOMUnitFix * equipe.HCGPnom
            equipe.OandMfix=OMFix

        self.calculateCostParameters(equipe, modelID)
        (turnKey, OMFix, OMVar) = self.estimateEconomicParameters(equipe)
        equipe.set(**{"TurnKeyPrice": turnKey,
                       "OandMfix":OMFix,
                       "OandMvar":OMVar})
        
                
        dbio.commit()

        Status.int.systemModel.updateEquipmentData()
        
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def setSolarSystemPars(self, verbose=False):
#------------------------------------------------------------------------------
        
        sysPars = Status.int.GData['ST SysPars']

        (STList, collectorList) = self.screenEquipments()

        createNewSSY = False
        
        if len (STList) > 0:
            changeInEq = False

            eq = Status.prj.getEquipe(cascadeIndex=self.cascadeIndex)
            if eq is not None:
                equipe = eq
                
                if equipe.HCGPnom is not None:
                    if fabs(float(sysPars[0]) - equipe.HCGPnom) > 1.e-6:
                        changeInEq = True
                else:
                    changeInEq = True
                
                equipe.HCGPnom = float(sysPars[0])

                if equipe.HCGPnom is not None:
                    equipe.ElectriConsum = PARASITIC_ELECTRICITY * equipe.HCGPnom

                if equipe.ST_SysEff is not None:
                    if fabs(float(sysPars[1]) - equipe.ST_SysEff) > 1.e-6:
                        changeInEq = True
                else:
                    changeInEq = True

                equipe.ST_SysEff = float(sysPars[1])       
                if equipe.ST_SysEff is None or not (MIN_SYSTEM_EFFICIENCY <= equipe.ST_SysEff <= 1.0):
                    logWarning(_("Invalid specification of system efficiency (value should be from 0.5 - 1.0). eta = %s assumed") % DEFAULT_SYSTEM_EFFICIENCY)
                    equipe.ST_SysEff = DEFAULT_SYSTEM_EFFICIENCY

                if equipe.ST_Volume is not None:
                    if fabs(float(sysPars[2]) - equipe.ST_Volume) > 1.e-6:
                        changeInEq = True
                else:
                    changeInEq = True

                equipe.ST_Volume = float(sysPars[2])

                dbio.commit()

            else:
                logDebug("ModuleST (setSolarSystemPars): STList > 0 but equipment not found")
                
            if changeInEq == True: 
                Status.int.changeInCascade(self.cascadeIndex)

            self.calculateCostParameters(equipe, self.selectedCollID)
                
#        elif self.selectedCollID is not None:
#            createNewSSY = True
#            print "ModuleST (setSolarSystemPars): mode 2 selectedCollID = ",self.selectedCollID
#            equipe = self.addEquipmentDummy()
#            self.setEquipmentFromDB(equipe,self.selectedCollID)
#            self.setSolarSystemPars()
            
        else:
            try:
                self.guiSysPars[0] = float(sysPars[0])
            except:
                self.guiSysPars[0] = None
                pass
                
            try:
                self.guiSysPars[1] = float(sysPars[1])
            except:
                self.guiSysPars[1] = None
                
            try:
                self.guiSysPars[2] = float(sysPars[2])
            except:
                self.guiSysPars[2] = None
                

            if self.guiSysPars[0] > 0:
                createNewSSY = True

                if self.NSurfaces <= 0 and verbose:
                    logWarning(_("No information on available surfaces is given !\n") + \
                               _("Is is assumed that there is no constraint due to limited space"))
            
        if createNewSSY and self.selectedCollID is None:
            showMessage(_("Please select a solar collector"))

        return createNewSSY
#------------------------------------------------------------------------------
    def setManualSelection(self, theId):

        previous = self.selectedCollID

        if theId != previous:
            self.manualChoosenColl = dbio.get("dbsolarthermal", id=theId)
            self.selectedCollID = theId
            self.collType = self.manualChoosenColl.STType

            self.avAreaFactor = self.manualChoosenColl.STAreaFactor
            self.calcSurfArea()

            # set flags that system has changed and call for update of panel
            if self.eqSQL is not None and self.eqSQL.HCGPnom is not None and self.eqSQL.HCGPnom > 0:
                Status.int.changeInCascade(self.cascadeIndex)

            self.updatePanel()
            
    def calculateEnergyFlows(self, equipe, demand_Tt, cascadeIndex):
        """Calculate the energy flows in the solar thermal module 
        identified by "cascadeIndex" and update the corresponding 
        global data.
        
        :returns: a pair consisting of the time and temperature 
        dependent numpy array containing the covered demand. Also known
        as USHj_Tt of the given equipment and the same data aggregated 
        over time and temperature USHj.
        
        """        
        logTrack("ModuleST (calculateEnergyFlows): starting (cascade no: %s)" % cascadeIndex)

        if Status.int.cascadeUpdateLevel < (cascadeIndex - 1) or Status.mod.moduleEnergy.outputLevel < 3:
            logDebug("ModuleST (calculateEnergyFlows): cannot calculate without previously updating the previous levels")
            Status.mod.moduleEnergy.runSimulation(last=(cascadeIndex - 1), outputLevel = 3)
        Status.int.extendCascadeArrays(cascadeIndex)

        if Status.SimTool == "Plus":
            USHj_Tt = Status.int.createQ_Tt()

            USHj = Status.mod.moduleEnergy.sim.systemModel.simulateEquipment(USHj_Tt, demand_Tt, cascadeIndex-1, equipe=equipe)
            self.TavCollMean = Status.mod.moduleEnergy.sim.systemModel.equipment[cascadeIndex-1].resume()
                
            return USHj_Tt, USHj

        else:
            # get equipment data from equipment list in SQL
            self.screenEquipments()

            STType = equipe.EquipType
            collType = getEquipmentSubClass(STType)
            PNom = equipe.HCGPnom

            if PNom is None or PNom == 0:   #security feature
                logDebug("ModuleST (calcEn.Flows): equipe with PNom = None or PNom = 0 detected\nCan't do anything with this !!!")
                return (Status.int.createQ_Tt(), 0)
                
            #XXX TOpMax or something similar should be defined in SQL        TMax = equipe.TOpMax
            Tmax = 200

            controlSchedule = Status.int.systemModel.eqObject[cascadeIndex-1].controlSchedule
            control = controlSchedule.schedule if controlSchedule else 1.0

            # get demand data for CascadeIndex/EquipmentNo from Interfaces
            # and create arrays for storing heat flow in equipment
            QD_Tt = demand_Tt.copy()
            QA_Tt = Status.int.QA_Tt_mod[cascadeIndex - 1].copy()

            QD_T = Status.int.calcQ_T(QD_Tt)
            if DEBUG in ("ALL", "BASIC"):
                print "ModuleST (cEF): QD_T"
                print QD_T

            USHj_Tt = Status.int.createQ_Tt()
            QHXj_Tt = Status.int.createQ_Tt()
            QWHj_Tt = Status.int.createQ_Tt()

            # Start hourly loop
            dlg = DialogGauge(Status.main, _("EINSTEIN system simulation"), _("simulating %s") % (_("solar system")))

            # update surface calculations if necessary
            if self.atmosfericFlag != 1:
                self.generalData()
            if self.surfaceFlag != 1:
                self.screenSurfaces()
                self.calcSurfArea()
                (self.inclination, self.azimuth, self.shLoss) = self.defineCalcParams()

            I = self.ST_IDefault

            # Solar system parameters
            SurfAreaSol = PNom / 0.7
            if equipe.ST_SysEff != None and MIN_SYSTEM_EFFICIENCY <= equipe.ST_SysEff <= 1.0:
                systemEfficiency = equipe.ST_SysEff
            else:
                logDebug(_("ModuleST (cEF): invalid specification of system efficiency (should be from 0.5 - 1.0). eta = %s assumed") % DEFAULT_SYSTEM_EFFICIENCY)
                systemEfficiency = DEFAULT_SYSTEM_EFFICIENCY

            # Solar collector parameters
            ST_C0 = equipe.ST_C0
            ST_C1 = equipe.ST_C1
            ST_C2 = equipe.ST_C2
            
            if equipe.ST_K50T > 0 and equipe.ST_K50T is not None: ST_K50T = equipe.ST_K50T
            else:
                logTrack("ModuleST (cEF): invalid specification of KT. KT = 0.95 assumed")
                ST_K50T = 0.95
            
            if equipe.ST_K50L > 0 and equipe.ST_K50L is not None: ST_K50L = equipe.ST_K50L
            else:
                logTrack("ModuleST (cEF): invalid specification of KT. KT = 0.95 assumed")
                ST_K50L = 0.95

            if collType == "Fresnel collector":
                tracking = 0
                diffuse = 0.0
            elif collType == "Parabolic trough collector":
                tracking = 1
                diffuse = 0.0
            else:
                tracking = 0
                diffuse = 1.0

            # Parameters for storage modelling
            TStorage = 60.0                 #initial temperature of the storage at 1st of January 0:00
                                            #HS2008-08-02. initial temperature changed to 60 �C
            TStorageMax = 250.0             #maximum allowed temperature in the storage
            
            if equipe.ST_Volume is not None:
                if equipe.ST_Volume < 0.01 * equipe.HCGPnom:
                    storageVolume = 0.01 * equipe.HCGPnom
                    logWarning(_("The heat storage volume is very small. For internal calculation it has been set to 0.01 m3/kW"))
                else:
                    storageVolume = equipe.ST_Volume
            else:
                logTrack("ModuleST (cEF): no storage Volume specified. Specific storage size of 0.05 m3/m2 assumed")
                storageVolume = 0.05 * SurfAreaSol       #in m3. for the moment fixed to 50 l/m2. can be an equipment parameter
                equipe.ST_Volume = storageVolume
            	dbio.commit()

            CStorage = 1.16 * storageVolume   #heat capacity of the storage in kWh/K
            QStorage = CStorage * TStorage    #total heat stored in the tank (reference temperature = 0�C)
            QStorageMax = CStorage * TStorageMax #maximum amount of heat that can be stored in the tank (ref. = 0�C)

            storageLoss = min(0.0005 * (10. * pow(storageVolume, 2.0 / 3.0)) / CStorage, 1.0 / Status.TimeStep)
                                            # A = 10 * V^2/3; U = 0.5 W/m2K converted to kW/m2K; 
            QStorageEnv = CStorage * 15.0     # 15.0 C ambient temperature of Storage for calculation of losses.
                                            # -> dotQStorageLoss = storageLoss*(QStorage - QStorageEnv)
                     
    #.............................................................................
    # HERE THE CYCLE OF CALCULATE ENERGY FLOWS STARTS
    #.............................................................................
            TIMESTEP = Status.TimeStep
            NT = Status.NT

            # setting counters and initial value of integrals (sums) to zero
            oldDay = 0

            QavColl = 0     #these two values only for calculating the weighted mean of the average collector
            QTavColl = 0    #temperature during the year

            USHj = 0
            HPerYear = 0
            annualDemand = 0

            self.location = Location(self.latitude, 2.67, 1.0)
            
            if Status.WeatherData.name == "TMY2File":
                wData_G = Status.WeatherData.getGlobalHorizontalRadiation()
                wData_Gd = Status.WeatherData.getDiffuseHorizontalRadiation()
                wData_TAmb = Status.WeatherData.getDryBulbTemperature()

                longitude = Status.WeatherData.getLongitude()    
                timeZone = 1.0
                
            else:
                longitude = 0.0
                timeZone = 1.0

            self.location = Location(self.latitude, longitude, timeZone)

            annualGbT = 0
            annualGdT = 0
            HTilted = 0
            
            for it in range(Status.Nt):
                fctrl = 1.0 if isinstance(control, float) else control[it]

                time_in_h = it * TIMESTEP    # start time of time interval !!!
                it_wData = int(time_in_h + 0.5)

                time_in_s = time_in_h * HOUR
                time_in_s1 = time_in_s + TIMESTEP * HOUR # end time of time interval
                nday = int(ceil(time_in_h / 24.0))
    #            print "DAY: %s Hour: %s Sec: %s"%(nday,time_in_h,time_in_s)

                # recalculate solar data for a new day

                #   The function PREP_SUN prepares the radiation calculations for a given day
                #   in the year
                if nday > oldDay:
                    oldDay = nday
                    
                #HS2008-07-10: Multiplication with EXTRAPOLATE_TO_YEAR added. so a full solar year will be
                # simulated also if calculation time is less than 8760!
                    nday_solar = int(nday * Status.EXTRAPOLATE_TO_YEAR + 0.1)
                    sunOnH.t = getDailyRadiation(I, nday_solar, self.latitude)

                    prep_sun(self.location,
                             self.inclination,
                             self.azimuth,
                             nday_solar,
                             sunAnglesOnSurface,
                             sunOnH,
                             tracking)

                    if DEBUG in ["ALL", "BASIC"]:
                        print "NEW DAY: %s H = %s Hb %s Hd %s I %s" % (nday_solar, sunOnH.t, sunOnH.b, sunOnH.d, I)
                        print "lat %s inc %s azi %s" % (self.latitude, self.inclination, self.azimuth)

                # now get instantaneous radiation (in W/m2)
                if Status.WeatherData.name=="TMY2File":
                    G = wData_G[it_wData] * 1000
                    Gd = wData_Gd[it_wData] * 1000

                    GT_W = sun_hourly_WDATA(sunAnglesOnSurface, G, Gd, time_in_s, time_in_s1)  #GT_W: radiation in [W] vs. GT in [kW]!!!
                    TAmb_t = wData_TAmb[it_wData]
                else:
                    GT_W = sun_hourly(sunAnglesOnSurface, sunOnH, time_in_s)  #GT_W: radiation in [W] vs. GT in [kW]!!!
                TAmb_t = Status.WeatherData.TAmb[it_wData]
                                
                GbT = GT_W.b / 1000.    #conversion to kW/m2
                GdT = GT_W.d / 1000.
                GT = GT_W.t / 1000.

                if DEBUG in ["ALL"]:
                    print "ModuleST.cEF: GT %s GbT %s GdT %s"%(GT,GbT,GdT)

                # calculation of tan(incidence_angle). needed later on for efficiency calculation.
                if GT_W.n[2] <= 0:
                    tanThetaT = INFINITE
                    tanThetaL = INFINITE
                else:
                    tanThetaT = fabs(GT_W.n[0] / GT_W.n[2])
                    tanThetaL = fabs(GT_W.n[1] / GT_W.n[2])

                if DEBUG in ["ALL"]:
                    print "ModuleST (cEF): dirSunOnT = %s" % GT_W.n
                    
                annualGbT += GbT * TIMESTEP
                annualGdT += GdT * TIMESTEP

                # calculate maximum solar system output, heat demand and storage capacity

                #   The working temperature of the solar system is the temperature
                #   of the storage tank at the beginning of the time-step plus a Delta_T
                #   accounting for the heat exchanger in the primary loop
                deltaT_primary = 7.0
                TStorage = QStorage / CStorage
                TavCollector = TStorage + deltaT_primary    #=working temperature of the collector in the present time step
                
                DeltaT = TavCollector - TAmb_t

                if collType == "Fresnel collector":
                    eff = collectorEfficiencyFresnel(GT_W.b, GT_W.d * diffuse, tanThetaL, tanThetaT, DeltaT, ST_C0, ST_C1, ST_C2, ST_K50L, ST_K50T)
                else:
                    eff = collectorEfficiencyBiaxial(GT_W.b, GT_W.d * diffuse, tanThetaL, tanThetaT, DeltaT, ST_C0, ST_C1, ST_C2, ST_K50L, ST_K50T)
                            
                if DEBUG in ["ALL"]:
                    print"ModuleST: tanThetaL %s tanThetaT %s eff %s" % (tanThetaL, tanThetaT, eff)

                dotQuSolarMax = eff * systemEfficiency * SurfAreaSol * (GbT + GdT * diffuse) * self.shLoss
                                                        #*shadingFactor!!!  if shaded then add shading factor.ADD an IF!
                dotQuSolarMax = max(dotQuSolarMax, 0)    #solar system regulation controls that there's no negative Qu !!!

                # determination of minimum and maximum demand and corresponding temperature levels
                # the minimum temperature of demand indicates to what temperature the storage can be cooled down.
                # if we have heat demand only above e.g. 60�C, we cannot cool down the storage lower than this,
                # as heat can not be transferred from cold to hot (at least not yet, maybe EINSTEIN succeeds to do it)
                DTStratification = 5
                Tmax_it = min(Tmax, TStorage + DTStratification)    #in the present time step, the maximum temperature
                                                #that can be supplied corresponds to the storage temperature

                iTmax = (int(floor(Tmax_it / Status.TemperatureInterval)))
            	dotQDemand = QD_Tt[iTmax][it] * fctrl / Status.TimeStep

                iTmin = NT + 1
                for iT in range(0, NT + 1):
                    if QD_Tt[iT][it] > 0:
                        iTmin = iT
                        break

                TMinDemand = max(iTmin - 1, 0) * Status.TemperatureInterval

                # calculate actual solar system production
                QStorageCapacity = max(QStorageMax - QStorage, 0.0)   #maximum amount of heat that still can be fed into the storage
                dotQStorageLoss = storageLoss * (QStorage - QStorageEnv)
                
                dotQuSolar = min(dotQDemand + QStorageCapacity / TIMESTEP + dotQStorageLoss, dotQuSolarMax)
                                                            #solar production constrained by demand + remaining storage capacity

                # calculate actual heat delivery to demand
                QStorageMin = CStorage * (TMinDemand + DTStratification)           #minimum heat that has to remain within the storage, as it can't be cooled down further
                dotQuSupply = min(dotQDemand, dotQuSolar + max((QStorage - QStorageMin) / TIMESTEP - dotQStorageLoss, 0))
                                                            #supply constrained by solar production + available heat stored at T > TMinDemand
    #            try:
    #                theta = (180./pi)*acos(GT_W.n[2])
    #            except:
    #                theta = 99.9
    #
    #            try:
    #                gamma =  (180./pi)*atan(GT_W.n[0]/GT_W.n[1])
    #            except:
    #                gamma = -180.0
    #                
    #            SolarData.append([time_in_h,it_wData,GT,GbT,GdT,G,Gd,dotQuSupply,dotQuSolar,dotQDemand,TStorage,dotQuSolarMax,TMinDemand, \
    #                              theta,gamma,tanThetaL,tanThetaT])

                if DEBUG in ["ALL", "BASIC"]:
                    print"ModuleST (calculateEnergyFlows): time: %s eff: %s DT: %s GT: %s dotQu: %s dotQuSolar: %s Tamb: %s" % \
                                    (time_in_h, eff, DeltaT, GT, dotQuSupply, dotQuSolar, self.TAmb)

                # now update heat stored in tank
                QStorage += (dotQuSolar - dotQStorageLoss - dotQuSupply) * TIMESTEP

                if dotQuSolar > 0:
    #                print "ModuleST (cEF): it %s dotQuSolar %s TavColl %s TStorage %s TMinDemand %s "%\
    #                      (it,dotQuSolar,TavCollector,TStorage,TMinDemand)
                    HPerYear += Status.TimeStep
                    
                USHj_t = dotQuSupply * TIMESTEP
                annualDemand += dotQDemand * TIMESTEP    

                QavColl += dotQuSolar
                QTavColl += dotQuSolar * TavCollector

                USHj += USHj_t
                HTilted += GT * TIMESTEP

                # now the remaining problem is to decide at WHAT temperature the solar system does supply heat (from USHj_t to USHj_Tt)
                # -> let's use the simplest possible solution (although it's not the most efficient one ...):
                # fill up the demand from below
                USHj_Tt[:, it] = QD_Tt[:, it].clip(max=USHj_t)
                QD_Tt[:, it] = QD_Tt[:, it] - USHj_Tt[:, it]

                if it % 200 == 0:
                    dlg.update(100.0 * it / Status.Nt)
            #.............................................................................
            # end of the year reached. now some final calculations
            dlg.Destroy()
            self.TavCollMean = QTavColl / max(QavColl, 0.000000001)

            #  Store results in interfaces
            #        for row in SolarData:
            #            print "%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|"%\
            #            (row[0],row[1],row[2],row[3],row[4],row[5],row[6],row[7],row[8],row[9],row[10],row[11],row[12],row[13],row[14],row[15],row[16])
            # remaining heat demand and availability for next equipment in cascade
            Status.int.QA_Tt_mod[cascadeIndex] = QA_Tt
            Status.int.QA_T_mod[cascadeIndex] = Status.int.calcQ_T(QA_Tt)

            # heat delivered by present equipment
            Status.int.USHj_Tt[cascadeIndex - 1] = USHj_Tt
            Status.int.USHj_T[cascadeIndex - 1] = Status.int.calcQ_T(USHj_Tt)
            Status.int.USHj_t[cascadeIndex - 1] = copy.deepcopy(USHj_Tt[Status.NT + 1])
            Status.int.USHjChanged[cascadeIndex - 1] = True

            # waste heat absorbed by present equipment
            Status.int.QHXj_Tt[cascadeIndex - 1] = QHXj_Tt
            Status.int.QHXj_T[cascadeIndex - 1] = Status.int.calcQ_T(QHXj_Tt)
            Status.int.QHXj_t[cascadeIndex - 1] = copy.deepcopy(QHXj_Tt[Status.NT + 1])

            Status.int.QHXEqReq_Tt[cascadeIndex - 1] = HeatFlow_t()
            Status.int.QHXEqReqChanged[cascadeIndex - 1] = False #no waste heat used

            # waste heat produced by present equipment
            Status.int.QWHj_Tt[cascadeIndex - 1] = QWHj_Tt
            Status.int.QWHj_T[cascadeIndex - 1] = Status.int.calcQ_T(QWHj_Tt)
            Status.int.QWHj_t[cascadeIndex - 1] = copy.deepcopy(QWHj_Tt[Status.NT + 1])
            Status.int.QWHjChanged[cascadeIndex - 1] = False #no waste heat generated

            Status.int.QWHEq_Tt[cascadeIndex - 1] = HeatFlow_t()

            # Indicate that present level of cascade has been updated
            Status.int.cascadeUpdateLevel = cascadeIndex

            # Global results (annual energy flows)
            HTilted *= Status.EXTRAPOLATE_TO_YEAR
            annualGbT *= Status.EXTRAPOLATE_TO_YEAR
            annualGdT *= Status.EXTRAPOLATE_TO_YEAR
            
            equipe.HTilted=dbio.check(HTilted)                                  #### E.F. 28/07
            dbio.commit()                                                 #### E.F. 28/07

            USHj *= Status.EXTRAPOLATE_TO_YEAR
            HPerYear *= Status.EXTRAPOLATE_TO_YEAR
            
            FETFuel_j = 0.0
            Status.int.USHj[cascadeIndex - 1] = USHj
            Status.int.FETFuel_j[cascadeIndex - 1] = FETFuel_j
            Status.int.FETel_j[cascadeIndex - 1] = PARASITIC_ELECTRICITY * USHj
            Status.int.FETel_j_t[cascadeIndex - 1] = PARASITIC_ELECTRICITY * np.array(Status.int.USHj_t[cascadeIndex - 1])
            Status.int.ElGen_j[cascadeIndex - 1] = 0.0
            Status.int.ElGen_t[cascadeIndex - 1] = Status.int.createQ_t()
            Status.int.HPerYearEq[cascadeIndex - 1] = HPerYear
            
            Status.int.QWHj[cascadeIndex - 1] = 0.0   # not considering the latent heat(condensing water)
            Status.int.QHXj[cascadeIndex - 1] = 0.0

            if DEBUG in ["ALL", "BASIC"]:
                print "ModuleST (cEF): USH: %s HT: %s HbT: %s HdT: %s" % \
                      (USHj,
                       HTilted,
                       annualGbT,
                       annualGdT)

            logTrack("USHj(solar) = %s" % USHj)
            
            return (np.array(USHj_Tt), USHj)    

#==============================================================================
#==============================================================================
#==============================================================================
#   FUNCTIONS FOR SURFACE CALCULATIONS AND WEATHER DATA 
#============================================================================== 
#============================================================================== 
#============================================================================== 
#------------------------------------------------------------------------------ 
    def resetST(self):
#------------------------------------------------------------------------------ 
#   resets the surface and weather flag. for calls outside from moduleST
#   at present reset is called in:
#   - setActiveProject: -> always when a new project is selected
#   - PanelQ7 - OK and DELETE buttons: -> always when the user changes some-
#       thing manually in the surface or weather data
#------------------------------------------------------------------------------ 
        self.atmosfericFlag = 0
        self.surfaceFlag = 0
#------------------------------------------------------------------------------ 
    def generalData(self):
        """ 
        Sets general data for module calculations (H, TAmb, latitude, etc.)
        """
        
        self.latitude = Status.WeatherData.latitude
        self.ST_IDefault = Status.WeatherData.I
        self.TAmb = Status.WeatherData.TAmbAvg
        
        # define selection parameters for surface orientation and inclination
        if self.latitude > MININCLINATION:     # northern hemisphere
            self.goodAzimuthClasses = ["S", "SE", "SW"]
            self.goodAzimuth = 0
            self.badAzimuthClasses = ["N", "NE", "NW"]
            self.badOrientation = _("North")
            self.goodOrientation = _("South")
            
        elif self.latitude < -(MININCLINATION): # southern hemisphere
            self.goodAzimuthClasses = ["N", "NE", "NW"]
            self.goodAzimuth = 180
            self.badAzimuthClasses = ["S", "SE", "SW"]
            self.badOrientation = _("South")
            self.goodOrientation = _("North")
            
        else:                               # close to equator
            self.goodAzimuthClasses = ["S", "SE", "SW", "E", "W", "N", "NE", "NW"]
            self.badAzimuthClasses = []
            
            if self.latitude >= 0:
                self.goodAzimuth = 0
                self.badOrientation = _("North")
                self.goodOrientation = _("South")
            else:
                self.goodAzimuth = 180
                self.badOrientation = _("South")
                self.goodOrientation = _("North")

        # finally set Flag
        self.atmosfericFlag = 1
#------------------------------------------------------------------------------ 
#------------------------------------------------------------------------------ 
    def defineCalcParams(self, verbose=False):
#------------------------------------------------------------------------------ 
#   sets inclination, azimuth and shLoss for calculations of system performance
#------------------------------------------------------------------------------

# set default inclination to optimum for collector type

        if self.collType in STTYPES[2:4]:
            inclinationOpt = 0
        else:
            inclinationOpt = max(abs(self.latitude) - 10, MININCLINATION_FPC)

        if len(self.usableSurfaces) == 1:

# in-roof mounting for sufficiently inclined S - SE - SW surfaces:

            if self.usableSurfaces[0]["inclination"] != None \
               and self.usableSurfaces[0]["inclination"] > MININCLINATION \
               and self.usableSurfaces[0]["azimuthClass"] in self.goodAzimuthClasses:
                
                inclinationOpt = self.usableSurfaces[0]["inclination"]
                inRoof = True
            else:
                inRoof = False

            if self.usableSurfaces[0]["azimuth"] != None:
                azimuthOpt = self.usableSurfaces[0]["azimuth"]

                if not inRoof: #rotate only if not in-roof-mounted
                    i_rot90 = int((azimuthOpt - self.goodAzimuth) / 90 + 0.5) # rotate by 90� (supposing rectangular surfaces)
                    azimuthOpt = azimuthOpt - i_rot90 * 90.                   # so that best orientation is obtained
            else:
                azimuthOpt = self.goodAzimuth
            if self.allShaded == 1:
                shLoss = 0.8
            else:
                shLoss = 1
        else:
            if verbose: logMessage(_("More than one surfaces are defined. Optimum inclination assumed for calculation of solar thermal system"))
            azimuthOpt = self.goodAzimuth
            shLoss = 1

####### CHECK CHECK CHECK
####### Instead of using the optimum azimuth, it would be better to use some average of the used surfaces

        self.surfaceFlag = 1

#        print "defineCalcParams: self.collType = %s inclinationOpt = %s"%(self.collType,inclinationOpt)

        return (inclinationOpt, azimuthOpt, shLoss)
        
            
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def screenSurfaces(self):
#------------------------------------------------------------------------------
#   screens the available surfaces for their aptitude for solar energy use
#------------------------------------------------------------------------------

# screen existing surfaces and add to self.surfaces

        self.qsurfareas = Status.prj.getSurfaces()

        self.surfaces = []
        for qsurfarea in self.qsurfareas:
            if qsurfarea.SurfArea is None:
                logWarning("No size specified for surface area %s. Infinite size assumed." % qsurfarea.SurfAreaName)
                area = INFINITE
            else:
                area = qsurfarea.SurfArea
            self.surfaces.append({"surfAreaName":qsurfarea.SurfAreaName, \
                                  "surfArea":area, \
                                  "inclination":qsurfarea.Inclination, \
                                  "azimuth":qsurfarea.Azimuth, \
                                  "azimuthClass":qsurfarea.AzimuthClass, \
                                  "shading":qsurfarea.Shading, \
                                  "distance":qsurfarea.Distance, \
                                  "roofType":qsurfarea.RoofType, \
                                  "roofStaticLoadCap":qsurfarea.RoofStaticLoadCap, \
                                  "enclBuildGroundSketch":qsurfarea.Sketch, \
                                  "ST_ibT":qsurfarea.ST_IbT, \
                                  "ST_IT":qsurfarea.ST_IT})

#if inclination, azimuth, azimuthClass, shading are None we assume some default value!!!

        for i in range(len(self.surfaces)):
            if self.surfaces[i]["inclination"] == None:
                self.surfaces[i]["inclination"] = 0
                logWarning(_("Surface %s: no inclination specified. 0° assumed") % (i + 1))
            if self.surfaces[i]["azimuth"] == None:
                self.surfaces[i]["azimuth"] = self.goodAzimuth
#                logWarning(_U("Surface %s: no azimuth specified. 0� assumed")%(i+1))
            if self.surfaces[i]["azimuthClass"] == None:
                self.surfaces[i]["azimuthClass"] = self.goodAzimuthClasses[0]
                logWarning(_("Surface %s: no azimuth specified. %s orientation assumed") % (i + 1, _(self.goodOrientation)))
                                  
        self.NSurfaces = len(self.surfaces)
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def grossSurfAreaFactor_FPC (self):
#------------------------------------------------------------------------------
#   calculates the minimum row separation for FPC and ETC collector arrays
#------------------------------------------------------------------------------

        inclination = max(abs(self.latitude) - 10., MININCLINATION_FPC)
        declination = 23.45

        cosb = cos(inclination * pi / 180.0)
        sinb = sin(inclination * pi / 180.0)
        
        cosd = cos(declination * pi / 180.0)
        sind = sin(declination * pi / 180.0)

        fsep = 0.9                      # fsep = 1 - 1/nrows; nrows = 10
                                        # gives a quadratic field for 2m high collectors
                                        # and approximately 1000 m2

##        print "cosb %s sinb %s tand %s pi %s" % (cosb, sinb, sind / cosd, pi)
        
        f = (cosb + sinb * cosd / sind * fsep)

##        print inclination, f
        return(f)

#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def calcGrossSurfAreaFactor (self, i):

        """
        Calculates the ratio of gross surface area to net effective collector area
        for a given collector type (self.collType)

        :params i: index of the surface (via self.surfaces defines inlcination and orientation of surface)
        

        """

# 1. special collector types

        if self.collType == "Parabolic trough collector":
            grossSurfAreaFactor = 3.0 * SURFACEFACTOR #separation = 3 + additional "dead area"
            
        elif self.collType == "Fresnel collector":
            grossSurfAreaFactor = SURFACEFACTOR
            
########## IDAE IDAE IDAE IDAE! in IDAE ETC are considered as horizontally mounted
        elif self.collType == "Evacuated tube collector":
            grossSurfAreaFactor = SURFACEFACTOR
            logTrack("einstein.modules.solar.moduleST: case 1c - ETC ||||||||| skip after IDAE! %s" % 
                     (grossSurfAreaFactor, ))
            
# 2. south oriented surfaces with sufficient inclination: in-roof mounting without separation of rows is supposed

        elif self.surfaces[i]["azimuthClass"] in self.goodAzimuthClasses and  self.surfaces[i]["inclination"] > MININCLINATION:
            grossSurfAreaFactor = SURFACEFACTOR
            logTrack("einstein.modules.solar.moduleST: case 2a %s" % (grossSurfAreaFactor,))

        else:
            grossSurfAreaFactor = self.grossSurfAreaFactor_FPC()*SURFACEFACTOR
            logTrack("einstein.modules.solar.moduleST: case 2b %s, %s" % 
                     (grossSurfAreaFactor, self.grossSurfAreaFactor_FPC()))

        return grossSurfAreaFactor
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------        
    def checkGrossCollArea(self, i):
#------------------------------------------------------------------------------
        GrossCollAreaPartial = self.surfaces[i]["surfArea"] / self.GrossSurfAreaFactor
        if GrossCollAreaPartial >= MINSURFACE:
            f1 = 1
        else:
            f1 = 0
            logMessage(_("In '%s' the surface available for the collectors mounting is too small.") % (self.surfaces[i]["surfAreaName"]))
        return (f1, GrossCollAreaPartial)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def checkDistance(self, i):
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
        DistanceFactor = (-2 * 0.0000001 * pow(self.surfaces[i]["surfArea"], 2)) + (0.0022 * self.surfaces[i]["surfArea"]) + 0.3938
        MaxDistance = self.surfaces[i]["surfArea"] / (DistanceFactor * self.GrossSurfAreaFactor)
        if self.surfaces[i]["distance"] != None and self.surfaces[i]["distance"] > min(DISTANCECONSTRAINT , MaxDistance):
            if self.surfaces[i]["distance"] != None and self.surfaces[i]["distance"] > 1.2 * min(DISTANCECONSTRAINT , MaxDistance):
                f2 = 0
            else:
                f2 = 1
                logWarning(_("Distance between the solar field'%s' and the technical room or process is too long") % \
                       (self.surfaces[i]["surfAreaName"]))
        else:
            f2 = 1
          
        
        return (f2)
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def checkOrientation(self, i):
#------------------------------------------------------------------------------
        if self.surfaces[i]["azimuthClass"] in self.badAzimuthClasses:
            if self.surfaces[i]["inclination"] < MININCLINATION:
                f3 = 1
            else:
                f3 = 0
                logMessage(_("Area '%s' oriented to %s:not suitable.") % \
                       (self.surfaces[i]["surfAreaName"], self.badOrientation))
        else:
            f3 = 1
        return (f3)       
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def checkShading(self, i):
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
        if self.surfaces[i]["shading"] == "Yes,fully shaded": #==SHADINGTYPES.keys()[2]
            f4 = 0
            fs = 1
            logMessage(_("Area '%s' shaded: not suitable for collectors mounting.") % (self.surfaces[i]["surfAreaName"]))
        else:
            f4 = 1
            if self.surfaces[i]["shading"] == "Yes,partially shaded" :
                fs = 1
            else:
                fs = 0
        return (f4, fs)          
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def checkLoad(self, i, verbose=False):
#------------------------------------------------------------------------------
#   checks static load capacity of the roof
#------------------------------------------------------------------------------

        if self.surfaces[i]["roofStaticLoadCap"] == None:
            f5 = 1
            if self.surfaces[i]["roofType"] in ["Corrugated metal roof", "Composite sandwich panels", "Other"]:
                if verbose: logMessage(_("Check the static load capacity of the roof '%s'and the collectors mounting feasibility.") % \
                           (self.surfaces[i]["surfAreaName"]))
        elif self.surfaces[i]["roofStaticLoadCap"] >= MINSTATICROOF:
            f5 = 1
        else:
            f5 = 0
        return(f5)
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def calcSurfArea(self, verbose=False):
#------------------------------------------------------------------------------
#   ???
#------------------------------------------------------------------------------
        
        self.allShaded = 1
        self.usableSurfacesArea = 0
        self.usableSurfaces = []
        self.usableSurfacesData = []
        self.TotAvailSurfArea = 0
        self.grossCollArea = 0
        self.screenSurfaces()
        
        for i in range(self.NSurfaces):
#            if latitude: if>70 and >20�data entering error
            if self.surfaces[i]["inclination"] > MAXINCLINATION:
                showWarning(_("Check surface '%s' inclination: out of range.") % (self.surfaces[i]["surfAreaName"]))
            self.GrossSurfAreaFactor = self.calcGrossSurfAreaFactor(i)
            (f1, grossCollAreaPartial) = self.checkGrossCollArea(i)
            f2 = self.checkDistance(i)
            f3 = self.checkOrientation(i)
            f4, fs = self.checkShading(i)
            if self.surfaces[i]["roofType"] in ["Corrugated metal roof", "Composite sandwich panels", "Other", "Concrete roof", "Tilted roof"]:
                f5 = self.checkLoad(i)
            else:
                f5 = 1
            self.grossCollArea += grossCollAreaPartial * f1 * f2 * f3 * f4 * f5
            self.TotAvailSurfArea += self.surfaces[i]["surfArea"]
            self.usableSurfacesArea += self.surfaces[i]["surfArea"] * f1 * f2 * f3 * f4 * f5
            if f1 * f2 * f3 * f4 * f5 == 1:
                self.allShaded = self.allShaded * fs
                self.usableSurfaces.append(self.surfaces[i])
                self.usableSurfacesData.append({"GrossSurfAreaFactor":self.GrossSurfAreaFactor, "GrossCollAreaPartial":grossCollAreaPartial})

            logTrack("Surface %s: grossCollAreaPartial %s" % (i, grossCollAreaPartial))

            if DEBUG in ("BASIC", "ALL"):
                print "Surface %s: grossCollAreaPartial %s" % (i, grossCollAreaPartial)

        avAreaFactor = self.avAreaFactor
        self.netCollArea = self.grossCollArea / avAreaFactor        

        if DEBUG in ("BASIC", "ALL"):
            print "avAreaFactor: %s grossCollArea %s netCollArea %s" % (avAreaFactor, self.grossCollArea, self.netCollArea)

        if self.netCollArea < MINTOTNETCOLLAREA and self.NSurfaces > 0:
            if verbose: showWarning(_("The surface available for the collectors mounting is quite small."))
            enoughSurface = False
        else:
            enoughSurface = True

        logTrack("The surfaces have a total area of '%s' and a usable (net) area of '%s'" % 
                 (self.TotAvailSurfArea, self.netCollArea))
        return enoughSurface

                
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def findIndex(self):
        """If a solar thermal equipment exists return its cascaseIndex - 1
        and set self.STExistance to True.
        If none exists return 0 and set self.STExistance to False.
        
        """
        
        index = 0
        Status.int.getEquipmentCascade()
        i = 0
        self.STExistance = False
        for row in Status.int.cascade:
            if getEquipmentClass(row["equipeType"]) == "ST" and index == 0:
                self.existingSTSystem = row
                index = i
#                print ("A solar sistem exist and its position in cascade is:", index+1)
                self.STExistance = True
            i += 1
        return (index)
        
#------------------------------------------------------------------------------
#          if Status.UserInteractionLevel == "interactive" or Status.UserInteractionLevel == "semi-automatic":


#------------------------------------------------------------------------------
    def selectST(self, GT, dT):
        """Return the data of the most efficient solar thermal
        system that fits the given parameters.
        
        :param GT: Solar radiation on tilted surface (SI - units: W/m2)
        :param dT: Temperature difference between collector average and ambient
        
        :returns: a pSQL database entry from dbsolarthermal
        
        """
        config = noneFilter(self.getUserDefinedPars())
        selectedCollector = None

# if a preselected collector is available and selection is set to --preselected-- then use the selected one

        if self.manualChoosenColl is not None:   #E.F. 28/07
            if config[1] == "preselected":   #E.F. 28/07
                selectedCollector = self.manualChoosenColl    #E.F. 28/07
            else:
                logWarning (_("When a particular collector has been selected in the database the collector type should be setted as 'preselected'"))  #E.F. 28/07

# otherwise apply search criteria

        if selectedCollector == None:
            if config[1] in STTYPES:
                possibleCollector = dbio.select("dbsolarthermal", STManufacturer="Einstein", STType = config[1])
            else:
                possibleCollector = dbio.select("dbsolarthermal", STManufacturer="Einstein")

#------------------------------------------------------------------------------
            self.a = 0
            STc0 = possibleCollector[0].STc0
            STc1 = possibleCollector[0].STc1
            STc2 = possibleCollector[0].STc2
            self.etaSelected = collectorEfficiency(GT, dT, STc0, STc1, STc2)
            for i in range (1, len(possibleCollector)):
                STc0 = possibleCollector[i].STc0
                STc1 = possibleCollector[i].STc1
                STc2 = possibleCollector[i].STc2
                etaCollector = collectorEfficiency(GT, dT, STc0, STc1, STc2)
                if possibleCollector[i].STType in ["Parabolic trough collector", "Fresnel collector"]:
                    etaCollector *= 0.8         #HS2008-09-13. some temporary solution for a more realistic comparison
                                                #of concentrating and non-concentrating collectors.
                logTrack("ModuleST.selectST: testing type %s eta=%s"%(possibleCollector[i].STType, etaCollector))
                if etaCollector > self.etaSelected:
                    self.etaSelected = etaCollector
                    self.a = i
            selectedCollector = possibleCollector[self.a]

        if selectedCollector.STAreaFactor is not None:
            self.avAreaFactor = selectedCollector.STAreaFactor
        else:
            self.avAreaFactor = 1.1

        self.collType = selectedCollector.STType

        logTrack("ModuleST.selectST: type(predef) %s selected %s"%(config[1], self.collType))
        return selectedCollector
        
#==============================================================================
#==============================================================================
#==============================================================================
#   PREDIMENSIONING AND DESIGN ASSISTANT 
#============================================================================== 
#============================================================================== 
#============================================================================== 
#------------------------------------------------------------------------------
    def preDimensioning(self):

        """
        carries out a very rough predimensioning of required solar system size

        :returns self.SurfAreaSol: proposed net (effective) area of solar system
        :returns self.TavCollector: estimated average working temperature
        """

        logTrack(" ModuleST: preDimensioning is starting") 

        self.solFraDefault = 0.5
        self.Tmax = 200
        SolFra = self.solFraDefault  # Later on solarFra will be entered from the GUI. Right?
        QuSolarUnitary = 700
        deltaT_Primary = 7
        
        self.findIndex() # check for existing ST equipment
        iTmax = int(self.Tmax / Status.TemperatureInterval + 0.5)
        
        self.QD_Tmax = Status.int.USHTotal_T[iTmax]
        
        QDTav = self.QD_Tmax * SolFra

        i = 0
        self.Tav = 0
        while Status.int.USHTotal_T[i] < QDTav:
            i += 1
            self.Tav = i * Status.TemperatureInterval
        
        if DEBUG in ("BASIC", "ALL"):
            print ("The temperature corresponding to %s of the demand at 200 C is: %s" % (SolFra, self.Tav))

        USHSolTot = QDTav
        self.SurfAreaSol = USHSolTot / (QuSolarUnitary * 0.7)
        
        if self.netCollArea < self.SurfAreaSol and self.NSurfaces > 0:   # limit SSY size by available area only if surfaces are defined
            self.SurfAreaSol = self.netCollArea
            self.QDT = self.SurfAreaSol * QuSolarUnitary * 0.7 / SolFra
            QDTav = self.QDT * SolFra
            i = 0
            while Status.int.USHTotal_T[i] < QDTav:
                i += 1
            self.Tav = i * Status.TemperatureInterval

        else:
            self.QDT = self.QD_Tmax

        self.TavCollector = self.Tav + deltaT_Primary

        logTrack("ModuleST (predimensioning): Area %s TavColl %s QDT %s" % \
                 (self.SurfAreaSol, self.TavCollector, self.QDT))

# NB       if da userDefinedPars arriva una potenza pi� picola ridurre direttamente qui la superficie...
#        print ("The solar area is:",self.SurfAreaSol)
#        print ("corresponding to '%s' kWh" % self.QDT)
#        print ("The temperature corresponding to half of the demand is:",self.Tav)
        
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------
    def designAssistant1(self):
#------------------------------------------------------------------------------
#   assisted design of a solar thermal system
#------------------------------------------------------------------------------
        logTrack("ModuleST (designAssistant1): starting")
        
        if self.NSurfaces <= 0:
            logWarning(_("No information on available surfaces is given !\n") + \
                       _("Is is assumed that there is no constraint due to limited space"))

        index = self.findIndex()   # Index = cascadeIndex-1

#........................................................................
# read in and check the design assistant configuration parameters
        config = self.getUserDefinedPars()       

        if config[0]is not None:
            try:    #if the field is empty division may given an error
                self.desiredSolarFraction = float(config[0]) / 100.0
            except:
                logDebug("ModuleST (designAssistant1): non-readable solar fraction [%s]" % config[0])
                self.desiredSolarFraction = 0.5 
        else:
            self.desiredSolarFraction = 0.5
        if config[2]is not None:
            try:
                self.QuSolarUnitaryMin = float(config[2])
            except:
                self.QuSolarUnitaryMin = 400.0
        else:
            self.QuSolarUnitaryMin = 400.0

        if self.STExistance is True and self.manualChoosenColl is None: #E.F. 28/07
            showWarning(_("A solar thermal system is in use! in order to design a new one you should delete the existent."))
        else:
           
            if Status.int.cascadeUpdateLevel < self.NEquipe or Status.mod.moduleEnergy.outputLevel < 3:
                Status.mod.moduleEnergy.runSimulation()
                
            self.preDimensioning()

            GT = 800
            dT = self.Tav - self.TAmb    
            selectedCollector = self.selectST(GT, dT)

            enoughSurface = self.calcSurfArea()

            if enoughSurface is True:
                # if everything's ok, start assisted design
                # TODO: Replace with dynamic search after ST equipment,
                # the global isn't always updated.
                if Status.SimTool!="Plus":
                    USHjRequired = Status.int.USHTotal_Tt
                    # these two lines are needed to update the plot
                    Status.int.QD_Tt_mod[0] = USHjRequired
                    Status.int.QD_T_mod[0] = QD_T

                if self.STExistance is True:
                    equipe = Status.prj.getEquipe(CascadeIndex=index + 1)
                else:
                    equipe = self.addEquipmentDummy()
                    
                modelID = selectedCollector.id
                
                equipe.HCGPnom = dbio.check(self.SurfAreaSol * 0.7)
                equipe.ST_Volume = dbio.check(self.SurfAreaSol * 0.05)
                equipe.ST_SysEff = dbio.check(DEFAULT_SYSTEM_EFFICIENCY)

                logTrack("ModuleST (designAssistant1): equipe %s"%equipe)
                
                self.setEquipmentFromDB(equipe, modelID)
                
                logTrack("ModuleST (designAssistant1): first estimate for dimensioning\n Pnom: %s Vol: %s" % (equipe.HCGPnom, equipe.ST_Volume))
                dbio.commit()

                if Status.SimTool == "Plus":
                    Status.mod.moduleEnergy.runSimulation(outputLevel=2)
                    self.TavCollMean = Status.mod.moduleEnergy.sim.systemModel.equipment[0].resume()
                    USHjRequired = Status.int.USHTotal_Tt
                    USHj = Status.int.USHj[0]
                else:
##                    USHjRequired = Status.mod.moduleEnergy.sim.getUSHjRequired(1)
                    (dummy, USHj) = self.calculateEnergyFlows(equipe,
                                                              USHjRequired,
                                                              1)
                SolFraCalc = USHj / self.QD_Tmax
                logTrack("ModuleST (designAssistant1): USHj %s QD_Tmax %s"%(USHj, self.QD_Tmax))

                logTrack("ModuleST (designAssistant1): solar fraction for first estimate: %s estimated average op. temp.: %s" % (SolFraCalc, self.TavCollMean))

                if self.manualChoosenColl is None:                    ########## E.F. 28/07
                    while abs (self.TavCollMean - self.Tav) > 10:
                        self.Tav = self.TavCollMean
                        GT = 800
                        dT = self.Tav - self.TAmb
                        
                        selectedCollector = self.selectST(GT, dT)
                        modelID = selectedCollector.id

                        self.setEquipmentFromDB(equipe, modelID)
                        (dummy, USHj) = self.calculateEnergyFlows(equipe,
                                                                 USHjRequired,
                                                                 1)
                
                    SolFraCalc = USHj / self.QD_Tmax

                logTrack("ModuleST (designAssistant1): selected collector type: %s" % selectedCollector.STModel)

                Pnom = equipe.HCGPnom
#                A=equipe.HCGPnom/0.7
                try:
                        QuSolarUnitary = USHj / Pnom
                except ZeroDivisionError:
                        logTrack("einstein.modules.solar.moduleST.ModuleST.designAssistant1.cummulativeDemand"\
                                 " division by zero, used default value 0 for QuSolarUnitary")
                        QuSolarUnitary = 0
                        
                logTrack("ModuleST (designAssistant1): size adjustment cycle starting with P: %s qu %s fsol: %s" % \
                        (Pnom, QuSolarUnitary, SolFraCalc))

                nMaxiter = 10
                bestPnom = Pnom
                bestQu = QuSolarUnitary
                alpha_Qu = 1.0
                
                for n in range(nMaxiter):
                    logTrack("ModuleST.designAssistant: iteration no. %s SolFraCalc%s qu %s" % (n, SolFraCalc, QuSolarUnitary))
#                    if QuSolarUnitary < self.QuSolarUnitaryMin:
                        
#                        A=A*QuSolarUnitary/(self.QuSolarUnitaryMin*1.02)
#                        Pnom=Pnom*QuSolarUnitary/(self.QuSolarUnitaryMin*1.02)
#                        A=max(A,MINTOTNETCOLLAREA)
#                        Pnom=max(Pnom,MINTOTNETCOLLAREA*0.7)
                        
#                        equipe.HCGPnom = Pnom
#                        equipe.ST_Volume = Pnom/0.7*0.05
#                        Status.SQL.commit()

                    if abs(SolFraCalc - self.desiredSolarFraction) > 0.005 or \
                       QuSolarUnitary < self.QuSolarUnitaryMin:   #  In this way we reduce the solar fraction when it exceed 0.5: even if the QuSolarUnitary is good enough
                        logTrack ("ModuleST designAssistant: SolFraCalc= '%s' self.desiredSolarFraction='%s'" % (SolFraCalc, self.desiredSolarFraction))

                        if (abs(QuSolarUnitary - self.QuSolarUnitaryMin) < 0.5) and SolFraCalc < self.desiredSolarFraction:
                            break           # in this case Qu is the dominant adjustment parameter and is already converged
                        
                        # linear extrapolation to desired solar fraction, with subrelaxation factor for assuring convergence
                        # A2 = A*self.desiredSolarFraction/SolFraCalc
                        if (SolFraCalc < 1):
                            logfaux = max(-log(1 - SolFraCalc), 0)
                        else:
                            logfaux = INFINITE

                        if (self.desiredSolarFraction < 1):
                            logfauxD = max(-log(1 - self.desiredSolarFraction), 0)
                        else:
                            logfauxD = INFINITE

                        if SolFraCalc == self.desiredSolarFraction == 0:
                            logfaux = INFINITE
                            logfauxD = INFINITE
                            
#                        Pnom2 = Pnom*self.desiredSolarFraction/SolFraCalc                       
#                        k=0.8
#                        A3 = k*A2+(1-k)*A
#                        Pnom3 = k*Pnom2+(1-k)*Pnom
                
                        k = 0.8

                        if logfaux == 0:
                            if logfauxD <= 0:
                                Pnom_Fsol = Pnom
                            else:
                                Pnom_Fsol = MINTOTNETCOLLAREA * 0.7 #try a minimum solar field in order to see if theres something to do 
                        else:
                            Pnom_Fsol = max(Pnom * (1 + k * (logfauxD / logfaux - 1)), 0)

# linear extrapolation to minimum unit solar gains
#                        Pnom_Qu = Pnom*pow(QuSolarUnitary/self.QuSolarUnitaryMin,2)

                        if (QuSolarUnitary - bestQu) <> 0:
                            Pnom_Qu = alpha_Qu * (Pnom + (self.QuSolarUnitaryMin - QuSolarUnitary) / (QuSolarUnitary - bestQu) * (Pnom - bestPnom)) \
                                      + (1 - alpha_Qu) * bestPnom
                        else:
                            Pnom_Qu = Pnom * pow(QuSolarUnitary / self.QuSolarUnitaryMin, 2)
                            

                        if abs(QuSolarUnitary - self.QuSolarUnitaryMin) <= abs(bestQu - self.QuSolarUnitaryMin):
                            bestQu = QuSolarUnitary
                            bestPnom = Pnom
                        else:
                            alpha_Qu *= 0.9
                    
#                        if QuSolarUnitary < self.QuSolarUnitaryMin:
#                            Pnom_Qu *= 1.02       #seems that this was necessary for a better convergence ???

                        Pnom = min(Pnom_Qu, Pnom_Fsol)

                        logTrack("Pnom %s Pnom_Qu %s Pnom_Fsol %s bestPnom %s bestQu %s" % (Pnom, Pnom_Qu, Pnom_Fsol, bestPnom, bestQu))
                        logTrack("netCollArea = %s" % self.netCollArea)
                        
# limit to available surface
                        if self.NSurfaces > 0:              # otherwise surface constraint is not active !!!
                            Pnom = min (Pnom, self.netCollArea * 0.7)

#                        A5= selectedCollector.STAreaAper * int((A4/selectedCollector.STAreaAper)+0.5)
# now force to an integer number of real collectors
                        Pnom = selectedCollector.STAreaAper * 0.7 * int(((Pnom / 0.7) / selectedCollector.STAreaAper) + 0.5)
                            
                    else:
                        break

# finally constrain by lower limit for reasonable system size

                    if Pnom < MINTOTNETCOLLAREA * 0.7:
                        Pnom = 0

                    equipe.HCGPnom = dbio.check(Pnom)
                    equipe.ST_Volume = dbio.check(Pnom / 0.7 * 0.05)
                    dbio.commit()
                    Status.int.changeInCascade(self.cascadeIndex)

                    (dummy, USHj) = self.calculateEnergyFlows(equipe,
                                                              USHjRequired,
                                                              1)

                    SolFraCalc = USHj / self.QD_Tmax
                    
                    Pnom = equipe.HCGPnom
#                    A = Pnom/0.7
                    try:
                        QuSolarUnitary = USHj / Pnom
                    except ZeroDivisionError:
                        logTrack("einstein.modules.solar.moduleST.ModuleST.designAssistant1.cummulativeDemand"\
                                 " division by zero, used default value 0 for QuSolarUnitary")
                        QuSolarUnitary = 0

                    logTrack("ModuleST (designAssistant1): size adjustment cycle %s. P: %s qu %s fsol %s USH %s" % \
                             ((n + 1), Pnom, QuSolarUnitary, SolFraCalc, USHj))                        
                
#                A=equipe.HCGPnom/0.7
#                NColl=int((A/selectedCollector.STAreaAper)+0.1)
#                equipe.HCGPnom = NColl*selectedCollector.STAreaAper*0.7
#                Status.SQL.commit()
#                USHj=self.calculateEnergyFlows(equipe,1)
                logTrack("ModuleST (designAssistant1): final selection. P: %s qu %s fsol %s" % (Pnom, QuSolarUnitary, SolFraCalc))
#------------------------------------------------------------------------------
#   Here all the data to be displayed or registered are calculated.

                k30 = selectedCollector.STUnitTurnKeyPrice30kW
                k300 = selectedCollector.STUnitTurnKeyPrice300kW
                k3000 = selectedCollector.STUnitTurnKeyPrice3000kW
                unitPrice300 = selectedCollector.STUnitPrice300kW
                OM = selectedCollector.STOMUnitFix
                scf = SolarCostFunction(k30, k300, k3000, unitPrice300, OM)

                (equipe.TurnKeyPrice, equipe.Price, equipe.OandMfix) = scf.calculate(equipe.HCGPnom)
                equipe.ST_Volume = dbio.check(0.05 * equipe.HCGPnom / 0.7)
                dbio.commit()

                try:
                    QuSolarUnitary = USHj / equipe.HCGPnom
                except ZeroDivisionError:
                    logTrack("einstein.modules.solar.moduleST.ModuleST.designAssistant1.cummulativeDemand"\
                                 " division by zero, used default value 0 for QuSolarUnitary")
                    QuSolarUnitary = 0

                self.updatePanel()
            else:
                showWarning(_("The suitable surfaces are quite small. Einstein suggests not to install a solar system"))

            Status.int.systemModel.changed()
            Status.prj.setStatus("Energy", 0)   # the accumulated energy statistics 
                                                # will be out of date

#------------------------------------------------------------------------------

    def calculateCostParameters(self, equipe, modelID):
        """Calculate the cost parameters of a given equipment 
        (depending on size) from the model
        
        """
        try:
            model = dbio.get("dbsolarthermal", id=modelID)
            k30 = model.STUnitTurnKeyPrice30kW
            k300 = model.STUnitTurnKeyPrice300kW
            k3000 = model.STUnitTurnKeyPrice3000kW
            unitPrice300 = model.STUnitPrice300kW
            OM = model.STOMUnitFix
            scf = SolarCostFunction(k30, k300, k3000, unitPrice300, OM)
            (equipe.TurnKeyPrice, equipe.Price, equipe.OandMfix) = scf.calculate(equipe.HCGPnom)
        except:
            logWarning(_("No valid cost parameters available for solar collector model %s") % model.STModel)

        (equipe.TurnKeyPrice, equipe.OandMfix, equipe.OandMvar) = self.estimateEconomicParameters(equipe)
        dbio.commit()

    def estimateEconomicParameters(self, eq):
        """ Default values for economic parameters
        """

        if hasattr(eq, "eqType"): # is an eqObject        
            eqType = eq.eqType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OMFix
            OMVar = eq.OMVar

            Pnom = eq.Pnom
            
        else:   # kept for backward compatibility - still used in setEquipmentFromDB
            eqType = eq.EquipType

            turnKey = eq.TurnKeyPrice
            OMFix = eq.OandMfix
            OMVar = eq.OandMvar

            Pnom = eq.HCGPnom
            

        # security check if the reference power is defined
        Pref = Pnom 
        if Pref is None:
            return (None, None, None)

        turnKey = noneCheck(turnKey, 700*Pnom)
        OMFix = noneCheck(OMFix, 7.0*Pnom)
        OMVar = noneCheck(OMVar, 0.0)      

        return (turnKey, OMFix, OMVar)


if __name__ == "__main__":
    pass
