# -*- coding: utf-8 -*-
#==============================================================================
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#    UPGRADE
#            
#------------------------------------------------------------------------------
#            
#    Functions for management of version upgrades
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------        
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from GUI.GUITools import check
from GUI.status import *
from constants import VERSION, DBSERVER, EPSILON
from modules.messageLogger import (logMessage, showError, showWarning,
                                            logTrack, logDebug, logWarning)
from auxiliary.auxiliary import toGettextAndUnicode as _U, noneCheck
from datetime import date
from sys import exit

import database.dbio as dbio

def upgradeEINSTEIN(targetVersion=VERSION, sourceVersion=None, projectIds=None):
    """Perform all defined upgrades to reach the specified version of EINSTEIN"""
    if projectIds is not None:
        target = (len(projectIds) == 1) and '1 project' or ('%d projects' % len(projectIds))
    else:
        target = 'data base'

    if sourceVersion is None: # determine source version
        if projectIds is None: # complete data base upgrade requested
            try:
                stool = dbio.get("stool")
                sourceVersion = stool.Version
            except (KeyError, IndexError): # default version for complete upgrade
                sourceVersion = '1.1'
        else: # default schema version for selected projects
            sourceVersion = '1.1'

    if lastUpgrade(targetVersion) == lastUpgrade(sourceVersion): # no upgrades necessary
        return
    elif lastUpgrade(targetVersion) < lastUpgrade(sourceVersion):     
        logWarning(_("Your database has already been updated with an EINSTEIN Version newer than the one you are currently using."))
        logWarning(_("Take care: this may eventually cause damage to your projects. Projects may also not be running correctly"))
        logWarning(_("Current tool version: %s Last database update by tool version: %s"%(targetVersion, sourceVersion)))
        return
##    else:
##        logMessage(_('Upgrading your database from version %s to version %s' % (sourceVersion, targetVersion)))

    # call all upgrade functions in order of ascending versions and then list order    
    version = sourceVersion
    for version in sorted(upgrades.keys()):
        if version > sourceVersion and version <= targetVersion:
            logMessage('Upgrade of %s to version %s starting' % (target, version))
            for upgrade in upgrades[version]:
                logMessage('%s starting' % upgrade.__name__)
                try:
                    upgrade(projectIds)
                except Exception, error:
                    raise
                    showError('EINSTEIN upgrade failed\n\nError in %s\n%s' % (upgrade.__name__, error))
                    exit('EINSTEIN fatal: upgrade failed')
                logMessage('%s complete' % upgrade.__name__)
            logMessage('Upgrade of %s to version %s complete' % (target, version))
    
        if (projectIds is None) and (sourceVersion < version): # store reached version in data base
            stool = dbio.get("stool")
            if stool is None:
                dbio.insertRow("stool", Version=targetVersion)
            else:
                stool.set(Version=targetVersion)

    if version < targetVersion: logMessage('Upgrade of %s to version %s complete' % (target, targetVersion))

    dbio.commit()

##def upgradeDBSchema(version):
##    """Upgrade data base schema to specified version
##       UNUSED for now, needs more work
##    """
##    import os
##    import fnmatch
##    import MySQLdb
##    schemaUpgrade = fnmatch.filter(os.listdir('../sql'), 'update_einsteinDB_schema_*_to_%s.sql' % version)
##    try: # only honor the first upgrade to given version
##        sqlSchemaUpgrade = open(os.path.join('../sql', schemaUpgrade[0])).read()
##        dbCursor = Status.SQL.cursor(MySQLdb.cursors.DictCursor)
##        logMessage('data base schema upgrade starting')
##        dbCursor.execute(sqlSchemaUpgrade)
##        logMessage('data base schema upgrade complete: %d rows affected' % dbCursor.rowcount)
##    except IndexError: # no upgrade found -> fine
##        pass

#=============================================================================================================
# Auxiliary functions
#=============================================================================================================

def getAllBuildings(projectIds):
    # set defaults in upgraded table qprocessdata
    if projectIds is None: # default to complete table
        dataRows = dbio.select("qbuildings")
    elif projectIds == []: # no projects -> nothing to do
        return
    else: # just process the selected projects
        dataRows = []
        for PId in projectIds:
            dataRows += dbio.select("qbuildings", Questionnaire_id = PId)
    return dataRows

def getAllEquipment(projectIds):
    # set defaults in upgraded table qprocessdata
    if projectIds is None: # default to complete table
        eqDataRows = dbio.select("qgeneraldata")
    elif projectIds == []: # no projects -> nothing to do
        return
    else: # just process the selected projects
        eqDataRows = []
        for PId in projectIds:
            eqDataRows += dbio.select("qgeneraldata", Questionnaire_id = PId)
    return eqDataRows

def getAllProcesses(projectIds):
    # set defaults in upgraded table qprocessdata
    if projectIds is None: # default to complete table
        processDataRows = dbio.select("qprocessdata")
    elif projectIds == []: # no projects -> nothing to do
        return
    else: # just process the selected projects
        processDataRows = []
        for PId in projectIds:
            processDataRows += dbio.select("qprocessdata", Questionnaire_id = PId)
    return processDataRows


#=============================================================================================================
# Below: single upgrade functions to be called
#=============================================================================================================

def upgradeDetailedSchedules(projectIds=None):
    """Parameter upgrades for compatibility with detailed schedules"""
    from modules.schedules     import DEFAULTCHARGETIME
    from modules.fluids        import Fluid

    processDataRows = getAllProcesses(projectIds)

    for processDataRow in processDataRows:
        # VInFlowCycle
        try:
            VInFlowCycle = processDataRow.VInFlowDay / processDataRow.NBatch
        except (TypeError, ZeroDivisionError): # one of the operands is None == NULL -> Store NULL
            VInFlowCycle = None

        # VOutFlowCycle
        try:
            VOutFlowCycle = processDataRow.VOutFlow / processDataRow.NBatch
        except (TypeError, ZeroDivisionError): # one of the operands is None == NULL -> Store NULL
            VOutFlowCycle = None

        # HPerYearInFlow
        try:
            HPerYearInFlow = processDataRow.HBatch * processDataRow.NBatch * processDataRow.NDaysProc
            if processDataRow.ProcType == 'batch':
                HPerYearInFlow = DEFAULTCHARGETIME * HPerYearInFlow
        except (TypeError, ZeroDivisionError): # one of the operands is None == NULL -> Store NULL
            HPerYearInFlow = None

        # HPerYearOutFlow
        HPerYearOutFlow = HPerYearInFlow

        # mInFlowNom
        if processDataRow.ProcMedDBFluid_id is not None:
            FluidDensity = Fluid(processDataRow.ProcMedDBFluid_id).rho
        else:
            FluidDensity = None
        try:
            mInFlowNom = (FluidDensity * processDataRow.VInFlowDay * processDataRow.NDaysProc) / HPerYearInFlow
        except (TypeError, ZeroDivisionError): # one of the operands is None == NULL -> Store NULL
            mInFlowNom = None

        # mOutFlowNom
        try:
            mOutFlowNom = (FluidDensity * processDataRow.VOutFlow * processDataRow.NDaysProc) / HPerYearOutFlow
        except (TypeError, ZeroDivisionError): # one of the operands is None == NULL -> Store NULL
            mOutFlowNom = None

        processDataRow.update({
                               'PartLoad'        : 1.0,
			                   'VInFlowCycle'    : VInFlowCycle,
			                   'VOutFlowCycle'   : VOutFlowCycle,
                               'mInFlowNom'      : mInFlowNom,
                               'mOutFlowNom'     : mOutFlowNom,
                               'HPerYearInFlow'  : HPerYearInFlow,
                               'HPerYearOutFlow' : HPerYearOutFlow
                               })
    dbio.commit()

def portProcessIOtoStreams(projectIds=None):
    """Map the inflow and outflow properties of the processes of a
    project to corresponding streams.
    
    If the process has in VInFlowDay, mInFlow, vInFlowCycle an 
    explicit 0 then no inflow stream is created.
    If the process has in VoutFlow, mOutFlow, vOutFlowNom an explicit
    0 then no outflow stream is created.
    In the other cases the values of the new streams are taken from 
    the process.
     
    :param projectIds: a list of projectIds of projects to be 
        converted.
    :type projectIds: List
    
    """
    logDebug("Updating processes to streams model.")
    from model.objects.processStreams import InflowingStream, OutflowingStream
    fluids = Status.prj.getFluidDict()
    fluids[None] = None

    processDataRows = getAllProcesses(projectIds)

    for row in processDataRows:
        logDebug("Updating process %s with id %s to streams model." % (row.Process, row.QProcessData_ID))
        # create the inflow stream, if necessary
        if any([item != 0 for item in (row.VInFlowDay, row.mInFlowNom,
                                      row.VInFlowCycle)]):
            logTrack("Creating instream for process %s." % (row.Process, ))
            instream = InflowingStream("generic inflow stream of "
                                       + str(row.Process))
            instream.Medium = fluids[row.ProcMedDBFluid_id]
            instream.PTInFlow = row.PTInFlow
            instream.PTInFlowRec = row.PTInFlowRec
            instream.VInFlowCycle = row.VInFlowCycle
            instream.mInFlowNom = row.mInFlowNom
            instream.HeatRecExist = row.HeatRecExist
            instream.HInFlow      = None
            instream.XInFlow      = None
            instream.UPHc = row.UPHc
            instream.QdotProc_c = row.QdotProc_c
            instream.saveToDB(row.QProcessData_ID, 
                              row.AlternativeProposalNo,
                              row.Questionnaire_id)
        # create the outflow stream, if necessary
        if any([item != 0 for item in (row.VOutFlow, row.mOutFlowNom,
                                       row.VOutFlowCycle)]):
            logDebug("Creating outstream for process %s." % (row.Process, ))
            outstream = OutflowingStream("generic outflow stream of "
                                         + str(row.Process))
            outstream.Medium = fluids[row.ProcMedDBFluid_id]
            outstream.PTOutFlow = row.PTOutFlow
            outstream.PTOutFlowRec = row.PTOutFlowRec
            outstream.PTFinal = row.PTFinal
            outstream.VOutFlowCycle = row.VOutFlowCycle
            outstream.mOutFlowNom = row.mOutFlowNom
            outstream.HeatRecOK = row.HeatRecOK
            outstream.HOutFlow = row.HOutFlow
            outstream.XOutFlow = row.XOutFlow
            outstream.UPHw = row.UPHw
            outstream.QdotProc_w = row.QdotProc_w
            outstream.saveToDB(row.QProcessData_ID, 
                              row.AlternativeProposalNo,
                              row.Questionnaire_id)
    dbio.commit()
    return

def createProcessesForBuildings(projectIds=None):
    """Create processes that represent the heating, cooling and
    domestic hot water demand of a building.

    :param projectIds: list of ProjectIDs

    """
    from modules.buildings import Building

    datarows = getAllBuildings(projectIds)
    if len(dataRows)>0:
        showWarning(_U("Buildings found. Maximal heating, maximal cooling and daily domestic hot water demand of these are set to 0."))
    for row in dataRows:
        building = Building(row.BuildName)
        try:
            building.loadFromDB(row.Questionnaire_id, row.AlternativeProposalNo)
        except:
            logWarning("Invalid building entry for %s" % (row.BuildName,))
            continue # invalid building entry, can be omitted
        building.buildMaxCP = 0
        building.buildMaxHP = 0 
        building.buildDailyDHW = 0
        building.buildAnnualAirCond = None
        building.buildAnnualHeating = None
        building.saveToDB(row.Questionnaire_id, row.AlternativeProposalNo)
    dbio.commit()
    
def changeSTCollectortype(projectIds=None):
    """Replace the type "Concentrating collector" with 
    "Parabolic trough collector" in the dbsolarthermal table.

    :param projectIds: is ignored as it does not touch projects

    """
    dbsolarthermal = dbio.select("dbsolarthermal", STType="Concentrating collector")
    for row in dbsolarthermal:
        row.STType = "Parabolic trough collector"
    dbio.commit()

def moveEconomicDataToCGeneralData(projectIds=None):
    """Move some economic parameters from the questionnaire to the
    cgeneraldata table, as they may be changed by the consistency check
    and questionnaire contains only data that is user input.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    db = Status.DB
    if projectIds is None:
        source = dbio.select("questionnaire")
    else:
        source = []
        for pid in projectIds:
            source += dbio.select("questionnaire", id=pid)

    for source_row in source:
        sinks = dbio.select("cgeneraldata", Questionnaire_id=source_row.Questionnaire_ID)
        for sink_row in sinks:
            sink_row.InflationRate = noneCheck(source_row.InflationRate, 0.02)
            sink_row.FuelPriceRate = noneCheck(source_row.FuelPriceRate, 0.03)
            sink_row.InterestExtFinancing = noneCheck(source_row.InterestExtFinancing, 0.06)
            sink_row.CompSpecificDiscountRate = noneCheck(source_row.CompSpecificDiscountRate, 0.03)
            sink_row.AmortisationTime = noneCheck(source_row.AmortisationTime, 20)
            
    dbio.commit()

def shiftTemperatureReCooling(projectIds=None):
    """Shift the parameter TemperatureReCooling for air and water 
    cooled compression chillers, as the simulation code has changed 
    to a different reference point.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    equipment = getAllEquipment(projectIds)                            
    for eq in equipment:
        if eq.EquipType == "compression chiller (air cooled)":
            try:
                eq.TemperatureReCooling = eq.TemperatureReCooling - 16.5
            except TypeError: 
                pass # None entries don't need modification
        elif eq.EquipType == "compression chiller (water cooled)":
            try:
                eq.TemperatureReCooling = eq.TemperatureReCooling - 7.5
            except TypeError: 
                pass # None entries don't need modification
        else:
            pass # other equipments are not of interest

    dbio.commit()

def assumeSoundEconomicData(projectIds=None):
    """Assume that the econmic data of old projects is fine as is.
    Otherwise all variable costs in old projects would be removed as
    the equipment would seem to have changed since the last TCA calculations.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    if projectIds is None:
        source = dbio.select("salternatives")
    else:
        source = []
        for PId in projectIds:
            source += dbio.select("salternatives", ProjectID=PId)
                                   
    for source_row in source:
        source_row.StatusECO = 2

    dbio.commit()
        
def connectTCADataToAlternatives(projectIds=None):
    """All TCA data sets now keep track of their project id and
    alternative number. Set these values for all old TCA data sets.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    if projectIds is None:
        TCAgeneral = dbio.select("tcageneraldata")
    else:
        TCAgeneral = []
        for PId in projectIds:
                TCAgeneral += dbio.select("tcageneraldata", ProjectID=PId)
                            
    tableNames = ("tcacontingencies", "tcadetailedopcost", "tcadetailedrevenue", "tcaenergy", "tcainvestments", "tcanonreoccuringcosts")
    for tca_set in TCAgeneral:
        PId = tca_set.ProjectID
        ANo = tca_set.AlternativeProposalNo
        tcaID = tca_set.id
        for tableName in tableNames:
            table = dbio.select(tableName, TcaID=tcaID)   
            for data_set in table:
                data_set.ProjectID = PId
                data_set.AlternativeProposalNo = ANo

    dbio.commit()

def addDPSTminValues(projectIds=None):
    """Add a variable for the before fixed parameters delta process/supply
    temperature.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    processes = getAllProcesses(projectIds)

    for process in processes:
        heating = process.heating_cooling == "heating"
        if heating is True:
            process.DPSTmin = 10
        else:
            process.DPSTmin = 5
        process.DPSTminMCD = 5
        
    dbio.commit()

def addExternalSupply(projectIds=None):
    """Add a variable for the before fixed parameters delta process/supply
    temperature.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    processes = getAllProcesses(projectIds)
    for process in processes:
        process.ExternalSupply = "yes"
    dbio.commit()

def resetHeatSourceLT(projectIds=None):
    """The type of the HeatSourceLT column has changed, all old values
    were ignored anyway, so reset them for imported projects.
    
    An update of the complete database is not needed as this is part 
    of the database schema changes. If projectIds is None nothing is
    done.
    
    :param projectIds: list of ProjectIDs
    
    """

    equipment = getAllEquipment(projectIds)
    for eq in equipment:
        eq.HeatSourceLT = None
    dbio.commit()
    
def addIndoorTemperatures(projectIds=None):
    """Add the target temperatures for building heating and cooling
    as derived from the temperatures of the supply.
    
    If no project IDs are given all projects are updated.
    
    :param projectIds: list of ProjectIDs
    
    """
    buildings = getAllBuildings(projectIds)
                                   
    for building in buildings:
        try:
            building.IndoorTHeating = building.BuildTHeating - 10
        except TypeError:
            pass
                                   
        try:
            building.IndoorTAirCond = building.BuildTAirCond + 5
        except TypeError:
            pass

        try:
            bui = Building(building.BuildName)
            bui.loadFromDB(building.Questionnaire_id, building.AlternativeProposalNo)
        except:
            logWarning("Invalid building entry for %s" % (building.BuildName,))
            continue # invalid building entry, can be omitted
        else:
            bui.saveToDB(buillding.Questionnaire_id, building.AlternativeProposalNo)

    dbio.commit()

def updateSolarThermalDB(projectIds=None):
    """Replace default data with more precise default data.
    
    If a projectId is given, nothing is done. This function is only 
    sensible for updates of the whole database.               

    """
    for name, STc0, K50L, K50T, STOMUnitFix in (("FPCEinsteinSelective", 0.78, 0.93, 0.93, 3.57),
                                                ("ETCEinstein", 0.76, 0.94, 1.1, 3.57)):
        dataset = dbio.get("dbsolarthermal", STModel=name)
        if dataset is not None:
            dataset.STc0 = STc0
            dataset.K50L = K50L
            dataset.K50T = K50T
            dataset.STOMUnitFix = STOMUnitFix 
        else:
            logWarning("Could not find the solar thermal type %s,\nbut it is a default database entry." % name)
            
    dbio.commit()

def updateChillerData(projectIds=None):
    """Adapt chiller data for versions 2.1 and before in order to
    reflect the changes in the chiller module:

    HCGTEfficiency -> copy into HCGTEffReal, if applies with partload correction.

    :param projectIds: list of ProjectIDs
    
    """
    logTrack("upgrade.updateChillerData: updating database for projectIds %s"%projectIds)
    db = Status.DB
    if projectIds is None:
        equipmentsOrig = dbio.select("qgenerationhc", AlternativeProposalNo = -1)
        equipmentsPS = dbio.select("qgenerationhc", AlternativeProposalNo = 0)
    else:
        equipmentsOrig = []
        equipmentsPS = []
        for PId in projectIds:
            equipmentsOrig += dbio.select("qgenerationhc", Questionnaire_id=PId, AlternativeProposalNo = -1)
            equipmentsPS += dbio.select("qgenerationhc", Questionnaire_id=PId, AlternativeProposalNo = 0)
    for eqOrig in equipmentsOrig:
        eqPS = None
        eqPSid = None
        for eq in equipmentsPS:
            if (eq.Questionnaire_id == eqOrig.Questionnaire_id) and (eq.Equipment == eqOrig.Equipment):
                eqPS = eq
                eqPSid = eq.QGenerationHC_ID
                
                break

        logTrack("Equipment pair %s orig: %s PS: %s"%(eqOrig.Equipment, eqOrig.QGenerationHC_ID,eqPSid))
                
        equipmentType = eqOrig.EquipType
        if equipmentType in ["thermal chiller (water cooled)",
                             "thermal chiller (air cooled)",
                             "compression chiller (air cooled)",
                             "compression chiller (water cooled)"]:
            partLoadCorrection = True
        else:
            partLoadCorrection = False

        # first check and update present state values
        partLoad = None
        if eqPS is not None:
            logTrack("updating equipment %s (PS) id %s OriginalId %s PartLoad %s"%(eqPS.Equipment,eqPS.QGenerationHC_ID,eqOrig.QGenerationHC_ID, eqPS.PartLoad))

            partLoad = eqPS.PartLoad
            if eqPS.HCGTEffReal is None and eqPS.HCGTEfficiency is not None:

                eff = eqPS.HCGTEfficiency
                if partLoadCorrection and partLoad is not None:
                    eff *= (0.9 + 0.1*partLoad)

                logTrack("HCGTEfficiency: %s HCGTEffReal %s -> %s"%(eqPS.HCGTEfficiency,eqPS.HCGTEffReal,eff))
                eqPS.HCGTEffReal = eff

        if partLoad is None:
            partLoad = 1. # this is not always correct, but the only approximation that can be done -> partLoad PS used as default value for correction in original state.
            
        logTrack("updating equipment %s (original)id %s PartLoad %s"%(eqOrig.Equipment,eqOrig.QGenerationHC_ID,eqOrig.PartLoad))
        if eqOrig.HCGTEffReal is None and eqOrig.HCGTEfficiency is not None:

            eff = eqOrig.HCGTEfficiency
            if partLoadCorrection:                
                partLoadOrig = eqOrig.PartLoad
                if partLoadOrig is None:
                    partLoadOrig = partLoad #use present state value, if partLoad not defined as user input
                eff *= (0.9 + 0.1*partLoadOrig)
                
            logTrack("HCGTEfficiency: %s HCGTEffReal %s -> %s"%(eqOrig.HCGTEfficiency,eqOrig.HCGTEffReal,eff))
            eqOrig.HCGTEffReal = eff
    dbio.commit()

def upgradeTimeSchedules(projectIds=None):
    """Adapt chiller data for versions 2.1 and before in order to
    reflect the changes in the chiller module:

    HCGTEfficiency -> copy into HCGTEffReal, if applies with partload correction.

    :param projectIds: list of ProjectIDs
    
    """
    from modules.project import copySQLRows
    
    logTrack("upgrade.updateTimeSchedules: updating database for projectIds %s"%projectIds)
    pIds = projectIds
    if pIds is None:
        projects = Status.DB.questionnaire.Questionnaire_ID['%']
        pIds = [p.Questionnaire_ID for p in projects]

    for PId in pIds:
        originalProcesses = Status.prj.getProcesses(PId,ANo=-1)
        processes = Status.prj.getProcesses(PId,ANo=None)
        for oproc in originalProcesses:
            oProcessId = oproc.QProcessData_ID
            oPeriods = dbio.select("process_periods", qprocessdata_QProcessData_ID=oProcessId)
            oprocHasDetailedSchedule = (len(oPeriods) > 0)
            if oprocHasDetailedSchedule:
                oProcNo = oproc.ProcNo
                oName = oproc.Process
                for proc in processes:
                    procId = proc.QProcessData_ID
                    if proc.Process == oName and proc.QProcessData_ID != oProcessId:
                        periods = dbio.select("process_periods", qprocessdata_QProcessData_ID=procId)
                        procHasDetailedSchedule = (len(periods)>0)
                        if not procHasDetailedSchedule:
                            logTrack("Copying detailed schedule: PId %s oProcNo %s oProc %s ANo %s"% (PId, oProcNo, oName, proc.AlternativeProposalNo))

                            # copying project detailed schedule tables
                            copiedProcessPeriods = copySQLRows("process_periods", "qprocessdata_QProcessData_ID = %d" % oProcessId, "id", "qprocessdata_QProcessData_ID", procId)
                            for (oldProcessPeriodsId, newProcessPeriodsId) in copiedProcessPeriods.items():
                                copySQLRows("process_period_profiles", "process_periods_id = %d" % oldProcessPeriodsId, "id", "process_periods_id", newProcessPeriodsId)

                            if proc.ProcNo != oProcNo:
                                logWarning("Project %s: process %s has oProcNo %s and ProcNo %s in alternative %s" %(PId, oName, oProcNo, proc.ProcNo, proc.AlternativeProposalNo))
            
    dbio.commit()
        
def upgradeProjectData(projectIds=None):
    """Updates the key project data
    creation date -> remains None
    last modification date -> set to current date
    

    :param projectIds: list of ProjectIDs
    
    """
    from modules.xmlIO import setUniqueName
    
    logTrack("upgrade.upgradeProjectData: updating project data for project IDs %s"%projectIds)
    pIds = projectIds
    if pIds is None:
        projects = dbio.select("questionnaire")
        pIds = [p.Questionnaire_ID for p in projects]

    for PId in pIds:
        projectStatus = Status.prj.getProjectStatus(PId)
        projectData, generalData = Status.prj.getProjectData(PId)

        if projectStatus.ProjectName is None:   
            projectStatus.ProjectName = setUniqueName("sproject", projectData.Name) #old projects have been identified only by explicit company name
            projectStatus.ModifiedBy = Status.AuditorId
            projectStatus.LastModified = date(2000,1,1)
            logTrack("PId %s name %s updated. AuditorID %s"%(PId, projectStatus.ProjectName, Status.AuditorId))

    dbio.commit()
    
def upgradeWaterCooledChillerData(projectIds=None):
    """Updates the key project data
    creation date -> remains None
    last modification date -> set to current date
    

    :param projectIds: list of ProjectIDs
    
    """

    logTrack("upgrade.upgradeWaterCooledChillerData: updating chiller data for project IDs %s"%projectIds)
    equipment = getAllEquipment(projectIds)
                                   
    count = 0
    for eq in equipment:
        if eq.EquipType in ["thermal chiller (water cooled)",
                            "compression chiller (water cooled)"]:
            eq.DestinationWasteHeat = "cooling tower (wet)"
            count += 1
        elif eq.EquipType in ["thermal chiller (air cooled)",
                            "compression chiller (air cooled)"]:
            eq.DestinationWasteHeat = "direct heat rejection"
            count += 1
        else:
            eq.DestinationWasteHeat = None #even if before something had been set, this has been ignored
    logTrack("upgrade.upgradeWaterCooledChillerData: %s utilities changed"%count)

    dbio.commit()

def upgradeHXConnections(projectIds):
    """
    Writes columns source_id and pID establishing the correct links formerly defined
    redundantly by entry of objectNames in table heat_exchanger and by pinchStreamName
    """
    
    connections = dbio.select("heatexchanger_pinchstream") if projectIds is None else \
                  dbio.selectByList("heatexchanger_pinchstream", "ProjectID", projectIds)
    
    for conn in connections:
        ps = dbio.get("pinchstream", id=conn.pinchstream_id)
        hx = dbio.get("qheatexchanger", id=conn.qheatexchanger_QHeatExchanger_Id)

        upstreamHXId = conn.outletOfHX_id

        if upstreamHXId is not None:
            ps.source_type="HX"
            ps.source_id=upstreamHXId
            continue

        if ps is None or hx is None:
            print "TEST TEST Upgrate hxconn: no linked tables found for connId %s ps %s hx %s"%(conn.id, ps, hx)
            continue
            
        if ps.Hot_Cold in ["Sink", "Cold"]:
            heating = True
        else:
            heating = False

        PId = hx.ProjectID
        ANo = hx.AlternativeProposalNo
        
        objectName = hx.HXSink if heating else hx.HXSource
        objectType = ps.source_type

        # first correct outdated objecttypes
        if objectType=="EQUIPMENT":
            objectType = "Equipment"
            ps.source_type=objectType
        elif objectType in ["PROCESSBATCH", "PROCESSCONTINUOUS"]:
            objectType = "ProcessH"
            ps.source_type=objectType
            
        if objectType in ["ProcessH", "ProcessC"]:
            obj = dbio.get("qprocessdata", Process=objectName, Questionnaire_id=PId, AlternativeProposalNo=ANo)
            if obj is not None and obj.heating_cooling == "cooling":
                objectType = "ProcessC"
                ps.source_type=objectType
        elif objectType in ["Equipment"]:
            obj = dbio.get("qgenerationhc", Equipment=objectName, Questionnaire_id=PId, AlternativeProposalNo=ANo)
        elif objectType in ["PipeC", "PipeH"]:
            obj = dbio.get("qdistributionhc", Pipeduct=objectName, Questionnaire_id=PId, AlternativeProposalNo=ANo)
##        elif objectType in ["HX"]:
##            obj = dbio.get("qheatexchanger", HXName=objectName, Questionnaire_id=PId, AlternativeProposalNo=ANo)
        elif objectType in ["WHEE"]:
            obj = dbio.get("qwasteheatelequip", WHEEName=objectName, ProjectID=PId, AlternativeProposalNo=ANo)
        else:
            continue

        if obj is not None:
            ps.source_id = obj.id
        else:
            ps.source_id = None

        ps.pID = None   # pID is set to None, which triggers loading pID from ps.Name (in HXConnection, loadFromDB; for backward compatibility)
        
        
    dbio.commit()

def cleanUpHXConnections(psIds):
    """
    Writes columns source_id and pID establishing the correct links formerly defined
    redundantly by entry of objectNames in table heat_exchanger and by pinchStreamName
    """
    for psId in psIds:
        ps = dbio.get("pinchstream", id=psId)
        ps.source_id = None
        ps.pID = None
        
    dbio.commit()

def correctRawStreamData(projectIds=None):
    """
    In some cases QdotProc_c data have been assigned to 0.0 in old projects
    although it was not really a user input befor V2.9.00
    """

    if projectIds is None:
        instreams = dbio.select("streams_in", AlternativeProposalNo=-1)
    else:
        instreams = dbio.selectByList("streams_in", "ProjectId", projectIds)
        instreams = [stream for stream in instreams if stream.AlternativeProposalNo==-1]

    for stream in instreams:
        stream.QdotProc_c = None

    dbio.commit()

def correctCOPex_HP(projectIds):
    """
    Correct COPex in projects created with older versions in order to compensate for
    the change of HPTDROP from 5 to 7 K in Version 3.0
    """
    from lib.physics.thermo import carnotCOPh
    eqType = "compression heat pump"

    hps = dbio.select("qgenerationhc", EquipType=eqType)
    if projectIds is not None:
        hps = [hp for hp in hps if hp.Questionnaire_id in projectIds]
        
    for hp in hps:
        Th = hp.TSupply
        Tc = hp.THeatSourceLT
        COPh = hp.HCGTEfficiency
        COPex = hp.HPExHeatCOP

        nominalConditionsGiven = not (Th is None or Tc is None or COPh is None)   

        Delta_T_default = 33      # default Delta_T for 7/40°C
        Th_i = noneCheck(Th, (Tc + Delta_T_default) if Tc is not None else 313.15) # default 40 °C
        Tc_i = noneCheck(Tc, (Th - Delta_T_default) if Th is not None else 280.15) # default  7 °C

        COPCarnot_old = carnotCOPh(Th_i + 5, Tc_i - 5)
        COPCarnot_new = carnotCOPh(Th_i + 7, Tc_i - 7)

        COPex_calculated = COPh / COPCarnot_new if (nominalConditionsGiven and COPCarnot_new > EPSILON) else None
        
        if COPex is None or COPex_calculated is not None: # no COPex defined -> write the calculated one
            COPex = COPex_calculated
##            print "Upgrade.correctCOPex_HP: COPex calculated to %s"%(f_corr, COPex)
        else:
            f_corr = COPCarnot_old / COPCarnot_new if COPCarnot_new > 0 else 0.0
            COPex *= f_corr
##            print "Upgrade.correctCOPex_HP: COPex corrected by f_corr %s to %s"%(f_corr, COPex)
            
        hp.HPExHeatCOP = COPex
    
# Upgrades to perform in order to reach a specific version of EINSTEIN
upgrades = {
            "1.2": [upgradeDetailedSchedules],
            "1.8": [portProcessIOtoStreams, createProcessesForBuildings],
            "1.9": [changeSTCollectortype],
            "2.0.1": [moveEconomicDataToCGeneralData],
            "2.0.2": [assumeSoundEconomicData, connectTCADataToAlternatives],
            "2.0.3": [addDPSTminValues, addExternalSupply],
            "2.0.4": [resetHeatSourceLT, addIndoorTemperatures],
            "2.0.5": [shiftTemperatureReCooling],
            "2.0.8": [updateSolarThermalDB],
            "2.1.01": [updateChillerData],
            "2.1.02": [upgradeTimeSchedules],
            "2.1.94": [upgradeProjectData, upgradeWaterCooledChillerData],
            "2.9.00": [upgradeHXConnections, correctRawStreamData],    
            "2.9.90": [correctCOPex_HP]      
            }


def lastUpgrade(version):
    upgradeList = ["1.0"]+[elem for elem in sorted(upgrades.keys()) if elem <= version]
    lastUpgrade = upgradeList[-1]
    return(lastUpgrade)
        

#if __name__ == "__main__":
#    upgradeEINSTEIN(targetVersion='1.2')
