# -*- coding: utf-8 -*-
#==============================================================================
#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   EINSTEIN Main
#
#------------------------------------------------------------------------------
#
#   GUI Main routines
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from datetime import datetime
import numpy as np
from GUI.status import Status
from modules.messageLogger import logWarning, logError, showError, logTrack
from constants import DEBUG, YEAR, APPDATA, DBSERVER
import database.dbio as dbio
import math
from auxiliary.auxiliary import toGettextAndUnicode as _U
from modules.energy.sunday import Location

class WeatherData:

    def __init__(self, PId):
        """weather data that should be defined for EINSTEIN calculations
        
        A filePath that consists only of whitespace is considered a None
        filePath.
        
        """
        self.Tdb = None
        self.Twb = None
        self.TAmbAvg = None
        self.I = None
        self.Idn = None
        self.latitude = None
        self.longitude = None

        self.PId = PId

        self.loadFromDB(PId)


    def loadFromDB(self, PId):
        """ loads data from DB """

        ren = Status.prj.getRenewables(PId)
        # read in paramters from SQL table            
        if ren is not None:
            self.latitude = ren.Latitude
            self.I = ren.ST_I
            self.Idn = ren.ST_Idn
            self.TAmbAvg = ren.TAmb
            self.dataFilePath = ren.WDataFile

            if self.dataFilePath is not None and self.dataFilePath.strip() == "":
                self.dataFilePath = None

        else:
            self.dataFilePath = None

        self.buildData(self.dataFilePath)

    def saveToDB(self):

        tmp = {
            "Latitude":dbio.check(self.latitude),
            "ST_I":dbio.check(self.I),
            "ST_Idn":dbio.check(self.Idn),
            "TAmb":dbio.check(self.TAmbAvg),
            "WDataFile": dbio.check(self.dataFilePath),
            }
        
        qrenewables = Status.prj.getRenewables(self.PId)
        if qrenewables is None:
            # register does not exist, so store also id
            tmp.update({"Questionnaire_id":self.PId}) 
            dbio.insertRow("qrenewables", **tmp)
        else:
            # register does exist
            qrenewables.set(**tmp)
        dbio.commit()

        
    def buildData(self, path):
        """
        Rebuilds the data from a data file if available (path), otherwise
        - or in case of a corrupt data file - generates simplified data
        from the global parameters as stored
        """
        
        self.dataFilePath = path
        if path is not None:
            self.name = 'TMY2File'
            try:
                self.loadFromTMY2File(path)
                dataOK = True
            except Exception, error:  
                showError('Error loading weather data\n%s' % error)
                dataOK = False
        else:
            dataOK = False
            
        if not dataOK:
            self.name='Simplified'
            self.generateSimplifiedData()

        self.buildHourlyData() # loads and stores hourly data to be used in simulations and sets up Location class
        logTrack("WeatherData.init: dataFile %s (%s) I %s Idn %s TambAvg %s"%(dataOK, self.dataFilePath, self.I, self.Idn, self.TAmbAvg))



### DATA FILE HANDLING -------------------------------------------------------------------------------        
    
    def loadFromTMY2File(self, path):
        self.data = {}
        """Typical Meteorological Year, version 2 (TMY2)"""

        if not path.startswith("/"):
            absolutePath = APPDATA + "/weather/" + path
        else:
            absolutePath = path
            
        try:
            dataFile = file(absolutePath)
        except IOError:
            try:
                defaultPath = self.defaultPath(path)
                logWarning(_("Datafile %s not found")%absolutePath + "\n" + _("Looking up in default path %s")%defaultPath)
                dataFile = file(defaultPath)
            except IOError:
                logWarning("saxndixn")
                raise
        
        try:
            # header elements
            header = dataFile.readline()
            # Station's Weather Bureau Army Navy number
            self.data['WBAN Number'] =     header[ 1: 6].strip()
            # City where the station is located (maximum of 22 characters)
            self.data['City']        =     header[ 7:29].strip()
            # State where the station is located (abbreviated to two letters)
            self.data['State']       =     header[30:32].strip()
            # Time zone is the number of hours by which the local standard time is ahead of or behind Universal Time. For example, Mountain Standard Time is designated -7 because it is 7 hours behind Universal Time.
            self.data['Time Zone']   = int(header[33:36])
            # Latitude of the station, N = North of equator, Degrees, Minutes 
            self.data['Latitude']    = (   header[37].strip() or 'S',
                                           int(header[39:41]),
                                           int(header[42:44])) # hemisphere, degrees and minutes
            #  Longitude of the station, W = West, E = East, Degrees, Minutes
            self.data['Longitude']   = (   header[45].strip(),
                                           int(header[47:50]),
                                           int(header[51:53])) # meridian direction, degrees and minutes
            # Elevation of station in meters above sea level
            self.data['Elevation']   = int(header[55:59])
            
            # data elements (selected values only)
            self.data['data'] = []
            for record in dataFile.readlines():
                recordData = {}
                # hour of year [0,8760]
                yearDelta = (datetime( year=int(record[ 1: 3]),
                                      month=int(record[ 3: 5]),
                                        day=int(record[ 5: 7]),
                                       hour=int(record[ 7: 9]) - 1)
                            -datetime( year=int(record[ 1: 3]),
                                      month=1,
                                        day=1))
                recordData['Local Standard Time'] = (yearDelta.days * 24.) + (yearDelta.seconds / 3600.)
                # Total amount of direct and diffuse solar radiation in Wh/m2 received on a horizontal surface during the 60 minutes preceding the hour indicated 
                recordData['Global Horizontal Radiation'] = (int(record[17:21]),     # Data Value: 0 - 1200
                                                                 record[21].strip(), # Flag for Data Source: 0 - 1200 
                                                             int(record[22]))        # Flag for Data Uncertainty: 0 - 9
                # Amount of solar radiation in Wh/m2 received within a 5.7° field of view centered on the sun during the 60 minutes preceding the hour indicated 
                recordData['Direct Normal Radiation'] = (int(record[23:27]),     # Data Value: 0 - 1100
                                                             record[27].strip(), # Flag for Data Source: A - H, ?
                                                         int(record[28]))        # Flag for Data Uncertainty: 0 - 9
                # Amount of solar radiation in Wh/m2 received from the sky (excluding the solar disk) on a horizontal surface during the 60 minutes preceding the hour indicated 
                recordData['Diffuse Horizontal Radiation'] = (int(record[29:33]),     # Data Value: 0 - 700
                                                                  record[33].strip(), # Flag for Data Source: A - H, ?
                                                              int(record[34]))        # Flag for Data Uncertainty: 0 - 9
                # Dry bulb temperature in tenths of °C at the hour indicated -500 to 500 = -50.0 to 50.0°C
                recordData['Dry Bulb Temperature'] = (int(record[67:71]),     # Data Value: -500 to 500
                                                          record[71].strip(), # Flag for Data Source: A - F
                                                      int(record[72]))        # Flag for Data Uncertainty: 0 - 9
                # Dew point temperature in tenths of °C at the hour indicated -600 to 300 = -60.0 to 30.0°C
                recordData['Dew Point Temperature'] = (int(record[73:77]),     #  Data Value: -600 to 300 
                                                           record[77].strip(), # Flag for Data Source: A - F
                                                       int(record[78]))        # Flag for Data Uncertainty: 0 - 9
                # Relative humidity in percent at the hour indicated
                recordData['Relative Humidity'] = (int(record[79:82]),     # Data Value: 0 - 100
                                                       record[82].strip(), # Flag for Data Source: A - F
                                                   int(record[83]))        # Flag for Data Uncertainty: 0 - 9                                             
                self.data['data'].append(recordData)
        finally:
            dataFile.close()

        self.Tdb = self.getDryBulbTemperature()
        self.Twb = self.getTAmbWetBulb()

        self.TAmbAvg = self.getMeanDryBulbTemperature()
        self.latitude = self.getLatitude()
        self.longitude = self.getLongitude()

        self.I = self.getTotalGlobalHorizontalRadiation()
        self.Idn = self.getTotalDirectNormalRadiation()

    def buildHourlyData(self):
        if self.name == "TMY2File":
            self.G = self.getGlobalHorizontalRadiation()
            self.Gd = self.getDiffuseHorizontalRadiation()
            self.TAmb = self.getDryBulbTemperature()

            self.longitude = self.getLongitude()    
            self.timeZone = 1.0
            
        else:
            self.TAmb = self.Tdb
            self.longitude = 0.0
            self.timeZone = 1.0

        self.location = Location(self.latitude, self.longitude, self.timeZone)

    def defaultPath(self, filePath):

        tmp = filePath.split("/")
        if len(tmp) > 0: 
            parts = tmp[-1].split("\\")
        else:
            parts = filePath.split("\\")
        if len(parts) > 0:
            fileName = parts[-1]
        else:
            fileName = ""

        return(APPDATA+"/weather/"+fileName)

### HELPER FUNCTIONS FOR reading self.data entries
    
    def getHourOfYear(self):
        return [d['Local Standard Time'] for d in self.data['data']]
    
    def getGlobalHorizontalRadiation(self):
        return [(d['Global Horizontal Radiation'][0] / 1000.) for d in self.data['data']]
    
    def getDirectNormalRadiation(self):
        return [(d['Direct Normal Radiation'][0] / 1000.) for d in self.data['data']]
    
    def getDiffuseHorizontalRadiation(self):
        return [(d['Diffuse Horizontal Radiation'][0] / 1000.) for d in self.data['data']]
    
    def getDryBulbTemperature(self):
        return np.array([(d['Dry Bulb Temperature'][0] / 10.) for d in self.data['data']])
    
    def getDewPointTemperature(self):
        return [(d['Dew Point Temperature'][0] / 10.) for d in self.data['data']]
    
    def getRelativeHumidity(self):
        return [(d['Relative Humidity'][0] / 100.) for d in self.data['data']]
            
    def getLatitude(self):
        return (self.data['Latitude'][0] == 'S' and -1. or 1.) * (float(self.data['Latitude'][1]) + (float(self.data['Latitude'][2]) / 60.))
    
    def getLongitude(self):
        return (self.data['Longitude'][0] == 'E' and -1. or 1.) * (float(self.data['Longitude'][1]) + (float(self.data['Longitude'][2]) / 60.))
    
    def getTotalGlobalHorizontalRadiation(self):
        return sum(self.getGlobalHorizontalRadiation())
    
    def getTotalDirectNormalRadiation(self):
        return sum(self.getDirectNormalRadiation())
    
    def getMeanDryBulbTemperature(self):
        return np.mean(self.getDryBulbTemperature())



### GENERATION OF SIMPLIFIED DATA  -------------------------------------------------------------------------------        

    def generateSimplifiedData(self):
        """Calculate a simplified data set based on global parameters."""

        # check and substitute by default values where necessar
        if self.latitude is None:
##            logWarning(_U("The geographic latitude has not been defined. \n40 deg North (Southern Europe) assumed as default"))
            self.latitude = 40.0

        if self.I is None or self.I < 100.0 or self.I > 3000.0:
            self.set_IDefault()

        if self.Idn < 100.0 or self.Idn > 3000.0:
            self.Idn = None

        if self.TAmbAvg is None:
            self.set_TAmbAvgDefault()
            
        # calculate hourly values for temperature
        self.Tdb = self.getTAmbDryBulb()
        self.Twb = self.getTAmbWetBulb()
            
    def set_IDefault(self):
        """Calculate the radiation estimates (yearly radiation by latitude)."""
        x = abs(self.latitude)
        self.I = -0.0002 * pow(x, 4.0) + \
                           0.2547 * pow(x, 3.0) - \
                           31.055 * pow(x, 2.0) + \
                           1245.4 * x - \
                           14419
##        logWarning(_U("The yearly incident solar radiation has not been defined. \n"\
##                     "Value calculated from default correlation: H [kWh/m2] = %s") % 
##                   self.I)
        
    def set_TAmbAvgDefault(self):
        x = abs(self.latitude)
        self.TAmbAvg = -0.0143 * pow(x, 2.0) + 0.4668 * x + 20.5
##        logWarning(_U("The yearly mean ambient temperature has not been defined. \nValue calculated from default correlation: TAmb [C] = %s") % \
##               self.TAmbAvg)
        
    def getTAmbDryBulb(self):
        """Creates a sine-profile for yearly and daily variation of ambient temperature"""
        TAmb = Status.int.createQ_t()
        
        amplitudeYear = 10.0 * self.latitude / 40.0
        shiftYear = 45.0 * 24.0
        
        amplitudeDay = 5.0
        shiftDay = 2.0
        
        for it in xrange(Status.Nt):
            xyear = (it - shiftYear + 0.5) / YEAR * math.pi * 2
            xday = (it % 24 - shiftDay + 0.5)*math.pi / 12
            TAmb[it] = self.TAmbAvg - amplitudeYear*math.cos(xyear) - amplitudeDay*math.cos(xday)

        return TAmb
                
    def getTAmbWetBulb(self):
        """For the moment default Twb = Tdb - 5. Should be substituted by a real correlation"""
        
        TAmb = self.Tdb - 5
        return TAmb


                                   


if __name__ == "__main__":
    filePath = "/home/david/Arbeit/energyXperts/Project/IDAE/ES-Barcelona-City-81800.tm2"
    weatherData = WeatherData('Testing Barcelona')
    weatherData.loadFromTMY2File(filePath)
