# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.einstein-energy.net)
#
#------------------------------------------------------------------------------
#
#	EXPORTDATA
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Exports/imports parts of the SQL database to XML files
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
import sys
import os
import xml.dom
import xml.dom.minidom
import wx
from GUI.status import Status
from messageLogger import *
from GUI.dialogImport import DialogImport
from auxiliary.auxiliary import toGettextAndUnicode as _U, toUTF8, toUnicode as U, toFloat
import constants as c
import xml.sax.saxutils as saxutils
from GUI.dialogGauge import DialogGauge
import database.dbio as dbio
from datetime import datetime, time
from lib.auxiliary.float import almostEqual

from sqlobject.converters import sqlrepr, lookupConverter


# Row names for those tables which should be unique. take care: all in lower case letters !!!
ROWNAMES = {"sproject":"ProjectName",
            "questionnaire": "Name",
            "dbfluid": "FluidName",
            "dbfuel": "FuelName",
            "auditor": "Name",
            "dbelectricitymix": "ShortName",
            "profiles": "name"}


ESCAPE_CHARS={'"': '&quot;', "'": '&apos;'}
UNESCAPE_CHARS={'&quot;': '"', '&apos;': "'"}

def openfilecreate(text, filetype ,style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT):
    # ask for file for exporting
    dialog = wx.FileDialog(parent=None,
                           message=text,
                           wildcard=filetype,
                           style=style)
    if dialog.ShowModal() != wx.ID_OK:
        return None

    return dialog.GetPath()


def error(text):
    frame = wx.GetApp().GetTopWindow()
    frame.showWarning(text)
    
def escape(text):
    """Return the text properly escaped for XML."""
    return saxutils.escape(text, ESCAPE_CHARS)

def unescape(text):
    """Return the escaped text properly unescaped."""
    return saxutils.unescape(text, UNESCAPE_CHARS)

class VersionError(Exception):
    """The given version is higher then the highest known one"""
    pass

#------------------------------------------------------------------------------
class ExportProject(object):
    """ exports einstein project in XML format, including the selected information
        required from the databases
    """
    
    def __init__(self, pid=None, outfile=None):
        PId = pid
        if PId is None:
            logError('ExportProject: PId missing')
            return

#..............................................................................
# creates a list of ID's of processes, fluids and fuels that are used in the project

        processIDs = Status.prj.getProcessList(key='QProcessData_ID', PId=pid)
        equipmentScheduleIDs = [-eq.id for eq in Status.prj.getEquipment(PId=pid)]
        (fluidIDs,fuelIDs) = Status.prj.getFluidAndFuelList(pid)
        eMixIDs = Status.prj.getElectricityMixList(pid)
        auditorIDs = Status.prj.getAuditorID(all=True) #gets all relevant auditor Ids (responsible, created by, modified by)
        dataSetIDs = Status.prj.getDataSetIDs(pid)
        nodeIDs = Status.prj.getNodeIDs(pid)

##        self.DBName = Status.main.conf.get('DB', 'DBName')

        if outfile is None:
            outfile = openfilecreate(text='Output file for exporting project',filetype='XML files (*.xml)|*.xml')
            if outfile is None:
                return None

        stool = dbio.get("stool")
        schemaVersion = stool.Version

        fd = open(outfile, 'w')
        fd.write('<?xml version="1.0" encoding="utf-8"?>\n')
        fd.write('<EinsteinProject pid="%s">\n' % (pid,))
        fd.write('<EinsteinSchema version="%s" />\n' % schemaVersion)

##        cursor.execute("SHOW TABLES FROM `%s`"%self.DBName)
##        tables = cursor.fetchall()

        TABLEINFO = [("questionnaire", "id"),
                      ("sproject", "ProjectID"),
                      ("salternatives", "ProjectID"),
                      ("cgeneraldata", "Questionnaire_id"),
                      ("qelectricity", "Questionnaire_id"),
                      ("uheatpump", "Questionnaire_id"),
                      ("qbuildings", "Questionnaire_id"),
                      ("qdistributionhc", "Questionnaire_id"),
                      ("qflow", "ProjectID"),
                      ("qfuel", "Questionnaire_id"),
                      ("qgenerationhc", "Questionnaire_id"),
                      ("qheatexchanger", "ProjectID"),
                      ("qprocessdata", "Questionnaire_id"),
                      ("qproduct", "Questionnaire_id"),
                      ("qrenewables", "Questionnaire_id"),
                      ("qsurfarea", "ProjectID"),
                      ("qwasteheatelequip", "ProjectID"),
                      ("tcacontingencies", "ProjectID"),
                      ("tcadetailedopcost", "ProjectID"),
                      ("tcadetailedrevenue", "ProjectID"),
                      ("tcaenergy", "ProjectID"),
                      ("tcainvestments", "ProjectID"),
                      ("tcanonreoccuringcosts", "ProjectID"),
                      ("tcageneraldata", "ProjectID"),
                      ("heatexchanger_pinchstream", "ProjectID"),
                      ("pinchstream", "ProjectID"),
                      ("process_streams_in", "ProjectID"),
                      ("streams_in", "ProjectID"),
                      ("process_streams_out", "ProjectID"),
                      ("streams_out", "ProjectID"),
                      ("qmdatafiles", "ProjectID")]



##        for field in tables:
##            tableName = field['Tables_in_%s'%self.DBName]
        for info in TABLEINFO:
            self.dumpProjectTable(fd, info, PId)
            
        objectIDs = processIDs + equipmentScheduleIDs
        if len(objectIDs)>0:
##            criterium = "WHERE qprocessdata_QProcessData_ID IN (%s)" % ','.join([str(id) for id in processIDs])
##            dumpTable( fd, 'process_schedules', criterium)
##            
##            criterium = "WHERE qprocessdata_QProcessData_ID IN (%s)" % ','.join([str(id) for id in processIDs])
##            dumpTable(fd, 'process_schedules_tolerance_offsets', criterium)
            
            criterium = "qprocessdata_QProcessData_ID IN (%s)" % ','.join([str(id) for id in objectIDs])
            keys = dumpTable(fd, 'process_periods', criterium, foreignKeys=['id','periods_id'])
            processPeriodIDs = keys['id']
            periodIDs        = keys['periods_id']
            
            if len(processPeriodIDs)>0:
                criterium = "process_periods_id IN (%s)" % ','.join([str(id) for id in processPeriodIDs])
                profileIDs = dumpTable(fd, 'process_period_profiles', criterium, foreignKeys=['profiles_id'])['profiles_id']
            
                if len(periodIDs)>0:
                    criterium = "id IN (%s)" % ','.join([str(id) for id in periodIDs])
                    dumpTable(fd, 'periods', criterium)
                
                if len(profileIDs)>0:
                    criterium = "id IN (%s)" % ','.join([str(id) for id in profileIDs])
                    dumpTable(fd, 'profiles', criterium)
                    
                    criterium = "profiles_id IN (%s)" % ','.join([str(id) for id in profileIDs])
                    intervalIDs = dumpTable(fd, 'profile_intervals', criterium, foreignKeys=['intervals_id'])['intervals_id']
                    
                    criterium = "id IN (%s)" % ','.join([str(id) for id in intervalIDs])
                    dumpTable(fd, 'intervals', criterium)

        if len(processIDs) > 0:
            criterium = "ProcessId IN %s" % (str(processIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'qcombschedules', criterium)
            
        if len(nodeIDs)>0:
            criterium = "id IN %s" % (str(nodeIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'qnode', criterium)
            
        if len(fuelIDs)>0:
            criterium = "id IN %s" % (str(fuelIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'dbfuel', criterium)

        if len(fluidIDs)>0:
            criterium = "id IN %s" % (str(fluidIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'dbfluid', criterium)
        
        if len(eMixIDs)>0:
            criterium = "id IN %s" % (str(eMixIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'dbelectricitymix', criterium)
             
        if len(dataSetIDs)>0:
            criterium = "id IN %s" % (str(dataSetIDs),)
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'qmdatasets', criterium)
             
        auditorList = []
        for auditorId in auditorIDs:
            if auditorId is not None:
                auditorList.append(int(auditorId)) #note: int was necessary, as otherwise 1L - values led to problems in sql export
                
        if len(auditorList) > 0:
            criterium = "id IN %s" % (str(auditorList))
            criterium = criterium.replace('[','(').replace(']',')')
            dumpTable(fd, 'auditor', criterium)

        fd.write('</EinsteinProject>\n')
        fd.close()
        logMessage(_("XML file %s created"%(outfile)))


    def dumpProjectTable(self, fd, tableInfo, PId):
        """Dump all the rows of the table table in which there is a 
        reference to the project ID pid into the file fd as xml. 
        
        :param cursor: The database cursor.
        :param fd: The out file file descriptor.
        :param table: The name of the table.
        :param PId: The ProjectID/Questionnaire_ID
        
        """
        table = tableInfo[0]
        PIdKey = tableInfo[1]
        
        if PIdKey == "ProjectID":
            rows = dbio.select(table, ProjectID=PId)
        elif PIdKey == "Questionnaire_id":
            rows = dbio.select(table, Questionnaire_id=PId)
        elif PIdKey == "Questionnaire_ID" or PIdKey == "id":
            rows = dbio.select(table, id=PId)
        else:
            rows = []

        dumpRows(fd, table, rows)
        
def dumpTable(fd, table, criterium=None, order=None, foreignKeys=[]):
    """ Writes the content of a table to the XML file filtered by "criterium"
    """

    if criterium is None or criterium=="":
        rows = dbio.select(table)
        
    else:
        if order:
            criterium += (' ' + order)

        rows = dbio.selectByQuery(table, criterium)
    foreignKeyValues = dumpRows(fd, table, rows, foreignKeys)
        
    return foreignKeyValues

def dumpRows(fd, table, rows, foreignKeys=[]):
    """ auxiliary function used in dumpTable and dumpProjectTable
        exports a list of selected rows to the xml-file
    """
    foreignKeyValues = {}
    for foreignKey in foreignKeys:
        foreignKeyValues[foreignKey] = []

    nrows = len(rows)
    if nrows <= 0:
        fd.write('<!-- table %s has no values -->\n' % (table,))
    else:
        fd.write('<table name="%s">\n' % (table,))
        nn = 0
        for row in rows:
            nn+=1
            fd.write('<row n="%s">\n' % (nn,))
            colInfo = dbio.getColInfo(table)
            for key, valType in colInfo:
                value = getattr(row, key)
                if key in foreignKeys:
                    foreignKeyValues[key].append(value)
                    
                if value is not None:
                    if isinstance(value, str):
                        value = escape(value)
                    if key == "id":
                        extra = "auto_increment"
                    else:
                        extra = ""
                    s = '<element name="%s" type="%s" auto="%s" value="%s" />\n' % (key,valType,extra,value)
#                        s = '<%s>%s</%s>\n' % (key,value,key)
                    fd.write(s)
            fd.write('</row>\n')
        fd.write('</table>\n')
        
    return foreignKeyValues

class ImportProject(object):
    """ complement to ExportProject: imports data from another project
    #------------------------------------------------------------------------------
    #
    #oldKey: es el PRIMARY KEY de la tabla en el file.xml
    #       (p.ej. tabla qdistributionhc -> oldKey = valor de QDistributionHC_ID)
    #newKey: es el PRIMARY KEY de la misma tabla, una vez importada y colocada en la SQL
    #       (allï¿œ se hace un auto-increment del QDistributionHC_ID).
    #
    #
    """
    
    def __init__(self,infile=None):
        self.pid = None
        self.newpid = None

        if infile is None:
            infile = openfilecreate('Choose a project file for importing','XML files (*.xml)|*.xml',
                           style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
            if infile is None:
                return None
        
        dlg = DialogGauge(Status.main, _U("EINSTEIN project import"), _U("loading file")) if Status.main else None
        self.document = xml.dom.minidom.parse(infile)
        if dlg: dlg.update(20)
        schema = self.document.getElementsByTagName('EinsteinSchema')
        if dlg: dlg.update(30, _U("loading file into the database"))
        
        try:
            schemaVersion = schema[0].getAttribute('version')
            if schemaVersion > c.VERSION:
                showWarning(_("The project to be imported has been created with a higher version (%s) than the one you are currently using (%s)\n")%(schemaVersion,c.VERSION)+
                            _("The imported project may not run correctly"))
        except IndexError:
            schemaVersion = None 

        # get the elements from the DOM
        el = XMLDataElement() # working space for the data readin
        self.projectdict = {}
        newPIds = {}
        
        projects = self.document.getElementsByTagName('EinsteinProject')
        increment = 50.0 / float(len(projects))
        subIncrementCtr = 0

        importErrors = []
        
        for project in projects:
            self.pid = project.getAttribute('pid')

            # create a new entry in the receiving data base
            self.newpid = dbio.insertRow("questionnaire")
            questionnaireTable = dbio.get("questionnaire", id=self.newpid)
            dbio.commit()
            newPIds.update({self.pid: self.newpid})

            existingProfiles = []   # temporary storage for already existing profile ids
            newProfileIntervals = []

            tables = project.getElementsByTagName("table")
            tabledict = {}
            subincrement = increment / float(len(tables))
            for table in tables:
                tableName =  str(table.getAttribute('name'))

                # backwards compatibility: import of XML-files from older versions, containing tables that do not exist
                if tableName in ["process_schedules"]:
                    logMessage("XML-Import: Data for table %s skipped (no longer used in current version)"%tableName)
                    continue
                
                tablelist = []
                rows = table.getElementsByTagName("row")
                for row in rows:
                    sqldict = {}
                    elements = row.getElementsByTagName("element")
                    rowName = None #characteristic name of the row
                    
                    for element in elements:
                        el.read(element)
                        fieldname = el.name.lower()
                        fieldnameCAPS = el.name

                        if tableName == "sproject" and fieldname == "projectname":
                            xmlName = el.value
                            rowName = setUniqueName("sproject", xmlName)
                            el.value = rowName
                            if rowName != xmlName:
                                showWarning(_U("Project with name %s already exists in database\nProject renamed")%(xmlName))
                                
                        elif tableName == "intervals":
                            if fieldname == "scale":
                                rowName = el.value
                            elif fieldname == "start":
                                rowPar1 = el.value
                            elif fieldname == "stop":
                                rowPar2 = el.value

                        elif tableName in ROWNAMES.keys() and fieldname == ROWNAMES[tableName].lower():
                            rowName = el.value
                                
                        # substitute new pid in id field
                        if fieldname == 'projectid' or \
                             fieldname == 'questionnaire_id':
                            el.value = self.newpid
                            
                        # special handling for primary key
                        if el.auto:
                            oldKey = el.id
                        else:
                            sqldict.update({el.name:el.value})

#......................................................................
# before inserting new entry, check if entry with the same name exists
# (only for questionnaire, dbfluid, dbfuel and some others ...

                    ignoreRow = False
                    differences = []
                    if tableName == "dbfluid":
                        existingRow = dbio.get("dbfluid", FluidName=dbio.check(rowName))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            if len(differences) > 0:
                                message = _U("Fluid %s already in database. Data from imported file will be ignored")%rowName
                            ignoreRow = True
                            newID = existingRow.id

                    elif tableName == "dbfuel":
                        existingRow = dbio.get("dbfuel", FuelName=dbio.check(rowName))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            if len(differences) > 0:
                                message = _U("Fuel %s already in database. Data from imported file will be ignored")%rowName
                            ignoreRow = True
                            newID = existingRow.id
                    elif tableName == "auditor":
                        existingRow = dbio.get("auditor", Name=dbio.check(rowName))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            if len(differences) > 0:
                                message = _U("Auditor %s already in database. Data from imported file will be ignored")%rowName
                            ignoreRow = True
                            newID = existingRow.id
                    elif tableName == "dbelectricitymix":
                        existingRow = dbio.get("dbelectricitymix", ShortName=dbio.check(rowName))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            if len(differences) > 0:
                                message = _U("Electricity mix %s is already in the database. This data from imported file will be ignored")%rowName
                            ignoreRow = True
                            newID = existingRow.id
                    elif tableName == "profiles":
                        existingRow = dbio.get("profiles", name=dbio.check(rowName))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            if len(differences) > 0:
                                message = _U("Profile %s is already in the database. This data from imported file will be ignored")%rowName
                            ignoreRow = True
                            newID = existingRow.id
                            existingProfiles.append(oldKey)
                    elif tableName == "intervals":
                        existingRow = dbio.get("intervals",
                                               scale=dbio.check(rowName),
                                               start=dbio.check(rowPar1),
                                               stop=dbio.check(rowPar2))
                        if existingRow is not None:
                            differences = compareRows(existingRow, sqldict)
                            ignoreRow = True
                            newID = existingRow.id
                    else:
                        existingRow = None

                    if ignoreRow == False:
                        if tableName == "questionnaire":
                            newKey = self.newpid
                            questionnaireTable.set(**sqldict)
                        elif tableName == "profile_intervals":
                            newKey = oldKey
                            newProfileIntervals.append(sqldict)
                        else:
                            try: # assure that import continues even if non-existing or corrupt entries are in the
                                 # existing file
                                newKey = dbio.insertRow(tableName, **sqldict)
                                assert(newKey is not None)  # raise exception if no new key could be assigned
                            except:
                                importErrors.append("%s: %s"%(tableName, sqldict))
                                newKey = None
                    else:
                        if len(differences) > 0:
                            msgDifferences = U("\n".join(differences))
                            showWarning(message)
                            showWarning(msgDifferences)

                        newKey = newID
                    tablelist.append((oldKey,newKey))
                tabledict[tableName] = tablelist
                subIncrementCtr += subincrement
                if subIncrementCtr > 5.0:
                    if dlg: dlg.add(subIncrementCtr)
                    subIncrementCtr = 0

            # now write entries for table profile_intervals for those profiles which are new
            tablelist = []
            for i, sqldict in enumerate(newProfileIntervals):
                oldProfileId = sqldict["profiles_id"]
                oldKey = tabledict["profile_intervals"][i][0]
                if oldProfileId not in existingProfiles:
                    newKey = dbio.insertRow("profile_intervals", **sqldict)
                    tablelist.append((oldKey,newKey))
            tabledict["profile_intervals"] = tablelist
            dbio.commit()
            
            self.projectdict[self.pid] = tabledict

        if dlg: dlg.update(80, _U("adjusting links"))


        openPId = None
        from modules.upgrade import cleanUpHXConnections
        for PId in self.projectdict.keys():
            # upgrade projects to current EINSTEIN version - preparation before restoring links
            if schemaVersion < "2.9.00":
                psKeys = self.projectdict[PId].get("pinchstream")
                if psKeys is not None:
                    psIds = [key[1] for key in psKeys]
                    cleanUpHXConnections(psIds)

            Status.prj.restoreLinks(newPIds[PId],self.projectdict[PId])
            openPId = int(newPIds[PId])
            
        # upgrade projects to current EINSTEIN version if necessary
        from modules.upgrade import upgradeEINSTEIN
        if dlg: dlg.update(90)
        upgradeEINSTEIN(sourceVersion=schemaVersion, projectIds=newPIds.values())
        if dlg: dlg.Destroy()

        for err in importErrors:
            logError("Import error: %s"%err)

        if openPId:
            Status.prj.setActiveProject(openPId)

    def getPid(self):
        # returns the pid as stored in the xml file,
        # and the new pid generated when stored in the database.
        return (self.pid, self.newpid)

    def getDict(self):
        # returns the id correspondence dictionary, which has the form:
        #{pid_1:
        # {table_1:
        #  [(oldkey_row_0, newkey_row_0), (oldkey_row_1, newkey_row_1), ... ]
        # {table_2:
        #  [(oldkey_row_0, newkey_row_0), (oldkey_row_1, newkey_row_1), ... ]
        #  .
        #  .
        #  .
        #  },
        #  pid_2:
        #  ...
        # }
        # that is:
        # 1. A dictionary of projects. There can be more than one project
        #    in the xml file.
        #    The key is the pid
        #    The value is:
        # 2.   A dictionary of tableNames.
        #      The key is the table name
        #      The value is a list of tuples, one for each row, containing:
        #      (old key, new key)
        #      for the principal key of the row (in general called 'tableName_id')
        #        The old key is the value contained in the xml file.
        #        The new key is the value autogenerated when the row has been stored.
        #
        return self.projectdict

#------------------------------------------------------------------------------
class ExportDB(object):
#------------------------------------------------------------------------------
#
# exports einstein databases
#
#------------------------------------------------------------------------------
    def __init__(self, dbname="all", outfile=None):

        if outfile is None:
            outfile = openfilecreate(text='Output file for exporting data base',filetype='XML files (*.xml)|*.xml')
            if outfile is None:
                return None

        stool = dbio.get("stool")
        schemaVersion = stool.Version
            
        fd = open(outfile, 'w')
        fd.write('<?xml version="1.0" encoding="utf-8"?>\n')
        fd.write('<EinsteinDataBase dbname="%s">\n' % (dbname,))
        fd.write('<EinsteinSchema version="%s" />\n' % schemaVersion)

        if dbname == "all":
            dumpTable(fd, "dbfluid")
            dumpTable(fd, "dbfuel")
            
            dumpTable(fd, "dbboiler")
            dumpTable(fd, "dbchp")
            dumpTable(fd, "dbheatpump")
            dumpTable(fd, "dbsolarthermal")
            
#            dumpTable(cursor, fd, "dbbenchmark", "")
            
        elif dbname == "all equipments":
            dumpTable(fd, "dbboiler")
            dumpTable(fd, "dbchp")
            dumpTable(fd, "dbheatpump")
            dumpTable(fd, "dbsolarthermal")
            
        else:
            dumpTable(fd, dbname)

        fd.write('</EinsteinDataBase>\n')
        fd.close()
        logMessage(_("XML file %s created"%(outfile)))

#------------------------------------------------------------------------------
class ImportDB(object):
    """ complement to ExportDB: imports database files
    """

    def __init__(self,mode="ignore",infile=None, auto=False):

        if infile is None:
            infile = openfilecreate('Choose a project file for importing','XML files (*.xml)|*.xml',
                           style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
            if infile is None:
                return None

#......................................................................
# ask for confirmation before causing a major desaster

        if not auto:
            dialog =  DialogImport(None,_U("import data"),\
                               _U("This will modify your databases\nIf You are sure, specify what to do with duplicate data")+\
                               _U("\nElse press CANCEL"))
            ret = dialog.ShowModal()
            if ret == wx.ID_OK:
                mode = "overwrite"
            elif ret == wx.ID_IGNORE:
                mode = "ignore"
            else:
                return

        logTrack("ImportDB in mode %s"%mode)

        el = XMLDataElement() # working space for the data readin
        # create a dom and import in it the xml project file
        self.document = xml.dom.minidom.parse(infile)
        # get the elements from the DOM
        dbs = self.document.getElementsByTagName('EinsteinDataBase')
        for db in dbs:
            tables = db.getElementsByTagName("table")
            tabledict = {}
            for table in tables:
                tableName =  str(table.getAttribute('name'))
                tablelist = []
                rows = table.getElementsByTagName("row")
                for row in rows:
                    sqlist = []
                    sqldict = {}
                    newdict = {}
                    nrow =  row.getAttribute('n')
                    elements = row.getElementsByTagName("element")

                    rowName = None #characteristic name of the row
                    rowPar1 = None
                    rowPar2 = None
                    
                    for element in elements:
                        el.read(element)
                        fieldname = el.name.lower()
                        fieldnameCAPS = str(el.name)
                        fieldnameCapitalLetters = el.name

                        if tableName == "dbfluid" and fieldname == "fluidname":
                            rowName = el.value
                        elif tableName == "dbfuel" and fieldname == "fuelname":
                            rowName = el.value
                        elif tableName == "auditor" and fieldname == "name":
                            rowName = el.value
                        elif tableName == "dbelectricitymix":
                            if fieldname == "shortname":
                                rowName = el.value
                            elif fieldname == "year":
                                rowPar1 = el.value
                        elif tableName == "dbboiler":
                            if fieldname == "boilermanufacturer":
                                rowName = el.value
                            elif fieldname == "boilermodel":
                                rowPar1 = el.value
                        elif tableName == "dbchp":
                            if fieldname == "chpequip":
                                rowName = el.value
                        elif tableName == "dbheatpump":
                            if fieldname == "hpmanufacturer":
                                rowName = el.value
                            elif fieldname == "hpmodel":
                                rowPar1 = el.value
                        elif tableName == "dbsolarthermal":
                            if fieldname == "stmanufacturer":
                                rowName = el.value
                            elif fieldname == "stmodel":
                                rowPar1 = el.value

                        # special handling for primary key
                        if el.auto:
                            oldKey = el.id
                        else:
                            sqldict.update({el.name:el.value})
                            newdict[fieldnameCapitalLetters] = el.value
                        
#......................................................................
# before inserting new entry, check if entry with the same name exists
# (only for questionnaire, dbfluid, dbfuel

                    ignoreRow = False
                    existingRow = None
                    if tableName == "dbfluid":
                        existingRow = dbio.get("dbfluid", FluidName=dbio.check(rowName))

                    elif tableName == "dbfuel":
                        existingRow = dbio.get("dbfuel", FuelName=dbio.check(rowName))

                    elif tableName == "auditor":
                        existingRow = dbio.get("auditor", Name=dbio.check(rowName))

                    elif tableName == "dbelectricitymix":
                        if rowPar1 is None:
                            existingRow = dbio.get("dbelectricitymix", Country=dbio.check(rowName))
                        else:
                            existingRow = dbio.get("dbelectricitymix", Country=dbio.check(rowName), Year=int(rowPar1))
                            
                    elif tableName == "dbboiler":
                        if rowPar1 is None:
                            existingRow = dbio.get("dbboiler", BoilerManufacturer=dbio.check(rowName))
                        else:
                            existingRow = dbio.get("dbboiler", BoilerManufacturer=dbio.check(rowName), BoilerModel=dbio.check(rowPar1))
                            
                    elif tableName == "dbchp":
                        existingRow = dbio.get("dbchp", CHPequip=dbio.check(rowName))
                            

                    elif tableName == "dbheatpump":
                        if rowPar1 is None:
                            existingRow = dbio.get("dbheatpump", HPManufacturer=dbio.check(rowName))
                        else:
                            existingRow = dbio.get("dbheatpump", HPManufacturer=dbio.check(rowName), HPModel=dbio.check(rowPar1))
                            
                    elif tableName == "dbsolarthermal":
                        if rowPar1 is None:
                            existingRow = dbio.get("dbsolarthermal", STManufacturer=dbio.check(rowName))
                        else:
                            existingRow = dbio.get("dbsolarthermal", STManufacturer=dbio.check(rowName),STModel=dbio.check(rowPar1))
                            
   
                    else:
                        existingRow = None

                    if existingRow is not None:
                        differences = compareRows(existingRow, sqldict)
                        
                        if rowPar2 is None:
                            par2 = ""
                        else:
                            par2 = " " + str(rowPar2)

                        if rowPar1 is None:
                            par1 = ""
                        else:
                            par1 = " "+str(rowPar1)

                        if len(differences) > 0:
                            if mode == "overwrite":
                                existingRow.set(**sqldict)
                                message = "%s: %s%s%s"%(tableName,dbio.check(rowName),par1,par2)+\
                                           _U(" already in database. Data in database will be updated")
                            else:
                                message = "%s: %s%s%s"%(tableName,dbio.check(rowName),par1,par2)+\
                                           _U(" already in database. Data from imported file will be ignored")
                            msgDifferences = "\n".join(differences)
                            showWarning(message + "\n" + msgDifferences)
                                             
                    else:
                        dbio.insertRow(tableName, **sqldict)
                        
            dbio.commit()                                                


def compareRows(actualRow, newRow, ignore="\r", replace=["\n", "\t"]):
    """ auxiliary function for importDB
        compares one by one the values for corresponding entries

    """
    differences = []
    for key in newRow.keys():
        actualValue = getattr(actualRow, key)
        newValue = newRow[key]

    ### TODO: currently \n characters are substituted by spaces in the xml-parser
    # this leads to loss of information when importing text !!!

##        if isinstance(actualValue, str) and actualValue.startswith("Ger"):
##            for i, c in enumerate(actualValue):
##                d = newValue[i] if i < len(newValue) else "#"
##                print "%s %2x - %s %2x"%(c, ord(c), d, ord(d))

        newValueFiltered = newValue
        if isinstance(newValue, str):
            for c in ignore:
                newValueFiltered = newValueFiltered.replace(c, "")
            for c in replace:
                newValueFiltered = newValueFiltered.replace(c, " ")

        actualValueFiltered = actualValue
        if isinstance(actualValue, str):
            for c in ignore:
                actualValueFiltered = actualValueFiltered.replace(c, "")
            for c in replace:
                actualValueFiltered = actualValueFiltered.replace(c, " ")

        if actualValueFiltered != newValueFiltered:
            if isinstance(actualValue, float) and almostEqual(actualValue, newValue):
                pass
            else:
                differences.append(_("Parameter %s: current %s | imported file %s")%(key, actualValue, newValue))
            
    return differences

class XMLDataElement():
    """ Class for conversion of data formats between the format as read by the xml-parser
        and the one used internally
    """
    
    def __init__(self):
        self.auto = None
        self.value = None
        self.xmlValue = None
        self.type = None
        self.name = None
        self.id = None

    def read(self, element):
        """Checks the data contained in element after read in from the
            XML file and converts appropriately to internal formats
        """
        self.name = toUTF8(element.getAttribute('name'))
        self.type = toUTF8(element.getAttribute('type')).lower()
        self.auto = (element.getAttribute('auto')=="auto_increment")
        self.xmlValue = toUTF8(element.getAttribute('value'))

        # substitute invalid chars in char fields and enclose in ''
        if self.type.startswith('char') or \
           self.type.startswith('varchar') or \
           self.type.startswith('text') or \
           self.type.startswith('enum'):

            self.value = unescape(self.xmlValue)

        elif self.type.startswith('date'):
            self.value = subsIllegal(self.xmlValue)
            self.value = datetime.strptime(self.value, "%Y-%m-%d")
            
        elif self.type.startswith('time'):
            self.value = subsIllegal(self.xmlValue)
            self.value = datetime.strptime(self.value, "%H:%M:%S").time()
            
        elif self.type.startswith('double') or self.type.startswith('float'):
            self.value = toFloat(self.xmlValue)

        elif self.type.startswith('int') or self.type.startswith('tinyint'):
            self.value = int(self.xmlValue)

        # substitute auto-increment value with NULL
        if self.auto:
            # main key field
            self.id = int(self.value)
            self.value = None
            
        return self.value
            
		
def subsIllegal(text):
    parts = text.split("'")
    newtext = "''".join(parts)
    return newtext

def setUniqueName(tableName, xmlName):
    """ Changes a name adding a number in parenthesis to the original name, so long until
        the name is unique in the data base
    """

    if tableName in ROWNAMES.keys():
        nameKey = ROWNAMES[tableName]
        
        for ctr in range(100):
            if ctr == 0:
                rowName = xmlName
            else:
                rowName = xmlName + " (%d)"%ctr

            existingRow = dbio.get(tableName, **{nameKey: dbio.check(rowName)})
            if existingRow is None:
                break
        return rowName
    
    else: # if there is no nameKey, no check can be carried out
        return xmlName
