# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.einstein-energy.net/" target="_blank">www.einstein-energy.net</a>)
#
#------------------------------------------------------------------------------
#
#	UNITS: dictionaries and functions for unit conversion
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from GUI.status import Status
from modules.messageLogger import *
import numpy as np

CURRENCY = "EUR"

UNITSDENSITY = 1.0

# Basic units conversion
UNITS_LB = 0.45359
UNITS_FT = 0.3048
UNITS_BTU = 0.0002930711
UNITS_F = 5.0/9.0

GRAVITY = 9.80665

UNITS = {
# conversion to internal unit: [°C]
    'TEMPERATURE' : {
        '°C':   (1.0,   0.0),
        '°F':   (UNITS_F, -32.0*UNITS_F),
        'K':    (1.0,   -273.15)
        },

# conversion to internal unit: [K]
    'TEMPERATUREDIFF' : {
        'K':    (1.0,   0.0),
        '°F':   (UNITS_F, 0.0),
        },

# conversion to internal unit: [1/K]
    'INVTEMP' : {
        '1/K':    (1.0, 0.0),
        '1/°F':   (1.0/UNITS_F, 0.0),
        },

# conversion to internal unit: [1/K^2]
    'INVTEMP2' : {
        '1/K^2':    (1.0, 0.0),
        '1/°F^2':   (pow(UNITS_F, -2), 0.0),
        },

# conversion to internal unit: [h]
    'TIME' : {
        'h' : (1.0,0.0),
        'min' : (1.0/60,0.0),
        's' : (1.0/3600,0.0)
        },

# conversion to internal unit: [a]
    'LONGTIME' : {
        'a' : (1.0,0.0),
        },

# conversion to internal unit: [m]
    'LENGTH' : {
        'm' :       (1.0,       0.0),
        'km' :      (1000.0,    0.0),
        'mm' :      (1.0e-3,    0.0),
        'yards':    (0.9144,    0.0),
        'ft'   :    (UNITS_FT,    0.0),
        'in'   :    (0.0254,    0.0)
        },
    
# conversion to internal unit: [bar]
    'PRESSURE' : {
        'bar' : (1.0,0.0),
        'mbar' : (1.0e-3,0.0),
        'atm' : (1.01325,0.0),
        'mm Hg' : (0.001333224,0.0),
        'Pa' : (1.0e-5,0.0),
        'kPa' : (1.0e-2,0.0),
        'MPa' : (10.0,0.0),
        'lb/ft2': (1.0e-5*UNITS_LB*GRAVITY/(UNITS_FT*UNITS_FT),0.0),
        },
    
# conversion to internal unit: [m2]
    'AREA': {
        'm2' : (1.0,0.0),
        'ft2' : (UNITS_FT*UNITS_FT,0.0),
        'ha' : (1.0e4,0.0),
        },

# conversion to internal unit: [m3]
    'VOLUME': {
        'm3' : (1.0,0.0),
        'l' : (1.0e-3,0.0),
        'ft3' : (0.028317,0.0),
        'U.S. gal' : (0.003785,0.0),
        'Brit. gal' : (0.004546,0.0),
        'barrel (U.S. pet.)' : (0.15898,0.0)
        },

# conversion to internal unit: [m3/h]
    'VOLUMEFLOW' : {
        'm3/h' : (1.0,0.0),
        'm3/s' : (3600.0,0.0),
        'l/h' : (0.001,0.0),
        'l/s' : (3.6,0.0),
        'ft3/h'  : (0.028317,0.0)
        },

# conversion to internal unit: [kg]
    'MASS': {
        'kg' : (1.0,0.0),
        'lb' : (UNITS_LB,0.0),
        't' : (1000.0,0.0)
        },

# conversion to internal unit: [kg]
    'THERMALMASS': {
        'kWh/K' : (1.0,0.0),
        'btu/°F' : (UNITS_BTU/UNITS_F,0.0)
        },

# conversion to internal unit: [kg/h]
    'MASSFLOW': {
        'kg/h' : (1.0,0.0),
        'kg/s' : (3600.0,0.0),
        'lb/h' : (UNITS_LB,0.0),
        'lb/s' : (1632.924,0.0)
        },

# conversion to internal unit: [kg or m3]
    'MASSORVOLUME': {
        'kg' : (1.0,0.0),
        'lb' : (UNITS_LB,0.0),
        't' : (1000.0,0.0),
        'm3' : (1.0*UNITSDENSITY,0.0),
        'l' : (1.0e-3*UNITSDENSITY,0.0),
        'ft3' : (0.028317*UNITSDENSITY,0.0),
        'U.S. gal' : (0.003785*UNITSDENSITY,0.0),
        'Brit. gal' : (0.004546*UNITSDENSITY,0.0),
        'barrel (U.S. pet.)' : (0.15898*UNITSDENSITY,0.0)
        },

# conversion to internal unit: [kg/h]
    'MASSORVOLUMEFLOW': {
        'kg/h' : (1.0,0.0),
        'kg/s' : (3600.0,0.0),
        'lb/h' : (UNITS_LB,0.0),
        'lb/s' : (1632.924,0.0),
        'm3/h' : (1.0*UNITSDENSITY,0.0),
        'm3/s' : (3600.0*UNITSDENSITY,0.0),
        'l/h' : (0.001*UNITSDENSITY,0.0),
        'l/s' : (3.6*UNITSDENSITY,0.0),
        'ft3/h'  : (0.028317*UNITSDENSITY,0.0)
        },

# conversion to internal unit: [kg/h]
    'VOLUMEORMASSFLOW': {
        'kg/h' : (1.0,0.0),
        'kg/s' : (3600.0,0.0),
        'lb/h' : (UNITS_LB,0.0),
        'lb/s' : (1632.924,0.0),
        'm3/h' : (1.0,0.0),
        'm3/s' : (3600.0,0.0),
        'l/h' : (0.001*UNITSDENSITY,0.0),
        'l/s' : (3.6*UNITSDENSITY,0.0),
        'ft3/h'  : (0.028317*UNITSDENSITY,0.0)
        },

# conversion to internal unit: [kg or m3]
    'VOLUMEORMASS': {
        'kg' : (1.0/UNITSDENSITY,0.0),
        'lb' : (UNITS_LB/UNITSDENSITY,0.0),
        't' : (1000.0/UNITSDENSITY,0.0),
        'm3' : (1.0,0.0),
        'l' : (1.0e-3,0.0),
        'ft3' : (0.028317,0.0),
        'U.S. gal' : (0.003785,0.0),
        'Brit. gal' : (0.004546,0.0),
        'barrel (U.S. pet.)' : (0.15898,0.0)
        },

# conversion to internal unit: [kWh]
    'ENERGY': {
        'kWh' : (1.0,0.0),
        'MWh' : (1.0e+3,0.0),
        'GWh' : (1.0e+6,0.0),
        'kJ': (1.0/3600,0.0),
        'MJ': (1.0/3.6,0.0),
        'GJ': (1.0e6/3600,0.0),
        'kcal': (1.163/1000,0.0),
        'btu': (UNITS_BTU,0.0)
        },

# conversion to internal unit: [kWh]
    'ENERGYFLOW': {
        'kWh/m2a' : (1.0,0.0),
        'MJ/m2a': (1.0/3.6,0.0),
        'btu/h.ft2.a': (UNITS_BTU/(UNITS_FT*UNITS_FT),0.0)
        },

# conversion to internal unit: [kW]
    'POWER' : {
        'kW' : (1.0,0.0),
        'MW' : (1.0e3,0.0),
        'GW' : (1.0e6,0.0),
        'W' : (1.0e-3,0.0),
        'kcal/h' : (1.163e-3,0.0),
        'btu/h' : (UNITS_BTU,0.0)
        },

# conversion to internal unit: [kWh/kg]
    'SPECIFICENTHALPY' : {
        'kWh/kg' : (1.0,0.0),
        'kJ/kg' : (1.0/3600,0.0),
        'kcal/kg' : (1.163e-3,0.0),
        'btu/lb' : (UNITS_BTU/UNITS_LB,0.0)
        },
        
# conversion to internal unit: [kg/kWh]
    'CO2RATIO' : {
        'kg/kWh' : (1.0,0.0),
        'kg/kJ' : (3600,0.0),
        'kg/kcal' : (1/1.163e-3,0.0),
        'lb/btu' : (UNITS_LB/UNITS_BTU,0.0)
        },
        
# conversion to internal unit: [mg/kWh]
    'NUKESRATIO' : {
        'mg/kWh' : (1.0,0.0),
        'mg/kJ' : (3600,0.0),
        'mg/kcal' : (1/1.163e-3,0.0),
        'lb/btu' : (UNITS_LB/0.2930711,0.0)
        },

# conversion to internal unit: [kWh/kgK]
    'SPECIFICHEAT' : {
        'kWh/kgK' : (1.0,0.0),
        'kJ/kgK' : (1.0/3600.0,0.0),
        'kcal/kgK' : (1.163e-3,0.0),
        'btu/lb°F' : ((UNITS_BTU/(UNITS_LB*UNITS_F)),0.0)
        },

# conversion to internal unit: [kW/K]
    'HEATTRANSFERCOEF' : {
        'kW/K' : (1.0,0.0),
        'W/K' : (1.0e-3,0.0),
        'btu/h°F' : (UNITS_BTU/UNITS_F,0.0)
        },

# conversion to internal unit: [W/m2K]
    'HEATLOSSCOEFF' : {
        'kW/m2K' : (1.0e+3, 0.0),
        'W/m2K' : (1.0, 0.0),
        'btu/h.ft2°F' : (UNITS_BTU/(UNITS_F*pow(UNITS_FT,2)), 0.0),
        },
        
# conversion to internal unit: [W/m2K2]^2
    'HEATLOSSCOEFF2' : {
        'kW/m2K^2' : (1.0e+3, 0.0),
        'W/m2K^2' : (1.0, 0.0),
        'btu/h.ft2°F2' : (UNITS_BTU/pow(UNITS_FT*UNITS_F, 2), 0.0),
        },

# conversion to internal unit: [-]
    'ANGLE' : {
        '°' : (1.0,0.0),
        'rad' : (0.017453,0.0),
        },

# conversion to internal unit: [-]
    'FRACTION' : {
        '-' : (1.0,0.0),
        '%' : (0.01,0.0),
        },

# conversion to internal unit: [-]
# note: internally stored the same as FRACTION, but different default display unit 
    'PERCENTAGE' : {
        '-' : (1.0,0.0),
        '%' : (0.01,0.0),
        },

# conversion to internal unit: [1/a]
    'GROWTHRATE' : {
        '1/a' : (1.0,0.0),
        '%/a' : (0.01,0.0),
        },

# conversion to internal unit: [€]
    'PRICE': {
        '%s'%CURRENCY : (1.0,0.0),
        'k%s'%CURRENCY: (1000.0,0.0),
        'M%s'%CURRENCY: (1.e+6,0.0),
        },

# conversion to internal unit: [€/kW]
    'UNITPRICE' : {
        '%s/kW'%CURRENCY : (1.0,0.0),
        '%s/MW'%CURRENCY: (0.001,0.0),
        },

# conversion to internal unit: [€/kWh]
    'ENERGYTARIFF': {
        '%s/kWh'%CURRENCY : (1.0,0.0),
        '%s/MWh'%CURRENCY : (0.001,0.0),
        '%s/kJ'%CURRENCY: (3600.0,0),
        '%s/MJ'%CURRENCY: (3.6,0.0),
        '%s/GJ'%CURRENCY: (3.6e-3,0.0),
        '%s/btu'%CURRENCY: (1.0/UNITS_BTU,0.0)
        },
        
    'CURRENCY' : {
        '%s'%CURRENCY : (1.0,0.0)
        },

# conversion to internal unit: [kg/m2]
    'MASSPERAREA' : {
        'kg/m2' : (1.0, 0.0),
        'lb/ft2' : (UNITS_LB/(UNITS_FT*UNITS_FT), 0.0),
        },

# conversion to internal unit: [X]
#TODO: ELIMINATE IN DBEditor-Panels AND SUBSTITUTE BY MASSFLOW
    'MASSFLOWRATE' : {
        'kg/h' : (1.0,0.0),
        'kg/s' : (3600.0,0.0),
        'lb/h' : (UNITS_LB,0.0),
        'lb/s' : (1632.924,0.0)
        },

# conversion to internal unit: [kg/m3]
    'DENSITY' : {
        'kg/m3' : (1.0, 0.0),
        'lb/ft3' : (UNITS_LB/pow(UNITS_FT, 3), 0.0),
        },

# conversion to internal unit: [kWh/kgK]
#TODO: ELIMINATE IN DBEditor-Panels AND SUBSTITUTE BY SPECIFICHEAT
    'HEATCAPACITY' : {
        'kWh/kgK' : (1.0,0.0),
        'kJ/kgK' : (1.0/3600.0,0.0),
        'kcal/kgK' : (1.163e-3,0.0),
        'btu/lb°F' : ((UNITS_BTU/(UNITS_LB*UNITS_F)),0.0)
        },

# conversion to internal unit: [M2/H]
    'CINEMATICVISCOSITY' : {
        'm2/h' : (1.0, 0.0),
        'mm2/s': (3.6e-3),
        'St'   : (3.6e-1),
        'cSt'  : (3.6e-3)
        },
# conversion to internal unit: [Pa s]
    'DYNAMICVISCOSITY' : {
        'Pa s' : (1.0, 0.0),
        'P' : (0.1, 0.0),
        },

# conversion to internal unit: [kW/mK]
    'CONDUCTIVITY' : {
        'kW/mK' : (1.0, 0.0),
        'W/mK'  : (0.001, 0.0),
        'btu/h.ft°F'  : (UNITS_BTU/(UNITS_FT*UNITS_F), 0.0),
        },

# conversion to internal unit: [unit]
    'PU' : {
        'units' : (1.0, 0.0),
        't'     : (1.0, 0.0),
        'm3'    : (1.0, 0.0)
        },

# conversion to internal unit: [kWh/CU]
    'ENERGYPERCU' : {
        'kWh/%s'%CURRENCY : (1.0,0.0),
        'MWh/%s'%CURRENCY: (1000.0,0.0),
        'btu/%s'%CURRENCY : (UNITS_BTU,0.0),
        },

# conversion to internal unit: [X]
    'ENERGYPERPU' : {
        'kWh/p.u.': (1.0,0.0),
        'MWh/p.u.': (1000.0,0.0),
        'kWh/p.u.': (UNITS_BTU,0.0),
        },

# conversion to internal unit: [m2/s]
    'CINEMATICVISCOSITY' : {
#TODO: RENAME here and IN DBEditor-Panels to KINEMATICVISCOSITY
        'm2/s' : (1.0, 0.0),
        'mm2/s' : (1.0e-6, 0.0),        
        'St' : (1.0e-4, 0.0),        
        'cSt' : (1.0e-6, 0.0),        
        },
    }

UNITSYSTEM = {
    'SI' :      {
                'TEMPERATURE':'°C',
                'TEMPERATUREDIFF':'K',
                'INVTEMP' : '1/K',
                'INVTEMP2' : '1/K^2',
                'TIME':'h',
                'LONGTIME':'a',
                'LENGTH': 'm',
                'PRESSURE':'bar',
                'AREA':'m2',
                'VOLUME':'m3',
                'VOLUMEFLOW':'m3/h',
                'MASS':'kg',
                'THERMALMASS':'kWh/K',
                'MASSFLOW':'kg/h',
                'MASSORVOLUME':'kg',
                'MASSORVOLUMEFLOW':'kg/h',
                'VOLUMEORMASS':'m3',
                'VOLUMEORMASSFLOW':'m3/h',
                'ENERGY':'GJ',
                'ENERGYFLOW':'MJ/m2a',
                'POWER':'kW',
                'SPECIFICENTHALPY':'kWh/kg',
                'CO2RATIO' : 'kg/kWh',
                'NUKESRATIO': 'mg/kWh',
                'SPECIFICHEAT':'kJ/kgK',
                'HEATTRANSFERCOEF':'kW/K',
                'ANGLE':'°',
                'HEATLOSSCOEFF' : 'W/m2K',
                'HEATLOSSCOEFF2' : 'W/m2K^2',
                'FRACTION':'-',
                'PERCENTAGE':'%',
                'GROWTHRATE':'%/a',
                'ENERGYTARIFF':'%s/GJ'%CURRENCY,
                'PRICE':'%s'%CURRENCY,
                'UNITPRICE' : '%s/kW'%CURRENCY,
                'CURRENCY' : '%s'%CURRENCY,
                'MASSPERAREA' : 'kg/m2',
                'MASSFLOWRATE' : 'kg/h',
                'DENSITY' : 'kg/m3',
                'HEATCAPACITY' : 'kWh/kgK',
                'DYNAMICVISCOSITY' : 'Pa s',
                'CONDUCTIVITY' : 'W/mK',
                'PU' : 'units',
                'ENERGYPERCU' : 'kWh/%s'%CURRENCY,
                'ENERGYPERPU' : 'kWh/p.u.',
                'CINEMATICVISCOSITY' : 'm2/s',
                },
    
    'SI-kWh' :  {
                'TEMPERATURE':'°C',
                'TEMPERATUREDIFF':'K',
                'INVTEMP' : '1/K',
                'INVTEMP2' : '1/K^2',
                'TIME':'h',
                'LONGTIME':'a',
                'LENGTH': 'm',
                'PRESSURE':'bar',
                'AREA':'m2',
                'VOLUME':'m3',
                'VOLUMEFLOW':'m3/h',
                'MASS':'kg',
                'THERMALMASS':'kWh/K',
                'MASSFLOW':'kg/h',
                'MASSORVOLUME':'kg',
                'MASSORVOLUMEFLOW':'kg/h',
                'VOLUMEORMASS':'m3',
                'VOLUMEORMASSFLOW':'m3/h',
                'ENERGY':'MWh',
                'ENERGYFLOW':'kWh/m2a',
                'POWER':'kW',
                'SPECIFICENTHALPY':'kJ/kg',
                'CO2RATIO' : 'kg/kWh',
                'NUKESRATIO': 'mg/kWh',
                'SPECIFICHEAT':'kJ/kgK',
                'HEATTRANSFERCOEF':'kW/K',
                'ANGLE':'°',
                'HEATLOSSCOEFF' : 'W/m2K',
                'HEATLOSSCOEFF2' : 'W/m2K^2',
                'ANGLE':'°',
                'FRACTION':'-',
                'PERCENTAGE':'%',
                'GROWTHRATE':'%/a',
                'ENERGYTARIFF':'%s/MWh'%CURRENCY,
                'PRICE':'%s'%CURRENCY,
                'UNITPRICE' : '%s/kW'%CURRENCY,
                'CURRENCY' : '%s'%CURRENCY,
                'MASSPERAREA' : 'kg/m2',
                'MASSFLOWRATE' : 'kg/h',
                'DENSITY' : 'kg/m3',
                'HEATCAPACITY' : 'kWh/kgK',
                'DYNAMICVISCOSITY' : 'Pa s',
                'CONDUCTIVITY' : 'W/mK',
                'PU' : 'units',
                'ENERGYPERCU' : 'kWh/%s'%CURRENCY,
                'ENERGYPERPU' : 'kWh/p.u.',
                'CINEMATICVISCOSITY' : 'm2/s',
                },

    'BTU' :     {
                'TEMPERATURE':'°F',
                'TEMPERATUREDIFF':'°F',
                'INVTEMP' : '1/K',
                'INVTEMP2' : '1/K^2',
                'TIME':'h',
                'LONGTIME':'a',
                'LENGTH': 'ft',
                'PRESSURE':'lb/ft2',
                'AREA':'ft2',
                'VOLUME':'ft3',
                'VOLUMEFLOW':'ft3/h',
                'MASS':'lb',
                'THERMALMASS':'btu/°F', 
                'MASSFLOW':'lb/h',
                'MASSORVOLUME':'kg',
                'MASSORVOLUMEFLOW':'lb/h',
                'VOLUMEORMASS':'ft3',
                'VOLUMEORMASSFLOW':'ft3/h',
                'ENERGY':'btu',
                'ENERGYFLOW':'btu/h.ft2.a',
                'POWER':'btu/h',
                'SPECIFICENTHALPY':'btu/lb',
                'SPECIFICHEAT':'btu/lb°F',
                'HEATTRANSFERCOEF':'btu/h°F',
                'ANGLE':'°',
                'HEATLOSSCOEFF' : 'btu/h.ft2°F',
                'HEATLOSSCOEFF2' : 'btu/h.ft2°F2',
                'FRACTION':'-',
                'PERCENTAGE':'%',
                'GROWTHRATE':'%/a',
                'ENERGYTARIFF':'%s/btu'%CURRENCY,
                'PRICE':'%s'%CURRENCY,
                'UNITPRICE' : '%s/kW'%CURRENCY,
                'CURRENCY' : '%s'%CURRENCY,
                'MASSPERAREA' : 'lb/ft2',
                'MASSFLOWRATE' : 'lb/h',
                'DENSITY' : 'lb/ft3',
                'HEATCAPACITY' : 'btu/h.lb°F',
                'DYNAMICVISCOSITY' : 'Pa s',
                'CONDUCTIVITY' : 'btu/h.ft°F',
                'PU' : 'unit',
                'ENERGYPERCU' : 'btu/%s'%CURRENCY,
                'ENERGYPERPU' : 'btu/unit',
                'CINEMATICVISCOSITY' : 'ft2/s',
                }
}

#==================================================================================
#   UnitConversion class
#==================================================================================

class UnitConversion():
    def __init__(self, currency="EUR"):
        self.units = UNITS.copy()
        self.setCurrency(currency)

    def setCurrency(self, currency):
        self.currency = currency
        for uType in UNITS.keys():
            unitDict = UNITS[uType]
            newDict = {}
            for unit in unitDict.keys():
                newUnit = unit.replace(CURRENCY, currency)
                newDict.update({newUnit:unitDict[unit]})
            self.units[uType] = newDict

    def defaultUnit(self, unitType):
        """
        returns the default unit for a given unitType
        """
        if unitType in UNITSYSTEM[Status.Units].keys():
            unit = UNITSYSTEM[Status.Units][unitType]
            unit = unit.replace(CURRENCY, self.currency)
        else:
            return None

        return unit

    def conversionFactor(self, unit, unitType=None, funit2=None):
        """
        wrapper for global function conversionFactor
        """
        result = None
        
        if unitType is None:
            for unitType in self.units.keys():
                if unit in self.units[unitType].keys():
                    (a,b) = self.units[unitType][unit]
                    result = a
                    break
        else:
            (a,b) = self.units[unitType][unit]
            result = a
        
        if result is None:
            logWarning(_("Units (conversionFactor): unknown unit <%s>")%unit)
            result = 1.0
            
        #addon for combined unit systems
        if funit2 is not None and funit2 > 0:
            if  (unitType == "MASSORVOLUME" and unit in self.units["VOLUME"].keys()) or \
                (unitType == "MASSORVOLUMEFLOW" and unit in self.units["VOLUMEFLOW"].keys()) or \
                (unitType == "VOLUMEORMASSFLOW" and unit in self.units["MASSFLOW"].keys()) or \
                (unitType == "VOLUMEORMASS" and unit in self.units["MASS"].keys()):
                result *= funit2

        return result

    def internalValue(self, displayValue,unit,unitType,funit2=None):
        """
        calculates the internal value from the value entered on GUI
        """
        (a,b) = self.units[unitType][unit]    
        internalValue = a*displayValue + b

        #addon for combined unit systems
        if funit2 is not None and funit2 > 0:
            if  (unitType == "MASSORVOLUME" and unit in self.units["VOLUME"].keys()) or \
                (unitType == "MASSORVOLUMEFLOW" and unit in self.units["VOLUMEFLOW"].keys()) or \
                (unitType == "VOLUMEORMASSFLOW" and unit in self.units["MASSFLOW"].keys()) or \
                (unitType == "VOLUMEORMASS" and unit in self.units["MASS"].keys()):
                internalValue *= funit2


    #    print "internalValue d: %s i: %s unit: %s unitType: %s a: %s"%(displayValue,
    #                                                     internalValue,
    #                                                     unit,
    #                                                     unitType,
    #                                                     a)

        return internalValue
    
    def displayValue(self, internalValue,unit,unitType,funit2=None):
        """
        calculates the value to be displayed from the internally stored value
        """
        (a,b) = self.units[unitType][unit]
        displayValue = (internalValue - b)/a

        #addon for combined unit systems
        if funit2 is not None:
            if  (unitType == "MASSORVOLUME" and unit in self.units["VOLUME"].keys()) or \
                (unitType == "MASSORVOLUMEFLOW" and unit in self.units["VOLUMEFLOW"].keys()) or \
                (unitType == "VOLUMEORMASSFLOW" and unit in self.units["MASSFLOW"].keys()) or \
                (unitType == "VOLUMEORMASS" and unit in self.units["MASS"].keys()):
                displayValue /= funit2

##        print "displayValue d: %s i: %s unit: %s unitType: %s a: %s"%(displayValue,
##                                                         internalValue,
##                                                         unit,
##                                                         unitType,
##                                                         a)
        return displayValue
        
    def displayDecimals(self, internalDecimals, unit, unitType=None, funit2=None):
        """
        adapts number of decimals to be displayed to the conversion factor
        """
        
        cf = self.conversionFactor(unit, unitType=unitType, funit2=funit2)
        displayDecimals = max(int(internalDecimals + np.log10(cf) + 0.5), 0)
        return displayDecimals

    def getUnitTypes(self, unit):
        """
        gets the unit type (TEMPERATURE, MASSFLOW, ...) from the unit (°C, kg/h, ...)
        used for distinction of mass flow and volume flow, mass and volume, ...
        """
        
        unitTypes = []
        for key in UNITS.keys():
            if unit in self.units[key].values():
                unitTypes.append(key)

        return unitTypes
    
#==================================================================================
