# -*- coding: utf-8 -*-
#==============================================================================
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       www.einstein-energy.net
#
#------------------------------------------------------------------------------
#
#	EditCell
#
#==============================================================================
#
#	Version 3.0
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================

from Globals import logger
from view.units import internalValue, displayValue, displayDecimals, UNITS
from lib.auxiliary.float import toFloat, FloatIO

class EditCell():
    """
    Helper class for DataEntry and TableEntry

    To be used for read/write of any value (float, text, choice) from/to the GUI
    (or elsewhere)

    Carries out also unit conversion

    Basic parameters:

    dstr: string on the display (in- or output)
    dval: value corresponding to dstr
    val:  internal value

    ctype: cell type ("float", "int", "text", "choice")
    unit: unit to be displayed
    unitdict: dictionary containing possible units
    decimals: number of decimals to be displayed (refers to the internal unit; number of decimals is adapted to the display unit)

    dval = unitConversion(val), f(unit)
    dstr = representation(dval)

    noneStr: representation of none in the display (e.g. "" (empty cell; default), "?" (explicit presentation), "-", "n/a", ...)

    Differences to displayClasses:
    - the internal value remains fixed and unchanged when switching units, independently of rounding in the display units
    
    """

    def __init__(self, ctype,
                 unitType=None, # for floats only
                 decimals=2,    # for floats only
                 choices=[],    # for choices only
                 noneStr=""):   # string to be displayed for none-values

        self.ctype = ctype
        self.unitType = unitType
        self.unit = None
        self.funit2 = 1.0      # conversion factor for secondary unit (float only). Use 1.0 as default in order to avoid crash

        self.choice = choices

        self.noneStr = noneStr

        self.val = None
        self.dval = None        # dval only used for unit conversion for float entries
        self.dstr = self.noneStr

        self.decimals = decimals
        self.displayDecimals = None

        self.floatIO = FloatIO(tsep=" ", dc=".", snone=noneStr) # use space and dot for uniformity with displayClasses.FloatEntry

    def __repr__(self):
        return "%s -> %s %s:[%s]"%(self.val, self.dval, self.unit, self.dstr)
        
    def setUnit(self, unit):
        """
        sets the actual unit to be displayed
        """
        if unit in UNITS[self.unitType].keys():
            self.unit = unit
            self.displayDecimals = displayDecimals(self.decimals, self.unit, self.unitType, self.funit2)
        else:
            logger.logError("EditCell: cannot select unit %s")

        self.write(self.val) # leaves the value unchanged, but updates the dval/dstr corresponding to the new units

    def updateSecondaryUnit(self, funit2):
        """
        funit2: conversion factor for secondary unit (e.g. Mass- or volume flow ...)
        """
        self.funit2 = funit2
        self.write(self.val) # writes the own (internal) value again (this activates unit conversion and regeneration of a new dstr)
            
    def write(self, val):
        """
        writes any value in the required format
        if val does not correspond to the format, None is written
        """
        
        if val is None:
            self.val  = None
            self.dval = None
            self.dstr = self.noneStr
            return self.dstr

        if self.ctype=="float":
            self.val = toFloat(val)
            if self.val is None:
                self.write(None)
            else:
                self.unitConv()
##                formatString = "%" + ".%df"%self.displayDecimals
##                self.dstr = formatString%self.dval
                self.dstr = self.floatIO.write(self.dval, decimals=self.displayDecimals)
        
        elif self.ctype=="text":
            self.val = val
            self.dstr = val
            
        elif self.ctype=="int":
            self.val = int(val+0.5)
            self.dstr = str(self.val)

        elif self.ctype=="choice":
            if val >= 0 and val <= len(self.choice):
                self.val = val
                self.dstr = self.choice[val]
            else:
                self.write(None)
##        except:
##            logger.logError("EditCell: erroneous value %s for type %s"%(val, self.ctype))
##            self.write(None)
            
        return self.dstr
        
    def read(self, dstr):
        """
        converts the string into a corresponding internal value,
        or None, if conversion not possible

        overWrite:  if True, overwrites the stored values (dval, val),
                    otherwise just converts
        """
        if dstr is None or dstr.strip() == "" or dstr == self.noneStr:
            self.write(None)
            return None

        self.dstr = dstr

        if self.ctype=="float":
            dval = toFloat(dstr)
            if dval is None:
                self.write(None)
            else:
                self.dval = dval
                self.unitConvReverse()

        elif self.ctype=="int":
            try:
                self.dval = int(dstr)  
                self.val = self.dval
            except:
                self.write(None)
            
        elif self.ctype=="text":
            self.dval = dstr
            self.val = self.dval

        elif self.ctype=="choice":
            if dstr in self.choice:
                self.dval = dstr
                self.val = self.choice.index(dstr)
            else:
                self.write(None)


        return self.val

        
    def unitConv(self):
        """
        converts val to dval
        """
        self.dval = displayValue(self.val, self.unit, self.unitType, self.funit2)

    def unitConvReverse(self):
        """
        inverse of unitConv: converts dval to val
        """

        self.val = internalValue(self.dval, self.unit, self.unitType, self.funit2)
    
#------------------------------------------------------------------------------

        
