# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#    Expert System for an Intelligent Supply of Thermal Energy in Industry
#    www.einstein-energy.net
#
#------------------------------------------------------------------------------
#
#    drawingFrame: Generic frame for drawings
#
#==============================================================================
#
#    (C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2017
#    http://www.energyxperts.net/
#
#    This program is free software: you can redistribute it or modify it under
#    the terms of the GNU general public license as published by the Free
#    Software Foundation (www.gnu.org).
#
#==============================================================================
import wx
from GUI.status import *
from auxiliary.auxiliary import toUnicode as U, toGettextAndUnicode as _U
import wx.lib.scrolledpanel as scrolled
from GUI.GUITools import ORANGE, SOFTORANGE

from moduleIcon import ModuleIcon
from nodeEditor import NodeEditor

XMAXDRAWING = 800
YMAXDRAWING = 600

ICONSIZE = wx.Size(116,80)
GRIDSIZE = wx.Size(220, 120)
GRIDORIGIN = (50, 20)
YPOSMARGIN = GRIDORIGIN[1]

CONSIZE = (40, 12)  # size of connection mark
CONLABELWIDTH = 60  # width of connection label
CONLABELSHIFT = 18  # shift of connection label in y direction
CONTEXTSHIFT = 6    # shift of connection text in y direction
CONYSEP = 24       # vertical separation of connections
CONPOSIN = (20, 0)  # position of first connection label relative to default position (out of border of icon, YSEPCONS*(n+1))
CONPOSOUT = (-20, 0)

POPUPMENU =[_("real flow"), _("information flow"), _("delete connection")]
POPUPMENUIDS = [wx.NewId() for item in POPUPMENU]

class DrawingPanel(wx.lib.scrolledpanel.ScrolledPanel):
    def __init__(self, parent, mod, conStyle=0, size=None):
        """
        parent:
        mod: background module for calculations
        """

        self.emptyDrawing() # create space for lines, icons and labels
        
        super(DrawingPanel, self).__init__(parent)

        self.mode = 0       # state of the drawing:
                            # 0: default mode. just display or edit objects
                            # 1: in the middle of building a connection (startPoint selected, looking for end point)

        self.conStart = None
        self.conEnd = None
        self.startIcon = None
        self.startNo = None
        self.endIcon = None
        self.endNo = None
        self.conStyle = conStyle # 0: no area around connections. 1: area around
        
        # grid parameters
        self.iconSize = ICONSIZE
        self.gridSize = GRIDSIZE
        self.gridOrigin = GRIDORIGIN

        self.conSize = CONSIZE
        self.conPosIn = CONPOSIN
        self.conPosOut = CONPOSOUT
        self.conYSep = CONYSEP
        self.conLabelWidth = CONLABELWIDTH
        self.conLabelShift = CONLABELSHIFT
        self.conTextShift = CONTEXTSHIFT

        if size is None:
            self.xmaxDrawing = XMAXDRAWING
            self.ymaxDrawing = YMAXDRAWING
        else:
            self.xmaxDrawing, self.ymaxDrawing = size

        self.activeNodeData = None

        self.mod = mod

        self._init_ctrls(parent)

##        self.display()

        self.actionCode = None # dummy, used for pop-up menu



    def _init_ctrls(self, prnt):
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnMouseLeft)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnMouseRight)
        self.Bind(wx.EVT_MOTION, self.OnMouseMove)

        self.EnableScrolling(True, True)
        self.SetScrollbars(20, 20, self.xmaxDrawing/20, self.ymaxDrawing/20)

        self.parent = prnt

# sizer issues from StreamPlot
##        self.SetClientSize(wx.Size(STREAMPLOTWINDOWWIDTH+100, 600))
##
##        self.scrollPanel = scrolled.ScrolledPanel(self, pos = (100,0))
##        self.panel = wx.Panel(self.scrollPanel, style=wx.SUNKEN_BORDER, size = wx.Size(STREAMPLOTWINDOWWIDTH, 600))
##        sizer = wx.GridBagSizer(1, 1)
##        self.scrollPanel.SetSizer(sizer)
##        sizer.Add(self.panel, (1, 1))

##        self.SetupScrolling()


    def display(self):
        """ function to be called from outside for updating the plot (e.g. when project data have changed)
        """
        self.data = self.mod.updatePanel()
        
##        self.data = Status.int.GData[self.mod.key]
        if self.data:
            self.buildDrawing() # user function to build
        else:
            self.emptyDrawing()
            
        self._draw()


# dummy functions to be substituted by custom functions in order to obtain functionality
    def emptyDrawing(self):
        """ this function should be substituted by a custom function buildDrawing
            building the drawing
        """
        self.lines = []
        self.icons = []
        self.labels = []
        self.connections = []


    def buildConnection(self, fromIcon, fromNo, toIcon, toNo):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        pass

    def deleteConnection(self, n):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        return False

    def changeConnection(self, n, mode):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        return False

    def insertIcon(self, pos):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        return False

    def deleteIcon(self, icon):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function
        """
        return False

    def editIcon(self, icon):
        """ this function is called from the event handler for mouse clicks
            should be substituted by a custom function

            function should return None if icon is not editable,
            otherwise any object (node) with attributes:
            - nodeType
            - scaleFactor
            - splitFraction
        """
        node = None
        return None

    def saveIcon(self, nodeData):
        return
    
    def _draw(self):        

        # now bring the drawing to the panel

            
        self.SetVirtualSize((self.xmaxDrawing, self.ymaxDrawing))

        self.buffer = wx.EmptyBitmap(self.xmaxDrawing, self.ymaxDrawing)
        cdc = wx.ClientDC(self)
        self.PrepareDC(cdc)
        self.dc = wx.BufferedDC(cdc, self.buffer)

        self.dc.Clear()
        self.dc.BeginDrawing()

        for icon in self.icons:
            if icon is not None: icon.draw()
##            self.drawIcon(icon)

        for data in self.lines:
            if len(data)==4: # for backward compatibility
                x, y, colour, style = data
                width = 1
            else:
                x, y, colour, style, width = data
            self.drawLine(x, y, colour, style, width)

        for label in self.labels:
            self.drawLabel(label)

        if self.mode==1: # building a connection
            self.dc.SetPen(wx.Pen(wx.BLACK))
            self.dc.SetBrush(wx.Brush(SOFTORANGE))
            self.dc.DrawLine(self.conStart[0], self.conStart[1], self.conEnd[0], self.conEnd[1])
            self.dc.DrawRectangle(self.conEnd[0]-5, self.conEnd[1]-5, 10, 10)
        
        self.dc.EndDrawing()

        self.Refresh()


    def OnPaint(self, event=None):
        dc = wx.BufferedPaintDC(self, self.buffer, wx.BUFFER_VIRTUAL_AREA)

    def OnMouseLeft(self, event):
        pos = self.CalcUnscrolledPosition(event.GetPosition())


        # check if click is at some connection point of any icon
        iconFound = False
        for n, icon in reversed(list(enumerate(self.icons))):    # use reversed order as icons with higher n are in the foreground
            if icon is None:
                continue
                
            res = icon.clickedCon(pos)

            if res is None:
                continue
            
            if self.mode==0 and res < 0:
                self.conStart = pos
                self.startIcon = icon
                self.startNo = -res
                self.mode = 1
                iconFound = True
                break

            elif self.mode==1 and res > 0:
                self.conEnd = pos
                self.endIcon = icon
                self.endNo = res
                self.mode = 0
                self.buildConnection(self.startIcon, self.startNo, self.endIcon, self.endNo)
                self.buildDrawing()
                wx.SafeYield()
                self._draw()
                iconFound = True
                break

            elif self.mode==0 and res >=0:  # here the icon is selected for editing
                node=self.editIcon(icon)
                if node is not None:
                    dlg = NodeEditor(self, node)
                    ret = dlg.ShowModal()
                    wx.SafeYield()
                    if ret == wx.ID_OK:
                        self.saveIcon(self.activeNodeData)
                        self.buildDrawing()
                        wx.SafeYield()
                        self._draw()
                iconFound = True
                break

                
        if not iconFound:
            if self.mode != 1: # not in connection build mode and 
                conFound = False
                for n, con in enumerate(self.connections):
                    for line in con:
                        r0, r1, colour, style = line
                        xmin = min(r0[0], r1[0]) - 2
                        xmax = max(r0[0], r1[0]) + 2
                        ymin = min(r0[1], r1[1]) - 2
                        ymax = max(r0[1], r1[1]) + 2

                        if pos[0] >= xmin and pos[0] <= xmax:
                            if pos[1] >= ymin and pos[1] <= ymax:
                                # create pop-up
                                menu = wx.Menu()
                                for i, title in enumerate(POPUPMENU):
                                    id = POPUPMENUIDS[i]
                                    menu.Append( id, title )
                                    wx.EVT_MENU(menu, id, self.OnPopUp)

                                # Display menu with call to PopupMenu, invoked on the source component, passing event's GetPoint. ###
                                self.PopupMenu(menu, event.GetPosition())
                                menu.Destroy() # destroy to avoid mem leak
                                wx.SafeYield()

                                if self.actionCode in [0, 1]:
                                    ret = self.changeConnection(n, self.actionCode)
                                    
                                elif self.actionCode == 2:
                                    ret = self.deleteConnection(n)

                                else:
                                    ret = None
                                
                                if ret:
                                    self.buildDrawing()
                                    wx.SafeYield()
                                    self._draw()
                                    conFound = True
                                    break
                                
                    if conFound: break

            elif self.mode == 1: # interrupt connection building by clicking to some nowhere
                self.mode = 0
                wx.SafeYield()
                self._draw()
            
        event.Skip()

    def OnMouseRight(self, event):
        pos = self.CalcUnscrolledPosition(event.GetPosition())

        selectedIcon = None
        if self.mode==0:
            # check if click is at any icon
            for n, icon in enumerate(self.icons):
                if icon is None:
                    continue

                res = icon.clickedCon(pos)

                if res is not None:
                    selectedIcon = icon
                    break

        if selectedIcon is None:        # insert a new icon
            node = self.insertIcon(pos)
            if node is not None:
                self.buildDrawing()
                wx.SafeYield()
                self._draw()
                dlg = NodeEditor(self, node)

                ret = dlg.ShowModal()
                if ret == wx.ID_OK:
                    self.saveIcon(self.activeNodeData)
                    self.buildDrawing()
                    wx.SafeYield()
                    self._draw()
        else:                           # delete the selected icon
            ret = self.deleteIcon(selectedIcon)
            if ret:
                self.buildDrawing()
                wx.SafeYield()
                self._draw()


    def OnMouseMove(self, event):

        if self.mode==0: return

        pos = self.CalcUnscrolledPosition(event.GetPosition())
        self.conEnd = pos
        self._draw()
        event.Skip()

    def OnScrollChanged(self, event):
        xView, yView = self.GetViewStart()
        
    def OnPopUp(self, event):
        actionCode = POPUPMENUIDS.index(event.GetId())
        self.actionCode = actionCode
        
    def drawLine(self,x,y,colour=wx.BLACK, style=wx.SOLID, width=1):
        self.dc.SetPen(wx.Pen(colour, width, style=style))
        self.dc.DrawLine(x[0], x[1], y[0], y[1])

    def drawLabel(self, label):

        text, x,y, colour, style = label
        if colour is None: colour=wx.BLACK
        if style is None: style=wx.SOLID
        
        font = wx.Font(8, wx.MODERN, wx.NORMAL, wx.BOLD, False, u'Arial')
        self.dc.SetFont(font)
        self.dc.SetTextForeground(colour)

        self.dc.SetPen(wx.Pen(colour, 1, style=style))
        self.dc.DrawText(U(text), x, y)

        

