# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (<a href="http://www.iee-einstein.org/" target="_blank">www.iee-einstein.org</a>)
#
#------------------------------------------------------------------------------
#
#    drawingFrame: Generic frame for drawings
#
#==============================================================================
#
#    (C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2017
#    http://www.energyxperts.net/
#
#    This program is free software: you can redistribute it or modify it under
#    the terms of the GNU general public license as published by the Free
#    Software Foundation (www.gnu.org).
#
#==============================================================================
import wx
from auxiliary.auxiliary import toUnicode as U
from GUI.GUITools import SOFTORANGE

YSEPCONS = 20
XSEPCONS = 120
##CLICKTOLX = 30
CLICKTOLX = 5
CLICKTOLY = 4

BOXWIDTH = 4    # width in pixels of box for highlighting of icons (set colour)
BOXFGWIDTH = 2

class ModuleIcon(wx.Panel):
    """representation of any object on the graphical window

    attributes:
    pos:        x, y:   position of upper left corner of the icon
    iconSize:   dx, dy: size in x and y direction
    conPosIn/Out:   relative coordinates of connection frame with respect to
                    default position just outside of icon border
    conSize:        size of connection frame
    in/outPos:      xcon, ycon:     coordinates of physical connection point
                                    calculated from conPosIn/Out and conSize
    
    """
    
    def __init__(self, parent, pos=wx.Point(0,0), iconSize=None,
                 name = "---", code=None, colour = wx.GREEN, inFlow = [], outFlow=[], bitmap=None, \
                 conSize=None, conPosIn=None, conPosOut=None,
                 conLabelWidth=None, conLabelShift=None, conTextShift=None, conStyle=None):

        self.pos = pos
        self.name = name
        self.code = code        # set of attributes which helps to associate the icon to the source object (e.g. (objectCode, objectNo))
        self.colour = colour
        self.inFlow = inFlow
        self.outFlow = outFlow
        self.parent = parent
        self.qin = None         # for mode = 1: inflow/outflow = label, qin/qout = value
        self.qout = None

        self.conSize = self.parent.conSize if conSize is None else conSize
        self.conPosIn = self.parent.conPosIn if conPosIn is None else conPosIn
        self.conPosOut = self.parent.conPosOut if conPosOut is None else conPosOut

        self.conStyle = self.parent.conStyle if conStyle is None else conStyle
        self.conLabelShift = self.parent.conLabelShift if conLabelShift is None else conLabelShift
        self.conLabelWidth = self.parent.conLabelWidth if conLabelWidth is None else conLabelWidth
        self.conTextShift = self.parent.conTextShift if conTextShift is None else conTextShift

        self.bitmap = bitmap
        self.iconSize = self.parent.iconSize if iconSize is None else iconSize
        parent.xmaxDrawing = max(parent.xmaxDrawing, self.pos[0] + self.parent.gridSize[0])
        parent.ymaxDrawing = max(parent.ymaxDrawing, self.pos[1] + self.parent.gridSize[1])

        self.boxColour = None
        self.boxFGColour = None

    def inPos(self, n):
        """ position of input connection
        """
        (x,y) = self.pos
        x += self.conPosIn[0] - self.conSize[0]
        y += self.conPosIn[1]
        y += self.parent.conYSep * (n+1.5)
        return (x,y)
    
    def outPos(self, n):
        (x,y) = self.pos
        x += self.conPosOut[0] + self.iconSize[0] + self.conSize[0]
        y += self.conPosOut[1]
        y += self.parent.conYSep * (n+1.5)
        return (x,y)

    def setColour(self, colour, fgcolour=None):
        """
        Sets the colours of the frame of the box
        """
        
        self.boxColour = colour
        self.boxFGColour = fgcolour

    def clickedCon(self, pos):
        """ checks if a click at position pos is on
        this icon, and if yes, at which connection
        returns:
        +1 ...+n if click at input
        -1 ...-n if click at output
        0 if click at icon, but not on input or output
        None if click out of icon
        """

        x, y = pos
        x0, y0 = self.pos
        x1 = x0 + self.iconSize[0]
        y1 = y0 + self.iconSize[1]

        if x >= x0 - CLICKTOLX + self.conPosIn[0] - self.conSize[0] \
           and x <= x1 + CLICKTOLX + self.conPosOut[0] + self.conSize[0]: # the 20 is for the labels of in/outflow
            if y >= y0 - CLICKTOLY and y <= y1 + CLICKTOLY:
                for n, flow in enumerate(self.inFlow):
                    x0con, y0con = self.inPos(n)
                    x1con = x0con + self.conSize[0]
                    y1con = y0con + self.conSize[1]
                    if x >= x0con - CLICKTOLX and x <= x1con + CLICKTOLX:
                        if y >= y0con - CLICKTOLY and y <= y1con + CLICKTOLY:
                            return n+1

                for n, flow in enumerate(self.outFlow):
                    x1con, y0con = self.outPos(n)
                    x0con = x1con - self.conSize[0]
                    y1con = y0con + self.conSize[1]
                    if x >= x0con - CLICKTOLX and x <= x1con + CLICKTOLX:
                        if y >= y0con - CLICKTOLY and y <= y1con + CLICKTOLY:
                            return -(n+1)
                    
                return 0

        return None

    def draw(self):
        # import some parameters from the parent drawing
        dc = self.parent.dc
        
        shiftText = -4
        
        (x,y) = self.pos
        if self.boxColour is not None:
            dc.SetPen(wx.TRANSPARENT_PEN)
            dc.SetBrush(wx.Brush(self.boxColour))
            dc.DrawRectangle(x - BOXWIDTH, y-BOXWIDTH, self.iconSize[0] + 2*BOXWIDTH, self.iconSize[1] + 2*BOXWIDTH)

        if self.boxFGColour is not None:
            dc.SetPen(wx.TRANSPARENT_PEN)
            dc.SetBrush(wx.Brush(self.boxFGColour))
            dc.DrawRectangle(x - BOXFGWIDTH, y-BOXFGWIDTH, self.iconSize[0] + 2*BOXFGWIDTH, self.iconSize[1] + 2*BOXFGWIDTH)

        if self.bitmap is not None:
            dc.DrawBitmap(self.bitmap, x, y, True)

        font = wx.Font(8, wx.MODERN, wx.NORMAL, wx.BOLD, False, u'Arial')
        dc.SetFont(font)

        # icon title
        (x,y) = self.pos
        if self.name is not None: dc.DrawText(U(self.name), x, y)

##        family can be:
##
##    wx.DECORATIVE, wx.DEFAULT,wx.MODERN, wx.ROMAN, wx.SCRIPT or wx.SWISS.
##
##style can be:
##
##    wx.NORMAL, wx.SLANT or wx.ITALIC.
##
##weight can be:
##
##    wx.NORMAL, wx.LIGHT, or wx.BOLD

        font = wx.Font(8, wx.MODERN, wx.NORMAL, wx.NORMAL, False, u'Arial')
        dc.SetFont(font)

        for n, sInFlow in enumerate(self.inFlow):
                
            x, y = self.inPos(n)
            if sInFlow not in ["", " "]:
                if self.conStyle==1:
                    dc.SetPen(wx.TRANSPARENT_PEN)
                    dc.SetBrush(wx.Brush(SOFTORANGE))
                    dc.DrawRectangle(x, y-0.5*self.conSize[1], self.conSize[0], self.conSize[1])

            xlabel = x - (self.conLabelWidth - self.conSize[0])

            if self.conStyle == 1:
                dc.DrawText(U(sInFlow), xlabel, y - self.conLabelShift)
                val = None if self.qin is None else self.qin[n] 
                if val is not None: dc.DrawText(val, x+2, y - self.conTextShift)
            else:
                dc.DrawText(U(sInFlow), x+2, y - self.conTextShift)
                

        for n, sOutFlow in enumerate(self.outFlow):
            x, y = self.outPos(n)
            if sOutFlow not in ["", " "]:
                if self.conStyle==1:
                    dc.SetPen(wx.TRANSPARENT_PEN)
                    dc.SetBrush(wx.Brush(SOFTORANGE))
                    dc.DrawRectangle(x - self.conSize[0], y-0.5*self.conSize[1], self.conSize[0], self.conSize[1])

            xlabel  = x - self.conSize[0]

            if self.conStyle == 1:
                dc.DrawText(U(sOutFlow), xlabel, y - self.conLabelShift)
                val = None if self.qout is None else self.qout[n]
                if val is not None: dc.DrawText(val, xlabel+2, y - self.conTextShift)
            else:
                dc.DrawText(U(sOutFlow), xlabel+2, y - self.conTextShift)


def connectIcons(fromIcon, toIcon, nv = 0, nhx = 0, nhy = 0, nvt=0, nhback=0,
                 colour = wx.BLACK, style=wx.SOLID, inputNo=1, outputNo=1):
    """
    nhx: vertical shift of outgoing horizontal line towards standard position
    nhy: vertical shift of ingoing horizontal line towards standard position
    nv:  horizontal position of vertical line
    """

    lines = []
    
    rs = fromIcon.outPos(outputNo-1)
    ri = fromIcon.inPos(0)
    rt = toIcon.inPos(inputNo-1)
##    xIcon = fromIcon.pos
##    xBack = xIcon[1] + fromIcon.parent.iconSize[1] + 4*(nhback+1)

    rIcon = toIcon.pos
    yBack = rIcon[1] - 4*(nhback+1)
    
    if rs[0] < rt[0] and rt[0] < rs[0] + fromIcon.parent.gridSize[0]: # straight connection from left to right (adjacent icons only)
        r0 = (rs[0], rs[1] + nhx*4)
        r1 = (rt[0] - 10 - nv*4, r0[1])
        r2 = (r1[0], rt[1] + nhy*4)
        r3 = (rt[0], r2[1])

        lines.append((r0,r1,colour, style))
        lines.append((r1,r2,colour, style))
        lines.append((r2,r3,colour, style))

    else:
        r0 = (rs[0], rs[1] + nhx*4)    # connection point
        r1 = (ri[0] + fromIcon.parent.gridSize[0] - 10 - 4*nv, r0[1])     # end horizontal line to vchannel
        r2 = (r1[0], yBack)
        r3 = (rt[0] - 10 - nvt*4, r2[1])
        r4 = (r3[0], rt[1] + nhy*4)
        r5 = (rt[0], r4[1])

        lines.append((r0,r1,colour, style))
        lines.append((r1,r2,colour, style))
        lines.append((r2,r3,colour, style))
        lines.append((r3,r4,colour, style))
        lines.append((r4,r5,colour, style))
        

    return lines


