# -*- coding: utf-8 -*-
#==============================================================================
#
#    E I N S T E I N
#
#    Expert System for an Intelligent Supply of Thermal Energy in Industry
#
#------------------------------------------------------------------------------
#
#    NodeEditor: editing of nodes in drawings
#
#==============================================================================
#
#    (C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2017
#    http://www.energyxperts.net/
#
#==============================================================================
import wx
from auxiliary.auxiliary import findKey
from GUI.dataEntry.tableEntry import TableEntry
from GUI.tools.dialogOKCancel import DialogOKCancel

NODETYPES = {"MIX": _("Mixer / Scaling"),
             "SPLIT": _("Splitter"),
             "SRC": _("Source")}
##             "GEN": _("Generic")} #TODO: generic node type not yet implemented
                    
DEFAULTSCALE = [1., 1., 1.]
DEFAULTSPLIT = [0.5, 0.5]

class NodeData():
    def __init__(self, nodeType, name=_("New node"), scaleFactor=DEFAULTSCALE, splitFraction=None):
        self.name = name
        self.nodeType = nodeType
        
        if self.nodeType=="MIX":
            self.scaleFactor = scaleFactor
            self.splitFraction = [1.0]
        elif self.nodeType=="SPLIT":
            self.scaleFactor = [1.0]
            self.splitFraction = DEFAULTSPLIT if splitFraction is None else splitFraction
        elif self.nodeType=="SRC":
            self.scaleFactor = []
            self.splitFraction = [1.0] if splitFraction is None else splitFraction
            

    def __repr__(self):
        return "NodeData: %s %s %s %s"%(self.name, self.nodeType, self.scaleFactor, self.splitFraction)
        
class NodeEditor(DialogOKCancel):
    def __init__(self, prnt, nodeData=None):

        if nodeData is None:
            self.nodeData = NodeData("MIX")
        else:
            self.nodeData = nodeData

        self.parent = prnt

        super(NodeEditor, self).__init__(prnt, title=self.nodeData.name, size=wx.Size(520,200))

        self._custom_ctrls(prnt)
        self._custom_layout()

        self.table.setExternalEventHandler(self.onTableClick)

        self.display()


    def _custom_ctrls(self, prnt):
        titles = [_("parameter"), _("value")]
        self.table = TableEntry(self, titles, self.changeData,
                                readOnly=[True, False],
                                colSize=250,
                                colLabelSize=0,
                                size=wx.Size(500, 150))
        self.table.setPopUpMenu([_("insert"), _("delete")])
##        self.table.setColSize(0, 300)
##        self.table.setAttribute("edit", [False, True])
            
    def _custom_layout(self):
        flagText = wx.ALIGN_CENTER_VERTICAL|wx.TOP

        # global sizer for panel.
        sizerGlobal = wx.BoxSizer(wx.VERTICAL)
        sizerGlobal.Add(self.table, 0, wx.ALIGN_CENTRE, 8)
        self._do_layout(sizerGlobal)
 
    def display(self):

        data, choices = self.buildData()
        self.table.display(data, choices=choices, selection=None)
        wx.SafeYield()

        self.Show()
        self.parent.activeNodeData = self.nodeData  # export node data to parent for saving after EndModal

    def changeData(self, index, mode, data=None, selection=None):
        if mode==-1 and index==0:
            self.nodeData.name = data[1]

        elif mode==-1 and index==1:
            newTypeStr = data[1]

            newType = findKey(NODETYPES, newTypeStr)
            if newType is not None and newType != self.nodeData.nodeType: # do something only if type has changed and has a correct value
                self.nodeData = NodeData(newType, name=self.nodeData.name)
                
##                self.display() #TODO: check whether display should be activated here or within the TableEntry - class

        elif mode==-1:
            startIndex = 2 if self.nodeData.nodeType == "SRC" else 3
            if index >= startIndex:
                con = index-startIndex
                newValStr = data[1]
                try:
                    newVal = float(newValStr)
                except:
                    newVal = None
                    
                print "NodeEditor.changeData: val %s nodeType %s"%(newVal, self.nodeData.nodeType)
                if self.nodeData.nodeType == "MIX":
                    self.nodeData.scaleFactor[con] = newVal
                elif self.nodeData.nodeType == "SPLIT":
                    self.nodeData.splitFraction[con] = newVal
                elif self.nodeData.nodeType == "SRC":
                    self.nodeData.splitFraction[con] = newVal
            
        elif mode==0: # insert
            self.nodeData.scaleFactor.append(self.nodeData.scaleFactor[-1])   # append same value as last entry
            self.nodeData.splitFraction.append(0.0)
##            self.display()
            
        elif mode==1: # delete
            if len(self.nodeData.scaleFactor) > 1: self.nodeData.scaleFactor.pop(-1)
            if len(self.nodeData.splitFraction) > 1: self.nodeData.splitFraction.pop(-1)
##            self.display()

        data, choices = self.buildData()

        self.parent.activeNodeData = self.nodeData  # export node data to parent for saving after EndModal
        
        return 0, data, choices

    def onTableClick(self, eventType, r, c, selection):
        if eventType == "LeftDclick":
            self.OnButtonOK(None)

    def buildData(self):
        """ converts the internal information into a data table to be represented on the display
        """
        
        data = [[_("Name"), self.nodeData.name],
                [_("Type of node"), NODETYPES[self.nodeData.nodeType]]]
        choices = [[None, None],
                   [None, NODETYPES.values()]]

        if self.nodeData.nodeType == "MIX":
            data.append([_("Scale factors for inputs:"), ""])
            for n, fscale in enumerate(self.nodeData.scaleFactor):
                data.append([_("Input %s")%(n+1), "" if fscale is None else "%s"%fscale])
                
        elif self.nodeData.nodeType == "SPLIT":
            data.append([_("Split fraction for outputs:"), ""])
            for n, fsplit in enumerate(self.nodeData.splitFraction):
                data.append([_("Output %s")%(n+1), "" if fsplit is None else "%6.4f"%fsplit])

        elif self.nodeData.nodeType == "SRC":
            fsplit = self.nodeData.splitFraction[0]
            data.append([_("Output: "), "" if fsplit is None else "%6.4f"%fsplit])

        else:
            logWarning("NodeEditor: node type %s not yet supported"%self.nodeType)
            data = [[]]

        for row in data[2:]: choices.append([None, None])
                
        return data, choices


            

