import wx

from lib.auxiliary.auxiliary import noneCheck
from lib.text.textConversion import toUnicode as U, toGettextAndUnicode as _U, toUTF8
import matPanel as Mp
import numpy as np

import matplotlib.font_manager as font
from matplotlib.ticker import FuncFormatter, FixedLocator
##from einstein.gui.tools.GUITools import *
# EINSTEIN colours
WHITE = '#FFFFFF'
LIGHTGREY = '#F6F6F6'
MIDDLEGREY = '#C0C0C0'
DARKGREY = '#000060'
BLACK = '#000000'

GRAPH_BACKGROUND_COLOR = WHITE

ORANGECASCADE = ['#FF6000','#EE5000','#DD4000','#CC2000','#BB1000',\
                 LIGHTGREY,MIDDLEGREY,DARKGREY,BLACK]
LINETYPES = ['-','--',':','.','-.',]

import textwrap

# fix parameters for plot
axeslabel_fontsize = 8
axesticks_fontsize = 8
legend_fontsize = 8
spacing_left = 0.12
spacing_right = 0.98
spacing_bottom = 0.16
spacing_top = 0.98

POPUPMENU =[_("export as csv-file"), _("export as image")]
POPUPMENUIDS = [wx.NewId() for item in POPUPMENU]

class XYPlot(wx.Panel):
    """
    eventHandler: external event handler for reacting on mouse clicks on the plot.
        receives a matplotlib-event object as input with the attributes (among others):
        - x, y: click position in pixels
        - xdata, ydata: click position in data points
        - button: button number (1 = left-click)
    """

    def __init__(self, parent, style=wx.SUNKEN_BORDER, size=wx.Size(800, 600), eventHandler=None, fileName=None):
        """
        fileName: data file name (w/o extension) for export of plots
        """
        super(wx.Panel, self).__init__(parent, style=style, size=size)

        self.eventHandler = eventHandler
        self.actionCode = None
        self.size = size if isinstance(size, wx.Size) else wx.Size(size[0], size[1])
        self.fileName = fileName # filename for export of plot
        
        # remap drawing methods to the wx widgets.
        paramList = {'backcolor'   : GRAPH_BACKGROUND_COLOR} # graph background color

        Mp.MatPanel(self, wx.Panel)
        self.figure.canvas.mpl_connect('button_release_event', self.onMouseClick)
        
        self.subplot = None

        # default values for axis-limits
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None

        self.axvLines = []
        self.xTicks = None
        self.minorXTicks = None

        self.grid = True

        self.lineTypes = None
        self.lineWidths = None
        self.colours = None

        self.markerSize = 8

        self.options = {"legend": True}

    def xAxis(self, label="x", xmin=None, xmax=None):
        """ sets Parameters of x-Axis
        """
        self.xlabel = label
        self.xmin = xmin
        self.xmax = xmax
    
    def yAxis(self, label="y", ymin=None, ymax=None):
        """ sets Parameters of y-Axis
        """
        self.ylabel = label
        self.ymin = ymin
        self.ymax = ymax

    def setDataLabels(self, labels, lineTypes=None, colours=None, lineWidths=None, markerSizes=None):
        n = len(labels)
        
        self.dataLabel = labels if labels else ["Row %s"%(i+1) for i in range(n)]
        self.lineTypes = lineTypes if lineTypes else [LINETYPES[j%len(LINETYPES)] for j in range(n)]
        self.colours = colours if colours else [ORANGECASCADE[j%len(ORANGECASCADE)] for j in range(n)]
        self.lineWidths = lineWidths if lineWidths else [2.0]*n
        self.markerSizes = markerSizes

    def setAxvLines(self, xvals):
        """ sets vertical lines within the plot """
        self.axvLines = xvals

    def setXTicks(self, xvals):
        """ sets vertical lines within the plot """
        self.xTicks = xvals

    def setMinorXTicks(self, xvals):
        """ sets vertical lines within the plot """
        self.minorXTicks = xvals

    def draw(self, xData, yData):
        """ Plotting method
        will be dynamically added as method of class XYPlot by Mp.matPanel
        """
        
        self.figure.clf()
        self.subplot = self.figure.add_subplot(1, 1, 1)
        self.figure.subplots_adjust(left=spacing_left, right=spacing_right, bottom=spacing_bottom, top=spacing_top)

        nx = len(xData)
        ny = len(yData)

        self.xData = xData
        self.yData = yData

        for j in range(ny):

            # common or separate x-data for each of the series in y ?
            jx = 0 if nx==1 else j
                
            self.subplot.plot(xData[jx],
                              yData[j],
                              self.lineTypes[j],
                              color=self.colours[j],
                              label=U(self.dataLabel[j]),
                              linewidth=self.lineWidths[j],
                              markersize=self.markerSize if self.markerSizes is None else self.markerSizes[j])
            
        # set limits of x- and y-axes
        if any([val is None for val in [self.xmin, self.xmax]]):
            xmin, xmax = self.getAutoLimits(xData)
            self.xmin = noneCheck(self.xmin, xmin)
            self.xmax = noneCheck(self.xmax, xmax)

        if any([val is None for val in [self.ymin, self.ymax]]):
            ymin, ymax = self.getAutoLimits(yData)
            self.ymin = noneCheck(self.ymin, ymin)
            self.ymax = noneCheck(self.ymax, ymax)

        self.subplot.axes.set_xlim([self.xmin, self.xmax])
        self.subplot.axes.set_ylim([self.ymin, self.ymax])

        if self.xTicks:
            ticks = [tick for tick in self.xTicks if tick >= self.xmin and tick <= self.xmax]
            self.subplot.xaxis.set_ticks(ticks)

        if self.minorXTicks:
            ticks = [tick for tick in self.minorXTicks if tick >= self.xmin and tick <= self.xmax]
            minor_locator = FixedLocator(ticks)
            self.subplot.xaxis.set_minor_locator(minor_locator)

        if self.grid:
            if self.minorXTicks:
                self.subplot.grid(which='minor')
            self.subplot.grid()

        # set labels of axes
    ##    major_formatter = FuncFormatter(format_int_wrapper)
    ##    self.subplot.axes.xaxis.set_major_formatter(major_formatter)
    ##    self.subplot.axes.yaxis.set_major_formatter(major_formatter)
        
        fp = font.FontProperties(size=axeslabel_fontsize)

        self.subplot.axes.set_xlabel(U(self.xlabel), fontproperties=fp)
        
        ylabel = self.ylabel
        padding = 10 if len(ylabel) > 40 else None
        self.subplot.axes.set_ylabel(textwrap.fill(U(ylabel), 40, break_long_words=False),
                                     fontproperties=fp,
                                     horizontalalignment='center',
                                     labelpad=padding)

        for label in self.subplot.axes.get_yticklabels():
    #        label.set_color(self.params['ytickscolor'])
            label.set_fontsize(axesticks_fontsize)
    #        label.set_rotation(self.params['yticksangle'])

        # properties of labels on the x axis
        for label in self.subplot.axes.get_xticklabels():
    #        label.set_color(self.params['xtickscolor'])
            label.set_fontsize(axesticks_fontsize)
    #        label.set_rotation(self.params['xticksangle'])

        if self.options["legend"]:
            # legend
            self.subplot.legend(loc=0)   #4: left lower; 0: best
            try:
                lg = self.subplot.get_legend()
                ltext = lg.get_texts()             # all the text.Text instance in the legend
                for txt in ltext:
                    txt.set_fontsize(legend_fontsize)  # the legend text fontsize
                # legend line thickness
                llines = lg.get_lines()             # all the lines.Line2D instance in the legend
                for lli in llines:
                    lli.set_linewidth(1.5)          # the legend linewidth
                # color of the legend frame
                # this only works when the frame is painted (see below draw_frame)
                frame = lg.get_frame()             # the patch.Rectangle instance surrounding the legend
                frame.set_facecolor('#F0F0F0')      # set the frame face color to light gray
                # should the legend frame be painted
                lg.draw_frame(False)
            except:
                # no legend
                pass

        lines = [line for line in self.axvLines if line >= self.xmin and line <= self.xmax]
        for xval in lines:
            self.subplot.axvline(x=xval)

        self.canvas.draw()
        self.canvas.flush_events()

    def SetValue(self, dataDict, emitEvent=True):
        """
        For call from EinsteinPanel - symmetry with other displayClasses
        """
        self.setDataLabels(dataDict.get("labels"),
                           lineTypes= dataDict.get("lineTypes"),
                           colours = dataDict.get("colours"),
                           lineWidths = dataDict.get("lineWidths"),
                           markerSizes = dataDict.get("markerSizes"))
                           
        self.draw(dataDict.get("xData"), dataDict.get("yData"))


    def getAutoLimits(self, data):
        """
        helper function for setting axis limits to round values around the data
        """
        if data is None or len(data) == 0: return 0, 0 # security feature

        fmin = np.min([(np.min(row) if len(row) > 0 else 0.0) for row in data]) 
        fmax = np.max([(np.max(row) if len(row) > 0 else 0.0) for row in data])

        df = fmax - fmin
        fabs = max(abs(fmax), abs(fmin))

        minDist = max(0.02*df, 0.001*fabs, 1.e-3)

        return fmin - minDist, fmax + minDist


        
        
        
    def onMouseClick(self, event):
        if event.button > 1:
            # create pop-up
            menu = wx.Menu()
            for i, title in enumerate(POPUPMENU):
                id = POPUPMENUIDS[i]
                menu.Append( id, title)
                wx.EVT_MENU(menu, id, self.onPopUp)

            # Display menu with call to PopupMenu, invoked on the source component, passing event's GetPoint. ###

            pos = (event.x, self.size.height - event.y)
            self.PopupMenu(menu, pos)            
            menu.Destroy() # destroy to avoid mem leak
            wx.SafeYield()

            if self.actionCode==0: # file export

                table = []
                nx = len(self.xData)
                ny = len(self.yData)

                for j in range(ny):
                    jx = 0 if nx==1 else j
                    if j==0 or nx>1:
                        xlabel = "%s[%s]"%(toUTF8(self.xlabel), toUTF8(self.dataLabel[j])) if nx>1 else toUTF8(self.xlabel)
                        table.append([xlabel]+[val for val in self.xData[j]])
                    table.append([toUTF8(self.dataLabel[j])]+[val for val in self.yData[j]])

                if self.fileName is None:
                    from lib.GUI.tools.filePicker import pickFile
                    fileName = pickFile(_("file for export of plot data"), wildcard="*.csv", defaultFile="xyPlot.csv", new=True)
                else:
                    fileName = self.fileName+".csv"

                if fileName is not None:
                    from lib.fileIO.fileIO import writeTableToFile
                    writeTableToFile(table, fileName, transpose=True)

            elif self.actionCode==1:
                from constants import APPDATA

                if self.fileName is None:
                    from lib.GUI.tools.filePicker import pickFile
                    fileName = pickFile(_("file for export of plot data"), wildcard="*.png", defaultFile="xyPlot.png", new=True)
                else:
                    fileName = self.fileName+".png"

                if fileName is not None:
                    import os
                    from constants import APPDATA
                    filePath = os.path.join(APPDATA, fileName)
                    self.figure.savefig(filePath)
            
##        elif self.eventHandler is not None:
##            self.eventHandler(event)

    def onPopUp(self, event):
        actionCode = POPUPMENUIDS.index(event.GetId())
        self.actionCode = actionCode
            
