# -*- coding: utf-8 -*-
#==============================================================================
#
#       energyXperts software library
#
#------------------------------------------------------------------------------		
#
#       Float: processing of floating point numbers
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
from math import floor, ceil
import bisect
import numpy as np
import locale
from constants import INFINITE, EPSILON

FRACTIONOFINFINITE = 1.e-12*INFINITE

def almostEqual(a, b, tol=-EPSILON):
    if a is None or b is None:
        return False
    elif tol < 0:   # relative tolerance
        ref = max(abs(a), abs(b))
        return abs(a - b) <= -tol*ref
    else:           # absolute tolerance
        return abs(a - b) <= tol

def checkLimits(x, xmin, xmax, default=None):   
    """Checks if a value is within limits
    If no value is given, the default value is assigned
    
    :param x: the value
    :param xmin: the lower limit
    :param xmax: the upper limit
    :param default: the default value for invalid x

    """
    if x is None:
        return default
    xl = min(x, xmax)
    xl = max(xl, xmin)
    return xl

def toFloat(val):
    """
    converts ISO31-0 strings to a floating point number:
    
    - comma or dot as decimal separator
    - space as optional thousands separator
    """

    if isinstance(val, list):
        return [toFloat(elem) for elem in val]
    else:
        try:
            return float(val)
        except:
            try:
                return float(val.replace(",",".").replace(" ",""))
            except:
                return None

def toInt(val):
    """
    converts strings to an integer:
    - comma or dot as decimal separator
    - space as optional thousands separator
    """

    if isinstance(val, list):
        return [toInt(elem) for elem in val]
    else:
        try:
            return int(val)
        except:
            try:
                return int(val.replace(",",".").replace(" ",""))
            except:
                return None

def scale(val, fscale):
    """ allows for multiplying a value by scale without checking before
    if its a valid number. If not, the value is returned as is
    """
    
    try:
        x = float(val) * fscale
    except:
        x = val
    return x

#-------------------------------------------------------------------------
# tools for conversion from / to string (GUI entry)

#TODO: check state in current version of EINSTEIN and unify

class FloatIO():
    """
    class for reading and writing floats to strings

    nomenclature:
    f: float value
    s: string value
    """

    def __init__(self, decimals=2, tsep=" ", dc=",", sinf="inf", snan="nan", snone=""):
        """
        Thousands separator (tsep) and decimal sign (dc) are set to ISO-31 by default
        Setting both to None loads from the locale
        """
        self.decimals = decimals    # number of decimals to be printed
        self.tsep = tsep
        self.dc = dc

        self.sinf = sinf
        self.snan = snan
        self.snone = snone
        
        if tsep is None or dc is None:
            # get a dictionary of local parameters
            loc = locale.localeconv()
            self.dc = loc['decimal_point']
            self.tsep = loc['thousands_sep']
  
# 1. conversion float to string    

    def write(self, f, decimals=None):
        """
        writes a float to a string
        """
        if isinstance(f, str):  # typical use is with float or None as input, but for security improvement in "strange" use cases
            f = self.read(f)
            
        # first check for special cases
        if f is None:
            return self.snone

        elif np.isnan(f):
            return self.snan

        elif (np.abs(f) >= FRACTIONOFINFINITE) or np.isinf(f):
            return ("-" if f<0 else "") + self.sinf
        
        dec = self.decimals if decimals is None else decimals
            
        sval = self.showFormatted(f, dec)
        return sval

    def showFormatted(self, f, dec):
        """
        Helper function. prints a float with thousands separator and changes decimal sign
        """
        #TODO: implement the python ":0.f".format(...) functions once this works for the separator types of interest

        # separate into sign and number
        sign = ''
        if f < 0.0:
            sign = '-'
            f = abs(f)

        # use default formatted printing
        if dec is None:
            try:
                dec = max(int(6-np.log10(f)), 0)
            except:
                dec = 0
            
        snum = "%.*f" % (dec, f)

        # separate into integer and fraction
        try:
            (integ, frac) = snum.split(".")
        except ValueError:
            integ = snum
            frac = ''

        # set thousands separator
        sn = ''
        r0 = integ[::-1] # reverse
        for i, c in enumerate(r0):
            if (i % 3) == 0 and len(sn) > 1:
                sn += self.tsep
            sn += c
            
        s = sign + sn[::-1]
        if dec > 0: s += self.dc + frac

        return s

# 2. conversion string to float        

    def read(self, s):
        """
        reads a float from a string
        """
        if not isinstance(s, str):
            s = str(s)

        # eliminate thousand separator and spaces
        s.replace(self.tsep, '')
        s = s.replace(' ', '')

        # convert decimal character to a dot
        s = s.replace(self.dc, '.')
        try:
            f = float(s)
        except ValueError:
            f = None

        return f


     

if __name__=="__main__":

    conv = FloatIO()
    print conv.write(-17411.235946, decimals=4)
