#==============================================================================
#
#	E4 library
#
#------------------------------------------------------------------------------
#
#	MESSAGE LOGGER
#			
#------------------------------------------------------------------------------
#			
#	Auxiliary module for using the main-fucntions
#       logMessage, logError and logWarning also from other modules
#
#       logMessage, logWarnings, logError: display on the GUI interface
#       logTrack: tracks messages in data file for further analysis
#
#==============================================================================
#
#	(C) E4-Experts GmbH
#	www.energyxperts.net / info@energyxperts.net
#
#==============================================================================

import logging
import logging.handlers
import time
import os

DEBUG_TO = "OFF"    #SCREEN: prints debug information on the screen
                    #LOGFILE: writes debug information to the log-File
                    #OFF: no debug information written

def setup_logger(name, logFile, fmt="%(message)s", level=logging.INFO):
    """
    Function setup as many loggers as you want
    Use this function instead of basicConfig in order to use different log files
    """

    handler = logging.handlers.TimedRotatingFileHandler(logFile,
                                              when="H",
                                              interval=1)
    
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger, handler

class MessageLogger():
    def __init__(self, name, logFile, timeStamp=False, GUI=None, recordLevel=2):
        """
        recordLevel: 0: nothing 1: only errors 2: errors and warnings
        """
        self.timeStamp = timeStamp
        self.name = name
        self.logFile = logFile
        #----- Prepare logging
        self.setUp()
        self.dateStr = None
        self.GUI = GUI  # any GUI module with dialog functions showMessage, showWarning, showError and askConfirmation

        self.recordLevel = recordLevel
        self.nerr = 0
        self.errors = []
        self.nwarn = 0
        self.warnings = []
        
    def setUp(self):
##        logging.basicConfig(level=logging.INFO, filename=logFile, format=fmt)
        self.logger, self.handler = setup_logger(self.name, self.logFile)

    def linkToGUI(self, GUI):
        self.GUI = GUI

    def log(self, msg, timeStamp=False):
        if timeStamp or self.timeStamp:
            dateStr = time.strftime("%Y-%m-%d")

##            self.logger.flush()
##            os.system("mv %s %s-%s"%(self.logFile, self.logFile, dateStr))
##            self.setUp()       
            
            if dateStr != self.dateStr or self.dateStr is None:
                self.logger.info('DATE=%s' % (dateStr))
                self.dateStr = dateStr
            tStr = time.strftime("%H:%M:%S")
            self.logger.info('%s %s' % (tStr, msg))
        else:
            self.logger.info(msg)

        self.handler.flush()

    def logTrack(self, msg):
        """		
        keeps track of what the user does within the EINSTEIN log-file
        this information can be very useful for debuggig
        e.g. each user interaction on the GUI should be logged
        """		
        self.log(msg)

        if DEBUG_TO == "SCREEN":
            print(msg)


    def logMessage(self, msg):
        """ identical to log, just for backward compatibility of function names
        """
        self.log(msg)
        if self.GUI is not None:
            self.GUI.logMessage(msg)
        
    def logWarning(self, msg):
        self.log('Warning:' + msg)
        if self.GUI is not None:
            self.GUI.logWarning(msg)

        if self.recordLevel >= 2:
            self.nwarn += 1
            self.warnings.append(msg)

    def logError(self, msg):
        """		
        writes a message to the GUI and parallely to the log-file
        highlights as ERROR
        """
        self.log('Error:' + msg)
        if self.GUI is not None:
            self.GUI.logError(msg)

        if self.recordLevel >= 1:
            self.nerr += 1
            self.errors.append(msg)

    def logDebug(self, msg):
        """		
        writes a message to the log-file or to the screen (depending on DEBUG mode)
        doesnt do anything if DEBUG = "OFF"
        """		
        if DEBUG_TO == "SCREEN":
            print msg
        elif DEBUG_TO == "LOGFILE":
            self.log("DEBUG: "+msg)

######### generic GUI functions interacting with message logger

    def showMessage(self, msg):
        if self.GUI is not None:
            self.GUI.showMessage(msg)
        else:
            self.logMessage(msg)

    def showWarning(self, msg):
        if self.GUI is not None:
            self.GUI.showWarning(msg)
        else:
            self.logWarning(msg)

    def showError(self, msg):
        if self.GUI is not None:
            self.GUI.showError(msg)
        else:
            self.logError(msg)
    
    def askConfirmation(self, msg, alwaysAsk=False):
        if self.GUI is not None:
            ret = self.GUI.askConfirmation(msg, alwaysAsk)
        else:
            ret = True
            
        self.log("@" + msg + " " + str(ret))
            
        return ret






