# -*- coding: utf-8 -*-

#==============================================================================
#
#   E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#   fileIO
#
#------------------------------------------------------------------------------
#
#   Functions for writing data to data files
#
#==============================================================================
#
#	Version 2.3plus, 30.06.2015
#
#------------------------------------------------------------------------------
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#==============================================================================
import os
import constants as c
from lib.text.textConversion import toUnicode as U, toUTF8
from auxiliary.auxiliary import transpose as transposeTable
from lib.auxiliary.float import toFloat
import csv
from modules.messageLogger import logError
import numpy as np

def writeTableToFile(table, fileName, separator=";", transpose=False):
    """Writes a data table as a comma-separated text-file

    params:: table: data table
    params:: fileName: file name (including extension) for the fil eto be created
    """

    datafile = os.path.join(c.APPDATA, fileName)
    fw = open(datafile, 'w')

    if transpose:
        outTable = transposeTable(table)
    else:
        outTable = table
        
    for row in outTable:
        string = ""
        for elem in row:
            string += "%s"%elem + separator
        string += "\n"
        fw.write(toUTF8(string))
    fw.close()


def readTableFromFile(fileName, separator=";", comment="#", transpose=False):
    """Reads a data table from a comma-separated text-file

    params:: table: data table
    params:: fileName: file name (including extension) for the fil eto be created
    """

    datafile = os.path.join(c.APPDATA, fileName)
    try:
        f = open(datafile, 'r')
    except:
        logError(_("Data file not found: %s"%fileName))
        return None
    
    csvReader = csv.reader(f, delimiter=separator)

    table = []
    for row in csvReader:
        if len(row) > 0:
            key = row[0]
            if key.startswith(comment): # comment line
                continue

        table.append([val.replace(",",".") for val in row if val != ""])

    if transpose:
        return transposeTable(table)
    else:
        return table
            

def readColumn(table, c, headerRows=1):
    return [table[r][c] if len(table[r]) > c else None for r in range(headerRows, len(table))]

def readHeaders(table, headerRows=1):
    return [table[r] for r in range(0, headerRows)]


#===============================================================================================================
#     CSVTables
#===============================================================================================================
#     Handler for CSV-files containing several tables of different size
#===============================================================================================================


class CSVTables():
    """ reads "AMECO-Like" datafiles with different time in columns and different countries/regions in rows, e.g.

        key
        ----- 1960 1961 1962 ...
        EU-15   55   66   77
        DE      11   12   13
        FR      14   15   16
    """

    def __init__(self, fileName, separator="\t", rowTitleCols=1, colTitleRows=1, symmetric=False, comment="#", transpose=False):
        """
        if ncols = None, ncols (max. number of cols is defined by the writeHeader
        """
        self.datafile = fileName
        self.separator = separator
        self.mode = "nc"
        self.symmetric = symmetric
        self.comment = comment
        self.transpose = transpose

        self.rowTitleCols = rowTitleCols
        self.colTitleRows = colTitleRows

        try:
            self.open()
        except:
            self.f = None

    def open(self, mode="r"):
        self.f = open(self.datafile, mode)
        self.mode = mode

    def close(self):
        if self.f: self.f.close()

    def read(self, tableKey=None, text=False):
        """
        Two operation modes:
        a) if tableKey defined -> reads one table with the specified key
        b) if no tableKey defined -> reads all tables, returning a list

        Symmetric=True: -> it is supposed that all tables have the same number and order of rows and cols
        and the same row- and column keys. Therefore, only the keys of the first table are returned

        text: if True, reads in values as string
        """
        if not self.f: return None

        csvReader = csv.reader(self.f, delimiter=self.separator)

        # first read in as arrays
        rowTitles = []
        table = []
        if tableKey or self.symmetric:
            colTitles = None
        else:
            colTitles = []
            
        inTable = False
        firstRow = False
        skipRows = 0

        iTable = -1

        keys = []
        
        for r, row in enumerate(csvReader):
            if row and len(row) > 0:
                if row[0].startswith(self.comment): continue
                key = row[0]
                
                if not inTable:
                    firstRow = True
                else:
                    firstRow = False
                    
            elif inTable:      # empty line after the table indicates end of relevant data
                inTable = False # continue looking for the next table
                firstRow = False
            else:
                continue

            if key==tableKey or (tableKey is None and firstRow):
                skipRows = r
                inTable = True
                keys.append(key)
                
            elif not inTable:
                continue
                
            if r - skipRows < self.colTitleRows - 1:
                continue

            if r - skipRows == self.colTitleRows-1:   # reading title row of table
                data = row[self.rowTitleCols:]
                if tableKey:
                    colTitles=np.array(data)
                else:   # new table
                    iTable += 1
                    if self.symmetric and iTable==0:
                        colTitles = np.array(data)
                    elif not self.symmetric:
                        colTitles.append(np.array(data))
                    if not self.symmetric:
                        rowTitles.append([])
                    table.append([])
            else:                                   # reading data row
                if text:
                    data = [val for val in row[self.rowTitleCols:]]
                else:
                    data = [toFloat(val) for val in row[self.rowTitleCols:]]
                    
                if tableKey:
                    rowTitles.append(key)
                    table.append(data)
                else:
                    if self.symmetric and iTable==0:
                        rowTitles.append(key)
                    elif not self.symmetric:
                        rowTitles[iTable].append(key)
                    table[iTable].append(data)
                    
        array = [np.array(block) for block in table]

        if tableKey:
            return colTitles, rowTitles, array
        else:
            return keys, colTitles, rowTitles, array


    def write(self, keys, colTitles, rowTitles, array):
        self.close()
        self.open(mode="w")

        for i, key in enumerate(keys):
            cTitles = colTitles if self.symmetric or not colTitles else colTitles[i]
            rTitles = rowTitles if self.symmetric or not rowTitles else rowTitles[i]
            table = array[i]

            if cTitles is not None:
                self.writeLine(cTitles, label=key)
            else:
                self.writeLine([key])
                
            self.writeTable(table, rTitles)
            self.f.write("\n")
            
 
    def writeLine(self, row, label=None):
        """ helper for writeTable """
        
        if not self.f: return

        string = ""
        if label is not None:
            string += label + self.separator

        for i, elem in enumerate(row):
            if i > 0: string += self.separator
            if isinstance(elem, float):
                string += "%12.4f"%elem
            elif isinstance(elem, int):
                string += "%12d"%elem
            elif isinstance(elem, unicode):
                string += toUTF8(elem)
            else:
                string += "%s"%elem
        string += "\n"
        self.f.write(string)
                

    def writeTable(self, table, rowTitles):
        """ helper for write """
        for r, row in enumerate(table):
            self.writeLine(row, label=rowTitles[r] if rowTitles else None)

        
    
