# -*- coding: utf-8 -*-
#============================================================================== 				
#
#	E I N S T E I N
#
#       Expert System for an Intelligent Supply of Thermal Energy in Industry
#       (www.iee-einstein.org)
#
#------------------------------------------------------------------------------
#
#	GenericNet
#			
#------------------------------------------------------------------------------
#			
#	Short description:
#	
#	Generic network formed by nodes (with in- and outputs and
#       (directed) flows
#
#==============================================================================
#	
#------------------------------------------------------------------------------		
#	(C) energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2017
#	www.energyxperts.net / info@energyxperts.net
#
#============================================================================== 				

##EPSILON = 1.e-3     # required accuracy for function "isequal"
##INFINITE = 1.e99    # numerical value assigned to "infinite"

import numpy as np

class GenericNet():
    """ builds the system of nodes and connections

    all information then is within
    node: list of nodes (class Node objects)
    flow: list of connections (class Flow objects)
    """

    def __init__(self, ANo=None):
        self.flow=[]
        self.node=[]
        self.deletedFlows = []
        self.deletedNodes = []

        self.nNodes = 0
        self.nFlows = 0

    def __repr__(self):
        output = "Network================\nNodes:\n"
        for node in self.node:
            output += "%s\n"%node

        output += "\nFlows:\n"

        for flow in self.flow:
            output += "%s\n"%flow
        output += "===================="

        return output

    def addNode(self, node):
        """
        adds a node to the network
        and returns the corresponding node no. in the network
        """

        node.no = self.nNodes
        node.net = self
        self.node.append(node)
        self.nNodes += 1
        
        return node.no

    def addFlow(self, flow):
        flow.no = self.nFlows
        flow.net = self
        self.flow.append(flow)
        self.nFlows += 1

    def connectNodes(self, fromNo, fromCon, toNo, toCon):
        """
        builds a connection from -> to for nodes
        in the network specified by the number of the node and the
        number of the in/out connection in the node
        """
        if fromNo >=0 and fromNo < self.nNodes:
            fromNode = self.node[fromNo]
            if fromCon < 0 or fromCon >= fromNode.nOut:
                return
        else:
            return
        
        if fromNo >=0 and fromNo < self.nNodes:
            fromNode = self.node[fromNo]
        else:
            return
       
        flow = Flow(fromNode, fromCon, self.node[toNo], toCon)
        flow.exportToNodes()    # creates links to the flow within the node objects
        self.addFlow(flow)

    def divideFlow(self, flow, obj=None, objectCode=None):
        """
        Divides a flow and adds a new node into the intersection
        """
        newNode = Node(1, 1, obj=obj, objectCode=objectCode)
        self.addNode(newNode)

        # store downstream connection
        toNode = flow.toNode
        toCon = flow.toCon

        # first change current flow to upstream connection
        flow.toNode = newNode
        flow.toCon = 0
        flow.exportToNodes()

        # then build a new downstream connection
        newFlow = Flow(newNode, 0, toNode, toCon)
        newFlow.exportToNodes()
        self.addFlow(newFlow)

    def splitFlow(self, flow, split, obj=None, objectCode=None):
        """
        Adds a splitter and creates a new downstream object for the split
        """
        # add a splitter node
        newNode = Node(1, len(split), split=split, obj=obj, objectCode=objectCode)
        self.addNode(newNode)

        # store downstream connection
        toNode = flow.toNode
        toCon = flow.toCon

        # first change current flow to upstream connection
        flow.toNode = newNode
        flow.toCon = 0
        flow.exportToNodes()

        downStreamNodes = [toNode]
        for split in split[1:]:
            new = toNode.copy()
            self.addNode(new)
            downStreamNodes.append(new)
        
        # then build downstream connections
        for i, node in enumerate(downStreamNodes):
            newFlow = Flow(newNode, i, node, toCon)
            newFlow.exportToNodes()
            self.addFlow(newFlow)

    def findNode(self, objectCode=None, findAll=False):
        """ helper function for finding nodes with certain attributes

        findAll: if true returns a list of all nodes
        """
        found = []
        if objectCode is not None:
            found = [node for node in self.node if node.objectCode==objectCode]
        if findAll:
            return found
        elif len(found) > 0:
            return found[0]
        else:
            return None

    def buildFlowChart(self, nodeList):
        """
        Builds a hierarchy tree assigning increasing x-pos for downstream objects
        y-pos is assigned so that trees for each source node in the node list are separated

        Note: works well only for networks without cycles !!!
        """

        pos = np.zeros(2, dtype=int)
        pointer = pos
        stack = []

        posMax = np.zeros(2, dtype=int)
        
        for node in nodeList:

            node.pos = pointer.copy()
            origin = node
            branch = 0
            ok = False

            its = 0            
            while not ok and its < 50:
                its += 1

                # lookup the downstream node
                flow = origin.qout[branch]
                flow.pos = pointer.copy()

                down = flow.toNode
                pointer[0] += 1

                down.pos = pointer.copy()
                posMax = np.maximum(down.pos, posMax)

                if down.nodeType=="SPLIT":      # splitter found: pop-up splitter and continue with main branch
                    stack.append([down, 1])    # bookmark the splitter, together with the next branch to be processed
                    origin = down

                elif down is None or down.nodeType=="SNK":
                    if down is not None: down.pos[0]=-1     # marker for end of line
                    while len(stack)> 0:           
                        origin, branch = stack[-1]
                        if branch < len(origin.qout):    # process next branch of the splitter
                            stack[-1][1] += 1
                            pointer[0] = origin.pos[0]  
                            break
                        else:                               # splitter finished, pop from stack and try the next one
                            stack.pop()
                    if len(stack)==0:
                        ok = True
                    else:
                        pointer[1] += 1
                else:                           # just continue
                    origin = down
                    branch = 0

            pointer[1] += 1
            pointer[0] = 0
            
        return posMax
            
class Node():
    def __init__(self, nIn, nOut, id=None, name=None,
                 obj=None, objectType=None, objectCode=None, objectId=None,
                 split=None):
        """ 
        qin, qout: vector of class Flow objects (inputs, outputs)

        nodeType:   INT     generic internal node of the network
                    MIX     mixing of flows with equation
                    SPLIT   split of flows
                    SRC     origin of some flow (nIn = 0, no flow balance)
                    SNK     end point of some flow (nOut = 0, no flow balance)

        obj:        object "behind" / represented by the node
        objectType/objectCode/objectId: identifiers for the object

        split: split fraction for splitter nodes (if None: equal fractions 1/Nout)
        
        """
        # node properties
        self.id = id        # identifier of node
        self.name = name

        # parent network
        self.net = None       # parent network to which the node belongs (set in addNode)
        self.no = None      # number of node within the network (from 0 to N-1)

        self.root = []          # first upstream node(s) in the network
        self.destination = []   # last downstream node(s) in the network
        

        # space for connections (to be filled with Flow-objects)
        
        self.qin = [None]*nIn       # flow-objects of in/outflows
        self.qout = [None]*nOut

        self.nIn = nIn
        self.nOut = nOut

        self.qinVal = [None]*nIn       # values of in/outflows
        self.qoutVal = [None]*nOut

        self.qinInfo = []       # flow objects of information inflows
        self.qinInfoVal = []

        self.q = None       # value of flow through the node (sum of in/outflows)

        self.connected = False

        # determine nodeType depending on the number of in- and outputs
        if nIn==0 and nOut > 0:
            self.nodeType="SRC"
            self.root = self
        elif nIn>0 and nOut==0:
            self.nodeType="SNK"
            self.destination = self
        elif nIn==1 and nOut > 1:
            self.nodeType="SPLIT"
        elif nIn>1 and nOut==1:
            self.nodeType="MIX"
        elif nIn==1 and nOut==1:
            self.nodeType="INT"

        self.scaleFactor = 1.0*np.ones(nIn)                         # list of scale factors for each input
        if split is not None:
            self.split = split
        else:
            self.split = np.ones(nOut, dtype=float)/nOut if nOut > 0 else [] # list of split fractions for each output

        # background object linking with the real system behind
        self.obj = obj      # optional: possibility for direct link to the represented object
        self.objectType=objectType
        self.objectCode = objectCode
        self.objectId = objectId  # identifier of corresponding object

    def __repr__(self):
        string = "Node(%s) [%s:%s] %s: q = %s"%(self.no, self.objectType, self.objectCode, self.name, self.q)
        return string

    def copy(self):
        """
        creates a copy of a node
        NOTE: this node is not automatically added to the network, therefore the attributes net and no are not assigned
        """
        
        new = Node(self.nIn, self.nOut,
                   split=self.split,
                   name=self.name,
                   obj=self.obj,
                   objectType=self.objectType,
                   objectCode=self.objectCode,
                   objectId=self.objectId)
        return new

    def up(self):
        """
        returns list of upstream nodes
        """
        return [q.fromNode for q in self.qin]
                
    def down(self):
        """
        returns list of downstream nodes
        """
        return [q.toNode for q in self.qout]
        
class Flow():
    def __init__(self, fromNode, fromCon, toNode, toCon, flow=None, id=None, real=True):
        self.net = None       # parent network to which the node belongs (set in connectNodes)

        self.fromNode = fromNode
        self.toNode = toNode
        
        self.fromCon = fromCon
        self.toCon = toCon

        self.real = real    # False: information flow only
        
        self.flow = flow
        self.id = id

    def __repr__(self):
        output = "Flow %3d.%1d > %3d.%1d"%(self.fromNode.no, self.fromCon, self.toNode.no, self.toCon)
        return output

    def exportToNodes(self, delete=False):
        """
        builds a link to itself at the connected nodes
        """

        if self.real:
            q = None if delete else self

            self.fromNode.qout[self.fromCon] = q
            self.toNode.qin[self.toCon] = q
        else:
            if self in self.toNode.qinInfo[self.toCon]:
                if delete:
                    self.toNode.qinInfo[self.toCon].remove(self)

            else:
                if not delete:
                    self.toNode.qinInfo[self.toCon].append(self)

    def setMode(self, mode):
        """
        sets mode to real or information flow

        """
        
        real = (mode==0)
        if real != self.real:
            if self.real:           # change from real to virtual
                self.exportToNode(delete=True) # delete as 
            self.real = real
            self.exportToNode()
            return True
        else:
            return False
        


    
        
if __name__ == "__main__":
    print "No unit-demo available"

