# -*- coding: utf-8 -*-
#==============================================================================#
#
#       OO Library
#
#==============================================================================
#
#	Tools for editing open office documents (spreadsheets)
#
#------------------------------------------------------------------------------		
#
#	(C) copyright energyXperts.NET (E4-Experts GmbH), Berlin, Germany 2008-2015
#	http://www.energyxperts.net
#
#	all rights reserved
#
#============================================================================== 
#
# Description of the process of report generation.
#
# 1.  The master document areas subject to change are marked with 'names',
#     which should be related to the keys to the data for the panels.
#     i.e. panelEA2 uses the key 'EA2', so the name will be 'KEY_EA2' (OO does
#     not accept simply 'EA2', so I added the ' KEY_' prefix.). These names
#     are saved in the document (contents.xml) and can be read from there.
#
# 1.1   Creation of a name.
#       a. A region is selected in the sheet.
#       b. Main menu Insert -> Names ->Define (or just Ctrl-F3) will open a dialog
#          where the name is written. The area is taken from the selection. After
#          writing the new name, press Add.
#
# 2.  When the Generate Report option is activated, this module (reporter.py) starts
#     to execute. The steps are:
#
# 2.1   The contents.xml component of the master document (Master) is read and stored
#       as a DOM tree in memory.
#
# 2.2   The names are extracted, together with the ranges they point to. Only the names
#       that start with the prefix 'KEY_' are kept. The prefix is discarded and the
#       rest (that will be the same as the key to the data in Interfaces.GData) is saved
#       as the key to a dictionary entry. The data is the range information. The
#       dictionaries 'names' and 'sheetnames' manage different views of this information.
#
# 2.2   The names dictionary is saved as a XML document, as a reference and debugging
#       aid.
#
# 2.3   Now contents.xml is scanned sequentially looking for worksheets and ranges
#       that exist in the names dictionary. When a correspondence is found, the data
#       is taken from Interfaces using the name (without the prefix) as key. Then
#       the valued are stored in the tree.
#
# 3.  Finally, the DOM tree is saved as a new OO document, which is the generated report.
#
#########################################################################################
import xml.dom.minidom
import os
import wx
import numpy as np

from globals import Globals

from zipfile import ZipFile, ZIP_DEFLATED
from StringIO import StringIO
from tempfile import mkdtemp
import shutil
#
# constants
#
COLUMNS     = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
PREFIX      = "KEY_"
TMPFILE     = "./workfile"
CONTENTFILE = "content.xml"



class OODocument():
    """

    attributes:

    document: the internal represenation of the content file
    """
    def __init__(self, master):
        self.master = master

    def readMaster(self):
        """
        reads in an scans the master document
        """
        infile = ZipFile(self.master, "r")
        dataSource = StringIO(infile.read(CONTENTFILE))
        self.content = xml.dom.minidom.parse(dataSource)
        self.worksheets = self.content.getElementsByTagName("table:table")
        self.sheetNames = [sheet.getAttribute('table:name') for sheet in self.worksheets]


        # create the names list
        self.createNamesTable()

    def writeContentFile(self):
        """
        generates a (temporary) new content file
        """
        self.tempdir = mkdtemp()
        dataOut = open(os.path.abspath(self.tempdir + "/" + CONTENTFILE), "wb")
        dataOut.write(self.content.toxml(encoding='utf-8'))
        dataOut.close()

        self.close()

    def buildNewDocument(self, fileName):
        """
        builds a copy of the original file, substituting CONTENTFILE by the new one
        """
        # Send the document tree to a temporary file
        # and destroy the tree
        # and store the file in a new OO doc
        
        try:
            outfile = ZipFile(fileName, "w", ZIP_DEFLATED)
        except IOError,e:
            logger.showError(_('Error opening file %s. You have probably another version opened: %s') 
                                % (fileName, e))
            return

#        self.main.logMessage('Adding OO component '+CONTENTFILE)
        outfile.write(os.path.abspath(self.tempdir + "/" + CONTENTFILE), CONTENTFILE)

        # copy the rest of the original components
        # to the new OO document
        infile = ZipFile(self.master, "r")
        for component in infile.namelist():
            if component == CONTENTFILE: continue
#                self.main.logMessage('Adding OO component '+component)
            dataIn = infile.read(component)
            dataOut = open(os.path.abspath(self.tempdir + "/" + TMPFILE), "wb")
            dataOut.write(dataIn)
            dataOut.close()
            outfile.write(os.path.abspath(self.tempdir + "/" + TMPFILE), component)

        outfile.close()
        infile.close()

        # remove the temporary directory        
        shutil.rmtree(self.tempdir)
        
        
    def generateReport(self): 
        self.main.logMessage('Generating report from ' + self.master)

        try:
            prepareDataForReport()
        except:
            showError(_U("The data collection for the report has failed. "\
                         "Look into the logs in %s "\
                         "if you want more information about the error.") %
                         (os.path.join(c.APPDATA, c.APPNAME), ))
            raise
        else:
            dlg = DialogGauge(None,_U("Report generation"),_U("creating copy of master"))
    
            thisreport = os.path.abspath(self.newreport)
    

            self.readMaster()   

            # replace values in 
            self.replaceData()

            self.writeContentFile()
            self.buildNewDocument()
            
            dlg.Destroy()
            self.main.showInfo(_U('Report generation is finished\n'
                                 'The generated report is:') + thisreport)

    def close(self):
        self.content.unlink()

    def getDOM(self):
        return self.content
    
	
    def getNameComponents(self, expr):
        """		
        auxiliary function for decomposition of Open-Office cell references
        of the type 'Sheetname'.$A$1:$Z$99
        """
        
        result = []
        e1 = expr.replace('.','')
        e2 = e1.replace(':','')
        e3 = e2.replace("'",'')
        e4 = e3.replace('&apos;','')
        comp = e4.split('$')
        for c in comp:
            s = c.strip()
            if s:
                result.append(s)

        if len(result) == 5:
            # range of cells ['Sheetname'.$A$1:$Z$99]
            return ((result[0],
                    [int(COLUMNS.find(result[1].upper())) + 1,
                     int(result[2]),
                     int(COLUMNS.find(result[3].upper())) + 1,
                     int(result[4])]))
        elif len(result) == 6:
            # range of cells, sheetname repeated ['Sheetname'.$A$1:'Sheetname'.$Z$99]
            return ((result[0],
                    [int(COLUMNS.find(result[1].upper())) + 1,
                     int(result[2]),
                     int(COLUMNS.find(result[4].upper())) + 1,
                     int(result[5])]))
        elif len(result) == 3:
            # one single cell ['Sheetname'.$A$1]
            return ((result[0],
                    [int(COLUMNS.find(result[1].upper())) + 1,
                     int(result[2])]))
        else:
            # not recognized
            logTrack('panelReport (getNameComponents): name format not recognized->'+repr(expr))
            return None

    def createNamesTable(self):
        """
        generates a dictionary for range names
        each entry contains a list of length 2: [sheetname, itemlist]

        sheetname: the name of the sheet where the range is located

        """

        self.names = {}
##        self.sheetnames = {}

        namelements = self.content.getElementsByTagName("table:named-range")

        nElements = len(namelements)
        i = 0
        for namelement in namelements:
            name = namelement.getAttribute('table:name')
            if not name.startswith(PREFIX):
                continue
            name = name[len(PREFIX):]
            range = namelement.getAttribute('table:cell-range-address')
            m = self.getNameComponents(range)
            if m is not None:
                (sheetname,itemlist) = m

#                print "createNamesTable %r %r - %r"%(name,sheetname,itemlist)
                
                self.names[name] = [sheetname, itemlist]


    def modifyCellRange(self, key, data):

##    def modifyCellRange(self, sheet, nrow0, nrows, ncol0, ncols, newdata):
        """
        traverse the DOM looking for one element, identified by the name of the sheet, the row and the column
        when found, modify the range element by element using self._changeOneCell
        """

        sheetName, coords = self.names[key]

        index = self.sheetNames.index(sheetName)
        sheet = self.worksheets[index]
        
##        print "ModifyCellRange: key %s sheetname %s coords %s"%(key, sheetName, coords)

        ncol0 = coords[0]
        nrow0 = coords[1]

        try:
            (nrows, ncols) = np.shape(data)
        except:
            logger.logDebug("OODocument.modifyCellRange: no data shape found for key %s"%key)
            return
##        print "ModifyCellRange: data shape = %s, %s"%(nrows, ncols)

        elementCounter = nrows*ncols
        
#                print "PanelReport (_findCellRange): sheetname %s found"%sheetname
#                print "NRow0 - Ncol0 = %s %s"%(nrow0,ncol0)
        nrow=0
        rowElements = sheet.getElementsByTagName("table:table-row")
        for row in rowElements:
            n = row.getAttribute('table:number-rows-repeated')
            if n:
                nrow += int(n)
#                        print "PanelReport: repeated rows found"

            else:
                nrow += 1
            
            if nrow < nrow0 or nrow >= (nrow0+nrows):
                continue

            ncol = 1        # merged cells count as one
            ncolGrid = 1    # merged cells count individually until ncol0, then as one
            
            for element in row.childNodes:
#                        print "NodeName(%s %s [%s]) = %s"%(nrow,ncol,ncolGrid,element.nodeName)
                if element.nodeName == 'table:covered-table-cell':
                    pass # looks like this doesn't count
                elif element.nodeName == 'table:table-cell':
                    if ncolGrid in range(ncol0,ncol0+ncols):
                        # found the place!
                        # replace and exit
#                                print 'PanelReport (_findCellRange): Call _changeOneCell with data', \
#                                      repr(newdata[nrow-nrow0][ncolGrid-ncol0])
                        self._changeOneCell(element, data[nrow-nrow0][ncolGrid-ncol0])
                        elementCounter -= 1
                        if elementCounter == 0: return
                    n = element.getAttribute('table:number-columns-repeated')
                    if n:
                        # if repeated, add the value

#                            if ncolGrid >= ncol0:
#                                print "PanelReport: WARNING - repeated cells in marked cell range"
                            
                        ncol += int(n)
                        ncolGrid += int(n)

                    else:
                        # if not, add 1
                        ncol += 1
                        ncolGrid += 1

                    n = element.getAttribute('table:number-columns-spanned')

                    if n:
                        # if spanned cells
                        if ncolGrid < ncol0: ncolGrid += int(n)-1   #once within the range, merged cells count as one column

    def _changeOneCell(self, element, newval):
        """
        helper function for self.modifyCellRange
        
        changes one cell value

        element: obtained in screening in modifyCellRange
        """
        
        # substitute Nones and similar by an empty string
        if newval is None:
            newval = ''
        elif (not newval) and newval <> 0:
            newval = ''

        # ok we have some value
        try:
            dummy = float(newval)
            aNumber = True
        except ValueError:
            aNumber = False
            
        if aNumber:
            newval = str(newval)
#            print "PanelReport: converting -> %s as a number"%newval
        
            element.setAttribute('office:value',newval)
            ##print "Do: element.setAttribute('office:value',%s)" % (newval,)
            element.setAttribute('office:value-type','float')
            # there probably is a text node also with the value
            # it is deleted here and OO will recreate it.
            child = element.firstChild
            while child is not None:
                next = child.nextSibling
                if (child.nodeType == child.ELEMENT_NODE and child.tagName == 'text:p') or \
                    (child.nodeType == child.TEXT_NODE):
                    element.removeChild(child)
                child = next
        else:
            if isinstance(newval,unicode):
                pass
            else:
                newval = unicode(newval,"utf-8")

#            print "PanelReport: converting -> %r as text"%newval
            # the value is a text element
            child = element.firstChild
            while child is not None:
                next = child.nextSibling
                if child.nodeType == child.ELEMENT_NODE:
                    try:
                        child.firstChild.data = newval
                    except:
                        logDebug("PanelReport (_changeOneCell): error writing to child.firstchild.data)")
#                        print "error writing to firstchild"
                    return
                child = next

#            print "creating new text element [%r]"%newval
            # if comes here, there was not a previous text node.
            # create a text child for the data
            dataelement = self.content.createElement('text:p')
            # now the text with the value
            text = self.content.createTextNode(newval)
            
            dataelement.appendChild(text)
            element.appendChild(dataelement)
            
        element.normalize()



