from scipy.optimize import minimize, basinhopping, brute
import numpy as np
from math import exp

class Optimiser():
    """ Wrapper integrating all the scipy.optimize functions plus some additional tools for multivariate optimisation problems
    """

    def __init__(self, f, scalar=None, method="SLSQP", globalMethod="brute", bounds=None, constraints=[], options=None):
        """
        scalar has to be defined only if f returns a vector instead of the scalar
        and the optimisation method is based on a scalar minimisation algorithm
        As a default, linear least squares are used

        bounds: bounds of the function in absolute units (not normalised by [scale]). Tuple of arrays (bL, bH)

        constraints: of type cdict = ({"type": "eq", "fun": res})

        """

        self.f = f
        self.scale = None # scale factor for each x_i: u_i = scale_i*x_i before passing u to the optimiser

        self.setBounds(bounds)
        self.constraints = constraints

        self.method = method
        self.globalMethod = globalMethod

        self.maxiter = None     # maximum number of iterations
        self.tol = None         # tolerance for break of iterative process

        self.options = options

    def setScale(self, scale=None):
        """
        scale for reduction of the optimisation problem f(x) to
        the equivalent problem
        
            g(u) = f(x(u)) with x = scale*u
            
        so that the variations (curvature) of the objective function is similar
        in all directions
        """
        self.scale = scale

    def setBounds(self, bounds):

        if bounds is None:
            self.bounds = None
            return
        
        xmin, xmax = bounds
        bounds = [(xmin[i], xmax[i]) for i in range(len(xmin))]
        self.bounds = bounds
        self.bdCheck = CheckBounds(np.array(xmin), np.array(xmax))

    def getLocalMin(self, xo=None):
        """ calculates the local minimum for an initial value xo

        Results:
        xmin: position of the minimum
        fvmin: function value (vector) at the minimum
        fmin: function value (sum square) at the minimum
        dev:  standard deviation of each of the parameters
        """
        if xo is None and self.bounds is not None:
            xo = np.array([0.5*(bd[0] + bd[1]) for bd in self.bounds])
            
        if self.method=="leastsq":
            output = leastsq(optFct,
                             xo,
                             full_output=True,
                             maxfev=10*(len(xo)+1),
                             factor=1,
                             diag=self.scale,
                             ftol=1.0)  # ??? TODO - CHECK. This is a hack in order to achieve a more stable cov_f calculation
            self.xmin = output[0]

            cov_f = output[1]
            self.fvmin = output[2]["fvec"]
            self.fmin = np.sum(np.square(self.fvmin))
            msg = output[3]
            ier = output[4]

    ##        print "cov_f: %s"%cov_f
    ##        print "fmin: %s"%fmin
    ##        print "IER: %s MSG: %s "%(ier, msg)
            
            self.dev = np.sqrt(2.0*fmin*cov_f.diagonal()) if cov_f is not None else None
        
        elif self.method=="least_square":
            pass

        elif self.method=="SLSQP":
            res = minimize(self.f, xo, method="SLSQP", bounds=self.bounds, constraints=self.constraints, options=self.options)
            self.xmin = res.x
            self.fmin = res.fun

        elif self.method=="L-BFGS-B":
            res = minimize(f, xo, method= "L-BFGS-B", bounds=self.bounds, tol=1.e-4, options={"maxiter":100})
            self.xmin = res.x
            self.fmin = res.fun
            
        ##    elif method == "DE":
        ##        bounds = [(xmin[i], xmax[i]) for i in range(len(xo))]
        ##        res = differential_evolution(f, bounds)
        ##        xbest = res.x
        ##        fbest = f(x)

        return self.xmin


    def getGlobalMin(self, xo=None):
        """ calculates the global minimum
        """
        if xo is None and self.bounds is not None:
            xo = np.array([0.5*(bd[0] + bd[1]) for bd in self.bounds])

        if self.globalMethod=="basinhopping":
            if self.bounds:
                res = basinhopping(self.f, xo, niter=20, T = 10000., stepsize=0.5, accept_test=self.bdCheck)
            else:
                res = basinhopping(self.f, xo, niter=20, T = 10000., stepsize=0.5)
            self.xmin = res.x
            self.fmin = res.fun

        elif self.globalMethod=="brute":
            # temporarily deactivated - there is some problem passing the range ... -> check source code of brute
            # for the moment substituted by a manual version
##            res = brute(self.f, np.mgrid[0:1.1:0.5], full_output=True)
##            self.xmin = res.x0
##            self.fmin = res.fval

            fglobal = 1.e+99
            xglobal = xo
            xtry = xo.copy()

            xmax = self.bdCheck.xmax
            xmin = self.bdCheck.xmin

            nits = pow(3, len(xo))
            dx = xmax - xmin

            
            for n in range(nits):
                for i in range(len(xo)):
                    dim = pow(3, i)
                    j = (n/dim)%3
                    if j==0:
                        xtry[i] = xmin[i]
                    elif j==1:
                        xtry[i] = xo[i]
                    elif j==2:
                        xtry[i] = xmax[i]
                
                self.getLocalMin(xtry)

                if self.fmin < fglobal:
                    fglobal = self.fmin
                    xglobal = self.xmin
                
            self.xmin = xglobal
            self.fmin = fglobal


        elif self.globalMethod=="ecopy":

            fglobal = 1.e+99
            xglobal = xo
            xtry = xo.copy()

            xmax = self.bdCheck.xmax
            xmin = self.bdCheck.xmin

            nits = 10
            
            dx = xmax - xmin
    
            for nit in xrange(nits):
                self.getLocalMin(xtry)
                if self.fmin < fglobal:
                    fglobal = self.fmin
                    xglobal = self.xmin
                
                # test new start value 
##                ksi = 0.2 + 0.8*np.random.random()    # gives z from 0 to 4.0 (1.0 at 
##                zo = 0.2*(1./ksi - 1.0)
                zo = 0.0
                xtry = zo*xglobal + (1-zo)*(xmin + dx*np.random.random(len(xo)))
                    
                xtry = np.minimum(xtry, xmax)
                xtry = np.maximum(xtry, xmin)

            self.xmin = xglobal
            self.fmin = fglobal


        return self.xmin

class CheckBounds(object):
    def __init__(self, xmin, xmax):
        self.xmax = xmax
        self.xmin = xmin
    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))
        return tmax and tmin            

# Helper function for reading opt-files
from lib.fileIO.fileIO import CSVTables
from lib.auxiliary.float import toFloat

def readOptPars(filename):
    optFile = CSVTables(filename, separator="\t", comment="#")
    optKeys = ["[target]", "[parameters]", "[options]"]

    optPars = optFile.read()
    keys, colTitles, rowTitles, array = optPars

    # first read optimisation target
    if "[target]" in keys:
        i = keys.index("[target]")
        target = rowTitles[i][0]                
        
    else:
        logError(_("No optimisation target given in %s")%(filename))
        return None

    # read independent parameters
    if "[parameters]" in keys:
        i = keys.index("[parameters]")
        parList = rowTitles[i]
        xmin = np.array([toFloat(row[0]) for row in array[i]])
        xmax = np.array([toFloat(row[1]) for row in array[i]])
    else:
        logError(_("No optimisation parameters given in %s")%(filename))
        return None
        

    # read options
    if "[options]" in keys:
        i = keys.index("[options]")
        optionTable = array[i]
    else:
        optionTable = []

    return (target, parList, xmin, xmax)
            
